(function ($$1) {
  $$1 = $$1 && $$1.hasOwnProperty('default') ? $$1['default'] : $$1;

  define("jquery", function() {
    return $;
  });
  define("underscore", function() {
    return {};
  });

  window.exports = {};

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.

  define('services/kernels/comm', [
      'jquery',
      'base/js/utils',
  ], function($, utils) {

      //-----------------------------------------------------------------------
      // CommManager class
      //-----------------------------------------------------------------------
      
      var CommManager = function (kernel) {
          this.comms = {};
          this.targets = {};
          if (kernel !== undefined) {
              this.init_kernel(kernel);
          }
      };
      
      CommManager.prototype.init_kernel = function (kernel) {
          /**
           * connect the kernel, and register message handlers
           */
          this.kernel = kernel;
          var msg_types = ['comm_open', 'comm_msg', 'comm_close'];
          for (var i = 0; i < msg_types.length; i++) {
              var msg_type = msg_types[i];
              kernel.register_iopub_handler(msg_type, $.proxy(this[msg_type], this));
          }
      };

      CommManager.prototype.new_comm = function (target_name, data, callbacks, metadata, comm_id, buffers) {
          /**
           * Create a new Comm, register it, and open its Kernel-side counterpart
           * Mimics the auto-registration in `Comm.__init__` in the Jupyter Comm.
           *
           * argument comm_id is optional
           */
          var comm = new Comm(target_name, comm_id);
          this.register_comm(comm);
          comm.open(data, callbacks, metadata, buffers);
          return comm;
      };
      
      CommManager.prototype.register_target = function (target_name, f) {
          /**
           * Register a target function for a given target name
           */
          this.targets[target_name] = f;
      };
      
      CommManager.prototype.unregister_target = function (target_name, f) {
          /**
           * Unregister a target function for a given target name
           */
          delete this.targets[target_name];
      };
      
      CommManager.prototype.register_comm = function (comm) {
          /**
           * Register a comm in the mapping
           */
          this.comms[comm.comm_id] = Promise.resolve(comm);
          comm.kernel = this.kernel;
          return comm.comm_id;
      };
      
      CommManager.prototype.unregister_comm = function (comm) {
          /**
           * Remove a comm from the mapping
           */
          delete this.comms[comm.comm_id];
      };
      
      // comm message handlers
      
      CommManager.prototype.comm_open = function (msg) {
          var content = msg.content;
          var that = this;
          var comm_id = content.comm_id;

          this.comms[comm_id] = utils.load_class(content.target_name, content.target_module, 
              this.targets).then(function(target) {
                  var comm = new Comm(content.target_name, comm_id);
                  comm.kernel = that.kernel;
                  try {
                      var response = target(comm, msg);
                  } catch (e) {
                      comm.close();
                      that.unregister_comm(comm);
                      var wrapped_error = new utils.WrappedError("Exception opening new comm", e);
                      console.error(wrapped_error);
                      return Promise.reject(wrapped_error);
                  }
                  // Regardless of the target return value, we need to
                  // then return the comm
                  return Promise.resolve(response).then(function() {return comm;});
              }, utils.reject('Could not open comm', true));
          return this.comms[comm_id];
      };
      
      CommManager.prototype.comm_close = function(msg) {
          var content = msg.content;
          if (this.comms[content.comm_id] === undefined) {
              console.error('Comm promise not found for comm id ' + content.comm_id);
              return;
          }
          var that = this;
          this.comms[content.comm_id] = this.comms[content.comm_id].then(function(comm) {
              that.unregister_comm(comm);
              try {
                  comm.handle_close(msg);
              } catch (e) {
                  console.log("Exception closing comm: ", e, e.stack, msg);
              }
              // don't return a comm, so that further .then() functions
              // get an undefined comm input
          });
          return this.comms[content.comm_id];
      };
      
      CommManager.prototype.comm_msg = function(msg) {
          var content = msg.content;
          if (this.comms[content.comm_id] === undefined) {
              console.error('Comm promise not found for comm id ' + content.comm_id);
              return;
          }

          this.comms[content.comm_id] = this.comms[content.comm_id].then(function(comm) {
              return (Promise.resolve(comm.handle_msg(msg))
                  .catch(utils.reject('Exception handling comm message'))
                  .then(function() {return comm;}));
          });
          return this.comms[content.comm_id];
      };
      
      //-----------------------------------------------------------------------
      // Comm base class
      //-----------------------------------------------------------------------
      
      var Comm = function (target_name, comm_id) {
          this.target_name = target_name;
          this.comm_id = comm_id || utils.uuid();
          this._msg_callback = this._close_callback = null;
      };
      
      // methods for sending messages
      Comm.prototype.open = function (data, callbacks, metadata, buffers) {
          var content = {
              comm_id : this.comm_id,
              target_name : this.target_name,
              data : data || {},
          };
          return this.kernel.send_shell_message("comm_open", content, callbacks, metadata, buffers);
      };
      
      Comm.prototype.send = function (data, callbacks, metadata, buffers) {
          var content = {
              comm_id : this.comm_id,
              data : data || {},
          };
          return this.kernel.send_shell_message("comm_msg", content, callbacks, metadata, buffers);
      };

      Comm.prototype.close = function (data, callbacks, metadata, buffers) {
          var content = {
              comm_id : this.comm_id,
              data : data || {},
          };
          return this.kernel.send_shell_message("comm_close", content, callbacks, metadata, buffers);
      };
      
      // methods for registering callbacks for incoming messages
      Comm.prototype._register_callback = function (key, callback) {
          this['_' + key + '_callback'] = callback;
      };
      
      Comm.prototype.on_msg = function (callback) {
          this._register_callback('msg', callback);
      };
      
      Comm.prototype.on_close = function (callback) {
          this._register_callback('close', callback);
      };
      
      // methods for handling incoming messages
      
      Comm.prototype._callback = function (key, msg) {
          var callback = this['_' + key + '_callback'];
          if (callback) {
              try {
                  return callback(msg);
              } catch (e) {
                  console.log("Exception in Comm callback", e, e.stack, msg);
              }
          }
      };
      
      Comm.prototype.handle_msg = function (msg) {
          return this._callback('msg', msg);
      };
      
      Comm.prototype.handle_close = function (msg) {
          this._callback('close', msg);
      };
      
      return {
          'CommManager': CommManager,
          'Comm': Comm
      };
  });

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.

  define('services/kernels/kernel', [
      'jquery',
      'base/js/utils',
      './comm',
      './serialize',
      'base/js/events'
  ], function($, utils, comm, serialize, events) {

      /**
       * A Kernel class to communicate with the Python kernel. This
       * should generally not be constructed directly, but be created
       * by.  the `Session` object. Once created, this object should be
       * used to communicate with the kernel.
       * 
       * Preliminary documentation for the REST API is at
       * https://github.com/ipython/ipython/wiki/IPEP-16%3A-Notebook-multi-directory-dashboard-and-URL-mapping#kernels-api
       * 
       * Documentation for the messaging specifications is at
       * https://jupyter-client.readthedocs.io/en/stable/messaging.html
       *
       * @class Kernel
       * @param {string} kernel_service_url - the URL to access the kernel REST api
       * @param {string} ws_url - the websockets URL
       * @param {string} name - the kernel type (e.g. python3)
       */
      var Kernel = function (kernel_service_url, ws_url, name) {
          this.events = events;

          this.id = null;
          this.name = name;
          this.ws = null;
          this._stopping = false;

          this.kernel_service_url = kernel_service_url;
          this.kernel_url = null;
          this.ws_url = ws_url || utils.get_body_data("wsUrl");
          if (!this.ws_url) {
              // trailing 's' in https will become wss for secure web sockets
              this.ws_url = location.protocol.replace('http', 'ws') + "//" + location.host;
          }

          this.username = "username";
          this.session_id = utils.uuid();
          this._msg_callbacks = {};
          this._msg_callbacks_overrides = {};
          this._display_id_to_parent_ids = {};
          this._msg_queue = Promise.resolve();
          this.info_reply = {}; // kernel_info_reply stored here after starting

          if (typeof(WebSocket) !== 'undefined') {
              this.WebSocket = WebSocket;
          } else if (typeof(MozWebSocket) !== 'undefined') {
              this.WebSocket = MozWebSocket;
          } else {
              alert('Your browser does not have WebSocket support, please try Chrome, Safari or Firefox ≥ 6. Firefox 4 and 5 are also supported by you have to enable WebSockets in about:config.');
          }
          
          this.bind_events();
          this.init_iopub_handlers();
          this.comm_manager = new comm.CommManager(this);
          
          this.last_msg_id = null;
          this.last_msg_callbacks = {};

          this._autorestart_attempt = 0;
          this._reconnect_attempt = 0;
          this.reconnect_limit = 7;
          
          this._pending_messages = [];
      };

      /**
       * @function _get_msg
       */
      Kernel.prototype._get_msg = function (msg_type, content, metadata, buffers) {
          var msg = {
              header : {
                  msg_id : utils.uuid(),
                  username : this.username,
                  session : this.session_id,
                  msg_type : msg_type,
                  version : "5.2",
              },
              metadata : metadata || {},
              content : content,
              buffers : buffers || [],
              parent_header : {}
          };
          return msg;
      };

      /**
       * @function bind_events
       */
      Kernel.prototype.bind_events = function () {
          var that = this;
          this.events.on('send_input_reply.Kernel', function(evt, data) { 
              that.send_input_reply(data);
          });

          var record_status = function (evt, info) {
              console.log('Kernel: ' + evt.type + ' (' + info.kernel.id + ')');
          };

          this.events.on('kernel_created.Kernel', record_status);
          this.events.on('kernel_reconnecting.Kernel', record_status);
          this.events.on('kernel_connected.Kernel', record_status);
          this.events.on('kernel_starting.Kernel', record_status);
          this.events.on('kernel_restarting.Kernel', record_status);
          this.events.on('kernel_autorestarting.Kernel', record_status);
          this.events.on('kernel_interrupting.Kernel', record_status);
          this.events.on('kernel_disconnected.Kernel', record_status);
          // these are commented out because they are triggered a lot, but can
          // be uncommented for debugging purposes
          //this.events.on('kernel_idle.Kernel', record_status);
          //this.events.on('kernel_busy.Kernel', record_status);
          this.events.on('kernel_ready.Kernel', record_status);
          this.events.on('kernel_killed.Kernel', record_status);
          this.events.on('kernel_dead.Kernel', record_status);

          this.events.on('kernel_ready.Kernel', function () {
              that._autorestart_attempt = 0;
          });
          this.events.on('kernel_connected.Kernel', function () {
              that._reconnect_attempt = 0;
          });
      };

      /**
       * Initialize the iopub handlers.
       *
       * @function init_iopub_handlers
       */
      Kernel.prototype.init_iopub_handlers = function () {
          var output_msg_types = ['stream', 'display_data', 'execute_result', 'error', 'update_display_data'];
          this._iopub_handlers = {};
          this.register_iopub_handler('status', $.proxy(this._handle_status_message, this));
          this.register_iopub_handler('clear_output', $.proxy(this._handle_clear_output, this));
          this.register_iopub_handler('execute_input', $.proxy(this._handle_input_message, this));
          
          for (var i=0; i < output_msg_types.length; i++) {
              this.register_iopub_handler(output_msg_types[i], $.proxy(this._handle_output_message, this));
          }
      };

      /**
       * GET /api/kernels
       *
       * Get the list of running kernels.
       *
       * @function list
       * @param {function} [success] - function executed on ajax success
       * @param {function} [error] - function executed on ajax error
       */
      Kernel.prototype.list = function (success, error) {
          utils.ajax(this.kernel_service_url, {
              processData: false,
              cache: false,
              type: "GET",
              dataType: "json",
              success: success,
              error: this._on_error(error)
          });
      };

      /**
       * POST /api/kernels
       *
       * Start a new kernel.
       *
       * In general this shouldn't be used -- the kernel should be
       * started through the session API. If you use this function and
       * are also using the session API then your session and kernel
       * WILL be out of sync!
       *
       * @function start
       * @param {params} [Object] - parameters to include in the query string
       * @param {function} [success] - function executed on ajax success
       * @param {function} [error] - function executed on ajax error
       */
      Kernel.prototype.start = function (params, success, error) {
          var url = this.kernel_service_url;
          var qs = $.param(params || {}); // query string for sage math stuff
          if (qs !== "") {
              url = url + "?" + qs;
          }

          this.events.trigger('kernel_starting.Kernel', {kernel: this});
          var that = this;
          var on_success = function (data, status, xhr) {
              that.events.trigger('kernel_created.Kernel', {kernel: that});
              that._kernel_created(data);
              if (success) {
                  success(data, status, xhr);
              }
          };

          utils.ajax(url, {
              processData: false,
              cache: false,
              type: "POST",
              data: JSON.stringify({name: this.name}),
              contentType: 'application/json',
              dataType: "json",
              success: this._on_success(on_success),
              error: this._on_error(error)
          });

          return url;
      };

      /**
       * GET /api/kernels/[:kernel_id]
       *
       * Get information about the kernel.
       *
       * @function get_info
       * @param {function} [success] - function executed on ajax success
       * @param {function} [error] - function executed on ajax error
       */
      Kernel.prototype.get_info = function (success, error) {
          utils.ajax(this.kernel_url, {
              processData: false,
              cache: false,
              type: "GET",
              dataType: "json",
              success: this._on_success(success),
              error: this._on_error(error)
          });
      };

      /**
       * DELETE /api/kernels/[:kernel_id]
       *
       * Shutdown the kernel.
       *
       * If you are also using sessions, then this function should NOT be
       * used. Instead, use Session.delete. Otherwise, the session and
       * kernel WILL be out of sync.
       *
       * @function kill
       * @param {function} [success] - function executed on ajax success
       * @param {function} [error] - function executed on ajax error
       */
      Kernel.prototype.kill = function (success, error) {
          this.events.trigger('kernel_killed.Kernel', {kernel: this});
          this._kernel_dead();
          utils.ajax(this.kernel_url, {
              processData: false,
              cache: false,
              type: "DELETE",
              dataType: "json",
              success: this._on_success(success),
              error: this._on_error(error)
          });
      };

      /**
       * POST /api/kernels/[:kernel_id]/interrupt
       *
       * Interrupt the kernel.
       *
       * @function interrupt
       * @param {function} [success] - function executed on ajax success
       * @param {function} [error] - function executed on ajax error
       */
      Kernel.prototype.interrupt = function (success, error) {
          this.events.trigger('kernel_interrupting.Kernel', {kernel: this});

          var that = this;
          var on_success = function (data, status, xhr) {
              /**
               * get kernel info so we know what state the kernel is in
               */
              that.kernel_info();
              if (success) {
                  success(data, status, xhr);
              }
          };

          var url = utils.url_path_join(this.kernel_url, 'interrupt');
          utils.ajax(url, {
              processData: false,
              cache: false,
              type: "POST",
              contentType: false,  // there's no data with this
              dataType: "json",
              success: this._on_success(on_success),
              error: this._on_error(error)
          });
      };

      Kernel.prototype.restart = function (success, error) {
          /**
           * POST /api/kernels/[:kernel_id]/restart
           *
           * Restart the kernel.
           *
           * @function interrupt
           * @param {function} [success] - function executed on ajax success
           * @param {function} [error] - function executed on ajax error
           */
          this.events.trigger('kernel_restarting.Kernel', {kernel: this});
          this.stop_channels();
          this._msg_callbacks = {};
          this._msg_callbacks_overrides = {};
          this._display_id_to_parent_ids = {};

          var that = this;
          var on_success = function (data, status, xhr) {
              that.events.trigger('kernel_created.Kernel', {kernel: that});
              that._kernel_created(data);
              if (success) {
                  success(data, status, xhr);
              }
          };

          var on_error = function (xhr, status, err) {
              that.events.trigger('kernel_failed_restart.Kernel', {kernel: that});
              that._kernel_dead();
              if (error) {
                  error(xhr, status, err);
              }
          };

          var url = utils.url_path_join(this.kernel_url, 'restart');
          utils.ajax(url, {
              processData: false,
              cache: false,
              type: "POST",
              contentType: false,  // there's no data with this
              dataType: "json",
              success: this._on_success(on_success),
              error: this._on_error(on_error)
          });
      };

      Kernel.prototype.reconnect = function () {
          /**
           * Reconnect to a disconnected kernel. This is not actually a
           * standard HTTP request, but useful function nonetheless for
           * reconnecting to the kernel if the connection is somehow lost.
           *
           * @function reconnect
           */
          if (this.is_connected()) {
              this.stop_channels();
          }
          this._reconnect_attempt = this._reconnect_attempt + 1;
          this.events.trigger('kernel_reconnecting.Kernel', {
              kernel: this,
              attempt: this._reconnect_attempt,
          });
          this.start_channels();
      };

      Kernel.prototype._on_success = function (success) {
          /**
           * Handle a successful AJAX request by updating the kernel id and
           * name from the response, and then optionally calling a provided
           * callback.
           *
           * @function _on_success
           * @param {function} success - callback
           */
          var that = this;
          return function (data, status, xhr) {
              if (data) {
                  that.id = data.id;
                  that.name = data.name;
              }
              that.kernel_url = utils.url_path_join(that.kernel_service_url,
                  encodeURIComponent(that.id));
              if (success) {
                  success(data, status, xhr);
              }
          };
      };

      Kernel.prototype._on_error = function (error) {
          /**
           * Handle a failed AJAX request by logging the error message, and
           * then optionally calling a provided callback.
           *
           * @function _on_error
           * @param {function} error - callback
           */
          return function (xhr, status, err) {
              utils.log_ajax_error(xhr, status, err);
              if (error) {
                  error(xhr, status, err);
              }
          };
      };

      Kernel.prototype._kernel_created = function (data) {
          /**
           * Perform necessary tasks once the kernel has been started,
           * including actually connecting to the kernel.
           *
           * @function _kernel_created
           * @param {Object} data - information about the kernel including id
           */
          this.id = data.id;
          this.kernel_url = utils.url_path_join(this.kernel_service_url,
              encodeURIComponent(this.id));
          this.start_channels();
      };

      Kernel.prototype._kernel_connected = function () {
          /**
           * Perform necessary tasks once the connection to the kernel has
           * been established. This includes requesting information about
           * the kernel.
           *
           * @function _kernel_connected
           */
          this.events.trigger('kernel_connected.Kernel', {kernel: this});

          // Send pending messages. We shift the message off the queue
          // after the message is sent so that if there is an exception,
          // the message is still pending.
          while (this._pending_messages.length > 0) {
            this.ws.send(this._pending_messages[0]);
            this._pending_messages.shift();
          }

          // get kernel info so we know what state the kernel is in
          var that = this;
          this.kernel_info(function (reply) {
              that.info_reply = reply.content;
              that.events.trigger('kernel_ready.Kernel', {kernel: that});
          });
      };

      Kernel.prototype._kernel_dead = function () {
          /**
           * Perform necessary tasks after the kernel has died. This closing
           * communication channels to the kernel if they are still somehow
           * open.
           *
           * @function _kernel_dead
           */
          this.stop_channels();
      };

      Kernel.prototype.start_channels = function () {
          /**
           * Start the websocket channels.
           * Will stop and restart them if they already exist.
           *
           * @function start_channels
           */
          var that = this;
          this.stop_channels();
          var ws_host_url = this.ws_url + this.kernel_url;

          console.log("Starting WebSockets:", ws_host_url);

          this.ws = new this.WebSocket([
                  that.ws_url,
                  utils.url_path_join(that.kernel_url, 'channels'),
                  "?session_id=" + that.session_id
              ].join('')
          );
          
          var already_called_onclose = false; // only alert once
          var ws_closed_early = function(evt){
              console.log("WebSocket closed early", evt);
              if (already_called_onclose){
                  return;
              }
              already_called_onclose = true;
              // If the websocket was closed early, that could mean
              // that the kernel is actually dead. Try getting
              // information about the kernel from the API call --
              // if that fails, then assume the kernel is dead,
              // otherwise just follow the typical websocket closed
              // protocol.
              that.get_info(function () {
                  that._ws_closed(ws_host_url, false);
              }, function () {
                  that.events.trigger('kernel_dead.Kernel', {kernel: that});
                  that._kernel_dead();
              });
          };
          var ws_closed_late = function(evt){
              console.log("WebSocket closed unexpectedly", evt);
              if (already_called_onclose){
                  return;
              }
              already_called_onclose = true;
              that._ws_closed(ws_host_url, false);
          };
          var ws_error = function(evt){
              if (already_called_onclose){
                  return;
              }
              already_called_onclose = true;
              that._ws_closed(ws_host_url, true);
          };

          this.ws.onopen = $.proxy(this._ws_opened, this);
          this.ws.onclose = ws_closed_early;
          this.ws.onerror = ws_error;
          // switch from early-close to late-close message after 1s
          setTimeout(function() {
              if (that.ws !== null && !that._stopping) {
                  that.ws.onclose = ws_closed_late;
              }
          }, 1000);
          this.ws.onmessage = $.proxy(this._handle_ws_message, this);
      };

      Kernel.prototype._ws_opened = function (evt) {
          /**
           * Handle a websocket entering the open state,
           * signaling that the kernel is connected when websocket is open.
           *
           * @function _ws_opened
           */
          if (this.is_connected()) {
              // all events ready, trigger started event.
              this._kernel_connected();
          }
      };

      Kernel.prototype._ws_closed = function(ws_url, error) {
          /**
           * Handle a websocket entering the closed state.  If the websocket
           * was not closed due to an error, try to reconnect to the kernel.
           *
           * @function _ws_closed
           * @param {string} ws_url - the websocket url
           * @param {bool} error - whether the connection was closed due to an error
           */
          this.stop_channels();

          this.events.trigger('kernel_disconnected.Kernel', {kernel: this});
          if (error) {
              console.log('WebSocket connection failed: ', ws_url, error);
              this.events.trigger('kernel_connection_failed.Kernel', {
                  kernel: this,
                  ws_url: ws_url,
                  attempt: this._reconnect_attempt,
                  error: error,
              });
          }
          this._schedule_reconnect();
      };
      
      Kernel.prototype._schedule_reconnect = function () {
          /**
           * function to call when kernel connection is lost
           * schedules reconnect, or fires 'connection_dead' if reconnect limit is hit
           */
          if (this._reconnect_attempt < this.reconnect_limit) {
              var timeout = Math.pow(2, this._reconnect_attempt);
              console.log("Connection lost, reconnecting in " + timeout + " seconds.");
              setTimeout($.proxy(this.reconnect, this), 1e3 * timeout);
          } else {
              this.events.trigger('kernel_connection_dead.Kernel', {
                  kernel: this,
                  reconnect_attempt: this._reconnect_attempt,
              });
              console.log("Failed to reconnect, giving up.");
          }
      };
      
      Kernel.prototype.stop_channels = function () {
          /**
           * Close the websocket. After successful close, the value
           * in `this.ws` will be null.
           *
           * @function stop_channels
           */
          var that = this;
          var close = function () {
              that._stopping = false;
              if (that.ws && that.ws.readyState === WebSocket.CLOSED) {
                  that.ws = null;
              }
          };
          if (this.ws !== null) {
              // flag to avoid races with on_close_late
              this._stopping = true;
              if (this.ws.readyState === WebSocket.OPEN) {
                  this.ws.onclose = close;
                  this.ws.close();
              } else {
                  close();
              }
          }
      };

      Kernel.prototype.is_connected = function () {
          /**
           * Check whether there is a connection to the kernel. This
           * function only returns true if websocket has been
           * created and has a state of WebSocket.OPEN.
           *
           * @function is_connected
           * @returns {bool} - whether there is a connection
           */
          // if any channel is not ready, then we're not connected
          if (this.ws === null) {
              return false;
          }
          if (this.ws.readyState !== WebSocket.OPEN) {
              return false;
          }
          return true;
      };

      Kernel.prototype.is_fully_disconnected = function () {
          /**
           * Check whether the connection to the kernel has been completely
           * severed. This function only returns true if all channel objects
           * are null.
           *
           * @function is_fully_disconnected
           * @returns {bool} - whether the kernel is fully disconnected
           */
          return (this.ws === null);
      };
      
      Kernel.prototype._send = function(msg) {
        /**
         * Send a message (if the kernel is connected) or queue the message for future delivery
         *
         * Pending messages will automatically be sent when a kernel becomes connected.
         *
         * @function _send
         * @param msg
         */
        if (this.is_connected()) {
              this.ws.send(msg);
          } else {
              this._pending_messages.push(msg);
          }
      };
      
      Kernel.prototype.send_shell_message = function (msg_type, content, callbacks, metadata, buffers) {
          /**
           * Send a message on the Kernel's shell channel
           *
           * If the kernel is not connected, the message will be buffered.
           * 
           * @function send_shell_message
           */
          var msg = this._get_msg(msg_type, content, metadata, buffers);
          msg.channel = 'shell';
          this.set_callbacks_for_msg(msg.header.msg_id, callbacks);
          this._send(serialize.serialize(msg));
          return msg.header.msg_id;
      };

      Kernel.prototype.kernel_info = function (callback) {
          /**
           * Get kernel info
           *
           * @function kernel_info
           * @param callback {function}
           *
           * When calling this method, pass a callback function that expects one argument.
           * The callback will be passed the complete `kernel_info_reply` message documented
           * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#kernel-info)
           */
          var callbacks;
          if (callback) {
              callbacks = { shell : { reply : callback } };
          }
          return this.send_shell_message("kernel_info_request", {}, callbacks);
      };

      Kernel.prototype.comm_info = function (target_name, callback) {
          /**
           * Get comm info
           *
           * @function comm_info
           * @param callback {function}
           *
           * When calling this method, pass a callback function that expects one argument.
           * The callback will be passed the complete `comm_info_reply` message documented
           * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#comm_info)
           */
          var callbacks;
          if (callback) {
              callbacks = { shell : { reply : callback } };
          }
          var content = {
              target_name : target_name,
          };
          return this.send_shell_message("comm_info_request", content, callbacks);
      };

      Kernel.prototype.inspect = function (code, cursor_pos, callback) {
          /**
           * Get info on an object
           *
           * When calling this method, pass a callback function that expects one argument.
           * The callback will be passed the complete `inspect_reply` message documented
           * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#object-information)
           *
           * @function inspect
           * @param code {string}
           * @param cursor_pos {integer}
           * @param callback {function}
           */
          var callbacks;
          if (callback) {
              callbacks = { shell : { reply : callback } };
          }
          
          var content = {
              code : code,
              cursor_pos : cursor_pos,
              detail_level : 0
          };
          return this.send_shell_message("inspect_request", content, callbacks);
      };

      Kernel.prototype.execute = function (code, callbacks, options) {
          /**
           * Execute given code into kernel, and pass result to callback.
           *
           * @async
           * @function execute
           * @param {string} code
           * @param [callbacks] {Object} With the following keys (all optional)
           *      @param callbacks.shell.reply {function}
           *      @param callbacks.shell.payload.[payload_name] {function}
           *      @param callbacks.iopub.output {function}
           *      @param callbacks.iopub.clear_output {function}
           *      @param callbacks.input {function}
           *      @param callbacks.clear_on_done=true {Boolean}
           * @param {object} [options]
           *      @param [options.silent=false] {Boolean}
           *      @param [options.user_expressions=empty_dict] {Dict}
           *      @param [options.allow_stdin=false] {Boolean} true|false
           *
           * @example
           *
           * The options object should contain the options for the execute
           * call. Its default values are:
           *
           *      options = {
           *        silent : true,
           *        user_expressions : {},
           *        allow_stdin : false
           *      }
           *
           * When calling this method pass a callbacks structure of the
           * form:
           *
           *      callbacks = {
           *       shell : {
           *         reply : execute_reply_callback,
           *         payload : {
           *           set_next_input : set_next_input_callback,
           *         }
           *       },
           *       iopub : {
           *         output : output_callback,
           *         clear_output : clear_output_callback,
           *       },
           *       input : raw_input_callback
           *      }
           *
           * Each callback will be passed the entire message as a single
           * argument.  Payload handlers will be passed the corresponding
           * payload and the execute_reply message.
           */
          var content = {
              code : code,
              silent : true,
              store_history : false,
              user_expressions : {},
              allow_stdin : false
          };
          callbacks = callbacks || {};
          if (callbacks.input !== undefined) {
              content.allow_stdin = true;
          }
          $.extend(true, content, options);
          this.events.trigger('execution_request.Kernel', {kernel: this, content: content});
          return this.send_shell_message("execute_request", content, callbacks);
      };

      /**
       * When calling this method, pass a function to be called with the
       * `complete_reply` message as its only argument when it arrives.
       *
       * `complete_reply` is documented
       * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#complete)
       *
       * @function complete
       * @param code {string}
       * @param cursor_pos {integer}
       * @param callback {function}
       */
      Kernel.prototype.complete = function (code, cursor_pos, callback) {
          var callbacks;
          if (callback) {
              callbacks = { shell : { reply : callback } };
          }
          var content = {
              code : code,
              cursor_pos : cursor_pos
          };
          return this.send_shell_message("complete_request", content, callbacks);
      };

      /**
       * @function send_input_reply
       */
      Kernel.prototype.send_input_reply = function (input) {
          var content = {
              value : input
          };
          this.events.trigger('input_reply.Kernel', {kernel: this, content: content});
          var msg = this._get_msg("input_reply", content);
          msg.channel = 'stdin';
          this._send(serialize.serialize(msg));
          return msg.header.msg_id;
      };

      /**
       * @function register_iopub_handler
       */
      Kernel.prototype.register_iopub_handler = function (msg_type, callback) {
          this._iopub_handlers[msg_type] = callback;
      };

      /**
       * Get the iopub handler for a specific message type.
       *
       * @function get_iopub_handler
       */
      Kernel.prototype.get_iopub_handler = function (msg_type) {
          return this._iopub_handlers[msg_type];
      };

      /**
       * Get callbacks for a specific message.
       *
       * @function get_callbacks_for_msg
       */
      Kernel.prototype.get_callbacks_for_msg = function (msg_id) {
          if (msg_id == this.last_msg_id) {
              return this.last_msg_callbacks;
          } else {
              return this._msg_callbacks[msg_id];
          }
      };

      /**
       * Get output callbacks for a specific message.
       *
       * @function get_output_callbacks_for_msg
       *
       * Since output callbacks can be overridden, we first check the override stack.
       */
      Kernel.prototype.get_output_callbacks_for_msg = function (msg_id) {
          return this.get_callbacks_for_msg(this.get_output_callback_id(msg_id));
      };


      /**
       * Get the output callback id for a message
       *
       * Since output callbacks can be redirected, this may not be the same as
       * the msg_id.
       *
       * @function get_output_callback_id
       */
      Kernel.prototype.get_output_callback_id = function (msg_id) {
          var callback_id = msg_id;
          var overrides = this._msg_callbacks_overrides[msg_id];
          if (overrides && overrides.length > 0) {
              callback_id = overrides[overrides.length-1];
          }
          return callback_id
      };

      /**
       * Clear callbacks for a specific message.
       *
       * @function clear_callbacks_for_msg
       */
      Kernel.prototype.clear_callbacks_for_msg = function (msg_id) {
          if (this._msg_callbacks[msg_id] !== undefined ) {
              var callbacks = this._msg_callbacks[msg_id];
              var kernel = this;
              // clear display_id:msg_id map for display_ids associated with this msg_id
              if (!callbacks) return;
              callbacks.display_ids.map(function (display_id) {
                  var msg_ids = kernel._display_id_to_parent_ids[display_id];
                  if (msg_ids) {
                      var idx = msg_ids.indexOf(msg_id);
                      if (idx === -1) {
                          return;
                      }
                      if (msg_ids.length === 1) {
                          delete kernel._display_id_to_parent_ids[display_id];
                      } else {
                          msg_ids.splice(idx, 1);
                          kernel._display_id_to_parent_ids[display_id] = msg_ids;
                      }
                  }
              });
              delete this._msg_callbacks[msg_id];
          }
      };
      
      /**
       * @function _finish_shell
       */
      Kernel.prototype._finish_shell = function (msg_id) {
          var callbacks = this._msg_callbacks[msg_id];
          if (callbacks !== undefined) {
              callbacks.shell_done = true;
              if (callbacks.clear_on_done && callbacks.iopub_done) {
                  this.clear_callbacks_for_msg(msg_id);
              }
          }
      };

      /**
       * @function _finish_iopub
       */
      Kernel.prototype._finish_iopub = function (msg_id) {
          var callbacks = this._msg_callbacks[msg_id];
          if (callbacks !== undefined) {
              callbacks.iopub_done = true;
              if (callbacks.clear_on_done && callbacks.shell_done) {
                  this.clear_callbacks_for_msg(msg_id);
              }
          }
          this.events.trigger('finished_iopub.Kernel', {kernel: this, msg_id: msg_id});
      };
      
      /**
       * Set callbacks for a particular message.
       * Callbacks should be a struct of the following form:
       * shell : {
       * 
       * }
       *
       * If the third parameter is truthy, the callback is set as the last
       * callback registered.
       *
       * @function set_callbacks_for_msg
       */
      Kernel.prototype.set_callbacks_for_msg = function (msg_id, callbacks, save) {
          var remember = save || true;
          if (remember) {
              this.last_msg_id = msg_id;
          }
          if (callbacks) {
              // shallow-copy mapping, because we will modify it at the top level
              var cbcopy = this._msg_callbacks[msg_id] = this.last_msg_callbacks = {};
              cbcopy.shell = callbacks.shell;
              cbcopy.iopub = callbacks.iopub;
              cbcopy.input = callbacks.input;
              cbcopy.clear_on_done = callbacks.clear_on_done;
              cbcopy.shell_done = (!callbacks.shell);
              cbcopy.iopub_done = (!callbacks.iopub);
              cbcopy.display_ids = [];
              if (callbacks.clear_on_done === undefined) {
                  // default to clear-on-done
                  cbcopy.clear_on_done = true;
              }
          } else if (remember) {
              this.last_msg_callbacks = {};
          }
      };

      /**
       * Override output callbacks for a particular msg_id
       */
      Kernel.prototype.output_callback_overrides_push = function(msg_id, callback_id) {
          var output_callbacks = this._msg_callbacks_overrides[msg_id];
          if (!output_callbacks) {
              this._msg_callbacks_overrides[msg_id] = output_callbacks = [];
          }
          output_callbacks.push(callback_id);
      };

      Kernel.prototype.output_callback_overrides_pop = function(msg_id) {
          var callback_ids = this._msg_callbacks_overrides[msg_id];
          if (!callback_ids) {
              console.error("Popping callback overrides, but none registered", msg_id);
              return;
          }
          return callback_ids.pop();
      };

      Kernel.prototype._handle_ws_message = function (e) {
          var that = this;
          this._msg_queue = this._msg_queue.then(function() {
              return serialize.deserialize(e.data);
          }).then(function(msg) {return that._finish_ws_message(msg);})
          .catch(function(error) { console.error("Couldn't process kernel message", error); });
      };

      Kernel.prototype._finish_ws_message = function (msg) {
          switch (msg.channel) {
              case 'shell':
                  return this._handle_shell_reply(msg);
              case 'iopub':
                  return this._handle_iopub_message(msg);
              case 'stdin':
                  return this._handle_input_request(msg);
              default:
                  console.error("unrecognized message channel", msg.channel, msg);
          }
      };
      
      Kernel.prototype._handle_shell_reply = function (reply) {
          this.events.trigger('shell_reply.Kernel', {kernel: this, reply:reply});
          var that = this;
          var content = reply.content;
          var metadata = reply.metadata;
          var parent_id = reply.parent_header.msg_id;
          var callbacks = this.get_callbacks_for_msg(parent_id);
          var promise = Promise.resolve();
          if (!callbacks || !callbacks.shell) {
              return;
          }
          var shell_callbacks = callbacks.shell;
          
          // signal that shell callbacks are done
          this._finish_shell(parent_id);
          
          if (shell_callbacks.reply !== undefined) {
              promise = promise.then(function() {return shell_callbacks.reply(reply);});
          }
          if (content.payload && shell_callbacks.payload) {
              promise = promise.then(function() {
                  return that._handle_payloads(content.payload, shell_callbacks.payload, reply);
              });
          }
          return promise;
      };

      /**
       * @function _handle_payloads
       */
      Kernel.prototype._handle_payloads = function (payloads, payload_callbacks, msg) {
          var promise = [];
          var l = payloads.length;
          // Payloads are handled by triggering events because we don't want the Kernel
          // to depend on the Notebook or Pager classes.
          for (var i=0; i<l; i++) {
              var payload = payloads[i];
              var callback = payload_callbacks[payload.source];
              if (callback) {
                  promise.push(callback(payload, msg));
              }
          }
          return Promise.all(promise);
      };

      /**
       * @function _handle_status_message
       */
      Kernel.prototype._handle_status_message = function (msg) {
          var execution_state = msg.content.execution_state;
          var parent_id = msg.parent_header.msg_id;
          
          // dispatch status msg callbacks, if any
          var callbacks = this.get_callbacks_for_msg(parent_id);
          if (callbacks && callbacks.iopub && callbacks.iopub.status) {
              try {
                  callbacks.iopub.status(msg);
              } catch (e) {
                  console.log("Exception in status msg handler", e, e.stack);
              }
          }
          
          if (execution_state === 'busy') {
              this.events.trigger('kernel_busy.Kernel', {kernel: this});

          } else if (execution_state === 'idle') {
              // signal that iopub callbacks are (probably) done
              // async output may still arrive,
              // but only for the most recent request
              this._finish_iopub(parent_id);
              
              // trigger status_idle event
              this.events.trigger('kernel_idle.Kernel', {kernel: this});

          } else if (execution_state === 'starting') {
              this.events.trigger('kernel_starting.Kernel', {kernel: this});
              var that = this;
              this.kernel_info(function (reply) {
                  that.info_reply = reply.content;
                  that.events.trigger('kernel_ready.Kernel', {kernel: that});
              });

          } else if (execution_state === 'restarting') {
              // autorestarting is distinct from restarting,
              // in that it means the kernel died and the server is restarting it.
              // kernel_restarting sets the notification widget,
              // autorestart shows the more prominent dialog.
              this._autorestart_attempt = this._autorestart_attempt + 1;
              this.events.trigger('kernel_restarting.Kernel', {kernel: this});
              this.events.trigger('kernel_autorestarting.Kernel', {kernel: this, attempt: this._autorestart_attempt});

          } else if (execution_state === 'dead') {
              this.events.trigger('kernel_dead.Kernel', {kernel: this});
              this._kernel_dead();
          }
      };
      
      /**
       * Handle clear_output message
       *
       * @function _handle_clear_output
       */
      Kernel.prototype._handle_clear_output = function (msg) {
          var callbacks = this.get_output_callbacks_for_msg(msg.parent_header.msg_id);
          if (!callbacks || !callbacks.iopub) {
              return;
          }
          var callback = callbacks.iopub.clear_output;
          if (callback) {
              callback(msg);
          }
      };

      /**
       * handle an output message (execute_result, display_data, etc.)
       *
       * @function _handle_output_message
       */
      Kernel.prototype._handle_output_message = function (msg) {
          var that = this;
          var msg_id = msg.parent_header.msg_id;
          var callbacks = this.get_output_callbacks_for_msg(msg_id);
          if (['display_data', 'update_display_data', 'execute_result'].indexOf(msg.header.msg_type) > -1) {
              // display_data messages may re-route based on their display_id
              var display_id = (msg.content.transient || {}).display_id;
              if (display_id) {
                  // it has a display_id
                  var parent_ids = this._display_id_to_parent_ids[display_id];
                  if (parent_ids) {
                      // we've seen it before, update existing outputs with same display_id
                      // by handling display_data as update_display_data
                      var update_msg = $.extend(true, {}, msg);
                      update_msg.header.msg_type = 'update_display_data';

                      parent_ids.map(function (parent_id) {
                          var callbacks = that.get_callbacks_for_msg(parent_id);
                          if (!callbacks) return;
                          var callback = callbacks.iopub.output;
                          if (callback) {
                              callback(update_msg);
                          }
                      });
                  }
                  // we're done here if it's update_display
                  if (msg.header.msg_type === 'update_display_data') {
                      // it's an update, don't proceed to the normal display
                      return;
                  }
                  // regular display_data with id, record it for future updating
                  // in _display_id_to_parent_ids for future lookup
                  if (this._display_id_to_parent_ids[display_id] === undefined) {
                      this._display_id_to_parent_ids[display_id] = [];
                  }
                  var callback_id = this.get_output_callback_id(msg_id);
                  if (this._display_id_to_parent_ids[display_id].indexOf(callback_id) === -1) {
                      this._display_id_to_parent_ids[display_id].push(callback_id);
                  }
                  // and in callbacks for cleanup on clear_callbacks_for_msg
                  if (callbacks && callbacks.display_ids.indexOf(display_id) === -1) {
                      callbacks.display_ids.push(display_id);
                  }
              }
          }


          if (!callbacks || !callbacks.iopub) {
              // The message came from another client. Let the UI decide what to
              // do with it.
              this.events.trigger('received_unsolicited_message.Kernel', msg);
              return;
          }
          var callback = callbacks.iopub.output;
          if (callback) {
              callback(msg);
          }
      };

      /**
       * Handle an input message (execute_input).
       *
       * @function _handle_input message
       */
      Kernel.prototype._handle_input_message = function (msg) {
          var callbacks = this.get_callbacks_for_msg(msg.parent_header.msg_id);
          if (!callbacks) {
              // The message came from another client. Let the UI decide what to
              // do with it.
              this.events.trigger('received_unsolicited_message.Kernel', msg);
          }
      };

      /**
       * Dispatch IOPub messages to respective handlers. Each message
       * type should have a handler.
       *
       * @function _handle_iopub_message
       */
      Kernel.prototype._handle_iopub_message = function (msg) {
          var handler = this.get_iopub_handler(msg.header.msg_type);
          if (handler !== undefined) {
              return handler(msg);
          }
      };

      /**
       * @function _handle_input_request
       */
      Kernel.prototype._handle_input_request = function (request) {
          var header = request.header;
          var content = request.content;
          var metadata = request.metadata;
          var msg_type = header.msg_type;
          if (msg_type !== 'input_request') {
              console.log("Invalid input request!", request);
              return;
          }
          var callbacks = this.get_callbacks_for_msg(request.parent_header.msg_id);
          if (callbacks) {
              if (callbacks.input) {
                  callbacks.input(request);
              }
          }
      };

      return {'Kernel': Kernel};
  });

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.

  define('services/kernels/serialize', [
      'underscore',
      ], function (_) {
      
      var _deserialize_array_buffer = function (buf) {
          var data = new DataView(buf);
          // read the header: 1 + nbufs 32b integers
          var nbufs = data.getUint32(0);
          var offsets = [];
          var i;
          for (i = 1; i <= nbufs; i++) {
              offsets.push(data.getUint32(i * 4));
          }
          var json_bytes = new Uint8Array(buf.slice(offsets[0], offsets[1]));
          var msg = JSON.parse(
              (new TextDecoder('utf8')).decode(json_bytes)
          );
          // the remaining chunks are stored as DataViews in msg.buffers
          msg.buffers = [];
          var start, stop;
          for (i = 1; i < nbufs; i++) {
              start = offsets[i];
              stop = offsets[i+1] || buf.byteLength;
              msg.buffers.push(new DataView(buf.slice(start, stop)));
          }
          return msg;
      };
      
      var _deserialize_binary = function(data) {
          /**
           * deserialize the binary message format
           * callback will be called with a message whose buffers attribute
           * will be an array of DataViews.
           */
          if (data instanceof Blob) {
              // data is Blob, have to deserialize from ArrayBuffer in reader callback
              var reader = new FileReader();
              var promise = new Promise(function(resolve, reject) {
                  reader.onload = function () {
                      var msg = _deserialize_array_buffer(this.result);
                      resolve(msg);
                  };
              });
              reader.readAsArrayBuffer(data);
              return promise;
          } else {
              // data is ArrayBuffer, can deserialize directly
              var msg = _deserialize_array_buffer(data);
              return msg;
          }
      };

      var deserialize = function (data) {
          /**
           * deserialize a message and return a promise for the unpacked message
           */
          if (typeof data === "string") {
              // text JSON message
              return Promise.resolve(JSON.parse(data));
          } else {
              // binary message
              return Promise.resolve(_deserialize_binary(data));
          }
      };
      
      var _serialize_binary = function (msg) {
          /**
           * implement the binary serialization protocol
           * serializes JSON message to ArrayBuffer
           */
          msg = _.clone(msg);
          var offsets = [];
          var buffers = [];
          var i;
          for (i = 0; i < msg.buffers.length; i++) {
              // msg.buffers elements could be either views or ArrayBuffers
              // buffers elements are ArrayBuffers
              var b = msg.buffers[i];
              buffers.push(b.buffer instanceof ArrayBuffer ? b.buffer : b);
          }
          delete msg.buffers;
          var json_utf8 = (new TextEncoder('utf8')).encode(JSON.stringify(msg));
          buffers.unshift(json_utf8);
          var nbufs = buffers.length;
          offsets.push(4 * (nbufs + 1));
          for (i = 0; i + 1 < buffers.length; i++) {
              offsets.push(offsets[offsets.length-1] + buffers[i].byteLength);
          }
          var msg_buf = new Uint8Array(
              offsets[offsets.length-1] + buffers[buffers.length-1].byteLength
          );
          // use DataView.setUint32 for network byte-order
          var view = new DataView(msg_buf.buffer);
          // write nbufs to first 4 bytes
          view.setUint32(0, nbufs);
          // write offsets to next 4 * nbufs bytes
          for (i = 0; i < offsets.length; i++) {
              view.setUint32(4 * (i+1), offsets[i]);
          }
          // write all the buffers at their respective offsets
          for (i = 0; i < buffers.length; i++) {
              msg_buf.set(new Uint8Array(buffers[i]), offsets[i]);
          }
          
          // return raw ArrayBuffer
          return msg_buf.buffer;
      };
      
      var serialize = function (msg) {
          if (msg.buffers && msg.buffers.length) {
              return _serialize_binary(msg);
          } else {
              return JSON.stringify(msg);
          }
      };
      
      var exports = {
          deserialize : deserialize,
          serialize: serialize
      };
      return exports;
  });

  define("notebook/js/outputarea", [
    "base/js/keyboard",
    "base/js/events"
  ], function (keyboard_manager, events) {

    function createElement() {
      var area = $('<div/>');
      document.body.append(area[0]);
      return area[0];
    }
    function OutputArea() {
      this.keyboard_manager = keyboard_manager;
      this.events = events;
      this.outputs = [];
      this.element = createElement();
    }

    OutputArea.__mime_types = {};

    OutputArea.prototype.register_mime_type = function (mimetype, append, options) {
      OutputArea.__mime_types[mimetype] = {
        append: append,
        options: options
      };
    };

    OutputArea.prototype.create_output_subarea = function (md, classes, mime) {
      var subarea = $('<div/>').addClass('output_subarea').addClass(classes);
      return subarea;
    };

    OutputArea.prototype.create_output_area = function () {
      return $("<div/>").addClass("output_area");
    };

    OutputArea.prototype.append_mime_type = function (json, element, handle_inserted) {
      let types = Object.keys(OutputArea.__mime_types);
      types.forEach(type => {
        var append = OutputArea.__mime_types[type].append;
        if ((json.data[type] !== undefined) && append) {
          var md = json.metadata || {};
          var value = json.data[type];
          return append.apply(this, [value, md, element, handle_inserted]);
        }
      });
      return null;
    };

    OutputArea.prototype.clear_output = function () {
      this.outputs.length = 0;
      this.element.innerHTML = "";
    };

    OutputArea.prototype.fromJSON = function (outputs, metadata) {
      var len = outputs.length;
      metadata = metadata || {};
      for (var i = 0; i < len; i++) {
        this.append_output(outputs[i]);
      }
    };

    OutputArea.prototype.handle_clear_output = function (msg) {
      this.clear_output(msg.content.wait || false);
    };

    OutputArea.prototype.handle_output = function (msg) {
      var json = {};
      var msg_type = json.output_type = msg.header.msg_type;
      var content = msg.content;
      switch (msg_type) {
        case "display_data":
          json.transient = content.transient;
          json.data = content.data;
          json.metadata = content.metadata;
          break;
        default:
          console.error("unhandled output message", msg);
          return;
      }
      this.append_output(json);
    };

    OutputArea.prototype.append_output = function (json) {
      this.append_display_data(json, this.handle_appended);
      this.outputs.push(json);
      this.events.trigger('output_added.OutputArea', {
        output: json,
        output_area: this,
      });
    };

    OutputArea.prototype.append_display_data = function (json, handle_inserted) {
      var toinsert = this.create_output_area();
      this.element.append(toinsert[0]);
      this.append_mime_type(json, toinsert, handle_inserted);
    };

    const outputarea = {
      OutputArea: OutputArea
    };

    return outputarea;

  });

  define("base/js/events", function () {

    const eventTarget = document.createElement("div");

    const events = {
      on: function (type, handler) {
        eventTarget.addEventListener(type, evt => {
          handler(evt, evt.detail);
        });
      },
      trigger: function (type, data) {
        eventTarget.dispatchEvent(new CustomEvent(type, {detail: data}));
      }
    };

    return events;
  });

  define("base/js/keyboard", function () {
    function KeyboardManager() {
    }
    KeyboardManager.prototype.actions = {
      register: function () {
      }
    };

    KeyboardManager.prototype.register_events = function () {
    };

    const keyboard_manager = new KeyboardManager();

    return keyboard_manager;
  });

  define("base/js/namespace", [
    "services/kernels/kernel",
    "notebook/js/outputarea",
    "base/js/keyboard",
    "base/js/events",
  ], function (kernel, outputarea, keyboard_manager, events) {

    const cells = [{
      output_area: new outputarea.OutputArea()
    }];

    const namespace = {
      notebook: {
        kernel: new kernel.Kernel(),
        keyboard_manager: keyboard_manager,
        container: {
          on: function () {
          }
        },
        metadata: {
          widgets: {}
        },
        get_cells: function () {
          return cells;
        },
        render_cell_output: function (cell) {
          const output_area = cell.output_area;
          const outputs = [...output_area.outputs];
          output_area.clear_output();
          output_area.fromJSON(outputs);
        },
        events: events
      },
      notification_area: {
        widget: function () {
        }
      },
      menubar: {
        actions: {
          register: function () {
          }
        }
      }
    };

    return namespace;
  });

  define("base/js/utils", function () {
    return {
      get_body_data: function () {
        return "";
      },
      uuid: function () {
        return Date.now() + "_" + Math.random();
      },
      load_class: function (class_name, module_name, registry) {
        return new Promise(function (resolve, reject) {
          if (registry && registry[class_name]) {
            resolve(registry[class_name]);
          } else {
            reject(new Error('Class ' + class_name + ' not found in registry '));
          }
        });
      },
      reject: function (message, log) {
        return function (error) {
          return Promise.reject(message + "\n" + error);
        }
      },
      debounse: function debounce(fun, delay) {
        var timerId;
        return function () {
          var args = arguments;
          if (timerId) {
            clearTimeout(timerId);
          }
          timerId = setTimeout(function () {
            fun.apply(null, args);
            timerId = null;
          }, delay);
        }
      }
    };
  });

  const eventTarget = document.createElement("div");

  const INIT_REQUEST = "init";
  const SAVE_STATE_REQUEST = "save_state";
  const COMM_MSG_REQUEST = "comm";
  const WIDGET_MODEL_SET_REQUEST = "widget_model_set";

  const DISPLAY_DATA_RESPONSE = "display_data";
  const KERNEL_STATE_RESPONSE = "kernel_state";
  const COMM_MSG_RESPONSE = "comm";
  const SET_STATE_RESPONSE = "set_state";
  const WIDGET_MODEL_SET_RESPONSE = "widget_model_set";

  const id = location.hash.replace("#", "");

  window.addEventListener("message", event => {
    const data = event.data;
    const message = data.message ? JSON.parse(data.message) : undefined;
    eventTarget.dispatchEvent(new CustomEvent(data.type, {detail: message}));
  });

  const channel = {
    send: function (type, message) {
      window.top.postMessage({
        type: type,
        sender: id,
        message: message,
      }, "*");
    },
    on: function (type, handler) {
      eventTarget.addEventListener(type, evt => {
        handler(evt.detail);
      });
    },
    start: function () {
      this.send(INIT_REQUEST);
    }
  };

  function patch(kernel) {

    kernel.set_kernel_state = function (connected) {
      if (!connected) {
        this.comm_info = function (target_name, callback) {
          callback({content: {comms: {}}});
        };
      }

      this._is_connected = true;
      this._kernel_connected();
    };

    kernel.is_connected = function () {
      return this._is_connected;
    };

    kernel._send = function (msg) {
      channel.send(COMM_MSG_REQUEST, msg);
    };

    channel.on(COMM_MSG_RESPONSE, msg => {
      if (msg && msg.content && msg.content.data) {
        msg.content.data.buffer_paths = [];
      }
      kernel._finish_ws_message(msg);
    });
  }

  function patch$1(WidgetModel, widgetManager) {

    const originalSet = WidgetModel.prototype.set;

    WidgetModel.prototype.set = function (key, val, options) {

      if (key == "value") {
        channel.send(WIDGET_MODEL_SET_REQUEST, JSON.stringify({
          model_id: this.model_id,
          val: val,
          key: key
        }));
      }

      return originalSet.apply(this, arguments);
    };

    channel.on(WIDGET_MODEL_SET_RESPONSE, msg => {
      const model_id = msg.model_id;
      const model_promise = widgetManager.get_model(model_id);

      if (model_promise) {
        model_promise.then(model => {
          originalSet.call(model, msg.key, msg.val);
        });
      }
    });
  }

  var toByteArray_1 = toByteArray;
  var fromByteArray_1 = fromByteArray;

  var lookup = [];
  var revLookup = [];
  var Arr = typeof Uint8Array !== 'undefined' ? Uint8Array : Array;

  var code = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
  for (var i = 0, len = code.length; i < len; ++i) {
    lookup[i] = code[i];
    revLookup[code.charCodeAt(i)] = i;
  }

  // Support decoding URL-safe base64 strings, as Node.js does.
  // See: https://en.wikipedia.org/wiki/Base64#URL_applications
  revLookup['-'.charCodeAt(0)] = 62;
  revLookup['_'.charCodeAt(0)] = 63;

  function getLens (b64) {
    var len = b64.length;

    if (len % 4 > 0) {
      throw new Error('Invalid string. Length must be a multiple of 4')
    }

    // Trim off extra bytes after placeholder bytes are found
    // See: https://github.com/beatgammit/base64-js/issues/42
    var validLen = b64.indexOf('=');
    if (validLen === -1) validLen = len;

    var placeHoldersLen = validLen === len
      ? 0
      : 4 - (validLen % 4);

    return [validLen, placeHoldersLen]
  }

  function _byteLength (b64, validLen, placeHoldersLen) {
    return ((validLen + placeHoldersLen) * 3 / 4) - placeHoldersLen
  }

  function toByteArray (b64) {
    var tmp;
    var lens = getLens(b64);
    var validLen = lens[0];
    var placeHoldersLen = lens[1];

    var arr = new Arr(_byteLength(b64, validLen, placeHoldersLen));

    var curByte = 0;

    // if there are placeholders, only get up to the last complete 4 chars
    var len = placeHoldersLen > 0
      ? validLen - 4
      : validLen;

    var i;
    for (i = 0; i < len; i += 4) {
      tmp =
        (revLookup[b64.charCodeAt(i)] << 18) |
        (revLookup[b64.charCodeAt(i + 1)] << 12) |
        (revLookup[b64.charCodeAt(i + 2)] << 6) |
        revLookup[b64.charCodeAt(i + 3)];
      arr[curByte++] = (tmp >> 16) & 0xFF;
      arr[curByte++] = (tmp >> 8) & 0xFF;
      arr[curByte++] = tmp & 0xFF;
    }

    if (placeHoldersLen === 2) {
      tmp =
        (revLookup[b64.charCodeAt(i)] << 2) |
        (revLookup[b64.charCodeAt(i + 1)] >> 4);
      arr[curByte++] = tmp & 0xFF;
    }

    if (placeHoldersLen === 1) {
      tmp =
        (revLookup[b64.charCodeAt(i)] << 10) |
        (revLookup[b64.charCodeAt(i + 1)] << 4) |
        (revLookup[b64.charCodeAt(i + 2)] >> 2);
      arr[curByte++] = (tmp >> 8) & 0xFF;
      arr[curByte++] = tmp & 0xFF;
    }

    return arr
  }

  function tripletToBase64 (num) {
    return lookup[num >> 18 & 0x3F] +
      lookup[num >> 12 & 0x3F] +
      lookup[num >> 6 & 0x3F] +
      lookup[num & 0x3F]
  }

  function encodeChunk (uint8, start, end) {
    var tmp;
    var output = [];
    for (var i = start; i < end; i += 3) {
      tmp =
        ((uint8[i] << 16) & 0xFF0000) +
        ((uint8[i + 1] << 8) & 0xFF00) +
        (uint8[i + 2] & 0xFF);
      output.push(tripletToBase64(tmp));
    }
    return output.join('')
  }

  function fromByteArray (uint8) {
    var tmp;
    var len = uint8.length;
    var extraBytes = len % 3; // if we have 1 byte left, pad 2 bytes
    var parts = [];
    var maxChunkLength = 16383; // must be multiple of 3

    // go through the array every three bytes, we'll deal with trailing stuff later
    for (var i = 0, len2 = len - extraBytes; i < len2; i += maxChunkLength) {
      parts.push(encodeChunk(
        uint8, i, (i + maxChunkLength) > len2 ? len2 : (i + maxChunkLength)
      ));
    }

    // pad the end with zeros, but make sure to not forget the extra bytes
    if (extraBytes === 1) {
      tmp = uint8[len - 1];
      parts.push(
        lookup[tmp >> 2] +
        lookup[(tmp << 4) & 0x3F] +
        '=='
      );
    } else if (extraBytes === 2) {
      tmp = (uint8[len - 2] << 8) + uint8[len - 1];
      parts.push(
        lookup[tmp >> 10] +
        lookup[(tmp >> 4) & 0x3F] +
        lookup[(tmp << 2) & 0x3F] +
        '='
      );
    }

    return parts.join('')
  }

  var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

  function createCommonjsModule(fn, basedir, module) {
  	return module = {
  	  path: basedir,
  	  exports: {},
  	  require: function (path, base) {
        return commonjsRequire(path, (base === undefined || base === null) ? module.path : base);
      }
  	}, fn(module, module.exports), module.exports;
  }

  function getCjsExportFromNamespace (n) {
  	return n && n['default'] || n;
  }

  function commonjsRequire () {
  	throw new Error('Dynamic requires are not currently supported by @rollup/plugin-commonjs');
  }

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for JSON-specific functions.
   */
  var JSONExt;
  (function (JSONExt) {
      /**
       * A shared frozen empty JSONObject
       */
      JSONExt.emptyObject = Object.freeze({});
      /**
       * A shared frozen empty JSONArray
       */
      JSONExt.emptyArray = Object.freeze([]);
      /**
       * Test whether a JSON value is a primitive.
       *
       * @param value - The JSON value of interest.
       *
       * @returns `true` if the value is a primitive,`false` otherwise.
       */
      function isPrimitive(value) {
          return (value === null ||
              typeof value === 'boolean' ||
              typeof value === 'number' ||
              typeof value === 'string');
      }
      JSONExt.isPrimitive = isPrimitive;
      function isArray(value) {
          return Array.isArray(value);
      }
      JSONExt.isArray = isArray;
      function isObject(value) {
          return !isPrimitive(value) && !isArray(value);
      }
      JSONExt.isObject = isObject;
      /**
       * Compare two JSON values for deep equality.
       *
       * @param first - The first JSON value of interest.
       *
       * @param second - The second JSON value of interest.
       *
       * @returns `true` if the values are equivalent, `false` otherwise.
       */
      function deepEqual(first, second) {
          // Check referential and primitive equality first.
          if (first === second) {
              return true;
          }
          // If one is a primitive, the `===` check ruled out the other.
          if (isPrimitive(first) || isPrimitive(second)) {
              return false;
          }
          // Test whether they are arrays.
          var a1 = isArray(first);
          var a2 = isArray(second);
          // Bail if the types are different.
          if (a1 !== a2) {
              return false;
          }
          // If they are both arrays, compare them.
          if (a1 && a2) {
              return deepArrayEqual(first, second);
          }
          // At this point, they must both be objects.
          return deepObjectEqual(first, second);
      }
      JSONExt.deepEqual = deepEqual;
      /**
       * Create a deep copy of a JSON value.
       *
       * @param value - The JSON value to copy.
       *
       * @returns A deep copy of the given JSON value.
       */
      function deepCopy(value) {
          // Do nothing for primitive values.
          if (isPrimitive(value)) {
              return value;
          }
          // Deep copy an array.
          if (isArray(value)) {
              return deepArrayCopy(value);
          }
          // Deep copy an object.
          return deepObjectCopy(value);
      }
      JSONExt.deepCopy = deepCopy;
      /**
       * Compare two JSON arrays for deep equality.
       */
      function deepArrayEqual(first, second) {
          // Check referential equality first.
          if (first === second) {
              return true;
          }
          // Test the arrays for equal length.
          if (first.length !== second.length) {
              return false;
          }
          // Compare the values for equality.
          for (var i = 0, n = first.length; i < n; ++i) {
              if (!deepEqual(first[i], second[i])) {
                  return false;
              }
          }
          // At this point, the arrays are equal.
          return true;
      }
      /**
       * Compare two JSON objects for deep equality.
       */
      function deepObjectEqual(first, second) {
          // Check referential equality first.
          if (first === second) {
              return true;
          }
          // Check for the first object's keys in the second object.
          for (var key in first) {
              if (first[key] !== undefined && !(key in second)) {
                  return false;
              }
          }
          // Check for the second object's keys in the first object.
          for (var key in second) {
              if (second[key] !== undefined && !(key in first)) {
                  return false;
              }
          }
          // Compare the values for equality.
          for (var key in first) {
              // Get the values.
              var firstValue = first[key];
              var secondValue = second[key];
              // If both are undefined, ignore the key.
              if (firstValue === undefined && secondValue === undefined) {
                  continue;
              }
              // If only one value is undefined, the objects are not equal.
              if (firstValue === undefined || secondValue === undefined) {
                  return false;
              }
              // Compare the values.
              if (!deepEqual(firstValue, secondValue)) {
                  return false;
              }
          }
          // At this point, the objects are equal.
          return true;
      }
      /**
       * Create a deep copy of a JSON array.
       */
      function deepArrayCopy(value) {
          var result = new Array(value.length);
          for (var i = 0, n = value.length; i < n; ++i) {
              result[i] = deepCopy(value[i]);
          }
          return result;
      }
      /**
       * Create a deep copy of a JSON object.
       */
      function deepObjectCopy(value) {
          var result = {};
          for (var key in value) {
              // Ignore undefined values.
              var subvalue = value[key];
              if (subvalue === undefined) {
                  continue;
              }
              result[key] = deepCopy(subvalue);
          }
          return result;
      }
  })(JSONExt || (JSONExt = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * An object which stores MIME data for general application use.
   *
   * #### Notes
   * This class does not attempt to enforce "correctness" of MIME types
   * and their associated data. Since this class is designed to transfer
   * arbitrary data and objects within the same application, it assumes
   * that the user provides correct and accurate data.
   */
  var MimeData = /** @class */ (function () {
      function MimeData() {
          this._types = [];
          this._values = [];
      }
      /**
       * Get an array of the MIME types contained within the dataset.
       *
       * @returns A new array of the MIME types, in order of insertion.
       */
      MimeData.prototype.types = function () {
          return this._types.slice();
      };
      /**
       * Test whether the dataset has an entry for the given type.
       *
       * @param mime - The MIME type of interest.
       *
       * @returns `true` if the dataset contains a value for the given
       *   MIME type, `false` otherwise.
       */
      MimeData.prototype.hasData = function (mime) {
          return this._types.indexOf(mime) !== -1;
      };
      /**
       * Get the data value for the given MIME type.
       *
       * @param mime - The MIME type of interest.
       *
       * @returns The value for the given MIME type, or `undefined` if
       *   the dataset does not contain a value for the type.
       */
      MimeData.prototype.getData = function (mime) {
          var i = this._types.indexOf(mime);
          return i !== -1 ? this._values[i] : undefined;
      };
      /**
       * Set the data value for the given MIME type.
       *
       * @param mime - The MIME type of interest.
       *
       * @param data - The data value for the given MIME type.
       *
       * #### Notes
       * This will overwrite any previous entry for the MIME type.
       */
      MimeData.prototype.setData = function (mime, data) {
          this.clearData(mime);
          this._types.push(mime);
          this._values.push(data);
      };
      /**
       * Remove the data entry for the given MIME type.
       *
       * @param mime - The MIME type of interest.
       *
       * #### Notes
       * This is a no-op if there is no entry for the given MIME type.
       */
      MimeData.prototype.clearData = function (mime) {
          var i = this._types.indexOf(mime);
          if (i !== -1) {
              this._types.splice(i, 1);
              this._values.splice(i, 1);
          }
      };
      /**
       * Remove all data entries from the dataset.
       */
      MimeData.prototype.clear = function () {
          this._types.length = 0;
          this._values.length = 0;
      };
      return MimeData;
  }());

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for random number related functionality.
   */
  var Random;
  (function (Random) {
      /**
       * A function which generates random bytes.
       *
       * @param buffer - The `Uint8Array` to fill with random bytes.
       *
       * #### Notes
       * A cryptographically strong random number generator will be used if
       * available. Otherwise, `Math.random` will be used as a fallback for
       * randomness.
       *
       * The following RNGs are supported, listed in order of precedence:
       *   - `window.crypto.getRandomValues`
       *   - `window.msCrypto.getRandomValues`
       *   - `require('crypto').randomFillSync
       *   - `require('crypto').randomBytes
       *   - `Math.random`
       */
      Random.getRandomValues = (function () {
          // Look up the crypto module if available.
          var crypto = ((typeof window !== 'undefined' && (window.crypto || window.msCrypto)) ||
              (typeof commonjsRequire !== 'undefined' && require('crypto')) || null);
          // Modern browsers and IE 11
          if (crypto && typeof crypto.getRandomValues === 'function') {
              return function getRandomValues(buffer) {
                  return crypto.getRandomValues(buffer);
              };
          }
          // Node 7+
          if (crypto && typeof crypto.randomFillSync === 'function') {
              return function getRandomValues(buffer) {
                  return crypto.randomFillSync(buffer);
              };
          }
          // Node 0.10+
          if (crypto && typeof crypto.randomBytes === 'function') {
              return function getRandomValues(buffer) {
                  var bytes = crypto.randomBytes(buffer.length);
                  for (var i = 0, n = bytes.length; i < n; ++i) {
                      buffer[i] = bytes[i];
                  }
              };
          }
          // Fallback
          return function getRandomValues(buffer) {
              var value = 0;
              for (var i = 0, n = buffer.length; i < n; ++i) {
                  if (i % 4 === 0) {
                      value = Math.random() * 0xFFFFFFFF >>> 0;
                  }
                  buffer[i] = value & 0xFF;
                  value >>>= 8;
              }
          };
      })();
  })(Random || (Random = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * A runtime object which captures compile-time type information.
   *
   * #### Notes
   * A token captures the compile-time type of an interface or class in
   * an object which can be used at runtime in a type-safe fashion.
   */
  var Token = /** @class */ (function () {
      /**
       * Construct a new token.
       *
       * @param name - A human readable name for the token.
       */
      function Token(name) {
          this.name = name;
          this._tokenStructuralPropertyT = null;
      }
      return Token;
  }());

  // Copyright (c) Jupyter Development Team.
  /**
   * The namespace for UUID related functionality.
   */
  var UUID;
  (function (UUID) {
      /**
       * A function which generates UUID v4 identifiers.
       *
       * @returns A new UUID v4 string.
       *
       * #### Notes
       * This implementation complies with RFC 4122.
       *
       * This uses `Random.getRandomValues()` for random bytes, which in
       * turn will use the underlying `crypto` module of the platform if
       * it is available. The fallback for randomness is `Math.random`.
       */
      UUID.uuid4 = (function () {
          // Create a 16 byte array to hold the random values.
          var bytes = new Uint8Array(16);
          // Create a look up table from bytes to hex strings.
          var lut = new Array(256);
          // Pad the single character hex digits with a leading zero.
          for (var i = 0; i < 16; ++i) {
              lut[i] = '0' + i.toString(16);
          }
          // Populate the rest of the hex digits.
          for (var i = 16; i < 256; ++i) {
              lut[i] = i.toString(16);
          }
          // Return a function which generates the UUID.
          return function uuid4() {
              // Get a new batch of random values.
              Random.getRandomValues(bytes);
              // Set the UUID version number to 4.
              bytes[6] = 0x40 | (bytes[6] & 0x0F);
              // Set the clock sequence bit to the RFC spec.
              bytes[8] = 0x80 | (bytes[8] & 0x3F);
              // Assemble the UUID string.
              return (lut[bytes[0]] +
                  lut[bytes[1]] +
                  lut[bytes[2]] +
                  lut[bytes[3]] +
                  '-' +
                  lut[bytes[4]] +
                  lut[bytes[5]] +
                  '-' +
                  lut[bytes[6]] +
                  lut[bytes[7]] +
                  '-' +
                  lut[bytes[8]] +
                  lut[bytes[9]] +
                  '-' +
                  lut[bytes[10]] +
                  lut[bytes[11]] +
                  lut[bytes[12]] +
                  lut[bytes[13]] +
                  lut[bytes[14]] +
                  lut[bytes[15]]);
          };
      })();
  })(UUID || (UUID = {}));

  /**
   * Removes all key-value entries from the list cache.
   *
   * @private
   * @name clear
   * @memberOf ListCache
   */
  function listCacheClear() {
    this.__data__ = [];
    this.size = 0;
  }

  var _listCacheClear = listCacheClear;

  /**
   * Performs a
   * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
   * comparison between two values to determine if they are equivalent.
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to compare.
   * @param {*} other The other value to compare.
   * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
   * @example
   *
   * var object = { 'a': 1 };
   * var other = { 'a': 1 };
   *
   * _.eq(object, object);
   * // => true
   *
   * _.eq(object, other);
   * // => false
   *
   * _.eq('a', 'a');
   * // => true
   *
   * _.eq('a', Object('a'));
   * // => false
   *
   * _.eq(NaN, NaN);
   * // => true
   */
  function eq(value, other) {
    return value === other || (value !== value && other !== other);
  }

  var eq_1 = eq;

  /**
   * Gets the index at which the `key` is found in `array` of key-value pairs.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {*} key The key to search for.
   * @returns {number} Returns the index of the matched value, else `-1`.
   */
  function assocIndexOf(array, key) {
    var length = array.length;
    while (length--) {
      if (eq_1(array[length][0], key)) {
        return length;
      }
    }
    return -1;
  }

  var _assocIndexOf = assocIndexOf;

  /** Used for built-in method references. */
  var arrayProto = Array.prototype;

  /** Built-in value references. */
  var splice = arrayProto.splice;

  /**
   * Removes `key` and its value from the list cache.
   *
   * @private
   * @name delete
   * @memberOf ListCache
   * @param {string} key The key of the value to remove.
   * @returns {boolean} Returns `true` if the entry was removed, else `false`.
   */
  function listCacheDelete(key) {
    var data = this.__data__,
        index = _assocIndexOf(data, key);

    if (index < 0) {
      return false;
    }
    var lastIndex = data.length - 1;
    if (index == lastIndex) {
      data.pop();
    } else {
      splice.call(data, index, 1);
    }
    --this.size;
    return true;
  }

  var _listCacheDelete = listCacheDelete;

  /**
   * Gets the list cache value for `key`.
   *
   * @private
   * @name get
   * @memberOf ListCache
   * @param {string} key The key of the value to get.
   * @returns {*} Returns the entry value.
   */
  function listCacheGet(key) {
    var data = this.__data__,
        index = _assocIndexOf(data, key);

    return index < 0 ? undefined : data[index][1];
  }

  var _listCacheGet = listCacheGet;

  /**
   * Checks if a list cache value for `key` exists.
   *
   * @private
   * @name has
   * @memberOf ListCache
   * @param {string} key The key of the entry to check.
   * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
   */
  function listCacheHas(key) {
    return _assocIndexOf(this.__data__, key) > -1;
  }

  var _listCacheHas = listCacheHas;

  /**
   * Sets the list cache `key` to `value`.
   *
   * @private
   * @name set
   * @memberOf ListCache
   * @param {string} key The key of the value to set.
   * @param {*} value The value to set.
   * @returns {Object} Returns the list cache instance.
   */
  function listCacheSet(key, value) {
    var data = this.__data__,
        index = _assocIndexOf(data, key);

    if (index < 0) {
      ++this.size;
      data.push([key, value]);
    } else {
      data[index][1] = value;
    }
    return this;
  }

  var _listCacheSet = listCacheSet;

  /**
   * Creates an list cache object.
   *
   * @private
   * @constructor
   * @param {Array} [entries] The key-value pairs to cache.
   */
  function ListCache(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }

  // Add methods to `ListCache`.
  ListCache.prototype.clear = _listCacheClear;
  ListCache.prototype['delete'] = _listCacheDelete;
  ListCache.prototype.get = _listCacheGet;
  ListCache.prototype.has = _listCacheHas;
  ListCache.prototype.set = _listCacheSet;

  var _ListCache = ListCache;

  /**
   * Removes all key-value entries from the stack.
   *
   * @private
   * @name clear
   * @memberOf Stack
   */
  function stackClear() {
    this.__data__ = new _ListCache;
    this.size = 0;
  }

  var _stackClear = stackClear;

  /**
   * Removes `key` and its value from the stack.
   *
   * @private
   * @name delete
   * @memberOf Stack
   * @param {string} key The key of the value to remove.
   * @returns {boolean} Returns `true` if the entry was removed, else `false`.
   */
  function stackDelete(key) {
    var data = this.__data__,
        result = data['delete'](key);

    this.size = data.size;
    return result;
  }

  var _stackDelete = stackDelete;

  /**
   * Gets the stack value for `key`.
   *
   * @private
   * @name get
   * @memberOf Stack
   * @param {string} key The key of the value to get.
   * @returns {*} Returns the entry value.
   */
  function stackGet(key) {
    return this.__data__.get(key);
  }

  var _stackGet = stackGet;

  /**
   * Checks if a stack value for `key` exists.
   *
   * @private
   * @name has
   * @memberOf Stack
   * @param {string} key The key of the entry to check.
   * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
   */
  function stackHas(key) {
    return this.__data__.has(key);
  }

  var _stackHas = stackHas;

  /** Detect free variable `global` from Node.js. */
  var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

  var _freeGlobal = freeGlobal;

  /** Detect free variable `self`. */
  var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

  /** Used as a reference to the global object. */
  var root = _freeGlobal || freeSelf || Function('return this')();

  var _root = root;

  /** Built-in value references. */
  var Symbol$1 = _root.Symbol;

  var _Symbol = Symbol$1;

  /** Used for built-in method references. */
  var objectProto = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty = objectProto.hasOwnProperty;

  /**
   * Used to resolve the
   * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
   * of values.
   */
  var nativeObjectToString = objectProto.toString;

  /** Built-in value references. */
  var symToStringTag = _Symbol ? _Symbol.toStringTag : undefined;

  /**
   * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
   *
   * @private
   * @param {*} value The value to query.
   * @returns {string} Returns the raw `toStringTag`.
   */
  function getRawTag(value) {
    var isOwn = hasOwnProperty.call(value, symToStringTag),
        tag = value[symToStringTag];

    try {
      value[symToStringTag] = undefined;
      var unmasked = true;
    } catch (e) {}

    var result = nativeObjectToString.call(value);
    if (unmasked) {
      if (isOwn) {
        value[symToStringTag] = tag;
      } else {
        delete value[symToStringTag];
      }
    }
    return result;
  }

  var _getRawTag = getRawTag;

  /** Used for built-in method references. */
  var objectProto$1 = Object.prototype;

  /**
   * Used to resolve the
   * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
   * of values.
   */
  var nativeObjectToString$1 = objectProto$1.toString;

  /**
   * Converts `value` to a string using `Object.prototype.toString`.
   *
   * @private
   * @param {*} value The value to convert.
   * @returns {string} Returns the converted string.
   */
  function objectToString(value) {
    return nativeObjectToString$1.call(value);
  }

  var _objectToString = objectToString;

  /** `Object#toString` result references. */
  var nullTag = '[object Null]',
      undefinedTag = '[object Undefined]';

  /** Built-in value references. */
  var symToStringTag$1 = _Symbol ? _Symbol.toStringTag : undefined;

  /**
   * The base implementation of `getTag` without fallbacks for buggy environments.
   *
   * @private
   * @param {*} value The value to query.
   * @returns {string} Returns the `toStringTag`.
   */
  function baseGetTag(value) {
    if (value == null) {
      return value === undefined ? undefinedTag : nullTag;
    }
    return (symToStringTag$1 && symToStringTag$1 in Object(value))
      ? _getRawTag(value)
      : _objectToString(value);
  }

  var _baseGetTag = baseGetTag;

  /**
   * Checks if `value` is the
   * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
   * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
   *
   * @static
   * @memberOf _
   * @since 0.1.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is an object, else `false`.
   * @example
   *
   * _.isObject({});
   * // => true
   *
   * _.isObject([1, 2, 3]);
   * // => true
   *
   * _.isObject(_.noop);
   * // => true
   *
   * _.isObject(null);
   * // => false
   */
  function isObject(value) {
    var type = typeof value;
    return value != null && (type == 'object' || type == 'function');
  }

  var isObject_1 = isObject;

  /** `Object#toString` result references. */
  var asyncTag = '[object AsyncFunction]',
      funcTag = '[object Function]',
      genTag = '[object GeneratorFunction]',
      proxyTag = '[object Proxy]';

  /**
   * Checks if `value` is classified as a `Function` object.
   *
   * @static
   * @memberOf _
   * @since 0.1.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a function, else `false`.
   * @example
   *
   * _.isFunction(_);
   * // => true
   *
   * _.isFunction(/abc/);
   * // => false
   */
  function isFunction(value) {
    if (!isObject_1(value)) {
      return false;
    }
    // The use of `Object#toString` avoids issues with the `typeof` operator
    // in Safari 9 which returns 'object' for typed arrays and other constructors.
    var tag = _baseGetTag(value);
    return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
  }

  var isFunction_1 = isFunction;

  /** Used to detect overreaching core-js shims. */
  var coreJsData = _root['__core-js_shared__'];

  var _coreJsData = coreJsData;

  /** Used to detect methods masquerading as native. */
  var maskSrcKey = (function() {
    var uid = /[^.]+$/.exec(_coreJsData && _coreJsData.keys && _coreJsData.keys.IE_PROTO || '');
    return uid ? ('Symbol(src)_1.' + uid) : '';
  }());

  /**
   * Checks if `func` has its source masked.
   *
   * @private
   * @param {Function} func The function to check.
   * @returns {boolean} Returns `true` if `func` is masked, else `false`.
   */
  function isMasked(func) {
    return !!maskSrcKey && (maskSrcKey in func);
  }

  var _isMasked = isMasked;

  /** Used for built-in method references. */
  var funcProto = Function.prototype;

  /** Used to resolve the decompiled source of functions. */
  var funcToString = funcProto.toString;

  /**
   * Converts `func` to its source code.
   *
   * @private
   * @param {Function} func The function to convert.
   * @returns {string} Returns the source code.
   */
  function toSource(func) {
    if (func != null) {
      try {
        return funcToString.call(func);
      } catch (e) {}
      try {
        return (func + '');
      } catch (e) {}
    }
    return '';
  }

  var _toSource = toSource;

  /**
   * Used to match `RegExp`
   * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
   */
  var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;

  /** Used to detect host constructors (Safari). */
  var reIsHostCtor = /^\[object .+?Constructor\]$/;

  /** Used for built-in method references. */
  var funcProto$1 = Function.prototype,
      objectProto$2 = Object.prototype;

  /** Used to resolve the decompiled source of functions. */
  var funcToString$1 = funcProto$1.toString;

  /** Used to check objects for own properties. */
  var hasOwnProperty$1 = objectProto$2.hasOwnProperty;

  /** Used to detect if a method is native. */
  var reIsNative = RegExp('^' +
    funcToString$1.call(hasOwnProperty$1).replace(reRegExpChar, '\\$&')
    .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
  );

  /**
   * The base implementation of `_.isNative` without bad shim checks.
   *
   * @private
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a native function,
   *  else `false`.
   */
  function baseIsNative(value) {
    if (!isObject_1(value) || _isMasked(value)) {
      return false;
    }
    var pattern = isFunction_1(value) ? reIsNative : reIsHostCtor;
    return pattern.test(_toSource(value));
  }

  var _baseIsNative = baseIsNative;

  /**
   * Gets the value at `key` of `object`.
   *
   * @private
   * @param {Object} [object] The object to query.
   * @param {string} key The key of the property to get.
   * @returns {*} Returns the property value.
   */
  function getValue(object, key) {
    return object == null ? undefined : object[key];
  }

  var _getValue = getValue;

  /**
   * Gets the native function at `key` of `object`.
   *
   * @private
   * @param {Object} object The object to query.
   * @param {string} key The key of the method to get.
   * @returns {*} Returns the function if it's native, else `undefined`.
   */
  function getNative(object, key) {
    var value = _getValue(object, key);
    return _baseIsNative(value) ? value : undefined;
  }

  var _getNative = getNative;

  /* Built-in method references that are verified to be native. */
  var Map$1 = _getNative(_root, 'Map');

  var _Map = Map$1;

  /* Built-in method references that are verified to be native. */
  var nativeCreate = _getNative(Object, 'create');

  var _nativeCreate = nativeCreate;

  /**
   * Removes all key-value entries from the hash.
   *
   * @private
   * @name clear
   * @memberOf Hash
   */
  function hashClear() {
    this.__data__ = _nativeCreate ? _nativeCreate(null) : {};
    this.size = 0;
  }

  var _hashClear = hashClear;

  /**
   * Removes `key` and its value from the hash.
   *
   * @private
   * @name delete
   * @memberOf Hash
   * @param {Object} hash The hash to modify.
   * @param {string} key The key of the value to remove.
   * @returns {boolean} Returns `true` if the entry was removed, else `false`.
   */
  function hashDelete(key) {
    var result = this.has(key) && delete this.__data__[key];
    this.size -= result ? 1 : 0;
    return result;
  }

  var _hashDelete = hashDelete;

  /** Used to stand-in for `undefined` hash values. */
  var HASH_UNDEFINED = '__lodash_hash_undefined__';

  /** Used for built-in method references. */
  var objectProto$3 = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$2 = objectProto$3.hasOwnProperty;

  /**
   * Gets the hash value for `key`.
   *
   * @private
   * @name get
   * @memberOf Hash
   * @param {string} key The key of the value to get.
   * @returns {*} Returns the entry value.
   */
  function hashGet(key) {
    var data = this.__data__;
    if (_nativeCreate) {
      var result = data[key];
      return result === HASH_UNDEFINED ? undefined : result;
    }
    return hasOwnProperty$2.call(data, key) ? data[key] : undefined;
  }

  var _hashGet = hashGet;

  /** Used for built-in method references. */
  var objectProto$4 = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$3 = objectProto$4.hasOwnProperty;

  /**
   * Checks if a hash value for `key` exists.
   *
   * @private
   * @name has
   * @memberOf Hash
   * @param {string} key The key of the entry to check.
   * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
   */
  function hashHas(key) {
    var data = this.__data__;
    return _nativeCreate ? (data[key] !== undefined) : hasOwnProperty$3.call(data, key);
  }

  var _hashHas = hashHas;

  /** Used to stand-in for `undefined` hash values. */
  var HASH_UNDEFINED$1 = '__lodash_hash_undefined__';

  /**
   * Sets the hash `key` to `value`.
   *
   * @private
   * @name set
   * @memberOf Hash
   * @param {string} key The key of the value to set.
   * @param {*} value The value to set.
   * @returns {Object} Returns the hash instance.
   */
  function hashSet(key, value) {
    var data = this.__data__;
    this.size += this.has(key) ? 0 : 1;
    data[key] = (_nativeCreate && value === undefined) ? HASH_UNDEFINED$1 : value;
    return this;
  }

  var _hashSet = hashSet;

  /**
   * Creates a hash object.
   *
   * @private
   * @constructor
   * @param {Array} [entries] The key-value pairs to cache.
   */
  function Hash(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }

  // Add methods to `Hash`.
  Hash.prototype.clear = _hashClear;
  Hash.prototype['delete'] = _hashDelete;
  Hash.prototype.get = _hashGet;
  Hash.prototype.has = _hashHas;
  Hash.prototype.set = _hashSet;

  var _Hash = Hash;

  /**
   * Removes all key-value entries from the map.
   *
   * @private
   * @name clear
   * @memberOf MapCache
   */
  function mapCacheClear() {
    this.size = 0;
    this.__data__ = {
      'hash': new _Hash,
      'map': new (_Map || _ListCache),
      'string': new _Hash
    };
  }

  var _mapCacheClear = mapCacheClear;

  /**
   * Checks if `value` is suitable for use as unique object key.
   *
   * @private
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is suitable, else `false`.
   */
  function isKeyable(value) {
    var type = typeof value;
    return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
      ? (value !== '__proto__')
      : (value === null);
  }

  var _isKeyable = isKeyable;

  /**
   * Gets the data for `map`.
   *
   * @private
   * @param {Object} map The map to query.
   * @param {string} key The reference key.
   * @returns {*} Returns the map data.
   */
  function getMapData(map, key) {
    var data = map.__data__;
    return _isKeyable(key)
      ? data[typeof key == 'string' ? 'string' : 'hash']
      : data.map;
  }

  var _getMapData = getMapData;

  /**
   * Removes `key` and its value from the map.
   *
   * @private
   * @name delete
   * @memberOf MapCache
   * @param {string} key The key of the value to remove.
   * @returns {boolean} Returns `true` if the entry was removed, else `false`.
   */
  function mapCacheDelete(key) {
    var result = _getMapData(this, key)['delete'](key);
    this.size -= result ? 1 : 0;
    return result;
  }

  var _mapCacheDelete = mapCacheDelete;

  /**
   * Gets the map value for `key`.
   *
   * @private
   * @name get
   * @memberOf MapCache
   * @param {string} key The key of the value to get.
   * @returns {*} Returns the entry value.
   */
  function mapCacheGet(key) {
    return _getMapData(this, key).get(key);
  }

  var _mapCacheGet = mapCacheGet;

  /**
   * Checks if a map value for `key` exists.
   *
   * @private
   * @name has
   * @memberOf MapCache
   * @param {string} key The key of the entry to check.
   * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
   */
  function mapCacheHas(key) {
    return _getMapData(this, key).has(key);
  }

  var _mapCacheHas = mapCacheHas;

  /**
   * Sets the map `key` to `value`.
   *
   * @private
   * @name set
   * @memberOf MapCache
   * @param {string} key The key of the value to set.
   * @param {*} value The value to set.
   * @returns {Object} Returns the map cache instance.
   */
  function mapCacheSet(key, value) {
    var data = _getMapData(this, key),
        size = data.size;

    data.set(key, value);
    this.size += data.size == size ? 0 : 1;
    return this;
  }

  var _mapCacheSet = mapCacheSet;

  /**
   * Creates a map cache object to store key-value pairs.
   *
   * @private
   * @constructor
   * @param {Array} [entries] The key-value pairs to cache.
   */
  function MapCache(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }

  // Add methods to `MapCache`.
  MapCache.prototype.clear = _mapCacheClear;
  MapCache.prototype['delete'] = _mapCacheDelete;
  MapCache.prototype.get = _mapCacheGet;
  MapCache.prototype.has = _mapCacheHas;
  MapCache.prototype.set = _mapCacheSet;

  var _MapCache = MapCache;

  /** Used as the size to enable large array optimizations. */
  var LARGE_ARRAY_SIZE = 200;

  /**
   * Sets the stack `key` to `value`.
   *
   * @private
   * @name set
   * @memberOf Stack
   * @param {string} key The key of the value to set.
   * @param {*} value The value to set.
   * @returns {Object} Returns the stack cache instance.
   */
  function stackSet(key, value) {
    var data = this.__data__;
    if (data instanceof _ListCache) {
      var pairs = data.__data__;
      if (!_Map || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
        pairs.push([key, value]);
        this.size = ++data.size;
        return this;
      }
      data = this.__data__ = new _MapCache(pairs);
    }
    data.set(key, value);
    this.size = data.size;
    return this;
  }

  var _stackSet = stackSet;

  /**
   * Creates a stack cache object to store key-value pairs.
   *
   * @private
   * @constructor
   * @param {Array} [entries] The key-value pairs to cache.
   */
  function Stack(entries) {
    var data = this.__data__ = new _ListCache(entries);
    this.size = data.size;
  }

  // Add methods to `Stack`.
  Stack.prototype.clear = _stackClear;
  Stack.prototype['delete'] = _stackDelete;
  Stack.prototype.get = _stackGet;
  Stack.prototype.has = _stackHas;
  Stack.prototype.set = _stackSet;

  var _Stack = Stack;

  /** Used to stand-in for `undefined` hash values. */
  var HASH_UNDEFINED$2 = '__lodash_hash_undefined__';

  /**
   * Adds `value` to the array cache.
   *
   * @private
   * @name add
   * @memberOf SetCache
   * @alias push
   * @param {*} value The value to cache.
   * @returns {Object} Returns the cache instance.
   */
  function setCacheAdd(value) {
    this.__data__.set(value, HASH_UNDEFINED$2);
    return this;
  }

  var _setCacheAdd = setCacheAdd;

  /**
   * Checks if `value` is in the array cache.
   *
   * @private
   * @name has
   * @memberOf SetCache
   * @param {*} value The value to search for.
   * @returns {number} Returns `true` if `value` is found, else `false`.
   */
  function setCacheHas(value) {
    return this.__data__.has(value);
  }

  var _setCacheHas = setCacheHas;

  /**
   *
   * Creates an array cache object to store unique values.
   *
   * @private
   * @constructor
   * @param {Array} [values] The values to cache.
   */
  function SetCache(values) {
    var index = -1,
        length = values == null ? 0 : values.length;

    this.__data__ = new _MapCache;
    while (++index < length) {
      this.add(values[index]);
    }
  }

  // Add methods to `SetCache`.
  SetCache.prototype.add = SetCache.prototype.push = _setCacheAdd;
  SetCache.prototype.has = _setCacheHas;

  var _SetCache = SetCache;

  /**
   * A specialized version of `_.some` for arrays without support for iteratee
   * shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} predicate The function invoked per iteration.
   * @returns {boolean} Returns `true` if any element passes the predicate check,
   *  else `false`.
   */
  function arraySome(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (predicate(array[index], index, array)) {
        return true;
      }
    }
    return false;
  }

  var _arraySome = arraySome;

  /**
   * Checks if a `cache` value for `key` exists.
   *
   * @private
   * @param {Object} cache The cache to query.
   * @param {string} key The key of the entry to check.
   * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
   */
  function cacheHas(cache, key) {
    return cache.has(key);
  }

  var _cacheHas = cacheHas;

  /** Used to compose bitmasks for value comparisons. */
  var COMPARE_PARTIAL_FLAG = 1,
      COMPARE_UNORDERED_FLAG = 2;

  /**
   * A specialized version of `baseIsEqualDeep` for arrays with support for
   * partial deep comparisons.
   *
   * @private
   * @param {Array} array The array to compare.
   * @param {Array} other The other array to compare.
   * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
   * @param {Function} customizer The function to customize comparisons.
   * @param {Function} equalFunc The function to determine equivalents of values.
   * @param {Object} stack Tracks traversed `array` and `other` objects.
   * @returns {boolean} Returns `true` if the arrays are equivalent, else `false`.
   */
  function equalArrays(array, other, bitmask, customizer, equalFunc, stack) {
    var isPartial = bitmask & COMPARE_PARTIAL_FLAG,
        arrLength = array.length,
        othLength = other.length;

    if (arrLength != othLength && !(isPartial && othLength > arrLength)) {
      return false;
    }
    // Assume cyclic values are equal.
    var stacked = stack.get(array);
    if (stacked && stack.get(other)) {
      return stacked == other;
    }
    var index = -1,
        result = true,
        seen = (bitmask & COMPARE_UNORDERED_FLAG) ? new _SetCache : undefined;

    stack.set(array, other);
    stack.set(other, array);

    // Ignore non-index properties.
    while (++index < arrLength) {
      var arrValue = array[index],
          othValue = other[index];

      if (customizer) {
        var compared = isPartial
          ? customizer(othValue, arrValue, index, other, array, stack)
          : customizer(arrValue, othValue, index, array, other, stack);
      }
      if (compared !== undefined) {
        if (compared) {
          continue;
        }
        result = false;
        break;
      }
      // Recursively compare arrays (susceptible to call stack limits).
      if (seen) {
        if (!_arraySome(other, function(othValue, othIndex) {
              if (!_cacheHas(seen, othIndex) &&
                  (arrValue === othValue || equalFunc(arrValue, othValue, bitmask, customizer, stack))) {
                return seen.push(othIndex);
              }
            })) {
          result = false;
          break;
        }
      } else if (!(
            arrValue === othValue ||
              equalFunc(arrValue, othValue, bitmask, customizer, stack)
          )) {
        result = false;
        break;
      }
    }
    stack['delete'](array);
    stack['delete'](other);
    return result;
  }

  var _equalArrays = equalArrays;

  /** Built-in value references. */
  var Uint8Array$1 = _root.Uint8Array;

  var _Uint8Array = Uint8Array$1;

  /**
   * Converts `map` to its key-value pairs.
   *
   * @private
   * @param {Object} map The map to convert.
   * @returns {Array} Returns the key-value pairs.
   */
  function mapToArray(map) {
    var index = -1,
        result = Array(map.size);

    map.forEach(function(value, key) {
      result[++index] = [key, value];
    });
    return result;
  }

  var _mapToArray = mapToArray;

  /**
   * Converts `set` to an array of its values.
   *
   * @private
   * @param {Object} set The set to convert.
   * @returns {Array} Returns the values.
   */
  function setToArray(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = value;
    });
    return result;
  }

  var _setToArray = setToArray;

  /** Used to compose bitmasks for value comparisons. */
  var COMPARE_PARTIAL_FLAG$1 = 1,
      COMPARE_UNORDERED_FLAG$1 = 2;

  /** `Object#toString` result references. */
  var boolTag = '[object Boolean]',
      dateTag = '[object Date]',
      errorTag = '[object Error]',
      mapTag = '[object Map]',
      numberTag = '[object Number]',
      regexpTag = '[object RegExp]',
      setTag = '[object Set]',
      stringTag = '[object String]',
      symbolTag = '[object Symbol]';

  var arrayBufferTag = '[object ArrayBuffer]',
      dataViewTag = '[object DataView]';

  /** Used to convert symbols to primitives and strings. */
  var symbolProto = _Symbol ? _Symbol.prototype : undefined,
      symbolValueOf = symbolProto ? symbolProto.valueOf : undefined;

  /**
   * A specialized version of `baseIsEqualDeep` for comparing objects of
   * the same `toStringTag`.
   *
   * **Note:** This function only supports comparing values with tags of
   * `Boolean`, `Date`, `Error`, `Number`, `RegExp`, or `String`.
   *
   * @private
   * @param {Object} object The object to compare.
   * @param {Object} other The other object to compare.
   * @param {string} tag The `toStringTag` of the objects to compare.
   * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
   * @param {Function} customizer The function to customize comparisons.
   * @param {Function} equalFunc The function to determine equivalents of values.
   * @param {Object} stack Tracks traversed `object` and `other` objects.
   * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
   */
  function equalByTag(object, other, tag, bitmask, customizer, equalFunc, stack) {
    switch (tag) {
      case dataViewTag:
        if ((object.byteLength != other.byteLength) ||
            (object.byteOffset != other.byteOffset)) {
          return false;
        }
        object = object.buffer;
        other = other.buffer;

      case arrayBufferTag:
        if ((object.byteLength != other.byteLength) ||
            !equalFunc(new _Uint8Array(object), new _Uint8Array(other))) {
          return false;
        }
        return true;

      case boolTag:
      case dateTag:
      case numberTag:
        // Coerce booleans to `1` or `0` and dates to milliseconds.
        // Invalid dates are coerced to `NaN`.
        return eq_1(+object, +other);

      case errorTag:
        return object.name == other.name && object.message == other.message;

      case regexpTag:
      case stringTag:
        // Coerce regexes to strings and treat strings, primitives and objects,
        // as equal. See http://www.ecma-international.org/ecma-262/7.0/#sec-regexp.prototype.tostring
        // for more details.
        return object == (other + '');

      case mapTag:
        var convert = _mapToArray;

      case setTag:
        var isPartial = bitmask & COMPARE_PARTIAL_FLAG$1;
        convert || (convert = _setToArray);

        if (object.size != other.size && !isPartial) {
          return false;
        }
        // Assume cyclic values are equal.
        var stacked = stack.get(object);
        if (stacked) {
          return stacked == other;
        }
        bitmask |= COMPARE_UNORDERED_FLAG$1;

        // Recursively compare objects (susceptible to call stack limits).
        stack.set(object, other);
        var result = _equalArrays(convert(object), convert(other), bitmask, customizer, equalFunc, stack);
        stack['delete'](object);
        return result;

      case symbolTag:
        if (symbolValueOf) {
          return symbolValueOf.call(object) == symbolValueOf.call(other);
        }
    }
    return false;
  }

  var _equalByTag = equalByTag;

  /**
   * Appends the elements of `values` to `array`.
   *
   * @private
   * @param {Array} array The array to modify.
   * @param {Array} values The values to append.
   * @returns {Array} Returns `array`.
   */
  function arrayPush(array, values) {
    var index = -1,
        length = values.length,
        offset = array.length;

    while (++index < length) {
      array[offset + index] = values[index];
    }
    return array;
  }

  var _arrayPush = arrayPush;

  /**
   * Checks if `value` is classified as an `Array` object.
   *
   * @static
   * @memberOf _
   * @since 0.1.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is an array, else `false`.
   * @example
   *
   * _.isArray([1, 2, 3]);
   * // => true
   *
   * _.isArray(document.body.children);
   * // => false
   *
   * _.isArray('abc');
   * // => false
   *
   * _.isArray(_.noop);
   * // => false
   */
  var isArray = Array.isArray;

  var isArray_1 = isArray;

  /**
   * The base implementation of `getAllKeys` and `getAllKeysIn` which uses
   * `keysFunc` and `symbolsFunc` to get the enumerable property names and
   * symbols of `object`.
   *
   * @private
   * @param {Object} object The object to query.
   * @param {Function} keysFunc The function to get the keys of `object`.
   * @param {Function} symbolsFunc The function to get the symbols of `object`.
   * @returns {Array} Returns the array of property names and symbols.
   */
  function baseGetAllKeys(object, keysFunc, symbolsFunc) {
    var result = keysFunc(object);
    return isArray_1(object) ? result : _arrayPush(result, symbolsFunc(object));
  }

  var _baseGetAllKeys = baseGetAllKeys;

  /**
   * A specialized version of `_.filter` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} predicate The function invoked per iteration.
   * @returns {Array} Returns the new filtered array.
   */
  function arrayFilter(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length,
        resIndex = 0,
        result = [];

    while (++index < length) {
      var value = array[index];
      if (predicate(value, index, array)) {
        result[resIndex++] = value;
      }
    }
    return result;
  }

  var _arrayFilter = arrayFilter;

  /**
   * This method returns a new empty array.
   *
   * @static
   * @memberOf _
   * @since 4.13.0
   * @category Util
   * @returns {Array} Returns the new empty array.
   * @example
   *
   * var arrays = _.times(2, _.stubArray);
   *
   * console.log(arrays);
   * // => [[], []]
   *
   * console.log(arrays[0] === arrays[1]);
   * // => false
   */
  function stubArray() {
    return [];
  }

  var stubArray_1 = stubArray;

  /** Used for built-in method references. */
  var objectProto$5 = Object.prototype;

  /** Built-in value references. */
  var propertyIsEnumerable = objectProto$5.propertyIsEnumerable;

  /* Built-in method references for those with the same name as other `lodash` methods. */
  var nativeGetSymbols = Object.getOwnPropertySymbols;

  /**
   * Creates an array of the own enumerable symbols of `object`.
   *
   * @private
   * @param {Object} object The object to query.
   * @returns {Array} Returns the array of symbols.
   */
  var getSymbols = !nativeGetSymbols ? stubArray_1 : function(object) {
    if (object == null) {
      return [];
    }
    object = Object(object);
    return _arrayFilter(nativeGetSymbols(object), function(symbol) {
      return propertyIsEnumerable.call(object, symbol);
    });
  };

  var _getSymbols = getSymbols;

  /**
   * The base implementation of `_.times` without support for iteratee shorthands
   * or max array length checks.
   *
   * @private
   * @param {number} n The number of times to invoke `iteratee`.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {Array} Returns the array of results.
   */
  function baseTimes(n, iteratee) {
    var index = -1,
        result = Array(n);

    while (++index < n) {
      result[index] = iteratee(index);
    }
    return result;
  }

  var _baseTimes = baseTimes;

  /**
   * Checks if `value` is object-like. A value is object-like if it's not `null`
   * and has a `typeof` result of "object".
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
   * @example
   *
   * _.isObjectLike({});
   * // => true
   *
   * _.isObjectLike([1, 2, 3]);
   * // => true
   *
   * _.isObjectLike(_.noop);
   * // => false
   *
   * _.isObjectLike(null);
   * // => false
   */
  function isObjectLike(value) {
    return value != null && typeof value == 'object';
  }

  var isObjectLike_1 = isObjectLike;

  /** `Object#toString` result references. */
  var argsTag = '[object Arguments]';

  /**
   * The base implementation of `_.isArguments`.
   *
   * @private
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is an `arguments` object,
   */
  function baseIsArguments(value) {
    return isObjectLike_1(value) && _baseGetTag(value) == argsTag;
  }

  var _baseIsArguments = baseIsArguments;

  /** Used for built-in method references. */
  var objectProto$6 = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$4 = objectProto$6.hasOwnProperty;

  /** Built-in value references. */
  var propertyIsEnumerable$1 = objectProto$6.propertyIsEnumerable;

  /**
   * Checks if `value` is likely an `arguments` object.
   *
   * @static
   * @memberOf _
   * @since 0.1.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is an `arguments` object,
   *  else `false`.
   * @example
   *
   * _.isArguments(function() { return arguments; }());
   * // => true
   *
   * _.isArguments([1, 2, 3]);
   * // => false
   */
  var isArguments = _baseIsArguments(function() { return arguments; }()) ? _baseIsArguments : function(value) {
    return isObjectLike_1(value) && hasOwnProperty$4.call(value, 'callee') &&
      !propertyIsEnumerable$1.call(value, 'callee');
  };

  var isArguments_1 = isArguments;

  /**
   * This method returns `false`.
   *
   * @static
   * @memberOf _
   * @since 4.13.0
   * @category Util
   * @returns {boolean} Returns `false`.
   * @example
   *
   * _.times(2, _.stubFalse);
   * // => [false, false]
   */
  function stubFalse() {
    return false;
  }

  var stubFalse_1 = stubFalse;

  var isBuffer_1 = createCommonjsModule(function (module, exports) {
  /** Detect free variable `exports`. */
  var freeExports = exports && !exports.nodeType && exports;

  /** Detect free variable `module`. */
  var freeModule = freeExports && 'object' == 'object' && module && !module.nodeType && module;

  /** Detect the popular CommonJS extension `module.exports`. */
  var moduleExports = freeModule && freeModule.exports === freeExports;

  /** Built-in value references. */
  var Buffer = moduleExports ? _root.Buffer : undefined;

  /* Built-in method references for those with the same name as other `lodash` methods. */
  var nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined;

  /**
   * Checks if `value` is a buffer.
   *
   * @static
   * @memberOf _
   * @since 4.3.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
   * @example
   *
   * _.isBuffer(new Buffer(2));
   * // => true
   *
   * _.isBuffer(new Uint8Array(2));
   * // => false
   */
  var isBuffer = nativeIsBuffer || stubFalse_1;

  module.exports = isBuffer;
  });

  /** Used as references for various `Number` constants. */
  var MAX_SAFE_INTEGER = 9007199254740991;

  /** Used to detect unsigned integer values. */
  var reIsUint = /^(?:0|[1-9]\d*)$/;

  /**
   * Checks if `value` is a valid array-like index.
   *
   * @private
   * @param {*} value The value to check.
   * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
   * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
   */
  function isIndex(value, length) {
    var type = typeof value;
    length = length == null ? MAX_SAFE_INTEGER : length;

    return !!length &&
      (type == 'number' ||
        (type != 'symbol' && reIsUint.test(value))) &&
          (value > -1 && value % 1 == 0 && value < length);
  }

  var _isIndex = isIndex;

  /** Used as references for various `Number` constants. */
  var MAX_SAFE_INTEGER$1 = 9007199254740991;

  /**
   * Checks if `value` is a valid array-like length.
   *
   * **Note:** This method is loosely based on
   * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
   * @example
   *
   * _.isLength(3);
   * // => true
   *
   * _.isLength(Number.MIN_VALUE);
   * // => false
   *
   * _.isLength(Infinity);
   * // => false
   *
   * _.isLength('3');
   * // => false
   */
  function isLength(value) {
    return typeof value == 'number' &&
      value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER$1;
  }

  var isLength_1 = isLength;

  /** `Object#toString` result references. */
  var argsTag$1 = '[object Arguments]',
      arrayTag = '[object Array]',
      boolTag$1 = '[object Boolean]',
      dateTag$1 = '[object Date]',
      errorTag$1 = '[object Error]',
      funcTag$1 = '[object Function]',
      mapTag$1 = '[object Map]',
      numberTag$1 = '[object Number]',
      objectTag = '[object Object]',
      regexpTag$1 = '[object RegExp]',
      setTag$1 = '[object Set]',
      stringTag$1 = '[object String]',
      weakMapTag = '[object WeakMap]';

  var arrayBufferTag$1 = '[object ArrayBuffer]',
      dataViewTag$1 = '[object DataView]',
      float32Tag = '[object Float32Array]',
      float64Tag = '[object Float64Array]',
      int8Tag = '[object Int8Array]',
      int16Tag = '[object Int16Array]',
      int32Tag = '[object Int32Array]',
      uint8Tag = '[object Uint8Array]',
      uint8ClampedTag = '[object Uint8ClampedArray]',
      uint16Tag = '[object Uint16Array]',
      uint32Tag = '[object Uint32Array]';

  /** Used to identify `toStringTag` values of typed arrays. */
  var typedArrayTags = {};
  typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
  typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
  typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
  typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
  typedArrayTags[uint32Tag] = true;
  typedArrayTags[argsTag$1] = typedArrayTags[arrayTag] =
  typedArrayTags[arrayBufferTag$1] = typedArrayTags[boolTag$1] =
  typedArrayTags[dataViewTag$1] = typedArrayTags[dateTag$1] =
  typedArrayTags[errorTag$1] = typedArrayTags[funcTag$1] =
  typedArrayTags[mapTag$1] = typedArrayTags[numberTag$1] =
  typedArrayTags[objectTag] = typedArrayTags[regexpTag$1] =
  typedArrayTags[setTag$1] = typedArrayTags[stringTag$1] =
  typedArrayTags[weakMapTag] = false;

  /**
   * The base implementation of `_.isTypedArray` without Node.js optimizations.
   *
   * @private
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
   */
  function baseIsTypedArray(value) {
    return isObjectLike_1(value) &&
      isLength_1(value.length) && !!typedArrayTags[_baseGetTag(value)];
  }

  var _baseIsTypedArray = baseIsTypedArray;

  /**
   * The base implementation of `_.unary` without support for storing metadata.
   *
   * @private
   * @param {Function} func The function to cap arguments for.
   * @returns {Function} Returns the new capped function.
   */
  function baseUnary(func) {
    return function(value) {
      return func(value);
    };
  }

  var _baseUnary = baseUnary;

  var _nodeUtil = createCommonjsModule(function (module, exports) {
  /** Detect free variable `exports`. */
  var freeExports = exports && !exports.nodeType && exports;

  /** Detect free variable `module`. */
  var freeModule = freeExports && 'object' == 'object' && module && !module.nodeType && module;

  /** Detect the popular CommonJS extension `module.exports`. */
  var moduleExports = freeModule && freeModule.exports === freeExports;

  /** Detect free variable `process` from Node.js. */
  var freeProcess = moduleExports && _freeGlobal.process;

  /** Used to access faster Node.js helpers. */
  var nodeUtil = (function() {
    try {
      // Use `util.types` for Node.js 10+.
      var types = freeModule && freeModule.require && freeModule.require('util').types;

      if (types) {
        return types;
      }

      // Legacy `process.binding('util')` for Node.js < 10.
      return freeProcess && freeProcess.binding && freeProcess.binding('util');
    } catch (e) {}
  }());

  module.exports = nodeUtil;
  });

  /* Node.js helper references. */
  var nodeIsTypedArray = _nodeUtil && _nodeUtil.isTypedArray;

  /**
   * Checks if `value` is classified as a typed array.
   *
   * @static
   * @memberOf _
   * @since 3.0.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
   * @example
   *
   * _.isTypedArray(new Uint8Array);
   * // => true
   *
   * _.isTypedArray([]);
   * // => false
   */
  var isTypedArray = nodeIsTypedArray ? _baseUnary(nodeIsTypedArray) : _baseIsTypedArray;

  var isTypedArray_1 = isTypedArray;

  /** Used for built-in method references. */
  var objectProto$7 = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$5 = objectProto$7.hasOwnProperty;

  /**
   * Creates an array of the enumerable property names of the array-like `value`.
   *
   * @private
   * @param {*} value The value to query.
   * @param {boolean} inherited Specify returning inherited property names.
   * @returns {Array} Returns the array of property names.
   */
  function arrayLikeKeys(value, inherited) {
    var isArr = isArray_1(value),
        isArg = !isArr && isArguments_1(value),
        isBuff = !isArr && !isArg && isBuffer_1(value),
        isType = !isArr && !isArg && !isBuff && isTypedArray_1(value),
        skipIndexes = isArr || isArg || isBuff || isType,
        result = skipIndexes ? _baseTimes(value.length, String) : [],
        length = result.length;

    for (var key in value) {
      if ((inherited || hasOwnProperty$5.call(value, key)) &&
          !(skipIndexes && (
             // Safari 9 has enumerable `arguments.length` in strict mode.
             key == 'length' ||
             // Node.js 0.10 has enumerable non-index properties on buffers.
             (isBuff && (key == 'offset' || key == 'parent')) ||
             // PhantomJS 2 has enumerable non-index properties on typed arrays.
             (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
             // Skip index properties.
             _isIndex(key, length)
          ))) {
        result.push(key);
      }
    }
    return result;
  }

  var _arrayLikeKeys = arrayLikeKeys;

  /** Used for built-in method references. */
  var objectProto$8 = Object.prototype;

  /**
   * Checks if `value` is likely a prototype object.
   *
   * @private
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
   */
  function isPrototype(value) {
    var Ctor = value && value.constructor,
        proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$8;

    return value === proto;
  }

  var _isPrototype = isPrototype;

  /**
   * Creates a unary function that invokes `func` with its argument transformed.
   *
   * @private
   * @param {Function} func The function to wrap.
   * @param {Function} transform The argument transform.
   * @returns {Function} Returns the new function.
   */
  function overArg(func, transform) {
    return function(arg) {
      return func(transform(arg));
    };
  }

  var _overArg = overArg;

  /* Built-in method references for those with the same name as other `lodash` methods. */
  var nativeKeys = _overArg(Object.keys, Object);

  var _nativeKeys = nativeKeys;

  /** Used for built-in method references. */
  var objectProto$9 = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$6 = objectProto$9.hasOwnProperty;

  /**
   * The base implementation of `_.keys` which doesn't treat sparse arrays as dense.
   *
   * @private
   * @param {Object} object The object to query.
   * @returns {Array} Returns the array of property names.
   */
  function baseKeys(object) {
    if (!_isPrototype(object)) {
      return _nativeKeys(object);
    }
    var result = [];
    for (var key in Object(object)) {
      if (hasOwnProperty$6.call(object, key) && key != 'constructor') {
        result.push(key);
      }
    }
    return result;
  }

  var _baseKeys = baseKeys;

  /**
   * Checks if `value` is array-like. A value is considered array-like if it's
   * not a function and has a `value.length` that's an integer greater than or
   * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
   * @example
   *
   * _.isArrayLike([1, 2, 3]);
   * // => true
   *
   * _.isArrayLike(document.body.children);
   * // => true
   *
   * _.isArrayLike('abc');
   * // => true
   *
   * _.isArrayLike(_.noop);
   * // => false
   */
  function isArrayLike(value) {
    return value != null && isLength_1(value.length) && !isFunction_1(value);
  }

  var isArrayLike_1 = isArrayLike;

  /**
   * Creates an array of the own enumerable property names of `object`.
   *
   * **Note:** Non-object values are coerced to objects. See the
   * [ES spec](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
   * for more details.
   *
   * @static
   * @since 0.1.0
   * @memberOf _
   * @category Object
   * @param {Object} object The object to query.
   * @returns {Array} Returns the array of property names.
   * @example
   *
   * function Foo() {
   *   this.a = 1;
   *   this.b = 2;
   * }
   *
   * Foo.prototype.c = 3;
   *
   * _.keys(new Foo);
   * // => ['a', 'b'] (iteration order is not guaranteed)
   *
   * _.keys('hi');
   * // => ['0', '1']
   */
  function keys(object) {
    return isArrayLike_1(object) ? _arrayLikeKeys(object) : _baseKeys(object);
  }

  var keys_1 = keys;

  /**
   * Creates an array of own enumerable property names and symbols of `object`.
   *
   * @private
   * @param {Object} object The object to query.
   * @returns {Array} Returns the array of property names and symbols.
   */
  function getAllKeys(object) {
    return _baseGetAllKeys(object, keys_1, _getSymbols);
  }

  var _getAllKeys = getAllKeys;

  /** Used to compose bitmasks for value comparisons. */
  var COMPARE_PARTIAL_FLAG$2 = 1;

  /** Used for built-in method references. */
  var objectProto$a = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$7 = objectProto$a.hasOwnProperty;

  /**
   * A specialized version of `baseIsEqualDeep` for objects with support for
   * partial deep comparisons.
   *
   * @private
   * @param {Object} object The object to compare.
   * @param {Object} other The other object to compare.
   * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
   * @param {Function} customizer The function to customize comparisons.
   * @param {Function} equalFunc The function to determine equivalents of values.
   * @param {Object} stack Tracks traversed `object` and `other` objects.
   * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
   */
  function equalObjects(object, other, bitmask, customizer, equalFunc, stack) {
    var isPartial = bitmask & COMPARE_PARTIAL_FLAG$2,
        objProps = _getAllKeys(object),
        objLength = objProps.length,
        othProps = _getAllKeys(other),
        othLength = othProps.length;

    if (objLength != othLength && !isPartial) {
      return false;
    }
    var index = objLength;
    while (index--) {
      var key = objProps[index];
      if (!(isPartial ? key in other : hasOwnProperty$7.call(other, key))) {
        return false;
      }
    }
    // Assume cyclic values are equal.
    var stacked = stack.get(object);
    if (stacked && stack.get(other)) {
      return stacked == other;
    }
    var result = true;
    stack.set(object, other);
    stack.set(other, object);

    var skipCtor = isPartial;
    while (++index < objLength) {
      key = objProps[index];
      var objValue = object[key],
          othValue = other[key];

      if (customizer) {
        var compared = isPartial
          ? customizer(othValue, objValue, key, other, object, stack)
          : customizer(objValue, othValue, key, object, other, stack);
      }
      // Recursively compare objects (susceptible to call stack limits).
      if (!(compared === undefined
            ? (objValue === othValue || equalFunc(objValue, othValue, bitmask, customizer, stack))
            : compared
          )) {
        result = false;
        break;
      }
      skipCtor || (skipCtor = key == 'constructor');
    }
    if (result && !skipCtor) {
      var objCtor = object.constructor,
          othCtor = other.constructor;

      // Non `Object` object instances with different constructors are not equal.
      if (objCtor != othCtor &&
          ('constructor' in object && 'constructor' in other) &&
          !(typeof objCtor == 'function' && objCtor instanceof objCtor &&
            typeof othCtor == 'function' && othCtor instanceof othCtor)) {
        result = false;
      }
    }
    stack['delete'](object);
    stack['delete'](other);
    return result;
  }

  var _equalObjects = equalObjects;

  /* Built-in method references that are verified to be native. */
  var DataView$1 = _getNative(_root, 'DataView');

  var _DataView = DataView$1;

  /* Built-in method references that are verified to be native. */
  var Promise$1 = _getNative(_root, 'Promise');

  var _Promise = Promise$1;

  /* Built-in method references that are verified to be native. */
  var Set$1 = _getNative(_root, 'Set');

  var _Set = Set$1;

  /* Built-in method references that are verified to be native. */
  var WeakMap$1 = _getNative(_root, 'WeakMap');

  var _WeakMap = WeakMap$1;

  /** `Object#toString` result references. */
  var mapTag$2 = '[object Map]',
      objectTag$1 = '[object Object]',
      promiseTag = '[object Promise]',
      setTag$2 = '[object Set]',
      weakMapTag$1 = '[object WeakMap]';

  var dataViewTag$2 = '[object DataView]';

  /** Used to detect maps, sets, and weakmaps. */
  var dataViewCtorString = _toSource(_DataView),
      mapCtorString = _toSource(_Map),
      promiseCtorString = _toSource(_Promise),
      setCtorString = _toSource(_Set),
      weakMapCtorString = _toSource(_WeakMap);

  /**
   * Gets the `toStringTag` of `value`.
   *
   * @private
   * @param {*} value The value to query.
   * @returns {string} Returns the `toStringTag`.
   */
  var getTag = _baseGetTag;

  // Fallback for data views, maps, sets, and weak maps in IE 11 and promises in Node.js < 6.
  if ((_DataView && getTag(new _DataView(new ArrayBuffer(1))) != dataViewTag$2) ||
      (_Map && getTag(new _Map) != mapTag$2) ||
      (_Promise && getTag(_Promise.resolve()) != promiseTag) ||
      (_Set && getTag(new _Set) != setTag$2) ||
      (_WeakMap && getTag(new _WeakMap) != weakMapTag$1)) {
    getTag = function(value) {
      var result = _baseGetTag(value),
          Ctor = result == objectTag$1 ? value.constructor : undefined,
          ctorString = Ctor ? _toSource(Ctor) : '';

      if (ctorString) {
        switch (ctorString) {
          case dataViewCtorString: return dataViewTag$2;
          case mapCtorString: return mapTag$2;
          case promiseCtorString: return promiseTag;
          case setCtorString: return setTag$2;
          case weakMapCtorString: return weakMapTag$1;
        }
      }
      return result;
    };
  }

  var _getTag = getTag;

  /** Used to compose bitmasks for value comparisons. */
  var COMPARE_PARTIAL_FLAG$3 = 1;

  /** `Object#toString` result references. */
  var argsTag$2 = '[object Arguments]',
      arrayTag$1 = '[object Array]',
      objectTag$2 = '[object Object]';

  /** Used for built-in method references. */
  var objectProto$b = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty$8 = objectProto$b.hasOwnProperty;

  /**
   * A specialized version of `baseIsEqual` for arrays and objects which performs
   * deep comparisons and tracks traversed objects enabling objects with circular
   * references to be compared.
   *
   * @private
   * @param {Object} object The object to compare.
   * @param {Object} other The other object to compare.
   * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
   * @param {Function} customizer The function to customize comparisons.
   * @param {Function} equalFunc The function to determine equivalents of values.
   * @param {Object} [stack] Tracks traversed `object` and `other` objects.
   * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
   */
  function baseIsEqualDeep(object, other, bitmask, customizer, equalFunc, stack) {
    var objIsArr = isArray_1(object),
        othIsArr = isArray_1(other),
        objTag = objIsArr ? arrayTag$1 : _getTag(object),
        othTag = othIsArr ? arrayTag$1 : _getTag(other);

    objTag = objTag == argsTag$2 ? objectTag$2 : objTag;
    othTag = othTag == argsTag$2 ? objectTag$2 : othTag;

    var objIsObj = objTag == objectTag$2,
        othIsObj = othTag == objectTag$2,
        isSameTag = objTag == othTag;

    if (isSameTag && isBuffer_1(object)) {
      if (!isBuffer_1(other)) {
        return false;
      }
      objIsArr = true;
      objIsObj = false;
    }
    if (isSameTag && !objIsObj) {
      stack || (stack = new _Stack);
      return (objIsArr || isTypedArray_1(object))
        ? _equalArrays(object, other, bitmask, customizer, equalFunc, stack)
        : _equalByTag(object, other, objTag, bitmask, customizer, equalFunc, stack);
    }
    if (!(bitmask & COMPARE_PARTIAL_FLAG$3)) {
      var objIsWrapped = objIsObj && hasOwnProperty$8.call(object, '__wrapped__'),
          othIsWrapped = othIsObj && hasOwnProperty$8.call(other, '__wrapped__');

      if (objIsWrapped || othIsWrapped) {
        var objUnwrapped = objIsWrapped ? object.value() : object,
            othUnwrapped = othIsWrapped ? other.value() : other;

        stack || (stack = new _Stack);
        return equalFunc(objUnwrapped, othUnwrapped, bitmask, customizer, stack);
      }
    }
    if (!isSameTag) {
      return false;
    }
    stack || (stack = new _Stack);
    return _equalObjects(object, other, bitmask, customizer, equalFunc, stack);
  }

  var _baseIsEqualDeep = baseIsEqualDeep;

  /**
   * The base implementation of `_.isEqual` which supports partial comparisons
   * and tracks traversed objects.
   *
   * @private
   * @param {*} value The value to compare.
   * @param {*} other The other value to compare.
   * @param {boolean} bitmask The bitmask flags.
   *  1 - Unordered comparison
   *  2 - Partial comparison
   * @param {Function} [customizer] The function to customize comparisons.
   * @param {Object} [stack] Tracks traversed `value` and `other` objects.
   * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
   */
  function baseIsEqual(value, other, bitmask, customizer, stack) {
    if (value === other) {
      return true;
    }
    if (value == null || other == null || (!isObjectLike_1(value) && !isObjectLike_1(other))) {
      return value !== value && other !== other;
    }
    return _baseIsEqualDeep(value, other, bitmask, customizer, baseIsEqual, stack);
  }

  var _baseIsEqual = baseIsEqual;

  /**
   * Performs a deep comparison between two values to determine if they are
   * equivalent.
   *
   * **Note:** This method supports comparing arrays, array buffers, booleans,
   * date objects, error objects, maps, numbers, `Object` objects, regexes,
   * sets, strings, symbols, and typed arrays. `Object` objects are compared
   * by their own, not inherited, enumerable properties. Functions and DOM
   * nodes are compared by strict equality, i.e. `===`.
   *
   * @static
   * @memberOf _
   * @since 0.1.0
   * @category Lang
   * @param {*} value The value to compare.
   * @param {*} other The other value to compare.
   * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
   * @example
   *
   * var object = { 'a': 1 };
   * var other = { 'a': 1 };
   *
   * _.isEqual(object, other);
   * // => true
   *
   * object === other;
   * // => false
   */
  function isEqual(value, other) {
    return _baseIsEqual(value, other);
  }

  var isEqual_1 = isEqual;

  /** Built-in value references. */
  var getPrototype = _overArg(Object.getPrototypeOf, Object);

  var _getPrototype = getPrototype;

  /** `Object#toString` result references. */
  var objectTag$3 = '[object Object]';

  /** Used for built-in method references. */
  var funcProto$2 = Function.prototype,
      objectProto$c = Object.prototype;

  /** Used to resolve the decompiled source of functions. */
  var funcToString$2 = funcProto$2.toString;

  /** Used to check objects for own properties. */
  var hasOwnProperty$9 = objectProto$c.hasOwnProperty;

  /** Used to infer the `Object` constructor. */
  var objectCtorString = funcToString$2.call(Object);

  /**
   * Checks if `value` is a plain object, that is, an object created by the
   * `Object` constructor or one with a `[[Prototype]]` of `null`.
   *
   * @static
   * @memberOf _
   * @since 0.8.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a plain object, else `false`.
   * @example
   *
   * function Foo() {
   *   this.a = 1;
   * }
   *
   * _.isPlainObject(new Foo);
   * // => false
   *
   * _.isPlainObject([1, 2, 3]);
   * // => false
   *
   * _.isPlainObject({ 'x': 0, 'y': 0 });
   * // => true
   *
   * _.isPlainObject(Object.create(null));
   * // => true
   */
  function isPlainObject(value) {
    if (!isObjectLike_1(value) || _baseGetTag(value) != objectTag$3) {
      return false;
    }
    var proto = _getPrototype(value);
    if (proto === null) {
      return true;
    }
    var Ctor = hasOwnProperty$9.call(proto, 'constructor') && proto.constructor;
    return typeof Ctor == 'function' && Ctor instanceof Ctor &&
      funcToString$2.call(Ctor) == objectCtorString;
  }

  var isPlainObject_1 = isPlainObject;

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign = (undefined && undefined.__assign) || function () {
      __assign = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign.apply(this, arguments);
  };
  /**
   * Find all strings in the first argument that are not in the second.
   */
  function difference(a, b) {
      return a.filter(function (v) { return b.indexOf(v) === -1; });
  }
  /**
   * Compare two objects deeply to see if they are equal.
   */
  function isEqual$1(a, b) {
      return isEqual_1(a, b);
  }
  /**
   * A polyfill for Object.assign
   *
   * This is from code that Typescript 2.4 generates for a polyfill.
   */
  var assign = Object.assign || function (t) {
      for (var i = 1; i < arguments.length; i++) {
          var s = arguments[i];
          for (var p in s) {
              if (Object.prototype.hasOwnProperty.call(s, p)) {
                  t[p] = s[p];
              }
          }
      }
      return t;
  };
  /**
   * Generate a UUID
   *
   * http://www.ietf.org/rfc/rfc4122.txt
   */
  function uuid() {
      return UUID.uuid4();
  }
  /**
   * Wrappable Error class
   *
   * The Error class doesn't actually act on `this`.  Instead it always
   * returns a new instance of Error.  Here we capture that instance so we
   * can apply it's properties to `this`.
   */
  var WrappedError = /** @class */ (function (_super) {
      __extends(WrappedError, _super);
      function WrappedError(message, error) {
          var _this = _super.call(this, message) || this;
          console.warn('WrappedError has been deprecated!');
          // Keep a stack of the original error messages.
          if (error instanceof WrappedError) {
              _this.error_stack = error.error_stack;
          }
          else {
              _this.error_stack = [error];
          }
          _this.error_stack.push(_this);
          return _this;
      }
      return WrappedError;
  }(Error));
  /**
   * Resolve a promiseful dictionary.
   * Returns a single Promise.
   */
  function resolvePromisesDict(d) {
      var keys = Object.keys(d);
      var values = [];
      keys.forEach(function (key) {
          values.push(d[key]);
      });
      return Promise.all(values).then(function (v) {
          var d = {};
          for (var i = 0; i < keys.length; i++) {
              d[keys[i]] = v[i];
          }
          return d;
      });
  }
  /**
   * Creates a wrappable Promise rejection function.
   *
   * Creates a function that logs an error message before rethrowing
   * the original error that caused the promise to reject.
   */
  function reject(message, log) {
      return function promiseRejection(error) {
          if (log) {
              console.error(new Error(message));
          }
          throw error;
      };
  }
  /**
   * Takes an object 'state' and fills in buffer[i] at 'path' buffer_paths[i]
   * where buffer_paths[i] is a list indicating where in the object buffer[i] should
   * be placed
   * Example: state = {a: 1, b: {}, c: [0, null]}
   * buffers = [array1, array2]
   * buffer_paths = [['b', 'data'], ['c', 1]]
   * Will lead to {a: 1, b: {data: array1}, c: [0, array2]}
   */
  function put_buffers(state, buffer_paths, buffers) {
      for (var i = 0; i < buffer_paths.length; i++) {
          var buffer_path = buffer_paths[i];
          // say we want to set state[x][y][z] = buffers[i]
          var obj = state;
          // we first get obj = state[x][y]
          for (var j = 0; j < buffer_path.length - 1; j++) {
              obj = obj[buffer_path[j]];
          }
          // and then set: obj[z] = buffers[i]
          obj[buffer_path[buffer_path.length - 1]] = buffers[i];
      }
  }
  /**
   * The inverse of put_buffers, return an objects with the new state where all buffers(ArrayBuffer)
   * are removed. If a buffer is a member of an object, that object is cloned, and the key removed. If a buffer
   * is an element of an array, that array is cloned, and the element is set to null.
   * See put_buffers for the meaning of buffer_paths
   * Returns an object with the new state (.state) an array with paths to the buffers (.buffer_paths),
   * and the buffers associated to those paths (.buffers).
   */
  function remove_buffers(state) {
      var buffers = [];
      var buffer_paths = [];
      // if we need to remove an object from a list, we need to clone that list, otherwise we may modify
      // the internal state of the widget model
      // however, we do not want to clone everything, for performance
      function remove(obj, path) {
          if (obj.toJSON) {
              // We need to get the JSON form of the object before recursing.
              // See https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#toJSON()_behavior
              obj = obj.toJSON();
          }
          if (Array.isArray(obj)) {
              var is_cloned = false;
              for (var i = 0; i < obj.length; i++) {
                  var value = obj[i];
                  if (value) {
                      if (value instanceof ArrayBuffer || ArrayBuffer.isView(value)) {
                          if (!is_cloned) {
                              obj = obj.slice();
                              is_cloned = true;
                          }
                          buffers.push(ArrayBuffer.isView(value) ? value.buffer : value);
                          buffer_paths.push(path.concat([i]));
                          // easier to just keep the array, but clear the entry, otherwise we have to think
                          // about array length, much easier this way
                          obj[i] = null;
                      }
                      else {
                          var new_value = remove(value, path.concat([i]));
                          // only assigned when the value changes, we may serialize objects that don't support assignment
                          if (new_value !== value) {
                              if (!is_cloned) {
                                  obj = obj.slice();
                                  is_cloned = true;
                              }
                              obj[i] = new_value;
                          }
                      }
                  }
              }
          }
          else if (isPlainObject_1(obj)) {
              for (var key in obj) {
                  var is_cloned = false;
                  if (obj.hasOwnProperty(key)) {
                      var value = obj[key];
                      if (value) {
                          if (value instanceof ArrayBuffer || ArrayBuffer.isView(value)) {
                              if (!is_cloned) {
                                  obj = __assign({}, obj);
                                  is_cloned = true;
                              }
                              buffers.push(ArrayBuffer.isView(value) ? value.buffer : value);
                              buffer_paths.push(path.concat([key]));
                              delete obj[key]; // for objects/dicts we just delete them
                          }
                          else {
                              var new_value = remove(value, path.concat([key]));
                              // only assigned when the value changes, we may serialize objects that don't support assignment
                              if (new_value !== value) {
                                  if (!is_cloned) {
                                      obj = __assign({}, obj);
                                      is_cloned = true;
                                  }
                                  obj[key] = new_value;
                              }
                          }
                      }
                  }
              }
          }
          return obj;
      }
      var new_state = remove(state, []);
      return { state: new_state, buffers: buffers, buffer_paths: buffer_paths };
  }
  var hexTable = [
      '00', '01', '02', '03', '04', '05', '06', '07', '08', '09', '0A', '0B', '0C', '0D', '0E', '0F',
      '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '1A', '1B', '1C', '1D', '1E', '1F',
      '20', '21', '22', '23', '24', '25', '26', '27', '28', '29', '2A', '2B', '2C', '2D', '2E', '2F',
      '30', '31', '32', '33', '34', '35', '36', '37', '38', '39', '3A', '3B', '3C', '3D', '3E', '3F',
      '40', '41', '42', '43', '44', '45', '46', '47', '48', '49', '4A', '4B', '4C', '4D', '4E', '4F',
      '50', '51', '52', '53', '54', '55', '56', '57', '58', '59', '5A', '5B', '5C', '5D', '5E', '5F',
      '60', '61', '62', '63', '64', '65', '66', '67', '68', '69', '6A', '6B', '6C', '6D', '6E', '6F',
      '70', '71', '72', '73', '74', '75', '76', '77', '78', '79', '7A', '7B', '7C', '7D', '7E', '7F',
      '80', '81', '82', '83', '84', '85', '86', '87', '88', '89', '8A', '8B', '8C', '8D', '8E', '8F',
      '90', '91', '92', '93', '94', '95', '96', '97', '98', '99', '9A', '9B', '9C', '9D', '9E', '9F',
      'A0', 'A1', 'A2', 'A3', 'A4', 'A5', 'A6', 'A7', 'A8', 'A9', 'AA', 'AB', 'AC', 'AD', 'AE', 'AF',
      'B0', 'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8', 'B9', 'BA', 'BB', 'BC', 'BD', 'BE', 'BF',
      'C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9', 'CA', 'CB', 'CC', 'CD', 'CE', 'CF',
      'D0', 'D1', 'D2', 'D3', 'D4', 'D5', 'D6', 'D7', 'D8', 'D9', 'DA', 'DB', 'DC', 'DD', 'DE', 'DF',
      'E0', 'E1', 'E2', 'E3', 'E4', 'E5', 'E6', 'E7', 'E8', 'E9', 'EA', 'EB', 'EC', 'ED', 'EE', 'EF',
      'F0', 'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'FA', 'FB', 'FC', 'FD', 'FE', 'FF'
  ];
  /**
   * Convert an ArrayBuffer to a hex string.
   */
  function bufferToHex(buffer) {
      var x = new Uint8Array(buffer);
      var s = [];
      for (var i = 0; i < x.length; i++) {
          s.push(hexTable[x[i]]);
      }
      return s.join('');
  }
  /**
   * Convert a hex string to an ArrayBuffer.
   */
  function hexToBuffer(hex) {
      var x = new Uint8Array(hex.length / 2);
      for (var i = 0; i < hex.length; i += 2) {
          x[i / 2] = parseInt(hex.slice(i, i + 2), 16);
      }
      return x.buffer;
  }
  /**
   * Convert an ArrayBuffer to a base64 string.
   */
  function bufferToBase64(buffer) {
      return fromByteArray_1(new Uint8Array(buffer));
  }
  /**
   * Convert a base64 string to an ArrayBuffer.
   */
  function base64ToBuffer(base64) {
      return toByteArray_1(base64).buffer;
  }

  var utils = /*#__PURE__*/Object.freeze({
    difference: difference,
    isEqual: isEqual$1,
    assign: assign,
    uuid: uuid,
    WrappedError: WrappedError,
    resolvePromisesDict: resolvePromisesDict,
    reject: reject,
    put_buffers: put_buffers,
    remove_buffers: remove_buffers,
    bufferToHex: bufferToHex,
    hexToBuffer: hexToBuffer,
    bufferToBase64: bufferToBase64,
    base64ToBuffer: base64ToBuffer
  });

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var JUPYTER_WIDGETS_VERSION = '1.2.0';
  var PROTOCOL_VERSION = '2.0.0';

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __assign$1 = (undefined && undefined.__assign) || function () {
      __assign$1 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$1.apply(this, arguments);
  };
  var __awaiter = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
      function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
      return new (P || (P = Promise))(function (resolve, reject) {
          function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
          function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
          function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
          step((generator = generator.apply(thisArg, _arguments || [])).next());
      });
  };
  var __generator = (undefined && undefined.__generator) || function (thisArg, body) {
      var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
      return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
      function verb(n) { return function (v) { return step([n, v]); }; }
      function step(op) {
          if (f) throw new TypeError("Generator is already executing.");
          while (_) try {
              if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
              if (y = 0, t) op = [op[0] & 2, t.value];
              switch (op[0]) {
                  case 0: case 1: t = op; break;
                  case 4: _.label++; return { value: op[1], done: false };
                  case 5: _.label++; y = op[1]; op = [0]; continue;
                  case 7: op = _.ops.pop(); _.trys.pop(); continue;
                  default:
                      if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                      if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                      if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                      if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                      if (t[2]) _.ops.pop();
                      _.trys.pop(); continue;
              }
              op = body.call(thisArg, _);
          } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
          if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
      }
  };
  var PROTOCOL_MAJOR_VERSION = PROTOCOL_VERSION.split('.', 1)[0];
  /**
   * Manager abstract base class
   */
  var ManagerBase = /** @class */ (function () {
      function ManagerBase() {
          /**
           * The comm target name to register
           */
          this.comm_target_name = 'jupyter.widget';
          /**
           * Dictionary of model ids and model instance promises
           */
          this._models = Object.create(null);
      }
      /**
       * Display a DOMWidgetView for a particular model.
       */
      ManagerBase.prototype.display_model = function (msg, model, options) {
          var _this = this;
          if (options === void 0) { options = {}; }
          return this.create_view(model, options).then(function (view) { return _this.display_view(msg, view, options); }).catch(reject('Could not create view', true));
      };
      /**
       * Modifies view options. Generally overloaded in custom widget manager
       * implementations.
       */
      ManagerBase.prototype.setViewOptions = function (options) {
          if (options === void 0) { options = {}; }
          return options;
      };
      ManagerBase.prototype.create_view = function (model, options) {
          var _this = this;
          if (options === void 0) { options = {}; }
          var viewPromise = model.state_change = model.state_change.then(function () {
              return _this.loadClass(model.get('_view_name'), model.get('_view_module'), model.get('_view_module_version')).then(function (ViewType) {
                  var view = new ViewType({
                      model: model,
                      options: _this.setViewOptions(options)
                  });
                  view.listenTo(model, 'destroy', view.remove);
                  return Promise.resolve(view.render()).then(function () { return view; });
              }).catch(reject('Could not create a view for model id ' + model.model_id, true));
          });
          var id = uuid();
          model.views[id] = viewPromise;
          viewPromise.then(function (view) {
              view.once('remove', function () { delete view.model.views[id]; }, _this);
          });
          return model.state_change;
      };
      /**
       * callback handlers specific to a view
       */
      ManagerBase.prototype.callbacks = function (view) {
          return {};
      };
      /**
       * Get a promise for a model by model id.
       *
       * #### Notes
       * If a model is not found, undefined is returned (NOT a promise). However,
       * the calling code should also deal with the case where a rejected promise
       * is returned, and should treat that also as a model not found.
       */
      ManagerBase.prototype.get_model = function (model_id) {
          // TODO: Perhaps we should return a Promise.reject if the model is not
          // found. Right now this isn't a true async function because it doesn't
          // always return a promise.
          return this._models[model_id];
      };
      /**
       * Handle when a comm is opened.
       */
      ManagerBase.prototype.handle_comm_open = function (comm, msg) {
          var protocolVersion = (msg.metadata || {}).version || '';
          if (protocolVersion.split('.', 1)[0] !== PROTOCOL_MAJOR_VERSION) {
              var error = "Wrong widget protocol version: received protocol version '" + protocolVersion + "', but was expecting major version '" + PROTOCOL_MAJOR_VERSION + "'";
              console.error(error);
              return Promise.reject(error);
          }
          var data = msg.content.data;
          var buffer_paths = data.buffer_paths || [];
          // Make sure the buffers are DataViews
          var buffers = (msg.buffers || []).map(function (b) {
              if (b instanceof DataView) {
                  return b;
              }
              else {
                  return new DataView(b instanceof ArrayBuffer ? b : b.buffer);
              }
          });
          put_buffers(data.state, buffer_paths, buffers);
          return this.new_model({
              model_name: data.state['_model_name'],
              model_module: data.state['_model_module'],
              model_module_version: data.state['_model_module_version'],
              comm: comm
          }, data.state).catch(reject('Could not create a model.', true));
      };
      /**
       * Create a comm and new widget model.
       * @param  options - same options as new_model but comm is not
       *                          required and additional options are available.
       * @param  serialized_state - serialized model attributes.
       */
      ManagerBase.prototype.new_widget = function (options, serialized_state) {
          var _this = this;
          if (serialized_state === void 0) { serialized_state = {}; }
          var commPromise;
          // we check to make sure the view information is provided, to help catch
          // backwards incompatibility errors.
          if (options.view_name === undefined
              || options.view_module === undefined
              || options.view_module_version === undefined) {
              return Promise.reject('new_widget(...) must be given view information in the options.');
          }
          // If no comm is provided, a new comm is opened for the jupyter.widget
          // target.
          if (options.comm) {
              commPromise = Promise.resolve(options.comm);
          }
          else {
              commPromise = this._create_comm(this.comm_target_name, options.model_id, {
                  state: {
                      _model_module: options.model_module,
                      _model_module_version: options.model_module_version,
                      _model_name: options.model_name,
                      _view_module: options.view_module,
                      _view_module_version: options.view_module_version,
                      _view_name: options.view_name
                  },
              }, { version: PROTOCOL_VERSION });
          }
          // The options dictionary is copied since data will be added to it.
          var options_clone = __assign$1({}, options);
          // Create the model. In the case where the comm promise is rejected a
          // comm-less model is still created with the required model id.
          return commPromise.then(function (comm) {
              // Comm Promise Resolved.
              options_clone.comm = comm;
              var widget_model = _this.new_model(options_clone, serialized_state);
              return widget_model.then(function (model) {
                  model.sync('create', model);
                  return model;
              });
          }, function () {
              // Comm Promise Rejected.
              if (!options_clone.model_id) {
                  options_clone.model_id = uuid();
              }
              return _this.new_model(options_clone, serialized_state);
          });
      };
      ManagerBase.prototype.register_model = function (model_id, modelPromise) {
          var _this = this;
          this._models[model_id] = modelPromise;
          modelPromise.then(function (model) {
              model.once('comm:close', function () {
                  delete _this._models[model_id];
              });
          });
      };
      /**
       * Create and return a promise for a new widget model
       *
       * @param options - the options for creating the model.
       * @param serialized_state - attribute values for the model.
       *
       * @example
       * widget_manager.new_model({
       *      model_name: 'IntSlider',
       *      model_module: '@jupyter-widgets/controls',
       *      model_module_version: '1.0.0',
       *      model_id: 'u-u-i-d'
       * }).then((model) => { console.log('Create success!', model); },
       *  (err) => {console.error(err)});
       *
       */
      ManagerBase.prototype.new_model = function (options, serialized_state) {
          if (serialized_state === void 0) { serialized_state = {}; }
          return __awaiter(this, void 0, void 0, function () {
              var model_id, modelPromise;
              return __generator(this, function (_a) {
                  switch (_a.label) {
                      case 0:
                          if (options.model_id) {
                              model_id = options.model_id;
                          }
                          else if (options.comm) {
                              model_id = options.model_id = options.comm.comm_id;
                          }
                          else {
                              throw new Error('Neither comm nor model_id provided in options object. At least one must exist.');
                          }
                          modelPromise = this._make_model(options, serialized_state);
                          // this call needs to happen before the first `await`, see note in `set_state`:
                          this.register_model(model_id, modelPromise);
                          return [4 /*yield*/, modelPromise];
                      case 1: return [2 /*return*/, _a.sent()];
                  }
              });
          });
      };
      ManagerBase.prototype._make_model = function (options, serialized_state) {
          if (serialized_state === void 0) { serialized_state = {}; }
          return __awaiter(this, void 0, void 0, function () {
              var model_id, model_promise, ModelType, error_1, attributes, modelOptions, widget_model;
              return __generator(this, function (_a) {
                  switch (_a.label) {
                      case 0:
                          model_id = options.model_id;
                          model_promise = this.loadClass(options.model_name, options.model_module, options.model_module_version);
                          _a.label = 1;
                      case 1:
                          _a.trys.push([1, 3, , 4]);
                          return [4 /*yield*/, model_promise];
                      case 2:
                          ModelType = _a.sent();
                          return [3 /*break*/, 4];
                      case 3:
                          error_1 = _a.sent();
                          console.error('Could not instantiate widget');
                          throw error_1;
                      case 4:
                          if (!ModelType) {
                              throw new Error("Cannot find model module " + options.model_module + "@" + options.model_module_version + ", " + options.model_name);
                          }
                          return [4 /*yield*/, ModelType._deserialize_state(serialized_state, this)];
                      case 5:
                          attributes = _a.sent();
                          modelOptions = {
                              widget_manager: this,
                              model_id: model_id,
                              comm: options.comm,
                          };
                          widget_model = new ModelType(attributes, modelOptions);
                          widget_model.name = options.model_name;
                          widget_model.module = options.model_module;
                          return [2 /*return*/, widget_model];
                  }
              });
          });
      };
      /**
       * Close all widgets and empty the widget state.
       * @return Promise that resolves when the widget state is cleared.
       */
      ManagerBase.prototype.clear_state = function () {
          var _this = this;
          return resolvePromisesDict(this._models).then(function (models) {
              Object.keys(models).forEach(function (id) { return models[id].close(); });
              _this._models = Object.create(null);
          });
      };
      /**
       * Asynchronously get the state of the widget manager.
       *
       * This includes all of the widget models, and follows the format given in
       * the @jupyter-widgets/schema package.
       *
       * @param options - The options for what state to return.
       * @returns Promise for a state dictionary
       */
      ManagerBase.prototype.get_state = function (options) {
          var _this = this;
          if (options === void 0) { options = {}; }
          var modelPromises = Object.keys(this._models).map(function (id) { return _this._models[id]; });
          return Promise.all(modelPromises).then(function (models) {
              return serialize_state(models, options);
          });
      };
      /**
       * Set the widget manager state.
       *
       * @param state - a Javascript object conforming to the application/vnd.jupyter.widget-state+json spec.
       *
       * Reconstructs all of the widget models in the state, merges that with the
       * current manager state, and then attempts to redisplay the widgets in the
       * state.
       */
      ManagerBase.prototype.set_state = function (state) {
          var _this = this;
          // Check to make sure that it's the same version we are parsing.
          if (!(state.version_major && state.version_major <= 2)) {
              throw 'Unsupported widget state format';
          }
          var models = state.state;
          // Recreate all the widget models for the given widget manager state.
          var all_models = this._get_comm_info().then(function (live_comms) {
              /* Note: It is currently safe to just loop over the models in any order,
                 given that the following holds (does at the time of writing):
                 1: any call to `new_model` with state registers the model promise (e.g. with `register_model`)
                    synchronously (before it's first `await` statement).
                 2: any calls to a model constructor or the `set_state` method on a model,
                    happens asynchronously (in a `then` clause, or after an `await` statement).

                Without these assumptions, one risks trying to set model state with a reference
                to another model that doesn't exist yet!
              */
              return Promise.all(Object.keys(models).map(function (model_id) {
                  // First put back the binary buffers
                  var decode = { 'base64': base64ToBuffer, 'hex': hexToBuffer };
                  var model = models[model_id];
                  var modelState = model.state;
                  if (model.buffers) {
                      var bufferPaths = model.buffers.map(function (b) { return b.path; });
                      // put_buffers expects buffers to be DataViews
                      var buffers = model.buffers.map(function (b) { return new DataView(decode[b.encoding](b.data)); });
                      put_buffers(model.state, bufferPaths, buffers);
                  }
                  // If the model has already been created, set its state and then
                  // return it.
                  if (_this._models[model_id]) {
                      return _this._models[model_id].then(function (model) {
                          // deserialize state
                          return model.constructor._deserialize_state(modelState || {}, _this).then(function (attributes) {
                              model.set_state(attributes); // case 2
                              return model;
                          });
                      });
                  }
                  var modelCreate = {
                      model_id: model_id,
                      model_name: model.model_name,
                      model_module: model.model_module,
                      model_module_version: model.model_module_version
                  };
                  if (live_comms.hasOwnProperty(model_id)) { // live comm
                      // This connects to an existing comm if it exists, and
                      // should *not* send a comm open message.
                      return _this._create_comm(_this.comm_target_name, model_id).then(function (comm) {
                          modelCreate.comm = comm;
                          return _this.new_model(modelCreate); // No state, so safe wrt. case 1
                      });
                  }
                  else {
                      return _this.new_model(modelCreate, modelState); // case 1
                  }
              }));
          });
          return all_models;
      };
      /**
       * Disconnect the widget manager from the kernel, setting each model's comm
       * as dead.
       */
      ManagerBase.prototype.disconnect = function () {
          var _this = this;
          Object.keys(this._models).forEach(function (i) {
              _this._models[i].then(function (model) { model.comm_live = false; });
          });
      };
      /**
       * Resolve a URL relative to the current notebook location.
       *
       * The default implementation just returns the original url.
       */
      ManagerBase.prototype.resolveUrl = function (url) {
          return Promise.resolve(url);
      };
      /**
       * Filter serialized widget state to remove any ID's already present in manager.
       *
       * @param {*} state Serialized state to filter
       *
       * @returns {*} A copy of the state, with its 'state' attribute filtered
       */
      ManagerBase.prototype.filterExistingModelState = function (serialized_state) {
          var _this = this;
          var models = serialized_state.state;
          models = Object.keys(models)
              .filter(function (model_id) {
              return !_this._models[model_id];
          })
              .reduce(function (res, model_id) {
              res[model_id] = models[model_id];
              return res;
          }, {});
          return __assign$1(__assign$1({}, serialized_state), { state: models });
      };
      return ManagerBase;
  }());
  /**
   * Serialize an array of widget models
   *
   * #### Notes
   * The return value follows the format given in the
   * @jupyter-widgets/schema package.
   */
  function serialize_state(models, options) {
      if (options === void 0) { options = {}; }
      var state = {};
      models.forEach(function (model) {
          var model_id = model.model_id;
          var split = remove_buffers(model.serialize(model.get_state(options.drop_defaults)));
          var buffers = split.buffers.map(function (buffer, index) {
              return {
                  data: bufferToBase64(buffer),
                  path: split.buffer_paths[index],
                  encoding: 'base64'
              };
          });
          state[model_id] = {
              model_name: model.name,
              model_module: model.module,
              model_module_version: model.get('_model_module_version'),
              state: split.state
          };
          // To save space, only include the buffers key if we have buffers
          if (buffers.length > 0) {
              state[model_id].buffers = buffers;
          }
      });
      return { version_major: 2, version_minor: 0, state: state };
  }

  // This file contains a modified version of the set function from the Backbone
  // (see
  // https://github.com/jashkenas/backbone/blob/05fde9e201f7e2137796663081105cd6dad12a98/backbone.js#L460,
  // with changes below marked with an EDIT comment). This file in Backbone has the following license.
  var __assign$2 = (undefined && undefined.__assign) || function () {
      __assign$2 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$2.apply(this, arguments);
  };
  // Set a hash of model attributes on the object, firing `"change"`. This is
  // the core primitive operation of a model, updating the data and notifying
  // anyone who needs to know about the change in state. The heart of the beast.
  // This *MUST* be called with the model as the `this` context.
  function set(key, val, options) {
      /* tslint:disable:no-invalid-this */
      if (key == null) {
          return this;
      }
      // Handle both `"key", value` and `{key: value}` -style arguments.
      var attrs;
      if (typeof key === 'object') {
          attrs = key;
          options = val;
      }
      else {
          (attrs = {})[key] = val;
      }
      options || (options = {});
      // Run validation.
      if (!this._validate(attrs, options)) {
          return false;
      }
      // Extract attributes and options.
      var unset = options.unset;
      var silent = options.silent;
      var changes = [];
      var changing = this._changing;
      this._changing = true;
      if (!changing) {
          // EDIT: changed to use object spread instead of _.clone
          this._previousAttributes = __assign$2({}, this.attributes);
          this.changed = {};
      }
      var current = this.attributes;
      var changed = this.changed;
      var prev = this._previousAttributes;
      // For each `set` attribute, update or delete the current value.
      for (var attr in attrs) {
          val = attrs[attr];
          // EDIT: the following two lines use our isEqual instead of _.isEqual
          if (!isEqual$1(current[attr], val)) {
              changes.push(attr);
          }
          if (!isEqual$1(prev[attr], val)) {
              changed[attr] = val;
          }
          else {
              delete changed[attr];
          }
          unset ? delete current[attr] : current[attr] = val;
      }
      // Update the `id`.
      this.id = this.get(this.idAttribute);
      // Trigger all relevant attribute changes.
      if (!silent) {
          if (changes.length) {
              this._pending = options;
          }
          for (var i = 0; i < changes.length; i++) {
              this.trigger('change:' + changes[i], this, current[changes[i]], options);
          }
      }
      // You might be wondering why there's a `while` loop here. Changes can
      // be recursively nested within `"change"` events.
      if (changing) {
          return this;
      }
      if (!silent) {
          while (this._pending) {
              options = this._pending;
              this._pending = false;
              this.trigger('change', this, options);
          }
      }
      this._pending = false;
      this._changing = false;
      return this;
      /* tslint:enable:no-invalid-this */
  }

  //     Underscore.js 1.10.2
  //     https://underscorejs.org
  //     (c) 2009-2020 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
  //     Underscore may be freely distributed under the MIT license.

  // Baseline setup
  // --------------

  // Establish the root object, `window` (`self`) in the browser, `global`
  // on the server, or `this` in some virtual machines. We use `self`
  // instead of `window` for `WebWorker` support.
  var root$1 = typeof self == 'object' && self.self === self && self ||
            typeof global == 'object' && global.global === global && global ||
            Function('return this')() ||
            {};

  // Save bytes in the minified (but not gzipped) version:
  var ArrayProto = Array.prototype, ObjProto = Object.prototype;
  var SymbolProto = typeof Symbol !== 'undefined' ? Symbol.prototype : null;

  // Create quick reference variables for speed access to core prototypes.
  var push = ArrayProto.push,
      slice = ArrayProto.slice,
      toString = ObjProto.toString,
      hasOwnProperty$a = ObjProto.hasOwnProperty;

  // All **ECMAScript 5** native function implementations that we hope to use
  // are declared here.
  var nativeIsArray = Array.isArray,
      nativeKeys$1 = Object.keys,
      nativeCreate$1 = Object.create;

  // Create references to these builtin functions because we override them.
  var _isNaN = root$1.isNaN,
      _isFinite = root$1.isFinite;

  // Naked function reference for surrogate-prototype-swapping.
  var Ctor = function(){};

  // The Underscore object. All exported functions below are added to it in the
  // modules/index-all.js using the mixin function.
  function _(obj) {
    if (obj instanceof _) return obj;
    if (!(this instanceof _)) return new _(obj);
    this._wrapped = obj;
  }

  // Current version.
  var VERSION = _.VERSION = '1.10.2';

  // Internal function that returns an efficient (for current engines) version
  // of the passed-in callback, to be repeatedly applied in other Underscore
  // functions.
  function optimizeCb(func, context, argCount) {
    if (context === void 0) return func;
    switch (argCount == null ? 3 : argCount) {
      case 1: return function(value) {
        return func.call(context, value);
      };
      // The 2-argument case is omitted because we’re not using it.
      case 3: return function(value, index, collection) {
        return func.call(context, value, index, collection);
      };
      case 4: return function(accumulator, value, index, collection) {
        return func.call(context, accumulator, value, index, collection);
      };
    }
    return function() {
      return func.apply(context, arguments);
    };
  }

  // An internal function to generate callbacks that can be applied to each
  // element in a collection, returning the desired result — either `identity`,
  // an arbitrary callback, a property matcher, or a property accessor.
  function baseIteratee(value, context, argCount) {
    if (value == null) return identity;
    if (isFunction$1(value)) return optimizeCb(value, context, argCount);
    if (isObject$1(value) && !isArray$1(value)) return matcher(value);
    return property(value);
  }

  // External wrapper for our callback generator. Users may customize
  // `_.iteratee` if they want additional predicate/iteratee shorthand styles.
  // This abstraction hides the internal-only argCount argument.
  _.iteratee = iteratee;
  function iteratee(value, context) {
    return baseIteratee(value, context, Infinity);
  }

  // The function we actually call internally. It invokes _.iteratee if
  // overridden, otherwise baseIteratee.
  function cb(value, context, argCount) {
    if (_.iteratee !== iteratee) return _.iteratee(value, context);
    return baseIteratee(value, context, argCount);
  }

  // Some functions take a variable number of arguments, or a few expected
  // arguments at the beginning and then a variable number of values to operate
  // on. This helper accumulates all remaining arguments past the function’s
  // argument length (or an explicit `startIndex`), into an array that becomes
  // the last argument. Similar to ES6’s "rest parameter".
  function restArguments(func, startIndex) {
    startIndex = startIndex == null ? func.length - 1 : +startIndex;
    return function() {
      var length = Math.max(arguments.length - startIndex, 0),
          rest = Array(length),
          index = 0;
      for (; index < length; index++) {
        rest[index] = arguments[index + startIndex];
      }
      switch (startIndex) {
        case 0: return func.call(this, rest);
        case 1: return func.call(this, arguments[0], rest);
        case 2: return func.call(this, arguments[0], arguments[1], rest);
      }
      var args = Array(startIndex + 1);
      for (index = 0; index < startIndex; index++) {
        args[index] = arguments[index];
      }
      args[startIndex] = rest;
      return func.apply(this, args);
    };
  }

  // An internal function for creating a new object that inherits from another.
  function baseCreate(prototype) {
    if (!isObject$1(prototype)) return {};
    if (nativeCreate$1) return nativeCreate$1(prototype);
    Ctor.prototype = prototype;
    var result = new Ctor;
    Ctor.prototype = null;
    return result;
  }

  function shallowProperty(key) {
    return function(obj) {
      return obj == null ? void 0 : obj[key];
    };
  }

  function _has(obj, path) {
    return obj != null && hasOwnProperty$a.call(obj, path);
  }

  function deepGet(obj, path) {
    var length = path.length;
    for (var i = 0; i < length; i++) {
      if (obj == null) return void 0;
      obj = obj[path[i]];
    }
    return length ? obj : void 0;
  }

  // Helper for collection methods to determine whether a collection
  // should be iterated as an array or as an object.
  // Related: https://people.mozilla.org/~jorendorff/es6-draft.html#sec-tolength
  // Avoids a very nasty iOS 8 JIT bug on ARM-64. #2094
  var MAX_ARRAY_INDEX = Math.pow(2, 53) - 1;
  var getLength = shallowProperty('length');
  function isArrayLike$1(collection) {
    var length = getLength(collection);
    return typeof length == 'number' && length >= 0 && length <= MAX_ARRAY_INDEX;
  }

  // Collection Functions
  // --------------------

  // The cornerstone, an `each` implementation, aka `forEach`.
  // Handles raw objects in addition to array-likes. Treats all
  // sparse array-likes as if they were dense.
  function each(obj, iteratee, context) {
    iteratee = optimizeCb(iteratee, context);
    var i, length;
    if (isArrayLike$1(obj)) {
      for (i = 0, length = obj.length; i < length; i++) {
        iteratee(obj[i], i, obj);
      }
    } else {
      var _keys = keys$1(obj);
      for (i = 0, length = _keys.length; i < length; i++) {
        iteratee(obj[_keys[i]], _keys[i], obj);
      }
    }
    return obj;
  }

  // Return the results of applying the iteratee to each element.
  function map(obj, iteratee, context) {
    iteratee = cb(iteratee, context);
    var _keys = !isArrayLike$1(obj) && keys$1(obj),
        length = (_keys || obj).length,
        results = Array(length);
    for (var index = 0; index < length; index++) {
      var currentKey = _keys ? _keys[index] : index;
      results[index] = iteratee(obj[currentKey], currentKey, obj);
    }
    return results;
  }

  // Create a reducing function iterating left or right.
  function createReduce(dir) {
    // Wrap code that reassigns argument variables in a separate function than
    // the one that accesses `arguments.length` to avoid a perf hit. (#1991)
    var reducer = function(obj, iteratee, memo, initial) {
      var _keys = !isArrayLike$1(obj) && keys$1(obj),
          length = (_keys || obj).length,
          index = dir > 0 ? 0 : length - 1;
      if (!initial) {
        memo = obj[_keys ? _keys[index] : index];
        index += dir;
      }
      for (; index >= 0 && index < length; index += dir) {
        var currentKey = _keys ? _keys[index] : index;
        memo = iteratee(memo, obj[currentKey], currentKey, obj);
      }
      return memo;
    };

    return function(obj, iteratee, memo, context) {
      var initial = arguments.length >= 3;
      return reducer(obj, optimizeCb(iteratee, context, 4), memo, initial);
    };
  }

  // **Reduce** builds up a single result from a list of values, aka `inject`,
  // or `foldl`.
  var reduce = createReduce(1);

  // The right-associative version of reduce, also known as `foldr`.
  var reduceRight = createReduce(-1);

  // Return the first value which passes a truth test.
  function find(obj, predicate, context) {
    var keyFinder = isArrayLike$1(obj) ? findIndex : findKey;
    var key = keyFinder(obj, predicate, context);
    if (key !== void 0 && key !== -1) return obj[key];
  }

  // Return all the elements that pass a truth test.
  function filter(obj, predicate, context) {
    var results = [];
    predicate = cb(predicate, context);
    each(obj, function(value, index, list) {
      if (predicate(value, index, list)) results.push(value);
    });
    return results;
  }

  // Return all the elements for which a truth test fails.
  function reject$1(obj, predicate, context) {
    return filter(obj, negate(cb(predicate)), context);
  }

  // Determine whether all of the elements match a truth test.
  function every(obj, predicate, context) {
    predicate = cb(predicate, context);
    var _keys = !isArrayLike$1(obj) && keys$1(obj),
        length = (_keys || obj).length;
    for (var index = 0; index < length; index++) {
      var currentKey = _keys ? _keys[index] : index;
      if (!predicate(obj[currentKey], currentKey, obj)) return false;
    }
    return true;
  }

  // Determine if at least one element in the object matches a truth test.
  function some(obj, predicate, context) {
    predicate = cb(predicate, context);
    var _keys = !isArrayLike$1(obj) && keys$1(obj),
        length = (_keys || obj).length;
    for (var index = 0; index < length; index++) {
      var currentKey = _keys ? _keys[index] : index;
      if (predicate(obj[currentKey], currentKey, obj)) return true;
    }
    return false;
  }

  // Determine if the array or object contains a given item (using `===`).
  function contains(obj, item, fromIndex, guard) {
    if (!isArrayLike$1(obj)) obj = values(obj);
    if (typeof fromIndex != 'number' || guard) fromIndex = 0;
    return indexOf(obj, item, fromIndex) >= 0;
  }

  // Invoke a method (with arguments) on every item in a collection.
  var invoke = restArguments(function(obj, path, args) {
    var contextPath, func;
    if (isFunction$1(path)) {
      func = path;
    } else if (isArray$1(path)) {
      contextPath = path.slice(0, -1);
      path = path[path.length - 1];
    }
    return map(obj, function(context) {
      var method = func;
      if (!method) {
        if (contextPath && contextPath.length) {
          context = deepGet(context, contextPath);
        }
        if (context == null) return void 0;
        method = context[path];
      }
      return method == null ? method : method.apply(context, args);
    });
  });

  // Convenience version of a common use case of `map`: fetching a property.
  function pluck(obj, key) {
    return map(obj, property(key));
  }

  // Convenience version of a common use case of `filter`: selecting only objects
  // containing specific `key:value` pairs.
  function where(obj, attrs) {
    return filter(obj, matcher(attrs));
  }

  // Convenience version of a common use case of `find`: getting the first object
  // containing specific `key:value` pairs.
  function findWhere(obj, attrs) {
    return find(obj, matcher(attrs));
  }

  // Return the maximum element (or element-based computation).
  function max(obj, iteratee, context) {
    var result = -Infinity, lastComputed = -Infinity,
        value, computed;
    if (iteratee == null || typeof iteratee == 'number' && typeof obj[0] != 'object' && obj != null) {
      obj = isArrayLike$1(obj) ? obj : values(obj);
      for (var i = 0, length = obj.length; i < length; i++) {
        value = obj[i];
        if (value != null && value > result) {
          result = value;
        }
      }
    } else {
      iteratee = cb(iteratee, context);
      each(obj, function(v, index, list) {
        computed = iteratee(v, index, list);
        if (computed > lastComputed || computed === -Infinity && result === -Infinity) {
          result = v;
          lastComputed = computed;
        }
      });
    }
    return result;
  }

  // Return the minimum element (or element-based computation).
  function min(obj, iteratee, context) {
    var result = Infinity, lastComputed = Infinity,
        value, computed;
    if (iteratee == null || typeof iteratee == 'number' && typeof obj[0] != 'object' && obj != null) {
      obj = isArrayLike$1(obj) ? obj : values(obj);
      for (var i = 0, length = obj.length; i < length; i++) {
        value = obj[i];
        if (value != null && value < result) {
          result = value;
        }
      }
    } else {
      iteratee = cb(iteratee, context);
      each(obj, function(v, index, list) {
        computed = iteratee(v, index, list);
        if (computed < lastComputed || computed === Infinity && result === Infinity) {
          result = v;
          lastComputed = computed;
        }
      });
    }
    return result;
  }

  // Shuffle a collection.
  function shuffle(obj) {
    return sample(obj, Infinity);
  }

  // Sample **n** random values from a collection using the modern version of the
  // [Fisher-Yates shuffle](https://en.wikipedia.org/wiki/Fisher–Yates_shuffle).
  // If **n** is not specified, returns a single random element.
  // The internal `guard` argument allows it to work with `map`.
  function sample(obj, n, guard) {
    if (n == null || guard) {
      if (!isArrayLike$1(obj)) obj = values(obj);
      return obj[random(obj.length - 1)];
    }
    var sample = isArrayLike$1(obj) ? clone(obj) : values(obj);
    var length = getLength(sample);
    n = Math.max(Math.min(n, length), 0);
    var last = length - 1;
    for (var index = 0; index < n; index++) {
      var rand = random(index, last);
      var temp = sample[index];
      sample[index] = sample[rand];
      sample[rand] = temp;
    }
    return sample.slice(0, n);
  }

  // Sort the object's values by a criterion produced by an iteratee.
  function sortBy(obj, iteratee, context) {
    var index = 0;
    iteratee = cb(iteratee, context);
    return pluck(map(obj, function(value, key, list) {
      return {
        value: value,
        index: index++,
        criteria: iteratee(value, key, list)
      };
    }).sort(function(left, right) {
      var a = left.criteria;
      var b = right.criteria;
      if (a !== b) {
        if (a > b || a === void 0) return 1;
        if (a < b || b === void 0) return -1;
      }
      return left.index - right.index;
    }), 'value');
  }

  // An internal function used for aggregate "group by" operations.
  function group(behavior, partition) {
    return function(obj, iteratee, context) {
      var result = partition ? [[], []] : {};
      iteratee = cb(iteratee, context);
      each(obj, function(value, index) {
        var key = iteratee(value, index, obj);
        behavior(result, value, key);
      });
      return result;
    };
  }

  // Groups the object's values by a criterion. Pass either a string attribute
  // to group by, or a function that returns the criterion.
  var groupBy = group(function(result, value, key) {
    if (_has(result, key)) result[key].push(value); else result[key] = [value];
  });

  // Indexes the object's values by a criterion, similar to `groupBy`, but for
  // when you know that your index values will be unique.
  var indexBy = group(function(result, value, key) {
    result[key] = value;
  });

  // Counts instances of an object that group by a certain criterion. Pass
  // either a string attribute to count by, or a function that returns the
  // criterion.
  var countBy = group(function(result, value, key) {
    if (_has(result, key)) result[key]++; else result[key] = 1;
  });

  var reStrSymbol = /[^\ud800-\udfff]|[\ud800-\udbff][\udc00-\udfff]|[\ud800-\udfff]/g;
  // Safely create a real, live array from anything iterable.
  function toArray(obj) {
    if (!obj) return [];
    if (isArray$1(obj)) return slice.call(obj);
    if (isString(obj)) {
      // Keep surrogate pair characters together
      return obj.match(reStrSymbol);
    }
    if (isArrayLike$1(obj)) return map(obj, identity);
    return values(obj);
  }

  // Return the number of elements in an object.
  function size(obj) {
    if (obj == null) return 0;
    return isArrayLike$1(obj) ? obj.length : keys$1(obj).length;
  }

  // Split a collection into two arrays: one whose elements all satisfy the given
  // predicate, and one whose elements all do not satisfy the predicate.
  var partition = group(function(result, value, pass) {
    result[pass ? 0 : 1].push(value);
  }, true);

  // Array Functions
  // ---------------

  // Get the first element of an array. Passing **n** will return the first N
  // values in the array. The **guard** check allows it to work with `map`.
  function first(array, n, guard) {
    if (array == null || array.length < 1) return n == null ? void 0 : [];
    if (n == null || guard) return array[0];
    return initial(array, array.length - n);
  }

  // Returns everything but the last entry of the array. Especially useful on
  // the arguments object. Passing **n** will return all the values in
  // the array, excluding the last N.
  function initial(array, n, guard) {
    return slice.call(array, 0, Math.max(0, array.length - (n == null || guard ? 1 : n)));
  }

  // Get the last element of an array. Passing **n** will return the last N
  // values in the array.
  function last(array, n, guard) {
    if (array == null || array.length < 1) return n == null ? void 0 : [];
    if (n == null || guard) return array[array.length - 1];
    return rest(array, Math.max(0, array.length - n));
  }

  // Returns everything but the first entry of the array. Especially useful on
  // the arguments object. Passing an **n** will return the rest N values in the
  // array.
  function rest(array, n, guard) {
    return slice.call(array, n == null || guard ? 1 : n);
  }

  // Trim out all falsy values from an array.
  function compact(array) {
    return filter(array, Boolean);
  }

  // Internal implementation of a recursive `flatten` function.
  function _flatten(input, shallow, strict, output) {
    output = output || [];
    var idx = output.length;
    for (var i = 0, length = getLength(input); i < length; i++) {
      var value = input[i];
      if (isArrayLike$1(value) && (isArray$1(value) || isArguments$1(value))) {
        // Flatten current level of array or arguments object.
        if (shallow) {
          var j = 0, len = value.length;
          while (j < len) output[idx++] = value[j++];
        } else {
          _flatten(value, shallow, strict, output);
          idx = output.length;
        }
      } else if (!strict) {
        output[idx++] = value;
      }
    }
    return output;
  }

  // Flatten out an array, either recursively (by default), or just one level.
  function flatten(array, shallow) {
    return _flatten(array, shallow, false);
  }

  // Return a version of the array that does not contain the specified value(s).
  var without = restArguments(function(array, otherArrays) {
    return difference$1(array, otherArrays);
  });

  // Produce a duplicate-free version of the array. If the array has already
  // been sorted, you have the option of using a faster algorithm.
  // The faster algorithm will not work with an iteratee if the iteratee
  // is not a one-to-one function, so providing an iteratee will disable
  // the faster algorithm.
  function uniq(array, isSorted, iteratee, context) {
    if (!isBoolean(isSorted)) {
      context = iteratee;
      iteratee = isSorted;
      isSorted = false;
    }
    if (iteratee != null) iteratee = cb(iteratee, context);
    var result = [];
    var seen = [];
    for (var i = 0, length = getLength(array); i < length; i++) {
      var value = array[i],
          computed = iteratee ? iteratee(value, i, array) : value;
      if (isSorted && !iteratee) {
        if (!i || seen !== computed) result.push(value);
        seen = computed;
      } else if (iteratee) {
        if (!contains(seen, computed)) {
          seen.push(computed);
          result.push(value);
        }
      } else if (!contains(result, value)) {
        result.push(value);
      }
    }
    return result;
  }

  // Produce an array that contains the union: each distinct element from all of
  // the passed-in arrays.
  var union = restArguments(function(arrays) {
    return uniq(_flatten(arrays, true, true));
  });

  // Produce an array that contains every item shared between all the
  // passed-in arrays.
  function intersection(array) {
    var result = [];
    var argsLength = arguments.length;
    for (var i = 0, length = getLength(array); i < length; i++) {
      var item = array[i];
      if (contains(result, item)) continue;
      var j;
      for (j = 1; j < argsLength; j++) {
        if (!contains(arguments[j], item)) break;
      }
      if (j === argsLength) result.push(item);
    }
    return result;
  }

  // Take the difference between one array and a number of other arrays.
  // Only the elements present in just the first array will remain.
  var difference$1 = restArguments(function(array, rest) {
    rest = _flatten(rest, true, true);
    return filter(array, function(value){
      return !contains(rest, value);
    });
  });

  // Complement of zip. Unzip accepts an array of arrays and groups
  // each array's elements on shared indices.
  function unzip(array) {
    var length = array && max(array, getLength).length || 0;
    var result = Array(length);

    for (var index = 0; index < length; index++) {
      result[index] = pluck(array, index);
    }
    return result;
  }

  // Zip together multiple lists into a single array -- elements that share
  // an index go together.
  var zip = restArguments(unzip);

  // Converts lists into objects. Pass either a single array of `[key, value]`
  // pairs, or two parallel arrays of the same length -- one of keys, and one of
  // the corresponding values. Passing by pairs is the reverse of pairs.
  function object(list, values) {
    var result = {};
    for (var i = 0, length = getLength(list); i < length; i++) {
      if (values) {
        result[list[i]] = values[i];
      } else {
        result[list[i][0]] = list[i][1];
      }
    }
    return result;
  }

  // Generator function to create the findIndex and findLastIndex functions.
  function createPredicateIndexFinder(dir) {
    return function(array, predicate, context) {
      predicate = cb(predicate, context);
      var length = getLength(array);
      var index = dir > 0 ? 0 : length - 1;
      for (; index >= 0 && index < length; index += dir) {
        if (predicate(array[index], index, array)) return index;
      }
      return -1;
    };
  }

  // Returns the first index on an array-like that passes a predicate test.
  var findIndex = createPredicateIndexFinder(1);
  var findLastIndex = createPredicateIndexFinder(-1);

  // Use a comparator function to figure out the smallest index at which
  // an object should be inserted so as to maintain order. Uses binary search.
  function sortedIndex(array, obj, iteratee, context) {
    iteratee = cb(iteratee, context, 1);
    var value = iteratee(obj);
    var low = 0, high = getLength(array);
    while (low < high) {
      var mid = Math.floor((low + high) / 2);
      if (iteratee(array[mid]) < value) low = mid + 1; else high = mid;
    }
    return low;
  }

  // Generator function to create the indexOf and lastIndexOf functions.
  function createIndexFinder(dir, predicateFind, sortedIndex) {
    return function(array, item, idx) {
      var i = 0, length = getLength(array);
      if (typeof idx == 'number') {
        if (dir > 0) {
          i = idx >= 0 ? idx : Math.max(idx + length, i);
        } else {
          length = idx >= 0 ? Math.min(idx + 1, length) : idx + length + 1;
        }
      } else if (sortedIndex && idx && length) {
        idx = sortedIndex(array, item);
        return array[idx] === item ? idx : -1;
      }
      if (item !== item) {
        idx = predicateFind(slice.call(array, i, length), isNaN$1);
        return idx >= 0 ? idx + i : -1;
      }
      for (idx = dir > 0 ? i : length - 1; idx >= 0 && idx < length; idx += dir) {
        if (array[idx] === item) return idx;
      }
      return -1;
    };
  }

  // Return the position of the first occurrence of an item in an array,
  // or -1 if the item is not included in the array.
  // If the array is large and already in sort order, pass `true`
  // for **isSorted** to use binary search.
  var indexOf = createIndexFinder(1, findIndex, sortedIndex);
  var lastIndexOf = createIndexFinder(-1, findLastIndex);

  // Generate an integer Array containing an arithmetic progression. A port of
  // the native Python `range()` function. See
  // [the Python documentation](https://docs.python.org/library/functions.html#range).
  function range(start, stop, step) {
    if (stop == null) {
      stop = start || 0;
      start = 0;
    }
    if (!step) {
      step = stop < start ? -1 : 1;
    }

    var length = Math.max(Math.ceil((stop - start) / step), 0);
    var range = Array(length);

    for (var idx = 0; idx < length; idx++, start += step) {
      range[idx] = start;
    }

    return range;
  }

  // Chunk a single array into multiple arrays, each containing `count` or fewer
  // items.
  function chunk(array, count) {
    if (count == null || count < 1) return [];
    var result = [];
    var i = 0, length = array.length;
    while (i < length) {
      result.push(slice.call(array, i, i += count));
    }
    return result;
  }

  // Function (ahem) Functions
  // ------------------

  // Determines whether to execute a function as a constructor
  // or a normal function with the provided arguments.
  function executeBound(sourceFunc, boundFunc, context, callingContext, args) {
    if (!(callingContext instanceof boundFunc)) return sourceFunc.apply(context, args);
    var self = baseCreate(sourceFunc.prototype);
    var result = sourceFunc.apply(self, args);
    if (isObject$1(result)) return result;
    return self;
  }

  // Create a function bound to a given object (assigning `this`, and arguments,
  // optionally). Delegates to **ECMAScript 5**'s native `Function.bind` if
  // available.
  var bind = restArguments(function(func, context, args) {
    if (!isFunction$1(func)) throw new TypeError('Bind must be called on a function');
    var bound = restArguments(function(callArgs) {
      return executeBound(func, bound, context, this, args.concat(callArgs));
    });
    return bound;
  });

  // Partially apply a function by creating a version that has had some of its
  // arguments pre-filled, without changing its dynamic `this` context. _ acts
  // as a placeholder by default, allowing any combination of arguments to be
  // pre-filled. Set `partial.placeholder` for a custom placeholder argument.
  var partial = restArguments(function(func, boundArgs) {
    var placeholder = partial.placeholder;
    var bound = function() {
      var position = 0, length = boundArgs.length;
      var args = Array(length);
      for (var i = 0; i < length; i++) {
        args[i] = boundArgs[i] === placeholder ? arguments[position++] : boundArgs[i];
      }
      while (position < arguments.length) args.push(arguments[position++]);
      return executeBound(func, bound, this, this, args);
    };
    return bound;
  });

  partial.placeholder = _;

  // Bind a number of an object's methods to that object. Remaining arguments
  // are the method names to be bound. Useful for ensuring that all callbacks
  // defined on an object belong to it.
  var bindAll = restArguments(function(obj, _keys) {
    _keys = _flatten(_keys, false, false);
    var index = _keys.length;
    if (index < 1) throw new Error('bindAll must be passed function names');
    while (index--) {
      var key = _keys[index];
      obj[key] = bind(obj[key], obj);
    }
  });

  // Memoize an expensive function by storing its results.
  function memoize(func, hasher) {
    var memoize = function(key) {
      var cache = memoize.cache;
      var address = '' + (hasher ? hasher.apply(this, arguments) : key);
      if (!_has(cache, address)) cache[address] = func.apply(this, arguments);
      return cache[address];
    };
    memoize.cache = {};
    return memoize;
  }

  // Delays a function for the given number of milliseconds, and then calls
  // it with the arguments supplied.
  var delay = restArguments(function(func, wait, args) {
    return setTimeout(function() {
      return func.apply(null, args);
    }, wait);
  });

  // Defers a function, scheduling it to run after the current call stack has
  // cleared.
  var defer = partial(delay, _, 1);

  // Returns a function, that, when invoked, will only be triggered at most once
  // during a given window of time. Normally, the throttled function will run
  // as much as it can, without ever going more than once per `wait` duration;
  // but if you'd like to disable the execution on the leading edge, pass
  // `{leading: false}`. To disable execution on the trailing edge, ditto.
  function throttle(func, wait, options) {
    var timeout, context, args, result;
    var previous = 0;
    if (!options) options = {};

    var later = function() {
      previous = options.leading === false ? 0 : now();
      timeout = null;
      result = func.apply(context, args);
      if (!timeout) context = args = null;
    };

    var throttled = function() {
      var _now = now();
      if (!previous && options.leading === false) previous = _now;
      var remaining = wait - (_now - previous);
      context = this;
      args = arguments;
      if (remaining <= 0 || remaining > wait) {
        if (timeout) {
          clearTimeout(timeout);
          timeout = null;
        }
        previous = _now;
        result = func.apply(context, args);
        if (!timeout) context = args = null;
      } else if (!timeout && options.trailing !== false) {
        timeout = setTimeout(later, remaining);
      }
      return result;
    };

    throttled.cancel = function() {
      clearTimeout(timeout);
      previous = 0;
      timeout = context = args = null;
    };

    return throttled;
  }

  // Returns a function, that, as long as it continues to be invoked, will not
  // be triggered. The function will be called after it stops being called for
  // N milliseconds. If `immediate` is passed, trigger the function on the
  // leading edge, instead of the trailing.
  function debounce(func, wait, immediate) {
    var timeout, result;

    var later = function(context, args) {
      timeout = null;
      if (args) result = func.apply(context, args);
    };

    var debounced = restArguments(function(args) {
      if (timeout) clearTimeout(timeout);
      if (immediate) {
        var callNow = !timeout;
        timeout = setTimeout(later, wait);
        if (callNow) result = func.apply(this, args);
      } else {
        timeout = delay(later, wait, this, args);
      }

      return result;
    });

    debounced.cancel = function() {
      clearTimeout(timeout);
      timeout = null;
    };

    return debounced;
  }

  // Returns the first function passed as an argument to the second,
  // allowing you to adjust arguments, run code before and after, and
  // conditionally execute the original function.
  function wrap(func, wrapper) {
    return partial(wrapper, func);
  }

  // Returns a negated version of the passed-in predicate.
  function negate(predicate) {
    return function() {
      return !predicate.apply(this, arguments);
    };
  }

  // Returns a function that is the composition of a list of functions, each
  // consuming the return value of the function that follows.
  function compose() {
    var args = arguments;
    var start = args.length - 1;
    return function() {
      var i = start;
      var result = args[start].apply(this, arguments);
      while (i--) result = args[i].call(this, result);
      return result;
    };
  }

  // Returns a function that will only be executed on and after the Nth call.
  function after(times, func) {
    return function() {
      if (--times < 1) {
        return func.apply(this, arguments);
      }
    };
  }

  // Returns a function that will only be executed up to (but not including) the Nth call.
  function before(times, func) {
    var memo;
    return function() {
      if (--times > 0) {
        memo = func.apply(this, arguments);
      }
      if (times <= 1) func = null;
      return memo;
    };
  }

  // Returns a function that will be executed at most one time, no matter how
  // often you call it. Useful for lazy initialization.
  var once = partial(before, 2);

  // Object Functions
  // ----------------

  // Keys in IE < 9 that won't be iterated by `for key in ...` and thus missed.
  var hasEnumBug = !{toString: null}.propertyIsEnumerable('toString');
  var nonEnumerableProps = ['valueOf', 'isPrototypeOf', 'toString',
    'propertyIsEnumerable', 'hasOwnProperty', 'toLocaleString'];

  function collectNonEnumProps(obj, _keys) {
    var nonEnumIdx = nonEnumerableProps.length;
    var constructor = obj.constructor;
    var proto = isFunction$1(constructor) && constructor.prototype || ObjProto;

    // Constructor is a special case.
    var prop = 'constructor';
    if (_has(obj, prop) && !contains(_keys, prop)) _keys.push(prop);

    while (nonEnumIdx--) {
      prop = nonEnumerableProps[nonEnumIdx];
      if (prop in obj && obj[prop] !== proto[prop] && !contains(_keys, prop)) {
        _keys.push(prop);
      }
    }
  }

  // Retrieve the names of an object's own properties.
  // Delegates to **ECMAScript 5**'s native `Object.keys`.
  function keys$1(obj) {
    if (!isObject$1(obj)) return [];
    if (nativeKeys$1) return nativeKeys$1(obj);
    var _keys = [];
    for (var key in obj) if (_has(obj, key)) _keys.push(key);
    // Ahem, IE < 9.
    if (hasEnumBug) collectNonEnumProps(obj, _keys);
    return _keys;
  }

  // Retrieve all the property names of an object.
  function allKeys(obj) {
    if (!isObject$1(obj)) return [];
    var _keys = [];
    for (var key in obj) _keys.push(key);
    // Ahem, IE < 9.
    if (hasEnumBug) collectNonEnumProps(obj, _keys);
    return _keys;
  }

  // Retrieve the values of an object's properties.
  function values(obj) {
    var _keys = keys$1(obj);
    var length = _keys.length;
    var values = Array(length);
    for (var i = 0; i < length; i++) {
      values[i] = obj[_keys[i]];
    }
    return values;
  }

  // Returns the results of applying the iteratee to each element of the object.
  // In contrast to map it returns an object.
  function mapObject(obj, iteratee, context) {
    iteratee = cb(iteratee, context);
    var _keys = keys$1(obj),
        length = _keys.length,
        results = {};
    for (var index = 0; index < length; index++) {
      var currentKey = _keys[index];
      results[currentKey] = iteratee(obj[currentKey], currentKey, obj);
    }
    return results;
  }

  // Convert an object into a list of `[key, value]` pairs.
  // The opposite of object.
  function pairs(obj) {
    var _keys = keys$1(obj);
    var length = _keys.length;
    var pairs = Array(length);
    for (var i = 0; i < length; i++) {
      pairs[i] = [_keys[i], obj[_keys[i]]];
    }
    return pairs;
  }

  // Invert the keys and values of an object. The values must be serializable.
  function invert(obj) {
    var result = {};
    var _keys = keys$1(obj);
    for (var i = 0, length = _keys.length; i < length; i++) {
      result[obj[_keys[i]]] = _keys[i];
    }
    return result;
  }

  // Return a sorted list of the function names available on the object.
  function functions(obj) {
    var names = [];
    for (var key in obj) {
      if (isFunction$1(obj[key])) names.push(key);
    }
    return names.sort();
  }

  // An internal function for creating assigner functions.
  function createAssigner(keysFunc, defaults) {
    return function(obj) {
      var length = arguments.length;
      if (defaults) obj = Object(obj);
      if (length < 2 || obj == null) return obj;
      for (var index = 1; index < length; index++) {
        var source = arguments[index],
            _keys = keysFunc(source),
            l = _keys.length;
        for (var i = 0; i < l; i++) {
          var key = _keys[i];
          if (!defaults || obj[key] === void 0) obj[key] = source[key];
        }
      }
      return obj;
    };
  }

  // Extend a given object with all the properties in passed-in object(s).
  var extend = createAssigner(allKeys);

  // Assigns a given object with all the own properties in the passed-in object(s).
  // (https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Object/assign)
  var extendOwn = createAssigner(keys$1);

  // Returns the first key on an object that passes a predicate test.
  function findKey(obj, predicate, context) {
    predicate = cb(predicate, context);
    var _keys = keys$1(obj), key;
    for (var i = 0, length = _keys.length; i < length; i++) {
      key = _keys[i];
      if (predicate(obj[key], key, obj)) return key;
    }
  }

  // Internal pick helper function to determine if `obj` has key `key`.
  function keyInObj(value, key, obj) {
    return key in obj;
  }

  // Return a copy of the object only containing the whitelisted properties.
  var pick = restArguments(function(obj, _keys) {
    var result = {}, iteratee = _keys[0];
    if (obj == null) return result;
    if (isFunction$1(iteratee)) {
      if (_keys.length > 1) iteratee = optimizeCb(iteratee, _keys[1]);
      _keys = allKeys(obj);
    } else {
      iteratee = keyInObj;
      _keys = _flatten(_keys, false, false);
      obj = Object(obj);
    }
    for (var i = 0, length = _keys.length; i < length; i++) {
      var key = _keys[i];
      var value = obj[key];
      if (iteratee(value, key, obj)) result[key] = value;
    }
    return result;
  });

  // Return a copy of the object without the blacklisted properties.
  var omit = restArguments(function(obj, _keys) {
    var iteratee = _keys[0], context;
    if (isFunction$1(iteratee)) {
      iteratee = negate(iteratee);
      if (_keys.length > 1) context = _keys[1];
    } else {
      _keys = map(_flatten(_keys, false, false), String);
      iteratee = function(value, key) {
        return !contains(_keys, key);
      };
    }
    return pick(obj, iteratee, context);
  });

  // Fill in a given object with default properties.
  var defaults = createAssigner(allKeys, true);

  // Creates an object that inherits from the given prototype object.
  // If additional properties are provided then they will be added to the
  // created object.
  function create(prototype, props) {
    var result = baseCreate(prototype);
    if (props) extendOwn(result, props);
    return result;
  }

  // Create a (shallow-cloned) duplicate of an object.
  function clone(obj) {
    if (!isObject$1(obj)) return obj;
    return isArray$1(obj) ? obj.slice() : extend({}, obj);
  }

  // Invokes interceptor with the obj, and then returns obj.
  // The primary purpose of this method is to "tap into" a method chain, in
  // order to perform operations on intermediate results within the chain.
  function tap(obj, interceptor) {
    interceptor(obj);
    return obj;
  }

  // Returns whether an object has a given set of `key:value` pairs.
  function isMatch(object, attrs) {
    var _keys = keys$1(attrs), length = _keys.length;
    if (object == null) return !length;
    var obj = Object(object);
    for (var i = 0; i < length; i++) {
      var key = _keys[i];
      if (attrs[key] !== obj[key] || !(key in obj)) return false;
    }
    return true;
  }


  // Internal recursive comparison function for `isEqual`.
  function eq$1(a, b, aStack, bStack) {
    // Identical objects are equal. `0 === -0`, but they aren't identical.
    // See the [Harmony `egal` proposal](https://wiki.ecmascript.org/doku.php?id=harmony:egal).
    if (a === b) return a !== 0 || 1 / a === 1 / b;
    // `null` or `undefined` only equal to itself (strict comparison).
    if (a == null || b == null) return false;
    // `NaN`s are equivalent, but non-reflexive.
    if (a !== a) return b !== b;
    // Exhaust primitive checks
    var type = typeof a;
    if (type !== 'function' && type !== 'object' && typeof b != 'object') return false;
    return deepEq(a, b, aStack, bStack);
  }

  // Internal recursive comparison function for `isEqual`.
  function deepEq(a, b, aStack, bStack) {
    // Unwrap any wrapped objects.
    if (a instanceof _) a = a._wrapped;
    if (b instanceof _) b = b._wrapped;
    // Compare `[[Class]]` names.
    var className = toString.call(a);
    if (className !== toString.call(b)) return false;
    switch (className) {
      // Strings, numbers, regular expressions, dates, and booleans are compared by value.
      case '[object RegExp]':
      // RegExps are coerced to strings for comparison (Note: '' + /a/i === '/a/i')
      case '[object String]':
        // Primitives and their corresponding object wrappers are equivalent; thus, `"5"` is
        // equivalent to `new String("5")`.
        return '' + a === '' + b;
      case '[object Number]':
        // `NaN`s are equivalent, but non-reflexive.
        // Object(NaN) is equivalent to NaN.
        if (+a !== +a) return +b !== +b;
        // An `egal` comparison is performed for other numeric values.
        return +a === 0 ? 1 / +a === 1 / b : +a === +b;
      case '[object Date]':
      case '[object Boolean]':
        // Coerce dates and booleans to numeric primitive values. Dates are compared by their
        // millisecond representations. Note that invalid dates with millisecond representations
        // of `NaN` are not equivalent.
        return +a === +b;
      case '[object Symbol]':
        return SymbolProto.valueOf.call(a) === SymbolProto.valueOf.call(b);
    }

    var areArrays = className === '[object Array]';
    if (!areArrays) {
      if (typeof a != 'object' || typeof b != 'object') return false;

      // Objects with different constructors are not equivalent, but `Object`s or `Array`s
      // from different frames are.
      var aCtor = a.constructor, bCtor = b.constructor;
      if (aCtor !== bCtor && !(isFunction$1(aCtor) && aCtor instanceof aCtor &&
                               isFunction$1(bCtor) && bCtor instanceof bCtor)
                          && ('constructor' in a && 'constructor' in b)) {
        return false;
      }
    }
    // Assume equality for cyclic structures. The algorithm for detecting cyclic
    // structures is adapted from ES 5.1 section 15.12.3, abstract operation `JO`.

    // Initializing stack of traversed objects.
    // It's done here since we only need them for objects and arrays comparison.
    aStack = aStack || [];
    bStack = bStack || [];
    var length = aStack.length;
    while (length--) {
      // Linear search. Performance is inversely proportional to the number of
      // unique nested structures.
      if (aStack[length] === a) return bStack[length] === b;
    }

    // Add the first object to the stack of traversed objects.
    aStack.push(a);
    bStack.push(b);

    // Recursively compare objects and arrays.
    if (areArrays) {
      // Compare array lengths to determine if a deep comparison is necessary.
      length = a.length;
      if (length !== b.length) return false;
      // Deep compare the contents, ignoring non-numeric properties.
      while (length--) {
        if (!eq$1(a[length], b[length], aStack, bStack)) return false;
      }
    } else {
      // Deep compare objects.
      var _keys = keys$1(a), key;
      length = _keys.length;
      // Ensure that both objects contain the same number of properties before comparing deep equality.
      if (keys$1(b).length !== length) return false;
      while (length--) {
        // Deep compare each member
        key = _keys[length];
        if (!(_has(b, key) && eq$1(a[key], b[key], aStack, bStack))) return false;
      }
    }
    // Remove the first object from the stack of traversed objects.
    aStack.pop();
    bStack.pop();
    return true;
  }

  // Perform a deep comparison to check if two objects are equal.
  function isEqual$2(a, b) {
    return eq$1(a, b);
  }

  // Is a given array, string, or object empty?
  // An "empty" object has no enumerable own-properties.
  function isEmpty(obj) {
    if (obj == null) return true;
    if (isArrayLike$1(obj) && (isArray$1(obj) || isString(obj) || isArguments$1(obj))) return obj.length === 0;
    return keys$1(obj).length === 0;
  }

  // Is a given value a DOM element?
  function isElement(obj) {
    return !!(obj && obj.nodeType === 1);
  }

  // Internal function for creating a toString-based type tester.
  function tagTester(name) {
    return function(obj) {
      return toString.call(obj) === '[object ' + name + ']';
    };
  }

  // Is a given value an array?
  // Delegates to ECMA5's native Array.isArray
  var isArray$1 = nativeIsArray || tagTester('Array');

  // Is a given variable an object?
  function isObject$1(obj) {
    var type = typeof obj;
    return type === 'function' || type === 'object' && !!obj;
  }

  // Add some isType methods: isArguments, isFunction, isString, isNumber, isDate, isRegExp, isError, isMap, isWeakMap, isSet, isWeakSet.
  var isArguments$1 = tagTester('Arguments');
  var isFunction$1 = tagTester('Function');
  var isString = tagTester('String');
  var isNumber = tagTester('Number');
  var isDate = tagTester('Date');
  var isRegExp = tagTester('RegExp');
  var isError = tagTester('Error');
  var isSymbol = tagTester('Symbol');
  var isMap = tagTester('Map');
  var isWeakMap = tagTester('WeakMap');
  var isSet = tagTester('Set');
  var isWeakSet = tagTester('WeakSet');

  // Define a fallback version of the method in browsers (ahem, IE < 9), where
  // there isn't any inspectable "Arguments" type.
  (function() {
    if (!isArguments$1(arguments)) {
      isArguments$1 = function(obj) {
        return _has(obj, 'callee');
      };
    }
  }());

  // Optimize `isFunction` if appropriate. Work around some typeof bugs in old v8,
  // IE 11 (#1621), Safari 8 (#1929), and PhantomJS (#2236).
  var nodelist = root$1.document && root$1.document.childNodes;
  if (typeof Int8Array != 'object' && typeof nodelist != 'function') {
    isFunction$1 = function(obj) {
      return typeof obj == 'function' || false;
    };
  }

  // Is a given object a finite number?
  function isFinite(obj) {
    return !isSymbol(obj) && _isFinite(obj) && !_isNaN(parseFloat(obj));
  }

  // Is the given value `NaN`?
  function isNaN$1(obj) {
    return isNumber(obj) && _isNaN(obj);
  }

  // Is a given value a boolean?
  function isBoolean(obj) {
    return obj === true || obj === false || toString.call(obj) === '[object Boolean]';
  }

  // Is a given value equal to null?
  function isNull(obj) {
    return obj === null;
  }

  // Is a given variable undefined?
  function isUndefined(obj) {
    return obj === void 0;
  }

  // Shortcut function for checking if an object has a given property directly
  // on itself (in other words, not on a prototype).
  function has(obj, path) {
    if (!isArray$1(path)) {
      return _has(obj, path);
    }
    var length = path.length;
    for (var i = 0; i < length; i++) {
      var key = path[i];
      if (obj == null || !hasOwnProperty$a.call(obj, key)) {
        return false;
      }
      obj = obj[key];
    }
    return !!length;
  }

  // Utility Functions
  // -----------------

  // Keep the identity function around for default iteratees.
  function identity(value) {
    return value;
  }

  // Predicate-generating functions. Often useful outside of Underscore.
  function constant(value) {
    return function() {
      return value;
    };
  }

  function noop(){}

  // Creates a function that, when passed an object, will traverse that object’s
  // properties down the given `path`, specified as an array of keys or indexes.
  function property(path) {
    if (!isArray$1(path)) {
      return shallowProperty(path);
    }
    return function(obj) {
      return deepGet(obj, path);
    };
  }

  // Generates a function for a given object that returns a given property.
  function propertyOf(obj) {
    if (obj == null) {
      return function(){};
    }
    return function(path) {
      return !isArray$1(path) ? obj[path] : deepGet(obj, path);
    };
  }

  // Returns a predicate for checking whether an object has a given set of
  // `key:value` pairs.
  function matcher(attrs) {
    attrs = extendOwn({}, attrs);
    return function(obj) {
      return isMatch(obj, attrs);
    };
  }

  // Run a function **n** times.
  function times(n, iteratee, context) {
    var accum = Array(Math.max(0, n));
    iteratee = optimizeCb(iteratee, context, 1);
    for (var i = 0; i < n; i++) accum[i] = iteratee(i);
    return accum;
  }

  // Return a random integer between min and max (inclusive).
  function random(min, max) {
    if (max == null) {
      max = min;
      min = 0;
    }
    return min + Math.floor(Math.random() * (max - min + 1));
  }

  // A (possibly faster) way to get the current timestamp as an integer.
  var now = Date.now || function() {
    return new Date().getTime();
  };

  // List of HTML entities for escaping.
  var escapeMap = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;',
    "'": '&#x27;',
    '`': '&#x60;'
  };
  var unescapeMap = invert(escapeMap);

  // Functions for escaping and unescaping strings to/from HTML interpolation.
  function createEscaper(map) {
    var escaper = function(match) {
      return map[match];
    };
    // Regexes for identifying a key that needs to be escaped.
    var source = '(?:' + keys$1(map).join('|') + ')';
    var testRegexp = RegExp(source);
    var replaceRegexp = RegExp(source, 'g');
    return function(string) {
      string = string == null ? '' : '' + string;
      return testRegexp.test(string) ? string.replace(replaceRegexp, escaper) : string;
    };
  }
  var escape = createEscaper(escapeMap);
  var unescape = createEscaper(unescapeMap);

  // Traverses the children of `obj` along `path`. If a child is a function, it
  // is invoked with its parent as context. Returns the value of the final
  // child, or `fallback` if any child is undefined.
  function result(obj, path, fallback) {
    if (!isArray$1(path)) path = [path];
    var length = path.length;
    if (!length) {
      return isFunction$1(fallback) ? fallback.call(obj) : fallback;
    }
    for (var i = 0; i < length; i++) {
      var prop = obj == null ? void 0 : obj[path[i]];
      if (prop === void 0) {
        prop = fallback;
        i = length; // Ensure we don't continue iterating.
      }
      obj = isFunction$1(prop) ? prop.call(obj) : prop;
    }
    return obj;
  }

  // Generate a unique integer id (unique within the entire client session).
  // Useful for temporary DOM ids.
  var idCounter = 0;
  function uniqueId(prefix) {
    var id = ++idCounter + '';
    return prefix ? prefix + id : id;
  }

  // By default, Underscore uses ERB-style template delimiters, change the
  // following template settings to use alternative delimiters.
  var templateSettings = _.templateSettings = {
    evaluate: /<%([\s\S]+?)%>/g,
    interpolate: /<%=([\s\S]+?)%>/g,
    escape: /<%-([\s\S]+?)%>/g
  };

  // When customizing `templateSettings`, if you don't want to define an
  // interpolation, evaluation or escaping regex, we need one that is
  // guaranteed not to match.
  var noMatch = /(.)^/;

  // Certain characters need to be escaped so that they can be put into a
  // string literal.
  var escapes = {
    "'": "'",
    '\\': '\\',
    '\r': 'r',
    '\n': 'n',
    '\u2028': 'u2028',
    '\u2029': 'u2029'
  };

  var escapeRegExp = /\\|'|\r|\n|\u2028|\u2029/g;

  var escapeChar = function(match) {
    return '\\' + escapes[match];
  };

  // JavaScript micro-templating, similar to John Resig's implementation.
  // Underscore templating handles arbitrary delimiters, preserves whitespace,
  // and correctly escapes quotes within interpolated code.
  // NB: `oldSettings` only exists for backwards compatibility.
  function template(text, settings, oldSettings) {
    if (!settings && oldSettings) settings = oldSettings;
    settings = defaults({}, settings, _.templateSettings);

    // Combine delimiters into one regular expression via alternation.
    var matcher = RegExp([
      (settings.escape || noMatch).source,
      (settings.interpolate || noMatch).source,
      (settings.evaluate || noMatch).source
    ].join('|') + '|$', 'g');

    // Compile the template source, escaping string literals appropriately.
    var index = 0;
    var source = "__p+='";
    text.replace(matcher, function(match, escape, interpolate, evaluate, offset) {
      source += text.slice(index, offset).replace(escapeRegExp, escapeChar);
      index = offset + match.length;

      if (escape) {
        source += "'+\n((__t=(" + escape + "))==null?'':_.escape(__t))+\n'";
      } else if (interpolate) {
        source += "'+\n((__t=(" + interpolate + "))==null?'':__t)+\n'";
      } else if (evaluate) {
        source += "';\n" + evaluate + "\n__p+='";
      }

      // Adobe VMs need the match returned to produce the correct offset.
      return match;
    });
    source += "';\n";

    // If a variable is not specified, place data values in local scope.
    if (!settings.variable) source = 'with(obj||{}){\n' + source + '}\n';

    source = "var __t,__p='',__j=Array.prototype.join," +
      "print=function(){__p+=__j.call(arguments,'');};\n" +
      source + 'return __p;\n';

    var render;
    try {
      render = new Function(settings.variable || 'obj', '_', source);
    } catch (e) {
      e.source = source;
      throw e;
    }

    var template = function(data) {
      return render.call(this, data, _);
    };

    // Provide the compiled source as a convenience for precompilation.
    var argument = settings.variable || 'obj';
    template.source = 'function(' + argument + '){\n' + source + '}';

    return template;
  }

  // Add a "chain" function. Start chaining a wrapped Underscore object.
  function chain(obj) {
    var instance = _(obj);
    instance._chain = true;
    return instance;
  }

  // OOP
  // ---------------
  // If Underscore is called as a function, it returns a wrapped object that
  // can be used OO-style. This wrapper holds altered versions of all the
  // underscore functions. Wrapped objects may be chained.

  // Helper function to continue chaining intermediate results.
  function chainResult(instance, obj) {
    return instance._chain ? _(obj).chain() : obj;
  }

  // Add your own custom functions to the Underscore object.
  function mixin(obj) {
    each(functions(obj), function(name) {
      var func = _[name] = obj[name];
      _.prototype[name] = function() {
        var args = [this._wrapped];
        push.apply(args, arguments);
        return chainResult(this, func.apply(_, args));
      };
    });
    return _;
  }

  // Add all mutator Array functions to the wrapper.
  each(['pop', 'push', 'reverse', 'shift', 'sort', 'splice', 'unshift'], function(name) {
    var method = ArrayProto[name];
    _.prototype[name] = function() {
      var obj = this._wrapped;
      method.apply(obj, arguments);
      if ((name === 'shift' || name === 'splice') && obj.length === 0) delete obj[0];
      return chainResult(this, obj);
    };
  });

  // Add all accessor Array functions to the wrapper.
  each(['concat', 'join', 'slice'], function(name) {
    var method = ArrayProto[name];
    _.prototype[name] = function() {
      return chainResult(this, method.apply(this._wrapped, arguments));
    };
  });

  // Extracts the result from a wrapped and chained object.
  _.prototype.value = function() {
    return this._wrapped;
  };

  // Provide unwrapping proxy for some methods used in engine operations
  // such as arithmetic and JSON stringification.
  _.prototype.valueOf = _.prototype.toJSON = _.prototype.value;

  _.prototype.toString = function() {
    return String(this._wrapped);
  };

  var allExports = /*#__PURE__*/Object.freeze({
    'default': _,
    VERSION: VERSION,
    iteratee: iteratee,
    restArguments: restArguments,
    each: each,
    forEach: each,
    map: map,
    collect: map,
    reduce: reduce,
    foldl: reduce,
    inject: reduce,
    reduceRight: reduceRight,
    foldr: reduceRight,
    find: find,
    detect: find,
    filter: filter,
    select: filter,
    reject: reject$1,
    every: every,
    all: every,
    some: some,
    any: some,
    contains: contains,
    includes: contains,
    include: contains,
    invoke: invoke,
    pluck: pluck,
    where: where,
    findWhere: findWhere,
    max: max,
    min: min,
    shuffle: shuffle,
    sample: sample,
    sortBy: sortBy,
    groupBy: groupBy,
    indexBy: indexBy,
    countBy: countBy,
    toArray: toArray,
    size: size,
    partition: partition,
    first: first,
    head: first,
    take: first,
    initial: initial,
    last: last,
    rest: rest,
    tail: rest,
    drop: rest,
    compact: compact,
    flatten: flatten,
    without: without,
    uniq: uniq,
    unique: uniq,
    union: union,
    intersection: intersection,
    difference: difference$1,
    unzip: unzip,
    zip: zip,
    object: object,
    findIndex: findIndex,
    findLastIndex: findLastIndex,
    sortedIndex: sortedIndex,
    indexOf: indexOf,
    lastIndexOf: lastIndexOf,
    range: range,
    chunk: chunk,
    bind: bind,
    partial: partial,
    bindAll: bindAll,
    memoize: memoize,
    delay: delay,
    defer: defer,
    throttle: throttle,
    debounce: debounce,
    wrap: wrap,
    negate: negate,
    compose: compose,
    after: after,
    before: before,
    once: once,
    keys: keys$1,
    allKeys: allKeys,
    values: values,
    mapObject: mapObject,
    pairs: pairs,
    invert: invert,
    functions: functions,
    methods: functions,
    extend: extend,
    extendOwn: extendOwn,
    assign: extendOwn,
    findKey: findKey,
    pick: pick,
    omit: omit,
    defaults: defaults,
    create: create,
    clone: clone,
    tap: tap,
    isMatch: isMatch,
    isEqual: isEqual$2,
    isEmpty: isEmpty,
    isElement: isElement,
    isArray: isArray$1,
    isObject: isObject$1,
    get isArguments () { return isArguments$1; },
    get isFunction () { return isFunction$1; },
    isString: isString,
    isNumber: isNumber,
    isDate: isDate,
    isRegExp: isRegExp,
    isError: isError,
    isSymbol: isSymbol,
    isMap: isMap,
    isWeakMap: isWeakMap,
    isSet: isSet,
    isWeakSet: isWeakSet,
    isFinite: isFinite,
    isNaN: isNaN$1,
    isBoolean: isBoolean,
    isNull: isNull,
    isUndefined: isUndefined,
    has: has,
    identity: identity,
    constant: constant,
    noop: noop,
    property: property,
    propertyOf: propertyOf,
    matcher: matcher,
    matches: matcher,
    times: times,
    random: random,
    now: now,
    escape: escape,
    unescape: unescape,
    result: result,
    uniqueId: uniqueId,
    templateSettings: templateSettings,
    template: template,
    chain: chain,
    mixin: mixin
  });

  // Add all of the Underscore functions to the wrapper object.
  var _$1 = mixin(allExports);
  // Legacy Node.js API
  _$1._ = _$1;

  //     Backbone.js 1.2.3

  //     (c) 2010-2015 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
  //     Backbone may be freely distributed under the MIT license.
  //     For all details and documentation:
  //     http://backbonejs.org

  (function(factory) {

    // Establish the root object, `window` (`self`) in the browser, or `global` on the server.
    // We use `self` instead of `window` for `WebWorker` support.
    var root = (typeof self == 'object' && self.self == self && self) ||
              (typeof global == 'object' && global.global == global && global);

    // Set up Backbone appropriately for the environment. Start with AMD.
    {
      var _ = _$1, $;
      try { $ = $$1; } catch(e) {}
      factory(root, exports, _, $);

    // Finally, as a browser global.
    }

  }(function(root, Backbone, _, $) {

    // Initial Setup
    // -------------

    // Save the previous value of the `Backbone` variable, so that it can be
    // restored later on, if `noConflict` is used.
    var previousBackbone = root.Backbone;

    // Create a local reference to a common array method we'll want to use later.
    var slice = Array.prototype.slice;

    // Current version of the library. Keep in sync with `package.json`.
    Backbone.VERSION = '1.2.3';

    // For Backbone's purposes, jQuery, Zepto, Ender, or My Library (kidding) owns
    // the `$` variable.
    Backbone.$ = $;

    // Runs Backbone.js in *noConflict* mode, returning the `Backbone` variable
    // to its previous owner. Returns a reference to this Backbone object.
    Backbone.noConflict = function() {
      root.Backbone = previousBackbone;
      return this;
    };

    // Turn on `emulateHTTP` to support legacy HTTP servers. Setting this option
    // will fake `"PATCH"`, `"PUT"` and `"DELETE"` requests via the `_method` parameter and
    // set a `X-Http-Method-Override` header.
    Backbone.emulateHTTP = false;

    // Turn on `emulateJSON` to support legacy servers that can't deal with direct
    // `application/json` requests ... this will encode the body as
    // `application/x-www-form-urlencoded` instead and will send the model in a
    // form param named `model`.
    Backbone.emulateJSON = false;

    // Proxy Backbone class methods to Underscore functions, wrapping the model's
    // `attributes` object or collection's `models` array behind the scenes.
    //
    // collection.filter(function(model) { return model.get('age') > 10 });
    // collection.each(this.addView);
    //
    // `Function#apply` can be slow so we use the method's arg count, if we know it.
    var addMethod = function(length, method, attribute) {
      switch (length) {
        case 1: return function() {
          return _[method](this[attribute]);
        };
        case 2: return function(value) {
          return _[method](this[attribute], value);
        };
        case 3: return function(iteratee, context) {
          return _[method](this[attribute], cb(iteratee, this), context);
        };
        case 4: return function(iteratee, defaultVal, context) {
          return _[method](this[attribute], cb(iteratee, this), defaultVal, context);
        };
        default: return function() {
          var args = slice.call(arguments);
          args.unshift(this[attribute]);
          return _[method].apply(_, args);
        };
      }
    };
    var addUnderscoreMethods = function(Class, methods, attribute) {
      _.each(methods, function(length, method) {
        if (_[method]) Class.prototype[method] = addMethod(length, method, attribute);
      });
    };

    // Support `collection.sortBy('attr')` and `collection.findWhere({id: 1})`.
    var cb = function(iteratee, instance) {
      if (_.isFunction(iteratee)) return iteratee;
      if (_.isObject(iteratee) && !instance._isModel(iteratee)) return modelMatcher(iteratee);
      if (_.isString(iteratee)) return function(model) { return model.get(iteratee); };
      return iteratee;
    };
    var modelMatcher = function(attrs) {
      var matcher = _.matches(attrs);
      return function(model) {
        return matcher(model.attributes);
      };
    };

    // Backbone.Events
    // ---------------

    // A module that can be mixed in to *any object* in order to provide it with
    // a custom event channel. You may bind a callback to an event with `on` or
    // remove with `off`; `trigger`-ing an event fires all callbacks in
    // succession.
    //
    //     var object = {};
    //     _.extend(object, Backbone.Events);
    //     object.on('expand', function(){ alert('expanded'); });
    //     object.trigger('expand');
    //
    var Events = Backbone.Events = {};

    // Regular expression used to split event strings.
    var eventSplitter = /\s+/;

    // Iterates over the standard `event, callback` (as well as the fancy multiple
    // space-separated events `"change blur", callback` and jQuery-style event
    // maps `{event: callback}`).
    var eventsApi = function(iteratee, events, name, callback, opts) {
      var i = 0, names;
      if (name && typeof name === 'object') {
        // Handle event maps.
        if (callback !== void 0 && 'context' in opts && opts.context === void 0) opts.context = callback;
        for (names = _.keys(name); i < names.length ; i++) {
          events = eventsApi(iteratee, events, names[i], name[names[i]], opts);
        }
      } else if (name && eventSplitter.test(name)) {
        // Handle space separated event names by delegating them individually.
        for (names = name.split(eventSplitter); i < names.length; i++) {
          events = iteratee(events, names[i], callback, opts);
        }
      } else {
        // Finally, standard events.
        events = iteratee(events, name, callback, opts);
      }
      return events;
    };

    // Bind an event to a `callback` function. Passing `"all"` will bind
    // the callback to all events fired.
    Events.on = function(name, callback, context) {
      return internalOn(this, name, callback, context);
    };

    // Guard the `listening` argument from the public API.
    var internalOn = function(obj, name, callback, context, listening) {
      obj._events = eventsApi(onApi, obj._events || {}, name, callback, {
          context: context,
          ctx: obj,
          listening: listening
      });

      if (listening) {
        var listeners = obj._listeners || (obj._listeners = {});
        listeners[listening.id] = listening;
      }

      return obj;
    };

    // Inversion-of-control versions of `on`. Tell *this* object to listen to
    // an event in another object... keeping track of what it's listening to
    // for easier unbinding later.
    Events.listenTo =  function(obj, name, callback) {
      if (!obj) return this;
      var id = obj._listenId || (obj._listenId = _.uniqueId('l'));
      var listeningTo = this._listeningTo || (this._listeningTo = {});
      var listening = listeningTo[id];

      // This object is not listening to any other events on `obj` yet.
      // Setup the necessary references to track the listening callbacks.
      if (!listening) {
        var thisId = this._listenId || (this._listenId = _.uniqueId('l'));
        listening = listeningTo[id] = {obj: obj, objId: id, id: thisId, listeningTo: listeningTo, count: 0};
      }

      // Bind callbacks on obj, and keep track of them on listening.
      internalOn(obj, name, callback, this, listening);
      return this;
    };

    // The reducing API that adds a callback to the `events` object.
    var onApi = function(events, name, callback, options) {
      if (callback) {
        var handlers = events[name] || (events[name] = []);
        var context = options.context, ctx = options.ctx, listening = options.listening;
        if (listening) listening.count++;

        handlers.push({ callback: callback, context: context, ctx: context || ctx, listening: listening });
      }
      return events;
    };

    // Remove one or many callbacks. If `context` is null, removes all
    // callbacks with that function. If `callback` is null, removes all
    // callbacks for the event. If `name` is null, removes all bound
    // callbacks for all events.
    Events.off =  function(name, callback, context) {
      if (!this._events) return this;
      this._events = eventsApi(offApi, this._events, name, callback, {
          context: context,
          listeners: this._listeners
      });
      return this;
    };

    // Tell this object to stop listening to either specific events ... or
    // to every object it's currently listening to.
    Events.stopListening =  function(obj, name, callback) {
      var listeningTo = this._listeningTo;
      if (!listeningTo) return this;

      var ids = obj ? [obj._listenId] : _.keys(listeningTo);

      for (var i = 0; i < ids.length; i++) {
        var listening = listeningTo[ids[i]];

        // If listening doesn't exist, this object is not currently
        // listening to obj. Break out early.
        if (!listening) break;

        listening.obj.off(name, callback, this);
      }
      if (_.isEmpty(listeningTo)) this._listeningTo = void 0;

      return this;
    };

    // The reducing API that removes a callback from the `events` object.
    var offApi = function(events, name, callback, options) {
      if (!events) return;

      var i = 0, listening;
      var context = options.context, listeners = options.listeners;

      // Delete all events listeners and "drop" events.
      if (!name && !callback && !context) {
        var ids = _.keys(listeners);
        for (; i < ids.length; i++) {
          listening = listeners[ids[i]];
          delete listeners[listening.id];
          delete listening.listeningTo[listening.objId];
        }
        return;
      }

      var names = name ? [name] : _.keys(events);
      for (; i < names.length; i++) {
        name = names[i];
        var handlers = events[name];

        // Bail out if there are no events stored.
        if (!handlers) break;

        // Replace events if there are any remaining.  Otherwise, clean up.
        var remaining = [];
        for (var j = 0; j < handlers.length; j++) {
          var handler = handlers[j];
          if (
            callback && callback !== handler.callback &&
              callback !== handler.callback._callback ||
                context && context !== handler.context
          ) {
            remaining.push(handler);
          } else {
            listening = handler.listening;
            if (listening && --listening.count === 0) {
              delete listeners[listening.id];
              delete listening.listeningTo[listening.objId];
            }
          }
        }

        // Update tail event if the list has any events.  Otherwise, clean up.
        if (remaining.length) {
          events[name] = remaining;
        } else {
          delete events[name];
        }
      }
      if (_.size(events)) return events;
    };

    // Bind an event to only be triggered a single time. After the first time
    // the callback is invoked, its listener will be removed. If multiple events
    // are passed in using the space-separated syntax, the handler will fire
    // once for each event, not once for a combination of all events.
    Events.once =  function(name, callback, context) {
      // Map the event into a `{event: once}` object.
      var events = eventsApi(onceMap, {}, name, callback, _.bind(this.off, this));
      return this.on(events, void 0, context);
    };

    // Inversion-of-control versions of `once`.
    Events.listenToOnce =  function(obj, name, callback) {
      // Map the event into a `{event: once}` object.
      var events = eventsApi(onceMap, {}, name, callback, _.bind(this.stopListening, this, obj));
      return this.listenTo(obj, events);
    };

    // Reduces the event callbacks into a map of `{event: onceWrapper}`.
    // `offer` unbinds the `onceWrapper` after it has been called.
    var onceMap = function(map, name, callback, offer) {
      if (callback) {
        var once = map[name] = _.once(function() {
          offer(name, once);
          callback.apply(this, arguments);
        });
        once._callback = callback;
      }
      return map;
    };

    // Trigger one or many events, firing all bound callbacks. Callbacks are
    // passed the same arguments as `trigger` is, apart from the event name
    // (unless you're listening on `"all"`, which will cause your callback to
    // receive the true name of the event as the first argument).
    Events.trigger =  function(name) {
      if (!this._events) return this;

      var length = Math.max(0, arguments.length - 1);
      var args = Array(length);
      for (var i = 0; i < length; i++) args[i] = arguments[i + 1];

      eventsApi(triggerApi, this._events, name, void 0, args);
      return this;
    };

    // Handles triggering the appropriate event callbacks.
    var triggerApi = function(objEvents, name, cb, args) {
      if (objEvents) {
        var events = objEvents[name];
        var allEvents = objEvents.all;
        if (events && allEvents) allEvents = allEvents.slice();
        if (events) triggerEvents(events, args);
        if (allEvents) triggerEvents(allEvents, [name].concat(args));
      }
      return objEvents;
    };

    // A difficult-to-believe, but optimized internal dispatch function for
    // triggering events. Tries to keep the usual cases speedy (most internal
    // Backbone events have 3 arguments).
    var triggerEvents = function(events, args) {
      var ev, i = -1, l = events.length, a1 = args[0], a2 = args[1], a3 = args[2];
      switch (args.length) {
        case 0: while (++i < l) (ev = events[i]).callback.call(ev.ctx); return;
        case 1: while (++i < l) (ev = events[i]).callback.call(ev.ctx, a1); return;
        case 2: while (++i < l) (ev = events[i]).callback.call(ev.ctx, a1, a2); return;
        case 3: while (++i < l) (ev = events[i]).callback.call(ev.ctx, a1, a2, a3); return;
        default: while (++i < l) (ev = events[i]).callback.apply(ev.ctx, args); return;
      }
    };

    // Aliases for backwards compatibility.
    Events.bind   = Events.on;
    Events.unbind = Events.off;

    // Allow the `Backbone` object to serve as a global event bus, for folks who
    // want global "pubsub" in a convenient place.
    _.extend(Backbone, Events);

    // Backbone.Model
    // --------------

    // Backbone **Models** are the basic data object in the framework --
    // frequently representing a row in a table in a database on your server.
    // A discrete chunk of data and a bunch of useful, related methods for
    // performing computations and transformations on that data.

    // Create a new model with the specified attributes. A client id (`cid`)
    // is automatically generated and assigned for you.
    var Model = Backbone.Model = function(attributes, options) {
      var attrs = attributes || {};
      options || (options = {});
      this.cid = _.uniqueId(this.cidPrefix);
      this.attributes = {};
      if (options.collection) this.collection = options.collection;
      if (options.parse) attrs = this.parse(attrs, options) || {};
      attrs = _.defaults({}, attrs, _.result(this, 'defaults'));
      this.set(attrs, options);
      this.changed = {};
      this.initialize.apply(this, arguments);
    };

    // Attach all inheritable methods to the Model prototype.
    _.extend(Model.prototype, Events, {

      // A hash of attributes whose current and previous value differ.
      changed: null,

      // The value returned during the last failed validation.
      validationError: null,

      // The default name for the JSON `id` attribute is `"id"`. MongoDB and
      // CouchDB users may want to set this to `"_id"`.
      idAttribute: 'id',

      // The prefix is used to create the client id which is used to identify models locally.
      // You may want to override this if you're experiencing name clashes with model ids.
      cidPrefix: 'c',

      // Initialize is an empty function by default. Override it with your own
      // initialization logic.
      initialize: function(){},

      // Return a copy of the model's `attributes` object.
      toJSON: function(options) {
        return _.clone(this.attributes);
      },

      // Proxy `Backbone.sync` by default -- but override this if you need
      // custom syncing semantics for *this* particular model.
      sync: function() {
        return Backbone.sync.apply(this, arguments);
      },

      // Get the value of an attribute.
      get: function(attr) {
        return this.attributes[attr];
      },

      // Get the HTML-escaped value of an attribute.
      escape: function(attr) {
        return _.escape(this.get(attr));
      },

      // Returns `true` if the attribute contains a value that is not null
      // or undefined.
      has: function(attr) {
        return this.get(attr) != null;
      },

      // Special-cased proxy to underscore's `_.matches` method.
      matches: function(attrs) {
        return !!_.iteratee(attrs, this)(this.attributes);
      },

      // Set a hash of model attributes on the object, firing `"change"`. This is
      // the core primitive operation of a model, updating the data and notifying
      // anyone who needs to know about the change in state. The heart of the beast.
      set: function(key, val, options) {
        if (key == null) return this;

        // Handle both `"key", value` and `{key: value}` -style arguments.
        var attrs;
        if (typeof key === 'object') {
          attrs = key;
          options = val;
        } else {
          (attrs = {})[key] = val;
        }

        options || (options = {});

        // Run validation.
        if (!this._validate(attrs, options)) return false;

        // Extract attributes and options.
        var unset      = options.unset;
        var silent     = options.silent;
        var changes    = [];
        var changing   = this._changing;
        this._changing = true;

        if (!changing) {
          this._previousAttributes = _.clone(this.attributes);
          this.changed = {};
        }

        var current = this.attributes;
        var changed = this.changed;
        var prev    = this._previousAttributes;

        // For each `set` attribute, update or delete the current value.
        for (var attr in attrs) {
          val = attrs[attr];
          if (!_.isEqual(current[attr], val)) changes.push(attr);
          if (!_.isEqual(prev[attr], val)) {
            changed[attr] = val;
          } else {
            delete changed[attr];
          }
          unset ? delete current[attr] : current[attr] = val;
        }

        // Update the `id`.
        this.id = this.get(this.idAttribute);

        // Trigger all relevant attribute changes.
        if (!silent) {
          if (changes.length) this._pending = options;
          for (var i = 0; i < changes.length; i++) {
            this.trigger('change:' + changes[i], this, current[changes[i]], options);
          }
        }

        // You might be wondering why there's a `while` loop here. Changes can
        // be recursively nested within `"change"` events.
        if (changing) return this;
        if (!silent) {
          while (this._pending) {
            options = this._pending;
            this._pending = false;
            this.trigger('change', this, options);
          }
        }
        this._pending = false;
        this._changing = false;
        return this;
      },

      // Remove an attribute from the model, firing `"change"`. `unset` is a noop
      // if the attribute doesn't exist.
      unset: function(attr, options) {
        return this.set(attr, void 0, _.extend({}, options, {unset: true}));
      },

      // Clear all attributes on the model, firing `"change"`.
      clear: function(options) {
        var attrs = {};
        for (var key in this.attributes) attrs[key] = void 0;
        return this.set(attrs, _.extend({}, options, {unset: true}));
      },

      // Determine if the model has changed since the last `"change"` event.
      // If you specify an attribute name, determine if that attribute has changed.
      hasChanged: function(attr) {
        if (attr == null) return !_.isEmpty(this.changed);
        return _.has(this.changed, attr);
      },

      // Return an object containing all the attributes that have changed, or
      // false if there are no changed attributes. Useful for determining what
      // parts of a view need to be updated and/or what attributes need to be
      // persisted to the server. Unset attributes will be set to undefined.
      // You can also pass an attributes object to diff against the model,
      // determining if there *would be* a change.
      changedAttributes: function(diff) {
        if (!diff) return this.hasChanged() ? _.clone(this.changed) : false;
        var old = this._changing ? this._previousAttributes : this.attributes;
        var changed = {};
        for (var attr in diff) {
          var val = diff[attr];
          if (_.isEqual(old[attr], val)) continue;
          changed[attr] = val;
        }
        return _.size(changed) ? changed : false;
      },

      // Get the previous value of an attribute, recorded at the time the last
      // `"change"` event was fired.
      previous: function(attr) {
        if (attr == null || !this._previousAttributes) return null;
        return this._previousAttributes[attr];
      },

      // Get all of the attributes of the model at the time of the previous
      // `"change"` event.
      previousAttributes: function() {
        return _.clone(this._previousAttributes);
      },

      // Fetch the model from the server, merging the response with the model's
      // local attributes. Any changed attributes will trigger a "change" event.
      fetch: function(options) {
        options = _.extend({parse: true}, options);
        var model = this;
        var success = options.success;
        options.success = function(resp) {
          var serverAttrs = options.parse ? model.parse(resp, options) : resp;
          if (!model.set(serverAttrs, options)) return false;
          if (success) success.call(options.context, model, resp, options);
          model.trigger('sync', model, resp, options);
        };
        wrapError(this, options);
        return this.sync('read', this, options);
      },

      // Set a hash of model attributes, and sync the model to the server.
      // If the server returns an attributes hash that differs, the model's
      // state will be `set` again.
      save: function(key, val, options) {
        // Handle both `"key", value` and `{key: value}` -style arguments.
        var attrs;
        if (key == null || typeof key === 'object') {
          attrs = key;
          options = val;
        } else {
          (attrs = {})[key] = val;
        }

        options = _.extend({validate: true, parse: true}, options);
        var wait = options.wait;

        // If we're not waiting and attributes exist, save acts as
        // `set(attr).save(null, opts)` with validation. Otherwise, check if
        // the model will be valid when the attributes, if any, are set.
        if (attrs && !wait) {
          if (!this.set(attrs, options)) return false;
        } else {
          if (!this._validate(attrs, options)) return false;
        }

        // After a successful server-side save, the client is (optionally)
        // updated with the server-side state.
        var model = this;
        var success = options.success;
        var attributes = this.attributes;
        options.success = function(resp) {
          // Ensure attributes are restored during synchronous saves.
          model.attributes = attributes;
          var serverAttrs = options.parse ? model.parse(resp, options) : resp;
          if (wait) serverAttrs = _.extend({}, attrs, serverAttrs);
          if (serverAttrs && !model.set(serverAttrs, options)) return false;
          if (success) success.call(options.context, model, resp, options);
          model.trigger('sync', model, resp, options);
        };
        wrapError(this, options);

        // Set temporary attributes if `{wait: true}` to properly find new ids.
        if (attrs && wait) this.attributes = _.extend({}, attributes, attrs);

        var method = this.isNew() ? 'create' : (options.patch ? 'patch' : 'update');
        if (method === 'patch' && !options.attrs) options.attrs = attrs;
        var xhr = this.sync(method, this, options);

        // Restore attributes.
        this.attributes = attributes;

        return xhr;
      },

      // Destroy this model on the server if it was already persisted.
      // Optimistically removes the model from its collection, if it has one.
      // If `wait: true` is passed, waits for the server to respond before removal.
      destroy: function(options) {
        options = options ? _.clone(options) : {};
        var model = this;
        var success = options.success;
        var wait = options.wait;

        var destroy = function() {
          model.stopListening();
          model.trigger('destroy', model, model.collection, options);
        };

        options.success = function(resp) {
          if (wait) destroy();
          if (success) success.call(options.context, model, resp, options);
          if (!model.isNew()) model.trigger('sync', model, resp, options);
        };

        var xhr = false;
        if (this.isNew()) {
          _.defer(options.success);
        } else {
          wrapError(this, options);
          xhr = this.sync('delete', this, options);
        }
        if (!wait) destroy();
        return xhr;
      },

      // Default URL for the model's representation on the server -- if you're
      // using Backbone's restful methods, override this to change the endpoint
      // that will be called.
      url: function() {
        var base =
          _.result(this, 'urlRoot') ||
          _.result(this.collection, 'url') ||
          urlError();
        if (this.isNew()) return base;
        var id = this.get(this.idAttribute);
        return base.replace(/[^\/]$/, '$&/') + encodeURIComponent(id);
      },

      // **parse** converts a response into the hash of attributes to be `set` on
      // the model. The default implementation is just to pass the response along.
      parse: function(resp, options) {
        return resp;
      },

      // Create a new model with identical attributes to this one.
      clone: function() {
        return new this.constructor(this.attributes);
      },

      // A model is new if it has never been saved to the server, and lacks an id.
      isNew: function() {
        return !this.has(this.idAttribute);
      },

      // Check if the model is currently in a valid state.
      isValid: function(options) {
        return this._validate({}, _.defaults({validate: true}, options));
      },

      // Run validation against the next complete set of model attributes,
      // returning `true` if all is well. Otherwise, fire an `"invalid"` event.
      _validate: function(attrs, options) {
        if (!options.validate || !this.validate) return true;
        attrs = _.extend({}, this.attributes, attrs);
        var error = this.validationError = this.validate(attrs, options) || null;
        if (!error) return true;
        this.trigger('invalid', this, error, _.extend(options, {validationError: error}));
        return false;
      }

    });

    // Underscore methods that we want to implement on the Model, mapped to the
    // number of arguments they take.
    var modelMethods = { keys: 1, values: 1, pairs: 1, invert: 1, pick: 0,
        omit: 0, chain: 1, isEmpty: 1 };

    // Mix in each Underscore method as a proxy to `Model#attributes`.
    addUnderscoreMethods(Model, modelMethods, 'attributes');

    // Backbone.Collection
    // -------------------

    // If models tend to represent a single row of data, a Backbone Collection is
    // more analogous to a table full of data ... or a small slice or page of that
    // table, or a collection of rows that belong together for a particular reason
    // -- all of the messages in this particular folder, all of the documents
    // belonging to this particular author, and so on. Collections maintain
    // indexes of their models, both in order, and for lookup by `id`.

    // Create a new **Collection**, perhaps to contain a specific type of `model`.
    // If a `comparator` is specified, the Collection will maintain
    // its models in sort order, as they're added and removed.
    var Collection = Backbone.Collection = function(models, options) {
      options || (options = {});
      if (options.model) this.model = options.model;
      if (options.comparator !== void 0) this.comparator = options.comparator;
      this._reset();
      this.initialize.apply(this, arguments);
      if (models) this.reset(models, _.extend({silent: true}, options));
    };

    // Default options for `Collection#set`.
    var setOptions = {add: true, remove: true, merge: true};
    var addOptions = {add: true, remove: false};

    // Splices `insert` into `array` at index `at`.
    var splice = function(array, insert, at) {
      at = Math.min(Math.max(at, 0), array.length);
      var tail = Array(array.length - at);
      var length = insert.length;
      for (var i = 0; i < tail.length; i++) tail[i] = array[i + at];
      for (i = 0; i < length; i++) array[i + at] = insert[i];
      for (i = 0; i < tail.length; i++) array[i + length + at] = tail[i];
    };

    // Define the Collection's inheritable methods.
    _.extend(Collection.prototype, Events, {

      // The default model for a collection is just a **Backbone.Model**.
      // This should be overridden in most cases.
      model: Model,

      // Initialize is an empty function by default. Override it with your own
      // initialization logic.
      initialize: function(){},

      // The JSON representation of a Collection is an array of the
      // models' attributes.
      toJSON: function(options) {
        return this.map(function(model) { return model.toJSON(options); });
      },

      // Proxy `Backbone.sync` by default.
      sync: function() {
        return Backbone.sync.apply(this, arguments);
      },

      // Add a model, or list of models to the set. `models` may be Backbone
      // Models or raw JavaScript objects to be converted to Models, or any
      // combination of the two.
      add: function(models, options) {
        return this.set(models, _.extend({merge: false}, options, addOptions));
      },

      // Remove a model, or a list of models from the set.
      remove: function(models, options) {
        options = _.extend({}, options);
        var singular = !_.isArray(models);
        models = singular ? [models] : _.clone(models);
        var removed = this._removeModels(models, options);
        if (!options.silent && removed) this.trigger('update', this, options);
        return singular ? removed[0] : removed;
      },

      // Update a collection by `set`-ing a new list of models, adding new ones,
      // removing models that are no longer present, and merging models that
      // already exist in the collection, as necessary. Similar to **Model#set**,
      // the core operation for updating the data contained by the collection.
      set: function(models, options) {
        if (models == null) return;

        options = _.defaults({}, options, setOptions);
        if (options.parse && !this._isModel(models)) models = this.parse(models, options);

        var singular = !_.isArray(models);
        models = singular ? [models] : models.slice();

        var at = options.at;
        if (at != null) at = +at;
        if (at < 0) at += this.length + 1;

        var set = [];
        var toAdd = [];
        var toRemove = [];
        var modelMap = {};

        var add = options.add;
        var merge = options.merge;
        var remove = options.remove;

        var sort = false;
        var sortable = this.comparator && (at == null) && options.sort !== false;
        var sortAttr = _.isString(this.comparator) ? this.comparator : null;

        // Turn bare objects into model references, and prevent invalid models
        // from being added.
        var model;
        for (var i = 0; i < models.length; i++) {
          model = models[i];

          // If a duplicate is found, prevent it from being added and
          // optionally merge it into the existing model.
          var existing = this.get(model);
          if (existing) {
            if (merge && model !== existing) {
              var attrs = this._isModel(model) ? model.attributes : model;
              if (options.parse) attrs = existing.parse(attrs, options);
              existing.set(attrs, options);
              if (sortable && !sort) sort = existing.hasChanged(sortAttr);
            }
            if (!modelMap[existing.cid]) {
              modelMap[existing.cid] = true;
              set.push(existing);
            }
            models[i] = existing;

          // If this is a new, valid model, push it to the `toAdd` list.
          } else if (add) {
            model = models[i] = this._prepareModel(model, options);
            if (model) {
              toAdd.push(model);
              this._addReference(model, options);
              modelMap[model.cid] = true;
              set.push(model);
            }
          }
        }

        // Remove stale models.
        if (remove) {
          for (i = 0; i < this.length; i++) {
            model = this.models[i];
            if (!modelMap[model.cid]) toRemove.push(model);
          }
          if (toRemove.length) this._removeModels(toRemove, options);
        }

        // See if sorting is needed, update `length` and splice in new models.
        var orderChanged = false;
        var replace = !sortable && add && remove;
        if (set.length && replace) {
          orderChanged = this.length != set.length || _.some(this.models, function(model, index) {
            return model !== set[index];
          });
          this.models.length = 0;
          splice(this.models, set, 0);
          this.length = this.models.length;
        } else if (toAdd.length) {
          if (sortable) sort = true;
          splice(this.models, toAdd, at == null ? this.length : at);
          this.length = this.models.length;
        }

        // Silently sort the collection if appropriate.
        if (sort) this.sort({silent: true});

        // Unless silenced, it's time to fire all appropriate add/sort events.
        if (!options.silent) {
          for (i = 0; i < toAdd.length; i++) {
            if (at != null) options.index = at + i;
            model = toAdd[i];
            model.trigger('add', model, this, options);
          }
          if (sort || orderChanged) this.trigger('sort', this, options);
          if (toAdd.length || toRemove.length) this.trigger('update', this, options);
        }

        // Return the added (or merged) model (or models).
        return singular ? models[0] : models;
      },

      // When you have more items than you want to add or remove individually,
      // you can reset the entire set with a new list of models, without firing
      // any granular `add` or `remove` events. Fires `reset` when finished.
      // Useful for bulk operations and optimizations.
      reset: function(models, options) {
        options = options ? _.clone(options) : {};
        for (var i = 0; i < this.models.length; i++) {
          this._removeReference(this.models[i], options);
        }
        options.previousModels = this.models;
        this._reset();
        models = this.add(models, _.extend({silent: true}, options));
        if (!options.silent) this.trigger('reset', this, options);
        return models;
      },

      // Add a model to the end of the collection.
      push: function(model, options) {
        return this.add(model, _.extend({at: this.length}, options));
      },

      // Remove a model from the end of the collection.
      pop: function(options) {
        var model = this.at(this.length - 1);
        return this.remove(model, options);
      },

      // Add a model to the beginning of the collection.
      unshift: function(model, options) {
        return this.add(model, _.extend({at: 0}, options));
      },

      // Remove a model from the beginning of the collection.
      shift: function(options) {
        var model = this.at(0);
        return this.remove(model, options);
      },

      // Slice out a sub-array of models from the collection.
      slice: function() {
        return slice.apply(this.models, arguments);
      },

      // Get a model from the set by id.
      get: function(obj) {
        if (obj == null) return void 0;
        var id = this.modelId(this._isModel(obj) ? obj.attributes : obj);
        return this._byId[obj] || this._byId[id] || this._byId[obj.cid];
      },

      // Get the model at the given index.
      at: function(index) {
        if (index < 0) index += this.length;
        return this.models[index];
      },

      // Return models with matching attributes. Useful for simple cases of
      // `filter`.
      where: function(attrs, first) {
        return this[first ? 'find' : 'filter'](attrs);
      },

      // Return the first model with matching attributes. Useful for simple cases
      // of `find`.
      findWhere: function(attrs) {
        return this.where(attrs, true);
      },

      // Force the collection to re-sort itself. You don't need to call this under
      // normal circumstances, as the set will maintain sort order as each item
      // is added.
      sort: function(options) {
        var comparator = this.comparator;
        if (!comparator) throw new Error('Cannot sort a set without a comparator');
        options || (options = {});

        var length = comparator.length;
        if (_.isFunction(comparator)) comparator = _.bind(comparator, this);

        // Run sort based on type of `comparator`.
        if (length === 1 || _.isString(comparator)) {
          this.models = this.sortBy(comparator);
        } else {
          this.models.sort(comparator);
        }
        if (!options.silent) this.trigger('sort', this, options);
        return this;
      },

      // Pluck an attribute from each model in the collection.
      pluck: function(attr) {
        return _.invoke(this.models, 'get', attr);
      },

      // Fetch the default set of models for this collection, resetting the
      // collection when they arrive. If `reset: true` is passed, the response
      // data will be passed through the `reset` method instead of `set`.
      fetch: function(options) {
        options = _.extend({parse: true}, options);
        var success = options.success;
        var collection = this;
        options.success = function(resp) {
          var method = options.reset ? 'reset' : 'set';
          collection[method](resp, options);
          if (success) success.call(options.context, collection, resp, options);
          collection.trigger('sync', collection, resp, options);
        };
        wrapError(this, options);
        return this.sync('read', this, options);
      },

      // Create a new instance of a model in this collection. Add the model to the
      // collection immediately, unless `wait: true` is passed, in which case we
      // wait for the server to agree.
      create: function(model, options) {
        options = options ? _.clone(options) : {};
        var wait = options.wait;
        model = this._prepareModel(model, options);
        if (!model) return false;
        if (!wait) this.add(model, options);
        var collection = this;
        var success = options.success;
        options.success = function(model, resp, callbackOpts) {
          if (wait) collection.add(model, callbackOpts);
          if (success) success.call(callbackOpts.context, model, resp, callbackOpts);
        };
        model.save(null, options);
        return model;
      },

      // **parse** converts a response into a list of models to be added to the
      // collection. The default implementation is just to pass it through.
      parse: function(resp, options) {
        return resp;
      },

      // Create a new collection with an identical list of models as this one.
      clone: function() {
        return new this.constructor(this.models, {
          model: this.model,
          comparator: this.comparator
        });
      },

      // Define how to uniquely identify models in the collection.
      modelId: function (attrs) {
        return attrs[this.model.prototype.idAttribute || 'id'];
      },

      // Private method to reset all internal state. Called when the collection
      // is first initialized or reset.
      _reset: function() {
        this.length = 0;
        this.models = [];
        this._byId  = {};
      },

      // Prepare a hash of attributes (or other model) to be added to this
      // collection.
      _prepareModel: function(attrs, options) {
        if (this._isModel(attrs)) {
          if (!attrs.collection) attrs.collection = this;
          return attrs;
        }
        options = options ? _.clone(options) : {};
        options.collection = this;
        var model = new this.model(attrs, options);
        if (!model.validationError) return model;
        this.trigger('invalid', this, model.validationError, options);
        return false;
      },

      // Internal method called by both remove and set.
      _removeModels: function(models, options) {
        var removed = [];
        for (var i = 0; i < models.length; i++) {
          var model = this.get(models[i]);
          if (!model) continue;

          var index = this.indexOf(model);
          this.models.splice(index, 1);
          this.length--;

          if (!options.silent) {
            options.index = index;
            model.trigger('remove', model, this, options);
          }

          removed.push(model);
          this._removeReference(model, options);
        }
        return removed.length ? removed : false;
      },

      // Method for checking whether an object should be considered a model for
      // the purposes of adding to the collection.
      _isModel: function (model) {
        return model instanceof Model;
      },

      // Internal method to create a model's ties to a collection.
      _addReference: function(model, options) {
        this._byId[model.cid] = model;
        var id = this.modelId(model.attributes);
        if (id != null) this._byId[id] = model;
        model.on('all', this._onModelEvent, this);
      },

      // Internal method to sever a model's ties to a collection.
      _removeReference: function(model, options) {
        delete this._byId[model.cid];
        var id = this.modelId(model.attributes);
        if (id != null) delete this._byId[id];
        if (this === model.collection) delete model.collection;
        model.off('all', this._onModelEvent, this);
      },

      // Internal method called every time a model in the set fires an event.
      // Sets need to update their indexes when models change ids. All other
      // events simply proxy through. "add" and "remove" events that originate
      // in other collections are ignored.
      _onModelEvent: function(event, model, collection, options) {
        if ((event === 'add' || event === 'remove') && collection !== this) return;
        if (event === 'destroy') this.remove(model, options);
        if (event === 'change') {
          var prevId = this.modelId(model.previousAttributes());
          var id = this.modelId(model.attributes);
          if (prevId !== id) {
            if (prevId != null) delete this._byId[prevId];
            if (id != null) this._byId[id] = model;
          }
        }
        this.trigger.apply(this, arguments);
      }

    });

    // Underscore methods that we want to implement on the Collection.
    // 90% of the core usefulness of Backbone Collections is actually implemented
    // right here:
    var collectionMethods = { forEach: 3, each: 3, map: 3, collect: 3, reduce: 4,
        foldl: 4, inject: 4, reduceRight: 4, foldr: 4, find: 3, detect: 3, filter: 3,
        select: 3, reject: 3, every: 3, all: 3, some: 3, any: 3, include: 3, includes: 3,
        contains: 3, invoke: 0, max: 3, min: 3, toArray: 1, size: 1, first: 3,
        head: 3, take: 3, initial: 3, rest: 3, tail: 3, drop: 3, last: 3,
        without: 0, difference: 0, indexOf: 3, shuffle: 1, lastIndexOf: 3,
        isEmpty: 1, chain: 1, sample: 3, partition: 3, groupBy: 3, countBy: 3,
        sortBy: 3, indexBy: 3};

    // Mix in each Underscore method as a proxy to `Collection#models`.
    addUnderscoreMethods(Collection, collectionMethods, 'models');

    // Backbone.View
    // -------------

    // Backbone Views are almost more convention than they are actual code. A View
    // is simply a JavaScript object that represents a logical chunk of UI in the
    // DOM. This might be a single item, an entire list, a sidebar or panel, or
    // even the surrounding frame which wraps your whole app. Defining a chunk of
    // UI as a **View** allows you to define your DOM events declaratively, without
    // having to worry about render order ... and makes it easy for the view to
    // react to specific changes in the state of your models.

    // Creating a Backbone.View creates its initial element outside of the DOM,
    // if an existing element is not provided...
    var View = Backbone.View = function(options) {
      this.cid = _.uniqueId('view');
      _.extend(this, _.pick(options, viewOptions));
      this._ensureElement();
      this.initialize.apply(this, arguments);
    };

    // Cached regex to split keys for `delegate`.
    var delegateEventSplitter = /^(\S+)\s*(.*)$/;

    // List of view options to be set as properties.
    var viewOptions = ['model', 'collection', 'el', 'id', 'attributes', 'className', 'tagName', 'events'];

    // Set up all inheritable **Backbone.View** properties and methods.
    _.extend(View.prototype, Events, {

      // The default `tagName` of a View's element is `"div"`.
      tagName: 'div',

      // jQuery delegate for element lookup, scoped to DOM elements within the
      // current view. This should be preferred to global lookups where possible.
      $: function(selector) {
        return this.$el.find(selector);
      },

      // Initialize is an empty function by default. Override it with your own
      // initialization logic.
      initialize: function(){},

      // **render** is the core function that your view should override, in order
      // to populate its element (`this.el`), with the appropriate HTML. The
      // convention is for **render** to always return `this`.
      render: function() {
        return this;
      },

      // Remove this view by taking the element out of the DOM, and removing any
      // applicable Backbone.Events listeners.
      remove: function() {
        this._removeElement();
        this.stopListening();
        return this;
      },

      // Remove this view's element from the document and all event listeners
      // attached to it. Exposed for subclasses using an alternative DOM
      // manipulation API.
      _removeElement: function() {
        this.$el.remove();
      },

      // Change the view's element (`this.el` property) and re-delegate the
      // view's events on the new element.
      setElement: function(element) {
        this.undelegateEvents();
        this._setElement(element);
        this.delegateEvents();
        return this;
      },

      // Creates the `this.el` and `this.$el` references for this view using the
      // given `el`. `el` can be a CSS selector or an HTML string, a jQuery
      // context or an element. Subclasses can override this to utilize an
      // alternative DOM manipulation API and are only required to set the
      // `this.el` property.
      _setElement: function(el) {
        this.$el = el instanceof Backbone.$ ? el : Backbone.$(el);
        this.el = this.$el[0];
      },

      // Set callbacks, where `this.events` is a hash of
      //
      // *{"event selector": "callback"}*
      //
      //     {
      //       'mousedown .title':  'edit',
      //       'click .button':     'save',
      //       'click .open':       function(e) { ... }
      //     }
      //
      // pairs. Callbacks will be bound to the view, with `this` set properly.
      // Uses event delegation for efficiency.
      // Omitting the selector binds the event to `this.el`.
      delegateEvents: function(events) {
        events || (events = _.result(this, 'events'));
        if (!events) return this;
        this.undelegateEvents();
        for (var key in events) {
          var method = events[key];
          if (!_.isFunction(method)) method = this[method];
          if (!method) continue;
          var match = key.match(delegateEventSplitter);
          this.delegate(match[1], match[2], _.bind(method, this));
        }
        return this;
      },

      // Add a single event listener to the view's element (or a child element
      // using `selector`). This only works for delegate-able events: not `focus`,
      // `blur`, and not `change`, `submit`, and `reset` in Internet Explorer.
      delegate: function(eventName, selector, listener) {
        this.$el.on(eventName + '.delegateEvents' + this.cid, selector, listener);
        return this;
      },

      // Clears all callbacks previously bound to the view by `delegateEvents`.
      // You usually don't need to use this, but may wish to if you have multiple
      // Backbone views attached to the same DOM element.
      undelegateEvents: function() {
        if (this.$el) this.$el.off('.delegateEvents' + this.cid);
        return this;
      },

      // A finer-grained `undelegateEvents` for removing a single delegated event.
      // `selector` and `listener` are both optional.
      undelegate: function(eventName, selector, listener) {
        this.$el.off(eventName + '.delegateEvents' + this.cid, selector, listener);
        return this;
      },

      // Produces a DOM element to be assigned to your view. Exposed for
      // subclasses using an alternative DOM manipulation API.
      _createElement: function(tagName) {
        return document.createElement(tagName);
      },

      // Ensure that the View has a DOM element to render into.
      // If `this.el` is a string, pass it through `$()`, take the first
      // matching element, and re-assign it to `el`. Otherwise, create
      // an element from the `id`, `className` and `tagName` properties.
      _ensureElement: function() {
        if (!this.el) {
          var attrs = _.extend({}, _.result(this, 'attributes'));
          if (this.id) attrs.id = _.result(this, 'id');
          if (this.className) attrs['class'] = _.result(this, 'className');
          this.setElement(this._createElement(_.result(this, 'tagName')));
          this._setAttributes(attrs);
        } else {
          this.setElement(_.result(this, 'el'));
        }
      },

      // Set attributes from a hash on this view's element.  Exposed for
      // subclasses using an alternative DOM manipulation API.
      _setAttributes: function(attributes) {
        this.$el.attr(attributes);
      }

    });

    // Backbone.sync
    // -------------

    // Override this function to change the manner in which Backbone persists
    // models to the server. You will be passed the type of request, and the
    // model in question. By default, makes a RESTful Ajax request
    // to the model's `url()`. Some possible customizations could be:
    //
    // * Use `setTimeout` to batch rapid-fire updates into a single request.
    // * Send up the models as XML instead of JSON.
    // * Persist models via WebSockets instead of Ajax.
    //
    // Turn on `Backbone.emulateHTTP` in order to send `PUT` and `DELETE` requests
    // as `POST`, with a `_method` parameter containing the true HTTP method,
    // as well as all requests with the body as `application/x-www-form-urlencoded`
    // instead of `application/json` with the model in a param named `model`.
    // Useful when interfacing with server-side languages like **PHP** that make
    // it difficult to read the body of `PUT` requests.
    Backbone.sync = function(method, model, options) {
      var type = methodMap[method];

      // Default options, unless specified.
      _.defaults(options || (options = {}), {
        emulateHTTP: Backbone.emulateHTTP,
        emulateJSON: Backbone.emulateJSON
      });

      // Default JSON-request options.
      var params = {type: type, dataType: 'json'};

      // Ensure that we have a URL.
      if (!options.url) {
        params.url = _.result(model, 'url') || urlError();
      }

      // Ensure that we have the appropriate request data.
      if (options.data == null && model && (method === 'create' || method === 'update' || method === 'patch')) {
        params.contentType = 'application/json';
        params.data = JSON.stringify(options.attrs || model.toJSON(options));
      }

      // For older servers, emulate JSON by encoding the request into an HTML-form.
      if (options.emulateJSON) {
        params.contentType = 'application/x-www-form-urlencoded';
        params.data = params.data ? {model: params.data} : {};
      }

      // For older servers, emulate HTTP by mimicking the HTTP method with `_method`
      // And an `X-HTTP-Method-Override` header.
      if (options.emulateHTTP && (type === 'PUT' || type === 'DELETE' || type === 'PATCH')) {
        params.type = 'POST';
        if (options.emulateJSON) params.data._method = type;
        var beforeSend = options.beforeSend;
        options.beforeSend = function(xhr) {
          xhr.setRequestHeader('X-HTTP-Method-Override', type);
          if (beforeSend) return beforeSend.apply(this, arguments);
        };
      }

      // Don't process data on a non-GET request.
      if (params.type !== 'GET' && !options.emulateJSON) {
        params.processData = false;
      }

      // Pass along `textStatus` and `errorThrown` from jQuery.
      var error = options.error;
      options.error = function(xhr, textStatus, errorThrown) {
        options.textStatus = textStatus;
        options.errorThrown = errorThrown;
        if (error) error.call(options.context, xhr, textStatus, errorThrown);
      };

      // Make the request, allowing the user to override any Ajax options.
      var xhr = options.xhr = Backbone.ajax(_.extend(params, options));
      model.trigger('request', model, xhr, options);
      return xhr;
    };

    // Map from CRUD to HTTP for our default `Backbone.sync` implementation.
    var methodMap = {
      'create': 'POST',
      'update': 'PUT',
      'patch':  'PATCH',
      'delete': 'DELETE',
      'read':   'GET'
    };

    // Set the default implementation of `Backbone.ajax` to proxy through to `$`.
    // Override this if you'd like to use a different library.
    Backbone.ajax = function() {
      return Backbone.$.ajax.apply(Backbone.$, arguments);
    };

    // Backbone.Router
    // ---------------

    // Routers map faux-URLs to actions, and fire events when routes are
    // matched. Creating a new one sets its `routes` hash, if not set statically.
    var Router = Backbone.Router = function(options) {
      options || (options = {});
      if (options.routes) this.routes = options.routes;
      this._bindRoutes();
      this.initialize.apply(this, arguments);
    };

    // Cached regular expressions for matching named param parts and splatted
    // parts of route strings.
    var optionalParam = /\((.*?)\)/g;
    var namedParam    = /(\(\?)?:\w+/g;
    var splatParam    = /\*\w+/g;
    var escapeRegExp  = /[\-{}\[\]+?.,\\\^$|#\s]/g;

    // Set up all inheritable **Backbone.Router** properties and methods.
    _.extend(Router.prototype, Events, {

      // Initialize is an empty function by default. Override it with your own
      // initialization logic.
      initialize: function(){},

      // Manually bind a single named route to a callback. For example:
      //
      //     this.route('search/:query/p:num', 'search', function(query, num) {
      //       ...
      //     });
      //
      route: function(route, name, callback) {
        if (!_.isRegExp(route)) route = this._routeToRegExp(route);
        if (_.isFunction(name)) {
          callback = name;
          name = '';
        }
        if (!callback) callback = this[name];
        var router = this;
        Backbone.history.route(route, function(fragment) {
          var args = router._extractParameters(route, fragment);
          if (router.execute(callback, args, name) !== false) {
            router.trigger.apply(router, ['route:' + name].concat(args));
            router.trigger('route', name, args);
            Backbone.history.trigger('route', router, name, args);
          }
        });
        return this;
      },

      // Execute a route handler with the provided parameters.  This is an
      // excellent place to do pre-route setup or post-route cleanup.
      execute: function(callback, args, name) {
        if (callback) callback.apply(this, args);
      },

      // Simple proxy to `Backbone.history` to save a fragment into the history.
      navigate: function(fragment, options) {
        Backbone.history.navigate(fragment, options);
        return this;
      },

      // Bind all defined routes to `Backbone.history`. We have to reverse the
      // order of the routes here to support behavior where the most general
      // routes can be defined at the bottom of the route map.
      _bindRoutes: function() {
        if (!this.routes) return;
        this.routes = _.result(this, 'routes');
        var route, routes = _.keys(this.routes);
        while ((route = routes.pop()) != null) {
          this.route(route, this.routes[route]);
        }
      },

      // Convert a route string into a regular expression, suitable for matching
      // against the current location hash.
      _routeToRegExp: function(route) {
        route = route.replace(escapeRegExp, '\\$&')
                     .replace(optionalParam, '(?:$1)?')
                     .replace(namedParam, function(match, optional) {
                       return optional ? match : '([^/?]+)';
                     })
                     .replace(splatParam, '([^?]*?)');
        return new RegExp('^' + route + '(?:\\?([\\s\\S]*))?$');
      },

      // Given a route, and a URL fragment that it matches, return the array of
      // extracted decoded parameters. Empty or unmatched parameters will be
      // treated as `null` to normalize cross-browser behavior.
      _extractParameters: function(route, fragment) {
        var params = route.exec(fragment).slice(1);
        return _.map(params, function(param, i) {
          // Don't decode the search params.
          if (i === params.length - 1) return param || null;
          return param ? decodeURIComponent(param) : null;
        });
      }

    });

    // Backbone.History
    // ----------------

    // Handles cross-browser history management, based on either
    // [pushState](http://diveintohtml5.info/history.html) and real URLs, or
    // [onhashchange](https://developer.mozilla.org/en-US/docs/DOM/window.onhashchange)
    // and URL fragments. If the browser supports neither (old IE, natch),
    // falls back to polling.
    var History = Backbone.History = function() {
      this.handlers = [];
      this.checkUrl = _.bind(this.checkUrl, this);

      // Ensure that `History` can be used outside of the browser.
      if (typeof window !== 'undefined') {
        this.location = window.location;
        this.history = window.history;
      }
    };

    // Cached regex for stripping a leading hash/slash and trailing space.
    var routeStripper = /^[#\/]|\s+$/g;

    // Cached regex for stripping leading and trailing slashes.
    var rootStripper = /^\/+|\/+$/g;

    // Cached regex for stripping urls of hash.
    var pathStripper = /#.*$/;

    // Has the history handling already been started?
    History.started = false;

    // Set up all inheritable **Backbone.History** properties and methods.
    _.extend(History.prototype, Events, {

      // The default interval to poll for hash changes, if necessary, is
      // twenty times a second.
      interval: 50,

      // Are we at the app root?
      atRoot: function() {
        var path = this.location.pathname.replace(/[^\/]$/, '$&/');
        return path === this.root && !this.getSearch();
      },

      // Does the pathname match the root?
      matchRoot: function() {
        var path = this.decodeFragment(this.location.pathname);
        var root = path.slice(0, this.root.length - 1) + '/';
        return root === this.root;
      },

      // Unicode characters in `location.pathname` are percent encoded so they're
      // decoded for comparison. `%25` should not be decoded since it may be part
      // of an encoded parameter.
      decodeFragment: function(fragment) {
        return decodeURI(fragment.replace(/%25/g, '%2525'));
      },

      // In IE6, the hash fragment and search params are incorrect if the
      // fragment contains `?`.
      getSearch: function() {
        var match = this.location.href.replace(/#.*/, '').match(/\?.+/);
        return match ? match[0] : '';
      },

      // Gets the true hash value. Cannot use location.hash directly due to bug
      // in Firefox where location.hash will always be decoded.
      getHash: function(window) {
        var match = (window || this).location.href.match(/#(.*)$/);
        return match ? match[1] : '';
      },

      // Get the pathname and search params, without the root.
      getPath: function() {
        var path = this.decodeFragment(
          this.location.pathname + this.getSearch()
        ).slice(this.root.length - 1);
        return path.charAt(0) === '/' ? path.slice(1) : path;
      },

      // Get the cross-browser normalized URL fragment from the path or hash.
      getFragment: function(fragment) {
        if (fragment == null) {
          if (this._usePushState || !this._wantsHashChange) {
            fragment = this.getPath();
          } else {
            fragment = this.getHash();
          }
        }
        return fragment.replace(routeStripper, '');
      },

      // Start the hash change handling, returning `true` if the current URL matches
      // an existing route, and `false` otherwise.
      start: function(options) {
        if (History.started) throw new Error('Backbone.history has already been started');
        History.started = true;

        // Figure out the initial configuration. Do we need an iframe?
        // Is pushState desired ... is it available?
        this.options          = _.extend({root: '/'}, this.options, options);
        this.root             = this.options.root;
        this._wantsHashChange = this.options.hashChange !== false;
        this._hasHashChange   = 'onhashchange' in window && (document.documentMode === void 0 || document.documentMode > 7);
        this._useHashChange   = this._wantsHashChange && this._hasHashChange;
        this._wantsPushState  = !!this.options.pushState;
        this._hasPushState    = !!(this.history && this.history.pushState);
        this._usePushState    = this._wantsPushState && this._hasPushState;
        this.fragment         = this.getFragment();

        // Normalize root to always include a leading and trailing slash.
        this.root = ('/' + this.root + '/').replace(rootStripper, '/');

        // Transition from hashChange to pushState or vice versa if both are
        // requested.
        if (this._wantsHashChange && this._wantsPushState) {

          // If we've started off with a route from a `pushState`-enabled
          // browser, but we're currently in a browser that doesn't support it...
          if (!this._hasPushState && !this.atRoot()) {
            var root = this.root.slice(0, -1) || '/';
            this.location.replace(root + '#' + this.getPath());
            // Return immediately as browser will do redirect to new url
            return true;

          // Or if we've started out with a hash-based route, but we're currently
          // in a browser where it could be `pushState`-based instead...
          } else if (this._hasPushState && this.atRoot()) {
            this.navigate(this.getHash(), {replace: true});
          }

        }

        // Proxy an iframe to handle location events if the browser doesn't
        // support the `hashchange` event, HTML5 history, or the user wants
        // `hashChange` but not `pushState`.
        if (!this._hasHashChange && this._wantsHashChange && !this._usePushState) {
          this.iframe = document.createElement('iframe');
          this.iframe.src = 'javascript:0';
          this.iframe.style.display = 'none';
          this.iframe.tabIndex = -1;
          var body = document.body;
          // Using `appendChild` will throw on IE < 9 if the document is not ready.
          var iWindow = body.insertBefore(this.iframe, body.firstChild).contentWindow;
          iWindow.document.open();
          iWindow.document.close();
          iWindow.location.hash = '#' + this.fragment;
        }

        // Add a cross-platform `addEventListener` shim for older browsers.
        var addEventListener = window.addEventListener || function (eventName, listener) {
          return attachEvent('on' + eventName, listener);
        };

        // Depending on whether we're using pushState or hashes, and whether
        // 'onhashchange' is supported, determine how we check the URL state.
        if (this._usePushState) {
          addEventListener('popstate', this.checkUrl, false);
        } else if (this._useHashChange && !this.iframe) {
          addEventListener('hashchange', this.checkUrl, false);
        } else if (this._wantsHashChange) {
          this._checkUrlInterval = setInterval(this.checkUrl, this.interval);
        }

        if (!this.options.silent) return this.loadUrl();
      },

      // Disable Backbone.history, perhaps temporarily. Not useful in a real app,
      // but possibly useful for unit testing Routers.
      stop: function() {
        // Add a cross-platform `removeEventListener` shim for older browsers.
        var removeEventListener = window.removeEventListener || function (eventName, listener) {
          return detachEvent('on' + eventName, listener);
        };

        // Remove window listeners.
        if (this._usePushState) {
          removeEventListener('popstate', this.checkUrl, false);
        } else if (this._useHashChange && !this.iframe) {
          removeEventListener('hashchange', this.checkUrl, false);
        }

        // Clean up the iframe if necessary.
        if (this.iframe) {
          document.body.removeChild(this.iframe);
          this.iframe = null;
        }

        // Some environments will throw when clearing an undefined interval.
        if (this._checkUrlInterval) clearInterval(this._checkUrlInterval);
        History.started = false;
      },

      // Add a route to be tested when the fragment changes. Routes added later
      // may override previous routes.
      route: function(route, callback) {
        this.handlers.unshift({route: route, callback: callback});
      },

      // Checks the current URL to see if it has changed, and if it has,
      // calls `loadUrl`, normalizing across the hidden iframe.
      checkUrl: function(e) {
        var current = this.getFragment();

        // If the user pressed the back button, the iframe's hash will have
        // changed and we should use that for comparison.
        if (current === this.fragment && this.iframe) {
          current = this.getHash(this.iframe.contentWindow);
        }

        if (current === this.fragment) return false;
        if (this.iframe) this.navigate(current);
        this.loadUrl();
      },

      // Attempt to load the current URL fragment. If a route succeeds with a
      // match, returns `true`. If no defined routes matches the fragment,
      // returns `false`.
      loadUrl: function(fragment) {
        // If the root doesn't match, no routes can match either.
        if (!this.matchRoot()) return false;
        fragment = this.fragment = this.getFragment(fragment);
        return _.some(this.handlers, function(handler) {
          if (handler.route.test(fragment)) {
            handler.callback(fragment);
            return true;
          }
        });
      },

      // Save a fragment into the hash history, or replace the URL state if the
      // 'replace' option is passed. You are responsible for properly URL-encoding
      // the fragment in advance.
      //
      // The options object can contain `trigger: true` if you wish to have the
      // route callback be fired (not usually desirable), or `replace: true`, if
      // you wish to modify the current URL without adding an entry to the history.
      navigate: function(fragment, options) {
        if (!History.started) return false;
        if (!options || options === true) options = {trigger: !!options};

        // Normalize the fragment.
        fragment = this.getFragment(fragment || '');

        // Don't include a trailing slash on the root.
        var root = this.root;
        if (fragment === '' || fragment.charAt(0) === '?') {
          root = root.slice(0, -1) || '/';
        }
        var url = root + fragment;

        // Strip the hash and decode for matching.
        fragment = this.decodeFragment(fragment.replace(pathStripper, ''));

        if (this.fragment === fragment) return;
        this.fragment = fragment;

        // If pushState is available, we use it to set the fragment as a real URL.
        if (this._usePushState) {
          this.history[options.replace ? 'replaceState' : 'pushState']({}, document.title, url);

        // If hash changes haven't been explicitly disabled, update the hash
        // fragment to store history.
        } else if (this._wantsHashChange) {
          this._updateHash(this.location, fragment, options.replace);
          if (this.iframe && (fragment !== this.getHash(this.iframe.contentWindow))) {
            var iWindow = this.iframe.contentWindow;

            // Opening and closing the iframe tricks IE7 and earlier to push a
            // history entry on hash-tag change.  When replace is true, we don't
            // want this.
            if (!options.replace) {
              iWindow.document.open();
              iWindow.document.close();
            }

            this._updateHash(iWindow.location, fragment, options.replace);
          }

        // If you've told us that you explicitly don't want fallback hashchange-
        // based history, then `navigate` becomes a page refresh.
        } else {
          return this.location.assign(url);
        }
        if (options.trigger) return this.loadUrl(fragment);
      },

      // Update the hash location, either replacing the current entry, or adding
      // a new one to the browser history.
      _updateHash: function(location, fragment, replace) {
        if (replace) {
          var href = location.href.replace(/(javascript:|#).*$/, '');
          location.replace(href + '#' + fragment);
        } else {
          // Some browsers require that `hash` contains a leading #.
          location.hash = '#' + fragment;
        }
      }

    });

    // Create the default Backbone.history.
    Backbone.history = new History;

    // Helpers
    // -------

    // Helper function to correctly set up the prototype chain for subclasses.
    // Similar to `goog.inherits`, but uses a hash of prototype properties and
    // class properties to be extended.
    var extend = function(protoProps, staticProps) {
      var parent = this;
      var child;

      // The constructor function for the new subclass is either defined by you
      // (the "constructor" property in your `extend` definition), or defaulted
      // by us to simply call the parent constructor.
      if (protoProps && _.has(protoProps, 'constructor')) {
        child = protoProps.constructor;
      } else {
        child = function(){ return parent.apply(this, arguments); };
      }

      // Add static properties to the constructor function, if supplied.
      _.extend(child, parent, staticProps);

      // Set the prototype chain to inherit from `parent`, without calling
      // `parent` constructor function.
      var Surrogate = function(){ this.constructor = child; };
      Surrogate.prototype = parent.prototype;
      child.prototype = new Surrogate;

      // Add prototype properties (instance properties) to the subclass,
      // if supplied.
      if (protoProps) _.extend(child.prototype, protoProps);

      // Set a convenience property in case the parent's prototype is needed
      // later.
      child.__super__ = parent.prototype;

      return child;
    };

    // Set up inheritance for the model, collection, router, view and history.
    Model.extend = Collection.extend = Router.extend = View.extend = History.extend = extend;

    // Throw an error when a URL is needed, and none is supplied.
    var urlError = function() {
      throw new Error('A "url" property or function must be specified');
    };

    // Wrap an optional error callback with a fallback error event.
    var wrapError = function(model, options) {
      var error = options.error;
      options.error = function(resp) {
        if (error) error.call(options.context, model, resp, options);
        model.trigger('error', model, resp, options);
      };
    };

    return Backbone;

  }));

  var __export1 = exports.View;
  var __export2 = exports.Model;

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$1 = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  // Caches a local reference to `Element.prototype` for faster access.
  var ElementProto = Element.prototype; // : typeof Element = (typeof Element !== 'undefined' && Element.prototype) || {};
  // Find the right `Element#matches` for IE>=9 and modern browsers.
  var matchesSelector = ElementProto.matches ||
      ElementProto['webkitMatchesSelector'] ||
      ElementProto['mozMatchesSelector'] ||
      ElementProto['msMatchesSelector'] ||
      ElementProto['oMatchesSelector'] ||
      function matches(selector) {
          /* tslint:disable:no-invalid-this */
          var matches = (this.document || this.ownerDocument).querySelectorAll(selector);
          var i = matches.length;
          while (--i >= 0 && matches.item(i) !== this) {
              continue;
          }
          return i > -1;
          /* tslint:enable:no-invalid-this */
      };
  var NativeView = /** @class */ (function (_super) {
      __extends$1(NativeView, _super);
      function NativeView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      NativeView.prototype._removeElement = function () {
          this.undelegateEvents();
          if (this.el.parentNode) {
              this.el.parentNode.removeChild(this.el);
          }
      };
      // Apply the `element` to the view.
      NativeView.prototype._setElement = function (element) {
          this.el = element;
      };
      // Set a hash of attributes to the view's `el`. We use the "prop" version
      // if available, falling back to `setAttribute` for the catch-all.
      NativeView.prototype._setAttributes = function (attrs) {
          for (var attr in attrs) {
              attr in this.el ? this.el[attr] = attrs[attr] : this.el.setAttribute(attr, attrs[attr]);
          }
      };
      /**
       * Make an event delegation handler for the given `eventName` and `selector`
       * and attach it to `this.el`.
       * If selector is empty, the listener will be bound to `this.el`. If not, a
       * new handler that will recursively traverse up the event target's DOM
       * hierarchy looking for a node that matches the selector. If one is found,
       * the event's `delegateTarget` property is set to it and the return the
       * result of calling bound `listener` with the parameters given to the
       * handler.
       *
       * This does not properly handle selectors for things like focus and blur (see
       * https://github.com/jquery/jquery/blob/7d21f02b9ec9f655583e898350badf89165ed4d5/src/event.js#L442
       * for some similar exceptional cases).
       */
      NativeView.prototype.delegate = function (eventName, selector, listener) {
          if (typeof selector !== 'string') {
              listener = selector;
              selector = null;
          }
          // We have to initialize this here, instead of in the constructor, because the
          // super constructor eventually calls this method before we get a chance to initialize
          // this._domEvents to an empty list.
          if (this._domEvents === void 0) {
              this._domEvents = [];
          }
          var root = this.el;
          var handler = selector ? function (e) {
              var node = e.target || e.srcElement;
              for (; node && node !== root; node = node.parentNode) {
                  if (matchesSelector.call(node, selector)) {
                      e.delegateTarget = node;
                      if (listener.handleEvent) {
                          return listener.handleEvent(e);
                      }
                      else {
                          return listener(e);
                      }
                  }
              }
          } : listener;
          this.el.addEventListener(eventName, handler, false);
          this._domEvents.push({ eventName: eventName, handler: handler, listener: listener, selector: selector });
          return handler;
      };
      // Remove a single delegated event. Either `eventName` or `selector` must
      // be included, `selector` and `listener` are optional.
      NativeView.prototype.undelegate = function (eventName, selector, listener) {
          if (typeof selector === 'function') {
              listener = selector;
              selector = null;
          }
          if (this.el && this._domEvents) {
              var handlers = this._domEvents.slice();
              var i = handlers.length;
              while (i--) {
                  var item = handlers[i];
                  var match = item.eventName === eventName &&
                      (listener ? item.listener === listener : true) &&
                      (selector ? item.selector === selector : true);
                  if (!match) {
                      continue;
                  }
                  this.el.removeEventListener(item.eventName, item.handler, false);
                  this._domEvents.splice(i, 1);
              }
          }
          return this;
      };
      // Remove all events created with `delegate` from `el`
      NativeView.prototype.undelegateEvents = function () {
          if (this.el && this._domEvents) {
              var len = this._domEvents.length;
              for (var i = 0; i < len; i++) {
                  var item = this._domEvents[i];
                  this.el.removeEventListener(item.eventName, item.handler, false);
              }
              this._domEvents.length = 0;
          }
          return this;
      };
      return NativeView;
  }(__export1));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for array-specific algorithms.
   */
  var ArrayExt;
  (function (ArrayExt) {
      /**
       * Find the index of the first occurrence of a value in an array.
       *
       * @param array - The array-like object to search.
       *
       * @param value - The value to locate in the array. Values are
       *   compared using strict `===` equality.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the first occurrence of the value, or `-1`
       *   if the value is not found.
       *
       * #### Notes
       * If `stop < start` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = ['one', 'two', 'three', 'four', 'one'];
       * ArrayExt.firstIndexOf(data, 'red');        // -1
       * ArrayExt.firstIndexOf(data, 'one');        // 0
       * ArrayExt.firstIndexOf(data, 'one', 1);     // 4
       * ArrayExt.firstIndexOf(data, 'two', 2);     // -1
       * ArrayExt.firstIndexOf(data, 'two', 2, 1);  // 1
       * ```
       */
      function firstIndexOf(array, value, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return -1;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var span;
          if (stop < start) {
              span = (stop + 1) + (n - start);
          }
          else {
              span = stop - start + 1;
          }
          for (var i = 0; i < span; ++i) {
              var j = (start + i) % n;
              if (array[j] === value) {
                  return j;
              }
          }
          return -1;
      }
      ArrayExt.firstIndexOf = firstIndexOf;
      /**
       * Find the index of the last occurrence of a value in an array.
       *
       * @param array - The array-like object to search.
       *
       * @param value - The value to locate in the array. Values are
       *   compared using strict `===` equality.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the last occurrence of the value, or `-1`
       *   if the value is not found.
       *
       * #### Notes
       * If `start < stop` the search will wrap at the front of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = ['one', 'two', 'three', 'four', 'one'];
       * ArrayExt.lastIndexOf(data, 'red');        // -1
       * ArrayExt.lastIndexOf(data, 'one');        // 4
       * ArrayExt.lastIndexOf(data, 'one', 1);     // 0
       * ArrayExt.lastIndexOf(data, 'two', 0);     // -1
       * ArrayExt.lastIndexOf(data, 'two', 0, 1);  // 1
       * ```
       */
      function lastIndexOf(array, value, start, stop) {
          if (start === void 0) { start = -1; }
          if (stop === void 0) { stop = 0; }
          var n = array.length;
          if (n === 0) {
              return -1;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var span;
          if (start < stop) {
              span = (start + 1) + (n - stop);
          }
          else {
              span = start - stop + 1;
          }
          for (var i = 0; i < span; ++i) {
              var j = (start - i + n) % n;
              if (array[j] === value) {
                  return j;
              }
          }
          return -1;
      }
      ArrayExt.lastIndexOf = lastIndexOf;
      /**
       * Find the index of the first value which matches a predicate.
       *
       * @param array - The array-like object to search.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the first matching value, or `-1` if no
       *   matching value is found.
       *
       * #### Notes
       * If `stop < start` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * Modifying the length of the array while searching.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * let data = [1, 2, 3, 4, 3, 2, 1];
       * ArrayExt.findFirstIndex(data, isEven);       // 1
       * ArrayExt.findFirstIndex(data, isEven, 4);    // 5
       * ArrayExt.findFirstIndex(data, isEven, 6);    // -1
       * ArrayExt.findFirstIndex(data, isEven, 6, 5); // 1
       * ```
       */
      function findFirstIndex(array, fn, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return -1;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var span;
          if (stop < start) {
              span = (stop + 1) + (n - start);
          }
          else {
              span = stop - start + 1;
          }
          for (var i = 0; i < span; ++i) {
              var j = (start + i) % n;
              if (fn(array[j], j)) {
                  return j;
              }
          }
          return -1;
      }
      ArrayExt.findFirstIndex = findFirstIndex;
      /**
       * Find the index of the last value which matches a predicate.
       *
       * @param object - The array-like object to search.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the last matching value, or `-1` if no
       *   matching value is found.
       *
       * #### Notes
       * If `start < stop` the search will wrap at the front of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * Modifying the length of the array while searching.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * let data = [1, 2, 3, 4, 3, 2, 1];
       * ArrayExt.findLastIndex(data, isEven);        // 5
       * ArrayExt.findLastIndex(data, isEven, 4);     // 3
       * ArrayExt.findLastIndex(data, isEven, 0);     // -1
       * ArrayExt.findLastIndex(data, isEven, 0, 1);  // 5
       * ```
       */
      function findLastIndex(array, fn, start, stop) {
          if (start === void 0) { start = -1; }
          if (stop === void 0) { stop = 0; }
          var n = array.length;
          if (n === 0) {
              return -1;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var d;
          if (start < stop) {
              d = (start + 1) + (n - stop);
          }
          else {
              d = start - stop + 1;
          }
          for (var i = 0; i < d; ++i) {
              var j = (start - i + n) % n;
              if (fn(array[j], j)) {
                  return j;
              }
          }
          return -1;
      }
      ArrayExt.findLastIndex = findLastIndex;
      /**
       * Find the first value which matches a predicate.
       *
       * @param array - The array-like object to search.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The first matching value, or `undefined` if no matching
       *   value is found.
       *
       * #### Notes
       * If `stop < start` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * Modifying the length of the array while searching.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * let data = [1, 2, 3, 4, 3, 2, 1];
       * ArrayExt.findFirstValue(data, isEven);       // 2
       * ArrayExt.findFirstValue(data, isEven, 2);    // 4
       * ArrayExt.findFirstValue(data, isEven, 6);    // undefined
       * ArrayExt.findFirstValue(data, isEven, 6, 5); // 2
       * ```
       */
      function findFirstValue(array, fn, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var index = findFirstIndex(array, fn, start, stop);
          return index !== -1 ? array[index] : undefined;
      }
      ArrayExt.findFirstValue = findFirstValue;
      /**
       * Find the last value which matches a predicate.
       *
       * @param object - The array-like object to search.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The last matching value, or `undefined` if no matching
       *   value is found.
       *
       * #### Notes
       * If `start < stop` the search will wrap at the front of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * Modifying the length of the array while searching.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * let data = [1, 2, 3, 4, 3, 2, 1];
       * ArrayExt.findLastValue(data, isEven);        // 2
       * ArrayExt.findLastValue(data, isEven, 4);     // 4
       * ArrayExt.findLastValue(data, isEven, 0);     // undefined
       * ArrayExt.findLastValue(data, isEven, 0, 1);  // 2
       * ```
       */
      function findLastValue(array, fn, start, stop) {
          if (start === void 0) { start = -1; }
          if (stop === void 0) { stop = 0; }
          var index = findLastIndex(array, fn, start, stop);
          return index !== -1 ? array[index] : undefined;
      }
      ArrayExt.findLastValue = findLastValue;
      /**
       * Find the index of the first element which compares `>=` to a value.
       *
       * @param array - The sorted array-like object to search.
       *
       * @param value - The value to locate in the array.
       *
       * @param fn - The 3-way comparison function to apply to the values.
       *   It should return `< 0` if an element is less than a value, `0` if
       *   an element is equal to a value, or `> 0` if an element is greater
       *   than a value.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the first element which compares `>=` to the
       *   value, or `length` if there is no such element. If the computed
       *   index for `stop` is less than `start`, then the computed index
       *   for `start` is returned.
       *
       * #### Notes
       * The array must already be sorted in ascending order according to
       * the comparison function.
       *
       * #### Complexity
       * Logarithmic.
       *
       * #### Undefined Behavior
       * Searching a range which is not sorted in ascending order.
       *
       * A `start` or `stop` which is non-integral.
       *
       * Modifying the length of the array while searching.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function numberCmp(a: number, b: number): number {
       *   return a - b;
       * }
       *
       * let data = [0, 3, 4, 7, 7, 9];
       * ArrayExt.lowerBound(data, 0, numberCmp);   // 0
       * ArrayExt.lowerBound(data, 6, numberCmp);   // 3
       * ArrayExt.lowerBound(data, 7, numberCmp);   // 3
       * ArrayExt.lowerBound(data, -1, numberCmp);  // 0
       * ArrayExt.lowerBound(data, 10, numberCmp);  // 6
       * ```
       */
      function lowerBound(array, value, fn, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return 0;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var begin = start;
          var span = stop - start + 1;
          while (span > 0) {
              var half = span >> 1;
              var middle = begin + half;
              if (fn(array[middle], value) < 0) {
                  begin = middle + 1;
                  span -= half + 1;
              }
              else {
                  span = half;
              }
          }
          return begin;
      }
      ArrayExt.lowerBound = lowerBound;
      /**
       * Find the index of the first element which compares `>` than a value.
       *
       * @param array - The sorted array-like object to search.
       *
       * @param value - The value to locate in the array.
       *
       * @param fn - The 3-way comparison function to apply to the values.
       *   It should return `< 0` if an element is less than a value, `0` if
       *   an element is equal to a value, or `> 0` if an element is greater
       *   than a value.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the first element which compares `>` than the
       *   value, or `length` if there is no such element. If the computed
       *   index for `stop` is less than `start`, then the computed index
       *   for `start` is returned.
       *
       * #### Notes
       * The array must already be sorted in ascending order according to
       * the comparison function.
       *
       * #### Complexity
       * Logarithmic.
       *
       * #### Undefined Behavior
       * Searching a range which is not sorted in ascending order.
       *
       * A `start` or `stop` which is non-integral.
       *
       * Modifying the length of the array while searching.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function numberCmp(a: number, b: number): number {
       *   return a - b;
       * }
       *
       * let data = [0, 3, 4, 7, 7, 9];
       * ArrayExt.upperBound(data, 0, numberCmp);   // 1
       * ArrayExt.upperBound(data, 6, numberCmp);   // 3
       * ArrayExt.upperBound(data, 7, numberCmp);   // 5
       * ArrayExt.upperBound(data, -1, numberCmp);  // 0
       * ArrayExt.upperBound(data, 10, numberCmp);  // 6
       * ```
       */
      function upperBound(array, value, fn, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return 0;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var begin = start;
          var span = stop - start + 1;
          while (span > 0) {
              var half = span >> 1;
              var middle = begin + half;
              if (fn(array[middle], value) > 0) {
                  span = half;
              }
              else {
                  begin = middle + 1;
                  span -= half + 1;
              }
          }
          return begin;
      }
      ArrayExt.upperBound = upperBound;
      /**
       * Test whether two arrays are shallowly equal.
       *
       * @param a - The first array-like object to compare.
       *
       * @param b - The second array-like object to compare.
       *
       * @param fn - The comparison function to apply to the elements. It
       *   should return `true` if the elements are "equal". The default
       *   compares elements using strict `===` equality.
       *
       * @returns Whether the two arrays are shallowly equal.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * Modifying the length of the arrays while comparing.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let d1 = [0, 3, 4, 7, 7, 9];
       * let d2 = [0, 3, 4, 7, 7, 9];
       * let d3 = [42];
       * ArrayExt.shallowEqual(d1, d2);  // true
       * ArrayExt.shallowEqual(d2, d3);  // false
       * ```
       */
      function shallowEqual(a, b, fn) {
          // Check for object identity first.
          if (a === b) {
              return true;
          }
          // Bail early if the lengths are different.
          if (a.length !== b.length) {
              return false;
          }
          // Compare each element for equality.
          for (var i = 0, n = a.length; i < n; ++i) {
              if (fn ? !fn(a[i], b[i]) : a[i] !== b[i]) {
                  return false;
              }
          }
          // The array are shallowly equal.
          return true;
      }
      ArrayExt.shallowEqual = shallowEqual;
      /**
       * Create a slice of an array subject to an optional step.
       *
       * @param array - The array-like object of interest.
       *
       * @param options - The options for configuring the slice.
       *
       * @returns A new array with the specified values.
       *
       * @throws An exception if the slice `step` is `0`.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start`, `stop`, or `step` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 3, 4, 7, 7, 9];
       * ArrayExt.slice(data);                         // [0, 3, 4, 7, 7, 9]
       * ArrayExt.slice(data, { start: 2 });           // [4, 7, 7, 9]
       * ArrayExt.slice(data, { start: 0, stop: 4 });  // [0, 3, 4, 7]
       * ArrayExt.slice(data, { step: 2 });            // [0, 4, 7]
       * ArrayExt.slice(data, { step: -1 });           // [9, 7, 7, 4, 3, 0]
       * ```
       */
      function slice(array, options) {
          if (options === void 0) { options = {}; }
          // Extract the options.
          var start = options.start, stop = options.stop, step = options.step;
          // Set up the `step` value.
          if (step === undefined) {
              step = 1;
          }
          // Validate the step size.
          if (step === 0) {
              throw new Error('Slice `step` cannot be zero.');
          }
          // Look up the length of the array.
          var n = array.length;
          // Set up the `start` value.
          if (start === undefined) {
              start = step < 0 ? n - 1 : 0;
          }
          else if (start < 0) {
              start = Math.max(start + n, step < 0 ? -1 : 0);
          }
          else if (start >= n) {
              start = step < 0 ? n - 1 : n;
          }
          // Set up the `stop` value.
          if (stop === undefined) {
              stop = step < 0 ? -1 : n;
          }
          else if (stop < 0) {
              stop = Math.max(stop + n, step < 0 ? -1 : 0);
          }
          else if (stop >= n) {
              stop = step < 0 ? n - 1 : n;
          }
          // Compute the slice length.
          var length;
          if ((step < 0 && stop >= start) || (step > 0 && start >= stop)) {
              length = 0;
          }
          else if (step < 0) {
              length = Math.floor((stop - start + 1) / step + 1);
          }
          else {
              length = Math.floor((stop - start - 1) / step + 1);
          }
          // Compute the sliced result.
          var result = [];
          for (var i = 0; i < length; ++i) {
              result[i] = array[start + i * step];
          }
          // Return the result.
          return result;
      }
      ArrayExt.slice = slice;
      /**
       * Move an element in an array from one index to another.
       *
       * @param array - The mutable array-like object of interest.
       *
       * @param fromIndex - The index of the element to move. Negative
       *   values are taken as an offset from the end of the array.
       *
       * @param toIndex - The target index of the element. Negative
       *   values are taken as an offset from the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `fromIndex` or `toIndex` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from from '@lumino/algorithm';
       *
       * let data = [0, 1, 2, 3, 4];
       * ArrayExt.move(data, 1, 2);  // [0, 2, 1, 3, 4]
       * ArrayExt.move(data, 4, 2);  // [0, 2, 4, 1, 3]
       * ```
       */
      function move(array, fromIndex, toIndex) {
          var n = array.length;
          if (n <= 1) {
              return;
          }
          if (fromIndex < 0) {
              fromIndex = Math.max(0, fromIndex + n);
          }
          else {
              fromIndex = Math.min(fromIndex, n - 1);
          }
          if (toIndex < 0) {
              toIndex = Math.max(0, toIndex + n);
          }
          else {
              toIndex = Math.min(toIndex, n - 1);
          }
          if (fromIndex === toIndex) {
              return;
          }
          var value = array[fromIndex];
          var d = fromIndex < toIndex ? 1 : -1;
          for (var i = fromIndex; i !== toIndex; i += d) {
              array[i] = array[i + d];
          }
          array[toIndex] = value;
      }
      ArrayExt.move = move;
      /**
       * Reverse an array in-place.
       *
       * @param array - The mutable array-like object of interest.
       *
       * @param start - The index of the first element in the range to be
       *   reversed, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   reversed, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or  `stop` index which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 1, 2, 3, 4];
       * ArrayExt.reverse(data, 1, 3);  // [0, 3, 2, 1, 4]
       * ArrayExt.reverse(data, 3);     // [0, 3, 2, 4, 1]
       * ArrayExt.reverse(data);        // [1, 4, 2, 3, 0]
       * ```
       */
      function reverse(array, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n <= 1) {
              return;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          while (start < stop) {
              var a = array[start];
              var b = array[stop];
              array[start++] = b;
              array[stop--] = a;
          }
      }
      ArrayExt.reverse = reverse;
      /**
       * Rotate the elements of an array in-place.
       *
       * @param array - The mutable array-like object of interest.
       *
       * @param delta - The amount of rotation to apply to the elements. A
       *   positive value will rotate the elements to the left. A negative
       *   value will rotate the elements to the right.
       *
       * @param start - The index of the first element in the range to be
       *   rotated, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   rotated, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `delta`, `start`, or `stop` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 1, 2, 3, 4];
       * ArrayExt.rotate(data, 2);        // [2, 3, 4, 0, 1]
       * ArrayExt.rotate(data, -2);       // [0, 1, 2, 3, 4]
       * ArrayExt.rotate(data, 10);       // [0, 1, 2, 3, 4]
       * ArrayExt.rotate(data, 9);        // [4, 0, 1, 2, 3]
       * ArrayExt.rotate(data, 2, 1, 3);  // [4, 2, 0, 1, 3]
       * ```
       */
      function rotate(array, delta, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n <= 1) {
              return;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          if (start >= stop) {
              return;
          }
          var length = stop - start + 1;
          if (delta > 0) {
              delta = delta % length;
          }
          else if (delta < 0) {
              delta = ((delta % length) + length) % length;
          }
          if (delta === 0) {
              return;
          }
          var pivot = start + delta;
          reverse(array, start, pivot - 1);
          reverse(array, pivot, stop);
          reverse(array, start, stop);
      }
      ArrayExt.rotate = rotate;
      /**
       * Fill an array with a static value.
       *
       * @param array - The mutable array-like object to fill.
       *
       * @param value - The static value to use to fill the array.
       *
       * @param start - The index of the first element in the range to be
       *   filled, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   filled, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * #### Notes
       * If `stop < start` the fill will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * A `start` or `stop` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = ['one', 'two', 'three', 'four'];
       * ArrayExt.fill(data, 'r');        // ['r', 'r', 'r', 'r']
       * ArrayExt.fill(data, 'g', 1);     // ['r', 'g', 'g', 'g']
       * ArrayExt.fill(data, 'b', 2, 3);  // ['r', 'g', 'b', 'b']
       * ArrayExt.fill(data, 'z', 3, 1);  // ['z', 'z', 'b', 'z']
       * ```
       */
      function fill(array, value, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var span;
          if (stop < start) {
              span = (stop + 1) + (n - start);
          }
          else {
              span = stop - start + 1;
          }
          for (var i = 0; i < span; ++i) {
              array[(start + i) % n] = value;
          }
      }
      ArrayExt.fill = fill;
      /**
       * Insert a value into an array at a specific index.
       *
       * @param array - The array of interest.
       *
       * @param index - The index at which to insert the value. Negative
       *   values are taken as an offset from the end of the array.
       *
       * @param value - The value to set at the specified index.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * An `index` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 1, 2];
       * ArrayExt.insert(data, 0, -1);  // [-1, 0, 1, 2]
       * ArrayExt.insert(data, 2, 12);  // [-1, 0, 12, 1, 2]
       * ArrayExt.insert(data, -1, 7);  // [-1, 0, 12, 1, 7, 2]
       * ArrayExt.insert(data, 6, 19);  // [-1, 0, 12, 1, 7, 2, 19]
       * ```
       */
      function insert(array, index, value) {
          var n = array.length;
          if (index < 0) {
              index = Math.max(0, index + n);
          }
          else {
              index = Math.min(index, n);
          }
          for (var i = n; i > index; --i) {
              array[i] = array[i - 1];
          }
          array[index] = value;
      }
      ArrayExt.insert = insert;
      /**
       * Remove and return a value at a specific index in an array.
       *
       * @param array - The array of interest.
       *
       * @param index - The index of the value to remove. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The value at the specified index, or `undefined` if the
       *   index is out of range.
       *
       * #### Complexity
       * Linear.
       *
       * #### Undefined Behavior
       * An `index` which is non-integral.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 12, 23, 39, 14, 12, 75];
       * ArrayExt.removeAt(data, 2);   // 23
       * ArrayExt.removeAt(data, -2);  // 12
       * ArrayExt.removeAt(data, 10);  // undefined;
       * ```
       */
      function removeAt(array, index) {
          var n = array.length;
          if (index < 0) {
              index += n;
          }
          if (index < 0 || index >= n) {
              return undefined;
          }
          var value = array[index];
          for (var i = index + 1; i < n; ++i) {
              array[i - 1] = array[i];
          }
          array.length = n - 1;
          return value;
      }
      ArrayExt.removeAt = removeAt;
      /**
       * Remove the first occurrence of a value from an array.
       *
       * @param array - The array of interest.
       *
       * @param value - The value to remove from the array. Values are
       *   compared using strict `===` equality.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the removed value, or `-1` if the value
       *   is not contained in the array.
       *
       * #### Notes
       * If `stop < start` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 12, 23, 39, 14, 12, 75];
       * ArrayExt.removeFirstOf(data, 12);        // 1
       * ArrayExt.removeFirstOf(data, 17);        // -1
       * ArrayExt.removeFirstOf(data, 39, 3);     // -1
       * ArrayExt.removeFirstOf(data, 39, 3, 2);  // 2
       * ```
       */
      function removeFirstOf(array, value, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var index = firstIndexOf(array, value, start, stop);
          if (index !== -1) {
              removeAt(array, index);
          }
          return index;
      }
      ArrayExt.removeFirstOf = removeFirstOf;
      /**
       * Remove the last occurrence of a value from an array.
       *
       * @param array - The array of interest.
       *
       * @param value - The value to remove from the array. Values are
       *   compared using strict `===` equality.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The index of the removed value, or `-1` if the value
       *   is not contained in the array.
       *
       * #### Notes
       * If `start < stop` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [0, 12, 23, 39, 14, 12, 75];
       * ArrayExt.removeLastOf(data, 12);        // 5
       * ArrayExt.removeLastOf(data, 17);        // -1
       * ArrayExt.removeLastOf(data, 39, 2);     // -1
       * ArrayExt.removeLastOf(data, 39, 2, 3);  // 3
       * ```
       */
      function removeLastOf(array, value, start, stop) {
          if (start === void 0) { start = -1; }
          if (stop === void 0) { stop = 0; }
          var index = lastIndexOf(array, value, start, stop);
          if (index !== -1) {
              removeAt(array, index);
          }
          return index;
      }
      ArrayExt.removeLastOf = removeLastOf;
      /**
       * Remove all occurrences of a value from an array.
       *
       * @param array - The array of interest.
       *
       * @param value - The value to remove from the array. Values are
       *   compared using strict `===` equality.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The number of elements removed from the array.
       *
       * #### Notes
       * If `stop < start` the search will conceptually wrap at the end of
       * the array, however the array will be traversed front-to-back.
       *
       * #### Complexity
       * Linear.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * let data = [14, 12, 23, 39, 14, 12, 19, 14];
       * ArrayExt.removeAllOf(data, 12);        // 2
       * ArrayExt.removeAllOf(data, 17);        // 0
       * ArrayExt.removeAllOf(data, 14, 1, 4);  // 1
       * ```
       */
      function removeAllOf(array, value, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return 0;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var count = 0;
          for (var i = 0; i < n; ++i) {
              if (start <= stop && (i >= start && i <= stop) && array[i] === value) {
                  count++;
              }
              else if (stop < start && (i <= stop || i >= start) && array[i] === value) {
                  count++;
              }
              else if (count > 0) {
                  array[i - count] = array[i];
              }
          }
          if (count > 0) {
              array.length = n - count;
          }
          return count;
      }
      ArrayExt.removeAllOf = removeAllOf;
      /**
       * Remove the first occurrence of a value which matches a predicate.
       *
       * @param array - The array of interest.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The removed `{ index, value }`, which will be `-1` and
       *   `undefined` if the value is not contained in the array.
       *
       * #### Notes
       * If `stop < start` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * let data = [0, 12, 23, 39, 14, 12, 75];
       * ArrayExt.removeFirstWhere(data, isEven);     // { index: 0, value: 0 }
       * ArrayExt.removeFirstWhere(data, isEven, 2);  // { index: 3, value: 14 }
       * ArrayExt.removeFirstWhere(data, isEven, 4);  // { index: -1, value: undefined }
       * ```
       */
      function removeFirstWhere(array, fn, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var value;
          var index = findFirstIndex(array, fn, start, stop);
          if (index !== -1) {
              value = removeAt(array, index);
          }
          return { index: index, value: value };
      }
      ArrayExt.removeFirstWhere = removeFirstWhere;
      /**
       * Remove the last occurrence of a value which matches a predicate.
       *
       * @param array - The array of interest.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The removed `{ index, value }`, which will be `-1` and
       *   `undefined` if the value is not contained in the array.
       *
       * #### Notes
       * If `start < stop` the search will wrap at the end of the array.
       *
       * #### Complexity
       * Linear.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * let data = [0, 12, 23, 39, 14, 12, 75];
       * ArrayExt.removeLastWhere(data, isEven);        // { index: 5, value: 12 }
       * ArrayExt.removeLastWhere(data, isEven, 2);     // { index: 1, value: 12 }
       * ArrayExt.removeLastWhere(data, isEven, 2, 1);  // { index: -1, value: undefined }
       * ```
       */
      function removeLastWhere(array, fn, start, stop) {
          if (start === void 0) { start = -1; }
          if (stop === void 0) { stop = 0; }
          var value;
          var index = findLastIndex(array, fn, start, stop);
          if (index !== -1) {
              value = removeAt(array, index);
          }
          return { index: index, value: value };
      }
      ArrayExt.removeLastWhere = removeLastWhere;
      /**
       * Remove all occurrences of values which match a predicate.
       *
       * @param array - The array of interest.
       *
       * @param fn - The predicate function to apply to the values.
       *
       * @param start - The index of the first element in the range to be
       *   searched, inclusive. The default value is `0`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @param stop - The index of the last element in the range to be
       *   searched, inclusive. The default value is `-1`. Negative values
       *   are taken as an offset from the end of the array.
       *
       * @returns The number of elements removed from the array.
       *
       * #### Notes
       * If `stop < start` the search will conceptually wrap at the end of
       * the array, however the array will be traversed front-to-back.
       *
       * #### Complexity
       * Linear.
       *
       * #### Example
       * ```typescript
       * import { ArrayExt } from '@lumino/algorithm';
       *
       * function isEven(value: number): boolean {
       *   return value % 2 === 0;
       * }
       *
       * function isNegative(value: number): boolean {
       *   return value < 0;
       * }
       *
       * let data = [0, 12, -13, -9, 23, 39, 14, -15, 12, 75];
       * ArrayExt.removeAllWhere(data, isEven);            // 4
       * ArrayExt.removeAllWhere(data, isNegative, 0, 3);  // 2
       * ```
       */
      function removeAllWhere(array, fn, start, stop) {
          if (start === void 0) { start = 0; }
          if (stop === void 0) { stop = -1; }
          var n = array.length;
          if (n === 0) {
              return 0;
          }
          if (start < 0) {
              start = Math.max(0, start + n);
          }
          else {
              start = Math.min(start, n - 1);
          }
          if (stop < 0) {
              stop = Math.max(0, stop + n);
          }
          else {
              stop = Math.min(stop, n - 1);
          }
          var count = 0;
          for (var i = 0; i < n; ++i) {
              if (start <= stop && (i >= start && i <= stop) && fn(array[i], i)) {
                  count++;
              }
              else if (stop < start && (i <= stop || i >= start) && fn(array[i], i)) {
                  count++;
              }
              else if (count > 0) {
                  array[i - count] = array[i];
              }
          }
          if (count > 0) {
              array.length = n - count;
          }
          return count;
      }
      ArrayExt.removeAllWhere = removeAllWhere;
  })(ArrayExt || (ArrayExt = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * Create an iterator for an iterable object.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @returns A new iterator for the given object.
   *
   * #### Notes
   * This function allows iteration algorithms to operate on user-defined
   * iterable types and builtin array-like objects in a uniform fashion.
   */
  function iter(object) {
      var it;
      if (typeof object.iter === 'function') {
          it = object.iter();
      }
      else {
          it = new ArrayIterator(object);
      }
      return it;
  }
  /**
   * Invoke a function for each value in an iterable.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @param fn - The callback function to invoke for each value.
   *
   * #### Notes
   * Iteration can be terminated early by returning `false` from the
   * callback function.
   *
   * #### Complexity
   * Linear.
   *
   * #### Example
   * ```typescript
   * import { each } from '@lumino/algorithm';
   *
   * let data = [5, 7, 0, -2, 9];
   *
   * each(data, value => { console.log(value); });
   * ```
   */
  function each$1(object, fn) {
      var index = 0;
      var it = iter(object);
      var value;
      while ((value = it.next()) !== undefined) {
          if (fn(value, index++) === false) {
              return;
          }
      }
  }
  /**
   * Test whether all values in an iterable satisfy a predicate.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @param fn - The predicate function to invoke for each value.
   *
   * @returns `true` if all values pass the test, `false` otherwise.
   *
   * #### Notes
   * Iteration terminates on the first `false` predicate result.
   *
   * #### Complexity
   * Linear.
   *
   * #### Example
   * ```typescript
   * import { every } from '@lumino/algorithm';
   *
   * let data = [5, 7, 1];
   *
   * every(data, value => value % 2 === 0);  // false
   * every(data, value => value % 2 === 1);  // true
   * ```
   */
  function every$1(object, fn) {
      var index = 0;
      var it = iter(object);
      var value;
      while ((value = it.next()) !== undefined) {
          if (!fn(value, index++)) {
              return false;
          }
      }
      return true;
  }
  /**
   * Test whether any value in an iterable satisfies a predicate.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @param fn - The predicate function to invoke for each value.
   *
   * @returns `true` if any value passes the test, `false` otherwise.
   *
   * #### Notes
   * Iteration terminates on the first `true` predicate result.
   *
   * #### Complexity
   * Linear.
   *
   * #### Example
   * ```typescript
   * import { some } from '@lumino/algorithm';
   *
   * let data = [5, 7, 1];
   *
   * some(data, value => value === 7);  // true
   * some(data, value => value === 3);  // false
   * ```
   */
  function some$1(object, fn) {
      var index = 0;
      var it = iter(object);
      var value;
      while ((value = it.next()) !== undefined) {
          if (fn(value, index++)) {
              return true;
          }
      }
      return false;
  }
  /**
   * Create an array from an iterable of values.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @returns A new array of values from the given object.
   *
   * #### Example
   * ```typescript
   * import { iter, toArray } from '@lumino/algorithm';
   *
   * let data = [1, 2, 3, 4, 5, 6];
   *
   * let stream = iter(data);
   *
   * toArray(stream);  // [1, 2, 3, 4, 5, 6];
   * ```
   */
  function toArray$1(object) {
      var index = 0;
      var result = [];
      var it = iter(object);
      var value;
      while ((value = it.next()) !== undefined) {
          result[index++] = value;
      }
      return result;
  }
  /**
   * An iterator for an array-like object.
   *
   * #### Notes
   * This iterator can be used for any builtin JS array-like object.
   */
  var ArrayIterator = /** @class */ (function () {
      /**
       * Construct a new array iterator.
       *
       * @param source - The array-like object of interest.
       */
      function ArrayIterator(source) {
          this._index = 0;
          this._source = source;
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      ArrayIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      ArrayIterator.prototype.clone = function () {
          var result = new ArrayIterator(this._source);
          result._index = this._index;
          return result;
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      ArrayIterator.prototype.next = function () {
          if (this._index >= this._source.length) {
              return undefined;
          }
          return this._source[this._index++];
      };
      return ArrayIterator;
  }());

  // Copyright (c) Jupyter Development Team.
  /**
   * Chain together several iterables.
   *
   * @param objects - The iterable or array-like objects of interest.
   *
   * @returns An iterator which yields the values of the iterables
   *   in the order in which they are supplied.
   *
   * #### Example
   * ```typescript
   * import { chain, toArray } from '@lumino/algorithm';
   *
   * let data1 = [1, 2, 3];
   * let data2 = [4, 5, 6];
   *
   * let stream = chain(data1, data2);
   *
   * toArray(stream);  // [1, 2, 3, 4, 5, 6]
   * ```
   */
  function chain$1() {
      var objects = [];
      for (var _i = 0; _i < arguments.length; _i++) {
          objects[_i] = arguments[_i];
      }
      return new ChainIterator(iter(objects.map(iter)));
  }
  /**
   * An iterator which chains together several iterators.
   */
  var ChainIterator = /** @class */ (function () {
      /**
       * Construct a new chain iterator.
       *
       * @param source - The iterator of iterators of interest.
       */
      function ChainIterator(source) {
          this._cloned = false;
          this._source = source;
          this._active = undefined;
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      ChainIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      ChainIterator.prototype.clone = function () {
          var result = new ChainIterator(this._source.clone());
          result._active = this._active && this._active.clone();
          result._cloned = true;
          this._cloned = true;
          return result;
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      ChainIterator.prototype.next = function () {
          if (this._active === undefined) {
              var active = this._source.next();
              if (active === undefined) {
                  return undefined;
              }
              this._active = this._cloned ? active.clone() : active;
          }
          var value = this._active.next();
          if (value !== undefined) {
              return value;
          }
          this._active = undefined;
          return this.next();
      };
      return ChainIterator;
  }());

  /**
   * Create an empty iterator.
   *
   * @returns A new iterator which yields nothing.
   *
   * #### Example
   * ```typescript
   * import { empty, toArray } from '@lumino/algorithm';
   *
   * let stream = empty<number>();
   *
   * toArray(stream);  // []
   * ```
   */
  function empty() {
      return new EmptyIterator();
  }
  /**
   * An iterator which is always empty.
   */
  var EmptyIterator = /** @class */ (function () {
      /**
       * Construct a new empty iterator.
       */
      function EmptyIterator() {
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      EmptyIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      EmptyIterator.prototype.clone = function () {
          return new EmptyIterator();
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      EmptyIterator.prototype.next = function () {
          return undefined;
      };
      return EmptyIterator;
  }());

  // Copyright (c) Jupyter Development Team.
  /**
   * Filter an iterable for values which pass a test.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @param fn - The predicate function to invoke for each value.
   *
   * @returns An iterator which yields the values which pass the test.
   *
   * #### Example
   * ```typescript
   * import { filter, toArray } from '@lumino/algorithm';
   *
   * let data = [1, 2, 3, 4, 5, 6];
   *
   * let stream = filter(data, value => value % 2 === 0);
   *
   * toArray(stream);  // [2, 4, 6]
   * ```
   */
  function filter$1(object, fn) {
      return new FilterIterator(iter(object), fn);
  }
  /**
   * An iterator which yields values which pass a test.
   */
  var FilterIterator = /** @class */ (function () {
      /**
       * Construct a new filter iterator.
       *
       * @param source - The iterator of values of interest.
       *
       * @param fn - The predicate function to invoke for each value.
       */
      function FilterIterator(source, fn) {
          this._index = 0;
          this._source = source;
          this._fn = fn;
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      FilterIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      FilterIterator.prototype.clone = function () {
          var result = new FilterIterator(this._source.clone(), this._fn);
          result._index = this._index;
          return result;
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      FilterIterator.prototype.next = function () {
          var fn = this._fn;
          var it = this._source;
          var value;
          while ((value = it.next()) !== undefined) {
              if (fn(value, this._index++)) {
                  return value;
              }
          }
          return undefined;
      };
      return FilterIterator;
  }());

  // Copyright (c) Jupyter Development Team.
  /**
   * Find the first value in an iterable which matches a predicate.
   *
   * @param object - The iterable or array-like object to search.
   *
   * @param fn - The predicate function to apply to the values.
   *
   * @returns The first matching value, or `undefined` if no matching
   *   value is found.
   *
   * #### Complexity
   * Linear.
   *
   * #### Example
   * ```typescript
   * import { find } from '@lumino/algorithm';
   *
   * interface IAnimal { species: string, name: string };
   *
   * function isCat(value: IAnimal): boolean {
   *   return value.species === 'cat';
   * }
   *
   * let data: IAnimal[] = [
   *   { species: 'dog', name: 'spot' },
   *   { species: 'cat', name: 'fluffy' },
   *   { species: 'alligator', name: 'pocho' }
   * ];
   *
   * find(data, isCat).name;  // 'fluffy'
   * ```
   */
  function find$1(object, fn) {
      var index = 0;
      var it = iter(object);
      var value;
      while ((value = it.next()) !== undefined) {
          if (fn(value, index++)) {
              return value;
          }
      }
      return undefined;
  }
  /**
   * Find the maximum value in an iterable.
   *
   * @param object - The iterable or array-like object to search.
   *
   * @param fn - The 3-way comparison function to apply to the values.
   *   It should return `< 0` if the first value is less than the second.
   *   `0` if the values are equivalent, or `> 0` if the first value is
   *   greater than the second.
   *
   * @returns The maximum value in the iterable. If multiple values are
   *   equivalent to the maximum, the left-most value is returned. If
   *   the iterable is empty, this returns `undefined`.
   *
   * #### Complexity
   * Linear.
   *
   * #### Example
   * ```typescript
   * import { max } from '@lumino/algorithm';
   *
   * function numberCmp(a: number, b: number): number {
   *   return a - b;
   * }
   *
   * max([7, 4, 0, 3, 9, 4], numberCmp);  // 9
   * ```
   */
  function max$1(object, fn) {
      var it = iter(object);
      var value = it.next();
      if (value === undefined) {
          return undefined;
      }
      var result = value;
      while ((value = it.next()) !== undefined) {
          if (fn(value, result) > 0) {
              result = value;
          }
      }
      return result;
  }

  // Copyright (c) Jupyter Development Team.
  /**
   * Transform the values of an iterable with a mapping function.
   *
   * @param object - The iterable or array-like object of interest.
   *
   * @param fn - The mapping function to invoke for each value.
   *
   * @returns An iterator which yields the transformed values.
   *
   * #### Example
   * ```typescript
   * import { map, toArray } from '@lumino/algorithm';
   *
   * let data = [1, 2, 3];
   *
   * let stream = map(data, value => value * 2);
   *
   * toArray(stream);  // [2, 4, 6]
   * ```
   */
  function map$1(object, fn) {
      return new MapIterator(iter(object), fn);
  }
  /**
   * An iterator which transforms values using a mapping function.
   */
  var MapIterator = /** @class */ (function () {
      /**
       * Construct a new map iterator.
       *
       * @param source - The iterator of values of interest.
       *
       * @param fn - The mapping function to invoke for each value.
       */
      function MapIterator(source, fn) {
          this._index = 0;
          this._source = source;
          this._fn = fn;
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      MapIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      MapIterator.prototype.clone = function () {
          var result = new MapIterator(this._source.clone(), this._fn);
          result._index = this._index;
          return result;
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      MapIterator.prototype.next = function () {
          var value = this._source.next();
          if (value === undefined) {
              return undefined;
          }
          return this._fn.call(undefined, value, this._index++);
      };
      return MapIterator;
  }());
  /**
   * The namespace for the module implementation details.
   */
  var Private;
  (function (Private) {
      /**
       * Compute the effective length of a range.
       *
       * @param start - The starting value for the range, inclusive.
       *
       * @param stop - The stopping value for the range, exclusive.
       *
       * @param step - The distance between each value.
       *
       * @returns The number of steps need to traverse the range.
       */
      function rangeLength(start, stop, step) {
          if (step === 0) {
              return Infinity;
          }
          if (start > stop && step > 0) {
              return 0;
          }
          if (start < stop && step < 0) {
              return 0;
          }
          return Math.ceil((stop - start) / step);
      }
      Private.rangeLength = rangeLength;
  })(Private || (Private = {}));

  // Copyright (c) Jupyter Development Team.
  function reduce$1(object, fn, initial) {
      // Setup the iterator and fetch the first value.
      var index = 0;
      var it = iter(object);
      var first = it.next();
      // An empty iterator and no initial value is an error.
      if (first === undefined && initial === undefined) {
          throw new TypeError('Reduce of empty iterable with no initial value.');
      }
      // If the iterator is empty, return the initial value.
      if (first === undefined) {
          return initial;
      }
      // If the iterator has a single item and no initial value, the
      // reducer is not invoked and the first item is the return value.
      var second = it.next();
      if (second === undefined && initial === undefined) {
          return first;
      }
      // If iterator has a single item and an initial value is provided,
      // the reducer is invoked and that result is the return value.
      if (second === undefined) {
          return fn(initial, first, index++);
      }
      // Setup the initial accumlated value.
      var accumulator;
      if (initial === undefined) {
          accumulator = fn(first, second, index++);
      }
      else {
          accumulator = fn(fn(initial, first, index++), second, index++);
      }
      // Iterate the rest of the values, updating the accumulator.
      var next;
      while ((next = it.next()) !== undefined) {
          accumulator = fn(accumulator, next, index++);
      }
      // Return the final accumulated value.
      return accumulator;
  }
  /**
   * Create an iterator which yields a value a single time.
   *
   * @param value - The value to wrap in an iterator.
   *
   * @returns A new iterator which yields the value a single time.
   *
   * #### Example
   * ```typescript
   * import { once, toArray } from '@lumino/algorithm';
   *
   * let stream = once(7);
   *
   * toArray(stream);  // [7]
   * ```
   */
  function once$1(value) {
      return new RepeatIterator(value, 1);
  }
  /**
   * An iterator which repeats a value a specified number of times.
   */
  var RepeatIterator = /** @class */ (function () {
      /**
       * Construct a new repeat iterator.
       *
       * @param value - The value to repeat.
       *
       * @param count - The number of times to repeat the value.
       */
      function RepeatIterator(value, count) {
          this._value = value;
          this._count = count;
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      RepeatIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      RepeatIterator.prototype.clone = function () {
          return new RepeatIterator(this._value, this._count);
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      RepeatIterator.prototype.next = function () {
          if (this._count <= 0) {
              return undefined;
          }
          this._count--;
          return this._value;
      };
      return RepeatIterator;
  }());

  /**
   * Create an iterator for a retroable object.
   *
   * @param object - The retroable or array-like object of interest.
   *
   * @returns An iterator which traverses the object's values in reverse.
   *
   * #### Example
   * ```typescript
   * import { retro, toArray } from '@lumino/algorithm';
   *
   * let data = [1, 2, 3, 4, 5, 6];
   *
   * let stream = retro(data);
   *
   * toArray(stream);  // [6, 5, 4, 3, 2, 1]
   * ```
   */
  function retro(object) {
      var it;
      if (typeof object.retro === 'function') {
          it = object.retro();
      }
      else {
          it = new RetroArrayIterator(object);
      }
      return it;
  }
  /**
   * An iterator which traverses an array-like object in reverse.
   *
   * #### Notes
   * This iterator can be used for any builtin JS array-like object.
   */
  var RetroArrayIterator = /** @class */ (function () {
      /**
       * Construct a new retro iterator.
       *
       * @param source - The array-like object of interest.
       */
      function RetroArrayIterator(source) {
          this._source = source;
          this._index = source.length - 1;
      }
      /**
       * Get an iterator over the object's values.
       *
       * @returns An iterator which yields the object's values.
       */
      RetroArrayIterator.prototype.iter = function () {
          return this;
      };
      /**
       * Create an independent clone of the iterator.
       *
       * @returns A new independent clone of the iterator.
       */
      RetroArrayIterator.prototype.clone = function () {
          var result = new RetroArrayIterator(this._source);
          result._index = this._index;
          return result;
      };
      /**
       * Get the next value from the iterator.
       *
       * @returns The next value from the iterator, or `undefined`.
       */
      RetroArrayIterator.prototype.next = function () {
          if (this._index < 0 || this._index >= this._source.length) {
              return undefined;
          }
          return this._source[this._index--];
      };
      return RetroArrayIterator;
  }());

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for string-specific algorithms.
   */
  var StringExt;
  (function (StringExt) {
      /**
       * Find the indices of characters in a source text.
       *
       * @param source - The source text which should be searched.
       *
       * @param query - The characters to locate in the source text.
       *
       * @param start - The index to start the search.
       *
       * @returns The matched indices, or `null` if there is no match.
       *
       * #### Complexity
       * Linear on `sourceText`.
       *
       * #### Notes
       * In order for there to be a match, all of the characters in `query`
       * **must** appear in `source` in the order given by `query`.
       *
       * Characters are matched using strict `===` equality.
       */
      function findIndices(source, query, start) {
          if (start === void 0) { start = 0; }
          var indices = new Array(query.length);
          for (var i = 0, j = start, n = query.length; i < n; ++i, ++j) {
              j = source.indexOf(query[i], j);
              if (j === -1) {
                  return null;
              }
              indices[i] = j;
          }
          return indices;
      }
      StringExt.findIndices = findIndices;
      /**
       * A string matcher which uses a sum-of-squares algorithm.
       *
       * @param source - The source text which should be searched.
       *
       * @param query - The characters to locate in the source text.
       *
       * @param start - The index to start the search.
       *
       * @returns The match result, or `null` if there is no match.
       *   A lower `score` represents a stronger match.
       *
       * #### Complexity
       * Linear on `sourceText`.
       *
       * #### Notes
       * This scoring algorithm uses a sum-of-squares approach to determine
       * the score. In order for there to be a match, all of the characters
       * in `query` **must** appear in `source` in order. The index of each
       * matching character is squared and added to the score. This means
       * that early and consecutive character matches are preferred, while
       * late matches are heavily penalized.
       */
      function matchSumOfSquares(source, query, start) {
          if (start === void 0) { start = 0; }
          var indices = findIndices(source, query, start);
          if (!indices) {
              return null;
          }
          var score = 0;
          for (var i = 0, n = indices.length; i < n; ++i) {
              var j = indices[i] - start;
              score += j * j;
          }
          return { score: score, indices: indices };
      }
      StringExt.matchSumOfSquares = matchSumOfSquares;
      /**
       * A string matcher which uses a sum-of-deltas algorithm.
       *
       * @param source - The source text which should be searched.
       *
       * @param query - The characters to locate in the source text.
       *
       * @param start - The index to start the search.
       *
       * @returns The match result, or `null` if there is no match.
       *   A lower `score` represents a stronger match.
       *
       * #### Complexity
       * Linear on `sourceText`.
       *
       * #### Notes
       * This scoring algorithm uses a sum-of-deltas approach to determine
       * the score. In order for there to be a match, all of the characters
       * in `query` **must** appear in `source` in order. The delta between
       * the indices are summed to create the score. This means that groups
       * of matched characters are preferred, while fragmented matches are
       * penalized.
       */
      function matchSumOfDeltas(source, query, start) {
          if (start === void 0) { start = 0; }
          var indices = findIndices(source, query, start);
          if (!indices) {
              return null;
          }
          var score = 0;
          var last = start - 1;
          for (var i = 0, n = indices.length; i < n; ++i) {
              var j = indices[i];
              score += j - last - 1;
              last = j;
          }
          return { score: score, indices: indices };
      }
      StringExt.matchSumOfDeltas = matchSumOfDeltas;
      /**
       * Highlight the matched characters of a source text.
       *
       * @param source - The text which should be highlighted.
       *
       * @param indices - The indices of the matched characters. They must
       *   appear in increasing order and must be in bounds of the source.
       *
       * @param fn - The function to apply to the matched chunks.
       *
       * @returns An array of unmatched and highlighted chunks.
       */
      function highlight(source, indices, fn) {
          // Set up the result array.
          var result = [];
          // Set up the counter variables.
          var k = 0;
          var last = 0;
          var n = indices.length;
          // Iterator over each index.
          while (k < n) {
              // Set up the chunk indices.
              var i = indices[k];
              var j = indices[k];
              // Advance the right chunk index until it's non-contiguous.
              while (++k < n && indices[k] === j + 1) {
                  j++;
              }
              // Extract the unmatched text.
              if (last < i) {
                  result.push(source.slice(last, i));
              }
              // Extract and highlight the matched text.
              if (i < j + 1) {
                  result.push(fn(source.slice(i, j + 1)));
              }
              // Update the last visited index.
              last = j + 1;
          }
          // Extract any remaining unmatched text.
          if (last < source.length) {
              result.push(source.slice(last));
          }
          // Return the highlighted result.
          return result;
      }
      StringExt.highlight = highlight;
      /**
       * A 3-way string comparison function.
       *
       * @param a - The first string of interest.
       *
       * @param b - The second string of interest.
       *
       * @returns `-1` if `a < b`, else `1` if `a > b`, else `0`.
       */
      function cmp(a, b) {
          return a < b ? -1 : a > b ? 1 : 0;
      }
      StringExt.cmp = cmp;
  })(StringExt || (StringExt = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2019, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for clipboard related functionality.
   */
  var ClipboardExt;
  (function (ClipboardExt) {
      /**
       * Copy text to the system clipboard.
       *
       * @param text - The text to copy to the clipboard.
       */
      function copyText(text) {
          // Fetch the document body.
          var body = document.body;
          // Set up the clipboard event listener.
          var handler = function (event) {
              // Stop the event propagation.
              event.preventDefault();
              event.stopPropagation();
              // Set the clipboard data.
              event.clipboardData.setData('text', text);
              // Remove the event listener.
              body.removeEventListener('copy', handler, true);
          };
          // Add the event listener.
          body.addEventListener('copy', handler, true);
          // Trigger the event.
          document.execCommand('copy');
      }
      ClipboardExt.copyText = copyText;
  })(ClipboardExt || (ClipboardExt = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for element related utilities.
   */
  var ElementExt;
  (function (ElementExt) {
      /**
       * Compute the box sizing for an element.
       *
       * @param element - The element of interest.
       *
       * @returns The box sizing data for the specified element.
       */
      function boxSizing(element) {
          var style = window.getComputedStyle(element);
          var bt = parseFloat(style.borderTopWidth) || 0;
          var bl = parseFloat(style.borderLeftWidth) || 0;
          var br = parseFloat(style.borderRightWidth) || 0;
          var bb = parseFloat(style.borderBottomWidth) || 0;
          var pt = parseFloat(style.paddingTop) || 0;
          var pl = parseFloat(style.paddingLeft) || 0;
          var pr = parseFloat(style.paddingRight) || 0;
          var pb = parseFloat(style.paddingBottom) || 0;
          var hs = bl + pl + pr + br;
          var vs = bt + pt + pb + bb;
          return {
              borderTop: bt,
              borderLeft: bl,
              borderRight: br,
              borderBottom: bb,
              paddingTop: pt,
              paddingLeft: pl,
              paddingRight: pr,
              paddingBottom: pb,
              horizontalSum: hs,
              verticalSum: vs
          };
      }
      ElementExt.boxSizing = boxSizing;
      /**
       * Compute the size limits for an element.
       *
       * @param element - The element of interest.
       *
       * @returns The size limit data for the specified element.
       */
      function sizeLimits(element) {
          var style = window.getComputedStyle(element);
          var minWidth = parseFloat(style.minWidth) || 0;
          var minHeight = parseFloat(style.minHeight) || 0;
          var maxWidth = parseFloat(style.maxWidth) || Infinity;
          var maxHeight = parseFloat(style.maxHeight) || Infinity;
          maxWidth = Math.max(minWidth, maxWidth);
          maxHeight = Math.max(minHeight, maxHeight);
          return { minWidth: minWidth, minHeight: minHeight, maxWidth: maxWidth, maxHeight: maxHeight };
      }
      ElementExt.sizeLimits = sizeLimits;
      /**
       * Test whether a client position lies within an element.
       *
       * @param element - The DOM element of interest.
       *
       * @param clientX - The client X coordinate of interest.
       *
       * @param clientY - The client Y coordinate of interest.
       *
       * @returns Whether the point is within the given element.
       */
      function hitTest(element, clientX, clientY) {
          var rect = element.getBoundingClientRect();
          return (clientX >= rect.left &&
              clientX < rect.right &&
              clientY >= rect.top &&
              clientY < rect.bottom);
      }
      ElementExt.hitTest = hitTest;
      /**
       * Vertically scroll an element into view if needed.
       *
       * @param area - The scroll area element.
       *
       * @param element - The element of interest.
       *
       * #### Notes
       * This follows the "nearest" behavior of the native `scrollIntoView`
       * method, which is not supported by all browsers.
       * https://drafts.csswg.org/cssom-view/#element-scrolling-members
       *
       * If the element fully covers the visible area or is fully contained
       * within the visible area, no scrolling will take place. Otherwise,
       * the nearest edges of the area and element are aligned.
       */
      function scrollIntoViewIfNeeded(area, element) {
          var ar = area.getBoundingClientRect();
          var er = element.getBoundingClientRect();
          if (er.top <= ar.top && er.bottom >= ar.bottom) {
              return;
          }
          if (er.top < ar.top && er.height <= ar.height) {
              area.scrollTop -= ar.top - er.top;
              return;
          }
          if (er.bottom > ar.bottom && er.height >= ar.height) {
              area.scrollTop -= ar.top - er.top;
              return;
          }
          if (er.top < ar.top && er.height > ar.height) {
              area.scrollTop -= ar.bottom - er.bottom;
              return;
          }
          if (er.bottom > ar.bottom && er.height < ar.height) {
              area.scrollTop -= ar.bottom - er.bottom;
              return;
          }
      }
      ElementExt.scrollIntoViewIfNeeded = scrollIntoViewIfNeeded;
  })(ElementExt || (ElementExt = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for platform related utilities.
   */
  var Platform;
  (function (Platform) {
      /**
       * A flag indicating whether the platform is Mac.
       */
      Platform.IS_MAC = !!navigator.platform.match(/Mac/i);
      /**
       * A flag indicating whether the platform is Windows.
       */
      Platform.IS_WIN = !!navigator.platform.match(/Win/i);
      /**
       * A flag indicating whether the browser is IE.
       */
      Platform.IS_IE = /Trident/.test(navigator.userAgent);
      /**
       * A flag indicating whether the browser is Edge.
       */
      Platform.IS_EDGE = /Edge/.test(navigator.userAgent);
      /**
       * Test whether the `accel` key is pressed.
       *
       * @param event - The keyboard or mouse event of interest.
       *
       * @returns Whether the `accel` key is pressed.
       *
       * #### Notes
       * On Mac the `accel` key is the command key. On all other
       * platforms the `accel` key is the control key.
       */
      function accelKey(event) {
          return Platform.IS_MAC ? event.metaKey : event.ctrlKey;
      }
      Platform.accelKey = accelKey;
  })(Platform || (Platform = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * The namespace for selector related utilities.
   */
  var Selector;
  (function (Selector) {
      /**
       * Calculate the specificity of a single CSS selector.
       *
       * @param selector - The CSS selector of interest.
       *
       * @returns The specificity of the selector.
       *
       * #### Undefined Behavior
       * The selector is invalid.
       *
       * #### Notes
       * This is based on https://www.w3.org/TR/css3-selectors/#specificity
       *
       * A larger number represents a more specific selector.
       *
       * The smallest possible specificity is `0`.
       *
       * The result is represented as a hex number `0x<aa><bb><cc>` where
       * each component is the count of the respective selector clause.
       *
       * If the selector contains commas, only the first clause is used.
       *
       * The computed result is cached, so subsequent calculations for the
       * same selector are extremely fast.
       */
      function calculateSpecificity(selector) {
          if (selector in Private$1.specificityCache) {
              return Private$1.specificityCache[selector];
          }
          var result = Private$1.calculateSingle(selector);
          return Private$1.specificityCache[selector] = result;
      }
      Selector.calculateSpecificity = calculateSpecificity;
      /**
       * Test whether a selector is a valid CSS selector.
       *
       * @param selector - The CSS selector of interest.
       *
       * @returns `true` if the selector is valid, `false` otherwise.
       *
       * #### Notes
       * The computed result is cached, so subsequent tests for the same
       * selector are extremely fast.
       */
      function isValid(selector) {
          if (selector in Private$1.validityCache) {
              return Private$1.validityCache[selector];
          }
          var result = true;
          try {
              Private$1.testElem.querySelector(selector);
          }
          catch (err) {
              result = false;
          }
          return Private$1.validityCache[selector] = result;
      }
      Selector.isValid = isValid;
      /**
       * Test whether an element matches a CSS selector.
       *
       * @param element - The element of interest.
       *
       * @param selector - The valid CSS selector of interest.
       *
       * @returns `true` if the element is a match, `false` otherwise.
       *
       * #### Notes
       * This function uses the builtin browser capabilities when possible,
       * falling back onto a document query otherwise.
       */
      function matches(element, selector) {
          return Private$1.protoMatchFunc.call(element, selector);
      }
      Selector.matches = matches;
  })(Selector || (Selector = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$1;
  (function (Private) {
      /**
       * A cache of computed selector specificity values.
       */
      Private.specificityCache = Object.create(null);
      /**
       * A cache of computed selector validity.
       */
      Private.validityCache = Object.create(null);
      /**
       * An empty element for testing selector validity.
       */
      Private.testElem = document.createElement('div');
      /**
       * A cross-browser CSS selector matching prototype function.
       */
      Private.protoMatchFunc = (function () {
          var proto = Element.prototype;
          return (proto.matches ||
              proto.matchesSelector ||
              proto.mozMatchesSelector ||
              proto.msMatchesSelector ||
              proto.oMatchesSelector ||
              proto.webkitMatchesSelector ||
              (function (selector) {
                  var elem = this;
                  var matches = elem.ownerDocument ? elem.ownerDocument.querySelectorAll(selector) : [];
                  return Array.prototype.indexOf.call(matches, elem) !== -1;
              }));
      })();
      /**
       * Calculate the specificity of a single selector.
       *
       * The behavior is undefined if the selector is invalid.
       */
      function calculateSingle(selector) {
          // Ignore anything after the first comma.
          selector = selector.split(',', 1)[0];
          // Setup the aggregate counters.
          var a = 0;
          var b = 0;
          var c = 0;
          // Apply a regex to the front of the selector. If it succeeds, that
          // portion of the selector is removed. Returns a success/fail flag.
          function match(re) {
              var match = selector.match(re);
              if (match === null) {
                  return false;
              }
              selector = selector.slice(match[0].length);
              return true;
          }
          // Replace the negation pseudo-class (which is ignored),
          // but keep its inner content (which is not ignored).
          selector = selector.replace(NEGATION_RE, ' $1 ');
          // Continue matching until the selector is consumed.
          while (selector.length > 0) {
              // Match an ID selector.
              if (match(ID_RE)) {
                  a++;
                  continue;
              }
              // Match a class selector.
              if (match(CLASS_RE)) {
                  b++;
                  continue;
              }
              // Match an attribute selector.
              if (match(ATTR_RE)) {
                  b++;
                  continue;
              }
              // Match a pseudo-element selector. This is done before matching
              // a pseudo-class since this regex overlaps with that regex.
              if (match(PSEUDO_ELEM_RE)) {
                  c++;
                  continue;
              }
              // Match a pseudo-class selector.
              if (match(PSEDUO_CLASS_RE)) {
                  b++;
                  continue;
              }
              // Match a plain type selector.
              if (match(TYPE_RE)) {
                  c++;
                  continue;
              }
              // Finally, match any ignored characters.
              if (match(IGNORE_RE)) {
                  continue;
              }
              // At this point, the selector is assumed to be invalid.
              return 0;
          }
          // Clamp each component to a reasonable base.
          a = Math.min(a, 0xFF);
          b = Math.min(b, 0xFF);
          c = Math.min(c, 0xFF);
          // Combine the components into a single result.
          return (a << 16) | (b << 8) | c;
      }
      Private.calculateSingle = calculateSingle;
      /**
       * A regex which matches an ID selector at string start.
       */
      var ID_RE = /^#[^\s\+>~#\.\[:]+/;
      /**
       * A regex which matches a class selector at string start.
       */
      var CLASS_RE = /^\.[^\s\+>~#\.\[:]+/;
      /**
       * A regex which matches an attribute selector at string start.
       */
      var ATTR_RE = /^\[[^\]]+\]/;
      /**
       * A regex which matches a type selector at string start.
       */
      var TYPE_RE = /^[^\s\+>~#\.\[:]+/;
      /**
       * A regex which matches a pseudo-element selector at string start.
       */
      var PSEUDO_ELEM_RE = /^(::[^\s\+>~#\.\[:]+|:first-line|:first-letter|:before|:after)/;
      /**
       * A regex which matches a pseudo-class selector at string start.
       */
      var PSEDUO_CLASS_RE = /^:[^\s\+>~#\.\[:]+/;
      /**
       * A regex which matches ignored characters at string start.
       */
      var IGNORE_RE = /^[\s\+>~\*]+/;
      /**
       * A regex which matches the negation pseudo-class globally.
       */
      var NEGATION_RE = /:not\(([^\)]+)\)/g;
  })(Private$1 || (Private$1 = {}));

  // Copyright (c) Jupyter Development Team.
  /**
   * A generic B+ tree.
   *
   * #### Notes
   * Most operations have `O(log32 n)` or better complexity.
   */
  var BPlusTree = /** @class */ (function () {
      /**
       * Construct a new B+ tree.
       *
       * @param cmp - The item comparison function for the tree.
       */
      function BPlusTree(cmp) {
          this._root = new Private$2.LeafNode();
          this.cmp = cmp;
      }
      Object.defineProperty(BPlusTree.prototype, "isEmpty", {
          /**
           * Whether the tree is empty.
           *
           * #### Complexity
           * `O(1)`
           */
          get: function () {
              return this._root.size === 0;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BPlusTree.prototype, "size", {
          /**
           * The size of the tree.
           *
           * #### Complexity
           * `O(1)`
           */
          get: function () {
              return this._root.size;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BPlusTree.prototype, "first", {
          /**
           * The first item in the tree.
           *
           * This is `undefined` if the tree is empty.
           *
           * #### Complexity
           * `O(log32 n)`
           */
          get: function () {
              var node = Private$2.firstLeaf(this._root);
              return node.size > 0 ? node.items[0] : undefined;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BPlusTree.prototype, "last", {
          /**
           * The last item in the tree.
           *
           * This is `undefined` if the tree is empty.
           *
           * #### Complexity
           * `O(log32 n)`
           */
          get: function () {
              var node = Private$2.lastLeaf(this._root);
              return node.size > 0 ? node.items[node.size - 1] : undefined;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over the items in the tree.
       *
       * @returns A new iterator starting with the first item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.iter = function () {
          return Private$2.iterItems(this._root);
      };
      /**
       * Create a reverse iterator over the items in the tree.
       *
       * @returns A new iterator starting with the last item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.retro = function () {
          return Private$2.retroItems(this._root);
      };
      /**
       * Create an iterator for a slice of items in the tree.
       *
       * @param start - The index of the first item, inclusive. This
       *   should be `< stop`. Negative values are taken as an offset
       *   from the end of the tree. The default is `0`.
       *
       * @param stop - The index of the last item, exclusive. This
       *   should be `> start`. Negative values are taken as an offset
       *   from the end of the tree. The default is `size`.
       *
       * @returns A new iterator starting with the specified item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.slice = function (start, stop) {
          return Private$2.sliceItems(this._root, start, stop);
      };
      /**
       * Create a reverse iterator for a slice of items in the tree.
       *
       * @param start - The index of the first item, inclusive. This
       *   should be `> stop`. Negative values are taken as an offset
       *   from the end of the tree. The default is `size - 1`.
       *
       * @param stop - The index of the last item, exclusive. This
       *   should be `< start`. Negative values are taken as an offset
       *   from the end of the tree. The default is `-size - 1`.
       *
       * @returns A new reverse iterator starting with the specified item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.retroSlice = function (start, stop) {
          return Private$2.retroSliceItems(this._root, start, stop);
      };
      /**
       * Get the item at a particular index.
       *
       * @param index - The index of the item of interest. Negative
       *   values are taken as an offset from the end of the tree.
       *
       * @returns The item at the specified index, or `undefined` if
       *   the index is out of range.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.at = function (index) {
          return Private$2.itemAt(this._root, index);
      };
      /**
       * Test whether the tree has an item which matches a key.
       *
       * @param key - The key of interest.
       *
       * @param cmp - A function which compares an item against the key.
       *
       * @returns `true` if the tree has an item which matches the given
       *   key, `false` otherwise.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.has = function (key, cmp) {
          return Private$2.hasItem(this._root, key, cmp);
      };
      /**
       * Get the index of an item which matches a key.
       *
       * @param key - The key of interest.
       *
       * @param cmp - A function which compares an item against the key.
       *
       * @returns The index of the item which matches the given key. A
       *   negative value means that a matching item does not exist in
       *   the tree, but if one did it would reside at `-index - 1`.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.indexOf = function (key, cmp) {
          return Private$2.indexOf(this._root, key, cmp);
      };
      /**
       * Get the item which matches a key.
       *
       * @param item - The key of interest.
       *
       * @param cmp - A function which compares an item against the key.
       *
       * @returns The item which matches the given key, or `undefined` if
       *   the tree does not have a matching item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.get = function (key, cmp) {
          return Private$2.getItem(this._root, key, cmp);
      };
      /**
       * Assign new items to the tree, replacing all current items.
       *
       * @param items - The items to assign to the tree.
       *
       * #### Complexity
       * `O(n log32 n)`
       */
      BPlusTree.prototype.assign = function (items) {
          this.clear();
          this.update(items);
      };
      /**
       * Insert an item into the tree.
       *
       * @param item - The item of interest.
       *
       * @returns If the given item matches an existing item in the tree,
       *   the given item will replace it, and the existing item will be
       *   returned. Otherwise, this method returns `undefined`.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.insert = function (item) {
          var existing = Private$2.insertItem(this._root, item, this.cmp);
          this._root = Private$2.maybeSplitRoot(this._root);
          return existing;
      };
      /**
       * Update the tree with multiple items.
       *
       * @param items - The items to insert into the tree.
       *
       * #### Complexity
       * `O(k log32 n)`
       */
      BPlusTree.prototype.update = function (items) {
          var _this = this;
          each$1(items, function (item) { _this.insert(item); });
      };
      /**
       * Delete an item which matches a particular key.
       *
       * @param key - The key of interest.
       *
       * @param cmp - A function which compares an item against the key.
       *
       * @returns The item removed from the tree, or `undefined` if no
       *   item matched the given key.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.delete = function (key, cmp) {
          var item = Private$2.deleteItem(this._root, key, cmp);
          this._root = Private$2.maybeExtractRoot(this._root);
          return item;
      };
      /**
       * Remove an item at a particular index.
       *
       * @param index - The index of the item to remove. Negative
       *   values are taken as an offset from the end of the tree.
       *
       * @returns The item removed from the tree, or `undefined` if
       *   the given index is out of range.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      BPlusTree.prototype.remove = function (index) {
          var item = Private$2.removeItem(this._root, index);
          this._root = Private$2.maybeExtractRoot(this._root);
          return item;
      };
      /**
       * Clear the contents of the tree.
       *
       * #### Complexity
       * `O(n)`
       */
      BPlusTree.prototype.clear = function () {
          Private$2.clear(this._root);
          this._root = new Private$2.LeafNode();
      };
      return BPlusTree;
  }());
  /**
   * The namespace for the `BPlusTree` class statics.
   */
  (function (BPlusTree) {
      /**
       * Create a new B+ tree populated with the given items.
       *
       * @param items - The items to add to the tree.
       *
       * @param cmp - The item comparison function for the tree.
       *
       * @returns A new B+ tree populated with the given items.
       *
       * #### Complexity
       * `O(n log32 n)`
       */
      function from(items, cmp) {
          var tree = new BPlusTree(cmp);
          tree.assign(items);
          return tree;
      }
      BPlusTree.from = from;
  })(BPlusTree || (BPlusTree = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$2;
  (function (Private) {
      /**
       * A branch node in a B+ tree.
       */
      var BranchNode = /** @class */ (function () {
          function BranchNode() {
              /**
               * The left-most item of each child subtree.
               */
              this.items = [];
              /**
               * The cumulative sizes of each child subtree.
               */
              this.sizes = [];
              /**
               * The child nodes of this branch node.
               */
              this.children = [];
          }
          Object.defineProperty(BranchNode.prototype, "type", {
              /**
               * The discriminated type of the node.
               */
              get: function () {
                  return 0 /* Branch */;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(BranchNode.prototype, "size", {
              /**
               * The total number of items in the subtree.
               */
              get: function () {
                  return this.sizes[this.sizes.length - 1];
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(BranchNode.prototype, "width", {
              /**
               * The tree width of the node.
               */
              get: function () {
                  return this.children.length;
              },
              enumerable: true,
              configurable: true
          });
          return BranchNode;
      }());
      Private.BranchNode = BranchNode;
      /**
       * A leaf node in a B+ tree.
       */
      var LeafNode = /** @class */ (function () {
          function LeafNode() {
              /**
               * The next sibling leaf node of this leaf node.
               */
              this.next = null;
              /**
               * The previous sibling leaf node of this leaf node.
               */
              this.prev = null;
              /**
               * The items of the leaf.
               */
              this.items = [];
          }
          Object.defineProperty(LeafNode.prototype, "type", {
              /**
               * The discriminated type of the node.
               */
              get: function () {
                  return 1 /* Leaf */;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(LeafNode.prototype, "size", {
              /**
               * The total number of items in the leaf.
               */
              get: function () {
                  return this.items.length;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(LeafNode.prototype, "width", {
              /**
               * The tree width of the node.
               */
              get: function () {
                  return this.items.length;
              },
              enumerable: true,
              configurable: true
          });
          return LeafNode;
      }());
      Private.LeafNode = LeafNode;
      /**
       * Get the first leaf node in the tree.
       *
       * @param node - The root node of interest.
       *
       * @returns The first leaf node in the tree.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function firstLeaf(node) {
          while (node.type === 0 /* Branch */) {
              node = node.children[0];
          }
          return node;
      }
      Private.firstLeaf = firstLeaf;
      /**
       * Get the last leaf node in the tree.
       *
       * @param node - The root node of interest.
       *
       * @returns The last leaf node in the tree.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function lastLeaf(node) {
          while (node.type === 0 /* Branch */) {
              node = node.children[node.children.length - 1];
          }
          return node;
      }
      Private.lastLeaf = lastLeaf;
      /**
       * Create a forward iterator for the items in the tree.
       *
       * @param node - The root node of interest.
       *
       * @returns A new forward iterator starting with the first item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function iterItems(node) {
          var leaf = firstLeaf(node);
          return new ForwardIterator(leaf, 0, -1);
      }
      Private.iterItems = iterItems;
      /**
       * Create a reverse iterator for the items in the tree.
       *
       * @param node - The root node of interest.
       *
       * @returns A new reverse iterator starting with the last item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function retroItems(node) {
          var leaf = lastLeaf(node);
          return new RetroIterator(leaf, leaf.size - 1, -1);
      }
      Private.retroItems = retroItems;
      /**
       * Create an iterator for a slice of items in the tree.
       *
       * @param node - The root node of interest.
       *
       * @param start - The index of the first item, inclusive. This
       *   should be `< stop`. Negative values are taken as an offset
       *   from the end of the tree. The default is `0`.
       *
       * @param stop - The index of the last item, exclusive. This
       *   should be `> start`. Negative values are taken as an offset
       *   from the end of the tree. The default is `size`.
       *
       * @returns A new iterator starting with the specified item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function sliceItems(node, start, stop) {
          // Normalize the start index.
          if (start === undefined) {
              start = 0;
          }
          else if (start < 0) {
              start = Math.max(0, start + node.size);
          }
          else {
              start = Math.min(start, node.size);
          }
          // Normalize the stop index.
          if (stop === undefined) {
              stop = node.size;
          }
          else if (stop < 0) {
              stop = Math.max(0, stop + node.size);
          }
          else {
              stop = Math.min(stop, node.size);
          }
          // Compute effective count.
          var count = Math.max(0, stop - start);
          // Bail early if there is nothing to iterate.
          if (count === 0) {
              return empty();
          }
          // Find the starting leaf node and local index.
          while (node.type === 0 /* Branch */) {
              var i = findPivotIndexByIndex(node.sizes, start);
              if (i > 0)
                  start -= node.sizes[i - 1];
              node = node.children[i];
          }
          // Return the forward iterator for the range.
          return new ForwardIterator(node, start, count);
      }
      Private.sliceItems = sliceItems;
      /**
       * Create a reverse iterator for a slice of items in the tree.
       *
       * @param node - The root node of interest.
       *
       * @param start - The index of the first item, inclusive. This
       *   should be `> stop`. Negative values are taken as an offset
       *   from the end of the tree. The default is `size - 1`.
       *
       * @param stop - The index of the last item, exclusive. This
       *   should be `< start`. Negative values are taken as an offset
       *   from the end of the tree. The default is `-size - 1`.
       *
       * @returns A new reverse iterator starting with the specified item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function retroSliceItems(node, start, stop) {
          // Normalize the start index.
          if (start === undefined) {
              start = node.size - 1;
          }
          else if (start < 0) {
              start = Math.max(-1, start + node.size);
          }
          else {
              start = Math.min(start, node.size - 1);
          }
          // Normalize the stop index.
          if (stop === undefined) {
              stop = -1;
          }
          else if (stop < 0) {
              stop = Math.max(-1, stop + node.size);
          }
          else {
              stop = Math.min(stop, node.size - 1);
          }
          // Compute the effective count.
          var count = Math.max(0, start - stop);
          // Bail early if there is nothing to iterate.
          if (count === 0) {
              return empty();
          }
          // Find the starting leaf node and local index.
          while (node.type === 0 /* Branch */) {
              var i = findPivotIndexByIndex(node.sizes, start);
              if (i > 0)
                  start -= node.sizes[i - 1];
              node = node.children[i];
          }
          // Return the retro iterator for the range.
          return new RetroIterator(node, start, count);
      }
      Private.retroSliceItems = retroSliceItems;
      /**
       * Get the item at the specified index.
       *
       * @param node - The root node of interest.
       *
       * @param index - The index of the item of interest. Negative
       *   values are taken as an offset from the end of the tree.
       *
       * @returns The item at the specified index, or `undefined` if
       *   the index is out of range.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function itemAt(node, index) {
          // Wrap negative indices.
          if (index < 0) {
              index += node.size;
          }
          // Bail early if the index is out of range.
          if (index < 0 || index >= node.size) {
              return undefined;
          }
          // Find the containing leaf node and local index.
          while (node.type === 0 /* Branch */) {
              var i = findPivotIndexByIndex(node.sizes, index);
              if (i > 0)
                  index -= node.sizes[i - 1];
              node = node.children[i];
          }
          // Return the item at the specified index.
          return node.items[index];
      }
      Private.itemAt = itemAt;
      /**
       * Test whether the tree contains an item which matches a key.
       *
       * @param node - The root node of interest.
       *
       * @param key - The key of interest.
       *
       * @param cmp - The key comparison function.
       *
       * @returns Whether the tree contains a matching item.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function hasItem(node, key, cmp) {
          // Find the containing leaf node.
          while (node.type === 0 /* Branch */) {
              var i_1 = findPivotIndexByKey(node.items, key, cmp);
              node = node.children[i_1];
          }
          // Find the key index.
          var i = findKeyIndex(node.items, key, cmp);
          // Return whether or not the node contains a matching item.
          return i >= 0;
      }
      Private.hasItem = hasItem;
      /**
       * Get the index of the item which matches a key.
       *
       * @param node - The node of interest.
       *
       * @param key - The key of interest.
       *
       * @param cmp - The key comparison function.
       *
       * @returns The index of the item which matches the given key. A
       *   negative value means that a matching item does not exist in
       *   the tree, but if one did it would reside at `-index - 1`.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function indexOf(node, key, cmp) {
          // Set up the global index.
          var index = 0;
          // Find the containing leaf node and global index.
          while (node.type === 0 /* Branch */) {
              var i_2 = findPivotIndexByKey(node.items, key, cmp);
              if (i_2 > 0)
                  index += node.sizes[i_2 - 1];
              node = node.children[i_2];
          }
          // Find the key index.
          var i = findKeyIndex(node.items, key, cmp);
          // Return the final computed index.
          return i >= 0 ? index + i : -index + i;
      }
      Private.indexOf = indexOf;
      /**
       * Get the item for a particular key.
       *
       * @param node - The node of interest.
       *
       * @param key - The key of interest.
       *
       * @param cmp - The key comparison function.
       *
       * @returns The item for the specified key, or `undefined` if
       *   the tree does not have a matching item for the key.
       *
       * #### Complexity
       * `O(log32 n)`
       */
      function getItem(node, key, cmp) {
          // Find the containing leaf node.
          while (node.type === 0 /* Branch */) {
              var i_3 = findPivotIndexByKey(node.items, key, cmp);
              node = node.children[i_3];
          }
          // Find the key index.
          var i = findKeyIndex(node.items, key, cmp);
          // Return the item for the given key.
          return i >= 0 ? node.items[i] : undefined;
      }
      Private.getItem = getItem;
      /**
       * Insert an item into the tree.
       *
       * @param node - The root node of interest.
       *
       * @param item - The item of interest.
       *
       * @param cmp - The item comparison function.
       *
       * @returns If the given item matches an existing item in the tree,
       *   the given item will replace it, and the existing item will be
       *   returned. Otherwise, this function returns `undefined`.
       *
       * #### Complexity
       * `O(log32 n)`
       *
       * #### Notes
       * The root may be overfull after calling this function.
       */
      function insertItem(node, item, cmp) {
          // Handle leaf nodes first.
          if (node.type === 1 /* Leaf */) {
              // Find the index for the given item.
              var i_4 = findKeyIndex(node.items, item, cmp);
              // Fetch the existing item and insert the new item.
              var existing_1;
              if (i_4 >= 0) {
                  existing_1 = node.items[i_4];
                  node.items[i_4] = item;
              }
              else {
                  existing_1 = undefined;
                  ArrayExt.insert(node.items, -i_4 - 1, item);
              }
              // Return the existing item.
              return existing_1;
          }
          // Find the pivot index for the insert.
          var i = findPivotIndexByKey(node.items, item, cmp);
          // Fetch the pivot child.
          var child = node.children[i];
          // Fetch the current size of the child.
          var prevSize = child.size;
          // Recursively insert the item into the child.
          var existing = insertItem(child, item, cmp);
          // Fetch the updated size of the child.
          var currSize = child.size;
          // Update the item state of the branch.
          node.items[i] = child.items[0];
          // Bail early if the child size did not change.
          if (prevSize === currSize) {
              return existing;
          }
          // Split the child if it's overfull.
          if (child.width > MAX_NODE_WIDTH) {
              var next = splitNode(child);
              ArrayExt.insert(node.children, i + 1, next);
              ArrayExt.insert(node.items, i + 1, next.items[0]);
          }
          // Update the dirty sizes of the branch.
          updateSizes(node, i);
          // Return the existing item.
          return existing;
      }
      Private.insertItem = insertItem;
      /**
       * Delete an item in the tree.
       *
       * @param node - The node of interest.
       *
       * @param key - The key of interest.
       *
       * @param cmp - The key comparison function.
       *
       * @returns The deleted item or `undefined`.
       *
       * #### Complexity
       * `O(log32 n)`
       *
       * #### Notes
       * The root may be underfull after calling this function.
       */
      function deleteItem(node, key, cmp) {
          // Handle leaf nodes first.
          if (node.type === 1 /* Leaf */) {
              // Find the index for the given key.
              var i_5 = findKeyIndex(node.items, key, cmp);
              // Bail early if the item does not exist.
              if (i_5 < 0) {
                  return undefined;
              }
              // Remove the item at the computed index.
              return ArrayExt.removeAt(node.items, i_5);
          }
          // Find the pivot index for the delete.
          var i = findPivotIndexByKey(node.items, key, cmp);
          // Fetch the pivot child.
          var child = node.children[i];
          // Fetch the current size of the child.
          var prevSize = child.size;
          // Recursively remove the item from the child.
          var item = deleteItem(child, key, cmp);
          // Fetch the updated size of the child.
          var currSize = child.size;
          // Bail early if the child size did not change.
          if (prevSize === currSize) {
              return item;
          }
          // Update the item state of the branch.
          node.items[i] = child.items[0];
          // Join the child if it's underfull.
          if (child.width < MIN_NODE_WIDTH) {
              i = joinChild(node, i);
          }
          // Update the dirty sizes of the branch.
          updateSizes(node, i);
          // Return the deleted item.
          return item;
      }
      Private.deleteItem = deleteItem;
      /**
       * Remove an item from the tree.
       *
       * @param node - The node of interest.
       *
       * @param index - The index of interest.
       *
       * @returns The removed item or `undefined`.
       *
       * #### Complexity
       * `O(log32 n)`
       *
       * #### Notes
       * The root may be underfull after calling this function.
       */
      function removeItem(node, index) {
          // Wrap negative indices.
          if (index < 0) {
              index += node.size;
          }
          // Bail early if the index is out of range.
          if (index < 0 || index >= node.size) {
              return undefined;
          }
          // Handle leaf nodes first.
          if (node.type === 1 /* Leaf */) {
              return ArrayExt.removeAt(node.items, index);
          }
          // Find the pivot index for the remove.
          var i = findPivotIndexByIndex(node.sizes, index);
          if (i > 0)
              index -= node.sizes[i];
          // Fetch the pivot child.
          var child = node.children[i];
          // Recursively remove the item from the child.
          var item = removeItem(child, index);
          // Update the item state of the branch.
          node.items[i] = child.items[0];
          // Join the child if it's underfull.
          if (child.width < MIN_NODE_WIDTH) {
              i = joinChild(node, i);
          }
          // Update the dirty sizes of the branch.
          updateSizes(node, i);
          // Return the removed item.
          return item;
      }
      Private.removeItem = removeItem;
      /**
       * Recursively clear the contents of a node.
       *
       * @param node - The node of interest.
       *
       * #### Complexity
       * `O(n)`
       */
      function clear(node) {
          if (node.type === 0 /* Branch */) {
              each$1(node.children, clear);
              node.children.length = 0;
              node.sizes.length = 0;
              node.items.length = 0;
          }
          else {
              node.items.length = 0;
              node.next = null;
              node.prev = null;
          }
      }
      Private.clear = clear;
      /**
       * Split a root node and create a new root, if needed.
       *
       * @param node - The root node of interest.
       *
       * @returns The new root node.
       */
      function maybeSplitRoot(node) {
          // Bail early if the current root is not overfull.
          if (node.width <= MAX_NODE_WIDTH) {
              return node;
          }
          // Create a new root branch node.
          var root = new BranchNode();
          // Split the node to the right and create a new sibling.
          var next = splitNode(node);
          // Add the sizes to the root.
          root.sizes[0] = node.size;
          root.sizes[1] = node.size + next.size;
          // Add the children to the root.
          root.children[0] = node;
          root.children[1] = next;
          // Add the items to the root.
          root.items[0] = node.items[0];
          root.items[1] = next.items[0];
          // Return the new root node.
          return root;
      }
      Private.maybeSplitRoot = maybeSplitRoot;
      /**
       * Extract a single node child as a new root, if needed.
       *
       * @param node - The root node of interest.
       *
       * @returns The new root node.
       */
      function maybeExtractRoot(node) {
          // Bail early if the node it already a leaf.
          if (node.type === 1 /* Leaf */) {
              return node;
          }
          // Bail early if the branch has more than one child.
          if (node.children.length > 1) {
              return node;
          }
          // Extract the sole remaining child as the new root.
          var root = node.children.pop();
          // Clear the rest of the node state.
          clear(node);
          // Return the new root.
          return root;
      }
      Private.maybeExtractRoot = maybeExtractRoot;
      /**
       * The maximum width for a node in the tree.
       */
      var MAX_NODE_WIDTH = 32;
      /**
       * The minimum width for a node in the tree.
       */
      var MIN_NODE_WIDTH = MAX_NODE_WIDTH >> 1;
      /**
       * A forward iterator for a B+ tree.
       */
      var ForwardIterator = /** @class */ (function () {
          /**
           * Construct a new forward iterator.
           *
           * @param node - The first leaf node in the chain.
           *
           * @param index - The local index of the first item.
           *
           * @param count - The number of items to iterate. A value `< 0`
           *   will iterate all available items.
           */
          function ForwardIterator(node, index, count) {
              this._node = node;
              this._index = index;
              this._count = count;
          }
          /**
           * Get an iterator over the object's values.
           *
           * @returns An iterator which yields the object's values.
           */
          ForwardIterator.prototype.iter = function () {
              return this;
          };
          /**
           * Create an independent clone of the iterator.
           *
           * @returns A new independent clone of the iterator.
           */
          ForwardIterator.prototype.clone = function () {
              return new ForwardIterator(this._node, this._index, this._count);
          };
          /**
           * Get the next value from the iterator.
           *
           * @returns The next value from the iterator, or `undefined`.
           */
          ForwardIterator.prototype.next = function () {
              if (this._node === null || this._count === 0) {
                  return undefined;
              }
              if (this._index >= this._node.size) {
                  this._node = this._node.next;
                  this._index = 0;
                  return this.next();
              }
              if (this._count > 0) {
                  this._count--;
              }
              return this._node.items[this._index++];
          };
          return ForwardIterator;
      }());
      /**
       * A reverse iterator for a B+ tree.
       */
      var RetroIterator = /** @class */ (function () {
          /**
           * Construct a new retro iterator.
           *
           * @param node - The last leaf node in the chain.
           *
           * @param index - The local index of the last item.
           *
           * @param count - The number of items to iterate. A value `< 0`
           *   will iterate all available items.
           */
          function RetroIterator(node, index, count) {
              this._node = node;
              this._index = index;
              this._count = count;
          }
          /**
           * Get an iterator over the object's values.
           *
           * @returns An iterator which yields the object's values.
           */
          RetroIterator.prototype.iter = function () {
              return this;
          };
          /**
           * Create an independent clone of the iterator.
           *
           * @returns A new independent clone of the iterator.
           */
          RetroIterator.prototype.clone = function () {
              return new RetroIterator(this._node, this._index, this._count);
          };
          /**
           * Get the next value from the iterator.
           *
           * @returns The next value from the iterator, or `undefined`.
           */
          RetroIterator.prototype.next = function () {
              if (this._node === null || this._count === 0) {
                  return undefined;
              }
              if (this._index >= this._node.size) {
                  this._index = this._node.size - 1;
              }
              if (this._index < 0) {
                  this._node = this._node.prev;
                  this._index = this._node ? this._node.size - 1 : -1;
                  return this.next();
              }
              if (this._count > 0) {
                  this._count--;
              }
              return this._node.items[this._index--];
          };
          return RetroIterator;
      }());
      /**
       * Find the pivot index for a particular local index.
       */
      function findPivotIndexByIndex(sizes, index) {
          var n = sizes.length;
          for (var i = 0; i < n; ++i) {
              if (sizes[i] > index) {
                  return i;
              }
          }
          return n - 1;
      }
      /**
       * Find the pivot index for a particular key.
       */
      function findPivotIndexByKey(items, key, cmp) {
          var n = items.length;
          for (var i = 1; i < n; ++i) {
              if (cmp(items[i], key) > 0) {
                  return i - 1;
              }
          }
          return n - 1;
      }
      /**
       * Find the key index for a particular key.
       */
      function findKeyIndex(items, key, cmp) {
          var n = items.length;
          for (var i = 0; i < n; ++i) {
              var c = cmp(items[i], key);
              if (c === 0) {
                  return i;
              }
              if (c > 0) {
                  return -i - 1;
              }
          }
          return -n - 1;
      }
      /**
       * Update the sizes of a branch node starting at the given index.
       */
      function updateSizes(node, i) {
          var sizes = node.sizes, children = node.children;
          var last = i > 0 ? sizes[i - 1] : 0;
          for (var n = children.length; i < n; ++i) {
              last = sizes[i] = last + children[i].size;
          }
          sizes.length = children.length;
      }
      /**
       * Split a node and return its new next sibling.
       *
       * @param node - The node of interest.
       *
       * @returns The new next sibling node.
       */
      function splitNode(node) {
          // Handle leaf nodes first.
          if (node.type === 1 /* Leaf */) {
              // Create the new sibling leaf node.
              var next_1 = new LeafNode();
              // Move the items to the new sibling.
              var v1_1 = node.items;
              var v2_1 = next_1.items;
              for (var i = MIN_NODE_WIDTH, n = v1_1.length; i < n; ++i) {
                  v2_1.push(v1_1[i]);
              }
              v1_1.length = MIN_NODE_WIDTH;
              // Patch up the sibling links.
              if (node.next)
                  node.next.prev = next_1;
              next_1.next = node.next;
              next_1.prev = node;
              node.next = next_1;
              // Return the new next sibling.
              return next_1;
          }
          // Create the new sibling branch node.
          var next = new BranchNode();
          // Move the children to the new sibling.
          var c1 = node.children;
          var c2 = next.children;
          for (var i = MIN_NODE_WIDTH, n = c1.length; i < n; ++i) {
              c2.push(c1[i]);
          }
          c1.length = MIN_NODE_WIDTH;
          // Move the items to the new sibling.
          var v1 = node.items;
          var v2 = next.items;
          for (var i = MIN_NODE_WIDTH, n = v1.length; i < n; ++i) {
              v2.push(v1[i]);
          }
          v1.length = MIN_NODE_WIDTH;
          // Update the dirty sizes of the nodes.
          updateSizes(node, MIN_NODE_WIDTH);
          updateSizes(next, 0);
          // Return the new next sibling.
          return next;
      }
      /**
       * Join a child node of a branch with one of its siblings.
       *
       * @param node - The branch node of interest.
       *
       * @param i - The index of the child node of interest.
       *
       * @returns The first modified index.
       *
       * #### Notes
       * This may cause the branch to become underfull.
       */
      function joinChild(node, i) {
          var _a, _b, _c, _d, _e, _f;
          // Fetch the child to be joined.
          var child = node.children[i];
          // Fetch the relevant sibling.
          var sibling = i === 0 ? node.children[i + 1] : node.children[i - 1];
          // Compute the flags which control the join behavior.
          var hasNext = i === 0;
          var isLeaf = child.type === 1 /* Leaf */;
          var hasExtra = sibling.width > MIN_NODE_WIDTH;
          // Join case #1: steal from next sibling leaf
          if (isLeaf && hasExtra && hasNext) {
              // Cast the children as leaves.
              var c = child;
              var s = sibling;
              // Steal an item.
              c.items.push(s.items.shift());
              // Update the branch items.
              node.items[i + 1] = s.items[0];
              // Return the first modified index.
              return i;
          }
          // Join case #2: steal from previous sibling leaf
          if (isLeaf && hasExtra && !hasNext) {
              // Cast the children as leaves.
              var c = child;
              var s = sibling;
              // Steal an item.
              c.items.unshift(s.items.pop());
              // Update the branch items.
              node.items[i] = c.items[0];
              // Return the first modified index.
              return i - 1;
          }
          // Join case #3: merge with next sibling leaf
          if (isLeaf && !hasExtra && hasNext) {
              // Cast the children as leaves.
              var c = child;
              var s = sibling;
              // Merge items.
              (_a = s.items).unshift.apply(_a, c.items);
              // Remove the old branch child.
              ArrayExt.removeAt(node.children, i);
              // Remove the stale branch item.
              ArrayExt.removeAt(node.items, i + 1);
              // Patch up the sibling links.
              if (c.prev)
                  c.prev.next = s;
              s.prev = c.prev;
              // Clear the original child.
              clear(c);
              // Return the first modified index.
              return i;
          }
          // Join case #4: merge with previous sibling leaf
          if (isLeaf && !hasExtra && !hasNext) {
              // Cast the children as leaves.
              var c = child;
              var s = sibling;
              // Merge items.
              (_b = s.items).push.apply(_b, c.items);
              // Remove the old branch child.
              ArrayExt.removeAt(node.children, i);
              // Remove the stale branch item.
              ArrayExt.removeAt(node.items, i);
              // Patch up the sibling links.
              if (c.next)
                  c.next.prev = s;
              s.next = c.next;
              // Clear the original child.
              clear(c);
              // Return the first modified index.
              return i - 1;
          }
          // Join case #5: steal from next sibling branch
          if (!isLeaf && hasExtra && hasNext) {
              // Cast the children to branches.
              var c = child;
              var s = sibling;
              // Steal a child from the next sibling.
              c.children.push(s.children.shift());
              // Steal an item from the next sibling.
              c.items.push(s.items.shift());
              // Update the branch items.
              node.items[i + 1] = s.items[0];
              // Update the sibling sizes.
              updateSizes(c, c.width - 1);
              updateSizes(s, 0);
              // Return the first modified index.
              return i;
          }
          // Join case #6: steal from previous sibling branch
          if (!isLeaf && hasExtra && !hasNext) {
              // Cast the children to branches.
              var c = child;
              var s = sibling;
              // Steal a child from the previous sibling.
              c.children.unshift(s.children.pop());
              // Steal an item from the previous sibling.
              c.items.unshift(s.items.pop());
              // Update the branch items.
              node.items[i] = c.items[0];
              // Update the sibling sizes.
              updateSizes(c, 0);
              updateSizes(s, s.width - 1);
              // Return the first modified index.
              return i - 1;
          }
          // Join case #7: merge with next sibling branch
          if (!isLeaf && !hasExtra && hasNext) {
              // Cast the children to branches.
              var c = child;
              var s = sibling;
              // Merge the children with the next sibling.
              (_c = s.children).unshift.apply(_c, c.children);
              // Merge the items with the next sibling.
              (_d = s.items).unshift.apply(_d, c.items);
              // Remove the old branch child.
              ArrayExt.removeAt(node.children, i);
              // Remove the stale branch item.
              ArrayExt.removeAt(node.items, i + 1);
              // Update the sibling sizes.
              updateSizes(s, 0);
              // Clear the original child but, not its children.
              c.children.length = 0;
              clear(c);
              // Return the first modified index.
              return i;
          }
          // Join case #8: merge with previous sibling branch
          if (!isLeaf && !hasExtra && !hasNext) {
              // Cast the children to branches.
              var c = child;
              var s = sibling;
              // Merge the children with the previous sibling.
              (_e = s.children).push.apply(_e, c.children);
              // Merge the items with the previous sibling.
              (_f = s.items).push.apply(_f, c.items);
              // Remove the old branch child.
              ArrayExt.removeAt(node.children, i);
              // Remove the stale branch item.
              ArrayExt.removeAt(node.items, i);
              // Update the sibling sizes.
              updateSizes(s, 0);
              // Clear the original child, but not its children.
              c.children.length = 0;
              clear(c);
              // Return the first modified index.
              return i - 1;
          }
          // One of the above cases must match.
          throw 'unreachable';
      }
  })(Private$2 || (Private$2 = {}));

  // Copyright (c) Jupyter Development Team.
  /**
   * A generic doubly-linked list.
   */
  var LinkedList = /** @class */ (function () {
      /**
       * Construct a new linked list.
       */
      function LinkedList() {
          this._first = null;
          this._last = null;
          this._size = 0;
      }
      Object.defineProperty(LinkedList.prototype, "isEmpty", {
          /**
           * Whether the list is empty.
           *
           * #### Complexity
           * Constant.
           */
          get: function () {
              return this._size === 0;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LinkedList.prototype, "size", {
          /**
           * The size of the list.
           *
           * #### Complexity
           * `O(1)`
           *
           * #### Notes
           * This is equivalent to `length`.
           */
          get: function () {
              return this._size;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LinkedList.prototype, "length", {
          /**
           * The length of the list.
           *
           * #### Complexity
           * Constant.
           *
           * #### Notes
           * This is equivalent to `size`.
           *
           * This property is deprecated.
           */
          get: function () {
              return this._size;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LinkedList.prototype, "first", {
          /**
           * The first value in the list.
           *
           * This is `undefined` if the list is empty.
           *
           * #### Complexity
           * Constant.
           */
          get: function () {
              return this._first ? this._first.value : undefined;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LinkedList.prototype, "last", {
          /**
           * The last value in the list.
           *
           * This is `undefined` if the list is empty.
           *
           * #### Complexity
           * Constant.
           */
          get: function () {
              return this._last ? this._last.value : undefined;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LinkedList.prototype, "firstNode", {
          /**
           * The first node in the list.
           *
           * This is `null` if the list is empty.
           *
           * #### Complexity
           * Constant.
           */
          get: function () {
              return this._first;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LinkedList.prototype, "lastNode", {
          /**
           * The last node in the list.
           *
           * This is `null` if the list is empty.
           *
           * #### Complexity
           * Constant.
           */
          get: function () {
              return this._last;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over the values in the list.
       *
       * @returns A new iterator starting with the first value.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.iter = function () {
          return new LinkedList.ForwardValueIterator(this._first);
      };
      /**
       * Create a reverse iterator over the values in the list.
       *
       * @returns A new iterator starting with the last value.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.retro = function () {
          return new LinkedList.RetroValueIterator(this._last);
      };
      /**
       * Create an iterator over the nodes in the list.
       *
       * @returns A new iterator starting with the first node.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.nodes = function () {
          return new LinkedList.ForwardNodeIterator(this._first);
      };
      /**
       * Create a reverse iterator over the nodes in the list.
       *
       * @returns A new iterator starting with the last node.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.retroNodes = function () {
          return new LinkedList.RetroNodeIterator(this._last);
      };
      /**
       * Assign new values to the list, replacing all current values.
       *
       * @param values - The values to assign to the list.
       *
       * #### Complexity
       * Linear.
       */
      LinkedList.prototype.assign = function (values) {
          var _this = this;
          this.clear();
          each$1(values, function (value) { _this.addLast(value); });
      };
      /**
       * Add a value to the end of the list.
       *
       * @param value - The value to add to the end of the list.
       *
       * #### Complexity
       * Constant.
       *
       * #### Notes
       * This is equivalent to `addLast`.
       */
      LinkedList.prototype.push = function (value) {
          this.addLast(value);
      };
      /**
       * Remove and return the value at the end of the list.
       *
       * @returns The removed value, or `undefined` if the list is empty.
       *
       * #### Complexity
       * Constant.
       *
       * #### Notes
       * This is equivalent to `removeLast`.
       */
      LinkedList.prototype.pop = function () {
          return this.removeLast();
      };
      /**
       * Add a value to the beginning of the list.
       *
       * @param value - The value to add to the beginning of the list.
       *
       * #### Complexity
       * Constant.
       *
       * #### Notes
       * This is equivalent to `addFirst`.
       */
      LinkedList.prototype.shift = function (value) {
          this.addFirst(value);
      };
      /**
       * Remove and return the value at the beginning of the list.
       *
       * @returns The removed value, or `undefined` if the list is empty.
       *
       * #### Complexity
       * Constant.
       *
       * #### Notes
       * This is equivalent to `removeFirst`.
       */
      LinkedList.prototype.unshift = function () {
          return this.removeFirst();
      };
      /**
       * Add a value to the beginning of the list.
       *
       * @param value - The value to add to the beginning of the list.
       *
       * @returns The list node which holds the value.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.addFirst = function (value) {
          var node = new Private$1$1.LinkedListNode(this, value);
          if (!this._first) {
              this._first = node;
              this._last = node;
          }
          else {
              node.next = this._first;
              this._first.prev = node;
              this._first = node;
          }
          this._size++;
          return node;
      };
      /**
       * Add a value to the end of the list.
       *
       * @param value - The value to add to the end of the list.
       *
       * @returns The list node which holds the value.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.addLast = function (value) {
          var node = new Private$1$1.LinkedListNode(this, value);
          if (!this._last) {
              this._first = node;
              this._last = node;
          }
          else {
              node.prev = this._last;
              this._last.next = node;
              this._last = node;
          }
          this._size++;
          return node;
      };
      /**
       * Insert a value before a specific node in the list.
       *
       * @param value - The value to insert before the reference node.
       *
       * @param ref - The reference node of interest. If this is `null`,
       *   the value will be added to the beginning of the list.
       *
       * @returns The list node which holds the value.
       *
       * #### Notes
       * The reference node must be owned by the list.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.insertBefore = function (value, ref) {
          if (!ref || ref === this._first) {
              return this.addFirst(value);
          }
          if (!(ref instanceof Private$1$1.LinkedListNode) || ref.list !== this) {
              throw new Error('Reference node is not owned by the list.');
          }
          var node = new Private$1$1.LinkedListNode(this, value);
          var _ref = ref;
          var prev = _ref.prev;
          node.next = _ref;
          node.prev = prev;
          _ref.prev = node;
          prev.next = node;
          this._size++;
          return node;
      };
      /**
       * Insert a value after a specific node in the list.
       *
       * @param value - The value to insert after the reference node.
       *
       * @param ref - The reference node of interest. If this is `null`,
       *   the value will be added to the end of the list.
       *
       * @returns The list node which holds the value.
       *
       * #### Notes
       * The reference node must be owned by the list.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.insertAfter = function (value, ref) {
          if (!ref || ref === this._last) {
              return this.addLast(value);
          }
          if (!(ref instanceof Private$1$1.LinkedListNode) || ref.list !== this) {
              throw new Error('Reference node is not owned by the list.');
          }
          var node = new Private$1$1.LinkedListNode(this, value);
          var _ref = ref;
          var next = _ref.next;
          node.next = next;
          node.prev = _ref;
          _ref.next = node;
          next.prev = node;
          this._size++;
          return node;
      };
      /**
       * Remove and return the value at the beginning of the list.
       *
       * @returns The removed value, or `undefined` if the list is empty.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.removeFirst = function () {
          var node = this._first;
          if (!node) {
              return undefined;
          }
          if (node === this._last) {
              this._first = null;
              this._last = null;
          }
          else {
              this._first = node.next;
              this._first.prev = null;
          }
          node.list = null;
          node.next = null;
          node.prev = null;
          this._size--;
          return node.value;
      };
      /**
       * Remove and return the value at the end of the list.
       *
       * @returns The removed value, or `undefined` if the list is empty.
       *
       * #### Complexity
       * Constant.
       */
      LinkedList.prototype.removeLast = function () {
          var node = this._last;
          if (!node) {
              return undefined;
          }
          if (node === this._first) {
              this._first = null;
              this._last = null;
          }
          else {
              this._last = node.prev;
              this._last.next = null;
          }
          node.list = null;
          node.next = null;
          node.prev = null;
          this._size--;
          return node.value;
      };
      /**
       * Remove a specific node from the list.
       *
       * @param node - The node to remove from the list.
       *
       * #### Complexity
       * Constant.
       *
       * #### Notes
       * The node must be owned by the list.
       */
      LinkedList.prototype.removeNode = function (node) {
          if (!(node instanceof Private$1$1.LinkedListNode) || node.list !== this) {
              throw new Error('Node is not owned by the list.');
          }
          var _node = node;
          if (_node === this._first && _node === this._last) {
              this._first = null;
              this._last = null;
          }
          else if (_node === this._first) {
              this._first = _node.next;
              this._first.prev = null;
          }
          else if (_node === this._last) {
              this._last = _node.prev;
              this._last.next = null;
          }
          else {
              _node.next.prev = _node.prev;
              _node.prev.next = _node.next;
          }
          _node.list = null;
          _node.next = null;
          _node.prev = null;
          this._size--;
      };
      /**
       * Remove all values from the list.
       *
       * #### Complexity
       * Linear.
       */
      LinkedList.prototype.clear = function () {
          var node = this._first;
          while (node) {
              var next = node.next;
              node.list = null;
              node.prev = null;
              node.next = null;
              node = next;
          }
          this._first = null;
          this._last = null;
          this._size = 0;
      };
      return LinkedList;
  }());
  /**
   * The namespace for the `LinkedList` class statics.
   */
  (function (LinkedList) {
      /**
       * Create a linked list from an iterable of values.
       *
       * @param values - The iterable or array-like object of interest.
       *
       * @returns A new linked list initialized with the given values.
       *
       * #### Complexity
       * Linear.
       */
      function from(values) {
          var list = new LinkedList();
          list.assign(values);
          return list;
      }
      LinkedList.from = from;
      /**
       * A forward iterator for values in a linked list.
       */
      var ForwardValueIterator = /** @class */ (function () {
          /**
           * Construct a forward value iterator.
           *
           * @param node - The first node in the list.
           */
          function ForwardValueIterator(node) {
              this._node = node;
          }
          /**
           * Get an iterator over the object's values.
           *
           * @returns An iterator which yields the object's values.
           */
          ForwardValueIterator.prototype.iter = function () {
              return this;
          };
          /**
           * Create an independent clone of the iterator.
           *
           * @returns A new independent clone of the iterator.
           */
          ForwardValueIterator.prototype.clone = function () {
              return new ForwardValueIterator(this._node);
          };
          /**
           * Get the next value from the iterator.
           *
           * @returns The next value from the iterator, or `undefined`.
           */
          ForwardValueIterator.prototype.next = function () {
              if (!this._node) {
                  return undefined;
              }
              var node = this._node;
              this._node = node.next;
              return node.value;
          };
          return ForwardValueIterator;
      }());
      LinkedList.ForwardValueIterator = ForwardValueIterator;
      /**
       * A reverse iterator for values in a linked list.
       */
      var RetroValueIterator = /** @class */ (function () {
          /**
           * Construct a retro value iterator.
           *
           * @param node - The last node in the list.
           */
          function RetroValueIterator(node) {
              this._node = node;
          }
          /**
           * Get an iterator over the object's values.
           *
           * @returns An iterator which yields the object's values.
           */
          RetroValueIterator.prototype.iter = function () {
              return this;
          };
          /**
           * Create an independent clone of the iterator.
           *
           * @returns A new independent clone of the iterator.
           */
          RetroValueIterator.prototype.clone = function () {
              return new RetroValueIterator(this._node);
          };
          /**
           * Get the next value from the iterator.
           *
           * @returns The next value from the iterator, or `undefined`.
           */
          RetroValueIterator.prototype.next = function () {
              if (!this._node) {
                  return undefined;
              }
              var node = this._node;
              this._node = node.prev;
              return node.value;
          };
          return RetroValueIterator;
      }());
      LinkedList.RetroValueIterator = RetroValueIterator;
      /**
       * A forward iterator for nodes in a linked list.
       */
      var ForwardNodeIterator = /** @class */ (function () {
          /**
           * Construct a forward node iterator.
           *
           * @param node - The first node in the list.
           */
          function ForwardNodeIterator(node) {
              this._node = node;
          }
          /**
           * Get an iterator over the object's values.
           *
           * @returns An iterator which yields the object's values.
           */
          ForwardNodeIterator.prototype.iter = function () {
              return this;
          };
          /**
           * Create an independent clone of the iterator.
           *
           * @returns A new independent clone of the iterator.
           */
          ForwardNodeIterator.prototype.clone = function () {
              return new ForwardNodeIterator(this._node);
          };
          /**
           * Get the next value from the iterator.
           *
           * @returns The next value from the iterator, or `undefined`.
           */
          ForwardNodeIterator.prototype.next = function () {
              if (!this._node) {
                  return undefined;
              }
              var node = this._node;
              this._node = node.next;
              return node;
          };
          return ForwardNodeIterator;
      }());
      LinkedList.ForwardNodeIterator = ForwardNodeIterator;
      /**
       * A reverse iterator for nodes in a linked list.
       */
      var RetroNodeIterator = /** @class */ (function () {
          /**
           * Construct a retro node iterator.
           *
           * @param node - The last node in the list.
           */
          function RetroNodeIterator(node) {
              this._node = node;
          }
          /**
           * Get an iterator over the object's values.
           *
           * @returns An iterator which yields the object's values.
           */
          RetroNodeIterator.prototype.iter = function () {
              return this;
          };
          /**
           * Create an independent clone of the iterator.
           *
           * @returns A new independent clone of the iterator.
           */
          RetroNodeIterator.prototype.clone = function () {
              return new RetroNodeIterator(this._node);
          };
          /**
           * Get the next value from the iterator.
           *
           * @returns The next value from the iterator, or `undefined`.
           */
          RetroNodeIterator.prototype.next = function () {
              if (!this._node) {
                  return undefined;
              }
              var node = this._node;
              this._node = node.prev;
              return node;
          };
          return RetroNodeIterator;
      }());
      LinkedList.RetroNodeIterator = RetroNodeIterator;
  })(LinkedList || (LinkedList = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$1$1;
  (function (Private) {
      /**
       * The internal linked list node implementation.
       */
      var LinkedListNode = /** @class */ (function () {
          /**
           * Construct a new linked list node.
           *
           * @param list - The list which owns the node.
           *
           * @param value - The value for the link.
           */
          function LinkedListNode(list, value) {
              /**
               * The linked list which created and owns the node.
               */
              this.list = null;
              /**
               * The next node in the list.
               */
              this.next = null;
              /**
               * The previous node in the list.
               */
              this.prev = null;
              this.list = list;
              this.value = value;
          }
          return LinkedListNode;
      }());
      Private.LinkedListNode = LinkedListNode;
  })(Private$1$1 || (Private$1$1 = {}));

  /*! *****************************************************************************
  Copyright (c) Microsoft Corporation. All rights reserved.
  Licensed under the Apache License, Version 2.0 (the "License"); you may not use
  this file except in compliance with the License. You may obtain a copy of the
  License at http://www.apache.org/licenses/LICENSE-2.0

  THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
  KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
  WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
  MERCHANTABLITY OR NON-INFRINGEMENT.

  See the Apache Version 2.0 License for specific language governing permissions
  and limitations under the License.
  ***************************************************************************** */
  /* global Reflect, Promise */

  var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf ||
          ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
          function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
      return extendStatics(d, b);
  };

  function __extends$2(d, b) {
      extendStatics(d, b);
      function __() { this.constructor = d; }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
  }

  /**
   * A message which can be delivered to a message handler.
   *
   * #### Notes
   * This class may be subclassed to create complex message types.
   */
  var Message = /** @class */ (function () {
      /**
       * Construct a new message.
       *
       * @param type - The type of the message.
       */
      function Message(type) {
          this.type = type;
      }
      Object.defineProperty(Message.prototype, "isConflatable", {
          /**
           * Test whether the message is conflatable.
           *
           * #### Notes
           * Message conflation is an advanced topic. Most message types will
           * not make use of this feature.
           *
           * If a conflatable message is posted to a handler while another
           * conflatable message of the same `type` has already been posted
           * to the handler, the `conflate()` method of the existing message
           * will be invoked. If that method returns `true`, the new message
           * will not be enqueued. This allows messages to be compressed, so
           * that only a single instance of the message type is processed per
           * cycle, no matter how many times messages of that type are posted.
           *
           * Custom message types may reimplement this property.
           *
           * The default implementation is always `false`.
           */
          get: function () {
              return false;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Conflate this message with another message of the same `type`.
       *
       * @param other - A conflatable message of the same `type`.
       *
       * @returns `true` if the message was successfully conflated, or
       *   `false` otherwise.
       *
       * #### Notes
       * Message conflation is an advanced topic. Most message types will
       * not make use of this feature.
       *
       * This method is called automatically by the message loop when the
       * given message is posted to the handler paired with this message.
       * This message will already be enqueued and conflatable, and the
       * given message will have the same `type` and also be conflatable.
       *
       * This method should merge the state of the other message into this
       * message as needed so that when this message is finally delivered
       * to the handler, it receives the most up-to-date information.
       *
       * If this method returns `true`, it signals that the other message
       * was successfully conflated and that message will not be enqueued.
       *
       * If this method returns `false`, the other message will be enqueued
       * for normal delivery.
       *
       * Custom message types may reimplement this method.
       *
       * The default implementation always returns `false`.
       */
      Message.prototype.conflate = function (other) {
          return false;
      };
      return Message;
  }());
  /**
   * A convenience message class which conflates automatically.
   *
   * #### Notes
   * Message conflation is an advanced topic. Most user code will not
   * make use of this class.
   *
   * This message class is useful for creating message instances which
   * should be conflated, but which have no state other than `type`.
   *
   * If conflation of stateful messages is required, a custom `Message`
   * subclass should be created.
   */
  var ConflatableMessage = /** @class */ (function (_super) {
      __extends$2(ConflatableMessage, _super);
      function ConflatableMessage() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      Object.defineProperty(ConflatableMessage.prototype, "isConflatable", {
          /**
           * Test whether the message is conflatable.
           *
           * #### Notes
           * This property is always `true`.
           */
          get: function () {
              return true;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Conflate this message with another message of the same `type`.
       *
       * #### Notes
       * This method always returns `true`.
       */
      ConflatableMessage.prototype.conflate = function (other) {
          return true;
      };
      return ConflatableMessage;
  }(Message));
  /**
   * The namespace for the global singleton message loop.
   */
  var MessageLoop;
  (function (MessageLoop) {
      /**
       * Send a message to a message handler to process immediately.
       *
       * @param handler - The handler which should process the message.
       *
       * @param msg - The message to deliver to the handler.
       *
       * #### Notes
       * The message will first be sent through any installed message hooks
       * for the handler. If the message passes all hooks, it will then be
       * delivered to the `processMessage` method of the handler.
       *
       * The message will not be conflated with pending posted messages.
       *
       * Exceptions in hooks and handlers will be caught and logged.
       */
      function sendMessage(handler, msg) {
          // Lookup the message hooks for the handler.
          var hooks = messageHooks.get(handler);
          // Handle the common case of no installed hooks.
          if (!hooks || hooks.length === 0) {
              invokeHandler(handler, msg);
              return;
          }
          // Invoke the message hooks starting with the newest first.
          var passed = every$1(retro(hooks), function (hook) {
              return hook ? invokeHook(hook, handler, msg) : true;
          });
          // Invoke the handler if the message passes all hooks.
          if (passed) {
              invokeHandler(handler, msg);
          }
      }
      MessageLoop.sendMessage = sendMessage;
      /**
       * Post a message to a message handler to process in the future.
       *
       * @param handler - The handler which should process the message.
       *
       * @param msg - The message to post to the handler.
       *
       * #### Notes
       * The message will be conflated with the pending posted messages for
       * the handler, if possible. If the message is not conflated, it will
       * be queued for normal delivery on the next cycle of the event loop.
       *
       * Exceptions in hooks and handlers will be caught and logged.
       */
      function postMessage(handler, msg) {
          // Handle the common case of a non-conflatable message.
          if (!msg.isConflatable) {
              enqueueMessage(handler, msg);
              return;
          }
          // Conflate the message with an existing message if possible.
          var conflated = some$1(messageQueue, function (posted) {
              if (posted.handler !== handler) {
                  return false;
              }
              if (!posted.msg) {
                  return false;
              }
              if (posted.msg.type !== msg.type) {
                  return false;
              }
              if (!posted.msg.isConflatable) {
                  return false;
              }
              return posted.msg.conflate(msg);
          });
          // Enqueue the message if it was not conflated.
          if (!conflated) {
              enqueueMessage(handler, msg);
          }
      }
      MessageLoop.postMessage = postMessage;
      /**
       * Install a message hook for a message handler.
       *
       * @param handler - The message handler of interest.
       *
       * @param hook - The message hook to install.
       *
       * #### Notes
       * A message hook is invoked before a message is delivered to the
       * handler. If the hook returns `false`, no other hooks will be
       * invoked and the message will not be delivered to the handler.
       *
       * The most recently installed message hook is executed first.
       *
       * If the hook is already installed, this is a no-op.
       */
      function installMessageHook(handler, hook) {
          // Lookup the hooks for the handler.
          var hooks = messageHooks.get(handler);
          // Bail early if the hook is already installed.
          if (hooks && hooks.indexOf(hook) !== -1) {
              return;
          }
          // Add the hook to the end, so it will be the first to execute.
          if (!hooks) {
              messageHooks.set(handler, [hook]);
          }
          else {
              hooks.push(hook);
          }
      }
      MessageLoop.installMessageHook = installMessageHook;
      /**
       * Remove an installed message hook for a message handler.
       *
       * @param handler - The message handler of interest.
       *
       * @param hook - The message hook to remove.
       *
       * #### Notes
       * It is safe to call this function while the hook is executing.
       *
       * If the hook is not installed, this is a no-op.
       */
      function removeMessageHook(handler, hook) {
          // Lookup the hooks for the handler.
          var hooks = messageHooks.get(handler);
          // Bail early if the hooks do not exist.
          if (!hooks) {
              return;
          }
          // Lookup the index of the hook and bail if not found.
          var i = hooks.indexOf(hook);
          if (i === -1) {
              return;
          }
          // Clear the hook and schedule a cleanup of the array.
          hooks[i] = null;
          scheduleCleanup(hooks);
      }
      MessageLoop.removeMessageHook = removeMessageHook;
      /**
       * Clear all message data associated with a message handler.
       *
       * @param handler - The message handler of interest.
       *
       * #### Notes
       * This will clear all posted messages and hooks for the handler.
       */
      function clearData(handler) {
          // Lookup the hooks for the handler.
          var hooks = messageHooks.get(handler);
          // Clear all messsage hooks for the handler.
          if (hooks && hooks.length > 0) {
              ArrayExt.fill(hooks, null);
              scheduleCleanup(hooks);
          }
          // Clear all posted messages for the handler.
          each$1(messageQueue, function (posted) {
              if (posted.handler === handler) {
                  posted.handler = null;
                  posted.msg = null;
              }
          });
      }
      MessageLoop.clearData = clearData;
      /**
       * Process the pending posted messages in the queue immediately.
       *
       * #### Notes
       * This function is useful when posted messages must be processed
       * immediately, instead of on the next animation frame.
       *
       * This function should normally not be needed, but it may be
       * required to work around certain browser idiosyncrasies.
       *
       * Recursing into this function is a no-op.
       */
      function flush() {
          // Bail if recursion is detected or if there is no pending task.
          if (flushGuard || loopTaskID === 0) {
              return;
          }
          // Unschedule the pending loop task.
          unschedule(loopTaskID);
          // Run the message loop within the recursion guard.
          flushGuard = true;
          runMessageLoop();
          flushGuard = false;
      }
      MessageLoop.flush = flush;
      /**
       * Get the message loop exception handler.
       *
       * @returns The current exception handler.
       *
       * #### Notes
       * The default exception handler is `console.error`.
       */
      function getExceptionHandler() {
          return exceptionHandler;
      }
      MessageLoop.getExceptionHandler = getExceptionHandler;
      /**
       * Set the message loop exception handler.
       *
       * @param handler - The function to use as the exception handler.
       *
       * @returns The old exception handler.
       *
       * #### Notes
       * The exception handler is invoked when a message handler or a
       * message hook throws an exception.
       */
      function setExceptionHandler(handler) {
          var old = exceptionHandler;
          exceptionHandler = handler;
          return old;
      }
      MessageLoop.setExceptionHandler = setExceptionHandler;
      /**
       * The queue of posted message pairs.
       */
      var messageQueue = new LinkedList();
      /**
       * A mapping of handler to array of installed message hooks.
       */
      var messageHooks = new WeakMap();
      /**
       * A set of message hook arrays which are pending cleanup.
       */
      var dirtySet = new Set();
      /**
       * The message loop exception handler.
       */
      var exceptionHandler = function (err) {
          console.error(err);
      };
      /**
       * The id of the pending loop task animation frame.
       */
      var loopTaskID = 0;
      /**
       * A guard flag to prevent flush recursion.
       */
      var flushGuard = false;
      /**
       * A function to schedule an event loop callback.
       */
      var schedule = (function () {
          var ok = typeof requestAnimationFrame === 'function';
          return ok ? requestAnimationFrame : setImmediate;
      })();
      /**
       * A function to unschedule an event loop callback.
       */
      var unschedule = (function () {
          var ok = typeof cancelAnimationFrame === 'function';
          return ok ? cancelAnimationFrame : clearImmediate;
      })();
      /**
       * Invoke a message hook with the specified handler and message.
       *
       * Returns the result of the hook, or `true` if the hook throws.
       *
       * Exceptions in the hook will be caught and logged.
       */
      function invokeHook(hook, handler, msg) {
          var result = true;
          try {
              if (typeof hook === 'function') {
                  result = hook(handler, msg);
              }
              else {
                  result = hook.messageHook(handler, msg);
              }
          }
          catch (err) {
              exceptionHandler(err);
          }
          return result;
      }
      /**
       * Invoke a message handler with the specified message.
       *
       * Exceptions in the handler will be caught and logged.
       */
      function invokeHandler(handler, msg) {
          try {
              handler.processMessage(msg);
          }
          catch (err) {
              exceptionHandler(err);
          }
      }
      /**
       * Add a message to the end of the message queue.
       *
       * This will automatically schedule a run of the message loop.
       */
      function enqueueMessage(handler, msg) {
          // Add the posted message to the queue.
          messageQueue.addLast({ handler: handler, msg: msg });
          // Bail if a loop task is already pending.
          if (loopTaskID !== 0) {
              return;
          }
          // Schedule a run of the message loop.
          loopTaskID = schedule(runMessageLoop);
      }
      /**
       * Run an iteration of the message loop.
       *
       * This will process all pending messages in the queue. If a message
       * is added to the queue while the message loop is running, it will
       * be processed on the next cycle of the loop.
       */
      function runMessageLoop() {
          // Clear the task ID so the next loop can be scheduled.
          loopTaskID = 0;
          // If the message queue is empty, there is nothing else to do.
          if (messageQueue.isEmpty) {
              return;
          }
          // Add a sentinel value to the end of the queue. The queue will
          // only be processed up to the sentinel. Messages posted during
          // this cycle will execute on the next cycle.
          var sentinel = { handler: null, msg: null };
          messageQueue.addLast(sentinel);
          // Enter the message loop.
          while (true) {
              // Remove the first posted message in the queue.
              var posted = messageQueue.removeFirst();
              // If the value is the sentinel, exit the loop.
              if (posted === sentinel) {
                  return;
              }
              // Dispatch the message if it has not been cleared.
              if (posted.handler && posted.msg) {
                  sendMessage(posted.handler, posted.msg);
              }
          }
      }
      /**
       * Schedule a cleanup of a message hooks array.
       *
       * This will add the array to the dirty set and schedule a deferred
       * cleanup of the array contents. On cleanup, any `null` hook will
       * be removed from the array.
       */
      function scheduleCleanup(hooks) {
          if (dirtySet.size === 0) {
              schedule(cleanupDirtySet);
          }
          dirtySet.add(hooks);
      }
      /**
       * Cleanup the message hook arrays in the dirty set.
       *
       * This function should only be invoked asynchronously, when the
       * stack frame is guaranteed to not be on the path of user code.
       */
      function cleanupDirtySet() {
          dirtySet.forEach(cleanupHooks);
          dirtySet.clear();
      }
      /**
       * Cleanup the dirty hooks in a message hooks array.
       *
       * This will remove any `null` hook from the array.
       *
       * This function should only be invoked asynchronously, when the
       * stack frame is guaranteed to not be on the path of user code.
       */
      function cleanupHooks(hooks) {
          ArrayExt.removeAllWhere(hooks, isNull);
      }
      /**
       * Test whether a value is `null`.
       */
      function isNull(value) {
          return value === null;
      }
  })(MessageLoop || (MessageLoop = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * A class which attaches a value to an external object.
   *
   * #### Notes
   * Attached properties are used to extend the state of an object with
   * semantic data from an unrelated class. They also encapsulate value
   * creation, coercion, and notification.
   *
   * Because attached property values are stored in a hash table, which
   * in turn is stored in a WeakMap keyed on the owner object, there is
   * non-trivial storage overhead involved in their use. The pattern is
   * therefore best used for the storage of rare data.
   */
  var AttachedProperty = /** @class */ (function () {
      /**
       * Construct a new attached property.
       *
       * @param options - The options for initializing the property.
       */
      function AttachedProperty(options) {
          this._pid = Private$3.nextPID();
          this.name = options.name;
          this._create = options.create;
          this._coerce = options.coerce || null;
          this._compare = options.compare || null;
          this._changed = options.changed || null;
      }
      /**
       * Get the current value of the property for a given owner.
       *
       * @param owner - The property owner of interest.
       *
       * @returns The current value of the property.
       *
       * #### Notes
       * If the value has not yet been set, the default value will be
       * computed and assigned as the current value of the property.
       */
      AttachedProperty.prototype.get = function (owner) {
          var value;
          var map = Private$3.ensureMap(owner);
          if (this._pid in map) {
              value = map[this._pid];
          }
          else {
              value = map[this._pid] = this._createValue(owner);
          }
          return value;
      };
      /**
       * Set the current value of the property for a given owner.
       *
       * @param owner - The property owner of interest.
       *
       * @param value - The value for the property.
       *
       * #### Notes
       * If the value has not yet been set, the default value will be
       * computed and used as the previous value for the comparison.
       */
      AttachedProperty.prototype.set = function (owner, value) {
          var oldValue;
          var map = Private$3.ensureMap(owner);
          if (this._pid in map) {
              oldValue = map[this._pid];
          }
          else {
              oldValue = map[this._pid] = this._createValue(owner);
          }
          var newValue = this._coerceValue(owner, value);
          this._maybeNotify(owner, oldValue, map[this._pid] = newValue);
      };
      /**
       * Explicitly coerce the current property value for a given owner.
       *
       * @param owner - The property owner of interest.
       *
       * #### Notes
       * If the value has not yet been set, the default value will be
       * computed and used as the previous value for the comparison.
       */
      AttachedProperty.prototype.coerce = function (owner) {
          var oldValue;
          var map = Private$3.ensureMap(owner);
          if (this._pid in map) {
              oldValue = map[this._pid];
          }
          else {
              oldValue = map[this._pid] = this._createValue(owner);
          }
          var newValue = this._coerceValue(owner, oldValue);
          this._maybeNotify(owner, oldValue, map[this._pid] = newValue);
      };
      /**
       * Get or create the default value for the given owner.
       */
      AttachedProperty.prototype._createValue = function (owner) {
          var create = this._create;
          return create(owner);
      };
      /**
       * Coerce the value for the given owner.
       */
      AttachedProperty.prototype._coerceValue = function (owner, value) {
          var coerce = this._coerce;
          return coerce ? coerce(owner, value) : value;
      };
      /**
       * Compare the old value and new value for equality.
       */
      AttachedProperty.prototype._compareValue = function (oldValue, newValue) {
          var compare = this._compare;
          return compare ? compare(oldValue, newValue) : oldValue === newValue;
      };
      /**
       * Run the change notification if the given values are different.
       */
      AttachedProperty.prototype._maybeNotify = function (owner, oldValue, newValue) {
          var changed = this._changed;
          if (changed && !this._compareValue(oldValue, newValue)) {
              changed(owner, oldValue, newValue);
          }
      };
      return AttachedProperty;
  }());
  /**
   * The namespace for the `AttachedProperty` class statics.
   */
  (function (AttachedProperty) {
      /**
       * Clear the stored property data for the given owner.
       *
       * @param owner - The property owner of interest.
       *
       * #### Notes
       * This will clear all property values for the owner, but it will
       * **not** run the change notification for any of the properties.
       */
      function clearData(owner) {
          Private$3.ownerData.delete(owner);
      }
      AttachedProperty.clearData = clearData;
  })(AttachedProperty || (AttachedProperty = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$3;
  (function (Private) {
      /**
       * A weak mapping of property owner to property map.
       */
      Private.ownerData = new WeakMap();
      /**
       * A function which computes successive unique property ids.
       */
      Private.nextPID = (function () {
          var id = 0;
          return function () {
              var rand = Math.random();
              var stem = ("" + rand).slice(2);
              return "pid-" + stem + "-" + id++;
          };
      })();
      /**
       * Lookup the data map for the property owner.
       *
       * This will create the map if one does not already exist.
       */
      function ensureMap(owner) {
          var map = Private.ownerData.get(owner);
          if (map) {
              return map;
          }
          map = Object.create(null);
          Private.ownerData.set(owner, map);
          return map;
      }
      Private.ensureMap = ensureMap;
  })(Private$3 || (Private$3 = {}));

  // Copyright (c) Jupyter Development Team.
  /**
   * A concrete implementation of `ISignal`.
   *
   * #### Example
   * ```typescript
   * import { ISignal, Signal } from '@lumino/signaling';
   *
   * class SomeClass {
   *
   *   constructor(name: string) {
   *     this.name = name;
   *   }
   *
   *   readonly name: string;
   *
   *   get valueChanged: ISignal<this, number> {
   *     return this._valueChanged;
   *   }
   *
   *   get value(): number {
   *     return this._value;
   *   }
   *
   *   set value(value: number) {
   *     if (value === this._value) {
   *       return;
   *     }
   *     this._value = value;
   *     this._valueChanged.emit(value);
   *   }
   *
   *   private _value = 0;
   *   private _valueChanged = new Signal<this, number>(this);
   * }
   *
   * function logger(sender: SomeClass, value: number): void {
   *   console.log(sender.name, value);
   * }
   *
   * let m1 = new SomeClass('foo');
   * let m2 = new SomeClass('bar');
   *
   * m1.valueChanged.connect(logger);
   * m2.valueChanged.connect(logger);
   *
   * m1.value = 42;  // logs: foo 42
   * m2.value = 17;  // logs: bar 17
   * ```
   */
  var Signal = /** @class */ (function () {
      /**
       * Construct a new signal.
       *
       * @param sender - The sender which owns the signal.
       */
      function Signal(sender) {
          this.sender = sender;
      }
      /**
       * Connect a slot to the signal.
       *
       * @param slot - The slot to invoke when the signal is emitted.
       *
       * @param thisArg - The `this` context for the slot. If provided,
       *   this must be a non-primitive object.
       *
       * @returns `true` if the connection succeeds, `false` otherwise.
       */
      Signal.prototype.connect = function (slot, thisArg) {
          return Private$4.connect(this, slot, thisArg);
      };
      /**
       * Disconnect a slot from the signal.
       *
       * @param slot - The slot to disconnect from the signal.
       *
       * @param thisArg - The `this` context for the slot. If provided,
       *   this must be a non-primitive object.
       *
       * @returns `true` if the connection is removed, `false` otherwise.
       */
      Signal.prototype.disconnect = function (slot, thisArg) {
          return Private$4.disconnect(this, slot, thisArg);
      };
      /**
       * Emit the signal and invoke the connected slots.
       *
       * @param args - The args to pass to the connected slots.
       *
       * #### Notes
       * Slots are invoked synchronously in connection order.
       *
       * Exceptions thrown by connected slots will be caught and logged.
       */
      Signal.prototype.emit = function (args) {
          Private$4.emit(this, args);
      };
      return Signal;
  }());
  /**
   * The namespace for the `Signal` class statics.
   */
  (function (Signal) {
      /**
       * Remove all connections between a sender and receiver.
       *
       * @param sender - The sender object of interest.
       *
       * @param receiver - The receiver object of interest.
       *
       * #### Notes
       * If a `thisArg` is provided when connecting a signal, that object
       * is considered the receiver. Otherwise, the `slot` is considered
       * the receiver.
       */
      function disconnectBetween(sender, receiver) {
          Private$4.disconnectBetween(sender, receiver);
      }
      Signal.disconnectBetween = disconnectBetween;
      /**
       * Remove all connections where the given object is the sender.
       *
       * @param sender - The sender object of interest.
       */
      function disconnectSender(sender) {
          Private$4.disconnectSender(sender);
      }
      Signal.disconnectSender = disconnectSender;
      /**
       * Remove all connections where the given object is the receiver.
       *
       * @param receiver - The receiver object of interest.
       *
       * #### Notes
       * If a `thisArg` is provided when connecting a signal, that object
       * is considered the receiver. Otherwise, the `slot` is considered
       * the receiver.
       */
      function disconnectReceiver(receiver) {
          Private$4.disconnectReceiver(receiver);
      }
      Signal.disconnectReceiver = disconnectReceiver;
      /**
       * Remove all connections where an object is the sender or receiver.
       *
       * @param object - The object of interest.
       *
       * #### Notes
       * If a `thisArg` is provided when connecting a signal, that object
       * is considered the receiver. Otherwise, the `slot` is considered
       * the receiver.
       */
      function disconnectAll(object) {
          Private$4.disconnectAll(object);
      }
      Signal.disconnectAll = disconnectAll;
      /**
       * Clear all signal data associated with the given object.
       *
       * @param object - The object for which the data should be cleared.
       *
       * #### Notes
       * This removes all signal connections and any other signal data
       * associated with the object.
       */
      function clearData(object) {
          Private$4.disconnectAll(object);
      }
      Signal.clearData = clearData;
      /**
       * Get the signal exception handler.
       *
       * @returns The current exception handler.
       *
       * #### Notes
       * The default exception handler is `console.error`.
       */
      function getExceptionHandler() {
          return Private$4.exceptionHandler;
      }
      Signal.getExceptionHandler = getExceptionHandler;
      /**
       * Set the signal exception handler.
       *
       * @param handler - The function to use as the exception handler.
       *
       * @returns The old exception handler.
       *
       * #### Notes
       * The exception handler is invoked when a slot throws an exception.
       */
      function setExceptionHandler(handler) {
          var old = Private$4.exceptionHandler;
          Private$4.exceptionHandler = handler;
          return old;
      }
      Signal.setExceptionHandler = setExceptionHandler;
  })(Signal || (Signal = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$4;
  (function (Private) {
      /**
       * The signal exception handler function.
       */
      Private.exceptionHandler = function (err) {
          console.error(err);
      };
      /**
       * Connect a slot to a signal.
       *
       * @param signal - The signal of interest.
       *
       * @param slot - The slot to invoke when the signal is emitted.
       *
       * @param thisArg - The `this` context for the slot. If provided,
       *   this must be a non-primitive object.
       *
       * @returns `true` if the connection succeeds, `false` otherwise.
       */
      function connect(signal, slot, thisArg) {
          // Coerce a `null` `thisArg` to `undefined`.
          thisArg = thisArg || undefined;
          // Ensure the sender's array of receivers is created.
          var receivers = receiversForSender.get(signal.sender);
          if (!receivers) {
              receivers = [];
              receiversForSender.set(signal.sender, receivers);
          }
          // Bail if a matching connection already exists.
          if (findConnection(receivers, signal, slot, thisArg)) {
              return false;
          }
          // Choose the best object for the receiver.
          var receiver = thisArg || slot;
          // Ensure the receiver's array of senders is created.
          var senders = sendersForReceiver.get(receiver);
          if (!senders) {
              senders = [];
              sendersForReceiver.set(receiver, senders);
          }
          // Create a new connection and add it to the end of each array.
          var connection = { signal: signal, slot: slot, thisArg: thisArg };
          receivers.push(connection);
          senders.push(connection);
          // Indicate a successful connection.
          return true;
      }
      Private.connect = connect;
      /**
       * Disconnect a slot from a signal.
       *
       * @param signal - The signal of interest.
       *
       * @param slot - The slot to disconnect from the signal.
       *
       * @param thisArg - The `this` context for the slot. If provided,
       *   this must be a non-primitive object.
       *
       * @returns `true` if the connection is removed, `false` otherwise.
       */
      function disconnect(signal, slot, thisArg) {
          // Coerce a `null` `thisArg` to `undefined`.
          thisArg = thisArg || undefined;
          // Lookup the list of receivers, and bail if none exist.
          var receivers = receiversForSender.get(signal.sender);
          if (!receivers || receivers.length === 0) {
              return false;
          }
          // Bail if no matching connection exits.
          var connection = findConnection(receivers, signal, slot, thisArg);
          if (!connection) {
              return false;
          }
          // Choose the best object for the receiver.
          var receiver = thisArg || slot;
          // Lookup the array of senders, which is now known to exist.
          var senders = sendersForReceiver.get(receiver);
          // Clear the connection and schedule cleanup of the arrays.
          connection.signal = null;
          scheduleCleanup(receivers);
          scheduleCleanup(senders);
          // Indicate a successful disconnection.
          return true;
      }
      Private.disconnect = disconnect;
      /**
       * Remove all connections between a sender and receiver.
       *
       * @param sender - The sender object of interest.
       *
       * @param receiver - The receiver object of interest.
       */
      function disconnectBetween(sender, receiver) {
          // If there are no receivers, there is nothing to do.
          var receivers = receiversForSender.get(sender);
          if (!receivers || receivers.length === 0) {
              return;
          }
          // If there are no senders, there is nothing to do.
          var senders = sendersForReceiver.get(receiver);
          if (!senders || senders.length === 0) {
              return;
          }
          // Clear each connection between the sender and receiver.
          each$1(senders, function (connection) {
              // Skip connections which have already been cleared.
              if (!connection.signal) {
                  return;
              }
              // Clear the connection if it matches the sender.
              if (connection.signal.sender === sender) {
                  connection.signal = null;
              }
          });
          // Schedule a cleanup of the senders and receivers.
          scheduleCleanup(receivers);
          scheduleCleanup(senders);
      }
      Private.disconnectBetween = disconnectBetween;
      /**
       * Remove all connections where the given object is the sender.
       *
       * @param sender - The sender object of interest.
       */
      function disconnectSender(sender) {
          // If there are no receivers, there is nothing to do.
          var receivers = receiversForSender.get(sender);
          if (!receivers || receivers.length === 0) {
              return;
          }
          // Clear each receiver connection.
          each$1(receivers, function (connection) {
              // Skip connections which have already been cleared.
              if (!connection.signal) {
                  return;
              }
              // Choose the best object for the receiver.
              var receiver = connection.thisArg || connection.slot;
              // Clear the connection.
              connection.signal = null;
              // Cleanup the array of senders, which is now known to exist.
              scheduleCleanup(sendersForReceiver.get(receiver));
          });
          // Schedule a cleanup of the receivers.
          scheduleCleanup(receivers);
      }
      Private.disconnectSender = disconnectSender;
      /**
       * Remove all connections where the given object is the receiver.
       *
       * @param receiver - The receiver object of interest.
       */
      function disconnectReceiver(receiver) {
          // If there are no senders, there is nothing to do.
          var senders = sendersForReceiver.get(receiver);
          if (!senders || senders.length === 0) {
              return;
          }
          // Clear each sender connection.
          each$1(senders, function (connection) {
              // Skip connections which have already been cleared.
              if (!connection.signal) {
                  return;
              }
              // Lookup the sender for the connection.
              var sender = connection.signal.sender;
              // Clear the connection.
              connection.signal = null;
              // Cleanup the array of receivers, which is now known to exist.
              scheduleCleanup(receiversForSender.get(sender));
          });
          // Schedule a cleanup of the list of senders.
          scheduleCleanup(senders);
      }
      Private.disconnectReceiver = disconnectReceiver;
      /**
       * Remove all connections where an object is the sender or receiver.
       *
       * @param object - The object of interest.
       */
      function disconnectAll(object) {
          // Remove all connections where the given object is the sender.
          disconnectSender(object);
          // Remove all connections where the given object is the receiver.
          disconnectReceiver(object);
      }
      Private.disconnectAll = disconnectAll;
      /**
       * Emit a signal and invoke its connected slots.
       *
       * @param signal - The signal of interest.
       *
       * @param args - The args to pass to the connected slots.
       *
       * #### Notes
       * Slots are invoked synchronously in connection order.
       *
       * Exceptions thrown by connected slots will be caught and logged.
       */
      function emit(signal, args) {
          // If there are no receivers, there is nothing to do.
          var receivers = receiversForSender.get(signal.sender);
          if (!receivers || receivers.length === 0) {
              return;
          }
          // Invoke the slots for connections with a matching signal.
          // Any connections added during emission are not invoked.
          for (var i = 0, n = receivers.length; i < n; ++i) {
              var connection = receivers[i];
              if (connection.signal === signal) {
                  invokeSlot(connection, args);
              }
          }
      }
      Private.emit = emit;
      /**
       * A weak mapping of sender to array of receiver connections.
       */
      var receiversForSender = new WeakMap();
      /**
       * A weak mapping of receiver to array of sender connections.
       */
      var sendersForReceiver = new WeakMap();
      /**
       * A set of connection arrays which are pending cleanup.
       */
      var dirtySet = new Set();
      /**
       * A function to schedule an event loop callback.
       */
      var schedule = (function () {
          var ok = typeof requestAnimationFrame === 'function';
          // @ts-ignore
          return ok ? requestAnimationFrame : setImmediate;
      })();
      /**
       * Find a connection which matches the given parameters.
       */
      function findConnection(connections, signal, slot, thisArg) {
          return find$1(connections, function (connection) { return (connection.signal === signal &&
              connection.slot === slot &&
              connection.thisArg === thisArg); });
      }
      /**
       * Invoke a slot with the given parameters.
       *
       * The connection is assumed to be valid.
       *
       * Exceptions in the slot will be caught and logged.
       */
      function invokeSlot(connection, args) {
          var signal = connection.signal, slot = connection.slot, thisArg = connection.thisArg;
          try {
              slot.call(thisArg, signal.sender, args);
          }
          catch (err) {
              Private.exceptionHandler(err);
          }
      }
      /**
       * Schedule a cleanup of a connection array.
       *
       * This will add the array to the dirty set and schedule a deferred
       * cleanup of the array contents. On cleanup, any connection with a
       * `null` signal will be removed from the array.
       */
      function scheduleCleanup(array) {
          if (dirtySet.size === 0) {
              schedule(cleanupDirtySet);
          }
          dirtySet.add(array);
      }
      /**
       * Cleanup the connection lists in the dirty set.
       *
       * This function should only be invoked asynchronously, when the
       * stack frame is guaranteed to not be on the path of user code.
       */
      function cleanupDirtySet() {
          dirtySet.forEach(cleanupConnections);
          dirtySet.clear();
      }
      /**
       * Cleanup the dirty connections in a connections array.
       *
       * This will remove any connection with a `null` signal.
       *
       * This function should only be invoked asynchronously, when the
       * stack frame is guaranteed to not be on the path of user code.
       */
      function cleanupConnections(connections) {
          ArrayExt.removeAllWhere(connections, isDeadConnection);
      }
      /**
       * Test whether a connection is dead.
       *
       * A dead connection has a `null` signal.
       */
      function isDeadConnection(connection) {
          return connection.signal === null;
      }
  })(Private$4 || (Private$4 = {}));

  /*! *****************************************************************************
  Copyright (c) Microsoft Corporation. All rights reserved.
  Licensed under the Apache License, Version 2.0 (the "License"); you may not use
  this file except in compliance with the License. You may obtain a copy of the
  License at http://www.apache.org/licenses/LICENSE-2.0

  THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
  KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
  WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
  MERCHANTABLITY OR NON-INFRINGEMENT.

  See the Apache Version 2.0 License for specific language governing permissions
  and limitations under the License.
  ***************************************************************************** */
  /* global Reflect, Promise */

  var extendStatics$1 = function(d, b) {
      extendStatics$1 = Object.setPrototypeOf ||
          ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
          function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
      return extendStatics$1(d, b);
  };

  function __extends$3(d, b) {
      extendStatics$1(d, b);
      function __() { this.constructor = d; }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
  }

  /**
   * A disposable object which delegates to a callback function.
   */
  var DisposableDelegate = /** @class */ (function () {
      /**
       * Construct a new disposable delegate.
       *
       * @param fn - The callback function to invoke on dispose.
       */
      function DisposableDelegate(fn) {
          this._fn = fn;
      }
      Object.defineProperty(DisposableDelegate.prototype, "isDisposed", {
          /**
           * Test whether the delegate has been disposed.
           */
          get: function () {
              return !this._fn;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Dispose of the delegate and invoke the callback function.
       */
      DisposableDelegate.prototype.dispose = function () {
          if (!this._fn) {
              return;
          }
          var fn = this._fn;
          this._fn = null;
          fn();
      };
      return DisposableDelegate;
  }());
  /**
   * An observable disposable object which delegates to a callback function.
   */
  var ObservableDisposableDelegate = /** @class */ (function (_super) {
      __extends$3(ObservableDisposableDelegate, _super);
      function ObservableDisposableDelegate() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._disposed = new Signal(_this);
          return _this;
      }
      Object.defineProperty(ObservableDisposableDelegate.prototype, "disposed", {
          /**
           * A signal emitted when the delegate is disposed.
           */
          get: function () {
              return this._disposed;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Dispose of the delegate and invoke the callback function.
       */
      ObservableDisposableDelegate.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          this._disposed.emit(undefined);
          Signal.clearData(this);
      };
      return ObservableDisposableDelegate;
  }(DisposableDelegate));
  /**
   * An object which manages a collection of disposable items.
   */
  var DisposableSet = /** @class */ (function () {
      /**
       * Construct a new disposable set.
       */
      function DisposableSet() {
          this._isDisposed = false;
          this._items = new Set();
      }
      Object.defineProperty(DisposableSet.prototype, "isDisposed", {
          /**
           * Test whether the set has been disposed.
           */
          get: function () {
              return this._isDisposed;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Dispose of the set and the items it contains.
       *
       * #### Notes
       * Items are disposed in the order they are added to the set.
       */
      DisposableSet.prototype.dispose = function () {
          if (this._isDisposed) {
              return;
          }
          this._isDisposed = true;
          this._items.forEach(function (item) { item.dispose(); });
          this._items.clear();
      };
      /**
       * Test whether the set contains a specific item.
       *
       * @param item - The item of interest.
       *
       * @returns `true` if the set contains the item, `false` otherwise.
       */
      DisposableSet.prototype.contains = function (item) {
          return this._items.has(item);
      };
      /**
       * Add a disposable item to the set.
       *
       * @param item - The item to add to the set.
       *
       * #### Notes
       * If the item is already contained in the set, this is a no-op.
       */
      DisposableSet.prototype.add = function (item) {
          this._items.add(item);
      };
      /**
       * Remove a disposable item from the set.
       *
       * @param item - The item to remove from the set.
       *
       * #### Notes
       * If the item is not contained in the set, this is a no-op.
       */
      DisposableSet.prototype.remove = function (item) {
          this._items.delete(item);
      };
      /**
       * Remove all items from the set.
       */
      DisposableSet.prototype.clear = function () {
          this._items.clear();
      };
      return DisposableSet;
  }());
  /**
   * The namespace for the `DisposableSet` class statics.
   */
  (function (DisposableSet) {
      /**
       * Create a disposable set from an iterable of items.
       *
       * @param items - The iterable or array-like object of interest.
       *
       * @returns A new disposable initialized with the given items.
       */
      function from(items) {
          var set = new DisposableSet();
          each$1(items, function (item) { set.add(item); });
          return set;
      }
      DisposableSet.from = from;
  })(DisposableSet || (DisposableSet = {}));
  /**
   * An observable object which manages a collection of disposable items.
   */
  var ObservableDisposableSet = /** @class */ (function (_super) {
      __extends$3(ObservableDisposableSet, _super);
      function ObservableDisposableSet() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._disposed = new Signal(_this);
          return _this;
      }
      Object.defineProperty(ObservableDisposableSet.prototype, "disposed", {
          /**
           * A signal emitted when the set is disposed.
           */
          get: function () {
              return this._disposed;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Dispose of the set and the items it contains.
       *
       * #### Notes
       * Items are disposed in the order they are added to the set.
       */
      ObservableDisposableSet.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          this._disposed.emit(undefined);
          Signal.clearData(this);
      };
      return ObservableDisposableSet;
  }(DisposableSet));
  /**
   * The namespace for the `ObservableDisposableSet` class statics.
   */
  (function (ObservableDisposableSet) {
      /**
       * Create an observable disposable set from an iterable of items.
       *
       * @param items - The iterable or array-like object of interest.
       *
       * @returns A new disposable initialized with the given items.
       */
      function from(items) {
          var set = new ObservableDisposableSet();
          each$1(items, function (item) { set.add(item); });
          return set;
      }
      ObservableDisposableSet.from = from;
  })(ObservableDisposableSet || (ObservableDisposableSet = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * Get the global application keyboard layout instance.
   *
   * @returns The keyboard layout for use by the application.
   *
   * #### Notes
   * The default keyboard layout is US-English.
   */
  function getKeyboardLayout() {
      return Private$5.keyboardLayout;
  }
  /**
   * A concrete implementation of [[IKeyboardLayout]] based on keycodes.
   *
   * The `keyCode` property of a `'keydown'` event is a browser and OS
   * specific representation of the physical key (not character) which
   * was pressed on a keyboard. While not the most convenient API, it
   * is currently the only one which works reliably on all browsers.
   *
   * This class accepts a user-defined mapping of keycode to key, which
   * allows for reliable shortcuts tailored to the user's system.
   */
  var KeycodeLayout = /** @class */ (function () {
      /**
       * Construct a new keycode layout.
       *
       * @param name - The human readable name for the layout.
       *
       * @param codes - A mapping of keycode to key value.
       */
      function KeycodeLayout(name, codes) {
          this.name = name;
          this._codes = codes;
          this._keys = KeycodeLayout.extractKeys(codes);
      }
      /**
       * Get an array of the key values supported by the layout.
       *
       * @returns A new array of the supported key values.
       */
      KeycodeLayout.prototype.keys = function () {
          return Object.keys(this._keys);
      };
      /**
       * Test whether the given key is a valid value for the layout.
       *
       * @param key - The user provided key to test for validity.
       *
       * @returns `true` if the key is valid, `false` otherwise.
       */
      KeycodeLayout.prototype.isValidKey = function (key) {
          return key in this._keys;
      };
      /**
       * Get the key for a `'keydown'` event.
       *
       * @param event - The event object for a `'keydown'` event.
       *
       * @returns The associated key value, or an empty string if
       *   the event does not represent a valid primary key.
       */
      KeycodeLayout.prototype.keyForKeydownEvent = function (event) {
          return this._codes[event.keyCode] || '';
      };
      return KeycodeLayout;
  }());
  /**
   * The namespace for the `KeycodeLayout` class statics.
   */
  (function (KeycodeLayout) {
      /**
       * Extract the set of keys from a code map.
       *
       * @param code - The code map of interest.
       *
       * @returns A set of the keys in the code map.
       */
      function extractKeys(codes) {
          var keys = Object.create(null);
          for (var c in codes) {
              keys[codes[c]] = true;
          }
          return keys;
      }
      KeycodeLayout.extractKeys = extractKeys;
  })(KeycodeLayout || (KeycodeLayout = {}));
  /**
   * A keycode-based keyboard layout for US English keyboards.
   *
   * This layout is valid for the following OS/Browser combinations.
   *
   * - Windows
   *   - Chrome
   *   - Firefox
   *   - IE
   *
   * - OSX
   *   - Chrome
   *   - Firefox
   *   - Safari
   *
   * - Linux
   *   - Chrome
   *   - Firefox
   *
   * Other combinations may also work, but are untested.
   */
  var EN_US = new KeycodeLayout('en-us', {
      8: 'Backspace',
      9: 'Tab',
      13: 'Enter',
      19: 'Pause',
      27: 'Escape',
      32: 'Space',
      33: 'PageUp',
      34: 'PageDown',
      35: 'End',
      36: 'Home',
      37: 'ArrowLeft',
      38: 'ArrowUp',
      39: 'ArrowRight',
      40: 'ArrowDown',
      45: 'Insert',
      46: 'Delete',
      48: '0',
      49: '1',
      50: '2',
      51: '3',
      52: '4',
      53: '5',
      54: '6',
      55: '7',
      56: '8',
      57: '9',
      59: ';',
      61: '=',
      65: 'A',
      66: 'B',
      67: 'C',
      68: 'D',
      69: 'E',
      70: 'F',
      71: 'G',
      72: 'H',
      73: 'I',
      74: 'J',
      75: 'K',
      76: 'L',
      77: 'M',
      78: 'N',
      79: 'O',
      80: 'P',
      81: 'Q',
      82: 'R',
      83: 'S',
      84: 'T',
      85: 'U',
      86: 'V',
      87: 'W',
      88: 'X',
      89: 'Y',
      90: 'Z',
      93: 'ContextMenu',
      96: '0',
      97: '1',
      98: '2',
      99: '3',
      100: '4',
      101: '5',
      102: '6',
      103: '7',
      104: '8',
      105: '9',
      106: '*',
      107: '+',
      109: '-',
      110: '.',
      111: '/',
      112: 'F1',
      113: 'F2',
      114: 'F3',
      115: 'F4',
      116: 'F5',
      117: 'F6',
      118: 'F7',
      119: 'F8',
      120: 'F9',
      121: 'F10',
      122: 'F11',
      123: 'F12',
      173: '-',
      186: ';',
      187: '=',
      188: ',',
      189: '-',
      190: '.',
      191: '/',
      192: '`',
      219: '[',
      220: '\\',
      221: ']',
      222: '\''
  });
  /**
   * The namespace for the module implementation details.
   */
  var Private$5;
  (function (Private) {
      /**
       * The global keyboard layout instance.
       */
      Private.keyboardLayout = EN_US;
  })(Private$5 || (Private$5 = {}));

  // Copyright (c) Jupyter Development Team.
  /**
   * An object which manages a collection of commands.
   *
   * #### Notes
   * A command registry can be used to populate a variety of action-based
   * widgets, such as command palettes, menus, and toolbars.
   */
  var CommandRegistry = /** @class */ (function () {
      /**
       * Construct a new command registry.
       */
      function CommandRegistry() {
          this._timerID = 0;
          this._replaying = false;
          this._keystrokes = [];
          this._keydownEvents = [];
          this._keyBindings = [];
          this._exactKeyMatch = null;
          this._commands = Object.create(null);
          this._commandChanged = new Signal(this);
          this._commandExecuted = new Signal(this);
          this._keyBindingChanged = new Signal(this);
      }
      Object.defineProperty(CommandRegistry.prototype, "commandChanged", {
          /**
           * A signal emitted when a command has changed.
           *
           * #### Notes
           * This signal is useful for visual representations of commands which
           * need to refresh when the state of a relevant command has changed.
           */
          get: function () {
              return this._commandChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(CommandRegistry.prototype, "commandExecuted", {
          /**
           * A signal emitted when a command has executed.
           *
           * #### Notes
           * Care should be taken when consuming this signal. It is intended to
           * be used largely for debugging and logging purposes. It should not
           * be (ab)used for general purpose spying on command execution.
           */
          get: function () {
              return this._commandExecuted;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(CommandRegistry.prototype, "keyBindingChanged", {
          /**
           * A signal emitted when a key binding is changed.
           */
          get: function () {
              return this._keyBindingChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(CommandRegistry.prototype, "keyBindings", {
          /**
           * A read-only array of the key bindings in the registry.
           */
          get: function () {
              return this._keyBindings;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * List the ids of the registered commands.
       *
       * @returns A new array of the registered command ids.
       */
      CommandRegistry.prototype.listCommands = function () {
          return Object.keys(this._commands);
      };
      /**
       * Test whether a specific command is registered.
       *
       * @param id - The id of the command of interest.
       *
       * @returns `true` if the command is registered, `false` otherwise.
       */
      CommandRegistry.prototype.hasCommand = function (id) {
          return id in this._commands;
      };
      /**
       * Add a command to the registry.
       *
       * @param id - The unique id of the command.
       *
       * @param options - The options for the command.
       *
       * @returns A disposable which will remove the command.
       *
       * @throws An error if the given `id` is already registered.
       */
      CommandRegistry.prototype.addCommand = function (id, options) {
          var _this = this;
          // Throw an error if the id is already registered.
          if (id in this._commands) {
              throw new Error("Command '" + id + "' already registered.");
          }
          // Add the command to the registry.
          this._commands[id] = Private$6.createCommand(options);
          // Emit the `commandChanged` signal.
          this._commandChanged.emit({ id: id, type: 'added' });
          // Return a disposable which will remove the command.
          return new DisposableDelegate(function () {
              // Remove the command from the registry.
              delete _this._commands[id];
              // Emit the `commandChanged` signal.
              _this._commandChanged.emit({ id: id, type: 'removed' });
          });
      };
      /**
       * Notify listeners that the state of a command has changed.
       *
       * @param id - The id of the command which has changed. If more than
       *   one command has changed, this argument should be omitted.
       *
       * @throws An error if the given `id` is not registered.
       *
       * #### Notes
       * This method should be called by the command author whenever the
       * application state changes such that the results of the command
       * metadata functions may have changed.
       *
       * This will cause the `commandChanged` signal to be emitted.
       */
      CommandRegistry.prototype.notifyCommandChanged = function (id) {
          if (id !== undefined && !(id in this._commands)) {
              throw new Error("Command '" + id + "' is not registered.");
          }
          this._commandChanged.emit({ id: id, type: id ? 'changed' : 'many-changed' });
      };
      /**
       * Get the display label for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The display label for the command, or an empty string
       *   if the command is not registered.
       */
      CommandRegistry.prototype.label = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.label.call(undefined, args) : '';
      };
      /**
       * Get the mnemonic index for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The mnemonic index for the command, or `-1` if the
       *   command is not registered.
       */
      CommandRegistry.prototype.mnemonic = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.mnemonic.call(undefined, args) : -1;
      };
      /**
       * Get the icon renderer for a specific command.
       *
       * DEPRECATED: if set to a string value, the .icon field will
       * function as an alias for the .iconClass field, for backwards
       * compatibility. In the future when this is removed, the default
       * return type will become undefined.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The icon renderer for the command, or
       *   an empty string if the command is not registered.
       */
      CommandRegistry.prototype.icon = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.icon.call(undefined, args) : /* <DEPRECATED> */ '' /* </DEPRECATED> */ /* <FUTURE> undefined </FUTURE> */;
      };
      /**
       * Get the icon class for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The icon class for the command, or an empty string if
       *   the command is not registered.
       */
      CommandRegistry.prototype.iconClass = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.iconClass.call(undefined, args) : '';
      };
      /**
       * Get the icon label for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The icon label for the command, or an empty string if
       *   the command is not registered.
       */
      CommandRegistry.prototype.iconLabel = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.iconLabel.call(undefined, args) : '';
      };
      /**
       * Get the short form caption for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The caption for the command, or an empty string if the
       *   command is not registered.
       */
      CommandRegistry.prototype.caption = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.caption.call(undefined, args) : '';
      };
      /**
       * Get the usage help text for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The usage text for the command, or an empty string if
       *   the command is not registered.
       */
      CommandRegistry.prototype.usage = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.usage.call(undefined, args) : '';
      };
      /**
       * Get the extra class name for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The class name for the command, or an empty string if
       *   the command is not registered.
       */
      CommandRegistry.prototype.className = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.className.call(undefined, args) : '';
      };
      /**
       * Get the dataset for a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns The dataset for the command, or an empty dataset if
       *   the command is not registered.
       */
      CommandRegistry.prototype.dataset = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.dataset.call(undefined, args) : {};
      };
      /**
       * Test whether a specific command is enabled.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns A boolean indicating whether the command is enabled,
       *   or `false` if the command is not registered.
       */
      CommandRegistry.prototype.isEnabled = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.isEnabled.call(undefined, args) : false;
      };
      /**
       * Test whether a specific command is toggled.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns A boolean indicating whether the command is toggled,
       *   or `false` if the command is not registered.
       */
      CommandRegistry.prototype.isToggled = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.isToggled.call(undefined, args) : false;
      };
      /**
       * Test whether a specific command is visible.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns A boolean indicating whether the command is visible,
       *   or `false` if the command is not registered.
       */
      CommandRegistry.prototype.isVisible = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          var cmd = this._commands[id];
          return cmd ? cmd.isVisible.call(undefined, args) : false;
      };
      /**
       * Execute a specific command.
       *
       * @param id - The id of the command of interest.
       *
       * @param args - The arguments for the command.
       *
       * @returns A promise which resolves with the result of the command.
       *
       * #### Notes
       * The promise will reject if the command throws an exception,
       * or if the command is not registered.
       */
      CommandRegistry.prototype.execute = function (id, args) {
          if (args === void 0) { args = JSONExt.emptyObject; }
          // Reject if the command is not registered.
          var cmd = this._commands[id];
          if (!cmd) {
              return Promise.reject(new Error("Command '" + id + "' not registered."));
          }
          // Execute the command and reject if an exception is thrown.
          var value;
          try {
              value = cmd.execute.call(undefined, args);
          }
          catch (err) {
              value = Promise.reject(err);
          }
          // Create the return promise which resolves the result.
          var result = Promise.resolve(value);
          // Emit the command executed signal.
          this._commandExecuted.emit({ id: id, args: args, result: result });
          // Return the result promise to the caller.
          return result;
      };
      /**
       * Add a key binding to the registry.
       *
       * @param options - The options for creating the key binding.
       *
       * @returns A disposable which removes the added key binding.
       *
       * #### Notes
       * If multiple key bindings are registered for the same sequence, the
       * binding with the highest selector specificity is executed first. A
       * tie is broken by using the most recently added key binding.
       *
       * Ambiguous key bindings are resolved with a timeout. As an example,
       * suppose two key bindings are registered: one with the key sequence
       * `['Ctrl D']`, and another with `['Ctrl D', 'Ctrl W']`. If the user
       * presses `Ctrl D`, the first binding cannot be immediately executed
       * since the user may intend to complete the chord with `Ctrl W`. For
       * such cases, a timer is used to allow the chord to be completed. If
       * the chord is not completed before the timeout, the first binding
       * is executed.
       */
      CommandRegistry.prototype.addKeyBinding = function (options) {
          var _this = this;
          // Create the binding for the given options.
          var binding = Private$6.createKeyBinding(options);
          // Add the key binding to the bindings array.
          this._keyBindings.push(binding);
          // Emit the `bindingChanged` signal.
          this._keyBindingChanged.emit({ binding: binding, type: 'added' });
          // Return a disposable which will remove the binding.
          return new DisposableDelegate(function () {
              // Remove the binding from the array.
              ArrayExt.removeFirstOf(_this._keyBindings, binding);
              // Emit the `bindingChanged` signal.
              _this._keyBindingChanged.emit({ binding: binding, type: 'removed' });
          });
      };
      /**
       * Process a `'keydown'` event and invoke a matching key binding.
       *
       * @param event - The event object for a `'keydown'` event.
       *
       * #### Notes
       * This should be called in response to a `'keydown'` event in order
       * to invoke the command for the best matching key binding.
       *
       * The registry **does not** install its own listener for `'keydown'`
       * events. This allows the application full control over the nodes
       * and phase for which the registry processes `'keydown'` events.
       *
       * When the keydown event is processed, if the event target or any of its
       * ancestor nodes has a `data-lm-suppress-shortcuts` attribute, its keydown
       * events will not invoke commands.
       */
      CommandRegistry.prototype.processKeydownEvent = function (event) {
          // Bail immediately if playing back keystrokes.
          if (this._replaying) {
              return;
          }
          // Get the normalized keystroke for the event.
          var keystroke = CommandRegistry.keystrokeForKeydownEvent(event);
          // If the keystroke is not valid for the keyboard layout, replay
          // any suppressed events and clear the pending state.
          if (!keystroke) {
              this._replayKeydownEvents();
              this._clearPendingState();
              return;
          }
          // Add the keystroke to the current key sequence.
          this._keystrokes.push(keystroke);
          // Find the exact and partial matches for the key sequence.
          var _a = Private$6.matchKeyBinding(this._keyBindings, this._keystrokes, event), exact = _a.exact, partial = _a.partial;
          // If there is no exact match and no partial match, replay
          // any suppressed events and clear the pending state.
          if (!exact && !partial) {
              this._replayKeydownEvents();
              this._clearPendingState();
              return;
          }
          // Stop propagation of the event. If there is only a partial match,
          // the event will be replayed if a final exact match never occurs.
          event.preventDefault();
          event.stopPropagation();
          // If there is an exact match but no partial match, the exact match
          // can be dispatched immediately. The pending state is cleared so
          // the next key press starts from the default state.
          if (exact && !partial) {
              this._executeKeyBinding(exact);
              this._clearPendingState();
              return;
          }
          // If there is both an exact match and a partial match, the exact
          // match is stored for future dispatch in case the timer expires
          // before a more specific match is triggered.
          if (exact) {
              this._exactKeyMatch = exact;
          }
          // Store the event for possible playback in the future.
          this._keydownEvents.push(event);
          // (Re)start the timer to dispatch the most recent exact match
          // in case the partial match fails to result in an exact match.
          this._startTimer();
      };
      /**
       * Start or restart the pending timeout.
       */
      CommandRegistry.prototype._startTimer = function () {
          var _this = this;
          this._clearTimer();
          this._timerID = window.setTimeout(function () {
              _this._onPendingTimeout();
          }, Private$6.CHORD_TIMEOUT);
      };
      /**
       * Clear the pending timeout.
       */
      CommandRegistry.prototype._clearTimer = function () {
          if (this._timerID !== 0) {
              clearTimeout(this._timerID);
              this._timerID = 0;
          }
      };
      /**
       * Replay the keydown events which were suppressed.
       */
      CommandRegistry.prototype._replayKeydownEvents = function () {
          if (this._keydownEvents.length === 0) {
              return;
          }
          this._replaying = true;
          this._keydownEvents.forEach(Private$6.replayKeyEvent);
          this._replaying = false;
      };
      /**
       * Execute the command for the given key binding.
       *
       * If the command is missing or disabled, a warning will be logged.
       */
      CommandRegistry.prototype._executeKeyBinding = function (binding) {
          var command = binding.command, args = binding.args;
          if (!this.hasCommand(command) || !this.isEnabled(command, args)) {
              var word = this.hasCommand(command) ? 'enabled' : 'registered';
              var keys = binding.keys.join(', ');
              var msg1 = "Cannot execute key binding '" + keys + "':";
              var msg2 = "command '" + command + "' is not " + word + ".";
              console.warn(msg1 + " " + msg2);
              return;
          }
          this.execute(command, args);
      };
      /**
       * Clear the internal pending state.
       */
      CommandRegistry.prototype._clearPendingState = function () {
          this._clearTimer();
          this._exactKeyMatch = null;
          this._keystrokes.length = 0;
          this._keydownEvents.length = 0;
      };
      /**
       * Handle the partial match timeout.
       */
      CommandRegistry.prototype._onPendingTimeout = function () {
          this._timerID = 0;
          if (this._exactKeyMatch) {
              this._executeKeyBinding(this._exactKeyMatch);
          }
          else {
              this._replayKeydownEvents();
          }
          this._clearPendingState();
      };
      return CommandRegistry;
  }());
  /**
   * The namespace for the `CommandRegistry` class statics.
   */
  (function (CommandRegistry) {
      /**
       * Parse a keystroke into its constituent components.
       *
       * @param keystroke - The keystroke of interest.
       *
       * @returns The parsed components of the keystroke.
       *
       * #### Notes
       * The keystroke should be of the form:
       *   `[<modifier 1> [<modifier 2> [<modifier N> ]]]<primary key>`
       *
       * The supported modifiers are: `Accel`, `Alt`, `Cmd`, `Ctrl`, and
       * `Shift`. The `Accel` modifier is translated to `Cmd` on Mac and
       * `Ctrl` on all other platforms.
       *
       * The parsing is tolerant and will not throw exceptions. Notably:
       *   - Duplicate modifiers are ignored.
       *   - Extra primary keys are ignored.
       *   - The order of modifiers and primary key is irrelevant.
       *   - The keystroke parts should be separated by whitespace.
       *   - The keystroke is case sensitive.
       */
      function parseKeystroke(keystroke) {
          var key = '';
          var alt = false;
          var cmd = false;
          var ctrl = false;
          var shift = false;
          for (var _i = 0, _a = keystroke.split(/\s+/); _i < _a.length; _i++) {
              var token = _a[_i];
              if (token === 'Accel') {
                  if (Platform.IS_MAC) {
                      cmd = true;
                  }
                  else {
                      ctrl = true;
                  }
              }
              else if (token === 'Alt') {
                  alt = true;
              }
              else if (token === 'Cmd') {
                  cmd = true;
              }
              else if (token === 'Ctrl') {
                  ctrl = true;
              }
              else if (token === 'Shift') {
                  shift = true;
              }
              else if (token.length > 0) {
                  key = token;
              }
          }
          return { cmd: cmd, ctrl: ctrl, alt: alt, shift: shift, key: key };
      }
      CommandRegistry.parseKeystroke = parseKeystroke;
      /**
       * Normalize a keystroke into a canonical representation.
       *
       * @param keystroke - The keystroke of interest.
       *
       * @returns The normalized representation of the keystroke.
       *
       * #### Notes
       * This normalizes the keystroke by removing duplicate modifiers and
       * extra primary keys, and assembling the parts in a canonical order.
       *
       * The `Cmd` modifier is ignored on non-Mac platforms.
       */
      function normalizeKeystroke(keystroke) {
          var mods = '';
          var parts = parseKeystroke(keystroke);
          if (parts.ctrl) {
              mods += 'Ctrl ';
          }
          if (parts.alt) {
              mods += 'Alt ';
          }
          if (parts.shift) {
              mods += 'Shift ';
          }
          if (parts.cmd && Platform.IS_MAC) {
              mods += 'Cmd ';
          }
          return mods + parts.key;
      }
      CommandRegistry.normalizeKeystroke = normalizeKeystroke;
      /**
       * Get the platform-specific normalized keys for an options object.
       *
       * @param options - The options for the key binding.
       *
       * @returns Array of combined, normalized keys.
       */
      function normalizeKeys(options) {
          var keys;
          if (Platform.IS_WIN) {
              keys = options.winKeys || options.keys;
          }
          else if (Platform.IS_MAC) {
              keys = options.macKeys || options.keys;
          }
          else {
              keys = options.linuxKeys || options.keys;
          }
          return keys.map(normalizeKeystroke);
      }
      CommandRegistry.normalizeKeys = normalizeKeys;
      /**
       * Format a keystroke for display on the local system.
       */
      function formatKeystroke(keystroke) {
          var mods = '';
          var parts = parseKeystroke(keystroke);
          if (Platform.IS_MAC) {
              if (parts.ctrl) {
                  mods += '\u2303 ';
              }
              if (parts.alt) {
                  mods += '\u2325 ';
              }
              if (parts.shift) {
                  mods += '\u21E7 ';
              }
              if (parts.cmd) {
                  mods += '\u2318 ';
              }
          }
          else {
              if (parts.ctrl) {
                  mods += 'Ctrl+';
              }
              if (parts.alt) {
                  mods += 'Alt+';
              }
              if (parts.shift) {
                  mods += 'Shift+';
              }
          }
          return mods + parts.key;
      }
      CommandRegistry.formatKeystroke = formatKeystroke;
      /**
       * Create a normalized keystroke for a `'keydown'` event.
       *
       * @param event - The event object for a `'keydown'` event.
       *
       * @returns A normalized keystroke, or an empty string if the event
       *   does not represent a valid keystroke for the given layout.
       */
      function keystrokeForKeydownEvent(event) {
          var key = getKeyboardLayout().keyForKeydownEvent(event);
          if (!key) {
              return '';
          }
          var mods = '';
          if (event.ctrlKey) {
              mods += 'Ctrl ';
          }
          if (event.altKey) {
              mods += 'Alt ';
          }
          if (event.shiftKey) {
              mods += 'Shift ';
          }
          if (event.metaKey && Platform.IS_MAC) {
              mods += 'Cmd ';
          }
          return mods + key;
      }
      CommandRegistry.keystrokeForKeydownEvent = keystrokeForKeydownEvent;
  })(CommandRegistry || (CommandRegistry = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$6;
  (function (Private) {
      /**
       * The timeout in ms for triggering a key binding chord.
       */
      Private.CHORD_TIMEOUT = 1000;
      /**
       * Create a normalized command from an options object.
       */
      function createCommand(options) {
          var icon;
          var iconClass;
          /* <DEPRECATED> */
          if (!(options.icon) || typeof options.icon === 'string') {
              // alias icon to iconClass
              iconClass = asFunc(options.iconClass || options.icon, emptyStringFunc);
              icon = iconClass;
          }
          else {
              /* /<DEPRECATED> */
              iconClass = asFunc(options.iconClass, emptyStringFunc);
              icon = asFunc(options.icon, undefinedFunc);
              /* <DEPRECATED> */
          }
          /* </DEPRECATED> */
          return {
              execute: options.execute,
              label: asFunc(options.label, emptyStringFunc),
              mnemonic: asFunc(options.mnemonic, negativeOneFunc),
              icon: icon,
              iconClass: iconClass,
              iconLabel: asFunc(options.iconLabel, emptyStringFunc),
              caption: asFunc(options.caption, emptyStringFunc),
              usage: asFunc(options.usage, emptyStringFunc),
              className: asFunc(options.className, emptyStringFunc),
              dataset: asFunc(options.dataset, emptyDatasetFunc),
              isEnabled: options.isEnabled || trueFunc,
              isToggled: options.isToggled || falseFunc,
              isVisible: options.isVisible || trueFunc
          };
      }
      Private.createCommand = createCommand;
      /**
       * Create a key binding object from key binding options.
       */
      function createKeyBinding(options) {
          return {
              keys: CommandRegistry.normalizeKeys(options),
              selector: validateSelector(options),
              command: options.command,
              args: options.args || JSONExt.emptyObject
          };
      }
      Private.createKeyBinding = createKeyBinding;
      /**
       * Find the key bindings which match a key sequence.
       *
       * This returns a match result which contains the best exact matching
       * binding, and a flag which indicates if there are partial matches.
       */
      function matchKeyBinding(bindings, keys, event) {
          // The current best exact match.
          var exact = null;
          // Whether a partial match has been found.
          var partial = false;
          // The match distance for the exact match.
          var distance = Infinity;
          // The specificity for the exact match.
          var specificity = 0;
          // Iterate over the bindings and search for the best match.
          for (var i = 0, n = bindings.length; i < n; ++i) {
              // Lookup the current binding.
              var binding = bindings[i];
              // Check whether the key binding sequence is a match.
              var sqm = matchSequence(binding.keys, keys);
              // If there is no match, the binding is ignored.
              if (sqm === 0 /* None */) {
                  continue;
              }
              // If it is a partial match and no other partial match has been
              // found, ensure the selector matches and set the partial flag.
              if (sqm === 2 /* Partial */) {
                  if (!partial && targetDistance(binding.selector, event) !== -1) {
                      partial = true;
                  }
                  continue;
              }
              // Ignore the match if the selector doesn't match, or if the
              // matched node is farther away than the current best match.
              var td = targetDistance(binding.selector, event);
              if (td === -1 || td > distance) {
                  continue;
              }
              // Get the specificity for the selector.
              var sp = Selector.calculateSpecificity(binding.selector);
              // Update the best match if this match is stronger.
              if (!exact || td < distance || sp >= specificity) {
                  exact = binding;
                  distance = td;
                  specificity = sp;
              }
          }
          // Return the match result.
          return { exact: exact, partial: partial };
      }
      Private.matchKeyBinding = matchKeyBinding;
      /**
       * Replay a keyboard event.
       *
       * This synthetically dispatches a clone of the keyboard event.
       */
      function replayKeyEvent(event) {
          event.target.dispatchEvent(cloneKeyboardEvent(event));
      }
      Private.replayKeyEvent = replayKeyEvent;
      /**
       * A singleton empty string function.
       */
      var emptyStringFunc = function () { return ''; };
      /**
       * A singleton `-1` number function
       */
      var negativeOneFunc = function () { return -1; };
      /**
       * A singleton true boolean function.
       */
      var trueFunc = function () { return true; };
      /**
       * A singleton false boolean function.
       */
      var falseFunc = function () { return false; };
      /**
       * A singleton empty dataset function.
       */
      var emptyDatasetFunc = function () { return ({}); };
      /**
       * A singleton undefined function
       */
      var undefinedFunc = function () { return undefined; };
      /**
       * Cast a value or command func to a command func.
       */
      function asFunc(value, dfault) {
          if (value === undefined) {
              return dfault;
          }
          if (typeof value === 'function') {
              return value;
          }
          return function () { return value; };
      }
      /**
       * Validate the selector for an options object.
       *
       * This returns the validated selector, or throws if the selector is
       * invalid or contains commas.
       */
      function validateSelector(options) {
          if (options.selector.indexOf(',') !== -1) {
              throw new Error("Selector cannot contain commas: " + options.selector);
          }
          if (!Selector.isValid(options.selector)) {
              throw new Error("Invalid selector: " + options.selector);
          }
          return options.selector;
      }
      /**
       * Test whether a key binding sequence matches a key sequence.
       *
       * Returns a `SequenceMatch` value indicating the type of match.
       */
      function matchSequence(bindKeys, userKeys) {
          if (bindKeys.length < userKeys.length) {
              return 0 /* None */;
          }
          for (var i = 0, n = userKeys.length; i < n; ++i) {
              if (bindKeys[i] !== userKeys[i]) {
                  return 0 /* None */;
              }
          }
          if (bindKeys.length > userKeys.length) {
              return 2 /* Partial */;
          }
          return 1 /* Exact */;
      }
      /**
       * Find the distance from the target node to the first matching node.
       *
       * This traverses the event path from `target` to `currentTarget` and
       * computes the distance from `target` to the first node which matches
       * the CSS selector. If no match is found, `-1` is returned.
       */
      function targetDistance(selector, event) {
          var targ = event.target;
          var curr = event.currentTarget;
          for (var dist = 0; targ !== null; targ = targ.parentElement, ++dist) {
              if (targ.hasAttribute('data-lm-suppress-shortcuts')) {
                  return -1;
              }
              /* <DEPRECATED> */
              if (targ.hasAttribute('data-p-suppress-shortcuts')) {
                  return -1;
              }
              /* </DEPRECATED> */
              if (Selector.matches(targ, selector)) {
                  return dist;
              }
              if (targ === curr) {
                  return -1;
              }
          }
          return -1;
      }
      /**
       * Clone a keyboard event.
       */
      function cloneKeyboardEvent(event) {
          // A custom event is required because Chrome nulls out the
          // `keyCode` field in user-generated `KeyboardEvent` types.
          var clone = document.createEvent('Event');
          var bubbles = event.bubbles || true;
          var cancelable = event.cancelable || true;
          clone.initEvent(event.type || 'keydown', bubbles, cancelable);
          clone.key = event.key || '';
          clone.keyCode = event.keyCode || 0;
          clone.which = event.keyCode || 0;
          clone.ctrlKey = event.ctrlKey || false;
          clone.altKey = event.altKey || false;
          clone.shiftKey = event.shiftKey || false;
          clone.metaKey = event.metaKey || false;
          clone.view = event.view || window;
          return clone;
      }
  })(Private$6 || (Private$6 = {}));

  /*! *****************************************************************************
  Copyright (c) Microsoft Corporation. All rights reserved.
  Licensed under the Apache License, Version 2.0 (the "License"); you may not use
  this file except in compliance with the License. You may obtain a copy of the
  License at http://www.apache.org/licenses/LICENSE-2.0

  THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
  KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
  WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
  MERCHANTABLITY OR NON-INFRINGEMENT.

  See the Apache Version 2.0 License for specific language governing permissions
  and limitations under the License.
  ***************************************************************************** */
  /* global Reflect, Promise */

  var extendStatics$2 = function(d, b) {
      extendStatics$2 = Object.setPrototypeOf ||
          ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
          function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
      return extendStatics$2(d, b);
  };

  function __extends$4(d, b) {
      extendStatics$2(d, b);
      function __() { this.constructor = d; }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
  }

  /**
   * A virtual node which represents plain text content.
   *
   * #### Notes
   * User code will not typically create a `VirtualText` node directly.
   * Instead, the `h()` function will be used to create an element tree.
   */
  var VirtualText = /** @class */ (function () {
      /**
       * Construct a new virtual text node.
       *
       * @param content - The text content for the node.
       */
      function VirtualText(content) {
          /**
           * The type of the node.
           *
           * This value can be used as a type guard for discriminating the
           * `VirtualNode` union type.
           */
          this.type = 'text';
          this.content = content;
      }
      return VirtualText;
  }());
  /**
   * A virtual node which represents an HTML element.
   *
   * #### Notes
   * User code will not typically create a `VirtualElement` node directly.
   * Instead, the `h()` function will be used to create an element tree.
   */
  var VirtualElement = /** @class */ (function () {
      /**
       * Construct a new virtual element node.
       *
       * @param tag - The element tag name.
       *
       * @param attrs - The element attributes.
       *
       * @param children - The element children.
       *
       * @param renderer - An optional custom renderer for the element.
       */
      function VirtualElement(tag, attrs, children, renderer) {
          /**
           * The type of the node.
           *
           * This value can be used as a type guard for discriminating the
           * `VirtualNode` union type.
           */
          this.type = 'element';
          this.tag = tag;
          this.attrs = attrs;
          this.children = children;
          this.renderer = renderer;
      }
      return VirtualElement;
  }());
  /**
   * DEPRECATED - use VirtualElement with a defined renderer param instead.
   * This class is provided as a backwards compatibility shim
   *
   * A "pass thru" virtual node whose children are managed by a render and an
   * unrender callback. The intent of this flavor of virtual node is to make
   * it easy to blend other kinds of virtualdom (eg React) into Phosphor's
   * virtualdom.
   *
   * #### Notes
   * User code will not typically create a `VirtualElementPass` node directly.
   * Instead, the `hpass()` function will be used to create an element tree.
   */
  var VirtualElementPass = /** @class */ (function (_super) {
      __extends$4(VirtualElementPass, _super);
      /**
       * DEPRECATED - use VirtualElement with a defined renderer param instead
       *
       * Construct a new virtual element pass thru node.
       *
       * @param tag - the tag of the parent element of this node. Once the parent
       * element is rendered, it will be passed as an argument to
       * renderer.render
       *
       * @param attrs - attributes that will assigned to the
       * parent element
       *
       * @param renderer - an object with render and unrender
       * functions, each of which should take a single argument of type
       * HTMLElement and return nothing. If null, the parent element
       * will be rendered barren without any children.
       */
      function VirtualElementPass(tag, attrs, renderer) {
          return _super.call(this, tag, attrs, [], renderer || undefined) || this;
      }
      return VirtualElementPass;
  }(VirtualElement));
  function h(tag) {
      var attrs = {};
      var renderer;
      var children = [];
      for (var i = 1, n = arguments.length; i < n; ++i) {
          var arg = arguments[i];
          if (typeof arg === 'string') {
              children.push(new VirtualText(arg));
          }
          else if (arg instanceof VirtualText) {
              children.push(arg);
          }
          else if (arg instanceof VirtualElement) {
              children.push(arg);
          }
          else if (arg instanceof Array) {
              extend(children, arg);
          }
          else if ((i === 1 || i === 2) && arg && typeof arg === 'object') {
              if ("render" in arg) {
                  renderer = arg;
              }
              else {
                  attrs = arg;
              }
          }
      }
      return new VirtualElement(tag, attrs, children, renderer);
      function extend(array, values) {
          for (var _i = 0, values_1 = values; _i < values_1.length; _i++) {
              var child = values_1[_i];
              if (typeof child === 'string') {
                  array.push(new VirtualText(child));
              }
              else if (child instanceof VirtualText) {
                  array.push(child);
              }
              else if (child instanceof VirtualElement) {
                  array.push(child);
              }
          }
      }
  }
  /**
   * The namespace for the `h` function statics.
   */
  (function (h) {
      h.a = h.bind(undefined, 'a');
      h.abbr = h.bind(undefined, 'abbr');
      h.address = h.bind(undefined, 'address');
      h.area = h.bind(undefined, 'area');
      h.article = h.bind(undefined, 'article');
      h.aside = h.bind(undefined, 'aside');
      h.audio = h.bind(undefined, 'audio');
      h.b = h.bind(undefined, 'b');
      h.bdi = h.bind(undefined, 'bdi');
      h.bdo = h.bind(undefined, 'bdo');
      h.blockquote = h.bind(undefined, 'blockquote');
      h.br = h.bind(undefined, 'br');
      h.button = h.bind(undefined, 'button');
      h.canvas = h.bind(undefined, 'canvas');
      h.caption = h.bind(undefined, 'caption');
      h.cite = h.bind(undefined, 'cite');
      h.code = h.bind(undefined, 'code');
      h.col = h.bind(undefined, 'col');
      h.colgroup = h.bind(undefined, 'colgroup');
      h.data = h.bind(undefined, 'data');
      h.datalist = h.bind(undefined, 'datalist');
      h.dd = h.bind(undefined, 'dd');
      h.del = h.bind(undefined, 'del');
      h.dfn = h.bind(undefined, 'dfn');
      h.div = h.bind(undefined, 'div');
      h.dl = h.bind(undefined, 'dl');
      h.dt = h.bind(undefined, 'dt');
      h.em = h.bind(undefined, 'em');
      h.embed = h.bind(undefined, 'embed');
      h.fieldset = h.bind(undefined, 'fieldset');
      h.figcaption = h.bind(undefined, 'figcaption');
      h.figure = h.bind(undefined, 'figure');
      h.footer = h.bind(undefined, 'footer');
      h.form = h.bind(undefined, 'form');
      h.h1 = h.bind(undefined, 'h1');
      h.h2 = h.bind(undefined, 'h2');
      h.h3 = h.bind(undefined, 'h3');
      h.h4 = h.bind(undefined, 'h4');
      h.h5 = h.bind(undefined, 'h5');
      h.h6 = h.bind(undefined, 'h6');
      h.header = h.bind(undefined, 'header');
      h.hr = h.bind(undefined, 'hr');
      h.i = h.bind(undefined, 'i');
      h.iframe = h.bind(undefined, 'iframe');
      h.img = h.bind(undefined, 'img');
      h.input = h.bind(undefined, 'input');
      h.ins = h.bind(undefined, 'ins');
      h.kbd = h.bind(undefined, 'kbd');
      h.label = h.bind(undefined, 'label');
      h.legend = h.bind(undefined, 'legend');
      h.li = h.bind(undefined, 'li');
      h.main = h.bind(undefined, 'main');
      h.map = h.bind(undefined, 'map');
      h.mark = h.bind(undefined, 'mark');
      h.meter = h.bind(undefined, 'meter');
      h.nav = h.bind(undefined, 'nav');
      h.noscript = h.bind(undefined, 'noscript');
      h.object = h.bind(undefined, 'object');
      h.ol = h.bind(undefined, 'ol');
      h.optgroup = h.bind(undefined, 'optgroup');
      h.option = h.bind(undefined, 'option');
      h.output = h.bind(undefined, 'output');
      h.p = h.bind(undefined, 'p');
      h.param = h.bind(undefined, 'param');
      h.pre = h.bind(undefined, 'pre');
      h.progress = h.bind(undefined, 'progress');
      h.q = h.bind(undefined, 'q');
      h.rp = h.bind(undefined, 'rp');
      h.rt = h.bind(undefined, 'rt');
      h.ruby = h.bind(undefined, 'ruby');
      h.s = h.bind(undefined, 's');
      h.samp = h.bind(undefined, 'samp');
      h.section = h.bind(undefined, 'section');
      h.select = h.bind(undefined, 'select');
      h.small = h.bind(undefined, 'small');
      h.source = h.bind(undefined, 'source');
      h.span = h.bind(undefined, 'span');
      h.strong = h.bind(undefined, 'strong');
      h.sub = h.bind(undefined, 'sub');
      h.summary = h.bind(undefined, 'summary');
      h.sup = h.bind(undefined, 'sup');
      h.table = h.bind(undefined, 'table');
      h.tbody = h.bind(undefined, 'tbody');
      h.td = h.bind(undefined, 'td');
      h.textarea = h.bind(undefined, 'textarea');
      h.tfoot = h.bind(undefined, 'tfoot');
      h.th = h.bind(undefined, 'th');
      h.thead = h.bind(undefined, 'thead');
      h.time = h.bind(undefined, 'time');
      h.title = h.bind(undefined, 'title');
      h.tr = h.bind(undefined, 'tr');
      h.track = h.bind(undefined, 'track');
      h.u = h.bind(undefined, 'u');
      h.ul = h.bind(undefined, 'ul');
      h.var_ = h.bind(undefined, 'var');
      h.video = h.bind(undefined, 'video');
      h.wbr = h.bind(undefined, 'wbr');
  })(h || (h = {}));
  /**
   * The namespace for the virtual DOM rendering functions.
   */
  var VirtualDOM;
  (function (VirtualDOM) {
      function realize(node) {
          return Private$7.createDOMNode(node);
      }
      VirtualDOM.realize = realize;
      /**
       * Render virtual DOM content into a host element.
       *
       * @param content - The virtual DOM content to render.
       *
       * @param host - The host element for the rendered content.
       *
       * #### Notes
       * This renders the delta from the previous rendering. It assumes that
       * the content of the host element is not manipulated by external code.
       *
       * Providing `null` content will clear the rendering.
       *
       * Externally modifying the provided content or the host element will
       * result in undefined rendering behavior.
       */
      function render(content, host) {
          var oldContent = Private$7.hostMap.get(host) || [];
          var newContent = Private$7.asContentArray(content);
          Private$7.hostMap.set(host, newContent);
          Private$7.updateContent(host, oldContent, newContent);
      }
      VirtualDOM.render = render;
  })(VirtualDOM || (VirtualDOM = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$7;
  (function (Private) {
      /**
       * A weak mapping of host element to virtual DOM content.
       */
      Private.hostMap = new WeakMap();
      /**
       * Cast a content value to a content array.
       */
      function asContentArray(value) {
          if (!value) {
              return [];
          }
          if (value instanceof Array) {
              return value;
          }
          return [value];
      }
      Private.asContentArray = asContentArray;
      function createDOMNode(node) {
          var host = arguments[1] || null;
          var before = arguments[2] || null;
          if (host) {
              host.insertBefore(createDOMNode(node), before);
          }
          else {
              // Create a text node for a virtual text node.
              if (node.type === 'text') {
                  return document.createTextNode(node.content);
              }
              // Create the HTML element with the specified tag.
              host = document.createElement(node.tag);
              // Add the attributes for the new element.
              addAttrs(host, node.attrs);
              if (node.renderer) {
                  node.renderer.render(host, { attrs: node.attrs, children: node.children });
                  return host;
              }
              // Recursively populate the element with child content.
              for (var i = 0, n = node.children.length; i < n; ++i) {
                  createDOMNode(node.children[i], host);
              }
          }
          return host;
      }
      Private.createDOMNode = createDOMNode;
      /**
       * Update a host element with the delta of the virtual content.
       *
       * This is the core "diff" algorithm. There is no explicit "patch"
       * phase. The host is patched at each step as the diff progresses.
       */
      function updateContent(host, oldContent, newContent) {
          // Bail early if the content is identical.
          if (oldContent === newContent) {
              return;
          }
          // Collect the old keyed elems into a mapping.
          var oldKeyed = collectKeys(host, oldContent);
          // Create a copy of the old content which can be modified in-place.
          var oldCopy = oldContent.slice();
          // Update the host with the new content. The diff always proceeds
          // forward and never modifies a previously visited index. The old
          // copy array is modified in-place to reflect the changes made to
          // the host children. This causes the stale nodes to be pushed to
          // the end of the host node and removed at the end of the loop.
          var currElem = host.firstChild;
          var newCount = newContent.length;
          for (var i = 0; i < newCount; ++i) {
              // If the old content is exhausted, create a new node.
              if (i >= oldCopy.length) {
                  createDOMNode(newContent[i], host);
                  continue;
              }
              // Lookup the old and new virtual nodes.
              var oldVNode = oldCopy[i];
              var newVNode = newContent[i];
              // If both elements are identical, there is nothing to do.
              if (oldVNode === newVNode) {
                  currElem = currElem.nextSibling;
                  continue;
              }
              // Handle the simplest case of in-place text update first.
              if (oldVNode.type === 'text' && newVNode.type === 'text') {
                  currElem.textContent = newVNode.content;
                  currElem = currElem.nextSibling;
                  continue;
              }
              // If the old or new node is a text node, the other node is now
              // known to be an element node, so create and insert a new node.
              if (oldVNode.type === 'text' || newVNode.type === 'text') {
                  ArrayExt.insert(oldCopy, i, newVNode);
                  createDOMNode(newVNode, host, currElem);
                  continue;
              }
              // If the old XOR new node has a custom renderer,
              // create and insert a new node.
              if (!(oldVNode.renderer) != !(newVNode.renderer)) {
                  ArrayExt.insert(oldCopy, i, newVNode);
                  createDOMNode(newVNode, host, currElem);
                  continue;
              }
              // At this point, both nodes are known to be element nodes.
              // If the new elem is keyed, move an old keyed elem to the proper
              // location before proceeding with the diff. The search can start
              // at the current index, since the unmatched old keyed elems are
              // pushed forward in the old copy array.
              var newKey = newVNode.attrs.key;
              if (newKey && newKey in oldKeyed) {
                  var pair = oldKeyed[newKey];
                  if (pair.vNode !== oldVNode) {
                      ArrayExt.move(oldCopy, oldCopy.indexOf(pair.vNode, i + 1), i);
                      host.insertBefore(pair.element, currElem);
                      oldVNode = pair.vNode;
                      currElem = pair.element;
                  }
              }
              // If both elements are identical, there is nothing to do.
              if (oldVNode === newVNode) {
                  currElem = currElem.nextSibling;
                  continue;
              }
              // If the old elem is keyed and does not match the new elem key,
              // create a new node. This is necessary since the old keyed elem
              // may be matched at a later point in the diff.
              var oldKey = oldVNode.attrs.key;
              if (oldKey && oldKey !== newKey) {
                  ArrayExt.insert(oldCopy, i, newVNode);
                  createDOMNode(newVNode, host, currElem);
                  continue;
              }
              // If the tags are different, create a new node.
              if (oldVNode.tag !== newVNode.tag) {
                  ArrayExt.insert(oldCopy, i, newVNode);
                  createDOMNode(newVNode, host, currElem);
                  continue;
              }
              // At this point, the element can be updated in-place.
              // Update the element attributes.
              updateAttrs(currElem, oldVNode.attrs, newVNode.attrs);
              // Update the element content.
              if (newVNode.renderer) {
                  newVNode.renderer.render(currElem, { attrs: newVNode.attrs, children: newVNode.children });
              }
              else {
                  updateContent(currElem, oldVNode.children, newVNode.children);
              }
              // Step to the next sibling element.
              currElem = currElem.nextSibling;
          }
          // Cleanup stale DOM
          removeContent(host, oldCopy, newCount, true);
      }
      Private.updateContent = updateContent;
      /**
       * Handle cleanup of stale vdom and its associated DOM. The host node is
       * traversed recursively (in depth-first order), and any explicit cleanup
       * required by a child node is carried out when it is visited (eg if a node
       * has a custom renderer, the renderer.unrender function will be called).
       * Once the subtree beneath each child of host has been completely visited,
       * that child will be removed via a call to host.removeChild.
       */
      function removeContent(host, oldContent, newCount, _sentinel) {
          // Dispose of the old nodes pushed to the end of the host.
          for (var i = oldContent.length - 1; i >= newCount; --i) {
              var oldNode = oldContent[i];
              var child = (_sentinel ? host.lastChild : host.childNodes[i]);
              // recursively clean up host children
              if (oldNode.type === 'text') ;
              else if (oldNode.renderer && oldNode.renderer.unrender) {
                  oldNode.renderer.unrender(child, { attrs: oldNode.attrs, children: oldNode.children });
              }
              else {
                  removeContent(child, oldNode.children, 0, false);
              }
              if (_sentinel) {
                  host.removeChild(child);
              }
          }
      }
      /**
       * A set of special-cased attribute names.
       */
      var specialAttrs = {
          'key': true,
          'className': true,
          'htmlFor': true,
          'dataset': true,
          'style': true,
      };
      /**
       * Add element attributes to a newly created HTML element.
       */
      function addAttrs(element, attrs) {
          // Add the inline event listeners and node attributes.
          for (var name_1 in attrs) {
              if (name_1 in specialAttrs) {
                  continue;
              }
              if (name_1.substr(0, 2) === 'on') {
                  element[name_1] = attrs[name_1];
              }
              else {
                  element.setAttribute(name_1, attrs[name_1]);
              }
          }
          // Add the element `class` attribute.
          if (attrs.className !== undefined) {
              element.setAttribute('class', attrs.className);
          }
          // Add the element `for` attribute.
          if (attrs.htmlFor !== undefined) {
              element.setAttribute('for', attrs.htmlFor);
          }
          // Add the dataset values.
          if (attrs.dataset) {
              addDataset(element, attrs.dataset);
          }
          // Add the inline styles.
          if (attrs.style) {
              addStyle(element, attrs.style);
          }
      }
      /**
       * Update the element attributes of an HTML element.
       */
      function updateAttrs(element, oldAttrs, newAttrs) {
          // Do nothing if the attrs are the same object.
          if (oldAttrs === newAttrs) {
              return;
          }
          // Setup the strongly typed loop variable.
          var name;
          // Remove attributes and listeners which no longer exist.
          for (name in oldAttrs) {
              if (name in specialAttrs || name in newAttrs) {
                  continue;
              }
              if (name.substr(0, 2) === 'on') {
                  element[name] = null;
              }
              else {
                  element.removeAttribute(name);
              }
          }
          // Add and update new and existing attributes and listeners.
          for (name in newAttrs) {
              if (name in specialAttrs || oldAttrs[name] === newAttrs[name]) {
                  continue;
              }
              if (name.substr(0, 2) === 'on') {
                  element[name] = newAttrs[name];
              }
              else {
                  element.setAttribute(name, newAttrs[name]);
              }
          }
          // Update the element `class` attribute.
          if (oldAttrs.className !== newAttrs.className) {
              if (newAttrs.className !== undefined) {
                  element.setAttribute('class', newAttrs.className);
              }
              else {
                  element.removeAttribute('class');
              }
          }
          // Add the element `for` attribute.
          if (oldAttrs.htmlFor !== newAttrs.htmlFor) {
              if (newAttrs.htmlFor !== undefined) {
                  element.setAttribute('for', newAttrs.htmlFor);
              }
              else {
                  element.removeAttribute('for');
              }
          }
          // Update the dataset values.
          if (oldAttrs.dataset !== newAttrs.dataset) {
              updateDataset(element, oldAttrs.dataset || {}, newAttrs.dataset || {});
          }
          // Update the inline styles.
          if (oldAttrs.style !== newAttrs.style) {
              updateStyle(element, oldAttrs.style || {}, newAttrs.style || {});
          }
      }
      /**
       * Add dataset values to a newly created HTML element.
       */
      function addDataset(element, dataset) {
          for (var name_2 in dataset) {
              element.setAttribute("data-" + name_2, dataset[name_2]);
          }
      }
      /**
       * Update the dataset values of an HTML element.
       */
      function updateDataset(element, oldDataset, newDataset) {
          for (var name_3 in oldDataset) {
              if (!(name_3 in newDataset)) {
                  element.removeAttribute("data-" + name_3);
              }
          }
          for (var name_4 in newDataset) {
              if (oldDataset[name_4] !== newDataset[name_4]) {
                  element.setAttribute("data-" + name_4, newDataset[name_4]);
              }
          }
      }
      /**
       * Add inline style values to a newly created HTML element.
       */
      function addStyle(element, style) {
          var elemStyle = element.style;
          var name;
          for (name in style) {
              elemStyle[name] = style[name];
          }
      }
      /**
       * Update the inline style values of an HTML element.
       */
      function updateStyle(element, oldStyle, newStyle) {
          var elemStyle = element.style;
          var name;
          for (name in oldStyle) {
              if (!(name in newStyle)) {
                  elemStyle[name] = '';
              }
          }
          for (name in newStyle) {
              if (oldStyle[name] !== newStyle[name]) {
                  elemStyle[name] = newStyle[name];
              }
          }
      }
      /**
       * Collect a mapping of keyed elements for the host content.
       */
      function collectKeys(host, content) {
          var node = host.firstChild;
          var keyMap = Object.create(null);
          for (var _i = 0, content_1 = content; _i < content_1.length; _i++) {
              var vNode = content_1[_i];
              if (vNode.type === 'element' && vNode.attrs.key) {
                  keyMap[vNode.attrs.key] = { vNode: vNode, element: node };
              }
              node = node.nextSibling;
          }
          return keyMap;
      }
  })(Private$7 || (Private$7 = {}));

  /**
   * An object which manages a drag-drop operation.
   *
   * A drag object dispatches four different events to drop targets:
   *
   * - `'lm-dragenter'` - Dispatched when the mouse enters the target
   *   element. This event must be canceled in order to receive any
   *   of the other events.
   *
   * - `'lm-dragover'` - Dispatched when the mouse moves over the drop
   *   target. It must cancel the event and set the `dropAction` to one
   *   of the supported actions in order to receive drop events.
   *
   * - `'lm-dragleave'` - Dispatched when the mouse leaves the target
   *   element. This includes moving the mouse into child elements.
   *
   * - `'lm-drop'`- Dispatched when the mouse is released over the target
   *   element when the target indicates an appropriate drop action. If
   *   the event is canceled, the indicated drop action is returned to
   *   the initiator through the resolved promise.
   *
   * A drag operation can be terminated at any time by pressing `Escape`
   * or by disposing the drag object.
   *
   * A drag object has the ability to automatically scroll a scrollable
   * element when the mouse is hovered near one of its edges. To enable
   * this, add the `data-lm-dragscroll` attribute to any element which
   * the drag object should consider for scrolling.
   *
   * #### Notes
   * This class is designed to be used when dragging and dropping custom
   * data *within* a single application. It is *not* a replacement for
   * the native drag-drop API. Instead, it provides an API which allows
   * drag operations to be initiated programmatically and enables the
   * transfer of arbitrary non-string objects; features which are not
   * possible with the native drag-drop API.
   */
  var Drag = /** @class */ (function () {
      /**
       * Construct a new drag object.
       *
       * @param options - The options for initializing the drag.
       */
      function Drag(options) {
          var _this = this;
          /**
           * The scroll loop handler function.
           */
          this._onScrollFrame = function () {
              // Bail early if there is no scroll target.
              if (!_this._scrollTarget) {
                  return;
              }
              // Unpack the scroll target.
              var _a = _this._scrollTarget, element = _a.element, edge = _a.edge, distance = _a.distance;
              // Calculate the scroll delta using nonlinear acceleration.
              var d = Private$8.SCROLL_EDGE_SIZE - distance;
              var f = Math.pow(d / Private$8.SCROLL_EDGE_SIZE, 2);
              var s = Math.max(1, Math.round(f * Private$8.SCROLL_EDGE_SIZE));
              // Scroll the element in the specified direction.
              switch (edge) {
                  case 'top':
                      element.scrollTop -= s;
                      break;
                  case 'left':
                      element.scrollLeft -= s;
                      break;
                  case 'right':
                      element.scrollLeft += s;
                      break;
                  case 'bottom':
                      element.scrollTop += s;
                      break;
              }
              // Request the next cycle of the scroll loop.
              requestAnimationFrame(_this._onScrollFrame);
          };
          this._disposed = false;
          this._dropAction = 'none';
          this._override = null;
          this._currentTarget = null;
          this._currentElement = null;
          this._promise = null;
          this._scrollTarget = null;
          this._resolve = null;
          this.mimeData = options.mimeData;
          this.dragImage = options.dragImage || null;
          this.proposedAction = options.proposedAction || 'copy';
          this.supportedActions = options.supportedActions || 'all';
          this.source = options.source || null;
      }
      /**
       * Dispose of the resources held by the drag object.
       *
       * #### Notes
       * This will cancel the drag operation if it is active.
       */
      Drag.prototype.dispose = function () {
          // Do nothing if the drag object is already disposed.
          if (this._disposed) {
              return;
          }
          this._disposed = true;
          // If there is a current target, dispatch a drag leave event.
          if (this._currentTarget) {
              var event_1 = Private$8.createMouseEvent('mouseup', -1, -1);
              Private$8.dispatchDragLeave(this, this._currentTarget, null, event_1);
          }
          // Finalize the drag object with `'none'`.
          this._finalize('none');
      };
      Object.defineProperty(Drag.prototype, "isDisposed", {
          /**
           * Test whether the drag object is disposed.
           */
          get: function () {
              return this._disposed;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Start the drag operation at the specified client position.
       *
       * @param clientX - The client X position for the drag start.
       *
       * @param clientY - The client Y position for the drag start.
       *
       * @returns A promise which resolves to the result of the drag.
       *
       * #### Notes
       * If the drag has already been started, the promise created by the
       * first call to `start` is returned.
       *
       * If the drag operation has ended, or if the drag object has been
       * disposed, the returned promise will resolve to `'none'`.
       *
       * The drag object will be automatically disposed when drag operation
       * completes. This means `Drag` objects are for single-use only.
       *
       * This method assumes the left mouse button is already held down.
       */
      Drag.prototype.start = function (clientX, clientY) {
          var _this = this;
          // If the drag object is already disposed, resolve to `None`.
          if (this._disposed) {
              return Promise.resolve('none');
          }
          // If the drag has already been started, return the promise.
          if (this._promise) {
              return this._promise;
          }
          // Install the document listeners for the drag object.
          this._addListeners();
          // Attach the drag image at the specified client position.
          this._attachDragImage(clientX, clientY);
          // Create the promise which will be resolved on completion.
          this._promise = new Promise(function (resolve, reject) {
              _this._resolve = resolve;
          });
          // Trigger a fake move event to kick off the drag operation.
          var event = Private$8.createMouseEvent('mousemove', clientX, clientY);
          document.dispatchEvent(event);
          // Return the pending promise for the drag operation.
          return this._promise;
      };
      /**
       * Handle the DOM events for the drag operation.
       *
       * @param event - The DOM event sent to the drag object.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the document. It should not be
       * called directly by user code.
       */
      Drag.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseup':
                  this._evtMouseUp(event);
                  break;
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              default:
                  // Stop all other events during drag-drop.
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * Handle the `'mousemove'` event for the drag object.
       */
      Drag.prototype._evtMouseMove = function (event) {
          // Stop all input events during drag-drop.
          event.preventDefault();
          event.stopPropagation();
          // Update the current target node and dispatch enter/leave events.
          this._updateCurrentTarget(event);
          // Update the drag scroll element.
          this._updateDragScroll(event);
          // Move the drag image to the specified client position. This is
          // performed *after* dispatching to prevent unnecessary reflows.
          this._moveDragImage(event.clientX, event.clientY);
      };
      /**
       * Handle the `'mouseup'` event for the drag object.
       */
      Drag.prototype._evtMouseUp = function (event) {
          // Stop all input events during drag-drop.
          event.preventDefault();
          event.stopPropagation();
          // Do nothing if the left button is not released.
          if (event.button !== 0) {
              return;
          }
          // Update the current target node and dispatch enter/leave events.
          // This prevents a subtle issue where the DOM mutates under the
          // cursor after the last move event but before the drop event.
          this._updateCurrentTarget(event);
          // If there is no current target, finalize with `'none'`.
          if (!this._currentTarget) {
              this._finalize('none');
              return;
          }
          // If the last drop action was `'none'`, dispatch a leave event
          // to the current target and finalize the drag with `'none'`.
          if (this._dropAction === 'none') {
              Private$8.dispatchDragLeave(this, this._currentTarget, null, event);
              this._finalize('none');
              return;
          }
          // Dispatch the drop event at the current target and finalize
          // with the resulting drop action.
          var action = Private$8.dispatchDrop(this, this._currentTarget, event);
          this._finalize(action);
      };
      /**
       * Handle the `'keydown'` event for the drag object.
       */
      Drag.prototype._evtKeyDown = function (event) {
          // Stop all input events during drag-drop.
          event.preventDefault();
          event.stopPropagation();
          // Cancel the drag if `Escape` is pressed.
          if (event.keyCode === 27) {
              this.dispose();
          }
      };
      /**
       * Add the document event listeners for the drag object.
       */
      Drag.prototype._addListeners = function () {
          document.addEventListener('mousedown', this, true);
          document.addEventListener('mousemove', this, true);
          document.addEventListener('mouseup', this, true);
          document.addEventListener('mouseenter', this, true);
          document.addEventListener('mouseleave', this, true);
          document.addEventListener('mouseover', this, true);
          document.addEventListener('mouseout', this, true);
          document.addEventListener('keydown', this, true);
          document.addEventListener('keyup', this, true);
          document.addEventListener('keypress', this, true);
          document.addEventListener('contextmenu', this, true);
      };
      /**
       * Remove the document event listeners for the drag object.
       */
      Drag.prototype._removeListeners = function () {
          document.removeEventListener('mousedown', this, true);
          document.removeEventListener('mousemove', this, true);
          document.removeEventListener('mouseup', this, true);
          document.removeEventListener('mouseenter', this, true);
          document.removeEventListener('mouseleave', this, true);
          document.removeEventListener('mouseover', this, true);
          document.removeEventListener('mouseout', this, true);
          document.removeEventListener('keydown', this, true);
          document.removeEventListener('keyup', this, true);
          document.removeEventListener('keypress', this, true);
          document.removeEventListener('contextmenu', this, true);
      };
      /**
       * Update the drag scroll element under the mouse.
       */
      Drag.prototype._updateDragScroll = function (event) {
          // Find the scroll target under the mouse.
          var target = Private$8.findScrollTarget(event);
          // Bail if there is nothing to scroll.
          if (!this._scrollTarget && !target) {
              return;
          }
          // Start the scroll loop if needed.
          if (!this._scrollTarget) {
              setTimeout(this._onScrollFrame, 500);
          }
          // Update the scroll target.
          this._scrollTarget = target;
      };
      /**
       * Update the current target node using the given mouse event.
       */
      Drag.prototype._updateCurrentTarget = function (event) {
          // Fetch common local state.
          var prevTarget = this._currentTarget;
          var currTarget = this._currentTarget;
          var prevElem = this._currentElement;
          // Find the current indicated element at the given position.
          var currElem = document.elementFromPoint(event.clientX, event.clientY);
          // Update the current element reference.
          this._currentElement = currElem;
          // If the indicated element changes from the previous iteration,
          // and is different from the current target, dispatch the exit
          // event to the target.
          if (currElem !== prevElem && currElem !== currTarget) {
              Private$8.dispatchDragExit(this, currTarget, currElem, event);
          }
          // If the indicated element changes from the previous iteration,
          // and is different from the current target, dispatch the enter
          // event and compute the new target element.
          if (currElem !== prevElem && currElem !== currTarget) {
              currTarget = Private$8.dispatchDragEnter(this, currElem, currTarget, event);
          }
          // If the current target element has changed, update the current
          // target reference and dispatch the leave event to the old target.
          if (currTarget !== prevTarget) {
              this._currentTarget = currTarget;
              Private$8.dispatchDragLeave(this, prevTarget, currTarget, event);
          }
          // Dispatch the drag over event and update the drop action.
          var action = Private$8.dispatchDragOver(this, currTarget, event);
          this._setDropAction(action);
      };
      /**
       * Attach the drag image element at the specified location.
       *
       * This is a no-op if there is no drag image element.
       */
      Drag.prototype._attachDragImage = function (clientX, clientY) {
          if (!this.dragImage) {
              return;
          }
          this.dragImage.classList.add('lm-mod-drag-image');
          /* <DEPRECATED> */
          this.dragImage.classList.add('p-mod-drag-image');
          /* </DEPRECATED> */
          var style = this.dragImage.style;
          style.pointerEvents = 'none';
          style.position = 'fixed';
          style.top = clientY + "px";
          style.left = clientX + "px";
          document.body.appendChild(this.dragImage);
      };
      /**
       * Move the drag image element to the specified location.
       *
       * This is a no-op if there is no drag image element.
       */
      Drag.prototype._moveDragImage = function (clientX, clientY) {
          if (!this.dragImage) {
              return;
          }
          var style = this.dragImage.style;
          style.top = clientY + "px";
          style.left = clientX + "px";
      };
      /**
       * Detach the drag image element from the DOM.
       *
       * This is a no-op if there is no drag image element.
       */
      Drag.prototype._detachDragImage = function () {
          if (!this.dragImage) {
              return;
          }
          var parent = this.dragImage.parentNode;
          if (!parent) {
              return;
          }
          parent.removeChild(this.dragImage);
      };
      /**
       * Set the internal drop action state and update the drag cursor.
       */
      Drag.prototype._setDropAction = function (action) {
          action = Private$8.validateAction(action, this.supportedActions);
          if (this._override && this._dropAction === action) {
              return;
          }
          switch (action) {
              case 'none':
                  this._dropAction = action;
                  this._override = Drag.overrideCursor('no-drop');
                  break;
              case 'copy':
                  this._dropAction = action;
                  this._override = Drag.overrideCursor('copy');
                  break;
              case 'link':
                  this._dropAction = action;
                  this._override = Drag.overrideCursor('alias');
                  break;
              case 'move':
                  this._dropAction = action;
                  this._override = Drag.overrideCursor('move');
                  break;
          }
      };
      /**
       * Finalize the drag operation and resolve the drag promise.
       */
      Drag.prototype._finalize = function (action) {
          // Store the resolve function as a temp variable.
          var resolve = this._resolve;
          // Remove the document event listeners.
          this._removeListeners();
          // Detach the drag image.
          this._detachDragImage();
          // Dispose of the cursor override.
          if (this._override) {
              this._override.dispose();
              this._override = null;
          }
          // Clear the mime data.
          this.mimeData.clear();
          // Clear the rest of the internal drag state.
          this._disposed = true;
          this._dropAction = 'none';
          this._currentTarget = null;
          this._currentElement = null;
          this._scrollTarget = null;
          this._promise = null;
          this._resolve = null;
          // Finally, resolve the promise to the given drop action.
          if (resolve) {
              resolve(action);
          }
      };
      return Drag;
  }());
  /**
   * The namespace for the `Drag` class statics.
   */
  (function (Drag) {
      /**
       * Override the cursor icon for the entire document.
       *
       * @param cursor - The string representing the cursor style.
       *
       * @returns A disposable which will clear the override when disposed.
       *
       * #### Notes
       * The most recent call to `overrideCursor` takes precedence.
       * Disposing an old override has no effect on the current override.
       *
       * This utility function is used by the `Drag` class to override the
       * mouse cursor during a drag-drop operation, but it can also be used
       * by other classes to fix the cursor icon during normal mouse drags.
       *
       * #### Example
       * ```typescript
       * import { Drag } from '@lumino/dragdrop';
       *
       * // Force the cursor to be 'wait' for the entire document.
       * let override = Drag.overrideCursor('wait');
       *
       * // Clear the override by disposing the return value.
       * override.dispose();
       * ```
       */
      function overrideCursor(cursor) {
          var id = ++overrideCursorID;
          document.body.style.cursor = cursor;
          document.body.classList.add('lm-mod-override-cursor');
          /* <DEPRECATED> */
          document.body.classList.add('p-mod-override-cursor');
          /* </DEPRECATED> */
          return new DisposableDelegate(function () {
              if (id === overrideCursorID) {
                  document.body.style.cursor = '';
                  document.body.classList.remove('lm-mod-override-cursor');
                  /* <DEPRECATED> */
                  document.body.classList.remove('p-mod-override-cursor');
                  /* </DEPRECATED> */
              }
          });
      }
      Drag.overrideCursor = overrideCursor;
      /**
       * The internal id for the active cursor override.
       */
      var overrideCursorID = 0;
  })(Drag || (Drag = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$8;
  (function (Private) {
      /**
       * The size of a drag scroll edge, in pixels.
       */
      Private.SCROLL_EDGE_SIZE = 20;
      /**
       * Validate the given action is one of the supported actions.
       *
       * Returns the given action or `'none'` if the action is unsupported.
       */
      function validateAction(action, supported) {
          return (actionTable[action] & supportedTable[supported]) ? action : 'none';
      }
      Private.validateAction = validateAction;
      /**
       * Create a left mouse event at the given position.
       *
       * @param type - The event type for the mouse event.
       *
       * @param clientX - The client X position.
       *
       * @param clientY - The client Y position.
       *
       * @returns A newly created and initialized mouse event.
       */
      function createMouseEvent(type, clientX, clientY) {
          var event = document.createEvent('MouseEvent');
          event.initMouseEvent(type, true, true, window, 0, 0, 0, clientX, clientY, false, false, false, false, 0, null);
          return event;
      }
      Private.createMouseEvent = createMouseEvent;
      /**
       * Find the drag scroll target under the mouse, if any.
       */
      function findScrollTarget(event) {
          // Look up the client mouse position.
          var x = event.clientX;
          var y = event.clientY;
          // Get the element under the mouse.
          var element = document.elementFromPoint(x, y);
          // Search for a scrollable target based on the mouse position.
          // The null assert in third clause of for-loop is required due to:
          // https://github.com/Microsoft/TypeScript/issues/14143
          for (; element; element = element.parentElement) {
              // Ignore elements which are not marked as scrollable.
              var scrollable = element.hasAttribute('data-lm-dragscroll');
              /* <DEPRECATED> */
              scrollable = scrollable || element.hasAttribute('data-p-dragscroll');
              /* </DEPRECATED> */
              if (!scrollable) {
                  continue;
              }
              // Set up the coordinate offsets for the element.
              var offsetX = 0;
              var offsetY = 0;
              if (element === document.body) {
                  offsetX = window.pageXOffset;
                  offsetY = window.pageYOffset;
              }
              // Get the element bounds in viewport coordinates.
              var r = element.getBoundingClientRect();
              var top_1 = r.top + offsetY;
              var left = r.left + offsetX;
              var right = left + r.width;
              var bottom = top_1 + r.height;
              // Skip the element if it's not under the mouse.
              if (x < left || x >= right || y < top_1 || y >= bottom) {
                  continue;
              }
              // Compute the distance to each edge.
              var dl = x - left + 1;
              var dt = y - top_1 + 1;
              var dr = right - x;
              var db = bottom - y;
              // Find the smallest of the edge distances.
              var distance = Math.min(dl, dt, dr, db);
              // Skip the element if the mouse is not within a scroll edge.
              if (distance > Private.SCROLL_EDGE_SIZE) {
                  continue;
              }
              // Set up the edge result variable.
              var edge = void 0;
              // Find the edge for the computed distance.
              switch (distance) {
                  case db:
                      edge = 'bottom';
                      break;
                  case dt:
                      edge = 'top';
                      break;
                  case dr:
                      edge = 'right';
                      break;
                  case dl:
                      edge = 'left';
                      break;
                  default:
                      throw 'unreachable';
              }
              // Compute how much the element can scroll in width and height.
              var dsw = element.scrollWidth - element.clientWidth;
              var dsh = element.scrollHeight - element.clientHeight;
              // Determine if the element should be scrolled for the edge.
              var shouldScroll = void 0;
              switch (edge) {
                  case 'top':
                      shouldScroll = dsh > 0 && element.scrollTop > 0;
                      break;
                  case 'left':
                      shouldScroll = dsw > 0 && element.scrollLeft > 0;
                      break;
                  case 'right':
                      shouldScroll = dsw > 0 && element.scrollLeft < dsw;
                      break;
                  case 'bottom':
                      shouldScroll = dsh > 0 && element.scrollTop < dsh;
                      break;
                  default:
                      throw 'unreachable';
              }
              // Skip the element if it should not be scrolled.
              if (!shouldScroll) {
                  continue;
              }
              // Return the drag scroll target.
              return { element: element, edge: edge, distance: distance };
          }
          // No drag scroll target was found.
          return null;
      }
      Private.findScrollTarget = findScrollTarget;
      /**
       * Dispatch a drag enter event to the indicated element.
       *
       * @param drag - The drag object associated with the action.
       *
       * @param currElem - The currently indicated element, or `null`. This
       *   is the "immediate user selection" from the whatwg spec.
       *
       * @param currTarget - The current drag target element, or `null`. This
       *   is the "current target element" from the whatwg spec.
       *
       * @param event - The mouse event related to the action.
       *
       * @returns The element to use as the current drag target. This is the
       *   "current target element" from the whatwg spec, and may be `null`.
       *
       * #### Notes
       * This largely implements the drag enter portion of the whatwg spec:
       * https://html.spec.whatwg.org/multipage/interaction.html#drag-and-drop-processing-model
       */
      function dispatchDragEnter(drag, currElem, currTarget, event) {
          // If the current element is null, return null as the new target.
          if (!currElem) {
              return null;
          }
          // Dispatch a drag enter event to the current element.
          var dragEvent = createDragEvent('lm-dragenter', drag, event, currTarget);
          var canceled = !currElem.dispatchEvent(dragEvent);
          // If the event was canceled, use the current element as the new target.
          if (canceled) {
              return currElem;
          }
          /* <DEPRECATED> */
          dragEvent = createDragEvent('p-dragenter', drag, event, currTarget);
          canceled = !currElem.dispatchEvent(dragEvent);
          if (canceled) {
              return currElem;
          }
          /* </DEPRECATED> */
          // If the current element is the document body, keep the original target.
          if (currElem === document.body) {
              return currTarget;
          }
          // Dispatch a drag enter event on the document body.
          dragEvent = createDragEvent('lm-dragenter', drag, event, currTarget);
          document.body.dispatchEvent(dragEvent);
          /* <DEPRECATED> */
          dragEvent = createDragEvent('p-dragenter', drag, event, currTarget);
          document.body.dispatchEvent(dragEvent);
          /* </DEPRECATED> */
          // Ignore the event cancellation, and use the body as the new target.
          return document.body;
      }
      Private.dispatchDragEnter = dispatchDragEnter;
      /**
       * Dispatch a drag exit event to the indicated element.
       *
       * @param drag - The drag object associated with the action.
       *
       * @param prevTarget - The previous target element, or `null`. This
       *   is the previous "current target element" from the whatwg spec.
       *
       * @param currTarget - The current drag target element, or `null`. This
       *   is the "current target element" from the whatwg spec.
       *
       * @param event - The mouse event related to the action.
       *
       * #### Notes
       * This largely implements the drag exit portion of the whatwg spec:
       * https://html.spec.whatwg.org/multipage/interaction.html#drag-and-drop-processing-model
       */
      function dispatchDragExit(drag, prevTarget, currTarget, event) {
          // If the previous target is null, do nothing.
          if (!prevTarget) {
              return;
          }
          // Dispatch the drag exit event to the previous target.
          var dragEvent = createDragEvent('lm-dragexit', drag, event, currTarget);
          prevTarget.dispatchEvent(dragEvent);
          /* <DEPRECATED> */
          dragEvent = createDragEvent('p-dragexit', drag, event, currTarget);
          prevTarget.dispatchEvent(dragEvent);
          /* </DEPRECATED> */
      }
      Private.dispatchDragExit = dispatchDragExit;
      /**
       * Dispatch a drag leave event to the indicated element.
       *
       * @param drag - The drag object associated with the action.
       *
       * @param prevTarget - The previous target element, or `null`. This
       *   is the previous "current target element" from the whatwg spec.
       *
       * @param currTarget - The current drag target element, or `null`. This
       *   is the "current target element" from the whatwg spec.
       *
       * @param event - The mouse event related to the action.
       *
       * #### Notes
       * This largely implements the drag leave portion of the whatwg spec:
       * https://html.spec.whatwg.org/multipage/interaction.html#drag-and-drop-processing-model
       */
      function dispatchDragLeave(drag, prevTarget, currTarget, event) {
          // If the previous target is null, do nothing.
          if (!prevTarget) {
              return;
          }
          // Dispatch the drag leave event to the previous target.
          var dragEvent = createDragEvent('lm-dragleave', drag, event, currTarget);
          prevTarget.dispatchEvent(dragEvent);
          /* <DEPRECATED> */
          dragEvent = createDragEvent('p-dragleave', drag, event, currTarget);
          prevTarget.dispatchEvent(dragEvent);
          /* </DEPRECATED> */
      }
      Private.dispatchDragLeave = dispatchDragLeave;
      /**
       * Dispatch a drag over event to the indicated element.
       *
       * @param drag - The drag object associated with the action.
       *
       * @param currTarget - The current drag target element, or `null`. This
       *   is the "current target element" from the whatwg spec.
       *
       * @param event - The mouse event related to the action.
       *
       * @returns The `DropAction` result of the drag over event.
       *
       * #### Notes
       * This largely implements the drag over portion of the whatwg spec:
       * https://html.spec.whatwg.org/multipage/interaction.html#drag-and-drop-processing-model
       */
      function dispatchDragOver(drag, currTarget, event) {
          // If there is no current target, the drop action is none.
          if (!currTarget) {
              return 'none';
          }
          // Dispatch the drag over event to the current target.
          var dragEvent = createDragEvent('lm-dragover', drag, event, null);
          var canceled = !currTarget.dispatchEvent(dragEvent);
          // If the event was canceled, return the drop action result.
          if (canceled) {
              return dragEvent.dropAction;
          }
          /* <DEPRECATED> */
          dragEvent = createDragEvent('p-dragover', drag, event, null);
          canceled = !currTarget.dispatchEvent(dragEvent);
          if (canceled) {
              return dragEvent.dropAction;
          }
          /* </DEPRECATED> */
          // Otherwise, the effective drop action is none.
          return 'none';
      }
      Private.dispatchDragOver = dispatchDragOver;
      /**
       * Dispatch a drop event to the indicated element.
       *
       * @param drag - The drag object associated with the action.
       *
       * @param currTarget - The current drag target element, or `null`. This
       *   is the "current target element" from the whatwg spec.
       *
       * @param event - The mouse event related to the action.
       *
       * @returns The `DropAction` result of the drop event.
       *
       * #### Notes
       * This largely implements the drag over portion of the whatwg spec:
       * https://html.spec.whatwg.org/multipage/interaction.html#drag-and-drop-processing-model
       */
      function dispatchDrop(drag, currTarget, event) {
          // If there is no current target, the drop action is none.
          if (!currTarget) {
              return 'none';
          }
          // Dispatch the drop event to the current target.
          var dragEvent = createDragEvent('lm-drop', drag, event, null);
          var canceled = !currTarget.dispatchEvent(dragEvent);
          // If the event was canceled, return the drop action result.
          if (canceled) {
              return dragEvent.dropAction;
          }
          /* <DEPRECATED> */
          dragEvent = createDragEvent('p-drop', drag, event, null);
          canceled = !currTarget.dispatchEvent(dragEvent);
          if (canceled) {
              return dragEvent.dropAction;
          }
          /* </DEPRECATED> */
          // Otherwise, the effective drop action is none.
          return 'none';
      }
      Private.dispatchDrop = dispatchDrop;
      /**
       * A lookup table from drop action to bit value.
       */
      var actionTable = {
          'none': 0x0,
          'copy': 0x1,
          'link': 0x2,
          'move': 0x4
      };
      /**
       * A lookup table from supported action to drop action bit mask.
       */
      var supportedTable = {
          'none': actionTable['none'],
          'copy': actionTable['copy'],
          'link': actionTable['link'],
          'move': actionTable['move'],
          'copy-link': actionTable['copy'] | actionTable['link'],
          'copy-move': actionTable['copy'] | actionTable['move'],
          'link-move': actionTable['link'] | actionTable['move'],
          'all': actionTable['copy'] | actionTable['link'] | actionTable['move']
      };
      /**
       * Create a new initialized `IDragEvent` from the given data.
       *
       * @param type - The event type for the drag event.
       *
       * @param drag - The drag object to use for seeding the drag data.
       *
       * @param event - The mouse event to use for seeding the mouse data.
       *
       * @param related - The related target for the event, or `null`.
       *
       * @returns A new object which implements `IDragEvent`.
       */
      function createDragEvent(type, drag, event, related) {
          // Create a new mouse event to use as the drag event. Currently,
          // JS engines do now allow user-defined Event subclasses.
          var dragEvent = document.createEvent('MouseEvent');
          // Initialize the mouse event data.
          dragEvent.initMouseEvent(type, true, true, window, 0, event.screenX, event.screenY, event.clientX, event.clientY, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, event.button, related);
          // Forcefully add the custom drag event properties.
          dragEvent.dropAction = 'none';
          dragEvent.mimeData = drag.mimeData;
          dragEvent.proposedAction = drag.proposedAction;
          dragEvent.supportedActions = drag.supportedActions;
          dragEvent.source = drag.source;
          // Return the fully initialized drag event.
          return dragEvent;
      }
  })(Private$8 || (Private$8 = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /*-----------------------------------------------------------------------------
  | Copyright (c) 2014-2017, PhosphorJS Contributors
  |
  | Distributed under the terms of the BSD 3-Clause License.
  |
  | The full license is in the file LICENSE, distributed with this software.
  |----------------------------------------------------------------------------*/
  /**
   * A sizer object for use with the box engine layout functions.
   *
   * #### Notes
   * A box sizer holds the geometry information for an object along an
   * arbitrary layout orientation.
   *
   * For best performance, this class should be treated as a raw data
   * struct. It should not typically be subclassed.
   */
  var BoxSizer = /** @class */ (function () {
      function BoxSizer() {
          /**
           * The preferred size for the sizer.
           *
           * #### Notes
           * The sizer will be given this initial size subject to its size
           * bounds. The sizer will not deviate from this size unless such
           * deviation is required to fit into the available layout space.
           *
           * There is no limit to this value, but it will be clamped to the
           * bounds defined by [[minSize]] and [[maxSize]].
           *
           * The default value is `0`.
           */
          this.sizeHint = 0;
          /**
           * The minimum size of the sizer.
           *
           * #### Notes
           * The sizer will never be sized less than this value, even if
           * it means the sizer will overflow the available layout space.
           *
           * It is assumed that this value lies in the range `[0, Infinity)`
           * and that it is `<=` to [[maxSize]]. Failure to adhere to this
           * constraint will yield undefined results.
           *
           * The default value is `0`.
           */
          this.minSize = 0;
          /**
           * The maximum size of the sizer.
           *
           * #### Notes
           * The sizer will never be sized greater than this value, even if
           * it means the sizer will underflow the available layout space.
           *
           * It is assumed that this value lies in the range `[0, Infinity]`
           * and that it is `>=` to [[minSize]]. Failure to adhere to this
           * constraint will yield undefined results.
           *
           * The default value is `Infinity`.
           */
          this.maxSize = Infinity;
          /**
           * The stretch factor for the sizer.
           *
           * #### Notes
           * This controls how much the sizer stretches relative to its sibling
           * sizers when layout space is distributed. A stretch factor of zero
           * is special and will cause the sizer to only be resized after all
           * other sizers with a stretch factor greater than zero have been
           * resized to their limits.
           *
           * It is assumed that this value is an integer that lies in the range
           * `[0, Infinity)`. Failure to adhere to this constraint will yield
           * undefined results.
           *
           * The default value is `1`.
           */
          this.stretch = 1;
          /**
           * The computed size of the sizer.
           *
           * #### Notes
           * This value is the output of a call to [[boxCalc]]. It represents
           * the computed size for the object along the layout orientation,
           * and will always lie in the range `[minSize, maxSize]`.
           *
           * This value is output only.
           *
           * Changing this value will have no effect.
           */
          this.size = 0;
          /**
           * An internal storage property for the layout algorithm.
           *
           * #### Notes
           * This value is used as temporary storage by the layout algorithm.
           *
           * Changing this value will have no effect.
           */
          this.done = false;
      }
      return BoxSizer;
  }());
  /**
   * The namespace for the box engine layout functions.
   */
  var BoxEngine;
  (function (BoxEngine) {
      /**
       * Calculate the optimal layout sizes for a sequence of box sizers.
       *
       * This distributes the available layout space among the box sizers
       * according to the following algorithm:
       *
       * 1. Initialize the sizers's size to its size hint and compute the
       *    sums for each of size hint, min size, and max size.
       *
       * 2. If the total size hint equals the available space, return.
       *
       * 3. If the available space is less than the total min size, set all
       *    sizers to their min size and return.
       *
       * 4. If the available space is greater than the total max size, set
       *    all sizers to their max size and return.
       *
       * 5. If the layout space is less than the total size hint, distribute
       *    the negative delta as follows:
       *
       *    a. Shrink each sizer with a stretch factor greater than zero by
       *       an amount proportional to the negative space and the sum of
       *       stretch factors. If the sizer reaches its min size, remove
       *       it and its stretch factor from the computation.
       *
       *    b. If after adjusting all stretch sizers there remains negative
       *       space, distribute the space equally among the sizers with a
       *       stretch factor of zero. If a sizer reaches its min size,
       *       remove it from the computation.
       *
       * 6. If the layout space is greater than the total size hint,
       *    distribute the positive delta as follows:
       *
       *    a. Expand each sizer with a stretch factor greater than zero by
       *       an amount proportional to the postive space and the sum of
       *       stretch factors. If the sizer reaches its max size, remove
       *       it and its stretch factor from the computation.
       *
       *    b. If after adjusting all stretch sizers there remains positive
       *       space, distribute the space equally among the sizers with a
       *       stretch factor of zero. If a sizer reaches its max size,
       *       remove it from the computation.
       *
       * 7. return
       *
       * @param sizers - The sizers for a particular layout line.
       *
       * @param space - The available layout space for the sizers.
       *
       * @returns The delta between the provided available space and the
       *   actual consumed space. This value will be zero if the sizers
       *   can be adjusted to fit, negative if the available space is too
       *   small, and positive if the available space is too large.
       *
       * #### Notes
       * The [[size]] of each sizer is updated with the computed size.
       *
       * This function can be called at any time to recompute the layout for
       * an existing sequence of sizers. The previously computed results will
       * have no effect on the new output. It is therefore not necessary to
       * create new sizer objects on each resize event.
       */
      function calc(sizers, space) {
          // Bail early if there is nothing to do.
          var count = sizers.length;
          if (count === 0) {
              return space;
          }
          // Setup the size and stretch counters.
          var totalMin = 0;
          var totalMax = 0;
          var totalSize = 0;
          var totalStretch = 0;
          var stretchCount = 0;
          // Setup the sizers and compute the totals.
          for (var i = 0; i < count; ++i) {
              var sizer = sizers[i];
              var min = sizer.minSize;
              var max = sizer.maxSize;
              var hint = sizer.sizeHint;
              sizer.done = false;
              sizer.size = Math.max(min, Math.min(hint, max));
              totalSize += sizer.size;
              totalMin += min;
              totalMax += max;
              if (sizer.stretch > 0) {
                  totalStretch += sizer.stretch;
                  stretchCount++;
              }
          }
          // If the space is equal to the total size, return early.
          if (space === totalSize) {
              return 0;
          }
          // If the space is less than the total min, minimize each sizer.
          if (space <= totalMin) {
              for (var i = 0; i < count; ++i) {
                  var sizer = sizers[i];
                  sizer.size = sizer.minSize;
              }
              return space - totalMin;
          }
          // If the space is greater than the total max, maximize each sizer.
          if (space >= totalMax) {
              for (var i = 0; i < count; ++i) {
                  var sizer = sizers[i];
                  sizer.size = sizer.maxSize;
              }
              return space - totalMax;
          }
          // The loops below perform sub-pixel precision sizing. A near zero
          // value is used for compares instead of zero to ensure that the
          // loop terminates when the subdivided space is reasonably small.
          var nearZero = 0.01;
          // A counter which is decremented each time a sizer is resized to
          // its limit. This ensures the loops terminate even if there is
          // space remaining to distribute.
          var notDoneCount = count;
          // Distribute negative delta space.
          if (space < totalSize) {
              // Shrink each stretchable sizer by an amount proportional to its
              // stretch factor. If a sizer reaches its min size it's marked as
              // done. The loop progresses in phases where each sizer is given
              // a chance to consume its fair share for the pass, regardless of
              // whether a sizer before it reached its limit. This continues
              // until the stretchable sizers or the free space is exhausted.
              var freeSpace = totalSize - space;
              while (stretchCount > 0 && freeSpace > nearZero) {
                  var distSpace = freeSpace;
                  var distStretch = totalStretch;
                  for (var i = 0; i < count; ++i) {
                      var sizer = sizers[i];
                      if (sizer.done || sizer.stretch === 0) {
                          continue;
                      }
                      var amt = sizer.stretch * distSpace / distStretch;
                      if (sizer.size - amt <= sizer.minSize) {
                          freeSpace -= sizer.size - sizer.minSize;
                          totalStretch -= sizer.stretch;
                          sizer.size = sizer.minSize;
                          sizer.done = true;
                          notDoneCount--;
                          stretchCount--;
                      }
                      else {
                          freeSpace -= amt;
                          sizer.size -= amt;
                      }
                  }
              }
              // Distribute any remaining space evenly among the non-stretchable
              // sizers. This progresses in phases in the same manner as above.
              while (notDoneCount > 0 && freeSpace > nearZero) {
                  var amt = freeSpace / notDoneCount;
                  for (var i = 0; i < count; ++i) {
                      var sizer = sizers[i];
                      if (sizer.done) {
                          continue;
                      }
                      if (sizer.size - amt <= sizer.minSize) {
                          freeSpace -= sizer.size - sizer.minSize;
                          sizer.size = sizer.minSize;
                          sizer.done = true;
                          notDoneCount--;
                      }
                      else {
                          freeSpace -= amt;
                          sizer.size -= amt;
                      }
                  }
              }
          }
          // Distribute positive delta space.
          else {
              // Expand each stretchable sizer by an amount proportional to its
              // stretch factor. If a sizer reaches its max size it's marked as
              // done. The loop progresses in phases where each sizer is given
              // a chance to consume its fair share for the pass, regardless of
              // whether a sizer before it reached its limit. This continues
              // until the stretchable sizers or the free space is exhausted.
              var freeSpace = space - totalSize;
              while (stretchCount > 0 && freeSpace > nearZero) {
                  var distSpace = freeSpace;
                  var distStretch = totalStretch;
                  for (var i = 0; i < count; ++i) {
                      var sizer = sizers[i];
                      if (sizer.done || sizer.stretch === 0) {
                          continue;
                      }
                      var amt = sizer.stretch * distSpace / distStretch;
                      if (sizer.size + amt >= sizer.maxSize) {
                          freeSpace -= sizer.maxSize - sizer.size;
                          totalStretch -= sizer.stretch;
                          sizer.size = sizer.maxSize;
                          sizer.done = true;
                          notDoneCount--;
                          stretchCount--;
                      }
                      else {
                          freeSpace -= amt;
                          sizer.size += amt;
                      }
                  }
              }
              // Distribute any remaining space evenly among the non-stretchable
              // sizers. This progresses in phases in the same manner as above.
              while (notDoneCount > 0 && freeSpace > nearZero) {
                  var amt = freeSpace / notDoneCount;
                  for (var i = 0; i < count; ++i) {
                      var sizer = sizers[i];
                      if (sizer.done) {
                          continue;
                      }
                      if (sizer.size + amt >= sizer.maxSize) {
                          freeSpace -= sizer.maxSize - sizer.size;
                          sizer.size = sizer.maxSize;
                          sizer.done = true;
                          notDoneCount--;
                      }
                      else {
                          freeSpace -= amt;
                          sizer.size += amt;
                      }
                  }
              }
          }
          // Indicate that the consumed space equals the available space.
          return 0;
      }
      BoxEngine.calc = calc;
      /**
       * Adjust a sizer by a delta and update its neighbors accordingly.
       *
       * @param sizers - The sizers which should be adjusted.
       *
       * @param index - The index of the sizer to grow.
       *
       * @param delta - The amount to adjust the sizer, positive or negative.
       *
       * #### Notes
       * This will adjust the indicated sizer by the specified amount, along
       * with the sizes of the appropriate neighbors, subject to the limits
       * specified by each of the sizers.
       *
       * This is useful when implementing box layouts where the boundaries
       * between the sizers are interactively adjustable by the user.
       */
      function adjust(sizers, index, delta) {
          // Bail early when there is nothing to do.
          if (sizers.length === 0 || delta === 0) {
              return;
          }
          // Dispatch to the proper implementation.
          if (delta > 0) {
              growSizer(sizers, index, delta);
          }
          else {
              shrinkSizer(sizers, index, -delta);
          }
      }
      BoxEngine.adjust = adjust;
      /**
       * Grow a sizer by a positive delta and adjust neighbors.
       */
      function growSizer(sizers, index, delta) {
          // Compute how much the items to the left can expand.
          var growLimit = 0;
          for (var i = 0; i <= index; ++i) {
              var sizer = sizers[i];
              growLimit += sizer.maxSize - sizer.size;
          }
          // Compute how much the items to the right can shrink.
          var shrinkLimit = 0;
          for (var i = index + 1, n = sizers.length; i < n; ++i) {
              var sizer = sizers[i];
              shrinkLimit += sizer.size - sizer.minSize;
          }
          // Clamp the delta adjustment to the limits.
          delta = Math.min(delta, growLimit, shrinkLimit);
          // Grow the sizers to the left by the delta.
          var grow = delta;
          for (var i = index; i >= 0 && grow > 0; --i) {
              var sizer = sizers[i];
              var limit = sizer.maxSize - sizer.size;
              if (limit >= grow) {
                  sizer.sizeHint = sizer.size + grow;
                  grow = 0;
              }
              else {
                  sizer.sizeHint = sizer.size + limit;
                  grow -= limit;
              }
          }
          // Shrink the sizers to the right by the delta.
          var shrink = delta;
          for (var i = index + 1, n = sizers.length; i < n && shrink > 0; ++i) {
              var sizer = sizers[i];
              var limit = sizer.size - sizer.minSize;
              if (limit >= shrink) {
                  sizer.sizeHint = sizer.size - shrink;
                  shrink = 0;
              }
              else {
                  sizer.sizeHint = sizer.size - limit;
                  shrink -= limit;
              }
          }
      }
      /**
       * Shrink a sizer by a positive delta and adjust neighbors.
       */
      function shrinkSizer(sizers, index, delta) {
          // Compute how much the items to the right can expand.
          var growLimit = 0;
          for (var i = index + 1, n = sizers.length; i < n; ++i) {
              var sizer = sizers[i];
              growLimit += sizer.maxSize - sizer.size;
          }
          // Compute how much the items to the left can shrink.
          var shrinkLimit = 0;
          for (var i = 0; i <= index; ++i) {
              var sizer = sizers[i];
              shrinkLimit += sizer.size - sizer.minSize;
          }
          // Clamp the delta adjustment to the limits.
          delta = Math.min(delta, growLimit, shrinkLimit);
          // Grow the sizers to the right by the delta.
          var grow = delta;
          for (var i = index + 1, n = sizers.length; i < n && grow > 0; ++i) {
              var sizer = sizers[i];
              var limit = sizer.maxSize - sizer.size;
              if (limit >= grow) {
                  sizer.sizeHint = sizer.size + grow;
                  grow = 0;
              }
              else {
                  sizer.sizeHint = sizer.size + limit;
                  grow -= limit;
              }
          }
          // Shrink the sizers to the left by the delta.
          var shrink = delta;
          for (var i = index; i >= 0 && shrink > 0; --i) {
              var sizer = sizers[i];
              var limit = sizer.size - sizer.minSize;
              if (limit >= shrink) {
                  sizer.sizeHint = sizer.size - shrink;
                  shrink = 0;
              }
              else {
                  sizer.sizeHint = sizer.size - limit;
                  shrink -= limit;
              }
          }
      }
  })(BoxEngine || (BoxEngine = {}));

  /*! *****************************************************************************
  Copyright (c) Microsoft Corporation. All rights reserved.
  Licensed under the Apache License, Version 2.0 (the "License"); you may not use
  this file except in compliance with the License. You may obtain a copy of the
  License at http://www.apache.org/licenses/LICENSE-2.0

  THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
  KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
  WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
  MERCHANTABLITY OR NON-INFRINGEMENT.

  See the Apache Version 2.0 License for specific language governing permissions
  and limitations under the License.
  ***************************************************************************** */
  /* global Reflect, Promise */

  var extendStatics$3 = function(d, b) {
      extendStatics$3 = Object.setPrototypeOf ||
          ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
          function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
      return extendStatics$3(d, b);
  };

  function __extends$5(d, b) {
      extendStatics$3(d, b);
      function __() { this.constructor = d; }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
  }

  var __assign$3 = function() {
      __assign$3 = Object.assign || function __assign(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
          }
          return t;
      };
      return __assign$3.apply(this, arguments);
  };

  // Copyright (c) Jupyter Development Team.
  /**
   * An object which holds data related to an object's title.
   *
   * #### Notes
   * A title object is intended to hold the data necessary to display a
   * header for a particular object. A common example is the `TabPanel`,
   * which uses the widget title to populate the tab for a child widget.
   */
  var Title = /** @class */ (function () {
      /**
       * Construct a new title.
       *
       * @param options - The options for initializing the title.
       */
      function Title(options) {
          this._label = '';
          this._caption = '';
          this._mnemonic = -1;
          this._iconClass = '';
          this._iconLabel = '';
          this._className = '';
          this._closable = false;
          this._changed = new Signal(this);
          this.owner = options.owner;
          if (options.label !== undefined) {
              this._label = options.label;
          }
          if (options.mnemonic !== undefined) {
              this._mnemonic = options.mnemonic;
          }
          if (options.icon !== undefined) {
              /* <DEPRECATED> */
              if (typeof options.icon === "string") {
                  // when ._icon is null, the .icon getter will alias .iconClass
                  this._icon = null;
                  this._iconClass = options.icon;
              }
              else {
                  /* </DEPRECATED> */
                  this._icon = options.icon;
                  /* <DEPRECATED> */
              }
              /* </DEPRECATED> */
          }
          /* <DEPRECATED> */
          else {
              // if unset, default to aliasing .iconClass
              this._icon = null;
          }
          /* </DEPRECATED> */
          if (options.iconClass !== undefined) {
              this._iconClass = options.iconClass;
          }
          if (options.iconLabel !== undefined) {
              this._iconLabel = options.iconLabel;
          }
          if (options.iconRenderer !== undefined) {
              this._icon = options.iconRenderer;
          }
          if (options.caption !== undefined) {
              this._caption = options.caption;
          }
          if (options.className !== undefined) {
              this._className = options.className;
          }
          if (options.closable !== undefined) {
              this._closable = options.closable;
          }
          this._dataset = options.dataset || {};
      }
      Object.defineProperty(Title.prototype, "changed", {
          /**
           * A signal emitted when the state of the title changes.
           */
          get: function () {
              return this._changed;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "label", {
          /**
           * Get the label for the title.
           *
           * #### Notes
           * The default value is an empty string.
           */
          get: function () {
              return this._label;
          },
          /**
           * Set the label for the title.
           */
          set: function (value) {
              if (this._label === value) {
                  return;
              }
              this._label = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "mnemonic", {
          /**
           * Get the mnemonic index for the title.
           *
           * #### Notes
           * The default value is `-1`.
           */
          get: function () {
              return this._mnemonic;
          },
          /**
           * Set the mnemonic index for the title.
           */
          set: function (value) {
              if (this._mnemonic === value) {
                  return;
              }
              this._mnemonic = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "icon", {
          /**
           * Get the icon renderer for the title.
           *
           * #### Notes
           * The default value is undefined.
           *
           * DEPRECATED: if set to a string value, the .icon field will function as
           * an alias for the .iconClass field, for backwards compatibility
           */
          get: function () {
              /* <DEPRECATED> */
              if (this._icon === null) {
                  // only alias .iconClass if ._icon has been explicitly nulled
                  return this.iconClass;
              }
              /* </DEPRECATED> */
              return this._icon;
          },
          /**
           * Set the icon renderer for the title.
           *
           * #### Notes
           * A renderer is an object that supplies a render and unrender function.
           *
           * DEPRECATED: if set to a string value, the .icon field will function as
           * an alias for the .iconClass field, for backwards compatibility
           */
          set: function (value /* </DEPRECATED> */) {
              /* <DEPRECATED> */
              if (typeof value === "string") {
                  // when ._icon is null, the .icon getter will alias .iconClass
                  this._icon = null;
                  this.iconClass = value;
              }
              else {
                  /* </DEPRECATED> */
                  if (this._icon === value) {
                      return;
                  }
                  this._icon = value;
                  this._changed.emit(undefined);
                  /* <DEPRECATED> */
              }
              /* </DEPRECATED> */
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "iconClass", {
          /**
           * Get the icon class name for the title.
           *
           * #### Notes
           * The default value is an empty string.
           */
          get: function () {
              return this._iconClass;
          },
          /**
           * Set the icon class name for the title.
           *
           * #### Notes
           * Multiple class names can be separated with whitespace.
           */
          set: function (value) {
              if (this._iconClass === value) {
                  return;
              }
              this._iconClass = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "iconLabel", {
          /**
           * Get the icon label for the title.
           *
           * #### Notes
           * The default value is an empty string.
           */
          get: function () {
              return this._iconLabel;
          },
          /**
           * Set the icon label for the title.
           *
           * #### Notes
           * Multiple class names can be separated with whitespace.
           */
          set: function (value) {
              if (this._iconLabel === value) {
                  return;
              }
              this._iconLabel = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "iconRenderer", {
          /**
           * @deprecated Use `icon` instead.
           */
          get: function () {
              return this._icon || undefined;
          },
          /**
           * @deprecated Use `icon` instead.
           */
          set: function (value) {
              this.icon = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "caption", {
          /**
           * Get the caption for the title.
           *
           * #### Notes
           * The default value is an empty string.
           */
          get: function () {
              return this._caption;
          },
          /**
           * Set the caption for the title.
           */
          set: function (value) {
              if (this._caption === value) {
                  return;
              }
              this._caption = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "className", {
          /**
           * Get the extra class name for the title.
           *
           * #### Notes
           * The default value is an empty string.
           */
          get: function () {
              return this._className;
          },
          /**
           * Set the extra class name for the title.
           *
           * #### Notes
           * Multiple class names can be separated with whitespace.
           */
          set: function (value) {
              if (this._className === value) {
                  return;
              }
              this._className = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "closable", {
          /**
           * Get the closable state for the title.
           *
           * #### Notes
           * The default value is `false`.
           */
          get: function () {
              return this._closable;
          },
          /**
           * Set the closable state for the title.
           *
           * #### Notes
           * This controls the presence of a close icon when applicable.
           */
          set: function (value) {
              if (this._closable === value) {
                  return;
              }
              this._closable = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Title.prototype, "dataset", {
          /**
           * Get the dataset for the title.
           *
           * #### Notes
           * The default value is an empty dataset.
           */
          get: function () {
              return this._dataset;
          },
          /**
           * Set the dataset for the title.
           *
           * #### Notes
           * This controls the data attributes when applicable.
           */
          set: function (value) {
              if (this._dataset === value) {
                  return;
              }
              this._dataset = value;
              this._changed.emit(undefined);
          },
          enumerable: true,
          configurable: true
      });
      return Title;
  }());

  /**
   * The base class of the lumino widget hierarchy.
   *
   * #### Notes
   * This class will typically be subclassed in order to create a useful
   * widget. However, it can be used directly to host externally created
   * content.
   */
  var Widget = /** @class */ (function () {
      /**
       * Construct a new widget.
       *
       * @param options - The options for initializing the widget.
       */
      function Widget(options) {
          if (options === void 0) { options = {}; }
          this._flags = 0;
          this._layout = null;
          this._parent = null;
          this._disposed = new Signal(this);
          this.node = Private$9.createNode(options);
          this.addClass('lm-Widget');
          /* <DEPRECATED> */
          this.addClass('p-Widget');
          /* </DEPRECATED> */
      }
      /**
       * Dispose of the widget and its descendant widgets.
       *
       * #### Notes
       * It is unsafe to use the widget after it has been disposed.
       *
       * All calls made to this method after the first are a no-op.
       */
      Widget.prototype.dispose = function () {
          // Do nothing if the widget is already disposed.
          if (this.isDisposed) {
              return;
          }
          // Set the disposed flag and emit the disposed signal.
          this.setFlag(Widget.Flag.IsDisposed);
          this._disposed.emit(undefined);
          // Remove or detach the widget if necessary.
          if (this.parent) {
              this.parent = null;
          }
          else if (this.isAttached) {
              Widget.detach(this);
          }
          // Dispose of the widget layout.
          if (this._layout) {
              this._layout.dispose();
              this._layout = null;
          }
          // Clear the extra data associated with the widget.
          Signal.clearData(this);
          MessageLoop.clearData(this);
          AttachedProperty.clearData(this);
      };
      Object.defineProperty(Widget.prototype, "disposed", {
          /**
           * A signal emitted when the widget is disposed.
           */
          get: function () {
              return this._disposed;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "isDisposed", {
          /**
           * Test whether the widget has been disposed.
           */
          get: function () {
              return this.testFlag(Widget.Flag.IsDisposed);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "isAttached", {
          /**
           * Test whether the widget's node is attached to the DOM.
           */
          get: function () {
              return this.testFlag(Widget.Flag.IsAttached);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "isHidden", {
          /**
           * Test whether the widget is explicitly hidden.
           */
          get: function () {
              return this.testFlag(Widget.Flag.IsHidden);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "isVisible", {
          /**
           * Test whether the widget is visible.
           *
           * #### Notes
           * A widget is visible when it is attached to the DOM, is not
           * explicitly hidden, and has no explicitly hidden ancestors.
           */
          get: function () {
              return this.testFlag(Widget.Flag.IsVisible);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "title", {
          /**
           * The title object for the widget.
           *
           * #### Notes
           * The title object is used by some container widgets when displaying
           * the widget alongside some title, such as a tab panel or side bar.
           *
           * Since not all widgets will use the title, it is created on demand.
           *
           * The `owner` property of the title is set to this widget.
           */
          get: function () {
              return Private$9.titleProperty.get(this);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "id", {
          /**
           * Get the id of the widget's DOM node.
           */
          get: function () {
              return this.node.id;
          },
          /**
           * Set the id of the widget's DOM node.
           */
          set: function (value) {
              this.node.id = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "dataset", {
          /**
           * The dataset for the widget's DOM node.
           */
          get: function () {
              return this.node.dataset;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "parent", {
          /**
           * Get the parent of the widget.
           */
          get: function () {
              return this._parent;
          },
          /**
           * Set the parent of the widget.
           *
           * #### Notes
           * Children are typically added to a widget by using a layout, which
           * means user code will not normally set the parent widget directly.
           *
           * The widget will be automatically removed from its old parent.
           *
           * This is a no-op if there is no effective parent change.
           */
          set: function (value) {
              if (this._parent === value) {
                  return;
              }
              if (value && this.contains(value)) {
                  throw new Error('Invalid parent widget.');
              }
              if (this._parent && !this._parent.isDisposed) {
                  var msg = new Widget.ChildMessage('child-removed', this);
                  MessageLoop.sendMessage(this._parent, msg);
              }
              this._parent = value;
              if (this._parent && !this._parent.isDisposed) {
                  var msg = new Widget.ChildMessage('child-added', this);
                  MessageLoop.sendMessage(this._parent, msg);
              }
              if (!this.isDisposed) {
                  MessageLoop.sendMessage(this, Widget.Msg.ParentChanged);
              }
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Widget.prototype, "layout", {
          /**
           * Get the layout for the widget.
           */
          get: function () {
              return this._layout;
          },
          /**
           * Set the layout for the widget.
           *
           * #### Notes
           * The layout is single-use only. It cannot be changed after the
           * first assignment.
           *
           * The layout is disposed automatically when the widget is disposed.
           */
          set: function (value) {
              if (this._layout === value) {
                  return;
              }
              if (this.testFlag(Widget.Flag.DisallowLayout)) {
                  throw new Error('Cannot set widget layout.');
              }
              if (this._layout) {
                  throw new Error('Cannot change widget layout.');
              }
              if (value.parent) {
                  throw new Error('Cannot change layout parent.');
              }
              this._layout = value;
              value.parent = this;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over the widget's children.
       *
       * @returns A new iterator over the children of the widget.
       *
       * #### Notes
       * The widget must have a populated layout in order to have children.
       *
       * If a layout is not installed, the returned iterator will be empty.
       */
      Widget.prototype.children = function () {
          return this._layout ? this._layout.iter() : empty();
      };
      /**
       * Test whether a widget is a descendant of this widget.
       *
       * @param widget - The descendant widget of interest.
       *
       * @returns `true` if the widget is a descendant, `false` otherwise.
       */
      Widget.prototype.contains = function (widget) {
          for (var value = widget; value; value = value._parent) {
              if (value === this) {
                  return true;
              }
          }
          return false;
      };
      /**
       * Test whether the widget's DOM node has the given class name.
       *
       * @param name - The class name of interest.
       *
       * @returns `true` if the node has the class, `false` otherwise.
       */
      Widget.prototype.hasClass = function (name) {
          return this.node.classList.contains(name);
      };
      /**
       * Add a class name to the widget's DOM node.
       *
       * @param name - The class name to add to the node.
       *
       * #### Notes
       * If the class name is already added to the node, this is a no-op.
       *
       * The class name must not contain whitespace.
       */
      Widget.prototype.addClass = function (name) {
          this.node.classList.add(name);
      };
      /**
       * Remove a class name from the widget's DOM node.
       *
       * @param name - The class name to remove from the node.
       *
       * #### Notes
       * If the class name is not yet added to the node, this is a no-op.
       *
       * The class name must not contain whitespace.
       */
      Widget.prototype.removeClass = function (name) {
          this.node.classList.remove(name);
      };
      /**
       * Toggle a class name on the widget's DOM node.
       *
       * @param name - The class name to toggle on the node.
       *
       * @param force - Whether to force add the class (`true`) or force
       *   remove the class (`false`). If not provided, the presence of
       *   the class will be toggled from its current state.
       *
       * @returns `true` if the class is now present, `false` otherwise.
       *
       * #### Notes
       * The class name must not contain whitespace.
       */
      Widget.prototype.toggleClass = function (name, force) {
          if (force === true) {
              this.node.classList.add(name);
              return true;
          }
          if (force === false) {
              this.node.classList.remove(name);
              return false;
          }
          return this.node.classList.toggle(name);
      };
      /**
       * Post an `'update-request'` message to the widget.
       *
       * #### Notes
       * This is a simple convenience method for posting the message.
       */
      Widget.prototype.update = function () {
          MessageLoop.postMessage(this, Widget.Msg.UpdateRequest);
      };
      /**
       * Post a `'fit-request'` message to the widget.
       *
       * #### Notes
       * This is a simple convenience method for posting the message.
       */
      Widget.prototype.fit = function () {
          MessageLoop.postMessage(this, Widget.Msg.FitRequest);
      };
      /**
       * Post an `'activate-request'` message to the widget.
       *
       * #### Notes
       * This is a simple convenience method for posting the message.
       */
      Widget.prototype.activate = function () {
          MessageLoop.postMessage(this, Widget.Msg.ActivateRequest);
      };
      /**
       * Send a `'close-request'` message to the widget.
       *
       * #### Notes
       * This is a simple convenience method for sending the message.
       */
      Widget.prototype.close = function () {
          MessageLoop.sendMessage(this, Widget.Msg.CloseRequest);
      };
      /**
       * Show the widget and make it visible to its parent widget.
       *
       * #### Notes
       * This causes the [[isHidden]] property to be `false`.
       *
       * If the widget is not explicitly hidden, this is a no-op.
       */
      Widget.prototype.show = function () {
          if (!this.testFlag(Widget.Flag.IsHidden)) {
              return;
          }
          if (this.isAttached && (!this.parent || this.parent.isVisible)) {
              MessageLoop.sendMessage(this, Widget.Msg.BeforeShow);
          }
          this.clearFlag(Widget.Flag.IsHidden);
          this.removeClass('lm-mod-hidden');
          /* <DEPRECATED> */
          this.removeClass('p-mod-hidden');
          /* </DEPRECATED> */
          if (this.isAttached && (!this.parent || this.parent.isVisible)) {
              MessageLoop.sendMessage(this, Widget.Msg.AfterShow);
          }
          if (this.parent) {
              var msg = new Widget.ChildMessage('child-shown', this);
              MessageLoop.sendMessage(this.parent, msg);
          }
      };
      /**
       * Hide the widget and make it hidden to its parent widget.
       *
       * #### Notes
       * This causes the [[isHidden]] property to be `true`.
       *
       * If the widget is explicitly hidden, this is a no-op.
       */
      Widget.prototype.hide = function () {
          if (this.testFlag(Widget.Flag.IsHidden)) {
              return;
          }
          if (this.isAttached && (!this.parent || this.parent.isVisible)) {
              MessageLoop.sendMessage(this, Widget.Msg.BeforeHide);
          }
          this.setFlag(Widget.Flag.IsHidden);
          this.addClass('lm-mod-hidden');
          /* <DEPRECATED> */
          this.addClass('p-mod-hidden');
          /* </DEPRECATED> */
          if (this.isAttached && (!this.parent || this.parent.isVisible)) {
              MessageLoop.sendMessage(this, Widget.Msg.AfterHide);
          }
          if (this.parent) {
              var msg = new Widget.ChildMessage('child-hidden', this);
              MessageLoop.sendMessage(this.parent, msg);
          }
      };
      /**
       * Show or hide the widget according to a boolean value.
       *
       * @param hidden - `true` to hide the widget, or `false` to show it.
       *
       * #### Notes
       * This is a convenience method for `hide()` and `show()`.
       */
      Widget.prototype.setHidden = function (hidden) {
          if (hidden) {
              this.hide();
          }
          else {
              this.show();
          }
      };
      /**
       * Test whether the given widget flag is set.
       *
       * #### Notes
       * This will not typically be called directly by user code.
       */
      Widget.prototype.testFlag = function (flag) {
          return (this._flags & flag) !== 0;
      };
      /**
       * Set the given widget flag.
       *
       * #### Notes
       * This will not typically be called directly by user code.
       */
      Widget.prototype.setFlag = function (flag) {
          this._flags |= flag;
      };
      /**
       * Clear the given widget flag.
       *
       * #### Notes
       * This will not typically be called directly by user code.
       */
      Widget.prototype.clearFlag = function (flag) {
          this._flags &= ~flag;
      };
      /**
       * Process a message sent to the widget.
       *
       * @param msg - The message sent to the widget.
       *
       * #### Notes
       * Subclasses may reimplement this method as needed.
       */
      Widget.prototype.processMessage = function (msg) {
          switch (msg.type) {
              case 'resize':
                  this.notifyLayout(msg);
                  this.onResize(msg);
                  break;
              case 'update-request':
                  this.notifyLayout(msg);
                  this.onUpdateRequest(msg);
                  break;
              case 'fit-request':
                  this.notifyLayout(msg);
                  this.onFitRequest(msg);
                  break;
              case 'before-show':
                  this.notifyLayout(msg);
                  this.onBeforeShow(msg);
                  break;
              case 'after-show':
                  this.setFlag(Widget.Flag.IsVisible);
                  this.notifyLayout(msg);
                  this.onAfterShow(msg);
                  break;
              case 'before-hide':
                  this.notifyLayout(msg);
                  this.onBeforeHide(msg);
                  break;
              case 'after-hide':
                  this.clearFlag(Widget.Flag.IsVisible);
                  this.notifyLayout(msg);
                  this.onAfterHide(msg);
                  break;
              case 'before-attach':
                  this.notifyLayout(msg);
                  this.onBeforeAttach(msg);
                  break;
              case 'after-attach':
                  if (!this.isHidden && (!this.parent || this.parent.isVisible)) {
                      this.setFlag(Widget.Flag.IsVisible);
                  }
                  this.setFlag(Widget.Flag.IsAttached);
                  this.notifyLayout(msg);
                  this.onAfterAttach(msg);
                  break;
              case 'before-detach':
                  this.notifyLayout(msg);
                  this.onBeforeDetach(msg);
                  break;
              case 'after-detach':
                  this.clearFlag(Widget.Flag.IsVisible);
                  this.clearFlag(Widget.Flag.IsAttached);
                  this.notifyLayout(msg);
                  this.onAfterDetach(msg);
                  break;
              case 'activate-request':
                  this.notifyLayout(msg);
                  this.onActivateRequest(msg);
                  break;
              case 'close-request':
                  this.notifyLayout(msg);
                  this.onCloseRequest(msg);
                  break;
              case 'child-added':
                  this.notifyLayout(msg);
                  this.onChildAdded(msg);
                  break;
              case 'child-removed':
                  this.notifyLayout(msg);
                  this.onChildRemoved(msg);
                  break;
              default:
                  this.notifyLayout(msg);
                  break;
          }
      };
      /**
       * Invoke the message processing routine of the widget's layout.
       *
       * @param msg - The message to dispatch to the layout.
       *
       * #### Notes
       * This is a no-op if the widget does not have a layout.
       *
       * This will not typically be called directly by user code.
       */
      Widget.prototype.notifyLayout = function (msg) {
          if (this._layout) {
              this._layout.processParentMessage(msg);
          }
      };
      /**
       * A message handler invoked on a `'close-request'` message.
       *
       * #### Notes
       * The default implementation unparents or detaches the widget.
       */
      Widget.prototype.onCloseRequest = function (msg) {
          if (this.parent) {
              this.parent = null;
          }
          else if (this.isAttached) {
              Widget.detach(this);
          }
      };
      /**
       * A message handler invoked on a `'resize'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onResize = function (msg) { };
      /**
       * A message handler invoked on an `'update-request'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onUpdateRequest = function (msg) { };
      /**
       * A message handler invoked on a `'fit-request'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onFitRequest = function (msg) { };
      /**
       * A message handler invoked on an `'activate-request'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onActivateRequest = function (msg) { };
      /**
       * A message handler invoked on a `'before-show'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onBeforeShow = function (msg) { };
      /**
       * A message handler invoked on an `'after-show'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onAfterShow = function (msg) { };
      /**
       * A message handler invoked on a `'before-hide'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onBeforeHide = function (msg) { };
      /**
       * A message handler invoked on an `'after-hide'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onAfterHide = function (msg) { };
      /**
       * A message handler invoked on a `'before-attach'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onBeforeAttach = function (msg) { };
      /**
       * A message handler invoked on an `'after-attach'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onAfterAttach = function (msg) { };
      /**
       * A message handler invoked on a `'before-detach'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onBeforeDetach = function (msg) { };
      /**
       * A message handler invoked on an `'after-detach'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onAfterDetach = function (msg) { };
      /**
       * A message handler invoked on a `'child-added'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onChildAdded = function (msg) { };
      /**
       * A message handler invoked on a `'child-removed'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Widget.prototype.onChildRemoved = function (msg) { };
      return Widget;
  }());
  /**
   * The namespace for the `Widget` class statics.
   */
  (function (Widget) {
      /**
       * An enum of widget bit flags.
       */
      var Flag;
      (function (Flag) {
          /**
           * The widget has been disposed.
           */
          Flag[Flag["IsDisposed"] = 1] = "IsDisposed";
          /**
           * The widget is attached to the DOM.
           */
          Flag[Flag["IsAttached"] = 2] = "IsAttached";
          /**
           * The widget is hidden.
           */
          Flag[Flag["IsHidden"] = 4] = "IsHidden";
          /**
           * The widget is visible.
           */
          Flag[Flag["IsVisible"] = 8] = "IsVisible";
          /**
           * A layout cannot be set on the widget.
           */
          Flag[Flag["DisallowLayout"] = 16] = "DisallowLayout";
      })(Flag = Widget.Flag || (Widget.Flag = {}));
      /**
       * A collection of stateless messages related to widgets.
       */
      var Msg;
      (function (Msg) {
          /**
           * A singleton `'before-show'` message.
           *
           * #### Notes
           * This message is sent to a widget before it becomes visible.
           *
           * This message is **not** sent when the widget is being attached.
           */
          Msg.BeforeShow = new Message('before-show');
          /**
           * A singleton `'after-show'` message.
           *
           * #### Notes
           * This message is sent to a widget after it becomes visible.
           *
           * This message is **not** sent when the widget is being attached.
           */
          Msg.AfterShow = new Message('after-show');
          /**
           * A singleton `'before-hide'` message.
           *
           * #### Notes
           * This message is sent to a widget before it becomes not-visible.
           *
           * This message is **not** sent when the widget is being detached.
           */
          Msg.BeforeHide = new Message('before-hide');
          /**
           * A singleton `'after-hide'` message.
           *
           * #### Notes
           * This message is sent to a widget after it becomes not-visible.
           *
           * This message is **not** sent when the widget is being detached.
           */
          Msg.AfterHide = new Message('after-hide');
          /**
           * A singleton `'before-attach'` message.
           *
           * #### Notes
           * This message is sent to a widget before it is attached.
           */
          Msg.BeforeAttach = new Message('before-attach');
          /**
           * A singleton `'after-attach'` message.
           *
           * #### Notes
           * This message is sent to a widget after it is attached.
           */
          Msg.AfterAttach = new Message('after-attach');
          /**
           * A singleton `'before-detach'` message.
           *
           * #### Notes
           * This message is sent to a widget before it is detached.
           */
          Msg.BeforeDetach = new Message('before-detach');
          /**
           * A singleton `'after-detach'` message.
           *
           * #### Notes
           * This message is sent to a widget after it is detached.
           */
          Msg.AfterDetach = new Message('after-detach');
          /**
           * A singleton `'parent-changed'` message.
           *
           * #### Notes
           * This message is sent to a widget when its parent has changed.
           */
          Msg.ParentChanged = new Message('parent-changed');
          /**
           * A singleton conflatable `'update-request'` message.
           *
           * #### Notes
           * This message can be dispatched to supporting widgets in order to
           * update their content based on the current widget state. Not all
           * widgets will respond to messages of this type.
           *
           * For widgets with a layout, this message will inform the layout to
           * update the position and size of its child widgets.
           */
          Msg.UpdateRequest = new ConflatableMessage('update-request');
          /**
           * A singleton conflatable `'fit-request'` message.
           *
           * #### Notes
           * For widgets with a layout, this message will inform the layout to
           * recalculate its size constraints to fit the space requirements of
           * its child widgets, and to update their position and size. Not all
           * layouts will respond to messages of this type.
           */
          Msg.FitRequest = new ConflatableMessage('fit-request');
          /**
           * A singleton conflatable `'activate-request'` message.
           *
           * #### Notes
           * This message should be dispatched to a widget when it should
           * perform the actions necessary to activate the widget, which
           * may include focusing its node or descendant node.
           */
          Msg.ActivateRequest = new ConflatableMessage('activate-request');
          /**
           * A singleton conflatable `'close-request'` message.
           *
           * #### Notes
           * This message should be dispatched to a widget when it should close
           * and remove itself from the widget hierarchy.
           */
          Msg.CloseRequest = new ConflatableMessage('close-request');
      })(Msg = Widget.Msg || (Widget.Msg = {}));
      /**
       * A message class for child related messages.
       */
      var ChildMessage = /** @class */ (function (_super) {
          __extends$5(ChildMessage, _super);
          /**
           * Construct a new child message.
           *
           * @param type - The message type.
           *
           * @param child - The child widget for the message.
           */
          function ChildMessage(type, child) {
              var _this = _super.call(this, type) || this;
              _this.child = child;
              return _this;
          }
          return ChildMessage;
      }(Message));
      Widget.ChildMessage = ChildMessage;
      /**
       * A message class for `'resize'` messages.
       */
      var ResizeMessage = /** @class */ (function (_super) {
          __extends$5(ResizeMessage, _super);
          /**
           * Construct a new resize message.
           *
           * @param width - The **offset width** of the widget, or `-1` if
           *   the width is not known.
           *
           * @param height - The **offset height** of the widget, or `-1` if
           *   the height is not known.
           */
          function ResizeMessage(width, height) {
              var _this = _super.call(this, 'resize') || this;
              _this.width = width;
              _this.height = height;
              return _this;
          }
          return ResizeMessage;
      }(Message));
      Widget.ResizeMessage = ResizeMessage;
      /**
       * The namespace for the `ResizeMessage` class statics.
       */
      (function (ResizeMessage) {
          /**
           * A singleton `'resize'` message with an unknown size.
           */
          ResizeMessage.UnknownSize = new ResizeMessage(-1, -1);
      })(ResizeMessage = Widget.ResizeMessage || (Widget.ResizeMessage = {}));
      /**
       * Attach a widget to a host DOM node.
       *
       * @param widget - The widget of interest.
       *
       * @param host - The DOM node to use as the widget's host.
       *
       * @param ref - The child of `host` to use as the reference element.
       *   If this is provided, the widget will be inserted before this
       *   node in the host. The default is `null`, which will cause the
       *   widget to be added as the last child of the host.
       *
       * #### Notes
       * This will throw an error if the widget is not a root widget, if
       * the widget is already attached, or if the host is not attached
       * to the DOM.
       */
      function attach(widget, host, ref) {
          if (ref === void 0) { ref = null; }
          if (widget.parent) {
              throw new Error('Cannot attach a child widget.');
          }
          if (widget.isAttached || document.body.contains(widget.node)) {
              throw new Error('Widget is already attached.');
          }
          if (!document.body.contains(host)) {
              throw new Error('Host is not attached.');
          }
          MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          host.insertBefore(widget.node, ref);
          MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
      }
      Widget.attach = attach;
      /**
       * Detach the widget from its host DOM node.
       *
       * @param widget - The widget of interest.
       *
       * #### Notes
       * This will throw an error if the widget is not a root widget,
       * or if the widget is not attached to the DOM.
       */
      function detach(widget) {
          if (widget.parent) {
              throw new Error('Cannot detach a child widget.');
          }
          if (!widget.isAttached || !document.body.contains(widget.node)) {
              throw new Error('Widget is not attached.');
          }
          MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          widget.node.parentNode.removeChild(widget.node);
          MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
      }
      Widget.detach = detach;
  })(Widget || (Widget = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$9;
  (function (Private) {
      /**
       * An attached property for the widget title object.
       */
      Private.titleProperty = new AttachedProperty({
          name: 'title',
          create: function (owner) { return new Title({ owner: owner }); },
      });
      /**
       * Create a DOM node for the given widget options.
       */
      function createNode(options) {
          return options.node || document.createElement('div');
      }
      Private.createNode = createNode;
  })(Private$9 || (Private$9 = {}));

  // Copyright (c) Jupyter Development Team.
  /**
   * An abstract base class for creating lumino layouts.
   *
   * #### Notes
   * A layout is used to add widgets to a parent and to arrange those
   * widgets within the parent's DOM node.
   *
   * This class implements the base functionality which is required of
   * nearly all layouts. It must be subclassed in order to be useful.
   *
   * Notably, this class does not define a uniform interface for adding
   * widgets to the layout. A subclass should define that API in a way
   * which is meaningful for its intended use.
   */
  var Layout = /** @class */ (function () {
      /**
       * Construct a new layout.
       *
       * @param options - The options for initializing the layout.
       */
      function Layout(options) {
          if (options === void 0) { options = {}; }
          this._disposed = false;
          this._parent = null;
          this._fitPolicy = options.fitPolicy || 'set-min-size';
      }
      /**
       * Dispose of the resources held by the layout.
       *
       * #### Notes
       * This should be reimplemented to clear and dispose of the widgets.
       *
       * All reimplementations should call the superclass method.
       *
       * This method is called automatically when the parent is disposed.
       */
      Layout.prototype.dispose = function () {
          this._parent = null;
          this._disposed = true;
          Signal.clearData(this);
          AttachedProperty.clearData(this);
      };
      Object.defineProperty(Layout.prototype, "isDisposed", {
          /**
           * Test whether the layout is disposed.
           */
          get: function () {
              return this._disposed;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Layout.prototype, "parent", {
          /**
           * Get the parent widget of the layout.
           */
          get: function () {
              return this._parent;
          },
          /**
           * Set the parent widget of the layout.
           *
           * #### Notes
           * This is set automatically when installing the layout on the parent
           * widget. The parent widget should not be set directly by user code.
           */
          set: function (value) {
              if (this._parent === value) {
                  return;
              }
              if (this._parent) {
                  throw new Error('Cannot change parent widget.');
              }
              if (value.layout !== this) {
                  throw new Error('Invalid parent widget.');
              }
              this._parent = value;
              this.init();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Layout.prototype, "fitPolicy", {
          /**
           * Get the fit policy for the layout.
           *
           * #### Notes
           * The fit policy controls the computed size constraints which are
           * applied to the parent widget by the layout.
           *
           * Some layout implementations may ignore the fit policy.
           */
          get: function () {
              return this._fitPolicy;
          },
          /**
           * Set the fit policy for the layout.
           *
           * #### Notes
           * The fit policy controls the computed size constraints which are
           * applied to the parent widget by the layout.
           *
           * Some layout implementations may ignore the fit policy.
           *
           * Changing the fit policy will clear the current size constraint
           * for the parent widget and then re-fit the parent.
           */
          set: function (value) {
              // Bail if the policy does not change
              if (this._fitPolicy === value) {
                  return;
              }
              // Update the internal policy.
              this._fitPolicy = value;
              // Clear the size constraints and schedule a fit of the parent.
              if (this._parent) {
                  var style = this._parent.node.style;
                  style.minWidth = '';
                  style.minHeight = '';
                  style.maxWidth = '';
                  style.maxHeight = '';
                  this._parent.fit();
              }
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Process a message sent to the parent widget.
       *
       * @param msg - The message sent to the parent widget.
       *
       * #### Notes
       * This method is called by the parent widget to process a message.
       *
       * Subclasses may reimplement this method as needed.
       */
      Layout.prototype.processParentMessage = function (msg) {
          switch (msg.type) {
              case 'resize':
                  this.onResize(msg);
                  break;
              case 'update-request':
                  this.onUpdateRequest(msg);
                  break;
              case 'fit-request':
                  this.onFitRequest(msg);
                  break;
              case 'before-show':
                  this.onBeforeShow(msg);
                  break;
              case 'after-show':
                  this.onAfterShow(msg);
                  break;
              case 'before-hide':
                  this.onBeforeHide(msg);
                  break;
              case 'after-hide':
                  this.onAfterHide(msg);
                  break;
              case 'before-attach':
                  this.onBeforeAttach(msg);
                  break;
              case 'after-attach':
                  this.onAfterAttach(msg);
                  break;
              case 'before-detach':
                  this.onBeforeDetach(msg);
                  break;
              case 'after-detach':
                  this.onAfterDetach(msg);
                  break;
              case 'child-removed':
                  this.onChildRemoved(msg);
                  break;
              case 'child-shown':
                  this.onChildShown(msg);
                  break;
              case 'child-hidden':
                  this.onChildHidden(msg);
                  break;
          }
      };
      /**
       * Perform layout initialization which requires the parent widget.
       *
       * #### Notes
       * This method is invoked immediately after the layout is installed
       * on the parent widget.
       *
       * The default implementation reparents all of the widgets to the
       * layout parent widget.
       *
       * Subclasses should reimplement this method and attach the child
       * widget nodes to the parent widget's node.
       */
      Layout.prototype.init = function () {
          var _this = this;
          each$1(this, function (widget) {
              widget.parent = _this.parent;
          });
      };
      /**
       * A message handler invoked on a `'resize'` message.
       *
       * #### Notes
       * The layout should ensure that its widgets are resized according
       * to the specified layout space, and that they are sent a `'resize'`
       * message if appropriate.
       *
       * The default implementation of this method sends an `UnknownSize`
       * resize message to all widgets.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onResize = function (msg) {
          each$1(this, function (widget) {
              MessageLoop.sendMessage(widget, Widget.ResizeMessage.UnknownSize);
          });
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       *
       * #### Notes
       * The layout should ensure that its widgets are resized according
       * to the available layout space, and that they are sent a `'resize'`
       * message if appropriate.
       *
       * The default implementation of this method sends an `UnknownSize`
       * resize message to all widgets.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onUpdateRequest = function (msg) {
          each$1(this, function (widget) {
              MessageLoop.sendMessage(widget, Widget.ResizeMessage.UnknownSize);
          });
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message
       * to all widgets. It assumes all widget nodes are attached to the
       * parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onBeforeAttach = function (msg) {
          each$1(this, function (widget) {
              MessageLoop.sendMessage(widget, msg);
          });
      };
      /**
       * A message handler invoked on an `'after-attach'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message
       * to all widgets. It assumes all widget nodes are attached to the
       * parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onAfterAttach = function (msg) {
          each$1(this, function (widget) {
              MessageLoop.sendMessage(widget, msg);
          });
      };
      /**
       * A message handler invoked on a `'before-detach'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message
       * to all widgets. It assumes all widget nodes are attached to the
       * parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onBeforeDetach = function (msg) {
          each$1(this, function (widget) {
              MessageLoop.sendMessage(widget, msg);
          });
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message
       * to all widgets. It assumes all widget nodes are attached to the
       * parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onAfterDetach = function (msg) {
          each$1(this, function (widget) {
              MessageLoop.sendMessage(widget, msg);
          });
      };
      /**
       * A message handler invoked on a `'before-show'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message to
       * all non-hidden widgets. It assumes all widget nodes are attached
       * to the parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onBeforeShow = function (msg) {
          each$1(this, function (widget) {
              if (!widget.isHidden) {
                  MessageLoop.sendMessage(widget, msg);
              }
          });
      };
      /**
       * A message handler invoked on an `'after-show'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message to
       * all non-hidden widgets. It assumes all widget nodes are attached
       * to the parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onAfterShow = function (msg) {
          each$1(this, function (widget) {
              if (!widget.isHidden) {
                  MessageLoop.sendMessage(widget, msg);
              }
          });
      };
      /**
       * A message handler invoked on a `'before-hide'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message to
       * all non-hidden widgets. It assumes all widget nodes are attached
       * to the parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onBeforeHide = function (msg) {
          each$1(this, function (widget) {
              if (!widget.isHidden) {
                  MessageLoop.sendMessage(widget, msg);
              }
          });
      };
      /**
       * A message handler invoked on an `'after-hide'` message.
       *
       * #### Notes
       * The default implementation of this method forwards the message to
       * all non-hidden widgets. It assumes all widget nodes are attached
       * to the parent widget node.
       *
       * This may be reimplemented by subclasses as needed.
       */
      Layout.prototype.onAfterHide = function (msg) {
          each$1(this, function (widget) {
              if (!widget.isHidden) {
                  MessageLoop.sendMessage(widget, msg);
              }
          });
      };
      /**
       * A message handler invoked on a `'child-removed'` message.
       *
       * #### Notes
       * This will remove the child widget from the layout.
       *
       * Subclasses should **not** typically reimplement this method.
       */
      Layout.prototype.onChildRemoved = function (msg) {
          this.removeWidget(msg.child);
      };
      /**
       * A message handler invoked on a `'fit-request'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Layout.prototype.onFitRequest = function (msg) { };
      /**
       * A message handler invoked on a `'child-shown'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Layout.prototype.onChildShown = function (msg) { };
      /**
       * A message handler invoked on a `'child-hidden'` message.
       *
       * #### Notes
       * The default implementation of this handler is a no-op.
       */
      Layout.prototype.onChildHidden = function (msg) { };
      return Layout;
  }());
  /**
   * The namespace for the `Layout` class statics.
   */
  (function (Layout) {
      /**
       * Get the horizontal alignment for a widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The horizontal alignment for the widget.
       *
       * #### Notes
       * If the layout width allocated to a widget is larger than its max
       * width, the horizontal alignment controls how the widget is placed
       * within the extra horizontal space.
       *
       * If the allocated width is less than the widget's max width, the
       * horizontal alignment has no effect.
       *
       * Some layout implementations may ignore horizontal alignment.
       */
      function getHorizontalAlignment(widget) {
          return Private$1$2.horizontalAlignmentProperty.get(widget);
      }
      Layout.getHorizontalAlignment = getHorizontalAlignment;
      /**
       * Set the horizontal alignment for a widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the horizontal alignment.
       *
       * #### Notes
       * If the layout width allocated to a widget is larger than its max
       * width, the horizontal alignment controls how the widget is placed
       * within the extra horizontal space.
       *
       * If the allocated width is less than the widget's max width, the
       * horizontal alignment has no effect.
       *
       * Some layout implementations may ignore horizontal alignment.
       *
       * Changing the horizontal alignment will post an `update-request`
       * message to widget's parent, provided the parent has a layout
       * installed.
       */
      function setHorizontalAlignment(widget, value) {
          Private$1$2.horizontalAlignmentProperty.set(widget, value);
      }
      Layout.setHorizontalAlignment = setHorizontalAlignment;
      /**
       * Get the vertical alignment for a widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The vertical alignment for the widget.
       *
       * #### Notes
       * If the layout height allocated to a widget is larger than its max
       * height, the vertical alignment controls how the widget is placed
       * within the extra vertical space.
       *
       * If the allocated height is less than the widget's max height, the
       * vertical alignment has no effect.
       *
       * Some layout implementations may ignore vertical alignment.
       */
      function getVerticalAlignment(widget) {
          return Private$1$2.verticalAlignmentProperty.get(widget);
      }
      Layout.getVerticalAlignment = getVerticalAlignment;
      /**
       * Set the vertical alignment for a widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the vertical alignment.
       *
       * #### Notes
       * If the layout height allocated to a widget is larger than its max
       * height, the vertical alignment controls how the widget is placed
       * within the extra vertical space.
       *
       * If the allocated height is less than the widget's max height, the
       * vertical alignment has no effect.
       *
       * Some layout implementations may ignore vertical alignment.
       *
       * Changing the horizontal alignment will post an `update-request`
       * message to widget's parent, provided the parent has a layout
       * installed.
       */
      function setVerticalAlignment(widget, value) {
          Private$1$2.verticalAlignmentProperty.set(widget, value);
      }
      Layout.setVerticalAlignment = setVerticalAlignment;
  })(Layout || (Layout = {}));
  /**
   * An object which assists in the absolute layout of widgets.
   *
   * #### Notes
   * This class is useful when implementing a layout which arranges its
   * widgets using absolute positioning.
   *
   * This class is used by nearly all of the built-in lumino layouts.
   */
  var LayoutItem = /** @class */ (function () {
      /**
       * Construct a new layout item.
       *
       * @param widget - The widget to be managed by the item.
       *
       * #### Notes
       * The widget will be set to absolute positioning.
       */
      function LayoutItem(widget) {
          this._top = NaN;
          this._left = NaN;
          this._width = NaN;
          this._height = NaN;
          this._minWidth = 0;
          this._minHeight = 0;
          this._maxWidth = Infinity;
          this._maxHeight = Infinity;
          this._disposed = false;
          this.widget = widget;
          this.widget.node.style.position = 'absolute';
      }
      /**
       * Dispose of the the layout item.
       *
       * #### Notes
       * This will reset the positioning of the widget.
       */
      LayoutItem.prototype.dispose = function () {
          // Do nothing if the item is already disposed.
          if (this._disposed) {
              return;
          }
          // Mark the item as disposed.
          this._disposed = true;
          // Reset the widget style.
          var style = this.widget.node.style;
          style.position = '';
          style.top = '';
          style.left = '';
          style.width = '';
          style.height = '';
      };
      Object.defineProperty(LayoutItem.prototype, "minWidth", {
          /**
           * The computed minimum width of the widget.
           *
           * #### Notes
           * This value can be updated by calling the `fit` method.
           */
          get: function () {
              return this._minWidth;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "minHeight", {
          /**
           * The computed minimum height of the widget.
           *
           * #### Notes
           * This value can be updated by calling the `fit` method.
           */
          get: function () {
              return this._minHeight;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "maxWidth", {
          /**
           * The computed maximum width of the widget.
           *
           * #### Notes
           * This value can be updated by calling the `fit` method.
           */
          get: function () {
              return this._maxWidth;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "maxHeight", {
          /**
           * The computed maximum height of the widget.
           *
           * #### Notes
           * This value can be updated by calling the `fit` method.
           */
          get: function () {
              return this._maxHeight;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "isDisposed", {
          /**
           * Whether the layout item is disposed.
           */
          get: function () {
              return this._disposed;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "isHidden", {
          /**
           * Whether the managed widget is hidden.
           */
          get: function () {
              return this.widget.isHidden;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "isVisible", {
          /**
           * Whether the managed widget is visible.
           */
          get: function () {
              return this.widget.isVisible;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(LayoutItem.prototype, "isAttached", {
          /**
           * Whether the managed widget is attached.
           */
          get: function () {
              return this.widget.isAttached;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Update the computed size limits of the managed widget.
       */
      LayoutItem.prototype.fit = function () {
          var limits = ElementExt.sizeLimits(this.widget.node);
          this._minWidth = limits.minWidth;
          this._minHeight = limits.minHeight;
          this._maxWidth = limits.maxWidth;
          this._maxHeight = limits.maxHeight;
      };
      /**
       * Update the position and size of the managed widget.
       *
       * @param left - The left edge position of the layout box.
       *
       * @param top - The top edge position of the layout box.
       *
       * @param width - The width of the layout box.
       *
       * @param height - The height of the layout box.
       */
      LayoutItem.prototype.update = function (left, top, width, height) {
          // Clamp the size to the computed size limits.
          var clampW = Math.max(this._minWidth, Math.min(width, this._maxWidth));
          var clampH = Math.max(this._minHeight, Math.min(height, this._maxHeight));
          // Adjust the left edge for the horizontal alignment, if needed.
          if (clampW < width) {
              switch (Layout.getHorizontalAlignment(this.widget)) {
                  case 'left':
                      break;
                  case 'center':
                      left += (width - clampW) / 2;
                      break;
                  case 'right':
                      left += width - clampW;
                      break;
                  default:
                      throw 'unreachable';
              }
          }
          // Adjust the top edge for the vertical alignment, if needed.
          if (clampH < height) {
              switch (Layout.getVerticalAlignment(this.widget)) {
                  case 'top':
                      break;
                  case 'center':
                      top += (height - clampH) / 2;
                      break;
                  case 'bottom':
                      top += height - clampH;
                      break;
                  default:
                      throw 'unreachable';
              }
          }
          // Set up the resize variables.
          var resized = false;
          var style = this.widget.node.style;
          // Update the top edge of the widget if needed.
          if (this._top !== top) {
              this._top = top;
              style.top = top + "px";
          }
          // Update the left edge of the widget if needed.
          if (this._left !== left) {
              this._left = left;
              style.left = left + "px";
          }
          // Update the width of the widget if needed.
          if (this._width !== clampW) {
              resized = true;
              this._width = clampW;
              style.width = clampW + "px";
          }
          // Update the height of the widget if needed.
          if (this._height !== clampH) {
              resized = true;
              this._height = clampH;
              style.height = clampH + "px";
          }
          // Send a resize message to the widget if needed.
          if (resized) {
              var msg = new Widget.ResizeMessage(clampW, clampH);
              MessageLoop.sendMessage(this.widget, msg);
          }
      };
      return LayoutItem;
  }());
  /**
   * The namespace for the module implementation details.
   */
  var Private$1$2;
  (function (Private) {
      /**
       * The attached property for a widget horizontal alignment.
       */
      Private.horizontalAlignmentProperty = new AttachedProperty({
          name: 'horizontalAlignment',
          create: function () { return 'center'; },
          changed: onAlignmentChanged
      });
      /**
       * The attached property for a widget vertical alignment.
       */
      Private.verticalAlignmentProperty = new AttachedProperty({
          name: 'verticalAlignment',
          create: function () { return 'top'; },
          changed: onAlignmentChanged
      });
      /**
       * The change handler for the attached alignment properties.
       */
      function onAlignmentChanged(child) {
          if (child.parent && child.parent.layout) {
              child.parent.update();
          }
      }
  })(Private$1$2 || (Private$1$2 = {}));

  /**
   * A concrete layout implementation suitable for many use cases.
   *
   * #### Notes
   * This class is suitable as a base class for implementing a variety of
   * layouts, but can also be used directly with standard CSS to layout a
   * collection of widgets.
   */
  var PanelLayout = /** @class */ (function (_super) {
      __extends$5(PanelLayout, _super);
      function PanelLayout() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._widgets = [];
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       *
       * #### Notes
       * This will clear and dispose all widgets in the layout.
       *
       * All reimplementations should call the superclass method.
       *
       * This method is called automatically when the parent is disposed.
       */
      PanelLayout.prototype.dispose = function () {
          while (this._widgets.length > 0) {
              this._widgets.pop().dispose();
          }
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(PanelLayout.prototype, "widgets", {
          /**
           * A read-only array of the widgets in the layout.
           */
          get: function () {
              return this._widgets;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over the widgets in the layout.
       *
       * @returns A new iterator over the widgets in the layout.
       */
      PanelLayout.prototype.iter = function () {
          return iter(this._widgets);
      };
      /**
       * Add a widget to the end of the layout.
       *
       * @param widget - The widget to add to the layout.
       *
       * #### Notes
       * If the widget is already contained in the layout, it will be moved.
       */
      PanelLayout.prototype.addWidget = function (widget) {
          this.insertWidget(this._widgets.length, widget);
      };
      /**
       * Insert a widget into the layout at the specified index.
       *
       * @param index - The index at which to insert the widget.
       *
       * @param widget - The widget to insert into the layout.
       *
       * #### Notes
       * The index will be clamped to the bounds of the widgets.
       *
       * If the widget is already added to the layout, it will be moved.
       *
       * #### Undefined Behavior
       * An `index` which is non-integral.
       */
      PanelLayout.prototype.insertWidget = function (index, widget) {
          // Remove the widget from its current parent. This is a no-op
          // if the widget's parent is already the layout parent widget.
          widget.parent = this.parent;
          // Look up the current index of the widget.
          var i = this._widgets.indexOf(widget);
          // Clamp the insert index to the array bounds.
          var j = Math.max(0, Math.min(index, this._widgets.length));
          // If the widget is not in the array, insert it.
          if (i === -1) {
              // Insert the widget into the array.
              ArrayExt.insert(this._widgets, j, widget);
              // If the layout is parented, attach the widget to the DOM.
              if (this.parent) {
                  this.attachWidget(j, widget);
              }
              // There is nothing more to do.
              return;
          }
          // Otherwise, the widget exists in the array and should be moved.
          // Adjust the index if the location is at the end of the array.
          if (j === this._widgets.length) {
              j--;
          }
          // Bail if there is no effective move.
          if (i === j) {
              return;
          }
          // Move the widget to the new location.
          ArrayExt.move(this._widgets, i, j);
          // If the layout is parented, move the widget in the DOM.
          if (this.parent) {
              this.moveWidget(i, j, widget);
          }
      };
      /**
       * Remove a widget from the layout.
       *
       * @param widget - The widget to remove from the layout.
       *
       * #### Notes
       * A widget is automatically removed from the layout when its `parent`
       * is set to `null`. This method should only be invoked directly when
       * removing a widget from a layout which has yet to be installed on a
       * parent widget.
       *
       * This method does *not* modify the widget's `parent`.
       */
      PanelLayout.prototype.removeWidget = function (widget) {
          this.removeWidgetAt(this._widgets.indexOf(widget));
      };
      /**
       * Remove the widget at a given index from the layout.
       *
       * @param index - The index of the widget to remove.
       *
       * #### Notes
       * A widget is automatically removed from the layout when its `parent`
       * is set to `null`. This method should only be invoked directly when
       * removing a widget from a layout which has yet to be installed on a
       * parent widget.
       *
       * This method does *not* modify the widget's `parent`.
       *
       * #### Undefined Behavior
       * An `index` which is non-integral.
       */
      PanelLayout.prototype.removeWidgetAt = function (index) {
          // Remove the widget from the array.
          var widget = ArrayExt.removeAt(this._widgets, index);
          // If the layout is parented, detach the widget from the DOM.
          if (widget && this.parent) {
              this.detachWidget(index, widget);
          }
      };
      /**
       * Perform layout initialization which requires the parent widget.
       */
      PanelLayout.prototype.init = function () {
          var _this = this;
          _super.prototype.init.call(this);
          each$1(this, function (widget, index) {
              _this.attachWidget(index, widget);
          });
      };
      /**
       * Attach a widget to the parent's DOM node.
       *
       * @param index - The current index of the widget in the layout.
       *
       * @param widget - The widget to attach to the parent.
       *
       * #### Notes
       * This method is called automatically by the panel layout at the
       * appropriate time. It should not be called directly by user code.
       *
       * The default implementation adds the widgets's node to the parent's
       * node at the proper location, and sends the appropriate attach
       * messages to the widget if the parent is attached to the DOM.
       *
       * Subclasses may reimplement this method to control how the widget's
       * node is added to the parent's node.
       */
      PanelLayout.prototype.attachWidget = function (index, widget) {
          // Look up the next sibling reference node.
          var ref = this.parent.node.children[index];
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Insert the widget's node before the sibling.
          this.parent.node.insertBefore(widget.node, ref);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
      };
      /**
       * Move a widget in the parent's DOM node.
       *
       * @param fromIndex - The previous index of the widget in the layout.
       *
       * @param toIndex - The current index of the widget in the layout.
       *
       * @param widget - The widget to move in the parent.
       *
       * #### Notes
       * This method is called automatically by the panel layout at the
       * appropriate time. It should not be called directly by user code.
       *
       * The default implementation moves the widget's node to the proper
       * location in the parent's node and sends the appropriate attach and
       * detach messages to the widget if the parent is attached to the DOM.
       *
       * Subclasses may reimplement this method to control how the widget's
       * node is moved in the parent's node.
       */
      PanelLayout.prototype.moveWidget = function (fromIndex, toIndex, widget) {
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` and  message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
          // Look up the next sibling reference node.
          var ref = this.parent.node.children[toIndex];
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Insert the widget's node before the sibling.
          this.parent.node.insertBefore(widget.node, ref);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
      };
      /**
       * Detach a widget from the parent's DOM node.
       *
       * @param index - The previous index of the widget in the layout.
       *
       * @param widget - The widget to detach from the parent.
       *
       * #### Notes
       * This method is called automatically by the panel layout at the
       * appropriate time. It should not be called directly by user code.
       *
       * The default implementation removes the widget's node from the
       * parent's node, and sends the appropriate detach messages to the
       * widget if the parent is attached to the DOM.
       *
       * Subclasses may reimplement this method to control how the widget's
       * node is removed from the parent's node.
       */
      PanelLayout.prototype.detachWidget = function (index, widget) {
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
      };
      return PanelLayout;
  }(Layout));

  /**
   * A layout which arranges its widgets in a single row or column.
   */
  var BoxLayout = /** @class */ (function (_super) {
      __extends$5(BoxLayout, _super);
      /**
       * Construct a new box layout.
       *
       * @param options - The options for initializing the layout.
       */
      function BoxLayout(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this) || this;
          _this._fixed = 0;
          _this._spacing = 4;
          _this._dirty = false;
          _this._sizers = [];
          _this._items = [];
          _this._box = null;
          _this._alignment = 'start';
          _this._direction = 'top-to-bottom';
          if (options.direction !== undefined) {
              _this._direction = options.direction;
          }
          if (options.alignment !== undefined) {
              _this._alignment = options.alignment;
          }
          if (options.spacing !== undefined) {
              _this._spacing = Private$2$1.clampSpacing(options.spacing);
          }
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       */
      BoxLayout.prototype.dispose = function () {
          // Dispose of the layout items.
          each$1(this._items, function (item) { item.dispose(); });
          // Clear the layout state.
          this._box = null;
          this._items.length = 0;
          this._sizers.length = 0;
          // Dispose of the rest of the layout.
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(BoxLayout.prototype, "direction", {
          /**
           * Get the layout direction for the box layout.
           */
          get: function () {
              return this._direction;
          },
          /**
           * Set the layout direction for the box layout.
           */
          set: function (value) {
              if (this._direction === value) {
                  return;
              }
              this._direction = value;
              if (!this.parent) {
                  return;
              }
              this.parent.dataset['direction'] = value;
              this.parent.fit();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BoxLayout.prototype, "alignment", {
          /**
           * Get the content alignment for the box layout.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand to fill the
           * entire box layout.
           */
          get: function () {
              return this._alignment;
          },
          /**
           * Set the content alignment for the box layout.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand to fill the
           * entire box layout.
           */
          set: function (value) {
              if (this._alignment === value) {
                  return;
              }
              this._alignment = value;
              if (!this.parent) {
                  return;
              }
              this.parent.dataset['alignment'] = value;
              this.parent.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BoxLayout.prototype, "spacing", {
          /**
           * Get the inter-element spacing for the box layout.
           */
          get: function () {
              return this._spacing;
          },
          /**
           * Set the inter-element spacing for the box layout.
           */
          set: function (value) {
              value = Private$2$1.clampSpacing(value);
              if (this._spacing === value) {
                  return;
              }
              this._spacing = value;
              if (!this.parent) {
                  return;
              }
              this.parent.fit();
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Perform layout initialization which requires the parent widget.
       */
      BoxLayout.prototype.init = function () {
          this.parent.dataset['direction'] = this.direction;
          this.parent.dataset['alignment'] = this.alignment;
          _super.prototype.init.call(this);
      };
      /**
       * Attach a widget to the parent's DOM node.
       *
       * @param index - The current index of the widget in the layout.
       *
       * @param widget - The widget to attach to the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      BoxLayout.prototype.attachWidget = function (index, widget) {
          // Create and add a new layout item for the widget.
          ArrayExt.insert(this._items, index, new LayoutItem(widget));
          // Create and add a new sizer for the widget.
          ArrayExt.insert(this._sizers, index, new BoxSizer());
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Add the widget's node to the parent.
          this.parent.node.appendChild(widget.node);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Move a widget in the parent's DOM node.
       *
       * @param fromIndex - The previous index of the widget in the layout.
       *
       * @param toIndex - The current index of the widget in the layout.
       *
       * @param widget - The widget to move in the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      BoxLayout.prototype.moveWidget = function (fromIndex, toIndex, widget) {
          // Move the layout item for the widget.
          ArrayExt.move(this._items, fromIndex, toIndex);
          // Move the sizer for the widget.
          ArrayExt.move(this._sizers, fromIndex, toIndex);
          // Post an update request for the parent widget.
          this.parent.update();
      };
      /**
       * Detach a widget from the parent's DOM node.
       *
       * @param index - The previous index of the widget in the layout.
       *
       * @param widget - The widget to detach from the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      BoxLayout.prototype.detachWidget = function (index, widget) {
          // Remove the layout item for the widget.
          var item = ArrayExt.removeAt(this._items, index);
          // Remove the sizer for the widget.
          ArrayExt.removeAt(this._sizers, index);
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
          // Dispose of the layout item.
          item.dispose();
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'before-show'` message.
       */
      BoxLayout.prototype.onBeforeShow = function (msg) {
          _super.prototype.onBeforeShow.call(this, msg);
          this.parent.update();
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      BoxLayout.prototype.onBeforeAttach = function (msg) {
          _super.prototype.onBeforeAttach.call(this, msg);
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-shown'` message.
       */
      BoxLayout.prototype.onChildShown = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-hidden'` message.
       */
      BoxLayout.prototype.onChildHidden = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'resize'` message.
       */
      BoxLayout.prototype.onResize = function (msg) {
          if (this.parent.isVisible) {
              this._update(msg.width, msg.height);
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      BoxLayout.prototype.onUpdateRequest = function (msg) {
          if (this.parent.isVisible) {
              this._update(-1, -1);
          }
      };
      /**
       * A message handler invoked on a `'fit-request'` message.
       */
      BoxLayout.prototype.onFitRequest = function (msg) {
          if (this.parent.isAttached) {
              this._fit();
          }
      };
      /**
       * Fit the layout to the total size required by the widgets.
       */
      BoxLayout.prototype._fit = function () {
          // Compute the visible item count.
          var nVisible = 0;
          for (var i = 0, n = this._items.length; i < n; ++i) {
              nVisible += +!this._items[i].isHidden;
          }
          // Update the fixed space for the visible items.
          this._fixed = this._spacing * Math.max(0, nVisible - 1);
          // Setup the computed minimum size.
          var horz = Private$2$1.isHorizontal(this._direction);
          var minW = horz ? this._fixed : 0;
          var minH = horz ? 0 : this._fixed;
          // Update the sizers and computed minimum size.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item and corresponding box sizer.
              var item = this._items[i];
              var sizer = this._sizers[i];
              // If the item is hidden, it should consume zero size.
              if (item.isHidden) {
                  sizer.minSize = 0;
                  sizer.maxSize = 0;
                  continue;
              }
              // Update the size limits for the item.
              item.fit();
              // Update the size basis and stretch factor.
              sizer.sizeHint = BoxLayout.getSizeBasis(item.widget);
              sizer.stretch = BoxLayout.getStretch(item.widget);
              // Update the sizer limits and computed min size.
              if (horz) {
                  sizer.minSize = item.minWidth;
                  sizer.maxSize = item.maxWidth;
                  minW += item.minWidth;
                  minH = Math.max(minH, item.minHeight);
              }
              else {
                  sizer.minSize = item.minHeight;
                  sizer.maxSize = item.maxHeight;
                  minH += item.minHeight;
                  minW = Math.max(minW, item.minWidth);
              }
          }
          // Update the box sizing and add it to the computed min size.
          var box = this._box = ElementExt.boxSizing(this.parent.node);
          minW += box.horizontalSum;
          minH += box.verticalSum;
          // Update the parent's min size constraints.
          var style = this.parent.node.style;
          style.minWidth = minW + "px";
          style.minHeight = minH + "px";
          // Set the dirty flag to ensure only a single update occurs.
          this._dirty = true;
          // Notify the ancestor that it should fit immediately. This may
          // cause a resize of the parent, fulfilling the required update.
          if (this.parent.parent) {
              MessageLoop.sendMessage(this.parent.parent, Widget.Msg.FitRequest);
          }
          // If the dirty flag is still set, the parent was not resized.
          // Trigger the required update on the parent widget immediately.
          if (this._dirty) {
              MessageLoop.sendMessage(this.parent, Widget.Msg.UpdateRequest);
          }
      };
      /**
       * Update the layout position and size of the widgets.
       *
       * The parent offset dimensions should be `-1` if unknown.
       */
      BoxLayout.prototype._update = function (offsetWidth, offsetHeight) {
          // Clear the dirty flag to indicate the update occurred.
          this._dirty = false;
          // Compute the visible item count.
          var nVisible = 0;
          for (var i = 0, n = this._items.length; i < n; ++i) {
              nVisible += +!this._items[i].isHidden;
          }
          // Bail early if there are no visible items to layout.
          if (nVisible === 0) {
              return;
          }
          // Measure the parent if the offset dimensions are unknown.
          if (offsetWidth < 0) {
              offsetWidth = this.parent.node.offsetWidth;
          }
          if (offsetHeight < 0) {
              offsetHeight = this.parent.node.offsetHeight;
          }
          // Ensure the parent box sizing data is computed.
          if (!this._box) {
              this._box = ElementExt.boxSizing(this.parent.node);
          }
          // Compute the layout area adjusted for border and padding.
          var top = this._box.paddingTop;
          var left = this._box.paddingLeft;
          var width = offsetWidth - this._box.horizontalSum;
          var height = offsetHeight - this._box.verticalSum;
          // Distribute the layout space and adjust the start position.
          var delta;
          switch (this._direction) {
              case 'left-to-right':
                  delta = BoxEngine.calc(this._sizers, Math.max(0, width - this._fixed));
                  break;
              case 'top-to-bottom':
                  delta = BoxEngine.calc(this._sizers, Math.max(0, height - this._fixed));
                  break;
              case 'right-to-left':
                  delta = BoxEngine.calc(this._sizers, Math.max(0, width - this._fixed));
                  left += width;
                  break;
              case 'bottom-to-top':
                  delta = BoxEngine.calc(this._sizers, Math.max(0, height - this._fixed));
                  top += height;
                  break;
              default:
                  throw 'unreachable';
          }
          // Setup the variables for justification and alignment offset.
          var extra = 0;
          var offset = 0;
          // Account for alignment if there is extra layout space.
          if (delta > 0) {
              switch (this._alignment) {
                  case 'start':
                      break;
                  case 'center':
                      extra = 0;
                      offset = delta / 2;
                      break;
                  case 'end':
                      extra = 0;
                      offset = delta;
                      break;
                  case 'justify':
                      extra = delta / nVisible;
                      offset = 0;
                      break;
                  default:
                      throw 'unreachable';
              }
          }
          // Layout the items using the computed box sizes.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item.
              var item = this._items[i];
              // Ignore hidden items.
              if (item.isHidden) {
                  continue;
              }
              // Fetch the computed size for the widget.
              var size = this._sizers[i].size;
              // Update the widget geometry and advance the relevant edge.
              switch (this._direction) {
                  case 'left-to-right':
                      item.update(left + offset, top, size + extra, height);
                      left += size + extra + this._spacing;
                      break;
                  case 'top-to-bottom':
                      item.update(left, top + offset, width, size + extra);
                      top += size + extra + this._spacing;
                      break;
                  case 'right-to-left':
                      item.update(left - offset - size - extra, top, size + extra, height);
                      left -= size + extra + this._spacing;
                      break;
                  case 'bottom-to-top':
                      item.update(left, top - offset - size - extra, width, size + extra);
                      top -= size + extra + this._spacing;
                      break;
                  default:
                      throw 'unreachable';
              }
          }
      };
      return BoxLayout;
  }(PanelLayout));
  /**
   * The namespace for the `BoxLayout` class statics.
   */
  (function (BoxLayout) {
      /**
       * Get the box layout stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The box layout stretch factor for the widget.
       */
      function getStretch(widget) {
          return Private$2$1.stretchProperty.get(widget);
      }
      BoxLayout.getStretch = getStretch;
      /**
       * Set the box layout stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the stretch factor.
       */
      function setStretch(widget, value) {
          Private$2$1.stretchProperty.set(widget, value);
      }
      BoxLayout.setStretch = setStretch;
      /**
       * Get the box layout size basis for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The box layout size basis for the widget.
       */
      function getSizeBasis(widget) {
          return Private$2$1.sizeBasisProperty.get(widget);
      }
      BoxLayout.getSizeBasis = getSizeBasis;
      /**
       * Set the box layout size basis for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the size basis.
       */
      function setSizeBasis(widget, value) {
          Private$2$1.sizeBasisProperty.set(widget, value);
      }
      BoxLayout.setSizeBasis = setSizeBasis;
  })(BoxLayout || (BoxLayout = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$2$1;
  (function (Private) {
      /**
       * The property descriptor for a widget stretch factor.
       */
      Private.stretchProperty = new AttachedProperty({
          name: 'stretch',
          create: function () { return 0; },
          coerce: function (owner, value) { return Math.max(0, Math.floor(value)); },
          changed: onChildSizingChanged
      });
      /**
       * The property descriptor for a widget size basis.
       */
      Private.sizeBasisProperty = new AttachedProperty({
          name: 'sizeBasis',
          create: function () { return 0; },
          coerce: function (owner, value) { return Math.max(0, Math.floor(value)); },
          changed: onChildSizingChanged
      });
      /**
       * Test whether a direction has horizontal orientation.
       */
      function isHorizontal(dir) {
          return dir === 'left-to-right' || dir === 'right-to-left';
      }
      Private.isHorizontal = isHorizontal;
      /**
       * Clamp a spacing value to an integer >= 0.
       */
      function clampSpacing(value) {
          return Math.max(0, Math.floor(value));
      }
      Private.clampSpacing = clampSpacing;
      /**
       * The change handler for the attached sizing properties.
       */
      function onChildSizingChanged(child) {
          if (child.parent && child.parent.layout instanceof BoxLayout) {
              child.parent.fit();
          }
      }
  })(Private$2$1 || (Private$2$1 = {}));

  /**
   * A simple and convenient panel widget class.
   *
   * #### Notes
   * This class is suitable as a base class for implementing a variety of
   * convenience panel widgets, but can also be used directly with CSS to
   * arrange a collection of widgets.
   *
   * This class provides a convenience wrapper around a [[PanelLayout]].
   */
  var Panel = /** @class */ (function (_super) {
      __extends$5(Panel, _super);
      /**
       * Construct a new panel.
       *
       * @param options - The options for initializing the panel.
       */
      function Panel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this) || this;
          _this.addClass('lm-Panel');
          /* <DEPRECATED> */
          _this.addClass('p-Panel');
          /* </DEPRECATED> */
          _this.layout = Private$3$1.createLayout(options);
          return _this;
      }
      Object.defineProperty(Panel.prototype, "widgets", {
          /**
           * A read-only array of the widgets in the panel.
           */
          get: function () {
              return this.layout.widgets;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Add a widget to the end of the panel.
       *
       * @param widget - The widget to add to the panel.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       */
      Panel.prototype.addWidget = function (widget) {
          this.layout.addWidget(widget);
      };
      /**
       * Insert a widget at the specified index.
       *
       * @param index - The index at which to insert the widget.
       *
       * @param widget - The widget to insert into to the panel.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       */
      Panel.prototype.insertWidget = function (index, widget) {
          this.layout.insertWidget(index, widget);
      };
      return Panel;
  }(Widget));
  /**
   * The namespace for the module implementation details.
   */
  var Private$3$1;
  (function (Private) {
      /**
       * Create a panel layout for the given panel options.
       */
      function createLayout(options) {
          return options.layout || new PanelLayout();
      }
      Private.createLayout = createLayout;
  })(Private$3$1 || (Private$3$1 = {}));

  /**
   * A panel which arranges its widgets in a single row or column.
   *
   * #### Notes
   * This class provides a convenience wrapper around a [[BoxLayout]].
   */
  var BoxPanel = /** @class */ (function (_super) {
      __extends$5(BoxPanel, _super);
      /**
       * Construct a new box panel.
       *
       * @param options - The options for initializing the box panel.
       */
      function BoxPanel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, { layout: Private$4$1.createLayout(options) }) || this;
          _this.addClass('lm-BoxPanel');
          /* <DEPRECATED> */
          _this.addClass('p-BoxPanel');
          return _this;
          /* </DEPRECATED> */
      }
      Object.defineProperty(BoxPanel.prototype, "direction", {
          /**
           * Get the layout direction for the box panel.
           */
          get: function () {
              return this.layout.direction;
          },
          /**
           * Set the layout direction for the box panel.
           */
          set: function (value) {
              this.layout.direction = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BoxPanel.prototype, "alignment", {
          /**
           * Get the content alignment for the box panel.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand to fill the
           * entire box layout.
           */
          get: function () {
              return this.layout.alignment;
          },
          /**
           * Set the content alignment for the box panel.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand to fill the
           * entire box layout.
           */
          set: function (value) {
              this.layout.alignment = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(BoxPanel.prototype, "spacing", {
          /**
           * Get the inter-element spacing for the box panel.
           */
          get: function () {
              return this.layout.spacing;
          },
          /**
           * Set the inter-element spacing for the box panel.
           */
          set: function (value) {
              this.layout.spacing = value;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * A message handler invoked on a `'child-added'` message.
       */
      BoxPanel.prototype.onChildAdded = function (msg) {
          msg.child.addClass('lm-BoxPanel-child');
          /* <DEPRECATED> */
          msg.child.addClass('p-BoxPanel-child');
          /* </DEPRECATED> */
      };
      /**
       * A message handler invoked on a `'child-removed'` message.
       */
      BoxPanel.prototype.onChildRemoved = function (msg) {
          msg.child.removeClass('lm-BoxPanel-child');
          /* <DEPRECATED> */
          msg.child.removeClass('p-BoxPanel-child');
          /* </DEPRECATED> */
      };
      return BoxPanel;
  }(Panel));
  /**
   * The namespace for the `BoxPanel` class statics.
   */
  (function (BoxPanel) {
      /**
       * Get the box panel stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The box panel stretch factor for the widget.
       */
      function getStretch(widget) {
          return BoxLayout.getStretch(widget);
      }
      BoxPanel.getStretch = getStretch;
      /**
       * Set the box panel stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the stretch factor.
       */
      function setStretch(widget, value) {
          BoxLayout.setStretch(widget, value);
      }
      BoxPanel.setStretch = setStretch;
      /**
       * Get the box panel size basis for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The box panel size basis for the widget.
       */
      function getSizeBasis(widget) {
          return BoxLayout.getSizeBasis(widget);
      }
      BoxPanel.getSizeBasis = getSizeBasis;
      /**
       * Set the box panel size basis for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the size basis.
       */
      function setSizeBasis(widget, value) {
          BoxLayout.setSizeBasis(widget, value);
      }
      BoxPanel.setSizeBasis = setSizeBasis;
  })(BoxPanel || (BoxPanel = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$4$1;
  (function (Private) {
      /**
       * Create a box layout for the given panel options.
       */
      function createLayout(options) {
          return options.layout || new BoxLayout(options);
      }
      Private.createLayout = createLayout;
  })(Private$4$1 || (Private$4$1 = {}));

  /**
   * A widget which displays command items as a searchable palette.
   */
  var CommandPalette = /** @class */ (function (_super) {
      __extends$5(CommandPalette, _super);
      /**
       * Construct a new command palette.
       *
       * @param options - The options for initializing the palette.
       */
      function CommandPalette(options) {
          var _this = _super.call(this, { node: Private$5$1.createNode() }) || this;
          _this._activeIndex = -1;
          _this._items = [];
          _this._results = null;
          _this.addClass('lm-CommandPalette');
          /* <DEPRECATED> */
          _this.addClass('p-CommandPalette');
          /* </DEPRECATED> */
          _this.setFlag(Widget.Flag.DisallowLayout);
          _this.commands = options.commands;
          _this.renderer = options.renderer || CommandPalette.defaultRenderer;
          _this.commands.commandChanged.connect(_this._onGenericChange, _this);
          _this.commands.keyBindingChanged.connect(_this._onGenericChange, _this);
          return _this;
      }
      /**
       * Dispose of the resources held by the widget.
       */
      CommandPalette.prototype.dispose = function () {
          this._items.length = 0;
          this._results = null;
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(CommandPalette.prototype, "searchNode", {
          /**
           * The command palette search node.
           *
           * #### Notes
           * This is the node which contains the search-related elements.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-CommandPalette-search')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(CommandPalette.prototype, "inputNode", {
          /**
           * The command palette input node.
           *
           * #### Notes
           * This is the actual input node for the search area.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-CommandPalette-input')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(CommandPalette.prototype, "contentNode", {
          /**
           * The command palette content node.
           *
           * #### Notes
           * This is the node which holds the command item nodes.
           *
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-CommandPalette-content')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(CommandPalette.prototype, "items", {
          /**
           * A read-only array of the command items in the palette.
           */
          get: function () {
              return this._items;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Add a command item to the command palette.
       *
       * @param options - The options for creating the command item.
       *
       * @returns The command item added to the palette.
       */
      CommandPalette.prototype.addItem = function (options) {
          // Create a new command item for the options.
          var item = Private$5$1.createItem(this.commands, options);
          // Add the item to the array.
          this._items.push(item);
          // Refresh the search results.
          this.refresh();
          // Return the item added to the palette.
          return item;
      };
      /**
       * Remove an item from the command palette.
       *
       * @param item - The item to remove from the palette.
       *
       * #### Notes
       * This is a no-op if the item is not in the palette.
       */
      CommandPalette.prototype.removeItem = function (item) {
          this.removeItemAt(this._items.indexOf(item));
      };
      /**
       * Remove the item at a given index from the command palette.
       *
       * @param index - The index of the item to remove.
       *
       * #### Notes
       * This is a no-op if the index is out of range.
       */
      CommandPalette.prototype.removeItemAt = function (index) {
          // Remove the item from the array.
          var item = ArrayExt.removeAt(this._items, index);
          // Bail if the index is out of range.
          if (!item) {
              return;
          }
          // Refresh the search results.
          this.refresh();
      };
      /**
       * Remove all items from the command palette.
       */
      CommandPalette.prototype.clearItems = function () {
          // Bail if there is nothing to remove.
          if (this._items.length === 0) {
              return;
          }
          // Clear the array of items.
          this._items.length = 0;
          // Refresh the search results.
          this.refresh();
      };
      /**
       * Clear the search results and schedule an update.
       *
       * #### Notes
       * This should be called whenever the search results of the palette
       * should be updated.
       *
       * This is typically called automatically by the palette as needed,
       * but can be called manually if the input text is programatically
       * changed.
       *
       * The rendered results are updated asynchronously.
       */
      CommandPalette.prototype.refresh = function () {
          this._results = null;
          if (this.inputNode.value !== '') {
              var clear = this.node.getElementsByClassName('lm-close-icon')[0];
              clear.style.display = 'inherit';
          }
          else {
              var clear = this.node.getElementsByClassName('lm-close-icon')[0];
              clear.style.display = 'none';
          }
          this.update();
      };
      /**
       * Handle the DOM events for the command palette.
       *
       * @param event - The DOM event sent to the command palette.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the command palette's DOM node.
       * It should not be called directly by user code.
       */
      CommandPalette.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'click':
                  this._evtClick(event);
                  break;
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'input':
                  this.refresh();
                  break;
              case 'focus':
              case 'blur':
                  this._toggleFocused();
                  break;
          }
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      CommandPalette.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('click', this);
          this.node.addEventListener('keydown', this);
          this.node.addEventListener('input', this);
          this.node.addEventListener('focus', this, true);
          this.node.addEventListener('blur', this, true);
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       */
      CommandPalette.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('click', this);
          this.node.removeEventListener('keydown', this);
          this.node.removeEventListener('input', this);
          this.node.removeEventListener('focus', this, true);
          this.node.removeEventListener('blur', this, true);
      };
      /**
       * A message handler invoked on an `'activate-request'` message.
       */
      CommandPalette.prototype.onActivateRequest = function (msg) {
          if (this.isAttached) {
              var input = this.inputNode;
              input.focus();
              input.select();
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      CommandPalette.prototype.onUpdateRequest = function (msg) {
          // Fetch the current query text and content node.
          var query = this.inputNode.value;
          var contentNode = this.contentNode;
          // Ensure the search results are generated.
          var results = this._results;
          if (!results) {
              // Generate and store the new search results.
              results = this._results = Private$5$1.search(this._items, query);
              // Reset the active index.
              this._activeIndex = (query ? ArrayExt.findFirstIndex(results, Private$5$1.canActivate) : -1);
          }
          // If there is no query and no results, clear the content.
          if (!query && results.length === 0) {
              VirtualDOM.render(null, contentNode);
              return;
          }
          // If the is a query but no results, render the empty message.
          if (query && results.length === 0) {
              var content_1 = this.renderer.renderEmptyMessage({ query: query });
              VirtualDOM.render(content_1, contentNode);
              return;
          }
          // Create the render content for the search results.
          var renderer = this.renderer;
          var activeIndex = this._activeIndex;
          var content = new Array(results.length);
          for (var i = 0, n = results.length; i < n; ++i) {
              var result = results[i];
              if (result.type === 'header') {
                  var indices = result.indices;
                  var category = result.category;
                  content[i] = renderer.renderHeader({ category: category, indices: indices });
              }
              else {
                  var item = result.item;
                  var indices = result.indices;
                  var active = i === activeIndex;
                  content[i] = renderer.renderItem({ item: item, indices: indices, active: active });
              }
          }
          // Render the search result content.
          VirtualDOM.render(content, contentNode);
          // Adjust the scroll position as needed.
          if (activeIndex < 0 || activeIndex >= results.length) {
              contentNode.scrollTop = 0;
          }
          else {
              var element = contentNode.children[activeIndex];
              ElementExt.scrollIntoViewIfNeeded(contentNode, element);
          }
      };
      /**
       * Handle the `'click'` event for the command palette.
       */
      CommandPalette.prototype._evtClick = function (event) {
          // Bail if the click is not the left button.
          if (event.button !== 0) {
              return;
          }
          // Clear input if the target is clear button
          if (event.target.classList.contains("lm-close-icon")) {
              this.inputNode.value = '';
              this.refresh();
              return;
          }
          // Find the index of the item which was clicked.
          var index = ArrayExt.findFirstIndex(this.contentNode.children, function (node) {
              return node.contains(event.target);
          });
          // Bail if the click was not on an item.
          if (index === -1) {
              return;
          }
          // Kill the event when a content item is clicked.
          event.preventDefault();
          event.stopPropagation();
          // Execute the item if possible.
          this._execute(index);
      };
      /**
       * Handle the `'keydown'` event for the command palette.
       */
      CommandPalette.prototype._evtKeyDown = function (event) {
          if (event.altKey || event.ctrlKey || event.metaKey || event.shiftKey) {
              return;
          }
          switch (event.keyCode) {
              case 13: // Enter
                  event.preventDefault();
                  event.stopPropagation();
                  this._execute(this._activeIndex);
                  break;
              case 38: // Up Arrow
                  event.preventDefault();
                  event.stopPropagation();
                  this._activatePreviousItem();
                  break;
              case 40: // Down Arrow
                  event.preventDefault();
                  event.stopPropagation();
                  this._activateNextItem();
                  break;
          }
      };
      /**
       * Activate the next enabled command item.
       */
      CommandPalette.prototype._activateNextItem = function () {
          // Bail if there are no search results.
          if (!this._results || this._results.length === 0) {
              return;
          }
          // Find the next enabled item index.
          var ai = this._activeIndex;
          var n = this._results.length;
          var start = ai < n - 1 ? ai + 1 : 0;
          var stop = start === 0 ? n - 1 : start - 1;
          this._activeIndex = ArrayExt.findFirstIndex(this._results, Private$5$1.canActivate, start, stop);
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Activate the previous enabled command item.
       */
      CommandPalette.prototype._activatePreviousItem = function () {
          // Bail if there are no search results.
          if (!this._results || this._results.length === 0) {
              return;
          }
          // Find the previous enabled item index.
          var ai = this._activeIndex;
          var n = this._results.length;
          var start = ai <= 0 ? n - 1 : ai - 1;
          var stop = start === n - 1 ? 0 : start + 1;
          this._activeIndex = ArrayExt.findLastIndex(this._results, Private$5$1.canActivate, start, stop);
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Execute the command item at the given index, if possible.
       */
      CommandPalette.prototype._execute = function (index) {
          // Bail if there are no search results.
          if (!this._results) {
              return;
          }
          // Bail if the index is out of range.
          var part = this._results[index];
          if (!part) {
              return;
          }
          // Update the search text if the item is a header.
          if (part.type === 'header') {
              var input = this.inputNode;
              input.value = part.category.toLowerCase() + " ";
              input.focus();
              this.refresh();
              return;
          }
          // Bail if item is not enabled.
          if (!part.item.isEnabled) {
              return;
          }
          // Execute the item.
          this.commands.execute(part.item.command, part.item.args);
          // Clear the query text.
          this.inputNode.value = '';
          // Refresh the search results.
          this.refresh();
      };
      /**
       * Toggle the focused modifier based on the input node focus state.
       */
      CommandPalette.prototype._toggleFocused = function () {
          var focused = document.activeElement === this.inputNode;
          this.toggleClass('lm-mod-focused', focused);
          /* <DEPRECATED> */
          this.toggleClass('p-mod-focused', focused);
          /* </DEPRECATED> */
      };
      /**
       * A signal handler for generic command changes.
       */
      CommandPalette.prototype._onGenericChange = function () {
          this.refresh();
      };
      return CommandPalette;
  }(Widget));
  /**
   * The namespace for the `CommandPalette` class statics.
   */
  (function (CommandPalette) {
      /**
       * The default implementation of `IRenderer`.
       */
      var Renderer = /** @class */ (function () {
          function Renderer() {
          }
          /**
           * Render the virtual element for a command palette header.
           *
           * @param data - The data to use for rendering the header.
           *
           * @returns A virtual element representing the header.
           */
          Renderer.prototype.renderHeader = function (data) {
              var content = this.formatHeader(data);
              return h.li({ className: 'lm-CommandPalette-header'
                      /* <DEPRECATED> */
                      + ' p-CommandPalette-header'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Render the virtual element for a command palette item.
           *
           * @param data - The data to use for rendering the item.
           *
           * @returns A virtual element representing the item.
           */
          Renderer.prototype.renderItem = function (data) {
              var className = this.createItemClass(data);
              var dataset = this.createItemDataset(data);
              return (h.li({ className: className, dataset: dataset }, this.renderItemIcon(data), this.renderItemContent(data), this.renderItemShortcut(data)));
          };
          /**
           * Render the empty results message for a command palette.
           *
           * @param data - The data to use for rendering the message.
           *
           * @returns A virtual element representing the message.
           */
          Renderer.prototype.renderEmptyMessage = function (data) {
              var content = this.formatEmptyMessage(data);
              return h.li({
                  className: 'lm-CommandPalette-emptyMessage'
                      /* <DEPRECATED> */
                      + ' p-CommandPalette-emptyMessage'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Render the icon for a command palette item.
           *
           * @param data - The data to use for rendering the icon.
           *
           * @returns A virtual element representing the icon.
           */
          Renderer.prototype.renderItemIcon = function (data) {
              var className = this.createIconClass(data);
              /* <DEPRECATED> */
              if (typeof data.item.icon === 'string') {
                  return h.div({ className: className }, data.item.iconLabel);
              }
              /* </DEPRECATED> */
              // if data.item.icon is undefined, it will be ignored
              return h.div({ className: className }, data.item.icon, data.item.iconLabel);
          };
          /**
           * Render the content for a command palette item.
           *
           * @param data - The data to use for rendering the content.
           *
           * @returns A virtual element representing the content.
           */
          Renderer.prototype.renderItemContent = function (data) {
              return (h.div({
                  className: 'lm-CommandPalette-itemContent'
                      /* <DEPRECATED> */
                      + ' p-CommandPalette-itemContent'
                  /* </DEPRECATED> */
              }, this.renderItemLabel(data), this.renderItemCaption(data)));
          };
          /**
           * Render the label for a command palette item.
           *
           * @param data - The data to use for rendering the label.
           *
           * @returns A virtual element representing the label.
           */
          Renderer.prototype.renderItemLabel = function (data) {
              var content = this.formatItemLabel(data);
              return h.div({
                  className: 'lm-CommandPalette-itemLabel'
                      /* <DEPRECATED> */
                      + ' p-CommandPalette-itemLabel'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Render the caption for a command palette item.
           *
           * @param data - The data to use for rendering the caption.
           *
           * @returns A virtual element representing the caption.
           */
          Renderer.prototype.renderItemCaption = function (data) {
              var content = this.formatItemCaption(data);
              return h.div({
                  className: 'lm-CommandPalette-itemCaption'
                      /* <DEPRECATED> */
                      + ' p-CommandPalette-itemCaption'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Render the shortcut for a command palette item.
           *
           * @param data - The data to use for rendering the shortcut.
           *
           * @returns A virtual element representing the shortcut.
           */
          Renderer.prototype.renderItemShortcut = function (data) {
              var content = this.formatItemShortcut(data);
              return h.div({
                  className: 'lm-CommandPalette-itemShortcut'
                      /* <DEPRECATED> */
                      + ' p-CommandPalette-itemShortcut'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Create the class name for the command palette item.
           *
           * @param data - The data to use for the class name.
           *
           * @returns The full class name for the command palette item.
           */
          Renderer.prototype.createItemClass = function (data) {
              // Set up the initial class name.
              var name = 'lm-CommandPalette-item';
              /* <DEPRECATED> */
              name += ' p-CommandPalette-item';
              /* </DEPRECATED> */
              // Add the boolean state classes.
              if (!data.item.isEnabled) {
                  name += ' lm-mod-disabled';
                  /* <DEPRECATED> */
                  name += ' p-mod-disabled';
                  /* </DEPRECATED> */
              }
              if (data.item.isToggled) {
                  name += ' lm-mod-toggled';
                  /* <DEPRECATED> */
                  name += ' p-mod-toggled';
                  /* </DEPRECATED> */
              }
              if (data.active) {
                  name += ' lm-mod-active';
                  /* <DEPRECATED> */
                  name += ' p-mod-active';
                  /* </DEPRECATED> */
              }
              // Add the extra class.
              var extra = data.item.className;
              if (extra) {
                  name += " " + extra;
              }
              // Return the complete class name.
              return name;
          };
          /**
           * Create the dataset for the command palette item.
           *
           * @param data - The data to use for creating the dataset.
           *
           * @returns The dataset for the command palette item.
           */
          Renderer.prototype.createItemDataset = function (data) {
              return __assign$3(__assign$3({}, data.item.dataset), { command: data.item.command });
          };
          /**
           * Create the class name for the command item icon.
           *
           * @param data - The data to use for the class name.
           *
           * @returns The full class name for the item icon.
           */
          Renderer.prototype.createIconClass = function (data) {
              var name = 'lm-CommandPalette-itemIcon';
              /* <DEPRECATED> */
              name += ' p-CommandPalette-itemIcon';
              /* </DEPRECATED> */
              var extra = data.item.iconClass;
              return extra ? name + " " + extra : name;
          };
          /**
           * Create the render content for the header node.
           *
           * @param data - The data to use for the header content.
           *
           * @returns The content to add to the header node.
           */
          Renderer.prototype.formatHeader = function (data) {
              if (!data.indices || data.indices.length === 0) {
                  return data.category;
              }
              return StringExt.highlight(data.category, data.indices, h.mark);
          };
          /**
           * Create the render content for the empty message node.
           *
           * @param data - The data to use for the empty message content.
           *
           * @returns The content to add to the empty message node.
           */
          Renderer.prototype.formatEmptyMessage = function (data) {
              return "No commands found that match '" + data.query + "'";
          };
          /**
           * Create the render content for the item shortcut node.
           *
           * @param data - The data to use for the shortcut content.
           *
           * @returns The content to add to the shortcut node.
           */
          Renderer.prototype.formatItemShortcut = function (data) {
              var kb = data.item.keyBinding;
              return kb ? kb.keys.map(CommandRegistry.formatKeystroke).join(', ') : null;
          };
          /**
           * Create the render content for the item label node.
           *
           * @param data - The data to use for the label content.
           *
           * @returns The content to add to the label node.
           */
          Renderer.prototype.formatItemLabel = function (data) {
              if (!data.indices || data.indices.length === 0) {
                  return data.item.label;
              }
              return StringExt.highlight(data.item.label, data.indices, h.mark);
          };
          /**
           * Create the render content for the item caption node.
           *
           * @param data - The data to use for the caption content.
           *
           * @returns The content to add to the caption node.
           */
          Renderer.prototype.formatItemCaption = function (data) {
              return data.item.caption;
          };
          return Renderer;
      }());
      CommandPalette.Renderer = Renderer;
      /**
       * The default `Renderer` instance.
       */
      CommandPalette.defaultRenderer = new Renderer();
  })(CommandPalette || (CommandPalette = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$5$1;
  (function (Private) {
      /**
       * Create the DOM node for a command palette.
       */
      function createNode() {
          var node = document.createElement('div');
          var search = document.createElement('div');
          var wrapper = document.createElement('div');
          var input = document.createElement('input');
          var content = document.createElement('ul');
          var clear = document.createElement('button');
          search.className = 'lm-CommandPalette-search';
          wrapper.className = 'lm-CommandPalette-wrapper';
          input.className = 'lm-CommandPalette-input';
          clear.className = 'lm-close-icon';
          content.className = 'lm-CommandPalette-content';
          /* <DEPRECATED> */
          search.classList.add('p-CommandPalette-search');
          wrapper.classList.add('p-CommandPalette-wrapper');
          input.classList.add('p-CommandPalette-input');
          content.classList.add('p-CommandPalette-content');
          /* </DEPRECATED> */
          input.spellcheck = false;
          wrapper.appendChild(input);
          wrapper.appendChild(clear);
          search.appendChild(wrapper);
          node.appendChild(search);
          node.appendChild(content);
          return node;
      }
      Private.createNode = createNode;
      /**
       * Create a new command item from a command registry and options.
       */
      function createItem(commands, options) {
          return new CommandItem(commands, options);
      }
      Private.createItem = createItem;
      /**
       * Search an array of command items for fuzzy matches.
       */
      function search(items, query) {
          // Fuzzy match the items for the query.
          var scores = matchItems(items, query);
          // Sort the items based on their score.
          scores.sort(scoreCmp);
          // Create the results for the search.
          return createResults(scores);
      }
      Private.search = search;
      /**
       * Test whether a result item can be activated.
       */
      function canActivate(result) {
          return result.type === 'item' && result.item.isEnabled;
      }
      Private.canActivate = canActivate;
      /**
       * Normalize a category for a command item.
       */
      function normalizeCategory(category) {
          return category.trim().replace(/\s+/g, ' ');
      }
      /**
       * Normalize the query text for a fuzzy search.
       */
      function normalizeQuery(text) {
          return text.replace(/\s+/g, '').toLowerCase();
      }
      /**
       * Perform a fuzzy match on an array of command items.
       */
      function matchItems(items, query) {
          // Normalize the query text to lower case with no whitespace.
          query = normalizeQuery(query);
          // Create the array to hold the scores.
          var scores = [];
          // Iterate over the items and match against the query.
          for (var i = 0, n = items.length; i < n; ++i) {
              // Ignore items which are not visible.
              var item = items[i];
              if (!item.isVisible) {
                  continue;
              }
              // If the query is empty, all items are matched by default.
              if (!query) {
                  scores.push({
                      matchType: 3 /* Default */,
                      categoryIndices: null,
                      labelIndices: null,
                      score: 0, item: item
                  });
                  continue;
              }
              // Run the fuzzy search for the item and query.
              var score = fuzzySearch(item, query);
              // Ignore the item if it is not a match.
              if (!score) {
                  continue;
              }
              // Penalize disabled items.
              // TODO - push disabled items all the way down in sort cmp?
              if (!item.isEnabled) {
                  score.score += 1000;
              }
              // Add the score to the results.
              scores.push(score);
          }
          // Return the final array of scores.
          return scores;
      }
      /**
       * Perform a fuzzy search on a single command item.
       */
      function fuzzySearch(item, query) {
          // Create the source text to be searched.
          var category = item.category.toLowerCase();
          var label = item.label.toLowerCase();
          var source = category + " " + label;
          // Set up the match score and indices array.
          var score = Infinity;
          var indices = null;
          // The regex for search word boundaries
          var rgx = /\b\w/g;
          // Search the source by word boundary.
          while (true) {
              // Find the next word boundary in the source.
              var rgxMatch = rgx.exec(source);
              // Break if there is no more source context.
              if (!rgxMatch) {
                  break;
              }
              // Run the string match on the relevant substring.
              var match = StringExt.matchSumOfDeltas(source, query, rgxMatch.index);
              // Break if there is no match.
              if (!match) {
                  break;
              }
              // Update the match if the score is better.
              if (match && match.score <= score) {
                  score = match.score;
                  indices = match.indices;
              }
          }
          // Bail if there was no match.
          if (!indices || score === Infinity) {
              return null;
          }
          // Compute the pivot index between category and label text.
          var pivot = category.length + 1;
          // Find the slice index to separate matched indices.
          var j = ArrayExt.lowerBound(indices, pivot, function (a, b) { return a - b; });
          // Extract the matched category and label indices.
          var categoryIndices = indices.slice(0, j);
          var labelIndices = indices.slice(j);
          // Adjust the label indices for the pivot offset.
          for (var i = 0, n = labelIndices.length; i < n; ++i) {
              labelIndices[i] -= pivot;
          }
          // Handle a pure label match.
          if (categoryIndices.length === 0) {
              return {
                  matchType: 0 /* Label */,
                  categoryIndices: null,
                  labelIndices: labelIndices,
                  score: score, item: item
              };
          }
          // Handle a pure category match.
          if (labelIndices.length === 0) {
              return {
                  matchType: 1 /* Category */,
                  categoryIndices: categoryIndices,
                  labelIndices: null,
                  score: score, item: item
              };
          }
          // Handle a split match.
          return {
              matchType: 2 /* Split */,
              categoryIndices: categoryIndices,
              labelIndices: labelIndices,
              score: score, item: item
          };
      }
      /**
       * A sort comparison function for a match score.
       */
      function scoreCmp(a, b) {
          // First compare based on the match type
          var m1 = a.matchType - b.matchType;
          if (m1 !== 0) {
              return m1;
          }
          // Otherwise, compare based on the match score.
          var d1 = a.score - b.score;
          if (d1 !== 0) {
              return d1;
          }
          // Find the match index based on the match type.
          var i1 = 0;
          var i2 = 0;
          switch (a.matchType) {
              case 0 /* Label */:
                  i1 = a.labelIndices[0];
                  i2 = b.labelIndices[0];
                  break;
              case 1 /* Category */:
              case 2 /* Split */:
                  i1 = a.categoryIndices[0];
                  i2 = b.categoryIndices[0];
                  break;
          }
          // Compare based on the match index.
          if (i1 !== i2) {
              return i1 - i2;
          }
          // Otherwise, compare by category.
          var d2 = a.item.category.localeCompare(b.item.category);
          if (d2 !== 0) {
              return d2;
          }
          // Otherwise, compare by rank.
          var r1 = a.item.rank;
          var r2 = b.item.rank;
          if (r1 !== r2) {
              return r1 < r2 ? -1 : 1; // Infinity safe
          }
          // Finally, compare by label.
          return a.item.label.localeCompare(b.item.label);
      }
      /**
       * Create the results from an array of sorted scores.
       */
      function createResults(scores) {
          // Set up an array to track which scores have been visited.
          var visited = new Array(scores.length);
          ArrayExt.fill(visited, false);
          // Set up the search results array.
          var results = [];
          // Iterate over each score in the array.
          for (var i = 0, n = scores.length; i < n; ++i) {
              // Ignore a score which has already been processed.
              if (visited[i]) {
                  continue;
              }
              // Extract the current item and indices.
              var _a = scores[i], item = _a.item, categoryIndices = _a.categoryIndices;
              // Extract the category for the current item.
              var category = item.category;
              // Add the header result for the category.
              results.push({ type: 'header', category: category, indices: categoryIndices });
              // Find the rest of the scores with the same category.
              for (var j = i; j < n; ++j) {
                  // Ignore a score which has already been processed.
                  if (visited[j]) {
                      continue;
                  }
                  // Extract the data for the current score.
                  var _b = scores[j], item_1 = _b.item, labelIndices = _b.labelIndices;
                  // Ignore an item with a different category.
                  if (item_1.category !== category) {
                      continue;
                  }
                  // Create the item result for the score.
                  results.push({ type: 'item', item: item_1, indices: labelIndices });
                  // Mark the score as processed.
                  visited[j] = true;
              }
          }
          // Return the final results.
          return results;
      }
      /**
       * A concrete implementation of `CommandPalette.IItem`.
       */
      var CommandItem = /** @class */ (function () {
          /**
           * Construct a new command item.
           */
          function CommandItem(commands, options) {
              this._commands = commands;
              this.category = normalizeCategory(options.category);
              this.command = options.command;
              this.args = options.args || JSONExt.emptyObject;
              this.rank = options.rank !== undefined ? options.rank : Infinity;
          }
          Object.defineProperty(CommandItem.prototype, "label", {
              /**
               * The display label for the command item.
               */
              get: function () {
                  return this._commands.label(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "icon", {
              /**
               * The icon renderer for the command item.
               */
              get: function () {
                  return this._commands.icon(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "iconClass", {
              /**
               * The icon class for the command item.
               */
              get: function () {
                  return this._commands.iconClass(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "iconLabel", {
              /**
               * The icon label for the command item.
               */
              get: function () {
                  return this._commands.iconLabel(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "caption", {
              /**
               * The display caption for the command item.
               */
              get: function () {
                  return this._commands.caption(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "className", {
              /**
               * The extra class name for the command item.
               */
              get: function () {
                  return this._commands.className(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "dataset", {
              /**
               * The dataset for the command item.
               */
              get: function () {
                  return this._commands.dataset(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "isEnabled", {
              /**
               * Whether the command item is enabled.
               */
              get: function () {
                  return this._commands.isEnabled(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "isToggled", {
              /**
               * Whether the command item is toggled.
               */
              get: function () {
                  return this._commands.isToggled(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "isVisible", {
              /**
               * Whether the command item is visible.
               */
              get: function () {
                  return this._commands.isVisible(this.command, this.args);
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(CommandItem.prototype, "keyBinding", {
              /**
               * The key binding for the command item.
               */
              get: function () {
                  var _a = this, command = _a.command, args = _a.args;
                  return ArrayExt.findLastValue(this._commands.keyBindings, function (kb) {
                      return kb.command === command && JSONExt.deepEqual(kb.args, args);
                  }) || null;
              },
              enumerable: true,
              configurable: true
          });
          return CommandItem;
      }());
  })(Private$5$1 || (Private$5$1 = {}));

  /**
   * A widget which displays items as a canonical menu.
   */
  var Menu = /** @class */ (function (_super) {
      __extends$5(Menu, _super);
      /**
       * Construct a new menu.
       *
       * @param options - The options for initializing the menu.
       */
      function Menu(options) {
          var _this = _super.call(this, { node: Private$6$1.createNode() }) || this;
          _this._childIndex = -1;
          _this._activeIndex = -1;
          _this._openTimerID = 0;
          _this._closeTimerID = 0;
          _this._items = [];
          _this._childMenu = null;
          _this._parentMenu = null;
          _this._aboutToClose = new Signal(_this);
          _this._menuRequested = new Signal(_this);
          _this.addClass('lm-Menu');
          /* <DEPRECATED> */
          _this.addClass('p-Menu');
          /* </DEPRECATED> */
          _this.setFlag(Widget.Flag.DisallowLayout);
          _this.commands = options.commands;
          _this.renderer = options.renderer || Menu.defaultRenderer;
          return _this;
      }
      /**
       * Dispose of the resources held by the menu.
       */
      Menu.prototype.dispose = function () {
          this.close();
          this._items.length = 0;
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(Menu.prototype, "aboutToClose", {
          /**
           * A signal emitted just before the menu is closed.
           *
           * #### Notes
           * This signal is emitted when the menu receives a `'close-request'`
           * message, just before it removes itself from the DOM.
           *
           * This signal is not emitted if the menu is already detached from
           * the DOM when it receives the `'close-request'` message.
           */
          get: function () {
              return this._aboutToClose;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "menuRequested", {
          /**
           * A signal emitted when a new menu is requested by the user.
           *
           * #### Notes
           * This signal is emitted whenever the user presses the right or left
           * arrow keys, and a submenu cannot be opened or closed in response.
           *
           * This signal is useful when implementing menu bars in order to open
           * the next or previous menu in response to a user key press.
           *
           * This signal is only emitted for the root menu in a hierarchy.
           */
          get: function () {
              return this._menuRequested;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "parentMenu", {
          /**
           * The parent menu of the menu.
           *
           * #### Notes
           * This is `null` unless the menu is an open submenu.
           */
          get: function () {
              return this._parentMenu;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "childMenu", {
          /**
           * The child menu of the menu.
           *
           * #### Notes
           * This is `null` unless the menu has an open submenu.
           */
          get: function () {
              return this._childMenu;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "rootMenu", {
          /**
           * The root menu of the menu hierarchy.
           */
          get: function () {
              var menu = this;
              while (menu._parentMenu) {
                  menu = menu._parentMenu;
              }
              return menu;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "leafMenu", {
          /**
           * The leaf menu of the menu hierarchy.
           */
          get: function () {
              var menu = this;
              while (menu._childMenu) {
                  menu = menu._childMenu;
              }
              return menu;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "contentNode", {
          /**
           * The menu content node.
           *
           * #### Notes
           * This is the node which holds the menu item nodes.
           *
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-Menu-content')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "activeItem", {
          /**
           * Get the currently active menu item.
           */
          get: function () {
              return this._items[this._activeIndex] || null;
          },
          /**
           * Set the currently active menu item.
           *
           * #### Notes
           * If the item cannot be activated, the item will be set to `null`.
           */
          set: function (value) {
              this.activeIndex = value ? this._items.indexOf(value) : -1;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "activeIndex", {
          /**
           * Get the index of the currently active menu item.
           *
           * #### Notes
           * This will be `-1` if no menu item is active.
           */
          get: function () {
              return this._activeIndex;
          },
          /**
           * Set the index of the currently active menu item.
           *
           * #### Notes
           * If the item cannot be activated, the index will be set to `-1`.
           */
          set: function (value) {
              // Adjust the value for an out of range index.
              if (value < 0 || value >= this._items.length) {
                  value = -1;
              }
              // Ensure the item can be activated.
              if (value !== -1 && !Private$6$1.canActivate(this._items[value])) {
                  value = -1;
              }
              // Bail if the index will not change.
              if (this._activeIndex === value) {
                  return;
              }
              // Update the active index.
              this._activeIndex = value;
              // schedule an update of the items.
              this.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Menu.prototype, "items", {
          /**
           * A read-only array of the menu items in the menu.
           */
          get: function () {
              return this._items;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Activate the next selectable item in the menu.
       *
       * #### Notes
       * If no item is selectable, the index will be set to `-1`.
       */
      Menu.prototype.activateNextItem = function () {
          var n = this._items.length;
          var ai = this._activeIndex;
          var start = ai < n - 1 ? ai + 1 : 0;
          var stop = start === 0 ? n - 1 : start - 1;
          this.activeIndex = ArrayExt.findFirstIndex(this._items, Private$6$1.canActivate, start, stop);
      };
      /**
       * Activate the previous selectable item in the menu.
       *
       * #### Notes
       * If no item is selectable, the index will be set to `-1`.
       */
      Menu.prototype.activatePreviousItem = function () {
          var n = this._items.length;
          var ai = this._activeIndex;
          var start = ai <= 0 ? n - 1 : ai - 1;
          var stop = start === n - 1 ? 0 : start + 1;
          this.activeIndex = ArrayExt.findLastIndex(this._items, Private$6$1.canActivate, start, stop);
      };
      /**
       * Trigger the active menu item.
       *
       * #### Notes
       * If the active item is a submenu, it will be opened and the first
       * item will be activated.
       *
       * If the active item is a command, the command will be executed.
       *
       * If the menu is not attached, this is a no-op.
       *
       * If there is no active item, this is a no-op.
       */
      Menu.prototype.triggerActiveItem = function () {
          // Bail if the menu is not attached.
          if (!this.isAttached) {
              return;
          }
          // Bail if there is no active item.
          var item = this.activeItem;
          if (!item) {
              return;
          }
          // Cancel the pending timers.
          this._cancelOpenTimer();
          this._cancelCloseTimer();
          // If the item is a submenu, open it.
          if (item.type === 'submenu') {
              this._openChildMenu(true);
              return;
          }
          // Close the root menu before executing the command.
          this.rootMenu.close();
          // Execute the command for the item.
          var command = item.command, args = item.args;
          if (this.commands.isEnabled(command, args)) {
              this.commands.execute(command, args);
          }
          else {
              console.log("Command '" + command + "' is disabled.");
          }
      };
      /**
       * Add a menu item to the end of the menu.
       *
       * @param options - The options for creating the menu item.
       *
       * @returns The menu item added to the menu.
       */
      Menu.prototype.addItem = function (options) {
          return this.insertItem(this._items.length, options);
      };
      /**
       * Insert a menu item into the menu at the specified index.
       *
       * @param index - The index at which to insert the item.
       *
       * @param options - The options for creating the menu item.
       *
       * @returns The menu item added to the menu.
       *
       * #### Notes
       * The index will be clamped to the bounds of the items.
       */
      Menu.prototype.insertItem = function (index, options) {
          // Close the menu if it's attached.
          if (this.isAttached) {
              this.close();
          }
          // Reset the active index.
          this.activeIndex = -1;
          // Clamp the insert index to the array bounds.
          var i = Math.max(0, Math.min(index, this._items.length));
          // Create the item for the options.
          var item = Private$6$1.createItem(this, options);
          // Insert the item into the array.
          ArrayExt.insert(this._items, i, item);
          // Schedule an update of the items.
          this.update();
          // Return the item added to the menu.
          return item;
      };
      /**
       * Remove an item from the menu.
       *
       * @param item - The item to remove from the menu.
       *
       * #### Notes
       * This is a no-op if the item is not in the menu.
       */
      Menu.prototype.removeItem = function (item) {
          this.removeItemAt(this._items.indexOf(item));
      };
      /**
       * Remove the item at a given index from the menu.
       *
       * @param index - The index of the item to remove.
       *
       * #### Notes
       * This is a no-op if the index is out of range.
       */
      Menu.prototype.removeItemAt = function (index) {
          // Close the menu if it's attached.
          if (this.isAttached) {
              this.close();
          }
          // Reset the active index.
          this.activeIndex = -1;
          // Remove the item from the array.
          var item = ArrayExt.removeAt(this._items, index);
          // Bail if the index is out of range.
          if (!item) {
              return;
          }
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Remove all menu items from the menu.
       */
      Menu.prototype.clearItems = function () {
          // Close the menu if it's attached.
          if (this.isAttached) {
              this.close();
          }
          // Reset the active index.
          this.activeIndex = -1;
          // Bail if there is nothing to remove.
          if (this._items.length === 0) {
              return;
          }
          // Clear the items.
          this._items.length = 0;
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Open the menu at the specified location.
       *
       * @param x - The client X coordinate of the menu location.
       *
       * @param y - The client Y coordinate of the menu location.
       *
       * @param options - The additional options for opening the menu.
       *
       * #### Notes
       * The menu will be opened at the given location unless it will not
       * fully fit on the screen. If it will not fit, it will be adjusted
       * to fit naturally on the screen.
       *
       * This is a no-op if the menu is already attached to the DOM.
       */
      Menu.prototype.open = function (x, y, options) {
          if (options === void 0) { options = {}; }
          // Bail early if the menu is already attached.
          if (this.isAttached) {
              return;
          }
          // Extract the position options.
          var forceX = options.forceX || false;
          var forceY = options.forceY || false;
          // Open the menu as a root menu.
          Private$6$1.openRootMenu(this, x, y, forceX, forceY);
          // Activate the menu to accept keyboard input.
          this.activate();
      };
      /**
       * Handle the DOM events for the menu.
       *
       * @param event - The DOM event sent to the menu.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the menu's DOM nodes. It should
       * not be called directly by user code.
       */
      Menu.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'mouseup':
                  this._evtMouseUp(event);
                  break;
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseenter':
                  this._evtMouseEnter(event);
                  break;
              case 'mouseleave':
                  this._evtMouseLeave(event);
                  break;
              case 'mousedown':
                  this._evtMouseDown(event);
                  break;
              case 'contextmenu':
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      Menu.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('keydown', this);
          this.node.addEventListener('mouseup', this);
          this.node.addEventListener('mousemove', this);
          this.node.addEventListener('mouseenter', this);
          this.node.addEventListener('mouseleave', this);
          this.node.addEventListener('contextmenu', this);
          document.addEventListener('mousedown', this, true);
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       */
      Menu.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('keydown', this);
          this.node.removeEventListener('mouseup', this);
          this.node.removeEventListener('mousemove', this);
          this.node.removeEventListener('mouseenter', this);
          this.node.removeEventListener('mouseleave', this);
          this.node.removeEventListener('contextmenu', this);
          document.removeEventListener('mousedown', this, true);
      };
      /**
       * A message handler invoked on an `'activate-request'` message.
       */
      Menu.prototype.onActivateRequest = function (msg) {
          if (this.isAttached) {
              this.node.focus();
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      Menu.prototype.onUpdateRequest = function (msg) {
          var items = this._items;
          var renderer = this.renderer;
          var activeIndex = this._activeIndex;
          var collapsedFlags = Private$6$1.computeCollapsed(items);
          var content = new Array(items.length);
          for (var i = 0, n = items.length; i < n; ++i) {
              var item = items[i];
              var active = i === activeIndex;
              var collapsed = collapsedFlags[i];
              content[i] = renderer.renderItem({ item: item, active: active, collapsed: collapsed });
          }
          VirtualDOM.render(content, this.contentNode);
      };
      /**
       * A message handler invoked on a `'close-request'` message.
       */
      Menu.prototype.onCloseRequest = function (msg) {
          // Cancel the pending timers.
          this._cancelOpenTimer();
          this._cancelCloseTimer();
          // Reset the active index.
          this.activeIndex = -1;
          // Close any open child menu.
          var childMenu = this._childMenu;
          if (childMenu) {
              this._childIndex = -1;
              this._childMenu = null;
              childMenu._parentMenu = null;
              childMenu.close();
          }
          // Remove this menu from its parent and activate the parent.
          var parentMenu = this._parentMenu;
          if (parentMenu) {
              this._parentMenu = null;
              parentMenu._childIndex = -1;
              parentMenu._childMenu = null;
              parentMenu.activate();
          }
          // Emit the `aboutToClose` signal if the menu is attached.
          if (this.isAttached) {
              this._aboutToClose.emit(undefined);
          }
          // Finish closing the menu.
          _super.prototype.onCloseRequest.call(this, msg);
      };
      /**
       * Handle the `'keydown'` event for the menu.
       *
       * #### Notes
       * This listener is attached to the menu node.
       */
      Menu.prototype._evtKeyDown = function (event) {
          // A menu handles all keydown events.
          event.preventDefault();
          event.stopPropagation();
          // Fetch the key code for the event.
          var kc = event.keyCode;
          // Enter
          if (kc === 13) {
              this.triggerActiveItem();
              return;
          }
          // Escape
          if (kc === 27) {
              this.close();
              return;
          }
          // Left Arrow
          if (kc === 37) {
              if (this._parentMenu) {
                  this.close();
              }
              else {
                  this._menuRequested.emit('previous');
              }
              return;
          }
          // Up Arrow
          if (kc === 38) {
              this.activatePreviousItem();
              return;
          }
          // Right Arrow
          if (kc === 39) {
              var item = this.activeItem;
              if (item && item.type === 'submenu') {
                  this.triggerActiveItem();
              }
              else {
                  this.rootMenu._menuRequested.emit('next');
              }
              return;
          }
          // Down Arrow
          if (kc === 40) {
              this.activateNextItem();
              return;
          }
          // Get the pressed key character.
          var key = getKeyboardLayout().keyForKeydownEvent(event);
          // Bail if the key is not valid.
          if (!key) {
              return;
          }
          // Search for the next best matching mnemonic item.
          var start = this._activeIndex + 1;
          var result = Private$6$1.findMnemonic(this._items, key, start);
          // Handle the requested mnemonic based on the search results.
          // If exactly one mnemonic is matched, that item is triggered.
          // Otherwise, the next mnemonic is activated if available,
          // followed by the auto mnemonic if available.
          if (result.index !== -1 && !result.multiple) {
              this.activeIndex = result.index;
              this.triggerActiveItem();
          }
          else if (result.index !== -1) {
              this.activeIndex = result.index;
          }
          else if (result.auto !== -1) {
              this.activeIndex = result.auto;
          }
      };
      /**
       * Handle the `'mouseup'` event for the menu.
       *
       * #### Notes
       * This listener is attached to the menu node.
       */
      Menu.prototype._evtMouseUp = function (event) {
          if (event.button !== 0) {
              return;
          }
          event.preventDefault();
          event.stopPropagation();
          this.triggerActiveItem();
      };
      /**
       * Handle the `'mousemove'` event for the menu.
       *
       * #### Notes
       * This listener is attached to the menu node.
       */
      Menu.prototype._evtMouseMove = function (event) {
          // Hit test the item nodes for the item under the mouse.
          var index = ArrayExt.findFirstIndex(this.contentNode.children, function (node) {
              return ElementExt.hitTest(node, event.clientX, event.clientY);
          });
          // Bail early if the mouse is already over the active index.
          if (index === this._activeIndex) {
              return;
          }
          // Update and coerce the active index.
          this.activeIndex = index;
          index = this.activeIndex;
          // If the index is the current child index, cancel the timers.
          if (index === this._childIndex) {
              this._cancelOpenTimer();
              this._cancelCloseTimer();
              return;
          }
          // If a child menu is currently open, start the close timer.
          if (this._childIndex !== -1) {
              this._startCloseTimer();
          }
          // Cancel the open timer to give a full delay for opening.
          this._cancelOpenTimer();
          // Bail if the active item is not a valid submenu item.
          var item = this.activeItem;
          if (!item || item.type !== 'submenu' || !item.submenu) {
              return;
          }
          // Start the open timer to open the active item submenu.
          this._startOpenTimer();
      };
      /**
       * Handle the `'mouseenter'` event for the menu.
       *
       * #### Notes
       * This listener is attached to the menu node.
       */
      Menu.prototype._evtMouseEnter = function (event) {
          // Synchronize the active ancestor items.
          for (var menu = this._parentMenu; menu; menu = menu._parentMenu) {
              menu._cancelOpenTimer();
              menu._cancelCloseTimer();
              menu.activeIndex = menu._childIndex;
          }
      };
      /**
       * Handle the `'mouseleave'` event for the menu.
       *
       * #### Notes
       * This listener is attached to the menu node.
       */
      Menu.prototype._evtMouseLeave = function (event) {
          // Cancel any pending submenu opening.
          this._cancelOpenTimer();
          // If there is no open child menu, just reset the active index.
          if (!this._childMenu) {
              this.activeIndex = -1;
              return;
          }
          // If the mouse is over the child menu, cancel the close timer.
          var clientX = event.clientX, clientY = event.clientY;
          if (ElementExt.hitTest(this._childMenu.node, clientX, clientY)) {
              this._cancelCloseTimer();
              return;
          }
          // Otherwise, reset the active index and start the close timer.
          this.activeIndex = -1;
          this._startCloseTimer();
      };
      /**
       * Handle the `'mousedown'` event for the menu.
       *
       * #### Notes
       * This listener is attached to the document node.
       */
      Menu.prototype._evtMouseDown = function (event) {
          // Bail if the menu is not a root menu.
          if (this._parentMenu) {
              return;
          }
          // The mouse button which is pressed is irrelevant. If the press
          // is not on a menu, the entire hierarchy is closed and the event
          // is allowed to propagate. This allows other code to act on the
          // event, such as focusing the clicked element.
          if (Private$6$1.hitTestMenus(this, event.clientX, event.clientY)) {
              event.preventDefault();
              event.stopPropagation();
          }
          else {
              this.close();
          }
      };
      /**
       * Open the child menu at the active index immediately.
       *
       * If a different child menu is already open, it will be closed,
       * even if the active item is not a valid submenu.
       */
      Menu.prototype._openChildMenu = function (activateFirst) {
          if (activateFirst === void 0) { activateFirst = false; }
          // If the item is not a valid submenu, close the child menu.
          var item = this.activeItem;
          if (!item || item.type !== 'submenu' || !item.submenu) {
              this._closeChildMenu();
              return;
          }
          // Do nothing if the child menu will not change.
          var submenu = item.submenu;
          if (submenu === this._childMenu) {
              return;
          }
          // Ensure the current child menu is closed.
          this._closeChildMenu();
          // Update the private child state.
          this._childMenu = submenu;
          this._childIndex = this._activeIndex;
          // Set the parent menu reference for the child.
          submenu._parentMenu = this;
          // Ensure the menu is updated and lookup the item node.
          MessageLoop.sendMessage(this, Widget.Msg.UpdateRequest);
          var itemNode = this.contentNode.children[this._activeIndex];
          // Open the submenu at the active node.
          Private$6$1.openSubmenu(submenu, itemNode);
          // Activate the first item if desired.
          if (activateFirst) {
              submenu.activeIndex = -1;
              submenu.activateNextItem();
          }
          // Activate the child menu.
          submenu.activate();
      };
      /**
       * Close the child menu immediately.
       *
       * This is a no-op if a child menu is not open.
       */
      Menu.prototype._closeChildMenu = function () {
          if (this._childMenu) {
              this._childMenu.close();
          }
      };
      /**
       * Start the open timer, unless it is already pending.
       */
      Menu.prototype._startOpenTimer = function () {
          var _this = this;
          if (this._openTimerID === 0) {
              this._openTimerID = window.setTimeout(function () {
                  _this._openTimerID = 0;
                  _this._openChildMenu();
              }, Private$6$1.TIMER_DELAY);
          }
      };
      /**
       * Start the close timer, unless it is already pending.
       */
      Menu.prototype._startCloseTimer = function () {
          var _this = this;
          if (this._closeTimerID === 0) {
              this._closeTimerID = window.setTimeout(function () {
                  _this._closeTimerID = 0;
                  _this._closeChildMenu();
              }, Private$6$1.TIMER_DELAY);
          }
      };
      /**
       * Cancel the open timer, if the timer is pending.
       */
      Menu.prototype._cancelOpenTimer = function () {
          if (this._openTimerID !== 0) {
              clearTimeout(this._openTimerID);
              this._openTimerID = 0;
          }
      };
      /**
       * Cancel the close timer, if the timer is pending.
       */
      Menu.prototype._cancelCloseTimer = function () {
          if (this._closeTimerID !== 0) {
              clearTimeout(this._closeTimerID);
              this._closeTimerID = 0;
          }
      };
      return Menu;
  }(Widget));
  /**
   * The namespace for the `Menu` class statics.
   */
  (function (Menu) {
      /**
       * The default implementation of `IRenderer`.
       *
       * #### Notes
       * Subclasses are free to reimplement rendering methods as needed.
       */
      var Renderer = /** @class */ (function () {
          /**
           * Construct a new renderer.
           */
          function Renderer() {
          }
          /**
           * Render the virtual element for a menu item.
           *
           * @param data - The data to use for rendering the item.
           *
           * @returns A virtual element representing the item.
           */
          Renderer.prototype.renderItem = function (data) {
              var className = this.createItemClass(data);
              var dataset = this.createItemDataset(data);
              return (h.li({ className: className, dataset: dataset }, this.renderIcon(data), this.renderLabel(data), this.renderShortcut(data), this.renderSubmenu(data)));
          };
          /**
           * Render the icon element for a menu item.
           *
           * @param data - The data to use for rendering the icon.
           *
           * @returns A virtual element representing the item icon.
           */
          Renderer.prototype.renderIcon = function (data) {
              var className = this.createIconClass(data);
              /* <DEPRECATED> */
              if (typeof data.item.icon === 'string') {
                  return h.div({ className: className }, data.item.iconLabel);
              }
              /* </DEPRECATED> */
              // if data.item.icon is undefined, it will be ignored
              return h.div({ className: className }, data.item.icon, data.item.iconLabel);
          };
          /**
           * Render the label element for a menu item.
           *
           * @param data - The data to use for rendering the label.
           *
           * @returns A virtual element representing the item label.
           */
          Renderer.prototype.renderLabel = function (data) {
              var content = this.formatLabel(data);
              return h.div({
                  className: 'lm-Menu-itemLabel'
                      /* <DEPRECATED> */
                      + ' p-Menu-itemLabel'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Render the shortcut element for a menu item.
           *
           * @param data - The data to use for rendering the shortcut.
           *
           * @returns A virtual element representing the item shortcut.
           */
          Renderer.prototype.renderShortcut = function (data) {
              var content = this.formatShortcut(data);
              return h.div({
                  className: 'lm-Menu-itemShortcut'
                      /* <DEPRECATED> */
                      + ' p-Menu-itemShortcut'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Render the submenu icon element for a menu item.
           *
           * @param data - The data to use for rendering the submenu icon.
           *
           * @returns A virtual element representing the submenu icon.
           */
          Renderer.prototype.renderSubmenu = function (data) {
              return h.div({
                  className: 'lm-Menu-itemSubmenuIcon'
                      /* <DEPRECATED> */
                      + ' p-Menu-itemSubmenuIcon'
                  /* </DEPRECATED> */
              });
          };
          /**
           * Create the class name for the menu item.
           *
           * @param data - The data to use for the class name.
           *
           * @returns The full class name for the menu item.
           */
          Renderer.prototype.createItemClass = function (data) {
              // Setup the initial class name.
              var name = 'lm-Menu-item';
              /* <DEPRECATED> */
              name += ' p-Menu-item';
              /* </DEPRECATED> */
              // Add the boolean state classes.
              if (!data.item.isEnabled) {
                  name += ' lm-mod-disabled';
                  /* <DEPRECATED> */
                  name += ' p-mod-disabled';
                  /* </DEPRECATED> */
              }
              if (data.item.isToggled) {
                  name += ' lm-mod-toggled';
                  /* <DEPRECATED> */
                  name += ' p-mod-toggled';
                  /* </DEPRECATED> */
              }
              if (!data.item.isVisible) {
                  name += ' lm-mod-hidden';
                  /* <DEPRECATED> */
                  name += ' p-mod-hidden';
                  /* </DEPRECATED> */
              }
              if (data.active) {
                  name += ' lm-mod-active';
                  /* <DEPRECATED> */
                  name += ' p-mod-active';
                  /* </DEPRECATED> */
              }
              if (data.collapsed) {
                  name += ' lm-mod-collapsed';
                  /* <DEPRECATED> */
                  name += ' p-mod-collapsed';
                  /* </DEPRECATED> */
              }
              // Add the extra class.
              var extra = data.item.className;
              if (extra) {
                  name += " " + extra;
              }
              // Return the complete class name.
              return name;
          };
          /**
           * Create the dataset for the menu item.
           *
           * @param data - The data to use for creating the dataset.
           *
           * @returns The dataset for the menu item.
           */
          Renderer.prototype.createItemDataset = function (data) {
              var result;
              var _a = data.item, type = _a.type, command = _a.command, dataset = _a.dataset;
              if (type === 'command') {
                  result = __assign$3(__assign$3({}, dataset), { type: type, command: command });
              }
              else {
                  result = __assign$3(__assign$3({}, dataset), { type: type });
              }
              return result;
          };
          /**
           * Create the class name for the menu item icon.
           *
           * @param data - The data to use for the class name.
           *
           * @returns The full class name for the item icon.
           */
          Renderer.prototype.createIconClass = function (data) {
              var name = 'lm-Menu-itemIcon';
              /* <DEPRECATED> */
              name += ' p-Menu-itemIcon';
              /* </DEPRECATED> */
              var extra = data.item.iconClass;
              return extra ? name + " " + extra : name;
          };
          /**
           * Create the render content for the label node.
           *
           * @param data - The data to use for the label content.
           *
           * @returns The content to add to the label node.
           */
          Renderer.prototype.formatLabel = function (data) {
              // Fetch the label text and mnemonic index.
              var _a = data.item, label = _a.label, mnemonic = _a.mnemonic;
              // If the index is out of range, do not modify the label.
              if (mnemonic < 0 || mnemonic >= label.length) {
                  return label;
              }
              // Split the label into parts.
              var prefix = label.slice(0, mnemonic);
              var suffix = label.slice(mnemonic + 1);
              var char = label[mnemonic];
              // Wrap the mnemonic character in a span.
              var span = h.span({
                  className: 'lm-Menu-itemMnemonic'
                      /* <DEPRECATED> */
                      + ' p-Menu-itemMnemonic'
                  /* </DEPRECATED> */
              }, char);
              // Return the content parts.
              return [prefix, span, suffix];
          };
          /**
           * Create the render content for the shortcut node.
           *
           * @param data - The data to use for the shortcut content.
           *
           * @returns The content to add to the shortcut node.
           */
          Renderer.prototype.formatShortcut = function (data) {
              var kb = data.item.keyBinding;
              return kb ? kb.keys.map(CommandRegistry.formatKeystroke).join(', ') : null;
          };
          return Renderer;
      }());
      Menu.Renderer = Renderer;
      /**
       * The default `Renderer` instance.
       */
      Menu.defaultRenderer = new Renderer();
  })(Menu || (Menu = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$6$1;
  (function (Private) {
      /**
       * The ms delay for opening and closing a submenu.
       */
      Private.TIMER_DELAY = 300;
      /**
       * The horizontal pixel overlap for an open submenu.
       */
      Private.SUBMENU_OVERLAP = 3;
      /**
       * Create the DOM node for a menu.
       */
      function createNode() {
          var node = document.createElement('div');
          var content = document.createElement('ul');
          content.className = 'lm-Menu-content';
          /* <DEPRECATED> */
          content.classList.add('p-Menu-content');
          /* </DEPRECATED> */
          node.appendChild(content);
          node.tabIndex = -1;
          return node;
      }
      Private.createNode = createNode;
      /**
       * Test whether a menu item can be activated.
       */
      function canActivate(item) {
          return item.type !== 'separator' && item.isEnabled && item.isVisible;
      }
      Private.canActivate = canActivate;
      /**
       * Create a new menu item for an owner menu.
       */
      function createItem(owner, options) {
          return new MenuItem(owner.commands, options);
      }
      Private.createItem = createItem;
      /**
       * Hit test a menu hierarchy starting at the given root.
       */
      function hitTestMenus(menu, x, y) {
          for (var temp = menu; temp; temp = temp.childMenu) {
              if (ElementExt.hitTest(temp.node, x, y)) {
                  return true;
              }
          }
          return false;
      }
      Private.hitTestMenus = hitTestMenus;
      /**
       * Compute which extra separator items should be collapsed.
       */
      function computeCollapsed(items) {
          // Allocate the return array and fill it with `false`.
          var result = new Array(items.length);
          ArrayExt.fill(result, false);
          // Collapse the leading separators.
          var k1 = 0;
          var n = items.length;
          for (; k1 < n; ++k1) {
              var item = items[k1];
              if (!item.isVisible) {
                  continue;
              }
              if (item.type !== 'separator') {
                  break;
              }
              result[k1] = true;
          }
          // Hide the trailing separators.
          var k2 = n - 1;
          for (; k2 >= 0; --k2) {
              var item = items[k2];
              if (!item.isVisible) {
                  continue;
              }
              if (item.type !== 'separator') {
                  break;
              }
              result[k2] = true;
          }
          // Hide the remaining consecutive separators.
          var hide = false;
          while (++k1 < k2) {
              var item = items[k1];
              if (!item.isVisible) {
                  continue;
              }
              if (item.type !== 'separator') {
                  hide = false;
              }
              else if (hide) {
                  result[k1] = true;
              }
              else {
                  hide = true;
              }
          }
          // Return the resulting flags.
          return result;
      }
      Private.computeCollapsed = computeCollapsed;
      /**
       * Open a menu as a root menu at the target location.
       */
      function openRootMenu(menu, x, y, forceX, forceY) {
          // Ensure the menu is updated before attaching and measuring.
          MessageLoop.sendMessage(menu, Widget.Msg.UpdateRequest);
          // Get the current position and size of the main viewport.
          var px = window.pageXOffset;
          var py = window.pageYOffset;
          var cw = document.documentElement.clientWidth;
          var ch = document.documentElement.clientHeight;
          // Compute the maximum allowed height for the menu.
          var maxHeight = ch - (forceY ? y : 0);
          // Fetch common variables.
          var node = menu.node;
          var style = node.style;
          // Clear the menu geometry and prepare it for measuring.
          style.top = '';
          style.left = '';
          style.width = '';
          style.height = '';
          style.visibility = 'hidden';
          style.maxHeight = maxHeight + "px";
          // Attach the menu to the document.
          Widget.attach(menu, document.body);
          // Measure the size of the menu.
          var _a = node.getBoundingClientRect(), width = _a.width, height = _a.height;
          // Adjust the X position of the menu to fit on-screen.
          if (!forceX && (x + width > px + cw)) {
              x = px + cw - width;
          }
          // Adjust the Y position of the menu to fit on-screen.
          if (!forceY && (y + height > py + ch)) {
              if (y > py + ch) {
                  y = py + ch - height;
              }
              else {
                  y = y - height;
              }
          }
          // Update the position of the menu to the computed position.
          style.top = Math.max(0, y) + "px";
          style.left = Math.max(0, x) + "px";
          // Finally, make the menu visible on the screen.
          style.visibility = '';
      }
      Private.openRootMenu = openRootMenu;
      /**
       * Open a menu as a submenu using an item node for positioning.
       */
      function openSubmenu(submenu, itemNode) {
          // Ensure the menu is updated before opening.
          MessageLoop.sendMessage(submenu, Widget.Msg.UpdateRequest);
          // Get the current position and size of the main viewport.
          var px = window.pageXOffset;
          var py = window.pageYOffset;
          var cw = document.documentElement.clientWidth;
          var ch = document.documentElement.clientHeight;
          // Compute the maximum allowed height for the menu.
          var maxHeight = ch;
          // Fetch common variables.
          var node = submenu.node;
          var style = node.style;
          // Clear the menu geometry and prepare it for measuring.
          style.top = '';
          style.left = '';
          style.width = '';
          style.height = '';
          style.visibility = 'hidden';
          style.maxHeight = maxHeight + "px";
          // Attach the menu to the document.
          Widget.attach(submenu, document.body);
          // Measure the size of the menu.
          var _a = node.getBoundingClientRect(), width = _a.width, height = _a.height;
          // Compute the box sizing for the menu.
          var box = ElementExt.boxSizing(submenu.node);
          // Get the bounding rect for the target item node.
          var itemRect = itemNode.getBoundingClientRect();
          // Compute the target X position.
          var x = itemRect.right - Private.SUBMENU_OVERLAP;
          // Adjust the X position to fit on the screen.
          if (x + width > px + cw) {
              x = itemRect.left + Private.SUBMENU_OVERLAP - width;
          }
          // Compute the target Y position.
          var y = itemRect.top - box.borderTop - box.paddingTop;
          // Adjust the Y position to fit on the screen.
          if (y + height > py + ch) {
              y = itemRect.bottom + box.borderBottom + box.paddingBottom - height;
          }
          // Update the position of the menu to the computed position.
          style.top = Math.max(0, y) + "px";
          style.left = Math.max(0, x) + "px";
          // Finally, make the menu visible on the screen.
          style.visibility = '';
      }
      Private.openSubmenu = openSubmenu;
      /**
       * Find the best matching mnemonic item.
       *
       * The search starts at the given index and wraps around.
       */
      function findMnemonic(items, key, start) {
          // Setup the result variables.
          var index = -1;
          var auto = -1;
          var multiple = false;
          // Normalize the key to upper case.
          var upperKey = key.toUpperCase();
          // Search the items from the given start index.
          for (var i = 0, n = items.length; i < n; ++i) {
              // Compute the wrapped index.
              var k = (i + start) % n;
              // Lookup the item
              var item = items[k];
              // Ignore items which cannot be activated.
              if (!canActivate(item)) {
                  continue;
              }
              // Ignore items with an empty label.
              var label = item.label;
              if (label.length === 0) {
                  continue;
              }
              // Lookup the mnemonic index for the label.
              var mn = item.mnemonic;
              // Handle a valid mnemonic index.
              if (mn >= 0 && mn < label.length) {
                  if (label[mn].toUpperCase() === upperKey) {
                      if (index === -1) {
                          index = k;
                      }
                      else {
                          multiple = true;
                      }
                  }
                  continue;
              }
              // Finally, handle the auto index if possible.
              if (auto === -1 && label[0].toUpperCase() === upperKey) {
                  auto = k;
              }
          }
          // Return the search results.
          return { index: index, multiple: multiple, auto: auto };
      }
      Private.findMnemonic = findMnemonic;
      /**
       * A concrete implementation of `Menu.IItem`.
       */
      var MenuItem = /** @class */ (function () {
          /**
           * Construct a new menu item.
           */
          function MenuItem(commands, options) {
              this._commands = commands;
              this.type = options.type || 'command';
              this.command = options.command || '';
              this.args = options.args || JSONExt.emptyObject;
              this.submenu = options.submenu || null;
          }
          Object.defineProperty(MenuItem.prototype, "label", {
              /**
               * The display label for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.label(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.label;
                  }
                  return '';
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "mnemonic", {
              /**
               * The mnemonic index for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.mnemonic(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.mnemonic;
                  }
                  return -1;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "icon", {
              /**
               * The icon renderer for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.icon(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.icon;
                  }
                  /* <DEPRECATED> */
                  // alias to icon class if not otherwise defined
                  return this.iconClass;
                  /* </DEPRECATED> */
                  /* <FUTURE>
                  return undefined;
                  </FUTURE> */
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "iconClass", {
              /**
               * The icon class for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.iconClass(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.iconClass;
                  }
                  return '';
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "iconLabel", {
              /**
               * The icon label for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.iconLabel(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.iconLabel;
                  }
                  return '';
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "caption", {
              /**
               * The display caption for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.caption(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.caption;
                  }
                  return '';
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "className", {
              /**
               * The extra class name for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.className(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.className;
                  }
                  return '';
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "dataset", {
              /**
               * The dataset for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.dataset(this.command, this.args);
                  }
                  if (this.type === 'submenu' && this.submenu) {
                      return this.submenu.title.dataset;
                  }
                  return {};
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "isEnabled", {
              /**
               * Whether the menu item is enabled.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.isEnabled(this.command, this.args);
                  }
                  if (this.type === 'submenu') {
                      return this.submenu !== null;
                  }
                  return true;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "isToggled", {
              /**
               * Whether the menu item is toggled.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.isToggled(this.command, this.args);
                  }
                  return false;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "isVisible", {
              /**
               * Whether the menu item is visible.
               */
              get: function () {
                  if (this.type === 'command') {
                      return this._commands.isVisible(this.command, this.args);
                  }
                  if (this.type === 'submenu') {
                      return this.submenu !== null;
                  }
                  return true;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(MenuItem.prototype, "keyBinding", {
              /**
               * The key binding for the menu item.
               */
              get: function () {
                  if (this.type === 'command') {
                      var _a = this, command_1 = _a.command, args_1 = _a.args;
                      return ArrayExt.findLastValue(this._commands.keyBindings, function (kb) {
                          return kb.command === command_1 && JSONExt.deepEqual(kb.args, args_1);
                      }) || null;
                  }
                  return null;
              },
              enumerable: true,
              configurable: true
          });
          return MenuItem;
      }());
  })(Private$6$1 || (Private$6$1 = {}));

  /**
   * An object which implements a universal context menu.
   *
   * #### Notes
   * The items shown in the context menu are determined by CSS selector
   * matching against the DOM hierarchy at the site of the mouse click.
   * This is similar in concept to how keyboard shortcuts are matched
   * in the command registry.
   */
  var ContextMenu = /** @class */ (function () {
      /**
       * Construct a new context menu.
       *
       * @param options - The options for initializing the menu.
       */
      function ContextMenu(options) {
          this._idTick = 0;
          this._items = [];
          this.menu = new Menu(options);
      }
      /**
       * Add an item to the context menu.
       *
       * @param options - The options for creating the item.
       *
       * @returns A disposable which will remove the item from the menu.
       */
      ContextMenu.prototype.addItem = function (options) {
          var _this = this;
          // Create an item from the given options.
          var item = Private$7$1.createItem(options, this._idTick++);
          // Add the item to the internal array.
          this._items.push(item);
          // Return a disposable which will remove the item.
          return new DisposableDelegate(function () {
              ArrayExt.removeFirstOf(_this._items, item);
          });
      };
      /**
       * Open the context menu in response to a `'contextmenu'` event.
       *
       * @param event - The `'contextmenu'` event of interest.
       *
       * @returns `true` if the menu was opened, or `false` if no items
       *   matched the event and the menu was not opened.
       *
       * #### Notes
       * This method will populate the context menu with items which match
       * the propagation path of the event, then open the menu at the mouse
       * position indicated by the event.
       */
      ContextMenu.prototype.open = function (event) {
          var _this = this;
          // Clear the current contents of the context menu.
          this.menu.clearItems();
          // Bail early if there are no items to match.
          if (this._items.length === 0) {
              return false;
          }
          // Find the matching items for the event.
          var items = Private$7$1.matchItems(this._items, event);
          // Bail if there are no matching items.
          if (!items || items.length === 0) {
              return false;
          }
          // Add the filtered items to the menu.
          each$1(items, function (item) { _this.menu.addItem(item); });
          // Open the context menu at the current mouse position.
          this.menu.open(event.clientX, event.clientY);
          // Indicate success.
          return true;
      };
      return ContextMenu;
  }());
  /**
   * The namespace for the module implementation details.
   */
  var Private$7$1;
  (function (Private) {
      /**
       * Create a normalized context menu item from an options object.
       */
      function createItem(options, id) {
          var selector = validateSelector(options.selector);
          var rank = options.rank !== undefined ? options.rank : Infinity;
          return __assign$3(__assign$3({}, options), { selector: selector, rank: rank, id: id });
      }
      Private.createItem = createItem;
      /**
       * Find the items which match a context menu event.
       *
       * The results are sorted by DOM level, specificity, and rank.
       */
      function matchItems(items, event) {
          // Look up the target of the event.
          var target = event.target;
          // Bail if there is no target.
          if (!target) {
              return null;
          }
          // Look up the current target of the event.
          var currentTarget = event.currentTarget;
          // Bail if there is no current target.
          if (!currentTarget) {
              return null;
          }
          // There are some third party libraries that cause the `target` to
          // be detached from the DOM before lumino can process the event.
          // If that happens, search for a new target node by point. If that
          // node is still dangling, bail.
          if (!currentTarget.contains(target)) {
              target = document.elementFromPoint(event.clientX, event.clientY);
              if (!target || !currentTarget.contains(target)) {
                  return null;
              }
          }
          // Set up the result array.
          var result = [];
          // Copy the items array to allow in-place modification.
          var availableItems = items.slice();
          // Walk up the DOM hierarchy searching for matches.
          while (target !== null) {
              // Set up the match array for this DOM level.
              var matches = [];
              // Search the remaining items for matches.
              for (var i = 0, n = availableItems.length; i < n; ++i) {
                  // Fetch the item.
                  var item = availableItems[i];
                  // Skip items which are already consumed.
                  if (!item) {
                      continue;
                  }
                  // Skip items which do not match the element.
                  if (!Selector.matches(target, item.selector)) {
                      continue;
                  }
                  // Add the matched item to the result for this DOM level.
                  matches.push(item);
                  // Mark the item as consumed.
                  availableItems[i] = null;
              }
              // Sort the matches for this level and add them to the results.
              if (matches.length !== 0) {
                  matches.sort(itemCmp);
                  result.push.apply(result, matches);
              }
              // Stop searching at the limits of the DOM range.
              if (target === currentTarget) {
                  break;
              }
              // Step to the parent DOM level.
              target = target.parentElement;
          }
          // Return the matched and sorted results.
          return result;
      }
      Private.matchItems = matchItems;
      /**
       * Validate the selector for a menu item.
       *
       * This returns the validated selector, or throws if the selector is
       * invalid or contains commas.
       */
      function validateSelector(selector) {
          if (selector.indexOf(',') !== -1) {
              throw new Error("Selector cannot contain commas: " + selector);
          }
          if (!Selector.isValid(selector)) {
              throw new Error("Invalid selector: " + selector);
          }
          return selector;
      }
      /**
       * A sort comparison function for a context menu item.
       */
      function itemCmp(a, b) {
          // Sort first based on selector specificity.
          var s1 = Selector.calculateSpecificity(a.selector);
          var s2 = Selector.calculateSpecificity(b.selector);
          if (s1 !== s2) {
              return s2 - s1;
          }
          // If specificities are equal, sort based on rank.
          var r1 = a.rank;
          var r2 = b.rank;
          if (r1 !== r2) {
              return r1 < r2 ? -1 : 1; // Infinity-safe
          }
          // When all else fails, sort by item id.
          return a.id - b.id;
      }
  })(Private$7$1 || (Private$7$1 = {}));

  /**
   * A layout which provides a flexible docking arrangement.
   *
   * #### Notes
   * The consumer of this layout is responsible for handling all signals
   * from the generated tab bars and managing the visibility of widgets
   * and tab bars as needed.
   */
  var DockLayout = /** @class */ (function (_super) {
      __extends$5(DockLayout, _super);
      /**
       * Construct a new dock layout.
       *
       * @param options - The options for initializing the layout.
       */
      function DockLayout(options) {
          var _this = _super.call(this) || this;
          _this._spacing = 4;
          _this._dirty = false;
          _this._root = null;
          _this._box = null;
          _this._items = new Map();
          _this.renderer = options.renderer;
          if (options.spacing !== undefined) {
              _this._spacing = Private$8$1.clampSpacing(options.spacing);
          }
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       *
       * #### Notes
       * This will clear and dispose all widgets in the layout.
       */
      DockLayout.prototype.dispose = function () {
          // Get an iterator over the widgets in the layout.
          var widgets = this.iter();
          // Dispose of the layout items.
          this._items.forEach(function (item) { item.dispose(); });
          // Clear the layout state before disposing the widgets.
          this._box = null;
          this._root = null;
          this._items.clear();
          // Dispose of the widgets contained in the old layout root.
          each$1(widgets, function (widget) { widget.dispose(); });
          // Dispose of the base class.
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(DockLayout.prototype, "spacing", {
          /**
           * Get the inter-element spacing for the dock layout.
           */
          get: function () {
              return this._spacing;
          },
          /**
           * Set the inter-element spacing for the dock layout.
           */
          set: function (value) {
              value = Private$8$1.clampSpacing(value);
              if (this._spacing === value) {
                  return;
              }
              this._spacing = value;
              if (!this.parent) {
                  return;
              }
              this.parent.fit();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(DockLayout.prototype, "isEmpty", {
          /**
           * Whether the dock layout is empty.
           */
          get: function () {
              return this._root === null;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over all widgets in the layout.
       *
       * @returns A new iterator over the widgets in the layout.
       *
       * #### Notes
       * This iterator includes the generated tab bars.
       */
      DockLayout.prototype.iter = function () {
          return this._root ? this._root.iterAllWidgets() : empty();
      };
      /**
       * Create an iterator over the user widgets in the layout.
       *
       * @returns A new iterator over the user widgets in the layout.
       *
       * #### Notes
       * This iterator does not include the generated tab bars.
       */
      DockLayout.prototype.widgets = function () {
          return this._root ? this._root.iterUserWidgets() : empty();
      };
      /**
       * Create an iterator over the selected widgets in the layout.
       *
       * @returns A new iterator over the selected user widgets.
       *
       * #### Notes
       * This iterator yields the widgets corresponding to the current tab
       * of each tab bar in the layout.
       */
      DockLayout.prototype.selectedWidgets = function () {
          return this._root ? this._root.iterSelectedWidgets() : empty();
      };
      /**
       * Create an iterator over the tab bars in the layout.
       *
       * @returns A new iterator over the tab bars in the layout.
       *
       * #### Notes
       * This iterator does not include the user widgets.
       */
      DockLayout.prototype.tabBars = function () {
          return this._root ? this._root.iterTabBars() : empty();
      };
      /**
       * Create an iterator over the handles in the layout.
       *
       * @returns A new iterator over the handles in the layout.
       */
      DockLayout.prototype.handles = function () {
          return this._root ? this._root.iterHandles() : empty();
      };
      /**
       * Move a handle to the given offset position.
       *
       * @param handle - The handle to move.
       *
       * @param offsetX - The desired offset X position of the handle.
       *
       * @param offsetY - The desired offset Y position of the handle.
       *
       * #### Notes
       * If the given handle is not contained in the layout, this is no-op.
       *
       * The handle will be moved as close as possible to the desired
       * position without violating any of the layout constraints.
       *
       * Only one of the coordinates is used depending on the orientation
       * of the handle. This method accepts both coordinates to make it
       * easy to invoke from a mouse move event without needing to know
       * the handle orientation.
       */
      DockLayout.prototype.moveHandle = function (handle, offsetX, offsetY) {
          // Bail early if there is no root or if the handle is hidden.
          var hidden = handle.classList.contains('lm-mod-hidden');
          /* <DEPRECATED> */
          hidden = hidden || handle.classList.contains('p-mod-hidden');
          /* </DEPRECATED> */
          if (!this._root || hidden) {
              return;
          }
          // Lookup the split node for the handle.
          var data = this._root.findSplitNode(handle);
          if (!data) {
              return;
          }
          // Compute the desired delta movement for the handle.
          var delta;
          if (data.node.orientation === 'horizontal') {
              delta = offsetX - handle.offsetLeft;
          }
          else {
              delta = offsetY - handle.offsetTop;
          }
          // Bail if there is no handle movement.
          if (delta === 0) {
              return;
          }
          // Prevent sibling resizing unless needed.
          data.node.holdSizes();
          // Adjust the sizers to reflect the handle movement.
          BoxEngine.adjust(data.node.sizers, data.index, delta);
          // Update the layout of the widgets.
          if (this.parent) {
              this.parent.update();
          }
      };
      /**
       * Save the current configuration of the dock layout.
       *
       * @returns A new config object for the current layout state.
       *
       * #### Notes
       * The return value can be provided to the `restoreLayout` method
       * in order to restore the layout to its current configuration.
       */
      DockLayout.prototype.saveLayout = function () {
          // Bail early if there is no root.
          if (!this._root) {
              return { main: null };
          }
          // Hold the current sizes in the layout tree.
          this._root.holdAllSizes();
          // Return the layout config.
          return { main: this._root.createConfig() };
      };
      /**
       * Restore the layout to a previously saved configuration.
       *
       * @param config - The layout configuration to restore.
       *
       * #### Notes
       * Widgets which currently belong to the layout but which are not
       * contained in the config will be unparented.
       */
      DockLayout.prototype.restoreLayout = function (config) {
          var _this = this;
          // Create the widget set for validating the config.
          var widgetSet = new Set();
          // Normalize the main area config and collect the widgets.
          var mainConfig;
          if (config.main) {
              mainConfig = Private$8$1.normalizeAreaConfig(config.main, widgetSet);
          }
          else {
              mainConfig = null;
          }
          // Create iterators over the old content.
          var oldWidgets = this.widgets();
          var oldTabBars = this.tabBars();
          var oldHandles = this.handles();
          // Clear the root before removing the old content.
          this._root = null;
          // Unparent the old widgets which are not in the new config.
          each$1(oldWidgets, function (widget) {
              if (!widgetSet.has(widget)) {
                  widget.parent = null;
              }
          });
          // Dispose of the old tab bars.
          each$1(oldTabBars, function (tabBar) {
              tabBar.dispose();
          });
          // Remove the old handles.
          each$1(oldHandles, function (handle) {
              if (handle.parentNode) {
                  handle.parentNode.removeChild(handle);
              }
          });
          // Reparent the new widgets to the current parent.
          widgetSet.forEach(function (widget) {
              widget.parent = _this.parent;
          });
          // Create the root node for the new config.
          if (mainConfig) {
              this._root = Private$8$1.realizeAreaConfig(mainConfig, {
                  createTabBar: function () { return _this._createTabBar(); },
                  createHandle: function () { return _this._createHandle(); }
              });
          }
          else {
              this._root = null;
          }
          // If there is no parent, there is nothing more to do.
          if (!this.parent) {
              return;
          }
          // Attach the new widgets to the parent.
          widgetSet.forEach(function (widget) {
              _this.attachWidget(widget);
          });
          // Post a fit request to the parent.
          this.parent.fit();
      };
      /**
       * Add a widget to the dock layout.
       *
       * @param widget - The widget to add to the dock layout.
       *
       * @param options - The additional options for adding the widget.
       *
       * #### Notes
       * The widget will be moved if it is already contained in the layout.
       *
       * An error will be thrown if the reference widget is invalid.
       */
      DockLayout.prototype.addWidget = function (widget, options) {
          if (options === void 0) { options = {}; }
          // Parse the options.
          var ref = options.ref || null;
          var mode = options.mode || 'tab-after';
          // Find the tab node which holds the reference widget.
          var refNode = null;
          if (this._root && ref) {
              refNode = this._root.findTabNode(ref);
          }
          // Throw an error if the reference widget is invalid.
          if (ref && !refNode) {
              throw new Error('Reference widget is not in the layout.');
          }
          // Reparent the widget to the current layout parent.
          widget.parent = this.parent;
          // Insert the widget according to the insert mode.
          switch (mode) {
              case 'tab-after':
                  this._insertTab(widget, ref, refNode, true);
                  break;
              case 'tab-before':
                  this._insertTab(widget, ref, refNode, false);
                  break;
              case 'split-top':
                  this._insertSplit(widget, ref, refNode, 'vertical', false);
                  break;
              case 'split-left':
                  this._insertSplit(widget, ref, refNode, 'horizontal', false);
                  break;
              case 'split-right':
                  this._insertSplit(widget, ref, refNode, 'horizontal', true);
                  break;
              case 'split-bottom':
                  this._insertSplit(widget, ref, refNode, 'vertical', true);
                  break;
          }
          // Do nothing else if there is no parent widget.
          if (!this.parent) {
              return;
          }
          // Ensure the widget is attached to the parent widget.
          this.attachWidget(widget);
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Remove a widget from the layout.
       *
       * @param widget - The widget to remove from the layout.
       *
       * #### Notes
       * A widget is automatically removed from the layout when its `parent`
       * is set to `null`. This method should only be invoked directly when
       * removing a widget from a layout which has yet to be installed on a
       * parent widget.
       *
       * This method does *not* modify the widget's `parent`.
       */
      DockLayout.prototype.removeWidget = function (widget) {
          // Remove the widget from its current layout location.
          this._removeWidget(widget);
          // Do nothing else if there is no parent widget.
          if (!this.parent) {
              return;
          }
          // Detach the widget from the parent widget.
          this.detachWidget(widget);
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Find the tab area which contains the given client position.
       *
       * @param clientX - The client X position of interest.
       *
       * @param clientY - The client Y position of interest.
       *
       * @returns The geometry of the tab area at the given position, or
       *   `null` if there is no tab area at the given position.
       */
      DockLayout.prototype.hitTestTabAreas = function (clientX, clientY) {
          // Bail early if hit testing cannot produce valid results.
          if (!this._root || !this.parent || !this.parent.isVisible) {
              return null;
          }
          // Ensure the parent box sizing data is computed.
          if (!this._box) {
              this._box = ElementExt.boxSizing(this.parent.node);
          }
          // Convert from client to local coordinates.
          var rect = this.parent.node.getBoundingClientRect();
          var x = clientX - rect.left - this._box.borderLeft;
          var y = clientY - rect.top - this._box.borderTop;
          // Find the tab layout node at the local position.
          var tabNode = this._root.hitTestTabNodes(x, y);
          // Bail if a tab layout node was not found.
          if (!tabNode) {
              return null;
          }
          // Extract the data from the tab node.
          var tabBar = tabNode.tabBar, top = tabNode.top, left = tabNode.left, width = tabNode.width, height = tabNode.height;
          // Compute the right and bottom edges of the tab area.
          var borderWidth = this._box.borderLeft + this._box.borderRight;
          var borderHeight = this._box.borderTop + this._box.borderBottom;
          var right = rect.width - borderWidth - (left + width);
          var bottom = rect.height - borderHeight - (top + height);
          // Return the hit test results.
          return { tabBar: tabBar, x: x, y: y, top: top, left: left, right: right, bottom: bottom, width: width, height: height };
      };
      /**
       * Perform layout initialization which requires the parent widget.
       */
      DockLayout.prototype.init = function () {
          var _this = this;
          // Perform superclass initialization.
          _super.prototype.init.call(this);
          // Attach each widget to the parent.
          each$1(this, function (widget) { _this.attachWidget(widget); });
          // Attach each handle to the parent.
          each$1(this.handles(), function (handle) { _this.parent.node.appendChild(handle); });
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Attach the widget to the layout parent widget.
       *
       * @param widget - The widget to attach to the parent.
       *
       * #### Notes
       * This is a no-op if the widget is already attached.
       */
      DockLayout.prototype.attachWidget = function (widget) {
          // Do nothing if the widget is already attached.
          if (this.parent.node === widget.node.parentNode) {
              return;
          }
          // Create the layout item for the widget.
          this._items.set(widget, new LayoutItem(widget));
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Add the widget's node to the parent.
          this.parent.node.appendChild(widget.node);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
      };
      /**
       * Detach the widget from the layout parent widget.
       *
       * @param widget - The widget to detach from the parent.
       *
       * #### Notes
       * This is a no-op if the widget is not attached.
       */
      DockLayout.prototype.detachWidget = function (widget) {
          // Do nothing if the widget is not attached.
          if (this.parent.node !== widget.node.parentNode) {
              return;
          }
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
          // Delete the layout item for the widget.
          var item = this._items.get(widget);
          if (item) {
              this._items.delete(widget);
              item.dispose();
          }
      };
      /**
       * A message handler invoked on a `'before-show'` message.
       */
      DockLayout.prototype.onBeforeShow = function (msg) {
          _super.prototype.onBeforeShow.call(this, msg);
          this.parent.update();
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      DockLayout.prototype.onBeforeAttach = function (msg) {
          _super.prototype.onBeforeAttach.call(this, msg);
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-shown'` message.
       */
      DockLayout.prototype.onChildShown = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-hidden'` message.
       */
      DockLayout.prototype.onChildHidden = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'resize'` message.
       */
      DockLayout.prototype.onResize = function (msg) {
          if (this.parent.isVisible) {
              this._update(msg.width, msg.height);
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      DockLayout.prototype.onUpdateRequest = function (msg) {
          if (this.parent.isVisible) {
              this._update(-1, -1);
          }
      };
      /**
       * A message handler invoked on a `'fit-request'` message.
       */
      DockLayout.prototype.onFitRequest = function (msg) {
          if (this.parent.isAttached) {
              this._fit();
          }
      };
      /**
       * Remove the specified widget from the layout structure.
       *
       * #### Notes
       * This is a no-op if the widget is not in the layout tree.
       *
       * This does not detach the widget from the parent node.
       */
      DockLayout.prototype._removeWidget = function (widget) {
          // Bail early if there is no layout root.
          if (!this._root) {
              return;
          }
          // Find the tab node which contains the given widget.
          var tabNode = this._root.findTabNode(widget);
          // Bail early if the tab node is not found.
          if (!tabNode) {
              return;
          }
          // If there are multiple tabs, just remove the widget's tab.
          if (tabNode.tabBar.titles.length > 1) {
              tabNode.tabBar.removeTab(widget.title);
              return;
          }
          // Otherwise, the tab node needs to be removed...
          // Dispose the tab bar.
          tabNode.tabBar.dispose();
          // Handle the case where the tab node is the root.
          if (this._root === tabNode) {
              this._root = null;
              return;
          }
          // Otherwise, remove the tab node from its parent...
          // Prevent widget resizing unless needed.
          this._root.holdAllSizes();
          // Clear the parent reference on the tab node.
          var splitNode = tabNode.parent;
          tabNode.parent = null;
          // Remove the tab node from its parent split node.
          var i = ArrayExt.removeFirstOf(splitNode.children, tabNode);
          var handle = ArrayExt.removeAt(splitNode.handles, i);
          ArrayExt.removeAt(splitNode.sizers, i);
          // Remove the handle from its parent DOM node.
          if (handle.parentNode) {
              handle.parentNode.removeChild(handle);
          }
          // If there are multiple children, just update the handles.
          if (splitNode.children.length > 1) {
              splitNode.syncHandles();
              return;
          }
          // Otherwise, the split node also needs to be removed...
          // Clear the parent reference on the split node.
          var maybeParent = splitNode.parent;
          splitNode.parent = null;
          // Lookup the remaining child node and handle.
          var childNode = splitNode.children[0];
          var childHandle = splitNode.handles[0];
          // Clear the split node data.
          splitNode.children.length = 0;
          splitNode.handles.length = 0;
          splitNode.sizers.length = 0;
          // Remove the child handle from its parent node.
          if (childHandle.parentNode) {
              childHandle.parentNode.removeChild(childHandle);
          }
          // Handle the case where the split node is the root.
          if (this._root === splitNode) {
              childNode.parent = null;
              this._root = childNode;
              return;
          }
          // Otherwise, move the child node to the parent node...
          var parentNode = maybeParent;
          // Lookup the index of the split node.
          var j = parentNode.children.indexOf(splitNode);
          // Handle the case where the child node is a tab node.
          if (childNode instanceof Private$8$1.TabLayoutNode) {
              childNode.parent = parentNode;
              parentNode.children[j] = childNode;
              return;
          }
          // Remove the split data from the parent.
          var splitHandle = ArrayExt.removeAt(parentNode.handles, j);
          ArrayExt.removeAt(parentNode.children, j);
          ArrayExt.removeAt(parentNode.sizers, j);
          // Remove the handle from its parent node.
          if (splitHandle.parentNode) {
              splitHandle.parentNode.removeChild(splitHandle);
          }
          // The child node and the split parent node will have the same
          // orientation. Merge the grand-children with the parent node.
          for (var i_1 = 0, n = childNode.children.length; i_1 < n; ++i_1) {
              var gChild = childNode.children[i_1];
              var gHandle = childNode.handles[i_1];
              var gSizer = childNode.sizers[i_1];
              ArrayExt.insert(parentNode.children, j + i_1, gChild);
              ArrayExt.insert(parentNode.handles, j + i_1, gHandle);
              ArrayExt.insert(parentNode.sizers, j + i_1, gSizer);
              gChild.parent = parentNode;
          }
          // Clear the child node.
          childNode.children.length = 0;
          childNode.handles.length = 0;
          childNode.sizers.length = 0;
          childNode.parent = null;
          // Sync the handles on the parent node.
          parentNode.syncHandles();
      };
      /**
       * Insert a widget next to an existing tab.
       *
       * #### Notes
       * This does not attach the widget to the parent widget.
       */
      DockLayout.prototype._insertTab = function (widget, ref, refNode, after) {
          // Do nothing if the tab is inserted next to itself.
          if (widget === ref) {
              return;
          }
          // Create the root if it does not exist.
          if (!this._root) {
              var tabNode = new Private$8$1.TabLayoutNode(this._createTabBar());
              tabNode.tabBar.addTab(widget.title);
              this._root = tabNode;
              return;
          }
          // Use the first tab node as the ref node if needed.
          if (!refNode) {
              refNode = this._root.findFirstTabNode();
          }
          // If the widget is not contained in the ref node, ensure it is
          // removed from the layout and hidden before being added again.
          if (refNode.tabBar.titles.indexOf(widget.title) === -1) {
              this._removeWidget(widget);
              widget.hide();
          }
          // Lookup the target index for inserting the tab.
          var index;
          if (ref) {
              index = refNode.tabBar.titles.indexOf(ref.title);
          }
          else {
              index = refNode.tabBar.currentIndex;
          }
          // Insert the widget's tab relative to the target index.
          refNode.tabBar.insertTab(index + (after ? 1 : 0), widget.title);
      };
      /**
       * Insert a widget as a new split area.
       *
       * #### Notes
       * This does not attach the widget to the parent widget.
       */
      DockLayout.prototype._insertSplit = function (widget, ref, refNode, orientation, after) {
          // Do nothing if there is no effective split.
          if (widget === ref && refNode && refNode.tabBar.titles.length === 1) {
              return;
          }
          // Ensure the widget is removed from the current layout.
          this._removeWidget(widget);
          // Create the tab layout node to hold the widget.
          var tabNode = new Private$8$1.TabLayoutNode(this._createTabBar());
          tabNode.tabBar.addTab(widget.title);
          // Set the root if it does not exist.
          if (!this._root) {
              this._root = tabNode;
              return;
          }
          // If the ref node parent is null, split the root.
          if (!refNode || !refNode.parent) {
              // Ensure the root is split with the correct orientation.
              var root = this._splitRoot(orientation);
              // Determine the insert index for the new tab node.
              var i_2 = after ? root.children.length : 0;
              // Normalize the split node.
              root.normalizeSizes();
              // Create the sizer for new tab node.
              var sizer = Private$8$1.createSizer(refNode ? 1 : Private$8$1.GOLDEN_RATIO);
              // Insert the tab node sized to the golden ratio.
              ArrayExt.insert(root.children, i_2, tabNode);
              ArrayExt.insert(root.sizers, i_2, sizer);
              ArrayExt.insert(root.handles, i_2, this._createHandle());
              tabNode.parent = root;
              // Re-normalize the split node to maintain the ratios.
              root.normalizeSizes();
              // Finally, synchronize the visibility of the handles.
              root.syncHandles();
              return;
          }
          // Lookup the split node for the ref widget.
          var splitNode = refNode.parent;
          // If the split node already had the correct orientation,
          // the widget can be inserted into the split node directly.
          if (splitNode.orientation === orientation) {
              // Find the index of the ref node.
              var i_3 = splitNode.children.indexOf(refNode);
              // Normalize the split node.
              splitNode.normalizeSizes();
              // Consume half the space for the insert location.
              var s = splitNode.sizers[i_3].sizeHint /= 2;
              // Insert the tab node sized to the other half.
              var j_1 = i_3 + (after ? 1 : 0);
              ArrayExt.insert(splitNode.children, j_1, tabNode);
              ArrayExt.insert(splitNode.sizers, j_1, Private$8$1.createSizer(s));
              ArrayExt.insert(splitNode.handles, j_1, this._createHandle());
              tabNode.parent = splitNode;
              // Finally, synchronize the visibility of the handles.
              splitNode.syncHandles();
              return;
          }
          // Remove the ref node from the split node.
          var i = ArrayExt.removeFirstOf(splitNode.children, refNode);
          // Create a new normalized split node for the children.
          var childNode = new Private$8$1.SplitLayoutNode(orientation);
          childNode.normalized = true;
          // Add the ref node sized to half the space.
          childNode.children.push(refNode);
          childNode.sizers.push(Private$8$1.createSizer(0.5));
          childNode.handles.push(this._createHandle());
          refNode.parent = childNode;
          // Add the tab node sized to the other half.
          var j = after ? 1 : 0;
          ArrayExt.insert(childNode.children, j, tabNode);
          ArrayExt.insert(childNode.sizers, j, Private$8$1.createSizer(0.5));
          ArrayExt.insert(childNode.handles, j, this._createHandle());
          tabNode.parent = childNode;
          // Synchronize the visibility of the handles.
          childNode.syncHandles();
          // Finally, add the new child node to the original split node.
          ArrayExt.insert(splitNode.children, i, childNode);
          childNode.parent = splitNode;
      };
      /**
       * Ensure the root is a split node with the given orientation.
       */
      DockLayout.prototype._splitRoot = function (orientation) {
          // Bail early if the root already meets the requirements.
          var oldRoot = this._root;
          if (oldRoot instanceof Private$8$1.SplitLayoutNode) {
              if (oldRoot.orientation === orientation) {
                  return oldRoot;
              }
          }
          // Create a new root node with the specified orientation.
          var newRoot = this._root = new Private$8$1.SplitLayoutNode(orientation);
          // Add the old root to the new root.
          if (oldRoot) {
              newRoot.children.push(oldRoot);
              newRoot.sizers.push(Private$8$1.createSizer(0));
              newRoot.handles.push(this._createHandle());
              oldRoot.parent = newRoot;
          }
          // Return the new root as a convenience.
          return newRoot;
      };
      /**
       * Fit the layout to the total size required by the widgets.
       */
      DockLayout.prototype._fit = function () {
          // Set up the computed minimum size.
          var minW = 0;
          var minH = 0;
          // Update the size limits for the layout tree.
          if (this._root) {
              var limits = this._root.fit(this._spacing, this._items);
              minW = limits.minWidth;
              minH = limits.minHeight;
          }
          // Update the box sizing and add it to the computed min size.
          var box = this._box = ElementExt.boxSizing(this.parent.node);
          minW += box.horizontalSum;
          minH += box.verticalSum;
          // Update the parent's min size constraints.
          var style = this.parent.node.style;
          style.minWidth = minW + "px";
          style.minHeight = minH + "px";
          // Set the dirty flag to ensure only a single update occurs.
          this._dirty = true;
          // Notify the ancestor that it should fit immediately. This may
          // cause a resize of the parent, fulfilling the required update.
          if (this.parent.parent) {
              MessageLoop.sendMessage(this.parent.parent, Widget.Msg.FitRequest);
          }
          // If the dirty flag is still set, the parent was not resized.
          // Trigger the required update on the parent widget immediately.
          if (this._dirty) {
              MessageLoop.sendMessage(this.parent, Widget.Msg.UpdateRequest);
          }
      };
      /**
       * Update the layout position and size of the widgets.
       *
       * The parent offset dimensions should be `-1` if unknown.
       */
      DockLayout.prototype._update = function (offsetWidth, offsetHeight) {
          // Clear the dirty flag to indicate the update occurred.
          this._dirty = false;
          // Bail early if there is no root layout node.
          if (!this._root) {
              return;
          }
          // Measure the parent if the offset dimensions are unknown.
          if (offsetWidth < 0) {
              offsetWidth = this.parent.node.offsetWidth;
          }
          if (offsetHeight < 0) {
              offsetHeight = this.parent.node.offsetHeight;
          }
          // Ensure the parent box sizing data is computed.
          if (!this._box) {
              this._box = ElementExt.boxSizing(this.parent.node);
          }
          // Compute the actual layout bounds adjusted for border and padding.
          var x = this._box.paddingTop;
          var y = this._box.paddingLeft;
          var width = offsetWidth - this._box.horizontalSum;
          var height = offsetHeight - this._box.verticalSum;
          // Update the geometry of the layout tree.
          this._root.update(x, y, width, height, this._spacing, this._items);
      };
      /**
       * Create a new tab bar for use by the dock layout.
       *
       * #### Notes
       * The tab bar will be attached to the parent if it exists.
       */
      DockLayout.prototype._createTabBar = function () {
          // Create the tab bar using the renderer.
          var tabBar = this.renderer.createTabBar();
          // Enforce necessary tab bar behavior.
          tabBar.orientation = 'horizontal';
          // Reparent and attach the tab bar to the parent if possible.
          if (this.parent) {
              tabBar.parent = this.parent;
              this.attachWidget(tabBar);
          }
          // Return the initialized tab bar.
          return tabBar;
      };
      /**
       * Create a new handle for the dock layout.
       *
       * #### Notes
       * The handle will be attached to the parent if it exists.
       */
      DockLayout.prototype._createHandle = function () {
          // Create the handle using the renderer.
          var handle = this.renderer.createHandle();
          // Initialize the handle layout behavior.
          var style = handle.style;
          style.position = 'absolute';
          style.top = '0';
          style.left = '0';
          style.width = '0';
          style.height = '0';
          // Attach the handle to the parent if it exists.
          if (this.parent) {
              this.parent.node.appendChild(handle);
          }
          // Return the initialized handle.
          return handle;
      };
      return DockLayout;
  }(Layout));
  /**
   * The namespace for the module implementation details.
   */
  var Private$8$1;
  (function (Private) {
      /**
       * A fraction used for sizing root panels; ~= `1 / golden_ratio`.
       */
      Private.GOLDEN_RATIO = 0.618;
      /**
       * Clamp a spacing value to an integer >= 0.
       */
      function clampSpacing(value) {
          return Math.max(0, Math.floor(value));
      }
      Private.clampSpacing = clampSpacing;
      /**
       * Create a box sizer with an initial size hint.
       */
      function createSizer(hint) {
          var sizer = new BoxSizer();
          sizer.sizeHint = hint;
          sizer.size = hint;
          return sizer;
      }
      Private.createSizer = createSizer;
      /**
       * Normalize an area config object and collect the visited widgets.
       */
      function normalizeAreaConfig(config, widgetSet) {
          var result;
          if (config.type === 'tab-area') {
              result = normalizeTabAreaConfig(config, widgetSet);
          }
          else {
              result = normalizeSplitAreaConfig(config, widgetSet);
          }
          return result;
      }
      Private.normalizeAreaConfig = normalizeAreaConfig;
      /**
       * Convert a normalized area config into a layout tree.
       */
      function realizeAreaConfig(config, renderer) {
          var node;
          if (config.type === 'tab-area') {
              node = realizeTabAreaConfig(config, renderer);
          }
          else {
              node = realizeSplitAreaConfig(config, renderer);
          }
          return node;
      }
      Private.realizeAreaConfig = realizeAreaConfig;
      /**
       * A layout node which holds the data for a tabbed area.
       */
      var TabLayoutNode = /** @class */ (function () {
          /**
           * Construct a new tab layout node.
           *
           * @param tabBar - The tab bar to use for the layout node.
           */
          function TabLayoutNode(tabBar) {
              /**
               * The parent of the layout node.
               */
              this.parent = null;
              this._top = 0;
              this._left = 0;
              this._width = 0;
              this._height = 0;
              var tabSizer = new BoxSizer();
              var widgetSizer = new BoxSizer();
              tabSizer.stretch = 0;
              widgetSizer.stretch = 1;
              this.tabBar = tabBar;
              this.sizers = [tabSizer, widgetSizer];
          }
          Object.defineProperty(TabLayoutNode.prototype, "top", {
              /**
               * The most recent value for the `top` edge of the layout box.
               */
              get: function () {
                  return this._top;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(TabLayoutNode.prototype, "left", {
              /**
               * The most recent value for the `left` edge of the layout box.
               */
              get: function () {
                  return this._left;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(TabLayoutNode.prototype, "width", {
              /**
               * The most recent value for the `width` of the layout box.
               */
              get: function () {
                  return this._width;
              },
              enumerable: true,
              configurable: true
          });
          Object.defineProperty(TabLayoutNode.prototype, "height", {
              /**
               * The most recent value for the `height` of the layout box.
               */
              get: function () {
                  return this._height;
              },
              enumerable: true,
              configurable: true
          });
          /**
           * Create an iterator for all widgets in the layout tree.
           */
          TabLayoutNode.prototype.iterAllWidgets = function () {
              return chain$1(once$1(this.tabBar), this.iterUserWidgets());
          };
          /**
           * Create an iterator for the user widgets in the layout tree.
           */
          TabLayoutNode.prototype.iterUserWidgets = function () {
              return map$1(this.tabBar.titles, function (title) { return title.owner; });
          };
          /**
           * Create an iterator for the selected widgets in the layout tree.
           */
          TabLayoutNode.prototype.iterSelectedWidgets = function () {
              var title = this.tabBar.currentTitle;
              return title ? once$1(title.owner) : empty();
          };
          /**
           * Create an iterator for the tab bars in the layout tree.
           */
          TabLayoutNode.prototype.iterTabBars = function () {
              return once$1(this.tabBar);
          };
          /**
           * Create an iterator for the handles in the layout tree.
           */
          TabLayoutNode.prototype.iterHandles = function () {
              return empty();
          };
          /**
           * Find the tab layout node which contains the given widget.
           */
          TabLayoutNode.prototype.findTabNode = function (widget) {
              return this.tabBar.titles.indexOf(widget.title) !== -1 ? this : null;
          };
          /**
           * Find the split layout node which contains the given handle.
           */
          TabLayoutNode.prototype.findSplitNode = function (handle) {
              return null;
          };
          /**
           * Find the first tab layout node in a layout tree.
           */
          TabLayoutNode.prototype.findFirstTabNode = function () {
              return this;
          };
          /**
           * Find the tab layout node which contains the local point.
           */
          TabLayoutNode.prototype.hitTestTabNodes = function (x, y) {
              if (x < this._left || x >= this._left + this._width) {
                  return null;
              }
              if (y < this._top || y >= this._top + this._height) {
                  return null;
              }
              return this;
          };
          /**
           * Create a configuration object for the layout tree.
           */
          TabLayoutNode.prototype.createConfig = function () {
              var widgets = this.tabBar.titles.map(function (title) { return title.owner; });
              var currentIndex = this.tabBar.currentIndex;
              return { type: 'tab-area', widgets: widgets, currentIndex: currentIndex };
          };
          /**
           * Recursively hold all of the sizes in the layout tree.
           *
           * This ignores the sizers of tab layout nodes.
           */
          TabLayoutNode.prototype.holdAllSizes = function () {
              return;
          };
          /**
           * Fit the layout tree.
           */
          TabLayoutNode.prototype.fit = function (spacing, items) {
              // Set up the limit variables.
              var minWidth = 0;
              var minHeight = 0;
              var maxWidth = Infinity;
              var maxHeight = Infinity;
              // Lookup the tab bar layout item.
              var tabBarItem = items.get(this.tabBar);
              // Lookup the widget layout item.
              var current = this.tabBar.currentTitle;
              var widgetItem = current ? items.get(current.owner) : undefined;
              // Lookup the tab bar and widget sizers.
              var _a = this.sizers, tabBarSizer = _a[0], widgetSizer = _a[1];
              // Update the tab bar limits.
              if (tabBarItem) {
                  tabBarItem.fit();
              }
              // Update the widget limits.
              if (widgetItem) {
                  widgetItem.fit();
              }
              // Update the results and sizer for the tab bar.
              if (tabBarItem && !tabBarItem.isHidden) {
                  minWidth = Math.max(minWidth, tabBarItem.minWidth);
                  minHeight += tabBarItem.minHeight;
                  tabBarSizer.minSize = tabBarItem.minHeight;
                  tabBarSizer.maxSize = tabBarItem.maxHeight;
              }
              else {
                  tabBarSizer.minSize = 0;
                  tabBarSizer.maxSize = 0;
              }
              // Update the results and sizer for the current widget.
              if (widgetItem && !widgetItem.isHidden) {
                  minWidth = Math.max(minWidth, widgetItem.minWidth);
                  minHeight += widgetItem.minHeight;
                  widgetSizer.minSize = widgetItem.minHeight;
                  widgetSizer.maxSize = Infinity;
              }
              else {
                  widgetSizer.minSize = 0;
                  widgetSizer.maxSize = Infinity;
              }
              // Return the computed size limits for the layout node.
              return { minWidth: minWidth, minHeight: minHeight, maxWidth: maxWidth, maxHeight: maxHeight };
          };
          /**
           * Update the layout tree.
           */
          TabLayoutNode.prototype.update = function (left, top, width, height, spacing, items) {
              // Update the layout box values.
              this._top = top;
              this._left = left;
              this._width = width;
              this._height = height;
              // Lookup the tab bar layout item.
              var tabBarItem = items.get(this.tabBar);
              // Lookup the widget layout item.
              var current = this.tabBar.currentTitle;
              var widgetItem = current ? items.get(current.owner) : undefined;
              // Distribute the layout space to the sizers.
              BoxEngine.calc(this.sizers, height);
              // Update the tab bar item using the computed size.
              if (tabBarItem && !tabBarItem.isHidden) {
                  var size = this.sizers[0].size;
                  tabBarItem.update(left, top, width, size);
                  top += size;
              }
              // Layout the widget using the computed size.
              if (widgetItem && !widgetItem.isHidden) {
                  var size = this.sizers[1].size;
                  widgetItem.update(left, top, width, size);
              }
          };
          return TabLayoutNode;
      }());
      Private.TabLayoutNode = TabLayoutNode;
      /**
       * A layout node which holds the data for a split area.
       */
      var SplitLayoutNode = /** @class */ (function () {
          /**
           * Construct a new split layout node.
           *
           * @param orientation - The orientation of the node.
           */
          function SplitLayoutNode(orientation) {
              /**
               * The parent of the layout node.
               */
              this.parent = null;
              /**
               * Whether the sizers have been normalized.
               */
              this.normalized = false;
              /**
               * The child nodes for the split node.
               */
              this.children = [];
              /**
               * The box sizers for the layout children.
               */
              this.sizers = [];
              /**
               * The handles for the layout children.
               */
              this.handles = [];
              this.orientation = orientation;
          }
          /**
           * Create an iterator for all widgets in the layout tree.
           */
          SplitLayoutNode.prototype.iterAllWidgets = function () {
              var children = map$1(this.children, function (child) { return child.iterAllWidgets(); });
              return new ChainIterator(children);
          };
          /**
           * Create an iterator for the user widgets in the layout tree.
           */
          SplitLayoutNode.prototype.iterUserWidgets = function () {
              var children = map$1(this.children, function (child) { return child.iterUserWidgets(); });
              return new ChainIterator(children);
          };
          /**
           * Create an iterator for the selected widgets in the layout tree.
           */
          SplitLayoutNode.prototype.iterSelectedWidgets = function () {
              var children = map$1(this.children, function (child) { return child.iterSelectedWidgets(); });
              return new ChainIterator(children);
          };
          /**
           * Create an iterator for the tab bars in the layout tree.
           */
          SplitLayoutNode.prototype.iterTabBars = function () {
              var children = map$1(this.children, function (child) { return child.iterTabBars(); });
              return new ChainIterator(children);
          };
          /**
           * Create an iterator for the handles in the layout tree.
           */
          SplitLayoutNode.prototype.iterHandles = function () {
              var children = map$1(this.children, function (child) { return child.iterHandles(); });
              return chain$1(this.handles, new ChainIterator(children));
          };
          /**
           * Find the tab layout node which contains the given widget.
           */
          SplitLayoutNode.prototype.findTabNode = function (widget) {
              for (var i = 0, n = this.children.length; i < n; ++i) {
                  var result = this.children[i].findTabNode(widget);
                  if (result) {
                      return result;
                  }
              }
              return null;
          };
          /**
           * Find the split layout node which contains the given handle.
           */
          SplitLayoutNode.prototype.findSplitNode = function (handle) {
              var index = this.handles.indexOf(handle);
              if (index !== -1) {
                  return { index: index, node: this };
              }
              for (var i = 0, n = this.children.length; i < n; ++i) {
                  var result = this.children[i].findSplitNode(handle);
                  if (result) {
                      return result;
                  }
              }
              return null;
          };
          /**
           * Find the first tab layout node in a layout tree.
           */
          SplitLayoutNode.prototype.findFirstTabNode = function () {
              if (this.children.length === 0) {
                  return null;
              }
              return this.children[0].findFirstTabNode();
          };
          /**
           * Find the tab layout node which contains the local point.
           */
          SplitLayoutNode.prototype.hitTestTabNodes = function (x, y) {
              for (var i = 0, n = this.children.length; i < n; ++i) {
                  var result = this.children[i].hitTestTabNodes(x, y);
                  if (result) {
                      return result;
                  }
              }
              return null;
          };
          /**
           * Create a configuration object for the layout tree.
           */
          SplitLayoutNode.prototype.createConfig = function () {
              var orientation = this.orientation;
              var sizes = this.createNormalizedSizes();
              var children = this.children.map(function (child) { return child.createConfig(); });
              return { type: 'split-area', orientation: orientation, children: children, sizes: sizes };
          };
          /**
           * Sync the visibility and orientation of the handles.
           */
          SplitLayoutNode.prototype.syncHandles = function () {
              var _this = this;
              each$1(this.handles, function (handle, i) {
                  handle.setAttribute('data-orientation', _this.orientation);
                  if (i === _this.handles.length - 1) {
                      handle.classList.add('lm-mod-hidden');
                      /* <DEPRECATED> */
                      handle.classList.add('p-mod-hidden');
                      /* </DEPRECATED> */
                  }
                  else {
                      handle.classList.remove('lm-mod-hidden');
                      /* <DEPRECATED> */
                      handle.classList.remove('p-mod-hidden');
                      /* </DEPRECATED> */
                  }
              });
          };
          /**
           * Hold the current sizes of the box sizers.
           *
           * This sets the size hint of each sizer to its current size.
           */
          SplitLayoutNode.prototype.holdSizes = function () {
              each$1(this.sizers, function (sizer) { sizer.sizeHint = sizer.size; });
          };
          /**
           * Recursively hold all of the sizes in the layout tree.
           *
           * This ignores the sizers of tab layout nodes.
           */
          SplitLayoutNode.prototype.holdAllSizes = function () {
              each$1(this.children, function (child) { return child.holdAllSizes(); });
              this.holdSizes();
          };
          /**
           * Normalize the sizes of the split layout node.
           */
          SplitLayoutNode.prototype.normalizeSizes = function () {
              // Bail early if the sizers are empty.
              var n = this.sizers.length;
              if (n === 0) {
                  return;
              }
              // Hold the current sizes of the sizers.
              this.holdSizes();
              // Compute the sum of the sizes.
              var sum = reduce$1(this.sizers, function (v, sizer) { return v + sizer.sizeHint; }, 0);
              // Normalize the sizes based on the sum.
              if (sum === 0) {
                  each$1(this.sizers, function (sizer) {
                      sizer.size = sizer.sizeHint = 1 / n;
                  });
              }
              else {
                  each$1(this.sizers, function (sizer) {
                      sizer.size = sizer.sizeHint /= sum;
                  });
              }
              // Mark the sizes as normalized.
              this.normalized = true;
          };
          /**
           * Snap the normalized sizes of the split layout node.
           */
          SplitLayoutNode.prototype.createNormalizedSizes = function () {
              // Bail early if the sizers are empty.
              var n = this.sizers.length;
              if (n === 0) {
                  return [];
              }
              // Grab the current sizes of the sizers.
              var sizes = this.sizers.map(function (sizer) { return sizer.size; });
              // Compute the sum of the sizes.
              var sum = reduce$1(sizes, function (v, size) { return v + size; }, 0);
              // Normalize the sizes based on the sum.
              if (sum === 0) {
                  each$1(sizes, function (size, i) { sizes[i] = 1 / n; });
              }
              else {
                  each$1(sizes, function (size, i) { sizes[i] = size / sum; });
              }
              // Return the normalized sizes.
              return sizes;
          };
          /**
           * Fit the layout tree.
           */
          SplitLayoutNode.prototype.fit = function (spacing, items) {
              // Compute the required fixed space.
              var horizontal = this.orientation === 'horizontal';
              var fixed = Math.max(0, this.children.length - 1) * spacing;
              // Set up the limit variables.
              var minWidth = horizontal ? fixed : 0;
              var minHeight = horizontal ? 0 : fixed;
              var maxWidth = Infinity;
              var maxHeight = Infinity;
              // Fit the children and update the limits.
              for (var i = 0, n = this.children.length; i < n; ++i) {
                  var limits = this.children[i].fit(spacing, items);
                  if (horizontal) {
                      minHeight = Math.max(minHeight, limits.minHeight);
                      minWidth += limits.minWidth;
                      this.sizers[i].minSize = limits.minWidth;
                  }
                  else {
                      minWidth = Math.max(minWidth, limits.minWidth);
                      minHeight += limits.minHeight;
                      this.sizers[i].minSize = limits.minHeight;
                  }
              }
              // Return the computed limits for the layout node.
              return { minWidth: minWidth, minHeight: minHeight, maxWidth: maxWidth, maxHeight: maxHeight };
          };
          /**
           * Update the layout tree.
           */
          SplitLayoutNode.prototype.update = function (left, top, width, height, spacing, items) {
              // Compute the available layout space.
              var horizontal = this.orientation === 'horizontal';
              var fixed = Math.max(0, this.children.length - 1) * spacing;
              var space = Math.max(0, (horizontal ? width : height) - fixed);
              // De-normalize the sizes if needed.
              if (this.normalized) {
                  each$1(this.sizers, function (sizer) { sizer.sizeHint *= space; });
                  this.normalized = false;
              }
              // Distribute the layout space to the sizers.
              BoxEngine.calc(this.sizers, space);
              // Update the geometry of the child nodes and handles.
              for (var i = 0, n = this.children.length; i < n; ++i) {
                  var child = this.children[i];
                  var size = this.sizers[i].size;
                  var handleStyle = this.handles[i].style;
                  if (horizontal) {
                      child.update(left, top, size, height, spacing, items);
                      left += size;
                      handleStyle.top = top + "px";
                      handleStyle.left = left + "px";
                      handleStyle.width = spacing + "px";
                      handleStyle.height = height + "px";
                      left += spacing;
                  }
                  else {
                      child.update(left, top, width, size, spacing, items);
                      top += size;
                      handleStyle.top = top + "px";
                      handleStyle.left = left + "px";
                      handleStyle.width = width + "px";
                      handleStyle.height = spacing + "px";
                      top += spacing;
                  }
              }
          };
          return SplitLayoutNode;
      }());
      Private.SplitLayoutNode = SplitLayoutNode;
      /**
       * Normalize a tab area config and collect the visited widgets.
       */
      function normalizeTabAreaConfig(config, widgetSet) {
          // Bail early if there is no content.
          if (config.widgets.length === 0) {
              return null;
          }
          // Setup the filtered widgets array.
          var widgets = [];
          // Filter the config for unique widgets.
          each$1(config.widgets, function (widget) {
              if (!widgetSet.has(widget)) {
                  widgetSet.add(widget);
                  widgets.push(widget);
              }
          });
          // Bail if there are no effective widgets.
          if (widgets.length === 0) {
              return null;
          }
          // Normalize the current index.
          var index = config.currentIndex;
          if (index !== -1 && (index < 0 || index >= widgets.length)) {
              index = 0;
          }
          // Return a normalized config object.
          return { type: 'tab-area', widgets: widgets, currentIndex: index };
      }
      /**
       * Normalize a split area config and collect the visited widgets.
       */
      function normalizeSplitAreaConfig(config, widgetSet) {
          // Set up the result variables.
          var orientation = config.orientation;
          var children = [];
          var sizes = [];
          // Normalize the config children.
          for (var i = 0, n = config.children.length; i < n; ++i) {
              // Normalize the child config.
              var child = normalizeAreaConfig(config.children[i], widgetSet);
              // Ignore an empty child.
              if (!child) {
                  continue;
              }
              // Add the child or hoist its content as appropriate.
              if (child.type === 'tab-area' || child.orientation !== orientation) {
                  children.push(child);
                  sizes.push(Math.abs(config.sizes[i] || 0));
              }
              else {
                  children.push.apply(children, child.children);
                  sizes.push.apply(sizes, child.sizes);
              }
          }
          // Bail if there are no effective children.
          if (children.length === 0) {
              return null;
          }
          // If there is only one effective child, return that child.
          if (children.length === 1) {
              return children[0];
          }
          // Return a normalized config object.
          return { type: 'split-area', orientation: orientation, children: children, sizes: sizes };
      }
      /**
       * Convert a normalized tab area config into a layout tree.
       */
      function realizeTabAreaConfig(config, renderer) {
          // Create the tab bar for the layout node.
          var tabBar = renderer.createTabBar();
          // Hide each widget and add it to the tab bar.
          each$1(config.widgets, function (widget) {
              widget.hide();
              tabBar.addTab(widget.title);
          });
          // Set the current index of the tab bar.
          tabBar.currentIndex = config.currentIndex;
          // Return the new tab layout node.
          return new TabLayoutNode(tabBar);
      }
      /**
       * Convert a normalized split area config into a layout tree.
       */
      function realizeSplitAreaConfig(config, renderer) {
          // Create the split layout node.
          var node = new SplitLayoutNode(config.orientation);
          // Add each child to the layout node.
          each$1(config.children, function (child, i) {
              // Create the child data for the layout node.
              var childNode = realizeAreaConfig(child, renderer);
              var sizer = createSizer(config.sizes[i]);
              var handle = renderer.createHandle();
              // Add the child data to the layout node.
              node.children.push(childNode);
              node.handles.push(handle);
              node.sizers.push(sizer);
              // Update the parent for the child node.
              childNode.parent = node;
          });
          // Synchronize the handle state for the layout node.
          node.syncHandles();
          // Normalize the sizes for the layout node.
          node.normalizeSizes();
          // Return the new layout node.
          return node;
      }
  })(Private$8$1 || (Private$8$1 = {}));

  /**
   * A widget which displays titles as a single row or column of tabs.
   *
   * #### Notes
   * If CSS transforms are used to rotate nodes for vertically oriented
   * text, then tab dragging will not work correctly. The `tabsMovable`
   * property should be set to `false` when rotating nodes from CSS.
   */
  var TabBar = /** @class */ (function (_super) {
      __extends$5(TabBar, _super);
      /**
       * Construct a new tab bar.
       *
       * @param options - The options for initializing the tab bar.
       */
      function TabBar(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, { node: Private$9$1.createNode() }) || this;
          _this._currentIndex = -1;
          _this._titles = [];
          _this._previousTitle = null;
          _this._dragData = null;
          _this._tabMoved = new Signal(_this);
          _this._currentChanged = new Signal(_this);
          _this._tabCloseRequested = new Signal(_this);
          _this._tabDetachRequested = new Signal(_this);
          _this._tabActivateRequested = new Signal(_this);
          _this.addClass('lm-TabBar');
          /* <DEPRECATED> */
          _this.addClass('p-TabBar');
          /* </DEPRECATED> */
          _this.setFlag(Widget.Flag.DisallowLayout);
          _this.tabsMovable = options.tabsMovable || false;
          _this.allowDeselect = options.allowDeselect || false;
          _this.insertBehavior = options.insertBehavior || 'select-tab-if-needed';
          _this.removeBehavior = options.removeBehavior || 'select-tab-after';
          _this.renderer = options.renderer || TabBar.defaultRenderer;
          _this._orientation = options.orientation || 'horizontal';
          _this.dataset['orientation'] = _this._orientation;
          return _this;
      }
      /**
       * Dispose of the resources held by the widget.
       */
      TabBar.prototype.dispose = function () {
          this._releaseMouse();
          this._titles.length = 0;
          this._previousTitle = null;
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(TabBar.prototype, "currentChanged", {
          /**
           * A signal emitted when the current tab is changed.
           *
           * #### Notes
           * This signal is emitted when the currently selected tab is changed
           * either through user or programmatic interaction.
           *
           * Notably, this signal is not emitted when the index of the current
           * tab changes due to tabs being inserted, removed, or moved. It is
           * only emitted when the actual current tab node is changed.
           */
          get: function () {
              return this._currentChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "tabMoved", {
          /**
           * A signal emitted when a tab is moved by the user.
           *
           * #### Notes
           * This signal is emitted when a tab is moved by user interaction.
           *
           * This signal is not emitted when a tab is moved programmatically.
           */
          get: function () {
              return this._tabMoved;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "tabActivateRequested", {
          /**
           * A signal emitted when a tab is clicked by the user.
           *
           * #### Notes
           * If the clicked tab is not the current tab, the clicked tab will be
           * made current and the `currentChanged` signal will be emitted first.
           *
           * This signal is emitted even if the clicked tab is the current tab.
           */
          get: function () {
              return this._tabActivateRequested;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "tabCloseRequested", {
          /**
           * A signal emitted when a tab close icon is clicked.
           *
           * #### Notes
           * This signal is not emitted unless the tab title is `closable`.
           */
          get: function () {
              return this._tabCloseRequested;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "tabDetachRequested", {
          /**
           * A signal emitted when a tab is dragged beyond the detach threshold.
           *
           * #### Notes
           * This signal is emitted when the user drags a tab with the mouse,
           * and mouse is dragged beyond the detach threshold.
           *
           * The consumer of the signal should call `releaseMouse` and remove
           * the tab in order to complete the detach.
           *
           * This signal is only emitted once per drag cycle.
           */
          get: function () {
              return this._tabDetachRequested;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "currentTitle", {
          /**
           * Get the currently selected title.
           *
           * #### Notes
           * This will be `null` if no tab is selected.
           */
          get: function () {
              return this._titles[this._currentIndex] || null;
          },
          /**
           * Set the currently selected title.
           *
           * #### Notes
           * If the title does not exist, the title will be set to `null`.
           */
          set: function (value) {
              this.currentIndex = value ? this._titles.indexOf(value) : -1;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "currentIndex", {
          /**
           * Get the index of the currently selected tab.
           *
           * #### Notes
           * This will be `-1` if no tab is selected.
           */
          get: function () {
              return this._currentIndex;
          },
          /**
           * Set the index of the currently selected tab.
           *
           * #### Notes
           * If the value is out of range, the index will be set to `-1`.
           */
          set: function (value) {
              // Adjust for an out of range index.
              if (value < 0 || value >= this._titles.length) {
                  value = -1;
              }
              // Bail early if the index will not change.
              if (this._currentIndex === value) {
                  return;
              }
              // Look up the previous index and title.
              var pi = this._currentIndex;
              var pt = this._titles[pi] || null;
              // Look up the current index and title.
              var ci = value;
              var ct = this._titles[ci] || null;
              // Update the current index and previous title.
              this._currentIndex = ci;
              this._previousTitle = pt;
              // Schedule an update of the tabs.
              this.update();
              // Emit the current changed signal.
              this._currentChanged.emit({
                  previousIndex: pi, previousTitle: pt,
                  currentIndex: ci, currentTitle: ct
              });
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "orientation", {
          /**
           * Get the orientation of the tab bar.
           *
           * #### Notes
           * This controls whether the tabs are arranged in a row or column.
           */
          get: function () {
              return this._orientation;
          },
          /**
           * Set the orientation of the tab bar.
           *
           * #### Notes
           * This controls whether the tabs are arranged in a row or column.
           */
          set: function (value) {
              // Do nothing if the orientation does not change.
              if (this._orientation === value) {
                  return;
              }
              // Release the mouse before making any changes.
              this._releaseMouse();
              // Toggle the orientation values.
              this._orientation = value;
              this.dataset['orientation'] = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "titles", {
          /**
           * A read-only array of the titles in the tab bar.
           */
          get: function () {
              return this._titles;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabBar.prototype, "contentNode", {
          /**
           * The tab bar content node.
           *
           * #### Notes
           * This is the node which holds the tab nodes.
           *
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-TabBar-content')[0];
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Add a tab to the end of the tab bar.
       *
       * @param value - The title which holds the data for the tab,
       *   or an options object to convert to a title.
       *
       * @returns The title object added to the tab bar.
       *
       * #### Notes
       * If the title is already added to the tab bar, it will be moved.
       */
      TabBar.prototype.addTab = function (value) {
          return this.insertTab(this._titles.length, value);
      };
      /**
       * Insert a tab into the tab bar at the specified index.
       *
       * @param index - The index at which to insert the tab.
       *
       * @param value - The title which holds the data for the tab,
       *   or an options object to convert to a title.
       *
       * @returns The title object added to the tab bar.
       *
       * #### Notes
       * The index will be clamped to the bounds of the tabs.
       *
       * If the title is already added to the tab bar, it will be moved.
       */
      TabBar.prototype.insertTab = function (index, value) {
          // Release the mouse before making any changes.
          this._releaseMouse();
          // Coerce the value to a title.
          var title = Private$9$1.asTitle(value);
          // Look up the index of the title.
          var i = this._titles.indexOf(title);
          // Clamp the insert index to the array bounds.
          var j = Math.max(0, Math.min(index, this._titles.length));
          // If the title is not in the array, insert it.
          if (i === -1) {
              // Insert the title into the array.
              ArrayExt.insert(this._titles, j, title);
              // Connect to the title changed signal.
              title.changed.connect(this._onTitleChanged, this);
              // Schedule an update of the tabs.
              this.update();
              // Adjust the current index for the insert.
              this._adjustCurrentForInsert(j, title);
              // Return the title added to the tab bar.
              return title;
          }
          // Otherwise, the title exists in the array and should be moved.
          // Adjust the index if the location is at the end of the array.
          if (j === this._titles.length) {
              j--;
          }
          // Bail if there is no effective move.
          if (i === j) {
              return title;
          }
          // Move the title to the new location.
          ArrayExt.move(this._titles, i, j);
          // Schedule an update of the tabs.
          this.update();
          // Adjust the current index for the move.
          this._adjustCurrentForMove(i, j);
          // Return the title added to the tab bar.
          return title;
      };
      /**
       * Remove a tab from the tab bar.
       *
       * @param title - The title for the tab to remove.
       *
       * #### Notes
       * This is a no-op if the title is not in the tab bar.
       */
      TabBar.prototype.removeTab = function (title) {
          this.removeTabAt(this._titles.indexOf(title));
      };
      /**
       * Remove the tab at a given index from the tab bar.
       *
       * @param index - The index of the tab to remove.
       *
       * #### Notes
       * This is a no-op if the index is out of range.
       */
      TabBar.prototype.removeTabAt = function (index) {
          // Release the mouse before making any changes.
          this._releaseMouse();
          // Remove the title from the array.
          var title = ArrayExt.removeAt(this._titles, index);
          // Bail if the index is out of range.
          if (!title) {
              return;
          }
          // Disconnect from the title changed signal.
          title.changed.disconnect(this._onTitleChanged, this);
          // Clear the previous title if it's being removed.
          if (title === this._previousTitle) {
              this._previousTitle = null;
          }
          // Schedule an update of the tabs.
          this.update();
          // Adjust the current index for the remove.
          this._adjustCurrentForRemove(index, title);
      };
      /**
       * Remove all tabs from the tab bar.
       */
      TabBar.prototype.clearTabs = function () {
          // Bail if there is nothing to remove.
          if (this._titles.length === 0) {
              return;
          }
          // Release the mouse before making any changes.
          this._releaseMouse();
          // Disconnect from the title changed signals.
          for (var _i = 0, _a = this._titles; _i < _a.length; _i++) {
              var title = _a[_i];
              title.changed.disconnect(this._onTitleChanged, this);
          }
          // Get the current index and title.
          var pi = this.currentIndex;
          var pt = this.currentTitle;
          // Reset the current index and previous title.
          this._currentIndex = -1;
          this._previousTitle = null;
          // Clear the title array.
          this._titles.length = 0;
          // Schedule an update of the tabs.
          this.update();
          // If no tab was selected, there's nothing else to do.
          if (pi === -1) {
              return;
          }
          // Emit the current changed signal.
          this._currentChanged.emit({
              previousIndex: pi, previousTitle: pt,
              currentIndex: -1, currentTitle: null
          });
      };
      /**
       * Release the mouse and restore the non-dragged tab positions.
       *
       * #### Notes
       * This will cause the tab bar to stop handling mouse events and to
       * restore the tabs to their non-dragged positions.
       */
      TabBar.prototype.releaseMouse = function () {
          this._releaseMouse();
      };
      /**
       * Handle the DOM events for the tab bar.
       *
       * @param event - The DOM event sent to the tab bar.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the tab bar's DOM node.
       *
       * This should not be called directly by user code.
       */
      TabBar.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'mousedown':
                  this._evtMouseDown(event);
                  break;
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseup':
                  this._evtMouseUp(event);
                  break;
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'contextmenu':
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      TabBar.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('mousedown', this);
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       */
      TabBar.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('mousedown', this);
          this._releaseMouse();
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      TabBar.prototype.onUpdateRequest = function (msg) {
          var titles = this._titles;
          var renderer = this.renderer;
          var currentTitle = this.currentTitle;
          var content = new Array(titles.length);
          for (var i = 0, n = titles.length; i < n; ++i) {
              var title = titles[i];
              var current = title === currentTitle;
              var zIndex = current ? n : n - i - 1;
              content[i] = renderer.renderTab({ title: title, current: current, zIndex: zIndex });
          }
          VirtualDOM.render(content, this.contentNode);
      };
      /**
       * Handle the `'keydown'` event for the tab bar.
       */
      TabBar.prototype._evtKeyDown = function (event) {
          // Stop all input events during drag.
          event.preventDefault();
          event.stopPropagation();
          // Release the mouse if `Escape` is pressed.
          if (event.keyCode === 27) {
              this._releaseMouse();
          }
      };
      /**
       * Handle the `'mousedown'` event for the tab bar.
       */
      TabBar.prototype._evtMouseDown = function (event) {
          // Do nothing if it's not a left or middle mouse press.
          if (event.button !== 0 && event.button !== 1) {
              return;
          }
          // Do nothing if a drag is in progress.
          if (this._dragData) {
              return;
          }
          // Lookup the tab nodes.
          var tabs = this.contentNode.children;
          // Find the index of the pressed tab.
          var index = ArrayExt.findFirstIndex(tabs, function (tab) {
              return ElementExt.hitTest(tab, event.clientX, event.clientY);
          });
          // Do nothing if the press is not on a tab.
          if (index === -1) {
              return;
          }
          // Pressing on a tab stops the event propagation.
          event.preventDefault();
          event.stopPropagation();
          // Initialize the non-measured parts of the drag data.
          this._dragData = {
              tab: tabs[index],
              index: index,
              pressX: event.clientX,
              pressY: event.clientY,
              tabPos: -1,
              tabSize: -1,
              tabPressPos: -1,
              targetIndex: -1,
              tabLayout: null,
              contentRect: null,
              override: null,
              dragActive: false,
              dragAborted: false,
              detachRequested: false
          };
          // Add the document mouse up listener.
          document.addEventListener('mouseup', this, true);
          // Do nothing else if the middle button is clicked.
          if (event.button === 1) {
              return;
          }
          // Do nothing else if the close icon is clicked.
          var icon = tabs[index].querySelector(this.renderer.closeIconSelector);
          if (icon && icon.contains(event.target)) {
              return;
          }
          // Add the extra listeners if the tabs are movable.
          if (this.tabsMovable) {
              document.addEventListener('mousemove', this, true);
              document.addEventListener('keydown', this, true);
              document.addEventListener('contextmenu', this, true);
          }
          // Update the current index as appropriate.
          if (this.allowDeselect && this.currentIndex === index) {
              this.currentIndex = -1;
          }
          else {
              this.currentIndex = index;
          }
          // Do nothing else if there is no current tab.
          if (this.currentIndex === -1) {
              return;
          }
          // Emit the tab activate request signal.
          this._tabActivateRequested.emit({
              index: this.currentIndex, title: this.currentTitle
          });
      };
      /**
       * Handle the `'mousemove'` event for the tab bar.
       */
      TabBar.prototype._evtMouseMove = function (event) {
          // Do nothing if no drag is in progress.
          var data = this._dragData;
          if (!data) {
              return;
          }
          // Suppress the event during a drag.
          event.preventDefault();
          event.stopPropagation();
          // Lookup the tab nodes.
          var tabs = this.contentNode.children;
          // Bail early if the drag threshold has not been met.
          if (!data.dragActive && !Private$9$1.dragExceeded(data, event)) {
              return;
          }
          // Activate the drag if necessary.
          if (!data.dragActive) {
              // Fill in the rest of the drag data measurements.
              var tabRect = data.tab.getBoundingClientRect();
              if (this._orientation === 'horizontal') {
                  data.tabPos = data.tab.offsetLeft;
                  data.tabSize = tabRect.width;
                  data.tabPressPos = data.pressX - tabRect.left;
              }
              else {
                  data.tabPos = data.tab.offsetTop;
                  data.tabSize = tabRect.height;
                  data.tabPressPos = data.pressY - tabRect.top;
              }
              data.tabLayout = Private$9$1.snapTabLayout(tabs, this._orientation);
              data.contentRect = this.contentNode.getBoundingClientRect();
              data.override = Drag.overrideCursor('default');
              // Add the dragging style classes.
              data.tab.classList.add('lm-mod-dragging');
              this.addClass('lm-mod-dragging');
              /* <DEPRECATED> */
              data.tab.classList.add('p-mod-dragging');
              this.addClass('p-mod-dragging');
              /* </DEPRECATED> */
              // Mark the drag as active.
              data.dragActive = true;
          }
          // Emit the detach requested signal if the threshold is exceeded.
          if (!data.detachRequested && Private$9$1.detachExceeded(data, event)) {
              // Only emit the signal once per drag cycle.
              data.detachRequested = true;
              // Setup the arguments for the signal.
              var index = data.index;
              var clientX = event.clientX;
              var clientY = event.clientY;
              var tab = tabs[index];
              var title = this._titles[index];
              // Emit the tab detach requested signal.
              this._tabDetachRequested.emit({ index: index, title: title, tab: tab, clientX: clientX, clientY: clientY });
              // Bail if the signal handler aborted the drag.
              if (data.dragAborted) {
                  return;
              }
          }
          // Update the positions of the tabs.
          Private$9$1.layoutTabs(tabs, data, event, this._orientation);
      };
      /**
       * Handle the `'mouseup'` event for the document.
       */
      TabBar.prototype._evtMouseUp = function (event) {
          var _this = this;
          // Do nothing if it's not a left or middle mouse release.
          if (event.button !== 0 && event.button !== 1) {
              return;
          }
          // Do nothing if no drag is in progress.
          var data = this._dragData;
          if (!data) {
              return;
          }
          // Stop the event propagation.
          event.preventDefault();
          event.stopPropagation();
          // Remove the extra mouse event listeners.
          document.removeEventListener('mousemove', this, true);
          document.removeEventListener('mouseup', this, true);
          document.removeEventListener('keydown', this, true);
          document.removeEventListener('contextmenu', this, true);
          // Handle a release when the drag is not active.
          if (!data.dragActive) {
              // Clear the drag data.
              this._dragData = null;
              // Lookup the tab nodes.
              var tabs = this.contentNode.children;
              // Find the index of the released tab.
              var index = ArrayExt.findFirstIndex(tabs, function (tab) {
                  return ElementExt.hitTest(tab, event.clientX, event.clientY);
              });
              // Do nothing if the release is not on the original pressed tab.
              if (index !== data.index) {
                  return;
              }
              // Ignore the release if the title is not closable.
              var title = this._titles[index];
              if (!title.closable) {
                  return;
              }
              // Emit the close requested signal if the middle button is released.
              if (event.button === 1) {
                  this._tabCloseRequested.emit({ index: index, title: title });
                  return;
              }
              // Emit the close requested signal if the close icon was released.
              var icon = tabs[index].querySelector(this.renderer.closeIconSelector);
              if (icon && icon.contains(event.target)) {
                  this._tabCloseRequested.emit({ index: index, title: title });
                  return;
              }
              // Otherwise, there is nothing left to do.
              return;
          }
          // Do nothing if the left button is not released.
          if (event.button !== 0) {
              return;
          }
          // Position the tab at its final resting position.
          Private$9$1.finalizeTabPosition(data, this._orientation);
          // Remove the dragging class from the tab so it can be transitioned.
          data.tab.classList.remove('lm-mod-dragging');
          /* <DEPRECATED> */
          data.tab.classList.remove('p-mod-dragging');
          /* </DEPRECATED> */
          // Parse the transition duration for releasing the tab.
          var duration = Private$9$1.parseTransitionDuration(data.tab);
          // Complete the release on a timer to allow the tab to transition.
          setTimeout(function () {
              // Do nothing if the drag has been aborted.
              if (data.dragAborted) {
                  return;
              }
              // Clear the drag data reference.
              _this._dragData = null;
              // Reset the positions of the tabs.
              Private$9$1.resetTabPositions(_this.contentNode.children, _this._orientation);
              // Clear the cursor grab.
              data.override.dispose();
              // Remove the remaining dragging style.
              _this.removeClass('lm-mod-dragging');
              /* <DEPRECATED> */
              _this.removeClass('p-mod-dragging');
              /* </DEPRECATED> */
              // If the tab was not moved, there is nothing else to do.
              var i = data.index;
              var j = data.targetIndex;
              if (j === -1 || i === j) {
                  return;
              }
              // Move the title to the new locations.
              ArrayExt.move(_this._titles, i, j);
              // Adjust the current index for the move.
              _this._adjustCurrentForMove(i, j);
              // Emit the tab moved signal.
              _this._tabMoved.emit({
                  fromIndex: i, toIndex: j, title: _this._titles[j]
              });
              // Update the tabs immediately to prevent flicker.
              MessageLoop.sendMessage(_this, Widget.Msg.UpdateRequest);
          }, duration);
      };
      /**
       * Release the mouse and restore the non-dragged tab positions.
       */
      TabBar.prototype._releaseMouse = function () {
          // Do nothing if no drag is in progress.
          var data = this._dragData;
          if (!data) {
              return;
          }
          // Clear the drag data reference.
          this._dragData = null;
          // Remove the extra mouse listeners.
          document.removeEventListener('mousemove', this, true);
          document.removeEventListener('mouseup', this, true);
          document.removeEventListener('keydown', this, true);
          document.removeEventListener('contextmenu', this, true);
          // Indicate the drag has been aborted. This allows the mouse
          // event handlers to return early when the drag is canceled.
          data.dragAborted = true;
          // If the drag is not active, there's nothing more to do.
          if (!data.dragActive) {
              return;
          }
          // Reset the tabs to their non-dragged positions.
          Private$9$1.resetTabPositions(this.contentNode.children, this._orientation);
          // Clear the cursor override.
          data.override.dispose();
          // Clear the dragging style classes.
          data.tab.classList.remove('lm-mod-dragging');
          this.removeClass('lm-mod-dragging');
          /* <DEPRECATED> */
          data.tab.classList.remove('p-mod-dragging');
          this.removeClass('p-mod-dragging');
          /* </DEPRECATED> */
      };
      /**
       * Adjust the current index for a tab insert operation.
       *
       * This method accounts for the tab bar's insertion behavior when
       * adjusting the current index and emitting the changed signal.
       */
      TabBar.prototype._adjustCurrentForInsert = function (i, title) {
          // Lookup commonly used variables.
          var ct = this.currentTitle;
          var ci = this._currentIndex;
          var bh = this.insertBehavior;
          // Handle the behavior where the new tab is always selected,
          // or the behavior where the new tab is selected if needed.
          if (bh === 'select-tab' || (bh === 'select-tab-if-needed' && ci === -1)) {
              this._currentIndex = i;
              this._previousTitle = ct;
              this._currentChanged.emit({
                  previousIndex: ci, previousTitle: ct,
                  currentIndex: i, currentTitle: title
              });
              return;
          }
          // Otherwise, silently adjust the current index if needed.
          if (ci >= i) {
              this._currentIndex++;
          }
      };
      /**
       * Adjust the current index for a tab move operation.
       *
       * This method will not cause the actual current tab to change.
       * It silently adjusts the index to account for the given move.
       */
      TabBar.prototype._adjustCurrentForMove = function (i, j) {
          if (this._currentIndex === i) {
              this._currentIndex = j;
          }
          else if (this._currentIndex < i && this._currentIndex >= j) {
              this._currentIndex++;
          }
          else if (this._currentIndex > i && this._currentIndex <= j) {
              this._currentIndex--;
          }
      };
      /**
       * Adjust the current index for a tab remove operation.
       *
       * This method accounts for the tab bar's remove behavior when
       * adjusting the current index and emitting the changed signal.
       */
      TabBar.prototype._adjustCurrentForRemove = function (i, title) {
          // Lookup commonly used variables.
          var ci = this._currentIndex;
          var bh = this.removeBehavior;
          // Silently adjust the index if the current tab is not removed.
          if (ci !== i) {
              if (ci > i) {
                  this._currentIndex--;
              }
              return;
          }
          // No tab gets selected if the tab bar is empty.
          if (this._titles.length === 0) {
              this._currentIndex = -1;
              this._currentChanged.emit({
                  previousIndex: i, previousTitle: title,
                  currentIndex: -1, currentTitle: null
              });
              return;
          }
          // Handle behavior where the next sibling tab is selected.
          if (bh === 'select-tab-after') {
              this._currentIndex = Math.min(i, this._titles.length - 1);
              this._currentChanged.emit({
                  previousIndex: i, previousTitle: title,
                  currentIndex: this._currentIndex, currentTitle: this.currentTitle
              });
              return;
          }
          // Handle behavior where the previous sibling tab is selected.
          if (bh === 'select-tab-before') {
              this._currentIndex = Math.max(0, i - 1);
              this._currentChanged.emit({
                  previousIndex: i, previousTitle: title,
                  currentIndex: this._currentIndex, currentTitle: this.currentTitle
              });
              return;
          }
          // Handle behavior where the previous history tab is selected.
          if (bh === 'select-previous-tab') {
              if (this._previousTitle) {
                  this._currentIndex = this._titles.indexOf(this._previousTitle);
                  this._previousTitle = null;
              }
              else {
                  this._currentIndex = Math.min(i, this._titles.length - 1);
              }
              this._currentChanged.emit({
                  previousIndex: i, previousTitle: title,
                  currentIndex: this._currentIndex, currentTitle: this.currentTitle
              });
              return;
          }
          // Otherwise, no tab gets selected.
          this._currentIndex = -1;
          this._currentChanged.emit({
              previousIndex: i, previousTitle: title,
              currentIndex: -1, currentTitle: null
          });
      };
      /**
       * Handle the `changed` signal of a title object.
       */
      TabBar.prototype._onTitleChanged = function (sender) {
          this.update();
      };
      return TabBar;
  }(Widget));
  /**
   * The namespace for the `TabBar` class statics.
   */
  (function (TabBar) {
      /**
       * The default implementation of `IRenderer`.
       *
       * #### Notes
       * Subclasses are free to reimplement rendering methods as needed.
       */
      var Renderer = /** @class */ (function () {
          /**
           * Construct a new renderer.
           */
          function Renderer() {
              /**
               * A selector which matches the close icon node in a tab.
               */
              this.closeIconSelector = '.lm-TabBar-tabCloseIcon';
              this._tabID = 0;
              this._tabKeys = new WeakMap();
          }
          /**
           * Render the virtual element for a tab.
           *
           * @param data - The data to use for rendering the tab.
           *
           * @returns A virtual element representing the tab.
           */
          Renderer.prototype.renderTab = function (data) {
              var title = data.title.caption;
              var key = this.createTabKey(data);
              var style = this.createTabStyle(data);
              var className = this.createTabClass(data);
              var dataset = this.createTabDataset(data);
              return (h.li({ key: key, className: className, title: title, style: style, dataset: dataset }, this.renderIcon(data), this.renderLabel(data), this.renderCloseIcon(data)));
          };
          /**
           * Render the icon element for a tab.
           *
           * @param data - The data to use for rendering the tab.
           *
           * @returns A virtual element representing the tab icon.
           */
          Renderer.prototype.renderIcon = function (data) {
              var title = data.title;
              var className = this.createIconClass(data);
              /* <DEPRECATED> */
              if (typeof title.icon === 'string') {
                  return h.div({ className: className }, title.iconLabel);
              }
              /* </DEPRECATED> */
              // if title.icon is undefined, it will be ignored
              return h.div({ className: className }, title.icon, title.iconLabel);
          };
          /**
           * Render the label element for a tab.
           *
           * @param data - The data to use for rendering the tab.
           *
           * @returns A virtual element representing the tab label.
           */
          Renderer.prototype.renderLabel = function (data) {
              return h.div({
                  className: 'lm-TabBar-tabLabel'
                      /* <DEPRECATED> */
                      + ' p-TabBar-tabLabel'
                  /* </DEPRECATED> */
              }, data.title.label);
          };
          /**
           * Render the close icon element for a tab.
           *
           * @param data - The data to use for rendering the tab.
           *
           * @returns A virtual element representing the tab close icon.
           */
          Renderer.prototype.renderCloseIcon = function (data) {
              return h.div({
                  className: 'lm-TabBar-tabCloseIcon'
                      /* <DEPRECATED> */
                      + ' p-TabBar-tabCloseIcon'
                  /* </DEPRECATED> */
              });
          };
          /**
           * Create a unique render key for the tab.
           *
           * @param data - The data to use for the tab.
           *
           * @returns The unique render key for the tab.
           *
           * #### Notes
           * This method caches the key against the tab title the first time
           * the key is generated. This enables efficient rendering of moved
           * tabs and avoids subtle hover style artifacts.
           */
          Renderer.prototype.createTabKey = function (data) {
              var key = this._tabKeys.get(data.title);
              if (key === undefined) {
                  key = "tab-key-" + this._tabID++;
                  this._tabKeys.set(data.title, key);
              }
              return key;
          };
          /**
           * Create the inline style object for a tab.
           *
           * @param data - The data to use for the tab.
           *
           * @returns The inline style data for the tab.
           */
          Renderer.prototype.createTabStyle = function (data) {
              return { zIndex: "" + data.zIndex };
          };
          /**
           * Create the class name for the tab.
           *
           * @param data - The data to use for the tab.
           *
           * @returns The full class name for the tab.
           */
          Renderer.prototype.createTabClass = function (data) {
              var name = 'lm-TabBar-tab';
              /* <DEPRECATED> */
              name += ' p-TabBar-tab';
              /* </DEPRECATED> */
              if (data.title.className) {
                  name += " " + data.title.className;
              }
              if (data.title.closable) {
                  name += ' lm-mod-closable';
                  /* <DEPRECATED> */
                  name += ' p-mod-closable';
                  /* </DEPRECATED> */
              }
              if (data.current) {
                  name += ' lm-mod-current';
                  /* <DEPRECATED> */
                  name += ' p-mod-current';
                  /* </DEPRECATED> */
              }
              return name;
          };
          /**
           * Create the dataset for a tab.
           *
           * @param data - The data to use for the tab.
           *
           * @returns The dataset for the tab.
           */
          Renderer.prototype.createTabDataset = function (data) {
              return data.title.dataset;
          };
          /**
           * Create the class name for the tab icon.
           *
           * @param data - The data to use for the tab.
           *
           * @returns The full class name for the tab icon.
           */
          Renderer.prototype.createIconClass = function (data) {
              var name = 'lm-TabBar-tabIcon';
              /* <DEPRECATED> */
              name += ' p-TabBar-tabIcon';
              /* </DEPRECATED> */
              var extra = data.title.iconClass;
              return extra ? name + " " + extra : name;
          };
          return Renderer;
      }());
      TabBar.Renderer = Renderer;
      /**
       * The default `Renderer` instance.
       */
      TabBar.defaultRenderer = new Renderer();
  })(TabBar || (TabBar = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$9$1;
  (function (Private) {
      /**
       * The start drag distance threshold.
       */
      Private.DRAG_THRESHOLD = 5;
      /**
       * The detach distance threshold.
       */
      Private.DETACH_THRESHOLD = 20;
      /**
       * Create the DOM node for a tab bar.
       */
      function createNode() {
          var node = document.createElement('div');
          var content = document.createElement('ul');
          content.className = 'lm-TabBar-content';
          /* <DEPRECATED> */
          content.classList.add('p-TabBar-content');
          /* </DEPRECATED> */
          node.appendChild(content);
          return node;
      }
      Private.createNode = createNode;
      /**
       * Coerce a title or options into a real title.
       */
      function asTitle(value) {
          return value instanceof Title ? value : new Title(value);
      }
      Private.asTitle = asTitle;
      /**
       * Parse the transition duration for a tab node.
       */
      function parseTransitionDuration(tab) {
          var style = window.getComputedStyle(tab);
          return 1000 * (parseFloat(style.transitionDuration) || 0);
      }
      Private.parseTransitionDuration = parseTransitionDuration;
      /**
       * Get a snapshot of the current tab layout values.
       */
      function snapTabLayout(tabs, orientation) {
          var layout = new Array(tabs.length);
          for (var i = 0, n = tabs.length; i < n; ++i) {
              var node = tabs[i];
              var style = window.getComputedStyle(node);
              if (orientation === 'horizontal') {
                  layout[i] = {
                      pos: node.offsetLeft,
                      size: node.offsetWidth,
                      margin: parseFloat(style.marginLeft) || 0
                  };
              }
              else {
                  layout[i] = {
                      pos: node.offsetTop,
                      size: node.offsetHeight,
                      margin: parseFloat(style.marginTop) || 0
                  };
              }
          }
          return layout;
      }
      Private.snapTabLayout = snapTabLayout;
      /**
       * Test if the event exceeds the drag threshold.
       */
      function dragExceeded(data, event) {
          var dx = Math.abs(event.clientX - data.pressX);
          var dy = Math.abs(event.clientY - data.pressY);
          return dx >= Private.DRAG_THRESHOLD || dy >= Private.DRAG_THRESHOLD;
      }
      Private.dragExceeded = dragExceeded;
      /**
       * Test if the event exceeds the drag detach threshold.
       */
      function detachExceeded(data, event) {
          var rect = data.contentRect;
          return ((event.clientX < rect.left - Private.DETACH_THRESHOLD) ||
              (event.clientX >= rect.right + Private.DETACH_THRESHOLD) ||
              (event.clientY < rect.top - Private.DETACH_THRESHOLD) ||
              (event.clientY >= rect.bottom + Private.DETACH_THRESHOLD));
      }
      Private.detachExceeded = detachExceeded;
      /**
       * Update the relative tab positions and computed target index.
       */
      function layoutTabs(tabs, data, event, orientation) {
          // Compute the orientation-sensitive values.
          var pressPos;
          var localPos;
          var clientPos;
          var clientSize;
          if (orientation === 'horizontal') {
              pressPos = data.pressX;
              localPos = event.clientX - data.contentRect.left;
              clientPos = event.clientX;
              clientSize = data.contentRect.width;
          }
          else {
              pressPos = data.pressY;
              localPos = event.clientY - data.contentRect.top;
              clientPos = event.clientY;
              clientSize = data.contentRect.height;
          }
          // Compute the target data.
          var targetIndex = data.index;
          var targetPos = localPos - data.tabPressPos;
          var targetEnd = targetPos + data.tabSize;
          // Update the relative tab positions.
          for (var i = 0, n = tabs.length; i < n; ++i) {
              var pxPos = void 0;
              var layout = data.tabLayout[i];
              var threshold = layout.pos + (layout.size >> 1);
              if (i < data.index && targetPos < threshold) {
                  pxPos = data.tabSize + data.tabLayout[i + 1].margin + "px";
                  targetIndex = Math.min(targetIndex, i);
              }
              else if (i > data.index && targetEnd > threshold) {
                  pxPos = -data.tabSize - layout.margin + "px";
                  targetIndex = Math.max(targetIndex, i);
              }
              else if (i === data.index) {
                  var ideal = clientPos - pressPos;
                  var limit = clientSize - (data.tabPos + data.tabSize);
                  pxPos = Math.max(-data.tabPos, Math.min(ideal, limit)) + "px";
              }
              else {
                  pxPos = '';
              }
              if (orientation === 'horizontal') {
                  tabs[i].style.left = pxPos;
              }
              else {
                  tabs[i].style.top = pxPos;
              }
          }
          // Update the computed target index.
          data.targetIndex = targetIndex;
      }
      Private.layoutTabs = layoutTabs;
      /**
       * Position the drag tab at its final resting relative position.
       */
      function finalizeTabPosition(data, orientation) {
          // Compute the orientation-sensitive client size.
          var clientSize;
          if (orientation === 'horizontal') {
              clientSize = data.contentRect.width;
          }
          else {
              clientSize = data.contentRect.height;
          }
          // Compute the ideal final tab position.
          var ideal;
          if (data.targetIndex === data.index) {
              ideal = 0;
          }
          else if (data.targetIndex > data.index) {
              var tgt = data.tabLayout[data.targetIndex];
              ideal = tgt.pos + tgt.size - data.tabSize - data.tabPos;
          }
          else {
              var tgt = data.tabLayout[data.targetIndex];
              ideal = tgt.pos - data.tabPos;
          }
          // Compute the tab position limit.
          var limit = clientSize - (data.tabPos + data.tabSize);
          var final = Math.max(-data.tabPos, Math.min(ideal, limit));
          // Set the final orientation-sensitive position.
          if (orientation === 'horizontal') {
              data.tab.style.left = final + "px";
          }
          else {
              data.tab.style.top = final + "px";
          }
      }
      Private.finalizeTabPosition = finalizeTabPosition;
      /**
       * Reset the relative positions of the given tabs.
       */
      function resetTabPositions(tabs, orientation) {
          each$1(tabs, function (tab) {
              if (orientation === 'horizontal') {
                  tab.style.left = '';
              }
              else {
                  tab.style.top = '';
              }
          });
      }
      Private.resetTabPositions = resetTabPositions;
  })(Private$9$1 || (Private$9$1 = {}));

  /**
   * A widget which provides a flexible docking area for widgets.
   */
  var DockPanel = /** @class */ (function (_super) {
      __extends$5(DockPanel, _super);
      /**
       * Construct a new dock panel.
       *
       * @param options - The options for initializing the panel.
       */
      function DockPanel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this) || this;
          _this._drag = null;
          _this._tabsMovable = true;
          _this._pressData = null;
          _this._layoutModified = new Signal(_this);
          _this.addClass('lm-DockPanel');
          /* <DEPRECATED> */
          _this.addClass('p-DockPanel');
          /* </DEPRECATED> */
          _this._mode = options.mode || 'multiple-document';
          _this._renderer = options.renderer || DockPanel.defaultRenderer;
          _this._edges = options.edges || Private$a.DEFAULT_EDGES;
          if (options.tabsMovable !== undefined) {
              _this._tabsMovable = options.tabsMovable;
          }
          // Toggle the CSS mode attribute.
          _this.dataset['mode'] = _this._mode;
          // Create the delegate renderer for the layout.
          var renderer = {
              createTabBar: function () { return _this._createTabBar(); },
              createHandle: function () { return _this._createHandle(); }
          };
          // Set up the dock layout for the panel.
          _this.layout = new DockLayout({ renderer: renderer, spacing: options.spacing });
          // Set up the overlay drop indicator.
          _this.overlay = options.overlay || new DockPanel.Overlay();
          _this.node.appendChild(_this.overlay.node);
          return _this;
      }
      /**
       * Dispose of the resources held by the panel.
       */
      DockPanel.prototype.dispose = function () {
          // Ensure the mouse is released.
          this._releaseMouse();
          // Hide the overlay.
          this.overlay.hide(0);
          // Cancel a drag if one is in progress.
          if (this._drag) {
              this._drag.dispose();
          }
          // Dispose of the base class.
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(DockPanel.prototype, "layoutModified", {
          /**
           * A signal emitted when the layout configuration is modified.
           *
           * #### Notes
           * This signal is emitted whenever the current layout configuration
           * may have changed.
           *
           * This signal is emitted asynchronously in a collapsed fashion, so
           * that multiple synchronous modifications results in only a single
           * emit of the signal.
           */
          get: function () {
              return this._layoutModified;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(DockPanel.prototype, "renderer", {
          /**
           * The renderer used by the dock panel.
           */
          get: function () {
              return this.layout.renderer;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(DockPanel.prototype, "spacing", {
          /**
           * Get the spacing between the widgets.
           */
          get: function () {
              return this.layout.spacing;
          },
          /**
           * Set the spacing between the widgets.
           */
          set: function (value) {
              this.layout.spacing = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(DockPanel.prototype, "mode", {
          /**
           * Get the mode for the dock panel.
           */
          get: function () {
              return this._mode;
          },
          /**
           * Set the mode for the dock panel.
           *
           * #### Notes
           * Changing the mode is a destructive operation with respect to the
           * panel's layout configuration. If layout state must be preserved,
           * save the current layout config before changing the mode.
           */
          set: function (value) {
              // Bail early if the mode does not change.
              if (this._mode === value) {
                  return;
              }
              // Update the internal mode.
              this._mode = value;
              // Toggle the CSS mode attribute.
              this.dataset['mode'] = value;
              // Get the layout for the panel.
              var layout = this.layout;
              // Configure the layout for the specified mode.
              switch (value) {
                  case 'multiple-document':
                      each$1(layout.tabBars(), function (tabBar) { tabBar.show(); });
                      break;
                  case 'single-document':
                      layout.restoreLayout(Private$a.createSingleDocumentConfig(this));
                      break;
                  default:
                      throw 'unreachable';
              }
              // Schedule an emit of the layout modified signal.
              MessageLoop.postMessage(this, Private$a.LayoutModified);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(DockPanel.prototype, "tabsMovable", {
          /**
           * Whether the tabs can be dragged / moved at runtime.
           */
          get: function () {
              return this._tabsMovable;
          },
          /**
           * Enable / Disable draggable / movable tabs.
           */
          set: function (value) {
              this._tabsMovable = value;
              each$1(this.tabBars(), function (tabbar) { return tabbar.tabsMovable = value; });
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(DockPanel.prototype, "isEmpty", {
          /**
           * Whether the dock panel is empty.
           */
          get: function () {
              return this.layout.isEmpty;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over the user widgets in the panel.
       *
       * @returns A new iterator over the user widgets in the panel.
       *
       * #### Notes
       * This iterator does not include the generated tab bars.
       */
      DockPanel.prototype.widgets = function () {
          return this.layout.widgets();
      };
      /**
       * Create an iterator over the selected widgets in the panel.
       *
       * @returns A new iterator over the selected user widgets.
       *
       * #### Notes
       * This iterator yields the widgets corresponding to the current tab
       * of each tab bar in the panel.
       */
      DockPanel.prototype.selectedWidgets = function () {
          return this.layout.selectedWidgets();
      };
      /**
       * Create an iterator over the tab bars in the panel.
       *
       * @returns A new iterator over the tab bars in the panel.
       *
       * #### Notes
       * This iterator does not include the user widgets.
       */
      DockPanel.prototype.tabBars = function () {
          return this.layout.tabBars();
      };
      /**
       * Create an iterator over the handles in the panel.
       *
       * @returns A new iterator over the handles in the panel.
       */
      DockPanel.prototype.handles = function () {
          return this.layout.handles();
      };
      /**
       * Select a specific widget in the dock panel.
       *
       * @param widget - The widget of interest.
       *
       * #### Notes
       * This will make the widget the current widget in its tab area.
       */
      DockPanel.prototype.selectWidget = function (widget) {
          // Find the tab bar which contains the widget.
          var tabBar = find$1(this.tabBars(), function (bar) {
              return bar.titles.indexOf(widget.title) !== -1;
          });
          // Throw an error if no tab bar is found.
          if (!tabBar) {
              throw new Error('Widget is not contained in the dock panel.');
          }
          // Ensure the widget is the current widget.
          tabBar.currentTitle = widget.title;
      };
      /**
       * Activate a specified widget in the dock panel.
       *
       * @param widget - The widget of interest.
       *
       * #### Notes
       * This will select and activate the given widget.
       */
      DockPanel.prototype.activateWidget = function (widget) {
          this.selectWidget(widget);
          widget.activate();
      };
      /**
       * Save the current layout configuration of the dock panel.
       *
       * @returns A new config object for the current layout state.
       *
       * #### Notes
       * The return value can be provided to the `restoreLayout` method
       * in order to restore the layout to its current configuration.
       */
      DockPanel.prototype.saveLayout = function () {
          return this.layout.saveLayout();
      };
      /**
       * Restore the layout to a previously saved configuration.
       *
       * @param config - The layout configuration to restore.
       *
       * #### Notes
       * Widgets which currently belong to the layout but which are not
       * contained in the config will be unparented.
       *
       * The dock panel automatically reverts to `'multiple-document'`
       * mode when a layout config is restored.
       */
      DockPanel.prototype.restoreLayout = function (config) {
          // Reset the mode.
          this._mode = 'multiple-document';
          // Restore the layout.
          this.layout.restoreLayout(config);
          // Flush the message loop on IE and Edge to prevent flicker.
          if (Platform.IS_EDGE || Platform.IS_IE) {
              MessageLoop.flush();
          }
          // Schedule an emit of the layout modified signal.
          MessageLoop.postMessage(this, Private$a.LayoutModified);
      };
      /**
       * Add a widget to the dock panel.
       *
       * @param widget - The widget to add to the dock panel.
       *
       * @param options - The additional options for adding the widget.
       *
       * #### Notes
       * If the panel is in single document mode, the options are ignored
       * and the widget is always added as tab in the hidden tab bar.
       */
      DockPanel.prototype.addWidget = function (widget, options) {
          if (options === void 0) { options = {}; }
          // Add the widget to the layout.
          if (this._mode === 'single-document') {
              this.layout.addWidget(widget);
          }
          else {
              this.layout.addWidget(widget, options);
          }
          // Schedule an emit of the layout modified signal.
          MessageLoop.postMessage(this, Private$a.LayoutModified);
      };
      /**
       * Process a message sent to the widget.
       *
       * @param msg - The message sent to the widget.
       */
      DockPanel.prototype.processMessage = function (msg) {
          if (msg.type === 'layout-modified') {
              this._layoutModified.emit(undefined);
          }
          else {
              _super.prototype.processMessage.call(this, msg);
          }
      };
      /**
       * Handle the DOM events for the dock panel.
       *
       * @param event - The DOM event sent to the panel.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the panel's DOM node. It should
       * not be called directly by user code.
       */
      DockPanel.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'lm-dragenter':
                  this._evtDragEnter(event);
                  break;
              case 'lm-dragleave':
                  this._evtDragLeave(event);
                  break;
              case 'lm-dragover':
                  this._evtDragOver(event);
                  break;
              case 'lm-drop':
                  this._evtDrop(event);
                  break;
              case 'mousedown':
                  this._evtMouseDown(event);
                  break;
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseup':
                  this._evtMouseUp(event);
                  break;
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'contextmenu':
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      DockPanel.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('lm-dragenter', this);
          this.node.addEventListener('lm-dragleave', this);
          this.node.addEventListener('lm-dragover', this);
          this.node.addEventListener('lm-drop', this);
          this.node.addEventListener('mousedown', this);
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       */
      DockPanel.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('lm-dragenter', this);
          this.node.removeEventListener('lm-dragleave', this);
          this.node.removeEventListener('lm-dragover', this);
          this.node.removeEventListener('lm-drop', this);
          this.node.removeEventListener('mousedown', this);
          this._releaseMouse();
      };
      /**
       * A message handler invoked on a `'child-added'` message.
       */
      DockPanel.prototype.onChildAdded = function (msg) {
          // Ignore the generated tab bars.
          if (Private$a.isGeneratedTabBarProperty.get(msg.child)) {
              return;
          }
          // Add the widget class to the child.
          msg.child.addClass('lm-DockPanel-widget');
          /* <DEPRECATED> */
          msg.child.addClass('p-DockPanel-widget');
          /* </DEPRECATED> */
      };
      /**
       * A message handler invoked on a `'child-removed'` message.
       */
      DockPanel.prototype.onChildRemoved = function (msg) {
          // Ignore the generated tab bars.
          if (Private$a.isGeneratedTabBarProperty.get(msg.child)) {
              return;
          }
          // Remove the widget class from the child.
          msg.child.removeClass('lm-DockPanel-widget');
          /* <DEPRECATED> */
          msg.child.removeClass('p-DockPanel-widget');
          /* </DEPRECATED> */
          // Schedule an emit of the layout modified signal.
          MessageLoop.postMessage(this, Private$a.LayoutModified);
      };
      /**
       * Handle the `'lm-dragenter'` event for the dock panel.
       */
      DockPanel.prototype._evtDragEnter = function (event) {
          // If the factory mime type is present, mark the event as
          // handled in order to get the rest of the drag events.
          if (event.mimeData.hasData('application/vnd.lumino.widget-factory')) {
              event.preventDefault();
              event.stopPropagation();
          }
      };
      /**
       * Handle the `'lm-dragleave'` event for the dock panel.
       */
      DockPanel.prototype._evtDragLeave = function (event) {
          // Mark the event as handled.
          event.preventDefault();
          event.stopPropagation();
          // The new target might be a descendant, so we might still handle the drop.
          // Hide asynchronously so that if a lm-dragover event bubbles up to us, the
          // hide is cancelled by the lm-dragover handler's show overlay logic.
          this.overlay.hide(1);
      };
      /**
       * Handle the `'lm-dragover'` event for the dock panel.
       */
      DockPanel.prototype._evtDragOver = function (event) {
          // Mark the event as handled.
          event.preventDefault();
          event.stopPropagation();
          // Show the drop indicator overlay and update the drop
          // action based on the drop target zone under the mouse.
          if (this._showOverlay(event.clientX, event.clientY) === 'invalid') {
              event.dropAction = 'none';
          }
          else {
              event.dropAction = event.proposedAction;
          }
      };
      /**
       * Handle the `'lm-drop'` event for the dock panel.
       */
      DockPanel.prototype._evtDrop = function (event) {
          // Mark the event as handled.
          event.preventDefault();
          event.stopPropagation();
          // Hide the drop indicator overlay.
          this.overlay.hide(0);
          // Bail if the proposed action is to do nothing.
          if (event.proposedAction === 'none') {
              event.dropAction = 'none';
              return;
          }
          // Find the drop target under the mouse.
          var clientX = event.clientX, clientY = event.clientY;
          var _a = Private$a.findDropTarget(this, clientX, clientY, this._edges), zone = _a.zone, target = _a.target;
          // Bail if the drop zone is invalid.
          if (zone === 'invalid') {
              event.dropAction = 'none';
              return;
          }
          // Bail if the factory mime type has invalid data.
          var mimeData = event.mimeData;
          var factory = mimeData.getData('application/vnd.lumino.widget-factory');
          if (typeof factory !== 'function') {
              event.dropAction = 'none';
              return;
          }
          // Bail if the factory does not produce a widget.
          var widget = factory();
          if (!(widget instanceof Widget)) {
              event.dropAction = 'none';
              return;
          }
          // Bail if the widget is an ancestor of the dock panel.
          if (widget.contains(this)) {
              event.dropAction = 'none';
              return;
          }
          // Find the reference widget for the drop target.
          var ref = target ? Private$a.getDropRef(target.tabBar) : null;
          // Add the widget according to the indicated drop zone.
          switch (zone) {
              case 'root-all':
                  this.addWidget(widget);
                  break;
              case 'root-top':
                  this.addWidget(widget, { mode: 'split-top' });
                  break;
              case 'root-left':
                  this.addWidget(widget, { mode: 'split-left' });
                  break;
              case 'root-right':
                  this.addWidget(widget, { mode: 'split-right' });
                  break;
              case 'root-bottom':
                  this.addWidget(widget, { mode: 'split-bottom' });
                  break;
              case 'widget-all':
                  this.addWidget(widget, { mode: 'tab-after', ref: ref });
                  break;
              case 'widget-top':
                  this.addWidget(widget, { mode: 'split-top', ref: ref });
                  break;
              case 'widget-left':
                  this.addWidget(widget, { mode: 'split-left', ref: ref });
                  break;
              case 'widget-right':
                  this.addWidget(widget, { mode: 'split-right', ref: ref });
                  break;
              case 'widget-bottom':
                  this.addWidget(widget, { mode: 'split-bottom', ref: ref });
                  break;
              case 'widget-tab':
                  this.addWidget(widget, { mode: 'tab-after', ref: ref });
                  break;
              default:
                  throw 'unreachable';
          }
          // Accept the proposed drop action.
          event.dropAction = event.proposedAction;
          // Activate the dropped widget.
          this.activateWidget(widget);
      };
      /**
       * Handle the `'keydown'` event for the dock panel.
       */
      DockPanel.prototype._evtKeyDown = function (event) {
          // Stop input events during drag.
          event.preventDefault();
          event.stopPropagation();
          // Release the mouse if `Escape` is pressed.
          if (event.keyCode === 27) {
              // Finalize the mouse release.
              this._releaseMouse();
              // Schedule an emit of the layout modified signal.
              MessageLoop.postMessage(this, Private$a.LayoutModified);
          }
      };
      /**
       * Handle the `'mousedown'` event for the dock panel.
       */
      DockPanel.prototype._evtMouseDown = function (event) {
          // Do nothing if the left mouse button is not pressed.
          if (event.button !== 0) {
              return;
          }
          // Find the handle which contains the mouse target, if any.
          var layout = this.layout;
          var target = event.target;
          var handle = find$1(layout.handles(), function (handle) { return handle.contains(target); });
          if (!handle) {
              return;
          }
          // Stop the event when a handle is pressed.
          event.preventDefault();
          event.stopPropagation();
          // Add the extra document listeners.
          document.addEventListener('keydown', this, true);
          document.addEventListener('mouseup', this, true);
          document.addEventListener('mousemove', this, true);
          document.addEventListener('contextmenu', this, true);
          // Compute the offset deltas for the handle press.
          var rect = handle.getBoundingClientRect();
          var deltaX = event.clientX - rect.left;
          var deltaY = event.clientY - rect.top;
          // Override the cursor and store the press data.
          var style = window.getComputedStyle(handle);
          var override = Drag.overrideCursor(style.cursor);
          this._pressData = { handle: handle, deltaX: deltaX, deltaY: deltaY, override: override };
      };
      /**
       * Handle the `'mousemove'` event for the dock panel.
       */
      DockPanel.prototype._evtMouseMove = function (event) {
          // Bail early if no drag is in progress.
          if (!this._pressData) {
              return;
          }
          // Stop the event when dragging a handle.
          event.preventDefault();
          event.stopPropagation();
          // Compute the desired offset position for the handle.
          var rect = this.node.getBoundingClientRect();
          var xPos = event.clientX - rect.left - this._pressData.deltaX;
          var yPos = event.clientY - rect.top - this._pressData.deltaY;
          // Set the handle as close to the desired position as possible.
          var layout = this.layout;
          layout.moveHandle(this._pressData.handle, xPos, yPos);
      };
      /**
       * Handle the `'mouseup'` event for the dock panel.
       */
      DockPanel.prototype._evtMouseUp = function (event) {
          // Do nothing if the left mouse button is not released.
          if (event.button !== 0) {
              return;
          }
          // Stop the event when releasing a handle.
          event.preventDefault();
          event.stopPropagation();
          // Finalize the mouse release.
          this._releaseMouse();
          // Schedule an emit of the layout modified signal.
          MessageLoop.postMessage(this, Private$a.LayoutModified);
      };
      /**
       * Release the mouse grab for the dock panel.
       */
      DockPanel.prototype._releaseMouse = function () {
          // Bail early if no drag is in progress.
          if (!this._pressData) {
              return;
          }
          // Clear the override cursor.
          this._pressData.override.dispose();
          this._pressData = null;
          // Remove the extra document listeners.
          document.removeEventListener('keydown', this, true);
          document.removeEventListener('mouseup', this, true);
          document.removeEventListener('mousemove', this, true);
          document.removeEventListener('contextmenu', this, true);
      };
      /**
       * Show the overlay indicator at the given client position.
       *
       * Returns the drop zone at the specified client position.
       *
       * #### Notes
       * If the position is not over a valid zone, the overlay is hidden.
       */
      DockPanel.prototype._showOverlay = function (clientX, clientY) {
          // Find the dock target for the given client position.
          var _a = Private$a.findDropTarget(this, clientX, clientY, this._edges), zone = _a.zone, target = _a.target;
          // If the drop zone is invalid, hide the overlay and bail.
          if (zone === 'invalid') {
              this.overlay.hide(100);
              return zone;
          }
          // Setup the variables needed to compute the overlay geometry.
          var top;
          var left;
          var right;
          var bottom;
          var box = ElementExt.boxSizing(this.node); // TODO cache this?
          var rect = this.node.getBoundingClientRect();
          // Compute the overlay geometry based on the dock zone.
          switch (zone) {
              case 'root-all':
                  top = box.paddingTop;
                  left = box.paddingLeft;
                  right = box.paddingRight;
                  bottom = box.paddingBottom;
                  break;
              case 'root-top':
                  top = box.paddingTop;
                  left = box.paddingLeft;
                  right = box.paddingRight;
                  bottom = rect.height * Private$a.GOLDEN_RATIO;
                  break;
              case 'root-left':
                  top = box.paddingTop;
                  left = box.paddingLeft;
                  right = rect.width * Private$a.GOLDEN_RATIO;
                  bottom = box.paddingBottom;
                  break;
              case 'root-right':
                  top = box.paddingTop;
                  left = rect.width * Private$a.GOLDEN_RATIO;
                  right = box.paddingRight;
                  bottom = box.paddingBottom;
                  break;
              case 'root-bottom':
                  top = rect.height * Private$a.GOLDEN_RATIO;
                  left = box.paddingLeft;
                  right = box.paddingRight;
                  bottom = box.paddingBottom;
                  break;
              case 'widget-all':
                  top = target.top;
                  left = target.left;
                  right = target.right;
                  bottom = target.bottom;
                  break;
              case 'widget-top':
                  top = target.top;
                  left = target.left;
                  right = target.right;
                  bottom = target.bottom + target.height / 2;
                  break;
              case 'widget-left':
                  top = target.top;
                  left = target.left;
                  right = target.right + target.width / 2;
                  bottom = target.bottom;
                  break;
              case 'widget-right':
                  top = target.top;
                  left = target.left + target.width / 2;
                  right = target.right;
                  bottom = target.bottom;
                  break;
              case 'widget-bottom':
                  top = target.top + target.height / 2;
                  left = target.left;
                  right = target.right;
                  bottom = target.bottom;
                  break;
              case 'widget-tab':
                  var tabHeight = target.tabBar.node.getBoundingClientRect().height;
                  top = target.top;
                  left = target.left;
                  right = target.right;
                  bottom = target.bottom + target.height - tabHeight;
                  break;
              default:
                  throw 'unreachable';
          }
          // Show the overlay with the computed geometry.
          this.overlay.show({ top: top, left: left, right: right, bottom: bottom });
          // Finally, return the computed drop zone.
          return zone;
      };
      /**
       * Create a new tab bar for use by the panel.
       */
      DockPanel.prototype._createTabBar = function () {
          // Create the tab bar.
          var tabBar = this._renderer.createTabBar();
          // Set the generated tab bar property for the tab bar.
          Private$a.isGeneratedTabBarProperty.set(tabBar, true);
          // Hide the tab bar when in single document mode.
          if (this._mode === 'single-document') {
              tabBar.hide();
          }
          // Enforce necessary tab bar behavior.
          // TODO do we really want to enforce *all* of these?
          tabBar.tabsMovable = this._tabsMovable;
          tabBar.allowDeselect = false;
          tabBar.removeBehavior = 'select-previous-tab';
          tabBar.insertBehavior = 'select-tab-if-needed';
          // Connect the signal handlers for the tab bar.
          tabBar.tabMoved.connect(this._onTabMoved, this);
          tabBar.currentChanged.connect(this._onCurrentChanged, this);
          tabBar.tabCloseRequested.connect(this._onTabCloseRequested, this);
          tabBar.tabDetachRequested.connect(this._onTabDetachRequested, this);
          tabBar.tabActivateRequested.connect(this._onTabActivateRequested, this);
          // Return the initialized tab bar.
          return tabBar;
      };
      /**
       * Create a new handle for use by the panel.
       */
      DockPanel.prototype._createHandle = function () {
          return this._renderer.createHandle();
      };
      /**
       * Handle the `tabMoved` signal from a tab bar.
       */
      DockPanel.prototype._onTabMoved = function () {
          MessageLoop.postMessage(this, Private$a.LayoutModified);
      };
      /**
       * Handle the `currentChanged` signal from a tab bar.
       */
      DockPanel.prototype._onCurrentChanged = function (sender, args) {
          // Extract the previous and current title from the args.
          var previousTitle = args.previousTitle, currentTitle = args.currentTitle;
          // Hide the previous widget.
          if (previousTitle) {
              previousTitle.owner.hide();
          }
          // Show the current widget.
          if (currentTitle) {
              currentTitle.owner.show();
          }
          // Flush the message loop on IE and Edge to prevent flicker.
          if (Platform.IS_EDGE || Platform.IS_IE) {
              MessageLoop.flush();
          }
          // Schedule an emit of the layout modified signal.
          MessageLoop.postMessage(this, Private$a.LayoutModified);
      };
      /**
       * Handle the `tabActivateRequested` signal from a tab bar.
       */
      DockPanel.prototype._onTabActivateRequested = function (sender, args) {
          args.title.owner.activate();
      };
      /**
       * Handle the `tabCloseRequested` signal from a tab bar.
       */
      DockPanel.prototype._onTabCloseRequested = function (sender, args) {
          args.title.owner.close();
      };
      /**
       * Handle the `tabDetachRequested` signal from a tab bar.
       */
      DockPanel.prototype._onTabDetachRequested = function (sender, args) {
          var _this = this;
          // Do nothing if a drag is already in progress.
          if (this._drag) {
              return;
          }
          // Release the tab bar's hold on the mouse.
          sender.releaseMouse();
          // Extract the data from the args.
          var title = args.title, tab = args.tab, clientX = args.clientX, clientY = args.clientY;
          // Setup the mime data for the drag operation.
          var mimeData = new MimeData();
          var factory = function () { return title.owner; };
          mimeData.setData('application/vnd.lumino.widget-factory', factory);
          // Create the drag image for the drag operation.
          var dragImage = tab.cloneNode(true);
          // Create the drag object to manage the drag-drop operation.
          this._drag = new Drag({
              mimeData: mimeData, dragImage: dragImage,
              proposedAction: 'move',
              supportedActions: 'move',
          });
          // Hide the tab node in the original tab.
          tab.classList.add('lm-mod-hidden');
          /* <DEPRECATED> */
          tab.classList.add('p-mod-hidden');
          // Create the cleanup callback.
          var cleanup = (function () {
              _this._drag = null;
              tab.classList.remove('lm-mod-hidden');
              /* <DEPRECATED> */
              tab.classList.remove('p-mod-hidden');
          });
          // Start the drag operation and cleanup when done.
          this._drag.start(clientX, clientY).then(cleanup);
      };
      return DockPanel;
  }(Widget));
  /**
   * The namespace for the `DockPanel` class statics.
   */
  (function (DockPanel) {
      /**
       * A concrete implementation of `IOverlay`.
       *
       * This is the default overlay implementation for a dock panel.
       */
      var Overlay = /** @class */ (function () {
          /**
           * Construct a new overlay.
           */
          function Overlay() {
              this._timer = -1;
              this._hidden = true;
              this.node = document.createElement('div');
              this.node.classList.add('lm-DockPanel-overlay');
              this.node.classList.add('lm-mod-hidden');
              /* <DEPRECATED> */
              this.node.classList.add('p-DockPanel-overlay');
              this.node.classList.add('p-mod-hidden');
              this.node.style.position = 'absolute';
          }
          /**
           * Show the overlay using the given overlay geometry.
           *
           * @param geo - The desired geometry for the overlay.
           */
          Overlay.prototype.show = function (geo) {
              // Update the position of the overlay.
              var style = this.node.style;
              style.top = geo.top + "px";
              style.left = geo.left + "px";
              style.right = geo.right + "px";
              style.bottom = geo.bottom + "px";
              // Clear any pending hide timer.
              clearTimeout(this._timer);
              this._timer = -1;
              // If the overlay is already visible, we're done.
              if (!this._hidden) {
                  return;
              }
              // Clear the hidden flag.
              this._hidden = false;
              // Finally, show the overlay.
              this.node.classList.remove('lm-mod-hidden');
              /* <DEPRECATED> */
              this.node.classList.remove('p-mod-hidden');
          };
          /**
           * Hide the overlay node.
           *
           * @param delay - The delay (in ms) before hiding the overlay.
           *   A delay value <= 0 will hide the overlay immediately.
           */
          Overlay.prototype.hide = function (delay) {
              var _this = this;
              // Do nothing if the overlay is already hidden.
              if (this._hidden) {
                  return;
              }
              // Hide immediately if the delay is <= 0.
              if (delay <= 0) {
                  clearTimeout(this._timer);
                  this._timer = -1;
                  this._hidden = true;
                  this.node.classList.add('lm-mod-hidden');
                  /* <DEPRECATED> */
                  this.node.classList.add('p-mod-hidden');
                  return;
              }
              // Do nothing if a hide is already pending.
              if (this._timer !== -1) {
                  return;
              }
              // Otherwise setup the hide timer.
              this._timer = window.setTimeout(function () {
                  _this._timer = -1;
                  _this._hidden = true;
                  _this.node.classList.add('lm-mod-hidden');
                  /* <DEPRECATED> */
                  _this.node.classList.add('p-mod-hidden');
              }, delay);
          };
          return Overlay;
      }());
      DockPanel.Overlay = Overlay;
      /**
       * The default implementation of `IRenderer`.
       */
      var Renderer = /** @class */ (function () {
          function Renderer() {
          }
          /**
           * Create a new tab bar for use with a dock panel.
           *
           * @returns A new tab bar for a dock panel.
           */
          Renderer.prototype.createTabBar = function () {
              var bar = new TabBar();
              bar.addClass('lm-DockPanel-tabBar');
              /* <DEPRECATED> */
              bar.addClass('p-DockPanel-tabBar');
              /* </DEPRECATED> */
              return bar;
          };
          /**
           * Create a new handle node for use with a dock panel.
           *
           * @returns A new handle node for a dock panel.
           */
          Renderer.prototype.createHandle = function () {
              var handle = document.createElement('div');
              handle.className = 'lm-DockPanel-handle';
              /* <DEPRECATED> */
              handle.classList.add('p-DockPanel-handle');
              return handle;
          };
          return Renderer;
      }());
      DockPanel.Renderer = Renderer;
      /**
       * The default `Renderer` instance.
       */
      DockPanel.defaultRenderer = new Renderer();
  })(DockPanel || (DockPanel = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$a;
  (function (Private) {
      /**
       * A fraction used for sizing root panels; ~= `1 / golden_ratio`.
       */
      Private.GOLDEN_RATIO = 0.618;
      /**
       * The default sizes for the edge drop zones, in pixels.
       */
      Private.DEFAULT_EDGES = {
          /**
           * The size of the top edge dock zone for the root panel, in pixels.
           * This is different from the others to distinguish between the top
           * tab bar and the top root zone.
           */
          top: 12,
          /**
           * The size of the edge dock zone for the root panel, in pixels.
           */
          right: 40,
          /**
           * The size of the edge dock zone for the root panel, in pixels.
           */
          bottom: 40,
          /**
           * The size of the edge dock zone for the root panel, in pixels.
           */
          left: 40
      };
      /**
       * A singleton `'layout-modified'` conflatable message.
       */
      Private.LayoutModified = new ConflatableMessage('layout-modified');
      /**
       * An attached property used to track generated tab bars.
       */
      Private.isGeneratedTabBarProperty = new AttachedProperty({
          name: 'isGeneratedTabBar',
          create: function () { return false; }
      });
      /**
       * Create a single document config for the widgets in a dock panel.
       */
      function createSingleDocumentConfig(panel) {
          // Return an empty config if the panel is empty.
          if (panel.isEmpty) {
              return { main: null };
          }
          // Get a flat array of the widgets in the panel.
          var widgets = toArray$1(panel.widgets());
          // Get the first selected widget in the panel.
          var selected = panel.selectedWidgets().next();
          // Compute the current index for the new config.
          var currentIndex = selected ? widgets.indexOf(selected) : -1;
          // Return the single document config.
          return { main: { type: 'tab-area', widgets: widgets, currentIndex: currentIndex } };
      }
      Private.createSingleDocumentConfig = createSingleDocumentConfig;
      /**
       * Find the drop target at the given client position.
       */
      function findDropTarget(panel, clientX, clientY, edges) {
          // Bail if the mouse is not over the dock panel.
          if (!ElementExt.hitTest(panel.node, clientX, clientY)) {
              return { zone: 'invalid', target: null };
          }
          // Look up the layout for the panel.
          var layout = panel.layout;
          // If the layout is empty, indicate the entire root drop zone.
          if (layout.isEmpty) {
              return { zone: 'root-all', target: null };
          }
          // Test the edge zones when in multiple document mode.
          if (panel.mode === 'multiple-document') {
              // Get the client rect for the dock panel.
              var panelRect = panel.node.getBoundingClientRect();
              // Compute the distance to each edge of the panel.
              var pl = clientX - panelRect.left + 1;
              var pt = clientY - panelRect.top + 1;
              var pr = panelRect.right - clientX;
              var pb = panelRect.bottom - clientY;
              // Find the minimum distance to an edge.
              var pd = Math.min(pt, pr, pb, pl);
              // Return a root zone if the mouse is within an edge.
              switch (pd) {
                  case pt:
                      if (pt < edges.top) {
                          return { zone: 'root-top', target: null };
                      }
                      break;
                  case pr:
                      if (pr < edges.right) {
                          return { zone: 'root-right', target: null };
                      }
                      break;
                  case pb:
                      if (pb < edges.bottom) {
                          return { zone: 'root-bottom', target: null };
                      }
                      break;
                  case pl:
                      if (pl < edges.left) {
                          return { zone: 'root-left', target: null };
                      }
                      break;
                  default:
                      throw 'unreachable';
              }
          }
          // Hit test the dock layout at the given client position.
          var target = layout.hitTestTabAreas(clientX, clientY);
          // Bail if no target area was found.
          if (!target) {
              return { zone: 'invalid', target: null };
          }
          // Return the whole tab area when in single document mode.
          if (panel.mode === 'single-document') {
              return { zone: 'widget-all', target: target };
          }
          // Compute the distance to each edge of the tab area.
          var al = target.x - target.left + 1;
          var at = target.y - target.top + 1;
          var ar = target.left + target.width - target.x;
          var ab = target.top + target.height - target.y;
          var tabHeight = target.tabBar.node.getBoundingClientRect().height;
          if (at < tabHeight) {
              return { zone: 'widget-tab', target: target };
          }
          // Get the X and Y edge sizes for the area.
          var rx = Math.round(target.width / 3);
          var ry = Math.round(target.height / 3);
          // If the mouse is not within an edge, indicate the entire area.
          if (al > rx && ar > rx && at > ry && ab > ry) {
              return { zone: 'widget-all', target: target };
          }
          // Scale the distances by the slenderness ratio.
          al /= rx;
          at /= ry;
          ar /= rx;
          ab /= ry;
          // Find the minimum distance to the area edge.
          var ad = Math.min(al, at, ar, ab);
          // Find the widget zone for the area edge.
          var zone;
          switch (ad) {
              case al:
                  zone = 'widget-left';
                  break;
              case at:
                  zone = 'widget-top';
                  break;
              case ar:
                  zone = 'widget-right';
                  break;
              case ab:
                  zone = 'widget-bottom';
                  break;
              default:
                  throw 'unreachable';
          }
          // Return the final drop target.
          return { zone: zone, target: target };
      }
      Private.findDropTarget = findDropTarget;
      /**
       * Get the drop reference widget for a tab bar.
       */
      function getDropRef(tabBar) {
          if (tabBar.titles.length === 0) {
              return null;
          }
          if (tabBar.currentTitle) {
              return tabBar.currentTitle.owner;
          }
          return tabBar.titles[tabBar.titles.length - 1].owner;
      }
      Private.getDropRef = getDropRef;
  })(Private$a || (Private$a = {}));

  // Copyright (c) Jupyter Development Team.
  /**
   * A class which tracks focus among a set of widgets.
   *
   * This class is useful when code needs to keep track of the most
   * recently focused widget(s) among a set of related widgets.
   */
  var FocusTracker = /** @class */ (function () {
      /**
       * Construct a new focus tracker.
       */
      function FocusTracker() {
          this._counter = 0;
          this._widgets = [];
          this._activeWidget = null;
          this._currentWidget = null;
          this._numbers = new Map();
          this._nodes = new Map();
          this._activeChanged = new Signal(this);
          this._currentChanged = new Signal(this);
      }
      /**
       * Dispose of the resources held by the tracker.
       */
      FocusTracker.prototype.dispose = function () {
          var _this = this;
          // Do nothing if the tracker is already disposed.
          if (this._counter < 0) {
              return;
          }
          // Mark the tracker as disposed.
          this._counter = -1;
          // Clear the connections for the tracker.
          Signal.clearData(this);
          // Remove all event listeners.
          each$1(this._widgets, function (w) {
              w.node.removeEventListener('focus', _this, true);
              w.node.removeEventListener('blur', _this, true);
          });
          // Clear the internal data structures.
          this._activeWidget = null;
          this._currentWidget = null;
          this._nodes.clear();
          this._numbers.clear();
          this._widgets.length = 0;
      };
      Object.defineProperty(FocusTracker.prototype, "currentChanged", {
          /**
           * A signal emitted when the current widget has changed.
           */
          get: function () {
              return this._currentChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(FocusTracker.prototype, "activeChanged", {
          /**
           * A signal emitted when the active widget has changed.
           */
          get: function () {
              return this._activeChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(FocusTracker.prototype, "isDisposed", {
          /**
           * A flag indicating whether the tracker is disposed.
           */
          get: function () {
              return this._counter < 0;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(FocusTracker.prototype, "currentWidget", {
          /**
           * The current widget in the tracker.
           *
           * #### Notes
           * The current widget is the widget among the tracked widgets which
           * has the *descendant node* which has most recently been focused.
           *
           * The current widget will not be updated if the node loses focus. It
           * will only be updated when a different tracked widget gains focus.
           *
           * If the current widget is removed from the tracker, the previous
           * current widget will be restored.
           *
           * This behavior is intended to follow a user's conceptual model of
           * a semantically "current" widget, where the "last thing of type X"
           * to be interacted with is the "current instance of X", regardless
           * of whether that instance still has focus.
           */
          get: function () {
              return this._currentWidget;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(FocusTracker.prototype, "activeWidget", {
          /**
           * The active widget in the tracker.
           *
           * #### Notes
           * The active widget is the widget among the tracked widgets which
           * has the *descendant node* which is currently focused.
           */
          get: function () {
              return this._activeWidget;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(FocusTracker.prototype, "widgets", {
          /**
           * A read only array of the widgets being tracked.
           */
          get: function () {
              return this._widgets;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Get the focus number for a particular widget in the tracker.
       *
       * @param widget - The widget of interest.
       *
       * @returns The focus number for the given widget, or `-1` if the
       *   widget has not had focus since being added to the tracker, or
       *   is not contained by the tracker.
       *
       * #### Notes
       * The focus number indicates the relative order in which the widgets
       * have gained focus. A widget with a larger number has gained focus
       * more recently than a widget with a smaller number.
       *
       * The `currentWidget` will always have the largest focus number.
       *
       * All widgets start with a focus number of `-1`, which indicates that
       * the widget has not been focused since being added to the tracker.
       */
      FocusTracker.prototype.focusNumber = function (widget) {
          var n = this._numbers.get(widget);
          return n === undefined ? -1 : n;
      };
      /**
       * Test whether the focus tracker contains a given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns `true` if the widget is tracked, `false` otherwise.
       */
      FocusTracker.prototype.has = function (widget) {
          return this._numbers.has(widget);
      };
      /**
       * Add a widget to the focus tracker.
       *
       * @param widget - The widget of interest.
       *
       * #### Notes
       * A widget will be automatically removed from the tracker if it
       * is disposed after being added.
       *
       * If the widget is already tracked, this is a no-op.
       */
      FocusTracker.prototype.add = function (widget) {
          // Do nothing if the widget is already tracked.
          if (this._numbers.has(widget)) {
              return;
          }
          // Test whether the widget has focus.
          var focused = widget.node.contains(document.activeElement);
          // Set up the initial focus number.
          var n = focused ? this._counter++ : -1;
          // Add the widget to the internal data structures.
          this._widgets.push(widget);
          this._numbers.set(widget, n);
          this._nodes.set(widget.node, widget);
          // Set up the event listeners. The capturing phase must be used
          // since the 'focus' and 'blur' events don't bubble and Firefox
          // doesn't support the 'focusin' or 'focusout' events.
          widget.node.addEventListener('focus', this, true);
          widget.node.addEventListener('blur', this, true);
          // Connect the disposed signal handler.
          widget.disposed.connect(this._onWidgetDisposed, this);
          // Set the current and active widgets if needed.
          if (focused) {
              this._setWidgets(widget, widget);
          }
      };
      /**
       * Remove a widget from the focus tracker.
       *
       * #### Notes
       * If the widget is the `currentWidget`, the previous current widget
       * will become the new `currentWidget`.
       *
       * A widget will be automatically removed from the tracker if it
       * is disposed after being added.
       *
       * If the widget is not tracked, this is a no-op.
       */
      FocusTracker.prototype.remove = function (widget) {
          var _this = this;
          // Bail early if the widget is not tracked.
          if (!this._numbers.has(widget)) {
              return;
          }
          // Disconnect the disposed signal handler.
          widget.disposed.disconnect(this._onWidgetDisposed, this);
          // Remove the event listeners.
          widget.node.removeEventListener('focus', this, true);
          widget.node.removeEventListener('blur', this, true);
          // Remove the widget from the internal data structures.
          ArrayExt.removeFirstOf(this._widgets, widget);
          this._nodes.delete(widget.node);
          this._numbers.delete(widget);
          // Bail early if the widget is not the current widget.
          if (this._currentWidget !== widget) {
              return;
          }
          // Filter the widgets for those which have had focus.
          var valid = filter$1(this._widgets, function (w) { return _this._numbers.get(w) !== -1; });
          // Get the valid widget with the max focus number.
          var previous = max$1(valid, function (first, second) {
              var a = _this._numbers.get(first);
              var b = _this._numbers.get(second);
              return a - b;
          }) || null;
          // Set the current and active widgets.
          this._setWidgets(previous, null);
      };
      /**
       * Handle the DOM events for the focus tracker.
       *
       * @param event - The DOM event sent to the panel.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the tracked nodes. It should
       * not be called directly by user code.
       */
      FocusTracker.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'focus':
                  this._evtFocus(event);
                  break;
              case 'blur':
                  this._evtBlur(event);
                  break;
          }
      };
      /**
       * Set the current and active widgets for the tracker.
       */
      FocusTracker.prototype._setWidgets = function (current, active) {
          // Swap the current widget.
          var oldCurrent = this._currentWidget;
          this._currentWidget = current;
          // Swap the active widget.
          var oldActive = this._activeWidget;
          this._activeWidget = active;
          // Emit the `currentChanged` signal if needed.
          if (oldCurrent !== current) {
              this._currentChanged.emit({ oldValue: oldCurrent, newValue: current });
          }
          // Emit the `activeChanged` signal if needed.
          if (oldActive !== active) {
              this._activeChanged.emit({ oldValue: oldActive, newValue: active });
          }
      };
      /**
       * Handle the `'focus'` event for a tracked widget.
       */
      FocusTracker.prototype._evtFocus = function (event) {
          // Find the widget which gained focus, which is known to exist.
          var widget = this._nodes.get(event.currentTarget);
          // Update the focus number if necessary.
          if (widget !== this._currentWidget) {
              this._numbers.set(widget, this._counter++);
          }
          // Set the current and active widgets.
          this._setWidgets(widget, widget);
      };
      /**
       * Handle the `'blur'` event for a tracked widget.
       */
      FocusTracker.prototype._evtBlur = function (event) {
          // Find the widget which lost focus, which is known to exist.
          var widget = this._nodes.get(event.currentTarget);
          // Get the node which being focused after this blur.
          var focusTarget = event.relatedTarget;
          // If no other node is being focused, clear the active widget.
          if (!focusTarget) {
              this._setWidgets(this._currentWidget, null);
              return;
          }
          // Bail if the focus widget is not changing.
          if (widget.node.contains(focusTarget)) {
              return;
          }
          // If no tracked widget is being focused, clear the active widget.
          if (!find$1(this._widgets, function (w) { return w.node.contains(focusTarget); })) {
              this._setWidgets(this._currentWidget, null);
              return;
          }
      };
      /**
       * Handle the `disposed` signal for a tracked widget.
       */
      FocusTracker.prototype._onWidgetDisposed = function (sender) {
          this.remove(sender);
      };
      return FocusTracker;
  }());

  /**
   * A layout which arranges its widgets in a grid.
   */
  var GridLayout = /** @class */ (function (_super) {
      __extends$5(GridLayout, _super);
      /**
       * Construct a new grid layout.
       *
       * @param options - The options for initializing the layout.
       */
      function GridLayout(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, options) || this;
          _this._dirty = false;
          _this._rowSpacing = 4;
          _this._columnSpacing = 4;
          _this._items = [];
          _this._rowStarts = [];
          _this._columnStarts = [];
          _this._rowSizers = [new BoxSizer()];
          _this._columnSizers = [new BoxSizer()];
          _this._box = null;
          if (options.rowCount !== undefined) {
              Private$b.reallocSizers(_this._rowSizers, options.rowCount);
          }
          if (options.columnCount !== undefined) {
              Private$b.reallocSizers(_this._columnSizers, options.columnCount);
          }
          if (options.rowSpacing !== undefined) {
              _this._rowSpacing = Private$b.clampValue(options.rowSpacing);
          }
          if (options.columnSpacing !== undefined) {
              _this._columnSpacing = Private$b.clampValue(options.columnSpacing);
          }
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       */
      GridLayout.prototype.dispose = function () {
          // Dispose of the widgets and layout items.
          each$1(this._items, function (item) {
              var widget = item.widget;
              item.dispose();
              widget.dispose();
          });
          // Clear the layout state.
          this._box = null;
          this._items.length = 0;
          this._rowStarts.length = 0;
          this._rowSizers.length = 0;
          this._columnStarts.length = 0;
          this._columnSizers.length = 0;
          // Dispose of the rest of the layout.
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(GridLayout.prototype, "rowCount", {
          /**
           * Get the number of rows in the layout.
           */
          get: function () {
              return this._rowSizers.length;
          },
          /**
           * Set the number of rows in the layout.
           *
           * #### Notes
           * The minimum row count is `1`.
           */
          set: function (value) {
              // Do nothing if the row count does not change.
              if (value === this.rowCount) {
                  return;
              }
              // Reallocate the row sizers.
              Private$b.reallocSizers(this._rowSizers, value);
              // Schedule a fit of the parent.
              if (this.parent) {
                  this.parent.fit();
              }
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(GridLayout.prototype, "columnCount", {
          /**
           * Get the number of columns in the layout.
           */
          get: function () {
              return this._columnSizers.length;
          },
          /**
           * Set the number of columns in the layout.
           *
           * #### Notes
           * The minimum column count is `1`.
           */
          set: function (value) {
              // Do nothing if the column count does not change.
              if (value === this.columnCount) {
                  return;
              }
              // Reallocate the column sizers.
              Private$b.reallocSizers(this._columnSizers, value);
              // Schedule a fit of the parent.
              if (this.parent) {
                  this.parent.fit();
              }
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(GridLayout.prototype, "rowSpacing", {
          /**
           * Get the row spacing for the layout.
           */
          get: function () {
              return this._rowSpacing;
          },
          /**
           * Set the row spacing for the layout.
           */
          set: function (value) {
              // Clamp the spacing to the allowed range.
              value = Private$b.clampValue(value);
              // Bail if the spacing does not change
              if (this._rowSpacing === value) {
                  return;
              }
              // Update the internal spacing.
              this._rowSpacing = value;
              // Schedule a fit of the parent.
              if (this.parent) {
                  this.parent.fit();
              }
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(GridLayout.prototype, "columnSpacing", {
          /**
           * Get the column spacing for the layout.
           */
          get: function () {
              return this._columnSpacing;
          },
          /**
           * Set the col spacing for the layout.
           */
          set: function (value) {
              // Clamp the spacing to the allowed range.
              value = Private$b.clampValue(value);
              // Bail if the spacing does not change
              if (this._columnSpacing === value) {
                  return;
              }
              // Update the internal spacing.
              this._columnSpacing = value;
              // Schedule a fit of the parent.
              if (this.parent) {
                  this.parent.fit();
              }
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Get the stretch factor for a specific row.
       *
       * @param index - The row index of interest.
       *
       * @returns The stretch factor for the row.
       *
       * #### Notes
       * This returns `-1` if the index is out of range.
       */
      GridLayout.prototype.rowStretch = function (index) {
          var sizer = this._rowSizers[index];
          return sizer ? sizer.stretch : -1;
      };
      /**
       * Set the stretch factor for a specific row.
       *
       * @param index - The row index of interest.
       *
       * @param value - The stretch factor for the row.
       *
       * #### Notes
       * This is a no-op if the index is out of range.
       */
      GridLayout.prototype.setRowStretch = function (index, value) {
          // Look up the row sizer.
          var sizer = this._rowSizers[index];
          // Bail if the index is out of range.
          if (!sizer) {
              return;
          }
          // Clamp the value to the allowed range.
          value = Private$b.clampValue(value);
          // Bail if the stretch does not change.
          if (sizer.stretch === value) {
              return;
          }
          // Update the sizer stretch.
          sizer.stretch = value;
          // Schedule an update of the parent.
          if (this.parent) {
              this.parent.update();
          }
      };
      /**
       * Get the stretch factor for a specific column.
       *
       * @param index - The column index of interest.
       *
       * @returns The stretch factor for the column.
       *
       * #### Notes
       * This returns `-1` if the index is out of range.
       */
      GridLayout.prototype.columnStretch = function (index) {
          var sizer = this._columnSizers[index];
          return sizer ? sizer.stretch : -1;
      };
      /**
       * Set the stretch factor for a specific column.
       *
       * @param index - The column index of interest.
       *
       * @param value - The stretch factor for the column.
       *
       * #### Notes
       * This is a no-op if the index is out of range.
       */
      GridLayout.prototype.setColumnStretch = function (index, value) {
          // Look up the column sizer.
          var sizer = this._columnSizers[index];
          // Bail if the index is out of range.
          if (!sizer) {
              return;
          }
          // Clamp the value to the allowed range.
          value = Private$b.clampValue(value);
          // Bail if the stretch does not change.
          if (sizer.stretch === value) {
              return;
          }
          // Update the sizer stretch.
          sizer.stretch = value;
          // Schedule an update of the parent.
          if (this.parent) {
              this.parent.update();
          }
      };
      /**
       * Create an iterator over the widgets in the layout.
       *
       * @returns A new iterator over the widgets in the layout.
       */
      GridLayout.prototype.iter = function () {
          return map$1(this._items, function (item) { return item.widget; });
      };
      /**
       * Add a widget to the grid layout.
       *
       * @param widget - The widget to add to the layout.
       *
       * #### Notes
       * If the widget is already contained in the layout, this is no-op.
       */
      GridLayout.prototype.addWidget = function (widget) {
          // Look up the index for the widget.
          var i = ArrayExt.findFirstIndex(this._items, function (it) { return it.widget === widget; });
          // Bail if the widget is already in the layout.
          if (i !== -1) {
              return;
          }
          // Add the widget to the layout.
          this._items.push(new LayoutItem(widget));
          // Attach the widget to the parent.
          if (this.parent) {
              this.attachWidget(widget);
          }
      };
      /**
       * Remove a widget from the grid layout.
       *
       * @param widget - The widget to remove from the layout.
       *
       * #### Notes
       * A widget is automatically removed from the layout when its `parent`
       * is set to `null`. This method should only be invoked directly when
       * removing a widget from a layout which has yet to be installed on a
       * parent widget.
       *
       * This method does *not* modify the widget's `parent`.
       */
      GridLayout.prototype.removeWidget = function (widget) {
          // Look up the index for the widget.
          var i = ArrayExt.findFirstIndex(this._items, function (it) { return it.widget === widget; });
          // Bail if the widget is not in the layout.
          if (i === -1) {
              return;
          }
          // Remove the widget from the layout.
          var item = ArrayExt.removeAt(this._items, i);
          // Detach the widget from the parent.
          if (this.parent) {
              this.detachWidget(widget);
          }
          // Dispose the layout item.
          item.dispose();
      };
      /**
       * Perform layout initialization which requires the parent widget.
       */
      GridLayout.prototype.init = function () {
          var _this = this;
          _super.prototype.init.call(this);
          each$1(this, function (widget) { _this.attachWidget(widget); });
      };
      /**
       * Attach a widget to the parent's DOM node.
       *
       * @param widget - The widget to attach to the parent.
       */
      GridLayout.prototype.attachWidget = function (widget) {
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Add the widget's node to the parent.
          this.parent.node.appendChild(widget.node);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Detach a widget from the parent's DOM node.
       *
       * @param widget - The widget to detach from the parent.
       */
      GridLayout.prototype.detachWidget = function (widget) {
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'before-show'` message.
       */
      GridLayout.prototype.onBeforeShow = function (msg) {
          _super.prototype.onBeforeShow.call(this, msg);
          this.parent.update();
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      GridLayout.prototype.onBeforeAttach = function (msg) {
          _super.prototype.onBeforeAttach.call(this, msg);
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-shown'` message.
       */
      GridLayout.prototype.onChildShown = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-hidden'` message.
       */
      GridLayout.prototype.onChildHidden = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'resize'` message.
       */
      GridLayout.prototype.onResize = function (msg) {
          if (this.parent.isVisible) {
              this._update(msg.width, msg.height);
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      GridLayout.prototype.onUpdateRequest = function (msg) {
          if (this.parent.isVisible) {
              this._update(-1, -1);
          }
      };
      /**
       * A message handler invoked on a `'fit-request'` message.
       */
      GridLayout.prototype.onFitRequest = function (msg) {
          if (this.parent.isAttached) {
              this._fit();
          }
      };
      /**
       * Fit the layout to the total size required by the widgets.
       */
      GridLayout.prototype._fit = function () {
          // Reset the min sizes of the sizers.
          for (var i = 0, n = this.rowCount; i < n; ++i) {
              this._rowSizers[i].minSize = 0;
          }
          for (var i = 0, n = this.columnCount; i < n; ++i) {
              this._columnSizers[i].minSize = 0;
          }
          // Filter for the visible layout items.
          var items = this._items.filter(function (it) { return !it.isHidden; });
          // Fit the layout items.
          for (var i = 0, n = items.length; i < n; ++i) {
              items[i].fit();
          }
          // Get the max row and column index.
          var maxRow = this.rowCount - 1;
          var maxCol = this.columnCount - 1;
          // Sort the items by row span.
          items.sort(Private$b.rowSpanCmp);
          // Update the min sizes of the row sizers.
          for (var i = 0, n = items.length; i < n; ++i) {
              // Fetch the item.
              var item = items[i];
              // Get the row bounds for the item.
              var config = GridLayout.getCellConfig(item.widget);
              var r1 = Math.min(config.row, maxRow);
              var r2 = Math.min(config.row + config.rowSpan - 1, maxRow);
              // Distribute the minimum height to the sizers as needed.
              Private$b.distributeMin(this._rowSizers, r1, r2, item.minHeight);
          }
          // Sort the items by column span.
          items.sort(Private$b.columnSpanCmp);
          // Update the min sizes of the column sizers.
          for (var i = 0, n = items.length; i < n; ++i) {
              // Fetch the item.
              var item = items[i];
              // Get the column bounds for the item.
              var config = GridLayout.getCellConfig(item.widget);
              var c1 = Math.min(config.column, maxCol);
              var c2 = Math.min(config.column + config.columnSpan - 1, maxCol);
              // Distribute the minimum width to the sizers as needed.
              Private$b.distributeMin(this._columnSizers, c1, c2, item.minWidth);
          }
          // If no size constraint is needed, just update the parent.
          if (this.fitPolicy === 'set-no-constraint') {
              MessageLoop.sendMessage(this.parent, Widget.Msg.UpdateRequest);
              return;
          }
          // Set up the computed min size.
          var minH = maxRow * this._rowSpacing;
          var minW = maxCol * this._columnSpacing;
          // Add the sizer minimums to the computed min size.
          for (var i = 0, n = this.rowCount; i < n; ++i) {
              minH += this._rowSizers[i].minSize;
          }
          for (var i = 0, n = this.columnCount; i < n; ++i) {
              minW += this._columnSizers[i].minSize;
          }
          // Update the box sizing and add it to the computed min size.
          var box = this._box = ElementExt.boxSizing(this.parent.node);
          minW += box.horizontalSum;
          minH += box.verticalSum;
          // Update the parent's min size constraints.
          var style = this.parent.node.style;
          style.minWidth = minW + "px";
          style.minHeight = minH + "px";
          // Set the dirty flag to ensure only a single update occurs.
          this._dirty = true;
          // Notify the ancestor that it should fit immediately. This may
          // cause a resize of the parent, fulfilling the required update.
          if (this.parent.parent) {
              MessageLoop.sendMessage(this.parent.parent, Widget.Msg.FitRequest);
          }
          // If the dirty flag is still set, the parent was not resized.
          // Trigger the required update on the parent widget immediately.
          if (this._dirty) {
              MessageLoop.sendMessage(this.parent, Widget.Msg.UpdateRequest);
          }
      };
      /**
       * Update the layout position and size of the widgets.
       *
       * The parent offset dimensions should be `-1` if unknown.
       */
      GridLayout.prototype._update = function (offsetWidth, offsetHeight) {
          // Clear the dirty flag to indicate the update occurred.
          this._dirty = false;
          // Measure the parent if the offset dimensions are unknown.
          if (offsetWidth < 0) {
              offsetWidth = this.parent.node.offsetWidth;
          }
          if (offsetHeight < 0) {
              offsetHeight = this.parent.node.offsetHeight;
          }
          // Ensure the parent box sizing data is computed.
          if (!this._box) {
              this._box = ElementExt.boxSizing(this.parent.node);
          }
          // Compute the layout area adjusted for border and padding.
          var top = this._box.paddingTop;
          var left = this._box.paddingLeft;
          var width = offsetWidth - this._box.horizontalSum;
          var height = offsetHeight - this._box.verticalSum;
          // Get the max row and column index.
          var maxRow = this.rowCount - 1;
          var maxCol = this.columnCount - 1;
          // Compute the total fixed row and column space.
          var fixedRowSpace = maxRow * this._rowSpacing;
          var fixedColSpace = maxCol * this._columnSpacing;
          // Distribute the available space to the box sizers.
          BoxEngine.calc(this._rowSizers, Math.max(0, height - fixedRowSpace));
          BoxEngine.calc(this._columnSizers, Math.max(0, width - fixedColSpace));
          // Update the row start positions.
          for (var i = 0, pos = top, n = this.rowCount; i < n; ++i) {
              this._rowStarts[i] = pos;
              pos += this._rowSizers[i].size + this._rowSpacing;
          }
          // Update the column start positions.
          for (var i = 0, pos = left, n = this.columnCount; i < n; ++i) {
              this._columnStarts[i] = pos;
              pos += this._columnSizers[i].size + this._columnSpacing;
          }
          // Update the geometry of the layout items.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item.
              var item = this._items[i];
              // Ignore hidden items.
              if (item.isHidden) {
                  continue;
              }
              // Fetch the cell bounds for the widget.
              var config = GridLayout.getCellConfig(item.widget);
              var r1 = Math.min(config.row, maxRow);
              var c1 = Math.min(config.column, maxCol);
              var r2 = Math.min(config.row + config.rowSpan - 1, maxRow);
              var c2 = Math.min(config.column + config.columnSpan - 1, maxCol);
              // Compute the cell geometry.
              var x = this._columnStarts[c1];
              var y = this._rowStarts[r1];
              var w = this._columnStarts[c2] + this._columnSizers[c2].size - x;
              var h = this._rowStarts[r2] + this._rowSizers[r2].size - y;
              // Update the geometry of the layout item.
              item.update(x, y, w, h);
          }
      };
      return GridLayout;
  }(Layout));
  /**
   * The namespace for the `GridLayout` class statics.
   */
  (function (GridLayout) {
      /**
       * Get the cell config for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The cell config for the widget.
       */
      function getCellConfig(widget) {
          return Private$b.cellConfigProperty.get(widget);
      }
      GridLayout.getCellConfig = getCellConfig;
      /**
       * Set the cell config for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the cell config.
       */
      function setCellConfig(widget, value) {
          Private$b.cellConfigProperty.set(widget, Private$b.normalizeConfig(value));
      }
      GridLayout.setCellConfig = setCellConfig;
  })(GridLayout || (GridLayout = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$b;
  (function (Private) {
      /**
       * The property descriptor for the widget cell config.
       */
      Private.cellConfigProperty = new AttachedProperty({
          name: 'cellConfig',
          create: function () { return ({ row: 0, column: 0, rowSpan: 1, columnSpan: 1 }); },
          changed: onChildCellConfigChanged
      });
      /**
       * Normalize a partial cell config object.
       */
      function normalizeConfig(config) {
          var row = Math.max(0, Math.floor(config.row || 0));
          var column = Math.max(0, Math.floor(config.column || 0));
          var rowSpan = Math.max(1, Math.floor(config.rowSpan || 0));
          var columnSpan = Math.max(1, Math.floor(config.columnSpan || 0));
          return { row: row, column: column, rowSpan: rowSpan, columnSpan: columnSpan };
      }
      Private.normalizeConfig = normalizeConfig;
      /**
       * Clamp a value to an integer >= 0.
       */
      function clampValue(value) {
          return Math.max(0, Math.floor(value));
      }
      Private.clampValue = clampValue;
      /**
       * A sort comparison function for row spans.
       */
      function rowSpanCmp(a, b) {
          var c1 = Private.cellConfigProperty.get(a.widget);
          var c2 = Private.cellConfigProperty.get(b.widget);
          return c1.rowSpan - c2.rowSpan;
      }
      Private.rowSpanCmp = rowSpanCmp;
      /**
       * A sort comparison function for column spans.
       */
      function columnSpanCmp(a, b) {
          var c1 = Private.cellConfigProperty.get(a.widget);
          var c2 = Private.cellConfigProperty.get(b.widget);
          return c1.columnSpan - c2.columnSpan;
      }
      Private.columnSpanCmp = columnSpanCmp;
      /**
       * Reallocate the box sizers for the given grid dimensions.
       */
      function reallocSizers(sizers, count) {
          // Coerce the count to the valid range.
          count = Math.max(1, Math.floor(count));
          // Add the missing sizers.
          while (sizers.length < count) {
              sizers.push(new BoxSizer());
          }
          // Remove the extra sizers.
          if (sizers.length > count) {
              sizers.length = count;
          }
      }
      Private.reallocSizers = reallocSizers;
      /**
       * Distribute a min size constraint across a range of sizers.
       */
      function distributeMin(sizers, i1, i2, minSize) {
          // Sanity check the indices.
          if (i2 < i1) {
              return;
          }
          // Handle the simple case of no cell span.
          if (i1 === i2) {
              var sizer = sizers[i1];
              sizer.minSize = Math.max(sizer.minSize, minSize);
              return;
          }
          // Compute the total current min size of the span.
          var totalMin = 0;
          for (var i = i1; i <= i2; ++i) {
              totalMin += sizers[i].minSize;
          }
          // Do nothing if the total is greater than the required.
          if (totalMin >= minSize) {
              return;
          }
          // Compute the portion of the space to allocate to each sizer.
          var portion = (minSize - totalMin) / (i2 - i1 + 1);
          // Add the portion to each sizer.
          for (var i = i1; i <= i2; ++i) {
              sizers[i].minSize += portion;
          }
      }
      Private.distributeMin = distributeMin;
      /**
       * The change handler for the child cell config property.
       */
      function onChildCellConfigChanged(child) {
          if (child.parent && child.parent.layout instanceof GridLayout) {
              child.parent.fit();
          }
      }
  })(Private$b || (Private$b = {}));

  /**
   * A widget which displays menus as a canonical menu bar.
   */
  var MenuBar = /** @class */ (function (_super) {
      __extends$5(MenuBar, _super);
      /**
       * Construct a new menu bar.
       *
       * @param options - The options for initializing the menu bar.
       */
      function MenuBar(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, { node: Private$c.createNode() }) || this;
          _this._activeIndex = -1;
          _this._menus = [];
          _this._childMenu = null;
          _this.addClass('lm-MenuBar');
          /* <DEPRECATED> */
          _this.addClass('p-MenuBar');
          /* </DEPRECATED> */
          _this.setFlag(Widget.Flag.DisallowLayout);
          _this.renderer = options.renderer || MenuBar.defaultRenderer;
          return _this;
      }
      /**
       * Dispose of the resources held by the widget.
       */
      MenuBar.prototype.dispose = function () {
          this._closeChildMenu();
          this._menus.length = 0;
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(MenuBar.prototype, "childMenu", {
          /**
           * The child menu of the menu bar.
           *
           * #### Notes
           * This will be `null` if the menu bar does not have an open menu.
           */
          get: function () {
              return this._childMenu;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(MenuBar.prototype, "contentNode", {
          /**
           * Get the menu bar content node.
           *
           * #### Notes
           * This is the node which holds the menu title nodes.
           *
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-MenuBar-content')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(MenuBar.prototype, "activeMenu", {
          /**
           * Get the currently active menu.
           */
          get: function () {
              return this._menus[this._activeIndex] || null;
          },
          /**
           * Set the currently active menu.
           *
           * #### Notes
           * If the menu does not exist, the menu will be set to `null`.
           */
          set: function (value) {
              this.activeIndex = value ? this._menus.indexOf(value) : -1;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(MenuBar.prototype, "activeIndex", {
          /**
           * Get the index of the currently active menu.
           *
           * #### Notes
           * This will be `-1` if no menu is active.
           */
          get: function () {
              return this._activeIndex;
          },
          /**
           * Set the index of the currently active menu.
           *
           * #### Notes
           * If the menu cannot be activated, the index will be set to `-1`.
           */
          set: function (value) {
              // Adjust the value for an out of range index.
              if (value < 0 || value >= this._menus.length) {
                  value = -1;
              }
              // Bail early if the index will not change.
              if (this._activeIndex === value) {
                  return;
              }
              // Update the active index.
              this._activeIndex = value;
              // Schedule an update of the items.
              this.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(MenuBar.prototype, "menus", {
          /**
           * A read-only array of the menus in the menu bar.
           */
          get: function () {
              return this._menus;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Open the active menu and activate its first menu item.
       *
       * #### Notes
       * If there is no active menu, this is a no-op.
       */
      MenuBar.prototype.openActiveMenu = function () {
          // Bail early if there is no active item.
          if (this._activeIndex === -1) {
              return;
          }
          // Open the child menu.
          this._openChildMenu();
          // Activate the first item in the child menu.
          if (this._childMenu) {
              this._childMenu.activeIndex = -1;
              this._childMenu.activateNextItem();
          }
      };
      /**
       * Add a menu to the end of the menu bar.
       *
       * @param menu - The menu to add to the menu bar.
       *
       * #### Notes
       * If the menu is already added to the menu bar, it will be moved.
       */
      MenuBar.prototype.addMenu = function (menu) {
          this.insertMenu(this._menus.length, menu);
      };
      /**
       * Insert a menu into the menu bar at the specified index.
       *
       * @param index - The index at which to insert the menu.
       *
       * @param menu - The menu to insert into the menu bar.
       *
       * #### Notes
       * The index will be clamped to the bounds of the menus.
       *
       * If the menu is already added to the menu bar, it will be moved.
       */
      MenuBar.prototype.insertMenu = function (index, menu) {
          // Close the child menu before making changes.
          this._closeChildMenu();
          // Look up the index of the menu.
          var i = this._menus.indexOf(menu);
          // Clamp the insert index to the array bounds.
          var j = Math.max(0, Math.min(index, this._menus.length));
          // If the menu is not in the array, insert it.
          if (i === -1) {
              // Insert the menu into the array.
              ArrayExt.insert(this._menus, j, menu);
              // Add the styling class to the menu.
              menu.addClass('lm-MenuBar-menu');
              /* <DEPRECATED> */
              menu.addClass('p-MenuBar-menu');
              /* </DEPRECATED> */
              // Connect to the menu signals.
              menu.aboutToClose.connect(this._onMenuAboutToClose, this);
              menu.menuRequested.connect(this._onMenuMenuRequested, this);
              menu.title.changed.connect(this._onTitleChanged, this);
              // Schedule an update of the items.
              this.update();
              // There is nothing more to do.
              return;
          }
          // Otherwise, the menu exists in the array and should be moved.
          // Adjust the index if the location is at the end of the array.
          if (j === this._menus.length) {
              j--;
          }
          // Bail if there is no effective move.
          if (i === j) {
              return;
          }
          // Move the menu to the new locations.
          ArrayExt.move(this._menus, i, j);
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Remove a menu from the menu bar.
       *
       * @param menu - The menu to remove from the menu bar.
       *
       * #### Notes
       * This is a no-op if the menu is not in the menu bar.
       */
      MenuBar.prototype.removeMenu = function (menu) {
          this.removeMenuAt(this._menus.indexOf(menu));
      };
      /**
       * Remove the menu at a given index from the menu bar.
       *
       * @param index - The index of the menu to remove.
       *
       * #### Notes
       * This is a no-op if the index is out of range.
       */
      MenuBar.prototype.removeMenuAt = function (index) {
          // Close the child menu before making changes.
          this._closeChildMenu();
          // Remove the menu from the array.
          var menu = ArrayExt.removeAt(this._menus, index);
          // Bail if the index is out of range.
          if (!menu) {
              return;
          }
          // Disconnect from the menu signals.
          menu.aboutToClose.disconnect(this._onMenuAboutToClose, this);
          menu.menuRequested.disconnect(this._onMenuMenuRequested, this);
          menu.title.changed.disconnect(this._onTitleChanged, this);
          // Remove the styling class from the menu.
          menu.removeClass('lm-MenuBar-menu');
          /* <DEPRECATED> */
          menu.removeClass('p-MenuBar-menu');
          /* </DEPRECATED> */
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Remove all menus from the menu bar.
       */
      MenuBar.prototype.clearMenus = function () {
          // Bail if there is nothing to remove.
          if (this._menus.length === 0) {
              return;
          }
          // Close the child menu before making changes.
          this._closeChildMenu();
          // Disconnect from the menu signals and remove the styling class.
          for (var _i = 0, _a = this._menus; _i < _a.length; _i++) {
              var menu = _a[_i];
              menu.aboutToClose.disconnect(this._onMenuAboutToClose, this);
              menu.menuRequested.disconnect(this._onMenuMenuRequested, this);
              menu.title.changed.disconnect(this._onTitleChanged, this);
              menu.removeClass('lm-MenuBar-menu');
              /* <DEPRECATED> */
              menu.removeClass('p-MenuBar-menu');
              /* </DEPRECATED> */
          }
          // Clear the menus array.
          this._menus.length = 0;
          // Schedule an update of the items.
          this.update();
      };
      /**
       * Handle the DOM events for the menu bar.
       *
       * @param event - The DOM event sent to the menu bar.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the menu bar's DOM nodes. It
       * should not be called directly by user code.
       */
      MenuBar.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'mousedown':
                  this._evtMouseDown(event);
                  break;
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseleave':
                  this._evtMouseLeave(event);
                  break;
              case 'contextmenu':
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      MenuBar.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('keydown', this);
          this.node.addEventListener('mousedown', this);
          this.node.addEventListener('mousemove', this);
          this.node.addEventListener('mouseleave', this);
          this.node.addEventListener('contextmenu', this);
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       */
      MenuBar.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('keydown', this);
          this.node.removeEventListener('mousedown', this);
          this.node.removeEventListener('mousemove', this);
          this.node.removeEventListener('mouseleave', this);
          this.node.removeEventListener('contextmenu', this);
          this._closeChildMenu();
      };
      /**
       * A message handler invoked on an `'activate-request'` message.
       */
      MenuBar.prototype.onActivateRequest = function (msg) {
          if (this.isAttached) {
              this.node.focus();
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      MenuBar.prototype.onUpdateRequest = function (msg) {
          var menus = this._menus;
          var renderer = this.renderer;
          var activeIndex = this._activeIndex;
          var content = new Array(menus.length);
          for (var i = 0, n = menus.length; i < n; ++i) {
              var title = menus[i].title;
              var active = i === activeIndex;
              content[i] = renderer.renderItem({ title: title, active: active });
          }
          VirtualDOM.render(content, this.contentNode);
      };
      /**
       * Handle the `'keydown'` event for the menu bar.
       */
      MenuBar.prototype._evtKeyDown = function (event) {
          // A menu bar handles all keydown events.
          event.preventDefault();
          event.stopPropagation();
          // Fetch the key code for the event.
          var kc = event.keyCode;
          // Enter, Up Arrow, Down Arrow
          if (kc === 13 || kc === 38 || kc === 40) {
              this.openActiveMenu();
              return;
          }
          // Escape
          if (kc === 27) {
              this._closeChildMenu();
              this.activeIndex = -1;
              this.node.blur();
              return;
          }
          // Left Arrow
          if (kc === 37) {
              var i = this._activeIndex;
              var n = this._menus.length;
              this.activeIndex = i === 0 ? n - 1 : i - 1;
              return;
          }
          // Right Arrow
          if (kc === 39) {
              var i = this._activeIndex;
              var n = this._menus.length;
              this.activeIndex = i === n - 1 ? 0 : i + 1;
              return;
          }
          // Get the pressed key character.
          var key = getKeyboardLayout().keyForKeydownEvent(event);
          // Bail if the key is not valid.
          if (!key) {
              return;
          }
          // Search for the next best matching mnemonic item.
          var start = this._activeIndex + 1;
          var result = Private$c.findMnemonic(this._menus, key, start);
          // Handle the requested mnemonic based on the search results.
          // If exactly one mnemonic is matched, that menu is opened.
          // Otherwise, the next mnemonic is activated if available,
          // followed by the auto mnemonic if available.
          if (result.index !== -1 && !result.multiple) {
              this.activeIndex = result.index;
              this.openActiveMenu();
          }
          else if (result.index !== -1) {
              this.activeIndex = result.index;
          }
          else if (result.auto !== -1) {
              this.activeIndex = result.auto;
          }
      };
      /**
       * Handle the `'mousedown'` event for the menu bar.
       */
      MenuBar.prototype._evtMouseDown = function (event) {
          // Bail if the mouse press was not on the menu bar. This can occur
          // when the document listener is installed for an active menu bar.
          if (!ElementExt.hitTest(this.node, event.clientX, event.clientY)) {
              return;
          }
          // Stop the propagation of the event. Immediate propagation is
          // also stopped so that an open menu does not handle the event.
          event.preventDefault();
          event.stopPropagation();
          event.stopImmediatePropagation();
          // Check if the mouse is over one of the menu items.
          var index = ArrayExt.findFirstIndex(this.contentNode.children, function (node) {
              return ElementExt.hitTest(node, event.clientX, event.clientY);
          });
          // If the press was not on an item, close the child menu.
          if (index === -1) {
              this._closeChildMenu();
              return;
          }
          // If the press was not the left mouse button, do nothing further.
          if (event.button !== 0) {
              return;
          }
          // Otherwise, toggle the open state of the child menu.
          if (this._childMenu) {
              this._closeChildMenu();
              this.activeIndex = index;
          }
          else {
              this.activeIndex = index;
              this._openChildMenu();
          }
      };
      /**
       * Handle the `'mousemove'` event for the menu bar.
       */
      MenuBar.prototype._evtMouseMove = function (event) {
          // Check if the mouse is over one of the menu items.
          var index = ArrayExt.findFirstIndex(this.contentNode.children, function (node) {
              return ElementExt.hitTest(node, event.clientX, event.clientY);
          });
          // Bail early if the active index will not change.
          if (index === this._activeIndex) {
              return;
          }
          // Bail early if a child menu is open and the mouse is not over
          // an item. This allows the child menu to be kept open when the
          // mouse is over the empty part of the menu bar.
          if (index === -1 && this._childMenu) {
              return;
          }
          // Update the active index to the hovered item.
          this.activeIndex = index;
          // Open the new menu if a menu is already open.
          if (this._childMenu) {
              this._openChildMenu();
          }
      };
      /**
       * Handle the `'mouseleave'` event for the menu bar.
       */
      MenuBar.prototype._evtMouseLeave = function (event) {
          // Reset the active index if there is no open menu.
          if (!this._childMenu) {
              this.activeIndex = -1;
          }
      };
      /**
       * Open the child menu at the active index immediately.
       *
       * If a different child menu is already open, it will be closed,
       * even if there is no active menu.
       */
      MenuBar.prototype._openChildMenu = function () {
          // If there is no active menu, close the current menu.
          var newMenu = this.activeMenu;
          if (!newMenu) {
              this._closeChildMenu();
              return;
          }
          // Bail if there is no effective menu change.
          var oldMenu = this._childMenu;
          if (oldMenu === newMenu) {
              return;
          }
          // Swap the internal menu reference.
          this._childMenu = newMenu;
          // Close the current menu, or setup for the new menu.
          if (oldMenu) {
              oldMenu.close();
          }
          else {
              this.addClass('lm-mod-active');
              /* <DEPRECATED> */
              this.addClass('p-mod-active');
              /* </DEPRECATED> */
              document.addEventListener('mousedown', this, true);
          }
          // Ensure the menu bar is updated and look up the item node.
          MessageLoop.sendMessage(this, Widget.Msg.UpdateRequest);
          var itemNode = this.contentNode.children[this._activeIndex];
          // Get the positioning data for the new menu.
          var _a = itemNode.getBoundingClientRect(), left = _a.left, bottom = _a.bottom;
          // Open the new menu at the computed location.
          newMenu.open(left, bottom, { forceX: true, forceY: true });
      };
      /**
       * Close the child menu immediately.
       *
       * This is a no-op if a child menu is not open.
       */
      MenuBar.prototype._closeChildMenu = function () {
          // Bail if no child menu is open.
          if (!this._childMenu) {
              return;
          }
          // Remove the active class from the menu bar.
          this.removeClass('lm-mod-active');
          /* <DEPRECATED> */
          this.removeClass('p-mod-active');
          /* </DEPRECATED> */
          // Remove the document listeners.
          document.removeEventListener('mousedown', this, true);
          // Clear the internal menu reference.
          var menu = this._childMenu;
          this._childMenu = null;
          // Close the menu.
          menu.close();
          // Reset the active index.
          this.activeIndex = -1;
      };
      /**
       * Handle the `aboutToClose` signal of a menu.
       */
      MenuBar.prototype._onMenuAboutToClose = function (sender) {
          // Bail if the sender is not the child menu.
          if (sender !== this._childMenu) {
              return;
          }
          // Remove the active class from the menu bar.
          this.removeClass('lm-mod-active');
          /* <DEPRECATED> */
          this.removeClass('p-mod-active');
          /* </DEPRECATED> */
          // Remove the document listeners.
          document.removeEventListener('mousedown', this, true);
          // Clear the internal menu reference.
          this._childMenu = null;
          // Reset the active index.
          this.activeIndex = -1;
      };
      /**
       * Handle the `menuRequested` signal of a child menu.
       */
      MenuBar.prototype._onMenuMenuRequested = function (sender, args) {
          // Bail if the sender is not the child menu.
          if (sender !== this._childMenu) {
              return;
          }
          // Look up the active index and menu count.
          var i = this._activeIndex;
          var n = this._menus.length;
          // Active the next requested index.
          switch (args) {
              case 'next':
                  this.activeIndex = i === n - 1 ? 0 : i + 1;
                  break;
              case 'previous':
                  this.activeIndex = i === 0 ? n - 1 : i - 1;
                  break;
          }
          // Open the active menu.
          this.openActiveMenu();
      };
      /**
       * Handle the `changed` signal of a title object.
       */
      MenuBar.prototype._onTitleChanged = function () {
          this.update();
      };
      return MenuBar;
  }(Widget));
  /**
   * The namespace for the `MenuBar` class statics.
   */
  (function (MenuBar) {
      /**
       * The default implementation of `IRenderer`.
       *
       * #### Notes
       * Subclasses are free to reimplement rendering methods as needed.
       */
      var Renderer = /** @class */ (function () {
          /**
           * Construct a new renderer.
           */
          function Renderer() {
          }
          /**
           * Render the virtual element for a menu bar item.
           *
           * @param data - The data to use for rendering the item.
           *
           * @returns A virtual element representing the item.
           */
          Renderer.prototype.renderItem = function (data) {
              var className = this.createItemClass(data);
              var dataset = this.createItemDataset(data);
              return (h.li({ className: className, dataset: dataset }, this.renderIcon(data), this.renderLabel(data)));
          };
          /**
           * Render the icon element for a menu bar item.
           *
           * @param data - The data to use for rendering the icon.
           *
           * @returns A virtual element representing the item icon.
           */
          Renderer.prototype.renderIcon = function (data) {
              var className = this.createIconClass(data);
              /* <DEPRECATED> */
              if (typeof data.title.icon === 'string') {
                  return h.div({ className: className }, data.title.iconLabel);
              }
              /* </DEPRECATED> */
              // if data.title.icon is undefined, it will be ignored
              return h.div({ className: className }, data.title.icon, data.title.iconLabel);
          };
          /**
           * Render the label element for a menu item.
           *
           * @param data - The data to use for rendering the label.
           *
           * @returns A virtual element representing the item label.
           */
          Renderer.prototype.renderLabel = function (data) {
              var content = this.formatLabel(data);
              return h.div({ className: 'lm-MenuBar-itemLabel'
                      /* <DEPRECATED> */
                      + ' p-MenuBar-itemLabel'
                  /* </DEPRECATED> */
              }, content);
          };
          /**
           * Create the class name for the menu bar item.
           *
           * @param data - The data to use for the class name.
           *
           * @returns The full class name for the menu item.
           */
          Renderer.prototype.createItemClass = function (data) {
              var name = 'lm-MenuBar-item';
              /* <DEPRECATED> */
              name += ' p-MenuBar-item';
              /* </DEPRECATED> */
              if (data.title.className) {
                  name += " " + data.title.className;
              }
              if (data.active) {
                  name += ' lm-mod-active';
                  /* <DEPRECATED> */
                  name += ' p-mod-active';
                  /* </DEPRECATED> */
              }
              return name;
          };
          /**
           * Create the dataset for a menu bar item.
           *
           * @param data - The data to use for the item.
           *
           * @returns The dataset for the menu bar item.
           */
          Renderer.prototype.createItemDataset = function (data) {
              return data.title.dataset;
          };
          /**
           * Create the class name for the menu bar item icon.
           *
           * @param data - The data to use for the class name.
           *
           * @returns The full class name for the item icon.
           */
          Renderer.prototype.createIconClass = function (data) {
              var name = 'lm-MenuBar-itemIcon';
              /* <DEPRECATED> */
              name += ' p-MenuBar-itemIcon';
              /* </DEPRECATED> */
              var extra = data.title.iconClass;
              return extra ? name + " " + extra : name;
          };
          /**
           * Create the render content for the label node.
           *
           * @param data - The data to use for the label content.
           *
           * @returns The content to add to the label node.
           */
          Renderer.prototype.formatLabel = function (data) {
              // Fetch the label text and mnemonic index.
              var _a = data.title, label = _a.label, mnemonic = _a.mnemonic;
              // If the index is out of range, do not modify the label.
              if (mnemonic < 0 || mnemonic >= label.length) {
                  return label;
              }
              // Split the label into parts.
              var prefix = label.slice(0, mnemonic);
              var suffix = label.slice(mnemonic + 1);
              var char = label[mnemonic];
              // Wrap the mnemonic character in a span.
              var span = h.span({
                  className: 'lm-MenuBar-itemMnemonic'
                      /* <DEPRECATED> */
                      + ' p-MenuBar-itemMnemonic'
                  /* </DEPRECATED> */
              }, char);
              // Return the content parts.
              return [prefix, span, suffix];
          };
          return Renderer;
      }());
      MenuBar.Renderer = Renderer;
      /**
       * The default `Renderer` instance.
       */
      MenuBar.defaultRenderer = new Renderer();
  })(MenuBar || (MenuBar = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$c;
  (function (Private) {
      /**
       * Create the DOM node for a menu bar.
       */
      function createNode() {
          var node = document.createElement('div');
          var content = document.createElement('ul');
          content.className = 'lm-MenuBar-content';
          /* <DEPRECATED> */
          content.classList.add('p-MenuBar-content');
          /* </DEPRECATED> */
          node.appendChild(content);
          node.tabIndex = -1;
          return node;
      }
      Private.createNode = createNode;
      /**
       * Find the best matching mnemonic item.
       *
       * The search starts at the given index and wraps around.
       */
      function findMnemonic(menus, key, start) {
          // Setup the result variables.
          var index = -1;
          var auto = -1;
          var multiple = false;
          // Normalize the key to upper case.
          var upperKey = key.toUpperCase();
          // Search the items from the given start index.
          for (var i = 0, n = menus.length; i < n; ++i) {
              // Compute the wrapped index.
              var k = (i + start) % n;
              // Look up the menu title.
              var title = menus[k].title;
              // Ignore titles with an empty label.
              if (title.label.length === 0) {
                  continue;
              }
              // Look up the mnemonic index for the label.
              var mn = title.mnemonic;
              // Handle a valid mnemonic index.
              if (mn >= 0 && mn < title.label.length) {
                  if (title.label[mn].toUpperCase() === upperKey) {
                      if (index === -1) {
                          index = k;
                      }
                      else {
                          multiple = true;
                      }
                  }
                  continue;
              }
              // Finally, handle the auto index if possible.
              if (auto === -1 && title.label[0].toUpperCase() === upperKey) {
                  auto = k;
              }
          }
          // Return the search results.
          return { index: index, multiple: multiple, auto: auto };
      }
      Private.findMnemonic = findMnemonic;
  })(Private$c || (Private$c = {}));

  /**
   * A widget which implements a canonical scroll bar.
   */
  var ScrollBar = /** @class */ (function (_super) {
      __extends$5(ScrollBar, _super);
      /**
       * Construct a new scroll bar.
       *
       * @param options - The options for initializing the scroll bar.
       */
      function ScrollBar(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, { node: Private$d.createNode() }) || this;
          /**
           * A timeout callback for repeating the mouse press.
           */
          _this._onRepeat = function () {
              // Clear the repeat timer id.
              _this._repeatTimer = -1;
              // Bail if the mouse has been released.
              if (!_this._pressData) {
                  return;
              }
              // Look up the part that was pressed.
              var part = _this._pressData.part;
              // Bail if the thumb was pressed.
              if (part === 'thumb') {
                  return;
              }
              // Schedule the timer for another repeat.
              _this._repeatTimer = window.setTimeout(_this._onRepeat, 20);
              // Get the current mouse position.
              var mouseX = _this._pressData.mouseX;
              var mouseY = _this._pressData.mouseY;
              // Handle a decrement button repeat.
              if (part === 'decrement') {
                  // Bail if the mouse is not over the button.
                  if (!ElementExt.hitTest(_this.decrementNode, mouseX, mouseY)) {
                      return;
                  }
                  // Emit the step requested signal.
                  _this._stepRequested.emit('decrement');
                  // Finished.
                  return;
              }
              // Handle an increment button repeat.
              if (part === 'increment') {
                  // Bail if the mouse is not over the button.
                  if (!ElementExt.hitTest(_this.incrementNode, mouseX, mouseY)) {
                      return;
                  }
                  // Emit the step requested signal.
                  _this._stepRequested.emit('increment');
                  // Finished.
                  return;
              }
              // Handle a track repeat.
              if (part === 'track') {
                  // Bail if the mouse is not over the track.
                  if (!ElementExt.hitTest(_this.trackNode, mouseX, mouseY)) {
                      return;
                  }
                  // Fetch the thumb node.
                  var thumbNode = _this.thumbNode;
                  // Bail if the mouse is over the thumb.
                  if (ElementExt.hitTest(thumbNode, mouseX, mouseY)) {
                      return;
                  }
                  // Fetch the client rect for the thumb.
                  var thumbRect = thumbNode.getBoundingClientRect();
                  // Determine the direction for the page request.
                  var dir = void 0;
                  if (_this._orientation === 'horizontal') {
                      dir = mouseX < thumbRect.left ? 'decrement' : 'increment';
                  }
                  else {
                      dir = mouseY < thumbRect.top ? 'decrement' : 'increment';
                  }
                  // Emit the page requested signal.
                  _this._pageRequested.emit(dir);
                  // Finished.
                  return;
              }
          };
          _this._value = 0;
          _this._page = 10;
          _this._maximum = 100;
          _this._repeatTimer = -1;
          _this._pressData = null;
          _this._thumbMoved = new Signal(_this);
          _this._stepRequested = new Signal(_this);
          _this._pageRequested = new Signal(_this);
          _this.addClass('lm-ScrollBar');
          /* <DEPRECATED> */
          _this.addClass('p-ScrollBar');
          /* </DEPRECATED> */
          _this.setFlag(Widget.Flag.DisallowLayout);
          // Set the orientation.
          _this._orientation = options.orientation || 'vertical';
          _this.dataset['orientation'] = _this._orientation;
          // Parse the rest of the options.
          if (options.maximum !== undefined) {
              _this._maximum = Math.max(0, options.maximum);
          }
          if (options.page !== undefined) {
              _this._page = Math.max(0, options.page);
          }
          if (options.value !== undefined) {
              _this._value = Math.max(0, Math.min(options.value, _this._maximum));
          }
          return _this;
      }
      Object.defineProperty(ScrollBar.prototype, "thumbMoved", {
          /**
           * A signal emitted when the user moves the scroll thumb.
           *
           * #### Notes
           * The payload is the current value of the scroll bar.
           */
          get: function () {
              return this._thumbMoved;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "stepRequested", {
          /**
           * A signal emitted when the user clicks a step button.
           *
           * #### Notes
           * The payload is whether a decrease or increase is requested.
           */
          get: function () {
              return this._stepRequested;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "pageRequested", {
          /**
           * A signal emitted when the user clicks the scroll track.
           *
           * #### Notes
           * The payload is whether a decrease or increase is requested.
           */
          get: function () {
              return this._pageRequested;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "orientation", {
          /**
           * Get the orientation of the scroll bar.
           */
          get: function () {
              return this._orientation;
          },
          /**
           * Set the orientation of the scroll bar.
           */
          set: function (value) {
              // Do nothing if the orientation does not change.
              if (this._orientation === value) {
                  return;
              }
              // Release the mouse before making changes.
              this._releaseMouse();
              // Update the internal orientation.
              this._orientation = value;
              this.dataset['orientation'] = value;
              // Schedule an update the scroll bar.
              this.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "value", {
          /**
           * Get the current value of the scroll bar.
           */
          get: function () {
              return this._value;
          },
          /**
           * Set the current value of the scroll bar.
           *
           * #### Notes
           * The value will be clamped to the range `[0, maximum]`.
           */
          set: function (value) {
              // Clamp the value to the allowable range.
              value = Math.max(0, Math.min(value, this._maximum));
              // Do nothing if the value does not change.
              if (this._value === value) {
                  return;
              }
              // Update the internal value.
              this._value = value;
              // Schedule an update the scroll bar.
              this.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "page", {
          /**
           * Get the page size of the scroll bar.
           *
           * #### Notes
           * The page size is the amount of visible content in the scrolled
           * region, expressed in data units. It determines the size of the
           * scroll bar thumb.
           */
          get: function () {
              return this._page;
          },
          /**
           * Set the page size of the scroll bar.
           *
           * #### Notes
           * The page size will be clamped to the range `[0, Infinity]`.
           */
          set: function (value) {
              // Clamp the page size to the allowable range.
              value = Math.max(0, value);
              // Do nothing if the value does not change.
              if (this._page === value) {
                  return;
              }
              // Update the internal page size.
              this._page = value;
              // Schedule an update the scroll bar.
              this.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "maximum", {
          /**
           * Get the maximum value of the scroll bar.
           */
          get: function () {
              return this._maximum;
          },
          /**
           * Set the maximum value of the scroll bar.
           *
           * #### Notes
           * The max size will be clamped to the range `[0, Infinity]`.
           */
          set: function (value) {
              // Clamp the value to the allowable range.
              value = Math.max(0, value);
              // Do nothing if the value does not change.
              if (this._maximum === value) {
                  return;
              }
              // Update the internal values.
              this._maximum = value;
              // Clamp the current value to the new range.
              this._value = Math.min(this._value, value);
              // Schedule an update the scroll bar.
              this.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "decrementNode", {
          /**
           * The scroll bar decrement button node.
           *
           * #### Notes
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-ScrollBar-button')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "incrementNode", {
          /**
           * The scroll bar increment button node.
           *
           * #### Notes
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-ScrollBar-button')[1];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "trackNode", {
          /**
           * The scroll bar track node.
           *
           * #### Notes
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-ScrollBar-track')[0];
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(ScrollBar.prototype, "thumbNode", {
          /**
           * The scroll bar thumb node.
           *
           * #### Notes
           * Modifying this node directly can lead to undefined behavior.
           */
          get: function () {
              return this.node.getElementsByClassName('lm-ScrollBar-thumb')[0];
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Handle the DOM events for the scroll bar.
       *
       * @param event - The DOM event sent to the scroll bar.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the scroll bar's DOM node.
       *
       * This should not be called directly by user code.
       */
      ScrollBar.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'mousedown':
                  this._evtMouseDown(event);
                  break;
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseup':
                  this._evtMouseUp(event);
                  break;
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'contextmenu':
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * A method invoked on a 'before-attach' message.
       */
      ScrollBar.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('mousedown', this);
          this.update();
      };
      /**
       * A method invoked on an 'after-detach' message.
       */
      ScrollBar.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('mousedown', this);
          this._releaseMouse();
      };
      /**
       * A method invoked on an 'update-request' message.
       */
      ScrollBar.prototype.onUpdateRequest = function (msg) {
          // Convert the value and page into percentages.
          var value = this._value * 100 / this._maximum;
          var page = this._page * 100 / (this._page + this._maximum);
          // Clamp the value and page to the relevant range.
          value = Math.max(0, Math.min(value, 100));
          page = Math.max(0, Math.min(page, 100));
          // Fetch the thumb style.
          var thumbStyle = this.thumbNode.style;
          // Update the thumb style for the current orientation.
          if (this._orientation === 'horizontal') {
              thumbStyle.top = '';
              thumbStyle.height = '';
              thumbStyle.left = value + "%";
              thumbStyle.width = page + "%";
              thumbStyle.transform = "translate(" + -value + "%, 0%)";
          }
          else {
              thumbStyle.left = '';
              thumbStyle.width = '';
              thumbStyle.top = value + "%";
              thumbStyle.height = page + "%";
              thumbStyle.transform = "translate(0%, " + -value + "%)";
          }
      };
      /**
       * Handle the `'keydown'` event for the scroll bar.
       */
      ScrollBar.prototype._evtKeyDown = function (event) {
          // Stop all input events during drag.
          event.preventDefault();
          event.stopPropagation();
          // Ignore anything except the `Escape` key.
          if (event.keyCode !== 27) {
              return;
          }
          // Fetch the previous scroll value.
          var value = this._pressData ? this._pressData.value : -1;
          // Release the mouse.
          this._releaseMouse();
          // Restore the old scroll value if possible.
          if (value !== -1) {
              this._moveThumb(value);
          }
      };
      /**
       * Handle the `'mousedown'` event for the scroll bar.
       */
      ScrollBar.prototype._evtMouseDown = function (event) {
          // Do nothing if it's not a left mouse press.
          if (event.button !== 0) {
              return;
          }
          // Send an activate request to the scroll bar. This can be
          // used by message hooks to activate something relevant.
          this.activate();
          // Do nothing if the mouse is already captured.
          if (this._pressData) {
              return;
          }
          // Find the pressed scroll bar part.
          var part = Private$d.findPart(this, event.target);
          // Do nothing if the part is not of interest.
          if (!part) {
              return;
          }
          // Stop the event propagation.
          event.preventDefault();
          event.stopPropagation();
          // Override the mouse cursor.
          var override = Drag.overrideCursor('default');
          // Set up the press data.
          this._pressData = {
              part: part, override: override,
              delta: -1, value: -1,
              mouseX: event.clientX,
              mouseY: event.clientY
          };
          // Add the extra event listeners.
          document.addEventListener('mousemove', this, true);
          document.addEventListener('mouseup', this, true);
          document.addEventListener('keydown', this, true);
          document.addEventListener('contextmenu', this, true);
          // Handle a thumb press.
          if (part === 'thumb') {
              // Fetch the thumb node.
              var thumbNode = this.thumbNode;
              // Fetch the client rect for the thumb.
              var thumbRect = thumbNode.getBoundingClientRect();
              // Update the press data delta for the current orientation.
              if (this._orientation === 'horizontal') {
                  this._pressData.delta = event.clientX - thumbRect.left;
              }
              else {
                  this._pressData.delta = event.clientY - thumbRect.top;
              }
              // Add the active class to the thumb node.
              thumbNode.classList.add('lm-mod-active');
              /* <DEPRECATED> */
              thumbNode.classList.add('p-mod-active');
              /* </DEPRECATED> */
              // Store the current value in the press data.
              this._pressData.value = this._value;
              // Finished.
              return;
          }
          // Handle a track press.
          if (part === 'track') {
              // Fetch the client rect for the thumb.
              var thumbRect = this.thumbNode.getBoundingClientRect();
              // Determine the direction for the page request.
              var dir = void 0;
              if (this._orientation === 'horizontal') {
                  dir = event.clientX < thumbRect.left ? 'decrement' : 'increment';
              }
              else {
                  dir = event.clientY < thumbRect.top ? 'decrement' : 'increment';
              }
              // Start the repeat timer.
              this._repeatTimer = window.setTimeout(this._onRepeat, 350);
              // Emit the page requested signal.
              this._pageRequested.emit(dir);
              // Finished.
              return;
          }
          // Handle a decrement button press.
          if (part === 'decrement') {
              // Add the active class to the decrement node.
              this.decrementNode.classList.add('lm-mod-active');
              /* <DEPRECATED> */
              this.decrementNode.classList.add('p-mod-active');
              /* </DEPRECATED> */
              // Start the repeat timer.
              this._repeatTimer = window.setTimeout(this._onRepeat, 350);
              // Emit the step requested signal.
              this._stepRequested.emit('decrement');
              // Finished.
              return;
          }
          // Handle an increment button press.
          if (part === 'increment') {
              // Add the active class to the increment node.
              this.incrementNode.classList.add('lm-mod-active');
              /* <DEPRECATED> */
              this.incrementNode.classList.add('p-mod-active');
              /* </DEPRECATED> */
              // Start the repeat timer.
              this._repeatTimer = window.setTimeout(this._onRepeat, 350);
              // Emit the step requested signal.
              this._stepRequested.emit('increment');
              // Finished.
              return;
          }
      };
      /**
       * Handle the `'mousemove'` event for the scroll bar.
       */
      ScrollBar.prototype._evtMouseMove = function (event) {
          // Do nothing if no drag is in progress.
          if (!this._pressData) {
              return;
          }
          // Stop the event propagation.
          event.preventDefault();
          event.stopPropagation();
          // Update the mouse position.
          this._pressData.mouseX = event.clientX;
          this._pressData.mouseY = event.clientY;
          // Bail if the thumb is not being dragged.
          if (this._pressData.part !== 'thumb') {
              return;
          }
          // Get the client rect for the thumb and track.
          var thumbRect = this.thumbNode.getBoundingClientRect();
          var trackRect = this.trackNode.getBoundingClientRect();
          // Fetch the scroll geometry based on the orientation.
          var trackPos;
          var trackSpan;
          if (this._orientation === 'horizontal') {
              trackPos = event.clientX - trackRect.left - this._pressData.delta;
              trackSpan = trackRect.width - thumbRect.width;
          }
          else {
              trackPos = event.clientY - trackRect.top - this._pressData.delta;
              trackSpan = trackRect.height - thumbRect.height;
          }
          // Compute the desired value from the scroll geometry.
          var value = trackSpan === 0 ? 0 : trackPos * this._maximum / trackSpan;
          // Move the thumb to the computed value.
          this._moveThumb(value);
      };
      /**
       * Handle the `'mouseup'` event for the scroll bar.
       */
      ScrollBar.prototype._evtMouseUp = function (event) {
          // Do nothing if it's not a left mouse release.
          if (event.button !== 0) {
              return;
          }
          // Stop the event propagation.
          event.preventDefault();
          event.stopPropagation();
          // Release the mouse.
          this._releaseMouse();
      };
      /**
       * Release the mouse and restore the node states.
       */
      ScrollBar.prototype._releaseMouse = function () {
          // Bail if there is no press data.
          if (!this._pressData) {
              return;
          }
          // Clear the repeat timer.
          clearTimeout(this._repeatTimer);
          this._repeatTimer = -1;
          // Clear the press data.
          this._pressData.override.dispose();
          this._pressData = null;
          // Remove the extra event listeners.
          document.removeEventListener('mousemove', this, true);
          document.removeEventListener('mouseup', this, true);
          document.removeEventListener('keydown', this, true);
          document.removeEventListener('contextmenu', this, true);
          // Remove the active classes from the nodes.
          this.thumbNode.classList.remove('lm-mod-active');
          this.decrementNode.classList.remove('lm-mod-active');
          this.incrementNode.classList.remove('lm-mod-active');
          /* <DEPRECATED> */
          this.thumbNode.classList.remove('p-mod-active');
          this.decrementNode.classList.remove('p-mod-active');
          this.incrementNode.classList.remove('p-mod-active');
          /* </DEPRECATED> */
      };
      /**
       * Move the thumb to the specified position.
       */
      ScrollBar.prototype._moveThumb = function (value) {
          // Clamp the value to the allowed range.
          value = Math.max(0, Math.min(value, this._maximum));
          // Bail if the value does not change.
          if (this._value === value) {
              return;
          }
          // Update the internal value.
          this._value = value;
          // Schedule an update of the scroll bar.
          this.update();
          // Emit the thumb moved signal.
          this._thumbMoved.emit(value);
      };
      return ScrollBar;
  }(Widget));
  /**
   * The namespace for the module implementation details.
   */
  var Private$d;
  (function (Private) {
      /**
       * Create the DOM node for a scroll bar.
       */
      function createNode() {
          var node = document.createElement('div');
          var decrement = document.createElement('div');
          var increment = document.createElement('div');
          var track = document.createElement('div');
          var thumb = document.createElement('div');
          decrement.className = 'lm-ScrollBar-button';
          increment.className = 'lm-ScrollBar-button';
          decrement.dataset['action'] = 'decrement';
          increment.dataset['action'] = 'increment';
          track.className = 'lm-ScrollBar-track';
          thumb.className = 'lm-ScrollBar-thumb';
          /* <DEPRECATED> */
          decrement.classList.add('p-ScrollBar-button');
          increment.classList.add('p-ScrollBar-button');
          track.classList.add('p-ScrollBar-track');
          thumb.classList.add('p-ScrollBar-thumb');
          /* </DEPRECATED> */
          track.appendChild(thumb);
          node.appendChild(decrement);
          node.appendChild(track);
          node.appendChild(increment);
          return node;
      }
      Private.createNode = createNode;
      /**
       * Find the scroll bar part which contains the given target.
       */
      function findPart(scrollBar, target) {
          // Test the thumb.
          if (scrollBar.thumbNode.contains(target)) {
              return 'thumb';
          }
          // Test the track.
          if (scrollBar.trackNode.contains(target)) {
              return 'track';
          }
          // Test the decrement button.
          if (scrollBar.decrementNode.contains(target)) {
              return 'decrement';
          }
          // Test the increment button.
          if (scrollBar.incrementNode.contains(target)) {
              return 'increment';
          }
          // Indicate no match.
          return null;
      }
      Private.findPart = findPart;
  })(Private$d || (Private$d = {}));

  /**
   * A concrete layout implementation which holds a single widget.
   *
   * #### Notes
   * This class is useful for creating simple container widgets which
   * hold a single child. The child should be positioned with CSS.
   */
  var SingletonLayout = /** @class */ (function (_super) {
      __extends$5(SingletonLayout, _super);
      function SingletonLayout() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._widget = null;
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       */
      SingletonLayout.prototype.dispose = function () {
          if (this._widget) {
              var widget = this._widget;
              this._widget = null;
              widget.dispose();
          }
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(SingletonLayout.prototype, "widget", {
          /**
           * Get the child widget for the layout.
           */
          get: function () {
              return this._widget;
          },
          /**
           * Set the child widget for the layout.
           *
           * #### Notes
           * Setting the child widget will cause the old child widget to be
           * automatically disposed. If that is not desired, set the parent
           * of the old child to `null` before assigning a new child.
           */
          set: function (widget) {
              // Remove the widget from its current parent. This is a no-op
              // if the widget's parent is already the layout parent widget.
              if (widget) {
                  widget.parent = this.parent;
              }
              // Bail early if the widget does not change.
              if (this._widget === widget) {
                  return;
              }
              // Dispose of the old child widget.
              if (this._widget) {
                  this._widget.dispose();
              }
              // Update the internal widget.
              this._widget = widget;
              // Attach the new child widget if needed.
              if (this.parent && widget) {
                  this.attachWidget(widget);
              }
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Create an iterator over the widgets in the layout.
       *
       * @returns A new iterator over the widgets in the layout.
       */
      SingletonLayout.prototype.iter = function () {
          return this._widget ? once$1(this._widget) : empty();
      };
      /**
       * Remove a widget from the layout.
       *
       * @param widget - The widget to remove from the layout.
       *
       * #### Notes
       * A widget is automatically removed from the layout when its `parent`
       * is set to `null`. This method should only be invoked directly when
       * removing a widget from a layout which has yet to be installed on a
       * parent widget.
       *
       * This method does *not* modify the widget's `parent`.
       */
      SingletonLayout.prototype.removeWidget = function (widget) {
          // Bail early if the widget does not exist in the layout.
          if (this._widget !== widget) {
              return;
          }
          // Clear the internal widget.
          this._widget = null;
          // If the layout is parented, detach the widget from the DOM.
          if (this.parent) {
              this.detachWidget(widget);
          }
      };
      /**
       * Perform layout initialization which requires the parent widget.
       */
      SingletonLayout.prototype.init = function () {
          var _this = this;
          _super.prototype.init.call(this);
          each$1(this, function (widget) { _this.attachWidget(widget); });
      };
      /**
       * Attach a widget to the parent's DOM node.
       *
       * @param index - The current index of the widget in the layout.
       *
       * @param widget - The widget to attach to the parent.
       *
       * #### Notes
       * This method is called automatically by the single layout at the
       * appropriate time. It should not be called directly by user code.
       *
       * The default implementation adds the widgets's node to the parent's
       * node at the proper location, and sends the appropriate attach
       * messages to the widget if the parent is attached to the DOM.
       *
       * Subclasses may reimplement this method to control how the widget's
       * node is added to the parent's node.
       */
      SingletonLayout.prototype.attachWidget = function (widget) {
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Add the widget's node to the parent.
          this.parent.node.appendChild(widget.node);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
      };
      /**
       * Detach a widget from the parent's DOM node.
       *
       * @param widget - The widget to detach from the parent.
       *
       * #### Notes
       * This method is called automatically by the single layout at the
       * appropriate time. It should not be called directly by user code.
       *
       * The default implementation removes the widget's node from the
       * parent's node, and sends the appropriate detach messages to the
       * widget if the parent is attached to the DOM.
       *
       * Subclasses may reimplement this method to control how the widget's
       * node is removed from the parent's node.
       */
      SingletonLayout.prototype.detachWidget = function (widget) {
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
      };
      return SingletonLayout;
  }(Layout));

  /**
   * A layout which arranges its widgets into resizable sections.
   */
  var SplitLayout = /** @class */ (function (_super) {
      __extends$5(SplitLayout, _super);
      /**
       * Construct a new split layout.
       *
       * @param options - The options for initializing the layout.
       */
      function SplitLayout(options) {
          var _this = _super.call(this) || this;
          _this._fixed = 0;
          _this._spacing = 4;
          _this._dirty = false;
          _this._hasNormedSizes = false;
          _this._sizers = [];
          _this._items = [];
          _this._handles = [];
          _this._box = null;
          _this._alignment = 'start';
          _this._orientation = 'horizontal';
          _this.renderer = options.renderer;
          if (options.orientation !== undefined) {
              _this._orientation = options.orientation;
          }
          if (options.alignment !== undefined) {
              _this._alignment = options.alignment;
          }
          if (options.spacing !== undefined) {
              _this._spacing = Private$e.clampSpacing(options.spacing);
          }
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       */
      SplitLayout.prototype.dispose = function () {
          // Dispose of the layout items.
          each$1(this._items, function (item) { item.dispose(); });
          // Clear the layout state.
          this._box = null;
          this._items.length = 0;
          this._sizers.length = 0;
          this._handles.length = 0;
          // Dispose of the rest of the layout.
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(SplitLayout.prototype, "orientation", {
          /**
           * Get the layout orientation for the split layout.
           */
          get: function () {
              return this._orientation;
          },
          /**
           * Set the layout orientation for the split layout.
           */
          set: function (value) {
              if (this._orientation === value) {
                  return;
              }
              this._orientation = value;
              if (!this.parent) {
                  return;
              }
              this.parent.dataset['orientation'] = value;
              this.parent.fit();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitLayout.prototype, "alignment", {
          /**
           * Get the content alignment for the split layout.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand  to fill the
           * entire split layout.
           */
          get: function () {
              return this._alignment;
          },
          /**
           * Set the content alignment for the split layout.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand  to fill the
           * entire split layout.
           */
          set: function (value) {
              if (this._alignment === value) {
                  return;
              }
              this._alignment = value;
              if (!this.parent) {
                  return;
              }
              this.parent.dataset['alignment'] = value;
              this.parent.update();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitLayout.prototype, "spacing", {
          /**
           * Get the inter-element spacing for the split layout.
           */
          get: function () {
              return this._spacing;
          },
          /**
           * Set the inter-element spacing for the split layout.
           */
          set: function (value) {
              value = Private$e.clampSpacing(value);
              if (this._spacing === value) {
                  return;
              }
              this._spacing = value;
              if (!this.parent) {
                  return;
              }
              this.parent.fit();
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitLayout.prototype, "handles", {
          /**
           * A read-only array of the split handles in the layout.
           */
          get: function () {
              return this._handles;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Get the relative sizes of the widgets in the layout.
       *
       * @returns A new array of the relative sizes of the widgets.
       *
       * #### Notes
       * The returned sizes reflect the sizes of the widgets normalized
       * relative to their siblings.
       *
       * This method **does not** measure the DOM nodes.
       */
      SplitLayout.prototype.relativeSizes = function () {
          return Private$e.normalize(this._sizers.map(function (sizer) { return sizer.size; }));
      };
      /**
       * Set the relative sizes for the widgets in the layout.
       *
       * @param sizes - The relative sizes for the widgets in the panel.
       *
       * #### Notes
       * Extra values are ignored, too few will yield an undefined layout.
       *
       * The actual geometry of the DOM nodes is updated asynchronously.
       */
      SplitLayout.prototype.setRelativeSizes = function (sizes) {
          // Copy the sizes and pad with zeros as needed.
          var n = this._sizers.length;
          var temp = sizes.slice(0, n);
          while (temp.length < n) {
              temp.push(0);
          }
          // Normalize the padded sizes.
          var normed = Private$e.normalize(temp);
          // Apply the normalized sizes to the sizers.
          for (var i = 0; i < n; ++i) {
              var sizer = this._sizers[i];
              sizer.sizeHint = normed[i];
              sizer.size = normed[i];
          }
          // Set the flag indicating the sizes are normalized.
          this._hasNormedSizes = true;
          // Trigger an update of the parent widget.
          if (this.parent) {
              this.parent.update();
          }
      };
      /**
       * Move the offset position of a split handle.
       *
       * @param index - The index of the handle of the interest.
       *
       * @param position - The desired offset position of the handle.
       *
       * #### Notes
       * The position is relative to the offset parent.
       *
       * This will move the handle as close as possible to the desired
       * position. The sibling widgets will be adjusted as necessary.
       */
      SplitLayout.prototype.moveHandle = function (index, position) {
          // Bail if the index is invalid or the handle is hidden.
          var handle = this._handles[index];
          if (!handle || handle.classList.contains('lm-mod-hidden')) {
              return;
          }
          // Compute the desired delta movement for the handle.
          var delta;
          if (this._orientation === 'horizontal') {
              delta = position - handle.offsetLeft;
          }
          else {
              delta = position - handle.offsetTop;
          }
          // Bail if there is no handle movement.
          if (delta === 0) {
              return;
          }
          // Prevent widget resizing unless needed.
          for (var _i = 0, _a = this._sizers; _i < _a.length; _i++) {
              var sizer = _a[_i];
              if (sizer.size > 0) {
                  sizer.sizeHint = sizer.size;
              }
          }
          // Adjust the sizers to reflect the handle movement.
          BoxEngine.adjust(this._sizers, index, delta);
          // Update the layout of the widgets.
          if (this.parent) {
              this.parent.update();
          }
      };
      /**
       * Perform layout initialization which requires the parent widget.
       */
      SplitLayout.prototype.init = function () {
          this.parent.dataset['orientation'] = this.orientation;
          this.parent.dataset['alignment'] = this.alignment;
          _super.prototype.init.call(this);
      };
      /**
       * Attach a widget to the parent's DOM node.
       *
       * @param index - The current index of the widget in the layout.
       *
       * @param widget - The widget to attach to the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      SplitLayout.prototype.attachWidget = function (index, widget) {
          // Create the item, handle, and sizer for the new widget.
          var item = new LayoutItem(widget);
          var handle = Private$e.createHandle(this.renderer);
          var average = Private$e.averageSize(this._sizers);
          var sizer = Private$e.createSizer(average);
          // Insert the item, handle, and sizer into the internal arrays.
          ArrayExt.insert(this._items, index, item);
          ArrayExt.insert(this._sizers, index, sizer);
          ArrayExt.insert(this._handles, index, handle);
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Add the widget and handle nodes to the parent.
          this.parent.node.appendChild(widget.node);
          this.parent.node.appendChild(handle);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Move a widget in the parent's DOM node.
       *
       * @param fromIndex - The previous index of the widget in the layout.
       *
       * @param toIndex - The current index of the widget in the layout.
       *
       * @param widget - The widget to move in the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      SplitLayout.prototype.moveWidget = function (fromIndex, toIndex, widget) {
          // Move the item, sizer, and handle for the widget.
          ArrayExt.move(this._items, fromIndex, toIndex);
          ArrayExt.move(this._sizers, fromIndex, toIndex);
          ArrayExt.move(this._handles, fromIndex, toIndex);
          // Post a fit request to the parent to show/hide last handle.
          this.parent.fit();
      };
      /**
       * Detach a widget from the parent's DOM node.
       *
       * @param index - The previous index of the widget in the layout.
       *
       * @param widget - The widget to detach from the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      SplitLayout.prototype.detachWidget = function (index, widget) {
          // Remove the item, handle, and sizer for the widget.
          var item = ArrayExt.removeAt(this._items, index);
          var handle = ArrayExt.removeAt(this._handles, index);
          ArrayExt.removeAt(this._sizers, index);
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget and handle nodes from the parent.
          this.parent.node.removeChild(widget.node);
          this.parent.node.removeChild(handle);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
          // Dispose of the layout item.
          item.dispose();
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'before-show'` message.
       */
      SplitLayout.prototype.onBeforeShow = function (msg) {
          _super.prototype.onBeforeShow.call(this, msg);
          this.parent.update();
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      SplitLayout.prototype.onBeforeAttach = function (msg) {
          _super.prototype.onBeforeAttach.call(this, msg);
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-shown'` message.
       */
      SplitLayout.prototype.onChildShown = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-hidden'` message.
       */
      SplitLayout.prototype.onChildHidden = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'resize'` message.
       */
      SplitLayout.prototype.onResize = function (msg) {
          if (this.parent.isVisible) {
              this._update(msg.width, msg.height);
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      SplitLayout.prototype.onUpdateRequest = function (msg) {
          if (this.parent.isVisible) {
              this._update(-1, -1);
          }
      };
      /**
       * A message handler invoked on a `'fit-request'` message.
       */
      SplitLayout.prototype.onFitRequest = function (msg) {
          if (this.parent.isAttached) {
              this._fit();
          }
      };
      /**
       * Fit the layout to the total size required by the widgets.
       */
      SplitLayout.prototype._fit = function () {
          // Update the handles and track the visible widget count.
          var nVisible = 0;
          var lastHandleIndex = -1;
          for (var i = 0, n = this._items.length; i < n; ++i) {
              if (this._items[i].isHidden) {
                  this._handles[i].classList.add('lm-mod-hidden');
                  /* <DEPRECATED> */
                  this._handles[i].classList.add('p-mod-hidden');
                  /* </DEPRECATED> */
              }
              else {
                  this._handles[i].classList.remove('lm-mod-hidden');
                  /* <DEPRECATED> */
                  this._handles[i].classList.remove('p-mod-hidden');
                  /* </DEPRECATED> */
                  lastHandleIndex = i;
                  nVisible++;
              }
          }
          // Hide the handle for the last visible widget.
          if (lastHandleIndex !== -1) {
              this._handles[lastHandleIndex].classList.add('lm-mod-hidden');
              /* <DEPRECATED> */
              this._handles[lastHandleIndex].classList.add('p-mod-hidden');
              /* </DEPRECATED> */
          }
          // Update the fixed space for the visible items.
          this._fixed = this._spacing * Math.max(0, nVisible - 1);
          // Setup the computed minimum size.
          var horz = this._orientation === 'horizontal';
          var minW = horz ? this._fixed : 0;
          var minH = horz ? 0 : this._fixed;
          // Update the sizers and computed size limits.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item and corresponding box sizer.
              var item = this._items[i];
              var sizer = this._sizers[i];
              // Prevent resizing unless necessary.
              if (sizer.size > 0) {
                  sizer.sizeHint = sizer.size;
              }
              // If the item is hidden, it should consume zero size.
              if (item.isHidden) {
                  sizer.minSize = 0;
                  sizer.maxSize = 0;
                  continue;
              }
              // Update the size limits for the item.
              item.fit();
              // Update the stretch factor.
              sizer.stretch = SplitLayout.getStretch(item.widget);
              // Update the sizer limits and computed min size.
              if (horz) {
                  sizer.minSize = item.minWidth;
                  sizer.maxSize = item.maxWidth;
                  minW += item.minWidth;
                  minH = Math.max(minH, item.minHeight);
              }
              else {
                  sizer.minSize = item.minHeight;
                  sizer.maxSize = item.maxHeight;
                  minH += item.minHeight;
                  minW = Math.max(minW, item.minWidth);
              }
          }
          // Update the box sizing and add it to the computed min size.
          var box = this._box = ElementExt.boxSizing(this.parent.node);
          minW += box.horizontalSum;
          minH += box.verticalSum;
          // Update the parent's min size constraints.
          var style = this.parent.node.style;
          style.minWidth = minW + "px";
          style.minHeight = minH + "px";
          // Set the dirty flag to ensure only a single update occurs.
          this._dirty = true;
          // Notify the ancestor that it should fit immediately. This may
          // cause a resize of the parent, fulfilling the required update.
          if (this.parent.parent) {
              MessageLoop.sendMessage(this.parent.parent, Widget.Msg.FitRequest);
          }
          // If the dirty flag is still set, the parent was not resized.
          // Trigger the required update on the parent widget immediately.
          if (this._dirty) {
              MessageLoop.sendMessage(this.parent, Widget.Msg.UpdateRequest);
          }
      };
      /**
       * Update the layout position and size of the widgets.
       *
       * The parent offset dimensions should be `-1` if unknown.
       */
      SplitLayout.prototype._update = function (offsetWidth, offsetHeight) {
          // Clear the dirty flag to indicate the update occurred.
          this._dirty = false;
          // Compute the visible item count.
          var nVisible = 0;
          for (var i = 0, n = this._items.length; i < n; ++i) {
              nVisible += +!this._items[i].isHidden;
          }
          // Bail early if there are no visible items to layout.
          if (nVisible === 0) {
              return;
          }
          // Measure the parent if the offset dimensions are unknown.
          if (offsetWidth < 0) {
              offsetWidth = this.parent.node.offsetWidth;
          }
          if (offsetHeight < 0) {
              offsetHeight = this.parent.node.offsetHeight;
          }
          // Ensure the parent box sizing data is computed.
          if (!this._box) {
              this._box = ElementExt.boxSizing(this.parent.node);
          }
          // Compute the actual layout bounds adjusted for border and padding.
          var top = this._box.paddingTop;
          var left = this._box.paddingLeft;
          var width = offsetWidth - this._box.horizontalSum;
          var height = offsetHeight - this._box.verticalSum;
          // Compute the adjusted layout space.
          var space;
          var horz = this._orientation === 'horizontal';
          if (horz) {
              space = Math.max(0, width - this._fixed);
          }
          else {
              space = Math.max(0, height - this._fixed);
          }
          // Scale the size hints if they are normalized.
          if (this._hasNormedSizes) {
              for (var _i = 0, _a = this._sizers; _i < _a.length; _i++) {
                  var sizer = _a[_i];
                  sizer.sizeHint *= space;
              }
              this._hasNormedSizes = false;
          }
          // Distribute the layout space to the box sizers.
          var delta = BoxEngine.calc(this._sizers, space);
          // Set up the variables for justification and alignment offset.
          var extra = 0;
          var offset = 0;
          // Account for alignment if there is extra layout space.
          if (delta > 0) {
              switch (this._alignment) {
                  case 'start':
                      break;
                  case 'center':
                      extra = 0;
                      offset = delta / 2;
                      break;
                  case 'end':
                      extra = 0;
                      offset = delta;
                      break;
                  case 'justify':
                      extra = delta / nVisible;
                      offset = 0;
                      break;
                  default:
                      throw 'unreachable';
              }
          }
          // Layout the items using the computed box sizes.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item.
              var item = this._items[i];
              // Ignore hidden items.
              if (item.isHidden) {
                  continue;
              }
              // Fetch the computed size for the widget.
              var size = this._sizers[i].size;
              // Fetch the style for the handle.
              var handleStyle = this._handles[i].style;
              // Update the widget and handle, and advance the relevant edge.
              if (horz) {
                  item.update(left + offset, top, size + extra, height);
                  left += size + extra;
                  handleStyle.top = top + "px";
                  handleStyle.left = left + offset + "px";
                  handleStyle.width = this._spacing + "px";
                  handleStyle.height = height + "px";
                  left += this._spacing;
              }
              else {
                  item.update(left, top + offset, width, size + extra);
                  top += size + extra;
                  handleStyle.top = top + offset + "px";
                  handleStyle.left = left + "px";
                  handleStyle.width = width + "px";
                  handleStyle.height = this._spacing + "px";
                  top += this._spacing;
              }
          }
      };
      return SplitLayout;
  }(PanelLayout));
  /**
   * The namespace for the `SplitLayout` class statics.
   */
  (function (SplitLayout) {
      /**
       * Get the split layout stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The split layout stretch factor for the widget.
       */
      function getStretch(widget) {
          return Private$e.stretchProperty.get(widget);
      }
      SplitLayout.getStretch = getStretch;
      /**
       * Set the split layout stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the stretch factor.
       */
      function setStretch(widget, value) {
          Private$e.stretchProperty.set(widget, value);
      }
      SplitLayout.setStretch = setStretch;
  })(SplitLayout || (SplitLayout = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$e;
  (function (Private) {
      /**
       * The property descriptor for a widget stretch factor.
       */
      Private.stretchProperty = new AttachedProperty({
          name: 'stretch',
          create: function () { return 0; },
          coerce: function (owner, value) { return Math.max(0, Math.floor(value)); },
          changed: onChildSizingChanged
      });
      /**
       * Create a new box sizer with the given size hint.
       */
      function createSizer(size) {
          var sizer = new BoxSizer();
          sizer.sizeHint = Math.floor(size);
          return sizer;
      }
      Private.createSizer = createSizer;
      /**
       * Create a new split handle node using the given renderer.
       */
      function createHandle(renderer) {
          var handle = renderer.createHandle();
          handle.style.position = 'absolute';
          return handle;
      }
      Private.createHandle = createHandle;
      /**
       * Clamp a spacing value to an integer >= 0.
       */
      function clampSpacing(value) {
          return Math.max(0, Math.floor(value));
      }
      Private.clampSpacing = clampSpacing;
      /**
       * Compute the average size of an array of box sizers.
       */
      function averageSize(sizers) {
          return sizers.reduce(function (v, s) { return v + s.size; }, 0) / sizers.length || 0;
      }
      Private.averageSize = averageSize;
      /**
       * Normalize an array of values.
       */
      function normalize(values) {
          var n = values.length;
          if (n === 0) {
              return [];
          }
          var sum = values.reduce(function (a, b) { return a + Math.abs(b); }, 0);
          return sum === 0 ? values.map(function (v) { return 1 / n; }) : values.map(function (v) { return v / sum; });
      }
      Private.normalize = normalize;
      /**
       * The change handler for the attached sizing properties.
       */
      function onChildSizingChanged(child) {
          if (child.parent && child.parent.layout instanceof SplitLayout) {
              child.parent.fit();
          }
      }
  })(Private$e || (Private$e = {}));

  /**
   * A panel which arranges its widgets into resizable sections.
   *
   * #### Notes
   * This class provides a convenience wrapper around a [[SplitLayout]].
   */
  var SplitPanel = /** @class */ (function (_super) {
      __extends$5(SplitPanel, _super);
      /**
       * Construct a new split panel.
       *
       * @param options - The options for initializing the split panel.
       */
      function SplitPanel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, { layout: Private$f.createLayout(options) }) || this;
          _this._pressData = null;
          _this.addClass('lm-SplitPanel');
          /* <DEPRECATED> */
          _this.addClass('p-SplitPanel');
          return _this;
          /* </DEPRECATED> */
      }
      /**
       * Dispose of the resources held by the panel.
       */
      SplitPanel.prototype.dispose = function () {
          this._releaseMouse();
          _super.prototype.dispose.call(this);
      };
      Object.defineProperty(SplitPanel.prototype, "orientation", {
          /**
           * Get the layout orientation for the split panel.
           */
          get: function () {
              return this.layout.orientation;
          },
          /**
           * Set the layout orientation for the split panel.
           */
          set: function (value) {
              this.layout.orientation = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitPanel.prototype, "alignment", {
          /**
           * Get the content alignment for the split panel.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand to fill the
           * entire split panel.
           */
          get: function () {
              return this.layout.alignment;
          },
          /**
           * Set the content alignment for the split panel.
           *
           * #### Notes
           * This is the alignment of the widgets in the layout direction.
           *
           * The alignment has no effect if the widgets can expand to fill the
           * entire split panel.
           */
          set: function (value) {
              this.layout.alignment = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitPanel.prototype, "spacing", {
          /**
           * Get the inter-element spacing for the split panel.
           */
          get: function () {
              return this.layout.spacing;
          },
          /**
           * Set the inter-element spacing for the split panel.
           */
          set: function (value) {
              this.layout.spacing = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitPanel.prototype, "renderer", {
          /**
           * The renderer used by the split panel.
           */
          get: function () {
              return this.layout.renderer;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(SplitPanel.prototype, "handles", {
          /**
           * A read-only array of the split handles in the panel.
           */
          get: function () {
              return this.layout.handles;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Get the relative sizes of the widgets in the panel.
       *
       * @returns A new array of the relative sizes of the widgets.
       *
       * #### Notes
       * The returned sizes reflect the sizes of the widgets normalized
       * relative to their siblings.
       *
       * This method **does not** measure the DOM nodes.
       */
      SplitPanel.prototype.relativeSizes = function () {
          return this.layout.relativeSizes();
      };
      /**
       * Set the relative sizes for the widgets in the panel.
       *
       * @param sizes - The relative sizes for the widgets in the panel.
       *
       * #### Notes
       * Extra values are ignored, too few will yield an undefined layout.
       *
       * The actual geometry of the DOM nodes is updated asynchronously.
       */
      SplitPanel.prototype.setRelativeSizes = function (sizes) {
          this.layout.setRelativeSizes(sizes);
      };
      /**
       * Handle the DOM events for the split panel.
       *
       * @param event - The DOM event sent to the panel.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the panel's DOM node. It should
       * not be called directly by user code.
       */
      SplitPanel.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'mousedown':
                  this._evtMouseDown(event);
                  break;
              case 'mousemove':
                  this._evtMouseMove(event);
                  break;
              case 'mouseup':
                  this._evtMouseUp(event);
                  break;
              case 'keydown':
                  this._evtKeyDown(event);
                  break;
              case 'contextmenu':
                  event.preventDefault();
                  event.stopPropagation();
                  break;
          }
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      SplitPanel.prototype.onBeforeAttach = function (msg) {
          this.node.addEventListener('mousedown', this);
      };
      /**
       * A message handler invoked on an `'after-detach'` message.
       */
      SplitPanel.prototype.onAfterDetach = function (msg) {
          this.node.removeEventListener('mousedown', this);
          this._releaseMouse();
      };
      /**
       * A message handler invoked on a `'child-added'` message.
       */
      SplitPanel.prototype.onChildAdded = function (msg) {
          msg.child.addClass('lm-SplitPanel-child');
          /* <DEPRECATED> */
          msg.child.addClass('p-SplitPanel-child');
          /* </DEPRECATED> */
          this._releaseMouse();
      };
      /**
       * A message handler invoked on a `'child-removed'` message.
       */
      SplitPanel.prototype.onChildRemoved = function (msg) {
          msg.child.removeClass('lm-SplitPanel-child');
          /* <DEPRECATED> */
          msg.child.removeClass('p-SplitPanel-child');
          /* </DEPRECATED> */
          this._releaseMouse();
      };
      /**
       * Handle the `'keydown'` event for the split panel.
       */
      SplitPanel.prototype._evtKeyDown = function (event) {
          // Stop input events during drag.
          event.preventDefault();
          event.stopPropagation();
          // Release the mouse if `Escape` is pressed.
          if (event.keyCode === 27) {
              this._releaseMouse();
          }
      };
      /**
       * Handle the `'mousedown'` event for the split panel.
       */
      SplitPanel.prototype._evtMouseDown = function (event) {
          // Do nothing if the left mouse button is not pressed.
          if (event.button !== 0) {
              return;
          }
          // Find the handle which contains the mouse target, if any.
          var layout = this.layout;
          var index = ArrayExt.findFirstIndex(layout.handles, function (handle) {
              return handle.contains(event.target);
          });
          // Bail early if the mouse press was not on a handle.
          if (index === -1) {
              return;
          }
          // Stop the event when a split handle is pressed.
          event.preventDefault();
          event.stopPropagation();
          // Add the extra document listeners.
          document.addEventListener('mouseup', this, true);
          document.addEventListener('mousemove', this, true);
          document.addEventListener('keydown', this, true);
          document.addEventListener('contextmenu', this, true);
          // Compute the offset delta for the handle press.
          var delta;
          var handle = layout.handles[index];
          var rect = handle.getBoundingClientRect();
          if (layout.orientation === 'horizontal') {
              delta = event.clientX - rect.left;
          }
          else {
              delta = event.clientY - rect.top;
          }
          // Override the cursor and store the press data.
          var style = window.getComputedStyle(handle);
          var override = Drag.overrideCursor(style.cursor);
          this._pressData = { index: index, delta: delta, override: override };
      };
      /**
       * Handle the `'mousemove'` event for the split panel.
       */
      SplitPanel.prototype._evtMouseMove = function (event) {
          // Stop the event when dragging a split handle.
          event.preventDefault();
          event.stopPropagation();
          // Compute the desired offset position for the handle.
          var pos;
          var layout = this.layout;
          var rect = this.node.getBoundingClientRect();
          if (layout.orientation === 'horizontal') {
              pos = event.clientX - rect.left - this._pressData.delta;
          }
          else {
              pos = event.clientY - rect.top - this._pressData.delta;
          }
          // Move the handle as close to the desired position as possible.
          layout.moveHandle(this._pressData.index, pos);
      };
      /**
       * Handle the `'mouseup'` event for the split panel.
       */
      SplitPanel.prototype._evtMouseUp = function (event) {
          // Do nothing if the left mouse button is not released.
          if (event.button !== 0) {
              return;
          }
          // Stop the event when releasing a handle.
          event.preventDefault();
          event.stopPropagation();
          // Finalize the mouse release.
          this._releaseMouse();
      };
      /**
       * Release the mouse grab for the split panel.
       */
      SplitPanel.prototype._releaseMouse = function () {
          // Bail early if no drag is in progress.
          if (!this._pressData) {
              return;
          }
          // Clear the override cursor.
          this._pressData.override.dispose();
          this._pressData = null;
          // Remove the extra document listeners.
          document.removeEventListener('mouseup', this, true);
          document.removeEventListener('mousemove', this, true);
          document.removeEventListener('keydown', this, true);
          document.removeEventListener('contextmenu', this, true);
      };
      return SplitPanel;
  }(Panel));
  /**
   * The namespace for the `SplitPanel` class statics.
   */
  (function (SplitPanel) {
      /**
       * The default implementation of `IRenderer`.
       */
      var Renderer = /** @class */ (function () {
          function Renderer() {
          }
          /**
           * Create a new handle for use with a split panel.
           *
           * @returns A new handle element for a split panel.
           */
          Renderer.prototype.createHandle = function () {
              var handle = document.createElement('div');
              handle.className = 'lm-SplitPanel-handle';
              /* <DEPRECATED> */
              handle.classList.add('p-SplitPanel-handle');
              /* </DEPRECATED> */
              return handle;
          };
          return Renderer;
      }());
      SplitPanel.Renderer = Renderer;
      /**
       * The default `Renderer` instance.
       */
      SplitPanel.defaultRenderer = new Renderer();
      /**
       * Get the split panel stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @returns The split panel stretch factor for the widget.
       */
      function getStretch(widget) {
          return SplitLayout.getStretch(widget);
      }
      SplitPanel.getStretch = getStretch;
      /**
       * Set the split panel stretch factor for the given widget.
       *
       * @param widget - The widget of interest.
       *
       * @param value - The value for the stretch factor.
       */
      function setStretch(widget, value) {
          SplitLayout.setStretch(widget, value);
      }
      SplitPanel.setStretch = setStretch;
  })(SplitPanel || (SplitPanel = {}));
  /**
   * The namespace for the module implementation details.
   */
  var Private$f;
  (function (Private) {
      /**
       * Create a split layout for the given panel options.
       */
      function createLayout(options) {
          return options.layout || new SplitLayout({
              renderer: options.renderer || SplitPanel.defaultRenderer,
              orientation: options.orientation,
              alignment: options.alignment,
              spacing: options.spacing
          });
      }
      Private.createLayout = createLayout;
  })(Private$f || (Private$f = {}));

  /**
   * A layout where visible widgets are stacked atop one another.
   *
   * #### Notes
   * The Z-order of the visible widgets follows their layout order.
   */
  var StackedLayout = /** @class */ (function (_super) {
      __extends$5(StackedLayout, _super);
      function StackedLayout() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._dirty = false;
          _this._items = [];
          _this._box = null;
          return _this;
      }
      /**
       * Dispose of the resources held by the layout.
       */
      StackedLayout.prototype.dispose = function () {
          // Dispose of the layout items.
          each$1(this._items, function (item) { item.dispose(); });
          // Clear the layout state.
          this._box = null;
          this._items.length = 0;
          // Dispose of the rest of the layout.
          _super.prototype.dispose.call(this);
      };
      /**
       * Attach a widget to the parent's DOM node.
       *
       * @param index - The current index of the widget in the layout.
       *
       * @param widget - The widget to attach to the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      StackedLayout.prototype.attachWidget = function (index, widget) {
          // Create and add a new layout item for the widget.
          ArrayExt.insert(this._items, index, new LayoutItem(widget));
          // Send a `'before-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeAttach);
          }
          // Add the widget's node to the parent.
          this.parent.node.appendChild(widget.node);
          // Send an `'after-attach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterAttach);
          }
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * Move a widget in the parent's DOM node.
       *
       * @param fromIndex - The previous index of the widget in the layout.
       *
       * @param toIndex - The current index of the widget in the layout.
       *
       * @param widget - The widget to move in the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      StackedLayout.prototype.moveWidget = function (fromIndex, toIndex, widget) {
          // Move the layout item for the widget.
          ArrayExt.move(this._items, fromIndex, toIndex);
          // Post an update request for the parent widget.
          this.parent.update();
      };
      /**
       * Detach a widget from the parent's DOM node.
       *
       * @param index - The previous index of the widget in the layout.
       *
       * @param widget - The widget to detach from the parent.
       *
       * #### Notes
       * This is a reimplementation of the superclass method.
       */
      StackedLayout.prototype.detachWidget = function (index, widget) {
          // Remove the layout item for the widget.
          var item = ArrayExt.removeAt(this._items, index);
          // Send a `'before-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.BeforeDetach);
          }
          // Remove the widget's node from the parent.
          this.parent.node.removeChild(widget.node);
          // Send an `'after-detach'` message if the parent is attached.
          if (this.parent.isAttached) {
              MessageLoop.sendMessage(widget, Widget.Msg.AfterDetach);
          }
          // Reset the z-index for the widget.
          item.widget.node.style.zIndex = '';
          // Dispose of the layout item.
          item.dispose();
          // Post a fit request for the parent widget.
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'before-show'` message.
       */
      StackedLayout.prototype.onBeforeShow = function (msg) {
          _super.prototype.onBeforeShow.call(this, msg);
          this.parent.update();
      };
      /**
       * A message handler invoked on a `'before-attach'` message.
       */
      StackedLayout.prototype.onBeforeAttach = function (msg) {
          _super.prototype.onBeforeAttach.call(this, msg);
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-shown'` message.
       */
      StackedLayout.prototype.onChildShown = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'child-hidden'` message.
       */
      StackedLayout.prototype.onChildHidden = function (msg) {
          this.parent.fit();
      };
      /**
       * A message handler invoked on a `'resize'` message.
       */
      StackedLayout.prototype.onResize = function (msg) {
          if (this.parent.isVisible) {
              this._update(msg.width, msg.height);
          }
      };
      /**
       * A message handler invoked on an `'update-request'` message.
       */
      StackedLayout.prototype.onUpdateRequest = function (msg) {
          if (this.parent.isVisible) {
              this._update(-1, -1);
          }
      };
      /**
       * A message handler invoked on a `'fit-request'` message.
       */
      StackedLayout.prototype.onFitRequest = function (msg) {
          if (this.parent.isAttached) {
              this._fit();
          }
      };
      /**
       * Fit the layout to the total size required by the widgets.
       */
      StackedLayout.prototype._fit = function () {
          // Set up the computed minimum size.
          var minW = 0;
          var minH = 0;
          // Update the computed minimum size.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item.
              var item = this._items[i];
              // Ignore hidden items.
              if (item.isHidden) {
                  continue;
              }
              // Update the size limits for the item.
              item.fit();
              // Update the computed minimum size.
              minW = Math.max(minW, item.minWidth);
              minH = Math.max(minH, item.minHeight);
          }
          // Update the box sizing and add it to the computed min size.
          var box = this._box = ElementExt.boxSizing(this.parent.node);
          minW += box.horizontalSum;
          minH += box.verticalSum;
          // Update the parent's min size constraints.
          var style = this.parent.node.style;
          style.minWidth = minW + "px";
          style.minHeight = minH + "px";
          // Set the dirty flag to ensure only a single update occurs.
          this._dirty = true;
          // Notify the ancestor that it should fit immediately. This may
          // cause a resize of the parent, fulfilling the required update.
          if (this.parent.parent) {
              MessageLoop.sendMessage(this.parent.parent, Widget.Msg.FitRequest);
          }
          // If the dirty flag is still set, the parent was not resized.
          // Trigger the required update on the parent widget immediately.
          if (this._dirty) {
              MessageLoop.sendMessage(this.parent, Widget.Msg.UpdateRequest);
          }
      };
      /**
       * Update the layout position and size of the widgets.
       *
       * The parent offset dimensions should be `-1` if unknown.
       */
      StackedLayout.prototype._update = function (offsetWidth, offsetHeight) {
          // Clear the dirty flag to indicate the update occurred.
          this._dirty = false;
          // Compute the visible item count.
          var nVisible = 0;
          for (var i = 0, n = this._items.length; i < n; ++i) {
              nVisible += +!this._items[i].isHidden;
          }
          // Bail early if there are no visible items to layout.
          if (nVisible === 0) {
              return;
          }
          // Measure the parent if the offset dimensions are unknown.
          if (offsetWidth < 0) {
              offsetWidth = this.parent.node.offsetWidth;
          }
          if (offsetHeight < 0) {
              offsetHeight = this.parent.node.offsetHeight;
          }
          // Ensure the parent box sizing data is computed.
          if (!this._box) {
              this._box = ElementExt.boxSizing(this.parent.node);
          }
          // Compute the actual layout bounds adjusted for border and padding.
          var top = this._box.paddingTop;
          var left = this._box.paddingLeft;
          var width = offsetWidth - this._box.horizontalSum;
          var height = offsetHeight - this._box.verticalSum;
          // Update the widget stacking order and layout geometry.
          for (var i = 0, n = this._items.length; i < n; ++i) {
              // Fetch the item.
              var item = this._items[i];
              // Ignore hidden items.
              if (item.isHidden) {
                  continue;
              }
              // Set the z-index for the widget.
              item.widget.node.style.zIndex = "" + i;
              // Update the item geometry.
              item.update(left, top, width, height);
          }
      };
      return StackedLayout;
  }(PanelLayout));

  /**
   * A panel where visible widgets are stacked atop one another.
   *
   * #### Notes
   * This class provides a convenience wrapper around a [[StackedLayout]].
   */
  var StackedPanel = /** @class */ (function (_super) {
      __extends$5(StackedPanel, _super);
      /**
       * Construct a new stacked panel.
       *
       * @param options - The options for initializing the panel.
       */
      function StackedPanel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this, { layout: Private$g.createLayout(options) }) || this;
          _this._widgetRemoved = new Signal(_this);
          _this.addClass('lm-StackedPanel');
          /* <DEPRECATED> */
          _this.addClass('p-StackedPanel');
          return _this;
          /* </DEPRECATED> */
      }
      Object.defineProperty(StackedPanel.prototype, "widgetRemoved", {
          /**
           * A signal emitted when a widget is removed from a stacked panel.
           */
          get: function () {
              return this._widgetRemoved;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * A message handler invoked on a `'child-added'` message.
       */
      StackedPanel.prototype.onChildAdded = function (msg) {
          msg.child.addClass('lm-StackedPanel-child');
          /* <DEPRECATED> */
          msg.child.addClass('p-StackedPanel-child');
          /* </DEPRECATED> */
      };
      /**
       * A message handler invoked on a `'child-removed'` message.
       */
      StackedPanel.prototype.onChildRemoved = function (msg) {
          msg.child.removeClass('lm-StackedPanel-child');
          /* <DEPRECATED> */
          msg.child.removeClass('p-StackedPanel-child');
          /* </DEPRECATED> */
          this._widgetRemoved.emit(msg.child);
      };
      return StackedPanel;
  }(Panel));
  /**
   * The namespace for the module implementation details.
   */
  var Private$g;
  (function (Private) {
      /**
       * Create a stacked layout for the given panel options.
       */
      function createLayout(options) {
          return options.layout || new StackedLayout();
      }
      Private.createLayout = createLayout;
  })(Private$g || (Private$g = {}));

  /**
   * A widget which combines a `TabBar` and a `StackedPanel`.
   *
   * #### Notes
   * This is a simple panel which handles the common case of a tab bar
   * placed next to a content area. The selected tab controls the widget
   * which is shown in the content area.
   *
   * For use cases which require more control than is provided by this
   * panel, the `TabBar` widget may be used independently.
   */
  var TabPanel = /** @class */ (function (_super) {
      __extends$5(TabPanel, _super);
      /**
       * Construct a new tab panel.
       *
       * @param options - The options for initializing the tab panel.
       */
      function TabPanel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this) || this;
          _this._currentChanged = new Signal(_this);
          _this.addClass('lm-TabPanel');
          /* <DEPRECATED> */
          _this.addClass('p-TabPanel');
          /* </DEPRECATED> */
          // Create the tab bar and stacked panel.
          _this.tabBar = new TabBar(options);
          _this.tabBar.addClass('lm-TabPanel-tabBar');
          _this.stackedPanel = new StackedPanel();
          _this.stackedPanel.addClass('lm-TabPanel-stackedPanel');
          /* <DEPRECATED> */
          _this.tabBar.addClass('p-TabPanel-tabBar');
          _this.stackedPanel.addClass('p-TabPanel-stackedPanel');
          /* </DEPRECATED> */
          // Connect the tab bar signal handlers.
          _this.tabBar.tabMoved.connect(_this._onTabMoved, _this);
          _this.tabBar.currentChanged.connect(_this._onCurrentChanged, _this);
          _this.tabBar.tabCloseRequested.connect(_this._onTabCloseRequested, _this);
          _this.tabBar.tabActivateRequested.connect(_this._onTabActivateRequested, _this);
          // Connect the stacked panel signal handlers.
          _this.stackedPanel.widgetRemoved.connect(_this._onWidgetRemoved, _this);
          // Get the data related to the placement.
          _this._tabPlacement = options.tabPlacement || 'top';
          var direction = Private$h.directionFromPlacement(_this._tabPlacement);
          var orientation = Private$h.orientationFromPlacement(_this._tabPlacement);
          // Configure the tab bar for the placement.
          _this.tabBar.orientation = orientation;
          _this.tabBar.dataset['placement'] = _this._tabPlacement;
          // Create the box layout.
          var layout = new BoxLayout({ direction: direction, spacing: 0 });
          // Set the stretch factors for the child widgets.
          BoxLayout.setStretch(_this.tabBar, 0);
          BoxLayout.setStretch(_this.stackedPanel, 1);
          // Add the child widgets to the layout.
          layout.addWidget(_this.tabBar);
          layout.addWidget(_this.stackedPanel);
          // Install the layout on the tab panel.
          _this.layout = layout;
          return _this;
      }
      Object.defineProperty(TabPanel.prototype, "currentChanged", {
          /**
           * A signal emitted when the current tab is changed.
           *
           * #### Notes
           * This signal is emitted when the currently selected tab is changed
           * either through user or programmatic interaction.
           *
           * Notably, this signal is not emitted when the index of the current
           * tab changes due to tabs being inserted, removed, or moved. It is
           * only emitted when the actual current tab node is changed.
           */
          get: function () {
              return this._currentChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "currentIndex", {
          /**
           * Get the index of the currently selected tab.
           *
           * #### Notes
           * This will be `-1` if no tab is selected.
           */
          get: function () {
              return this.tabBar.currentIndex;
          },
          /**
           * Set the index of the currently selected tab.
           *
           * #### Notes
           * If the index is out of range, it will be set to `-1`.
           */
          set: function (value) {
              this.tabBar.currentIndex = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "currentWidget", {
          /**
           * Get the currently selected widget.
           *
           * #### Notes
           * This will be `null` if there is no selected tab.
           */
          get: function () {
              var title = this.tabBar.currentTitle;
              return title ? title.owner : null;
          },
          /**
           * Set the currently selected widget.
           *
           * #### Notes
           * If the widget is not in the panel, it will be set to `null`.
           */
          set: function (value) {
              this.tabBar.currentTitle = value ? value.title : null;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "tabsMovable", {
          /**
           * Get the whether the tabs are movable by the user.
           *
           * #### Notes
           * Tabs can always be moved programmatically.
           */
          get: function () {
              return this.tabBar.tabsMovable;
          },
          /**
           * Set the whether the tabs are movable by the user.
           *
           * #### Notes
           * Tabs can always be moved programmatically.
           */
          set: function (value) {
              this.tabBar.tabsMovable = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "tabPlacement", {
          /**
           * Get the tab placement for the tab panel.
           *
           * #### Notes
           * This controls the position of the tab bar relative to the content.
           */
          get: function () {
              return this._tabPlacement;
          },
          /**
           * Set the tab placement for the tab panel.
           *
           * #### Notes
           * This controls the position of the tab bar relative to the content.
           */
          set: function (value) {
              // Bail if the placement does not change.
              if (this._tabPlacement === value) {
                  return;
              }
              // Update the internal value.
              this._tabPlacement = value;
              // Get the values related to the placement.
              var direction = Private$h.directionFromPlacement(value);
              var orientation = Private$h.orientationFromPlacement(value);
              // Configure the tab bar for the placement.
              this.tabBar.orientation = orientation;
              this.tabBar.dataset['placement'] = value;
              // Update the layout direction.
              this.layout.direction = direction;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "widgets", {
          /**
           * A read-only array of the widgets in the panel.
           */
          get: function () {
              return this.stackedPanel.widgets;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Add a widget to the end of the tab panel.
       *
       * @param widget - The widget to add to the tab panel.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       *
       * The widget's `title` is used to populate the tab.
       */
      TabPanel.prototype.addWidget = function (widget) {
          this.insertWidget(this.widgets.length, widget);
      };
      /**
       * Insert a widget into the tab panel at a specified index.
       *
       * @param index - The index at which to insert the widget.
       *
       * @param widget - The widget to insert into to the tab panel.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       *
       * The widget's `title` is used to populate the tab.
       */
      TabPanel.prototype.insertWidget = function (index, widget) {
          if (widget !== this.currentWidget) {
              widget.hide();
          }
          this.stackedPanel.insertWidget(index, widget);
          this.tabBar.insertTab(index, widget.title);
      };
      /**
       * Handle the `currentChanged` signal from the tab bar.
       */
      TabPanel.prototype._onCurrentChanged = function (sender, args) {
          // Extract the previous and current title from the args.
          var previousIndex = args.previousIndex, previousTitle = args.previousTitle, currentIndex = args.currentIndex, currentTitle = args.currentTitle;
          // Extract the widgets from the titles.
          var previousWidget = previousTitle ? previousTitle.owner : null;
          var currentWidget = currentTitle ? currentTitle.owner : null;
          // Hide the previous widget.
          if (previousWidget) {
              previousWidget.hide();
          }
          // Show the current widget.
          if (currentWidget) {
              currentWidget.show();
          }
          // Emit the `currentChanged` signal for the tab panel.
          this._currentChanged.emit({
              previousIndex: previousIndex, previousWidget: previousWidget, currentIndex: currentIndex, currentWidget: currentWidget
          });
          // Flush the message loop on IE and Edge to prevent flicker.
          if (Platform.IS_EDGE || Platform.IS_IE) {
              MessageLoop.flush();
          }
      };
      /**
       * Handle the `tabActivateRequested` signal from the tab bar.
       */
      TabPanel.prototype._onTabActivateRequested = function (sender, args) {
          args.title.owner.activate();
      };
      /**
       * Handle the `tabCloseRequested` signal from the tab bar.
       */
      TabPanel.prototype._onTabCloseRequested = function (sender, args) {
          args.title.owner.close();
      };
      /**
       * Handle the `tabMoved` signal from the tab bar.
       */
      TabPanel.prototype._onTabMoved = function (sender, args) {
          this.stackedPanel.insertWidget(args.toIndex, args.title.owner);
      };
      /**
       * Handle the `widgetRemoved` signal from the stacked panel.
       */
      TabPanel.prototype._onWidgetRemoved = function (sender, widget) {
          this.tabBar.removeTab(widget.title);
      };
      return TabPanel;
  }(Widget));
  /**
   * The namespace for the module implementation details.
   */
  var Private$h;
  (function (Private) {
      /**
       * Convert a tab placement to tab bar orientation.
       */
      function orientationFromPlacement(plc) {
          return placementToOrientationMap[plc];
      }
      Private.orientationFromPlacement = orientationFromPlacement;
      /**
       * Convert a tab placement to a box layout direction.
       */
      function directionFromPlacement(plc) {
          return placementToDirectionMap[plc];
      }
      Private.directionFromPlacement = directionFromPlacement;
      /**
       * A mapping of tab placement to tab bar orientation.
       */
      var placementToOrientationMap = {
          'top': 'horizontal',
          'left': 'vertical',
          'right': 'vertical',
          'bottom': 'horizontal'
      };
      /**
       * A mapping of tab placement to box layout direction.
       */
      var placementToDirectionMap = {
          'top': 'top-to-bottom',
          'left': 'left-to-right',
          'right': 'right-to-left',
          'bottom': 'bottom-to-top'
      };
  })(Private$h || (Private$h = {}));

  var index_es6 = /*#__PURE__*/Object.freeze({
    get BoxEngine () { return BoxEngine; },
    get BoxLayout () { return BoxLayout; },
    get BoxPanel () { return BoxPanel; },
    BoxSizer: BoxSizer,
    get CommandPalette () { return CommandPalette; },
    ContextMenu: ContextMenu,
    DockLayout: DockLayout,
    get DockPanel () { return DockPanel; },
    FocusTracker: FocusTracker,
    get GridLayout () { return GridLayout; },
    get Layout () { return Layout; },
    LayoutItem: LayoutItem,
    get Menu () { return Menu; },
    get MenuBar () { return MenuBar; },
    Panel: Panel,
    PanelLayout: PanelLayout,
    ScrollBar: ScrollBar,
    SingletonLayout: SingletonLayout,
    get SplitLayout () { return SplitLayout; },
    get SplitPanel () { return SplitPanel; },
    StackedLayout: StackedLayout,
    StackedPanel: StackedPanel,
    get TabBar () { return TabBar; },
    TabPanel: TabPanel,
    Title: Title,
    get Widget () { return Widget; }
  });

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$6 = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$4 = (undefined && undefined.__assign) || function () {
      __assign$4 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$4.apply(this, arguments);
  };
  /**
   * Replace model ids with models recursively.
   */
  function unpack_models(value, manager) {
      if (Array.isArray(value)) {
          var unpacked_1 = [];
          value.forEach(function (sub_value, key) {
              unpacked_1.push(unpack_models(sub_value, manager));
          });
          return Promise.all(unpacked_1);
      }
      else if (value instanceof Object) {
          var unpacked_2 = {};
          Object.keys(value).forEach(function (key) {
              unpacked_2[key] = unpack_models(value[key], manager);
          });
          return resolvePromisesDict(unpacked_2);
      }
      else if (typeof value === 'string' && value.slice(0, 10) === 'IPY_MODEL_') {
          // get_model returns a promise already
          return manager.get_model(value.slice(10, value.length));
      }
      else {
          return Promise.resolve(value);
      }
  }
  var WidgetModel = /** @class */ (function (_super) {
      __extends$6(WidgetModel, _super);
      function WidgetModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * The default attributes.
       */
      WidgetModel.prototype.defaults = function () {
          return {
              _model_module: '@jupyter-widgets/base',
              _model_name: 'WidgetModel',
              _model_module_version: JUPYTER_WIDGETS_VERSION,
              _view_module: '@jupyter-widgets/base',
              _view_name: null,
              _view_module_version: JUPYTER_WIDGETS_VERSION,
              _view_count: null,
          };
      };
      /**
       * Test to see if the model has been synced with the server.
       *
       * #### Notes
       * As of backbone 1.1, backbone ignores `patch` if it thinks the
       * model has never been pushed.
       */
      WidgetModel.prototype.isNew = function () {
          return false;
      };
      /**
       * Constructor
       *
       * Initializes a WidgetModel instance. Called by the Backbone constructor.
       *
       * Parameters
       * ----------
       * widget_manager : WidgetManager instance
       * model_id : string
       *      An ID unique to this model.
       * comm : Comm instance (optional)
       */
      WidgetModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
          // Attributes should be initialized here, since user initialization may depend on it
          this.widget_manager = options.widget_manager;
          this.model_id = options.model_id;
          var comm = options.comm;
          this.views = Object.create(null);
          this.state_change = Promise.resolve();
          this._closed = false;
          this._state_lock = null;
          this._msg_buffer = null;
          this._msg_buffer_callbacks = null;
          this._pending_msgs = 0;
          // _buffered_state_diff must be created *after* the super.initialize
          // call above. See the note in the set() method below.
          this._buffered_state_diff = {};
          if (comm) {
              // Remember comm associated with the model.
              this.comm = comm;
              // Hook comm messages up to model.
              comm.on_close(this._handle_comm_closed.bind(this));
              comm.on_msg(this._handle_comm_msg.bind(this));
              this.comm_live = true;
          }
          else {
              this.comm_live = false;
          }
      };
      Object.defineProperty(WidgetModel.prototype, "comm_live", {
          get: function () {
              return this._comm_live;
          },
          set: function (x) {
              this._comm_live = x;
              this.trigger('comm_live_update');
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Send a custom msg over the comm.
       */
      WidgetModel.prototype.send = function (content, callbacks, buffers) {
          if (this.comm !== undefined) {
              var data = { method: 'custom', content: content };
              this.comm.send(data, callbacks, {}, buffers);
          }
      };
      /**
       * Close model
       *
       * @param comm_closed - true if the comm is already being closed. If false, the comm will be closed.
       *
       * @returns - a promise that is fulfilled when all the associated views have been removed.
       */
      WidgetModel.prototype.close = function (comm_closed) {
          var _this = this;
          if (comm_closed === void 0) { comm_closed = false; }
          // can only be closed once.
          if (this._closed) {
              return;
          }
          this._closed = true;
          if (this.comm && !comm_closed) {
              this.comm.close();
          }
          this.stopListening();
          this.trigger('destroy', this);
          if (this.comm) {
              delete this.comm;
          }
          // Delete all views of this model
          var views = Object.keys(this.views).map(function (id) {
              return _this.views[id].then(function (view) { return view.remove(); });
          });
          delete this.views;
          return Promise.all(views).then(function () { return; });
      };
      /**
       * Handle when a widget comm is closed.
       */
      WidgetModel.prototype._handle_comm_closed = function (msg) {
          this.trigger('comm:close');
          this.close(true);
      };
      /**
       * Handle incoming comm msg.
       */
      WidgetModel.prototype._handle_comm_msg = function (msg) {
          var _this = this;
          var data = msg.content.data;
          var method = data.method;
          // tslint:disable-next-line:switch-default
          switch (method) {
              case 'update':
                  this.state_change = this.state_change
                      .then(function () {
                      var state = data.state;
                      var buffer_paths = data.buffer_paths || [];
                      // Make sure the buffers are DataViews
                      var buffers = (msg.buffers || []).map(function (b) {
                          if (b instanceof DataView) {
                              return b;
                          }
                          else {
                              return new DataView(b instanceof ArrayBuffer ? b : b.buffer);
                          }
                      });
                      put_buffers(state, buffer_paths, buffers);
                      return _this.constructor._deserialize_state(state, _this.widget_manager);
                  }).then(function (state) {
                      _this.set_state(state);
                  }).catch(reject("Could not process update msg for model id: " + this.model_id, true));
                  return this.state_change;
              case 'custom':
                  this.trigger('msg:custom', data.content, msg.buffers);
                  return Promise.resolve();
          }
      };
      /**
       * Handle when a widget is updated from the backend.
       *
       * This function is meant for internal use only. Values set here will not be propagated on a sync.
       */
      WidgetModel.prototype.set_state = function (state) {
          this._state_lock = state;
          try {
              this.set(state);
          }
          catch (e) {
              console.error("Error setting state: " + e.message);
          }
          finally {
              this._state_lock = null;
          }
      };
      /**
       * Get the serializable state of the model.
       *
       * If drop_default is truthy, attributes that are equal to their default
       * values are dropped.
       */
      WidgetModel.prototype.get_state = function (drop_defaults) {
          var fullState = this.attributes;
          if (drop_defaults) {
              // if defaults is a function, call it
              var d = this.defaults;
              var defaults_1 = (typeof d === 'function') ? d.call(this) : d;
              var state_1 = {};
              Object.keys(fullState).forEach(function (key) {
                  if (!(isEqual$1(fullState[key], defaults_1[key]))) {
                      state_1[key] = fullState[key];
                  }
              });
              return state_1;
          }
          else {
              return __assign$4({}, fullState);
          }
      };
      /**
       * Handle status msgs.
       *
       * execution_state : ('busy', 'idle', 'starting')
       */
      WidgetModel.prototype._handle_status = function (msg) {
          if (this.comm !== void 0) {
              if (msg.content.execution_state === 'idle') {
                  this._pending_msgs--;
                  // Send buffer if one is waiting and we are below the throttle.
                  if (this._msg_buffer !== null
                      && this._pending_msgs < 1) {
                      this.send_sync_message(this._msg_buffer, this._msg_buffer_callbacks);
                      this._msg_buffer = null;
                      this._msg_buffer_callbacks = null;
                  }
              }
          }
      };
      /**
       * Create msg callbacks for a comm msg.
       */
      WidgetModel.prototype.callbacks = function (view) {
          return this.widget_manager.callbacks(view);
      };
      /**
       * Set one or more values.
       *
       * We just call the super method, in which val and options are optional.
       * Handles both "key", value and {key: value} -style arguments.
       */
      WidgetModel.prototype.set = function (key, val, options) {
          // Call our patched backbone set. See #1642 and #1643.
          var return_value = set.call(this, key, val, options);
          // Backbone only remembers the diff of the most recent set()
          // operation.  Calling set multiple times in a row results in a
          // loss of change information.  Here we keep our own running diff.
          //
          // We don't buffer the state set in the constructor (including
          // defaults), so we first check to see if we've initialized _buffered_state_diff.
          // which happens after the constructor sets attributes at creation.
          if (this._buffered_state_diff !== void 0) {
              var attrs = this.changedAttributes() || {};
              // The state_lock lists attributes that are currently being changed
              // right now from a kernel message. We don't want to send these
              // non-changes back to the kernel, so we delete them out of attrs if
              // they haven't changed from their state_lock value.
              // The state lock could be null or undefined (if set is being called from
              // the initializer).
              if (this._state_lock) {
                  for (var _i = 0, _a = Object.keys(this._state_lock); _i < _a.length; _i++) {
                      var key_1 = _a[_i];
                      if (attrs[key_1] === this._state_lock[key_1]) {
                          delete attrs[key_1];
                      }
                  }
              }
              this._buffered_state_diff = assign(this._buffered_state_diff, attrs);
          }
          return return_value;
      };
      /**
       * Handle sync to the back-end.  Called when a model.save() is called.
       *
       * Make sure a comm exists.
       *
       * Parameters
       * ----------
       * method : create, update, patch, delete, read
       *   create/update always send the full attribute set
       *   patch - only send attributes listed in options.attrs, and if we
       *   are queuing up messages, combine with previous messages that have
       *   not been sent yet
       * model : the model we are syncing
       *   will normally be the same as `this`
       * options : dict
       *   the `attrs` key, if it exists, gives an {attr: value} dict that
       *   should be synced, otherwise, sync all attributes.
       *
       */
      WidgetModel.prototype.sync = function (method, model, options) {
          if (options === void 0) { options = {}; }
          // the typing is to return `any` since the super.sync method returns a JqXHR, but we just return false if there is an error.
          if (this.comm === undefined) {
              throw 'Syncing error: no comm channel defined';
          }
          var attrs = (method === 'patch') ? options.attrs : model.get_state(options.drop_defaults);
          // The state_lock lists attributes that are currently being changed
          // right now from a kernel message. We don't want to send these
          // non-changes back to the kernel, so we delete them out of attrs if
          // they haven't changed from their state_lock value.
          // The state lock could be null or undefined (if this is triggered
          // from the initializer).
          if (this._state_lock) {
              for (var _i = 0, _a = Object.keys(this._state_lock); _i < _a.length; _i++) {
                  var key = _a[_i];
                  if (attrs[key] === this._state_lock[key]) {
                      delete attrs[key];
                  }
              }
          }
          var msgState = this.serialize(attrs);
          if (Object.keys(msgState).length > 0) {
              // If this message was sent via backbone itself, it will not
              // have any callbacks.  It's important that we create callbacks
              // so we can listen for status messages, etc...
              var callbacks = options.callbacks || this.callbacks();
              // Check throttle.
              if (this._pending_msgs >= 1) {
                  // The throttle has been exceeded, buffer the current msg so
                  // it can be sent once the kernel has finished processing
                  // some of the existing messages.
                  // Combine updates if it is a 'patch' sync, otherwise replace updates
                  switch (method) {
                      case 'patch':
                          this._msg_buffer = assign(this._msg_buffer || {}, msgState);
                          break;
                      case 'update':
                      case 'create':
                          this._msg_buffer = msgState;
                          break;
                      default:
                          throw 'unrecognized syncing method';
                  }
                  this._msg_buffer_callbacks = callbacks;
              }
              else {
                  // We haven't exceeded the throttle, send the message like
                  // normal.
                  this.send_sync_message(attrs, callbacks);
                  // Since the comm is a one-way communication, assume the message
                  // arrived and was processed successfully.
                  // Don't call options.success since we don't have a model back from
                  // the server. Note that this means we don't have the Backbone
                  // 'sync' event.
              }
          }
      };
      /**
       * Serialize widget state.
       *
       * A serializer is a function which takes in a state attribute and a widget,
       * and synchronously returns a JSONable object. The returned object will
       * have toJSON called if possible, and the final result should be a
       * primitive object that is a snapshot of the widget state that may have
       * binary array buffers.
       */
      WidgetModel.prototype.serialize = function (state) {
          var serializers = this.constructor.serializers || {};
          for (var _i = 0, _a = Object.keys(state); _i < _a.length; _i++) {
              var k = _a[_i];
              try {
                  if (serializers[k] && serializers[k].serialize) {
                      state[k] = (serializers[k].serialize)(state[k], this);
                  }
                  else {
                      // the default serializer just deep-copies the object
                      state[k] = JSON.parse(JSON.stringify(state[k]));
                  }
                  if (state[k] && state[k].toJSON) {
                      state[k] = state[k].toJSON();
                  }
              }
              catch (e) {
                  console.error('Error serializing widget state attribute: ', k);
                  throw e;
              }
          }
          return state;
      };
      /**
       * Send a sync message to the kernel.
       */
      WidgetModel.prototype.send_sync_message = function (state, callbacks) {
          var _this = this;
          if (callbacks === void 0) { callbacks = {}; }
          try {
              callbacks.iopub = callbacks.iopub || {};
              var statuscb_1 = callbacks.iopub.status;
              callbacks.iopub.status = function (msg) {
                  _this._handle_status(msg);
                  if (statuscb_1) {
                      statuscb_1(msg);
                  }
              };
              // split out the binary buffers
              var split = remove_buffers(state);
              this.comm.send({
                  method: 'update',
                  state: split.state,
                  buffer_paths: split.buffer_paths
              }, callbacks, {}, split.buffers);
              this._pending_msgs++;
          }
          catch (e) {
              console.error('Could not send widget sync message', e);
          }
      };
      /**
       * Push this model's state to the back-end
       *
       * This invokes a Backbone.Sync.
       */
      WidgetModel.prototype.save_changes = function (callbacks) {
          if (this.comm_live) {
              var options = { patch: true };
              if (callbacks) {
                  options.callbacks = callbacks;
              }
              this.save(this._buffered_state_diff, options);
              this._buffered_state_diff = {};
          }
      };
      /**
       * on_some_change(['key1', 'key2'], foo, context) differs from
       * on('change:key1 change:key2', foo, context).
       * If the widget attributes key1 and key2 are both modified,
       * the second form will result in foo being called twice
       * while the first will call foo only once.
       */
      WidgetModel.prototype.on_some_change = function (keys, callback, context) {
          var scope = this;
          this.on('change', function () {
              if (keys.some(scope.hasChanged, scope)) {
                  callback.apply(context, arguments);
              }
          }, this);
      };
      /**
       * Serialize the model.  See the deserialization function at the top of this file
       * and the kernel-side serializer/deserializer.
       */
      WidgetModel.prototype.toJSON = function (options) {
          return "IPY_MODEL_" + this.model_id;
      };
      /**
       * Returns a promise for the deserialized state. The second argument
       * is an instance of widget manager, which is required for the
       * deserialization of widget models.
       */
      WidgetModel._deserialize_state = function (state, manager) {
          var serializers = this.serializers;
          var deserialized;
          if (serializers) {
              deserialized = {};
              for (var k in state) {
                  if (serializers[k] && serializers[k].deserialize) {
                      deserialized[k] = (serializers[k].deserialize)(state[k], manager);
                  }
                  else {
                      deserialized[k] = state[k];
                  }
              }
          }
          else {
              deserialized = state;
          }
          return resolvePromisesDict(deserialized);
      };
      return WidgetModel;
  }(__export2));
  var DOMWidgetModel = /** @class */ (function (_super) {
      __extends$6(DOMWidgetModel, _super);
      function DOMWidgetModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DOMWidgetModel.prototype.defaults = function () {
          return assign(_super.prototype.defaults.call(this), {
              _dom_classes: []
              // We do not declare defaults for the layout and style attributes.
              // Those defaults are constructed on the kernel side and synced here
              // as needed, and our code here copes with those attributes being
              // undefined. See
              // https://github.com/jupyter-widgets/ipywidgets/issues/1620 and
              // https://github.com/jupyter-widgets/ipywidgets/pull/1621
          });
      };
      DOMWidgetModel.serializers = __assign$4(__assign$4({}, WidgetModel.serializers), { layout: { deserialize: unpack_models }, style: { deserialize: unpack_models } });
      return DOMWidgetModel;
  }(WidgetModel));
  var WidgetView = /** @class */ (function (_super) {
      __extends$6(WidgetView, _super);
      /**
       * Public constructor.
       */
      function WidgetView(options) {
          return _super.call(this, options) || this;
      }
      /**
       * Initializer, called at the end of the constructor.
       */
      WidgetView.prototype.initialize = function (parameters) {
          var _this = this;
          this.listenTo(this.model, 'change', function () {
              var changed = Object.keys(_this.model.changedAttributes() || {});
              if (changed[0] === '_view_count' && changed.length === 1) {
                  // Just the view count was updated
                  return;
              }
              _this.update();
          });
          this.options = parameters.options;
          this.once('remove', function () {
              if (typeof (_this.model.get('_view_count')) === 'number') {
                  _this.model.set('_view_count', _this.model.get('_view_count') - 1);
                  _this.model.save_changes();
              }
          });
          this.once('displayed', function () {
              if (typeof (_this.model.get('_view_count')) === 'number') {
                  _this.model.set('_view_count', _this.model.get('_view_count') + 1);
                  _this.model.save_changes();
              }
          });
          this.displayed = new Promise(function (resolve, reject) {
              _this.once('displayed', resolve);
          });
      };
      /**
       * Triggered on model change.
       *
       * Update view to be consistent with this.model
       */
      WidgetView.prototype.update = function (options) {
          return;
      };
      /**
       * Render a view
       *
       * @returns the view or a promise to the view.
       */
      WidgetView.prototype.render = function () {
          return;
      };
      /**
       * Create and promise that resolves to a child view of a given model
       */
      WidgetView.prototype.create_child_view = function (child_model, options) {
          if (options === void 0) { options = {}; }
          options = __assign$4({ parent: this }, options);
          return this.model.widget_manager.create_view(child_model, options)
              .catch(reject('Could not create child view', true));
      };
      /**
       * Create msg callbacks for a comm msg.
       */
      WidgetView.prototype.callbacks = function () {
          return this.model.callbacks(this);
      };
      /**
       * Send a custom msg associated with this view.
       */
      WidgetView.prototype.send = function (content, buffers) {
          this.model.send(content, this.callbacks(), buffers);
      };
      WidgetView.prototype.touch = function () {
          this.model.save_changes(this.callbacks());
      };
      WidgetView.prototype.remove = function () {
          // Raise a remove event when the view is removed.
          _super.prototype.remove.call(this);
          this.trigger('remove');
          return this;
      };
      return WidgetView;
  }(NativeView));
  var JupyterPhosphorWidget = /** @class */ (function (_super) {
      __extends$6(JupyterPhosphorWidget, _super);
      function JupyterPhosphorWidget(options) {
          var _this = this;
          var view = options.view;
          delete options.view;
          _this = _super.call(this, options) || this;
          _this._view = view;
          return _this;
      }
      /**
       * Dispose the widget.
       *
       * This causes the view to be destroyed as well with 'remove'
       */
      JupyterPhosphorWidget.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          if (this._view) {
              this._view.remove();
          }
          this._view = null;
      };
      /**
       * Process the phosphor message.
       *
       * Any custom phosphor widget used inside a Jupyter widget should override
       * the processMessage function like this.
       */
      JupyterPhosphorWidget.prototype.processMessage = function (msg) {
          _super.prototype.processMessage.call(this, msg);
          this._view.processPhosphorMessage(msg);
      };
      return JupyterPhosphorWidget;
  }(Widget));
  var JupyterPhosphorPanelWidget = /** @class */ (function (_super) {
      __extends$6(JupyterPhosphorPanelWidget, _super);
      function JupyterPhosphorPanelWidget(options) {
          var _this = this;
          var view = options.view;
          delete options.view;
          _this = _super.call(this, options) || this;
          _this._view = view;
          return _this;
      }
      /**
       * Process the phosphor message.
       *
       * Any custom phosphor widget used inside a Jupyter widget should override
       * the processMessage function like this.
       */
      JupyterPhosphorPanelWidget.prototype.processMessage = function (msg) {
          _super.prototype.processMessage.call(this, msg);
          this._view.processPhosphorMessage(msg);
      };
      /**
       * Dispose the widget.
       *
       * This causes the view to be destroyed as well with 'remove'
       */
      JupyterPhosphorPanelWidget.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          if (this._view) {
              this._view.remove();
          }
          this._view = null;
      };
      return JupyterPhosphorPanelWidget;
  }(Panel));
  var DOMWidgetView = /** @class */ (function (_super) {
      __extends$6(DOMWidgetView, _super);
      function DOMWidgetView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor
       */
      DOMWidgetView.prototype.initialize = function (parameters) {
          var _this = this;
          _super.prototype.initialize.call(this, parameters);
          this.listenTo(this.model, 'change:_dom_classes', function (model, new_classes) {
              var old_classes = model.previous('_dom_classes');
              _this.update_classes(old_classes, new_classes);
          });
          this.layoutPromise = Promise.resolve();
          this.listenTo(this.model, 'change:layout', function (model, value) {
              _this.setLayout(value, model.previous('layout'));
          });
          this.stylePromise = Promise.resolve();
          this.listenTo(this.model, 'change:style', function (model, value) {
              _this.setStyle(value, model.previous('style'));
          });
          this.displayed.then(function () {
              _this.update_classes([], _this.model.get('_dom_classes'));
              _this.setLayout(_this.model.get('layout'));
              _this.setStyle(_this.model.get('style'));
          });
          this._comm_live_update();
          this.listenTo(this.model, 'comm_live_update', function () {
              _this._comm_live_update();
          });
      };
      DOMWidgetView.prototype.setLayout = function (layout, oldLayout) {
          var _this = this;
          if (layout) {
              this.layoutPromise = this.layoutPromise.then(function (oldLayoutView) {
                  if (oldLayoutView) {
                      oldLayoutView.unlayout();
                      _this.stopListening(oldLayoutView.model);
                      oldLayoutView.remove();
                  }
                  return _this.create_child_view(layout).then(function (view) {
                      // Trigger the displayed event of the child view.
                      return _this.displayed.then(function () {
                          view.trigger('displayed');
                          _this.listenTo(view.model, 'change', function () {
                              // Post (asynchronous) so layout changes can take
                              // effect first.
                              MessageLoop.postMessage(_this.pWidget, Widget.ResizeMessage.UnknownSize);
                          });
                          MessageLoop.postMessage(_this.pWidget, Widget.ResizeMessage.UnknownSize);
                          return view;
                      });
                  }).catch(reject('Could not add LayoutView to DOMWidgetView', true));
              });
          }
      };
      DOMWidgetView.prototype.setStyle = function (style, oldStyle) {
          var _this = this;
          if (style) {
              this.stylePromise = this.stylePromise.then(function (oldStyleView) {
                  if (oldStyleView) {
                      oldStyleView.unstyle();
                      _this.stopListening(oldStyleView.model);
                      oldStyleView.remove();
                  }
                  return _this.create_child_view(style).then(function (view) {
                      // Trigger the displayed event of the child view.
                      return _this.displayed.then(function () {
                          view.trigger('displayed');
                          // Unlike for the layout attribute, style changes don't
                          // trigger phosphor resize messages.
                          return view;
                      });
                  }).catch(reject('Could not add styleView to DOMWidgetView', true));
              });
          }
      };
      /**
       * Update the DOM classes applied to an element, default to this.el.
       */
      DOMWidgetView.prototype.update_classes = function (old_classes, new_classes, el) {
          if (el === undefined) {
              el = this.el;
          }
          difference(old_classes, new_classes).map(function (c) {
              if (el.classList) { // classList is not supported by IE for svg elements
                  el.classList.remove(c);
              }
              else {
                  el.setAttribute('class', el.getAttribute('class').replace(c, ''));
              }
          });
          difference(new_classes, old_classes).map(function (c) {
              if (el.classList) { // classList is not supported by IE for svg elements
                  el.classList.add(c);
              }
              else {
                  el.setAttribute('class', el.getAttribute('class').concat(' ', c));
              }
          });
      };
      /**
       * Update the DOM classes applied to the widget based on a single
       * trait's value.
       *
       * Given a trait value classes map, this function automatically
       * handles applying the appropriate classes to the widget element
       * and removing classes that are no longer valid.
       *
       * Parameters
       * ----------
       * class_map: dictionary
       *  Dictionary of trait values to class lists.
       *  Example:
       *      {
       *          success: ['alert', 'alert-success'],
       *          info: ['alert', 'alert-info'],
       *          warning: ['alert', 'alert-warning'],
       *          danger: ['alert', 'alert-danger']
       *      };
       * trait_name: string
       *  Name of the trait to check the value of.
       * el: optional DOM element handle, defaults to this.el
       *  Element that the classes are applied to.
       */
      DOMWidgetView.prototype.update_mapped_classes = function (class_map, trait_name, el) {
          var key = this.model.previous(trait_name);
          var old_classes = class_map[key] ? class_map[key] : [];
          key = this.model.get(trait_name);
          var new_classes = class_map[key] ? class_map[key] : [];
          this.update_classes(old_classes, new_classes, el || this.el);
      };
      DOMWidgetView.prototype.set_mapped_classes = function (class_map, trait_name, el) {
          var key = this.model.get(trait_name);
          var new_classes = class_map[key] ? class_map[key] : [];
          this.update_classes([], new_classes, el || this.el);
      };
      DOMWidgetView.prototype._setElement = function (el) {
          if (this.pWidget) {
              this.pWidget.dispose();
          }
          this.$el = el instanceof $$1 ? el : $$1(el);
          this.el = this.$el[0];
          this.pWidget = new JupyterPhosphorWidget({
              node: el,
              view: this
          });
      };
      DOMWidgetView.prototype.remove = function () {
          if (this.pWidget) {
              this.pWidget.dispose();
          }
          return _super.prototype.remove.call(this);
      };
      DOMWidgetView.prototype.processPhosphorMessage = function (msg) {
          // tslint:disable-next-line:switch-default
          switch (msg.type) {
              case 'after-attach':
                  this.trigger('displayed');
                  break;
          }
      };
      DOMWidgetView.prototype._comm_live_update = function () {
          if (this.model.comm_live) {
              this.pWidget.removeClass('jupyter-widgets-disconnected');
          }
          else {
              this.pWidget.addClass('jupyter-widgets-disconnected');
          }
      };
      return DOMWidgetView;
  }(WidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$7 = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  /**
   * css properties exposed by the layout widget with their default values.
   */
  var css_properties = {
      align_content: null,
      align_items: null,
      align_self: null,
      border: null,
      bottom: null,
      display: null,
      flex: null,
      flex_flow: null,
      height: null,
      justify_content: null,
      justify_items: null,
      left: null,
      margin: null,
      max_height: null,
      max_width: null,
      min_height: null,
      min_width: null,
      overflow: null,
      overflow_x: null,
      overflow_y: null,
      order: null,
      padding: null,
      right: null,
      top: null,
      visibility: null,
      width: null,
      // image-specific
      object_fit: null,
      object_position: null,
      // container
      grid_auto_columns: null,
      grid_auto_flow: null,
      grid_auto_rows: null,
      grid_gap: null,
      grid_template_rows: null,
      grid_template_columns: null,
      grid_template_areas: null,
      // items
      grid_row: null,
      grid_column: null,
      grid_area: null
  };
  var LayoutModel = /** @class */ (function (_super) {
      __extends$7(LayoutModel, _super);
      function LayoutModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      LayoutModel.prototype.defaults = function () {
          return assign(_super.prototype.defaults.call(this), {
              _model_name: 'LayoutModel',
              _view_name: 'LayoutView'
          }, css_properties);
      };
      return LayoutModel;
  }(WidgetModel));
  var LayoutView = /** @class */ (function (_super) {
      __extends$7(LayoutView, _super);
      function LayoutView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor
       */
      LayoutView.prototype.initialize = function (parameters) {
          this._traitNames = [];
          _super.prototype.initialize.call(this, parameters);
          // Register the traits that live on the Python side
          for (var _i = 0, _a = Object.keys(css_properties); _i < _a.length; _i++) {
              var key = _a[_i];
              this.registerTrait(key);
          }
      };
      /**
       * Register a CSS trait that is known by the model
       * @param trait
       */
      LayoutView.prototype.registerTrait = function (trait) {
          var _this = this;
          this._traitNames.push(trait);
          // Treat overflow_x and overflow_y as a special case since they are deprecated
          // and interact in special ways with the overflow attribute.
          if (trait === 'overflow_x' || trait === 'overflow_y') {
              // Listen to changes, and set the value on change.
              this.listenTo(this.model, 'change:' + trait, function (model, value) {
                  _this.handleOverflowChange(trait, value);
              });
              // Set the initial value on display.
              this.handleOverflowChange(trait, this.model.get(trait));
              return;
          }
          // Listen to changes, and set the value on change.
          this.listenTo(this.model, 'change:' + trait, function (model, value) {
              _this.handleChange(trait, value);
          });
          // Set the initial value on display.
          this.handleChange(trait, this.model.get(trait));
      };
      /**
       * Get the the name of the css property from the trait name
       * @param  model attribute name
       * @return css property name
       */
      LayoutView.prototype.css_name = function (trait) {
          return trait.replace(/_/g, '-');
      };
      /**
       * Handles when a trait value changes
       */
      LayoutView.prototype.handleChange = function (trait, value) {
          // should be synchronous so that we can measure later.
          var parent = this.options.parent;
          if (parent) {
              if (value === null) {
                  parent.el.style.removeProperty(this.css_name(trait));
              }
              else {
                  parent.el.style[this.css_name(trait)] = value;
              }
          }
          else {
              console.warn('Style not applied because a parent view does not exist');
          }
      };
      /**
       * Handles when the value of overflow_x or overflow_y changes
       */
      LayoutView.prototype.handleOverflowChange = function (trait, value) {
          // This differs from the default handleChange method
          // in that setting `overflow_x` or `overflow_y` to null
          // when `overflow` is null removes the attribute.
          var parent = this.options.parent;
          if (parent) {
              if (value === null) {
                  if (this.model.get('overflow') === null) {
                      parent.el.style.removeProperty(this.css_name(trait));
                  }
              }
              else {
                  parent.el.style[this.css_name(trait)] = value;
              }
          }
          else {
              console.warn('Style not applied because a parent view does not exist');
          }
      };
      /**
       * Remove the styling from the parent view.
       */
      LayoutView.prototype.unlayout = function () {
          var _this = this;
          var parent = this.options.parent;
          this._traitNames.forEach(function (trait) {
              if (parent) {
                  parent.el.style.removeProperty(_this.css_name(trait));
              }
              else {
                  console.warn('Style not removed because a parent view does not exist');
              }
          }, this);
      };
      return LayoutView;
  }(WidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$8 = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var StyleModel = /** @class */ (function (_super) {
      __extends$8(StyleModel, _super);
      function StyleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      StyleModel.prototype.defaults = function () {
          var Derived = this.constructor;
          return assign(_super.prototype.defaults.call(this), {
              _model_name: 'StyleModel',
              _view_name: 'StyleView',
          }, Object.keys(Derived.styleProperties).reduce(function (obj, key) {
              obj[key] = Derived.styleProperties[key].default;
              return obj;
          }, {}));
      };
      StyleModel.styleProperties = {};
      return StyleModel;
  }(WidgetModel));
  var StyleView = /** @class */ (function (_super) {
      __extends$8(StyleView, _super);
      function StyleView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor
       */
      StyleView.prototype.initialize = function (parameters) {
          this._traitNames = [];
          _super.prototype.initialize.call(this, parameters);
          // Register the traits that live on the Python side
          var ModelType = this.model.constructor;
          for (var _i = 0, _a = Object.keys(ModelType.styleProperties); _i < _a.length; _i++) {
              var key = _a[_i];
              this.registerTrait(key);
          }
          // Set the initial styles
          this.style();
      };
      /**
       * Register a CSS trait that is known by the model
       * @param trait
       */
      StyleView.prototype.registerTrait = function (trait) {
          var _this = this;
          this._traitNames.push(trait);
          // Listen to changes, and set the value on change.
          this.listenTo(this.model, 'change:' + trait, function (model, value) {
              _this.handleChange(trait, value);
          });
      };
      /**
       * Handles when a trait value changes
       */
      StyleView.prototype.handleChange = function (trait, value) {
          // should be synchronous so that we can measure later.
          var parent = this.options.parent;
          if (parent) {
              var ModelType = this.model.constructor;
              var styleProperties = ModelType.styleProperties;
              var attribute = styleProperties[trait].attribute;
              var selector = styleProperties[trait].selector;
              var elements = selector ? parent.el.querySelectorAll(selector) : [parent.el];
              if (value === null) {
                  for (var i = 0; i !== elements.length; ++i) {
                      elements[i].style.removeProperty(attribute);
                  }
              }
              else {
                  for (var i = 0; i !== elements.length; ++i) {
                      elements[i].style[attribute] = value;
                  }
              }
          }
          else {
              console.warn('Style not applied because a parent view does not exist');
          }
      };
      /**
       * Apply styles for all registered traits
       */
      StyleView.prototype.style = function () {
          for (var _i = 0, _a = this._traitNames; _i < _a.length; _i++) {
              var trait = _a[_i];
              this.handleChange(trait, this.model.get(trait));
          }
      };
      /**
       * Remove the styling from the parent view.
       */
      StyleView.prototype.unstyle = function () {
          var parent = this.options.parent;
          var ModelType = this.model.constructor;
          var styleProperties = ModelType.styleProperties;
          this._traitNames.forEach(function (trait) {
              if (parent) {
                  var attribute = styleProperties[trait].attribute;
                  var selector = styleProperties[trait].selector;
                  var elements = selector ? parent.el.querySelectorAll(selector) : [parent.el];
                  for (var i = 0; i !== elements.length; ++i) {
                      elements[i].style.removeProperty(attribute);
                  }
              }
              else {
                  console.warn('Style not removed because a parent view does not exist');
              }
          }, this);
      };
      return StyleView;
  }(WidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __awaiter$1 = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
      function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
      return new (P || (P = Promise))(function (resolve, reject) {
          function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
          function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
          function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
          step((generator = generator.apply(thisArg, _arguments || [])).next());
      });
  };
  var __generator$1 = (undefined && undefined.__generator) || function (thisArg, body) {
      var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
      return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
      function verb(n) { return function (v) { return step([n, v]); }; }
      function step(op) {
          if (f) throw new TypeError("Generator is already executing.");
          while (_) try {
              if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
              if (y = 0, t) op = [op[0] & 2, t.value];
              switch (op[0]) {
                  case 0: case 1: t = op; break;
                  case 4: _.label++; return { value: op[1], done: false };
                  case 5: _.label++; y = op[1]; op = [0]; continue;
                  case 7: op = _.ops.pop(); _.trys.pop(); continue;
                  default:
                      if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                      if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                      if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                      if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                      if (t[2]) _.ops.pop();
                      _.trys.pop(); continue;
              }
              op = body.call(thisArg, _);
          } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
          if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
      }
  };
  var shims;
  (function (shims) {
      var services;
      (function (services) {
          /**
           * Public constructor
           * @param jsServicesKernel - @jupyterlab/services Kernel.IKernel instance
           */
          var CommManager = /** @class */ (function () {
              function CommManager(jsServicesKernel) {
                  this.targets = Object.create(null);
                  this.comms = Object.create(null);
                  this.kernel = null;
                  this.jsServicesKernel = null;
                  this.init_kernel(jsServicesKernel);
              }
              /**
               * Hookup kernel events.
               * @param  {Kernel.IKernel} jsServicesKernel - @jupyterlab/services Kernel.IKernel instance
               */
              CommManager.prototype.init_kernel = function (jsServicesKernel) {
                  this.kernel = jsServicesKernel; // These aren't really the same.
                  this.jsServicesKernel = jsServicesKernel;
              };
              /**
               * Creates a new connected comm
               */
              CommManager.prototype.new_comm = function (target_name, data, callbacks, metadata, comm_id, buffers) {
                  return __awaiter$1(this, void 0, void 0, function () {
                      var c, comm;
                      return __generator$1(this, function (_a) {
                          c = this.jsServicesKernel.createComm(target_name, comm_id);
                          comm = new Comm(c);
                          this.register_comm(comm);
                          comm.open(data, callbacks, metadata, buffers);
                          return [2 /*return*/, comm];
                      });
                  });
              };
              /**
               * Register a comm target
               * @param  {string} target_name
               * @param  {(Comm, object) => void} f - callback that is called when the
               *                         comm is made.  Signature of f(comm, msg).
               */
              CommManager.prototype.register_target = function (target_name, f) {
                  var _this = this;
                  var handle = this.jsServicesKernel.registerCommTarget(target_name, function (jsServicesComm, msg) {
                      // Create the comm.
                      var comm = new Comm(jsServicesComm);
                      _this.register_comm(comm);
                      // Call the callback for the comm.
                      try {
                          return f(comm, msg);
                      }
                      catch (e) {
                          comm.close();
                          console.error(e);
                          console.error(new Error('Exception opening new comm'));
                      }
                  });
                  this.targets[target_name] = handle;
              };
              /**
               * Unregisters a comm target
               * @param  {string} target_name
               */
              CommManager.prototype.unregister_target = function (target_name, f) {
                  var handle = this.targets[target_name];
                  handle.dispose();
                  delete this.targets[target_name];
              };
              /**
               * Register a comm in the mapping
               */
              CommManager.prototype.register_comm = function (comm) {
                  this.comms[comm.comm_id] = Promise.resolve(comm);
                  comm.kernel = this.kernel;
                  return comm.comm_id;
              };
              return CommManager;
          }());
          services.CommManager = CommManager;
          /**
           * Public constructor
           * @param  {IComm} jsServicesComm - @jupyterlab/services IComm instance
           */
          var Comm = /** @class */ (function () {
              function Comm(jsServicesComm) {
                  this.jsServicesComm = null;
                  this.kernel = null;
                  this.jsServicesComm = jsServicesComm;
              }
              Object.defineProperty(Comm.prototype, "comm_id", {
                  /**
                   * Comm id
                   * @return {string}
                   */
                  get: function () {
                      return this.jsServicesComm.commId;
                  },
                  enumerable: true,
                  configurable: true
              });
              Object.defineProperty(Comm.prototype, "target_name", {
                  /**
                   * Target name
                   * @return {string}
                   */
                  get: function () {
                      return this.jsServicesComm.targetName;
                  },
                  enumerable: true,
                  configurable: true
              });
              /**
               * Opens a sibling comm in the backend
               * @param  data
               * @param  callbacks
               * @param  metadata
               * @return msg id
               */
              Comm.prototype.open = function (data, callbacks, metadata, buffers) {
                  var future = this.jsServicesComm.open(data, metadata, buffers);
                  this._hookupCallbacks(future, callbacks);
                  return future.msg.header.msg_id;
              };
              /**
               * Sends a message to the sibling comm in the backend
               * @param  data
               * @param  callbacks
               * @param  metadata
               * @param  buffers
               * @return message id
               */
              Comm.prototype.send = function (data, callbacks, metadata, buffers) {
                  var future = this.jsServicesComm.send(data, metadata, buffers);
                  this._hookupCallbacks(future, callbacks);
                  return future.msg.header.msg_id;
              };
              /**
               * Closes the sibling comm in the backend
               * @param  data
               * @param  callbacks
               * @param  metadata
               * @return msg id
               */
              Comm.prototype.close = function (data, callbacks, metadata, buffers) {
                  var future = this.jsServicesComm.close(data, metadata, buffers);
                  this._hookupCallbacks(future, callbacks);
                  return future.msg.header.msg_id;
              };
              /**
               * Register a message handler
               * @param  callback, which is given a message
               */
              Comm.prototype.on_msg = function (callback) {
                  this.jsServicesComm.onMsg = callback.bind(this);
              };
              /**
               * Register a handler for when the comm is closed by the backend
               * @param  callback, which is given a message
               */
              Comm.prototype.on_close = function (callback) {
                  this.jsServicesComm.onClose = callback.bind(this);
              };
              /**
               * Hooks callback object up with @jupyterlab/services IKernelFuture
               * @param  @jupyterlab/services IKernelFuture instance
               * @param  callbacks
               */
              Comm.prototype._hookupCallbacks = function (future, callbacks) {
                  if (callbacks) {
                      future.onReply = function (msg) {
                          if (callbacks.shell && callbacks.shell.reply) {
                              callbacks.shell.reply(msg);
                          }
                          // TODO: Handle payloads.  See https://github.com/jupyter/notebook/blob/master/notebook/static/services/kernels/kernel.js#L923-L947
                      };
                      future.onStdin = function (msg) {
                          if (callbacks.input) {
                              callbacks.input(msg);
                          }
                      };
                      future.onIOPub = function (msg) {
                          if (callbacks.iopub) {
                              if (callbacks.iopub.status && msg.header.msg_type === 'status') {
                                  callbacks.iopub.status(msg);
                              }
                              else if (callbacks.iopub.clear_output && msg.header.msg_type === 'clear_output') {
                                  callbacks.iopub.clear_output(msg);
                              }
                              else if (callbacks.iopub.output) {
                                  switch (msg.header.msg_type) {
                                      case 'display_data':
                                      case 'execute_result':
                                      case 'stream':
                                      case 'error':
                                          callbacks.iopub.output(msg);
                                          break;
                                      default: break;
                                  }
                              }
                          }
                      };
                  }
              };
              return Comm;
          }());
          services.Comm = Comm;
      })(services = shims.services || (shims.services = {}));
  })(shims || (shims = {}));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /**
   * - create_view and remove_view are default functions called when adding or removing views
   * - create_view takes a model and an index and returns a view or a promise for a view for that model
   * - remove_view takes a view and destroys it (including calling `view.remove()`)
   * - each time the update() function is called with a new list, the create and remove
   *   callbacks will be called in an order so that if you append the views created in the
   *   create callback and remove the views in the remove callback, you will duplicate
   *   the order of the list.
   * - the remove callback defaults to just removing the view (e.g., pass in null for the second parameter)
   * - the context defaults to the created ViewList.  If you pass another context, the create and remove
   *   will be called in that context.
   */
  var ViewList = /** @class */ (function () {
      function ViewList(create_view, remove_view, context) {
          this.initialize(create_view, remove_view, context);
      }
      ViewList.prototype.initialize = function (create_view, remove_view, context) {
          this._handler_context = context || this;
          this._models = [];
          this.views = []; // list of promises for views
          this._create_view = create_view;
          this._remove_view = remove_view || function (view) { view.remove(); };
      };
      /**
       * the create_view, remove_view, and context arguments override the defaults
       * specified when the list is created.
       * after this function, the .views attribute is a list of promises for views
       * if you want to perform some action on the list of views, do something like
       * `Promise.all(myviewlist.views).then(function(views) {...});`
       */
      ViewList.prototype.update = function (new_models, create_view, remove_view, context) {
          var remove = remove_view || this._remove_view;
          var create = create_view || this._create_view;
          context = context || this._handler_context;
          var i = 0;
          // first, skip past the beginning of the lists if they are identical
          for (; i < new_models.length; i++) {
              if (i >= this._models.length || new_models[i] !== this._models[i]) {
                  break;
              }
          }
          var first_removed = i;
          // Remove the non-matching items from the old list.
          var removed = this.views.splice(first_removed, this.views.length - first_removed);
          for (var j = 0; j < removed.length; j++) {
              removed[j].then(function (view) {
                  remove.call(context, view);
              });
          }
          // Add the rest of the new list items.
          for (; i < new_models.length; i++) {
              this.views.push(Promise.resolve(create.call(context, new_models[i], i)));
          }
          // make a copy of the input array
          this._models = new_models.slice();
          // return a promise that resolves to all of the resolved views
          return Promise.all(this.views);
      };
      /**
       * removes every view in the list; convenience function for `.update([])`
       * that should be faster
       * returns a promise that resolves after this removal is done
       */
      ViewList.prototype.remove = function () {
          var _this = this;
          return Promise.all(this.views).then(function (views) {
              views.forEach(function (value) { return _this._remove_view.call(_this._handler_context, value); });
              _this.views = [];
              _this._models = [];
          });
      };
      /**
       * Dispose this viewlist.
       *
       * A synchronous function which just deletes references to child views. This
       * function does not call .remove() on child views because that is
       * asynchronous. Use this in cases where child views will be removed in
       * another way.
       */
      ViewList.prototype.dispose = function () {
          this.views = null;
          this._models = null;
      };
      return ViewList;
  }());

  // Copyright (c) Jupyter Development Team.
  /**
   * A runtime interface token for a widget registry.
   */
  var IJupyterWidgetRegistry = new Token('jupyter.extensions.jupyterWidgetRegistry');

  // Copyright (c) Jupyter Development Team.

  var lib$1 = /*#__PURE__*/Object.freeze({
    ManagerBase: ManagerBase,
    serialize_state: serialize_state,
    unpack_models: unpack_models,
    WidgetModel: WidgetModel,
    DOMWidgetModel: DOMWidgetModel,
    WidgetView: WidgetView,
    JupyterPhosphorWidget: JupyterPhosphorWidget,
    JupyterPhosphorPanelWidget: JupyterPhosphorPanelWidget,
    DOMWidgetView: DOMWidgetView,
    LayoutModel: LayoutModel,
    LayoutView: LayoutView,
    StyleModel: StyleModel,
    StyleView: StyleView,
    get shims () { return shims; },
    ViewList: ViewList,
    JUPYTER_WIDGETS_VERSION: JUPYTER_WIDGETS_VERSION,
    PROTOCOL_VERSION: PROTOCOL_VERSION,
    difference: difference,
    isEqual: isEqual$1,
    assign: assign,
    uuid: uuid,
    WrappedError: WrappedError,
    resolvePromisesDict: resolvePromisesDict,
    reject: reject,
    put_buffers: put_buffers,
    remove_buffers: remove_buffers,
    bufferToHex: bufferToHex,
    hexToBuffer: hexToBuffer,
    bufferToBase64: bufferToBase64,
    base64ToBuffer: base64ToBuffer,
    IJupyterWidgetRegistry: IJupyterWidgetRegistry
  });

  // Copyright (c) Jupyter Development Team.
  const hexTable$1 = [
      '00',
      '01',
      '02',
      '03',
      '04',
      '05',
      '06',
      '07',
      '08',
      '09',
      '0A',
      '0B',
      '0C',
      '0D',
      '0E',
      '0F',
      '10',
      '11',
      '12',
      '13',
      '14',
      '15',
      '16',
      '17',
      '18',
      '19',
      '1A',
      '1B',
      '1C',
      '1D',
      '1E',
      '1F',
      '20',
      '21',
      '22',
      '23',
      '24',
      '25',
      '26',
      '27',
      '28',
      '29',
      '2A',
      '2B',
      '2C',
      '2D',
      '2E',
      '2F',
      '30',
      '31',
      '32',
      '33',
      '34',
      '35',
      '36',
      '37',
      '38',
      '39',
      '3A',
      '3B',
      '3C',
      '3D',
      '3E',
      '3F',
      '40',
      '41',
      '42',
      '43',
      '44',
      '45',
      '46',
      '47',
      '48',
      '49',
      '4A',
      '4B',
      '4C',
      '4D',
      '4E',
      '4F',
      '50',
      '51',
      '52',
      '53',
      '54',
      '55',
      '56',
      '57',
      '58',
      '59',
      '5A',
      '5B',
      '5C',
      '5D',
      '5E',
      '5F',
      '60',
      '61',
      '62',
      '63',
      '64',
      '65',
      '66',
      '67',
      '68',
      '69',
      '6A',
      '6B',
      '6C',
      '6D',
      '6E',
      '6F',
      '70',
      '71',
      '72',
      '73',
      '74',
      '75',
      '76',
      '77',
      '78',
      '79',
      '7A',
      '7B',
      '7C',
      '7D',
      '7E',
      '7F',
      '80',
      '81',
      '82',
      '83',
      '84',
      '85',
      '86',
      '87',
      '88',
      '89',
      '8A',
      '8B',
      '8C',
      '8D',
      '8E',
      '8F',
      '90',
      '91',
      '92',
      '93',
      '94',
      '95',
      '96',
      '97',
      '98',
      '99',
      '9A',
      '9B',
      '9C',
      '9D',
      '9E',
      '9F',
      'A0',
      'A1',
      'A2',
      'A3',
      'A4',
      'A5',
      'A6',
      'A7',
      'A8',
      'A9',
      'AA',
      'AB',
      'AC',
      'AD',
      'AE',
      'AF',
      'B0',
      'B1',
      'B2',
      'B3',
      'B4',
      'B5',
      'B6',
      'B7',
      'B8',
      'B9',
      'BA',
      'BB',
      'BC',
      'BD',
      'BE',
      'BF',
      'C0',
      'C1',
      'C2',
      'C3',
      'C4',
      'C5',
      'C6',
      'C7',
      'C8',
      'C9',
      'CA',
      'CB',
      'CC',
      'CD',
      'CE',
      'CF',
      'D0',
      'D1',
      'D2',
      'D3',
      'D4',
      'D5',
      'D6',
      'D7',
      'D8',
      'D9',
      'DA',
      'DB',
      'DC',
      'DD',
      'DE',
      'DF',
      'E0',
      'E1',
      'E2',
      'E3',
      'E4',
      'E5',
      'E6',
      'E7',
      'E8',
      'E9',
      'EA',
      'EB',
      'EC',
      'ED',
      'EE',
      'EF',
      'F0',
      'F1',
      'F2',
      'F3',
      'F4',
      'F5',
      'F6',
      'F7',
      'F8',
      'F9',
      'FA',
      'FB',
      'FC',
      'FD',
      'FE',
      'FF'
  ];
  /**
   * Convert an ArrayBuffer to a hex string.
   */
  function bufferToHex$1(buffer) {
      const x = new Uint8Array(buffer);
      const s = [];
      for (let i = 0; i < x.length; i++) {
          s.push(hexTable$1[x[i]]);
      }
      return s.join('');
  }
  /**
   * Convert a hex string to an ArrayBuffer.
   */
  function hexToBuffer$1(hex) {
      const x = new Uint8Array(hex.length / 2);
      for (let i = 0; i < hex.length; i += 2) {
          x[i / 2] = parseInt(hex.slice(i, i + 2), 16);
      }
      return x.buffer;
  }
  /**
   * Convert an ArrayBuffer to a base64 string.
   */
  function bufferToBase64$1(buffer) {
      return fromByteArray_1(new Uint8Array(buffer));
  }
  /**
   * Convert a base64 string to an ArrayBuffer.
   */
  function base64ToBuffer$1(base64) {
      return toByteArray_1(base64).buffer;
  }

  // Copyright (c) Jupyter Development Team.
  const PROTOCOL_MAJOR_VERSION$1 = PROTOCOL_VERSION.split('.', 1)[0];
  /**
   * Manager abstract base class
   */
  class ManagerBase$1 {
      constructor() {
          /**
           * The comm target name to register
           */
          this.comm_target_name = 'jupyter.widget';
          /**
           * Dictionary of model ids and model instance promises
           */
          this._models = Object.create(null);
      }
      /**
       * Modifies view options. Generally overloaded in custom widget manager
       * implementations.
       */
      setViewOptions(options = {}) {
          return options;
      }
      create_view(model, options = {}) {
          const id = uuid();
          const viewPromise = (model.state_change = model.state_change.then(async () => {
              try {
                  const ViewType = (await this.loadClass(model.get('_view_name'), model.get('_view_module'), model.get('_view_module_version')));
                  const view = new ViewType({
                      model: model,
                      options: this.setViewOptions(options)
                  });
                  view.listenTo(model, 'destroy', view.remove);
                  await view.render();
                  // This presumes the view is added to the list of model views below
                  view.once('remove', () => {
                      delete model.views[id];
                  });
                  return view;
              }
              catch (e) {
                  console.error(`Could not create a view for model id ${model.model_id}`);
                  throw e;
              }
          }));
          model.views[id] = viewPromise;
          return viewPromise;
      }
      /**
       * callback handlers specific to a view
       */
      callbacks(view) {
          return {};
      }
      /**
       * Get a promise for a model by model id.
       *
       * #### Notes
       * If a model is not found, undefined is returned (NOT a promise). However,
       * the calling code should also deal with the case where a rejected promise
       * is returned, and should treat that also as a model not found.
       */
      get_model(model_id) {
          // TODO: Perhaps we should return a Promise.reject if the model is not
          // found. Right now this isn't a true async function because it doesn't
          // always return a promise.
          return this._models[model_id];
      }
      /**
       * Handle when a comm is opened.
       */
      handle_comm_open(comm, msg) {
          const protocolVersion = (msg.metadata || {})['version'] || '';
          if (protocolVersion.split('.', 1)[0] !== PROTOCOL_MAJOR_VERSION$1) {
              const error = `Wrong widget protocol version: received protocol version '${protocolVersion}', but was expecting major version '${PROTOCOL_MAJOR_VERSION$1}'`;
              console.error(error);
              return Promise.reject(error);
          }
          const data = msg.content.data;
          const buffer_paths = data.buffer_paths || [];
          // Make sure the buffers are DataViews
          const buffers = (msg.buffers || []).map(b => {
              if (b instanceof DataView) {
                  return b;
              }
              else {
                  return new DataView(b instanceof ArrayBuffer ? b : b.buffer);
              }
          });
          put_buffers(data.state, buffer_paths, buffers);
          return this.new_model({
              model_name: data.state['_model_name'],
              model_module: data.state['_model_module'],
              model_module_version: data.state['_model_module_version'],
              comm: comm
          }, data.state).catch(reject('Could not create a model.', true));
      }
      /**
       * Create a comm and new widget model.
       * @param  options - same options as new_model but comm is not
       *                          required and additional options are available.
       * @param  serialized_state - serialized model attributes.
       */
      new_widget(options, serialized_state = {}) {
          let commPromise;
          // we check to make sure the view information is provided, to help catch
          // backwards incompatibility errors.
          if (options.view_name === undefined ||
              options.view_module === undefined ||
              options.view_module_version === undefined) {
              return Promise.reject('new_widget(...) must be given view information in the options.');
          }
          // If no comm is provided, a new comm is opened for the jupyter.widget
          // target.
          if (options.comm) {
              commPromise = Promise.resolve(options.comm);
          }
          else {
              commPromise = this._create_comm(this.comm_target_name, options.model_id, {
                  state: {
                      _model_module: options.model_module,
                      _model_module_version: options.model_module_version,
                      _model_name: options.model_name,
                      _view_module: options.view_module,
                      _view_module_version: options.view_module_version,
                      _view_name: options.view_name
                  }
              }, { version: PROTOCOL_VERSION });
          }
          // The options dictionary is copied since data will be added to it.
          const options_clone = Object.assign({}, options);
          // Create the model. In the case where the comm promise is rejected a
          // comm-less model is still created with the required model id.
          return commPromise.then(comm => {
              // Comm Promise Resolved.
              options_clone.comm = comm;
              const widget_model = this.new_model(options_clone, serialized_state);
              return widget_model.then(model => {
                  model.sync('create', model);
                  return model;
              });
          }, () => {
              // Comm Promise Rejected.
              if (!options_clone.model_id) {
                  options_clone.model_id = uuid();
              }
              return this.new_model(options_clone, serialized_state);
          });
      }
      register_model(model_id, modelPromise) {
          this._models[model_id] = modelPromise;
          modelPromise.then(model => {
              model.once('comm:close', () => {
                  delete this._models[model_id];
              });
          });
      }
      /**
       * Create and return a promise for a new widget model
       *
       * @param options - the options for creating the model.
       * @param serialized_state - attribute values for the model.
       *
       * @example
       * widget_manager.new_model({
       *      model_name: 'IntSlider',
       *      model_module: '@jupyter-widgets/controls',
       *      model_module_version: '1.0.0',
       *      model_id: 'u-u-i-d'
       * }).then((model) => { console.log('Create success!', model); },
       *  (err) => {console.error(err)});
       *
       */
      async new_model(options, serialized_state = {}) {
          let model_id;
          if (options.model_id) {
              model_id = options.model_id;
          }
          else if (options.comm) {
              model_id = options.model_id = options.comm.comm_id;
          }
          else {
              throw new Error('Neither comm nor model_id provided in options object. At least one must exist.');
          }
          const modelPromise = this._make_model(options, serialized_state);
          // this call needs to happen before the first `await`, see note in `set_state`:
          this.register_model(model_id, modelPromise);
          return await modelPromise;
      }
      async _make_model(options, serialized_state = {}) {
          const model_id = options.model_id;
          const model_promise = this.loadClass(options.model_name, options.model_module, options.model_module_version);
          let ModelType;
          try {
              ModelType = await model_promise;
          }
          catch (error) {
              console.error('Could not instantiate widget');
              throw error;
          }
          if (!ModelType) {
              throw new Error(`Cannot find model module ${options.model_module}@${options.model_module_version}, ${options.model_name}`);
          }
          const attributes = await ModelType._deserialize_state(serialized_state, this);
          const modelOptions = {
              widget_manager: this,
              model_id: model_id,
              comm: options.comm
          };
          const widget_model = new ModelType(attributes, modelOptions);
          widget_model.name = options.model_name;
          widget_model.module = options.model_module;
          return widget_model;
      }
      /**
       * Close all widgets and empty the widget state.
       * @return Promise that resolves when the widget state is cleared.
       */
      clear_state() {
          return resolvePromisesDict(this._models).then(models => {
              Object.keys(models).forEach(id => models[id].close());
              this._models = Object.create(null);
          });
      }
      /**
       * Asynchronously get the state of the widget manager.
       *
       * This includes all of the widget models, and follows the format given in
       * the @jupyter-widgets/schema package.
       *
       * @param options - The options for what state to return.
       * @returns Promise for a state dictionary
       */
      get_state(options = {}) {
          const modelPromises = Object.keys(this._models).map(id => this._models[id]);
          return Promise.all(modelPromises).then(models => {
              return serialize_state$1(models, options);
          });
      }
      /**
       * Set the widget manager state.
       *
       * @param state - a Javascript object conforming to the application/vnd.jupyter.widget-state+json spec.
       *
       * Reconstructs all of the widget models in the state, merges that with the
       * current manager state, and then attempts to redisplay the widgets in the
       * state.
       */
      set_state(state) {
          // Check to make sure that it's the same version we are parsing.
          if (!(state.version_major && state.version_major <= 2)) {
              throw 'Unsupported widget state format';
          }
          const models = state.state;
          // Recreate all the widget models for the given widget manager state.
          const all_models = this._get_comm_info().then(live_comms => {
              /* Note: It is currently safe to just loop over the models in any order,
                       given that the following holds (does at the time of writing):
                       1: any call to `new_model` with state registers the model promise (e.g. with `register_model`)
                          synchronously (before it's first `await` statement).
                       2: any calls to a model constructor or the `set_state` method on a model,
                          happens asynchronously (in a `then` clause, or after an `await` statement).
        
                      Without these assumptions, one risks trying to set model state with a reference
                      to another model that doesn't exist yet!
                    */
              return Promise.all(Object.keys(models).map(model_id => {
                  // First put back the binary buffers
                  const decode = {
                      base64: base64ToBuffer$1,
                      hex: hexToBuffer$1
                  };
                  const model = models[model_id];
                  const modelState = model.state;
                  if (model.buffers) {
                      const bufferPaths = model.buffers.map((b) => b.path);
                      // put_buffers expects buffers to be DataViews
                      const buffers = model.buffers.map((b) => new DataView(decode[b.encoding](b.data)));
                      put_buffers(model.state, bufferPaths, buffers);
                  }
                  // If the model has already been created, set its state and then
                  // return it.
                  if (this._models[model_id]) {
                      return this._models[model_id].then(model => {
                          // deserialize state
                          return model.constructor
                              ._deserialize_state(modelState || {}, this)
                              .then(attributes => {
                              model.set_state(attributes); // case 2
                              return model;
                          });
                      });
                  }
                  const modelCreate = {
                      model_id: model_id,
                      model_name: model.model_name,
                      model_module: model.model_module,
                      model_module_version: model.model_module_version
                  };
                  if (Object.prototype.hasOwnProperty.call(live_comms, 'model_id')) {
                      // live comm
                      // This connects to an existing comm if it exists, and
                      // should *not* send a comm open message.
                      return this._create_comm(this.comm_target_name, model_id).then(comm => {
                          modelCreate.comm = comm;
                          return this.new_model(modelCreate); // No state, so safe wrt. case 1
                      });
                  }
                  else {
                      return this.new_model(modelCreate, modelState); // case 1
                  }
              }));
          });
          return all_models;
      }
      /**
       * Disconnect the widget manager from the kernel, setting each model's comm
       * as dead.
       */
      disconnect() {
          Object.keys(this._models).forEach(i => {
              this._models[i].then(model => {
                  model.comm_live = false;
              });
          });
      }
      /**
       * Resolve a URL relative to the current notebook location.
       *
       * The default implementation just returns the original url.
       */
      resolveUrl(url) {
          return Promise.resolve(url);
      }
      /**
       * Filter serialized widget state to remove any ID's already present in manager.
       *
       * @param {*} state Serialized state to filter
       *
       * @returns {*} A copy of the state, with its 'state' attribute filtered
       */
      filterExistingModelState(serialized_state) {
          let models = serialized_state.state;
          models = Object.keys(models)
              .filter(model_id => {
              return !this._models[model_id];
          })
              .reduce((res, model_id) => {
              res[model_id] = models[model_id];
              return res;
          }, {});
          return Object.assign(Object.assign({}, serialized_state), { state: models });
      }
  }
  /**
   * Serialize an array of widget models
   *
   * #### Notes
   * The return value follows the format given in the
   * @jupyter-widgets/schema package.
   */
  function serialize_state$1(models, options = {}) {
      const state = {};
      models.forEach(model => {
          const model_id = model.model_id;
          const split = remove_buffers(model.serialize(model.get_state(options.drop_defaults)));
          const buffers = split.buffers.map((buffer, index) => {
              return {
                  data: bufferToBase64$1(buffer),
                  path: split.buffer_paths[index],
                  encoding: 'base64'
              };
          });
          state[model_id] = {
              model_name: model.name,
              model_module: model.module,
              model_module_version: model.get('_model_module_version'),
              state: split.state
          };
          // To save space, only include the buffers key if we have buffers
          if (buffers.length > 0) {
              state[model_id].buffers = buffers;
          }
      });
      return { version_major: 2, version_minor: 0, state: state };
  }

  // Copyright (c) Jupyter Development Team.

  var lib$2 = /*#__PURE__*/Object.freeze({
    ManagerBase: ManagerBase$1,
    serialize_state: serialize_state$1,
    bufferToHex: bufferToHex$1,
    hexToBuffer: hexToBuffer$1,
    bufferToBase64: bufferToBase64$1,
    base64ToBuffer: base64ToBuffer$1
  });

  var author = {
  	name: "Project Jupyter"
  };
  var bugs = {
  	url: "https://github.com/jupyter-widgets/ipywidgets/issues"
  };
  var dependencies = {
  	"@jupyter-widgets/base": "^3.0.0",
  	"@lumino/algorithm": "^1.1.0",
  	"@lumino/domutils": "^1.1.0",
  	"@lumino/messaging": "^1.2.1",
  	"@lumino/signaling": "^1.2.0",
  	"@lumino/widgets": "^1.3.0",
  	"d3-format": "^1.3.0",
  	jquery: "^3.1.1",
  	"jquery-ui": "^1.12.1",
  	underscore: "^1.8.3"
  };
  var description = "Jupyter interactive widgets";
  var devDependencies = {
  	"@jupyterlab/services": "^5.0.0",
  	"@types/d3-format": "^1.3.1",
  	"@types/expect.js": "^0.3.29",
  	"@types/mathjax": "^0.0.35",
  	"@types/mocha": "^5.2.7",
  	"@types/node": "^12.0.10",
  	chai: "^4.0.0",
  	"css-loader": "^3.4.0",
  	"expect.js": "^0.3.1",
  	"file-loader": "^5.0.2",
  	"istanbul-instrumenter-loader": "^3.0.1",
  	karma: "^4.1.0",
  	"karma-chrome-launcher": "^2.2.0",
  	"karma-coverage": "^1.1.2",
  	"karma-firefox-launcher": "^1.1.0",
  	"karma-ie-launcher": "^1.0.0",
  	"karma-mocha": "^1.3.0",
  	"karma-mocha-reporter": "^2.2.5",
  	"karma-webpack": "^4.0.2",
  	mocha: "^6.1.4",
  	"npm-run-all": "^4.1.5",
  	"postcss-cli": "^6.1.2",
  	"postcss-cssnext": "^3.1.0",
  	"postcss-import": "^12.0.1",
  	"postcss-loader": "^3.0.0",
  	rimraf: "^2.6.1",
  	sinon: "^7.3.2",
  	"sinon-chai": "^3.3.0",
  	"style-loader": "^1.1.2",
  	tslint: "^5.20.1",
  	typescript: "~3.7.4",
  	"url-loader": "^3.0.0",
  	webpack: "^4.41.5"
  };
  var files = [
  	"lib/**/*.d.ts",
  	"lib/**/*.js",
  	"css/*.css",
  	"dist/"
  ];
  var gitHead = "bed7b2c0458330b1e5a427882ccfc4ea5d3b1e16";
  var homepage = "https://github.com/jupyter-widgets/ipywidgets#readme";
  var license = "BSD-3-Clause";
  var main = "lib/index.js";
  var name = "@jupyter-widgets/controls";
  var repository = {
  	type: "git",
  	url: "git+https://github.com/jupyter-widgets/ipywidgets.git"
  };
  var scripts = {
  	build: "npm run build:src && npm run build:css",
  	"build:css": "postcss --use postcss-import --use postcss-cssnext -o css/widgets.built.css css/widgets.css",
  	"build:src": "tsc",
  	"build:test": "tsc --project test && webpack --config test/webpack.conf.js",
  	clean: "npm run clean:src",
  	"clean:src": "rimraf lib && rimraf tsconfig.tsbuildinfo",
  	lint: "tslint --project tslint.json --format stylish",
  	prepublish: "npm run clean && npm run build",
  	test: "npm run test:unit",
  	"test:coverage": "npm run build:test && webpack --config test/webpack-cov.conf.js && karma start test/karma-cov.conf.js",
  	"test:unit": "npm run test:unit:firefox && npm run test:unit:chrome",
  	"test:unit:chrome": "npm run test:unit:default -- --browsers=Chrome",
  	"test:unit:default": "npm run build:test && karma start test/karma.conf.js --log-level debug",
  	"test:unit:firefox": "npm run test:unit:default -- --browsers=Firefox",
  	"test:unit:ie": "npm run test:unit:default -- --browsers=IE"
  };
  var typings = "lib/index.d.ts";
  var version = "2.0.0";
  var _package = {
  	author: author,
  	bugs: bugs,
  	dependencies: dependencies,
  	description: description,
  	devDependencies: devDependencies,
  	files: files,
  	gitHead: gitHead,
  	homepage: homepage,
  	license: license,
  	main: main,
  	name: name,
  	repository: repository,
  	scripts: scripts,
  	typings: typings,
  	version: version
  };

  var _package$1 = /*#__PURE__*/Object.freeze({
    author: author,
    bugs: bugs,
    dependencies: dependencies,
    description: description,
    devDependencies: devDependencies,
    files: files,
    gitHead: gitHead,
    homepage: homepage,
    license: license,
    main: main,
    name: name,
    repository: repository,
    scripts: scripts,
    typings: typings,
    version: version,
    'default': _package
  });

  var require$$0 = getCjsExportFromNamespace(_package$1);

  // Copyright (c) Jupyter Development Team.
  /**
   * Creates a wrappable Promise rejection function.
   *
   * Creates a function that returns a Promise.reject with a new WrappedError
   * that has the provided message and wraps the original error that
   * caused the promise to reject.
   */
  function reject$2(message, log) {
      return function promiseRejection(error) {
          var wrapped_error = new WrappedError(message, error);
          if (log) {
              console.error(wrapped_error);
          }
          return Promise.reject(wrapped_error);
      };
  }
  /**
   * Apply MathJax rendering to an element, and optionally set its text.
   *
   * If MathJax is not available, make no changes.
   *
   * Parameters
   * ----------
   * element: Node
   * text: optional string
   */
  function typeset(element, text) {
      if (text !== void 0) {
          element.textContent = text;
      }
      if (window.MathJax !== void 0) {
          MathJax.Hub.Queue(['Typeset', MathJax.Hub, element]);
      }
  }
  /**
   * escape text to HTML
   */
  function escape_html(text) {
      var esc = document.createElement('div');
      esc.textContent = text;
      return esc.innerHTML;
  }

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  /**
   * The version of the Jupyter controls widget attribute spec that this package
   * implements.
   */
  var JUPYTER_CONTROLS_VERSION = '1.5.0';

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$9 = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$5 = (undefined && undefined.__assign) || function () {
      __assign$5 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$5.apply(this, arguments);
  };
  var DescriptionStyleModel = /** @class */ (function (_super) {
      __extends$9(DescriptionStyleModel, _super);
      function DescriptionStyleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DescriptionStyleModel.prototype.defaults = function () {
          return __assign$5(__assign$5({}, _super.prototype.defaults.call(this)), { _model_name: 'DescriptionStyleModel', _model_module: '@jupyter-widgets/controls', _model_module_version: JUPYTER_CONTROLS_VERSION });
      };
      DescriptionStyleModel.styleProperties = {
          description_width: {
              selector: '.widget-label',
              attribute: 'width',
              default: null
          },
      };
      return DescriptionStyleModel;
  }(StyleModel));
  var DescriptionModel = /** @class */ (function (_super) {
      __extends$9(DescriptionModel, _super);
      function DescriptionModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DescriptionModel.prototype.defaults = function () {
          return __assign$5(__assign$5({}, _super.prototype.defaults.call(this)), { _model_name: 'DescriptionModel', _view_name: 'DescriptionView', _view_module: '@jupyter-widgets/controls', _model_module: '@jupyter-widgets/controls', _view_module_version: JUPYTER_CONTROLS_VERSION, _model_module_version: JUPYTER_CONTROLS_VERSION, description: '', description_tooltip: null });
      };
      return DescriptionModel;
  }(DOMWidgetModel));
  var DescriptionView = /** @class */ (function (_super) {
      __extends$9(DescriptionView, _super);
      function DescriptionView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DescriptionView.prototype.render = function () {
          this.label = document.createElement('label');
          this.el.appendChild(this.label);
          this.label.className = 'widget-label';
          this.label.style.display = 'none';
          this.listenTo(this.model, 'change:description', this.updateDescription);
          this.listenTo(this.model, 'change:description_tooltip', this.updateDescription);
          this.updateDescription();
      };
      DescriptionView.prototype.typeset = function (element, text) {
          this.displayed.then(function () { return typeset(element, text); });
      };
      DescriptionView.prototype.updateDescription = function () {
          var description = this.model.get('description');
          var description_tooltip = this.model.get('description_tooltip');
          if (description_tooltip === null) {
              description_tooltip = description;
          }
          if (description.length === 0) {
              this.label.style.display = 'none';
          }
          else {
              this.label.innerHTML = description;
              this.typeset(this.label);
              this.label.style.display = '';
          }
          this.label.title = description_tooltip;
      };
      return DescriptionView;
  }(DOMWidgetView));
  /**
   * For backwards compatibility with jupyter-js-widgets 2.x.
   *
   * Use DescriptionModel instead.
   */
  var LabeledDOMWidgetModel = /** @class */ (function (_super) {
      __extends$9(LabeledDOMWidgetModel, _super);
      function LabeledDOMWidgetModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      return LabeledDOMWidgetModel;
  }(DescriptionModel));
  /**
   * For backwards compatibility with jupyter-js-widgets 2.x.
   *
   * Use DescriptionView instead.
   */
  var LabeledDOMWidgetView = /** @class */ (function (_super) {
      __extends$9(LabeledDOMWidgetView, _super);
      function LabeledDOMWidgetView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      return LabeledDOMWidgetView;
  }(DescriptionView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$a = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var CoreWidgetModel = /** @class */ (function (_super) {
      __extends$a(CoreWidgetModel, _super);
      function CoreWidgetModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      CoreWidgetModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'CoreWidgetModel',
              _view_module: '@jupyter-widgets/controls',
              _model_module: '@jupyter-widgets/controls',
              _view_module_version: JUPYTER_CONTROLS_VERSION,
              _model_module_version: JUPYTER_CONTROLS_VERSION,
          });
      };
      return CoreWidgetModel;
  }(WidgetModel));
  var CoreDOMWidgetModel = /** @class */ (function (_super) {
      __extends$a(CoreDOMWidgetModel, _super);
      function CoreDOMWidgetModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      CoreDOMWidgetModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'CoreDOMWidgetModel',
              _view_module: '@jupyter-widgets/controls',
              _model_module: '@jupyter-widgets/controls',
              _view_module_version: JUPYTER_CONTROLS_VERSION,
              _model_module_version: JUPYTER_CONTROLS_VERSION,
          });
      };
      return CoreDOMWidgetModel;
  }(DOMWidgetModel));
  var CoreDescriptionModel = /** @class */ (function (_super) {
      __extends$a(CoreDescriptionModel, _super);
      function CoreDescriptionModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      CoreDescriptionModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'CoreDescriptionModel',
              _view_module: '@jupyter-widgets/controls',
              _model_module: '@jupyter-widgets/controls',
              _view_module_version: JUPYTER_CONTROLS_VERSION,
              _model_module_version: JUPYTER_CONTROLS_VERSION,
          });
      };
      return CoreDescriptionModel;
  }(DescriptionModel));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$b = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$6 = (undefined && undefined.__assign) || function () {
      __assign$6 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$6.apply(this, arguments);
  };
  var DirectionalLinkModel = /** @class */ (function (_super) {
      __extends$b(DirectionalLinkModel, _super);
      function DirectionalLinkModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DirectionalLinkModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              target: undefined,
              source: undefined,
              _model_name: 'DirectionalLinkModel'
          });
      };
      DirectionalLinkModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
          this.on('change', this.updateBindings, this);
          this.updateBindings();
      };
      DirectionalLinkModel.prototype.updateValue = function (sourceModel, sourceAttr, targetModel, targetAttr) {
          if (this._updating) {
              return;
          }
          this._updating = true;
          try {
              if (targetModel) {
                  targetModel.set(targetAttr, sourceModel.get(sourceAttr));
                  targetModel.save_changes();
              }
          }
          finally {
              this._updating = false;
          }
      };
      DirectionalLinkModel.prototype.updateBindings = function () {
          var _a, _b;
          var _this = this;
          this.cleanup();
          _a = this.get('source') || [null, null], this.sourceModel = _a[0], this.sourceAttr = _a[1];
          _b = this.get('target') || [null, null], this.targetModel = _b[0], this.targetAttr = _b[1];
          if (this.sourceModel) {
              this.listenTo(this.sourceModel, 'change:' + this.sourceAttr, function () {
                  _this.updateValue(_this.sourceModel, _this.sourceAttr, _this.targetModel, _this.targetAttr);
              });
              this.updateValue(this.sourceModel, this.sourceAttr, this.targetModel, this.targetAttr);
              this.listenToOnce(this.sourceModel, 'destroy', this.cleanup);
          }
          if (this.targetModel) {
              this.listenToOnce(this.targetModel, 'destroy', this.cleanup);
          }
      };
      DirectionalLinkModel.prototype.cleanup = function () {
          // Stop listening to 'change' and 'destroy' events of the source and target
          if (this.sourceModel) {
              this.stopListening(this.sourceModel, 'change:' + this.sourceAttr, null);
              this.stopListening(this.sourceModel, 'destroy', null);
          }
          if (this.targetModel) {
              this.stopListening(this.targetModel, 'destroy', null);
          }
      };
      DirectionalLinkModel.serializers = __assign$6(__assign$6({}, CoreWidgetModel.serializers), { target: { deserialize: unpack_models }, source: { deserialize: unpack_models } });
      return DirectionalLinkModel;
  }(CoreWidgetModel));
  var LinkModel = /** @class */ (function (_super) {
      __extends$b(LinkModel, _super);
      function LinkModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      LinkModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'LinkModel'
          });
      };
      LinkModel.prototype.updateBindings = function () {
          var _this = this;
          _super.prototype.updateBindings.call(this);
          if (this.targetModel) {
              this.listenTo(this.targetModel, 'change:' + this.targetAttr, function () {
                  _this.updateValue(_this.targetModel, _this.targetAttr, _this.sourceModel, _this.sourceAttr);
              });
          }
      };
      LinkModel.prototype.cleanup = function () {
          _super.prototype.cleanup.call(this);
          if (this.targetModel) {
              this.stopListening(this.targetModel, 'change:' + this.targetAttr, null);
          }
      };
      return LinkModel;
  }(DirectionalLinkModel));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$c = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var BoolModel = /** @class */ (function (_super) {
      __extends$c(BoolModel, _super);
      function BoolModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoolModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              value: false,
              disabled: false,
              _model_name: 'BoolModel'
          });
      };
      return BoolModel;
  }(CoreDescriptionModel));
  var CheckboxModel = /** @class */ (function (_super) {
      __extends$c(CheckboxModel, _super);
      function CheckboxModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      CheckboxModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              indent: true,
              _view_name: 'CheckboxView',
              _model_name: 'CheckboxModel'
          });
      };
      return CheckboxModel;
  }(CoreDescriptionModel));
  var CheckboxView = /** @class */ (function (_super) {
      __extends$c(CheckboxView, _super);
      function CheckboxView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      CheckboxView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-checkbox');
          // adding a zero-width space to the label to help
          // the browser set the baseline correctly
          this.label.innerHTML = '&#8203;';
          // label containing the checkbox and description span
          this.checkboxLabel = document.createElement('label');
          this.checkboxLabel.classList.add('widget-label-basic');
          this.el.appendChild(this.checkboxLabel);
          // checkbox
          this.checkbox = document.createElement('input');
          this.checkbox.setAttribute('type', 'checkbox');
          this.checkboxLabel.appendChild(this.checkbox);
          // span to the right of the checkbox that will render the description
          this.descriptionSpan = document.createElement('span');
          this.checkboxLabel.appendChild(this.descriptionSpan);
          this.listenTo(this.model, 'change:indent', this.updateIndent);
          this.update(); // Set defaults.
          this.updateDescription();
          this.updateIndent();
      };
      /**
       * Overriden from super class
       *
       * Update the description span (rather than the label) since
       * we want the description to the right of the checkbox.
       */
      CheckboxView.prototype.updateDescription = function () {
          // can be called before the view is fully initialized
          if (this.checkboxLabel == null) {
              return;
          }
          var description = this.model.get('description');
          this.descriptionSpan.innerHTML = description;
          this.typeset(this.descriptionSpan);
          this.descriptionSpan.title = description;
          this.checkbox.title = description;
      };
      /**
       * Update the visibility of the label in the super class
       * to provide the optional indent.
       */
      CheckboxView.prototype.updateIndent = function () {
          var indent = this.model.get('indent');
          this.label.style.display = indent ? '' : 'none';
      };
      CheckboxView.prototype.events = function () {
          return {
              'click input[type="checkbox"]': '_handle_click'
          };
      };
      /**
       * Handles when the checkbox is clicked.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      CheckboxView.prototype._handle_click = function () {
          var value = this.model.get('value');
          this.model.set('value', !value, { updated_view: this });
          this.touch();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed. The model may have been
       * changed by another view or by a state update from the back-end.
       */
      CheckboxView.prototype.update = function (options) {
          this.checkbox.checked = this.model.get('value');
          if (options === undefined || options.updated_view != this) {
              this.checkbox.disabled = this.model.get('disabled');
          }
          return _super.prototype.update.call(this);
      };
      return CheckboxView;
  }(DescriptionView));
  var ToggleButtonModel = /** @class */ (function (_super) {
      __extends$c(ToggleButtonModel, _super);
      function ToggleButtonModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ToggleButtonModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'ToggleButtonView',
              _model_name: 'ToggleButtonModel',
              tooltip: '',
              icon: '',
              button_style: ''
          });
      };
      return ToggleButtonModel;
  }(BoolModel));
  var ToggleButtonView = /** @class */ (function (_super) {
      __extends$c(ToggleButtonView, _super);
      function ToggleButtonView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      ToggleButtonView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('jupyter-button');
          this.el.classList.add('widget-toggle-button');
          this.listenTo(this.model, 'change:button_style', this.update_button_style);
          this.set_button_style();
          this.update(); // Set defaults.
      };
      ToggleButtonView.prototype.update_button_style = function () {
          this.update_mapped_classes(ToggleButtonView.class_map, 'button_style');
      };
      ToggleButtonView.prototype.set_button_style = function () {
          this.set_mapped_classes(ToggleButtonView.class_map, 'button_style');
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed. The model may have been
       * changed by another view or by a state update from the back-end.
       */
      ToggleButtonView.prototype.update = function (options) {
          if (this.model.get('value')) {
              this.el.classList.add('mod-active');
          }
          else {
              this.el.classList.remove('mod-active');
          }
          if (options === undefined || options.updated_view !== this) {
              this.el.disabled = this.model.get('disabled');
              this.el.setAttribute('title', this.model.get('tooltip'));
              var description = this.model.get('description');
              var icon = this.model.get('icon');
              if (description.trim().length === 0 && icon.trim().length === 0) {
                  this.el.innerHTML = '&nbsp;'; // Preserve button height
              }
              else {
                  this.el.textContent = '';
                  if (icon.trim().length) {
                      var i = document.createElement('i');
                      this.el.appendChild(i);
                      i.classList.add('fa');
                      i.classList.add('fa-' + icon);
                  }
                  this.el.appendChild(document.createTextNode(description));
              }
          }
          return _super.prototype.update.call(this);
      };
      ToggleButtonView.prototype.events = function () {
          return {
              // Dictionary of events and their handlers.
              'click': '_handle_click'
          };
      };
      /**
       * Handles and validates user input.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      ToggleButtonView.prototype._handle_click = function (event) {
          event.preventDefault();
          var value = this.model.get('value');
          this.model.set('value', !value, { updated_view: this });
          this.touch();
      };
      Object.defineProperty(ToggleButtonView.prototype, "tagName", {
          /**
           * The default tag name.
           *
           * #### Notes
           * This is a read-only attribute.
           */
          get: function () {
              // We can't make this an attribute with a default value
              // since it would be set after it is needed in the
              // constructor.
              return 'button';
          },
          enumerable: true,
          configurable: true
      });
      ToggleButtonView.class_map = {
          primary: ['mod-primary'],
          success: ['mod-success'],
          info: ['mod-info'],
          warning: ['mod-warning'],
          danger: ['mod-danger']
      };
      return ToggleButtonView;
  }(DOMWidgetView));
  var ValidModel = /** @class */ (function (_super) {
      __extends$c(ValidModel, _super);
      function ValidModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ValidModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              readout: 'Invalid',
              _view_name: 'ValidView',
              _model_name: 'ValidModel'
          });
      };
      return ValidModel;
  }(BoolModel));
  var ValidView = /** @class */ (function (_super) {
      __extends$c(ValidView, _super);
      function ValidView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      ValidView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-valid');
          this.el.classList.add('widget-inline-hbox');
          this.icon = document.createElement('i');
          this.icon.classList.add('fa', 'fa-fw');
          this.el.appendChild(this.icon);
          this.readout = document.createElement('span');
          this.readout.classList.add('widget-valid-readout');
          this.readout.classList.add('widget-readout');
          this.el.appendChild(this.readout);
          this.update();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      ValidView.prototype.update = function () {
          this.el.classList.remove('mod-valid');
          this.el.classList.remove('mod-invalid');
          this.icon.classList.remove('fa-check');
          this.icon.classList.remove('fa-times');
          this.readout.textContent = this.model.get('readout');
          if (this.model.get('value')) {
              this.el.classList.add('mod-valid');
              this.icon.classList.add('fa-check');
          }
          else {
              this.el.classList.add('mod-invalid');
              this.icon.classList.add('fa-times');
          }
      };
      return ValidView;
  }(DescriptionView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$d = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var ButtonStyleModel = /** @class */ (function (_super) {
      __extends$d(ButtonStyleModel, _super);
      function ButtonStyleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ButtonStyleModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'ButtonStyleModel',
              _model_module: '@jupyter-widgets/controls',
              _model_module_version: JUPYTER_CONTROLS_VERSION,
          });
      };
      ButtonStyleModel.styleProperties = {
          button_color: {
              selector: '',
              attribute: 'background-color',
              default: null
          },
          font_weight: {
              selector: '',
              attribute: 'font-weight',
              default: ''
          }
      };
      return ButtonStyleModel;
  }(StyleModel));
  var ButtonModel = /** @class */ (function (_super) {
      __extends$d(ButtonModel, _super);
      function ButtonModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ButtonModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              description: '',
              tooltip: '',
              disabled: false,
              icon: '',
              button_style: '',
              _view_name: 'ButtonView',
              _model_name: 'ButtonModel',
              style: null
          });
      };
      return ButtonModel;
  }(CoreDOMWidgetModel));
  var ButtonView = /** @class */ (function (_super) {
      __extends$d(ButtonView, _super);
      function ButtonView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      ButtonView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('jupyter-button');
          this.el.classList.add('widget-button');
          this.listenTo(this.model, 'change:button_style', this.update_button_style);
          this.set_button_style();
          this.update(); // Set defaults.
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed. The model may have been
       * changed by another view or by a state update from the back-end.
       */
      ButtonView.prototype.update = function () {
          this.el.disabled = this.model.get('disabled');
          this.el.setAttribute('title', this.model.get('tooltip'));
          var description = this.model.get('description');
          var icon = this.model.get('icon');
          if (description.length || icon.length) {
              this.el.textContent = '';
              if (icon.length) {
                  var i = document.createElement('i');
                  i.classList.add('fa');
                  i.classList.add('fa-' + icon);
                  if (description.length === 0) {
                      i.classList.add('center');
                  }
                  this.el.appendChild(i);
              }
              this.el.appendChild(document.createTextNode(description));
          }
          return _super.prototype.update.call(this);
      };
      ButtonView.prototype.update_button_style = function () {
          this.update_mapped_classes(ButtonView.class_map, 'button_style');
      };
      ButtonView.prototype.set_button_style = function () {
          this.set_mapped_classes(ButtonView.class_map, 'button_style');
      };
      /**
       * Dictionary of events and handlers
       */
      ButtonView.prototype.events = function () {
          // TODO: return typing not needed in Typescript later than 1.8.x
          // See http://stackoverflow.com/questions/22077023/why-cant-i-indirectly-return-an-object-literal-to-satisfy-an-index-signature-re and https://github.com/Microsoft/TypeScript/pull/7029
          return { 'click': '_handle_click' };
      };
      /**
       * Handles when the button is clicked.
       */
      ButtonView.prototype._handle_click = function (event) {
          event.preventDefault();
          this.send({ event: 'click' });
      };
      Object.defineProperty(ButtonView.prototype, "tagName", {
          /**
           * The default tag name.
           *
           * #### Notes
           * This is a read-only attribute.
           */
          get: function () {
              // We can't make this an attribute with a default value
              // since it would be set after it is needed in the
              // constructor.
              return 'button';
          },
          enumerable: true,
          configurable: true
      });
      ButtonView.class_map = {
          primary: ['mod-primary'],
          success: ['mod-success'],
          info: ['mod-info'],
          warning: ['mod-warning'],
          danger: ['mod-danger']
      };
      return ButtonView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$e = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$7 = (undefined && undefined.__assign) || function () {
      __assign$7 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$7.apply(this, arguments);
  };
  var BoxModel = /** @class */ (function (_super) {
      __extends$e(BoxModel, _super);
      function BoxModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoxModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'BoxView',
              _model_name: 'BoxModel',
              children: [],
              box_style: ''
          });
      };
      BoxModel.serializers = __assign$7(__assign$7({}, CoreDOMWidgetModel.serializers), { children: { deserialize: unpack_models } });
      return BoxModel;
  }(CoreDOMWidgetModel));
  var HBoxModel = /** @class */ (function (_super) {
      __extends$e(HBoxModel, _super);
      function HBoxModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      HBoxModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'HBoxView',
              _model_name: 'HBoxModel',
          });
      };
      return HBoxModel;
  }(BoxModel));
  var VBoxModel = /** @class */ (function (_super) {
      __extends$e(VBoxModel, _super);
      function VBoxModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      VBoxModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'VBoxView',
              _model_name: 'VBoxModel',
          });
      };
      return VBoxModel;
  }(BoxModel));
  var BoxView = /** @class */ (function (_super) {
      __extends$e(BoxView, _super);
      function BoxView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoxView.prototype._createElement = function (tagName) {
          this.pWidget = new JupyterPhosphorPanelWidget({ view: this });
          return this.pWidget.node;
      };
      BoxView.prototype._setElement = function (el) {
          if (this.el || el !== this.pWidget.node) {
              // Boxes don't allow setting the element beyond the initial creation.
              throw new Error('Cannot reset the DOM element.');
          }
          this.el = this.pWidget.node;
          this.$el = $$1(this.pWidget.node);
      };
      BoxView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.children_views = new ViewList(this.add_child_model, null, this);
          this.listenTo(this.model, 'change:children', this.update_children);
          this.listenTo(this.model, 'change:box_style', this.update_box_style);
          this.pWidget.addClass('jupyter-widgets');
          this.pWidget.addClass('widget-container');
          this.pWidget.addClass('widget-box');
      };
      BoxView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.update_children();
          this.set_box_style();
      };
      BoxView.prototype.update_children = function () {
          this.children_views.update(this.model.get('children')).then(function (views) {
              // Notify all children that their sizes may have changed.
              views.forEach(function (view) {
                  MessageLoop.postMessage(view.pWidget, Widget.ResizeMessage.UnknownSize);
              });
          });
      };
      BoxView.prototype.update_box_style = function () {
          this.update_mapped_classes(BoxView.class_map, 'box_style');
      };
      BoxView.prototype.set_box_style = function () {
          this.set_mapped_classes(BoxView.class_map, 'box_style');
      };
      BoxView.prototype.add_child_model = function (model) {
          var _this = this;
          // we insert a dummy element so the order is preserved when we add
          // the rendered content later.
          var dummy = new Widget();
          this.pWidget.addWidget(dummy);
          return this.create_child_view(model).then(function (view) {
              // replace the dummy widget with the new one.
              var i = ArrayExt.firstIndexOf(_this.pWidget.widgets, dummy);
              _this.pWidget.insertWidget(i, view.pWidget);
              dummy.dispose();
              return view;
          }).catch(reject$2('Could not add child view to box', true));
      };
      BoxView.prototype.remove = function () {
          this.children_views = null;
          _super.prototype.remove.call(this);
      };
      BoxView.class_map = {
          success: ['alert', 'alert-success'],
          info: ['alert', 'alert-info'],
          warning: ['alert', 'alert-warning'],
          danger: ['alert', 'alert-danger']
      };
      return BoxView;
  }(DOMWidgetView));
  var HBoxView = /** @class */ (function (_super) {
      __extends$e(HBoxView, _super);
      function HBoxView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor
       */
      HBoxView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.pWidget.addClass('widget-hbox');
      };
      return HBoxView;
  }(BoxView));
  var VBoxView = /** @class */ (function (_super) {
      __extends$e(VBoxView, _super);
      function VBoxView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor
       */
      VBoxView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.pWidget.addClass('widget-vbox');
      };
      return VBoxView;
  }(BoxView));
  var GridBoxView = /** @class */ (function (_super) {
      __extends$e(GridBoxView, _super);
      function GridBoxView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor
       */
      GridBoxView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.pWidget.addClass('widget-gridbox');
          // display needn't be set to flex and grid 
          this.pWidget.removeClass('widget-box');
      };
      return GridBoxView;
  }(BoxView));
  var GridBoxModel = /** @class */ (function (_super) {
      __extends$e(GridBoxModel, _super);
      function GridBoxModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      GridBoxModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'GridBoxView',
              _model_name: 'GridBoxModel',
          });
      };
      return GridBoxModel;
  }(BoxModel));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$f = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$8 = (undefined && undefined.__assign) || function () {
      __assign$8 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$8.apply(this, arguments);
  };
  var ImageModel = /** @class */ (function (_super) {
      __extends$f(ImageModel, _super);
      function ImageModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ImageModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'ImageModel',
              _view_name: 'ImageView',
              format: 'png',
              width: '',
              height: '',
              value: new DataView(new ArrayBuffer(0))
          });
      };
      ImageModel.serializers = __assign$8(__assign$8({}, CoreDOMWidgetModel.serializers), { value: { serialize: function (value) {
                  return new DataView(value.buffer.slice(0));
              } } });
      return ImageModel;
  }(CoreDOMWidgetModel));
  var ImageView = /** @class */ (function (_super) {
      __extends$f(ImageView, _super);
      function ImageView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ImageView.prototype.render = function () {
          /**
           * Called when view is rendered.
           */
          _super.prototype.render.call(this);
          this.pWidget.addClass('jupyter-widgets');
          this.pWidget.addClass('widget-image');
          this.update(); // Set defaults.
      };
      ImageView.prototype.update = function () {
          /**
           * Update the contents of this view
           *
           * Called when the model is changed.  The model may have been
           * changed by another view or by a state update from the back-end.
           */
          var url;
          var format = this.model.get('format');
          var value = this.model.get('value');
          if (format !== 'url') {
              var blob = new Blob([value], { type: "image/" + this.model.get('format') });
              url = URL.createObjectURL(blob);
          }
          else {
              url = (new TextDecoder('utf-8')).decode(value.buffer);
          }
          // Clean up the old objectURL
          var oldurl = this.el.src;
          this.el.src = url;
          if (oldurl && typeof oldurl !== 'string') {
              URL.revokeObjectURL(oldurl);
          }
          var width = this.model.get('width');
          if (width !== undefined && width.length > 0) {
              this.el.setAttribute('width', width);
          }
          else {
              this.el.removeAttribute('width');
          }
          var height = this.model.get('height');
          if (height !== undefined && height.length > 0) {
              this.el.setAttribute('height', height);
          }
          else {
              this.el.removeAttribute('height');
          }
          return _super.prototype.update.call(this);
      };
      ImageView.prototype.remove = function () {
          if (this.el.src) {
              URL.revokeObjectURL(this.el.src);
          }
          _super.prototype.remove.call(this);
      };
      Object.defineProperty(ImageView.prototype, "tagName", {
          /**
           * The default tag name.
           *
           * #### Notes
           * This is a read-only attribute.
           */
          get: function () {
              // We can't make this an attribute with a default value
              // since it would be set after it is needed in the
              // constructor.
              return 'img';
          },
          enumerable: true,
          configurable: true
      });
      return ImageView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$g = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$9 = (undefined && undefined.__assign) || function () {
      __assign$9 = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$9.apply(this, arguments);
  };
  var VideoModel = /** @class */ (function (_super) {
      __extends$g(VideoModel, _super);
      function VideoModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      VideoModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'VideoModel',
              _view_name: 'VideoView',
              format: 'mp4',
              width: '',
              height: '',
              autoplay: true,
              loop: true,
              controls: true,
              value: new DataView(new ArrayBuffer(0))
          });
      };
      VideoModel.serializers = __assign$9(__assign$9({}, CoreDOMWidgetModel.serializers), { value: { serialize: function (value) {
                  return new DataView(value.buffer.slice(0));
              } } });
      return VideoModel;
  }(CoreDOMWidgetModel));
  var VideoView = /** @class */ (function (_super) {
      __extends$g(VideoView, _super);
      function VideoView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      VideoView.prototype.render = function () {
          /**
           * Called when view is rendered.
           */
          _super.prototype.render.call(this);
          this.pWidget.addClass('jupyter-widgets');
          this.pWidget.addClass('widget-image');
          this.update(); // Set defaults.
      };
      VideoView.prototype.update = function () {
          /**
           * Update the contents of this view
           *
           * Called when the model is changed.  The model may have been
           * changed by another view or by a state update from the back-end.
           */
          var url;
          var format = this.model.get('format');
          var value = this.model.get('value');
          if (format !== 'url') {
              var blob = new Blob([value], { type: "video/" + this.model.get('format') });
              url = URL.createObjectURL(blob);
          }
          else {
              url = (new TextDecoder('utf-8')).decode(value.buffer);
          }
          // Clean up the old objectURL
          var oldurl = this.el.src;
          this.el.src = url;
          if (oldurl && typeof oldurl !== 'string') {
              URL.revokeObjectURL(oldurl);
          }
          // Height and width
          var width = this.model.get('width');
          if (width !== undefined && width.length > 0) {
              this.el.setAttribute('width', width);
          }
          else {
              this.el.removeAttribute('width');
          }
          var height = this.model.get('height');
          if (height !== undefined && height.length > 0) {
              this.el.setAttribute('height', height);
          }
          else {
              this.el.removeAttribute('height');
          }
          // Video attributes
          this.el.loop = this.model.get('loop');
          this.el.autoplay = this.model.get('autoplay');
          this.el.controls = this.model.get('controls');
          return _super.prototype.update.call(this);
      };
      VideoView.prototype.remove = function () {
          if (this.el.src) {
              URL.revokeObjectURL(this.el.src);
          }
          _super.prototype.remove.call(this);
      };
      Object.defineProperty(VideoView.prototype, "tagName", {
          /**
           * The default tag name.
           *
           * #### Notes
           * This is a read-only attribute.
           */
          get: function () {
              // We can't make this an attribute with a default value
              // since it would be set after it is needed in the
              // constructor.
              return 'video';
          },
          enumerable: true,
          configurable: true
      });
      return VideoView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$h = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$a = (undefined && undefined.__assign) || function () {
      __assign$a = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$a.apply(this, arguments);
  };
  var AudioModel = /** @class */ (function (_super) {
      __extends$h(AudioModel, _super);
      function AudioModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      AudioModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'AudioModel',
              _view_name: 'AudioView',
              format: 'mp3',
              autoplay: true,
              loop: true,
              controls: true,
              value: new DataView(new ArrayBuffer(0))
          });
      };
      AudioModel.serializers = __assign$a(__assign$a({}, CoreDOMWidgetModel.serializers), { value: { serialize: function (value) {
                  return new DataView(value.buffer.slice(0));
              } } });
      return AudioModel;
  }(CoreDOMWidgetModel));
  var AudioView = /** @class */ (function (_super) {
      __extends$h(AudioView, _super);
      function AudioView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      AudioView.prototype.render = function () {
          /**
           * Called when view is rendered.
           */
          _super.prototype.render.call(this);
          this.pWidget.addClass('jupyter-widgets');
          this.update(); // Set defaults.
      };
      AudioView.prototype.update = function () {
          /**
           * Update the contents of this view
           *
           * Called when the model is changed.  The model may have been
           * changed by another view or by a state update from the back-end.
           */
          var url;
          var format = this.model.get('format');
          var value = this.model.get('value');
          if (format !== 'url') {
              var blob = new Blob([value], { type: "audio/" + this.model.get('format') });
              url = URL.createObjectURL(blob);
          }
          else {
              url = (new TextDecoder('utf-8')).decode(value.buffer);
          }
          // Clean up the old objectURL
          var oldurl = this.el.src;
          this.el.src = url;
          if (oldurl && typeof oldurl !== 'string') {
              URL.revokeObjectURL(oldurl);
          }
          // Audio attributes
          this.el.loop = this.model.get('loop');
          this.el.autoplay = this.model.get('autoplay');
          this.el.controls = this.model.get('controls');
          return _super.prototype.update.call(this);
      };
      AudioView.prototype.remove = function () {
          if (this.el.src) {
              URL.revokeObjectURL(this.el.src);
          }
          _super.prototype.remove.call(this);
      };
      Object.defineProperty(AudioView.prototype, "tagName", {
          /**
           * The default tag name.
           *
           * #### Notes
           * This is a read-only attribute.
           */
          get: function () {
              // We can't make this an attribute with a default value
              // since it would be set after it is needed in the
              // constructor.
              return 'audio';
          },
          enumerable: true,
          configurable: true
      });
      return AudioView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$i = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var ColorPickerModel = /** @class */ (function (_super) {
      __extends$i(ColorPickerModel, _super);
      function ColorPickerModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ColorPickerModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              value: 'black',
              concise: false,
              _model_name: 'ColorPickerModel',
              _view_name: 'ColorPickerView'
          });
      };
      return ColorPickerModel;
  }(CoreDescriptionModel));
  var ColorPickerView = /** @class */ (function (_super) {
      __extends$i(ColorPickerView, _super);
      function ColorPickerView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ColorPickerView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-colorpicker');
          this._color_container = document.createElement('div');
          this._color_container.className = 'widget-inline-hbox widget-colorpicker-input';
          this.el.appendChild(this._color_container);
          this._textbox = document.createElement('input');
          this._textbox.setAttribute('type', 'text');
          this._textbox.id = this.label.htmlFor = uuid();
          this._color_container.appendChild(this._textbox);
          this._textbox.value = this.model.get('value');
          this._colorpicker = document.createElement('input');
          this._colorpicker.setAttribute('type', 'color');
          this._color_container.appendChild(this._colorpicker);
          this.listenTo(this.model, 'change:value', this._update_value);
          this.listenTo(this.model, 'change:concise', this._update_concise);
          this._update_concise();
          this._update_value();
          this.update();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed. The model may have been
       * changed by another view or by a state update from the back-end.
       */
      ColorPickerView.prototype.update = function (options) {
          if (options === undefined || options.updated_view != this) {
              var disabled = this.model.get('disabled');
              this._textbox.disabled = disabled;
              this._colorpicker.disabled = disabled;
          }
          return _super.prototype.update.call(this);
      };
      ColorPickerView.prototype.events = function () {
          // Typescript doesn't understand that these functions are called, so we
          // specifically use them here so it knows they are being used.
          void this._picker_change;
          void this._text_change;
          return {
              'change [type="color"]': '_picker_change',
              'change [type="text"]': '_text_change'
          };
      };
      ColorPickerView.prototype._update_value = function () {
          var value = this.model.get('value');
          this._colorpicker.value = color2hex(value);
          this._textbox.value = value;
      };
      ColorPickerView.prototype._update_concise = function () {
          var concise = this.model.get('concise');
          if (concise) {
              this.el.classList.add('concise');
              this._textbox.style.display = 'none';
          }
          else {
              this.el.classList.remove('concise');
              this._textbox.style.display = '';
          }
      };
      ColorPickerView.prototype._picker_change = function () {
          this.model.set('value', this._colorpicker.value);
          this.touch();
      };
      ColorPickerView.prototype._text_change = function () {
          var value = this._validate_color(this._textbox.value, this.model.get('value'));
          this.model.set('value', value);
          this.touch();
      };
      ColorPickerView.prototype._validate_color = function (color, fallback) {
          return color.match(/#[a-fA-F0-9]{3}(?:[a-fA-F0-9]{3})?$/) ||
              named_colors[color.toLowerCase()] ? color : fallback;
      };
      return ColorPickerView;
  }(DescriptionView));
  var named_colors = { aliceblue: '#f0f8ff', antiquewhite: '#faebd7', aqua: '#00ffff', aquamarine: '#7fffd4', azure: '#f0ffff', beige: '#f5f5dc', bisque: '#ffe4c4', black: '#000000', blanchedalmond: '#ffebcd', blue: '#0000ff', blueviolet: '#8a2be2', brown: '#a52a2a', burlywood: '#deb887', cadetblue: '#5f9ea0', chartreuse: '#7fff00', chocolate: '#d2691e', coral: '#ff7f50', cornflowerblue: '#6495ed', cornsilk: '#fff8dc', crimson: '#dc143c', cyan: '#00ffff', darkblue: '#00008b', darkcyan: '#008b8b', darkgoldenrod: '#b8860b', darkgray: '#a9a9a9', darkgrey: '#a9a9a9', darkgreen: '#006400', darkkhaki: '#bdb76b', darkmagenta: '#8b008b', darkolivegreen: '#556b2f', darkorange: '#ff8c00', darkorchid: '#9932cc', darkred: '#8b0000', darksalmon: '#e9967a', darkseagreen: '#8fbc8f', darkslateblue: '#483d8b', darkslategray: '#2f4f4f', darkslategrey: '#2f4f4f', darkturquoise: '#00ced1', darkviolet: '#9400d3', deeppink: '#ff1493', deepskyblue: '#00bfff', dimgray: '#696969', dimgrey: '#696969', dodgerblue: '#1e90ff', firebrick: '#b22222', floralwhite: '#fffaf0', forestgreen: '#228b22', fuchsia: '#ff00ff', gainsboro: '#dcdcdc', ghostwhite: '#f8f8ff', gold: '#ffd700', goldenrod: '#daa520', gray: '#808080', grey: '#808080', green: '#008000', greenyellow: '#adff2f', honeydew: '#f0fff0', hotpink: '#ff69b4', indianred: '#cd5c5c', indigo: '#4b0082', ivory: '#fffff0', khaki: '#f0e68c', lavender: '#e6e6fa', lavenderblush: '#fff0f5', lawngreen: '#7cfc00', lemonchiffon: '#fffacd', lightblue: '#add8e6', lightcoral: '#f08080', lightcyan: '#e0ffff', lightgoldenrodyellow: '#fafad2', lightgreen: '#90ee90', lightgray: '#d3d3d3', lightgrey: '#d3d3d3', lightpink: '#ffb6c1', lightsalmon: '#ffa07a', lightseagreen: '#20b2aa', lightskyblue: '#87cefa', lightslategray: '#778899', lightslategrey: '#778899', lightsteelblue: '#b0c4de', lightyellow: '#ffffe0', lime: '#00ff00', limegreen: '#32cd32', linen: '#faf0e6', magenta: '#ff00ff', maroon: '#800000', mediumaquamarine: '#66cdaa', mediumblue: '#0000cd', mediumorchid: '#ba55d3', mediumpurple: '#9370db', mediumseagreen: '#3cb371', mediumslateblue: '#7b68ee', mediumspringgreen: '#00fa9a', mediumturquoise: '#48d1cc', mediumvioletred: '#c71585', midnightblue: '#191970', mintcream: '#f5fffa', mistyrose: '#ffe4e1', moccasin: '#ffe4b5', navajowhite: '#ffdead', navy: '#000080', oldlace: '#fdf5e6', olive: '#808000', olivedrab: '#6b8e23', orange: '#ffa500', orangered: '#ff4500', orchid: '#da70d6', palegoldenrod: '#eee8aa', palegreen: '#98fb98', paleturquoise: '#afeeee', palevioletred: '#db7093', papayawhip: '#ffefd5', peachpuff: '#ffdab9', peru: '#cd853f', pink: '#ffc0cb', plum: '#dda0dd', powderblue: '#b0e0e6', purple: '#800080', red: '#ff0000', rosybrown: '#bc8f8f', royalblue: '#4169e1', saddlebrown: '#8b4513', salmon: '#fa8072', sandybrown: '#f4a460', seagreen: '#2e8b57', seashell: '#fff5ee', sienna: '#a0522d', silver: '#c0c0c0', skyblue: '#87ceeb', slateblue: '#6a5acd', slategray: '#708090', slategrey: '#708090', snow: '#fffafa', springgreen: '#00ff7f', steelblue: '#4682b4', tan: '#d2b48c', teal: '#008080', thistle: '#d8bfd8', tomato: '#ff6347', turquoise: '#40e0d0', violet: '#ee82ee', wheat: '#f5deb3', white: '#ffffff', whitesmoke: '#f5f5f5', yellow: '#ffff00', yellowgreen: '#9acd32', };
  /*
   * From a valid html color (named color, 6-digits or 3-digits hex format)
   * return a 6-digits hexadecimal color #rrggbb.
   */
  function color2hex(color) {
      return named_colors[color.toLowerCase()] || rgb3_to_rgb6(color);
  }
  function rgb3_to_rgb6(rgb) {
      if (rgb.length === 7) {
          return rgb;
      }
      else {
          return '#' + rgb.charAt(1) + rgb.charAt(1) +
              rgb.charAt(2) + rgb.charAt(2) +
              rgb.charAt(3) + rgb.charAt(3);
      }
  }

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$j = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$b = (undefined && undefined.__assign) || function () {
      __assign$b = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$b.apply(this, arguments);
  };
  function serialize_date(value) {
      if (value === null) {
          return null;
      }
      else {
          return {
              year: value.getUTCFullYear(),
              month: value.getUTCMonth(),
              date: value.getUTCDate()
          };
      }
  }
  function deserialize_date(value) {
      if (value === null) {
          return null;
      }
      else {
          var date = new Date();
          date.setUTCFullYear(value.year, value.month, value.date);
          date.setUTCHours(0, 0, 0, 0);
          return date;
      }
  }
  var DatePickerModel = /** @class */ (function (_super) {
      __extends$j(DatePickerModel, _super);
      function DatePickerModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DatePickerModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              value: null,
              _model_name: 'DatePickerModel',
              _view_name: 'DatePickerView'
          });
      };
      DatePickerModel.serializers = __assign$b(__assign$b({}, CoreDescriptionModel.serializers), { value: {
              serialize: serialize_date,
              deserialize: deserialize_date
          } });
      return DatePickerModel;
  }(CoreDescriptionModel));
  var DatePickerView = /** @class */ (function (_super) {
      __extends$j(DatePickerView, _super);
      function DatePickerView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DatePickerView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-datepicker');
          this._datepicker = document.createElement('input');
          this._datepicker.setAttribute('type', 'date');
          this._datepicker.id = this.label.htmlFor = uuid();
          this.el.appendChild(this._datepicker);
          this.listenTo(this.model, 'change:value', this._update_value);
          this._update_value();
          this.update();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed. The model may have been
       * changed by another view or by a state update from the back-end.
       */
      DatePickerView.prototype.update = function (options) {
          if (options === undefined || options.updated_view !== this) {
              this._datepicker.disabled = this.model.get('disabled');
          }
          return _super.prototype.update.call(this);
      };
      DatePickerView.prototype.events = function () {
          // Typescript doesn't understand that these functions are called, so we
          // specifically use them here so it knows they are being used.
          void this._picker_change;
          void this._picker_focusout;
          return {
              'change [type="date"]': '_picker_change',
              'focusout [type="date"]': '_picker_focusout'
          };
      };
      DatePickerView.prototype._update_value = function () {
          var value = this.model.get('value');
          this._datepicker.valueAsDate = value;
      };
      DatePickerView.prototype._picker_change = function () {
          if (!this._datepicker.validity.badInput) {
              this.model.set('value', this._datepicker.valueAsDate);
              this.touch();
          }
      };
      DatePickerView.prototype._picker_focusout = function () {
          if (this._datepicker.validity.badInput) {
              this.model.set('value', null);
              this.touch();
          }
      };
      return DatePickerView;
  }(DescriptionView));

  // Computes the decimal coefficient and exponent of the specified number x with
  // significant digits p, where x is positive and p is in [1, 21] or undefined.
  // For example, formatDecimal(1.23) returns ["123", 0].
  function formatDecimal(x, p) {
    if ((i = (x = p ? x.toExponential(p - 1) : x.toExponential()).indexOf("e")) < 0) return null; // NaN, ±Infinity
    var i, coefficient = x.slice(0, i);

    // The string returned by toExponential either has the form \d\.\d+e[-+]\d+
    // (e.g., 1.2e+3) or the form \de[-+]\d+ (e.g., 1e+3).
    return [
      coefficient.length > 1 ? coefficient[0] + coefficient.slice(2) : coefficient,
      +x.slice(i + 1)
    ];
  }

  function exponent(x) {
    return x = formatDecimal(Math.abs(x)), x ? x[1] : NaN;
  }

  function formatGroup(grouping, thousands) {
    return function(value, width) {
      var i = value.length,
          t = [],
          j = 0,
          g = grouping[0],
          length = 0;

      while (i > 0 && g > 0) {
        if (length + g + 1 > width) g = Math.max(1, width - length);
        t.push(value.substring(i -= g, i + g));
        if ((length += g + 1) > width) break;
        g = grouping[j = (j + 1) % grouping.length];
      }

      return t.reverse().join(thousands);
    };
  }

  function formatNumerals(numerals) {
    return function(value) {
      return value.replace(/[0-9]/g, function(i) {
        return numerals[+i];
      });
    };
  }

  // [[fill]align][sign][symbol][0][width][,][.precision][~][type]
  var re = /^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;

  function formatSpecifier(specifier) {
    if (!(match = re.exec(specifier))) throw new Error("invalid format: " + specifier);
    var match;
    return new FormatSpecifier({
      fill: match[1],
      align: match[2],
      sign: match[3],
      symbol: match[4],
      zero: match[5],
      width: match[6],
      comma: match[7],
      precision: match[8] && match[8].slice(1),
      trim: match[9],
      type: match[10]
    });
  }

  formatSpecifier.prototype = FormatSpecifier.prototype; // instanceof

  function FormatSpecifier(specifier) {
    this.fill = specifier.fill === undefined ? " " : specifier.fill + "";
    this.align = specifier.align === undefined ? ">" : specifier.align + "";
    this.sign = specifier.sign === undefined ? "-" : specifier.sign + "";
    this.symbol = specifier.symbol === undefined ? "" : specifier.symbol + "";
    this.zero = !!specifier.zero;
    this.width = specifier.width === undefined ? undefined : +specifier.width;
    this.comma = !!specifier.comma;
    this.precision = specifier.precision === undefined ? undefined : +specifier.precision;
    this.trim = !!specifier.trim;
    this.type = specifier.type === undefined ? "" : specifier.type + "";
  }

  FormatSpecifier.prototype.toString = function() {
    return this.fill
        + this.align
        + this.sign
        + this.symbol
        + (this.zero ? "0" : "")
        + (this.width === undefined ? "" : Math.max(1, this.width | 0))
        + (this.comma ? "," : "")
        + (this.precision === undefined ? "" : "." + Math.max(0, this.precision | 0))
        + (this.trim ? "~" : "")
        + this.type;
  };

  // Trims insignificant zeros, e.g., replaces 1.2000k with 1.2k.
  function formatTrim(s) {
    out: for (var n = s.length, i = 1, i0 = -1, i1; i < n; ++i) {
      switch (s[i]) {
        case ".": i0 = i1 = i; break;
        case "0": if (i0 === 0) i0 = i; i1 = i; break;
        default: if (!+s[i]) break out; if (i0 > 0) i0 = 0; break;
      }
    }
    return i0 > 0 ? s.slice(0, i0) + s.slice(i1 + 1) : s;
  }

  var prefixExponent;

  function formatPrefixAuto(x, p) {
    var d = formatDecimal(x, p);
    if (!d) return x + "";
    var coefficient = d[0],
        exponent = d[1],
        i = exponent - (prefixExponent = Math.max(-8, Math.min(8, Math.floor(exponent / 3))) * 3) + 1,
        n = coefficient.length;
    return i === n ? coefficient
        : i > n ? coefficient + new Array(i - n + 1).join("0")
        : i > 0 ? coefficient.slice(0, i) + "." + coefficient.slice(i)
        : "0." + new Array(1 - i).join("0") + formatDecimal(x, Math.max(0, p + i - 1))[0]; // less than 1y!
  }

  function formatRounded(x, p) {
    var d = formatDecimal(x, p);
    if (!d) return x + "";
    var coefficient = d[0],
        exponent = d[1];
    return exponent < 0 ? "0." + new Array(-exponent).join("0") + coefficient
        : coefficient.length > exponent + 1 ? coefficient.slice(0, exponent + 1) + "." + coefficient.slice(exponent + 1)
        : coefficient + new Array(exponent - coefficient.length + 2).join("0");
  }

  var formatTypes = {
    "%": function(x, p) { return (x * 100).toFixed(p); },
    "b": function(x) { return Math.round(x).toString(2); },
    "c": function(x) { return x + ""; },
    "d": function(x) { return Math.round(x).toString(10); },
    "e": function(x, p) { return x.toExponential(p); },
    "f": function(x, p) { return x.toFixed(p); },
    "g": function(x, p) { return x.toPrecision(p); },
    "o": function(x) { return Math.round(x).toString(8); },
    "p": function(x, p) { return formatRounded(x * 100, p); },
    "r": formatRounded,
    "s": formatPrefixAuto,
    "X": function(x) { return Math.round(x).toString(16).toUpperCase(); },
    "x": function(x) { return Math.round(x).toString(16); }
  };

  function identity$1(x) {
    return x;
  }

  var map$2 = Array.prototype.map,
      prefixes = ["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];

  function formatLocale(locale) {
    var group = locale.grouping === undefined || locale.thousands === undefined ? identity$1 : formatGroup(map$2.call(locale.grouping, Number), locale.thousands + ""),
        currencyPrefix = locale.currency === undefined ? "" : locale.currency[0] + "",
        currencySuffix = locale.currency === undefined ? "" : locale.currency[1] + "",
        decimal = locale.decimal === undefined ? "." : locale.decimal + "",
        numerals = locale.numerals === undefined ? identity$1 : formatNumerals(map$2.call(locale.numerals, String)),
        percent = locale.percent === undefined ? "%" : locale.percent + "",
        minus = locale.minus === undefined ? "-" : locale.minus + "",
        nan = locale.nan === undefined ? "NaN" : locale.nan + "";

    function newFormat(specifier) {
      specifier = formatSpecifier(specifier);

      var fill = specifier.fill,
          align = specifier.align,
          sign = specifier.sign,
          symbol = specifier.symbol,
          zero = specifier.zero,
          width = specifier.width,
          comma = specifier.comma,
          precision = specifier.precision,
          trim = specifier.trim,
          type = specifier.type;

      // The "n" type is an alias for ",g".
      if (type === "n") comma = true, type = "g";

      // The "" type, and any invalid type, is an alias for ".12~g".
      else if (!formatTypes[type]) precision === undefined && (precision = 12), trim = true, type = "g";

      // If zero fill is specified, padding goes after sign and before digits.
      if (zero || (fill === "0" && align === "=")) zero = true, fill = "0", align = "=";

      // Compute the prefix and suffix.
      // For SI-prefix, the suffix is lazily computed.
      var prefix = symbol === "$" ? currencyPrefix : symbol === "#" && /[boxX]/.test(type) ? "0" + type.toLowerCase() : "",
          suffix = symbol === "$" ? currencySuffix : /[%p]/.test(type) ? percent : "";

      // What format function should we use?
      // Is this an integer type?
      // Can this type generate exponential notation?
      var formatType = formatTypes[type],
          maybeSuffix = /[defgprs%]/.test(type);

      // Set the default precision if not specified,
      // or clamp the specified precision to the supported range.
      // For significant precision, it must be in [1, 21].
      // For fixed precision, it must be in [0, 20].
      precision = precision === undefined ? 6
          : /[gprs]/.test(type) ? Math.max(1, Math.min(21, precision))
          : Math.max(0, Math.min(20, precision));

      function format(value) {
        var valuePrefix = prefix,
            valueSuffix = suffix,
            i, n, c;

        if (type === "c") {
          valueSuffix = formatType(value) + valueSuffix;
          value = "";
        } else {
          value = +value;

          // Determine the sign. -0 is not less than 0, but 1 / -0 is!
          var valueNegative = value < 0 || 1 / value < 0;

          // Perform the initial formatting.
          value = isNaN(value) ? nan : formatType(Math.abs(value), precision);

          // Trim insignificant zeros.
          if (trim) value = formatTrim(value);

          // If a negative value rounds to zero after formatting, and no explicit positive sign is requested, hide the sign.
          if (valueNegative && +value === 0 && sign !== "+") valueNegative = false;

          // Compute the prefix and suffix.
          valuePrefix = (valueNegative ? (sign === "(" ? sign : minus) : sign === "-" || sign === "(" ? "" : sign) + valuePrefix;
          valueSuffix = (type === "s" ? prefixes[8 + prefixExponent / 3] : "") + valueSuffix + (valueNegative && sign === "(" ? ")" : "");

          // Break the formatted value into the integer “value” part that can be
          // grouped, and fractional or exponential “suffix” part that is not.
          if (maybeSuffix) {
            i = -1, n = value.length;
            while (++i < n) {
              if (c = value.charCodeAt(i), 48 > c || c > 57) {
                valueSuffix = (c === 46 ? decimal + value.slice(i + 1) : value.slice(i)) + valueSuffix;
                value = value.slice(0, i);
                break;
              }
            }
          }
        }

        // If the fill character is not "0", grouping is applied before padding.
        if (comma && !zero) value = group(value, Infinity);

        // Compute the padding.
        var length = valuePrefix.length + value.length + valueSuffix.length,
            padding = length < width ? new Array(width - length + 1).join(fill) : "";

        // If the fill character is "0", grouping is applied after padding.
        if (comma && zero) value = group(padding + value, padding.length ? width - valueSuffix.length : Infinity), padding = "";

        // Reconstruct the final output based on the desired alignment.
        switch (align) {
          case "<": value = valuePrefix + value + valueSuffix + padding; break;
          case "=": value = valuePrefix + padding + value + valueSuffix; break;
          case "^": value = padding.slice(0, length = padding.length >> 1) + valuePrefix + value + valueSuffix + padding.slice(length); break;
          default: value = padding + valuePrefix + value + valueSuffix; break;
        }

        return numerals(value);
      }

      format.toString = function() {
        return specifier + "";
      };

      return format;
    }

    function formatPrefix(specifier, value) {
      var f = newFormat((specifier = formatSpecifier(specifier), specifier.type = "f", specifier)),
          e = Math.max(-8, Math.min(8, Math.floor(exponent(value) / 3))) * 3,
          k = Math.pow(10, -e),
          prefix = prefixes[8 + e / 3];
      return function(value) {
        return f(k * value) + prefix;
      };
    }

    return {
      format: newFormat,
      formatPrefix: formatPrefix
    };
  }

  var locale;
  var format;
  var formatPrefix;

  defaultLocale({
    decimal: ".",
    thousands: ",",
    grouping: [3],
    currency: ["$", ""],
    minus: "-"
  });

  function defaultLocale(definition) {
    locale = formatLocale(definition);
    format = locale.format;
    formatPrefix = locale.formatPrefix;
    return locale;
  }

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$k = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$c = (undefined && undefined.__assign) || function () {
      __assign$c = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$c.apply(this, arguments);
  };

  var IntModel = /** @class */ (function (_super) {
      __extends$k(IntModel, _super);
      function IntModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      IntModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'IntModel',
              value: 0,
          });
      };
      return IntModel;
  }(CoreDescriptionModel));
  var BoundedIntModel = /** @class */ (function (_super) {
      __extends$k(BoundedIntModel, _super);
      function BoundedIntModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoundedIntModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'BoundedIntModel',
              max: 100,
              min: 0
          });
      };
      return BoundedIntModel;
  }(IntModel));
  var SliderStyleModel = /** @class */ (function (_super) {
      __extends$k(SliderStyleModel, _super);
      function SliderStyleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SliderStyleModel.prototype.defaults = function () {
          return __assign$c(__assign$c({}, _super.prototype.defaults.call(this)), { _model_name: 'SliderStyleModel' });
      };
      SliderStyleModel.styleProperties = __assign$c(__assign$c({}, DescriptionStyleModel.styleProperties), { handle_color: {
              selector: '.ui-slider-handle',
              attribute: 'background-color',
              default: null
          } });
      return SliderStyleModel;
  }(DescriptionStyleModel));
  var IntSliderModel = /** @class */ (function (_super) {
      __extends$k(IntSliderModel, _super);
      function IntSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      IntSliderModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'IntSliderModel',
              _view_name: 'IntSliderView',
              step: 1,
              orientation: 'horizontal',
              readout: true,
              readout_format: 'd',
              continuous_update: true,
              style: null,
              disabled: false,
          });
      };
      IntSliderModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
          this.on('change:readout_format', this.update_readout_format, this);
          this.update_readout_format();
      };
      IntSliderModel.prototype.update_readout_format = function () {
          this.readout_formatter = format(this.get('readout_format'));
      };
      return IntSliderModel;
  }(BoundedIntModel));
  var IntRangeSliderModel = /** @class */ (function (_super) {
      __extends$k(IntRangeSliderModel, _super);
      function IntRangeSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      return IntRangeSliderModel;
  }(IntSliderModel));
  var BaseIntSliderView = /** @class */ (function (_super) {
      __extends$k(BaseIntSliderView, _super);
      function BaseIntSliderView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._parse_value = parseInt;
          return _this;
      }
      BaseIntSliderView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-slider');
          this.el.classList.add('widget-hslider');
          (this.$slider = $$1('<div />'))
              .slider({
              slide: this.handleSliderChange.bind(this),
              stop: this.handleSliderChanged.bind(this)
          })
              .addClass('slider');
          // Put the slider in a container
          this.slider_container = document.createElement('div');
          this.slider_container.classList.add('slider-container');
          this.slider_container.appendChild(this.$slider[0]);
          this.el.appendChild(this.slider_container);
          this.readout = document.createElement('div');
          this.el.appendChild(this.readout);
          this.readout.classList.add('widget-readout');
          this.readout.contentEditable = 'true';
          this.readout.style.display = 'none';
          // Set defaults.
          this.update();
      };
      BaseIntSliderView.prototype.update = function (options) {
          /**
           * Update the contents of this view
           *
           * Called when the model is changed.  The model may have been
           * changed by another view or by a state update from the back-end.
           */
          if (options === undefined || options.updated_view !== this) {
              // JQuery slider option keys.  These keys happen to have a
              // one-to-one mapping with the corresponding keys of the model.
              var jquery_slider_keys = ['step', 'disabled'];
              var that_1 = this;
              that_1.$slider.slider({});
              jquery_slider_keys.forEach(function (key) {
                  var model_value = that_1.model.get(key);
                  if (model_value !== undefined) {
                      that_1.$slider.slider('option', key, model_value);
                  }
              });
              if (this.model.get('disabled')) {
                  this.readout.contentEditable = 'false';
              }
              else {
                  this.readout.contentEditable = 'true';
              }
              var max = this.model.get('max');
              var min = this.model.get('min');
              if (min <= max) {
                  if (max !== undefined) {
                      this.$slider.slider('option', 'max', max);
                  }
                  if (min !== undefined) {
                      this.$slider.slider('option', 'min', min);
                  }
              }
              // WORKAROUND FOR JQUERY SLIDER BUG.
              // The horizontal position of the slider handle
              // depends on the value of the slider at the time
              // of orientation change.  Before applying the new
              // workaround, we set the value to the minimum to
              // make sure that the horizontal placement of the
              // handle in the vertical slider is always
              // consistent.
              var orientation_1 = this.model.get('orientation');
              this.$slider.slider('option', 'orientation', orientation_1);
              // Use the right CSS classes for vertical & horizontal sliders
              if (orientation_1 === 'vertical') {
                  this.el.classList.remove('widget-hslider');
                  this.el.classList.add('widget-vslider');
                  this.el.classList.remove('widget-inline-hbox');
                  this.el.classList.add('widget-inline-vbox');
              }
              else {
                  this.el.classList.remove('widget-vslider');
                  this.el.classList.add('widget-hslider');
                  this.el.classList.remove('widget-inline-vbox');
                  this.el.classList.add('widget-inline-hbox');
              }
              var readout = this.model.get('readout');
              if (readout) {
                  this.readout.style.display = '';
                  this.displayed.then(function () {
                      if (that_1.readout_overflow()) {
                          that_1.readout.classList.add('overflow');
                      }
                      else {
                          that_1.readout.classList.remove('overflow');
                      }
                  });
              }
              else {
                  this.readout.style.display = 'none';
              }
          }
          return _super.prototype.update.call(this);
      };
      /**
       * Returns true if the readout box content overflows.
       */
      BaseIntSliderView.prototype.readout_overflow = function () {
          return this.readout.scrollWidth > this.readout.clientWidth;
      };
      BaseIntSliderView.prototype.events = function () {
          return {
              // Dictionary of events and their handlers.
              'slide': 'handleSliderChange',
              'slidestop': 'handleSliderChanged',
              'blur [contentEditable=true]': 'handleTextChange',
              'keydown [contentEditable=true]': 'handleKeyDown'
          };
      };
      BaseIntSliderView.prototype.handleKeyDown = function (e) {
          if (e.keyCode === 13) { /* keyboard keycodes `enter` */
              e.preventDefault();
              this.handleTextChange();
          }
      };
      /**
       * Validate the value of the slider before sending it to the back-end
       * and applying it to the other views on the page.
       */
      BaseIntSliderView.prototype._validate_slide_value = function (x) {
          return Math.floor(x);
      };
      return BaseIntSliderView;
  }(DescriptionView));
  var IntRangeSliderView = /** @class */ (function (_super) {
      __extends$k(IntRangeSliderView, _super);
      function IntRangeSliderView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          // range numbers can be separated by a hyphen, colon, or an en-dash
          _this._range_regex = /^\s*([+-]?\d+)\s*[-:–]\s*([+-]?\d+)/;
          return _this;
      }
      IntRangeSliderView.prototype.update = function (options) {
          _super.prototype.update.call(this, options);
          this.$slider.slider('option', 'range', true);
          // values for the range case are validated python-side in
          // _Bounded{Int,Float}RangeWidget._validate
          var value = this.model.get('value');
          this.$slider.slider('option', 'values', value.slice());
          this.readout.textContent = this.valueToString(value);
          if (this.model.get('value') !== value) {
              this.model.set('value', value, { updated_view: this });
              this.touch();
          }
      };
      /**
       * Write value to a string
       */
      IntRangeSliderView.prototype.valueToString = function (value) {
          var format = this.model.readout_formatter;
          return value.map(function (v) {
              return format(v);
          }).join(' – ');
      };
      /**
       * Parse value from a string
       */
      IntRangeSliderView.prototype.stringToValue = function (text) {
          // ranges can be expressed either 'val-val' or 'val:val' (+spaces)
          var match = this._range_regex.exec(text);
          if (match) {
              return [this._parse_value(match[1]), this._parse_value(match[2])];
          }
          else {
              return null;
          }
      };
      /**
       * this handles the entry of text into the contentEditable label first, the
       * value is checked if it contains a parseable value then it is clamped
       * within the min-max range of the slider finally, the model is updated if
       * the value is to be changed
       *
       * if any of these conditions are not met, the text is reset
       */
      IntRangeSliderView.prototype.handleTextChange = function () {
          var value = this.stringToValue(this.readout.textContent);
          var vmin = this.model.get('min');
          var vmax = this.model.get('max');
          // reject input where NaN or lower > upper
          if (value === null ||
              isNaN(value[0]) ||
              isNaN(value[1]) ||
              (value[0] > value[1])) {
              this.readout.textContent = this.valueToString(this.model.get('value'));
          }
          else {
              // clamp to range
              value = [Math.max(Math.min(value[0], vmax), vmin),
                  Math.max(Math.min(value[1], vmax), vmin)];
              if ((value[0] !== this.model.get('value')[0]) ||
                  (value[1] !== this.model.get('value')[1])) {
                  this.readout.textContent = this.valueToString(value);
                  this.model.set('value', value, { updated_view: this });
                  this.touch();
              }
              else {
                  this.readout.textContent = this.valueToString(this.model.get('value'));
              }
          }
      };
      /**
       * Called when the slider value is changing.
       */
      IntRangeSliderView.prototype.handleSliderChange = function (e, ui) {
          var actual_value = ui.values.map(this._validate_slide_value);
          this.readout.textContent = this.valueToString(actual_value);
          // Only persist the value while sliding if the continuous_update
          // trait is set to true.
          if (this.model.get('continuous_update')) {
              this.handleSliderChanged(e, ui);
          }
      };
      /**
       * Called when the slider value has changed.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      IntRangeSliderView.prototype.handleSliderChanged = function (e, ui) {
          var actual_value = ui.values.map(this._validate_slide_value);
          this.model.set('value', actual_value, { updated_view: this });
          this.touch();
      };
      return IntRangeSliderView;
  }(BaseIntSliderView));
  var IntSliderView = /** @class */ (function (_super) {
      __extends$k(IntSliderView, _super);
      function IntSliderView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      IntSliderView.prototype.update = function (options) {
          _super.prototype.update.call(this, options);
          var min = this.model.get('min');
          var max = this.model.get('max');
          var value = this.model.get('value');
          if (value > max) {
              value = max;
          }
          else if (value < min) {
              value = min;
          }
          this.$slider.slider('option', 'value', value);
          this.readout.textContent = this.valueToString(value);
          if (this.model.get('value') !== value) {
              this.model.set('value', value, { updated_view: this });
              this.touch();
          }
      };
      /**
       * Write value to a string
       */
      IntSliderView.prototype.valueToString = function (value) {
          var format = this.model.readout_formatter;
          return format(value);
      };
      /**
       * Parse value from a string
       */
      IntSliderView.prototype.stringToValue = function (text) {
          return this._parse_value(text);
      };
      /**
       * this handles the entry of text into the contentEditable label first, the
       * value is checked if it contains a parseable value then it is clamped
       * within the min-max range of the slider finally, the model is updated if
       * the value is to be changed
       *
       * if any of these conditions are not met, the text is reset
       */
      IntSliderView.prototype.handleTextChange = function () {
          var value = this.stringToValue(this.readout.textContent);
          var vmin = this.model.get('min');
          var vmax = this.model.get('max');
          if (isNaN(value)) {
              this.readout.textContent = this.valueToString(this.model.get('value'));
          }
          else {
              value = Math.max(Math.min(value, vmax), vmin);
              if (value !== this.model.get('value')) {
                  this.readout.textContent = this.valueToString(value);
                  this.model.set('value', value, { updated_view: this });
                  this.touch();
              }
              else {
                  this.readout.textContent = this.valueToString(this.model.get('value'));
              }
          }
      };
      /**
       * Called when the slider value is changing.
       */
      IntSliderView.prototype.handleSliderChange = function (e, ui) {
          var actual_value = this._validate_slide_value(ui.value);
          this.readout.textContent = this.valueToString(actual_value);
          // Only persist the value while sliding if the continuous_update
          // trait is set to true.
          if (this.model.get('continuous_update')) {
              this.handleSliderChanged(e, ui);
          }
      };
      /**
       * Called when the slider value has changed.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      IntSliderView.prototype.handleSliderChanged = function (e, ui) {
          var actual_value = this._validate_slide_value(ui.value);
          this.model.set('value', actual_value, { updated_view: this });
          this.touch();
      };
      return IntSliderView;
  }(BaseIntSliderView));
  var IntTextModel = /** @class */ (function (_super) {
      __extends$k(IntTextModel, _super);
      function IntTextModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      IntTextModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'IntTextModel',
              _view_name: 'IntTextView',
              disabled: false,
              continuous_update: false,
          });
      };
      return IntTextModel;
  }(IntModel));
  var BoundedIntTextModel = /** @class */ (function (_super) {
      __extends$k(BoundedIntTextModel, _super);
      function BoundedIntTextModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoundedIntTextModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'BoundedIntTextModel',
              _view_name: 'IntTextView',
              disabled: false,
              continuous_update: false,
              step: 1,
          });
      };
      return BoundedIntTextModel;
  }(BoundedIntModel));
  var IntTextView = /** @class */ (function (_super) {
      __extends$k(IntTextView, _super);
      function IntTextView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._parse_value = parseInt;
          _this._default_step = '1';
          return _this;
      }
      IntTextView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-text');
          this.textbox = document.createElement('input');
          this.textbox.type = 'number';
          this.textbox.required = true;
          this.textbox.id = this.label.htmlFor = uuid();
          this.el.appendChild(this.textbox);
          this.update(); // Set defaults.
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      IntTextView.prototype.update = function (options) {
          if (options === undefined || options.updated_view !== this) {
              var value = this.model.get('value');
              if (this._parse_value(this.textbox.value) !== value) {
                  this.textbox.value = value.toString();
              }
              if (this.model.get('min') !== undefined) {
                  this.textbox.min = this.model.get('min');
              }
              if (this.model.get('max') !== undefined) {
                  this.textbox.max = this.model.get('max');
              }
              if (this.model.get('step') !== undefined
                  && this.model.get('step') !== null) {
                  this.textbox.step = this.model.get('step');
              }
              else {
                  this.textbox.step = this._default_step;
              }
              this.textbox.disabled = this.model.get('disabled');
          }
          return _super.prototype.update.call(this);
      };
      IntTextView.prototype.events = function () {
          return {
              'keydown input': 'handleKeyDown',
              'keypress input': 'handleKeypress',
              'keyup input': 'handleKeyUp',
              'input input': 'handleChanging',
              'change input': 'handleChanged'
          };
      };
      /**
       * Handle key down
       *
       * Stop propagation so the event isn't sent to the application.
       */
      IntTextView.prototype.handleKeyDown = function (e) {
          e.stopPropagation();
      };
      /**
       * Handles key press
       */
      IntTextView.prototype.handleKeypress = function (e) {
          if (/[e,.\s]/.test(String.fromCharCode(e.keyCode))) {
              e.preventDefault();
          }
      };
      /**
       * Handle key up
       */
      IntTextView.prototype.handleKeyUp = function (e) {
          if (e.altKey || e.ctrlKey) {
              return;
          }
          var target = e.target;
          /* remove invalid characters */
          var value = target.value;
          value = value.replace(/[e,.\s]/g, "");
          if (value.length >= 1) {
              var subvalue = value.substr(1);
              value = value[0] + subvalue.replace(/[+-]/g, "");
          }
          if (target.value != value) {
              e.preventDefault();
              target.value = value;
          }
      };
      /**
       * Call the submit handler if continuous update is true and we are not
       * obviously incomplete.
       */
      IntTextView.prototype.handleChanging = function (e) {
          var target = e.target;
          var trimmed = target.value.trim();
          if (trimmed === '' || (['-', '-.', '.', '+.', '+'].indexOf(trimmed) >= 0)) {
              // incomplete number
              return;
          }
          if (this.model.get('continuous_update')) {
              this.handleChanged(e);
          }
      };
      /**
       * Applies validated input.
       */
      IntTextView.prototype.handleChanged = function (e) {
          var target = e.target;
          var numericalValue = this._parse_value(target.value);
          // If parse failed, reset value to value stored in model.
          if (isNaN(numericalValue)) {
              target.value = this.model.get('value');
          }
          else {
              // Handle both the unbounded and bounded case by
              // checking to see if the max/min properties are defined
              var boundedValue = numericalValue;
              if (this.model.get('max') !== undefined) {
                  boundedValue = Math.min(this.model.get('max'), boundedValue);
              }
              if (this.model.get('min') !== undefined) {
                  boundedValue = Math.max(this.model.get('min'), boundedValue);
              }
              if (boundedValue !== numericalValue) {
                  target.value = boundedValue;
                  numericalValue = boundedValue;
              }
              // Apply the value if it has changed.
              if (numericalValue !== this.model.get('value')) {
                  this.model.set('value', numericalValue, { updated_view: this });
                  this.touch();
              }
          }
      };
      return IntTextView;
  }(DescriptionView));
  var ProgressStyleModel = /** @class */ (function (_super) {
      __extends$k(ProgressStyleModel, _super);
      function ProgressStyleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ProgressStyleModel.prototype.defaults = function () {
          return __assign$c(__assign$c({}, _super.prototype.defaults.call(this)), { _model_name: 'ProgressStyleModel' });
      };
      ProgressStyleModel.styleProperties = __assign$c(__assign$c({}, DescriptionStyleModel.styleProperties), { bar_color: {
              selector: '.progress-bar',
              attribute: 'background-color',
              default: null
          } });
      return ProgressStyleModel;
  }(DescriptionStyleModel));
  var IntProgressModel = /** @class */ (function (_super) {
      __extends$k(IntProgressModel, _super);
      function IntProgressModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      IntProgressModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'IntProgressModel',
              _view_name: 'ProgressView',
              orientation: 'horizontal',
              bar_style: '',
              style: null
          });
      };
      return IntProgressModel;
  }(BoundedIntModel));
  var ProgressView = /** @class */ (function (_super) {
      __extends$k(ProgressView, _super);
      function ProgressView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ProgressView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.listenTo(this.model, 'change:bar_style', this.update_bar_style);
          this.pWidget.addClass('jupyter-widgets');
      };
      ProgressView.prototype.render = function () {
          _super.prototype.render.call(this);
          var orientation = this.model.get('orientation');
          var className = orientation === 'horizontal' ?
              'widget-hprogress' : 'widget-vprogress';
          this.el.classList.add(className);
          this.progress = document.createElement('div');
          this.progress.classList.add('progress');
          this.progress.style.position = 'relative';
          this.el.appendChild(this.progress);
          this.bar = document.createElement('div');
          this.bar.classList.add('progress-bar');
          this.bar.style.position = 'absolute';
          this.bar.style.bottom = '0px';
          this.bar.style.left = '0px';
          this.progress.appendChild(this.bar);
          // Set defaults.
          this.update();
          this.set_bar_style();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      ProgressView.prototype.update = function () {
          var value = this.model.get('value');
          var max = this.model.get('max');
          var min = this.model.get('min');
          var orientation = this.model.get('orientation');
          var percent = 100.0 * (value - min) / (max - min);
          if (orientation === 'horizontal') {
              this.el.classList.remove('widget-inline-vbox');
              this.el.classList.remove('widget-vprogress');
              this.el.classList.add('widget-inline-hbox');
              this.el.classList.add('widget-hprogress');
              this.bar.style.width = percent + '%';
              this.bar.style.height = '100%';
          }
          else {
              this.el.classList.remove('widget-inline-hbox');
              this.el.classList.remove('widget-hprogress');
              this.el.classList.add('widget-inline-vbox');
              this.el.classList.add('widget-vprogress');
              this.bar.style.width = '100%';
              this.bar.style.height = percent + '%';
          }
          return _super.prototype.update.call(this);
      };
      ProgressView.prototype.update_bar_style = function () {
          this.update_mapped_classes(ProgressView.class_map, 'bar_style', this.bar);
      };
      ProgressView.prototype.set_bar_style = function () {
          this.set_mapped_classes(ProgressView.class_map, 'bar_style', this.bar);
      };
      ProgressView.class_map = {
          success: ['progress-bar-success'],
          info: ['progress-bar-info'],
          warning: ['progress-bar-warning'],
          danger: ['progress-bar-danger']
      };
      return ProgressView;
  }(DescriptionView));
  var PlayModel = /** @class */ (function (_super) {
      __extends$k(PlayModel, _super);
      function PlayModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      PlayModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'PlayModel',
              _view_name: 'PlayView',
              _playing: false,
              _repeat: false,
              show_repeat: true,
              interval: 100,
              step: 1,
              disabled: false,
          });
      };
      PlayModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
      };
      PlayModel.prototype.loop = function () {
          if (this.get('_playing')) {
              var next_value = this.get('value') + this.get('step');
              if (next_value <= this.get('max')) {
                  this.set('value', next_value);
                  this.schedule_next();
              }
              else {
                  if (this.get('_repeat')) {
                      this.set('value', this.get('min'));
                      this.schedule_next();
                  }
                  else {
                      this.set('_playing', false);
                  }
              }
              this.save_changes();
          }
      };
      PlayModel.prototype.schedule_next = function () {
          window.setTimeout(this.loop.bind(this), this.get('interval'));
      };
      PlayModel.prototype.stop = function () {
          this.set('_playing', false);
          this.set('value', this.get('min'));
          this.save_changes();
      };
      PlayModel.prototype.pause = function () {
          this.set('_playing', false);
          this.save_changes();
      };
      PlayModel.prototype.play = function () {
          this.set('_playing', true);
          if (this.get('value') == this.get('max')) {
              // if the value is at the end, reset if first, and then schedule the next
              this.set('value', this.get('min'));
              this.schedule_next();
              this.save_changes();
          }
          else {
              // otherwise directly start with the next value
              // loop will call save_changes in this case
              this.loop();
          }
      };
      PlayModel.prototype.repeat = function () {
          this.set('_repeat', !this.get('_repeat'));
          this.save_changes();
      };
      return PlayModel;
  }(BoundedIntModel));
  var PlayView = /** @class */ (function (_super) {
      __extends$k(PlayView, _super);
      function PlayView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      PlayView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-play');
          this.playButton = document.createElement('button');
          this.pauseButton = document.createElement('button');
          this.stopButton = document.createElement('button');
          this.repeatButton = document.createElement('button');
          this.playButton.className = 'jupyter-button';
          this.pauseButton.className = 'jupyter-button';
          this.stopButton.className = 'jupyter-button';
          this.repeatButton.className = 'jupyter-button';
          this.el.appendChild(this.playButton); // Toggle button with playing
          this.el.appendChild(this.pauseButton); // Disable if not playing
          this.el.appendChild(this.stopButton); // Disable if not playing
          this.el.appendChild(this.repeatButton); // Always enabled, but may be hidden
          var playIcon = document.createElement('i');
          playIcon.className = 'fa fa-play';
          this.playButton.appendChild(playIcon);
          var pauseIcon = document.createElement('i');
          pauseIcon.className = 'fa fa-pause';
          this.pauseButton.appendChild(pauseIcon);
          var stopIcon = document.createElement('i');
          stopIcon.className = 'fa fa-stop';
          this.stopButton.appendChild(stopIcon);
          var repeatIcon = document.createElement('i');
          repeatIcon.className = 'fa fa-retweet';
          this.repeatButton.appendChild(repeatIcon);
          this.playButton.onclick = this.model.play.bind(this.model);
          this.pauseButton.onclick = this.model.pause.bind(this.model);
          this.stopButton.onclick = this.model.stop.bind(this.model);
          this.repeatButton.onclick = this.model.repeat.bind(this.model);
          this.listenTo(this.model, 'change:_playing', this.update_playing);
          this.listenTo(this.model, 'change:_repeat', this.update_repeat);
          this.listenTo(this.model, 'change:show_repeat', this.update_repeat);
          this.update_playing();
          this.update_repeat();
          this.update();
      };
      PlayView.prototype.update = function () {
          var disabled = this.model.get('disabled');
          this.playButton.disabled = disabled;
          this.pauseButton.disabled = disabled;
          this.stopButton.disabled = disabled;
          this.repeatButton.disabled = disabled;
          this.update_playing();
      };
      PlayView.prototype.update_playing = function () {
          var playing = this.model.get('_playing');
          var disabled = this.model.get('disabled');
          if (playing) {
              if (!disabled) {
                  this.pauseButton.disabled = false;
              }
              this.playButton.classList.add('mod-active');
          }
          else {
              if (!disabled) {
                  this.pauseButton.disabled = true;
              }
              this.playButton.classList.remove('mod-active');
          }
      };
      PlayView.prototype.update_repeat = function () {
          var repeat = this.model.get('_repeat');
          this.repeatButton.style.display = this.model.get('show_repeat') ? this.playButton.style.display : 'none';
          if (repeat) {
              this.repeatButton.classList.add('mod-active');
          }
          else {
              this.repeatButton.classList.remove('mod-active');
          }
      };
      return PlayView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$l = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var FloatModel = /** @class */ (function (_super) {
      __extends$l(FloatModel, _super);
      function FloatModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      FloatModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'FloatModel',
              value: 0,
          });
      };
      return FloatModel;
  }(CoreDescriptionModel));
  var BoundedFloatModel = /** @class */ (function (_super) {
      __extends$l(BoundedFloatModel, _super);
      function BoundedFloatModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoundedFloatModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'BoundedFloatModel',
              max: 100.0,
              min: 0.0
          });
      };
      return BoundedFloatModel;
  }(FloatModel));
  var FloatSliderModel = /** @class */ (function (_super) {
      __extends$l(FloatSliderModel, _super);
      function FloatSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      FloatSliderModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'FloatSliderModel',
              _view_name: 'FloatSliderView',
              step: 1.0,
              orientation: 'horizontal',
              _range: false,
              readout: true,
              readout_format: '.2f',
              slider_color: null,
              continuous_update: true,
              disabled: false,
          });
      };
      FloatSliderModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
          this.on('change:readout_format', this.update_readout_format, this);
          this.update_readout_format();
      };
      FloatSliderModel.prototype.update_readout_format = function () {
          this.readout_formatter = format(this.get('readout_format'));
      };
      return FloatSliderModel;
  }(BoundedFloatModel));
  var FloatLogSliderModel = /** @class */ (function (_super) {
      __extends$l(FloatLogSliderModel, _super);
      function FloatLogSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      FloatLogSliderModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'FloatLogSliderModel',
              _view_name: 'FloatLogSliderView',
              step: 0.1,
              orientation: 'horizontal',
              _range: false,
              readout: true,
              readout_format: '.3g',
              slider_color: null,
              continuous_update: true,
              disabled: false,
              base: 10.,
              value: 1.0,
              min: 0,
              max: 4
          });
      };
      FloatLogSliderModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
          this.on('change:readout_format', this.update_readout_format, this);
          this.update_readout_format();
      };
      FloatLogSliderModel.prototype.update_readout_format = function () {
          this.readout_formatter = format(this.get('readout_format'));
      };
      return FloatLogSliderModel;
  }(BoundedFloatModel));
  var FloatRangeSliderModel = /** @class */ (function (_super) {
      __extends$l(FloatRangeSliderModel, _super);
      function FloatRangeSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      return FloatRangeSliderModel;
  }(FloatSliderModel));
  var FloatSliderView = /** @class */ (function (_super) {
      __extends$l(FloatSliderView, _super);
      function FloatSliderView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._parse_value = parseFloat;
          return _this;
      }
      /**
       * Validate the value of the slider before sending it to the back-end
       * and applying it to the other views on the page.
       */
      FloatSliderView.prototype._validate_slide_value = function (x) {
          return x;
      };
      return FloatSliderView;
  }(IntSliderView));
  var FloatLogSliderView = /** @class */ (function (_super) {
      __extends$l(FloatLogSliderView, _super);
      function FloatLogSliderView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._parse_value = parseFloat;
          return _this;
      }
      FloatLogSliderView.prototype.update = function (options) {
          _super.prototype.update.call(this, options);
          var min = this.model.get('min');
          var max = this.model.get('max');
          var value = this.model.get('value');
          var base = this.model.get('base');
          var log_value = Math.log(value) / Math.log(base);
          if (log_value > max) {
              log_value = max;
          }
          else if (log_value < min) {
              log_value = min;
          }
          this.$slider.slider('option', 'value', log_value);
          this.readout.textContent = this.valueToString(value);
          if (this.model.get('value') !== value) {
              this.model.set('value', value, { updated_view: this });
              this.touch();
          }
      };
      /**
       * Write value to a string
       */
      FloatLogSliderView.prototype.valueToString = function (value) {
          var format = this.model.readout_formatter;
          return format(value);
      };
      /**
       * Parse value from a string
       */
      FloatLogSliderView.prototype.stringToValue = function (text) {
          return this._parse_value(text);
      };
      /**
       * this handles the entry of text into the contentEditable label first, the
       * value is checked if it contains a parseable value then it is clamped
       * within the min-max range of the slider finally, the model is updated if
       * the value is to be changed
       *
       * if any of these conditions are not met, the text is reset
       */
      FloatLogSliderView.prototype.handleTextChange = function () {
          var value = this.stringToValue(this.readout.textContent);
          var vmin = this.model.get('min');
          var vmax = this.model.get('max');
          var base = this.model.get('base');
          if (isNaN(value)) {
              this.readout.textContent = this.valueToString(this.model.get('value'));
          }
          else {
              value = Math.max(Math.min(value, Math.pow(base, vmax)), Math.pow(base, vmin));
              if (value !== this.model.get('value')) {
                  this.readout.textContent = this.valueToString(value);
                  this.model.set('value', value, { updated_view: this });
                  this.touch();
              }
              else {
                  this.readout.textContent = this.valueToString(this.model.get('value'));
              }
          }
      };
      /**
       * Called when the slider value is changing.
       */
      FloatLogSliderView.prototype.handleSliderChange = function (e, ui) {
          var base = this.model.get('base');
          var actual_value = Math.pow(base, this._validate_slide_value(ui.value));
          this.readout.textContent = this.valueToString(actual_value);
          // Only persist the value while sliding if the continuous_update
          // trait is set to true.
          if (this.model.get('continuous_update')) {
              this.handleSliderChanged(e, ui);
          }
      };
      /**
       * Called when the slider value has changed.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      FloatLogSliderView.prototype.handleSliderChanged = function (e, ui) {
          var base = this.model.get('base');
          var actual_value = Math.pow(base, this._validate_slide_value(ui.value));
          this.model.set('value', actual_value, { updated_view: this });
          this.touch();
      };
      FloatLogSliderView.prototype._validate_slide_value = function (x) {
          return x;
      };
      return FloatLogSliderView;
  }(BaseIntSliderView));
  var FloatRangeSliderView = /** @class */ (function (_super) {
      __extends$l(FloatRangeSliderView, _super);
      function FloatRangeSliderView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._parse_value = parseFloat;
          // matches: whitespace?, float, whitespace?, (hyphen, colon, or en-dash), whitespace?, float
          _this._range_regex = /^\s*([+-]?(?:\d*\.?\d+|\d+\.)(?:[eE][-:]?\d+)?)\s*[-:–]\s*([+-]?(?:\d*\.?\d+|\d+\.)(?:[eE][+-]?\d+)?)/;
          return _this;
      }
      /**
       * Validate the value of the slider before sending it to the back-end
       * and applying it to the other views on the page.
       */
      FloatRangeSliderView.prototype._validate_slide_value = function (x) {
          return x;
      };
      return FloatRangeSliderView;
  }(IntRangeSliderView));
  var FloatTextModel = /** @class */ (function (_super) {
      __extends$l(FloatTextModel, _super);
      function FloatTextModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      FloatTextModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'FloatTextModel',
              _view_name: 'FloatTextView',
              disabled: false,
              continuous_update: false,
          });
      };
      return FloatTextModel;
  }(FloatModel));
  var BoundedFloatTextModel = /** @class */ (function (_super) {
      __extends$l(BoundedFloatTextModel, _super);
      function BoundedFloatTextModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      BoundedFloatTextModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'BoundedFloatTextModel',
              _view_name: 'FloatTextView',
              disabled: false,
              continuous_update: false,
              step: 0.1
          });
      };
      return BoundedFloatTextModel;
  }(BoundedFloatModel));
  var FloatTextView = /** @class */ (function (_super) {
      __extends$l(FloatTextView, _super);
      function FloatTextView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._parse_value = parseFloat;
          _this._default_step = 'any';
          return _this;
      }
      /**
       * Handle key press
       */
      FloatTextView.prototype.handleKeypress = function (e) {
          // Overwrite IntTextView's handleKeypress
          // which prevents decimal points.
          e.stopPropagation();
      };
      /**
       * Handle key up
       */
      FloatTextView.prototype.handleKeyUp = function (e) {
          // Overwrite IntTextView's handleKeyUp
          // which prevents decimal points.
      };
      return FloatTextView;
  }(IntTextView));
  var FloatProgressModel = /** @class */ (function (_super) {
      __extends$l(FloatProgressModel, _super);
      function FloatProgressModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      FloatProgressModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'FloatProgressModel',
              _view_name: 'ProgressView',
              orientation: 'horizontal',
              bar_style: '',
              style: null
          });
      };
      return FloatProgressModel;
  }(BoundedFloatModel));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$m = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$d = (undefined && undefined.__assign) || function () {
      __assign$d = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$d.apply(this, arguments);
  };
  var ControllerButtonModel = /** @class */ (function (_super) {
      __extends$m(ControllerButtonModel, _super);
      function ControllerButtonModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ControllerButtonModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'ControllerButtonModel',
              _view_name: 'ControllerButtonView',
              value: 0.0,
              pressed: false
          });
      };
      return ControllerButtonModel;
  }(CoreDOMWidgetModel));
  /**
   * Very simple view for a gamepad button.
   */
  var ControllerButtonView = /** @class */ (function (_super) {
      __extends$m(ControllerButtonView, _super);
      function ControllerButtonView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ControllerButtonView.prototype.render = function () {
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-controller-button');
          this.el.style.width = 'fit-content';
          this.support = document.createElement('div');
          this.support.style.position = 'relative';
          this.support.style.margin = '1px';
          this.support.style.width = '16px';
          this.support.style.height = '16px';
          this.support.style.border = '1px solid black';
          this.support.style.background = 'lightgray';
          this.el.appendChild(this.support);
          this.bar = document.createElement('div');
          this.bar.style.position = 'absolute';
          this.bar.style.width = '100%';
          this.bar.style.bottom = '0px';
          this.bar.style.background = 'gray';
          this.support.appendChild(this.bar);
          this.update();
          this.label = document.createElement('div');
          this.label.textContent = this.model.get('description');
          this.label.style.textAlign = 'center';
          this.el.appendChild(this.label);
      };
      ControllerButtonView.prototype.update = function () {
          this.bar.style.height = (100 * this.model.get('value')) + '%';
      };
      return ControllerButtonView;
  }(DOMWidgetView));
  var ControllerAxisModel = /** @class */ (function (_super) {
      __extends$m(ControllerAxisModel, _super);
      function ControllerAxisModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ControllerAxisModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'ControllerAxisModel',
              _view_name: 'ControllerAxisView',
              value: 0.0
          });
      };
      return ControllerAxisModel;
  }(CoreDOMWidgetModel));
  /**
   * Very simple view for a gamepad axis.
   */
  var ControllerAxisView = /** @class */ (function (_super) {
      __extends$m(ControllerAxisView, _super);
      function ControllerAxisView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ControllerAxisView.prototype.render = function () {
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-controller-axis');
          this.el.style.width = '16px';
          this.el.style.padding = '4px';
          this.support = document.createElement('div');
          this.support.style.position = 'relative';
          this.support.style.margin = '1px';
          this.support.style.width = '4px';
          this.support.style.height = '64px';
          this.support.style.border = '1px solid black';
          this.support.style.background = 'lightgray';
          this.bullet = document.createElement('div');
          this.bullet.style.position = 'absolute';
          this.bullet.style.margin = '-3px';
          this.bullet.style.boxSizing = 'unset';
          this.bullet.style.width = '10px';
          this.bullet.style.height = '10px';
          this.bullet.style.background = 'gray';
          this.label = document.createElement('div');
          this.label.textContent = this.model.get('description');
          this.label.style.textAlign = 'center';
          this.support.appendChild(this.bullet);
          this.el.appendChild(this.support);
          this.el.appendChild(this.label);
          this.update();
      };
      ControllerAxisView.prototype.update = function () {
          this.bullet.style.top = (50 * (this.model.get('value') + 1)) + '%';
      };
      return ControllerAxisView;
  }(DOMWidgetView));
  var ControllerModel = /** @class */ (function (_super) {
      __extends$m(ControllerModel, _super);
      function ControllerModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ControllerModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'ControllerModel',
              _view_name: 'ControllerView',
              index: 0,
              name: '',
              mapping: '',
              connected: false,
              timestamp: 0,
              buttons: [],
              axes: []
          });
      };
      ControllerModel.prototype.initialize = function (attributes, options) {
          _super.prototype.initialize.call(this, attributes, options);
          if (navigator.getGamepads === void 0) {
              // Checks if the browser supports the gamepad API
              this.readout = 'This browser does not support gamepads.';
              console.error(this.readout);
          }
          else {
              // Start the wait loop, and listen to updates of the only
              // user-provided attribute, the gamepad index.
              this.readout = 'Connect gamepad and press any button.';
              if (this.get('connected')) {
                  // No need to re-create Button and Axis widgets, re-use
                  // the models provided by the backend which may already
                  // be wired to other things.
                  this.update_loop();
              }
              else {
                  // Wait for a gamepad to be connected.
                  this.wait_loop();
              }
          }
      };
      /**
       * Waits for a gamepad to be connected at the provided index.
       * Once one is connected, it will start the update loop, which
       * populates the update of axes and button values.
       */
      ControllerModel.prototype.wait_loop = function () {
          var index = this.get('index');
          var pad = navigator.getGamepads()[index];
          if (pad) {
              var that_1 = this;
              this.setup(pad).then(function (controls) {
                  that_1.set(controls);
                  that_1.save_changes();
                  window.requestAnimationFrame(that_1.update_loop.bind(that_1));
              });
          }
          else {
              window.requestAnimationFrame(this.wait_loop.bind(this));
          }
      };
      /**
       * Given a native gamepad object, returns a promise for a dictionary of
       * controls, of the form
       * {
       *     buttons: list of Button models,
       *     axes: list of Axis models,
       * }
       */
      ControllerModel.prototype.setup = function (pad) {
          // Set up the main gamepad attributes
          this.set({
              name: pad.id,
              mapping: pad.mapping,
              connected: pad.connected,
              timestamp: pad.timestamp
          });
          // Create buttons and axes. When done, start the update loop
          var that = this;
          return resolvePromisesDict({
              buttons: Promise.all(pad.buttons.map(function (btn, index) {
                  return that._create_button_model(index);
              })),
              axes: Promise.all(pad.axes.map(function (axis, index) {
                  return that._create_axis_model(index);
              })),
          });
      };
      /**
       * Update axes and buttons values, until the gamepad is disconnected.
       * When the gamepad is disconnected, this.reset_gamepad is called.
       */
      ControllerModel.prototype.update_loop = function () {
          var index = this.get('index');
          var id = this.get('name');
          var pad = navigator.getGamepads()[index];
          if (pad && index === pad.index && id === pad.id) {
              this.set({
                  timestamp: pad.timestamp,
                  connected: pad.connected
              });
              this.save_changes();
              this.get('buttons').forEach(function (model, index) {
                  model.set({
                      value: pad.buttons[index].value,
                      pressed: pad.buttons[index].pressed
                  });
                  model.save_changes();
              });
              this.get('axes').forEach(function (model, index) {
                  model.set('value', pad.axes[index]);
                  model.save_changes();
              });
              window.requestAnimationFrame(this.update_loop.bind(this));
          }
          else {
              this.reset_gamepad();
          }
      };
      /**
       * Resets the gamepad attributes, and start the wait_loop.
       */
      ControllerModel.prototype.reset_gamepad = function () {
          this.get('buttons').forEach(function (button) {
              button.close();
          });
          this.get('axes').forEach(function (axis) {
              axis.close();
          });
          this.set({
              name: '',
              mapping: '',
              connected: false,
              timestamp: 0.0,
              buttons: [],
              axes: []
          });
          this.save_changes();
          window.requestAnimationFrame(this.wait_loop.bind(this));
      };
      /**
       * Creates a gamepad button widget.
       */
      ControllerModel.prototype._create_button_model = function (index) {
          return this.widget_manager.new_widget({
              model_name: 'ControllerButtonModel',
              model_module: '@jupyter-widgets/controls',
              model_module_version: this.get('_model_module_version'),
              view_name: 'ControllerButtonView',
              view_module: '@jupyter-widgets/controls',
              view_module_version: this.get('_view_module_version'),
          }).then(function (model) {
              model.set('description', index);
              return model;
          });
      };
      /**
       * Creates a gamepad axis widget.
       */
      ControllerModel.prototype._create_axis_model = function (index) {
          return this.widget_manager.new_widget({
              model_name: 'ControllerAxisModel',
              model_module: '@jupyter-widgets/controls',
              model_module_version: this.get('_model_module_version'),
              view_name: 'ControllerAxisView',
              view_module: '@jupyter-widgets/controls',
              view_module_version: this.get('_view_module_version'),
          }).then(function (model) {
              model.set('description', index);
              return model;
          });
      };
      ControllerModel.serializers = __assign$d(__assign$d({}, CoreDOMWidgetModel.serializers), { buttons: { deserialize: unpack_models }, axes: { deserialize: unpack_models } });
      return ControllerModel;
  }(CoreDOMWidgetModel));
  /**
   * A simple view for a gamepad.
   */
  var ControllerView = /** @class */ (function (_super) {
      __extends$m(ControllerView, _super);
      function ControllerView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ControllerView.prototype._createElement = function (tagName) {
          this.pWidget = new JupyterPhosphorPanelWidget({ view: this });
          return this.pWidget.node;
      };
      ControllerView.prototype._setElement = function (el) {
          if (this.el || el !== this.pWidget.node) {
              // Boxes don't allow setting the element beyond the initial creation.
              throw new Error('Cannot reset the DOM element.');
          }
          this.el = this.pWidget.node;
          this.$el = $$1(this.pWidget.node);
      };
      ControllerView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.button_views = new ViewList(this.add_button, null, this);
          this.listenTo(this.model, 'change:buttons', function (model, value) {
              this.button_views.update(value);
          });
          this.axis_views = new ViewList(this.add_axis, null, this);
          this.listenTo(this.model, 'change:axes', function (model, value) {
              this.axis_views.update(value);
          });
          this.listenTo(this.model, 'change:name', this.update_label);
      };
      ControllerView.prototype.render = function () {
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-controller');
          this.label = document.createElement('div');
          this.el.appendChild(this.label);
          this.axis_box = new Panel();
          this.axis_box.node.style.display = 'flex';
          this.pWidget.addWidget(this.axis_box);
          this.button_box = new Panel();
          this.button_box.node.style.display = 'flex';
          this.pWidget.addWidget(this.button_box);
          this.button_views.update(this.model.get('buttons'));
          this.axis_views.update(this.model.get('axes'));
          this.update_label();
      };
      ControllerView.prototype.update_label = function () {
          this.label.textContent = this.model.get('name') || this.model.readout;
      };
      ControllerView.prototype.add_button = function (model) {
          var _this = this;
          // we insert a dummy element so the order is preserved when we add
          // the rendered content later.
          var dummy = new Widget();
          this.button_box.addWidget(dummy);
          return this.create_child_view(model).then(function (view) {
              // replace the dummy widget with the new one.
              var i = ArrayExt.firstIndexOf(_this.button_box.widgets, dummy);
              _this.button_box.insertWidget(i, view.pWidget);
              dummy.dispose();
              return view;
          }).catch(reject$2('Could not add child button view to controller', true));
      };
      ControllerView.prototype.add_axis = function (model) {
          var _this = this;
          // we insert a dummy element so the order is preserved when we add
          // the rendered content later.
          var dummy = new Widget();
          this.axis_box.addWidget(dummy);
          return this.create_child_view(model).then(function (view) {
              // replace the dummy widget with the new one.
              var i = ArrayExt.firstIndexOf(_this.axis_box.widgets, dummy);
              _this.axis_box.insertWidget(i, view.pWidget);
              dummy.dispose();
              return view;
          }).catch(reject$2('Could not add child axis view to controller', true));
      };
      ControllerView.prototype.remove = function () {
          _super.prototype.remove.call(this);
          this.button_views.remove();
          this.axis_views.remove();
      };
      return ControllerView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$n = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$e = (undefined && undefined.__assign) || function () {
      __assign$e = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$e.apply(this, arguments);
  };
  var SelectionModel = /** @class */ (function (_super) {
      __extends$n(SelectionModel, _super);
      function SelectionModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SelectionModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'SelectionModel', index: '', _options_labels: [], disabled: false });
      };
      return SelectionModel;
  }(CoreDescriptionModel));
  var DropdownModel = /** @class */ (function (_super) {
      __extends$n(DropdownModel, _super);
      function DropdownModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      DropdownModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'DropdownModel', _view_name: 'DropdownView', button_style: '' });
      };
      return DropdownModel;
  }(SelectionModel));
  // TODO: Make a phosphor dropdown control, wrapped in DropdownView. Also, fix
  // bugs in keyboard handling. See
  // https://github.com/jupyter-widgets/ipywidgets/issues/1055 and
  // https://github.com/jupyter-widgets/ipywidgets/issues/1049
  // For now, we subclass SelectView to provide DropdownView
  // For the old code, see commit f68bfbc566f3a78a8f3350b438db8ed523ce3642
  var DropdownView = /** @class */ (function (_super) {
      __extends$n(DropdownView, _super);
      function DropdownView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor.
       */
      DropdownView.prototype.initialize = function (parameters) {
          var _this = this;
          _super.prototype.initialize.call(this, parameters);
          this.listenTo(this.model, 'change:_options_labels', function () { return _this._updateOptions(); });
      };
      /**
       * Called when view is rendered.
       */
      DropdownView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-dropdown');
          this.listbox = document.createElement('select');
          this.listbox.id = this.label.htmlFor = uuid();
          this.el.appendChild(this.listbox);
          this._updateOptions();
          this.update();
      };
      /**
       * Update the contents of this view
       */
      DropdownView.prototype.update = function () {
          // Disable listbox if needed
          this.listbox.disabled = this.model.get('disabled');
          // Select the correct element
          var index = this.model.get('index');
          this.listbox.selectedIndex = index === null ? -1 : index;
          return _super.prototype.update.call(this);
      };
      DropdownView.prototype._updateOptions = function () {
          this.listbox.textContent = '';
          var items = this.model.get('_options_labels');
          for (var i = 0; i < items.length; i++) {
              var item = items[i];
              var option = document.createElement('option');
              option.textContent = item.replace(/ /g, '\xa0'); // space -> &nbsp;
              option.setAttribute('data-value', encodeURIComponent(item));
              option.value = item;
              this.listbox.appendChild(option);
          }
      };
      DropdownView.prototype.events = function () {
          return {
              'change select': '_handle_change'
          };
      };
      /**
       * Handle when a new value is selected.
       */
      DropdownView.prototype._handle_change = function () {
          this.model.set('index', this.listbox.selectedIndex === -1 ? null : this.listbox.selectedIndex);
          this.touch();
      };
      return DropdownView;
  }(DescriptionView));
  var SelectModel = /** @class */ (function (_super) {
      __extends$n(SelectModel, _super);
      function SelectModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SelectModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'SelectModel', _view_name: 'SelectView', rows: 5 });
      };
      return SelectModel;
  }(SelectionModel));
  var SelectView = /** @class */ (function (_super) {
      __extends$n(SelectView, _super);
      function SelectView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor.
       */
      SelectView.prototype.initialize = function (parameters) {
          var _this = this;
          _super.prototype.initialize.call(this, parameters);
          this.listenTo(this.model, 'change:_options_labels', function () { return _this._updateOptions(); });
          this.listenTo(this.model, 'change:index', function (model, value, options) { return _this.updateSelection(options); });
          // Create listbox here so that subclasses can modify it before it is populated in render()
          this.listbox = document.createElement('select');
      };
      /**
       * Called when view is rendered.
       */
      SelectView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-select');
          this.listbox.id = this.label.htmlFor = uuid();
          this.el.appendChild(this.listbox);
          this._updateOptions();
          this.update();
          this.updateSelection();
      };
      /**
       * Update the contents of this view
       */
      SelectView.prototype.update = function () {
          _super.prototype.update.call(this);
          this.listbox.disabled = this.model.get('disabled');
          var rows = this.model.get('rows');
          if (rows === null) {
              rows = '';
          }
          this.listbox.setAttribute('size', rows);
      };
      SelectView.prototype.updateSelection = function (options) {
          if (options === void 0) { options = {}; }
          if (options.updated_view === this) {
              return;
          }
          var index = this.model.get('index');
          this.listbox.selectedIndex = index === null ? -1 : index;
      };
      SelectView.prototype._updateOptions = function () {
          this.listbox.textContent = '';
          var items = this.model.get('_options_labels');
          for (var i = 0; i < items.length; i++) {
              var item = items[i];
              var option = document.createElement('option');
              option.textContent = item.replace(/ /g, '\xa0'); // space -> &nbsp;
              option.setAttribute('data-value', encodeURIComponent(item));
              option.value = item;
              this.listbox.appendChild(option);
          }
      };
      SelectView.prototype.events = function () {
          return {
              'change select': '_handle_change'
          };
      };
      /**
       * Handle when a new value is selected.
       */
      SelectView.prototype._handle_change = function () {
          this.model.set('index', this.listbox.selectedIndex, { updated_view: this });
          this.touch();
      };
      return SelectView;
  }(DescriptionView));
  var RadioButtonsModel = /** @class */ (function (_super) {
      __extends$n(RadioButtonsModel, _super);
      function RadioButtonsModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      RadioButtonsModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'RadioButtonsModel', _view_name: 'RadioButtonsView', tooltips: [], icons: [], button_style: '' });
      };
      return RadioButtonsModel;
  }(SelectionModel));
  var RadioButtonsView = /** @class */ (function (_super) {
      __extends$n(RadioButtonsView, _super);
      function RadioButtonsView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      RadioButtonsView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-radio');
          this.container = document.createElement('div');
          this.el.appendChild(this.container);
          this.container.classList.add('widget-radio-box');
          this.update();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      RadioButtonsView.prototype.update = function (options) {
          var view = this;
          var items = this.model.get('_options_labels');
          var radios = pluck(this.container.querySelectorAll('input[type="radio"]'), 'value');
          var stale = items.length != radios.length;
          if (!stale) {
              for (var i = 0, len = items.length; i < len; ++i) {
                  if (radios[i] !== items[i]) {
                      stale = true;
                      break;
                  }
              }
          }
          if (stale && (options === undefined || options.updated_view !== this)) {
              // Add items to the DOM.
              this.container.textContent = '';
              items.forEach(function (item, index) {
                  var label = document.createElement('label');
                  label.textContent = item;
                  view.container.appendChild(label);
                  var radio = document.createElement('input');
                  radio.setAttribute('type', 'radio');
                  radio.value = index.toString();
                  radio.setAttribute('data-value', encodeURIComponent(item));
                  label.appendChild(radio);
              });
          }
          items.forEach(function (item, index) {
              var item_query = 'input[data-value="' +
                  encodeURIComponent(item) + '"]';
              var radio = view.container.querySelectorAll(item_query);
              if (radio.length > 0) {
                  var radio_el = radio[0];
                  radio_el.checked = view.model.get('index') === index;
                  radio_el.disabled = view.model.get('disabled');
              }
          });
          // Schedule adjustPadding asynchronously to
          // allow dom elements to be created properly
          setTimeout(this.adjustPadding, 0, this);
          return _super.prototype.update.call(this, options);
      };
      /**
       * Adjust Padding to Multiple of Line Height
       *
       * Adjust margins so that the overall height
       * is a multiple of a single line height.
       *
       * This widget needs it because radio options
       * are spaced tighter than individual widgets
       * yet we would like the full widget line up properly
       * when displayed side-by-side with other widgets.
       */
      RadioButtonsView.prototype.adjustPadding = function (e) {
          // Vertical margins on a widget
          var elStyles = window.getComputedStyle(e.el);
          var margins = parseInt(elStyles.marginTop, 10) + parseInt(elStyles.marginBottom, 10);
          // Total spaces taken by a single-line widget
          var lineHeight = e.label.offsetHeight + margins;
          // Current adjustment value on this widget
          var cStyles = window.getComputedStyle(e.container);
          var containerMargin = parseInt(cStyles.marginBottom);
          // How far we are off from a multiple of single windget lines
          var diff = (e.el.offsetHeight + margins - containerMargin) % lineHeight;
          // Apply the new adjustment
          var extraMargin = diff == 0 ? 0 : (lineHeight - diff);
          e.container.style.marginBottom = extraMargin + 'px';
      };
      RadioButtonsView.prototype.events = function () {
          return {
              'click input[type="radio"]': '_handle_click'
          };
      };
      /**
       * Handle when a value is clicked.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      RadioButtonsView.prototype._handle_click = function (event) {
          var target = event.target;
          this.model.set('index', parseInt(target.value), { updated_view: this });
          this.touch();
      };
      return RadioButtonsView;
  }(DescriptionView));
  var ToggleButtonsStyleModel = /** @class */ (function (_super) {
      __extends$n(ToggleButtonsStyleModel, _super);
      function ToggleButtonsStyleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ToggleButtonsStyleModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'ToggleButtonsStyleModel',
          });
      };
      ToggleButtonsStyleModel.styleProperties = __assign$e(__assign$e({}, DescriptionStyleModel.styleProperties), { button_width: {
              selector: '.widget-toggle-button',
              attribute: 'width',
              default: null
          }, font_weight: {
              selector: '.widget-toggle-button',
              attribute: 'font-weight',
              default: ''
          } });
      return ToggleButtonsStyleModel;
  }(DescriptionStyleModel));
  var ToggleButtonsModel = /** @class */ (function (_super) {
      __extends$n(ToggleButtonsModel, _super);
      function ToggleButtonsModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ToggleButtonsModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'ToggleButtonsModel', _view_name: 'ToggleButtonsView' });
      };
      return ToggleButtonsModel;
  }(SelectionModel));
  var ToggleButtonsView = /** @class */ (function (_super) {
      __extends$n(ToggleButtonsView, _super);
      function ToggleButtonsView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ToggleButtonsView.prototype.initialize = function (options) {
          this._css_state = {};
          _super.prototype.initialize.call(this, options);
          this.listenTo(this.model, 'change:button_style', this.update_button_style);
      };
      /**
       * Called when view is rendered.
       */
      ToggleButtonsView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-toggle-buttons');
          this.buttongroup = document.createElement('div');
          this.el.appendChild(this.buttongroup);
          this.update();
          this.set_button_style();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      ToggleButtonsView.prototype.update = function (options) {
          var view = this;
          var items = this.model.get('_options_labels');
          var icons = this.model.get('icons') || [];
          var previous_icons = this.model.previous('icons') || [];
          var previous_bstyle = ToggleButtonsView.classMap[this.model.previous('button_style')] || '';
          var tooltips = view.model.get('tooltips') || [];
          var disabled = this.model.get('disabled');
          var buttons = this.buttongroup.querySelectorAll('button');
          var values = pluck(buttons, 'value');
          var stale = false;
          for (var i = 0, len = items.length; i < len; ++i) {
              if (values[i] !== items[i] || icons[i] !== previous_icons[i]) {
                  stale = true;
                  break;
              }
          }
          if (stale && (options === undefined || options.updated_view !== this)) {
              // Add items to the DOM.
              this.buttongroup.textContent = '';
              items.forEach(function (item, index) {
                  var item_html;
                  var empty = item.trim().length === 0 &&
                      (!icons[index] || icons[index].trim().length === 0);
                  if (empty) {
                      item_html = '&nbsp;';
                  }
                  else {
                      item_html = escape_html(item);
                  }
                  var icon = document.createElement('i');
                  var button = document.createElement('button');
                  if (icons[index]) {
                      icon.className = 'fa fa-' + icons[index];
                  }
                  button.setAttribute('type', 'button');
                  button.className = 'widget-toggle-button jupyter-button';
                  if (previous_bstyle) {
                      button.classList.add(previous_bstyle);
                  }
                  button.innerHTML = item_html;
                  button.setAttribute('data-value', encodeURIComponent(item));
                  button.setAttribute('value', index.toString());
                  button.appendChild(icon);
                  button.disabled = disabled;
                  if (tooltips[index]) {
                      button.setAttribute('title', tooltips[index]);
                  }
                  view.update_style_traits(button);
                  view.buttongroup.appendChild(button);
              });
          }
          // Select active button.
          items.forEach(function (item, index) {
              var item_query = '[data-value="' + encodeURIComponent(item) + '"]';
              var button = view.buttongroup.querySelector(item_query);
              if (view.model.get('index') === index) {
                  button.classList.add('mod-active');
              }
              else {
                  button.classList.remove('mod-active');
              }
          });
          this.stylePromise.then(function (style) {
              if (style) {
                  style.style();
              }
          });
          return _super.prototype.update.call(this, options);
      };
      ToggleButtonsView.prototype.update_style_traits = function (button) {
          for (var name_1 in this._css_state) {
              if (this._css_state.hasOwnProperty(name_1)) {
                  if (name_1 === 'margin') {
                      this.buttongroup.style[name_1] = this._css_state[name_1];
                  }
                  else if (name_1 !== 'width') {
                      if (button) {
                          button.style[name_1] = this._css_state[name_1];
                      }
                      else {
                          var buttons = this.buttongroup
                              .querySelectorAll('button');
                          if (buttons.length) {
                              (buttons[0]).style[name_1] = this._css_state[name_1];
                          }
                      }
                  }
              }
          }
      };
      ToggleButtonsView.prototype.update_button_style = function () {
          var buttons = this.buttongroup.querySelectorAll('button');
          for (var i = 0; i < buttons.length; i++) {
              this.update_mapped_classes(ToggleButtonsView.classMap, 'button_style', buttons[i]);
          }
      };
      ToggleButtonsView.prototype.set_button_style = function () {
          var buttons = this.buttongroup.querySelectorAll('button');
          for (var i = 0; i < buttons.length; i++) {
              this.set_mapped_classes(ToggleButtonsView.classMap, 'button_style', buttons[i]);
          }
      };
      ToggleButtonsView.prototype.events = function () {
          return {
              'click button': '_handle_click'
          };
      };
      /**
       * Handle when a value is clicked.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      ToggleButtonsView.prototype._handle_click = function (event) {
          var target = event.target;
          this.model.set('index', parseInt(target.value, 10), { updated_view: this });
          this.touch();
          // We also send a clicked event, since the value is only set if it changed.
          // See https://github.com/jupyter-widgets/ipywidgets/issues/763
          this.send({ event: 'click' });
      };
      return ToggleButtonsView;
  }(DescriptionView));
  (function (ToggleButtonsView) {
      ToggleButtonsView.classMap = {
          primary: ['mod-primary'],
          success: ['mod-success'],
          info: ['mod-info'],
          warning: ['mod-warning'],
          danger: ['mod-danger']
      };
  })(ToggleButtonsView || (ToggleButtonsView = {}));
  var SelectionSliderModel = /** @class */ (function (_super) {
      __extends$n(SelectionSliderModel, _super);
      function SelectionSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SelectionSliderModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'SelectionSliderModel', _view_name: 'SelectionSliderView', orientation: 'horizontal', readout: true, continuous_update: true });
      };
      return SelectionSliderModel;
  }(SelectionModel));
  var SelectionSliderView = /** @class */ (function (_super) {
      __extends$n(SelectionSliderView, _super);
      function SelectionSliderView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      SelectionSliderView.prototype.render = function () {
          var _this = this;
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-hslider');
          this.el.classList.add('widget-slider');
          (this.$slider = $$1('<div />'))
              .slider({
              slide: this.handleSliderChange.bind(this),
              stop: this.handleSliderChanged.bind(this)
          })
              .addClass('slider');
          // Put the slider in a container
          this.slider_container = document.createElement('div');
          this.slider_container.classList.add('slider-container');
          this.slider_container.appendChild(this.$slider[0]);
          this.el.appendChild(this.slider_container);
          this.readout = document.createElement('div');
          this.el.appendChild(this.readout);
          this.readout.classList.add('widget-readout');
          this.readout.style.display = 'none';
          this.listenTo(this.model, 'change:slider_color', function (sender, value) {
              _this.$slider.find('a').css('background', value);
          });
          this.$slider.find('a').css('background', this.model.get('slider_color'));
          // Set defaults.
          this.update();
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      SelectionSliderView.prototype.update = function (options) {
          if (options === undefined || options.updated_view !== this) {
              var labels = this.model.get('_options_labels');
              var max = labels.length - 1;
              var min = 0;
              this.$slider.slider('option', 'step', 1);
              this.$slider.slider('option', 'max', max);
              this.$slider.slider('option', 'min', min);
              // WORKAROUND FOR JQUERY SLIDER BUG.
              // The horizontal position of the slider handle
              // depends on the value of the slider at the time
              // of orientation change.  Before applying the new
              // workaround, we set the value to the minimum to
              // make sure that the horizontal placement of the
              // handle in the vertical slider is always
              // consistent.
              var orientation_1 = this.model.get('orientation');
              this.$slider.slider('option', 'value', min);
              this.$slider.slider('option', 'orientation', orientation_1);
              var disabled = this.model.get('disabled');
              this.$slider.slider('option', 'disabled', disabled);
              if (disabled) {
                  this.readout.contentEditable = 'false';
              }
              else {
                  this.readout.contentEditable = 'true';
              }
              // Use the right CSS classes for vertical & horizontal sliders
              if (orientation_1 === 'vertical') {
                  this.el.classList.remove('widget-hslider');
                  this.el.classList.remove('widget-inline-hbox');
                  this.el.classList.add('widget-vslider');
                  this.el.classList.add('widget-inline-vbox');
              }
              else {
                  this.el.classList.remove('widget-vslider');
                  this.el.classList.remove('widget-inline-vbox');
                  this.el.classList.add('widget-hslider');
                  this.el.classList.add('widget-inline-hbox');
              }
              var readout = this.model.get('readout');
              if (readout) {
                  // this.$readout.show();
                  this.readout.style.display = '';
              }
              else {
                  // this.$readout.hide();
                  this.readout.style.display = 'none';
              }
              this.updateSelection();
          }
          return _super.prototype.update.call(this, options);
      };
      SelectionSliderView.prototype.events = function () {
          return {
              'slide': 'handleSliderChange',
              'slidestop': 'handleSliderChanged'
          };
      };
      SelectionSliderView.prototype.updateSelection = function () {
          var index = this.model.get('index');
          this.$slider.slider('option', 'value', index);
          this.updateReadout(index);
      };
      SelectionSliderView.prototype.updateReadout = function (index) {
          var value = this.model.get('_options_labels')[index];
          this.readout.textContent = value;
      };
      /**
       * Called when the slider value is changing.
       */
      SelectionSliderView.prototype.handleSliderChange = function (e, ui) {
          this.updateReadout(ui.value);
          // Only persist the value while sliding if the continuous_update
          // trait is set to true.
          if (this.model.get('continuous_update')) {
              this.handleSliderChanged(e, ui);
          }
      };
      /**
       * Called when the slider value has changed.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      SelectionSliderView.prototype.handleSliderChanged = function (e, ui) {
          this.updateReadout(ui.value);
          this.model.set('index', ui.value, { updated_view: this });
          this.touch();
      };
      return SelectionSliderView;
  }(DescriptionView));
  var MultipleSelectionModel = /** @class */ (function (_super) {
      __extends$n(MultipleSelectionModel, _super);
      function MultipleSelectionModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      MultipleSelectionModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'MultipleSelectionModel' });
      };
      return MultipleSelectionModel;
  }(SelectionModel));
  var SelectMultipleModel = /** @class */ (function (_super) {
      __extends$n(SelectMultipleModel, _super);
      function SelectMultipleModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SelectMultipleModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'SelectMultipleModel', _view_name: 'SelectMultipleView', rows: null });
      };
      return SelectMultipleModel;
  }(MultipleSelectionModel));
  var SelectMultipleView = /** @class */ (function (_super) {
      __extends$n(SelectMultipleView, _super);
      function SelectMultipleView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Public constructor.
       */
      SelectMultipleView.prototype.initialize = function (parameters) {
          _super.prototype.initialize.call(this, parameters);
          this.listbox.multiple = true;
      };
      /**
       * Called when view is rendered.
       */
      SelectMultipleView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('widget-select-multiple');
      };
      SelectMultipleView.prototype.updateSelection = function (options) {
          if (options === void 0) { options = {}; }
          if (options.updated_view === this) {
              return;
          }
          var selected = this.model.get('index') || [];
          var listboxOptions = this.listbox.options;
          // Clear the selection
          this.listbox.selectedIndex = -1;
          // Select the appropriate options
          selected.forEach(function (i) {
              listboxOptions[i].selected = true;
          });
      };
      /**
       * Handle when a new value is selected.
       */
      SelectMultipleView.prototype._handle_change = function () {
          var index = Array.prototype.map
              .call(this.listbox.selectedOptions || [], function (option) {
              return option.index;
          });
          this.model.set('index', index, { updated_view: this });
          this.touch();
      };
      return SelectMultipleView;
  }(SelectView));
  var SelectionRangeSliderModel = /** @class */ (function (_super) {
      __extends$n(SelectionRangeSliderModel, _super);
      function SelectionRangeSliderModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SelectionRangeSliderModel.prototype.defaults = function () {
          return __assign$e(__assign$e({}, _super.prototype.defaults.call(this)), { _model_name: 'SelectionSliderModel', _view_name: 'SelectionSliderView', orientation: 'horizontal', readout: true, continuous_update: true });
      };
      return SelectionRangeSliderModel;
  }(MultipleSelectionModel));
  var SelectionRangeSliderView = /** @class */ (function (_super) {
      __extends$n(SelectionRangeSliderView, _super);
      function SelectionRangeSliderView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      SelectionRangeSliderView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.$slider.slider('option', 'range', true);
      };
      SelectionRangeSliderView.prototype.updateSelection = function () {
          var index = this.model.get('index');
          this.$slider.slider('option', 'values', index.slice());
          this.updateReadout(index);
      };
      SelectionRangeSliderView.prototype.updateReadout = function (index) {
          var labels = this.model.get('_options_labels');
          var minValue = labels[index[0]];
          var maxValue = labels[index[1]];
          this.readout.textContent = minValue + "-" + maxValue;
      };
      /**
       * Called when the slider value is changing.
       */
      SelectionRangeSliderView.prototype.handleSliderChange = function (e, ui) {
          this.updateReadout(ui.values);
          // Only persist the value while sliding if the continuous_update
          // trait is set to true.
          if (this.model.get('continuous_update')) {
              this.handleSliderChanged(e, ui);
          }
      };
      /**
       * Called when the slider value has changed.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      SelectionRangeSliderView.prototype.handleSliderChanged = function (e, ui) {
          // The jqueryui documentation indicates ui.values doesn't exist on the slidestop event,
          // but it appears that it actually does: https://github.com/jquery/jquery-ui/blob/ae31f2b3b478975f70526bdf3299464b9afa8bb1/ui/widgets/slider.js#L313
          this.updateReadout(ui.values);
          this.model.set('index', ui.values.slice(), { updated_view: this });
          this.touch();
      };
      return SelectionRangeSliderView;
  }(SelectionSliderView));

  /* This file has code derived from PhosphorJS. The license for this PhosphorJS code is:

  Copyright (c) 2014-2017, PhosphorJS Contributors
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

  * Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  */
  var __extends$o = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  /**
   * A panel where visible widgets are stacked atop one another.
   *
   * #### Notes
   * This class provides a convenience wrapper around a [[PanelLayout]].
   */
  var EventedPanel = /** @class */ (function (_super) {
      __extends$o(EventedPanel, _super);
      function EventedPanel() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this._widgetRemoved = new Signal(_this);
          return _this;
      }
      Object.defineProperty(EventedPanel.prototype, "widgetRemoved", {
          /**
           * A signal emitted when a widget is removed from the panel.
           */
          get: function () {
              return this._widgetRemoved;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * A message handler invoked on a `'child-removed'` message.
       */
      EventedPanel.prototype.onChildRemoved = function (msg) {
          this._widgetRemoved.emit(msg.child);
      };
      return EventedPanel;
  }(Panel));
  /**
   * A widget which combines a `TabBar` and a `EventedPanel`.
   *
   * #### Notes
   * This is a simple panel which handles the common case of a tab bar
   * placed next to a content area. The selected tab controls the widget
   * which is shown in the content area.
   *
   * For use cases which require more control than is provided by this
   * panel, the `TabBar` widget may be used independently.
   *
   * TODO: Support setting the direction??
   */
  var TabPanel$1 = /** @class */ (function (_super) {
      __extends$o(TabPanel, _super);
      /**
       * Construct a new tab panel.
       *
       * @param options - The options for initializing the tab panel.
       */
      function TabPanel(options) {
          if (options === void 0) { options = {}; }
          var _this = _super.call(this) || this;
          _this._currentChanged = new Signal(_this);
          _this.addClass('p-TabPanel');
          // Create the tab bar and contents panel.
          _this.tabBar = new TabBar(options);
          _this.tabBar.addClass('p-TabPanel-tabBar');
          _this.tabContents = new EventedPanel();
          _this.tabContents.addClass('p-TabPanel-tabContents');
          // Connect the tab bar signal handlers.
          _this.tabBar.tabMoved.connect(_this._onTabMoved, _this);
          _this.tabBar.currentChanged.connect(_this._onCurrentChanged, _this);
          _this.tabBar.tabCloseRequested.connect(_this._onTabCloseRequested, _this);
          _this.tabBar.tabActivateRequested.connect(_this._onTabActivateRequested, _this);
          // Connect the evented panel signal handlers.
          _this.tabContents.widgetRemoved.connect(_this._onWidgetRemoved, _this);
          // Create the layout.
          var layout = new PanelLayout();
          // Add the child widgets to the layout.
          layout.addWidget(_this.tabBar);
          layout.addWidget(_this.tabContents);
          // Install the layout on the tab panel.
          _this.layout = layout;
          return _this;
      }
      Object.defineProperty(TabPanel.prototype, "currentChanged", {
          /**
           * A signal emitted when the current tab is changed.
           *
           * #### Notes
           * This signal is emitted when the currently selected tab is changed
           * either through user or programmatic interaction.
           *
           * Notably, this signal is not emitted when the index of the current
           * tab changes due to tabs being inserted, removed, or moved. It is
           * only emitted when the actual current tab node is changed.
           */
          get: function () {
              return this._currentChanged;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "currentIndex", {
          /**
           * Get the index of the currently selected tab.
           *
           * #### Notes
           * This will be `null` if no tab is selected.
           */
          get: function () {
              var currentIndex = this.tabBar.currentIndex;
              // Phosphor tab bars have an index of -1 if no tab is selected
              return (currentIndex === -1 ? null : currentIndex);
          },
          /**
           * Set the index of the currently selected tab.
           *
           * #### Notes
           * If the index is out of range, it will be set to `null`.
           */
          set: function (value) {
              this.tabBar.currentIndex = (value === null ? -1 : value);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "currentWidget", {
          /**
           * Get the currently selected widget.
           *
           * #### Notes
           * This will be `null` if there is no selected tab.
           */
          get: function () {
              var title = this.tabBar.currentTitle;
              return title ? title.owner : null;
          },
          /**
           * Set the currently selected widget.
           *
           * #### Notes
           * If the widget is not in the panel, it will be set to `null`.
           */
          set: function (value) {
              this.tabBar.currentTitle = value ? value.title : null;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "tabsMovable", {
          /**
           * Get the whether the tabs are movable by the user.
           *
           * #### Notes
           * Tabs can always be moved programmatically.
           */
          get: function () {
              return this.tabBar.tabsMovable;
          },
          /**
           * Set the whether the tabs are movable by the user.
           *
           * #### Notes
           * Tabs can always be moved programmatically.
           */
          set: function (value) {
              this.tabBar.tabsMovable = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(TabPanel.prototype, "widgets", {
          /**
           * A read-only array of the widgets in the panel.
           */
          get: function () {
              return this.tabContents.widgets;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Add a widget to the end of the tab panel.
       *
       * @param widget - The widget to add to the tab panel.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       *
       * The widget's `title` is used to populate the tab.
       */
      TabPanel.prototype.addWidget = function (widget) {
          this.insertWidget(this.widgets.length, widget);
      };
      /**
       * Insert a widget into the tab panel at a specified index.
       *
       * @param index - The index at which to insert the widget.
       *
       * @param widget - The widget to insert into to the tab panel.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       *
       * The widget's `title` is used to populate the tab.
       */
      TabPanel.prototype.insertWidget = function (index, widget) {
          if (widget !== this.currentWidget) {
              widget.hide();
          }
          this.tabContents.insertWidget(index, widget);
          this.tabBar.insertTab(index, widget.title);
      };
      /**
       * Handle the `currentChanged` signal from the tab bar.
       */
      TabPanel.prototype._onCurrentChanged = function (sender, args) {
          // Extract the previous and current title from the args.
          var previousIndex = args.previousIndex, previousTitle = args.previousTitle, currentIndex = args.currentIndex, currentTitle = args.currentTitle;
          // Extract the widgets from the titles.
          var previousWidget = previousTitle ? previousTitle.owner : null;
          var currentWidget = currentTitle ? currentTitle.owner : null;
          // Hide the previous widget.
          if (previousWidget) {
              previousWidget.hide();
          }
          // Show the current widget.
          if (currentWidget) {
              currentWidget.show();
          }
          // Emit the `currentChanged` signal for the tab panel.
          this._currentChanged.emit({
              previousIndex: previousIndex, previousWidget: previousWidget, currentIndex: currentIndex, currentWidget: currentWidget
          });
          // Flush the message loop on IE and Edge to prevent flicker.
          if (Platform.IS_EDGE || Platform.IS_IE) {
              MessageLoop.flush();
          }
      };
      /**
       * Handle the `tabActivateRequested` signal from the tab bar.
       */
      TabPanel.prototype._onTabActivateRequested = function (sender, args) {
          args.title.owner.activate();
      };
      /**
       * Handle the `tabCloseRequested` signal from the tab bar.
       */
      TabPanel.prototype._onTabCloseRequested = function (sender, args) {
          args.title.owner.close();
      };
      /**
       * Handle the `tabMoved` signal from the tab bar.
       */
      TabPanel.prototype._onTabMoved = function (sender, args) {
          this.tabContents.insertWidget(args.toIndex, args.title.owner);
      };
      /**
       * Handle the `widgetRemoved` signal from the stacked panel.
       */
      TabPanel.prototype._onWidgetRemoved = function (sender, widget) {
          this.tabBar.removeTab(widget.title);
      };
      return TabPanel;
  }(Widget));

  // Copyright (c) Jupyter Development Team.
  var Selection = /** @class */ (function () {
      function Selection(sequence, options) {
          if (options === void 0) { options = {}; }
          this._array = null;
          this._value = null;
          this._previousValue = null;
          this._selectionChanged = new Signal(this);
          this._array = sequence;
          this._insertBehavior = options.insertBehavior || 'select-item-if-needed';
          this._removeBehavior = options.removeBehavior || 'select-item-after';
      }
      Object.defineProperty(Selection.prototype, "selectionChanged", {
          /**
           * A signal emitted when the current item is changed.
           *
           * #### Notes
           * This signal is emitted when the currently selected item is changed either
           * through user or programmatic interaction.
           *
           * Notably, this signal is not emitted when the index of the current item
           * changes due to other items being inserted, removed, or moved, but the
           * current item remains the same. It is only emitted when the actual current
           * item is changed.
           */
          get: function () {
              return this._selectionChanged;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Adjust for setting an item.
       *
       * This should be called *after* the set.
       *
       * @param index - The index set.
       * @param oldValue - The old value at the index.
       */
      Selection.prototype.adjustSelectionForSet = function (index) {
          // We just need to send a signal if the currentValue changed.
          // Get the current index and value.
          var pi = this.index;
          var pv = this.value;
          // Exit early if this doesn't affect the selection
          if (index !== pi) {
              return;
          }
          this._updateSelectedValue();
          var cv = this.value;
          // The previous item is now null, since it is no longer in the array.
          this._previousValue = null;
          // Send signal if there was a change
          if (pv !== cv) {
              // Emit the current changed signal.
              this._selectionChanged.emit({
                  previousIndex: pi, previousValue: pv,
                  currentIndex: pi, currentValue: cv
              });
          }
      };
      Object.defineProperty(Selection.prototype, "value", {
          /**
           * Get the currently selected item.
           *
           * #### Notes
           * This will be `null` if no item is selected.
           */
          get: function () {
              return this._value;
          },
          /**
           * Set the currently selected item.
           *
           * #### Notes
           * If the item does not exist in the vector, the currentValue will be set to
           * `null`. This selects the first entry equal to the desired item.
           */
          set: function (value) {
              if (value === null) {
                  this.index = null;
              }
              else {
                  this.index = ArrayExt.firstIndexOf(this._array, value);
              }
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Selection.prototype, "index", {
          /**
           * Get the index of the currently selected item.
           *
           * #### Notes
           * This will be `null` if no item is selected.
           */
          get: function () {
              return this._index;
          },
          /**
           * Set the index of the currently selected tab.
           *
           * @param index - The index to select.
           *
           * #### Notes
           * If the value is out of range, the index will be set to `null`, which
           * indicates no item is selected.
           */
          set: function (index) {
              // Coerce the value to an index.
              var i;
              if (index !== null) {
                  i = Math.floor(index);
                  if (i < 0 || i >= this._array.length) {
                      i = null;
                  }
              }
              else {
                  i = null;
              }
              // Bail early if the index will not change.
              if (this._index === i) {
                  return;
              }
              // Look up the previous index and item.
              var pi = this._index;
              var pv = this._value;
              // Update the state
              this._index = i;
              this._updateSelectedValue();
              this._previousValue = pv;
              // Emit the current changed signal.
              this._selectionChanged.emit({
                  previousIndex: pi, previousValue: pv,
                  currentIndex: i, currentValue: this._value
              });
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Selection.prototype, "insertBehavior", {
          /**
           * Get the selection behavior when inserting a tab.
           */
          get: function () {
              return this._insertBehavior;
          },
          /**
           * Set the selection behavior when inserting a tab.
           */
          set: function (value) {
              this._insertBehavior = value;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Selection.prototype, "removeBehavior", {
          /**
           * Get the selection behavior when removing a tab.
           */
          get: function () {
              return this._removeBehavior;
          },
          /**
           * Set the selection behavior when removing a tab.
           */
          set: function (value) {
              this._removeBehavior = value;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * Adjust the current index for a tab insert operation.
       *
       * @param i - The new index of the inserted item.
       * @param j - The inserted item.
       *
       * #### Notes
       * This method accounts for the tab bar's insertion behavior when adjusting
       * the current index and emitting the changed signal. This should be called
       * after the insertion.
       */
      Selection.prototype.adjustSelectionForInsert = function (i, item) {
          // Lookup commonly used variables.
          var cv = this._value;
          var ci = this._index;
          var bh = this._insertBehavior;
          // Handle the behavior where the new item is always selected,
          // or the behavior where the new item is selected if needed.
          if (bh === 'select-item' || (bh === 'select-item-if-needed' && ci === null)) {
              this._index = i;
              this._value = item;
              this._previousValue = cv;
              this._selectionChanged.emit({
                  previousIndex: ci, previousValue: cv,
                  currentIndex: i, currentValue: item
              });
              return;
          }
          // Otherwise, silently adjust the current index if needed.
          if (ci >= i) {
              this._index++;
          }
      };
      /**
       * Adjust the current index for move operation.
       *
       * @param i - The previous index of the item.
       * @param j - The new index of the item.
       *
       * #### Notes
       * This method will not cause the actual current item to change. It silently
       * adjusts the current index to account for the given move.
       */
      Selection.prototype.adjustSelectionForMove = function (i, j) {
          if (this._index === i) {
              this._index = j;
          }
          else if (this._index < i && this._index >= j) {
              this._index++;
          }
          else if (this._index > i && this._index <= j) {
              this._index--;
          }
      };
      /**
       * Clear the selection and history.
       */
      Selection.prototype.clearSelection = function () {
          // Get the current index and item.
          var pi = this._index;
          var pv = this._value;
          // Reset the current index and previous item.
          this._index = null;
          this._value = null;
          this._previousValue = null;
          // If no item was selected, there's nothing else to do.
          if (pi === null) {
              return;
          }
          // Emit the current changed signal.
          this._selectionChanged.emit({
              previousIndex: pi, previousValue: pv,
              currentIndex: this._index, currentValue: this._value
          });
      };
      /**
       * Adjust the current index for an item remove operation.
       *
       * @param i - The former index of the removed item.
       * @param item - The removed item.
       *
       * #### Notes
       * This method accounts for the remove behavior when adjusting the current
       * index and emitting the changed signal. It should be called after the item
       * is removed.
       */
      Selection.prototype.adjustSelectionForRemove = function (i, item) {
          // Lookup commonly used variables.
          var ci = this._index;
          var bh = this._removeBehavior;
          // Silently adjust the index if the current item is not removed.
          if (ci !== i) {
              if (ci > i) {
                  this._index--;
              }
              return;
          }
          // No item gets selected if the vector is empty.
          if (this._array.length === 0) {
              // Reset the current index and previous item.
              this._index = null;
              this._value = null;
              this._previousValue = null;
              this._selectionChanged.emit({
                  previousIndex: i, previousValue: item,
                  currentIndex: this._index, currentValue: this._value
              });
              return;
          }
          // Handle behavior where the next sibling item is selected.
          if (bh === 'select-item-after') {
              this._index = Math.min(i, this._array.length - 1);
              this._updateSelectedValue();
              this._previousValue = null;
              this._selectionChanged.emit({
                  previousIndex: i, previousValue: item,
                  currentIndex: this._index, currentValue: this._value
              });
              return;
          }
          // Handle behavior where the previous sibling item is selected.
          if (bh === 'select-item-before') {
              this._index = Math.max(0, i - 1);
              this._updateSelectedValue();
              this._previousValue = null;
              this._selectionChanged.emit({
                  previousIndex: i, previousValue: item,
                  currentIndex: this._index, currentValue: this._value
              });
              return;
          }
          // Handle behavior where the previous history item is selected.
          if (bh === 'select-previous-item') {
              if (this._previousValue) {
                  this.value = this._previousValue;
              }
              else {
                  this._index = Math.min(i, this._array.length - 1);
                  this._updateSelectedValue();
              }
              this._previousValue = null;
              this._selectionChanged.emit({
                  previousIndex: i, previousValue: item,
                  currentIndex: this._index, currentValue: this.value
              });
              return;
          }
          // Otherwise, no item gets selected.
          this._index = null;
          this._value = null;
          this._previousValue = null;
          this._selectionChanged.emit({
              previousIndex: i, previousValue: item,
              currentIndex: this._index, currentValue: this._value
          });
      };
      /**
       * Set the current value based on the current index.
       */
      Selection.prototype._updateSelectedValue = function () {
          var i = this._index;
          this._value = i !== null ? this._array[i] : null;
      };
      return Selection;
  }());

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$p = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  /**
   * The class name added to Collapse instances.
   */
  var COLLAPSE_CLASS = 'p-Collapse';
  /**
   * The class name added to a Collapse's header.
   */
  var COLLAPSE_HEADER_CLASS = 'p-Collapse-header';
  /**
   * The class name added to a Collapse's contents.
   */
  var COLLAPSE_CONTENTS_CLASS = 'p-Collapse-contents';
  /**
   * The class name added to a Collapse when it is opened
   */
  var COLLAPSE_CLASS_OPEN = 'p-Collapse-open';
  /**
   * A panel that supports a collapsible header, made from the widget's title.
   * Clicking on the title expands or contracts the widget.
   */
  var Collapse = /** @class */ (function (_super) {
      __extends$p(Collapse, _super);
      function Collapse(options) {
          var _this = _super.call(this, options) || this;
          _this._collapseChanged = new Signal(_this);
          _this.addClass(COLLAPSE_CLASS);
          _this._header = new Widget();
          _this._header.addClass(COLLAPSE_HEADER_CLASS);
          _this._header.node.addEventListener('click', _this);
          // Fontawesome icon for caret
          var icon = document.createElement('i');
          icon.classList.add('fa', 'fa-fw', 'fa-caret-right');
          _this._header.node.appendChild(icon);
          // Label content
          _this._header.node.appendChild(document.createElement('span'));
          _this._content = new Panel();
          _this._content.addClass(COLLAPSE_CONTENTS_CLASS);
          var layout = new PanelLayout();
          _this.layout = layout;
          layout.addWidget(_this._header);
          layout.addWidget(_this._content);
          if (options.widget) {
              _this.widget = options.widget;
          }
          _this.collapsed = false;
          return _this;
      }
      Collapse.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          this._header = null;
          this._widget = null;
          this._content = null;
      };
      Object.defineProperty(Collapse.prototype, "widget", {
          get: function () {
              return this._widget;
          },
          set: function (widget) {
              var oldWidget = this._widget;
              if (oldWidget) {
                  oldWidget.disposed.disconnect(this._onChildDisposed, this);
                  oldWidget.title.changed.disconnect(this._onTitleChanged, this);
                  oldWidget.parent = null;
              }
              this._widget = widget;
              widget.disposed.connect(this._onChildDisposed, this);
              widget.title.changed.connect(this._onTitleChanged, this);
              this._onTitleChanged(widget.title);
              this._content.addWidget(widget);
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Collapse.prototype, "collapsed", {
          get: function () {
              return this._collapsed;
          },
          set: function (value) {
              // TODO: should we have this check here?
              if (value === this._collapsed) {
                  return;
              }
              if (value) {
                  this._collapse();
              }
              else {
                  this._uncollapse();
              }
          },
          enumerable: true,
          configurable: true
      });
      Collapse.prototype.toggle = function () {
          this.collapsed = !this.collapsed;
      };
      Object.defineProperty(Collapse.prototype, "collapseChanged", {
          get: function () {
              return this._collapseChanged;
          },
          enumerable: true,
          configurable: true
      });
      Collapse.prototype._collapse = function () {
          this._collapsed = true;
          if (this._content) {
              this._content.hide();
          }
          this.removeClass(COLLAPSE_CLASS_OPEN);
          this._header.node.children[0].classList.add('fa-caret-right');
          this._header.node.children[0].classList.remove('fa-caret-down');
          this._collapseChanged.emit(void 0);
      };
      Collapse.prototype._uncollapse = function () {
          this._collapsed = false;
          if (this._content) {
              this._content.show();
          }
          this.addClass(COLLAPSE_CLASS_OPEN);
          this._header.node.children[0].classList.add('fa-caret-down');
          this._header.node.children[0].classList.remove('fa-caret-right');
          this._collapseChanged.emit(void 0);
      };
      /**
       * Handle the DOM events for the Collapse widget.
       *
       * @param event - The DOM event sent to the panel.
       *
       * #### Notes
       * This method implements the DOM `EventListener` interface and is
       * called in response to events on the panel's DOM node. It should
       * not be called directly by user code.
       */
      Collapse.prototype.handleEvent = function (event) {
          switch (event.type) {
              case 'click':
                  this._evtClick(event);
                  break;
              default:
                  break;
          }
      };
      Collapse.prototype._evtClick = function (event) {
          this.toggle();
      };
      /**
       * Handle the `changed` signal of a title object.
       */
      Collapse.prototype._onTitleChanged = function (sender) {
          this._header.node.children[1].textContent = this._widget.title.label;
      };
      Collapse.prototype._onChildDisposed = function (sender) {
          this.dispose();
      };
      return Collapse;
  }(Widget));
  /**
   * The class name added to Accordion instances.
   */
  var ACCORDION_CLASS = 'p-Accordion';
  /**
   * The class name added to an Accordion child.
   */
  var ACCORDION_CHILD_CLASS = 'p-Accordion-child';
  var ACCORDION_CHILD_ACTIVE_CLASS = 'p-Accordion-child-active';
  /**
   * A panel that supports a collapsible header, made from the widget's title.
   * Clicking on the title expands or contracts the widget.
   */
  var Accordion = /** @class */ (function (_super) {
      __extends$p(Accordion, _super);
      function Accordion(options) {
          var _this = _super.call(this, options) || this;
          _this._selection = new Selection(_this.widgets);
          _this._selection.selectionChanged.connect(_this._onSelectionChanged, _this);
          _this.addClass(ACCORDION_CLASS);
          return _this;
      }
      Object.defineProperty(Accordion.prototype, "collapseWidgets", {
          /**
           * A read-only sequence of the widgets in the panel.
           *
           * #### Notes
           * This is a read-only property.
           */
          /*  get widgets(): ISequence<Widget> {
              return new ArraySequence(toArray(map((this.layout as PanelLayout).widgets, (w: Collapse) => w.widget)));
            }
          */
          get: function () {
              return this.layout.widgets;
          },
          enumerable: true,
          configurable: true
      });
      Object.defineProperty(Accordion.prototype, "selection", {
          get: function () {
              return this._selection;
          },
          enumerable: true,
          configurable: true
      });
      Accordion.prototype.indexOf = function (widget) {
          return ArrayExt.findFirstIndex(this.collapseWidgets, function (w) { return w.widget === widget; });
      };
      /**
       * Add a widget to the end of the accordion.
       *
       * @param widget - The widget to add to the accordion.
       *
       * @returns The Collapse widget wrapping the added widget.
       *
       * #### Notes
       * The widget will be wrapped in a CollapsedWidget.
       */
      Accordion.prototype.addWidget = function (widget) {
          var collapse = this._wrapWidget(widget);
          collapse.collapsed = true;
          _super.prototype.addWidget.call(this, collapse);
          this._selection.adjustSelectionForInsert(this.widgets.length - 1, collapse);
          return collapse;
      };
      /**
       * Insert a widget at the specified index.
       *
       * @param index - The index at which to insert the widget.
       *
       * @param widget - The widget to insert into to the accordion.
       *
       * #### Notes
       * If the widget is already contained in the panel, it will be moved.
       */
      Accordion.prototype.insertWidget = function (index, widget) {
          var collapse = this._wrapWidget(widget);
          collapse.collapsed = true;
          _super.prototype.insertWidget.call(this, index, collapse);
          this._selection.adjustSelectionForInsert(index, collapse);
      };
      Accordion.prototype.removeWidget = function (widget) {
          var index = this.indexOf(widget);
          if (index >= 0) {
              var collapse = this.collapseWidgets[index];
              widget.parent = null;
              collapse.dispose();
              this._selection.adjustSelectionForRemove(index, null);
          }
      };
      Accordion.prototype._wrapWidget = function (widget) {
          var collapse = new Collapse({ widget: widget });
          collapse.addClass(ACCORDION_CHILD_CLASS);
          collapse.collapseChanged.connect(this._onCollapseChange, this);
          return collapse;
      };
      Accordion.prototype._onCollapseChange = function (sender) {
          if (!sender.collapsed) {
              this._selection.value = sender;
          }
          else if (this._selection.value === sender && sender.collapsed) {
              this._selection.value = null;
          }
      };
      Accordion.prototype._onSelectionChanged = function (sender, change) {
          // Collapse previous widget, open current widget
          var pv = change.previousValue;
          var cv = change.currentValue;
          if (pv) {
              pv.collapsed = true;
              pv.removeClass(ACCORDION_CHILD_ACTIVE_CLASS);
          }
          if (cv) {
              cv.collapsed = false;
              cv.addClass(ACCORDION_CHILD_ACTIVE_CLASS);
          }
      };
      return Accordion;
  }(Panel));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$q = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var SelectionContainerModel = /** @class */ (function (_super) {
      __extends$q(SelectionContainerModel, _super);
      function SelectionContainerModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      SelectionContainerModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'SelectionContainerModel',
              selected_index: 0,
              _titles: {}
          });
      };
      return SelectionContainerModel;
  }(BoxModel));
  var AccordionModel = /** @class */ (function (_super) {
      __extends$q(AccordionModel, _super);
      function AccordionModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      AccordionModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'AccordionModel',
              _view_name: 'AccordionView'
          });
      };
      return AccordionModel;
  }(SelectionContainerModel));
  // We implement our own tab widget since Phoshpor's TabPanel uses an absolute
  // positioning BoxLayout, but we want a more an html/css-based Panel layout.
  var JupyterPhosphorAccordionWidget = /** @class */ (function (_super) {
      __extends$q(JupyterPhosphorAccordionWidget, _super);
      function JupyterPhosphorAccordionWidget(options) {
          var _this = this;
          var view = options.view;
          delete options.view;
          _this = _super.call(this, options) || this;
          _this._view = view;
          return _this;
      }
      /**
       * Process the phosphor message.
       *
       * Any custom phosphor widget used inside a Jupyter widget should override
       * the processMessage function like this.
       */
      JupyterPhosphorAccordionWidget.prototype.processMessage = function (msg) {
          _super.prototype.processMessage.call(this, msg);
          this._view.processPhosphorMessage(msg);
      };
      /**
       * Dispose the widget.
       *
       * This causes the view to be destroyed as well with 'remove'
       */
      JupyterPhosphorAccordionWidget.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          if (this._view) {
              this._view.remove();
          }
          this._view = null;
      };
      return JupyterPhosphorAccordionWidget;
  }(Accordion));
  var AccordionView = /** @class */ (function (_super) {
      __extends$q(AccordionView, _super);
      function AccordionView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      AccordionView.prototype._createElement = function (tagName) {
          this.pWidget = new JupyterPhosphorAccordionWidget({ view: this });
          return this.pWidget.node;
      };
      AccordionView.prototype._setElement = function (el) {
          if (this.el || el !== this.pWidget.node) {
              // Accordions don't allow setting the element beyond the initial creation.
              throw new Error('Cannot reset the DOM element.');
          }
          this.el = this.pWidget.node;
          this.$el = $$1(this.pWidget.node);
      };
      AccordionView.prototype.initialize = function (parameters) {
          var _this = this;
          _super.prototype.initialize.call(this, parameters);
          this.children_views = new ViewList(this.add_child_view, this.remove_child_view, this);
          this.listenTo(this.model, 'change:children', function () { return _this.updateChildren(); });
          this.listenTo(this.model, 'change:selected_index', function () { return _this.update_selected_index(); });
          this.listenTo(this.model, 'change:_titles', function () { return _this.update_titles(); });
      };
      /**
       * Called when view is rendered.
       */
      AccordionView.prototype.render = function () {
          var _this = this;
          _super.prototype.render.call(this);
          var accordion = this.pWidget;
          accordion.addClass('jupyter-widgets');
          accordion.addClass('widget-accordion');
          accordion.addClass('widget-container');
          accordion.selection.selectionChanged.connect(function (sender) {
              if (!_this.updatingChildren) {
                  _this.model.set('selected_index', accordion.selection.index);
                  _this.touch();
              }
          });
          this.children_views.update(this.model.get('children'));
          this.update_titles();
          this.update_selected_index();
      };
      /**
       * Update children
       */
      AccordionView.prototype.updateChildren = function () {
          // While we are updating, the index may not be valid, so deselect the
          // tabs before updating so we don't get spurious changes in the index,
          // which would then set off another sync cycle.
          this.updatingChildren = true;
          this.pWidget.selection.index = null;
          this.children_views.update(this.model.get('children'));
          this.update_selected_index();
          this.updatingChildren = false;
      };
      /**
       * Set header titles
       */
      AccordionView.prototype.update_titles = function () {
          var collapsed = this.pWidget.collapseWidgets;
          var titles = this.model.get('_titles');
          for (var i = 0; i < collapsed.length; i++) {
              if (titles[i] !== void 0) {
                  collapsed[i].widget.title.label = titles[i];
              }
          }
      };
      /**
       * Make the rendering and selected index consistent.
       */
      AccordionView.prototype.update_selected_index = function () {
          this.pWidget.selection.index = this.model.get('selected_index');
      };
      /**
       * Called when a child is removed from children list.
       */
      AccordionView.prototype.remove_child_view = function (view) {
          this.pWidget.removeWidget(view.pWidget);
          view.remove();
      };
      /**
       * Called when a child is added to children list.
       */
      AccordionView.prototype.add_child_view = function (model, index) {
          // Placeholder widget to keep our position in the tab panel while we create the view.
          var accordion = this.pWidget;
          var placeholder = new Widget();
          placeholder.title.label = this.model.get('_titles')[index] || '';
          accordion.addWidget(placeholder);
          return this.create_child_view(model).then(function (view) {
              var widget = view.pWidget;
              widget.title.label = placeholder.title.label;
              var collapse = accordion.collapseWidgets[accordion.indexOf(placeholder)];
              collapse.widget = widget;
              placeholder.dispose();
              return view;
          }).catch(reject$2('Could not add child view to box', true));
      };
      AccordionView.prototype.remove = function () {
          this.children_views = null;
          _super.prototype.remove.call(this);
      };
      return AccordionView;
  }(DOMWidgetView));
  var TabModel = /** @class */ (function (_super) {
      __extends$q(TabModel, _super);
      function TabModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      TabModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'TabModel',
              _view_name: 'TabView'
          });
      };
      return TabModel;
  }(SelectionContainerModel));
  // We implement our own tab widget since Phoshpor's TabPanel uses an absolute
  // positioning BoxLayout, but we want a more an html/css-based Panel layout.
  var JupyterPhosphorTabPanelWidget = /** @class */ (function (_super) {
      __extends$q(JupyterPhosphorTabPanelWidget, _super);
      function JupyterPhosphorTabPanelWidget(options) {
          var _this = this;
          var view = options.view;
          delete options.view;
          _this = _super.call(this, options) || this;
          _this._view = view;
          // We want the view's messages to be the messages the tabContents panel
          // gets.
          MessageLoop.installMessageHook(_this.tabContents, function (handler, msg) {
              // There may be times when we want the view's handler to be called
              // *after* the message has been processed by the widget, in which
              // case we'll need to revisit using a message hook.
              _this._view.processPhosphorMessage(msg);
              return true;
          });
          return _this;
      }
      /**
       * Dispose the widget.
       *
       * This causes the view to be destroyed as well with 'remove'
       */
      JupyterPhosphorTabPanelWidget.prototype.dispose = function () {
          if (this.isDisposed) {
              return;
          }
          _super.prototype.dispose.call(this);
          if (this._view) {
              this._view.remove();
          }
          this._view = null;
      };
      return JupyterPhosphorTabPanelWidget;
  }(TabPanel$1));
  var TabView = /** @class */ (function (_super) {
      __extends$q(TabView, _super);
      function TabView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this.updatingTabs = false;
          return _this;
      }
      TabView.prototype._createElement = function (tagName) {
          this.pWidget = new JupyterPhosphorTabPanelWidget({
              view: this,
          });
          return this.pWidget.node;
      };
      TabView.prototype._setElement = function (el) {
          if (this.el || el !== this.pWidget.node) {
              // TabViews don't allow setting the element beyond the initial creation.
              throw new Error('Cannot reset the DOM element.');
          }
          this.el = this.pWidget.node;
          this.$el = $$1(this.pWidget.node);
      };
      /**
       * Public constructor.
       */
      TabView.prototype.initialize = function (parameters) {
          var _this = this;
          _super.prototype.initialize.call(this, parameters);
          this.childrenViews = new ViewList(this.addChildView, function (view) { view.remove(); }, this);
          this.listenTo(this.model, 'change:children', function () { return _this.updateTabs(); });
          this.listenTo(this.model, 'change:_titles', function () { return _this.updateTitles(); });
      };
      /**
       * Called when view is rendered.
       */
      TabView.prototype.render = function () {
          _super.prototype.render.call(this);
          var tabs = this.pWidget;
          tabs.addClass('jupyter-widgets');
          tabs.addClass('widget-container');
          tabs.addClass('widget-tab');
          tabs.tabsMovable = true;
          tabs.tabBar.insertBehavior = 'none'; // needed for insert behavior, see below.
          tabs.tabBar.currentChanged.connect(this._onTabChanged, this);
          tabs.tabBar.tabMoved.connect(this._onTabMoved, this);
          tabs.tabBar.addClass('widget-tab-bar');
          tabs.tabContents.addClass('widget-tab-contents');
          // TODO: expose this option in python
          tabs.tabBar.tabsMovable = false;
          this.updateTabs();
          this.update();
      };
      /**
       * Render tab views based on the current model's children.
       */
      TabView.prototype.updateTabs = function () {
          // While we are updating, the index may not be valid, so deselect the
          // tabs before updating so we don't get spurious changes in the index,
          // which would then set off another sync cycle.
          this.updatingTabs = true;
          this.pWidget.currentIndex = null;
          this.childrenViews.update(this.model.get('children'));
          this.pWidget.currentIndex = this.model.get('selected_index');
          this.updatingTabs = false;
      };
      /**
       * Called when a child is added to children list.
       */
      TabView.prototype.addChildView = function (model, index) {
          // Placeholder widget to keep our position in the tab panel while we create the view.
          var label = this.model.get('_titles')[index] || '';
          var tabs = this.pWidget;
          var placeholder = new Widget();
          placeholder.title.label = label;
          tabs.addWidget(placeholder);
          return this.create_child_view(model).then(function (view) {
              var widget = view.pWidget;
              widget.title.label = placeholder.title.label;
              widget.title.closable = false;
              var i = ArrayExt.firstIndexOf(tabs.widgets, placeholder);
              // insert after placeholder so that if placholder is selected, the
              // real widget will be selected now (this depends on the tab bar
              // insert behavior)
              tabs.insertWidget(i + 1, widget);
              placeholder.dispose();
              return view;
          }).catch(reject$2('Could not add child view to box', true));
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      TabView.prototype.update = function () {
          // Update the selected index in the overall update method because it
          // should be run after the tabs have been updated. Otherwise the
          // selected index may not be a valid tab in the tab bar.
          this.updateSelectedIndex();
          return _super.prototype.update.call(this);
      };
      /**
       * Updates the tab page titles.
       */
      TabView.prototype.updateTitles = function () {
          var titles = this.model.get('_titles') || {};
          each$1(this.pWidget.widgets, function (widget, i) {
              widget.title.label = titles[i] || '';
          });
      };
      /**
       * Updates the selected index.
       */
      TabView.prototype.updateSelectedIndex = function () {
          this.pWidget.currentIndex = this.model.get('selected_index');
      };
      TabView.prototype.remove = function () {
          this.childrenViews = null;
          _super.prototype.remove.call(this);
      };
      TabView.prototype._onTabChanged = function (sender, args) {
          if (!this.updatingTabs) {
              var i = args.currentIndex;
              this.model.set('selected_index', i === -1 ? null : i);
              this.touch();
          }
      };
      /**
       * Handle the `tabMoved` signal from the tab bar.
       */
      TabView.prototype._onTabMoved = function (sender, args) {
          var children = this.model.get('children').slice();
          ArrayExt.move(children, args.fromIndex, args.toIndex);
          this.model.set('children', children);
          this.touch();
      };
      return TabView;
  }(DOMWidgetView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$r = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$f = (undefined && undefined.__assign) || function () {
      __assign$f = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$f.apply(this, arguments);
  };
  /**
   * Class name for a combobox with an invlid value.
   */
  var INVALID_VALUE_CLASS = 'jpwidgets-invalidComboValue';
  var StringModel = /** @class */ (function (_super) {
      __extends$r(StringModel, _super);
      function StringModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      StringModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              value: '',
              disabled: false,
              placeholder: '\u200b',
              _model_name: 'StringModel'
          });
      };
      return StringModel;
  }(CoreDescriptionModel));
  var HTMLModel = /** @class */ (function (_super) {
      __extends$r(HTMLModel, _super);
      function HTMLModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      HTMLModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'HTMLView',
              _model_name: 'HTMLModel'
          });
      };
      return HTMLModel;
  }(StringModel));
  var HTMLView = /** @class */ (function (_super) {
      __extends$r(HTMLView, _super);
      function HTMLView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      HTMLView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-html');
          this.content = document.createElement('div');
          this.content.classList.add('widget-html-content');
          this.el.appendChild(this.content);
          this.update(); // Set defaults.
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      HTMLView.prototype.update = function () {
          this.content.innerHTML = this.model.get('value');
          return _super.prototype.update.call(this);
      };
      return HTMLView;
  }(DescriptionView));
  var HTMLMathModel = /** @class */ (function (_super) {
      __extends$r(HTMLMathModel, _super);
      function HTMLMathModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      HTMLMathModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'HTMLMathView',
              _model_name: 'HTMLMathModel'
          });
      };
      return HTMLMathModel;
  }(StringModel));
  var HTMLMathView = /** @class */ (function (_super) {
      __extends$r(HTMLMathView, _super);
      function HTMLMathView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      HTMLMathView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-htmlmath');
          this.content = document.createElement('div');
          this.content.classList.add('widget-htmlmath-content');
          this.el.appendChild(this.content);
          this.update(); // Set defaults.
      };
      /**
       * Update the contents of this view
       */
      HTMLMathView.prototype.update = function () {
          this.content.innerHTML = this.model.get('value');
          this.typeset(this.content);
          return _super.prototype.update.call(this);
      };
      return HTMLMathView;
  }(DescriptionView));
  var LabelModel = /** @class */ (function (_super) {
      __extends$r(LabelModel, _super);
      function LabelModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      LabelModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'LabelView',
              _model_name: 'LabelModel'
          });
      };
      return LabelModel;
  }(StringModel));
  var LabelView = /** @class */ (function (_super) {
      __extends$r(LabelView, _super);
      function LabelView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      LabelView.prototype.render = function () {
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-label');
          this.update(); // Set defaults.
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      LabelView.prototype.update = function () {
          this.typeset(this.el, this.model.get('value'));
          return _super.prototype.update.call(this);
      };
      return LabelView;
  }(DescriptionView));
  var TextareaModel = /** @class */ (function (_super) {
      __extends$r(TextareaModel, _super);
      function TextareaModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      TextareaModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'TextareaView',
              _model_name: 'TextareaModel',
              rows: null,
              continuous_update: true,
          });
      };
      return TextareaModel;
  }(StringModel));
  var TextareaView = /** @class */ (function (_super) {
      __extends$r(TextareaView, _super);
      function TextareaView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      /**
       * Called when view is rendered.
       */
      TextareaView.prototype.render = function () {
          var _this = this;
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-textarea');
          this.textbox = document.createElement('textarea');
          this.textbox.setAttribute('rows', '5');
          this.textbox.id = this.label.htmlFor = uuid();
          this.el.appendChild(this.textbox);
          this.update(); // Set defaults.
          this.listenTo(this.model, 'change:placeholder', function (model, value, options) {
              _this.update_placeholder(value);
          });
          this.update_placeholder();
      };
      TextareaView.prototype.update_placeholder = function (value) {
          value = value || this.model.get('placeholder');
          this.textbox.setAttribute('placeholder', value.toString());
      };
      /**
       * Update the contents of this view
       *
       * Called when the model is changed.  The model may have been
       * changed by another view or by a state update from the back-end.
       */
      TextareaView.prototype.update = function (options) {
          if (options === undefined || options.updated_view != this) {
              this.textbox.value = this.model.get('value');
              var rows = this.model.get('rows');
              if (rows === null) {
                  rows = '';
              }
              this.textbox.setAttribute('rows', rows);
              this.textbox.disabled = this.model.get('disabled');
          }
          return _super.prototype.update.call(this);
      };
      TextareaView.prototype.events = function () {
          return {
              'keydown input': 'handleKeyDown',
              'keypress input': 'handleKeypress',
              'input textarea': 'handleChanging',
              'change textarea': 'handleChanged'
          };
      };
      /**
       * Handle key down
       *
       * Stop propagation so the event isn't sent to the application.
       */
      TextareaView.prototype.handleKeyDown = function (e) {
          e.stopPropagation();
      };
      /**
       * Handles key press
       *
       * Stop propagation so the keypress isn't sent to the application.
       */
      TextareaView.prototype.handleKeypress = function (e) {
          e.stopPropagation();
      };
      /**
       * Triggered on input change
       */
      TextareaView.prototype.handleChanging = function (e) {
          if (this.model.get('continuous_update')) {
              this.handleChanged(e);
          }
      };
      /**
       * Sync the value with the kernel.
       *
       * @param e Event
       */
      TextareaView.prototype.handleChanged = function (e) {
          var target = e.target;
          this.model.set('value', target.value, { updated_view: this });
          this.touch();
      };
      return TextareaView;
  }(DescriptionView));
  var TextModel = /** @class */ (function (_super) {
      __extends$r(TextModel, _super);
      function TextModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      TextModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'TextView',
              _model_name: 'TextModel',
              continuous_update: true,
          });
      };
      return TextModel;
  }(StringModel));
  var TextView = /** @class */ (function (_super) {
      __extends$r(TextView, _super);
      function TextView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this.inputType = 'text';
          return _this;
      }
      /**
       * Called when view is rendered.
       */
      TextView.prototype.render = function () {
          var _this = this;
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-inline-hbox');
          this.el.classList.add('widget-text');
          this.textbox = document.createElement('input');
          this.textbox.setAttribute('type', this.inputType);
          this.textbox.id = this.label.htmlFor = uuid();
          this.el.appendChild(this.textbox);
          this.update(); // Set defaults.
          this.listenTo(this.model, 'change:placeholder', function (model, value, options) {
              _this.update_placeholder(value);
          });
          this.listenTo(this.model, 'change:description_tooltip', this.update_title);
          this.listenTo(this.model, 'change:description', this.update_title);
          this.update_placeholder();
          this.update_title();
      };
      TextView.prototype.update_placeholder = function (value) {
          this.textbox.setAttribute('placeholder', value || this.model.get('placeholder'));
      };
      TextView.prototype.update_title = function () {
          var title = this.model.get('description_tooltip');
          if (!title) {
              this.textbox.removeAttribute('title');
          }
          else if (this.model.get('description').length === 0) {
              this.textbox.setAttribute('title', title);
          }
      };
      TextView.prototype.update = function (options) {
          /**
           * Update the contents of this view
           *
           * Called when the model is changed.  The model may have been
           * changed by another view or by a state update from the back-end.
           */
          if (options === undefined || options.updated_view !== this) {
              if (this.textbox.value !== this.model.get('value')) {
                  this.textbox.value = this.model.get('value');
              }
              this.textbox.disabled = this.model.get('disabled');
          }
          return _super.prototype.update.call(this);
      };
      TextView.prototype.events = function () {
          return {
              'keydown input': 'handleKeyDown',
              'keypress input': 'handleKeypress',
              'input input': 'handleChanging',
              'change input': 'handleChanged'
          };
      };
      /**
       * Handle key down
       *
       * Stop propagation so the keypress isn't sent to the application.
       */
      TextView.prototype.handleKeyDown = function (e) {
          e.stopPropagation();
      };
      /**
       * Handles text submission
       */
      TextView.prototype.handleKeypress = function (e) {
          e.stopPropagation();
          // The submit message is deprecated in widgets 7
          if (e.keyCode === 13) { // Return key
              this.send({ event: 'submit' });
          }
      };
      /**
       * Handles user input.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      TextView.prototype.handleChanging = function (e) {
          if (this.model.get('continuous_update')) {
              this.handleChanged(e);
          }
      };
      /**
       * Handles user input.
       *
       * Calling model.set will trigger all of the other views of the
       * model to update.
       */
      TextView.prototype.handleChanged = function (e) {
          var target = e.target;
          this.model.set('value', target.value, { updated_view: this });
          this.touch();
      };
      return TextView;
  }(DescriptionView));
  var PasswordModel = /** @class */ (function (_super) {
      __extends$r(PasswordModel, _super);
      function PasswordModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      PasswordModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _view_name: 'PasswordView',
              _model_name: 'PasswordModel'
          });
      };
      return PasswordModel;
  }(TextModel));
  var PasswordView = /** @class */ (function (_super) {
      __extends$r(PasswordView, _super);
      function PasswordView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this.inputType = 'password';
          return _this;
      }
      return PasswordView;
  }(TextView));
  /**
   * Combobox widget model class.
   */
  var ComboboxModel = /** @class */ (function (_super) {
      __extends$r(ComboboxModel, _super);
      function ComboboxModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      ComboboxModel.prototype.defaults = function () {
          return __assign$f(__assign$f({}, _super.prototype.defaults.call(this)), { _model_name: 'ComboboxModel', _view_name: 'ComboboxView', options: [], ensure_options: false });
      };
      return ComboboxModel;
  }(TextModel));
  /**
   * Combobox widget view class.
   */
  var ComboboxView = /** @class */ (function (_super) {
      __extends$r(ComboboxView, _super);
      function ComboboxView() {
          var _this = _super !== null && _super.apply(this, arguments) || this;
          _this.isInitialRender = true;
          return _this;
      }
      ComboboxView.prototype.render = function () {
          this.datalist = document.createElement('datalist');
          this.datalist.id = uuid();
          _super.prototype.render.call(this);
          this.textbox.setAttribute('list', this.datalist.id);
          this.el.appendChild(this.datalist);
      };
      ComboboxView.prototype.update = function (options) {
          _super.prototype.update.call(this, options);
          if (!this.datalist) {
              return;
          }
          var valid = this.isValid(this.model.get('value'));
          this.highlightValidState(valid);
          // Check if we need to update options
          if ((options !== undefined && options.updated_view) || (!this.model.hasChanged('options') &&
              !this.isInitialRender)) {
              // Value update only, keep current options
              return;
          }
          this.isInitialRender = false;
          var opts = this.model.get('options');
          var optLines = opts.map(function (o) {
              return "<option value=\"" + o + "\"></option>";
          });
          this.datalist.innerHTML = optLines.join('\n');
      };
      ComboboxView.prototype.isValid = function (value) {
          if (true === this.model.get('ensure_option')) {
              var options = this.model.get('options');
              if (options.indexOf(value) === -1) {
                  return false;
              }
          }
          return true;
      };
      ComboboxView.prototype.handleChanging = function (e) {
          // Override to validate value
          var target = e.target;
          var valid = this.isValid(target.value);
          this.highlightValidState(valid);
          if (valid) {
              _super.prototype.handleChanging.call(this, e);
          }
      };
      ComboboxView.prototype.handleChanged = function (e) {
          // Override to validate value
          var target = e.target;
          var valid = this.isValid(target.value);
          this.highlightValidState(valid);
          if (valid) {
              _super.prototype.handleChanged.call(this, e);
          }
      };
      ComboboxView.prototype.highlightValidState = function (valid) {
          this.textbox.classList.toggle(INVALID_VALUE_CLASS, !valid);
      };
      return ComboboxView;
  }(TextView));

  // Copyright (c) Jupyter Development Team.
  // Distributed under the terms of the Modified BSD License.
  var __extends$s = (undefined && undefined.__extends) || (function () {
      var extendStatics = function (d, b) {
          extendStatics = Object.setPrototypeOf ||
              ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
              function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
          return extendStatics(d, b);
      };
      return function (d, b) {
          extendStatics(d, b);
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  var __assign$g = (undefined && undefined.__assign) || function () {
      __assign$g = Object.assign || function(t) {
          for (var s, i = 1, n = arguments.length; i < n; i++) {
              s = arguments[i];
              for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                  t[p] = s[p];
          }
          return t;
      };
      return __assign$g.apply(this, arguments);
  };
  var __spreadArrays = (undefined && undefined.__spreadArrays) || function () {
      for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
      for (var r = Array(s), k = 0, i = 0; i < il; i++)
          for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
              r[k] = a[j];
      return r;
  };
  var FileUploadModel = /** @class */ (function (_super) {
      __extends$s(FileUploadModel, _super);
      function FileUploadModel() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      FileUploadModel.prototype.defaults = function () {
          return extend(_super.prototype.defaults.call(this), {
              _model_name: 'FileUploadModel',
              _view_name: 'FileUploadView',
              _counter: 0,
              accept: '',
              description: 'Upload',
              tooltip: '',
              disabled: false,
              icon: 'upload',
              button_style: '',
              multiple: false,
              metadata: [],
              data: [],
              error: '',
              style: null
          });
      };
      FileUploadModel.serializers = __assign$g(__assign$g({}, CoreDOMWidgetModel.serializers), { data: { serialize: function (buffers) { return __spreadArrays(buffers); } } });
      return FileUploadModel;
  }(CoreDOMWidgetModel));
  var FileUploadView = /** @class */ (function (_super) {
      __extends$s(FileUploadView, _super);
      function FileUploadView() {
          return _super !== null && _super.apply(this, arguments) || this;
      }
      Object.defineProperty(FileUploadView.prototype, "tagName", {
          get: function () {
              return 'button';
          },
          enumerable: true,
          configurable: true
      });
      FileUploadView.prototype.render = function () {
          var _this = this;
          _super.prototype.render.call(this);
          this.el.classList.add('jupyter-widgets');
          this.el.classList.add('widget-upload');
          this.el.classList.add('jupyter-button');
          this.fileInput = document.createElement('input');
          this.fileInput.type = 'file';
          this.fileInput.style.display = 'none';
          this.el.appendChild(this.fileInput);
          this.el.addEventListener('click', function () {
              _this.fileInput.click();
          });
          this.fileInput.addEventListener('click', function () {
              _this.fileInput.value = '';
          });
          this.fileInput.addEventListener('change', function () {
              var promisesFile = [];
              Array.from(_this.fileInput.files).forEach(function (file) {
                  promisesFile.push(new Promise(function (resolve, reject) {
                      var metadata = {
                          name: file.name,
                          type: file.type,
                          size: file.size,
                          lastModified: file.lastModified,
                      };
                      _this.fileReader = new FileReader();
                      _this.fileReader.onload = function (event) {
                          var buffer = event.target.result;
                          resolve({
                              buffer: buffer,
                              metadata: metadata,
                              error: '',
                          });
                      };
                      _this.fileReader.onerror = function () {
                          reject();
                      };
                      _this.fileReader.onabort = _this.fileReader.onerror;
                      _this.fileReader.readAsArrayBuffer(file);
                  }));
              });
              Promise.all(promisesFile)
                  .then(function (contents) {
                  var metadata = [];
                  var li_buffer = [];
                  contents.forEach(function (c) {
                      metadata.push(c.metadata);
                      li_buffer.push(c.buffer);
                  });
                  var counter = _this.model.get('_counter');
                  _this.model.set({
                      _counter: counter + contents.length,
                      metadata: metadata,
                      data: li_buffer,
                      error: '',
                  });
                  _this.touch();
              })
                  .catch(function (err) {
                  console.error('error in file upload: %o', err);
                  _this.model.set({
                      error: err,
                  });
                  _this.touch();
              });
          });
          this.listenTo(this.model, 'change:button_style', this.update_button_style);
          this.set_button_style();
          this.update(); // Set defaults.
      };
      FileUploadView.prototype.update = function () {
          this.el.disabled = this.model.get('disabled');
          this.el.setAttribute('title', this.model.get('tooltip'));
          var description = this.model.get('description') + " (" + this.model.get('_counter') + ")";
          var icon = this.model.get('icon');
          if (description.length || icon.length) {
              this.el.textContent = '';
              if (icon.length) {
                  var i = document.createElement('i');
                  i.classList.add('fa');
                  i.classList.add('fa-' + icon);
                  if (description.length === 0) {
                      i.classList.add('center');
                  }
                  this.el.appendChild(i);
              }
              this.el.appendChild(document.createTextNode(description));
          }
          this.fileInput.accept = this.model.get('accept');
          this.fileInput.multiple = this.model.get('multiple');
          return _super.prototype.update.call(this);
      };
      FileUploadView.prototype.update_button_style = function () {
          this.update_mapped_classes(FileUploadView.class_map, 'button_style', this.el);
      };
      FileUploadView.prototype.set_button_style = function () {
          this.set_mapped_classes(FileUploadView.class_map, 'button_style', this.el);
      };
      FileUploadView.class_map = {
          primary: ['mod-primary'],
          success: ['mod-success'],
          info: ['mod-info'],
          warning: ['mod-warning'],
          danger: ['mod-danger']
      };
      return FileUploadView;
  }(DOMWidgetView));

  var version$1 = require$$0.version;

  var lib$3 = /*#__PURE__*/Object.freeze({
    version: version$1,
    reject: reject$2,
    typeset: typeset,
    escape_html: escape_html,
    uuid: uuid,
    WrappedError: WrappedError,
    resolvePromisesDict: resolvePromisesDict,
    JUPYTER_CONTROLS_VERSION: JUPYTER_CONTROLS_VERSION,
    DirectionalLinkModel: DirectionalLinkModel,
    LinkModel: LinkModel,
    BoolModel: BoolModel,
    CheckboxModel: CheckboxModel,
    CheckboxView: CheckboxView,
    ToggleButtonModel: ToggleButtonModel,
    ToggleButtonView: ToggleButtonView,
    ValidModel: ValidModel,
    ValidView: ValidView,
    ButtonStyleModel: ButtonStyleModel,
    ButtonModel: ButtonModel,
    ButtonView: ButtonView,
    BoxModel: BoxModel,
    HBoxModel: HBoxModel,
    VBoxModel: VBoxModel,
    BoxView: BoxView,
    HBoxView: HBoxView,
    VBoxView: VBoxView,
    GridBoxView: GridBoxView,
    GridBoxModel: GridBoxModel,
    ImageModel: ImageModel,
    ImageView: ImageView,
    VideoModel: VideoModel,
    VideoView: VideoView,
    AudioModel: AudioModel,
    AudioView: AudioView,
    ColorPickerModel: ColorPickerModel,
    ColorPickerView: ColorPickerView,
    serialize_date: serialize_date,
    deserialize_date: deserialize_date,
    DatePickerModel: DatePickerModel,
    DatePickerView: DatePickerView,
    IntModel: IntModel,
    BoundedIntModel: BoundedIntModel,
    SliderStyleModel: SliderStyleModel,
    IntSliderModel: IntSliderModel,
    IntRangeSliderModel: IntRangeSliderModel,
    BaseIntSliderView: BaseIntSliderView,
    IntRangeSliderView: IntRangeSliderView,
    IntSliderView: IntSliderView,
    IntTextModel: IntTextModel,
    BoundedIntTextModel: BoundedIntTextModel,
    IntTextView: IntTextView,
    ProgressStyleModel: ProgressStyleModel,
    IntProgressModel: IntProgressModel,
    ProgressView: ProgressView,
    PlayModel: PlayModel,
    PlayView: PlayView,
    FloatModel: FloatModel,
    BoundedFloatModel: BoundedFloatModel,
    FloatSliderModel: FloatSliderModel,
    FloatLogSliderModel: FloatLogSliderModel,
    FloatRangeSliderModel: FloatRangeSliderModel,
    FloatSliderView: FloatSliderView,
    FloatLogSliderView: FloatLogSliderView,
    FloatRangeSliderView: FloatRangeSliderView,
    FloatTextModel: FloatTextModel,
    BoundedFloatTextModel: BoundedFloatTextModel,
    FloatTextView: FloatTextView,
    FloatProgressModel: FloatProgressModel,
    ControllerButtonModel: ControllerButtonModel,
    ControllerButtonView: ControllerButtonView,
    ControllerAxisModel: ControllerAxisModel,
    ControllerAxisView: ControllerAxisView,
    ControllerModel: ControllerModel,
    ControllerView: ControllerView,
    SelectionModel: SelectionModel,
    DropdownModel: DropdownModel,
    DropdownView: DropdownView,
    SelectModel: SelectModel,
    SelectView: SelectView,
    RadioButtonsModel: RadioButtonsModel,
    RadioButtonsView: RadioButtonsView,
    ToggleButtonsStyleModel: ToggleButtonsStyleModel,
    ToggleButtonsModel: ToggleButtonsModel,
    get ToggleButtonsView () { return ToggleButtonsView; },
    SelectionSliderModel: SelectionSliderModel,
    SelectionSliderView: SelectionSliderView,
    MultipleSelectionModel: MultipleSelectionModel,
    SelectMultipleModel: SelectMultipleModel,
    SelectMultipleView: SelectMultipleView,
    SelectionRangeSliderModel: SelectionRangeSliderModel,
    SelectionRangeSliderView: SelectionRangeSliderView,
    SelectionContainerModel: SelectionContainerModel,
    AccordionModel: AccordionModel,
    JupyterPhosphorAccordionWidget: JupyterPhosphorAccordionWidget,
    AccordionView: AccordionView,
    TabModel: TabModel,
    JupyterPhosphorTabPanelWidget: JupyterPhosphorTabPanelWidget,
    TabView: TabView,
    StringModel: StringModel,
    HTMLModel: HTMLModel,
    HTMLView: HTMLView,
    HTMLMathModel: HTMLMathModel,
    HTMLMathView: HTMLMathView,
    LabelModel: LabelModel,
    LabelView: LabelView,
    TextareaModel: TextareaModel,
    TextareaView: TextareaView,
    TextModel: TextModel,
    TextView: TextView,
    PasswordModel: PasswordModel,
    PasswordView: PasswordView,
    ComboboxModel: ComboboxModel,
    ComboboxView: ComboboxView,
    DescriptionStyleModel: DescriptionStyleModel,
    DescriptionModel: DescriptionModel,
    DescriptionView: DescriptionView,
    LabeledDOMWidgetModel: LabeledDOMWidgetModel,
    LabeledDOMWidgetView: LabeledDOMWidgetView,
    FileUploadModel: FileUploadModel,
    FileUploadView: FileUploadView
  });

  const OUTPUT_WIDGET_VERSION = '1.0.0';
  class OutputModel extends DOMWidgetModel {
      defaults() {
          return Object.assign(Object.assign({}, super.defaults()), { _model_name: 'OutputModel', _view_name: 'OutputView', _model_module: '@jupyter-widgets/output', _view_module: '@jupyter-widgets/output', _model_module_version: OUTPUT_WIDGET_VERSION, _view_module_version: OUTPUT_WIDGET_VERSION });
      }
  }
  class OutputView extends DOMWidgetView {
  }



  var lib$4 = /*#__PURE__*/Object.freeze({
    OUTPUT_WIDGET_VERSION: OUTPUT_WIDGET_VERSION,
    OutputModel: OutputModel,
    OutputView: OutputView
  });

  function styleInject(css, ref) {
    if ( ref === void 0 ) ref = {};
    var insertAt = ref.insertAt;

    if (!css || typeof document === 'undefined') { return; }

    var head = document.head || document.getElementsByTagName('head')[0];
    var style = document.createElement('style');
    style.type = 'text/css';

    if (insertAt === 'top') {
      if (head.firstChild) {
        head.insertBefore(style, head.firstChild);
      } else {
        head.appendChild(style);
      }
    } else {
      head.appendChild(style);
    }

    if (style.styleSheet) {
      style.styleSheet.cssText = css;
    } else {
      style.appendChild(document.createTextNode(css));
    }
  }

  var css = ".jupyter-widgets-output-area div.output_subarea{max-width:100%}.jupyter-widgets-output-area>.out_prompt_overlay{display:none}";
  styleInject(css);

  var widget_output = /*#__PURE__*/Object.freeze({
    'default': css
  });

  getCjsExportFromNamespace(widget_output);

  // This widget is strongly coupled to the notebook because of the outputarea
  // dependency.



  var outputArea = new Promise(function(resolve, reject) {
    requirejs(['notebook/js/outputarea'], resolve, reject);
  });

  class OutputModel$1 extends lib$4.OutputModel {
    defaults() {
      return {
        ...super.defaults(),
        msg_id: '',
        outputs: []
      };
    }

    initialize(attributes, options) {
      super.initialize(attributes, options);
      this.listenTo(this, 'change:msg_id', this.reset_msg_id);

      if (this.comm && this.comm.kernel) {
        this.kernel = this.comm.kernel;
        this.kernel.set_callbacks_for_msg(this.model_id, this.callbacks(), false);
      }

      var that = this;
      // Create an output area to handle the data model part
      outputArea.then(function(outputArea) {
        that.output_area = new outputArea.OutputArea({
          selector: document.createElement('div'),
          config: { data: { OutputArea: {} } },
          prompt_area: false,
          events: that.widget_manager.notebook.events,
          keyboard_manager: that.widget_manager.keyboard_manager
        });
        that.listenTo(
          that,
          'new_message',
          function(msg) {
            that.output_area.handle_output(msg);
            that.set('outputs', that.output_area.toJSON(), { newMessage: true });
            that.save_changes();
          },
          that
        );
        that.listenTo(that, 'clear_output', function(msg) {
          that.output_area.handle_clear_output(msg);
          that.set('outputs', that.output_area.toJSON(), { newMessage: true });
          that.save_changes();
        });
        that.listenTo(that, 'change:outputs', that.setOutputs);
        that.setOutputs();
      });
    }

    // make callbacks
    callbacks() {
      // Merge our callbacks with the base class callbacks.
      var cb = super.callbacks();
      var iopub = cb.iopub || {};
      var iopubCallbacks = {
        ...iopub,
        output: function(msg) {
          this.trigger('new_message', msg);
          if (iopub.output) {
            iopub.output.apply(this, arguments);
          }
        }.bind(this),
        clear_output: function(msg) {
          this.trigger('clear_output', msg);
          if (iopub.clear_output) {
            iopub.clear_output.apply(this, arguments);
          }
        }.bind(this)
      };
      return { ...cb, iopub: iopubCallbacks };
    }

    reset_msg_id() {
      var kernel = this.kernel;
      // Pop previous message id
      var prev_msg_id = this.previous('msg_id');
      if (prev_msg_id && kernel) {
        var previous_callback = kernel.output_callback_overrides_pop(prev_msg_id);
        if (previous_callback !== this.model_id) {
          console.error(
            'Popped wrong message (' +
              previous_callback +
              ' instead of ' +
              this.model_id +
              ') - likely the stack was not maintained in kernel.'
          );
        }
      }
      var msg_id = this.get('msg_id');
      if (msg_id && kernel) {
        kernel.output_callback_overrides_push(msg_id, this.model_id);
      }
    }

    setOutputs(model, value, options) {
      if (!(options && options.newMessage)) {
        // fromJSON does not clear the existing output
        this.output_area.clear_output();
        // fromJSON does not copy the message, so we make a deep copy
        this.output_area.fromJSON(
          JSON.parse(JSON.stringify(this.get('outputs')))
        );
      }
    }
  }

  class OutputView$1 extends lib$4.OutputView {
    render() {
      var that = this;
      this.el.classList.add('jupyter-widgets-output-area');
      outputArea.then(function(outputArea) {
        that.output_area = new outputArea.OutputArea({
          selector: that.el,
          // use default values for the output area config
          config: { data: { OutputArea: {} } },
          prompt_area: false,
          events: that.model.widget_manager.notebook.events,
          keyboard_manager: that.model.widget_manager.keyboard_manager
        });
        that.listenTo(
          that.model,
          'new_message',
          function(msg) {
            that.output_area.handle_output(msg);
          },
          that
        );
        that.listenTo(that.model, 'clear_output', function(msg) {
          that.output_area.handle_clear_output(msg);
          // fake the event on the output area element. This can be
          // deleted when we can rely on
          // https://github.com/jupyter/notebook/pull/2411 being
          // available.
          that.output_area.element.trigger('clearing', { output_area: this });
        });
        // Render initial contents from the current model
        that.listenTo(that.model, 'change:outputs', that.setOutputs);
        that.setOutputs();
      });
      super.render();
    }

    setOutputs(model, value, options) {
      if (!(options && options.newMessage)) {
        // fromJSON does not clear the existing output
        this.output_area.clear_output();
        // fromJSON does not copy the message, so we make a deep copy
        this.output_area.fromJSON(
          JSON.parse(JSON.stringify(this.model.get('outputs')))
        );
      }
    }
  }

  var widget_output$1 = /*#__PURE__*/Object.freeze({
    OutputModel: OutputModel$1,
    OutputView: OutputView$1
  });

  // Copyright (c) Jupyter Development Team.

  var save_state = function() {
    return new Promise(function(resolve, reject) {
      requirejs(['base/js/namespace'], function(Jupyter) {
        var wm = Jupyter.WidgetManager._managers[0];
        if (!wm) {
          reject('No widget manager');
        }
        return wm
          .get_state({
            drop_defaults: true
          })
          .then(function(state) {
            var data =
              'text/json;charset=utf-8,' +
              encodeURIComponent(JSON.stringify(state, null, '    '));
            var a = document.createElement('a');
            a.download = 'widget_state.json';
            a.href = 'data:' + data;
            a.click();
            resolve();
          });
      });
    });
  };

  var action = {
    help: 'Download the widget state as a JSON file',
    icon: 'fa-sliders',
    help_index: 'zz',
    handler: save_state
  };

  var action_name = 'save-widget-state';
  var prefix = 'widgets';
  requirejs(['base/js/namespace'], function(Jupyter) {
    Jupyter.notebook.keyboard_manager.actions.register(
      action,
      action_name,
      prefix
    );
  });

  var save_state_1 = { action: action };

  var author$1 = {
  	name: "Jupyter Project"
  };
  var bugs$1 = {
  	url: "https://github.com/jupyter-widgets/ipywidgets/issues"
  };
  var dependencies$1 = {
  	"@fortawesome/fontawesome-free": "^5.12.0",
  	"@jupyter-widgets/base": "^4.0.0-alpha.0",
  	"@jupyter-widgets/base-manager": "^1.0.0-alpha.0",
  	"@jupyter-widgets/controls": "^3.0.0-alpha.0",
  	"@jupyter-widgets/output": "^4.0.0-alpha.0",
  	"@jupyter-widgets/schema": "^0.5.0-alpha.0",
  	"@jupyterlab/outputarea": "^2.0.2",
  	"@jupyterlab/rendermime": "^2.0.2",
  	"@jupyterlab/rendermime-interfaces": "^2.0.1",
  	"@lumino/widgets": "^1.11.1",
  	ajv: "^6.10.0",
  	jquery: "^3.1.1"
  };
  var description$1 = "Standalone package for rendering Jupyter widgets outside notebooks";
  var devDependencies$1 = {
  	"@types/mocha": "^5.2.7",
  	"@types/node": "^12.7.0",
  	chai: "^4.0.0",
  	"css-loader": "^3.4.0",
  	"file-loader": "^5.0.2",
  	karma: "^4.1.0",
  	"karma-chrome-launcher": "^2.2.0",
  	"karma-firefox-launcher": "^1.1.0",
  	"karma-ie-launcher": "^1.0.0",
  	"karma-mocha": "^1.3.0",
  	"karma-mocha-reporter": "^2.2.5",
  	mocha: "^6.1.4",
  	rimraf: "^2.6.1",
  	"style-loader": "^1.1.2",
  	typescript: "~3.7.4",
  	"url-loader": "^3.0.0",
  	webpack: "^4.41.5",
  	"webpack-cli": "^3.3.4"
  };
  var files$1 = [
  	"lib/**/*.d.ts",
  	"lib/**/*.js",
  	"dist/",
  	"css/*.css"
  ];
  var gitHead$1 = "6fbe5cb7ab0cc997de5b468efaa7225cff1f340e";
  var homepage$1 = "https://github.com/jupyter-widgets/ipywidgets#readme";
  var license$1 = "BSD-3-Clause";
  var main$1 = "lib/index.js";
  var name$1 = "@jupyter-widgets/html-manager";
  var repository$1 = {
  	type: "git",
  	url: "git+https://github.com/jupyter-widgets/ipywidgets.git"
  };
  var scripts$1 = {
  	build: "npm run build:src && webpack && npm run build:embed-amd",
  	"build:embed-amd": "node scripts/concat-amd-build.js && rimraf dist/amd",
  	"build:src": "tsc --build",
  	"build:test": "tsc --build test && webpack --config test/webpack.conf.js",
  	clean: "rimraf lib && rimraf tsconfig.tsbuildinfo && rimraf dist",
  	prepublish: "npm run clean && npm run build",
  	test: "npm run test:unit",
  	"test:unit": "npm run test:unit:firefox && npm run test:unit:chrome",
  	"test:unit:chrome": "npm run test:unit:default -- --browsers=Chrome",
  	"test:unit:default": "npm run build:test && karma start test/karma.conf.js --log-level debug --browsers=Firefox",
  	"test:unit:firefox": "npm run test:unit:default -- --browsers=Firefox",
  	"test:unit:firefox:headless": "npm run test:unit:default -- --browsers=FirefoxHeadless"
  };
  var typings$1 = "lib/index.d.ts";
  var version$2 = "0.20.0-alpha.0";
  var _package$2 = {
  	author: author$1,
  	bugs: bugs$1,
  	dependencies: dependencies$1,
  	description: description$1,
  	devDependencies: devDependencies$1,
  	files: files$1,
  	gitHead: gitHead$1,
  	homepage: homepage$1,
  	license: license$1,
  	main: main$1,
  	name: name$1,
  	repository: repository$1,
  	scripts: scripts$1,
  	typings: typings$1,
  	version: version$2
  };

  var _package$3 = /*#__PURE__*/Object.freeze({
    author: author$1,
    bugs: bugs$1,
    dependencies: dependencies$1,
    description: description$1,
    devDependencies: devDependencies$1,
    files: files$1,
    gitHead: gitHead$1,
    homepage: homepage$1,
    license: license$1,
    main: main$1,
    name: name$1,
    repository: repository$1,
    scripts: scripts$1,
    typings: typings$1,
    version: version$2,
    'default': _package$2
  });

  var require$$0$1 = getCjsExportFromNamespace(_package$3);

  var VIEW_MIME_TYPE = 'application/vnd.jupyter.widget-view+json';

  var htmlManagerVersion = require$$0$1
    .version;

  var embed_widgets = function() {
    /**
     * Escape a string that will be the content of an HTML script tag.
     *
     * We replace the opening bracket of <script, </script, and <!-- with the
     * unicode equivalent. This is inspired by the documentation for the script
     * tag at
     * https://html.spec.whatwg.org/multipage/scripting.html#restrictions-for-contents-of-script-elements
     *
     * We only replace these three cases so that most html or other content
     * involving `<` is readable.
     */
    function escapeScript(s) {
      return s.replace(/<(script|\/script|!--)/gi, '\\u003c$1');
    }

    return new Promise(function(resolve, reject) {
      requirejs(
        ['base/js/namespace', 'base/js/dialog', '@jupyter-widgets/controls'],
        function(Jupyter, dialog, widgets) {
          var wm = Jupyter.WidgetManager._managers[0];
          if (!wm) {
            reject('No widget manager');
          }
          wm.get_state({
            drop_defaults: true
          }).then(function(state) {
            var data = escapeScript(JSON.stringify(state, null, '    '));
            var value = [
              '<html><head>',
              '',
              '',
              '<!-- Load require.js. Delete this if your page already loads require.js -->',
              '<script src="https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js" integrity="sha256-Ae2Vz/4ePdIu6ZyI/5ZGsYnb+m0JlOmKPjt6XZ9JJkA=" crossorigin="anonymous"></script>',
              '<script src="https://unpkg.com/@jupyter-widgets/html-manager@*/dist/embed-amd.js" crossorigin="anonymous"></script>',
              '<script type="application/vnd.jupyter.widget-state+json">',
              data,
              '</script>',
              '</head>',
              '<body>',
              ''
            ].join('\n');
            var views = [];
            var cells = Jupyter.notebook.get_cells();
            Jupyter.notebook.get_cells().forEach(function(cell) {
              if (cell.output_area) {
                cell.output_area.outputs.forEach(function(output) {
                  if (
                    output.data &&
                    output.data[VIEW_MIME_TYPE] &&
                    state.state[output.data[VIEW_MIME_TYPE].model_id]
                  ) {
                    views.push(
                      '\n<script type="' +
                        VIEW_MIME_TYPE +
                        '">\n' +
                        escapeScript(
                          JSON.stringify(
                            output.data[VIEW_MIME_TYPE],
                            null,
                            '    '
                          )
                        ) +
                        '\n</script>'
                    );
                  }
                });
              }
            });
            value += views.join('\n');
            value += '\n\n</body>\n</html>\n';
            var content = document.createElement('textarea');
            content.setAttribute('readonly', 'true');
            content.style.width = '100%';
            content.style.minHeight = '250px';
            content.value = value;

            var mod = dialog.modal({
              show: true,
              title: 'Embed widgets',
              body: content,
              keyboard_manager: Jupyter.notebook.keyboard_manager,
              notebook: Jupyter.notebook,
              buttons: {
                'Copy to Clipboard': {
                  class: 'btn-primary',
                  click: function(event) {
                    content.select();
                    return document.execCommand('copy');
                  }
                }
              }
            });
          });
        }
      );
    });
  };

  var action$1 = {
    help: 'Embed interactive widgets',
    icon: 'fa-sliders',
    help_index: 'zz',
    handler: embed_widgets
  };

  var action_name$1 = 'embed-interactive-widgets';
  var prefix$1 = 'widgets';
  requirejs(['base/js/namespace'], function(Jupyter) {
    Jupyter.notebook.keyboard_manager.actions.register(
      action$1,
      action_name$1,
      prefix$1
    );
  });

  var embed_widgets_1 = {
    action: action$1
  };

  var ManagerBase$2 = lib$2.ManagerBase;





  var MIME_TYPE = 'application/vnd.jupyter.widget-view+json';

  function polyfill_new_comm_buffers(
    manager,
    target_name,
    data,
    callbacks,
    metadata,
    comm_id,
    buffers
  ) {
    /**
     * This polyfills services/kernel/comm/CommManager.new_comm to
     * accept the buffers argument.
     *
     * argument comm_id is optional
     */
    return new Promise(function(resolve) {
      requirejs(['services/kernels/comm'], function(comm) {
        var comm = new comm.Comm(target_name, comm_id);
        manager.register_comm(comm);
        // inline Comm.open(), but with buffers
        var content = {
          comm_id: comm_id,
          target_name: target_name,
          data: data || {}
        };
        comm.kernel.send_shell_message(
          'comm_open',
          content,
          callbacks,
          metadata,
          buffers
        );
        resolve(comm);
      });
    });
  }

  function new_comm(
    manager,
    target_name,
    data,
    callbacks,
    metadata,
    comm_id,
    buffers
  ) {
    // Checks whether new_comm needs a polyfill, and calls the correct version
    // Polyfill needed for notebook <5.1, in which the new_comm method does not support a buffers argument.
    // See https://github.com/jupyter-widgets/ipywidgets/pull/1817
    var need_polyfill = manager.new_comm.length < 6;
    if (need_polyfill) {
      return polyfill_new_comm_buffers.apply(null, arguments);
    }
    return manager.new_comm.apply(
      manager,
      Array.prototype.slice.call(arguments, 1)
    );
  }

  //--------------------------------------------------------------------
  // WidgetManager class
  //--------------------------------------------------------------------

  class WidgetManager extends ManagerBase$2 {
    constructor(comm_manager, notebook) {
      super();
      // Managers are stored in *reverse* order, so that _managers[0] is the most recent.
      WidgetManager._managers.unshift(this);

      // Attach a comm manager
      this.notebook = notebook;
      this.keyboard_manager = notebook.keyboard_manager;
      this.comm_manager = comm_manager;

      var widget_md = notebook.metadata.widgets;

      // Steps that needs to be done:
      // 1. Register comm target
      // 2. Get any widget state from the kernel and open comms with existing state
      // 3. Check saved state for widgets, and restore any that would not overwrite
      //    any live widgets.

      // Register with the comm manager. (1)
      this.comm_manager.register_target(
        this.comm_target_name,
        this.handle_comm_open.bind(this)
      );

      // Attempt to reconstruct any live comms by requesting them from the back-end (2).
      var that = this;
      this._get_comm_info().then(function(comm_ids) {
        // Create comm class instances from comm ids (2).
        var comm_promises = Object.keys(comm_ids).map(function(comm_id) {
          return that._create_comm(that.comm_target_name, comm_id);
        });

        // Send a state request message out for each widget comm and wait
        // for the responses (2).
        return Promise.all(comm_promises)
          .then(function(comms) {
            return Promise.all(
              comms.map(function(comm) {
                var update_promise = new Promise(function(resolve, reject) {
                  comm.on_msg(function(msg) {
                    lib$1.put_buffers(
                      msg.content.data.state,
                      msg.content.data.buffer_paths,
                      msg.buffers
                    );
                    // A suspected response was received, check to see if
                    // it's a state update. If so, resolve.
                    if (msg.content.data.method === 'update') {
                      resolve({
                        comm: comm,
                        msg: msg
                      });
                    }
                  });
                });
                comm.send(
                  {
                    method: 'request_state'
                  },
                  that.callbacks()
                );
                return update_promise;
              })
            );
          })
          .then(function(widgets_info) {
            return Promise.all(
              widgets_info.map(function(widget_info) {
                return that.new_model(
                  {
                    model_name: widget_info.msg.content.data.state._model_name,
                    model_module:
                      widget_info.msg.content.data.state._model_module,
                    model_module_version:
                      widget_info.msg.content.data.state._model_module_version,
                    comm: widget_info.comm
                  },
                  widget_info.msg.content.data.state
                );
              })
            );
          })
          .then(function() {
            // Now that we have mirrored any widgets from the kernel...
            // Restore any widgets from saved state that are not live (3)
            if (
              widget_md &&
              widget_md['application/vnd.jupyter.widget-state+json']
            ) {
              var state =
                notebook.metadata.widgets[
                  'application/vnd.jupyter.widget-state+json'
                ];
              state = that.filterExistingModelState(state);
              return that.set_state(state);
            }
          })
          .then(function() {
            // Rerender cells that have widget data
            that.notebook.get_cells().forEach(function(cell) {
              var rerender =
                cell.output_area &&
                cell.output_area.outputs.find(function(output) {
                  return output.data && output.data[MIME_TYPE];
                });
              if (rerender) {
                that.notebook.render_cell_output(cell);
              }
            });
          });
      });

      // Create the actions and menu
      this._init_actions();
      this._init_menu();
    }

    loadClass(className, moduleName, moduleVersion) {
      if (moduleName === '@jupyter-widgets/controls') {
        return Promise.resolve(lib$3[className]);
      } else if (moduleName === '@jupyter-widgets/base') {
        return Promise.resolve(lib$1[className]);
      } else if (moduleName == '@jupyter-widgets/output') {
        return Promise.resolve(widget_output$1[className]);
      } else {
        return new Promise(function(resolve, reject) {
          window.require([moduleName], resolve, reject);
        }).then(function(mod) {
          if (mod[className]) {
            return mod[className];
          } else {
            return Promise.reject(
              'Class ' + className + ' not found in module ' + moduleName
            );
          }
        });
      }
    }

    /**
     * Registers manager level actions with the notebook actions list
     */
    _init_actions() {
      var notifier = Jupyter.notification_area.widget('widgets');
      this.saveWidgetsAction = {
        handler: function() {
          this.get_state({
            drop_defaults: true
          }).then(function(state) {
            Jupyter.notebook.metadata.widgets = {
              'application/vnd.jupyter.widget-state+json': state
            };
            Jupyter.menubar.actions
              .get('jupyter-notebook:save-notebook')
              .handler({
                notebook: Jupyter.notebook
              });
          });
        }.bind(this),
        icon: 'fa-truck',
        help:
          'Save the notebook with the widget state information for static rendering'
      };
      Jupyter.menubar.actions.register(
        this.saveWidgetsAction,
        'save-with-widgets',
        'widgets'
      );

      this.clearWidgetsAction = {
        handler: function() {
          delete Jupyter.notebook.metadata.widgets;
          Jupyter.menubar.actions.get('jupyter-notebook:save-notebook').handler({
            notebook: Jupyter.notebook
          });
        },
        help: 'Clear the widget state information from the notebook'
      };
      Jupyter.menubar.actions.register(
        this.saveWidgetsAction,
        'save-clear-widgets',
        'widgets'
      );
    }

    /**
     * Initialize the widget menu
     */
    _init_menu() {
      // Add a widgets menubar item, before help.
      var widgetsMenu = document.createElement('li');
      widgetsMenu.classList.add('dropdown');
      var helpMenu = document.querySelector('#help_menu').parentElement;
      helpMenu.parentElement.insertBefore(widgetsMenu, helpMenu);

      var widgetsMenuLink = document.createElement('a');
      widgetsMenuLink.setAttribute('href', '#');
      widgetsMenuLink.setAttribute('data-toggle', 'dropdown');
      widgetsMenuLink.classList.add('dropdown-toggle');
      widgetsMenuLink.innerText = 'Widgets';
      widgetsMenu.appendChild(widgetsMenuLink);

      var widgetsSubmenu = document.createElement('ul');
      widgetsSubmenu.setAttribute('id', 'widget-submenu');
      widgetsSubmenu.classList.add('dropdown-menu');
      widgetsMenu.appendChild(widgetsSubmenu);

      var divider = document.createElement('ul');
      divider.classList.add('divider');

      widgetsSubmenu.appendChild(
        this._createMenuItem('Save Notebook Widget State', this.saveWidgetsAction)
      );
      widgetsSubmenu.appendChild(
        this._createMenuItem(
          'Clear Notebook Widget State',
          this.clearWidgetsAction
        )
      );
      widgetsSubmenu.appendChild(divider);
      widgetsSubmenu.appendChild(
        this._createMenuItem('Download Widget State', save_state_1.action)
      );
      widgetsSubmenu.appendChild(
        this._createMenuItem('Embed Widgets', embed_widgets_1.action)
      );
    }

    /**
     * Creates a menu item for an action.
     * @param  {string} title - display string for the menu item
     * @param  {Action} action
     * @return {HTMLElement} menu item
     */
    _createMenuItem(title, action) {
      var item = document.createElement('li');
      item.setAttribute('title', action.help);

      var itemLink = document.createElement('a');
      itemLink.setAttribute('href', '#');
      itemLink.innerText = title;
      item.appendChild(itemLink);

      item.onclick = action.handler;
      return item;
    }

    _create_comm(comm_target_name, comm_id, data, metadata, buffers) {
      var that = this;
      return this._get_connected_kernel().then(function(kernel) {
        if (data || metadata) {
          return new_comm(
            kernel.comm_manager,
            comm_target_name,
            data,
            that.callbacks(),
            metadata,
            comm_id,
            buffers
          );
        } else {
          // Construct a comm that already is open on the kernel side. We
          // don't want to send an open message, which would supersede the
          // kernel comm object, so we instead do by hand the necessary parts
          // of the new_comm call above.
          return new Promise(function(resolve) {
            requirejs(['services/kernels/comm'], function(comm) {
              var new_comm = new comm.Comm(comm_target_name, comm_id);
              kernel.comm_manager.register_comm(new_comm);
              resolve(new_comm);
            });
          });
        }
      });
    }

    _get_comm_info() {
      return this._get_connected_kernel().then(function(kernel) {
        return new Promise(function(resolve, reject) {
          kernel.comm_info('jupyter.widget', function(msg) {
            resolve(msg['content']['comms']);
          });
        });
      });
    }

    _get_connected_kernel() {
      /**
       * Gets a promise for a connected kernel
       */
      var that = this;
      return new Promise(function(resolve, reject) {
        if (
          that.comm_manager &&
          that.comm_manager.kernel &&
          that.comm_manager.kernel.is_connected()
        ) {
          resolve(that.comm_manager.kernel);
        } else {
          that.notebook.events.on('kernel_connected.Kernel', function(
            event,
            data
          ) {
            resolve(data.kernel);
          });
        }
      });
    }

    setViewOptions(options) {
      var options = options || {};
      if (!options.output && options.parent) {
        // use the parent output if we don't have one
        options.output = options.parent.options.output;
      }
      if (options.output) {
        options.iopub_callbacks = {
          output: options.output.handle_output.bind(options.output),
          clear_output: options.output.handle_clear_output.bind(options.output)
        };
      }
      return options;
    }

    /**
     * Callback handlers for a specific view
     */
    callbacks(view) {
      var callbacks = ManagerBase$2.prototype.callbacks.call(this, view);
      if (view && view.options.iopub_callbacks) {
        callbacks.iopub = view.options.iopub_callbacks;
      }
      return callbacks;
    }
  }

  /**
   * List of widget managers in *reverse* order
   * (_managers[0] is the most recent)
   */
  WidgetManager._managers = [];

  var manager = /*#__PURE__*/Object.freeze({
    WidgetManager: WidgetManager
  });

  var css$1 = ".lm-Widget,.p-Widget{box-sizing:border-box;position:relative;overflow:hidden;cursor:default}.lm-Widget.lm-mod-hidden,.p-Widget.p-mod-hidden{display:none!important}.lm-CommandPalette,.p-CommandPalette{display:flex;flex-direction:column;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}.lm-CommandPalette-search,.p-CommandPalette-search{flex:0 0 auto}.lm-CommandPalette-content,.p-CommandPalette-content{flex:1 1 auto;margin:0;padding:0;min-height:0;overflow:auto;list-style-type:none}.lm-CommandPalette-header,.p-CommandPalette-header{overflow:hidden;white-space:nowrap;text-overflow:ellipsis}.lm-CommandPalette-item,.p-CommandPalette-item{display:flex;flex-direction:row}.lm-CommandPalette-itemIcon,.p-CommandPalette-itemIcon{flex:0 0 auto}.lm-CommandPalette-itemContent,.p-CommandPalette-itemContent{flex:1 1 auto;overflow:hidden}.lm-CommandPalette-itemShortcut,.p-CommandPalette-itemShortcut{flex:0 0 auto}.lm-CommandPalette-itemLabel,.p-CommandPalette-itemLabel{overflow:hidden;white-space:nowrap;text-overflow:ellipsis}.lm-close-icon{border:1px solid transparent;background-color:transparent;position:absolute;z-index:1;right:3%;top:0;bottom:0;margin:auto;padding:7px 0;display:none;vertical-align:middle;outline:0;cursor:pointer}.lm-close-icon:after{content:\"X\";display:block;width:15px;height:15px;text-align:center;color:#000;font-weight:400;font-size:12px;cursor:pointer}.lm-DockPanel,.lm-DockPanel-widget,.p-DockPanel,.p-DockPanel-widget{z-index:0}.lm-DockPanel-tabBar,.p-DockPanel-tabBar{z-index:1}.lm-DockPanel-handle,.p-DockPanel-handle{z-index:2}.lm-DockPanel-handle.lm-mod-hidden,.p-DockPanel-handle.p-mod-hidden{display:none!important}.lm-DockPanel-handle:after,.p-DockPanel-handle:after{position:absolute;top:0;left:0;width:100%;height:100%;content:\"\"}.lm-DockPanel-handle[data-orientation=horizontal],.p-DockPanel-handle[data-orientation=horizontal]{cursor:ew-resize}.lm-DockPanel-handle[data-orientation=vertical],.p-DockPanel-handle[data-orientation=vertical]{cursor:ns-resize}.lm-DockPanel-handle[data-orientation=horizontal]:after,.p-DockPanel-handle[data-orientation=horizontal]:after{left:50%;min-width:8px;transform:translateX(-50%)}.lm-DockPanel-handle[data-orientation=vertical]:after,.p-DockPanel-handle[data-orientation=vertical]:after{top:50%;min-height:8px;transform:translateY(-50%)}.lm-DockPanel-overlay,.p-DockPanel-overlay{z-index:3;box-sizing:border-box;pointer-events:none}.lm-DockPanel-overlay.lm-mod-hidden,.p-DockPanel-overlay.p-mod-hidden{display:none!important}.lm-Menu,.p-Menu{z-index:10000;position:absolute;white-space:nowrap;overflow-x:hidden;overflow-y:auto;outline:none;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}.lm-Menu-content,.p-Menu-content{margin:0;padding:0;display:table;list-style-type:none}.lm-Menu-item,.p-Menu-item{display:table-row}.lm-Menu-item.lm-mod-collapsed,.lm-Menu-item.lm-mod-hidden,.p-Menu-item.p-mod-collapsed,.p-Menu-item.p-mod-hidden{display:none!important}.lm-Menu-itemIcon,.lm-Menu-itemSubmenuIcon,.p-Menu-itemIcon,.p-Menu-itemSubmenuIcon{display:table-cell;text-align:center}.lm-Menu-itemLabel,.p-Menu-itemLabel{display:table-cell;text-align:left}.lm-Menu-itemShortcut,.p-Menu-itemShortcut{display:table-cell;text-align:right}.lm-MenuBar,.p-MenuBar{outline:none;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}.lm-MenuBar-content,.p-MenuBar-content{margin:0;padding:0;display:flex;flex-direction:row;list-style-type:none}.lm-MenuBar-item,.p--MenuBar-item{box-sizing:border-box}.lm-MenuBar-itemIcon,.lm-MenuBar-itemLabel,.p-MenuBar-itemIcon,.p-MenuBar-itemLabel{display:inline-block}.lm-ScrollBar,.p-ScrollBar{display:flex;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}.lm-ScrollBar[data-orientation=horizontal],.p-ScrollBar[data-orientation=horizontal]{flex-direction:row}.lm-ScrollBar[data-orientation=vertical],.p-ScrollBar[data-orientation=vertical]{flex-direction:column}.lm-ScrollBar-button,.p-ScrollBar-button{box-sizing:border-box;flex:0 0 auto}.lm-ScrollBar-track,.p-ScrollBar-track{box-sizing:border-box;position:relative;overflow:hidden;flex:1 1 auto}.lm-ScrollBar-thumb,.p-ScrollBar-thumb{box-sizing:border-box;position:absolute}.lm-SplitPanel-child,.p-SplitPanel-child{z-index:0}.lm-SplitPanel-handle,.p-SplitPanel-handle{z-index:1}.lm-SplitPanel-handle.lm-mod-hidden,.p-SplitPanel-handle.p-mod-hidden{display:none!important}.lm-SplitPanel-handle:after,.p-SplitPanel-handle:after{position:absolute;top:0;left:0;width:100%;height:100%;content:\"\"}.lm-SplitPanel[data-orientation=horizontal]>.lm-SplitPanel-handle,.p-SplitPanel[data-orientation=horizontal]>.p-SplitPanel-handle{cursor:ew-resize}.lm-SplitPanel[data-orientation=vertical]>.lm-SplitPanel-handle,.p-SplitPanel[data-orientation=vertical]>.p-SplitPanel-handle{cursor:ns-resize}.lm-SplitPanel[data-orientation=horizontal]>.lm-SplitPanel-handle:after,.p-SplitPanel[data-orientation=horizontal]>.p-SplitPanel-handle:after{left:50%;min-width:8px;transform:translateX(-50%)}.lm-SplitPanel[data-orientation=vertical]>.lm-SplitPanel-handle:after,.p-SplitPanel[data-orientation=vertical]>.p-SplitPanel-handle:after{top:50%;min-height:8px;transform:translateY(-50%)}.lm-TabBar,.p-TabBar{display:flex;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}.lm-TabBar[data-orientation=horizontal],.p-TabBar[data-orientation=horizontal]{flex-direction:row}.lm-TabBar[data-orientation=vertical],.p-TabBar[data-orientation=vertical]{flex-direction:column}.lm-TabBar-content,.p-TabBar-content{margin:0;padding:0;display:flex;flex:1 1 auto;list-style-type:none}.lm-TabBar[data-orientation=horizontal]>.lm-TabBar-content,.p-TabBar[data-orientation=horizontal]>.p-TabBar-content{flex-direction:row}.lm-TabBar[data-orientation=vertical]>.lm-TabBar-content,.p-TabBar[data-orientation=vertical]>.p-TabBar-content{flex-direction:column}.lm-TabBar-tab,.p-TabBar-tab{display:flex;flex-direction:row;box-sizing:border-box;overflow:hidden}.lm-TabBar-tabCloseIcon,.lm-TabBar-tabIcon,.p-TabBar-tabCloseIcon,.p-TabBar-tabIcon{flex:0 0 auto}.lm-TabBar-tabLabel,.p-TabBar-tabLabel{flex:1 1 auto;overflow:hidden;white-space:nowrap}.lm-TabBar-tab.lm-mod-hidden,.p-TabBar-tab.p-mod-hidden{display:none!important}.lm-TabBar.lm-mod-dragging .lm-TabBar-tab,.p-TabBar.p-mod-dragging .p-TabBar-tab{position:relative}.lm-TabBar.lm-mod-dragging[data-orientation=horizontal] .lm-TabBar-tab,.p-TabBar.p-mod-dragging[data-orientation=horizontal] .p-TabBar-tab{left:0;transition:left .15s ease}.lm-TabBar.lm-mod-dragging[data-orientation=vertical] .lm-TabBar-tab,.p-TabBar.p-mod-dragging[data-orientation=vertical] .p-TabBar-tab{top:0;transition:top .15s ease}.p-TabBar.p-mod-dragging .p-TabBar-tab.p-mod-dragging .lm-TabBar.lm-mod-dragging .lm-TabBar-tab.lm-mod-dragging{transition:none}.lm-TabPanel-tabBar,.p-TabPanel-tabBar{z-index:1}.lm-TabPanel-stackedPanel,.p-TabPanel-stackedPanel{z-index:0}";
  styleInject(css$1);

  var style = /*#__PURE__*/Object.freeze({
    'default': css$1
  });

  var css$2 = ".jupyter-widgets-disconnected:before{content:\"\\f127\";display:inline-block;font:normal normal normal 14px/1 FontAwesome;font-size:inherit;text-rendering:auto;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;color:#d9534f;padding:3px;align-self:flex-start}";
  styleInject(css$2);

  var css$3 = /*#__PURE__*/Object.freeze({
    'default': css$2
  });

  var css$4 = ":root{--md-red-50:#ffebee;--md-red-100:#ffcdd2;--md-red-200:#ef9a9a;--md-red-300:#e57373;--md-red-400:#ef5350;--md-red-500:#f44336;--md-red-600:#e53935;--md-red-700:#d32f2f;--md-red-800:#c62828;--md-red-900:#b71c1c;--md-red-A100:#ff8a80;--md-red-A200:#ff5252;--md-red-A400:#ff1744;--md-red-A700:#d50000;--md-pink-50:#fce4ec;--md-pink-100:#f8bbd0;--md-pink-200:#f48fb1;--md-pink-300:#f06292;--md-pink-400:#ec407a;--md-pink-500:#e91e63;--md-pink-600:#d81b60;--md-pink-700:#c2185b;--md-pink-800:#ad1457;--md-pink-900:#880e4f;--md-pink-A100:#ff80ab;--md-pink-A200:#ff4081;--md-pink-A400:#f50057;--md-pink-A700:#c51162;--md-purple-50:#f3e5f5;--md-purple-100:#e1bee7;--md-purple-200:#ce93d8;--md-purple-300:#ba68c8;--md-purple-400:#ab47bc;--md-purple-500:#9c27b0;--md-purple-600:#8e24aa;--md-purple-700:#7b1fa2;--md-purple-800:#6a1b9a;--md-purple-900:#4a148c;--md-purple-A100:#ea80fc;--md-purple-A200:#e040fb;--md-purple-A400:#d500f9;--md-purple-A700:#a0f;--md-deep-purple-50:#ede7f6;--md-deep-purple-100:#d1c4e9;--md-deep-purple-200:#b39ddb;--md-deep-purple-300:#9575cd;--md-deep-purple-400:#7e57c2;--md-deep-purple-500:#673ab7;--md-deep-purple-600:#5e35b1;--md-deep-purple-700:#512da8;--md-deep-purple-800:#4527a0;--md-deep-purple-900:#311b92;--md-deep-purple-A100:#b388ff;--md-deep-purple-A200:#7c4dff;--md-deep-purple-A400:#651fff;--md-deep-purple-A700:#6200ea;--md-indigo-50:#e8eaf6;--md-indigo-100:#c5cae9;--md-indigo-200:#9fa8da;--md-indigo-300:#7986cb;--md-indigo-400:#5c6bc0;--md-indigo-500:#3f51b5;--md-indigo-600:#3949ab;--md-indigo-700:#303f9f;--md-indigo-800:#283593;--md-indigo-900:#1a237e;--md-indigo-A100:#8c9eff;--md-indigo-A200:#536dfe;--md-indigo-A400:#3d5afe;--md-indigo-A700:#304ffe;--md-blue-50:#e3f2fd;--md-blue-100:#bbdefb;--md-blue-200:#90caf9;--md-blue-300:#64b5f6;--md-blue-400:#42a5f5;--md-blue-500:#2196f3;--md-blue-600:#1e88e5;--md-blue-700:#1976d2;--md-blue-800:#1565c0;--md-blue-900:#0d47a1;--md-blue-A100:#82b1ff;--md-blue-A200:#448aff;--md-blue-A400:#2979ff;--md-blue-A700:#2962ff;--md-light-blue-50:#e1f5fe;--md-light-blue-100:#b3e5fc;--md-light-blue-200:#81d4fa;--md-light-blue-300:#4fc3f7;--md-light-blue-400:#29b6f6;--md-light-blue-500:#03a9f4;--md-light-blue-600:#039be5;--md-light-blue-700:#0288d1;--md-light-blue-800:#0277bd;--md-light-blue-900:#01579b;--md-light-blue-A100:#80d8ff;--md-light-blue-A200:#40c4ff;--md-light-blue-A400:#00b0ff;--md-light-blue-A700:#0091ea;--md-cyan-50:#e0f7fa;--md-cyan-100:#b2ebf2;--md-cyan-200:#80deea;--md-cyan-300:#4dd0e1;--md-cyan-400:#26c6da;--md-cyan-500:#00bcd4;--md-cyan-600:#00acc1;--md-cyan-700:#0097a7;--md-cyan-800:#00838f;--md-cyan-900:#006064;--md-cyan-A100:#84ffff;--md-cyan-A200:#18ffff;--md-cyan-A400:#00e5ff;--md-cyan-A700:#00b8d4;--md-teal-50:#e0f2f1;--md-teal-100:#b2dfdb;--md-teal-200:#80cbc4;--md-teal-300:#4db6ac;--md-teal-400:#26a69a;--md-teal-500:#009688;--md-teal-600:#00897b;--md-teal-700:#00796b;--md-teal-800:#00695c;--md-teal-900:#004d40;--md-teal-A100:#a7ffeb;--md-teal-A200:#64ffda;--md-teal-A400:#1de9b6;--md-teal-A700:#00bfa5;--md-green-50:#e8f5e9;--md-green-100:#c8e6c9;--md-green-200:#a5d6a7;--md-green-300:#81c784;--md-green-400:#66bb6a;--md-green-500:#4caf50;--md-green-600:#43a047;--md-green-700:#388e3c;--md-green-800:#2e7d32;--md-green-900:#1b5e20;--md-green-A100:#b9f6ca;--md-green-A200:#69f0ae;--md-green-A400:#00e676;--md-green-A700:#00c853;--md-light-green-50:#f1f8e9;--md-light-green-100:#dcedc8;--md-light-green-200:#c5e1a5;--md-light-green-300:#aed581;--md-light-green-400:#9ccc65;--md-light-green-500:#8bc34a;--md-light-green-600:#7cb342;--md-light-green-700:#689f38;--md-light-green-800:#558b2f;--md-light-green-900:#33691e;--md-light-green-A100:#ccff90;--md-light-green-A200:#b2ff59;--md-light-green-A400:#76ff03;--md-light-green-A700:#64dd17;--md-lime-50:#f9fbe7;--md-lime-100:#f0f4c3;--md-lime-200:#e6ee9c;--md-lime-300:#dce775;--md-lime-400:#d4e157;--md-lime-500:#cddc39;--md-lime-600:#c0ca33;--md-lime-700:#afb42b;--md-lime-800:#9e9d24;--md-lime-900:#827717;--md-lime-A100:#f4ff81;--md-lime-A200:#eeff41;--md-lime-A400:#c6ff00;--md-lime-A700:#aeea00;--md-yellow-50:#fffde7;--md-yellow-100:#fff9c4;--md-yellow-200:#fff59d;--md-yellow-300:#fff176;--md-yellow-400:#ffee58;--md-yellow-500:#ffeb3b;--md-yellow-600:#fdd835;--md-yellow-700:#fbc02d;--md-yellow-800:#f9a825;--md-yellow-900:#f57f17;--md-yellow-A100:#ffff8d;--md-yellow-A200:#ff0;--md-yellow-A400:#ffea00;--md-yellow-A700:#ffd600;--md-amber-50:#fff8e1;--md-amber-100:#ffecb3;--md-amber-200:#ffe082;--md-amber-300:#ffd54f;--md-amber-400:#ffca28;--md-amber-500:#ffc107;--md-amber-600:#ffb300;--md-amber-700:#ffa000;--md-amber-800:#ff8f00;--md-amber-900:#ff6f00;--md-amber-A100:#ffe57f;--md-amber-A200:#ffd740;--md-amber-A400:#ffc400;--md-amber-A700:#ffab00;--md-orange-50:#fff3e0;--md-orange-100:#ffe0b2;--md-orange-200:#ffcc80;--md-orange-300:#ffb74d;--md-orange-400:#ffa726;--md-orange-500:#ff9800;--md-orange-600:#fb8c00;--md-orange-700:#f57c00;--md-orange-800:#ef6c00;--md-orange-900:#e65100;--md-orange-A100:#ffd180;--md-orange-A200:#ffab40;--md-orange-A400:#ff9100;--md-orange-A700:#ff6d00;--md-deep-orange-50:#fbe9e7;--md-deep-orange-100:#ffccbc;--md-deep-orange-200:#ffab91;--md-deep-orange-300:#ff8a65;--md-deep-orange-400:#ff7043;--md-deep-orange-500:#ff5722;--md-deep-orange-600:#f4511e;--md-deep-orange-700:#e64a19;--md-deep-orange-800:#d84315;--md-deep-orange-900:#bf360c;--md-deep-orange-A100:#ff9e80;--md-deep-orange-A200:#ff6e40;--md-deep-orange-A400:#ff3d00;--md-deep-orange-A700:#dd2c00;--md-brown-50:#efebe9;--md-brown-100:#d7ccc8;--md-brown-200:#bcaaa4;--md-brown-300:#a1887f;--md-brown-400:#8d6e63;--md-brown-500:#795548;--md-brown-600:#6d4c41;--md-brown-700:#5d4037;--md-brown-800:#4e342e;--md-brown-900:#3e2723;--md-grey-50:#fafafa;--md-grey-100:#f5f5f5;--md-grey-200:#eee;--md-grey-300:#e0e0e0;--md-grey-400:#bdbdbd;--md-grey-500:#9e9e9e;--md-grey-600:#757575;--md-grey-700:#616161;--md-grey-800:#424242;--md-grey-900:#212121;--md-blue-grey-50:#eceff1;--md-blue-grey-100:#cfd8dc;--md-blue-grey-200:#b0bec5;--md-blue-grey-300:#90a4ae;--md-blue-grey-400:#78909c;--md-blue-grey-500:#607d8b;--md-blue-grey-600:#546e7a;--md-blue-grey-700:#455a64;--md-blue-grey-800:#37474f;--md-blue-grey-900:#263238;--jp-icon-search:none;--jp-ui-select-caret:none;--jp-border-width:1px;--jp-border-color0:var(--md-grey-700);--jp-border-color1:var(--md-grey-500);--jp-border-color2:var(--md-grey-300);--jp-border-color3:var(--md-grey-100);--jp-ui-icon-font-size:14px;--jp-ui-font-family:\"Helvetica Neue\",Helvetica,Arial,sans-serif;--jp-ui-font-color0:#000;--jp-ui-font-color1:rgba(0,0,0,0.8);--jp-ui-font-color2:rgba(0,0,0,0.5);--jp-ui-font-color3:rgba(0,0,0,0.3);--jp-ui-inverse-font-color0:#fff;--jp-ui-inverse-font-color1:#fff;--jp-ui-inverse-font-color2:hsla(0,0%,100%,0.7);--jp-ui-inverse-font-color3:hsla(0,0%,100%,0.5);--jp-inverse-ui-font-color0:#fff;--jp-inverse-ui-font-color1:#fff;--jp-inverse-ui-font-color2:hsla(0,0%,100%,0.7);--jp-inverse-ui-font-color3:hsla(0,0%,100%,0.5);--jp-content-font-size:13px;--jp-content-line-height:1.5;--jp-content-font-color0:#000;--jp-content-font-color1:#000;--jp-content-font-color2:var(--md-grey-700);--jp-content-font-color3:var(--md-grey-500);--jp-ui-font-scale-factor:1.2;--jp-ui-font-size0:calc(var(--jp-ui-font-size1)/var(--jp-ui-font-scale-factor));--jp-ui-font-size1:13px;--jp-ui-font-size2:calc(var(--jp-ui-font-size1)*var(--jp-ui-font-scale-factor));--jp-ui-font-size3:calc(var(--jp-ui-font-size2)*var(--jp-ui-font-scale-factor));--jp-code-font-size:13px;--jp-code-line-height:1.307;--jp-code-padding:5px;--jp-code-font-family:monospace;--jp-layout-color0:#fff;--jp-layout-color1:#fff;--jp-layout-color2:var(--md-grey-200);--jp-layout-color3:var(--md-grey-400);--jp-brand-color0:var(--md-blue-700);--jp-brand-color1:var(--md-blue-500);--jp-brand-color2:var(--md-blue-300);--jp-brand-color3:var(--md-blue-100);--jp-accent-color0:var(--md-green-700);--jp-accent-color1:var(--md-green-500);--jp-accent-color2:var(--md-green-300);--jp-accent-color3:var(--md-green-100);--jp-warn-color0:var(--md-orange-700);--jp-warn-color1:var(--md-orange-500);--jp-warn-color2:var(--md-orange-300);--jp-warn-color3:var(--md-orange-100);--jp-error-color0:var(--md-red-700);--jp-error-color1:var(--md-red-500);--jp-error-color2:var(--md-red-300);--jp-error-color3:var(--md-red-100);--jp-success-color0:var(--md-green-700);--jp-success-color1:var(--md-green-500);--jp-success-color2:var(--md-green-300);--jp-success-color3:var(--md-green-100);--jp-info-color0:var(--md-cyan-700);--jp-info-color1:var(--md-cyan-500);--jp-info-color2:var(--md-cyan-300);--jp-info-color3:var(--md-cyan-100);--jp-cell-padding:5px;--jp-cell-editor-background:#f7f7f7;--jp-cell-editor-border-color:#cfcfcf;--jp-cell-editor-background-edit:var(--jp-ui-layout-color1);--jp-cell-editor-border-color-edit:var(--jp-brand-color1);--jp-cell-prompt-width:100px;--jp-cell-prompt-font-family:\"Roboto Mono\",monospace;--jp-cell-prompt-letter-spacing:0px;--jp-cell-prompt-opacity:1.0;--jp-cell-prompt-opacity-not-active:0.4;--jp-cell-prompt-font-color-not-active:var(--md-grey-700);--jp-cell-inprompt-font-color:#307fc1;--jp-cell-outprompt-font-color:#bf5b3d;--jp-notebook-padding:10px;--jp-notebook-scroll-padding:100px;--jp-console-background:var(--md-grey-100);--jp-toolbar-border-color:var(--md-grey-400);--jp-toolbar-micro-height:8px;--jp-toolbar-background:var(--jp-layout-color0);--jp-toolbar-box-shadow:0px 0px 2px 0px rgba(0,0,0,0.24);--jp-toolbar-header-margin:4px 4px 0px 4px;--jp-toolbar-active-background:var(--md-grey-300)}.jupyter-widgets.widget-tab>.p-TabBar{display:flex;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}.jupyter-widgets.widget-tab>.p-TabBar[data-orientation=horizontal]{flex-direction:row}.jupyter-widgets.widget-tab>.p-TabBar[data-orientation=vertical]{flex-direction:column}.jupyter-widgets.widget-tab>.p-TabBar>.p-TabBar-content{margin:0;padding:0;display:flex;flex:1 1 auto;list-style-type:none}.jupyter-widgets.widget-tab>.p-TabBar[data-orientation=horizontal]>.p-TabBar-content{flex-direction:row}.jupyter-widgets.widget-tab>.p-TabBar[data-orientation=vertical]>.p-TabBar-content{flex-direction:column}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab{display:flex;flex-direction:row;box-sizing:border-box;overflow:hidden}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tabCloseIcon,.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tabIcon{flex:0 0 auto}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tabLabel{flex:1 1 auto;overflow:hidden;white-space:nowrap}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab.p-mod-hidden{display:none!important}.jupyter-widgets.widget-tab>.p-TabBar.p-mod-dragging .p-TabBar-tab{position:relative}.jupyter-widgets.widget-tab>.p-TabBar.p-mod-dragging[data-orientation=horizontal] .p-TabBar-tab{left:0;transition:left .15s ease}.jupyter-widgets.widget-tab>.p-TabBar.p-mod-dragging[data-orientation=vertical] .p-TabBar-tab{top:0;transition:top .15s ease}.jupyter-widgets.widget-tab>.p-TabBar.p-mod-dragging .p-TabBar-tab.p-mod-dragging{transition:none}:root{--jp-widgets-color:var(--jp-content-font-color1);--jp-widgets-label-color:var(--jp-widgets-color);--jp-widgets-readout-color:var(--jp-widgets-color);--jp-widgets-font-size:var(--jp-ui-font-size1);--jp-widgets-margin:2px;--jp-widgets-inline-height:28px;--jp-widgets-inline-width:300px;--jp-widgets-inline-width-short:calc(var(--jp-widgets-inline-width)/2 - var(--jp-widgets-margin));--jp-widgets-inline-width-tiny:calc(var(--jp-widgets-inline-width-short)/2 - var(--jp-widgets-margin));--jp-widgets-inline-margin:4px;--jp-widgets-inline-label-width:80px;--jp-widgets-border-width:var(--jp-border-width);--jp-widgets-vertical-height:200px;--jp-widgets-horizontal-tab-height:24px;--jp-widgets-horizontal-tab-width:144px;--jp-widgets-horizontal-tab-top-border:2px;--jp-widgets-progress-thickness:20px;--jp-widgets-container-padding:15px;--jp-widgets-input-padding:4px;--jp-widgets-radio-item-height-adjustment:8px;--jp-widgets-radio-item-height:calc(var(--jp-widgets-inline-height) - var(--jp-widgets-radio-item-height-adjustment));--jp-widgets-slider-track-thickness:4px;--jp-widgets-slider-border-width:var(--jp-widgets-border-width);--jp-widgets-slider-handle-size:16px;--jp-widgets-slider-handle-border-color:var(--jp-border-color1);--jp-widgets-slider-handle-background-color:var(--jp-layout-color1);--jp-widgets-slider-active-handle-color:var(--jp-brand-color1);--jp-widgets-menu-item-height:24px;--jp-widgets-dropdown-arrow:url(\"data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCAxOCAxOCI+PHBhdGggZD0iTTUuMiA1LjlMOSA5LjdsMy44LTMuOEwxNCA3LjFsLTQuOSA1LTQuOS01IDEtMS4yeiIvPjxwYXRoIGQ9Ik0wLS42aDE4djE4SDB2LTE4eiIgZmlsbD0ibm9uZSIvPjwvc3ZnPg==\");--jp-widgets-input-color:var(--jp-ui-font-color1);--jp-widgets-input-background-color:var(--jp-layout-color1);--jp-widgets-input-border-color:var(--jp-border-color1);--jp-widgets-input-focus-border-color:var(--jp-brand-color2);--jp-widgets-input-border-width:var(--jp-widgets-border-width);--jp-widgets-disabled-opacity:0.6;--md-shadow-key-umbra-opacity:0.2;--md-shadow-key-penumbra-opacity:0.14;--md-shadow-ambient-shadow-opacity:0.12}.jupyter-widgets{margin:var(--jp-widgets-margin);box-sizing:border-box;color:var(--jp-widgets-color);overflow:visible}.jupyter-widgets.jupyter-widgets-disconnected:before{line-height:var(--jp-widgets-inline-height);height:var(--jp-widgets-inline-height)}.jp-Output-result>.jupyter-widgets{margin-left:0;margin-right:0}.widget-inline-hbox{box-sizing:border-box;display:flex;flex-direction:row;align-items:baseline}.widget-inline-vbox{box-sizing:border-box;display:flex;flex-direction:column;align-items:center}.widget-box{display:flex}.widget-box,.widget-gridbox{box-sizing:border-box;margin:0;overflow:auto}.widget-gridbox{display:grid}.widget-hbox{flex-direction:row}.widget-vbox{flex-direction:column}.jupyter-button{padding:0 10px;display:inline-block;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;text-align:center;font-size:var(--jp-widgets-font-size);cursor:pointer;height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height);box-shadow:none;color:var(--jp-ui-font-color1);background-color:var(--jp-layout-color2);border:none;user-select:none}.jupyter-button i.fa{margin-right:var(--jp-widgets-inline-margin);pointer-events:none}.jupyter-button:empty:before{content:\"\\200b\"}.jupyter-widgets.jupyter-button:disabled{opacity:var(--jp-widgets-disabled-opacity)}.jupyter-button i.fa.center{margin-right:0}.jupyter-button:focus:enabled,.jupyter-button:hover:enabled{box-shadow:0 2px 2px 0 rgba(0,0,0,var(--md-shadow-key-penumbra-opacity)),0 3px 1px -2px rgba(0,0,0,var(--md-shadow-key-umbra-opacity)),0 1px 5px 0 rgba(0,0,0,var(--md-shadow-ambient-shadow-opacity))}.jupyter-button.mod-active,.jupyter-button:active{box-shadow:0 4px 5px 0 rgba(0,0,0,var(--md-shadow-key-penumbra-opacity)),0 1px 10px 0 rgba(0,0,0,var(--md-shadow-ambient-shadow-opacity)),0 2px 4px -1px rgba(0,0,0,var(--md-shadow-key-umbra-opacity));color:var(--jp-ui-font-color1);background-color:var(--jp-layout-color3)}.jupyter-button:focus:enabled{outline:1px solid var(--jp-widgets-input-focus-border-color)}.jupyter-button.mod-primary{color:var(--jp-ui-inverse-font-color1,var(--jp-inverse-ui-font-color1));background-color:var(--jp-brand-color1)}.jupyter-button.mod-primary.mod-active,.jupyter-button.mod-primary:active{color:var(--jp-ui-inverse-font-color0,var(--jp-inverse-ui-font-color0));background-color:var(--jp-brand-color0)}.jupyter-button.mod-success{color:var(--jp-ui-inverse-font-color1,var(--jp-inverse-ui-font-color1));background-color:var(--jp-success-color1)}.jupyter-button.mod-success.mod-active,.jupyter-button.mod-success:active{color:var(--jp-ui-inverse-font-color0,var(--jp-inverse-ui-font-color0));background-color:var(--jp-success-color0)}.jupyter-button.mod-info{color:var(--jp-ui-inverse-font-color1,var(--jp-inverse-ui-font-color1));background-color:var(--jp-info-color1)}.jupyter-button.mod-info.mod-active,.jupyter-button.mod-info:active{color:var(--jp-ui-inverse-font-color0,var(--jp-inverse-ui-font-color0));background-color:var(--jp-info-color0)}.jupyter-button.mod-warning{color:var(--jp-ui-inverse-font-color1,var(--jp-inverse-ui-font-color1));background-color:var(--jp-warn-color1)}.jupyter-button.mod-warning.mod-active,.jupyter-button.mod-warning:active{color:var(--jp-ui-inverse-font-color0,var(--jp-inverse-ui-font-color0));background-color:var(--jp-warn-color0)}.jupyter-button.mod-danger{color:var(--jp-ui-inverse-font-color1,var(--jp-inverse-ui-font-color1));background-color:var(--jp-error-color1)}.jupyter-button.mod-danger.mod-active,.jupyter-button.mod-danger:active{color:var(--jp-ui-inverse-font-color0,var(--jp-inverse-ui-font-color0));background-color:var(--jp-error-color0)}.widget-button,.widget-toggle-button,.widget-upload{width:var(--jp-widgets-inline-width-short)}.jupyter-widgets label{margin-bottom:0}.widget-label,.widget-label-basic{color:var(--jp-widgets-label-color);font-size:var(--jp-widgets-font-size);overflow:hidden;text-overflow:ellipsis;white-space:nowrap;line-height:var(--jp-widgets-inline-height)}.widget-inline-hbox .widget-label{color:var(--jp-widgets-label-color);text-align:right;margin-right:calc(var(--jp-widgets-inline-margin)*2);width:var(--jp-widgets-inline-label-width);flex-shrink:0}.widget-inline-vbox .widget-label{color:var(--jp-widgets-label-color)}.widget-inline-vbox .widget-label,.widget-readout{text-align:center;line-height:var(--jp-widgets-inline-height)}.widget-readout{color:var(--jp-widgets-readout-color);font-size:var(--jp-widgets-font-size);height:var(--jp-widgets-inline-height);overflow:hidden;white-space:nowrap}.widget-readout.overflow{-webkit-box-shadow:0 2px 2px 0 rgba(0,0,0,.2),0 3px 1px -2px rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12);-moz-box-shadow:0 2px 2px 0 rgba(0,0,0,.2),0 3px 1px -2px rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12);box-shadow:0 2px 2px 0 rgba(0,0,0,.2),0 3px 1px -2px rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12)}.widget-inline-hbox .widget-readout{text-align:center;max-width:var(--jp-widgets-inline-width-short);min-width:var(--jp-widgets-inline-width-tiny);margin-left:var(--jp-widgets-inline-margin)}.widget-inline-vbox .widget-readout{margin-top:var(--jp-widgets-inline-margin);width:inherit}.widget-checkbox{width:var(--jp-widgets-inline-width);height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height)}.widget-checkbox input[type=checkbox]{margin:0 calc(var(--jp-widgets-inline-margin)*2) 0 0;line-height:var(--jp-widgets-inline-height);font-size:large;flex-grow:1;flex-shrink:0;align-self:center}.widget-valid{height:var(--jp-widgets-inline-height);width:var(--jp-widgets-inline-width-short);font-size:var(--jp-widgets-font-size)}.widget-valid,.widget-valid i:before{line-height:var(--jp-widgets-inline-height)}.widget-valid i:before{margin-right:var(--jp-widgets-inline-margin);margin-left:var(--jp-widgets-inline-margin)}.widget-valid.mod-valid i:before{color:green}.widget-valid.mod-invalid i:before{color:red}.widget-valid.mod-valid .widget-valid-readout{display:none}.widget-text,.widget-textarea{width:var(--jp-widgets-inline-width)}.widget-text input[type=number],.widget-text input[type=password],.widget-text input[type=text]{height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height)}.widget-textarea textarea:disabled,.widget-text input[type=number]:disabled,.widget-text input[type=password]:disabled,.widget-text input[type=text]:disabled{opacity:var(--jp-widgets-disabled-opacity)}.widget-textarea textarea,.widget-text input[type=number],.widget-text input[type=password],.widget-text input[type=text]{box-sizing:border-box;border:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color);background-color:var(--jp-widgets-input-background-color);color:var(--jp-widgets-input-color);font-size:var(--jp-widgets-font-size);flex-grow:1;min-width:0;flex-shrink:1;outline:none!important}.widget-textarea textarea,.widget-text input[type=password],.widget-text input[type=text]{padding:var(--jp-widgets-input-padding) calc(var(--jp-widgets-input-padding)*2)}.widget-text input[type=number]{padding:var(--jp-widgets-input-padding) 0 var(--jp-widgets-input-padding) calc(var(--jp-widgets-input-padding)*2)}.widget-textarea textarea{height:inherit;width:inherit}.widget-textarea textarea:focus,.widget-text input:focus{border-color:var(--jp-widgets-input-focus-border-color)}.widget-slider .ui-slider{border:var(--jp-widgets-slider-border-width) solid var(--jp-layout-color3);background:var(--jp-layout-color3);box-sizing:border-box;position:relative;border-radius:0}.widget-slider .ui-slider .ui-slider-handle{outline:none!important;position:absolute;background-color:var(--jp-widgets-slider-handle-background-color);border:var(--jp-widgets-slider-border-width) solid var(--jp-widgets-slider-handle-border-color);box-sizing:border-box;z-index:1;background-image:none}.widget-slider .ui-slider .ui-slider-handle:focus,.widget-slider .ui-slider .ui-slider-handle:hover{background-color:var(--jp-widgets-slider-active-handle-color);border:var(--jp-widgets-slider-border-width) solid var(--jp-widgets-slider-active-handle-color)}.widget-slider .ui-slider .ui-slider-handle:active{background-color:var(--jp-widgets-slider-active-handle-color);border-color:var(--jp-widgets-slider-active-handle-color);z-index:2;transform:scale(1.2)}.widget-slider .ui-slider .ui-slider-range{position:absolute;background:var(--jp-widgets-slider-active-handle-color);z-index:0}.widget-hslider .ui-slider .ui-slider-handle{margin-top:calc((var(--jp-widgets-slider-track-thickness) - var(--jp-widgets-slider-handle-size))/2 - var(--jp-widgets-slider-border-width));margin-left:calc(var(--jp-widgets-slider-handle-size)/-2 + var(--jp-widgets-slider-border-width));top:0}.widget-hslider .ui-slider .ui-slider-handle,.widget-vslider .ui-slider .ui-slider-handle{width:var(--jp-widgets-slider-handle-size);height:var(--jp-widgets-slider-handle-size);border-radius:50%}.widget-vslider .ui-slider .ui-slider-handle{margin-bottom:calc(var(--jp-widgets-slider-handle-size)/-2 + var(--jp-widgets-slider-border-width));margin-left:calc((var(--jp-widgets-slider-track-thickness) - var(--jp-widgets-slider-handle-size))/2 - var(--jp-widgets-slider-border-width));left:0}.widget-hslider .ui-slider .ui-slider-range{height:calc(var(--jp-widgets-slider-track-thickness)*2);margin-top:calc((var(--jp-widgets-slider-track-thickness) - var(--jp-widgets-slider-track-thickness)*2)/2 - var(--jp-widgets-slider-border-width))}.widget-vslider .ui-slider .ui-slider-range{width:calc(var(--jp-widgets-slider-track-thickness)*2);margin-left:calc((var(--jp-widgets-slider-track-thickness) - var(--jp-widgets-slider-track-thickness)*2)/2 - var(--jp-widgets-slider-border-width))}.widget-hslider{width:var(--jp-widgets-inline-width);height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height);align-items:center}.widgets-slider .slider-container{overflow:visible}.widget-hslider .slider-container{height:var(--jp-widgets-inline-height);margin-left:calc(var(--jp-widgets-slider-handle-size) - var(--jp-widgets-slider-border-width));margin-right:calc(var(--jp-widgets-slider-handle-size) - var(--jp-widgets-slider-border-width));flex:1 1 var(--jp-widgets-inline-width-short)}.widget-hslider .ui-slider{height:var(--jp-widgets-slider-track-thickness);margin-top:calc((var(--jp-widgets-inline-height) - var(--jp-widgets-slider-track-thickness))/2);width:100%}.widget-vbox .widget-label{height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height)}.widget-vslider{height:var(--jp-widgets-vertical-height);width:var(--jp-widgets-inline-width-tiny)}.widget-vslider .slider-container{flex:1 1 var(--jp-widgets-inline-width-short);margin-left:auto;margin-right:auto;margin-bottom:calc(var(--jp-widgets-slider-handle-size) - var(--jp-widgets-slider-border-width));margin-top:calc(var(--jp-widgets-slider-handle-size) - var(--jp-widgets-slider-border-width));display:flex;flex-direction:column}.widget-vslider .ui-slider-vertical{width:var(--jp-widgets-slider-track-thickness);flex-grow:1;margin-left:auto;margin-right:auto}.progress-bar{-webkit-transition:none;-moz-transition:none;-ms-transition:none;-o-transition:none;transition:none;height:var(--jp-widgets-inline-height);background-color:var(--jp-brand-color1)}.progress-bar-success{background-color:var(--jp-success-color1)}.progress-bar-info{background-color:var(--jp-info-color1)}.progress-bar-warning{background-color:var(--jp-warn-color1)}.progress-bar-danger{background-color:var(--jp-error-color1)}.progress{background-color:var(--jp-layout-color2);border:none;box-shadow:none}.widget-hprogress{height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height);width:var(--jp-widgets-inline-width);align-items:center}.widget-hprogress .progress{flex-grow:1;margin-top:var(--jp-widgets-input-padding);margin-bottom:var(--jp-widgets-input-padding);align-self:stretch;height:auto}.widget-vprogress{height:var(--jp-widgets-vertical-height);width:var(--jp-widgets-inline-width-tiny)}.widget-vprogress .progress{flex-grow:1;width:var(--jp-widgets-progress-thickness);margin-left:auto;margin-right:auto;margin-bottom:0}.widget-dropdown{height:var(--jp-widgets-inline-height);width:var(--jp-widgets-inline-width);line-height:var(--jp-widgets-inline-height)}.widget-dropdown>select{padding-right:20px;border:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color);border-radius:0;height:inherit;flex:1 1 var(--jp-widgets-inline-width-short);min-width:0;box-sizing:border-box;outline:none!important;box-shadow:none;background-color:var(--jp-widgets-input-background-color);color:var(--jp-widgets-input-color);font-size:var(--jp-widgets-font-size);vertical-align:top;padding-left:calc(var(--jp-widgets-input-padding)*2);appearance:none;-webkit-appearance:none;-moz-appearance:none;background-repeat:no-repeat;background-size:20px;background-position:100%;background-image:var(--jp-widgets-dropdown-arrow)}.widget-dropdown>select:focus{border-color:var(--jp-widgets-input-focus-border-color)}.widget-dropdown>select:disabled{opacity:var(--jp-widgets-disabled-opacity)}.widget-dropdown>select:-moz-focusring{color:transparent;text-shadow:0 0 0 #000}.widget-select{width:var(--jp-widgets-inline-width);line-height:var(--jp-widgets-inline-height);align-items:flex-start}.widget-select>select{border:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color);background-color:var(--jp-widgets-input-background-color);color:var(--jp-widgets-input-color);font-size:var(--jp-widgets-font-size);flex:1 1 var(--jp-widgets-inline-width-short);outline:none!important;overflow:auto;height:inherit;padding-top:5px}.widget-select>select:focus{border-color:var(--jp-widgets-input-focus-border-color)}.wiget-select>select>option{padding-left:var(--jp-widgets-input-padding);padding-top:calc(var(--jp-widgets-inline-height) - var(--jp-widgets-font-size)/2);padding-bottom:calc(var(--jp-widgets-inline-height) - var(--jp-widgets-font-size)/2)}.widget-toggle-buttons,.wiget-select>select>option{line-height:var(--jp-widgets-inline-height)}.widget-toggle-buttons .widget-toggle-button{margin-left:var(--jp-widgets-margin);margin-right:var(--jp-widgets-margin)}.widget-toggle-buttons .jupyter-button:disabled{opacity:var(--jp-widgets-disabled-opacity)}.widget-radio{width:var(--jp-widgets-inline-width);line-height:var(--jp-widgets-inline-height)}.widget-radio-box{display:flex;flex-direction:column;align-items:stretch;box-sizing:border-box;flex-grow:1;margin-bottom:var(--jp-widgets-radio-item-height-adjustment)}.widget-radio-box label{font-size:var(--jp-widgets-font-size)}.widget-radio-box input,.widget-radio-box label{height:var(--jp-widgets-radio-item-height);line-height:var(--jp-widgets-radio-item-height)}.widget-radio-box input{margin:0 calc(var(--jp-widgets-input-padding)*2) 0 1px;float:left}.widget-colorpicker{width:var(--jp-widgets-inline-width);height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height)}.widget-colorpicker>.widget-colorpicker-input{flex-grow:1;flex-shrink:1;min-width:var(--jp-widgets-inline-width-tiny)}.widget-colorpicker input[type=color]{width:var(--jp-widgets-inline-height);height:var(--jp-widgets-inline-height);padding:0 2px;background:var(--jp-widgets-input-background-color);color:var(--jp-widgets-input-color);border:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color);border-left:none;flex-grow:0;flex-shrink:0;box-sizing:border-box;align-self:stretch;outline:none!important}.widget-colorpicker.concise input[type=color]{border-left:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color)}.widget-colorpicker input[type=color]:focus,.widget-colorpicker input[type=text]:focus{border-color:var(--jp-widgets-input-focus-border-color)}.widget-colorpicker input[type=text]{flex-grow:1;outline:none!important;height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height);background:var(--jp-widgets-input-background-color);color:var(--jp-widgets-input-color);border:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color);font-size:var(--jp-widgets-font-size);padding:var(--jp-widgets-input-padding) calc(var(--jp-widgets-input-padding)*2);min-width:0;flex-shrink:1;box-sizing:border-box}.widget-colorpicker input[type=text]:disabled{opacity:var(--jp-widgets-disabled-opacity)}.widget-datepicker{width:var(--jp-widgets-inline-width);height:var(--jp-widgets-inline-height);line-height:var(--jp-widgets-inline-height)}.widget-datepicker input[type=date]{flex-grow:1;flex-shrink:1;min-width:0;outline:none!important;height:var(--jp-widgets-inline-height);border:var(--jp-widgets-input-border-width) solid var(--jp-widgets-input-border-color);background-color:var(--jp-widgets-input-background-color);color:var(--jp-widgets-input-color);font-size:var(--jp-widgets-font-size);padding:var(--jp-widgets-input-padding) calc(var(--jp-widgets-input-padding)*2);box-sizing:border-box}.widget-datepicker input[type=date]:focus{border-color:var(--jp-widgets-input-focus-border-color)}.widget-datepicker input[type=date]:invalid{border-color:var(--jp-warn-color1)}.widget-datepicker input[type=date]:disabled{opacity:var(--jp-widgets-disabled-opacity)}.widget-play{width:var(--jp-widgets-inline-width-short);display:flex;align-items:stretch}.widget-play .jupyter-button{flex-grow:1;height:auto}.widget-play .jupyter-button:disabled{opacity:var(--jp-widgets-disabled-opacity)}.jupyter-widgets.widget-tab{display:flex;flex-direction:column}.jupyter-widgets.widget-tab>.p-TabBar{overflow-x:visible;overflow-y:visible}.jupyter-widgets.widget-tab>.p-TabBar>.p-TabBar-content{align-items:flex-end;min-width:0;min-height:0}.jupyter-widgets.widget-tab>.widget-tab-contents{width:100%;box-sizing:border-box;margin:0;background:var(--jp-layout-color1);color:var(--jp-ui-font-color1);border:var(--jp-border-width) solid var(--jp-border-color1);padding:var(--jp-widgets-container-padding);flex-grow:1;overflow:auto}.jupyter-widgets.widget-tab>.p-TabBar{font:var(--jp-widgets-font-size),Helvetica,Arial,sans-serif;min-height:calc(var(--jp-widgets-horizontal-tab-height) + var(--jp-border-width))}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab{flex:0 1 var(--jp-widgets-horizontal-tab-width);min-width:35px;min-height:calc(var(--jp-widgets-horizontal-tab-height) + var(--jp-border-width));line-height:var(--jp-widgets-horizontal-tab-height);margin-left:calc(-1*var(--jp-border-width));padding:0 10px;background:var(--jp-layout-color2);color:var(--jp-ui-font-color2);border:var(--jp-border-width) solid var(--jp-border-color1);border-bottom:none;position:relative}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab.p-mod-current{color:var(--jp-ui-font-color0);background:var(--jp-layout-color1);min-height:calc(var(--jp-widgets-horizontal-tab-height) + 2*var(--jp-border-width));transform:translateY(var(--jp-border-width));overflow:visible}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab.p-mod-current:before{position:absolute;top:calc(-1*var(--jp-border-width));left:calc(-1*var(--jp-border-width));content:\"\";height:var(--jp-widgets-horizontal-tab-top-border);width:calc(100% + 2*var(--jp-border-width));background:var(--jp-brand-color1)}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab:first-child{margin-left:0}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tab:hover:not(.p-mod-current){background:var(--jp-layout-color1);color:var(--jp-ui-font-color1)}.jupyter-widgets.widget-tab>.p-TabBar .p-mod-closable>.p-TabBar-tabCloseIcon{margin-left:4px}.jupyter-widgets.widget-tab>.p-TabBar .p-mod-closable>.p-TabBar-tabCloseIcon:before{font-family:FontAwesome;content:\"\\f00d\"}.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tabCloseIcon,.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tabIcon,.jupyter-widgets.widget-tab>.p-TabBar .p-TabBar-tabLabel{line-height:var(--jp-widgets-horizontal-tab-height)}.p-Collapse{display:flex;flex-direction:column;align-items:stretch}.p-Collapse-header{cursor:pointer;color:var(--jp-ui-font-color2);background-color:var(--jp-layout-color2);border:var(--jp-widgets-border-width) solid var(--jp-border-color1);padding:calc(var(--jp-widgets-container-padding)*2/3) var(--jp-widgets-container-padding);font-weight:700}.p-Collapse-header:hover{background-color:var(--jp-layout-color1);color:var(--jp-ui-font-color1)}.p-Collapse-open>.p-Collapse-header{background-color:var(--jp-layout-color1);color:var(--jp-ui-font-color0);cursor:default;border-bottom:none}.p-Collapse-contents{padding:var(--jp-widgets-container-padding);background-color:var(--jp-layout-color1);color:var(--jp-ui-font-color1);border-left:var(--jp-widgets-border-width) solid var(--jp-border-color1);border-right:var(--jp-widgets-border-width) solid var(--jp-border-color1);border-bottom:var(--jp-widgets-border-width) solid var(--jp-border-color1);overflow:auto}.p-Accordion{display:flex;flex-direction:column;align-items:stretch}.p-Accordion .p-Collapse{margin-bottom:0}.p-Accordion .p-Collapse+.p-Collapse{margin-top:4px}.widget-html,.widget-htmlmath{font-size:var(--jp-widgets-font-size)}.widget-html>.widget-html-content,.widget-htmlmath>.widget-html-content{align-self:stretch;flex-grow:1;flex-shrink:1;line-height:var(--jp-widgets-inline-height);position:relative}.widget-image{max-width:100%;height:auto}";
  styleInject(css$4);

  var widgets = /*#__PURE__*/Object.freeze({
    'default': css$4
  });

  getCjsExportFromNamespace(style);

  getCjsExportFromNamespace(css$3);

  getCjsExportFromNamespace(widgets);

  window['requirejs'].config({
    map: {
      '*': {
        '@jupyter-widgets/controls': 'nbextensions/jupyter-js-widgets/extension',
        '@jupyter-widgets/base': 'nbextensions/jupyter-js-widgets/extension',
        '@jupyter-widgets/output': 'nbextensions/jupyter-js-widgets/extension'
      }
    }
  });

  var MIME_TYPE$1 = 'application/vnd.jupyter.widget-view+json';
  var CLASS_NAME = 'jupyter-widgets-view';






  /**
   * Create a widget manager for a kernel instance.
   */
  var handle_kernel = function(Jupyter, kernel) {
    if (kernel.comm_manager && kernel.widget_manager === undefined) {
      // Clear any old widget manager
      if (Jupyter.WidgetManager) {
        Jupyter.WidgetManager._managers[0].clear_state();
      }

      // Create a new widget manager instance. Use the global
      // Jupyter.notebook handle.
      var manager$1 = new manager.WidgetManager(kernel.comm_manager, Jupyter.notebook);

      // For backwards compatibility and interactive use.
      Jupyter.WidgetManager = manager.WidgetManager;

      // Store a handle to the manager so we know not to
      // another for this kernel. This also is a convenience
      // for the user.
      kernel.widget_manager = manager$1;
    }
  };

  function register_events(Jupyter, events, outputarea) {
    // If a kernel already exists, create a widget manager.
    if (Jupyter.notebook && Jupyter.notebook.kernel) {
      handle_kernel(Jupyter, Jupyter.notebook.kernel);
    }
    // When the kernel is created, create a widget manager.
    events.on('kernel_created.Kernel kernel_created.Session', function(
      event,
      data
    ) {
      handle_kernel(Jupyter, data.kernel);
    });

    // When a kernel dies, disconnect the widgets.
    events.on(
      'kernel_killed.Session kernel_killed.Kernel kernel_restarting.Kernel',
      function(event, data) {
        var kernel = data.kernel;
        if (kernel && kernel.widget_manager) {
          kernel.widget_manager.disconnect();
        }
      }
    );

    /**
     * The views on this page. We keep this list so that we can call the view.remove()
     * method when a view is removed from the page.
     */
    var views = {};
    var removeView = function(event, data) {
      var output = data.cell ? data.cell.output_area : data.output_area;
      var viewids = output ? output._jupyterWidgetViews : void 0;
      if (viewids) {
        viewids.forEach(function(id) {
          // this may be called after the widget is pulled off the page
          // so we temporarily put it back on the page as a kludge
          // so that Lumino can trigger the appropriate detach signals
          var view = views[id];
          view.el.style.display = 'none';
          document.body.appendChild(view.el);
          view.remove();
          delete views[id];
        });
        output._jupyterWidgetViews = [];
      }
    };

    // Deleting a cell does *not* clear the outputs first.
    events.on('delete.Cell', removeView);
    // add an event to the notebook element for *any* outputs that are cleared.
    Jupyter.notebook.container.on('clearing', '.output', removeView);

    // For before https://github.com/jupyter/notebook/pull/2411 is merged and
    // released. This does not handle the case where an empty cell is executed
    // to clear input.
    events.on('execute.CodeCell', removeView);
    events.on('clear_output.CodeCell', removeView);

    /**
     * Render data to the output area.
     */
    function render(output, data, node) {
      // data is a model id
      var manager =
        Jupyter.notebook &&
        Jupyter.notebook.kernel &&
        Jupyter.notebook.kernel.widget_manager;
      if (!manager) {
        node.textContent =
          'Error rendering Jupyter widget: missing widget manager';
        return;
      }

      // Missing model id means the view was removed. Hide this element.
      if (data.model_id === '') {
        node.style.display = 'none';
        return;
      }

      var model = manager.get_model(data.model_id);
      if (model) {
        model
          .then(function(model) {
            return manager.create_view(model, { output: output });
          })
          .then(function(view) {
            var id = view.cid;
            output._jupyterWidgetViews = output._jupyterWidgetViews || [];
            output._jupyterWidgetViews.push(id);
            views[id] = view;
            index_es6.Widget.attach(view.pWidget, node);

            // Make the node completely disappear if the view is removed.
            view.once('remove', () => {
              // Since we have a mutable reference to the data, delete the
              // model id to indicate the view is removed.
              data.model_id = '';
              node.style.display = 'none';
            });
          });
      } else {
        node.textContent =
          'A Jupyter widget could not be displayed because the widget state could not be found. This could happen if the kernel storing the widget is no longer available, or if the widget state was not saved in the notebook. You may be able to create the widget by running the appropriate cells.';
      }
    }

    // `this` is the output area we are appending to
    var append_mime = function(json, md, element) {
      var toinsert = this.create_output_subarea(md, CLASS_NAME, MIME_TYPE$1);
      this.keyboard_manager.register_events(toinsert);
      render(this, json, toinsert[0]);
      element.append(toinsert);
      return toinsert;
    };
    // Register mime type with the output area
    outputarea.OutputArea.prototype.register_mime_type(MIME_TYPE$1, append_mime, {
      // An output widget could contain arbitrary user javascript
      safe: false,
      // Index of renderer in `output_area.display_order`
      index: 0
    });
  }

  function load_ipython_extension() {
    return new Promise(function(resolve) {
      requirejs(
        ['base/js/namespace', 'base/js/events', 'notebook/js/outputarea'],
        function(Jupyter, events, outputarea) {
          
          
          
          register_events(Jupyter, events, outputarea);
          resolve();
        }
      );
    });
  }

  var extension = {
    load_ipython_extension: load_ipython_extension,
    ...lib$3,
    ...lib$1,
    ...widget_output$1
  };
  var extension_1 = extension.load_ipython_extension;

  var lodash = createCommonjsModule(function (module, exports) {
  (function() {

    /** Used as a safe reference for `undefined` in pre-ES5 environments. */
    var undefined$1;

    /** Used as the semantic version number. */
    var VERSION = '4.17.15';

    /** Used as the size to enable large array optimizations. */
    var LARGE_ARRAY_SIZE = 200;

    /** Error message constants. */
    var CORE_ERROR_TEXT = 'Unsupported core-js use. Try https://npms.io/search?q=ponyfill.',
        FUNC_ERROR_TEXT = 'Expected a function';

    /** Used to stand-in for `undefined` hash values. */
    var HASH_UNDEFINED = '__lodash_hash_undefined__';

    /** Used as the maximum memoize cache size. */
    var MAX_MEMOIZE_SIZE = 500;

    /** Used as the internal argument placeholder. */
    var PLACEHOLDER = '__lodash_placeholder__';

    /** Used to compose bitmasks for cloning. */
    var CLONE_DEEP_FLAG = 1,
        CLONE_FLAT_FLAG = 2,
        CLONE_SYMBOLS_FLAG = 4;

    /** Used to compose bitmasks for value comparisons. */
    var COMPARE_PARTIAL_FLAG = 1,
        COMPARE_UNORDERED_FLAG = 2;

    /** Used to compose bitmasks for function metadata. */
    var WRAP_BIND_FLAG = 1,
        WRAP_BIND_KEY_FLAG = 2,
        WRAP_CURRY_BOUND_FLAG = 4,
        WRAP_CURRY_FLAG = 8,
        WRAP_CURRY_RIGHT_FLAG = 16,
        WRAP_PARTIAL_FLAG = 32,
        WRAP_PARTIAL_RIGHT_FLAG = 64,
        WRAP_ARY_FLAG = 128,
        WRAP_REARG_FLAG = 256,
        WRAP_FLIP_FLAG = 512;

    /** Used as default options for `_.truncate`. */
    var DEFAULT_TRUNC_LENGTH = 30,
        DEFAULT_TRUNC_OMISSION = '...';

    /** Used to detect hot functions by number of calls within a span of milliseconds. */
    var HOT_COUNT = 800,
        HOT_SPAN = 16;

    /** Used to indicate the type of lazy iteratees. */
    var LAZY_FILTER_FLAG = 1,
        LAZY_MAP_FLAG = 2,
        LAZY_WHILE_FLAG = 3;

    /** Used as references for various `Number` constants. */
    var INFINITY = 1 / 0,
        MAX_SAFE_INTEGER = 9007199254740991,
        MAX_INTEGER = 1.7976931348623157e+308,
        NAN = 0 / 0;

    /** Used as references for the maximum length and index of an array. */
    var MAX_ARRAY_LENGTH = 4294967295,
        MAX_ARRAY_INDEX = MAX_ARRAY_LENGTH - 1,
        HALF_MAX_ARRAY_LENGTH = MAX_ARRAY_LENGTH >>> 1;

    /** Used to associate wrap methods with their bit flags. */
    var wrapFlags = [
      ['ary', WRAP_ARY_FLAG],
      ['bind', WRAP_BIND_FLAG],
      ['bindKey', WRAP_BIND_KEY_FLAG],
      ['curry', WRAP_CURRY_FLAG],
      ['curryRight', WRAP_CURRY_RIGHT_FLAG],
      ['flip', WRAP_FLIP_FLAG],
      ['partial', WRAP_PARTIAL_FLAG],
      ['partialRight', WRAP_PARTIAL_RIGHT_FLAG],
      ['rearg', WRAP_REARG_FLAG]
    ];

    /** `Object#toString` result references. */
    var argsTag = '[object Arguments]',
        arrayTag = '[object Array]',
        asyncTag = '[object AsyncFunction]',
        boolTag = '[object Boolean]',
        dateTag = '[object Date]',
        domExcTag = '[object DOMException]',
        errorTag = '[object Error]',
        funcTag = '[object Function]',
        genTag = '[object GeneratorFunction]',
        mapTag = '[object Map]',
        numberTag = '[object Number]',
        nullTag = '[object Null]',
        objectTag = '[object Object]',
        promiseTag = '[object Promise]',
        proxyTag = '[object Proxy]',
        regexpTag = '[object RegExp]',
        setTag = '[object Set]',
        stringTag = '[object String]',
        symbolTag = '[object Symbol]',
        undefinedTag = '[object Undefined]',
        weakMapTag = '[object WeakMap]',
        weakSetTag = '[object WeakSet]';

    var arrayBufferTag = '[object ArrayBuffer]',
        dataViewTag = '[object DataView]',
        float32Tag = '[object Float32Array]',
        float64Tag = '[object Float64Array]',
        int8Tag = '[object Int8Array]',
        int16Tag = '[object Int16Array]',
        int32Tag = '[object Int32Array]',
        uint8Tag = '[object Uint8Array]',
        uint8ClampedTag = '[object Uint8ClampedArray]',
        uint16Tag = '[object Uint16Array]',
        uint32Tag = '[object Uint32Array]';

    /** Used to match empty string literals in compiled template source. */
    var reEmptyStringLeading = /\b__p \+= '';/g,
        reEmptyStringMiddle = /\b(__p \+=) '' \+/g,
        reEmptyStringTrailing = /(__e\(.*?\)|\b__t\)) \+\n'';/g;

    /** Used to match HTML entities and HTML characters. */
    var reEscapedHtml = /&(?:amp|lt|gt|quot|#39);/g,
        reUnescapedHtml = /[&<>"']/g,
        reHasEscapedHtml = RegExp(reEscapedHtml.source),
        reHasUnescapedHtml = RegExp(reUnescapedHtml.source);

    /** Used to match template delimiters. */
    var reEscape = /<%-([\s\S]+?)%>/g,
        reEvaluate = /<%([\s\S]+?)%>/g,
        reInterpolate = /<%=([\s\S]+?)%>/g;

    /** Used to match property names within property paths. */
    var reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,
        reIsPlainProp = /^\w*$/,
        rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;

    /**
     * Used to match `RegExp`
     * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
     */
    var reRegExpChar = /[\\^$.*+?()[\]{}|]/g,
        reHasRegExpChar = RegExp(reRegExpChar.source);

    /** Used to match leading and trailing whitespace. */
    var reTrim = /^\s+|\s+$/g,
        reTrimStart = /^\s+/,
        reTrimEnd = /\s+$/;

    /** Used to match wrap detail comments. */
    var reWrapComment = /\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,
        reWrapDetails = /\{\n\/\* \[wrapped with (.+)\] \*/,
        reSplitDetails = /,? & /;

    /** Used to match words composed of alphanumeric characters. */
    var reAsciiWord = /[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g;

    /** Used to match backslashes in property paths. */
    var reEscapeChar = /\\(\\)?/g;

    /**
     * Used to match
     * [ES template delimiters](http://ecma-international.org/ecma-262/7.0/#sec-template-literal-lexical-components).
     */
    var reEsTemplate = /\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g;

    /** Used to match `RegExp` flags from their coerced string values. */
    var reFlags = /\w*$/;

    /** Used to detect bad signed hexadecimal string values. */
    var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;

    /** Used to detect binary string values. */
    var reIsBinary = /^0b[01]+$/i;

    /** Used to detect host constructors (Safari). */
    var reIsHostCtor = /^\[object .+?Constructor\]$/;

    /** Used to detect octal string values. */
    var reIsOctal = /^0o[0-7]+$/i;

    /** Used to detect unsigned integer values. */
    var reIsUint = /^(?:0|[1-9]\d*)$/;

    /** Used to match Latin Unicode letters (excluding mathematical operators). */
    var reLatin = /[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g;

    /** Used to ensure capturing order of template delimiters. */
    var reNoMatch = /($^)/;

    /** Used to match unescaped characters in compiled string literals. */
    var reUnescapedString = /['\n\r\u2028\u2029\\]/g;

    /** Used to compose unicode character classes. */
    var rsAstralRange = '\\ud800-\\udfff',
        rsComboMarksRange = '\\u0300-\\u036f',
        reComboHalfMarksRange = '\\ufe20-\\ufe2f',
        rsComboSymbolsRange = '\\u20d0-\\u20ff',
        rsComboRange = rsComboMarksRange + reComboHalfMarksRange + rsComboSymbolsRange,
        rsDingbatRange = '\\u2700-\\u27bf',
        rsLowerRange = 'a-z\\xdf-\\xf6\\xf8-\\xff',
        rsMathOpRange = '\\xac\\xb1\\xd7\\xf7',
        rsNonCharRange = '\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf',
        rsPunctuationRange = '\\u2000-\\u206f',
        rsSpaceRange = ' \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000',
        rsUpperRange = 'A-Z\\xc0-\\xd6\\xd8-\\xde',
        rsVarRange = '\\ufe0e\\ufe0f',
        rsBreakRange = rsMathOpRange + rsNonCharRange + rsPunctuationRange + rsSpaceRange;

    /** Used to compose unicode capture groups. */
    var rsApos = "['\u2019]",
        rsAstral = '[' + rsAstralRange + ']',
        rsBreak = '[' + rsBreakRange + ']',
        rsCombo = '[' + rsComboRange + ']',
        rsDigits = '\\d+',
        rsDingbat = '[' + rsDingbatRange + ']',
        rsLower = '[' + rsLowerRange + ']',
        rsMisc = '[^' + rsAstralRange + rsBreakRange + rsDigits + rsDingbatRange + rsLowerRange + rsUpperRange + ']',
        rsFitz = '\\ud83c[\\udffb-\\udfff]',
        rsModifier = '(?:' + rsCombo + '|' + rsFitz + ')',
        rsNonAstral = '[^' + rsAstralRange + ']',
        rsRegional = '(?:\\ud83c[\\udde6-\\uddff]){2}',
        rsSurrPair = '[\\ud800-\\udbff][\\udc00-\\udfff]',
        rsUpper = '[' + rsUpperRange + ']',
        rsZWJ = '\\u200d';

    /** Used to compose unicode regexes. */
    var rsMiscLower = '(?:' + rsLower + '|' + rsMisc + ')',
        rsMiscUpper = '(?:' + rsUpper + '|' + rsMisc + ')',
        rsOptContrLower = '(?:' + rsApos + '(?:d|ll|m|re|s|t|ve))?',
        rsOptContrUpper = '(?:' + rsApos + '(?:D|LL|M|RE|S|T|VE))?',
        reOptMod = rsModifier + '?',
        rsOptVar = '[' + rsVarRange + ']?',
        rsOptJoin = '(?:' + rsZWJ + '(?:' + [rsNonAstral, rsRegional, rsSurrPair].join('|') + ')' + rsOptVar + reOptMod + ')*',
        rsOrdLower = '\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])',
        rsOrdUpper = '\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])',
        rsSeq = rsOptVar + reOptMod + rsOptJoin,
        rsEmoji = '(?:' + [rsDingbat, rsRegional, rsSurrPair].join('|') + ')' + rsSeq,
        rsSymbol = '(?:' + [rsNonAstral + rsCombo + '?', rsCombo, rsRegional, rsSurrPair, rsAstral].join('|') + ')';

    /** Used to match apostrophes. */
    var reApos = RegExp(rsApos, 'g');

    /**
     * Used to match [combining diacritical marks](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks) and
     * [combining diacritical marks for symbols](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks_for_Symbols).
     */
    var reComboMark = RegExp(rsCombo, 'g');

    /** Used to match [string symbols](https://mathiasbynens.be/notes/javascript-unicode). */
    var reUnicode = RegExp(rsFitz + '(?=' + rsFitz + ')|' + rsSymbol + rsSeq, 'g');

    /** Used to match complex or compound words. */
    var reUnicodeWord = RegExp([
      rsUpper + '?' + rsLower + '+' + rsOptContrLower + '(?=' + [rsBreak, rsUpper, '$'].join('|') + ')',
      rsMiscUpper + '+' + rsOptContrUpper + '(?=' + [rsBreak, rsUpper + rsMiscLower, '$'].join('|') + ')',
      rsUpper + '?' + rsMiscLower + '+' + rsOptContrLower,
      rsUpper + '+' + rsOptContrUpper,
      rsOrdUpper,
      rsOrdLower,
      rsDigits,
      rsEmoji
    ].join('|'), 'g');

    /** Used to detect strings with [zero-width joiners or code points from the astral planes](http://eev.ee/blog/2015/09/12/dark-corners-of-unicode/). */
    var reHasUnicode = RegExp('[' + rsZWJ + rsAstralRange  + rsComboRange + rsVarRange + ']');

    /** Used to detect strings that need a more robust regexp to match words. */
    var reHasUnicodeWord = /[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/;

    /** Used to assign default `context` object properties. */
    var contextProps = [
      'Array', 'Buffer', 'DataView', 'Date', 'Error', 'Float32Array', 'Float64Array',
      'Function', 'Int8Array', 'Int16Array', 'Int32Array', 'Map', 'Math', 'Object',
      'Promise', 'RegExp', 'Set', 'String', 'Symbol', 'TypeError', 'Uint8Array',
      'Uint8ClampedArray', 'Uint16Array', 'Uint32Array', 'WeakMap',
      '_', 'clearTimeout', 'isFinite', 'parseInt', 'setTimeout'
    ];

    /** Used to make template sourceURLs easier to identify. */
    var templateCounter = -1;

    /** Used to identify `toStringTag` values of typed arrays. */
    var typedArrayTags = {};
    typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
    typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
    typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
    typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
    typedArrayTags[uint32Tag] = true;
    typedArrayTags[argsTag] = typedArrayTags[arrayTag] =
    typedArrayTags[arrayBufferTag] = typedArrayTags[boolTag] =
    typedArrayTags[dataViewTag] = typedArrayTags[dateTag] =
    typedArrayTags[errorTag] = typedArrayTags[funcTag] =
    typedArrayTags[mapTag] = typedArrayTags[numberTag] =
    typedArrayTags[objectTag] = typedArrayTags[regexpTag] =
    typedArrayTags[setTag] = typedArrayTags[stringTag] =
    typedArrayTags[weakMapTag] = false;

    /** Used to identify `toStringTag` values supported by `_.clone`. */
    var cloneableTags = {};
    cloneableTags[argsTag] = cloneableTags[arrayTag] =
    cloneableTags[arrayBufferTag] = cloneableTags[dataViewTag] =
    cloneableTags[boolTag] = cloneableTags[dateTag] =
    cloneableTags[float32Tag] = cloneableTags[float64Tag] =
    cloneableTags[int8Tag] = cloneableTags[int16Tag] =
    cloneableTags[int32Tag] = cloneableTags[mapTag] =
    cloneableTags[numberTag] = cloneableTags[objectTag] =
    cloneableTags[regexpTag] = cloneableTags[setTag] =
    cloneableTags[stringTag] = cloneableTags[symbolTag] =
    cloneableTags[uint8Tag] = cloneableTags[uint8ClampedTag] =
    cloneableTags[uint16Tag] = cloneableTags[uint32Tag] = true;
    cloneableTags[errorTag] = cloneableTags[funcTag] =
    cloneableTags[weakMapTag] = false;

    /** Used to map Latin Unicode letters to basic Latin letters. */
    var deburredLetters = {
      // Latin-1 Supplement block.
      '\xc0': 'A',  '\xc1': 'A', '\xc2': 'A', '\xc3': 'A', '\xc4': 'A', '\xc5': 'A',
      '\xe0': 'a',  '\xe1': 'a', '\xe2': 'a', '\xe3': 'a', '\xe4': 'a', '\xe5': 'a',
      '\xc7': 'C',  '\xe7': 'c',
      '\xd0': 'D',  '\xf0': 'd',
      '\xc8': 'E',  '\xc9': 'E', '\xca': 'E', '\xcb': 'E',
      '\xe8': 'e',  '\xe9': 'e', '\xea': 'e', '\xeb': 'e',
      '\xcc': 'I',  '\xcd': 'I', '\xce': 'I', '\xcf': 'I',
      '\xec': 'i',  '\xed': 'i', '\xee': 'i', '\xef': 'i',
      '\xd1': 'N',  '\xf1': 'n',
      '\xd2': 'O',  '\xd3': 'O', '\xd4': 'O', '\xd5': 'O', '\xd6': 'O', '\xd8': 'O',
      '\xf2': 'o',  '\xf3': 'o', '\xf4': 'o', '\xf5': 'o', '\xf6': 'o', '\xf8': 'o',
      '\xd9': 'U',  '\xda': 'U', '\xdb': 'U', '\xdc': 'U',
      '\xf9': 'u',  '\xfa': 'u', '\xfb': 'u', '\xfc': 'u',
      '\xdd': 'Y',  '\xfd': 'y', '\xff': 'y',
      '\xc6': 'Ae', '\xe6': 'ae',
      '\xde': 'Th', '\xfe': 'th',
      '\xdf': 'ss',
      // Latin Extended-A block.
      '\u0100': 'A',  '\u0102': 'A', '\u0104': 'A',
      '\u0101': 'a',  '\u0103': 'a', '\u0105': 'a',
      '\u0106': 'C',  '\u0108': 'C', '\u010a': 'C', '\u010c': 'C',
      '\u0107': 'c',  '\u0109': 'c', '\u010b': 'c', '\u010d': 'c',
      '\u010e': 'D',  '\u0110': 'D', '\u010f': 'd', '\u0111': 'd',
      '\u0112': 'E',  '\u0114': 'E', '\u0116': 'E', '\u0118': 'E', '\u011a': 'E',
      '\u0113': 'e',  '\u0115': 'e', '\u0117': 'e', '\u0119': 'e', '\u011b': 'e',
      '\u011c': 'G',  '\u011e': 'G', '\u0120': 'G', '\u0122': 'G',
      '\u011d': 'g',  '\u011f': 'g', '\u0121': 'g', '\u0123': 'g',
      '\u0124': 'H',  '\u0126': 'H', '\u0125': 'h', '\u0127': 'h',
      '\u0128': 'I',  '\u012a': 'I', '\u012c': 'I', '\u012e': 'I', '\u0130': 'I',
      '\u0129': 'i',  '\u012b': 'i', '\u012d': 'i', '\u012f': 'i', '\u0131': 'i',
      '\u0134': 'J',  '\u0135': 'j',
      '\u0136': 'K',  '\u0137': 'k', '\u0138': 'k',
      '\u0139': 'L',  '\u013b': 'L', '\u013d': 'L', '\u013f': 'L', '\u0141': 'L',
      '\u013a': 'l',  '\u013c': 'l', '\u013e': 'l', '\u0140': 'l', '\u0142': 'l',
      '\u0143': 'N',  '\u0145': 'N', '\u0147': 'N', '\u014a': 'N',
      '\u0144': 'n',  '\u0146': 'n', '\u0148': 'n', '\u014b': 'n',
      '\u014c': 'O',  '\u014e': 'O', '\u0150': 'O',
      '\u014d': 'o',  '\u014f': 'o', '\u0151': 'o',
      '\u0154': 'R',  '\u0156': 'R', '\u0158': 'R',
      '\u0155': 'r',  '\u0157': 'r', '\u0159': 'r',
      '\u015a': 'S',  '\u015c': 'S', '\u015e': 'S', '\u0160': 'S',
      '\u015b': 's',  '\u015d': 's', '\u015f': 's', '\u0161': 's',
      '\u0162': 'T',  '\u0164': 'T', '\u0166': 'T',
      '\u0163': 't',  '\u0165': 't', '\u0167': 't',
      '\u0168': 'U',  '\u016a': 'U', '\u016c': 'U', '\u016e': 'U', '\u0170': 'U', '\u0172': 'U',
      '\u0169': 'u',  '\u016b': 'u', '\u016d': 'u', '\u016f': 'u', '\u0171': 'u', '\u0173': 'u',
      '\u0174': 'W',  '\u0175': 'w',
      '\u0176': 'Y',  '\u0177': 'y', '\u0178': 'Y',
      '\u0179': 'Z',  '\u017b': 'Z', '\u017d': 'Z',
      '\u017a': 'z',  '\u017c': 'z', '\u017e': 'z',
      '\u0132': 'IJ', '\u0133': 'ij',
      '\u0152': 'Oe', '\u0153': 'oe',
      '\u0149': "'n", '\u017f': 's'
    };

    /** Used to map characters to HTML entities. */
    var htmlEscapes = {
      '&': '&amp;',
      '<': '&lt;',
      '>': '&gt;',
      '"': '&quot;',
      "'": '&#39;'
    };

    /** Used to map HTML entities to characters. */
    var htmlUnescapes = {
      '&amp;': '&',
      '&lt;': '<',
      '&gt;': '>',
      '&quot;': '"',
      '&#39;': "'"
    };

    /** Used to escape characters for inclusion in compiled string literals. */
    var stringEscapes = {
      '\\': '\\',
      "'": "'",
      '\n': 'n',
      '\r': 'r',
      '\u2028': 'u2028',
      '\u2029': 'u2029'
    };

    /** Built-in method references without a dependency on `root`. */
    var freeParseFloat = parseFloat,
        freeParseInt = parseInt;

    /** Detect free variable `global` from Node.js. */
    var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

    /** Detect free variable `self`. */
    var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

    /** Used as a reference to the global object. */
    var root = freeGlobal || freeSelf || Function('return this')();

    /** Detect free variable `exports`. */
    var freeExports = exports && !exports.nodeType && exports;

    /** Detect free variable `module`. */
    var freeModule = freeExports && 'object' == 'object' && module && !module.nodeType && module;

    /** Detect the popular CommonJS extension `module.exports`. */
    var moduleExports = freeModule && freeModule.exports === freeExports;

    /** Detect free variable `process` from Node.js. */
    var freeProcess = moduleExports && freeGlobal.process;

    /** Used to access faster Node.js helpers. */
    var nodeUtil = (function() {
      try {
        // Use `util.types` for Node.js 10+.
        var types = freeModule && freeModule.require && freeModule.require('util').types;

        if (types) {
          return types;
        }

        // Legacy `process.binding('util')` for Node.js < 10.
        return freeProcess && freeProcess.binding && freeProcess.binding('util');
      } catch (e) {}
    }());

    /* Node.js helper references. */
    var nodeIsArrayBuffer = nodeUtil && nodeUtil.isArrayBuffer,
        nodeIsDate = nodeUtil && nodeUtil.isDate,
        nodeIsMap = nodeUtil && nodeUtil.isMap,
        nodeIsRegExp = nodeUtil && nodeUtil.isRegExp,
        nodeIsSet = nodeUtil && nodeUtil.isSet,
        nodeIsTypedArray = nodeUtil && nodeUtil.isTypedArray;

    /*--------------------------------------------------------------------------*/

    /**
     * A faster alternative to `Function#apply`, this function invokes `func`
     * with the `this` binding of `thisArg` and the arguments of `args`.
     *
     * @private
     * @param {Function} func The function to invoke.
     * @param {*} thisArg The `this` binding of `func`.
     * @param {Array} args The arguments to invoke `func` with.
     * @returns {*} Returns the result of `func`.
     */
    function apply(func, thisArg, args) {
      switch (args.length) {
        case 0: return func.call(thisArg);
        case 1: return func.call(thisArg, args[0]);
        case 2: return func.call(thisArg, args[0], args[1]);
        case 3: return func.call(thisArg, args[0], args[1], args[2]);
      }
      return func.apply(thisArg, args);
    }

    /**
     * A specialized version of `baseAggregator` for arrays.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} setter The function to set `accumulator` values.
     * @param {Function} iteratee The iteratee to transform keys.
     * @param {Object} accumulator The initial aggregated object.
     * @returns {Function} Returns `accumulator`.
     */
    function arrayAggregator(array, setter, iteratee, accumulator) {
      var index = -1,
          length = array == null ? 0 : array.length;

      while (++index < length) {
        var value = array[index];
        setter(accumulator, value, iteratee(value), array);
      }
      return accumulator;
    }

    /**
     * A specialized version of `_.forEach` for arrays without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array} Returns `array`.
     */
    function arrayEach(array, iteratee) {
      var index = -1,
          length = array == null ? 0 : array.length;

      while (++index < length) {
        if (iteratee(array[index], index, array) === false) {
          break;
        }
      }
      return array;
    }

    /**
     * A specialized version of `_.forEachRight` for arrays without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array} Returns `array`.
     */
    function arrayEachRight(array, iteratee) {
      var length = array == null ? 0 : array.length;

      while (length--) {
        if (iteratee(array[length], length, array) === false) {
          break;
        }
      }
      return array;
    }

    /**
     * A specialized version of `_.every` for arrays without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} predicate The function invoked per iteration.
     * @returns {boolean} Returns `true` if all elements pass the predicate check,
     *  else `false`.
     */
    function arrayEvery(array, predicate) {
      var index = -1,
          length = array == null ? 0 : array.length;

      while (++index < length) {
        if (!predicate(array[index], index, array)) {
          return false;
        }
      }
      return true;
    }

    /**
     * A specialized version of `_.filter` for arrays without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} predicate The function invoked per iteration.
     * @returns {Array} Returns the new filtered array.
     */
    function arrayFilter(array, predicate) {
      var index = -1,
          length = array == null ? 0 : array.length,
          resIndex = 0,
          result = [];

      while (++index < length) {
        var value = array[index];
        if (predicate(value, index, array)) {
          result[resIndex++] = value;
        }
      }
      return result;
    }

    /**
     * A specialized version of `_.includes` for arrays without support for
     * specifying an index to search from.
     *
     * @private
     * @param {Array} [array] The array to inspect.
     * @param {*} target The value to search for.
     * @returns {boolean} Returns `true` if `target` is found, else `false`.
     */
    function arrayIncludes(array, value) {
      var length = array == null ? 0 : array.length;
      return !!length && baseIndexOf(array, value, 0) > -1;
    }

    /**
     * This function is like `arrayIncludes` except that it accepts a comparator.
     *
     * @private
     * @param {Array} [array] The array to inspect.
     * @param {*} target The value to search for.
     * @param {Function} comparator The comparator invoked per element.
     * @returns {boolean} Returns `true` if `target` is found, else `false`.
     */
    function arrayIncludesWith(array, value, comparator) {
      var index = -1,
          length = array == null ? 0 : array.length;

      while (++index < length) {
        if (comparator(value, array[index])) {
          return true;
        }
      }
      return false;
    }

    /**
     * A specialized version of `_.map` for arrays without support for iteratee
     * shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array} Returns the new mapped array.
     */
    function arrayMap(array, iteratee) {
      var index = -1,
          length = array == null ? 0 : array.length,
          result = Array(length);

      while (++index < length) {
        result[index] = iteratee(array[index], index, array);
      }
      return result;
    }

    /**
     * Appends the elements of `values` to `array`.
     *
     * @private
     * @param {Array} array The array to modify.
     * @param {Array} values The values to append.
     * @returns {Array} Returns `array`.
     */
    function arrayPush(array, values) {
      var index = -1,
          length = values.length,
          offset = array.length;

      while (++index < length) {
        array[offset + index] = values[index];
      }
      return array;
    }

    /**
     * A specialized version of `_.reduce` for arrays without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @param {*} [accumulator] The initial value.
     * @param {boolean} [initAccum] Specify using the first element of `array` as
     *  the initial value.
     * @returns {*} Returns the accumulated value.
     */
    function arrayReduce(array, iteratee, accumulator, initAccum) {
      var index = -1,
          length = array == null ? 0 : array.length;

      if (initAccum && length) {
        accumulator = array[++index];
      }
      while (++index < length) {
        accumulator = iteratee(accumulator, array[index], index, array);
      }
      return accumulator;
    }

    /**
     * A specialized version of `_.reduceRight` for arrays without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @param {*} [accumulator] The initial value.
     * @param {boolean} [initAccum] Specify using the last element of `array` as
     *  the initial value.
     * @returns {*} Returns the accumulated value.
     */
    function arrayReduceRight(array, iteratee, accumulator, initAccum) {
      var length = array == null ? 0 : array.length;
      if (initAccum && length) {
        accumulator = array[--length];
      }
      while (length--) {
        accumulator = iteratee(accumulator, array[length], length, array);
      }
      return accumulator;
    }

    /**
     * A specialized version of `_.some` for arrays without support for iteratee
     * shorthands.
     *
     * @private
     * @param {Array} [array] The array to iterate over.
     * @param {Function} predicate The function invoked per iteration.
     * @returns {boolean} Returns `true` if any element passes the predicate check,
     *  else `false`.
     */
    function arraySome(array, predicate) {
      var index = -1,
          length = array == null ? 0 : array.length;

      while (++index < length) {
        if (predicate(array[index], index, array)) {
          return true;
        }
      }
      return false;
    }

    /**
     * Gets the size of an ASCII `string`.
     *
     * @private
     * @param {string} string The string inspect.
     * @returns {number} Returns the string size.
     */
    var asciiSize = baseProperty('length');

    /**
     * Converts an ASCII `string` to an array.
     *
     * @private
     * @param {string} string The string to convert.
     * @returns {Array} Returns the converted array.
     */
    function asciiToArray(string) {
      return string.split('');
    }

    /**
     * Splits an ASCII `string` into an array of its words.
     *
     * @private
     * @param {string} The string to inspect.
     * @returns {Array} Returns the words of `string`.
     */
    function asciiWords(string) {
      return string.match(reAsciiWord) || [];
    }

    /**
     * The base implementation of methods like `_.findKey` and `_.findLastKey`,
     * without support for iteratee shorthands, which iterates over `collection`
     * using `eachFunc`.
     *
     * @private
     * @param {Array|Object} collection The collection to inspect.
     * @param {Function} predicate The function invoked per iteration.
     * @param {Function} eachFunc The function to iterate over `collection`.
     * @returns {*} Returns the found element or its key, else `undefined`.
     */
    function baseFindKey(collection, predicate, eachFunc) {
      var result;
      eachFunc(collection, function(value, key, collection) {
        if (predicate(value, key, collection)) {
          result = key;
          return false;
        }
      });
      return result;
    }

    /**
     * The base implementation of `_.findIndex` and `_.findLastIndex` without
     * support for iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {Function} predicate The function invoked per iteration.
     * @param {number} fromIndex The index to search from.
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function baseFindIndex(array, predicate, fromIndex, fromRight) {
      var length = array.length,
          index = fromIndex + (fromRight ? 1 : -1);

      while ((fromRight ? index-- : ++index < length)) {
        if (predicate(array[index], index, array)) {
          return index;
        }
      }
      return -1;
    }

    /**
     * The base implementation of `_.indexOf` without `fromIndex` bounds checks.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @param {number} fromIndex The index to search from.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function baseIndexOf(array, value, fromIndex) {
      return value === value
        ? strictIndexOf(array, value, fromIndex)
        : baseFindIndex(array, baseIsNaN, fromIndex);
    }

    /**
     * This function is like `baseIndexOf` except that it accepts a comparator.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @param {number} fromIndex The index to search from.
     * @param {Function} comparator The comparator invoked per element.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function baseIndexOfWith(array, value, fromIndex, comparator) {
      var index = fromIndex - 1,
          length = array.length;

      while (++index < length) {
        if (comparator(array[index], value)) {
          return index;
        }
      }
      return -1;
    }

    /**
     * The base implementation of `_.isNaN` without support for number objects.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is `NaN`, else `false`.
     */
    function baseIsNaN(value) {
      return value !== value;
    }

    /**
     * The base implementation of `_.mean` and `_.meanBy` without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {number} Returns the mean.
     */
    function baseMean(array, iteratee) {
      var length = array == null ? 0 : array.length;
      return length ? (baseSum(array, iteratee) / length) : NAN;
    }

    /**
     * The base implementation of `_.property` without support for deep paths.
     *
     * @private
     * @param {string} key The key of the property to get.
     * @returns {Function} Returns the new accessor function.
     */
    function baseProperty(key) {
      return function(object) {
        return object == null ? undefined$1 : object[key];
      };
    }

    /**
     * The base implementation of `_.propertyOf` without support for deep paths.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Function} Returns the new accessor function.
     */
    function basePropertyOf(object) {
      return function(key) {
        return object == null ? undefined$1 : object[key];
      };
    }

    /**
     * The base implementation of `_.reduce` and `_.reduceRight`, without support
     * for iteratee shorthands, which iterates over `collection` using `eachFunc`.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @param {*} accumulator The initial value.
     * @param {boolean} initAccum Specify using the first or last element of
     *  `collection` as the initial value.
     * @param {Function} eachFunc The function to iterate over `collection`.
     * @returns {*} Returns the accumulated value.
     */
    function baseReduce(collection, iteratee, accumulator, initAccum, eachFunc) {
      eachFunc(collection, function(value, index, collection) {
        accumulator = initAccum
          ? (initAccum = false, value)
          : iteratee(accumulator, value, index, collection);
      });
      return accumulator;
    }

    /**
     * The base implementation of `_.sortBy` which uses `comparer` to define the
     * sort order of `array` and replaces criteria objects with their corresponding
     * values.
     *
     * @private
     * @param {Array} array The array to sort.
     * @param {Function} comparer The function to define sort order.
     * @returns {Array} Returns `array`.
     */
    function baseSortBy(array, comparer) {
      var length = array.length;

      array.sort(comparer);
      while (length--) {
        array[length] = array[length].value;
      }
      return array;
    }

    /**
     * The base implementation of `_.sum` and `_.sumBy` without support for
     * iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {number} Returns the sum.
     */
    function baseSum(array, iteratee) {
      var result,
          index = -1,
          length = array.length;

      while (++index < length) {
        var current = iteratee(array[index]);
        if (current !== undefined$1) {
          result = result === undefined$1 ? current : (result + current);
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.times` without support for iteratee shorthands
     * or max array length checks.
     *
     * @private
     * @param {number} n The number of times to invoke `iteratee`.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array} Returns the array of results.
     */
    function baseTimes(n, iteratee) {
      var index = -1,
          result = Array(n);

      while (++index < n) {
        result[index] = iteratee(index);
      }
      return result;
    }

    /**
     * The base implementation of `_.toPairs` and `_.toPairsIn` which creates an array
     * of key-value pairs for `object` corresponding to the property names of `props`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Array} props The property names to get values for.
     * @returns {Object} Returns the key-value pairs.
     */
    function baseToPairs(object, props) {
      return arrayMap(props, function(key) {
        return [key, object[key]];
      });
    }

    /**
     * The base implementation of `_.unary` without support for storing metadata.
     *
     * @private
     * @param {Function} func The function to cap arguments for.
     * @returns {Function} Returns the new capped function.
     */
    function baseUnary(func) {
      return function(value) {
        return func(value);
      };
    }

    /**
     * The base implementation of `_.values` and `_.valuesIn` which creates an
     * array of `object` property values corresponding to the property names
     * of `props`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Array} props The property names to get values for.
     * @returns {Object} Returns the array of property values.
     */
    function baseValues(object, props) {
      return arrayMap(props, function(key) {
        return object[key];
      });
    }

    /**
     * Checks if a `cache` value for `key` exists.
     *
     * @private
     * @param {Object} cache The cache to query.
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function cacheHas(cache, key) {
      return cache.has(key);
    }

    /**
     * Used by `_.trim` and `_.trimStart` to get the index of the first string symbol
     * that is not found in the character symbols.
     *
     * @private
     * @param {Array} strSymbols The string symbols to inspect.
     * @param {Array} chrSymbols The character symbols to find.
     * @returns {number} Returns the index of the first unmatched string symbol.
     */
    function charsStartIndex(strSymbols, chrSymbols) {
      var index = -1,
          length = strSymbols.length;

      while (++index < length && baseIndexOf(chrSymbols, strSymbols[index], 0) > -1) {}
      return index;
    }

    /**
     * Used by `_.trim` and `_.trimEnd` to get the index of the last string symbol
     * that is not found in the character symbols.
     *
     * @private
     * @param {Array} strSymbols The string symbols to inspect.
     * @param {Array} chrSymbols The character symbols to find.
     * @returns {number} Returns the index of the last unmatched string symbol.
     */
    function charsEndIndex(strSymbols, chrSymbols) {
      var index = strSymbols.length;

      while (index-- && baseIndexOf(chrSymbols, strSymbols[index], 0) > -1) {}
      return index;
    }

    /**
     * Gets the number of `placeholder` occurrences in `array`.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} placeholder The placeholder to search for.
     * @returns {number} Returns the placeholder count.
     */
    function countHolders(array, placeholder) {
      var length = array.length,
          result = 0;

      while (length--) {
        if (array[length] === placeholder) {
          ++result;
        }
      }
      return result;
    }

    /**
     * Used by `_.deburr` to convert Latin-1 Supplement and Latin Extended-A
     * letters to basic Latin letters.
     *
     * @private
     * @param {string} letter The matched letter to deburr.
     * @returns {string} Returns the deburred letter.
     */
    var deburrLetter = basePropertyOf(deburredLetters);

    /**
     * Used by `_.escape` to convert characters to HTML entities.
     *
     * @private
     * @param {string} chr The matched character to escape.
     * @returns {string} Returns the escaped character.
     */
    var escapeHtmlChar = basePropertyOf(htmlEscapes);

    /**
     * Used by `_.template` to escape characters for inclusion in compiled string literals.
     *
     * @private
     * @param {string} chr The matched character to escape.
     * @returns {string} Returns the escaped character.
     */
    function escapeStringChar(chr) {
      return '\\' + stringEscapes[chr];
    }

    /**
     * Gets the value at `key` of `object`.
     *
     * @private
     * @param {Object} [object] The object to query.
     * @param {string} key The key of the property to get.
     * @returns {*} Returns the property value.
     */
    function getValue(object, key) {
      return object == null ? undefined$1 : object[key];
    }

    /**
     * Checks if `string` contains Unicode symbols.
     *
     * @private
     * @param {string} string The string to inspect.
     * @returns {boolean} Returns `true` if a symbol is found, else `false`.
     */
    function hasUnicode(string) {
      return reHasUnicode.test(string);
    }

    /**
     * Checks if `string` contains a word composed of Unicode symbols.
     *
     * @private
     * @param {string} string The string to inspect.
     * @returns {boolean} Returns `true` if a word is found, else `false`.
     */
    function hasUnicodeWord(string) {
      return reHasUnicodeWord.test(string);
    }

    /**
     * Converts `iterator` to an array.
     *
     * @private
     * @param {Object} iterator The iterator to convert.
     * @returns {Array} Returns the converted array.
     */
    function iteratorToArray(iterator) {
      var data,
          result = [];

      while (!(data = iterator.next()).done) {
        result.push(data.value);
      }
      return result;
    }

    /**
     * Converts `map` to its key-value pairs.
     *
     * @private
     * @param {Object} map The map to convert.
     * @returns {Array} Returns the key-value pairs.
     */
    function mapToArray(map) {
      var index = -1,
          result = Array(map.size);

      map.forEach(function(value, key) {
        result[++index] = [key, value];
      });
      return result;
    }

    /**
     * Creates a unary function that invokes `func` with its argument transformed.
     *
     * @private
     * @param {Function} func The function to wrap.
     * @param {Function} transform The argument transform.
     * @returns {Function} Returns the new function.
     */
    function overArg(func, transform) {
      return function(arg) {
        return func(transform(arg));
      };
    }

    /**
     * Replaces all `placeholder` elements in `array` with an internal placeholder
     * and returns an array of their indexes.
     *
     * @private
     * @param {Array} array The array to modify.
     * @param {*} placeholder The placeholder to replace.
     * @returns {Array} Returns the new array of placeholder indexes.
     */
    function replaceHolders(array, placeholder) {
      var index = -1,
          length = array.length,
          resIndex = 0,
          result = [];

      while (++index < length) {
        var value = array[index];
        if (value === placeholder || value === PLACEHOLDER) {
          array[index] = PLACEHOLDER;
          result[resIndex++] = index;
        }
      }
      return result;
    }

    /**
     * Converts `set` to an array of its values.
     *
     * @private
     * @param {Object} set The set to convert.
     * @returns {Array} Returns the values.
     */
    function setToArray(set) {
      var index = -1,
          result = Array(set.size);

      set.forEach(function(value) {
        result[++index] = value;
      });
      return result;
    }

    /**
     * Converts `set` to its value-value pairs.
     *
     * @private
     * @param {Object} set The set to convert.
     * @returns {Array} Returns the value-value pairs.
     */
    function setToPairs(set) {
      var index = -1,
          result = Array(set.size);

      set.forEach(function(value) {
        result[++index] = [value, value];
      });
      return result;
    }

    /**
     * A specialized version of `_.indexOf` which performs strict equality
     * comparisons of values, i.e. `===`.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @param {number} fromIndex The index to search from.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function strictIndexOf(array, value, fromIndex) {
      var index = fromIndex - 1,
          length = array.length;

      while (++index < length) {
        if (array[index] === value) {
          return index;
        }
      }
      return -1;
    }

    /**
     * A specialized version of `_.lastIndexOf` which performs strict equality
     * comparisons of values, i.e. `===`.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @param {number} fromIndex The index to search from.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function strictLastIndexOf(array, value, fromIndex) {
      var index = fromIndex + 1;
      while (index--) {
        if (array[index] === value) {
          return index;
        }
      }
      return index;
    }

    /**
     * Gets the number of symbols in `string`.
     *
     * @private
     * @param {string} string The string to inspect.
     * @returns {number} Returns the string size.
     */
    function stringSize(string) {
      return hasUnicode(string)
        ? unicodeSize(string)
        : asciiSize(string);
    }

    /**
     * Converts `string` to an array.
     *
     * @private
     * @param {string} string The string to convert.
     * @returns {Array} Returns the converted array.
     */
    function stringToArray(string) {
      return hasUnicode(string)
        ? unicodeToArray(string)
        : asciiToArray(string);
    }

    /**
     * Used by `_.unescape` to convert HTML entities to characters.
     *
     * @private
     * @param {string} chr The matched character to unescape.
     * @returns {string} Returns the unescaped character.
     */
    var unescapeHtmlChar = basePropertyOf(htmlUnescapes);

    /**
     * Gets the size of a Unicode `string`.
     *
     * @private
     * @param {string} string The string inspect.
     * @returns {number} Returns the string size.
     */
    function unicodeSize(string) {
      var result = reUnicode.lastIndex = 0;
      while (reUnicode.test(string)) {
        ++result;
      }
      return result;
    }

    /**
     * Converts a Unicode `string` to an array.
     *
     * @private
     * @param {string} string The string to convert.
     * @returns {Array} Returns the converted array.
     */
    function unicodeToArray(string) {
      return string.match(reUnicode) || [];
    }

    /**
     * Splits a Unicode `string` into an array of its words.
     *
     * @private
     * @param {string} The string to inspect.
     * @returns {Array} Returns the words of `string`.
     */
    function unicodeWords(string) {
      return string.match(reUnicodeWord) || [];
    }

    /*--------------------------------------------------------------------------*/

    /**
     * Create a new pristine `lodash` function using the `context` object.
     *
     * @static
     * @memberOf _
     * @since 1.1.0
     * @category Util
     * @param {Object} [context=root] The context object.
     * @returns {Function} Returns a new `lodash` function.
     * @example
     *
     * _.mixin({ 'foo': _.constant('foo') });
     *
     * var lodash = _.runInContext();
     * lodash.mixin({ 'bar': lodash.constant('bar') });
     *
     * _.isFunction(_.foo);
     * // => true
     * _.isFunction(_.bar);
     * // => false
     *
     * lodash.isFunction(lodash.foo);
     * // => false
     * lodash.isFunction(lodash.bar);
     * // => true
     *
     * // Create a suped-up `defer` in Node.js.
     * var defer = _.runInContext({ 'setTimeout': setImmediate }).defer;
     */
    var runInContext = (function runInContext(context) {
      context = context == null ? root : _.defaults(root.Object(), context, _.pick(root, contextProps));

      /** Built-in constructor references. */
      var Array = context.Array,
          Date = context.Date,
          Error = context.Error,
          Function = context.Function,
          Math = context.Math,
          Object = context.Object,
          RegExp = context.RegExp,
          String = context.String,
          TypeError = context.TypeError;

      /** Used for built-in method references. */
      var arrayProto = Array.prototype,
          funcProto = Function.prototype,
          objectProto = Object.prototype;

      /** Used to detect overreaching core-js shims. */
      var coreJsData = context['__core-js_shared__'];

      /** Used to resolve the decompiled source of functions. */
      var funcToString = funcProto.toString;

      /** Used to check objects for own properties. */
      var hasOwnProperty = objectProto.hasOwnProperty;

      /** Used to generate unique IDs. */
      var idCounter = 0;

      /** Used to detect methods masquerading as native. */
      var maskSrcKey = (function() {
        var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
        return uid ? ('Symbol(src)_1.' + uid) : '';
      }());

      /**
       * Used to resolve the
       * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
       * of values.
       */
      var nativeObjectToString = objectProto.toString;

      /** Used to infer the `Object` constructor. */
      var objectCtorString = funcToString.call(Object);

      /** Used to restore the original `_` reference in `_.noConflict`. */
      var oldDash = root._;

      /** Used to detect if a method is native. */
      var reIsNative = RegExp('^' +
        funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
        .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
      );

      /** Built-in value references. */
      var Buffer = moduleExports ? context.Buffer : undefined$1,
          Symbol = context.Symbol,
          Uint8Array = context.Uint8Array,
          allocUnsafe = Buffer ? Buffer.allocUnsafe : undefined$1,
          getPrototype = overArg(Object.getPrototypeOf, Object),
          objectCreate = Object.create,
          propertyIsEnumerable = objectProto.propertyIsEnumerable,
          splice = arrayProto.splice,
          spreadableSymbol = Symbol ? Symbol.isConcatSpreadable : undefined$1,
          symIterator = Symbol ? Symbol.iterator : undefined$1,
          symToStringTag = Symbol ? Symbol.toStringTag : undefined$1;

      var defineProperty = (function() {
        try {
          var func = getNative(Object, 'defineProperty');
          func({}, '', {});
          return func;
        } catch (e) {}
      }());

      /** Mocked built-ins. */
      var ctxClearTimeout = context.clearTimeout !== root.clearTimeout && context.clearTimeout,
          ctxNow = Date && Date.now !== root.Date.now && Date.now,
          ctxSetTimeout = context.setTimeout !== root.setTimeout && context.setTimeout;

      /* Built-in method references for those with the same name as other `lodash` methods. */
      var nativeCeil = Math.ceil,
          nativeFloor = Math.floor,
          nativeGetSymbols = Object.getOwnPropertySymbols,
          nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined$1,
          nativeIsFinite = context.isFinite,
          nativeJoin = arrayProto.join,
          nativeKeys = overArg(Object.keys, Object),
          nativeMax = Math.max,
          nativeMin = Math.min,
          nativeNow = Date.now,
          nativeParseInt = context.parseInt,
          nativeRandom = Math.random,
          nativeReverse = arrayProto.reverse;

      /* Built-in method references that are verified to be native. */
      var DataView = getNative(context, 'DataView'),
          Map = getNative(context, 'Map'),
          Promise = getNative(context, 'Promise'),
          Set = getNative(context, 'Set'),
          WeakMap = getNative(context, 'WeakMap'),
          nativeCreate = getNative(Object, 'create');

      /** Used to store function metadata. */
      var metaMap = WeakMap && new WeakMap;

      /** Used to lookup unminified function names. */
      var realNames = {};

      /** Used to detect maps, sets, and weakmaps. */
      var dataViewCtorString = toSource(DataView),
          mapCtorString = toSource(Map),
          promiseCtorString = toSource(Promise),
          setCtorString = toSource(Set),
          weakMapCtorString = toSource(WeakMap);

      /** Used to convert symbols to primitives and strings. */
      var symbolProto = Symbol ? Symbol.prototype : undefined$1,
          symbolValueOf = symbolProto ? symbolProto.valueOf : undefined$1,
          symbolToString = symbolProto ? symbolProto.toString : undefined$1;

      /*------------------------------------------------------------------------*/

      /**
       * Creates a `lodash` object which wraps `value` to enable implicit method
       * chain sequences. Methods that operate on and return arrays, collections,
       * and functions can be chained together. Methods that retrieve a single value
       * or may return a primitive value will automatically end the chain sequence
       * and return the unwrapped value. Otherwise, the value must be unwrapped
       * with `_#value`.
       *
       * Explicit chain sequences, which must be unwrapped with `_#value`, may be
       * enabled using `_.chain`.
       *
       * The execution of chained methods is lazy, that is, it's deferred until
       * `_#value` is implicitly or explicitly called.
       *
       * Lazy evaluation allows several methods to support shortcut fusion.
       * Shortcut fusion is an optimization to merge iteratee calls; this avoids
       * the creation of intermediate arrays and can greatly reduce the number of
       * iteratee executions. Sections of a chain sequence qualify for shortcut
       * fusion if the section is applied to an array and iteratees accept only
       * one argument. The heuristic for whether a section qualifies for shortcut
       * fusion is subject to change.
       *
       * Chaining is supported in custom builds as long as the `_#value` method is
       * directly or indirectly included in the build.
       *
       * In addition to lodash methods, wrappers have `Array` and `String` methods.
       *
       * The wrapper `Array` methods are:
       * `concat`, `join`, `pop`, `push`, `shift`, `sort`, `splice`, and `unshift`
       *
       * The wrapper `String` methods are:
       * `replace` and `split`
       *
       * The wrapper methods that support shortcut fusion are:
       * `at`, `compact`, `drop`, `dropRight`, `dropWhile`, `filter`, `find`,
       * `findLast`, `head`, `initial`, `last`, `map`, `reject`, `reverse`, `slice`,
       * `tail`, `take`, `takeRight`, `takeRightWhile`, `takeWhile`, and `toArray`
       *
       * The chainable wrapper methods are:
       * `after`, `ary`, `assign`, `assignIn`, `assignInWith`, `assignWith`, `at`,
       * `before`, `bind`, `bindAll`, `bindKey`, `castArray`, `chain`, `chunk`,
       * `commit`, `compact`, `concat`, `conforms`, `constant`, `countBy`, `create`,
       * `curry`, `debounce`, `defaults`, `defaultsDeep`, `defer`, `delay`,
       * `difference`, `differenceBy`, `differenceWith`, `drop`, `dropRight`,
       * `dropRightWhile`, `dropWhile`, `extend`, `extendWith`, `fill`, `filter`,
       * `flatMap`, `flatMapDeep`, `flatMapDepth`, `flatten`, `flattenDeep`,
       * `flattenDepth`, `flip`, `flow`, `flowRight`, `fromPairs`, `functions`,
       * `functionsIn`, `groupBy`, `initial`, `intersection`, `intersectionBy`,
       * `intersectionWith`, `invert`, `invertBy`, `invokeMap`, `iteratee`, `keyBy`,
       * `keys`, `keysIn`, `map`, `mapKeys`, `mapValues`, `matches`, `matchesProperty`,
       * `memoize`, `merge`, `mergeWith`, `method`, `methodOf`, `mixin`, `negate`,
       * `nthArg`, `omit`, `omitBy`, `once`, `orderBy`, `over`, `overArgs`,
       * `overEvery`, `overSome`, `partial`, `partialRight`, `partition`, `pick`,
       * `pickBy`, `plant`, `property`, `propertyOf`, `pull`, `pullAll`, `pullAllBy`,
       * `pullAllWith`, `pullAt`, `push`, `range`, `rangeRight`, `rearg`, `reject`,
       * `remove`, `rest`, `reverse`, `sampleSize`, `set`, `setWith`, `shuffle`,
       * `slice`, `sort`, `sortBy`, `splice`, `spread`, `tail`, `take`, `takeRight`,
       * `takeRightWhile`, `takeWhile`, `tap`, `throttle`, `thru`, `toArray`,
       * `toPairs`, `toPairsIn`, `toPath`, `toPlainObject`, `transform`, `unary`,
       * `union`, `unionBy`, `unionWith`, `uniq`, `uniqBy`, `uniqWith`, `unset`,
       * `unshift`, `unzip`, `unzipWith`, `update`, `updateWith`, `values`,
       * `valuesIn`, `without`, `wrap`, `xor`, `xorBy`, `xorWith`, `zip`,
       * `zipObject`, `zipObjectDeep`, and `zipWith`
       *
       * The wrapper methods that are **not** chainable by default are:
       * `add`, `attempt`, `camelCase`, `capitalize`, `ceil`, `clamp`, `clone`,
       * `cloneDeep`, `cloneDeepWith`, `cloneWith`, `conformsTo`, `deburr`,
       * `defaultTo`, `divide`, `each`, `eachRight`, `endsWith`, `eq`, `escape`,
       * `escapeRegExp`, `every`, `find`, `findIndex`, `findKey`, `findLast`,
       * `findLastIndex`, `findLastKey`, `first`, `floor`, `forEach`, `forEachRight`,
       * `forIn`, `forInRight`, `forOwn`, `forOwnRight`, `get`, `gt`, `gte`, `has`,
       * `hasIn`, `head`, `identity`, `includes`, `indexOf`, `inRange`, `invoke`,
       * `isArguments`, `isArray`, `isArrayBuffer`, `isArrayLike`, `isArrayLikeObject`,
       * `isBoolean`, `isBuffer`, `isDate`, `isElement`, `isEmpty`, `isEqual`,
       * `isEqualWith`, `isError`, `isFinite`, `isFunction`, `isInteger`, `isLength`,
       * `isMap`, `isMatch`, `isMatchWith`, `isNaN`, `isNative`, `isNil`, `isNull`,
       * `isNumber`, `isObject`, `isObjectLike`, `isPlainObject`, `isRegExp`,
       * `isSafeInteger`, `isSet`, `isString`, `isUndefined`, `isTypedArray`,
       * `isWeakMap`, `isWeakSet`, `join`, `kebabCase`, `last`, `lastIndexOf`,
       * `lowerCase`, `lowerFirst`, `lt`, `lte`, `max`, `maxBy`, `mean`, `meanBy`,
       * `min`, `minBy`, `multiply`, `noConflict`, `noop`, `now`, `nth`, `pad`,
       * `padEnd`, `padStart`, `parseInt`, `pop`, `random`, `reduce`, `reduceRight`,
       * `repeat`, `result`, `round`, `runInContext`, `sample`, `shift`, `size`,
       * `snakeCase`, `some`, `sortedIndex`, `sortedIndexBy`, `sortedLastIndex`,
       * `sortedLastIndexBy`, `startCase`, `startsWith`, `stubArray`, `stubFalse`,
       * `stubObject`, `stubString`, `stubTrue`, `subtract`, `sum`, `sumBy`,
       * `template`, `times`, `toFinite`, `toInteger`, `toJSON`, `toLength`,
       * `toLower`, `toNumber`, `toSafeInteger`, `toString`, `toUpper`, `trim`,
       * `trimEnd`, `trimStart`, `truncate`, `unescape`, `uniqueId`, `upperCase`,
       * `upperFirst`, `value`, and `words`
       *
       * @name _
       * @constructor
       * @category Seq
       * @param {*} value The value to wrap in a `lodash` instance.
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * function square(n) {
       *   return n * n;
       * }
       *
       * var wrapped = _([1, 2, 3]);
       *
       * // Returns an unwrapped value.
       * wrapped.reduce(_.add);
       * // => 6
       *
       * // Returns a wrapped value.
       * var squares = wrapped.map(square);
       *
       * _.isArray(squares);
       * // => false
       *
       * _.isArray(squares.value());
       * // => true
       */
      function lodash(value) {
        if (isObjectLike(value) && !isArray(value) && !(value instanceof LazyWrapper)) {
          if (value instanceof LodashWrapper) {
            return value;
          }
          if (hasOwnProperty.call(value, '__wrapped__')) {
            return wrapperClone(value);
          }
        }
        return new LodashWrapper(value);
      }

      /**
       * The base implementation of `_.create` without support for assigning
       * properties to the created object.
       *
       * @private
       * @param {Object} proto The object to inherit from.
       * @returns {Object} Returns the new object.
       */
      var baseCreate = (function() {
        function object() {}
        return function(proto) {
          if (!isObject(proto)) {
            return {};
          }
          if (objectCreate) {
            return objectCreate(proto);
          }
          object.prototype = proto;
          var result = new object;
          object.prototype = undefined$1;
          return result;
        };
      }());

      /**
       * The function whose prototype chain sequence wrappers inherit from.
       *
       * @private
       */
      function baseLodash() {
        // No operation performed.
      }

      /**
       * The base constructor for creating `lodash` wrapper objects.
       *
       * @private
       * @param {*} value The value to wrap.
       * @param {boolean} [chainAll] Enable explicit method chain sequences.
       */
      function LodashWrapper(value, chainAll) {
        this.__wrapped__ = value;
        this.__actions__ = [];
        this.__chain__ = !!chainAll;
        this.__index__ = 0;
        this.__values__ = undefined$1;
      }

      /**
       * By default, the template delimiters used by lodash are like those in
       * embedded Ruby (ERB) as well as ES2015 template strings. Change the
       * following template settings to use alternative delimiters.
       *
       * @static
       * @memberOf _
       * @type {Object}
       */
      lodash.templateSettings = {

        /**
         * Used to detect `data` property values to be HTML-escaped.
         *
         * @memberOf _.templateSettings
         * @type {RegExp}
         */
        'escape': reEscape,

        /**
         * Used to detect code to be evaluated.
         *
         * @memberOf _.templateSettings
         * @type {RegExp}
         */
        'evaluate': reEvaluate,

        /**
         * Used to detect `data` property values to inject.
         *
         * @memberOf _.templateSettings
         * @type {RegExp}
         */
        'interpolate': reInterpolate,

        /**
         * Used to reference the data object in the template text.
         *
         * @memberOf _.templateSettings
         * @type {string}
         */
        'variable': '',

        /**
         * Used to import variables into the compiled template.
         *
         * @memberOf _.templateSettings
         * @type {Object}
         */
        'imports': {

          /**
           * A reference to the `lodash` function.
           *
           * @memberOf _.templateSettings.imports
           * @type {Function}
           */
          '_': lodash
        }
      };

      // Ensure wrappers are instances of `baseLodash`.
      lodash.prototype = baseLodash.prototype;
      lodash.prototype.constructor = lodash;

      LodashWrapper.prototype = baseCreate(baseLodash.prototype);
      LodashWrapper.prototype.constructor = LodashWrapper;

      /*------------------------------------------------------------------------*/

      /**
       * Creates a lazy wrapper object which wraps `value` to enable lazy evaluation.
       *
       * @private
       * @constructor
       * @param {*} value The value to wrap.
       */
      function LazyWrapper(value) {
        this.__wrapped__ = value;
        this.__actions__ = [];
        this.__dir__ = 1;
        this.__filtered__ = false;
        this.__iteratees__ = [];
        this.__takeCount__ = MAX_ARRAY_LENGTH;
        this.__views__ = [];
      }

      /**
       * Creates a clone of the lazy wrapper object.
       *
       * @private
       * @name clone
       * @memberOf LazyWrapper
       * @returns {Object} Returns the cloned `LazyWrapper` object.
       */
      function lazyClone() {
        var result = new LazyWrapper(this.__wrapped__);
        result.__actions__ = copyArray(this.__actions__);
        result.__dir__ = this.__dir__;
        result.__filtered__ = this.__filtered__;
        result.__iteratees__ = copyArray(this.__iteratees__);
        result.__takeCount__ = this.__takeCount__;
        result.__views__ = copyArray(this.__views__);
        return result;
      }

      /**
       * Reverses the direction of lazy iteration.
       *
       * @private
       * @name reverse
       * @memberOf LazyWrapper
       * @returns {Object} Returns the new reversed `LazyWrapper` object.
       */
      function lazyReverse() {
        if (this.__filtered__) {
          var result = new LazyWrapper(this);
          result.__dir__ = -1;
          result.__filtered__ = true;
        } else {
          result = this.clone();
          result.__dir__ *= -1;
        }
        return result;
      }

      /**
       * Extracts the unwrapped value from its lazy wrapper.
       *
       * @private
       * @name value
       * @memberOf LazyWrapper
       * @returns {*} Returns the unwrapped value.
       */
      function lazyValue() {
        var array = this.__wrapped__.value(),
            dir = this.__dir__,
            isArr = isArray(array),
            isRight = dir < 0,
            arrLength = isArr ? array.length : 0,
            view = getView(0, arrLength, this.__views__),
            start = view.start,
            end = view.end,
            length = end - start,
            index = isRight ? end : (start - 1),
            iteratees = this.__iteratees__,
            iterLength = iteratees.length,
            resIndex = 0,
            takeCount = nativeMin(length, this.__takeCount__);

        if (!isArr || (!isRight && arrLength == length && takeCount == length)) {
          return baseWrapperValue(array, this.__actions__);
        }
        var result = [];

        outer:
        while (length-- && resIndex < takeCount) {
          index += dir;

          var iterIndex = -1,
              value = array[index];

          while (++iterIndex < iterLength) {
            var data = iteratees[iterIndex],
                iteratee = data.iteratee,
                type = data.type,
                computed = iteratee(value);

            if (type == LAZY_MAP_FLAG) {
              value = computed;
            } else if (!computed) {
              if (type == LAZY_FILTER_FLAG) {
                continue outer;
              } else {
                break outer;
              }
            }
          }
          result[resIndex++] = value;
        }
        return result;
      }

      // Ensure `LazyWrapper` is an instance of `baseLodash`.
      LazyWrapper.prototype = baseCreate(baseLodash.prototype);
      LazyWrapper.prototype.constructor = LazyWrapper;

      /*------------------------------------------------------------------------*/

      /**
       * Creates a hash object.
       *
       * @private
       * @constructor
       * @param {Array} [entries] The key-value pairs to cache.
       */
      function Hash(entries) {
        var index = -1,
            length = entries == null ? 0 : entries.length;

        this.clear();
        while (++index < length) {
          var entry = entries[index];
          this.set(entry[0], entry[1]);
        }
      }

      /**
       * Removes all key-value entries from the hash.
       *
       * @private
       * @name clear
       * @memberOf Hash
       */
      function hashClear() {
        this.__data__ = nativeCreate ? nativeCreate(null) : {};
        this.size = 0;
      }

      /**
       * Removes `key` and its value from the hash.
       *
       * @private
       * @name delete
       * @memberOf Hash
       * @param {Object} hash The hash to modify.
       * @param {string} key The key of the value to remove.
       * @returns {boolean} Returns `true` if the entry was removed, else `false`.
       */
      function hashDelete(key) {
        var result = this.has(key) && delete this.__data__[key];
        this.size -= result ? 1 : 0;
        return result;
      }

      /**
       * Gets the hash value for `key`.
       *
       * @private
       * @name get
       * @memberOf Hash
       * @param {string} key The key of the value to get.
       * @returns {*} Returns the entry value.
       */
      function hashGet(key) {
        var data = this.__data__;
        if (nativeCreate) {
          var result = data[key];
          return result === HASH_UNDEFINED ? undefined$1 : result;
        }
        return hasOwnProperty.call(data, key) ? data[key] : undefined$1;
      }

      /**
       * Checks if a hash value for `key` exists.
       *
       * @private
       * @name has
       * @memberOf Hash
       * @param {string} key The key of the entry to check.
       * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
       */
      function hashHas(key) {
        var data = this.__data__;
        return nativeCreate ? (data[key] !== undefined$1) : hasOwnProperty.call(data, key);
      }

      /**
       * Sets the hash `key` to `value`.
       *
       * @private
       * @name set
       * @memberOf Hash
       * @param {string} key The key of the value to set.
       * @param {*} value The value to set.
       * @returns {Object} Returns the hash instance.
       */
      function hashSet(key, value) {
        var data = this.__data__;
        this.size += this.has(key) ? 0 : 1;
        data[key] = (nativeCreate && value === undefined$1) ? HASH_UNDEFINED : value;
        return this;
      }

      // Add methods to `Hash`.
      Hash.prototype.clear = hashClear;
      Hash.prototype['delete'] = hashDelete;
      Hash.prototype.get = hashGet;
      Hash.prototype.has = hashHas;
      Hash.prototype.set = hashSet;

      /*------------------------------------------------------------------------*/

      /**
       * Creates an list cache object.
       *
       * @private
       * @constructor
       * @param {Array} [entries] The key-value pairs to cache.
       */
      function ListCache(entries) {
        var index = -1,
            length = entries == null ? 0 : entries.length;

        this.clear();
        while (++index < length) {
          var entry = entries[index];
          this.set(entry[0], entry[1]);
        }
      }

      /**
       * Removes all key-value entries from the list cache.
       *
       * @private
       * @name clear
       * @memberOf ListCache
       */
      function listCacheClear() {
        this.__data__ = [];
        this.size = 0;
      }

      /**
       * Removes `key` and its value from the list cache.
       *
       * @private
       * @name delete
       * @memberOf ListCache
       * @param {string} key The key of the value to remove.
       * @returns {boolean} Returns `true` if the entry was removed, else `false`.
       */
      function listCacheDelete(key) {
        var data = this.__data__,
            index = assocIndexOf(data, key);

        if (index < 0) {
          return false;
        }
        var lastIndex = data.length - 1;
        if (index == lastIndex) {
          data.pop();
        } else {
          splice.call(data, index, 1);
        }
        --this.size;
        return true;
      }

      /**
       * Gets the list cache value for `key`.
       *
       * @private
       * @name get
       * @memberOf ListCache
       * @param {string} key The key of the value to get.
       * @returns {*} Returns the entry value.
       */
      function listCacheGet(key) {
        var data = this.__data__,
            index = assocIndexOf(data, key);

        return index < 0 ? undefined$1 : data[index][1];
      }

      /**
       * Checks if a list cache value for `key` exists.
       *
       * @private
       * @name has
       * @memberOf ListCache
       * @param {string} key The key of the entry to check.
       * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
       */
      function listCacheHas(key) {
        return assocIndexOf(this.__data__, key) > -1;
      }

      /**
       * Sets the list cache `key` to `value`.
       *
       * @private
       * @name set
       * @memberOf ListCache
       * @param {string} key The key of the value to set.
       * @param {*} value The value to set.
       * @returns {Object} Returns the list cache instance.
       */
      function listCacheSet(key, value) {
        var data = this.__data__,
            index = assocIndexOf(data, key);

        if (index < 0) {
          ++this.size;
          data.push([key, value]);
        } else {
          data[index][1] = value;
        }
        return this;
      }

      // Add methods to `ListCache`.
      ListCache.prototype.clear = listCacheClear;
      ListCache.prototype['delete'] = listCacheDelete;
      ListCache.prototype.get = listCacheGet;
      ListCache.prototype.has = listCacheHas;
      ListCache.prototype.set = listCacheSet;

      /*------------------------------------------------------------------------*/

      /**
       * Creates a map cache object to store key-value pairs.
       *
       * @private
       * @constructor
       * @param {Array} [entries] The key-value pairs to cache.
       */
      function MapCache(entries) {
        var index = -1,
            length = entries == null ? 0 : entries.length;

        this.clear();
        while (++index < length) {
          var entry = entries[index];
          this.set(entry[0], entry[1]);
        }
      }

      /**
       * Removes all key-value entries from the map.
       *
       * @private
       * @name clear
       * @memberOf MapCache
       */
      function mapCacheClear() {
        this.size = 0;
        this.__data__ = {
          'hash': new Hash,
          'map': new (Map || ListCache),
          'string': new Hash
        };
      }

      /**
       * Removes `key` and its value from the map.
       *
       * @private
       * @name delete
       * @memberOf MapCache
       * @param {string} key The key of the value to remove.
       * @returns {boolean} Returns `true` if the entry was removed, else `false`.
       */
      function mapCacheDelete(key) {
        var result = getMapData(this, key)['delete'](key);
        this.size -= result ? 1 : 0;
        return result;
      }

      /**
       * Gets the map value for `key`.
       *
       * @private
       * @name get
       * @memberOf MapCache
       * @param {string} key The key of the value to get.
       * @returns {*} Returns the entry value.
       */
      function mapCacheGet(key) {
        return getMapData(this, key).get(key);
      }

      /**
       * Checks if a map value for `key` exists.
       *
       * @private
       * @name has
       * @memberOf MapCache
       * @param {string} key The key of the entry to check.
       * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
       */
      function mapCacheHas(key) {
        return getMapData(this, key).has(key);
      }

      /**
       * Sets the map `key` to `value`.
       *
       * @private
       * @name set
       * @memberOf MapCache
       * @param {string} key The key of the value to set.
       * @param {*} value The value to set.
       * @returns {Object} Returns the map cache instance.
       */
      function mapCacheSet(key, value) {
        var data = getMapData(this, key),
            size = data.size;

        data.set(key, value);
        this.size += data.size == size ? 0 : 1;
        return this;
      }

      // Add methods to `MapCache`.
      MapCache.prototype.clear = mapCacheClear;
      MapCache.prototype['delete'] = mapCacheDelete;
      MapCache.prototype.get = mapCacheGet;
      MapCache.prototype.has = mapCacheHas;
      MapCache.prototype.set = mapCacheSet;

      /*------------------------------------------------------------------------*/

      /**
       *
       * Creates an array cache object to store unique values.
       *
       * @private
       * @constructor
       * @param {Array} [values] The values to cache.
       */
      function SetCache(values) {
        var index = -1,
            length = values == null ? 0 : values.length;

        this.__data__ = new MapCache;
        while (++index < length) {
          this.add(values[index]);
        }
      }

      /**
       * Adds `value` to the array cache.
       *
       * @private
       * @name add
       * @memberOf SetCache
       * @alias push
       * @param {*} value The value to cache.
       * @returns {Object} Returns the cache instance.
       */
      function setCacheAdd(value) {
        this.__data__.set(value, HASH_UNDEFINED);
        return this;
      }

      /**
       * Checks if `value` is in the array cache.
       *
       * @private
       * @name has
       * @memberOf SetCache
       * @param {*} value The value to search for.
       * @returns {number} Returns `true` if `value` is found, else `false`.
       */
      function setCacheHas(value) {
        return this.__data__.has(value);
      }

      // Add methods to `SetCache`.
      SetCache.prototype.add = SetCache.prototype.push = setCacheAdd;
      SetCache.prototype.has = setCacheHas;

      /*------------------------------------------------------------------------*/

      /**
       * Creates a stack cache object to store key-value pairs.
       *
       * @private
       * @constructor
       * @param {Array} [entries] The key-value pairs to cache.
       */
      function Stack(entries) {
        var data = this.__data__ = new ListCache(entries);
        this.size = data.size;
      }

      /**
       * Removes all key-value entries from the stack.
       *
       * @private
       * @name clear
       * @memberOf Stack
       */
      function stackClear() {
        this.__data__ = new ListCache;
        this.size = 0;
      }

      /**
       * Removes `key` and its value from the stack.
       *
       * @private
       * @name delete
       * @memberOf Stack
       * @param {string} key The key of the value to remove.
       * @returns {boolean} Returns `true` if the entry was removed, else `false`.
       */
      function stackDelete(key) {
        var data = this.__data__,
            result = data['delete'](key);

        this.size = data.size;
        return result;
      }

      /**
       * Gets the stack value for `key`.
       *
       * @private
       * @name get
       * @memberOf Stack
       * @param {string} key The key of the value to get.
       * @returns {*} Returns the entry value.
       */
      function stackGet(key) {
        return this.__data__.get(key);
      }

      /**
       * Checks if a stack value for `key` exists.
       *
       * @private
       * @name has
       * @memberOf Stack
       * @param {string} key The key of the entry to check.
       * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
       */
      function stackHas(key) {
        return this.__data__.has(key);
      }

      /**
       * Sets the stack `key` to `value`.
       *
       * @private
       * @name set
       * @memberOf Stack
       * @param {string} key The key of the value to set.
       * @param {*} value The value to set.
       * @returns {Object} Returns the stack cache instance.
       */
      function stackSet(key, value) {
        var data = this.__data__;
        if (data instanceof ListCache) {
          var pairs = data.__data__;
          if (!Map || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
            pairs.push([key, value]);
            this.size = ++data.size;
            return this;
          }
          data = this.__data__ = new MapCache(pairs);
        }
        data.set(key, value);
        this.size = data.size;
        return this;
      }

      // Add methods to `Stack`.
      Stack.prototype.clear = stackClear;
      Stack.prototype['delete'] = stackDelete;
      Stack.prototype.get = stackGet;
      Stack.prototype.has = stackHas;
      Stack.prototype.set = stackSet;

      /*------------------------------------------------------------------------*/

      /**
       * Creates an array of the enumerable property names of the array-like `value`.
       *
       * @private
       * @param {*} value The value to query.
       * @param {boolean} inherited Specify returning inherited property names.
       * @returns {Array} Returns the array of property names.
       */
      function arrayLikeKeys(value, inherited) {
        var isArr = isArray(value),
            isArg = !isArr && isArguments(value),
            isBuff = !isArr && !isArg && isBuffer(value),
            isType = !isArr && !isArg && !isBuff && isTypedArray(value),
            skipIndexes = isArr || isArg || isBuff || isType,
            result = skipIndexes ? baseTimes(value.length, String) : [],
            length = result.length;

        for (var key in value) {
          if ((inherited || hasOwnProperty.call(value, key)) &&
              !(skipIndexes && (
                 // Safari 9 has enumerable `arguments.length` in strict mode.
                 key == 'length' ||
                 // Node.js 0.10 has enumerable non-index properties on buffers.
                 (isBuff && (key == 'offset' || key == 'parent')) ||
                 // PhantomJS 2 has enumerable non-index properties on typed arrays.
                 (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
                 // Skip index properties.
                 isIndex(key, length)
              ))) {
            result.push(key);
          }
        }
        return result;
      }

      /**
       * A specialized version of `_.sample` for arrays.
       *
       * @private
       * @param {Array} array The array to sample.
       * @returns {*} Returns the random element.
       */
      function arraySample(array) {
        var length = array.length;
        return length ? array[baseRandom(0, length - 1)] : undefined$1;
      }

      /**
       * A specialized version of `_.sampleSize` for arrays.
       *
       * @private
       * @param {Array} array The array to sample.
       * @param {number} n The number of elements to sample.
       * @returns {Array} Returns the random elements.
       */
      function arraySampleSize(array, n) {
        return shuffleSelf(copyArray(array), baseClamp(n, 0, array.length));
      }

      /**
       * A specialized version of `_.shuffle` for arrays.
       *
       * @private
       * @param {Array} array The array to shuffle.
       * @returns {Array} Returns the new shuffled array.
       */
      function arrayShuffle(array) {
        return shuffleSelf(copyArray(array));
      }

      /**
       * This function is like `assignValue` except that it doesn't assign
       * `undefined` values.
       *
       * @private
       * @param {Object} object The object to modify.
       * @param {string} key The key of the property to assign.
       * @param {*} value The value to assign.
       */
      function assignMergeValue(object, key, value) {
        if ((value !== undefined$1 && !eq(object[key], value)) ||
            (value === undefined$1 && !(key in object))) {
          baseAssignValue(object, key, value);
        }
      }

      /**
       * Assigns `value` to `key` of `object` if the existing value is not equivalent
       * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons.
       *
       * @private
       * @param {Object} object The object to modify.
       * @param {string} key The key of the property to assign.
       * @param {*} value The value to assign.
       */
      function assignValue(object, key, value) {
        var objValue = object[key];
        if (!(hasOwnProperty.call(object, key) && eq(objValue, value)) ||
            (value === undefined$1 && !(key in object))) {
          baseAssignValue(object, key, value);
        }
      }

      /**
       * Gets the index at which the `key` is found in `array` of key-value pairs.
       *
       * @private
       * @param {Array} array The array to inspect.
       * @param {*} key The key to search for.
       * @returns {number} Returns the index of the matched value, else `-1`.
       */
      function assocIndexOf(array, key) {
        var length = array.length;
        while (length--) {
          if (eq(array[length][0], key)) {
            return length;
          }
        }
        return -1;
      }

      /**
       * Aggregates elements of `collection` on `accumulator` with keys transformed
       * by `iteratee` and values set by `setter`.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} setter The function to set `accumulator` values.
       * @param {Function} iteratee The iteratee to transform keys.
       * @param {Object} accumulator The initial aggregated object.
       * @returns {Function} Returns `accumulator`.
       */
      function baseAggregator(collection, setter, iteratee, accumulator) {
        baseEach(collection, function(value, key, collection) {
          setter(accumulator, value, iteratee(value), collection);
        });
        return accumulator;
      }

      /**
       * The base implementation of `_.assign` without support for multiple sources
       * or `customizer` functions.
       *
       * @private
       * @param {Object} object The destination object.
       * @param {Object} source The source object.
       * @returns {Object} Returns `object`.
       */
      function baseAssign(object, source) {
        return object && copyObject(source, keys(source), object);
      }

      /**
       * The base implementation of `_.assignIn` without support for multiple sources
       * or `customizer` functions.
       *
       * @private
       * @param {Object} object The destination object.
       * @param {Object} source The source object.
       * @returns {Object} Returns `object`.
       */
      function baseAssignIn(object, source) {
        return object && copyObject(source, keysIn(source), object);
      }

      /**
       * The base implementation of `assignValue` and `assignMergeValue` without
       * value checks.
       *
       * @private
       * @param {Object} object The object to modify.
       * @param {string} key The key of the property to assign.
       * @param {*} value The value to assign.
       */
      function baseAssignValue(object, key, value) {
        if (key == '__proto__' && defineProperty) {
          defineProperty(object, key, {
            'configurable': true,
            'enumerable': true,
            'value': value,
            'writable': true
          });
        } else {
          object[key] = value;
        }
      }

      /**
       * The base implementation of `_.at` without support for individual paths.
       *
       * @private
       * @param {Object} object The object to iterate over.
       * @param {string[]} paths The property paths to pick.
       * @returns {Array} Returns the picked elements.
       */
      function baseAt(object, paths) {
        var index = -1,
            length = paths.length,
            result = Array(length),
            skip = object == null;

        while (++index < length) {
          result[index] = skip ? undefined$1 : get(object, paths[index]);
        }
        return result;
      }

      /**
       * The base implementation of `_.clamp` which doesn't coerce arguments.
       *
       * @private
       * @param {number} number The number to clamp.
       * @param {number} [lower] The lower bound.
       * @param {number} upper The upper bound.
       * @returns {number} Returns the clamped number.
       */
      function baseClamp(number, lower, upper) {
        if (number === number) {
          if (upper !== undefined$1) {
            number = number <= upper ? number : upper;
          }
          if (lower !== undefined$1) {
            number = number >= lower ? number : lower;
          }
        }
        return number;
      }

      /**
       * The base implementation of `_.clone` and `_.cloneDeep` which tracks
       * traversed objects.
       *
       * @private
       * @param {*} value The value to clone.
       * @param {boolean} bitmask The bitmask flags.
       *  1 - Deep clone
       *  2 - Flatten inherited properties
       *  4 - Clone symbols
       * @param {Function} [customizer] The function to customize cloning.
       * @param {string} [key] The key of `value`.
       * @param {Object} [object] The parent object of `value`.
       * @param {Object} [stack] Tracks traversed objects and their clone counterparts.
       * @returns {*} Returns the cloned value.
       */
      function baseClone(value, bitmask, customizer, key, object, stack) {
        var result,
            isDeep = bitmask & CLONE_DEEP_FLAG,
            isFlat = bitmask & CLONE_FLAT_FLAG,
            isFull = bitmask & CLONE_SYMBOLS_FLAG;

        if (customizer) {
          result = object ? customizer(value, key, object, stack) : customizer(value);
        }
        if (result !== undefined$1) {
          return result;
        }
        if (!isObject(value)) {
          return value;
        }
        var isArr = isArray(value);
        if (isArr) {
          result = initCloneArray(value);
          if (!isDeep) {
            return copyArray(value, result);
          }
        } else {
          var tag = getTag(value),
              isFunc = tag == funcTag || tag == genTag;

          if (isBuffer(value)) {
            return cloneBuffer(value, isDeep);
          }
          if (tag == objectTag || tag == argsTag || (isFunc && !object)) {
            result = (isFlat || isFunc) ? {} : initCloneObject(value);
            if (!isDeep) {
              return isFlat
                ? copySymbolsIn(value, baseAssignIn(result, value))
                : copySymbols(value, baseAssign(result, value));
            }
          } else {
            if (!cloneableTags[tag]) {
              return object ? value : {};
            }
            result = initCloneByTag(value, tag, isDeep);
          }
        }
        // Check for circular references and return its corresponding clone.
        stack || (stack = new Stack);
        var stacked = stack.get(value);
        if (stacked) {
          return stacked;
        }
        stack.set(value, result);

        if (isSet(value)) {
          value.forEach(function(subValue) {
            result.add(baseClone(subValue, bitmask, customizer, subValue, value, stack));
          });
        } else if (isMap(value)) {
          value.forEach(function(subValue, key) {
            result.set(key, baseClone(subValue, bitmask, customizer, key, value, stack));
          });
        }

        var keysFunc = isFull
          ? (isFlat ? getAllKeysIn : getAllKeys)
          : (isFlat ? keysIn : keys);

        var props = isArr ? undefined$1 : keysFunc(value);
        arrayEach(props || value, function(subValue, key) {
          if (props) {
            key = subValue;
            subValue = value[key];
          }
          // Recursively populate clone (susceptible to call stack limits).
          assignValue(result, key, baseClone(subValue, bitmask, customizer, key, value, stack));
        });
        return result;
      }

      /**
       * The base implementation of `_.conforms` which doesn't clone `source`.
       *
       * @private
       * @param {Object} source The object of property predicates to conform to.
       * @returns {Function} Returns the new spec function.
       */
      function baseConforms(source) {
        var props = keys(source);
        return function(object) {
          return baseConformsTo(object, source, props);
        };
      }

      /**
       * The base implementation of `_.conformsTo` which accepts `props` to check.
       *
       * @private
       * @param {Object} object The object to inspect.
       * @param {Object} source The object of property predicates to conform to.
       * @returns {boolean} Returns `true` if `object` conforms, else `false`.
       */
      function baseConformsTo(object, source, props) {
        var length = props.length;
        if (object == null) {
          return !length;
        }
        object = Object(object);
        while (length--) {
          var key = props[length],
              predicate = source[key],
              value = object[key];

          if ((value === undefined$1 && !(key in object)) || !predicate(value)) {
            return false;
          }
        }
        return true;
      }

      /**
       * The base implementation of `_.delay` and `_.defer` which accepts `args`
       * to provide to `func`.
       *
       * @private
       * @param {Function} func The function to delay.
       * @param {number} wait The number of milliseconds to delay invocation.
       * @param {Array} args The arguments to provide to `func`.
       * @returns {number|Object} Returns the timer id or timeout object.
       */
      function baseDelay(func, wait, args) {
        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        return setTimeout(function() { func.apply(undefined$1, args); }, wait);
      }

      /**
       * The base implementation of methods like `_.difference` without support
       * for excluding multiple arrays or iteratee shorthands.
       *
       * @private
       * @param {Array} array The array to inspect.
       * @param {Array} values The values to exclude.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of filtered values.
       */
      function baseDifference(array, values, iteratee, comparator) {
        var index = -1,
            includes = arrayIncludes,
            isCommon = true,
            length = array.length,
            result = [],
            valuesLength = values.length;

        if (!length) {
          return result;
        }
        if (iteratee) {
          values = arrayMap(values, baseUnary(iteratee));
        }
        if (comparator) {
          includes = arrayIncludesWith;
          isCommon = false;
        }
        else if (values.length >= LARGE_ARRAY_SIZE) {
          includes = cacheHas;
          isCommon = false;
          values = new SetCache(values);
        }
        outer:
        while (++index < length) {
          var value = array[index],
              computed = iteratee == null ? value : iteratee(value);

          value = (comparator || value !== 0) ? value : 0;
          if (isCommon && computed === computed) {
            var valuesIndex = valuesLength;
            while (valuesIndex--) {
              if (values[valuesIndex] === computed) {
                continue outer;
              }
            }
            result.push(value);
          }
          else if (!includes(values, computed, comparator)) {
            result.push(value);
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.forEach` without support for iteratee shorthands.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @returns {Array|Object} Returns `collection`.
       */
      var baseEach = createBaseEach(baseForOwn);

      /**
       * The base implementation of `_.forEachRight` without support for iteratee shorthands.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @returns {Array|Object} Returns `collection`.
       */
      var baseEachRight = createBaseEach(baseForOwnRight, true);

      /**
       * The base implementation of `_.every` without support for iteratee shorthands.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} predicate The function invoked per iteration.
       * @returns {boolean} Returns `true` if all elements pass the predicate check,
       *  else `false`
       */
      function baseEvery(collection, predicate) {
        var result = true;
        baseEach(collection, function(value, index, collection) {
          result = !!predicate(value, index, collection);
          return result;
        });
        return result;
      }

      /**
       * The base implementation of methods like `_.max` and `_.min` which accepts a
       * `comparator` to determine the extremum value.
       *
       * @private
       * @param {Array} array The array to iterate over.
       * @param {Function} iteratee The iteratee invoked per iteration.
       * @param {Function} comparator The comparator used to compare values.
       * @returns {*} Returns the extremum value.
       */
      function baseExtremum(array, iteratee, comparator) {
        var index = -1,
            length = array.length;

        while (++index < length) {
          var value = array[index],
              current = iteratee(value);

          if (current != null && (computed === undefined$1
                ? (current === current && !isSymbol(current))
                : comparator(current, computed)
              )) {
            var computed = current,
                result = value;
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.fill` without an iteratee call guard.
       *
       * @private
       * @param {Array} array The array to fill.
       * @param {*} value The value to fill `array` with.
       * @param {number} [start=0] The start position.
       * @param {number} [end=array.length] The end position.
       * @returns {Array} Returns `array`.
       */
      function baseFill(array, value, start, end) {
        var length = array.length;

        start = toInteger(start);
        if (start < 0) {
          start = -start > length ? 0 : (length + start);
        }
        end = (end === undefined$1 || end > length) ? length : toInteger(end);
        if (end < 0) {
          end += length;
        }
        end = start > end ? 0 : toLength(end);
        while (start < end) {
          array[start++] = value;
        }
        return array;
      }

      /**
       * The base implementation of `_.filter` without support for iteratee shorthands.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} predicate The function invoked per iteration.
       * @returns {Array} Returns the new filtered array.
       */
      function baseFilter(collection, predicate) {
        var result = [];
        baseEach(collection, function(value, index, collection) {
          if (predicate(value, index, collection)) {
            result.push(value);
          }
        });
        return result;
      }

      /**
       * The base implementation of `_.flatten` with support for restricting flattening.
       *
       * @private
       * @param {Array} array The array to flatten.
       * @param {number} depth The maximum recursion depth.
       * @param {boolean} [predicate=isFlattenable] The function invoked per iteration.
       * @param {boolean} [isStrict] Restrict to values that pass `predicate` checks.
       * @param {Array} [result=[]] The initial result value.
       * @returns {Array} Returns the new flattened array.
       */
      function baseFlatten(array, depth, predicate, isStrict, result) {
        var index = -1,
            length = array.length;

        predicate || (predicate = isFlattenable);
        result || (result = []);

        while (++index < length) {
          var value = array[index];
          if (depth > 0 && predicate(value)) {
            if (depth > 1) {
              // Recursively flatten arrays (susceptible to call stack limits).
              baseFlatten(value, depth - 1, predicate, isStrict, result);
            } else {
              arrayPush(result, value);
            }
          } else if (!isStrict) {
            result[result.length] = value;
          }
        }
        return result;
      }

      /**
       * The base implementation of `baseForOwn` which iterates over `object`
       * properties returned by `keysFunc` and invokes `iteratee` for each property.
       * Iteratee functions may exit iteration early by explicitly returning `false`.
       *
       * @private
       * @param {Object} object The object to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @param {Function} keysFunc The function to get the keys of `object`.
       * @returns {Object} Returns `object`.
       */
      var baseFor = createBaseFor();

      /**
       * This function is like `baseFor` except that it iterates over properties
       * in the opposite order.
       *
       * @private
       * @param {Object} object The object to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @param {Function} keysFunc The function to get the keys of `object`.
       * @returns {Object} Returns `object`.
       */
      var baseForRight = createBaseFor(true);

      /**
       * The base implementation of `_.forOwn` without support for iteratee shorthands.
       *
       * @private
       * @param {Object} object The object to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @returns {Object} Returns `object`.
       */
      function baseForOwn(object, iteratee) {
        return object && baseFor(object, iteratee, keys);
      }

      /**
       * The base implementation of `_.forOwnRight` without support for iteratee shorthands.
       *
       * @private
       * @param {Object} object The object to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @returns {Object} Returns `object`.
       */
      function baseForOwnRight(object, iteratee) {
        return object && baseForRight(object, iteratee, keys);
      }

      /**
       * The base implementation of `_.functions` which creates an array of
       * `object` function property names filtered from `props`.
       *
       * @private
       * @param {Object} object The object to inspect.
       * @param {Array} props The property names to filter.
       * @returns {Array} Returns the function names.
       */
      function baseFunctions(object, props) {
        return arrayFilter(props, function(key) {
          return isFunction(object[key]);
        });
      }

      /**
       * The base implementation of `_.get` without support for default values.
       *
       * @private
       * @param {Object} object The object to query.
       * @param {Array|string} path The path of the property to get.
       * @returns {*} Returns the resolved value.
       */
      function baseGet(object, path) {
        path = castPath(path, object);

        var index = 0,
            length = path.length;

        while (object != null && index < length) {
          object = object[toKey(path[index++])];
        }
        return (index && index == length) ? object : undefined$1;
      }

      /**
       * The base implementation of `getAllKeys` and `getAllKeysIn` which uses
       * `keysFunc` and `symbolsFunc` to get the enumerable property names and
       * symbols of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @param {Function} keysFunc The function to get the keys of `object`.
       * @param {Function} symbolsFunc The function to get the symbols of `object`.
       * @returns {Array} Returns the array of property names and symbols.
       */
      function baseGetAllKeys(object, keysFunc, symbolsFunc) {
        var result = keysFunc(object);
        return isArray(object) ? result : arrayPush(result, symbolsFunc(object));
      }

      /**
       * The base implementation of `getTag` without fallbacks for buggy environments.
       *
       * @private
       * @param {*} value The value to query.
       * @returns {string} Returns the `toStringTag`.
       */
      function baseGetTag(value) {
        if (value == null) {
          return value === undefined$1 ? undefinedTag : nullTag;
        }
        return (symToStringTag && symToStringTag in Object(value))
          ? getRawTag(value)
          : objectToString(value);
      }

      /**
       * The base implementation of `_.gt` which doesn't coerce arguments.
       *
       * @private
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if `value` is greater than `other`,
       *  else `false`.
       */
      function baseGt(value, other) {
        return value > other;
      }

      /**
       * The base implementation of `_.has` without support for deep paths.
       *
       * @private
       * @param {Object} [object] The object to query.
       * @param {Array|string} key The key to check.
       * @returns {boolean} Returns `true` if `key` exists, else `false`.
       */
      function baseHas(object, key) {
        return object != null && hasOwnProperty.call(object, key);
      }

      /**
       * The base implementation of `_.hasIn` without support for deep paths.
       *
       * @private
       * @param {Object} [object] The object to query.
       * @param {Array|string} key The key to check.
       * @returns {boolean} Returns `true` if `key` exists, else `false`.
       */
      function baseHasIn(object, key) {
        return object != null && key in Object(object);
      }

      /**
       * The base implementation of `_.inRange` which doesn't coerce arguments.
       *
       * @private
       * @param {number} number The number to check.
       * @param {number} start The start of the range.
       * @param {number} end The end of the range.
       * @returns {boolean} Returns `true` if `number` is in the range, else `false`.
       */
      function baseInRange(number, start, end) {
        return number >= nativeMin(start, end) && number < nativeMax(start, end);
      }

      /**
       * The base implementation of methods like `_.intersection`, without support
       * for iteratee shorthands, that accepts an array of arrays to inspect.
       *
       * @private
       * @param {Array} arrays The arrays to inspect.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of shared values.
       */
      function baseIntersection(arrays, iteratee, comparator) {
        var includes = comparator ? arrayIncludesWith : arrayIncludes,
            length = arrays[0].length,
            othLength = arrays.length,
            othIndex = othLength,
            caches = Array(othLength),
            maxLength = Infinity,
            result = [];

        while (othIndex--) {
          var array = arrays[othIndex];
          if (othIndex && iteratee) {
            array = arrayMap(array, baseUnary(iteratee));
          }
          maxLength = nativeMin(array.length, maxLength);
          caches[othIndex] = !comparator && (iteratee || (length >= 120 && array.length >= 120))
            ? new SetCache(othIndex && array)
            : undefined$1;
        }
        array = arrays[0];

        var index = -1,
            seen = caches[0];

        outer:
        while (++index < length && result.length < maxLength) {
          var value = array[index],
              computed = iteratee ? iteratee(value) : value;

          value = (comparator || value !== 0) ? value : 0;
          if (!(seen
                ? cacheHas(seen, computed)
                : includes(result, computed, comparator)
              )) {
            othIndex = othLength;
            while (--othIndex) {
              var cache = caches[othIndex];
              if (!(cache
                    ? cacheHas(cache, computed)
                    : includes(arrays[othIndex], computed, comparator))
                  ) {
                continue outer;
              }
            }
            if (seen) {
              seen.push(computed);
            }
            result.push(value);
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.invert` and `_.invertBy` which inverts
       * `object` with values transformed by `iteratee` and set by `setter`.
       *
       * @private
       * @param {Object} object The object to iterate over.
       * @param {Function} setter The function to set `accumulator` values.
       * @param {Function} iteratee The iteratee to transform values.
       * @param {Object} accumulator The initial inverted object.
       * @returns {Function} Returns `accumulator`.
       */
      function baseInverter(object, setter, iteratee, accumulator) {
        baseForOwn(object, function(value, key, object) {
          setter(accumulator, iteratee(value), key, object);
        });
        return accumulator;
      }

      /**
       * The base implementation of `_.invoke` without support for individual
       * method arguments.
       *
       * @private
       * @param {Object} object The object to query.
       * @param {Array|string} path The path of the method to invoke.
       * @param {Array} args The arguments to invoke the method with.
       * @returns {*} Returns the result of the invoked method.
       */
      function baseInvoke(object, path, args) {
        path = castPath(path, object);
        object = parent(object, path);
        var func = object == null ? object : object[toKey(last(path))];
        return func == null ? undefined$1 : apply(func, object, args);
      }

      /**
       * The base implementation of `_.isArguments`.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an `arguments` object,
       */
      function baseIsArguments(value) {
        return isObjectLike(value) && baseGetTag(value) == argsTag;
      }

      /**
       * The base implementation of `_.isArrayBuffer` without Node.js optimizations.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an array buffer, else `false`.
       */
      function baseIsArrayBuffer(value) {
        return isObjectLike(value) && baseGetTag(value) == arrayBufferTag;
      }

      /**
       * The base implementation of `_.isDate` without Node.js optimizations.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a date object, else `false`.
       */
      function baseIsDate(value) {
        return isObjectLike(value) && baseGetTag(value) == dateTag;
      }

      /**
       * The base implementation of `_.isEqual` which supports partial comparisons
       * and tracks traversed objects.
       *
       * @private
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @param {boolean} bitmask The bitmask flags.
       *  1 - Unordered comparison
       *  2 - Partial comparison
       * @param {Function} [customizer] The function to customize comparisons.
       * @param {Object} [stack] Tracks traversed `value` and `other` objects.
       * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
       */
      function baseIsEqual(value, other, bitmask, customizer, stack) {
        if (value === other) {
          return true;
        }
        if (value == null || other == null || (!isObjectLike(value) && !isObjectLike(other))) {
          return value !== value && other !== other;
        }
        return baseIsEqualDeep(value, other, bitmask, customizer, baseIsEqual, stack);
      }

      /**
       * A specialized version of `baseIsEqual` for arrays and objects which performs
       * deep comparisons and tracks traversed objects enabling objects with circular
       * references to be compared.
       *
       * @private
       * @param {Object} object The object to compare.
       * @param {Object} other The other object to compare.
       * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
       * @param {Function} customizer The function to customize comparisons.
       * @param {Function} equalFunc The function to determine equivalents of values.
       * @param {Object} [stack] Tracks traversed `object` and `other` objects.
       * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
       */
      function baseIsEqualDeep(object, other, bitmask, customizer, equalFunc, stack) {
        var objIsArr = isArray(object),
            othIsArr = isArray(other),
            objTag = objIsArr ? arrayTag : getTag(object),
            othTag = othIsArr ? arrayTag : getTag(other);

        objTag = objTag == argsTag ? objectTag : objTag;
        othTag = othTag == argsTag ? objectTag : othTag;

        var objIsObj = objTag == objectTag,
            othIsObj = othTag == objectTag,
            isSameTag = objTag == othTag;

        if (isSameTag && isBuffer(object)) {
          if (!isBuffer(other)) {
            return false;
          }
          objIsArr = true;
          objIsObj = false;
        }
        if (isSameTag && !objIsObj) {
          stack || (stack = new Stack);
          return (objIsArr || isTypedArray(object))
            ? equalArrays(object, other, bitmask, customizer, equalFunc, stack)
            : equalByTag(object, other, objTag, bitmask, customizer, equalFunc, stack);
        }
        if (!(bitmask & COMPARE_PARTIAL_FLAG)) {
          var objIsWrapped = objIsObj && hasOwnProperty.call(object, '__wrapped__'),
              othIsWrapped = othIsObj && hasOwnProperty.call(other, '__wrapped__');

          if (objIsWrapped || othIsWrapped) {
            var objUnwrapped = objIsWrapped ? object.value() : object,
                othUnwrapped = othIsWrapped ? other.value() : other;

            stack || (stack = new Stack);
            return equalFunc(objUnwrapped, othUnwrapped, bitmask, customizer, stack);
          }
        }
        if (!isSameTag) {
          return false;
        }
        stack || (stack = new Stack);
        return equalObjects(object, other, bitmask, customizer, equalFunc, stack);
      }

      /**
       * The base implementation of `_.isMap` without Node.js optimizations.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a map, else `false`.
       */
      function baseIsMap(value) {
        return isObjectLike(value) && getTag(value) == mapTag;
      }

      /**
       * The base implementation of `_.isMatch` without support for iteratee shorthands.
       *
       * @private
       * @param {Object} object The object to inspect.
       * @param {Object} source The object of property values to match.
       * @param {Array} matchData The property names, values, and compare flags to match.
       * @param {Function} [customizer] The function to customize comparisons.
       * @returns {boolean} Returns `true` if `object` is a match, else `false`.
       */
      function baseIsMatch(object, source, matchData, customizer) {
        var index = matchData.length,
            length = index,
            noCustomizer = !customizer;

        if (object == null) {
          return !length;
        }
        object = Object(object);
        while (index--) {
          var data = matchData[index];
          if ((noCustomizer && data[2])
                ? data[1] !== object[data[0]]
                : !(data[0] in object)
              ) {
            return false;
          }
        }
        while (++index < length) {
          data = matchData[index];
          var key = data[0],
              objValue = object[key],
              srcValue = data[1];

          if (noCustomizer && data[2]) {
            if (objValue === undefined$1 && !(key in object)) {
              return false;
            }
          } else {
            var stack = new Stack;
            if (customizer) {
              var result = customizer(objValue, srcValue, key, object, source, stack);
            }
            if (!(result === undefined$1
                  ? baseIsEqual(srcValue, objValue, COMPARE_PARTIAL_FLAG | COMPARE_UNORDERED_FLAG, customizer, stack)
                  : result
                )) {
              return false;
            }
          }
        }
        return true;
      }

      /**
       * The base implementation of `_.isNative` without bad shim checks.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a native function,
       *  else `false`.
       */
      function baseIsNative(value) {
        if (!isObject(value) || isMasked(value)) {
          return false;
        }
        var pattern = isFunction(value) ? reIsNative : reIsHostCtor;
        return pattern.test(toSource(value));
      }

      /**
       * The base implementation of `_.isRegExp` without Node.js optimizations.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a regexp, else `false`.
       */
      function baseIsRegExp(value) {
        return isObjectLike(value) && baseGetTag(value) == regexpTag;
      }

      /**
       * The base implementation of `_.isSet` without Node.js optimizations.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a set, else `false`.
       */
      function baseIsSet(value) {
        return isObjectLike(value) && getTag(value) == setTag;
      }

      /**
       * The base implementation of `_.isTypedArray` without Node.js optimizations.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
       */
      function baseIsTypedArray(value) {
        return isObjectLike(value) &&
          isLength(value.length) && !!typedArrayTags[baseGetTag(value)];
      }

      /**
       * The base implementation of `_.iteratee`.
       *
       * @private
       * @param {*} [value=_.identity] The value to convert to an iteratee.
       * @returns {Function} Returns the iteratee.
       */
      function baseIteratee(value) {
        // Don't store the `typeof` result in a variable to avoid a JIT bug in Safari 9.
        // See https://bugs.webkit.org/show_bug.cgi?id=156034 for more details.
        if (typeof value == 'function') {
          return value;
        }
        if (value == null) {
          return identity;
        }
        if (typeof value == 'object') {
          return isArray(value)
            ? baseMatchesProperty(value[0], value[1])
            : baseMatches(value);
        }
        return property(value);
      }

      /**
       * The base implementation of `_.keys` which doesn't treat sparse arrays as dense.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names.
       */
      function baseKeys(object) {
        if (!isPrototype(object)) {
          return nativeKeys(object);
        }
        var result = [];
        for (var key in Object(object)) {
          if (hasOwnProperty.call(object, key) && key != 'constructor') {
            result.push(key);
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.keysIn` which doesn't treat sparse arrays as dense.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names.
       */
      function baseKeysIn(object) {
        if (!isObject(object)) {
          return nativeKeysIn(object);
        }
        var isProto = isPrototype(object),
            result = [];

        for (var key in object) {
          if (!(key == 'constructor' && (isProto || !hasOwnProperty.call(object, key)))) {
            result.push(key);
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.lt` which doesn't coerce arguments.
       *
       * @private
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if `value` is less than `other`,
       *  else `false`.
       */
      function baseLt(value, other) {
        return value < other;
      }

      /**
       * The base implementation of `_.map` without support for iteratee shorthands.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} iteratee The function invoked per iteration.
       * @returns {Array} Returns the new mapped array.
       */
      function baseMap(collection, iteratee) {
        var index = -1,
            result = isArrayLike(collection) ? Array(collection.length) : [];

        baseEach(collection, function(value, key, collection) {
          result[++index] = iteratee(value, key, collection);
        });
        return result;
      }

      /**
       * The base implementation of `_.matches` which doesn't clone `source`.
       *
       * @private
       * @param {Object} source The object of property values to match.
       * @returns {Function} Returns the new spec function.
       */
      function baseMatches(source) {
        var matchData = getMatchData(source);
        if (matchData.length == 1 && matchData[0][2]) {
          return matchesStrictComparable(matchData[0][0], matchData[0][1]);
        }
        return function(object) {
          return object === source || baseIsMatch(object, source, matchData);
        };
      }

      /**
       * The base implementation of `_.matchesProperty` which doesn't clone `srcValue`.
       *
       * @private
       * @param {string} path The path of the property to get.
       * @param {*} srcValue The value to match.
       * @returns {Function} Returns the new spec function.
       */
      function baseMatchesProperty(path, srcValue) {
        if (isKey(path) && isStrictComparable(srcValue)) {
          return matchesStrictComparable(toKey(path), srcValue);
        }
        return function(object) {
          var objValue = get(object, path);
          return (objValue === undefined$1 && objValue === srcValue)
            ? hasIn(object, path)
            : baseIsEqual(srcValue, objValue, COMPARE_PARTIAL_FLAG | COMPARE_UNORDERED_FLAG);
        };
      }

      /**
       * The base implementation of `_.merge` without support for multiple sources.
       *
       * @private
       * @param {Object} object The destination object.
       * @param {Object} source The source object.
       * @param {number} srcIndex The index of `source`.
       * @param {Function} [customizer] The function to customize merged values.
       * @param {Object} [stack] Tracks traversed source values and their merged
       *  counterparts.
       */
      function baseMerge(object, source, srcIndex, customizer, stack) {
        if (object === source) {
          return;
        }
        baseFor(source, function(srcValue, key) {
          stack || (stack = new Stack);
          if (isObject(srcValue)) {
            baseMergeDeep(object, source, key, srcIndex, baseMerge, customizer, stack);
          }
          else {
            var newValue = customizer
              ? customizer(safeGet(object, key), srcValue, (key + ''), object, source, stack)
              : undefined$1;

            if (newValue === undefined$1) {
              newValue = srcValue;
            }
            assignMergeValue(object, key, newValue);
          }
        }, keysIn);
      }

      /**
       * A specialized version of `baseMerge` for arrays and objects which performs
       * deep merges and tracks traversed objects enabling objects with circular
       * references to be merged.
       *
       * @private
       * @param {Object} object The destination object.
       * @param {Object} source The source object.
       * @param {string} key The key of the value to merge.
       * @param {number} srcIndex The index of `source`.
       * @param {Function} mergeFunc The function to merge values.
       * @param {Function} [customizer] The function to customize assigned values.
       * @param {Object} [stack] Tracks traversed source values and their merged
       *  counterparts.
       */
      function baseMergeDeep(object, source, key, srcIndex, mergeFunc, customizer, stack) {
        var objValue = safeGet(object, key),
            srcValue = safeGet(source, key),
            stacked = stack.get(srcValue);

        if (stacked) {
          assignMergeValue(object, key, stacked);
          return;
        }
        var newValue = customizer
          ? customizer(objValue, srcValue, (key + ''), object, source, stack)
          : undefined$1;

        var isCommon = newValue === undefined$1;

        if (isCommon) {
          var isArr = isArray(srcValue),
              isBuff = !isArr && isBuffer(srcValue),
              isTyped = !isArr && !isBuff && isTypedArray(srcValue);

          newValue = srcValue;
          if (isArr || isBuff || isTyped) {
            if (isArray(objValue)) {
              newValue = objValue;
            }
            else if (isArrayLikeObject(objValue)) {
              newValue = copyArray(objValue);
            }
            else if (isBuff) {
              isCommon = false;
              newValue = cloneBuffer(srcValue, true);
            }
            else if (isTyped) {
              isCommon = false;
              newValue = cloneTypedArray(srcValue, true);
            }
            else {
              newValue = [];
            }
          }
          else if (isPlainObject(srcValue) || isArguments(srcValue)) {
            newValue = objValue;
            if (isArguments(objValue)) {
              newValue = toPlainObject(objValue);
            }
            else if (!isObject(objValue) || isFunction(objValue)) {
              newValue = initCloneObject(srcValue);
            }
          }
          else {
            isCommon = false;
          }
        }
        if (isCommon) {
          // Recursively merge objects and arrays (susceptible to call stack limits).
          stack.set(srcValue, newValue);
          mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
          stack['delete'](srcValue);
        }
        assignMergeValue(object, key, newValue);
      }

      /**
       * The base implementation of `_.nth` which doesn't coerce arguments.
       *
       * @private
       * @param {Array} array The array to query.
       * @param {number} n The index of the element to return.
       * @returns {*} Returns the nth element of `array`.
       */
      function baseNth(array, n) {
        var length = array.length;
        if (!length) {
          return;
        }
        n += n < 0 ? length : 0;
        return isIndex(n, length) ? array[n] : undefined$1;
      }

      /**
       * The base implementation of `_.orderBy` without param guards.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function[]|Object[]|string[]} iteratees The iteratees to sort by.
       * @param {string[]} orders The sort orders of `iteratees`.
       * @returns {Array} Returns the new sorted array.
       */
      function baseOrderBy(collection, iteratees, orders) {
        var index = -1;
        iteratees = arrayMap(iteratees.length ? iteratees : [identity], baseUnary(getIteratee()));

        var result = baseMap(collection, function(value, key, collection) {
          var criteria = arrayMap(iteratees, function(iteratee) {
            return iteratee(value);
          });
          return { 'criteria': criteria, 'index': ++index, 'value': value };
        });

        return baseSortBy(result, function(object, other) {
          return compareMultiple(object, other, orders);
        });
      }

      /**
       * The base implementation of `_.pick` without support for individual
       * property identifiers.
       *
       * @private
       * @param {Object} object The source object.
       * @param {string[]} paths The property paths to pick.
       * @returns {Object} Returns the new object.
       */
      function basePick(object, paths) {
        return basePickBy(object, paths, function(value, path) {
          return hasIn(object, path);
        });
      }

      /**
       * The base implementation of  `_.pickBy` without support for iteratee shorthands.
       *
       * @private
       * @param {Object} object The source object.
       * @param {string[]} paths The property paths to pick.
       * @param {Function} predicate The function invoked per property.
       * @returns {Object} Returns the new object.
       */
      function basePickBy(object, paths, predicate) {
        var index = -1,
            length = paths.length,
            result = {};

        while (++index < length) {
          var path = paths[index],
              value = baseGet(object, path);

          if (predicate(value, path)) {
            baseSet(result, castPath(path, object), value);
          }
        }
        return result;
      }

      /**
       * A specialized version of `baseProperty` which supports deep paths.
       *
       * @private
       * @param {Array|string} path The path of the property to get.
       * @returns {Function} Returns the new accessor function.
       */
      function basePropertyDeep(path) {
        return function(object) {
          return baseGet(object, path);
        };
      }

      /**
       * The base implementation of `_.pullAllBy` without support for iteratee
       * shorthands.
       *
       * @private
       * @param {Array} array The array to modify.
       * @param {Array} values The values to remove.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns `array`.
       */
      function basePullAll(array, values, iteratee, comparator) {
        var indexOf = comparator ? baseIndexOfWith : baseIndexOf,
            index = -1,
            length = values.length,
            seen = array;

        if (array === values) {
          values = copyArray(values);
        }
        if (iteratee) {
          seen = arrayMap(array, baseUnary(iteratee));
        }
        while (++index < length) {
          var fromIndex = 0,
              value = values[index],
              computed = iteratee ? iteratee(value) : value;

          while ((fromIndex = indexOf(seen, computed, fromIndex, comparator)) > -1) {
            if (seen !== array) {
              splice.call(seen, fromIndex, 1);
            }
            splice.call(array, fromIndex, 1);
          }
        }
        return array;
      }

      /**
       * The base implementation of `_.pullAt` without support for individual
       * indexes or capturing the removed elements.
       *
       * @private
       * @param {Array} array The array to modify.
       * @param {number[]} indexes The indexes of elements to remove.
       * @returns {Array} Returns `array`.
       */
      function basePullAt(array, indexes) {
        var length = array ? indexes.length : 0,
            lastIndex = length - 1;

        while (length--) {
          var index = indexes[length];
          if (length == lastIndex || index !== previous) {
            var previous = index;
            if (isIndex(index)) {
              splice.call(array, index, 1);
            } else {
              baseUnset(array, index);
            }
          }
        }
        return array;
      }

      /**
       * The base implementation of `_.random` without support for returning
       * floating-point numbers.
       *
       * @private
       * @param {number} lower The lower bound.
       * @param {number} upper The upper bound.
       * @returns {number} Returns the random number.
       */
      function baseRandom(lower, upper) {
        return lower + nativeFloor(nativeRandom() * (upper - lower + 1));
      }

      /**
       * The base implementation of `_.range` and `_.rangeRight` which doesn't
       * coerce arguments.
       *
       * @private
       * @param {number} start The start of the range.
       * @param {number} end The end of the range.
       * @param {number} step The value to increment or decrement by.
       * @param {boolean} [fromRight] Specify iterating from right to left.
       * @returns {Array} Returns the range of numbers.
       */
      function baseRange(start, end, step, fromRight) {
        var index = -1,
            length = nativeMax(nativeCeil((end - start) / (step || 1)), 0),
            result = Array(length);

        while (length--) {
          result[fromRight ? length : ++index] = start;
          start += step;
        }
        return result;
      }

      /**
       * The base implementation of `_.repeat` which doesn't coerce arguments.
       *
       * @private
       * @param {string} string The string to repeat.
       * @param {number} n The number of times to repeat the string.
       * @returns {string} Returns the repeated string.
       */
      function baseRepeat(string, n) {
        var result = '';
        if (!string || n < 1 || n > MAX_SAFE_INTEGER) {
          return result;
        }
        // Leverage the exponentiation by squaring algorithm for a faster repeat.
        // See https://en.wikipedia.org/wiki/Exponentiation_by_squaring for more details.
        do {
          if (n % 2) {
            result += string;
          }
          n = nativeFloor(n / 2);
          if (n) {
            string += string;
          }
        } while (n);

        return result;
      }

      /**
       * The base implementation of `_.rest` which doesn't validate or coerce arguments.
       *
       * @private
       * @param {Function} func The function to apply a rest parameter to.
       * @param {number} [start=func.length-1] The start position of the rest parameter.
       * @returns {Function} Returns the new function.
       */
      function baseRest(func, start) {
        return setToString(overRest(func, start, identity), func + '');
      }

      /**
       * The base implementation of `_.sample`.
       *
       * @private
       * @param {Array|Object} collection The collection to sample.
       * @returns {*} Returns the random element.
       */
      function baseSample(collection) {
        return arraySample(values(collection));
      }

      /**
       * The base implementation of `_.sampleSize` without param guards.
       *
       * @private
       * @param {Array|Object} collection The collection to sample.
       * @param {number} n The number of elements to sample.
       * @returns {Array} Returns the random elements.
       */
      function baseSampleSize(collection, n) {
        var array = values(collection);
        return shuffleSelf(array, baseClamp(n, 0, array.length));
      }

      /**
       * The base implementation of `_.set`.
       *
       * @private
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to set.
       * @param {*} value The value to set.
       * @param {Function} [customizer] The function to customize path creation.
       * @returns {Object} Returns `object`.
       */
      function baseSet(object, path, value, customizer) {
        if (!isObject(object)) {
          return object;
        }
        path = castPath(path, object);

        var index = -1,
            length = path.length,
            lastIndex = length - 1,
            nested = object;

        while (nested != null && ++index < length) {
          var key = toKey(path[index]),
              newValue = value;

          if (index != lastIndex) {
            var objValue = nested[key];
            newValue = customizer ? customizer(objValue, key, nested) : undefined$1;
            if (newValue === undefined$1) {
              newValue = isObject(objValue)
                ? objValue
                : (isIndex(path[index + 1]) ? [] : {});
            }
          }
          assignValue(nested, key, newValue);
          nested = nested[key];
        }
        return object;
      }

      /**
       * The base implementation of `setData` without support for hot loop shorting.
       *
       * @private
       * @param {Function} func The function to associate metadata with.
       * @param {*} data The metadata.
       * @returns {Function} Returns `func`.
       */
      var baseSetData = !metaMap ? identity : function(func, data) {
        metaMap.set(func, data);
        return func;
      };

      /**
       * The base implementation of `setToString` without support for hot loop shorting.
       *
       * @private
       * @param {Function} func The function to modify.
       * @param {Function} string The `toString` result.
       * @returns {Function} Returns `func`.
       */
      var baseSetToString = !defineProperty ? identity : function(func, string) {
        return defineProperty(func, 'toString', {
          'configurable': true,
          'enumerable': false,
          'value': constant(string),
          'writable': true
        });
      };

      /**
       * The base implementation of `_.shuffle`.
       *
       * @private
       * @param {Array|Object} collection The collection to shuffle.
       * @returns {Array} Returns the new shuffled array.
       */
      function baseShuffle(collection) {
        return shuffleSelf(values(collection));
      }

      /**
       * The base implementation of `_.slice` without an iteratee call guard.
       *
       * @private
       * @param {Array} array The array to slice.
       * @param {number} [start=0] The start position.
       * @param {number} [end=array.length] The end position.
       * @returns {Array} Returns the slice of `array`.
       */
      function baseSlice(array, start, end) {
        var index = -1,
            length = array.length;

        if (start < 0) {
          start = -start > length ? 0 : (length + start);
        }
        end = end > length ? length : end;
        if (end < 0) {
          end += length;
        }
        length = start > end ? 0 : ((end - start) >>> 0);
        start >>>= 0;

        var result = Array(length);
        while (++index < length) {
          result[index] = array[index + start];
        }
        return result;
      }

      /**
       * The base implementation of `_.some` without support for iteratee shorthands.
       *
       * @private
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} predicate The function invoked per iteration.
       * @returns {boolean} Returns `true` if any element passes the predicate check,
       *  else `false`.
       */
      function baseSome(collection, predicate) {
        var result;

        baseEach(collection, function(value, index, collection) {
          result = predicate(value, index, collection);
          return !result;
        });
        return !!result;
      }

      /**
       * The base implementation of `_.sortedIndex` and `_.sortedLastIndex` which
       * performs a binary search of `array` to determine the index at which `value`
       * should be inserted into `array` in order to maintain its sort order.
       *
       * @private
       * @param {Array} array The sorted array to inspect.
       * @param {*} value The value to evaluate.
       * @param {boolean} [retHighest] Specify returning the highest qualified index.
       * @returns {number} Returns the index at which `value` should be inserted
       *  into `array`.
       */
      function baseSortedIndex(array, value, retHighest) {
        var low = 0,
            high = array == null ? low : array.length;

        if (typeof value == 'number' && value === value && high <= HALF_MAX_ARRAY_LENGTH) {
          while (low < high) {
            var mid = (low + high) >>> 1,
                computed = array[mid];

            if (computed !== null && !isSymbol(computed) &&
                (retHighest ? (computed <= value) : (computed < value))) {
              low = mid + 1;
            } else {
              high = mid;
            }
          }
          return high;
        }
        return baseSortedIndexBy(array, value, identity, retHighest);
      }

      /**
       * The base implementation of `_.sortedIndexBy` and `_.sortedLastIndexBy`
       * which invokes `iteratee` for `value` and each element of `array` to compute
       * their sort ranking. The iteratee is invoked with one argument; (value).
       *
       * @private
       * @param {Array} array The sorted array to inspect.
       * @param {*} value The value to evaluate.
       * @param {Function} iteratee The iteratee invoked per element.
       * @param {boolean} [retHighest] Specify returning the highest qualified index.
       * @returns {number} Returns the index at which `value` should be inserted
       *  into `array`.
       */
      function baseSortedIndexBy(array, value, iteratee, retHighest) {
        value = iteratee(value);

        var low = 0,
            high = array == null ? 0 : array.length,
            valIsNaN = value !== value,
            valIsNull = value === null,
            valIsSymbol = isSymbol(value),
            valIsUndefined = value === undefined$1;

        while (low < high) {
          var mid = nativeFloor((low + high) / 2),
              computed = iteratee(array[mid]),
              othIsDefined = computed !== undefined$1,
              othIsNull = computed === null,
              othIsReflexive = computed === computed,
              othIsSymbol = isSymbol(computed);

          if (valIsNaN) {
            var setLow = retHighest || othIsReflexive;
          } else if (valIsUndefined) {
            setLow = othIsReflexive && (retHighest || othIsDefined);
          } else if (valIsNull) {
            setLow = othIsReflexive && othIsDefined && (retHighest || !othIsNull);
          } else if (valIsSymbol) {
            setLow = othIsReflexive && othIsDefined && !othIsNull && (retHighest || !othIsSymbol);
          } else if (othIsNull || othIsSymbol) {
            setLow = false;
          } else {
            setLow = retHighest ? (computed <= value) : (computed < value);
          }
          if (setLow) {
            low = mid + 1;
          } else {
            high = mid;
          }
        }
        return nativeMin(high, MAX_ARRAY_INDEX);
      }

      /**
       * The base implementation of `_.sortedUniq` and `_.sortedUniqBy` without
       * support for iteratee shorthands.
       *
       * @private
       * @param {Array} array The array to inspect.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @returns {Array} Returns the new duplicate free array.
       */
      function baseSortedUniq(array, iteratee) {
        var index = -1,
            length = array.length,
            resIndex = 0,
            result = [];

        while (++index < length) {
          var value = array[index],
              computed = iteratee ? iteratee(value) : value;

          if (!index || !eq(computed, seen)) {
            var seen = computed;
            result[resIndex++] = value === 0 ? 0 : value;
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.toNumber` which doesn't ensure correct
       * conversions of binary, hexadecimal, or octal string values.
       *
       * @private
       * @param {*} value The value to process.
       * @returns {number} Returns the number.
       */
      function baseToNumber(value) {
        if (typeof value == 'number') {
          return value;
        }
        if (isSymbol(value)) {
          return NAN;
        }
        return +value;
      }

      /**
       * The base implementation of `_.toString` which doesn't convert nullish
       * values to empty strings.
       *
       * @private
       * @param {*} value The value to process.
       * @returns {string} Returns the string.
       */
      function baseToString(value) {
        // Exit early for strings to avoid a performance hit in some environments.
        if (typeof value == 'string') {
          return value;
        }
        if (isArray(value)) {
          // Recursively convert values (susceptible to call stack limits).
          return arrayMap(value, baseToString) + '';
        }
        if (isSymbol(value)) {
          return symbolToString ? symbolToString.call(value) : '';
        }
        var result = (value + '');
        return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
      }

      /**
       * The base implementation of `_.uniqBy` without support for iteratee shorthands.
       *
       * @private
       * @param {Array} array The array to inspect.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new duplicate free array.
       */
      function baseUniq(array, iteratee, comparator) {
        var index = -1,
            includes = arrayIncludes,
            length = array.length,
            isCommon = true,
            result = [],
            seen = result;

        if (comparator) {
          isCommon = false;
          includes = arrayIncludesWith;
        }
        else if (length >= LARGE_ARRAY_SIZE) {
          var set = iteratee ? null : createSet(array);
          if (set) {
            return setToArray(set);
          }
          isCommon = false;
          includes = cacheHas;
          seen = new SetCache;
        }
        else {
          seen = iteratee ? [] : result;
        }
        outer:
        while (++index < length) {
          var value = array[index],
              computed = iteratee ? iteratee(value) : value;

          value = (comparator || value !== 0) ? value : 0;
          if (isCommon && computed === computed) {
            var seenIndex = seen.length;
            while (seenIndex--) {
              if (seen[seenIndex] === computed) {
                continue outer;
              }
            }
            if (iteratee) {
              seen.push(computed);
            }
            result.push(value);
          }
          else if (!includes(seen, computed, comparator)) {
            if (seen !== result) {
              seen.push(computed);
            }
            result.push(value);
          }
        }
        return result;
      }

      /**
       * The base implementation of `_.unset`.
       *
       * @private
       * @param {Object} object The object to modify.
       * @param {Array|string} path The property path to unset.
       * @returns {boolean} Returns `true` if the property is deleted, else `false`.
       */
      function baseUnset(object, path) {
        path = castPath(path, object);
        object = parent(object, path);
        return object == null || delete object[toKey(last(path))];
      }

      /**
       * The base implementation of `_.update`.
       *
       * @private
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to update.
       * @param {Function} updater The function to produce the updated value.
       * @param {Function} [customizer] The function to customize path creation.
       * @returns {Object} Returns `object`.
       */
      function baseUpdate(object, path, updater, customizer) {
        return baseSet(object, path, updater(baseGet(object, path)), customizer);
      }

      /**
       * The base implementation of methods like `_.dropWhile` and `_.takeWhile`
       * without support for iteratee shorthands.
       *
       * @private
       * @param {Array} array The array to query.
       * @param {Function} predicate The function invoked per iteration.
       * @param {boolean} [isDrop] Specify dropping elements instead of taking them.
       * @param {boolean} [fromRight] Specify iterating from right to left.
       * @returns {Array} Returns the slice of `array`.
       */
      function baseWhile(array, predicate, isDrop, fromRight) {
        var length = array.length,
            index = fromRight ? length : -1;

        while ((fromRight ? index-- : ++index < length) &&
          predicate(array[index], index, array)) {}

        return isDrop
          ? baseSlice(array, (fromRight ? 0 : index), (fromRight ? index + 1 : length))
          : baseSlice(array, (fromRight ? index + 1 : 0), (fromRight ? length : index));
      }

      /**
       * The base implementation of `wrapperValue` which returns the result of
       * performing a sequence of actions on the unwrapped `value`, where each
       * successive action is supplied the return value of the previous.
       *
       * @private
       * @param {*} value The unwrapped value.
       * @param {Array} actions Actions to perform to resolve the unwrapped value.
       * @returns {*} Returns the resolved value.
       */
      function baseWrapperValue(value, actions) {
        var result = value;
        if (result instanceof LazyWrapper) {
          result = result.value();
        }
        return arrayReduce(actions, function(result, action) {
          return action.func.apply(action.thisArg, arrayPush([result], action.args));
        }, result);
      }

      /**
       * The base implementation of methods like `_.xor`, without support for
       * iteratee shorthands, that accepts an array of arrays to inspect.
       *
       * @private
       * @param {Array} arrays The arrays to inspect.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of values.
       */
      function baseXor(arrays, iteratee, comparator) {
        var length = arrays.length;
        if (length < 2) {
          return length ? baseUniq(arrays[0]) : [];
        }
        var index = -1,
            result = Array(length);

        while (++index < length) {
          var array = arrays[index],
              othIndex = -1;

          while (++othIndex < length) {
            if (othIndex != index) {
              result[index] = baseDifference(result[index] || array, arrays[othIndex], iteratee, comparator);
            }
          }
        }
        return baseUniq(baseFlatten(result, 1), iteratee, comparator);
      }

      /**
       * This base implementation of `_.zipObject` which assigns values using `assignFunc`.
       *
       * @private
       * @param {Array} props The property identifiers.
       * @param {Array} values The property values.
       * @param {Function} assignFunc The function to assign values.
       * @returns {Object} Returns the new object.
       */
      function baseZipObject(props, values, assignFunc) {
        var index = -1,
            length = props.length,
            valsLength = values.length,
            result = {};

        while (++index < length) {
          var value = index < valsLength ? values[index] : undefined$1;
          assignFunc(result, props[index], value);
        }
        return result;
      }

      /**
       * Casts `value` to an empty array if it's not an array like object.
       *
       * @private
       * @param {*} value The value to inspect.
       * @returns {Array|Object} Returns the cast array-like object.
       */
      function castArrayLikeObject(value) {
        return isArrayLikeObject(value) ? value : [];
      }

      /**
       * Casts `value` to `identity` if it's not a function.
       *
       * @private
       * @param {*} value The value to inspect.
       * @returns {Function} Returns cast function.
       */
      function castFunction(value) {
        return typeof value == 'function' ? value : identity;
      }

      /**
       * Casts `value` to a path array if it's not one.
       *
       * @private
       * @param {*} value The value to inspect.
       * @param {Object} [object] The object to query keys on.
       * @returns {Array} Returns the cast property path array.
       */
      function castPath(value, object) {
        if (isArray(value)) {
          return value;
        }
        return isKey(value, object) ? [value] : stringToPath(toString(value));
      }

      /**
       * A `baseRest` alias which can be replaced with `identity` by module
       * replacement plugins.
       *
       * @private
       * @type {Function}
       * @param {Function} func The function to apply a rest parameter to.
       * @returns {Function} Returns the new function.
       */
      var castRest = baseRest;

      /**
       * Casts `array` to a slice if it's needed.
       *
       * @private
       * @param {Array} array The array to inspect.
       * @param {number} start The start position.
       * @param {number} [end=array.length] The end position.
       * @returns {Array} Returns the cast slice.
       */
      function castSlice(array, start, end) {
        var length = array.length;
        end = end === undefined$1 ? length : end;
        return (!start && end >= length) ? array : baseSlice(array, start, end);
      }

      /**
       * A simple wrapper around the global [`clearTimeout`](https://mdn.io/clearTimeout).
       *
       * @private
       * @param {number|Object} id The timer id or timeout object of the timer to clear.
       */
      var clearTimeout = ctxClearTimeout || function(id) {
        return root.clearTimeout(id);
      };

      /**
       * Creates a clone of  `buffer`.
       *
       * @private
       * @param {Buffer} buffer The buffer to clone.
       * @param {boolean} [isDeep] Specify a deep clone.
       * @returns {Buffer} Returns the cloned buffer.
       */
      function cloneBuffer(buffer, isDeep) {
        if (isDeep) {
          return buffer.slice();
        }
        var length = buffer.length,
            result = allocUnsafe ? allocUnsafe(length) : new buffer.constructor(length);

        buffer.copy(result);
        return result;
      }

      /**
       * Creates a clone of `arrayBuffer`.
       *
       * @private
       * @param {ArrayBuffer} arrayBuffer The array buffer to clone.
       * @returns {ArrayBuffer} Returns the cloned array buffer.
       */
      function cloneArrayBuffer(arrayBuffer) {
        var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
        new Uint8Array(result).set(new Uint8Array(arrayBuffer));
        return result;
      }

      /**
       * Creates a clone of `dataView`.
       *
       * @private
       * @param {Object} dataView The data view to clone.
       * @param {boolean} [isDeep] Specify a deep clone.
       * @returns {Object} Returns the cloned data view.
       */
      function cloneDataView(dataView, isDeep) {
        var buffer = isDeep ? cloneArrayBuffer(dataView.buffer) : dataView.buffer;
        return new dataView.constructor(buffer, dataView.byteOffset, dataView.byteLength);
      }

      /**
       * Creates a clone of `regexp`.
       *
       * @private
       * @param {Object} regexp The regexp to clone.
       * @returns {Object} Returns the cloned regexp.
       */
      function cloneRegExp(regexp) {
        var result = new regexp.constructor(regexp.source, reFlags.exec(regexp));
        result.lastIndex = regexp.lastIndex;
        return result;
      }

      /**
       * Creates a clone of the `symbol` object.
       *
       * @private
       * @param {Object} symbol The symbol object to clone.
       * @returns {Object} Returns the cloned symbol object.
       */
      function cloneSymbol(symbol) {
        return symbolValueOf ? Object(symbolValueOf.call(symbol)) : {};
      }

      /**
       * Creates a clone of `typedArray`.
       *
       * @private
       * @param {Object} typedArray The typed array to clone.
       * @param {boolean} [isDeep] Specify a deep clone.
       * @returns {Object} Returns the cloned typed array.
       */
      function cloneTypedArray(typedArray, isDeep) {
        var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
        return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
      }

      /**
       * Compares values to sort them in ascending order.
       *
       * @private
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {number} Returns the sort order indicator for `value`.
       */
      function compareAscending(value, other) {
        if (value !== other) {
          var valIsDefined = value !== undefined$1,
              valIsNull = value === null,
              valIsReflexive = value === value,
              valIsSymbol = isSymbol(value);

          var othIsDefined = other !== undefined$1,
              othIsNull = other === null,
              othIsReflexive = other === other,
              othIsSymbol = isSymbol(other);

          if ((!othIsNull && !othIsSymbol && !valIsSymbol && value > other) ||
              (valIsSymbol && othIsDefined && othIsReflexive && !othIsNull && !othIsSymbol) ||
              (valIsNull && othIsDefined && othIsReflexive) ||
              (!valIsDefined && othIsReflexive) ||
              !valIsReflexive) {
            return 1;
          }
          if ((!valIsNull && !valIsSymbol && !othIsSymbol && value < other) ||
              (othIsSymbol && valIsDefined && valIsReflexive && !valIsNull && !valIsSymbol) ||
              (othIsNull && valIsDefined && valIsReflexive) ||
              (!othIsDefined && valIsReflexive) ||
              !othIsReflexive) {
            return -1;
          }
        }
        return 0;
      }

      /**
       * Used by `_.orderBy` to compare multiple properties of a value to another
       * and stable sort them.
       *
       * If `orders` is unspecified, all values are sorted in ascending order. Otherwise,
       * specify an order of "desc" for descending or "asc" for ascending sort order
       * of corresponding values.
       *
       * @private
       * @param {Object} object The object to compare.
       * @param {Object} other The other object to compare.
       * @param {boolean[]|string[]} orders The order to sort by for each property.
       * @returns {number} Returns the sort order indicator for `object`.
       */
      function compareMultiple(object, other, orders) {
        var index = -1,
            objCriteria = object.criteria,
            othCriteria = other.criteria,
            length = objCriteria.length,
            ordersLength = orders.length;

        while (++index < length) {
          var result = compareAscending(objCriteria[index], othCriteria[index]);
          if (result) {
            if (index >= ordersLength) {
              return result;
            }
            var order = orders[index];
            return result * (order == 'desc' ? -1 : 1);
          }
        }
        // Fixes an `Array#sort` bug in the JS engine embedded in Adobe applications
        // that causes it, under certain circumstances, to provide the same value for
        // `object` and `other`. See https://github.com/jashkenas/underscore/pull/1247
        // for more details.
        //
        // This also ensures a stable sort in V8 and other engines.
        // See https://bugs.chromium.org/p/v8/issues/detail?id=90 for more details.
        return object.index - other.index;
      }

      /**
       * Creates an array that is the composition of partially applied arguments,
       * placeholders, and provided arguments into a single array of arguments.
       *
       * @private
       * @param {Array} args The provided arguments.
       * @param {Array} partials The arguments to prepend to those provided.
       * @param {Array} holders The `partials` placeholder indexes.
       * @params {boolean} [isCurried] Specify composing for a curried function.
       * @returns {Array} Returns the new array of composed arguments.
       */
      function composeArgs(args, partials, holders, isCurried) {
        var argsIndex = -1,
            argsLength = args.length,
            holdersLength = holders.length,
            leftIndex = -1,
            leftLength = partials.length,
            rangeLength = nativeMax(argsLength - holdersLength, 0),
            result = Array(leftLength + rangeLength),
            isUncurried = !isCurried;

        while (++leftIndex < leftLength) {
          result[leftIndex] = partials[leftIndex];
        }
        while (++argsIndex < holdersLength) {
          if (isUncurried || argsIndex < argsLength) {
            result[holders[argsIndex]] = args[argsIndex];
          }
        }
        while (rangeLength--) {
          result[leftIndex++] = args[argsIndex++];
        }
        return result;
      }

      /**
       * This function is like `composeArgs` except that the arguments composition
       * is tailored for `_.partialRight`.
       *
       * @private
       * @param {Array} args The provided arguments.
       * @param {Array} partials The arguments to append to those provided.
       * @param {Array} holders The `partials` placeholder indexes.
       * @params {boolean} [isCurried] Specify composing for a curried function.
       * @returns {Array} Returns the new array of composed arguments.
       */
      function composeArgsRight(args, partials, holders, isCurried) {
        var argsIndex = -1,
            argsLength = args.length,
            holdersIndex = -1,
            holdersLength = holders.length,
            rightIndex = -1,
            rightLength = partials.length,
            rangeLength = nativeMax(argsLength - holdersLength, 0),
            result = Array(rangeLength + rightLength),
            isUncurried = !isCurried;

        while (++argsIndex < rangeLength) {
          result[argsIndex] = args[argsIndex];
        }
        var offset = argsIndex;
        while (++rightIndex < rightLength) {
          result[offset + rightIndex] = partials[rightIndex];
        }
        while (++holdersIndex < holdersLength) {
          if (isUncurried || argsIndex < argsLength) {
            result[offset + holders[holdersIndex]] = args[argsIndex++];
          }
        }
        return result;
      }

      /**
       * Copies the values of `source` to `array`.
       *
       * @private
       * @param {Array} source The array to copy values from.
       * @param {Array} [array=[]] The array to copy values to.
       * @returns {Array} Returns `array`.
       */
      function copyArray(source, array) {
        var index = -1,
            length = source.length;

        array || (array = Array(length));
        while (++index < length) {
          array[index] = source[index];
        }
        return array;
      }

      /**
       * Copies properties of `source` to `object`.
       *
       * @private
       * @param {Object} source The object to copy properties from.
       * @param {Array} props The property identifiers to copy.
       * @param {Object} [object={}] The object to copy properties to.
       * @param {Function} [customizer] The function to customize copied values.
       * @returns {Object} Returns `object`.
       */
      function copyObject(source, props, object, customizer) {
        var isNew = !object;
        object || (object = {});

        var index = -1,
            length = props.length;

        while (++index < length) {
          var key = props[index];

          var newValue = customizer
            ? customizer(object[key], source[key], key, object, source)
            : undefined$1;

          if (newValue === undefined$1) {
            newValue = source[key];
          }
          if (isNew) {
            baseAssignValue(object, key, newValue);
          } else {
            assignValue(object, key, newValue);
          }
        }
        return object;
      }

      /**
       * Copies own symbols of `source` to `object`.
       *
       * @private
       * @param {Object} source The object to copy symbols from.
       * @param {Object} [object={}] The object to copy symbols to.
       * @returns {Object} Returns `object`.
       */
      function copySymbols(source, object) {
        return copyObject(source, getSymbols(source), object);
      }

      /**
       * Copies own and inherited symbols of `source` to `object`.
       *
       * @private
       * @param {Object} source The object to copy symbols from.
       * @param {Object} [object={}] The object to copy symbols to.
       * @returns {Object} Returns `object`.
       */
      function copySymbolsIn(source, object) {
        return copyObject(source, getSymbolsIn(source), object);
      }

      /**
       * Creates a function like `_.groupBy`.
       *
       * @private
       * @param {Function} setter The function to set accumulator values.
       * @param {Function} [initializer] The accumulator object initializer.
       * @returns {Function} Returns the new aggregator function.
       */
      function createAggregator(setter, initializer) {
        return function(collection, iteratee) {
          var func = isArray(collection) ? arrayAggregator : baseAggregator,
              accumulator = initializer ? initializer() : {};

          return func(collection, setter, getIteratee(iteratee, 2), accumulator);
        };
      }

      /**
       * Creates a function like `_.assign`.
       *
       * @private
       * @param {Function} assigner The function to assign values.
       * @returns {Function} Returns the new assigner function.
       */
      function createAssigner(assigner) {
        return baseRest(function(object, sources) {
          var index = -1,
              length = sources.length,
              customizer = length > 1 ? sources[length - 1] : undefined$1,
              guard = length > 2 ? sources[2] : undefined$1;

          customizer = (assigner.length > 3 && typeof customizer == 'function')
            ? (length--, customizer)
            : undefined$1;

          if (guard && isIterateeCall(sources[0], sources[1], guard)) {
            customizer = length < 3 ? undefined$1 : customizer;
            length = 1;
          }
          object = Object(object);
          while (++index < length) {
            var source = sources[index];
            if (source) {
              assigner(object, source, index, customizer);
            }
          }
          return object;
        });
      }

      /**
       * Creates a `baseEach` or `baseEachRight` function.
       *
       * @private
       * @param {Function} eachFunc The function to iterate over a collection.
       * @param {boolean} [fromRight] Specify iterating from right to left.
       * @returns {Function} Returns the new base function.
       */
      function createBaseEach(eachFunc, fromRight) {
        return function(collection, iteratee) {
          if (collection == null) {
            return collection;
          }
          if (!isArrayLike(collection)) {
            return eachFunc(collection, iteratee);
          }
          var length = collection.length,
              index = fromRight ? length : -1,
              iterable = Object(collection);

          while ((fromRight ? index-- : ++index < length)) {
            if (iteratee(iterable[index], index, iterable) === false) {
              break;
            }
          }
          return collection;
        };
      }

      /**
       * Creates a base function for methods like `_.forIn` and `_.forOwn`.
       *
       * @private
       * @param {boolean} [fromRight] Specify iterating from right to left.
       * @returns {Function} Returns the new base function.
       */
      function createBaseFor(fromRight) {
        return function(object, iteratee, keysFunc) {
          var index = -1,
              iterable = Object(object),
              props = keysFunc(object),
              length = props.length;

          while (length--) {
            var key = props[fromRight ? length : ++index];
            if (iteratee(iterable[key], key, iterable) === false) {
              break;
            }
          }
          return object;
        };
      }

      /**
       * Creates a function that wraps `func` to invoke it with the optional `this`
       * binding of `thisArg`.
       *
       * @private
       * @param {Function} func The function to wrap.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @param {*} [thisArg] The `this` binding of `func`.
       * @returns {Function} Returns the new wrapped function.
       */
      function createBind(func, bitmask, thisArg) {
        var isBind = bitmask & WRAP_BIND_FLAG,
            Ctor = createCtor(func);

        function wrapper() {
          var fn = (this && this !== root && this instanceof wrapper) ? Ctor : func;
          return fn.apply(isBind ? thisArg : this, arguments);
        }
        return wrapper;
      }

      /**
       * Creates a function like `_.lowerFirst`.
       *
       * @private
       * @param {string} methodName The name of the `String` case method to use.
       * @returns {Function} Returns the new case function.
       */
      function createCaseFirst(methodName) {
        return function(string) {
          string = toString(string);

          var strSymbols = hasUnicode(string)
            ? stringToArray(string)
            : undefined$1;

          var chr = strSymbols
            ? strSymbols[0]
            : string.charAt(0);

          var trailing = strSymbols
            ? castSlice(strSymbols, 1).join('')
            : string.slice(1);

          return chr[methodName]() + trailing;
        };
      }

      /**
       * Creates a function like `_.camelCase`.
       *
       * @private
       * @param {Function} callback The function to combine each word.
       * @returns {Function} Returns the new compounder function.
       */
      function createCompounder(callback) {
        return function(string) {
          return arrayReduce(words(deburr(string).replace(reApos, '')), callback, '');
        };
      }

      /**
       * Creates a function that produces an instance of `Ctor` regardless of
       * whether it was invoked as part of a `new` expression or by `call` or `apply`.
       *
       * @private
       * @param {Function} Ctor The constructor to wrap.
       * @returns {Function} Returns the new wrapped function.
       */
      function createCtor(Ctor) {
        return function() {
          // Use a `switch` statement to work with class constructors. See
          // http://ecma-international.org/ecma-262/7.0/#sec-ecmascript-function-objects-call-thisargument-argumentslist
          // for more details.
          var args = arguments;
          switch (args.length) {
            case 0: return new Ctor;
            case 1: return new Ctor(args[0]);
            case 2: return new Ctor(args[0], args[1]);
            case 3: return new Ctor(args[0], args[1], args[2]);
            case 4: return new Ctor(args[0], args[1], args[2], args[3]);
            case 5: return new Ctor(args[0], args[1], args[2], args[3], args[4]);
            case 6: return new Ctor(args[0], args[1], args[2], args[3], args[4], args[5]);
            case 7: return new Ctor(args[0], args[1], args[2], args[3], args[4], args[5], args[6]);
          }
          var thisBinding = baseCreate(Ctor.prototype),
              result = Ctor.apply(thisBinding, args);

          // Mimic the constructor's `return` behavior.
          // See https://es5.github.io/#x13.2.2 for more details.
          return isObject(result) ? result : thisBinding;
        };
      }

      /**
       * Creates a function that wraps `func` to enable currying.
       *
       * @private
       * @param {Function} func The function to wrap.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @param {number} arity The arity of `func`.
       * @returns {Function} Returns the new wrapped function.
       */
      function createCurry(func, bitmask, arity) {
        var Ctor = createCtor(func);

        function wrapper() {
          var length = arguments.length,
              args = Array(length),
              index = length,
              placeholder = getHolder(wrapper);

          while (index--) {
            args[index] = arguments[index];
          }
          var holders = (length < 3 && args[0] !== placeholder && args[length - 1] !== placeholder)
            ? []
            : replaceHolders(args, placeholder);

          length -= holders.length;
          if (length < arity) {
            return createRecurry(
              func, bitmask, createHybrid, wrapper.placeholder, undefined$1,
              args, holders, undefined$1, undefined$1, arity - length);
          }
          var fn = (this && this !== root && this instanceof wrapper) ? Ctor : func;
          return apply(fn, this, args);
        }
        return wrapper;
      }

      /**
       * Creates a `_.find` or `_.findLast` function.
       *
       * @private
       * @param {Function} findIndexFunc The function to find the collection index.
       * @returns {Function} Returns the new find function.
       */
      function createFind(findIndexFunc) {
        return function(collection, predicate, fromIndex) {
          var iterable = Object(collection);
          if (!isArrayLike(collection)) {
            var iteratee = getIteratee(predicate, 3);
            collection = keys(collection);
            predicate = function(key) { return iteratee(iterable[key], key, iterable); };
          }
          var index = findIndexFunc(collection, predicate, fromIndex);
          return index > -1 ? iterable[iteratee ? collection[index] : index] : undefined$1;
        };
      }

      /**
       * Creates a `_.flow` or `_.flowRight` function.
       *
       * @private
       * @param {boolean} [fromRight] Specify iterating from right to left.
       * @returns {Function} Returns the new flow function.
       */
      function createFlow(fromRight) {
        return flatRest(function(funcs) {
          var length = funcs.length,
              index = length,
              prereq = LodashWrapper.prototype.thru;

          if (fromRight) {
            funcs.reverse();
          }
          while (index--) {
            var func = funcs[index];
            if (typeof func != 'function') {
              throw new TypeError(FUNC_ERROR_TEXT);
            }
            if (prereq && !wrapper && getFuncName(func) == 'wrapper') {
              var wrapper = new LodashWrapper([], true);
            }
          }
          index = wrapper ? index : length;
          while (++index < length) {
            func = funcs[index];

            var funcName = getFuncName(func),
                data = funcName == 'wrapper' ? getData(func) : undefined$1;

            if (data && isLaziable(data[0]) &&
                  data[1] == (WRAP_ARY_FLAG | WRAP_CURRY_FLAG | WRAP_PARTIAL_FLAG | WRAP_REARG_FLAG) &&
                  !data[4].length && data[9] == 1
                ) {
              wrapper = wrapper[getFuncName(data[0])].apply(wrapper, data[3]);
            } else {
              wrapper = (func.length == 1 && isLaziable(func))
                ? wrapper[funcName]()
                : wrapper.thru(func);
            }
          }
          return function() {
            var args = arguments,
                value = args[0];

            if (wrapper && args.length == 1 && isArray(value)) {
              return wrapper.plant(value).value();
            }
            var index = 0,
                result = length ? funcs[index].apply(this, args) : value;

            while (++index < length) {
              result = funcs[index].call(this, result);
            }
            return result;
          };
        });
      }

      /**
       * Creates a function that wraps `func` to invoke it with optional `this`
       * binding of `thisArg`, partial application, and currying.
       *
       * @private
       * @param {Function|string} func The function or method name to wrap.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @param {*} [thisArg] The `this` binding of `func`.
       * @param {Array} [partials] The arguments to prepend to those provided to
       *  the new function.
       * @param {Array} [holders] The `partials` placeholder indexes.
       * @param {Array} [partialsRight] The arguments to append to those provided
       *  to the new function.
       * @param {Array} [holdersRight] The `partialsRight` placeholder indexes.
       * @param {Array} [argPos] The argument positions of the new function.
       * @param {number} [ary] The arity cap of `func`.
       * @param {number} [arity] The arity of `func`.
       * @returns {Function} Returns the new wrapped function.
       */
      function createHybrid(func, bitmask, thisArg, partials, holders, partialsRight, holdersRight, argPos, ary, arity) {
        var isAry = bitmask & WRAP_ARY_FLAG,
            isBind = bitmask & WRAP_BIND_FLAG,
            isBindKey = bitmask & WRAP_BIND_KEY_FLAG,
            isCurried = bitmask & (WRAP_CURRY_FLAG | WRAP_CURRY_RIGHT_FLAG),
            isFlip = bitmask & WRAP_FLIP_FLAG,
            Ctor = isBindKey ? undefined$1 : createCtor(func);

        function wrapper() {
          var length = arguments.length,
              args = Array(length),
              index = length;

          while (index--) {
            args[index] = arguments[index];
          }
          if (isCurried) {
            var placeholder = getHolder(wrapper),
                holdersCount = countHolders(args, placeholder);
          }
          if (partials) {
            args = composeArgs(args, partials, holders, isCurried);
          }
          if (partialsRight) {
            args = composeArgsRight(args, partialsRight, holdersRight, isCurried);
          }
          length -= holdersCount;
          if (isCurried && length < arity) {
            var newHolders = replaceHolders(args, placeholder);
            return createRecurry(
              func, bitmask, createHybrid, wrapper.placeholder, thisArg,
              args, newHolders, argPos, ary, arity - length
            );
          }
          var thisBinding = isBind ? thisArg : this,
              fn = isBindKey ? thisBinding[func] : func;

          length = args.length;
          if (argPos) {
            args = reorder(args, argPos);
          } else if (isFlip && length > 1) {
            args.reverse();
          }
          if (isAry && ary < length) {
            args.length = ary;
          }
          if (this && this !== root && this instanceof wrapper) {
            fn = Ctor || createCtor(fn);
          }
          return fn.apply(thisBinding, args);
        }
        return wrapper;
      }

      /**
       * Creates a function like `_.invertBy`.
       *
       * @private
       * @param {Function} setter The function to set accumulator values.
       * @param {Function} toIteratee The function to resolve iteratees.
       * @returns {Function} Returns the new inverter function.
       */
      function createInverter(setter, toIteratee) {
        return function(object, iteratee) {
          return baseInverter(object, setter, toIteratee(iteratee), {});
        };
      }

      /**
       * Creates a function that performs a mathematical operation on two values.
       *
       * @private
       * @param {Function} operator The function to perform the operation.
       * @param {number} [defaultValue] The value used for `undefined` arguments.
       * @returns {Function} Returns the new mathematical operation function.
       */
      function createMathOperation(operator, defaultValue) {
        return function(value, other) {
          var result;
          if (value === undefined$1 && other === undefined$1) {
            return defaultValue;
          }
          if (value !== undefined$1) {
            result = value;
          }
          if (other !== undefined$1) {
            if (result === undefined$1) {
              return other;
            }
            if (typeof value == 'string' || typeof other == 'string') {
              value = baseToString(value);
              other = baseToString(other);
            } else {
              value = baseToNumber(value);
              other = baseToNumber(other);
            }
            result = operator(value, other);
          }
          return result;
        };
      }

      /**
       * Creates a function like `_.over`.
       *
       * @private
       * @param {Function} arrayFunc The function to iterate over iteratees.
       * @returns {Function} Returns the new over function.
       */
      function createOver(arrayFunc) {
        return flatRest(function(iteratees) {
          iteratees = arrayMap(iteratees, baseUnary(getIteratee()));
          return baseRest(function(args) {
            var thisArg = this;
            return arrayFunc(iteratees, function(iteratee) {
              return apply(iteratee, thisArg, args);
            });
          });
        });
      }

      /**
       * Creates the padding for `string` based on `length`. The `chars` string
       * is truncated if the number of characters exceeds `length`.
       *
       * @private
       * @param {number} length The padding length.
       * @param {string} [chars=' '] The string used as padding.
       * @returns {string} Returns the padding for `string`.
       */
      function createPadding(length, chars) {
        chars = chars === undefined$1 ? ' ' : baseToString(chars);

        var charsLength = chars.length;
        if (charsLength < 2) {
          return charsLength ? baseRepeat(chars, length) : chars;
        }
        var result = baseRepeat(chars, nativeCeil(length / stringSize(chars)));
        return hasUnicode(chars)
          ? castSlice(stringToArray(result), 0, length).join('')
          : result.slice(0, length);
      }

      /**
       * Creates a function that wraps `func` to invoke it with the `this` binding
       * of `thisArg` and `partials` prepended to the arguments it receives.
       *
       * @private
       * @param {Function} func The function to wrap.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @param {*} thisArg The `this` binding of `func`.
       * @param {Array} partials The arguments to prepend to those provided to
       *  the new function.
       * @returns {Function} Returns the new wrapped function.
       */
      function createPartial(func, bitmask, thisArg, partials) {
        var isBind = bitmask & WRAP_BIND_FLAG,
            Ctor = createCtor(func);

        function wrapper() {
          var argsIndex = -1,
              argsLength = arguments.length,
              leftIndex = -1,
              leftLength = partials.length,
              args = Array(leftLength + argsLength),
              fn = (this && this !== root && this instanceof wrapper) ? Ctor : func;

          while (++leftIndex < leftLength) {
            args[leftIndex] = partials[leftIndex];
          }
          while (argsLength--) {
            args[leftIndex++] = arguments[++argsIndex];
          }
          return apply(fn, isBind ? thisArg : this, args);
        }
        return wrapper;
      }

      /**
       * Creates a `_.range` or `_.rangeRight` function.
       *
       * @private
       * @param {boolean} [fromRight] Specify iterating from right to left.
       * @returns {Function} Returns the new range function.
       */
      function createRange(fromRight) {
        return function(start, end, step) {
          if (step && typeof step != 'number' && isIterateeCall(start, end, step)) {
            end = step = undefined$1;
          }
          // Ensure the sign of `-0` is preserved.
          start = toFinite(start);
          if (end === undefined$1) {
            end = start;
            start = 0;
          } else {
            end = toFinite(end);
          }
          step = step === undefined$1 ? (start < end ? 1 : -1) : toFinite(step);
          return baseRange(start, end, step, fromRight);
        };
      }

      /**
       * Creates a function that performs a relational operation on two values.
       *
       * @private
       * @param {Function} operator The function to perform the operation.
       * @returns {Function} Returns the new relational operation function.
       */
      function createRelationalOperation(operator) {
        return function(value, other) {
          if (!(typeof value == 'string' && typeof other == 'string')) {
            value = toNumber(value);
            other = toNumber(other);
          }
          return operator(value, other);
        };
      }

      /**
       * Creates a function that wraps `func` to continue currying.
       *
       * @private
       * @param {Function} func The function to wrap.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @param {Function} wrapFunc The function to create the `func` wrapper.
       * @param {*} placeholder The placeholder value.
       * @param {*} [thisArg] The `this` binding of `func`.
       * @param {Array} [partials] The arguments to prepend to those provided to
       *  the new function.
       * @param {Array} [holders] The `partials` placeholder indexes.
       * @param {Array} [argPos] The argument positions of the new function.
       * @param {number} [ary] The arity cap of `func`.
       * @param {number} [arity] The arity of `func`.
       * @returns {Function} Returns the new wrapped function.
       */
      function createRecurry(func, bitmask, wrapFunc, placeholder, thisArg, partials, holders, argPos, ary, arity) {
        var isCurry = bitmask & WRAP_CURRY_FLAG,
            newHolders = isCurry ? holders : undefined$1,
            newHoldersRight = isCurry ? undefined$1 : holders,
            newPartials = isCurry ? partials : undefined$1,
            newPartialsRight = isCurry ? undefined$1 : partials;

        bitmask |= (isCurry ? WRAP_PARTIAL_FLAG : WRAP_PARTIAL_RIGHT_FLAG);
        bitmask &= ~(isCurry ? WRAP_PARTIAL_RIGHT_FLAG : WRAP_PARTIAL_FLAG);

        if (!(bitmask & WRAP_CURRY_BOUND_FLAG)) {
          bitmask &= ~(WRAP_BIND_FLAG | WRAP_BIND_KEY_FLAG);
        }
        var newData = [
          func, bitmask, thisArg, newPartials, newHolders, newPartialsRight,
          newHoldersRight, argPos, ary, arity
        ];

        var result = wrapFunc.apply(undefined$1, newData);
        if (isLaziable(func)) {
          setData(result, newData);
        }
        result.placeholder = placeholder;
        return setWrapToString(result, func, bitmask);
      }

      /**
       * Creates a function like `_.round`.
       *
       * @private
       * @param {string} methodName The name of the `Math` method to use when rounding.
       * @returns {Function} Returns the new round function.
       */
      function createRound(methodName) {
        var func = Math[methodName];
        return function(number, precision) {
          number = toNumber(number);
          precision = precision == null ? 0 : nativeMin(toInteger(precision), 292);
          if (precision && nativeIsFinite(number)) {
            // Shift with exponential notation to avoid floating-point issues.
            // See [MDN](https://mdn.io/round#Examples) for more details.
            var pair = (toString(number) + 'e').split('e'),
                value = func(pair[0] + 'e' + (+pair[1] + precision));

            pair = (toString(value) + 'e').split('e');
            return +(pair[0] + 'e' + (+pair[1] - precision));
          }
          return func(number);
        };
      }

      /**
       * Creates a set object of `values`.
       *
       * @private
       * @param {Array} values The values to add to the set.
       * @returns {Object} Returns the new set.
       */
      var createSet = !(Set && (1 / setToArray(new Set([,-0]))[1]) == INFINITY) ? noop : function(values) {
        return new Set(values);
      };

      /**
       * Creates a `_.toPairs` or `_.toPairsIn` function.
       *
       * @private
       * @param {Function} keysFunc The function to get the keys of a given object.
       * @returns {Function} Returns the new pairs function.
       */
      function createToPairs(keysFunc) {
        return function(object) {
          var tag = getTag(object);
          if (tag == mapTag) {
            return mapToArray(object);
          }
          if (tag == setTag) {
            return setToPairs(object);
          }
          return baseToPairs(object, keysFunc(object));
        };
      }

      /**
       * Creates a function that either curries or invokes `func` with optional
       * `this` binding and partially applied arguments.
       *
       * @private
       * @param {Function|string} func The function or method name to wrap.
       * @param {number} bitmask The bitmask flags.
       *    1 - `_.bind`
       *    2 - `_.bindKey`
       *    4 - `_.curry` or `_.curryRight` of a bound function
       *    8 - `_.curry`
       *   16 - `_.curryRight`
       *   32 - `_.partial`
       *   64 - `_.partialRight`
       *  128 - `_.rearg`
       *  256 - `_.ary`
       *  512 - `_.flip`
       * @param {*} [thisArg] The `this` binding of `func`.
       * @param {Array} [partials] The arguments to be partially applied.
       * @param {Array} [holders] The `partials` placeholder indexes.
       * @param {Array} [argPos] The argument positions of the new function.
       * @param {number} [ary] The arity cap of `func`.
       * @param {number} [arity] The arity of `func`.
       * @returns {Function} Returns the new wrapped function.
       */
      function createWrap(func, bitmask, thisArg, partials, holders, argPos, ary, arity) {
        var isBindKey = bitmask & WRAP_BIND_KEY_FLAG;
        if (!isBindKey && typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        var length = partials ? partials.length : 0;
        if (!length) {
          bitmask &= ~(WRAP_PARTIAL_FLAG | WRAP_PARTIAL_RIGHT_FLAG);
          partials = holders = undefined$1;
        }
        ary = ary === undefined$1 ? ary : nativeMax(toInteger(ary), 0);
        arity = arity === undefined$1 ? arity : toInteger(arity);
        length -= holders ? holders.length : 0;

        if (bitmask & WRAP_PARTIAL_RIGHT_FLAG) {
          var partialsRight = partials,
              holdersRight = holders;

          partials = holders = undefined$1;
        }
        var data = isBindKey ? undefined$1 : getData(func);

        var newData = [
          func, bitmask, thisArg, partials, holders, partialsRight, holdersRight,
          argPos, ary, arity
        ];

        if (data) {
          mergeData(newData, data);
        }
        func = newData[0];
        bitmask = newData[1];
        thisArg = newData[2];
        partials = newData[3];
        holders = newData[4];
        arity = newData[9] = newData[9] === undefined$1
          ? (isBindKey ? 0 : func.length)
          : nativeMax(newData[9] - length, 0);

        if (!arity && bitmask & (WRAP_CURRY_FLAG | WRAP_CURRY_RIGHT_FLAG)) {
          bitmask &= ~(WRAP_CURRY_FLAG | WRAP_CURRY_RIGHT_FLAG);
        }
        if (!bitmask || bitmask == WRAP_BIND_FLAG) {
          var result = createBind(func, bitmask, thisArg);
        } else if (bitmask == WRAP_CURRY_FLAG || bitmask == WRAP_CURRY_RIGHT_FLAG) {
          result = createCurry(func, bitmask, arity);
        } else if ((bitmask == WRAP_PARTIAL_FLAG || bitmask == (WRAP_BIND_FLAG | WRAP_PARTIAL_FLAG)) && !holders.length) {
          result = createPartial(func, bitmask, thisArg, partials);
        } else {
          result = createHybrid.apply(undefined$1, newData);
        }
        var setter = data ? baseSetData : setData;
        return setWrapToString(setter(result, newData), func, bitmask);
      }

      /**
       * Used by `_.defaults` to customize its `_.assignIn` use to assign properties
       * of source objects to the destination object for all destination properties
       * that resolve to `undefined`.
       *
       * @private
       * @param {*} objValue The destination value.
       * @param {*} srcValue The source value.
       * @param {string} key The key of the property to assign.
       * @param {Object} object The parent object of `objValue`.
       * @returns {*} Returns the value to assign.
       */
      function customDefaultsAssignIn(objValue, srcValue, key, object) {
        if (objValue === undefined$1 ||
            (eq(objValue, objectProto[key]) && !hasOwnProperty.call(object, key))) {
          return srcValue;
        }
        return objValue;
      }

      /**
       * Used by `_.defaultsDeep` to customize its `_.merge` use to merge source
       * objects into destination objects that are passed thru.
       *
       * @private
       * @param {*} objValue The destination value.
       * @param {*} srcValue The source value.
       * @param {string} key The key of the property to merge.
       * @param {Object} object The parent object of `objValue`.
       * @param {Object} source The parent object of `srcValue`.
       * @param {Object} [stack] Tracks traversed source values and their merged
       *  counterparts.
       * @returns {*} Returns the value to assign.
       */
      function customDefaultsMerge(objValue, srcValue, key, object, source, stack) {
        if (isObject(objValue) && isObject(srcValue)) {
          // Recursively merge objects and arrays (susceptible to call stack limits).
          stack.set(srcValue, objValue);
          baseMerge(objValue, srcValue, undefined$1, customDefaultsMerge, stack);
          stack['delete'](srcValue);
        }
        return objValue;
      }

      /**
       * Used by `_.omit` to customize its `_.cloneDeep` use to only clone plain
       * objects.
       *
       * @private
       * @param {*} value The value to inspect.
       * @param {string} key The key of the property to inspect.
       * @returns {*} Returns the uncloned value or `undefined` to defer cloning to `_.cloneDeep`.
       */
      function customOmitClone(value) {
        return isPlainObject(value) ? undefined$1 : value;
      }

      /**
       * A specialized version of `baseIsEqualDeep` for arrays with support for
       * partial deep comparisons.
       *
       * @private
       * @param {Array} array The array to compare.
       * @param {Array} other The other array to compare.
       * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
       * @param {Function} customizer The function to customize comparisons.
       * @param {Function} equalFunc The function to determine equivalents of values.
       * @param {Object} stack Tracks traversed `array` and `other` objects.
       * @returns {boolean} Returns `true` if the arrays are equivalent, else `false`.
       */
      function equalArrays(array, other, bitmask, customizer, equalFunc, stack) {
        var isPartial = bitmask & COMPARE_PARTIAL_FLAG,
            arrLength = array.length,
            othLength = other.length;

        if (arrLength != othLength && !(isPartial && othLength > arrLength)) {
          return false;
        }
        // Assume cyclic values are equal.
        var stacked = stack.get(array);
        if (stacked && stack.get(other)) {
          return stacked == other;
        }
        var index = -1,
            result = true,
            seen = (bitmask & COMPARE_UNORDERED_FLAG) ? new SetCache : undefined$1;

        stack.set(array, other);
        stack.set(other, array);

        // Ignore non-index properties.
        while (++index < arrLength) {
          var arrValue = array[index],
              othValue = other[index];

          if (customizer) {
            var compared = isPartial
              ? customizer(othValue, arrValue, index, other, array, stack)
              : customizer(arrValue, othValue, index, array, other, stack);
          }
          if (compared !== undefined$1) {
            if (compared) {
              continue;
            }
            result = false;
            break;
          }
          // Recursively compare arrays (susceptible to call stack limits).
          if (seen) {
            if (!arraySome(other, function(othValue, othIndex) {
                  if (!cacheHas(seen, othIndex) &&
                      (arrValue === othValue || equalFunc(arrValue, othValue, bitmask, customizer, stack))) {
                    return seen.push(othIndex);
                  }
                })) {
              result = false;
              break;
            }
          } else if (!(
                arrValue === othValue ||
                  equalFunc(arrValue, othValue, bitmask, customizer, stack)
              )) {
            result = false;
            break;
          }
        }
        stack['delete'](array);
        stack['delete'](other);
        return result;
      }

      /**
       * A specialized version of `baseIsEqualDeep` for comparing objects of
       * the same `toStringTag`.
       *
       * **Note:** This function only supports comparing values with tags of
       * `Boolean`, `Date`, `Error`, `Number`, `RegExp`, or `String`.
       *
       * @private
       * @param {Object} object The object to compare.
       * @param {Object} other The other object to compare.
       * @param {string} tag The `toStringTag` of the objects to compare.
       * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
       * @param {Function} customizer The function to customize comparisons.
       * @param {Function} equalFunc The function to determine equivalents of values.
       * @param {Object} stack Tracks traversed `object` and `other` objects.
       * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
       */
      function equalByTag(object, other, tag, bitmask, customizer, equalFunc, stack) {
        switch (tag) {
          case dataViewTag:
            if ((object.byteLength != other.byteLength) ||
                (object.byteOffset != other.byteOffset)) {
              return false;
            }
            object = object.buffer;
            other = other.buffer;

          case arrayBufferTag:
            if ((object.byteLength != other.byteLength) ||
                !equalFunc(new Uint8Array(object), new Uint8Array(other))) {
              return false;
            }
            return true;

          case boolTag:
          case dateTag:
          case numberTag:
            // Coerce booleans to `1` or `0` and dates to milliseconds.
            // Invalid dates are coerced to `NaN`.
            return eq(+object, +other);

          case errorTag:
            return object.name == other.name && object.message == other.message;

          case regexpTag:
          case stringTag:
            // Coerce regexes to strings and treat strings, primitives and objects,
            // as equal. See http://www.ecma-international.org/ecma-262/7.0/#sec-regexp.prototype.tostring
            // for more details.
            return object == (other + '');

          case mapTag:
            var convert = mapToArray;

          case setTag:
            var isPartial = bitmask & COMPARE_PARTIAL_FLAG;
            convert || (convert = setToArray);

            if (object.size != other.size && !isPartial) {
              return false;
            }
            // Assume cyclic values are equal.
            var stacked = stack.get(object);
            if (stacked) {
              return stacked == other;
            }
            bitmask |= COMPARE_UNORDERED_FLAG;

            // Recursively compare objects (susceptible to call stack limits).
            stack.set(object, other);
            var result = equalArrays(convert(object), convert(other), bitmask, customizer, equalFunc, stack);
            stack['delete'](object);
            return result;

          case symbolTag:
            if (symbolValueOf) {
              return symbolValueOf.call(object) == symbolValueOf.call(other);
            }
        }
        return false;
      }

      /**
       * A specialized version of `baseIsEqualDeep` for objects with support for
       * partial deep comparisons.
       *
       * @private
       * @param {Object} object The object to compare.
       * @param {Object} other The other object to compare.
       * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
       * @param {Function} customizer The function to customize comparisons.
       * @param {Function} equalFunc The function to determine equivalents of values.
       * @param {Object} stack Tracks traversed `object` and `other` objects.
       * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
       */
      function equalObjects(object, other, bitmask, customizer, equalFunc, stack) {
        var isPartial = bitmask & COMPARE_PARTIAL_FLAG,
            objProps = getAllKeys(object),
            objLength = objProps.length,
            othProps = getAllKeys(other),
            othLength = othProps.length;

        if (objLength != othLength && !isPartial) {
          return false;
        }
        var index = objLength;
        while (index--) {
          var key = objProps[index];
          if (!(isPartial ? key in other : hasOwnProperty.call(other, key))) {
            return false;
          }
        }
        // Assume cyclic values are equal.
        var stacked = stack.get(object);
        if (stacked && stack.get(other)) {
          return stacked == other;
        }
        var result = true;
        stack.set(object, other);
        stack.set(other, object);

        var skipCtor = isPartial;
        while (++index < objLength) {
          key = objProps[index];
          var objValue = object[key],
              othValue = other[key];

          if (customizer) {
            var compared = isPartial
              ? customizer(othValue, objValue, key, other, object, stack)
              : customizer(objValue, othValue, key, object, other, stack);
          }
          // Recursively compare objects (susceptible to call stack limits).
          if (!(compared === undefined$1
                ? (objValue === othValue || equalFunc(objValue, othValue, bitmask, customizer, stack))
                : compared
              )) {
            result = false;
            break;
          }
          skipCtor || (skipCtor = key == 'constructor');
        }
        if (result && !skipCtor) {
          var objCtor = object.constructor,
              othCtor = other.constructor;

          // Non `Object` object instances with different constructors are not equal.
          if (objCtor != othCtor &&
              ('constructor' in object && 'constructor' in other) &&
              !(typeof objCtor == 'function' && objCtor instanceof objCtor &&
                typeof othCtor == 'function' && othCtor instanceof othCtor)) {
            result = false;
          }
        }
        stack['delete'](object);
        stack['delete'](other);
        return result;
      }

      /**
       * A specialized version of `baseRest` which flattens the rest array.
       *
       * @private
       * @param {Function} func The function to apply a rest parameter to.
       * @returns {Function} Returns the new function.
       */
      function flatRest(func) {
        return setToString(overRest(func, undefined$1, flatten), func + '');
      }

      /**
       * Creates an array of own enumerable property names and symbols of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names and symbols.
       */
      function getAllKeys(object) {
        return baseGetAllKeys(object, keys, getSymbols);
      }

      /**
       * Creates an array of own and inherited enumerable property names and
       * symbols of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names and symbols.
       */
      function getAllKeysIn(object) {
        return baseGetAllKeys(object, keysIn, getSymbolsIn);
      }

      /**
       * Gets metadata for `func`.
       *
       * @private
       * @param {Function} func The function to query.
       * @returns {*} Returns the metadata for `func`.
       */
      var getData = !metaMap ? noop : function(func) {
        return metaMap.get(func);
      };

      /**
       * Gets the name of `func`.
       *
       * @private
       * @param {Function} func The function to query.
       * @returns {string} Returns the function name.
       */
      function getFuncName(func) {
        var result = (func.name + ''),
            array = realNames[result],
            length = hasOwnProperty.call(realNames, result) ? array.length : 0;

        while (length--) {
          var data = array[length],
              otherFunc = data.func;
          if (otherFunc == null || otherFunc == func) {
            return data.name;
          }
        }
        return result;
      }

      /**
       * Gets the argument placeholder value for `func`.
       *
       * @private
       * @param {Function} func The function to inspect.
       * @returns {*} Returns the placeholder value.
       */
      function getHolder(func) {
        var object = hasOwnProperty.call(lodash, 'placeholder') ? lodash : func;
        return object.placeholder;
      }

      /**
       * Gets the appropriate "iteratee" function. If `_.iteratee` is customized,
       * this function returns the custom method, otherwise it returns `baseIteratee`.
       * If arguments are provided, the chosen function is invoked with them and
       * its result is returned.
       *
       * @private
       * @param {*} [value] The value to convert to an iteratee.
       * @param {number} [arity] The arity of the created iteratee.
       * @returns {Function} Returns the chosen function or its result.
       */
      function getIteratee() {
        var result = lodash.iteratee || iteratee;
        result = result === iteratee ? baseIteratee : result;
        return arguments.length ? result(arguments[0], arguments[1]) : result;
      }

      /**
       * Gets the data for `map`.
       *
       * @private
       * @param {Object} map The map to query.
       * @param {string} key The reference key.
       * @returns {*} Returns the map data.
       */
      function getMapData(map, key) {
        var data = map.__data__;
        return isKeyable(key)
          ? data[typeof key == 'string' ? 'string' : 'hash']
          : data.map;
      }

      /**
       * Gets the property names, values, and compare flags of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the match data of `object`.
       */
      function getMatchData(object) {
        var result = keys(object),
            length = result.length;

        while (length--) {
          var key = result[length],
              value = object[key];

          result[length] = [key, value, isStrictComparable(value)];
        }
        return result;
      }

      /**
       * Gets the native function at `key` of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @param {string} key The key of the method to get.
       * @returns {*} Returns the function if it's native, else `undefined`.
       */
      function getNative(object, key) {
        var value = getValue(object, key);
        return baseIsNative(value) ? value : undefined$1;
      }

      /**
       * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
       *
       * @private
       * @param {*} value The value to query.
       * @returns {string} Returns the raw `toStringTag`.
       */
      function getRawTag(value) {
        var isOwn = hasOwnProperty.call(value, symToStringTag),
            tag = value[symToStringTag];

        try {
          value[symToStringTag] = undefined$1;
          var unmasked = true;
        } catch (e) {}

        var result = nativeObjectToString.call(value);
        if (unmasked) {
          if (isOwn) {
            value[symToStringTag] = tag;
          } else {
            delete value[symToStringTag];
          }
        }
        return result;
      }

      /**
       * Creates an array of the own enumerable symbols of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of symbols.
       */
      var getSymbols = !nativeGetSymbols ? stubArray : function(object) {
        if (object == null) {
          return [];
        }
        object = Object(object);
        return arrayFilter(nativeGetSymbols(object), function(symbol) {
          return propertyIsEnumerable.call(object, symbol);
        });
      };

      /**
       * Creates an array of the own and inherited enumerable symbols of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of symbols.
       */
      var getSymbolsIn = !nativeGetSymbols ? stubArray : function(object) {
        var result = [];
        while (object) {
          arrayPush(result, getSymbols(object));
          object = getPrototype(object);
        }
        return result;
      };

      /**
       * Gets the `toStringTag` of `value`.
       *
       * @private
       * @param {*} value The value to query.
       * @returns {string} Returns the `toStringTag`.
       */
      var getTag = baseGetTag;

      // Fallback for data views, maps, sets, and weak maps in IE 11 and promises in Node.js < 6.
      if ((DataView && getTag(new DataView(new ArrayBuffer(1))) != dataViewTag) ||
          (Map && getTag(new Map) != mapTag) ||
          (Promise && getTag(Promise.resolve()) != promiseTag) ||
          (Set && getTag(new Set) != setTag) ||
          (WeakMap && getTag(new WeakMap) != weakMapTag)) {
        getTag = function(value) {
          var result = baseGetTag(value),
              Ctor = result == objectTag ? value.constructor : undefined$1,
              ctorString = Ctor ? toSource(Ctor) : '';

          if (ctorString) {
            switch (ctorString) {
              case dataViewCtorString: return dataViewTag;
              case mapCtorString: return mapTag;
              case promiseCtorString: return promiseTag;
              case setCtorString: return setTag;
              case weakMapCtorString: return weakMapTag;
            }
          }
          return result;
        };
      }

      /**
       * Gets the view, applying any `transforms` to the `start` and `end` positions.
       *
       * @private
       * @param {number} start The start of the view.
       * @param {number} end The end of the view.
       * @param {Array} transforms The transformations to apply to the view.
       * @returns {Object} Returns an object containing the `start` and `end`
       *  positions of the view.
       */
      function getView(start, end, transforms) {
        var index = -1,
            length = transforms.length;

        while (++index < length) {
          var data = transforms[index],
              size = data.size;

          switch (data.type) {
            case 'drop':      start += size; break;
            case 'dropRight': end -= size; break;
            case 'take':      end = nativeMin(end, start + size); break;
            case 'takeRight': start = nativeMax(start, end - size); break;
          }
        }
        return { 'start': start, 'end': end };
      }

      /**
       * Extracts wrapper details from the `source` body comment.
       *
       * @private
       * @param {string} source The source to inspect.
       * @returns {Array} Returns the wrapper details.
       */
      function getWrapDetails(source) {
        var match = source.match(reWrapDetails);
        return match ? match[1].split(reSplitDetails) : [];
      }

      /**
       * Checks if `path` exists on `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @param {Array|string} path The path to check.
       * @param {Function} hasFunc The function to check properties.
       * @returns {boolean} Returns `true` if `path` exists, else `false`.
       */
      function hasPath(object, path, hasFunc) {
        path = castPath(path, object);

        var index = -1,
            length = path.length,
            result = false;

        while (++index < length) {
          var key = toKey(path[index]);
          if (!(result = object != null && hasFunc(object, key))) {
            break;
          }
          object = object[key];
        }
        if (result || ++index != length) {
          return result;
        }
        length = object == null ? 0 : object.length;
        return !!length && isLength(length) && isIndex(key, length) &&
          (isArray(object) || isArguments(object));
      }

      /**
       * Initializes an array clone.
       *
       * @private
       * @param {Array} array The array to clone.
       * @returns {Array} Returns the initialized clone.
       */
      function initCloneArray(array) {
        var length = array.length,
            result = new array.constructor(length);

        // Add properties assigned by `RegExp#exec`.
        if (length && typeof array[0] == 'string' && hasOwnProperty.call(array, 'index')) {
          result.index = array.index;
          result.input = array.input;
        }
        return result;
      }

      /**
       * Initializes an object clone.
       *
       * @private
       * @param {Object} object The object to clone.
       * @returns {Object} Returns the initialized clone.
       */
      function initCloneObject(object) {
        return (typeof object.constructor == 'function' && !isPrototype(object))
          ? baseCreate(getPrototype(object))
          : {};
      }

      /**
       * Initializes an object clone based on its `toStringTag`.
       *
       * **Note:** This function only supports cloning values with tags of
       * `Boolean`, `Date`, `Error`, `Map`, `Number`, `RegExp`, `Set`, or `String`.
       *
       * @private
       * @param {Object} object The object to clone.
       * @param {string} tag The `toStringTag` of the object to clone.
       * @param {boolean} [isDeep] Specify a deep clone.
       * @returns {Object} Returns the initialized clone.
       */
      function initCloneByTag(object, tag, isDeep) {
        var Ctor = object.constructor;
        switch (tag) {
          case arrayBufferTag:
            return cloneArrayBuffer(object);

          case boolTag:
          case dateTag:
            return new Ctor(+object);

          case dataViewTag:
            return cloneDataView(object, isDeep);

          case float32Tag: case float64Tag:
          case int8Tag: case int16Tag: case int32Tag:
          case uint8Tag: case uint8ClampedTag: case uint16Tag: case uint32Tag:
            return cloneTypedArray(object, isDeep);

          case mapTag:
            return new Ctor;

          case numberTag:
          case stringTag:
            return new Ctor(object);

          case regexpTag:
            return cloneRegExp(object);

          case setTag:
            return new Ctor;

          case symbolTag:
            return cloneSymbol(object);
        }
      }

      /**
       * Inserts wrapper `details` in a comment at the top of the `source` body.
       *
       * @private
       * @param {string} source The source to modify.
       * @returns {Array} details The details to insert.
       * @returns {string} Returns the modified source.
       */
      function insertWrapDetails(source, details) {
        var length = details.length;
        if (!length) {
          return source;
        }
        var lastIndex = length - 1;
        details[lastIndex] = (length > 1 ? '& ' : '') + details[lastIndex];
        details = details.join(length > 2 ? ', ' : ' ');
        return source.replace(reWrapComment, '{\n/* [wrapped with ' + details + '] */\n');
      }

      /**
       * Checks if `value` is a flattenable `arguments` object or array.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is flattenable, else `false`.
       */
      function isFlattenable(value) {
        return isArray(value) || isArguments(value) ||
          !!(spreadableSymbol && value && value[spreadableSymbol]);
      }

      /**
       * Checks if `value` is a valid array-like index.
       *
       * @private
       * @param {*} value The value to check.
       * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
       * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
       */
      function isIndex(value, length) {
        var type = typeof value;
        length = length == null ? MAX_SAFE_INTEGER : length;

        return !!length &&
          (type == 'number' ||
            (type != 'symbol' && reIsUint.test(value))) &&
              (value > -1 && value % 1 == 0 && value < length);
      }

      /**
       * Checks if the given arguments are from an iteratee call.
       *
       * @private
       * @param {*} value The potential iteratee value argument.
       * @param {*} index The potential iteratee index or key argument.
       * @param {*} object The potential iteratee object argument.
       * @returns {boolean} Returns `true` if the arguments are from an iteratee call,
       *  else `false`.
       */
      function isIterateeCall(value, index, object) {
        if (!isObject(object)) {
          return false;
        }
        var type = typeof index;
        if (type == 'number'
              ? (isArrayLike(object) && isIndex(index, object.length))
              : (type == 'string' && index in object)
            ) {
          return eq(object[index], value);
        }
        return false;
      }

      /**
       * Checks if `value` is a property name and not a property path.
       *
       * @private
       * @param {*} value The value to check.
       * @param {Object} [object] The object to query keys on.
       * @returns {boolean} Returns `true` if `value` is a property name, else `false`.
       */
      function isKey(value, object) {
        if (isArray(value)) {
          return false;
        }
        var type = typeof value;
        if (type == 'number' || type == 'symbol' || type == 'boolean' ||
            value == null || isSymbol(value)) {
          return true;
        }
        return reIsPlainProp.test(value) || !reIsDeepProp.test(value) ||
          (object != null && value in Object(object));
      }

      /**
       * Checks if `value` is suitable for use as unique object key.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is suitable, else `false`.
       */
      function isKeyable(value) {
        var type = typeof value;
        return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
          ? (value !== '__proto__')
          : (value === null);
      }

      /**
       * Checks if `func` has a lazy counterpart.
       *
       * @private
       * @param {Function} func The function to check.
       * @returns {boolean} Returns `true` if `func` has a lazy counterpart,
       *  else `false`.
       */
      function isLaziable(func) {
        var funcName = getFuncName(func),
            other = lodash[funcName];

        if (typeof other != 'function' || !(funcName in LazyWrapper.prototype)) {
          return false;
        }
        if (func === other) {
          return true;
        }
        var data = getData(other);
        return !!data && func === data[0];
      }

      /**
       * Checks if `func` has its source masked.
       *
       * @private
       * @param {Function} func The function to check.
       * @returns {boolean} Returns `true` if `func` is masked, else `false`.
       */
      function isMasked(func) {
        return !!maskSrcKey && (maskSrcKey in func);
      }

      /**
       * Checks if `func` is capable of being masked.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `func` is maskable, else `false`.
       */
      var isMaskable = coreJsData ? isFunction : stubFalse;

      /**
       * Checks if `value` is likely a prototype object.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
       */
      function isPrototype(value) {
        var Ctor = value && value.constructor,
            proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto;

        return value === proto;
      }

      /**
       * Checks if `value` is suitable for strict equality comparisons, i.e. `===`.
       *
       * @private
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` if suitable for strict
       *  equality comparisons, else `false`.
       */
      function isStrictComparable(value) {
        return value === value && !isObject(value);
      }

      /**
       * A specialized version of `matchesProperty` for source values suitable
       * for strict equality comparisons, i.e. `===`.
       *
       * @private
       * @param {string} key The key of the property to get.
       * @param {*} srcValue The value to match.
       * @returns {Function} Returns the new spec function.
       */
      function matchesStrictComparable(key, srcValue) {
        return function(object) {
          if (object == null) {
            return false;
          }
          return object[key] === srcValue &&
            (srcValue !== undefined$1 || (key in Object(object)));
        };
      }

      /**
       * A specialized version of `_.memoize` which clears the memoized function's
       * cache when it exceeds `MAX_MEMOIZE_SIZE`.
       *
       * @private
       * @param {Function} func The function to have its output memoized.
       * @returns {Function} Returns the new memoized function.
       */
      function memoizeCapped(func) {
        var result = memoize(func, function(key) {
          if (cache.size === MAX_MEMOIZE_SIZE) {
            cache.clear();
          }
          return key;
        });

        var cache = result.cache;
        return result;
      }

      /**
       * Merges the function metadata of `source` into `data`.
       *
       * Merging metadata reduces the number of wrappers used to invoke a function.
       * This is possible because methods like `_.bind`, `_.curry`, and `_.partial`
       * may be applied regardless of execution order. Methods like `_.ary` and
       * `_.rearg` modify function arguments, making the order in which they are
       * executed important, preventing the merging of metadata. However, we make
       * an exception for a safe combined case where curried functions have `_.ary`
       * and or `_.rearg` applied.
       *
       * @private
       * @param {Array} data The destination metadata.
       * @param {Array} source The source metadata.
       * @returns {Array} Returns `data`.
       */
      function mergeData(data, source) {
        var bitmask = data[1],
            srcBitmask = source[1],
            newBitmask = bitmask | srcBitmask,
            isCommon = newBitmask < (WRAP_BIND_FLAG | WRAP_BIND_KEY_FLAG | WRAP_ARY_FLAG);

        var isCombo =
          ((srcBitmask == WRAP_ARY_FLAG) && (bitmask == WRAP_CURRY_FLAG)) ||
          ((srcBitmask == WRAP_ARY_FLAG) && (bitmask == WRAP_REARG_FLAG) && (data[7].length <= source[8])) ||
          ((srcBitmask == (WRAP_ARY_FLAG | WRAP_REARG_FLAG)) && (source[7].length <= source[8]) && (bitmask == WRAP_CURRY_FLAG));

        // Exit early if metadata can't be merged.
        if (!(isCommon || isCombo)) {
          return data;
        }
        // Use source `thisArg` if available.
        if (srcBitmask & WRAP_BIND_FLAG) {
          data[2] = source[2];
          // Set when currying a bound function.
          newBitmask |= bitmask & WRAP_BIND_FLAG ? 0 : WRAP_CURRY_BOUND_FLAG;
        }
        // Compose partial arguments.
        var value = source[3];
        if (value) {
          var partials = data[3];
          data[3] = partials ? composeArgs(partials, value, source[4]) : value;
          data[4] = partials ? replaceHolders(data[3], PLACEHOLDER) : source[4];
        }
        // Compose partial right arguments.
        value = source[5];
        if (value) {
          partials = data[5];
          data[5] = partials ? composeArgsRight(partials, value, source[6]) : value;
          data[6] = partials ? replaceHolders(data[5], PLACEHOLDER) : source[6];
        }
        // Use source `argPos` if available.
        value = source[7];
        if (value) {
          data[7] = value;
        }
        // Use source `ary` if it's smaller.
        if (srcBitmask & WRAP_ARY_FLAG) {
          data[8] = data[8] == null ? source[8] : nativeMin(data[8], source[8]);
        }
        // Use source `arity` if one is not provided.
        if (data[9] == null) {
          data[9] = source[9];
        }
        // Use source `func` and merge bitmasks.
        data[0] = source[0];
        data[1] = newBitmask;

        return data;
      }

      /**
       * This function is like
       * [`Object.keys`](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
       * except that it includes inherited enumerable properties.
       *
       * @private
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names.
       */
      function nativeKeysIn(object) {
        var result = [];
        if (object != null) {
          for (var key in Object(object)) {
            result.push(key);
          }
        }
        return result;
      }

      /**
       * Converts `value` to a string using `Object.prototype.toString`.
       *
       * @private
       * @param {*} value The value to convert.
       * @returns {string} Returns the converted string.
       */
      function objectToString(value) {
        return nativeObjectToString.call(value);
      }

      /**
       * A specialized version of `baseRest` which transforms the rest array.
       *
       * @private
       * @param {Function} func The function to apply a rest parameter to.
       * @param {number} [start=func.length-1] The start position of the rest parameter.
       * @param {Function} transform The rest array transform.
       * @returns {Function} Returns the new function.
       */
      function overRest(func, start, transform) {
        start = nativeMax(start === undefined$1 ? (func.length - 1) : start, 0);
        return function() {
          var args = arguments,
              index = -1,
              length = nativeMax(args.length - start, 0),
              array = Array(length);

          while (++index < length) {
            array[index] = args[start + index];
          }
          index = -1;
          var otherArgs = Array(start + 1);
          while (++index < start) {
            otherArgs[index] = args[index];
          }
          otherArgs[start] = transform(array);
          return apply(func, this, otherArgs);
        };
      }

      /**
       * Gets the parent value at `path` of `object`.
       *
       * @private
       * @param {Object} object The object to query.
       * @param {Array} path The path to get the parent value of.
       * @returns {*} Returns the parent value.
       */
      function parent(object, path) {
        return path.length < 2 ? object : baseGet(object, baseSlice(path, 0, -1));
      }

      /**
       * Reorder `array` according to the specified indexes where the element at
       * the first index is assigned as the first element, the element at
       * the second index is assigned as the second element, and so on.
       *
       * @private
       * @param {Array} array The array to reorder.
       * @param {Array} indexes The arranged array indexes.
       * @returns {Array} Returns `array`.
       */
      function reorder(array, indexes) {
        var arrLength = array.length,
            length = nativeMin(indexes.length, arrLength),
            oldArray = copyArray(array);

        while (length--) {
          var index = indexes[length];
          array[length] = isIndex(index, arrLength) ? oldArray[index] : undefined$1;
        }
        return array;
      }

      /**
       * Gets the value at `key`, unless `key` is "__proto__" or "constructor".
       *
       * @private
       * @param {Object} object The object to query.
       * @param {string} key The key of the property to get.
       * @returns {*} Returns the property value.
       */
      function safeGet(object, key) {
        if (key === 'constructor' && typeof object[key] === 'function') {
          return;
        }

        if (key == '__proto__') {
          return;
        }

        return object[key];
      }

      /**
       * Sets metadata for `func`.
       *
       * **Note:** If this function becomes hot, i.e. is invoked a lot in a short
       * period of time, it will trip its breaker and transition to an identity
       * function to avoid garbage collection pauses in V8. See
       * [V8 issue 2070](https://bugs.chromium.org/p/v8/issues/detail?id=2070)
       * for more details.
       *
       * @private
       * @param {Function} func The function to associate metadata with.
       * @param {*} data The metadata.
       * @returns {Function} Returns `func`.
       */
      var setData = shortOut(baseSetData);

      /**
       * A simple wrapper around the global [`setTimeout`](https://mdn.io/setTimeout).
       *
       * @private
       * @param {Function} func The function to delay.
       * @param {number} wait The number of milliseconds to delay invocation.
       * @returns {number|Object} Returns the timer id or timeout object.
       */
      var setTimeout = ctxSetTimeout || function(func, wait) {
        return root.setTimeout(func, wait);
      };

      /**
       * Sets the `toString` method of `func` to return `string`.
       *
       * @private
       * @param {Function} func The function to modify.
       * @param {Function} string The `toString` result.
       * @returns {Function} Returns `func`.
       */
      var setToString = shortOut(baseSetToString);

      /**
       * Sets the `toString` method of `wrapper` to mimic the source of `reference`
       * with wrapper details in a comment at the top of the source body.
       *
       * @private
       * @param {Function} wrapper The function to modify.
       * @param {Function} reference The reference function.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @returns {Function} Returns `wrapper`.
       */
      function setWrapToString(wrapper, reference, bitmask) {
        var source = (reference + '');
        return setToString(wrapper, insertWrapDetails(source, updateWrapDetails(getWrapDetails(source), bitmask)));
      }

      /**
       * Creates a function that'll short out and invoke `identity` instead
       * of `func` when it's called `HOT_COUNT` or more times in `HOT_SPAN`
       * milliseconds.
       *
       * @private
       * @param {Function} func The function to restrict.
       * @returns {Function} Returns the new shortable function.
       */
      function shortOut(func) {
        var count = 0,
            lastCalled = 0;

        return function() {
          var stamp = nativeNow(),
              remaining = HOT_SPAN - (stamp - lastCalled);

          lastCalled = stamp;
          if (remaining > 0) {
            if (++count >= HOT_COUNT) {
              return arguments[0];
            }
          } else {
            count = 0;
          }
          return func.apply(undefined$1, arguments);
        };
      }

      /**
       * A specialized version of `_.shuffle` which mutates and sets the size of `array`.
       *
       * @private
       * @param {Array} array The array to shuffle.
       * @param {number} [size=array.length] The size of `array`.
       * @returns {Array} Returns `array`.
       */
      function shuffleSelf(array, size) {
        var index = -1,
            length = array.length,
            lastIndex = length - 1;

        size = size === undefined$1 ? length : size;
        while (++index < size) {
          var rand = baseRandom(index, lastIndex),
              value = array[rand];

          array[rand] = array[index];
          array[index] = value;
        }
        array.length = size;
        return array;
      }

      /**
       * Converts `string` to a property path array.
       *
       * @private
       * @param {string} string The string to convert.
       * @returns {Array} Returns the property path array.
       */
      var stringToPath = memoizeCapped(function(string) {
        var result = [];
        if (string.charCodeAt(0) === 46 /* . */) {
          result.push('');
        }
        string.replace(rePropName, function(match, number, quote, subString) {
          result.push(quote ? subString.replace(reEscapeChar, '$1') : (number || match));
        });
        return result;
      });

      /**
       * Converts `value` to a string key if it's not a string or symbol.
       *
       * @private
       * @param {*} value The value to inspect.
       * @returns {string|symbol} Returns the key.
       */
      function toKey(value) {
        if (typeof value == 'string' || isSymbol(value)) {
          return value;
        }
        var result = (value + '');
        return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
      }

      /**
       * Converts `func` to its source code.
       *
       * @private
       * @param {Function} func The function to convert.
       * @returns {string} Returns the source code.
       */
      function toSource(func) {
        if (func != null) {
          try {
            return funcToString.call(func);
          } catch (e) {}
          try {
            return (func + '');
          } catch (e) {}
        }
        return '';
      }

      /**
       * Updates wrapper `details` based on `bitmask` flags.
       *
       * @private
       * @returns {Array} details The details to modify.
       * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
       * @returns {Array} Returns `details`.
       */
      function updateWrapDetails(details, bitmask) {
        arrayEach(wrapFlags, function(pair) {
          var value = '_.' + pair[0];
          if ((bitmask & pair[1]) && !arrayIncludes(details, value)) {
            details.push(value);
          }
        });
        return details.sort();
      }

      /**
       * Creates a clone of `wrapper`.
       *
       * @private
       * @param {Object} wrapper The wrapper to clone.
       * @returns {Object} Returns the cloned wrapper.
       */
      function wrapperClone(wrapper) {
        if (wrapper instanceof LazyWrapper) {
          return wrapper.clone();
        }
        var result = new LodashWrapper(wrapper.__wrapped__, wrapper.__chain__);
        result.__actions__ = copyArray(wrapper.__actions__);
        result.__index__  = wrapper.__index__;
        result.__values__ = wrapper.__values__;
        return result;
      }

      /*------------------------------------------------------------------------*/

      /**
       * Creates an array of elements split into groups the length of `size`.
       * If `array` can't be split evenly, the final chunk will be the remaining
       * elements.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to process.
       * @param {number} [size=1] The length of each chunk
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the new array of chunks.
       * @example
       *
       * _.chunk(['a', 'b', 'c', 'd'], 2);
       * // => [['a', 'b'], ['c', 'd']]
       *
       * _.chunk(['a', 'b', 'c', 'd'], 3);
       * // => [['a', 'b', 'c'], ['d']]
       */
      function chunk(array, size, guard) {
        if ((guard ? isIterateeCall(array, size, guard) : size === undefined$1)) {
          size = 1;
        } else {
          size = nativeMax(toInteger(size), 0);
        }
        var length = array == null ? 0 : array.length;
        if (!length || size < 1) {
          return [];
        }
        var index = 0,
            resIndex = 0,
            result = Array(nativeCeil(length / size));

        while (index < length) {
          result[resIndex++] = baseSlice(array, index, (index += size));
        }
        return result;
      }

      /**
       * Creates an array with all falsey values removed. The values `false`, `null`,
       * `0`, `""`, `undefined`, and `NaN` are falsey.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to compact.
       * @returns {Array} Returns the new array of filtered values.
       * @example
       *
       * _.compact([0, 1, false, 2, '', 3]);
       * // => [1, 2, 3]
       */
      function compact(array) {
        var index = -1,
            length = array == null ? 0 : array.length,
            resIndex = 0,
            result = [];

        while (++index < length) {
          var value = array[index];
          if (value) {
            result[resIndex++] = value;
          }
        }
        return result;
      }

      /**
       * Creates a new array concatenating `array` with any additional arrays
       * and/or values.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to concatenate.
       * @param {...*} [values] The values to concatenate.
       * @returns {Array} Returns the new concatenated array.
       * @example
       *
       * var array = [1];
       * var other = _.concat(array, 2, [3], [[4]]);
       *
       * console.log(other);
       * // => [1, 2, 3, [4]]
       *
       * console.log(array);
       * // => [1]
       */
      function concat() {
        var length = arguments.length;
        if (!length) {
          return [];
        }
        var args = Array(length - 1),
            array = arguments[0],
            index = length;

        while (index--) {
          args[index - 1] = arguments[index];
        }
        return arrayPush(isArray(array) ? copyArray(array) : [array], baseFlatten(args, 1));
      }

      /**
       * Creates an array of `array` values not included in the other given arrays
       * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons. The order and references of result values are
       * determined by the first array.
       *
       * **Note:** Unlike `_.pullAll`, this method returns a new array.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {...Array} [values] The values to exclude.
       * @returns {Array} Returns the new array of filtered values.
       * @see _.without, _.xor
       * @example
       *
       * _.difference([2, 1], [2, 3]);
       * // => [1]
       */
      var difference = baseRest(function(array, values) {
        return isArrayLikeObject(array)
          ? baseDifference(array, baseFlatten(values, 1, isArrayLikeObject, true))
          : [];
      });

      /**
       * This method is like `_.difference` except that it accepts `iteratee` which
       * is invoked for each element of `array` and `values` to generate the criterion
       * by which they're compared. The order and references of result values are
       * determined by the first array. The iteratee is invoked with one argument:
       * (value).
       *
       * **Note:** Unlike `_.pullAllBy`, this method returns a new array.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {...Array} [values] The values to exclude.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Array} Returns the new array of filtered values.
       * @example
       *
       * _.differenceBy([2.1, 1.2], [2.3, 3.4], Math.floor);
       * // => [1.2]
       *
       * // The `_.property` iteratee shorthand.
       * _.differenceBy([{ 'x': 2 }, { 'x': 1 }], [{ 'x': 1 }], 'x');
       * // => [{ 'x': 2 }]
       */
      var differenceBy = baseRest(function(array, values) {
        var iteratee = last(values);
        if (isArrayLikeObject(iteratee)) {
          iteratee = undefined$1;
        }
        return isArrayLikeObject(array)
          ? baseDifference(array, baseFlatten(values, 1, isArrayLikeObject, true), getIteratee(iteratee, 2))
          : [];
      });

      /**
       * This method is like `_.difference` except that it accepts `comparator`
       * which is invoked to compare elements of `array` to `values`. The order and
       * references of result values are determined by the first array. The comparator
       * is invoked with two arguments: (arrVal, othVal).
       *
       * **Note:** Unlike `_.pullAllWith`, this method returns a new array.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {...Array} [values] The values to exclude.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of filtered values.
       * @example
       *
       * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
       *
       * _.differenceWith(objects, [{ 'x': 1, 'y': 2 }], _.isEqual);
       * // => [{ 'x': 2, 'y': 1 }]
       */
      var differenceWith = baseRest(function(array, values) {
        var comparator = last(values);
        if (isArrayLikeObject(comparator)) {
          comparator = undefined$1;
        }
        return isArrayLikeObject(array)
          ? baseDifference(array, baseFlatten(values, 1, isArrayLikeObject, true), undefined$1, comparator)
          : [];
      });

      /**
       * Creates a slice of `array` with `n` elements dropped from the beginning.
       *
       * @static
       * @memberOf _
       * @since 0.5.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {number} [n=1] The number of elements to drop.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * _.drop([1, 2, 3]);
       * // => [2, 3]
       *
       * _.drop([1, 2, 3], 2);
       * // => [3]
       *
       * _.drop([1, 2, 3], 5);
       * // => []
       *
       * _.drop([1, 2, 3], 0);
       * // => [1, 2, 3]
       */
      function drop(array, n, guard) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return [];
        }
        n = (guard || n === undefined$1) ? 1 : toInteger(n);
        return baseSlice(array, n < 0 ? 0 : n, length);
      }

      /**
       * Creates a slice of `array` with `n` elements dropped from the end.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {number} [n=1] The number of elements to drop.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * _.dropRight([1, 2, 3]);
       * // => [1, 2]
       *
       * _.dropRight([1, 2, 3], 2);
       * // => [1]
       *
       * _.dropRight([1, 2, 3], 5);
       * // => []
       *
       * _.dropRight([1, 2, 3], 0);
       * // => [1, 2, 3]
       */
      function dropRight(array, n, guard) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return [];
        }
        n = (guard || n === undefined$1) ? 1 : toInteger(n);
        n = length - n;
        return baseSlice(array, 0, n < 0 ? 0 : n);
      }

      /**
       * Creates a slice of `array` excluding elements dropped from the end.
       * Elements are dropped until `predicate` returns falsey. The predicate is
       * invoked with three arguments: (value, index, array).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'active': true },
       *   { 'user': 'fred',    'active': false },
       *   { 'user': 'pebbles', 'active': false }
       * ];
       *
       * _.dropRightWhile(users, function(o) { return !o.active; });
       * // => objects for ['barney']
       *
       * // The `_.matches` iteratee shorthand.
       * _.dropRightWhile(users, { 'user': 'pebbles', 'active': false });
       * // => objects for ['barney', 'fred']
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.dropRightWhile(users, ['active', false]);
       * // => objects for ['barney']
       *
       * // The `_.property` iteratee shorthand.
       * _.dropRightWhile(users, 'active');
       * // => objects for ['barney', 'fred', 'pebbles']
       */
      function dropRightWhile(array, predicate) {
        return (array && array.length)
          ? baseWhile(array, getIteratee(predicate, 3), true, true)
          : [];
      }

      /**
       * Creates a slice of `array` excluding elements dropped from the beginning.
       * Elements are dropped until `predicate` returns falsey. The predicate is
       * invoked with three arguments: (value, index, array).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'active': false },
       *   { 'user': 'fred',    'active': false },
       *   { 'user': 'pebbles', 'active': true }
       * ];
       *
       * _.dropWhile(users, function(o) { return !o.active; });
       * // => objects for ['pebbles']
       *
       * // The `_.matches` iteratee shorthand.
       * _.dropWhile(users, { 'user': 'barney', 'active': false });
       * // => objects for ['fred', 'pebbles']
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.dropWhile(users, ['active', false]);
       * // => objects for ['pebbles']
       *
       * // The `_.property` iteratee shorthand.
       * _.dropWhile(users, 'active');
       * // => objects for ['barney', 'fred', 'pebbles']
       */
      function dropWhile(array, predicate) {
        return (array && array.length)
          ? baseWhile(array, getIteratee(predicate, 3), true)
          : [];
      }

      /**
       * Fills elements of `array` with `value` from `start` up to, but not
       * including, `end`.
       *
       * **Note:** This method mutates `array`.
       *
       * @static
       * @memberOf _
       * @since 3.2.0
       * @category Array
       * @param {Array} array The array to fill.
       * @param {*} value The value to fill `array` with.
       * @param {number} [start=0] The start position.
       * @param {number} [end=array.length] The end position.
       * @returns {Array} Returns `array`.
       * @example
       *
       * var array = [1, 2, 3];
       *
       * _.fill(array, 'a');
       * console.log(array);
       * // => ['a', 'a', 'a']
       *
       * _.fill(Array(3), 2);
       * // => [2, 2, 2]
       *
       * _.fill([4, 6, 8, 10], '*', 1, 3);
       * // => [4, '*', '*', 10]
       */
      function fill(array, value, start, end) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return [];
        }
        if (start && typeof start != 'number' && isIterateeCall(array, value, start)) {
          start = 0;
          end = length;
        }
        return baseFill(array, value, start, end);
      }

      /**
       * This method is like `_.find` except that it returns the index of the first
       * element `predicate` returns truthy for instead of the element itself.
       *
       * @static
       * @memberOf _
       * @since 1.1.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @param {number} [fromIndex=0] The index to search from.
       * @returns {number} Returns the index of the found element, else `-1`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'active': false },
       *   { 'user': 'fred',    'active': false },
       *   { 'user': 'pebbles', 'active': true }
       * ];
       *
       * _.findIndex(users, function(o) { return o.user == 'barney'; });
       * // => 0
       *
       * // The `_.matches` iteratee shorthand.
       * _.findIndex(users, { 'user': 'fred', 'active': false });
       * // => 1
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.findIndex(users, ['active', false]);
       * // => 0
       *
       * // The `_.property` iteratee shorthand.
       * _.findIndex(users, 'active');
       * // => 2
       */
      function findIndex(array, predicate, fromIndex) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return -1;
        }
        var index = fromIndex == null ? 0 : toInteger(fromIndex);
        if (index < 0) {
          index = nativeMax(length + index, 0);
        }
        return baseFindIndex(array, getIteratee(predicate, 3), index);
      }

      /**
       * This method is like `_.findIndex` except that it iterates over elements
       * of `collection` from right to left.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @param {number} [fromIndex=array.length-1] The index to search from.
       * @returns {number} Returns the index of the found element, else `-1`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'active': true },
       *   { 'user': 'fred',    'active': false },
       *   { 'user': 'pebbles', 'active': false }
       * ];
       *
       * _.findLastIndex(users, function(o) { return o.user == 'pebbles'; });
       * // => 2
       *
       * // The `_.matches` iteratee shorthand.
       * _.findLastIndex(users, { 'user': 'barney', 'active': true });
       * // => 0
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.findLastIndex(users, ['active', false]);
       * // => 2
       *
       * // The `_.property` iteratee shorthand.
       * _.findLastIndex(users, 'active');
       * // => 0
       */
      function findLastIndex(array, predicate, fromIndex) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return -1;
        }
        var index = length - 1;
        if (fromIndex !== undefined$1) {
          index = toInteger(fromIndex);
          index = fromIndex < 0
            ? nativeMax(length + index, 0)
            : nativeMin(index, length - 1);
        }
        return baseFindIndex(array, getIteratee(predicate, 3), index, true);
      }

      /**
       * Flattens `array` a single level deep.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to flatten.
       * @returns {Array} Returns the new flattened array.
       * @example
       *
       * _.flatten([1, [2, [3, [4]], 5]]);
       * // => [1, 2, [3, [4]], 5]
       */
      function flatten(array) {
        var length = array == null ? 0 : array.length;
        return length ? baseFlatten(array, 1) : [];
      }

      /**
       * Recursively flattens `array`.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to flatten.
       * @returns {Array} Returns the new flattened array.
       * @example
       *
       * _.flattenDeep([1, [2, [3, [4]], 5]]);
       * // => [1, 2, 3, 4, 5]
       */
      function flattenDeep(array) {
        var length = array == null ? 0 : array.length;
        return length ? baseFlatten(array, INFINITY) : [];
      }

      /**
       * Recursively flatten `array` up to `depth` times.
       *
       * @static
       * @memberOf _
       * @since 4.4.0
       * @category Array
       * @param {Array} array The array to flatten.
       * @param {number} [depth=1] The maximum recursion depth.
       * @returns {Array} Returns the new flattened array.
       * @example
       *
       * var array = [1, [2, [3, [4]], 5]];
       *
       * _.flattenDepth(array, 1);
       * // => [1, 2, [3, [4]], 5]
       *
       * _.flattenDepth(array, 2);
       * // => [1, 2, 3, [4], 5]
       */
      function flattenDepth(array, depth) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return [];
        }
        depth = depth === undefined$1 ? 1 : toInteger(depth);
        return baseFlatten(array, depth);
      }

      /**
       * The inverse of `_.toPairs`; this method returns an object composed
       * from key-value `pairs`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} pairs The key-value pairs.
       * @returns {Object} Returns the new object.
       * @example
       *
       * _.fromPairs([['a', 1], ['b', 2]]);
       * // => { 'a': 1, 'b': 2 }
       */
      function fromPairs(pairs) {
        var index = -1,
            length = pairs == null ? 0 : pairs.length,
            result = {};

        while (++index < length) {
          var pair = pairs[index];
          result[pair[0]] = pair[1];
        }
        return result;
      }

      /**
       * Gets the first element of `array`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @alias first
       * @category Array
       * @param {Array} array The array to query.
       * @returns {*} Returns the first element of `array`.
       * @example
       *
       * _.head([1, 2, 3]);
       * // => 1
       *
       * _.head([]);
       * // => undefined
       */
      function head(array) {
        return (array && array.length) ? array[0] : undefined$1;
      }

      /**
       * Gets the index at which the first occurrence of `value` is found in `array`
       * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons. If `fromIndex` is negative, it's used as the
       * offset from the end of `array`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {*} value The value to search for.
       * @param {number} [fromIndex=0] The index to search from.
       * @returns {number} Returns the index of the matched value, else `-1`.
       * @example
       *
       * _.indexOf([1, 2, 1, 2], 2);
       * // => 1
       *
       * // Search from the `fromIndex`.
       * _.indexOf([1, 2, 1, 2], 2, 2);
       * // => 3
       */
      function indexOf(array, value, fromIndex) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return -1;
        }
        var index = fromIndex == null ? 0 : toInteger(fromIndex);
        if (index < 0) {
          index = nativeMax(length + index, 0);
        }
        return baseIndexOf(array, value, index);
      }

      /**
       * Gets all but the last element of `array`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to query.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * _.initial([1, 2, 3]);
       * // => [1, 2]
       */
      function initial(array) {
        var length = array == null ? 0 : array.length;
        return length ? baseSlice(array, 0, -1) : [];
      }

      /**
       * Creates an array of unique values that are included in all given arrays
       * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons. The order and references of result values are
       * determined by the first array.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @returns {Array} Returns the new array of intersecting values.
       * @example
       *
       * _.intersection([2, 1], [2, 3]);
       * // => [2]
       */
      var intersection = baseRest(function(arrays) {
        var mapped = arrayMap(arrays, castArrayLikeObject);
        return (mapped.length && mapped[0] === arrays[0])
          ? baseIntersection(mapped)
          : [];
      });

      /**
       * This method is like `_.intersection` except that it accepts `iteratee`
       * which is invoked for each element of each `arrays` to generate the criterion
       * by which they're compared. The order and references of result values are
       * determined by the first array. The iteratee is invoked with one argument:
       * (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Array} Returns the new array of intersecting values.
       * @example
       *
       * _.intersectionBy([2.1, 1.2], [2.3, 3.4], Math.floor);
       * // => [2.1]
       *
       * // The `_.property` iteratee shorthand.
       * _.intersectionBy([{ 'x': 1 }], [{ 'x': 2 }, { 'x': 1 }], 'x');
       * // => [{ 'x': 1 }]
       */
      var intersectionBy = baseRest(function(arrays) {
        var iteratee = last(arrays),
            mapped = arrayMap(arrays, castArrayLikeObject);

        if (iteratee === last(mapped)) {
          iteratee = undefined$1;
        } else {
          mapped.pop();
        }
        return (mapped.length && mapped[0] === arrays[0])
          ? baseIntersection(mapped, getIteratee(iteratee, 2))
          : [];
      });

      /**
       * This method is like `_.intersection` except that it accepts `comparator`
       * which is invoked to compare elements of `arrays`. The order and references
       * of result values are determined by the first array. The comparator is
       * invoked with two arguments: (arrVal, othVal).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of intersecting values.
       * @example
       *
       * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
       * var others = [{ 'x': 1, 'y': 1 }, { 'x': 1, 'y': 2 }];
       *
       * _.intersectionWith(objects, others, _.isEqual);
       * // => [{ 'x': 1, 'y': 2 }]
       */
      var intersectionWith = baseRest(function(arrays) {
        var comparator = last(arrays),
            mapped = arrayMap(arrays, castArrayLikeObject);

        comparator = typeof comparator == 'function' ? comparator : undefined$1;
        if (comparator) {
          mapped.pop();
        }
        return (mapped.length && mapped[0] === arrays[0])
          ? baseIntersection(mapped, undefined$1, comparator)
          : [];
      });

      /**
       * Converts all elements in `array` into a string separated by `separator`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to convert.
       * @param {string} [separator=','] The element separator.
       * @returns {string} Returns the joined string.
       * @example
       *
       * _.join(['a', 'b', 'c'], '~');
       * // => 'a~b~c'
       */
      function join(array, separator) {
        return array == null ? '' : nativeJoin.call(array, separator);
      }

      /**
       * Gets the last element of `array`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to query.
       * @returns {*} Returns the last element of `array`.
       * @example
       *
       * _.last([1, 2, 3]);
       * // => 3
       */
      function last(array) {
        var length = array == null ? 0 : array.length;
        return length ? array[length - 1] : undefined$1;
      }

      /**
       * This method is like `_.indexOf` except that it iterates over elements of
       * `array` from right to left.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {*} value The value to search for.
       * @param {number} [fromIndex=array.length-1] The index to search from.
       * @returns {number} Returns the index of the matched value, else `-1`.
       * @example
       *
       * _.lastIndexOf([1, 2, 1, 2], 2);
       * // => 3
       *
       * // Search from the `fromIndex`.
       * _.lastIndexOf([1, 2, 1, 2], 2, 2);
       * // => 1
       */
      function lastIndexOf(array, value, fromIndex) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return -1;
        }
        var index = length;
        if (fromIndex !== undefined$1) {
          index = toInteger(fromIndex);
          index = index < 0 ? nativeMax(length + index, 0) : nativeMin(index, length - 1);
        }
        return value === value
          ? strictLastIndexOf(array, value, index)
          : baseFindIndex(array, baseIsNaN, index, true);
      }

      /**
       * Gets the element at index `n` of `array`. If `n` is negative, the nth
       * element from the end is returned.
       *
       * @static
       * @memberOf _
       * @since 4.11.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {number} [n=0] The index of the element to return.
       * @returns {*} Returns the nth element of `array`.
       * @example
       *
       * var array = ['a', 'b', 'c', 'd'];
       *
       * _.nth(array, 1);
       * // => 'b'
       *
       * _.nth(array, -2);
       * // => 'c';
       */
      function nth(array, n) {
        return (array && array.length) ? baseNth(array, toInteger(n)) : undefined$1;
      }

      /**
       * Removes all given values from `array` using
       * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons.
       *
       * **Note:** Unlike `_.without`, this method mutates `array`. Use `_.remove`
       * to remove elements from an array by predicate.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Array
       * @param {Array} array The array to modify.
       * @param {...*} [values] The values to remove.
       * @returns {Array} Returns `array`.
       * @example
       *
       * var array = ['a', 'b', 'c', 'a', 'b', 'c'];
       *
       * _.pull(array, 'a', 'c');
       * console.log(array);
       * // => ['b', 'b']
       */
      var pull = baseRest(pullAll);

      /**
       * This method is like `_.pull` except that it accepts an array of values to remove.
       *
       * **Note:** Unlike `_.difference`, this method mutates `array`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to modify.
       * @param {Array} values The values to remove.
       * @returns {Array} Returns `array`.
       * @example
       *
       * var array = ['a', 'b', 'c', 'a', 'b', 'c'];
       *
       * _.pullAll(array, ['a', 'c']);
       * console.log(array);
       * // => ['b', 'b']
       */
      function pullAll(array, values) {
        return (array && array.length && values && values.length)
          ? basePullAll(array, values)
          : array;
      }

      /**
       * This method is like `_.pullAll` except that it accepts `iteratee` which is
       * invoked for each element of `array` and `values` to generate the criterion
       * by which they're compared. The iteratee is invoked with one argument: (value).
       *
       * **Note:** Unlike `_.differenceBy`, this method mutates `array`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to modify.
       * @param {Array} values The values to remove.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Array} Returns `array`.
       * @example
       *
       * var array = [{ 'x': 1 }, { 'x': 2 }, { 'x': 3 }, { 'x': 1 }];
       *
       * _.pullAllBy(array, [{ 'x': 1 }, { 'x': 3 }], 'x');
       * console.log(array);
       * // => [{ 'x': 2 }]
       */
      function pullAllBy(array, values, iteratee) {
        return (array && array.length && values && values.length)
          ? basePullAll(array, values, getIteratee(iteratee, 2))
          : array;
      }

      /**
       * This method is like `_.pullAll` except that it accepts `comparator` which
       * is invoked to compare elements of `array` to `values`. The comparator is
       * invoked with two arguments: (arrVal, othVal).
       *
       * **Note:** Unlike `_.differenceWith`, this method mutates `array`.
       *
       * @static
       * @memberOf _
       * @since 4.6.0
       * @category Array
       * @param {Array} array The array to modify.
       * @param {Array} values The values to remove.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns `array`.
       * @example
       *
       * var array = [{ 'x': 1, 'y': 2 }, { 'x': 3, 'y': 4 }, { 'x': 5, 'y': 6 }];
       *
       * _.pullAllWith(array, [{ 'x': 3, 'y': 4 }], _.isEqual);
       * console.log(array);
       * // => [{ 'x': 1, 'y': 2 }, { 'x': 5, 'y': 6 }]
       */
      function pullAllWith(array, values, comparator) {
        return (array && array.length && values && values.length)
          ? basePullAll(array, values, undefined$1, comparator)
          : array;
      }

      /**
       * Removes elements from `array` corresponding to `indexes` and returns an
       * array of removed elements.
       *
       * **Note:** Unlike `_.at`, this method mutates `array`.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to modify.
       * @param {...(number|number[])} [indexes] The indexes of elements to remove.
       * @returns {Array} Returns the new array of removed elements.
       * @example
       *
       * var array = ['a', 'b', 'c', 'd'];
       * var pulled = _.pullAt(array, [1, 3]);
       *
       * console.log(array);
       * // => ['a', 'c']
       *
       * console.log(pulled);
       * // => ['b', 'd']
       */
      var pullAt = flatRest(function(array, indexes) {
        var length = array == null ? 0 : array.length,
            result = baseAt(array, indexes);

        basePullAt(array, arrayMap(indexes, function(index) {
          return isIndex(index, length) ? +index : index;
        }).sort(compareAscending));

        return result;
      });

      /**
       * Removes all elements from `array` that `predicate` returns truthy for
       * and returns an array of the removed elements. The predicate is invoked
       * with three arguments: (value, index, array).
       *
       * **Note:** Unlike `_.filter`, this method mutates `array`. Use `_.pull`
       * to pull elements from an array by value.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Array
       * @param {Array} array The array to modify.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the new array of removed elements.
       * @example
       *
       * var array = [1, 2, 3, 4];
       * var evens = _.remove(array, function(n) {
       *   return n % 2 == 0;
       * });
       *
       * console.log(array);
       * // => [1, 3]
       *
       * console.log(evens);
       * // => [2, 4]
       */
      function remove(array, predicate) {
        var result = [];
        if (!(array && array.length)) {
          return result;
        }
        var index = -1,
            indexes = [],
            length = array.length;

        predicate = getIteratee(predicate, 3);
        while (++index < length) {
          var value = array[index];
          if (predicate(value, index, array)) {
            result.push(value);
            indexes.push(index);
          }
        }
        basePullAt(array, indexes);
        return result;
      }

      /**
       * Reverses `array` so that the first element becomes the last, the second
       * element becomes the second to last, and so on.
       *
       * **Note:** This method mutates `array` and is based on
       * [`Array#reverse`](https://mdn.io/Array/reverse).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to modify.
       * @returns {Array} Returns `array`.
       * @example
       *
       * var array = [1, 2, 3];
       *
       * _.reverse(array);
       * // => [3, 2, 1]
       *
       * console.log(array);
       * // => [3, 2, 1]
       */
      function reverse(array) {
        return array == null ? array : nativeReverse.call(array);
      }

      /**
       * Creates a slice of `array` from `start` up to, but not including, `end`.
       *
       * **Note:** This method is used instead of
       * [`Array#slice`](https://mdn.io/Array/slice) to ensure dense arrays are
       * returned.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to slice.
       * @param {number} [start=0] The start position.
       * @param {number} [end=array.length] The end position.
       * @returns {Array} Returns the slice of `array`.
       */
      function slice(array, start, end) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return [];
        }
        if (end && typeof end != 'number' && isIterateeCall(array, start, end)) {
          start = 0;
          end = length;
        }
        else {
          start = start == null ? 0 : toInteger(start);
          end = end === undefined$1 ? length : toInteger(end);
        }
        return baseSlice(array, start, end);
      }

      /**
       * Uses a binary search to determine the lowest index at which `value`
       * should be inserted into `array` in order to maintain its sort order.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The sorted array to inspect.
       * @param {*} value The value to evaluate.
       * @returns {number} Returns the index at which `value` should be inserted
       *  into `array`.
       * @example
       *
       * _.sortedIndex([30, 50], 40);
       * // => 1
       */
      function sortedIndex(array, value) {
        return baseSortedIndex(array, value);
      }

      /**
       * This method is like `_.sortedIndex` except that it accepts `iteratee`
       * which is invoked for `value` and each element of `array` to compute their
       * sort ranking. The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The sorted array to inspect.
       * @param {*} value The value to evaluate.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {number} Returns the index at which `value` should be inserted
       *  into `array`.
       * @example
       *
       * var objects = [{ 'x': 4 }, { 'x': 5 }];
       *
       * _.sortedIndexBy(objects, { 'x': 4 }, function(o) { return o.x; });
       * // => 0
       *
       * // The `_.property` iteratee shorthand.
       * _.sortedIndexBy(objects, { 'x': 4 }, 'x');
       * // => 0
       */
      function sortedIndexBy(array, value, iteratee) {
        return baseSortedIndexBy(array, value, getIteratee(iteratee, 2));
      }

      /**
       * This method is like `_.indexOf` except that it performs a binary
       * search on a sorted `array`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {*} value The value to search for.
       * @returns {number} Returns the index of the matched value, else `-1`.
       * @example
       *
       * _.sortedIndexOf([4, 5, 5, 5, 6], 5);
       * // => 1
       */
      function sortedIndexOf(array, value) {
        var length = array == null ? 0 : array.length;
        if (length) {
          var index = baseSortedIndex(array, value);
          if (index < length && eq(array[index], value)) {
            return index;
          }
        }
        return -1;
      }

      /**
       * This method is like `_.sortedIndex` except that it returns the highest
       * index at which `value` should be inserted into `array` in order to
       * maintain its sort order.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The sorted array to inspect.
       * @param {*} value The value to evaluate.
       * @returns {number} Returns the index at which `value` should be inserted
       *  into `array`.
       * @example
       *
       * _.sortedLastIndex([4, 5, 5, 5, 6], 5);
       * // => 4
       */
      function sortedLastIndex(array, value) {
        return baseSortedIndex(array, value, true);
      }

      /**
       * This method is like `_.sortedLastIndex` except that it accepts `iteratee`
       * which is invoked for `value` and each element of `array` to compute their
       * sort ranking. The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The sorted array to inspect.
       * @param {*} value The value to evaluate.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {number} Returns the index at which `value` should be inserted
       *  into `array`.
       * @example
       *
       * var objects = [{ 'x': 4 }, { 'x': 5 }];
       *
       * _.sortedLastIndexBy(objects, { 'x': 4 }, function(o) { return o.x; });
       * // => 1
       *
       * // The `_.property` iteratee shorthand.
       * _.sortedLastIndexBy(objects, { 'x': 4 }, 'x');
       * // => 1
       */
      function sortedLastIndexBy(array, value, iteratee) {
        return baseSortedIndexBy(array, value, getIteratee(iteratee, 2), true);
      }

      /**
       * This method is like `_.lastIndexOf` except that it performs a binary
       * search on a sorted `array`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {*} value The value to search for.
       * @returns {number} Returns the index of the matched value, else `-1`.
       * @example
       *
       * _.sortedLastIndexOf([4, 5, 5, 5, 6], 5);
       * // => 3
       */
      function sortedLastIndexOf(array, value) {
        var length = array == null ? 0 : array.length;
        if (length) {
          var index = baseSortedIndex(array, value, true) - 1;
          if (eq(array[index], value)) {
            return index;
          }
        }
        return -1;
      }

      /**
       * This method is like `_.uniq` except that it's designed and optimized
       * for sorted arrays.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @returns {Array} Returns the new duplicate free array.
       * @example
       *
       * _.sortedUniq([1, 1, 2]);
       * // => [1, 2]
       */
      function sortedUniq(array) {
        return (array && array.length)
          ? baseSortedUniq(array)
          : [];
      }

      /**
       * This method is like `_.uniqBy` except that it's designed and optimized
       * for sorted arrays.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {Function} [iteratee] The iteratee invoked per element.
       * @returns {Array} Returns the new duplicate free array.
       * @example
       *
       * _.sortedUniqBy([1.1, 1.2, 2.3, 2.4], Math.floor);
       * // => [1.1, 2.3]
       */
      function sortedUniqBy(array, iteratee) {
        return (array && array.length)
          ? baseSortedUniq(array, getIteratee(iteratee, 2))
          : [];
      }

      /**
       * Gets all but the first element of `array`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * _.tail([1, 2, 3]);
       * // => [2, 3]
       */
      function tail(array) {
        var length = array == null ? 0 : array.length;
        return length ? baseSlice(array, 1, length) : [];
      }

      /**
       * Creates a slice of `array` with `n` elements taken from the beginning.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {number} [n=1] The number of elements to take.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * _.take([1, 2, 3]);
       * // => [1]
       *
       * _.take([1, 2, 3], 2);
       * // => [1, 2]
       *
       * _.take([1, 2, 3], 5);
       * // => [1, 2, 3]
       *
       * _.take([1, 2, 3], 0);
       * // => []
       */
      function take(array, n, guard) {
        if (!(array && array.length)) {
          return [];
        }
        n = (guard || n === undefined$1) ? 1 : toInteger(n);
        return baseSlice(array, 0, n < 0 ? 0 : n);
      }

      /**
       * Creates a slice of `array` with `n` elements taken from the end.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {number} [n=1] The number of elements to take.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * _.takeRight([1, 2, 3]);
       * // => [3]
       *
       * _.takeRight([1, 2, 3], 2);
       * // => [2, 3]
       *
       * _.takeRight([1, 2, 3], 5);
       * // => [1, 2, 3]
       *
       * _.takeRight([1, 2, 3], 0);
       * // => []
       */
      function takeRight(array, n, guard) {
        var length = array == null ? 0 : array.length;
        if (!length) {
          return [];
        }
        n = (guard || n === undefined$1) ? 1 : toInteger(n);
        n = length - n;
        return baseSlice(array, n < 0 ? 0 : n, length);
      }

      /**
       * Creates a slice of `array` with elements taken from the end. Elements are
       * taken until `predicate` returns falsey. The predicate is invoked with
       * three arguments: (value, index, array).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'active': true },
       *   { 'user': 'fred',    'active': false },
       *   { 'user': 'pebbles', 'active': false }
       * ];
       *
       * _.takeRightWhile(users, function(o) { return !o.active; });
       * // => objects for ['fred', 'pebbles']
       *
       * // The `_.matches` iteratee shorthand.
       * _.takeRightWhile(users, { 'user': 'pebbles', 'active': false });
       * // => objects for ['pebbles']
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.takeRightWhile(users, ['active', false]);
       * // => objects for ['fred', 'pebbles']
       *
       * // The `_.property` iteratee shorthand.
       * _.takeRightWhile(users, 'active');
       * // => []
       */
      function takeRightWhile(array, predicate) {
        return (array && array.length)
          ? baseWhile(array, getIteratee(predicate, 3), false, true)
          : [];
      }

      /**
       * Creates a slice of `array` with elements taken from the beginning. Elements
       * are taken until `predicate` returns falsey. The predicate is invoked with
       * three arguments: (value, index, array).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Array
       * @param {Array} array The array to query.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the slice of `array`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'active': false },
       *   { 'user': 'fred',    'active': false },
       *   { 'user': 'pebbles', 'active': true }
       * ];
       *
       * _.takeWhile(users, function(o) { return !o.active; });
       * // => objects for ['barney', 'fred']
       *
       * // The `_.matches` iteratee shorthand.
       * _.takeWhile(users, { 'user': 'barney', 'active': false });
       * // => objects for ['barney']
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.takeWhile(users, ['active', false]);
       * // => objects for ['barney', 'fred']
       *
       * // The `_.property` iteratee shorthand.
       * _.takeWhile(users, 'active');
       * // => []
       */
      function takeWhile(array, predicate) {
        return (array && array.length)
          ? baseWhile(array, getIteratee(predicate, 3))
          : [];
      }

      /**
       * Creates an array of unique values, in order, from all given arrays using
       * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @returns {Array} Returns the new array of combined values.
       * @example
       *
       * _.union([2], [1, 2]);
       * // => [2, 1]
       */
      var union = baseRest(function(arrays) {
        return baseUniq(baseFlatten(arrays, 1, isArrayLikeObject, true));
      });

      /**
       * This method is like `_.union` except that it accepts `iteratee` which is
       * invoked for each element of each `arrays` to generate the criterion by
       * which uniqueness is computed. Result values are chosen from the first
       * array in which the value occurs. The iteratee is invoked with one argument:
       * (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Array} Returns the new array of combined values.
       * @example
       *
       * _.unionBy([2.1], [1.2, 2.3], Math.floor);
       * // => [2.1, 1.2]
       *
       * // The `_.property` iteratee shorthand.
       * _.unionBy([{ 'x': 1 }], [{ 'x': 2 }, { 'x': 1 }], 'x');
       * // => [{ 'x': 1 }, { 'x': 2 }]
       */
      var unionBy = baseRest(function(arrays) {
        var iteratee = last(arrays);
        if (isArrayLikeObject(iteratee)) {
          iteratee = undefined$1;
        }
        return baseUniq(baseFlatten(arrays, 1, isArrayLikeObject, true), getIteratee(iteratee, 2));
      });

      /**
       * This method is like `_.union` except that it accepts `comparator` which
       * is invoked to compare elements of `arrays`. Result values are chosen from
       * the first array in which the value occurs. The comparator is invoked
       * with two arguments: (arrVal, othVal).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of combined values.
       * @example
       *
       * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
       * var others = [{ 'x': 1, 'y': 1 }, { 'x': 1, 'y': 2 }];
       *
       * _.unionWith(objects, others, _.isEqual);
       * // => [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }, { 'x': 1, 'y': 1 }]
       */
      var unionWith = baseRest(function(arrays) {
        var comparator = last(arrays);
        comparator = typeof comparator == 'function' ? comparator : undefined$1;
        return baseUniq(baseFlatten(arrays, 1, isArrayLikeObject, true), undefined$1, comparator);
      });

      /**
       * Creates a duplicate-free version of an array, using
       * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons, in which only the first occurrence of each element
       * is kept. The order of result values is determined by the order they occur
       * in the array.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @returns {Array} Returns the new duplicate free array.
       * @example
       *
       * _.uniq([2, 1, 2]);
       * // => [2, 1]
       */
      function uniq(array) {
        return (array && array.length) ? baseUniq(array) : [];
      }

      /**
       * This method is like `_.uniq` except that it accepts `iteratee` which is
       * invoked for each element in `array` to generate the criterion by which
       * uniqueness is computed. The order of result values is determined by the
       * order they occur in the array. The iteratee is invoked with one argument:
       * (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Array} Returns the new duplicate free array.
       * @example
       *
       * _.uniqBy([2.1, 1.2, 2.3], Math.floor);
       * // => [2.1, 1.2]
       *
       * // The `_.property` iteratee shorthand.
       * _.uniqBy([{ 'x': 1 }, { 'x': 2 }, { 'x': 1 }], 'x');
       * // => [{ 'x': 1 }, { 'x': 2 }]
       */
      function uniqBy(array, iteratee) {
        return (array && array.length) ? baseUniq(array, getIteratee(iteratee, 2)) : [];
      }

      /**
       * This method is like `_.uniq` except that it accepts `comparator` which
       * is invoked to compare elements of `array`. The order of result values is
       * determined by the order they occur in the array.The comparator is invoked
       * with two arguments: (arrVal, othVal).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new duplicate free array.
       * @example
       *
       * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }, { 'x': 1, 'y': 2 }];
       *
       * _.uniqWith(objects, _.isEqual);
       * // => [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }]
       */
      function uniqWith(array, comparator) {
        comparator = typeof comparator == 'function' ? comparator : undefined$1;
        return (array && array.length) ? baseUniq(array, undefined$1, comparator) : [];
      }

      /**
       * This method is like `_.zip` except that it accepts an array of grouped
       * elements and creates an array regrouping the elements to their pre-zip
       * configuration.
       *
       * @static
       * @memberOf _
       * @since 1.2.0
       * @category Array
       * @param {Array} array The array of grouped elements to process.
       * @returns {Array} Returns the new array of regrouped elements.
       * @example
       *
       * var zipped = _.zip(['a', 'b'], [1, 2], [true, false]);
       * // => [['a', 1, true], ['b', 2, false]]
       *
       * _.unzip(zipped);
       * // => [['a', 'b'], [1, 2], [true, false]]
       */
      function unzip(array) {
        if (!(array && array.length)) {
          return [];
        }
        var length = 0;
        array = arrayFilter(array, function(group) {
          if (isArrayLikeObject(group)) {
            length = nativeMax(group.length, length);
            return true;
          }
        });
        return baseTimes(length, function(index) {
          return arrayMap(array, baseProperty(index));
        });
      }

      /**
       * This method is like `_.unzip` except that it accepts `iteratee` to specify
       * how regrouped values should be combined. The iteratee is invoked with the
       * elements of each group: (...group).
       *
       * @static
       * @memberOf _
       * @since 3.8.0
       * @category Array
       * @param {Array} array The array of grouped elements to process.
       * @param {Function} [iteratee=_.identity] The function to combine
       *  regrouped values.
       * @returns {Array} Returns the new array of regrouped elements.
       * @example
       *
       * var zipped = _.zip([1, 2], [10, 20], [100, 200]);
       * // => [[1, 10, 100], [2, 20, 200]]
       *
       * _.unzipWith(zipped, _.add);
       * // => [3, 30, 300]
       */
      function unzipWith(array, iteratee) {
        if (!(array && array.length)) {
          return [];
        }
        var result = unzip(array);
        if (iteratee == null) {
          return result;
        }
        return arrayMap(result, function(group) {
          return apply(iteratee, undefined$1, group);
        });
      }

      /**
       * Creates an array excluding all given values using
       * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * for equality comparisons.
       *
       * **Note:** Unlike `_.pull`, this method returns a new array.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {Array} array The array to inspect.
       * @param {...*} [values] The values to exclude.
       * @returns {Array} Returns the new array of filtered values.
       * @see _.difference, _.xor
       * @example
       *
       * _.without([2, 1, 2, 3], 1, 2);
       * // => [3]
       */
      var without = baseRest(function(array, values) {
        return isArrayLikeObject(array)
          ? baseDifference(array, values)
          : [];
      });

      /**
       * Creates an array of unique values that is the
       * [symmetric difference](https://en.wikipedia.org/wiki/Symmetric_difference)
       * of the given arrays. The order of result values is determined by the order
       * they occur in the arrays.
       *
       * @static
       * @memberOf _
       * @since 2.4.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @returns {Array} Returns the new array of filtered values.
       * @see _.difference, _.without
       * @example
       *
       * _.xor([2, 1], [2, 3]);
       * // => [1, 3]
       */
      var xor = baseRest(function(arrays) {
        return baseXor(arrayFilter(arrays, isArrayLikeObject));
      });

      /**
       * This method is like `_.xor` except that it accepts `iteratee` which is
       * invoked for each element of each `arrays` to generate the criterion by
       * which by which they're compared. The order of result values is determined
       * by the order they occur in the arrays. The iteratee is invoked with one
       * argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Array} Returns the new array of filtered values.
       * @example
       *
       * _.xorBy([2.1, 1.2], [2.3, 3.4], Math.floor);
       * // => [1.2, 3.4]
       *
       * // The `_.property` iteratee shorthand.
       * _.xorBy([{ 'x': 1 }], [{ 'x': 2 }, { 'x': 1 }], 'x');
       * // => [{ 'x': 2 }]
       */
      var xorBy = baseRest(function(arrays) {
        var iteratee = last(arrays);
        if (isArrayLikeObject(iteratee)) {
          iteratee = undefined$1;
        }
        return baseXor(arrayFilter(arrays, isArrayLikeObject), getIteratee(iteratee, 2));
      });

      /**
       * This method is like `_.xor` except that it accepts `comparator` which is
       * invoked to compare elements of `arrays`. The order of result values is
       * determined by the order they occur in the arrays. The comparator is invoked
       * with two arguments: (arrVal, othVal).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Array
       * @param {...Array} [arrays] The arrays to inspect.
       * @param {Function} [comparator] The comparator invoked per element.
       * @returns {Array} Returns the new array of filtered values.
       * @example
       *
       * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
       * var others = [{ 'x': 1, 'y': 1 }, { 'x': 1, 'y': 2 }];
       *
       * _.xorWith(objects, others, _.isEqual);
       * // => [{ 'x': 2, 'y': 1 }, { 'x': 1, 'y': 1 }]
       */
      var xorWith = baseRest(function(arrays) {
        var comparator = last(arrays);
        comparator = typeof comparator == 'function' ? comparator : undefined$1;
        return baseXor(arrayFilter(arrays, isArrayLikeObject), undefined$1, comparator);
      });

      /**
       * Creates an array of grouped elements, the first of which contains the
       * first elements of the given arrays, the second of which contains the
       * second elements of the given arrays, and so on.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Array
       * @param {...Array} [arrays] The arrays to process.
       * @returns {Array} Returns the new array of grouped elements.
       * @example
       *
       * _.zip(['a', 'b'], [1, 2], [true, false]);
       * // => [['a', 1, true], ['b', 2, false]]
       */
      var zip = baseRest(unzip);

      /**
       * This method is like `_.fromPairs` except that it accepts two arrays,
       * one of property identifiers and one of corresponding values.
       *
       * @static
       * @memberOf _
       * @since 0.4.0
       * @category Array
       * @param {Array} [props=[]] The property identifiers.
       * @param {Array} [values=[]] The property values.
       * @returns {Object} Returns the new object.
       * @example
       *
       * _.zipObject(['a', 'b'], [1, 2]);
       * // => { 'a': 1, 'b': 2 }
       */
      function zipObject(props, values) {
        return baseZipObject(props || [], values || [], assignValue);
      }

      /**
       * This method is like `_.zipObject` except that it supports property paths.
       *
       * @static
       * @memberOf _
       * @since 4.1.0
       * @category Array
       * @param {Array} [props=[]] The property identifiers.
       * @param {Array} [values=[]] The property values.
       * @returns {Object} Returns the new object.
       * @example
       *
       * _.zipObjectDeep(['a.b[0].c', 'a.b[1].d'], [1, 2]);
       * // => { 'a': { 'b': [{ 'c': 1 }, { 'd': 2 }] } }
       */
      function zipObjectDeep(props, values) {
        return baseZipObject(props || [], values || [], baseSet);
      }

      /**
       * This method is like `_.zip` except that it accepts `iteratee` to specify
       * how grouped values should be combined. The iteratee is invoked with the
       * elements of each group: (...group).
       *
       * @static
       * @memberOf _
       * @since 3.8.0
       * @category Array
       * @param {...Array} [arrays] The arrays to process.
       * @param {Function} [iteratee=_.identity] The function to combine
       *  grouped values.
       * @returns {Array} Returns the new array of grouped elements.
       * @example
       *
       * _.zipWith([1, 2], [10, 20], [100, 200], function(a, b, c) {
       *   return a + b + c;
       * });
       * // => [111, 222]
       */
      var zipWith = baseRest(function(arrays) {
        var length = arrays.length,
            iteratee = length > 1 ? arrays[length - 1] : undefined$1;

        iteratee = typeof iteratee == 'function' ? (arrays.pop(), iteratee) : undefined$1;
        return unzipWith(arrays, iteratee);
      });

      /*------------------------------------------------------------------------*/

      /**
       * Creates a `lodash` wrapper instance that wraps `value` with explicit method
       * chain sequences enabled. The result of such sequences must be unwrapped
       * with `_#value`.
       *
       * @static
       * @memberOf _
       * @since 1.3.0
       * @category Seq
       * @param {*} value The value to wrap.
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'age': 36 },
       *   { 'user': 'fred',    'age': 40 },
       *   { 'user': 'pebbles', 'age': 1 }
       * ];
       *
       * var youngest = _
       *   .chain(users)
       *   .sortBy('age')
       *   .map(function(o) {
       *     return o.user + ' is ' + o.age;
       *   })
       *   .head()
       *   .value();
       * // => 'pebbles is 1'
       */
      function chain(value) {
        var result = lodash(value);
        result.__chain__ = true;
        return result;
      }

      /**
       * This method invokes `interceptor` and returns `value`. The interceptor
       * is invoked with one argument; (value). The purpose of this method is to
       * "tap into" a method chain sequence in order to modify intermediate results.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Seq
       * @param {*} value The value to provide to `interceptor`.
       * @param {Function} interceptor The function to invoke.
       * @returns {*} Returns `value`.
       * @example
       *
       * _([1, 2, 3])
       *  .tap(function(array) {
       *    // Mutate input array.
       *    array.pop();
       *  })
       *  .reverse()
       *  .value();
       * // => [2, 1]
       */
      function tap(value, interceptor) {
        interceptor(value);
        return value;
      }

      /**
       * This method is like `_.tap` except that it returns the result of `interceptor`.
       * The purpose of this method is to "pass thru" values replacing intermediate
       * results in a method chain sequence.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Seq
       * @param {*} value The value to provide to `interceptor`.
       * @param {Function} interceptor The function to invoke.
       * @returns {*} Returns the result of `interceptor`.
       * @example
       *
       * _('  abc  ')
       *  .chain()
       *  .trim()
       *  .thru(function(value) {
       *    return [value];
       *  })
       *  .value();
       * // => ['abc']
       */
      function thru(value, interceptor) {
        return interceptor(value);
      }

      /**
       * This method is the wrapper version of `_.at`.
       *
       * @name at
       * @memberOf _
       * @since 1.0.0
       * @category Seq
       * @param {...(string|string[])} [paths] The property paths to pick.
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': 3 } }, 4] };
       *
       * _(object).at(['a[0].b.c', 'a[1]']).value();
       * // => [3, 4]
       */
      var wrapperAt = flatRest(function(paths) {
        var length = paths.length,
            start = length ? paths[0] : 0,
            value = this.__wrapped__,
            interceptor = function(object) { return baseAt(object, paths); };

        if (length > 1 || this.__actions__.length ||
            !(value instanceof LazyWrapper) || !isIndex(start)) {
          return this.thru(interceptor);
        }
        value = value.slice(start, +start + (length ? 1 : 0));
        value.__actions__.push({
          'func': thru,
          'args': [interceptor],
          'thisArg': undefined$1
        });
        return new LodashWrapper(value, this.__chain__).thru(function(array) {
          if (length && !array.length) {
            array.push(undefined$1);
          }
          return array;
        });
      });

      /**
       * Creates a `lodash` wrapper instance with explicit method chain sequences enabled.
       *
       * @name chain
       * @memberOf _
       * @since 0.1.0
       * @category Seq
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * var users = [
       *   { 'user': 'barney', 'age': 36 },
       *   { 'user': 'fred',   'age': 40 }
       * ];
       *
       * // A sequence without explicit chaining.
       * _(users).head();
       * // => { 'user': 'barney', 'age': 36 }
       *
       * // A sequence with explicit chaining.
       * _(users)
       *   .chain()
       *   .head()
       *   .pick('user')
       *   .value();
       * // => { 'user': 'barney' }
       */
      function wrapperChain() {
        return chain(this);
      }

      /**
       * Executes the chain sequence and returns the wrapped result.
       *
       * @name commit
       * @memberOf _
       * @since 3.2.0
       * @category Seq
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * var array = [1, 2];
       * var wrapped = _(array).push(3);
       *
       * console.log(array);
       * // => [1, 2]
       *
       * wrapped = wrapped.commit();
       * console.log(array);
       * // => [1, 2, 3]
       *
       * wrapped.last();
       * // => 3
       *
       * console.log(array);
       * // => [1, 2, 3]
       */
      function wrapperCommit() {
        return new LodashWrapper(this.value(), this.__chain__);
      }

      /**
       * Gets the next value on a wrapped object following the
       * [iterator protocol](https://mdn.io/iteration_protocols#iterator).
       *
       * @name next
       * @memberOf _
       * @since 4.0.0
       * @category Seq
       * @returns {Object} Returns the next iterator value.
       * @example
       *
       * var wrapped = _([1, 2]);
       *
       * wrapped.next();
       * // => { 'done': false, 'value': 1 }
       *
       * wrapped.next();
       * // => { 'done': false, 'value': 2 }
       *
       * wrapped.next();
       * // => { 'done': true, 'value': undefined }
       */
      function wrapperNext() {
        if (this.__values__ === undefined$1) {
          this.__values__ = toArray(this.value());
        }
        var done = this.__index__ >= this.__values__.length,
            value = done ? undefined$1 : this.__values__[this.__index__++];

        return { 'done': done, 'value': value };
      }

      /**
       * Enables the wrapper to be iterable.
       *
       * @name Symbol.iterator
       * @memberOf _
       * @since 4.0.0
       * @category Seq
       * @returns {Object} Returns the wrapper object.
       * @example
       *
       * var wrapped = _([1, 2]);
       *
       * wrapped[Symbol.iterator]() === wrapped;
       * // => true
       *
       * Array.from(wrapped);
       * // => [1, 2]
       */
      function wrapperToIterator() {
        return this;
      }

      /**
       * Creates a clone of the chain sequence planting `value` as the wrapped value.
       *
       * @name plant
       * @memberOf _
       * @since 3.2.0
       * @category Seq
       * @param {*} value The value to plant.
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * function square(n) {
       *   return n * n;
       * }
       *
       * var wrapped = _([1, 2]).map(square);
       * var other = wrapped.plant([3, 4]);
       *
       * other.value();
       * // => [9, 16]
       *
       * wrapped.value();
       * // => [1, 4]
       */
      function wrapperPlant(value) {
        var result,
            parent = this;

        while (parent instanceof baseLodash) {
          var clone = wrapperClone(parent);
          clone.__index__ = 0;
          clone.__values__ = undefined$1;
          if (result) {
            previous.__wrapped__ = clone;
          } else {
            result = clone;
          }
          var previous = clone;
          parent = parent.__wrapped__;
        }
        previous.__wrapped__ = value;
        return result;
      }

      /**
       * This method is the wrapper version of `_.reverse`.
       *
       * **Note:** This method mutates the wrapped array.
       *
       * @name reverse
       * @memberOf _
       * @since 0.1.0
       * @category Seq
       * @returns {Object} Returns the new `lodash` wrapper instance.
       * @example
       *
       * var array = [1, 2, 3];
       *
       * _(array).reverse().value()
       * // => [3, 2, 1]
       *
       * console.log(array);
       * // => [3, 2, 1]
       */
      function wrapperReverse() {
        var value = this.__wrapped__;
        if (value instanceof LazyWrapper) {
          var wrapped = value;
          if (this.__actions__.length) {
            wrapped = new LazyWrapper(this);
          }
          wrapped = wrapped.reverse();
          wrapped.__actions__.push({
            'func': thru,
            'args': [reverse],
            'thisArg': undefined$1
          });
          return new LodashWrapper(wrapped, this.__chain__);
        }
        return this.thru(reverse);
      }

      /**
       * Executes the chain sequence to resolve the unwrapped value.
       *
       * @name value
       * @memberOf _
       * @since 0.1.0
       * @alias toJSON, valueOf
       * @category Seq
       * @returns {*} Returns the resolved unwrapped value.
       * @example
       *
       * _([1, 2, 3]).value();
       * // => [1, 2, 3]
       */
      function wrapperValue() {
        return baseWrapperValue(this.__wrapped__, this.__actions__);
      }

      /*------------------------------------------------------------------------*/

      /**
       * Creates an object composed of keys generated from the results of running
       * each element of `collection` thru `iteratee`. The corresponding value of
       * each key is the number of times the key was returned by `iteratee`. The
       * iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 0.5.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee to transform keys.
       * @returns {Object} Returns the composed aggregate object.
       * @example
       *
       * _.countBy([6.1, 4.2, 6.3], Math.floor);
       * // => { '4': 1, '6': 2 }
       *
       * // The `_.property` iteratee shorthand.
       * _.countBy(['one', 'two', 'three'], 'length');
       * // => { '3': 2, '5': 1 }
       */
      var countBy = createAggregator(function(result, value, key) {
        if (hasOwnProperty.call(result, key)) {
          ++result[key];
        } else {
          baseAssignValue(result, key, 1);
        }
      });

      /**
       * Checks if `predicate` returns truthy for **all** elements of `collection`.
       * Iteration is stopped once `predicate` returns falsey. The predicate is
       * invoked with three arguments: (value, index|key, collection).
       *
       * **Note:** This method returns `true` for
       * [empty collections](https://en.wikipedia.org/wiki/Empty_set) because
       * [everything is true](https://en.wikipedia.org/wiki/Vacuous_truth) of
       * elements of empty collections.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {boolean} Returns `true` if all elements pass the predicate check,
       *  else `false`.
       * @example
       *
       * _.every([true, 1, null, 'yes'], Boolean);
       * // => false
       *
       * var users = [
       *   { 'user': 'barney', 'age': 36, 'active': false },
       *   { 'user': 'fred',   'age': 40, 'active': false }
       * ];
       *
       * // The `_.matches` iteratee shorthand.
       * _.every(users, { 'user': 'barney', 'active': false });
       * // => false
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.every(users, ['active', false]);
       * // => true
       *
       * // The `_.property` iteratee shorthand.
       * _.every(users, 'active');
       * // => false
       */
      function every(collection, predicate, guard) {
        var func = isArray(collection) ? arrayEvery : baseEvery;
        if (guard && isIterateeCall(collection, predicate, guard)) {
          predicate = undefined$1;
        }
        return func(collection, getIteratee(predicate, 3));
      }

      /**
       * Iterates over elements of `collection`, returning an array of all elements
       * `predicate` returns truthy for. The predicate is invoked with three
       * arguments: (value, index|key, collection).
       *
       * **Note:** Unlike `_.remove`, this method returns a new array.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the new filtered array.
       * @see _.reject
       * @example
       *
       * var users = [
       *   { 'user': 'barney', 'age': 36, 'active': true },
       *   { 'user': 'fred',   'age': 40, 'active': false }
       * ];
       *
       * _.filter(users, function(o) { return !o.active; });
       * // => objects for ['fred']
       *
       * // The `_.matches` iteratee shorthand.
       * _.filter(users, { 'age': 36, 'active': true });
       * // => objects for ['barney']
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.filter(users, ['active', false]);
       * // => objects for ['fred']
       *
       * // The `_.property` iteratee shorthand.
       * _.filter(users, 'active');
       * // => objects for ['barney']
       */
      function filter(collection, predicate) {
        var func = isArray(collection) ? arrayFilter : baseFilter;
        return func(collection, getIteratee(predicate, 3));
      }

      /**
       * Iterates over elements of `collection`, returning the first element
       * `predicate` returns truthy for. The predicate is invoked with three
       * arguments: (value, index|key, collection).
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to inspect.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @param {number} [fromIndex=0] The index to search from.
       * @returns {*} Returns the matched element, else `undefined`.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'age': 36, 'active': true },
       *   { 'user': 'fred',    'age': 40, 'active': false },
       *   { 'user': 'pebbles', 'age': 1,  'active': true }
       * ];
       *
       * _.find(users, function(o) { return o.age < 40; });
       * // => object for 'barney'
       *
       * // The `_.matches` iteratee shorthand.
       * _.find(users, { 'age': 1, 'active': true });
       * // => object for 'pebbles'
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.find(users, ['active', false]);
       * // => object for 'fred'
       *
       * // The `_.property` iteratee shorthand.
       * _.find(users, 'active');
       * // => object for 'barney'
       */
      var find = createFind(findIndex);

      /**
       * This method is like `_.find` except that it iterates over elements of
       * `collection` from right to left.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to inspect.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @param {number} [fromIndex=collection.length-1] The index to search from.
       * @returns {*} Returns the matched element, else `undefined`.
       * @example
       *
       * _.findLast([1, 2, 3, 4], function(n) {
       *   return n % 2 == 1;
       * });
       * // => 3
       */
      var findLast = createFind(findLastIndex);

      /**
       * Creates a flattened array of values by running each element in `collection`
       * thru `iteratee` and flattening the mapped results. The iteratee is invoked
       * with three arguments: (value, index|key, collection).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the new flattened array.
       * @example
       *
       * function duplicate(n) {
       *   return [n, n];
       * }
       *
       * _.flatMap([1, 2], duplicate);
       * // => [1, 1, 2, 2]
       */
      function flatMap(collection, iteratee) {
        return baseFlatten(map(collection, iteratee), 1);
      }

      /**
       * This method is like `_.flatMap` except that it recursively flattens the
       * mapped results.
       *
       * @static
       * @memberOf _
       * @since 4.7.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the new flattened array.
       * @example
       *
       * function duplicate(n) {
       *   return [[[n, n]]];
       * }
       *
       * _.flatMapDeep([1, 2], duplicate);
       * // => [1, 1, 2, 2]
       */
      function flatMapDeep(collection, iteratee) {
        return baseFlatten(map(collection, iteratee), INFINITY);
      }

      /**
       * This method is like `_.flatMap` except that it recursively flattens the
       * mapped results up to `depth` times.
       *
       * @static
       * @memberOf _
       * @since 4.7.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @param {number} [depth=1] The maximum recursion depth.
       * @returns {Array} Returns the new flattened array.
       * @example
       *
       * function duplicate(n) {
       *   return [[[n, n]]];
       * }
       *
       * _.flatMapDepth([1, 2], duplicate, 2);
       * // => [[1, 1], [2, 2]]
       */
      function flatMapDepth(collection, iteratee, depth) {
        depth = depth === undefined$1 ? 1 : toInteger(depth);
        return baseFlatten(map(collection, iteratee), depth);
      }

      /**
       * Iterates over elements of `collection` and invokes `iteratee` for each element.
       * The iteratee is invoked with three arguments: (value, index|key, collection).
       * Iteratee functions may exit iteration early by explicitly returning `false`.
       *
       * **Note:** As with other "Collections" methods, objects with a "length"
       * property are iterated like arrays. To avoid this behavior use `_.forIn`
       * or `_.forOwn` for object iteration.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @alias each
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Array|Object} Returns `collection`.
       * @see _.forEachRight
       * @example
       *
       * _.forEach([1, 2], function(value) {
       *   console.log(value);
       * });
       * // => Logs `1` then `2`.
       *
       * _.forEach({ 'a': 1, 'b': 2 }, function(value, key) {
       *   console.log(key);
       * });
       * // => Logs 'a' then 'b' (iteration order is not guaranteed).
       */
      function forEach(collection, iteratee) {
        var func = isArray(collection) ? arrayEach : baseEach;
        return func(collection, getIteratee(iteratee, 3));
      }

      /**
       * This method is like `_.forEach` except that it iterates over elements of
       * `collection` from right to left.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @alias eachRight
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Array|Object} Returns `collection`.
       * @see _.forEach
       * @example
       *
       * _.forEachRight([1, 2], function(value) {
       *   console.log(value);
       * });
       * // => Logs `2` then `1`.
       */
      function forEachRight(collection, iteratee) {
        var func = isArray(collection) ? arrayEachRight : baseEachRight;
        return func(collection, getIteratee(iteratee, 3));
      }

      /**
       * Creates an object composed of keys generated from the results of running
       * each element of `collection` thru `iteratee`. The order of grouped values
       * is determined by the order they occur in `collection`. The corresponding
       * value of each key is an array of elements responsible for generating the
       * key. The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee to transform keys.
       * @returns {Object} Returns the composed aggregate object.
       * @example
       *
       * _.groupBy([6.1, 4.2, 6.3], Math.floor);
       * // => { '4': [4.2], '6': [6.1, 6.3] }
       *
       * // The `_.property` iteratee shorthand.
       * _.groupBy(['one', 'two', 'three'], 'length');
       * // => { '3': ['one', 'two'], '5': ['three'] }
       */
      var groupBy = createAggregator(function(result, value, key) {
        if (hasOwnProperty.call(result, key)) {
          result[key].push(value);
        } else {
          baseAssignValue(result, key, [value]);
        }
      });

      /**
       * Checks if `value` is in `collection`. If `collection` is a string, it's
       * checked for a substring of `value`, otherwise
       * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * is used for equality comparisons. If `fromIndex` is negative, it's used as
       * the offset from the end of `collection`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object|string} collection The collection to inspect.
       * @param {*} value The value to search for.
       * @param {number} [fromIndex=0] The index to search from.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.reduce`.
       * @returns {boolean} Returns `true` if `value` is found, else `false`.
       * @example
       *
       * _.includes([1, 2, 3], 1);
       * // => true
       *
       * _.includes([1, 2, 3], 1, 2);
       * // => false
       *
       * _.includes({ 'a': 1, 'b': 2 }, 1);
       * // => true
       *
       * _.includes('abcd', 'bc');
       * // => true
       */
      function includes(collection, value, fromIndex, guard) {
        collection = isArrayLike(collection) ? collection : values(collection);
        fromIndex = (fromIndex && !guard) ? toInteger(fromIndex) : 0;

        var length = collection.length;
        if (fromIndex < 0) {
          fromIndex = nativeMax(length + fromIndex, 0);
        }
        return isString(collection)
          ? (fromIndex <= length && collection.indexOf(value, fromIndex) > -1)
          : (!!length && baseIndexOf(collection, value, fromIndex) > -1);
      }

      /**
       * Invokes the method at `path` of each element in `collection`, returning
       * an array of the results of each invoked method. Any additional arguments
       * are provided to each invoked method. If `path` is a function, it's invoked
       * for, and `this` bound to, each element in `collection`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Array|Function|string} path The path of the method to invoke or
       *  the function invoked per iteration.
       * @param {...*} [args] The arguments to invoke each method with.
       * @returns {Array} Returns the array of results.
       * @example
       *
       * _.invokeMap([[5, 1, 7], [3, 2, 1]], 'sort');
       * // => [[1, 5, 7], [1, 2, 3]]
       *
       * _.invokeMap([123, 456], String.prototype.split, '');
       * // => [['1', '2', '3'], ['4', '5', '6']]
       */
      var invokeMap = baseRest(function(collection, path, args) {
        var index = -1,
            isFunc = typeof path == 'function',
            result = isArrayLike(collection) ? Array(collection.length) : [];

        baseEach(collection, function(value) {
          result[++index] = isFunc ? apply(path, value, args) : baseInvoke(value, path, args);
        });
        return result;
      });

      /**
       * Creates an object composed of keys generated from the results of running
       * each element of `collection` thru `iteratee`. The corresponding value of
       * each key is the last element responsible for generating the key. The
       * iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee to transform keys.
       * @returns {Object} Returns the composed aggregate object.
       * @example
       *
       * var array = [
       *   { 'dir': 'left', 'code': 97 },
       *   { 'dir': 'right', 'code': 100 }
       * ];
       *
       * _.keyBy(array, function(o) {
       *   return String.fromCharCode(o.code);
       * });
       * // => { 'a': { 'dir': 'left', 'code': 97 }, 'd': { 'dir': 'right', 'code': 100 } }
       *
       * _.keyBy(array, 'dir');
       * // => { 'left': { 'dir': 'left', 'code': 97 }, 'right': { 'dir': 'right', 'code': 100 } }
       */
      var keyBy = createAggregator(function(result, value, key) {
        baseAssignValue(result, key, value);
      });

      /**
       * Creates an array of values by running each element in `collection` thru
       * `iteratee`. The iteratee is invoked with three arguments:
       * (value, index|key, collection).
       *
       * Many lodash methods are guarded to work as iteratees for methods like
       * `_.every`, `_.filter`, `_.map`, `_.mapValues`, `_.reject`, and `_.some`.
       *
       * The guarded methods are:
       * `ary`, `chunk`, `curry`, `curryRight`, `drop`, `dropRight`, `every`,
       * `fill`, `invert`, `parseInt`, `random`, `range`, `rangeRight`, `repeat`,
       * `sampleSize`, `slice`, `some`, `sortBy`, `split`, `take`, `takeRight`,
       * `template`, `trim`, `trimEnd`, `trimStart`, and `words`
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the new mapped array.
       * @example
       *
       * function square(n) {
       *   return n * n;
       * }
       *
       * _.map([4, 8], square);
       * // => [16, 64]
       *
       * _.map({ 'a': 4, 'b': 8 }, square);
       * // => [16, 64] (iteration order is not guaranteed)
       *
       * var users = [
       *   { 'user': 'barney' },
       *   { 'user': 'fred' }
       * ];
       *
       * // The `_.property` iteratee shorthand.
       * _.map(users, 'user');
       * // => ['barney', 'fred']
       */
      function map(collection, iteratee) {
        var func = isArray(collection) ? arrayMap : baseMap;
        return func(collection, getIteratee(iteratee, 3));
      }

      /**
       * This method is like `_.sortBy` except that it allows specifying the sort
       * orders of the iteratees to sort by. If `orders` is unspecified, all values
       * are sorted in ascending order. Otherwise, specify an order of "desc" for
       * descending or "asc" for ascending sort order of corresponding values.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Array[]|Function[]|Object[]|string[]} [iteratees=[_.identity]]
       *  The iteratees to sort by.
       * @param {string[]} [orders] The sort orders of `iteratees`.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.reduce`.
       * @returns {Array} Returns the new sorted array.
       * @example
       *
       * var users = [
       *   { 'user': 'fred',   'age': 48 },
       *   { 'user': 'barney', 'age': 34 },
       *   { 'user': 'fred',   'age': 40 },
       *   { 'user': 'barney', 'age': 36 }
       * ];
       *
       * // Sort by `user` in ascending order and by `age` in descending order.
       * _.orderBy(users, ['user', 'age'], ['asc', 'desc']);
       * // => objects for [['barney', 36], ['barney', 34], ['fred', 48], ['fred', 40]]
       */
      function orderBy(collection, iteratees, orders, guard) {
        if (collection == null) {
          return [];
        }
        if (!isArray(iteratees)) {
          iteratees = iteratees == null ? [] : [iteratees];
        }
        orders = guard ? undefined$1 : orders;
        if (!isArray(orders)) {
          orders = orders == null ? [] : [orders];
        }
        return baseOrderBy(collection, iteratees, orders);
      }

      /**
       * Creates an array of elements split into two groups, the first of which
       * contains elements `predicate` returns truthy for, the second of which
       * contains elements `predicate` returns falsey for. The predicate is
       * invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the array of grouped elements.
       * @example
       *
       * var users = [
       *   { 'user': 'barney',  'age': 36, 'active': false },
       *   { 'user': 'fred',    'age': 40, 'active': true },
       *   { 'user': 'pebbles', 'age': 1,  'active': false }
       * ];
       *
       * _.partition(users, function(o) { return o.active; });
       * // => objects for [['fred'], ['barney', 'pebbles']]
       *
       * // The `_.matches` iteratee shorthand.
       * _.partition(users, { 'age': 1, 'active': false });
       * // => objects for [['pebbles'], ['barney', 'fred']]
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.partition(users, ['active', false]);
       * // => objects for [['barney', 'pebbles'], ['fred']]
       *
       * // The `_.property` iteratee shorthand.
       * _.partition(users, 'active');
       * // => objects for [['fred'], ['barney', 'pebbles']]
       */
      var partition = createAggregator(function(result, value, key) {
        result[key ? 0 : 1].push(value);
      }, function() { return [[], []]; });

      /**
       * Reduces `collection` to a value which is the accumulated result of running
       * each element in `collection` thru `iteratee`, where each successive
       * invocation is supplied the return value of the previous. If `accumulator`
       * is not given, the first element of `collection` is used as the initial
       * value. The iteratee is invoked with four arguments:
       * (accumulator, value, index|key, collection).
       *
       * Many lodash methods are guarded to work as iteratees for methods like
       * `_.reduce`, `_.reduceRight`, and `_.transform`.
       *
       * The guarded methods are:
       * `assign`, `defaults`, `defaultsDeep`, `includes`, `merge`, `orderBy`,
       * and `sortBy`
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @param {*} [accumulator] The initial value.
       * @returns {*} Returns the accumulated value.
       * @see _.reduceRight
       * @example
       *
       * _.reduce([1, 2], function(sum, n) {
       *   return sum + n;
       * }, 0);
       * // => 3
       *
       * _.reduce({ 'a': 1, 'b': 2, 'c': 1 }, function(result, value, key) {
       *   (result[value] || (result[value] = [])).push(key);
       *   return result;
       * }, {});
       * // => { '1': ['a', 'c'], '2': ['b'] } (iteration order is not guaranteed)
       */
      function reduce(collection, iteratee, accumulator) {
        var func = isArray(collection) ? arrayReduce : baseReduce,
            initAccum = arguments.length < 3;

        return func(collection, getIteratee(iteratee, 4), accumulator, initAccum, baseEach);
      }

      /**
       * This method is like `_.reduce` except that it iterates over elements of
       * `collection` from right to left.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @param {*} [accumulator] The initial value.
       * @returns {*} Returns the accumulated value.
       * @see _.reduce
       * @example
       *
       * var array = [[0, 1], [2, 3], [4, 5]];
       *
       * _.reduceRight(array, function(flattened, other) {
       *   return flattened.concat(other);
       * }, []);
       * // => [4, 5, 2, 3, 0, 1]
       */
      function reduceRight(collection, iteratee, accumulator) {
        var func = isArray(collection) ? arrayReduceRight : baseReduce,
            initAccum = arguments.length < 3;

        return func(collection, getIteratee(iteratee, 4), accumulator, initAccum, baseEachRight);
      }

      /**
       * The opposite of `_.filter`; this method returns the elements of `collection`
       * that `predicate` does **not** return truthy for.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the new filtered array.
       * @see _.filter
       * @example
       *
       * var users = [
       *   { 'user': 'barney', 'age': 36, 'active': false },
       *   { 'user': 'fred',   'age': 40, 'active': true }
       * ];
       *
       * _.reject(users, function(o) { return !o.active; });
       * // => objects for ['fred']
       *
       * // The `_.matches` iteratee shorthand.
       * _.reject(users, { 'age': 40, 'active': true });
       * // => objects for ['barney']
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.reject(users, ['active', false]);
       * // => objects for ['fred']
       *
       * // The `_.property` iteratee shorthand.
       * _.reject(users, 'active');
       * // => objects for ['barney']
       */
      function reject(collection, predicate) {
        var func = isArray(collection) ? arrayFilter : baseFilter;
        return func(collection, negate(getIteratee(predicate, 3)));
      }

      /**
       * Gets a random element from `collection`.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to sample.
       * @returns {*} Returns the random element.
       * @example
       *
       * _.sample([1, 2, 3, 4]);
       * // => 2
       */
      function sample(collection) {
        var func = isArray(collection) ? arraySample : baseSample;
        return func(collection);
      }

      /**
       * Gets `n` random elements at unique keys from `collection` up to the
       * size of `collection`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Collection
       * @param {Array|Object} collection The collection to sample.
       * @param {number} [n=1] The number of elements to sample.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the random elements.
       * @example
       *
       * _.sampleSize([1, 2, 3], 2);
       * // => [3, 1]
       *
       * _.sampleSize([1, 2, 3], 4);
       * // => [2, 3, 1]
       */
      function sampleSize(collection, n, guard) {
        if ((guard ? isIterateeCall(collection, n, guard) : n === undefined$1)) {
          n = 1;
        } else {
          n = toInteger(n);
        }
        var func = isArray(collection) ? arraySampleSize : baseSampleSize;
        return func(collection, n);
      }

      /**
       * Creates an array of shuffled values, using a version of the
       * [Fisher-Yates shuffle](https://en.wikipedia.org/wiki/Fisher-Yates_shuffle).
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to shuffle.
       * @returns {Array} Returns the new shuffled array.
       * @example
       *
       * _.shuffle([1, 2, 3, 4]);
       * // => [4, 1, 3, 2]
       */
      function shuffle(collection) {
        var func = isArray(collection) ? arrayShuffle : baseShuffle;
        return func(collection);
      }

      /**
       * Gets the size of `collection` by returning its length for array-like
       * values or the number of own enumerable string keyed properties for objects.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object|string} collection The collection to inspect.
       * @returns {number} Returns the collection size.
       * @example
       *
       * _.size([1, 2, 3]);
       * // => 3
       *
       * _.size({ 'a': 1, 'b': 2 });
       * // => 2
       *
       * _.size('pebbles');
       * // => 7
       */
      function size(collection) {
        if (collection == null) {
          return 0;
        }
        if (isArrayLike(collection)) {
          return isString(collection) ? stringSize(collection) : collection.length;
        }
        var tag = getTag(collection);
        if (tag == mapTag || tag == setTag) {
          return collection.size;
        }
        return baseKeys(collection).length;
      }

      /**
       * Checks if `predicate` returns truthy for **any** element of `collection`.
       * Iteration is stopped once `predicate` returns truthy. The predicate is
       * invoked with three arguments: (value, index|key, collection).
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {boolean} Returns `true` if any element passes the predicate check,
       *  else `false`.
       * @example
       *
       * _.some([null, 0, 'yes', false], Boolean);
       * // => true
       *
       * var users = [
       *   { 'user': 'barney', 'active': true },
       *   { 'user': 'fred',   'active': false }
       * ];
       *
       * // The `_.matches` iteratee shorthand.
       * _.some(users, { 'user': 'barney', 'active': false });
       * // => false
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.some(users, ['active', false]);
       * // => true
       *
       * // The `_.property` iteratee shorthand.
       * _.some(users, 'active');
       * // => true
       */
      function some(collection, predicate, guard) {
        var func = isArray(collection) ? arraySome : baseSome;
        if (guard && isIterateeCall(collection, predicate, guard)) {
          predicate = undefined$1;
        }
        return func(collection, getIteratee(predicate, 3));
      }

      /**
       * Creates an array of elements, sorted in ascending order by the results of
       * running each element in a collection thru each iteratee. This method
       * performs a stable sort, that is, it preserves the original sort order of
       * equal elements. The iteratees are invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Collection
       * @param {Array|Object} collection The collection to iterate over.
       * @param {...(Function|Function[])} [iteratees=[_.identity]]
       *  The iteratees to sort by.
       * @returns {Array} Returns the new sorted array.
       * @example
       *
       * var users = [
       *   { 'user': 'fred',   'age': 48 },
       *   { 'user': 'barney', 'age': 36 },
       *   { 'user': 'fred',   'age': 40 },
       *   { 'user': 'barney', 'age': 34 }
       * ];
       *
       * _.sortBy(users, [function(o) { return o.user; }]);
       * // => objects for [['barney', 36], ['barney', 34], ['fred', 48], ['fred', 40]]
       *
       * _.sortBy(users, ['user', 'age']);
       * // => objects for [['barney', 34], ['barney', 36], ['fred', 40], ['fred', 48]]
       */
      var sortBy = baseRest(function(collection, iteratees) {
        if (collection == null) {
          return [];
        }
        var length = iteratees.length;
        if (length > 1 && isIterateeCall(collection, iteratees[0], iteratees[1])) {
          iteratees = [];
        } else if (length > 2 && isIterateeCall(iteratees[0], iteratees[1], iteratees[2])) {
          iteratees = [iteratees[0]];
        }
        return baseOrderBy(collection, baseFlatten(iteratees, 1), []);
      });

      /*------------------------------------------------------------------------*/

      /**
       * Gets the timestamp of the number of milliseconds that have elapsed since
       * the Unix epoch (1 January 1970 00:00:00 UTC).
       *
       * @static
       * @memberOf _
       * @since 2.4.0
       * @category Date
       * @returns {number} Returns the timestamp.
       * @example
       *
       * _.defer(function(stamp) {
       *   console.log(_.now() - stamp);
       * }, _.now());
       * // => Logs the number of milliseconds it took for the deferred invocation.
       */
      var now = ctxNow || function() {
        return root.Date.now();
      };

      /*------------------------------------------------------------------------*/

      /**
       * The opposite of `_.before`; this method creates a function that invokes
       * `func` once it's called `n` or more times.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {number} n The number of calls before `func` is invoked.
       * @param {Function} func The function to restrict.
       * @returns {Function} Returns the new restricted function.
       * @example
       *
       * var saves = ['profile', 'settings'];
       *
       * var done = _.after(saves.length, function() {
       *   console.log('done saving!');
       * });
       *
       * _.forEach(saves, function(type) {
       *   asyncSave({ 'type': type, 'complete': done });
       * });
       * // => Logs 'done saving!' after the two async saves have completed.
       */
      function after(n, func) {
        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        n = toInteger(n);
        return function() {
          if (--n < 1) {
            return func.apply(this, arguments);
          }
        };
      }

      /**
       * Creates a function that invokes `func`, with up to `n` arguments,
       * ignoring any additional arguments.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Function
       * @param {Function} func The function to cap arguments for.
       * @param {number} [n=func.length] The arity cap.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Function} Returns the new capped function.
       * @example
       *
       * _.map(['6', '8', '10'], _.ary(parseInt, 1));
       * // => [6, 8, 10]
       */
      function ary(func, n, guard) {
        n = guard ? undefined$1 : n;
        n = (func && n == null) ? func.length : n;
        return createWrap(func, WRAP_ARY_FLAG, undefined$1, undefined$1, undefined$1, undefined$1, n);
      }

      /**
       * Creates a function that invokes `func`, with the `this` binding and arguments
       * of the created function, while it's called less than `n` times. Subsequent
       * calls to the created function return the result of the last `func` invocation.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Function
       * @param {number} n The number of calls at which `func` is no longer invoked.
       * @param {Function} func The function to restrict.
       * @returns {Function} Returns the new restricted function.
       * @example
       *
       * jQuery(element).on('click', _.before(5, addContactToList));
       * // => Allows adding up to 4 contacts to the list.
       */
      function before(n, func) {
        var result;
        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        n = toInteger(n);
        return function() {
          if (--n > 0) {
            result = func.apply(this, arguments);
          }
          if (n <= 1) {
            func = undefined$1;
          }
          return result;
        };
      }

      /**
       * Creates a function that invokes `func` with the `this` binding of `thisArg`
       * and `partials` prepended to the arguments it receives.
       *
       * The `_.bind.placeholder` value, which defaults to `_` in monolithic builds,
       * may be used as a placeholder for partially applied arguments.
       *
       * **Note:** Unlike native `Function#bind`, this method doesn't set the "length"
       * property of bound functions.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to bind.
       * @param {*} thisArg The `this` binding of `func`.
       * @param {...*} [partials] The arguments to be partially applied.
       * @returns {Function} Returns the new bound function.
       * @example
       *
       * function greet(greeting, punctuation) {
       *   return greeting + ' ' + this.user + punctuation;
       * }
       *
       * var object = { 'user': 'fred' };
       *
       * var bound = _.bind(greet, object, 'hi');
       * bound('!');
       * // => 'hi fred!'
       *
       * // Bound with placeholders.
       * var bound = _.bind(greet, object, _, '!');
       * bound('hi');
       * // => 'hi fred!'
       */
      var bind = baseRest(function(func, thisArg, partials) {
        var bitmask = WRAP_BIND_FLAG;
        if (partials.length) {
          var holders = replaceHolders(partials, getHolder(bind));
          bitmask |= WRAP_PARTIAL_FLAG;
        }
        return createWrap(func, bitmask, thisArg, partials, holders);
      });

      /**
       * Creates a function that invokes the method at `object[key]` with `partials`
       * prepended to the arguments it receives.
       *
       * This method differs from `_.bind` by allowing bound functions to reference
       * methods that may be redefined or don't yet exist. See
       * [Peter Michaux's article](http://peter.michaux.ca/articles/lazy-function-definition-pattern)
       * for more details.
       *
       * The `_.bindKey.placeholder` value, which defaults to `_` in monolithic
       * builds, may be used as a placeholder for partially applied arguments.
       *
       * @static
       * @memberOf _
       * @since 0.10.0
       * @category Function
       * @param {Object} object The object to invoke the method on.
       * @param {string} key The key of the method.
       * @param {...*} [partials] The arguments to be partially applied.
       * @returns {Function} Returns the new bound function.
       * @example
       *
       * var object = {
       *   'user': 'fred',
       *   'greet': function(greeting, punctuation) {
       *     return greeting + ' ' + this.user + punctuation;
       *   }
       * };
       *
       * var bound = _.bindKey(object, 'greet', 'hi');
       * bound('!');
       * // => 'hi fred!'
       *
       * object.greet = function(greeting, punctuation) {
       *   return greeting + 'ya ' + this.user + punctuation;
       * };
       *
       * bound('!');
       * // => 'hiya fred!'
       *
       * // Bound with placeholders.
       * var bound = _.bindKey(object, 'greet', _, '!');
       * bound('hi');
       * // => 'hiya fred!'
       */
      var bindKey = baseRest(function(object, key, partials) {
        var bitmask = WRAP_BIND_FLAG | WRAP_BIND_KEY_FLAG;
        if (partials.length) {
          var holders = replaceHolders(partials, getHolder(bindKey));
          bitmask |= WRAP_PARTIAL_FLAG;
        }
        return createWrap(key, bitmask, object, partials, holders);
      });

      /**
       * Creates a function that accepts arguments of `func` and either invokes
       * `func` returning its result, if at least `arity` number of arguments have
       * been provided, or returns a function that accepts the remaining `func`
       * arguments, and so on. The arity of `func` may be specified if `func.length`
       * is not sufficient.
       *
       * The `_.curry.placeholder` value, which defaults to `_` in monolithic builds,
       * may be used as a placeholder for provided arguments.
       *
       * **Note:** This method doesn't set the "length" property of curried functions.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Function
       * @param {Function} func The function to curry.
       * @param {number} [arity=func.length] The arity of `func`.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Function} Returns the new curried function.
       * @example
       *
       * var abc = function(a, b, c) {
       *   return [a, b, c];
       * };
       *
       * var curried = _.curry(abc);
       *
       * curried(1)(2)(3);
       * // => [1, 2, 3]
       *
       * curried(1, 2)(3);
       * // => [1, 2, 3]
       *
       * curried(1, 2, 3);
       * // => [1, 2, 3]
       *
       * // Curried with placeholders.
       * curried(1)(_, 3)(2);
       * // => [1, 2, 3]
       */
      function curry(func, arity, guard) {
        arity = guard ? undefined$1 : arity;
        var result = createWrap(func, WRAP_CURRY_FLAG, undefined$1, undefined$1, undefined$1, undefined$1, undefined$1, arity);
        result.placeholder = curry.placeholder;
        return result;
      }

      /**
       * This method is like `_.curry` except that arguments are applied to `func`
       * in the manner of `_.partialRight` instead of `_.partial`.
       *
       * The `_.curryRight.placeholder` value, which defaults to `_` in monolithic
       * builds, may be used as a placeholder for provided arguments.
       *
       * **Note:** This method doesn't set the "length" property of curried functions.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Function
       * @param {Function} func The function to curry.
       * @param {number} [arity=func.length] The arity of `func`.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Function} Returns the new curried function.
       * @example
       *
       * var abc = function(a, b, c) {
       *   return [a, b, c];
       * };
       *
       * var curried = _.curryRight(abc);
       *
       * curried(3)(2)(1);
       * // => [1, 2, 3]
       *
       * curried(2, 3)(1);
       * // => [1, 2, 3]
       *
       * curried(1, 2, 3);
       * // => [1, 2, 3]
       *
       * // Curried with placeholders.
       * curried(3)(1, _)(2);
       * // => [1, 2, 3]
       */
      function curryRight(func, arity, guard) {
        arity = guard ? undefined$1 : arity;
        var result = createWrap(func, WRAP_CURRY_RIGHT_FLAG, undefined$1, undefined$1, undefined$1, undefined$1, undefined$1, arity);
        result.placeholder = curryRight.placeholder;
        return result;
      }

      /**
       * Creates a debounced function that delays invoking `func` until after `wait`
       * milliseconds have elapsed since the last time the debounced function was
       * invoked. The debounced function comes with a `cancel` method to cancel
       * delayed `func` invocations and a `flush` method to immediately invoke them.
       * Provide `options` to indicate whether `func` should be invoked on the
       * leading and/or trailing edge of the `wait` timeout. The `func` is invoked
       * with the last arguments provided to the debounced function. Subsequent
       * calls to the debounced function return the result of the last `func`
       * invocation.
       *
       * **Note:** If `leading` and `trailing` options are `true`, `func` is
       * invoked on the trailing edge of the timeout only if the debounced function
       * is invoked more than once during the `wait` timeout.
       *
       * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
       * until to the next tick, similar to `setTimeout` with a timeout of `0`.
       *
       * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
       * for details over the differences between `_.debounce` and `_.throttle`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to debounce.
       * @param {number} [wait=0] The number of milliseconds to delay.
       * @param {Object} [options={}] The options object.
       * @param {boolean} [options.leading=false]
       *  Specify invoking on the leading edge of the timeout.
       * @param {number} [options.maxWait]
       *  The maximum time `func` is allowed to be delayed before it's invoked.
       * @param {boolean} [options.trailing=true]
       *  Specify invoking on the trailing edge of the timeout.
       * @returns {Function} Returns the new debounced function.
       * @example
       *
       * // Avoid costly calculations while the window size is in flux.
       * jQuery(window).on('resize', _.debounce(calculateLayout, 150));
       *
       * // Invoke `sendMail` when clicked, debouncing subsequent calls.
       * jQuery(element).on('click', _.debounce(sendMail, 300, {
       *   'leading': true,
       *   'trailing': false
       * }));
       *
       * // Ensure `batchLog` is invoked once after 1 second of debounced calls.
       * var debounced = _.debounce(batchLog, 250, { 'maxWait': 1000 });
       * var source = new EventSource('/stream');
       * jQuery(source).on('message', debounced);
       *
       * // Cancel the trailing debounced invocation.
       * jQuery(window).on('popstate', debounced.cancel);
       */
      function debounce(func, wait, options) {
        var lastArgs,
            lastThis,
            maxWait,
            result,
            timerId,
            lastCallTime,
            lastInvokeTime = 0,
            leading = false,
            maxing = false,
            trailing = true;

        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        wait = toNumber(wait) || 0;
        if (isObject(options)) {
          leading = !!options.leading;
          maxing = 'maxWait' in options;
          maxWait = maxing ? nativeMax(toNumber(options.maxWait) || 0, wait) : maxWait;
          trailing = 'trailing' in options ? !!options.trailing : trailing;
        }

        function invokeFunc(time) {
          var args = lastArgs,
              thisArg = lastThis;

          lastArgs = lastThis = undefined$1;
          lastInvokeTime = time;
          result = func.apply(thisArg, args);
          return result;
        }

        function leadingEdge(time) {
          // Reset any `maxWait` timer.
          lastInvokeTime = time;
          // Start the timer for the trailing edge.
          timerId = setTimeout(timerExpired, wait);
          // Invoke the leading edge.
          return leading ? invokeFunc(time) : result;
        }

        function remainingWait(time) {
          var timeSinceLastCall = time - lastCallTime,
              timeSinceLastInvoke = time - lastInvokeTime,
              timeWaiting = wait - timeSinceLastCall;

          return maxing
            ? nativeMin(timeWaiting, maxWait - timeSinceLastInvoke)
            : timeWaiting;
        }

        function shouldInvoke(time) {
          var timeSinceLastCall = time - lastCallTime,
              timeSinceLastInvoke = time - lastInvokeTime;

          // Either this is the first call, activity has stopped and we're at the
          // trailing edge, the system time has gone backwards and we're treating
          // it as the trailing edge, or we've hit the `maxWait` limit.
          return (lastCallTime === undefined$1 || (timeSinceLastCall >= wait) ||
            (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait));
        }

        function timerExpired() {
          var time = now();
          if (shouldInvoke(time)) {
            return trailingEdge(time);
          }
          // Restart the timer.
          timerId = setTimeout(timerExpired, remainingWait(time));
        }

        function trailingEdge(time) {
          timerId = undefined$1;

          // Only invoke if we have `lastArgs` which means `func` has been
          // debounced at least once.
          if (trailing && lastArgs) {
            return invokeFunc(time);
          }
          lastArgs = lastThis = undefined$1;
          return result;
        }

        function cancel() {
          if (timerId !== undefined$1) {
            clearTimeout(timerId);
          }
          lastInvokeTime = 0;
          lastArgs = lastCallTime = lastThis = timerId = undefined$1;
        }

        function flush() {
          return timerId === undefined$1 ? result : trailingEdge(now());
        }

        function debounced() {
          var time = now(),
              isInvoking = shouldInvoke(time);

          lastArgs = arguments;
          lastThis = this;
          lastCallTime = time;

          if (isInvoking) {
            if (timerId === undefined$1) {
              return leadingEdge(lastCallTime);
            }
            if (maxing) {
              // Handle invocations in a tight loop.
              clearTimeout(timerId);
              timerId = setTimeout(timerExpired, wait);
              return invokeFunc(lastCallTime);
            }
          }
          if (timerId === undefined$1) {
            timerId = setTimeout(timerExpired, wait);
          }
          return result;
        }
        debounced.cancel = cancel;
        debounced.flush = flush;
        return debounced;
      }

      /**
       * Defers invoking the `func` until the current call stack has cleared. Any
       * additional arguments are provided to `func` when it's invoked.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to defer.
       * @param {...*} [args] The arguments to invoke `func` with.
       * @returns {number} Returns the timer id.
       * @example
       *
       * _.defer(function(text) {
       *   console.log(text);
       * }, 'deferred');
       * // => Logs 'deferred' after one millisecond.
       */
      var defer = baseRest(function(func, args) {
        return baseDelay(func, 1, args);
      });

      /**
       * Invokes `func` after `wait` milliseconds. Any additional arguments are
       * provided to `func` when it's invoked.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to delay.
       * @param {number} wait The number of milliseconds to delay invocation.
       * @param {...*} [args] The arguments to invoke `func` with.
       * @returns {number} Returns the timer id.
       * @example
       *
       * _.delay(function(text) {
       *   console.log(text);
       * }, 1000, 'later');
       * // => Logs 'later' after one second.
       */
      var delay = baseRest(function(func, wait, args) {
        return baseDelay(func, toNumber(wait) || 0, args);
      });

      /**
       * Creates a function that invokes `func` with arguments reversed.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Function
       * @param {Function} func The function to flip arguments for.
       * @returns {Function} Returns the new flipped function.
       * @example
       *
       * var flipped = _.flip(function() {
       *   return _.toArray(arguments);
       * });
       *
       * flipped('a', 'b', 'c', 'd');
       * // => ['d', 'c', 'b', 'a']
       */
      function flip(func) {
        return createWrap(func, WRAP_FLIP_FLAG);
      }

      /**
       * Creates a function that memoizes the result of `func`. If `resolver` is
       * provided, it determines the cache key for storing the result based on the
       * arguments provided to the memoized function. By default, the first argument
       * provided to the memoized function is used as the map cache key. The `func`
       * is invoked with the `this` binding of the memoized function.
       *
       * **Note:** The cache is exposed as the `cache` property on the memoized
       * function. Its creation may be customized by replacing the `_.memoize.Cache`
       * constructor with one whose instances implement the
       * [`Map`](http://ecma-international.org/ecma-262/7.0/#sec-properties-of-the-map-prototype-object)
       * method interface of `clear`, `delete`, `get`, `has`, and `set`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to have its output memoized.
       * @param {Function} [resolver] The function to resolve the cache key.
       * @returns {Function} Returns the new memoized function.
       * @example
       *
       * var object = { 'a': 1, 'b': 2 };
       * var other = { 'c': 3, 'd': 4 };
       *
       * var values = _.memoize(_.values);
       * values(object);
       * // => [1, 2]
       *
       * values(other);
       * // => [3, 4]
       *
       * object.a = 2;
       * values(object);
       * // => [1, 2]
       *
       * // Modify the result cache.
       * values.cache.set(object, ['a', 'b']);
       * values(object);
       * // => ['a', 'b']
       *
       * // Replace `_.memoize.Cache`.
       * _.memoize.Cache = WeakMap;
       */
      function memoize(func, resolver) {
        if (typeof func != 'function' || (resolver != null && typeof resolver != 'function')) {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        var memoized = function() {
          var args = arguments,
              key = resolver ? resolver.apply(this, args) : args[0],
              cache = memoized.cache;

          if (cache.has(key)) {
            return cache.get(key);
          }
          var result = func.apply(this, args);
          memoized.cache = cache.set(key, result) || cache;
          return result;
        };
        memoized.cache = new (memoize.Cache || MapCache);
        return memoized;
      }

      // Expose `MapCache`.
      memoize.Cache = MapCache;

      /**
       * Creates a function that negates the result of the predicate `func`. The
       * `func` predicate is invoked with the `this` binding and arguments of the
       * created function.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Function
       * @param {Function} predicate The predicate to negate.
       * @returns {Function} Returns the new negated function.
       * @example
       *
       * function isEven(n) {
       *   return n % 2 == 0;
       * }
       *
       * _.filter([1, 2, 3, 4, 5, 6], _.negate(isEven));
       * // => [1, 3, 5]
       */
      function negate(predicate) {
        if (typeof predicate != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        return function() {
          var args = arguments;
          switch (args.length) {
            case 0: return !predicate.call(this);
            case 1: return !predicate.call(this, args[0]);
            case 2: return !predicate.call(this, args[0], args[1]);
            case 3: return !predicate.call(this, args[0], args[1], args[2]);
          }
          return !predicate.apply(this, args);
        };
      }

      /**
       * Creates a function that is restricted to invoking `func` once. Repeat calls
       * to the function return the value of the first invocation. The `func` is
       * invoked with the `this` binding and arguments of the created function.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to restrict.
       * @returns {Function} Returns the new restricted function.
       * @example
       *
       * var initialize = _.once(createApplication);
       * initialize();
       * initialize();
       * // => `createApplication` is invoked once
       */
      function once(func) {
        return before(2, func);
      }

      /**
       * Creates a function that invokes `func` with its arguments transformed.
       *
       * @static
       * @since 4.0.0
       * @memberOf _
       * @category Function
       * @param {Function} func The function to wrap.
       * @param {...(Function|Function[])} [transforms=[_.identity]]
       *  The argument transforms.
       * @returns {Function} Returns the new function.
       * @example
       *
       * function doubled(n) {
       *   return n * 2;
       * }
       *
       * function square(n) {
       *   return n * n;
       * }
       *
       * var func = _.overArgs(function(x, y) {
       *   return [x, y];
       * }, [square, doubled]);
       *
       * func(9, 3);
       * // => [81, 6]
       *
       * func(10, 5);
       * // => [100, 10]
       */
      var overArgs = castRest(function(func, transforms) {
        transforms = (transforms.length == 1 && isArray(transforms[0]))
          ? arrayMap(transforms[0], baseUnary(getIteratee()))
          : arrayMap(baseFlatten(transforms, 1), baseUnary(getIteratee()));

        var funcsLength = transforms.length;
        return baseRest(function(args) {
          var index = -1,
              length = nativeMin(args.length, funcsLength);

          while (++index < length) {
            args[index] = transforms[index].call(this, args[index]);
          }
          return apply(func, this, args);
        });
      });

      /**
       * Creates a function that invokes `func` with `partials` prepended to the
       * arguments it receives. This method is like `_.bind` except it does **not**
       * alter the `this` binding.
       *
       * The `_.partial.placeholder` value, which defaults to `_` in monolithic
       * builds, may be used as a placeholder for partially applied arguments.
       *
       * **Note:** This method doesn't set the "length" property of partially
       * applied functions.
       *
       * @static
       * @memberOf _
       * @since 0.2.0
       * @category Function
       * @param {Function} func The function to partially apply arguments to.
       * @param {...*} [partials] The arguments to be partially applied.
       * @returns {Function} Returns the new partially applied function.
       * @example
       *
       * function greet(greeting, name) {
       *   return greeting + ' ' + name;
       * }
       *
       * var sayHelloTo = _.partial(greet, 'hello');
       * sayHelloTo('fred');
       * // => 'hello fred'
       *
       * // Partially applied with placeholders.
       * var greetFred = _.partial(greet, _, 'fred');
       * greetFred('hi');
       * // => 'hi fred'
       */
      var partial = baseRest(function(func, partials) {
        var holders = replaceHolders(partials, getHolder(partial));
        return createWrap(func, WRAP_PARTIAL_FLAG, undefined$1, partials, holders);
      });

      /**
       * This method is like `_.partial` except that partially applied arguments
       * are appended to the arguments it receives.
       *
       * The `_.partialRight.placeholder` value, which defaults to `_` in monolithic
       * builds, may be used as a placeholder for partially applied arguments.
       *
       * **Note:** This method doesn't set the "length" property of partially
       * applied functions.
       *
       * @static
       * @memberOf _
       * @since 1.0.0
       * @category Function
       * @param {Function} func The function to partially apply arguments to.
       * @param {...*} [partials] The arguments to be partially applied.
       * @returns {Function} Returns the new partially applied function.
       * @example
       *
       * function greet(greeting, name) {
       *   return greeting + ' ' + name;
       * }
       *
       * var greetFred = _.partialRight(greet, 'fred');
       * greetFred('hi');
       * // => 'hi fred'
       *
       * // Partially applied with placeholders.
       * var sayHelloTo = _.partialRight(greet, 'hello', _);
       * sayHelloTo('fred');
       * // => 'hello fred'
       */
      var partialRight = baseRest(function(func, partials) {
        var holders = replaceHolders(partials, getHolder(partialRight));
        return createWrap(func, WRAP_PARTIAL_RIGHT_FLAG, undefined$1, partials, holders);
      });

      /**
       * Creates a function that invokes `func` with arguments arranged according
       * to the specified `indexes` where the argument value at the first index is
       * provided as the first argument, the argument value at the second index is
       * provided as the second argument, and so on.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Function
       * @param {Function} func The function to rearrange arguments for.
       * @param {...(number|number[])} indexes The arranged argument indexes.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var rearged = _.rearg(function(a, b, c) {
       *   return [a, b, c];
       * }, [2, 0, 1]);
       *
       * rearged('b', 'c', 'a')
       * // => ['a', 'b', 'c']
       */
      var rearg = flatRest(function(func, indexes) {
        return createWrap(func, WRAP_REARG_FLAG, undefined$1, undefined$1, undefined$1, indexes);
      });

      /**
       * Creates a function that invokes `func` with the `this` binding of the
       * created function and arguments from `start` and beyond provided as
       * an array.
       *
       * **Note:** This method is based on the
       * [rest parameter](https://mdn.io/rest_parameters).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Function
       * @param {Function} func The function to apply a rest parameter to.
       * @param {number} [start=func.length-1] The start position of the rest parameter.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var say = _.rest(function(what, names) {
       *   return what + ' ' + _.initial(names).join(', ') +
       *     (_.size(names) > 1 ? ', & ' : '') + _.last(names);
       * });
       *
       * say('hello', 'fred', 'barney', 'pebbles');
       * // => 'hello fred, barney, & pebbles'
       */
      function rest(func, start) {
        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        start = start === undefined$1 ? start : toInteger(start);
        return baseRest(func, start);
      }

      /**
       * Creates a function that invokes `func` with the `this` binding of the
       * create function and an array of arguments much like
       * [`Function#apply`](http://www.ecma-international.org/ecma-262/7.0/#sec-function.prototype.apply).
       *
       * **Note:** This method is based on the
       * [spread operator](https://mdn.io/spread_operator).
       *
       * @static
       * @memberOf _
       * @since 3.2.0
       * @category Function
       * @param {Function} func The function to spread arguments over.
       * @param {number} [start=0] The start position of the spread.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var say = _.spread(function(who, what) {
       *   return who + ' says ' + what;
       * });
       *
       * say(['fred', 'hello']);
       * // => 'fred says hello'
       *
       * var numbers = Promise.all([
       *   Promise.resolve(40),
       *   Promise.resolve(36)
       * ]);
       *
       * numbers.then(_.spread(function(x, y) {
       *   return x + y;
       * }));
       * // => a Promise of 76
       */
      function spread(func, start) {
        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        start = start == null ? 0 : nativeMax(toInteger(start), 0);
        return baseRest(function(args) {
          var array = args[start],
              otherArgs = castSlice(args, 0, start);

          if (array) {
            arrayPush(otherArgs, array);
          }
          return apply(func, this, otherArgs);
        });
      }

      /**
       * Creates a throttled function that only invokes `func` at most once per
       * every `wait` milliseconds. The throttled function comes with a `cancel`
       * method to cancel delayed `func` invocations and a `flush` method to
       * immediately invoke them. Provide `options` to indicate whether `func`
       * should be invoked on the leading and/or trailing edge of the `wait`
       * timeout. The `func` is invoked with the last arguments provided to the
       * throttled function. Subsequent calls to the throttled function return the
       * result of the last `func` invocation.
       *
       * **Note:** If `leading` and `trailing` options are `true`, `func` is
       * invoked on the trailing edge of the timeout only if the throttled function
       * is invoked more than once during the `wait` timeout.
       *
       * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
       * until to the next tick, similar to `setTimeout` with a timeout of `0`.
       *
       * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
       * for details over the differences between `_.throttle` and `_.debounce`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {Function} func The function to throttle.
       * @param {number} [wait=0] The number of milliseconds to throttle invocations to.
       * @param {Object} [options={}] The options object.
       * @param {boolean} [options.leading=true]
       *  Specify invoking on the leading edge of the timeout.
       * @param {boolean} [options.trailing=true]
       *  Specify invoking on the trailing edge of the timeout.
       * @returns {Function} Returns the new throttled function.
       * @example
       *
       * // Avoid excessively updating the position while scrolling.
       * jQuery(window).on('scroll', _.throttle(updatePosition, 100));
       *
       * // Invoke `renewToken` when the click event is fired, but not more than once every 5 minutes.
       * var throttled = _.throttle(renewToken, 300000, { 'trailing': false });
       * jQuery(element).on('click', throttled);
       *
       * // Cancel the trailing throttled invocation.
       * jQuery(window).on('popstate', throttled.cancel);
       */
      function throttle(func, wait, options) {
        var leading = true,
            trailing = true;

        if (typeof func != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        if (isObject(options)) {
          leading = 'leading' in options ? !!options.leading : leading;
          trailing = 'trailing' in options ? !!options.trailing : trailing;
        }
        return debounce(func, wait, {
          'leading': leading,
          'maxWait': wait,
          'trailing': trailing
        });
      }

      /**
       * Creates a function that accepts up to one argument, ignoring any
       * additional arguments.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Function
       * @param {Function} func The function to cap arguments for.
       * @returns {Function} Returns the new capped function.
       * @example
       *
       * _.map(['6', '8', '10'], _.unary(parseInt));
       * // => [6, 8, 10]
       */
      function unary(func) {
        return ary(func, 1);
      }

      /**
       * Creates a function that provides `value` to `wrapper` as its first
       * argument. Any additional arguments provided to the function are appended
       * to those provided to the `wrapper`. The wrapper is invoked with the `this`
       * binding of the created function.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Function
       * @param {*} value The value to wrap.
       * @param {Function} [wrapper=identity] The wrapper function.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var p = _.wrap(_.escape, function(func, text) {
       *   return '<p>' + func(text) + '</p>';
       * });
       *
       * p('fred, barney, & pebbles');
       * // => '<p>fred, barney, &amp; pebbles</p>'
       */
      function wrap(value, wrapper) {
        return partial(castFunction(wrapper), value);
      }

      /*------------------------------------------------------------------------*/

      /**
       * Casts `value` as an array if it's not one.
       *
       * @static
       * @memberOf _
       * @since 4.4.0
       * @category Lang
       * @param {*} value The value to inspect.
       * @returns {Array} Returns the cast array.
       * @example
       *
       * _.castArray(1);
       * // => [1]
       *
       * _.castArray({ 'a': 1 });
       * // => [{ 'a': 1 }]
       *
       * _.castArray('abc');
       * // => ['abc']
       *
       * _.castArray(null);
       * // => [null]
       *
       * _.castArray(undefined);
       * // => [undefined]
       *
       * _.castArray();
       * // => []
       *
       * var array = [1, 2, 3];
       * console.log(_.castArray(array) === array);
       * // => true
       */
      function castArray() {
        if (!arguments.length) {
          return [];
        }
        var value = arguments[0];
        return isArray(value) ? value : [value];
      }

      /**
       * Creates a shallow clone of `value`.
       *
       * **Note:** This method is loosely based on the
       * [structured clone algorithm](https://mdn.io/Structured_clone_algorithm)
       * and supports cloning arrays, array buffers, booleans, date objects, maps,
       * numbers, `Object` objects, regexes, sets, strings, symbols, and typed
       * arrays. The own enumerable properties of `arguments` objects are cloned
       * as plain objects. An empty object is returned for uncloneable values such
       * as error objects, functions, DOM nodes, and WeakMaps.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to clone.
       * @returns {*} Returns the cloned value.
       * @see _.cloneDeep
       * @example
       *
       * var objects = [{ 'a': 1 }, { 'b': 2 }];
       *
       * var shallow = _.clone(objects);
       * console.log(shallow[0] === objects[0]);
       * // => true
       */
      function clone(value) {
        return baseClone(value, CLONE_SYMBOLS_FLAG);
      }

      /**
       * This method is like `_.clone` except that it accepts `customizer` which
       * is invoked to produce the cloned value. If `customizer` returns `undefined`,
       * cloning is handled by the method instead. The `customizer` is invoked with
       * up to four arguments; (value [, index|key, object, stack]).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to clone.
       * @param {Function} [customizer] The function to customize cloning.
       * @returns {*} Returns the cloned value.
       * @see _.cloneDeepWith
       * @example
       *
       * function customizer(value) {
       *   if (_.isElement(value)) {
       *     return value.cloneNode(false);
       *   }
       * }
       *
       * var el = _.cloneWith(document.body, customizer);
       *
       * console.log(el === document.body);
       * // => false
       * console.log(el.nodeName);
       * // => 'BODY'
       * console.log(el.childNodes.length);
       * // => 0
       */
      function cloneWith(value, customizer) {
        customizer = typeof customizer == 'function' ? customizer : undefined$1;
        return baseClone(value, CLONE_SYMBOLS_FLAG, customizer);
      }

      /**
       * This method is like `_.clone` except that it recursively clones `value`.
       *
       * @static
       * @memberOf _
       * @since 1.0.0
       * @category Lang
       * @param {*} value The value to recursively clone.
       * @returns {*} Returns the deep cloned value.
       * @see _.clone
       * @example
       *
       * var objects = [{ 'a': 1 }, { 'b': 2 }];
       *
       * var deep = _.cloneDeep(objects);
       * console.log(deep[0] === objects[0]);
       * // => false
       */
      function cloneDeep(value) {
        return baseClone(value, CLONE_DEEP_FLAG | CLONE_SYMBOLS_FLAG);
      }

      /**
       * This method is like `_.cloneWith` except that it recursively clones `value`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to recursively clone.
       * @param {Function} [customizer] The function to customize cloning.
       * @returns {*} Returns the deep cloned value.
       * @see _.cloneWith
       * @example
       *
       * function customizer(value) {
       *   if (_.isElement(value)) {
       *     return value.cloneNode(true);
       *   }
       * }
       *
       * var el = _.cloneDeepWith(document.body, customizer);
       *
       * console.log(el === document.body);
       * // => false
       * console.log(el.nodeName);
       * // => 'BODY'
       * console.log(el.childNodes.length);
       * // => 20
       */
      function cloneDeepWith(value, customizer) {
        customizer = typeof customizer == 'function' ? customizer : undefined$1;
        return baseClone(value, CLONE_DEEP_FLAG | CLONE_SYMBOLS_FLAG, customizer);
      }

      /**
       * Checks if `object` conforms to `source` by invoking the predicate
       * properties of `source` with the corresponding property values of `object`.
       *
       * **Note:** This method is equivalent to `_.conforms` when `source` is
       * partially applied.
       *
       * @static
       * @memberOf _
       * @since 4.14.0
       * @category Lang
       * @param {Object} object The object to inspect.
       * @param {Object} source The object of property predicates to conform to.
       * @returns {boolean} Returns `true` if `object` conforms, else `false`.
       * @example
       *
       * var object = { 'a': 1, 'b': 2 };
       *
       * _.conformsTo(object, { 'b': function(n) { return n > 1; } });
       * // => true
       *
       * _.conformsTo(object, { 'b': function(n) { return n > 2; } });
       * // => false
       */
      function conformsTo(object, source) {
        return source == null || baseConformsTo(object, source, keys(source));
      }

      /**
       * Performs a
       * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
       * comparison between two values to determine if they are equivalent.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
       * @example
       *
       * var object = { 'a': 1 };
       * var other = { 'a': 1 };
       *
       * _.eq(object, object);
       * // => true
       *
       * _.eq(object, other);
       * // => false
       *
       * _.eq('a', 'a');
       * // => true
       *
       * _.eq('a', Object('a'));
       * // => false
       *
       * _.eq(NaN, NaN);
       * // => true
       */
      function eq(value, other) {
        return value === other || (value !== value && other !== other);
      }

      /**
       * Checks if `value` is greater than `other`.
       *
       * @static
       * @memberOf _
       * @since 3.9.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if `value` is greater than `other`,
       *  else `false`.
       * @see _.lt
       * @example
       *
       * _.gt(3, 1);
       * // => true
       *
       * _.gt(3, 3);
       * // => false
       *
       * _.gt(1, 3);
       * // => false
       */
      var gt = createRelationalOperation(baseGt);

      /**
       * Checks if `value` is greater than or equal to `other`.
       *
       * @static
       * @memberOf _
       * @since 3.9.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if `value` is greater than or equal to
       *  `other`, else `false`.
       * @see _.lte
       * @example
       *
       * _.gte(3, 1);
       * // => true
       *
       * _.gte(3, 3);
       * // => true
       *
       * _.gte(1, 3);
       * // => false
       */
      var gte = createRelationalOperation(function(value, other) {
        return value >= other;
      });

      /**
       * Checks if `value` is likely an `arguments` object.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an `arguments` object,
       *  else `false`.
       * @example
       *
       * _.isArguments(function() { return arguments; }());
       * // => true
       *
       * _.isArguments([1, 2, 3]);
       * // => false
       */
      var isArguments = baseIsArguments(function() { return arguments; }()) ? baseIsArguments : function(value) {
        return isObjectLike(value) && hasOwnProperty.call(value, 'callee') &&
          !propertyIsEnumerable.call(value, 'callee');
      };

      /**
       * Checks if `value` is classified as an `Array` object.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an array, else `false`.
       * @example
       *
       * _.isArray([1, 2, 3]);
       * // => true
       *
       * _.isArray(document.body.children);
       * // => false
       *
       * _.isArray('abc');
       * // => false
       *
       * _.isArray(_.noop);
       * // => false
       */
      var isArray = Array.isArray;

      /**
       * Checks if `value` is classified as an `ArrayBuffer` object.
       *
       * @static
       * @memberOf _
       * @since 4.3.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an array buffer, else `false`.
       * @example
       *
       * _.isArrayBuffer(new ArrayBuffer(2));
       * // => true
       *
       * _.isArrayBuffer(new Array(2));
       * // => false
       */
      var isArrayBuffer = nodeIsArrayBuffer ? baseUnary(nodeIsArrayBuffer) : baseIsArrayBuffer;

      /**
       * Checks if `value` is array-like. A value is considered array-like if it's
       * not a function and has a `value.length` that's an integer greater than or
       * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
       * @example
       *
       * _.isArrayLike([1, 2, 3]);
       * // => true
       *
       * _.isArrayLike(document.body.children);
       * // => true
       *
       * _.isArrayLike('abc');
       * // => true
       *
       * _.isArrayLike(_.noop);
       * // => false
       */
      function isArrayLike(value) {
        return value != null && isLength(value.length) && !isFunction(value);
      }

      /**
       * This method is like `_.isArrayLike` except that it also checks if `value`
       * is an object.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an array-like object,
       *  else `false`.
       * @example
       *
       * _.isArrayLikeObject([1, 2, 3]);
       * // => true
       *
       * _.isArrayLikeObject(document.body.children);
       * // => true
       *
       * _.isArrayLikeObject('abc');
       * // => false
       *
       * _.isArrayLikeObject(_.noop);
       * // => false
       */
      function isArrayLikeObject(value) {
        return isObjectLike(value) && isArrayLike(value);
      }

      /**
       * Checks if `value` is classified as a boolean primitive or object.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a boolean, else `false`.
       * @example
       *
       * _.isBoolean(false);
       * // => true
       *
       * _.isBoolean(null);
       * // => false
       */
      function isBoolean(value) {
        return value === true || value === false ||
          (isObjectLike(value) && baseGetTag(value) == boolTag);
      }

      /**
       * Checks if `value` is a buffer.
       *
       * @static
       * @memberOf _
       * @since 4.3.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
       * @example
       *
       * _.isBuffer(new Buffer(2));
       * // => true
       *
       * _.isBuffer(new Uint8Array(2));
       * // => false
       */
      var isBuffer = nativeIsBuffer || stubFalse;

      /**
       * Checks if `value` is classified as a `Date` object.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a date object, else `false`.
       * @example
       *
       * _.isDate(new Date);
       * // => true
       *
       * _.isDate('Mon April 23 2012');
       * // => false
       */
      var isDate = nodeIsDate ? baseUnary(nodeIsDate) : baseIsDate;

      /**
       * Checks if `value` is likely a DOM element.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a DOM element, else `false`.
       * @example
       *
       * _.isElement(document.body);
       * // => true
       *
       * _.isElement('<body>');
       * // => false
       */
      function isElement(value) {
        return isObjectLike(value) && value.nodeType === 1 && !isPlainObject(value);
      }

      /**
       * Checks if `value` is an empty object, collection, map, or set.
       *
       * Objects are considered empty if they have no own enumerable string keyed
       * properties.
       *
       * Array-like values such as `arguments` objects, arrays, buffers, strings, or
       * jQuery-like collections are considered empty if they have a `length` of `0`.
       * Similarly, maps and sets are considered empty if they have a `size` of `0`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is empty, else `false`.
       * @example
       *
       * _.isEmpty(null);
       * // => true
       *
       * _.isEmpty(true);
       * // => true
       *
       * _.isEmpty(1);
       * // => true
       *
       * _.isEmpty([1, 2, 3]);
       * // => false
       *
       * _.isEmpty({ 'a': 1 });
       * // => false
       */
      function isEmpty(value) {
        if (value == null) {
          return true;
        }
        if (isArrayLike(value) &&
            (isArray(value) || typeof value == 'string' || typeof value.splice == 'function' ||
              isBuffer(value) || isTypedArray(value) || isArguments(value))) {
          return !value.length;
        }
        var tag = getTag(value);
        if (tag == mapTag || tag == setTag) {
          return !value.size;
        }
        if (isPrototype(value)) {
          return !baseKeys(value).length;
        }
        for (var key in value) {
          if (hasOwnProperty.call(value, key)) {
            return false;
          }
        }
        return true;
      }

      /**
       * Performs a deep comparison between two values to determine if they are
       * equivalent.
       *
       * **Note:** This method supports comparing arrays, array buffers, booleans,
       * date objects, error objects, maps, numbers, `Object` objects, regexes,
       * sets, strings, symbols, and typed arrays. `Object` objects are compared
       * by their own, not inherited, enumerable properties. Functions and DOM
       * nodes are compared by strict equality, i.e. `===`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
       * @example
       *
       * var object = { 'a': 1 };
       * var other = { 'a': 1 };
       *
       * _.isEqual(object, other);
       * // => true
       *
       * object === other;
       * // => false
       */
      function isEqual(value, other) {
        return baseIsEqual(value, other);
      }

      /**
       * This method is like `_.isEqual` except that it accepts `customizer` which
       * is invoked to compare values. If `customizer` returns `undefined`, comparisons
       * are handled by the method instead. The `customizer` is invoked with up to
       * six arguments: (objValue, othValue [, index|key, object, other, stack]).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @param {Function} [customizer] The function to customize comparisons.
       * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
       * @example
       *
       * function isGreeting(value) {
       *   return /^h(?:i|ello)$/.test(value);
       * }
       *
       * function customizer(objValue, othValue) {
       *   if (isGreeting(objValue) && isGreeting(othValue)) {
       *     return true;
       *   }
       * }
       *
       * var array = ['hello', 'goodbye'];
       * var other = ['hi', 'goodbye'];
       *
       * _.isEqualWith(array, other, customizer);
       * // => true
       */
      function isEqualWith(value, other, customizer) {
        customizer = typeof customizer == 'function' ? customizer : undefined$1;
        var result = customizer ? customizer(value, other) : undefined$1;
        return result === undefined$1 ? baseIsEqual(value, other, undefined$1, customizer) : !!result;
      }

      /**
       * Checks if `value` is an `Error`, `EvalError`, `RangeError`, `ReferenceError`,
       * `SyntaxError`, `TypeError`, or `URIError` object.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an error object, else `false`.
       * @example
       *
       * _.isError(new Error);
       * // => true
       *
       * _.isError(Error);
       * // => false
       */
      function isError(value) {
        if (!isObjectLike(value)) {
          return false;
        }
        var tag = baseGetTag(value);
        return tag == errorTag || tag == domExcTag ||
          (typeof value.message == 'string' && typeof value.name == 'string' && !isPlainObject(value));
      }

      /**
       * Checks if `value` is a finite primitive number.
       *
       * **Note:** This method is based on
       * [`Number.isFinite`](https://mdn.io/Number/isFinite).
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a finite number, else `false`.
       * @example
       *
       * _.isFinite(3);
       * // => true
       *
       * _.isFinite(Number.MIN_VALUE);
       * // => true
       *
       * _.isFinite(Infinity);
       * // => false
       *
       * _.isFinite('3');
       * // => false
       */
      function isFinite(value) {
        return typeof value == 'number' && nativeIsFinite(value);
      }

      /**
       * Checks if `value` is classified as a `Function` object.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a function, else `false`.
       * @example
       *
       * _.isFunction(_);
       * // => true
       *
       * _.isFunction(/abc/);
       * // => false
       */
      function isFunction(value) {
        if (!isObject(value)) {
          return false;
        }
        // The use of `Object#toString` avoids issues with the `typeof` operator
        // in Safari 9 which returns 'object' for typed arrays and other constructors.
        var tag = baseGetTag(value);
        return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
      }

      /**
       * Checks if `value` is an integer.
       *
       * **Note:** This method is based on
       * [`Number.isInteger`](https://mdn.io/Number/isInteger).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an integer, else `false`.
       * @example
       *
       * _.isInteger(3);
       * // => true
       *
       * _.isInteger(Number.MIN_VALUE);
       * // => false
       *
       * _.isInteger(Infinity);
       * // => false
       *
       * _.isInteger('3');
       * // => false
       */
      function isInteger(value) {
        return typeof value == 'number' && value == toInteger(value);
      }

      /**
       * Checks if `value` is a valid array-like length.
       *
       * **Note:** This method is loosely based on
       * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
       * @example
       *
       * _.isLength(3);
       * // => true
       *
       * _.isLength(Number.MIN_VALUE);
       * // => false
       *
       * _.isLength(Infinity);
       * // => false
       *
       * _.isLength('3');
       * // => false
       */
      function isLength(value) {
        return typeof value == 'number' &&
          value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
      }

      /**
       * Checks if `value` is the
       * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
       * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is an object, else `false`.
       * @example
       *
       * _.isObject({});
       * // => true
       *
       * _.isObject([1, 2, 3]);
       * // => true
       *
       * _.isObject(_.noop);
       * // => true
       *
       * _.isObject(null);
       * // => false
       */
      function isObject(value) {
        var type = typeof value;
        return value != null && (type == 'object' || type == 'function');
      }

      /**
       * Checks if `value` is object-like. A value is object-like if it's not `null`
       * and has a `typeof` result of "object".
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
       * @example
       *
       * _.isObjectLike({});
       * // => true
       *
       * _.isObjectLike([1, 2, 3]);
       * // => true
       *
       * _.isObjectLike(_.noop);
       * // => false
       *
       * _.isObjectLike(null);
       * // => false
       */
      function isObjectLike(value) {
        return value != null && typeof value == 'object';
      }

      /**
       * Checks if `value` is classified as a `Map` object.
       *
       * @static
       * @memberOf _
       * @since 4.3.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a map, else `false`.
       * @example
       *
       * _.isMap(new Map);
       * // => true
       *
       * _.isMap(new WeakMap);
       * // => false
       */
      var isMap = nodeIsMap ? baseUnary(nodeIsMap) : baseIsMap;

      /**
       * Performs a partial deep comparison between `object` and `source` to
       * determine if `object` contains equivalent property values.
       *
       * **Note:** This method is equivalent to `_.matches` when `source` is
       * partially applied.
       *
       * Partial comparisons will match empty array and empty object `source`
       * values against any array or object value, respectively. See `_.isEqual`
       * for a list of supported value comparisons.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Lang
       * @param {Object} object The object to inspect.
       * @param {Object} source The object of property values to match.
       * @returns {boolean} Returns `true` if `object` is a match, else `false`.
       * @example
       *
       * var object = { 'a': 1, 'b': 2 };
       *
       * _.isMatch(object, { 'b': 2 });
       * // => true
       *
       * _.isMatch(object, { 'b': 1 });
       * // => false
       */
      function isMatch(object, source) {
        return object === source || baseIsMatch(object, source, getMatchData(source));
      }

      /**
       * This method is like `_.isMatch` except that it accepts `customizer` which
       * is invoked to compare values. If `customizer` returns `undefined`, comparisons
       * are handled by the method instead. The `customizer` is invoked with five
       * arguments: (objValue, srcValue, index|key, object, source).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {Object} object The object to inspect.
       * @param {Object} source The object of property values to match.
       * @param {Function} [customizer] The function to customize comparisons.
       * @returns {boolean} Returns `true` if `object` is a match, else `false`.
       * @example
       *
       * function isGreeting(value) {
       *   return /^h(?:i|ello)$/.test(value);
       * }
       *
       * function customizer(objValue, srcValue) {
       *   if (isGreeting(objValue) && isGreeting(srcValue)) {
       *     return true;
       *   }
       * }
       *
       * var object = { 'greeting': 'hello' };
       * var source = { 'greeting': 'hi' };
       *
       * _.isMatchWith(object, source, customizer);
       * // => true
       */
      function isMatchWith(object, source, customizer) {
        customizer = typeof customizer == 'function' ? customizer : undefined$1;
        return baseIsMatch(object, source, getMatchData(source), customizer);
      }

      /**
       * Checks if `value` is `NaN`.
       *
       * **Note:** This method is based on
       * [`Number.isNaN`](https://mdn.io/Number/isNaN) and is not the same as
       * global [`isNaN`](https://mdn.io/isNaN) which returns `true` for
       * `undefined` and other non-number values.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is `NaN`, else `false`.
       * @example
       *
       * _.isNaN(NaN);
       * // => true
       *
       * _.isNaN(new Number(NaN));
       * // => true
       *
       * isNaN(undefined);
       * // => true
       *
       * _.isNaN(undefined);
       * // => false
       */
      function isNaN(value) {
        // An `NaN` primitive is the only value that is not equal to itself.
        // Perform the `toStringTag` check first to avoid errors with some
        // ActiveX objects in IE.
        return isNumber(value) && value != +value;
      }

      /**
       * Checks if `value` is a pristine native function.
       *
       * **Note:** This method can't reliably detect native functions in the presence
       * of the core-js package because core-js circumvents this kind of detection.
       * Despite multiple requests, the core-js maintainer has made it clear: any
       * attempt to fix the detection will be obstructed. As a result, we're left
       * with little choice but to throw an error. Unfortunately, this also affects
       * packages, like [babel-polyfill](https://www.npmjs.com/package/babel-polyfill),
       * which rely on core-js.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a native function,
       *  else `false`.
       * @example
       *
       * _.isNative(Array.prototype.push);
       * // => true
       *
       * _.isNative(_);
       * // => false
       */
      function isNative(value) {
        if (isMaskable(value)) {
          throw new Error(CORE_ERROR_TEXT);
        }
        return baseIsNative(value);
      }

      /**
       * Checks if `value` is `null`.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is `null`, else `false`.
       * @example
       *
       * _.isNull(null);
       * // => true
       *
       * _.isNull(void 0);
       * // => false
       */
      function isNull(value) {
        return value === null;
      }

      /**
       * Checks if `value` is `null` or `undefined`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is nullish, else `false`.
       * @example
       *
       * _.isNil(null);
       * // => true
       *
       * _.isNil(void 0);
       * // => true
       *
       * _.isNil(NaN);
       * // => false
       */
      function isNil(value) {
        return value == null;
      }

      /**
       * Checks if `value` is classified as a `Number` primitive or object.
       *
       * **Note:** To exclude `Infinity`, `-Infinity`, and `NaN`, which are
       * classified as numbers, use the `_.isFinite` method.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a number, else `false`.
       * @example
       *
       * _.isNumber(3);
       * // => true
       *
       * _.isNumber(Number.MIN_VALUE);
       * // => true
       *
       * _.isNumber(Infinity);
       * // => true
       *
       * _.isNumber('3');
       * // => false
       */
      function isNumber(value) {
        return typeof value == 'number' ||
          (isObjectLike(value) && baseGetTag(value) == numberTag);
      }

      /**
       * Checks if `value` is a plain object, that is, an object created by the
       * `Object` constructor or one with a `[[Prototype]]` of `null`.
       *
       * @static
       * @memberOf _
       * @since 0.8.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a plain object, else `false`.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       * }
       *
       * _.isPlainObject(new Foo);
       * // => false
       *
       * _.isPlainObject([1, 2, 3]);
       * // => false
       *
       * _.isPlainObject({ 'x': 0, 'y': 0 });
       * // => true
       *
       * _.isPlainObject(Object.create(null));
       * // => true
       */
      function isPlainObject(value) {
        if (!isObjectLike(value) || baseGetTag(value) != objectTag) {
          return false;
        }
        var proto = getPrototype(value);
        if (proto === null) {
          return true;
        }
        var Ctor = hasOwnProperty.call(proto, 'constructor') && proto.constructor;
        return typeof Ctor == 'function' && Ctor instanceof Ctor &&
          funcToString.call(Ctor) == objectCtorString;
      }

      /**
       * Checks if `value` is classified as a `RegExp` object.
       *
       * @static
       * @memberOf _
       * @since 0.1.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a regexp, else `false`.
       * @example
       *
       * _.isRegExp(/abc/);
       * // => true
       *
       * _.isRegExp('/abc/');
       * // => false
       */
      var isRegExp = nodeIsRegExp ? baseUnary(nodeIsRegExp) : baseIsRegExp;

      /**
       * Checks if `value` is a safe integer. An integer is safe if it's an IEEE-754
       * double precision number which isn't the result of a rounded unsafe integer.
       *
       * **Note:** This method is based on
       * [`Number.isSafeInteger`](https://mdn.io/Number/isSafeInteger).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a safe integer, else `false`.
       * @example
       *
       * _.isSafeInteger(3);
       * // => true
       *
       * _.isSafeInteger(Number.MIN_VALUE);
       * // => false
       *
       * _.isSafeInteger(Infinity);
       * // => false
       *
       * _.isSafeInteger('3');
       * // => false
       */
      function isSafeInteger(value) {
        return isInteger(value) && value >= -MAX_SAFE_INTEGER && value <= MAX_SAFE_INTEGER;
      }

      /**
       * Checks if `value` is classified as a `Set` object.
       *
       * @static
       * @memberOf _
       * @since 4.3.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a set, else `false`.
       * @example
       *
       * _.isSet(new Set);
       * // => true
       *
       * _.isSet(new WeakSet);
       * // => false
       */
      var isSet = nodeIsSet ? baseUnary(nodeIsSet) : baseIsSet;

      /**
       * Checks if `value` is classified as a `String` primitive or object.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a string, else `false`.
       * @example
       *
       * _.isString('abc');
       * // => true
       *
       * _.isString(1);
       * // => false
       */
      function isString(value) {
        return typeof value == 'string' ||
          (!isArray(value) && isObjectLike(value) && baseGetTag(value) == stringTag);
      }

      /**
       * Checks if `value` is classified as a `Symbol` primitive or object.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
       * @example
       *
       * _.isSymbol(Symbol.iterator);
       * // => true
       *
       * _.isSymbol('abc');
       * // => false
       */
      function isSymbol(value) {
        return typeof value == 'symbol' ||
          (isObjectLike(value) && baseGetTag(value) == symbolTag);
      }

      /**
       * Checks if `value` is classified as a typed array.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
       * @example
       *
       * _.isTypedArray(new Uint8Array);
       * // => true
       *
       * _.isTypedArray([]);
       * // => false
       */
      var isTypedArray = nodeIsTypedArray ? baseUnary(nodeIsTypedArray) : baseIsTypedArray;

      /**
       * Checks if `value` is `undefined`.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is `undefined`, else `false`.
       * @example
       *
       * _.isUndefined(void 0);
       * // => true
       *
       * _.isUndefined(null);
       * // => false
       */
      function isUndefined(value) {
        return value === undefined$1;
      }

      /**
       * Checks if `value` is classified as a `WeakMap` object.
       *
       * @static
       * @memberOf _
       * @since 4.3.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a weak map, else `false`.
       * @example
       *
       * _.isWeakMap(new WeakMap);
       * // => true
       *
       * _.isWeakMap(new Map);
       * // => false
       */
      function isWeakMap(value) {
        return isObjectLike(value) && getTag(value) == weakMapTag;
      }

      /**
       * Checks if `value` is classified as a `WeakSet` object.
       *
       * @static
       * @memberOf _
       * @since 4.3.0
       * @category Lang
       * @param {*} value The value to check.
       * @returns {boolean} Returns `true` if `value` is a weak set, else `false`.
       * @example
       *
       * _.isWeakSet(new WeakSet);
       * // => true
       *
       * _.isWeakSet(new Set);
       * // => false
       */
      function isWeakSet(value) {
        return isObjectLike(value) && baseGetTag(value) == weakSetTag;
      }

      /**
       * Checks if `value` is less than `other`.
       *
       * @static
       * @memberOf _
       * @since 3.9.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if `value` is less than `other`,
       *  else `false`.
       * @see _.gt
       * @example
       *
       * _.lt(1, 3);
       * // => true
       *
       * _.lt(3, 3);
       * // => false
       *
       * _.lt(3, 1);
       * // => false
       */
      var lt = createRelationalOperation(baseLt);

      /**
       * Checks if `value` is less than or equal to `other`.
       *
       * @static
       * @memberOf _
       * @since 3.9.0
       * @category Lang
       * @param {*} value The value to compare.
       * @param {*} other The other value to compare.
       * @returns {boolean} Returns `true` if `value` is less than or equal to
       *  `other`, else `false`.
       * @see _.gte
       * @example
       *
       * _.lte(1, 3);
       * // => true
       *
       * _.lte(3, 3);
       * // => true
       *
       * _.lte(3, 1);
       * // => false
       */
      var lte = createRelationalOperation(function(value, other) {
        return value <= other;
      });

      /**
       * Converts `value` to an array.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {Array} Returns the converted array.
       * @example
       *
       * _.toArray({ 'a': 1, 'b': 2 });
       * // => [1, 2]
       *
       * _.toArray('abc');
       * // => ['a', 'b', 'c']
       *
       * _.toArray(1);
       * // => []
       *
       * _.toArray(null);
       * // => []
       */
      function toArray(value) {
        if (!value) {
          return [];
        }
        if (isArrayLike(value)) {
          return isString(value) ? stringToArray(value) : copyArray(value);
        }
        if (symIterator && value[symIterator]) {
          return iteratorToArray(value[symIterator]());
        }
        var tag = getTag(value),
            func = tag == mapTag ? mapToArray : (tag == setTag ? setToArray : values);

        return func(value);
      }

      /**
       * Converts `value` to a finite number.
       *
       * @static
       * @memberOf _
       * @since 4.12.0
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {number} Returns the converted number.
       * @example
       *
       * _.toFinite(3.2);
       * // => 3.2
       *
       * _.toFinite(Number.MIN_VALUE);
       * // => 5e-324
       *
       * _.toFinite(Infinity);
       * // => 1.7976931348623157e+308
       *
       * _.toFinite('3.2');
       * // => 3.2
       */
      function toFinite(value) {
        if (!value) {
          return value === 0 ? value : 0;
        }
        value = toNumber(value);
        if (value === INFINITY || value === -INFINITY) {
          var sign = (value < 0 ? -1 : 1);
          return sign * MAX_INTEGER;
        }
        return value === value ? value : 0;
      }

      /**
       * Converts `value` to an integer.
       *
       * **Note:** This method is loosely based on
       * [`ToInteger`](http://www.ecma-international.org/ecma-262/7.0/#sec-tointeger).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {number} Returns the converted integer.
       * @example
       *
       * _.toInteger(3.2);
       * // => 3
       *
       * _.toInteger(Number.MIN_VALUE);
       * // => 0
       *
       * _.toInteger(Infinity);
       * // => 1.7976931348623157e+308
       *
       * _.toInteger('3.2');
       * // => 3
       */
      function toInteger(value) {
        var result = toFinite(value),
            remainder = result % 1;

        return result === result ? (remainder ? result - remainder : result) : 0;
      }

      /**
       * Converts `value` to an integer suitable for use as the length of an
       * array-like object.
       *
       * **Note:** This method is based on
       * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {number} Returns the converted integer.
       * @example
       *
       * _.toLength(3.2);
       * // => 3
       *
       * _.toLength(Number.MIN_VALUE);
       * // => 0
       *
       * _.toLength(Infinity);
       * // => 4294967295
       *
       * _.toLength('3.2');
       * // => 3
       */
      function toLength(value) {
        return value ? baseClamp(toInteger(value), 0, MAX_ARRAY_LENGTH) : 0;
      }

      /**
       * Converts `value` to a number.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to process.
       * @returns {number} Returns the number.
       * @example
       *
       * _.toNumber(3.2);
       * // => 3.2
       *
       * _.toNumber(Number.MIN_VALUE);
       * // => 5e-324
       *
       * _.toNumber(Infinity);
       * // => Infinity
       *
       * _.toNumber('3.2');
       * // => 3.2
       */
      function toNumber(value) {
        if (typeof value == 'number') {
          return value;
        }
        if (isSymbol(value)) {
          return NAN;
        }
        if (isObject(value)) {
          var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
          value = isObject(other) ? (other + '') : other;
        }
        if (typeof value != 'string') {
          return value === 0 ? value : +value;
        }
        value = value.replace(reTrim, '');
        var isBinary = reIsBinary.test(value);
        return (isBinary || reIsOctal.test(value))
          ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
          : (reIsBadHex.test(value) ? NAN : +value);
      }

      /**
       * Converts `value` to a plain object flattening inherited enumerable string
       * keyed properties of `value` to own properties of the plain object.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {Object} Returns the converted plain object.
       * @example
       *
       * function Foo() {
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.assign({ 'a': 1 }, new Foo);
       * // => { 'a': 1, 'b': 2 }
       *
       * _.assign({ 'a': 1 }, _.toPlainObject(new Foo));
       * // => { 'a': 1, 'b': 2, 'c': 3 }
       */
      function toPlainObject(value) {
        return copyObject(value, keysIn(value));
      }

      /**
       * Converts `value` to a safe integer. A safe integer can be compared and
       * represented correctly.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {number} Returns the converted integer.
       * @example
       *
       * _.toSafeInteger(3.2);
       * // => 3
       *
       * _.toSafeInteger(Number.MIN_VALUE);
       * // => 0
       *
       * _.toSafeInteger(Infinity);
       * // => 9007199254740991
       *
       * _.toSafeInteger('3.2');
       * // => 3
       */
      function toSafeInteger(value) {
        return value
          ? baseClamp(toInteger(value), -MAX_SAFE_INTEGER, MAX_SAFE_INTEGER)
          : (value === 0 ? value : 0);
      }

      /**
       * Converts `value` to a string. An empty string is returned for `null`
       * and `undefined` values. The sign of `-0` is preserved.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Lang
       * @param {*} value The value to convert.
       * @returns {string} Returns the converted string.
       * @example
       *
       * _.toString(null);
       * // => ''
       *
       * _.toString(-0);
       * // => '-0'
       *
       * _.toString([1, 2, 3]);
       * // => '1,2,3'
       */
      function toString(value) {
        return value == null ? '' : baseToString(value);
      }

      /*------------------------------------------------------------------------*/

      /**
       * Assigns own enumerable string keyed properties of source objects to the
       * destination object. Source objects are applied from left to right.
       * Subsequent sources overwrite property assignments of previous sources.
       *
       * **Note:** This method mutates `object` and is loosely based on
       * [`Object.assign`](https://mdn.io/Object/assign).
       *
       * @static
       * @memberOf _
       * @since 0.10.0
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} [sources] The source objects.
       * @returns {Object} Returns `object`.
       * @see _.assignIn
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       * }
       *
       * function Bar() {
       *   this.c = 3;
       * }
       *
       * Foo.prototype.b = 2;
       * Bar.prototype.d = 4;
       *
       * _.assign({ 'a': 0 }, new Foo, new Bar);
       * // => { 'a': 1, 'c': 3 }
       */
      var assign = createAssigner(function(object, source) {
        if (isPrototype(source) || isArrayLike(source)) {
          copyObject(source, keys(source), object);
          return;
        }
        for (var key in source) {
          if (hasOwnProperty.call(source, key)) {
            assignValue(object, key, source[key]);
          }
        }
      });

      /**
       * This method is like `_.assign` except that it iterates over own and
       * inherited source properties.
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @alias extend
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} [sources] The source objects.
       * @returns {Object} Returns `object`.
       * @see _.assign
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       * }
       *
       * function Bar() {
       *   this.c = 3;
       * }
       *
       * Foo.prototype.b = 2;
       * Bar.prototype.d = 4;
       *
       * _.assignIn({ 'a': 0 }, new Foo, new Bar);
       * // => { 'a': 1, 'b': 2, 'c': 3, 'd': 4 }
       */
      var assignIn = createAssigner(function(object, source) {
        copyObject(source, keysIn(source), object);
      });

      /**
       * This method is like `_.assignIn` except that it accepts `customizer`
       * which is invoked to produce the assigned values. If `customizer` returns
       * `undefined`, assignment is handled by the method instead. The `customizer`
       * is invoked with five arguments: (objValue, srcValue, key, object, source).
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @alias extendWith
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} sources The source objects.
       * @param {Function} [customizer] The function to customize assigned values.
       * @returns {Object} Returns `object`.
       * @see _.assignWith
       * @example
       *
       * function customizer(objValue, srcValue) {
       *   return _.isUndefined(objValue) ? srcValue : objValue;
       * }
       *
       * var defaults = _.partialRight(_.assignInWith, customizer);
       *
       * defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
       * // => { 'a': 1, 'b': 2 }
       */
      var assignInWith = createAssigner(function(object, source, srcIndex, customizer) {
        copyObject(source, keysIn(source), object, customizer);
      });

      /**
       * This method is like `_.assign` except that it accepts `customizer`
       * which is invoked to produce the assigned values. If `customizer` returns
       * `undefined`, assignment is handled by the method instead. The `customizer`
       * is invoked with five arguments: (objValue, srcValue, key, object, source).
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} sources The source objects.
       * @param {Function} [customizer] The function to customize assigned values.
       * @returns {Object} Returns `object`.
       * @see _.assignInWith
       * @example
       *
       * function customizer(objValue, srcValue) {
       *   return _.isUndefined(objValue) ? srcValue : objValue;
       * }
       *
       * var defaults = _.partialRight(_.assignWith, customizer);
       *
       * defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
       * // => { 'a': 1, 'b': 2 }
       */
      var assignWith = createAssigner(function(object, source, srcIndex, customizer) {
        copyObject(source, keys(source), object, customizer);
      });

      /**
       * Creates an array of values corresponding to `paths` of `object`.
       *
       * @static
       * @memberOf _
       * @since 1.0.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {...(string|string[])} [paths] The property paths to pick.
       * @returns {Array} Returns the picked values.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': 3 } }, 4] };
       *
       * _.at(object, ['a[0].b.c', 'a[1]']);
       * // => [3, 4]
       */
      var at = flatRest(baseAt);

      /**
       * Creates an object that inherits from the `prototype` object. If a
       * `properties` object is given, its own enumerable string keyed properties
       * are assigned to the created object.
       *
       * @static
       * @memberOf _
       * @since 2.3.0
       * @category Object
       * @param {Object} prototype The object to inherit from.
       * @param {Object} [properties] The properties to assign to the object.
       * @returns {Object} Returns the new object.
       * @example
       *
       * function Shape() {
       *   this.x = 0;
       *   this.y = 0;
       * }
       *
       * function Circle() {
       *   Shape.call(this);
       * }
       *
       * Circle.prototype = _.create(Shape.prototype, {
       *   'constructor': Circle
       * });
       *
       * var circle = new Circle;
       * circle instanceof Circle;
       * // => true
       *
       * circle instanceof Shape;
       * // => true
       */
      function create(prototype, properties) {
        var result = baseCreate(prototype);
        return properties == null ? result : baseAssign(result, properties);
      }

      /**
       * Assigns own and inherited enumerable string keyed properties of source
       * objects to the destination object for all destination properties that
       * resolve to `undefined`. Source objects are applied from left to right.
       * Once a property is set, additional values of the same property are ignored.
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} [sources] The source objects.
       * @returns {Object} Returns `object`.
       * @see _.defaultsDeep
       * @example
       *
       * _.defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
       * // => { 'a': 1, 'b': 2 }
       */
      var defaults = baseRest(function(object, sources) {
        object = Object(object);

        var index = -1;
        var length = sources.length;
        var guard = length > 2 ? sources[2] : undefined$1;

        if (guard && isIterateeCall(sources[0], sources[1], guard)) {
          length = 1;
        }

        while (++index < length) {
          var source = sources[index];
          var props = keysIn(source);
          var propsIndex = -1;
          var propsLength = props.length;

          while (++propsIndex < propsLength) {
            var key = props[propsIndex];
            var value = object[key];

            if (value === undefined$1 ||
                (eq(value, objectProto[key]) && !hasOwnProperty.call(object, key))) {
              object[key] = source[key];
            }
          }
        }

        return object;
      });

      /**
       * This method is like `_.defaults` except that it recursively assigns
       * default properties.
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 3.10.0
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} [sources] The source objects.
       * @returns {Object} Returns `object`.
       * @see _.defaults
       * @example
       *
       * _.defaultsDeep({ 'a': { 'b': 2 } }, { 'a': { 'b': 1, 'c': 3 } });
       * // => { 'a': { 'b': 2, 'c': 3 } }
       */
      var defaultsDeep = baseRest(function(args) {
        args.push(undefined$1, customDefaultsMerge);
        return apply(mergeWith, undefined$1, args);
      });

      /**
       * This method is like `_.find` except that it returns the key of the first
       * element `predicate` returns truthy for instead of the element itself.
       *
       * @static
       * @memberOf _
       * @since 1.1.0
       * @category Object
       * @param {Object} object The object to inspect.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {string|undefined} Returns the key of the matched element,
       *  else `undefined`.
       * @example
       *
       * var users = {
       *   'barney':  { 'age': 36, 'active': true },
       *   'fred':    { 'age': 40, 'active': false },
       *   'pebbles': { 'age': 1,  'active': true }
       * };
       *
       * _.findKey(users, function(o) { return o.age < 40; });
       * // => 'barney' (iteration order is not guaranteed)
       *
       * // The `_.matches` iteratee shorthand.
       * _.findKey(users, { 'age': 1, 'active': true });
       * // => 'pebbles'
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.findKey(users, ['active', false]);
       * // => 'fred'
       *
       * // The `_.property` iteratee shorthand.
       * _.findKey(users, 'active');
       * // => 'barney'
       */
      function findKey(object, predicate) {
        return baseFindKey(object, getIteratee(predicate, 3), baseForOwn);
      }

      /**
       * This method is like `_.findKey` except that it iterates over elements of
       * a collection in the opposite order.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Object
       * @param {Object} object The object to inspect.
       * @param {Function} [predicate=_.identity] The function invoked per iteration.
       * @returns {string|undefined} Returns the key of the matched element,
       *  else `undefined`.
       * @example
       *
       * var users = {
       *   'barney':  { 'age': 36, 'active': true },
       *   'fred':    { 'age': 40, 'active': false },
       *   'pebbles': { 'age': 1,  'active': true }
       * };
       *
       * _.findLastKey(users, function(o) { return o.age < 40; });
       * // => returns 'pebbles' assuming `_.findKey` returns 'barney'
       *
       * // The `_.matches` iteratee shorthand.
       * _.findLastKey(users, { 'age': 36, 'active': true });
       * // => 'barney'
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.findLastKey(users, ['active', false]);
       * // => 'fred'
       *
       * // The `_.property` iteratee shorthand.
       * _.findLastKey(users, 'active');
       * // => 'pebbles'
       */
      function findLastKey(object, predicate) {
        return baseFindKey(object, getIteratee(predicate, 3), baseForOwnRight);
      }

      /**
       * Iterates over own and inherited enumerable string keyed properties of an
       * object and invokes `iteratee` for each property. The iteratee is invoked
       * with three arguments: (value, key, object). Iteratee functions may exit
       * iteration early by explicitly returning `false`.
       *
       * @static
       * @memberOf _
       * @since 0.3.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Object} Returns `object`.
       * @see _.forInRight
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.forIn(new Foo, function(value, key) {
       *   console.log(key);
       * });
       * // => Logs 'a', 'b', then 'c' (iteration order is not guaranteed).
       */
      function forIn(object, iteratee) {
        return object == null
          ? object
          : baseFor(object, getIteratee(iteratee, 3), keysIn);
      }

      /**
       * This method is like `_.forIn` except that it iterates over properties of
       * `object` in the opposite order.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Object} Returns `object`.
       * @see _.forIn
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.forInRight(new Foo, function(value, key) {
       *   console.log(key);
       * });
       * // => Logs 'c', 'b', then 'a' assuming `_.forIn` logs 'a', 'b', then 'c'.
       */
      function forInRight(object, iteratee) {
        return object == null
          ? object
          : baseForRight(object, getIteratee(iteratee, 3), keysIn);
      }

      /**
       * Iterates over own enumerable string keyed properties of an object and
       * invokes `iteratee` for each property. The iteratee is invoked with three
       * arguments: (value, key, object). Iteratee functions may exit iteration
       * early by explicitly returning `false`.
       *
       * @static
       * @memberOf _
       * @since 0.3.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Object} Returns `object`.
       * @see _.forOwnRight
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.forOwn(new Foo, function(value, key) {
       *   console.log(key);
       * });
       * // => Logs 'a' then 'b' (iteration order is not guaranteed).
       */
      function forOwn(object, iteratee) {
        return object && baseForOwn(object, getIteratee(iteratee, 3));
      }

      /**
       * This method is like `_.forOwn` except that it iterates over properties of
       * `object` in the opposite order.
       *
       * @static
       * @memberOf _
       * @since 2.0.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Object} Returns `object`.
       * @see _.forOwn
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.forOwnRight(new Foo, function(value, key) {
       *   console.log(key);
       * });
       * // => Logs 'b' then 'a' assuming `_.forOwn` logs 'a' then 'b'.
       */
      function forOwnRight(object, iteratee) {
        return object && baseForOwnRight(object, getIteratee(iteratee, 3));
      }

      /**
       * Creates an array of function property names from own enumerable properties
       * of `object`.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The object to inspect.
       * @returns {Array} Returns the function names.
       * @see _.functionsIn
       * @example
       *
       * function Foo() {
       *   this.a = _.constant('a');
       *   this.b = _.constant('b');
       * }
       *
       * Foo.prototype.c = _.constant('c');
       *
       * _.functions(new Foo);
       * // => ['a', 'b']
       */
      function functions(object) {
        return object == null ? [] : baseFunctions(object, keys(object));
      }

      /**
       * Creates an array of function property names from own and inherited
       * enumerable properties of `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The object to inspect.
       * @returns {Array} Returns the function names.
       * @see _.functions
       * @example
       *
       * function Foo() {
       *   this.a = _.constant('a');
       *   this.b = _.constant('b');
       * }
       *
       * Foo.prototype.c = _.constant('c');
       *
       * _.functionsIn(new Foo);
       * // => ['a', 'b', 'c']
       */
      function functionsIn(object) {
        return object == null ? [] : baseFunctions(object, keysIn(object));
      }

      /**
       * Gets the value at `path` of `object`. If the resolved value is
       * `undefined`, the `defaultValue` is returned in its place.
       *
       * @static
       * @memberOf _
       * @since 3.7.0
       * @category Object
       * @param {Object} object The object to query.
       * @param {Array|string} path The path of the property to get.
       * @param {*} [defaultValue] The value returned for `undefined` resolved values.
       * @returns {*} Returns the resolved value.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': 3 } }] };
       *
       * _.get(object, 'a[0].b.c');
       * // => 3
       *
       * _.get(object, ['a', '0', 'b', 'c']);
       * // => 3
       *
       * _.get(object, 'a.b.c', 'default');
       * // => 'default'
       */
      function get(object, path, defaultValue) {
        var result = object == null ? undefined$1 : baseGet(object, path);
        return result === undefined$1 ? defaultValue : result;
      }

      /**
       * Checks if `path` is a direct property of `object`.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The object to query.
       * @param {Array|string} path The path to check.
       * @returns {boolean} Returns `true` if `path` exists, else `false`.
       * @example
       *
       * var object = { 'a': { 'b': 2 } };
       * var other = _.create({ 'a': _.create({ 'b': 2 }) });
       *
       * _.has(object, 'a');
       * // => true
       *
       * _.has(object, 'a.b');
       * // => true
       *
       * _.has(object, ['a', 'b']);
       * // => true
       *
       * _.has(other, 'a');
       * // => false
       */
      function has(object, path) {
        return object != null && hasPath(object, path, baseHas);
      }

      /**
       * Checks if `path` is a direct or inherited property of `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The object to query.
       * @param {Array|string} path The path to check.
       * @returns {boolean} Returns `true` if `path` exists, else `false`.
       * @example
       *
       * var object = _.create({ 'a': _.create({ 'b': 2 }) });
       *
       * _.hasIn(object, 'a');
       * // => true
       *
       * _.hasIn(object, 'a.b');
       * // => true
       *
       * _.hasIn(object, ['a', 'b']);
       * // => true
       *
       * _.hasIn(object, 'b');
       * // => false
       */
      function hasIn(object, path) {
        return object != null && hasPath(object, path, baseHasIn);
      }

      /**
       * Creates an object composed of the inverted keys and values of `object`.
       * If `object` contains duplicate values, subsequent values overwrite
       * property assignments of previous values.
       *
       * @static
       * @memberOf _
       * @since 0.7.0
       * @category Object
       * @param {Object} object The object to invert.
       * @returns {Object} Returns the new inverted object.
       * @example
       *
       * var object = { 'a': 1, 'b': 2, 'c': 1 };
       *
       * _.invert(object);
       * // => { '1': 'c', '2': 'b' }
       */
      var invert = createInverter(function(result, value, key) {
        if (value != null &&
            typeof value.toString != 'function') {
          value = nativeObjectToString.call(value);
        }

        result[value] = key;
      }, constant(identity));

      /**
       * This method is like `_.invert` except that the inverted object is generated
       * from the results of running each element of `object` thru `iteratee`. The
       * corresponding inverted value of each inverted key is an array of keys
       * responsible for generating the inverted value. The iteratee is invoked
       * with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.1.0
       * @category Object
       * @param {Object} object The object to invert.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {Object} Returns the new inverted object.
       * @example
       *
       * var object = { 'a': 1, 'b': 2, 'c': 1 };
       *
       * _.invertBy(object);
       * // => { '1': ['a', 'c'], '2': ['b'] }
       *
       * _.invertBy(object, function(value) {
       *   return 'group' + value;
       * });
       * // => { 'group1': ['a', 'c'], 'group2': ['b'] }
       */
      var invertBy = createInverter(function(result, value, key) {
        if (value != null &&
            typeof value.toString != 'function') {
          value = nativeObjectToString.call(value);
        }

        if (hasOwnProperty.call(result, value)) {
          result[value].push(key);
        } else {
          result[value] = [key];
        }
      }, getIteratee);

      /**
       * Invokes the method at `path` of `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The object to query.
       * @param {Array|string} path The path of the method to invoke.
       * @param {...*} [args] The arguments to invoke the method with.
       * @returns {*} Returns the result of the invoked method.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': [1, 2, 3, 4] } }] };
       *
       * _.invoke(object, 'a[0].b.c.slice', 1, 3);
       * // => [2, 3]
       */
      var invoke = baseRest(baseInvoke);

      /**
       * Creates an array of the own enumerable property names of `object`.
       *
       * **Note:** Non-object values are coerced to objects. See the
       * [ES spec](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
       * for more details.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.keys(new Foo);
       * // => ['a', 'b'] (iteration order is not guaranteed)
       *
       * _.keys('hi');
       * // => ['0', '1']
       */
      function keys(object) {
        return isArrayLike(object) ? arrayLikeKeys(object) : baseKeys(object);
      }

      /**
       * Creates an array of the own and inherited enumerable property names of `object`.
       *
       * **Note:** Non-object values are coerced to objects.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Object
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property names.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.keysIn(new Foo);
       * // => ['a', 'b', 'c'] (iteration order is not guaranteed)
       */
      function keysIn(object) {
        return isArrayLike(object) ? arrayLikeKeys(object, true) : baseKeysIn(object);
      }

      /**
       * The opposite of `_.mapValues`; this method creates an object with the
       * same values as `object` and keys generated by running each own enumerable
       * string keyed property of `object` thru `iteratee`. The iteratee is invoked
       * with three arguments: (value, key, object).
       *
       * @static
       * @memberOf _
       * @since 3.8.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Object} Returns the new mapped object.
       * @see _.mapValues
       * @example
       *
       * _.mapKeys({ 'a': 1, 'b': 2 }, function(value, key) {
       *   return key + value;
       * });
       * // => { 'a1': 1, 'b2': 2 }
       */
      function mapKeys(object, iteratee) {
        var result = {};
        iteratee = getIteratee(iteratee, 3);

        baseForOwn(object, function(value, key, object) {
          baseAssignValue(result, iteratee(value, key, object), value);
        });
        return result;
      }

      /**
       * Creates an object with the same keys as `object` and values generated
       * by running each own enumerable string keyed property of `object` thru
       * `iteratee`. The iteratee is invoked with three arguments:
       * (value, key, object).
       *
       * @static
       * @memberOf _
       * @since 2.4.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Object} Returns the new mapped object.
       * @see _.mapKeys
       * @example
       *
       * var users = {
       *   'fred':    { 'user': 'fred',    'age': 40 },
       *   'pebbles': { 'user': 'pebbles', 'age': 1 }
       * };
       *
       * _.mapValues(users, function(o) { return o.age; });
       * // => { 'fred': 40, 'pebbles': 1 } (iteration order is not guaranteed)
       *
       * // The `_.property` iteratee shorthand.
       * _.mapValues(users, 'age');
       * // => { 'fred': 40, 'pebbles': 1 } (iteration order is not guaranteed)
       */
      function mapValues(object, iteratee) {
        var result = {};
        iteratee = getIteratee(iteratee, 3);

        baseForOwn(object, function(value, key, object) {
          baseAssignValue(result, key, iteratee(value, key, object));
        });
        return result;
      }

      /**
       * This method is like `_.assign` except that it recursively merges own and
       * inherited enumerable string keyed properties of source objects into the
       * destination object. Source properties that resolve to `undefined` are
       * skipped if a destination value exists. Array and plain object properties
       * are merged recursively. Other objects and value types are overridden by
       * assignment. Source objects are applied from left to right. Subsequent
       * sources overwrite property assignments of previous sources.
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 0.5.0
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} [sources] The source objects.
       * @returns {Object} Returns `object`.
       * @example
       *
       * var object = {
       *   'a': [{ 'b': 2 }, { 'd': 4 }]
       * };
       *
       * var other = {
       *   'a': [{ 'c': 3 }, { 'e': 5 }]
       * };
       *
       * _.merge(object, other);
       * // => { 'a': [{ 'b': 2, 'c': 3 }, { 'd': 4, 'e': 5 }] }
       */
      var merge = createAssigner(function(object, source, srcIndex) {
        baseMerge(object, source, srcIndex);
      });

      /**
       * This method is like `_.merge` except that it accepts `customizer` which
       * is invoked to produce the merged values of the destination and source
       * properties. If `customizer` returns `undefined`, merging is handled by the
       * method instead. The `customizer` is invoked with six arguments:
       * (objValue, srcValue, key, object, source, stack).
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The destination object.
       * @param {...Object} sources The source objects.
       * @param {Function} customizer The function to customize assigned values.
       * @returns {Object} Returns `object`.
       * @example
       *
       * function customizer(objValue, srcValue) {
       *   if (_.isArray(objValue)) {
       *     return objValue.concat(srcValue);
       *   }
       * }
       *
       * var object = { 'a': [1], 'b': [2] };
       * var other = { 'a': [3], 'b': [4] };
       *
       * _.mergeWith(object, other, customizer);
       * // => { 'a': [1, 3], 'b': [2, 4] }
       */
      var mergeWith = createAssigner(function(object, source, srcIndex, customizer) {
        baseMerge(object, source, srcIndex, customizer);
      });

      /**
       * The opposite of `_.pick`; this method creates an object composed of the
       * own and inherited enumerable property paths of `object` that are not omitted.
       *
       * **Note:** This method is considerably slower than `_.pick`.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The source object.
       * @param {...(string|string[])} [paths] The property paths to omit.
       * @returns {Object} Returns the new object.
       * @example
       *
       * var object = { 'a': 1, 'b': '2', 'c': 3 };
       *
       * _.omit(object, ['a', 'c']);
       * // => { 'b': '2' }
       */
      var omit = flatRest(function(object, paths) {
        var result = {};
        if (object == null) {
          return result;
        }
        var isDeep = false;
        paths = arrayMap(paths, function(path) {
          path = castPath(path, object);
          isDeep || (isDeep = path.length > 1);
          return path;
        });
        copyObject(object, getAllKeysIn(object), result);
        if (isDeep) {
          result = baseClone(result, CLONE_DEEP_FLAG | CLONE_FLAT_FLAG | CLONE_SYMBOLS_FLAG, customOmitClone);
        }
        var length = paths.length;
        while (length--) {
          baseUnset(result, paths[length]);
        }
        return result;
      });

      /**
       * The opposite of `_.pickBy`; this method creates an object composed of
       * the own and inherited enumerable string keyed properties of `object` that
       * `predicate` doesn't return truthy for. The predicate is invoked with two
       * arguments: (value, key).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The source object.
       * @param {Function} [predicate=_.identity] The function invoked per property.
       * @returns {Object} Returns the new object.
       * @example
       *
       * var object = { 'a': 1, 'b': '2', 'c': 3 };
       *
       * _.omitBy(object, _.isNumber);
       * // => { 'b': '2' }
       */
      function omitBy(object, predicate) {
        return pickBy(object, negate(getIteratee(predicate)));
      }

      /**
       * Creates an object composed of the picked `object` properties.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The source object.
       * @param {...(string|string[])} [paths] The property paths to pick.
       * @returns {Object} Returns the new object.
       * @example
       *
       * var object = { 'a': 1, 'b': '2', 'c': 3 };
       *
       * _.pick(object, ['a', 'c']);
       * // => { 'a': 1, 'c': 3 }
       */
      var pick = flatRest(function(object, paths) {
        return object == null ? {} : basePick(object, paths);
      });

      /**
       * Creates an object composed of the `object` properties `predicate` returns
       * truthy for. The predicate is invoked with two arguments: (value, key).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The source object.
       * @param {Function} [predicate=_.identity] The function invoked per property.
       * @returns {Object} Returns the new object.
       * @example
       *
       * var object = { 'a': 1, 'b': '2', 'c': 3 };
       *
       * _.pickBy(object, _.isNumber);
       * // => { 'a': 1, 'c': 3 }
       */
      function pickBy(object, predicate) {
        if (object == null) {
          return {};
        }
        var props = arrayMap(getAllKeysIn(object), function(prop) {
          return [prop];
        });
        predicate = getIteratee(predicate);
        return basePickBy(object, props, function(value, path) {
          return predicate(value, path[0]);
        });
      }

      /**
       * This method is like `_.get` except that if the resolved value is a
       * function it's invoked with the `this` binding of its parent object and
       * its result is returned.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The object to query.
       * @param {Array|string} path The path of the property to resolve.
       * @param {*} [defaultValue] The value returned for `undefined` resolved values.
       * @returns {*} Returns the resolved value.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c1': 3, 'c2': _.constant(4) } }] };
       *
       * _.result(object, 'a[0].b.c1');
       * // => 3
       *
       * _.result(object, 'a[0].b.c2');
       * // => 4
       *
       * _.result(object, 'a[0].b.c3', 'default');
       * // => 'default'
       *
       * _.result(object, 'a[0].b.c3', _.constant('default'));
       * // => 'default'
       */
      function result(object, path, defaultValue) {
        path = castPath(path, object);

        var index = -1,
            length = path.length;

        // Ensure the loop is entered when path is empty.
        if (!length) {
          length = 1;
          object = undefined$1;
        }
        while (++index < length) {
          var value = object == null ? undefined$1 : object[toKey(path[index])];
          if (value === undefined$1) {
            index = length;
            value = defaultValue;
          }
          object = isFunction(value) ? value.call(object) : value;
        }
        return object;
      }

      /**
       * Sets the value at `path` of `object`. If a portion of `path` doesn't exist,
       * it's created. Arrays are created for missing index properties while objects
       * are created for all other missing properties. Use `_.setWith` to customize
       * `path` creation.
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 3.7.0
       * @category Object
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to set.
       * @param {*} value The value to set.
       * @returns {Object} Returns `object`.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': 3 } }] };
       *
       * _.set(object, 'a[0].b.c', 4);
       * console.log(object.a[0].b.c);
       * // => 4
       *
       * _.set(object, ['x', '0', 'y', 'z'], 5);
       * console.log(object.x[0].y.z);
       * // => 5
       */
      function set(object, path, value) {
        return object == null ? object : baseSet(object, path, value);
      }

      /**
       * This method is like `_.set` except that it accepts `customizer` which is
       * invoked to produce the objects of `path`.  If `customizer` returns `undefined`
       * path creation is handled by the method instead. The `customizer` is invoked
       * with three arguments: (nsValue, key, nsObject).
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to set.
       * @param {*} value The value to set.
       * @param {Function} [customizer] The function to customize assigned values.
       * @returns {Object} Returns `object`.
       * @example
       *
       * var object = {};
       *
       * _.setWith(object, '[0][1]', 'a', Object);
       * // => { '0': { '1': 'a' } }
       */
      function setWith(object, path, value, customizer) {
        customizer = typeof customizer == 'function' ? customizer : undefined$1;
        return object == null ? object : baseSet(object, path, value, customizer);
      }

      /**
       * Creates an array of own enumerable string keyed-value pairs for `object`
       * which can be consumed by `_.fromPairs`. If `object` is a map or set, its
       * entries are returned.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @alias entries
       * @category Object
       * @param {Object} object The object to query.
       * @returns {Array} Returns the key-value pairs.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.toPairs(new Foo);
       * // => [['a', 1], ['b', 2]] (iteration order is not guaranteed)
       */
      var toPairs = createToPairs(keys);

      /**
       * Creates an array of own and inherited enumerable string keyed-value pairs
       * for `object` which can be consumed by `_.fromPairs`. If `object` is a map
       * or set, its entries are returned.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @alias entriesIn
       * @category Object
       * @param {Object} object The object to query.
       * @returns {Array} Returns the key-value pairs.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.toPairsIn(new Foo);
       * // => [['a', 1], ['b', 2], ['c', 3]] (iteration order is not guaranteed)
       */
      var toPairsIn = createToPairs(keysIn);

      /**
       * An alternative to `_.reduce`; this method transforms `object` to a new
       * `accumulator` object which is the result of running each of its own
       * enumerable string keyed properties thru `iteratee`, with each invocation
       * potentially mutating the `accumulator` object. If `accumulator` is not
       * provided, a new object with the same `[[Prototype]]` will be used. The
       * iteratee is invoked with four arguments: (accumulator, value, key, object).
       * Iteratee functions may exit iteration early by explicitly returning `false`.
       *
       * @static
       * @memberOf _
       * @since 1.3.0
       * @category Object
       * @param {Object} object The object to iterate over.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @param {*} [accumulator] The custom accumulator value.
       * @returns {*} Returns the accumulated value.
       * @example
       *
       * _.transform([2, 3, 4], function(result, n) {
       *   result.push(n *= n);
       *   return n % 2 == 0;
       * }, []);
       * // => [4, 9]
       *
       * _.transform({ 'a': 1, 'b': 2, 'c': 1 }, function(result, value, key) {
       *   (result[value] || (result[value] = [])).push(key);
       * }, {});
       * // => { '1': ['a', 'c'], '2': ['b'] }
       */
      function transform(object, iteratee, accumulator) {
        var isArr = isArray(object),
            isArrLike = isArr || isBuffer(object) || isTypedArray(object);

        iteratee = getIteratee(iteratee, 4);
        if (accumulator == null) {
          var Ctor = object && object.constructor;
          if (isArrLike) {
            accumulator = isArr ? new Ctor : [];
          }
          else if (isObject(object)) {
            accumulator = isFunction(Ctor) ? baseCreate(getPrototype(object)) : {};
          }
          else {
            accumulator = {};
          }
        }
        (isArrLike ? arrayEach : baseForOwn)(object, function(value, index, object) {
          return iteratee(accumulator, value, index, object);
        });
        return accumulator;
      }

      /**
       * Removes the property at `path` of `object`.
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Object
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to unset.
       * @returns {boolean} Returns `true` if the property is deleted, else `false`.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': 7 } }] };
       * _.unset(object, 'a[0].b.c');
       * // => true
       *
       * console.log(object);
       * // => { 'a': [{ 'b': {} }] };
       *
       * _.unset(object, ['a', '0', 'b', 'c']);
       * // => true
       *
       * console.log(object);
       * // => { 'a': [{ 'b': {} }] };
       */
      function unset(object, path) {
        return object == null ? true : baseUnset(object, path);
      }

      /**
       * This method is like `_.set` except that accepts `updater` to produce the
       * value to set. Use `_.updateWith` to customize `path` creation. The `updater`
       * is invoked with one argument: (value).
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.6.0
       * @category Object
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to set.
       * @param {Function} updater The function to produce the updated value.
       * @returns {Object} Returns `object`.
       * @example
       *
       * var object = { 'a': [{ 'b': { 'c': 3 } }] };
       *
       * _.update(object, 'a[0].b.c', function(n) { return n * n; });
       * console.log(object.a[0].b.c);
       * // => 9
       *
       * _.update(object, 'x[0].y.z', function(n) { return n ? n + 1 : 0; });
       * console.log(object.x[0].y.z);
       * // => 0
       */
      function update(object, path, updater) {
        return object == null ? object : baseUpdate(object, path, castFunction(updater));
      }

      /**
       * This method is like `_.update` except that it accepts `customizer` which is
       * invoked to produce the objects of `path`.  If `customizer` returns `undefined`
       * path creation is handled by the method instead. The `customizer` is invoked
       * with three arguments: (nsValue, key, nsObject).
       *
       * **Note:** This method mutates `object`.
       *
       * @static
       * @memberOf _
       * @since 4.6.0
       * @category Object
       * @param {Object} object The object to modify.
       * @param {Array|string} path The path of the property to set.
       * @param {Function} updater The function to produce the updated value.
       * @param {Function} [customizer] The function to customize assigned values.
       * @returns {Object} Returns `object`.
       * @example
       *
       * var object = {};
       *
       * _.updateWith(object, '[0][1]', _.constant('a'), Object);
       * // => { '0': { '1': 'a' } }
       */
      function updateWith(object, path, updater, customizer) {
        customizer = typeof customizer == 'function' ? customizer : undefined$1;
        return object == null ? object : baseUpdate(object, path, castFunction(updater), customizer);
      }

      /**
       * Creates an array of the own enumerable string keyed property values of `object`.
       *
       * **Note:** Non-object values are coerced to objects.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Object
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property values.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.values(new Foo);
       * // => [1, 2] (iteration order is not guaranteed)
       *
       * _.values('hi');
       * // => ['h', 'i']
       */
      function values(object) {
        return object == null ? [] : baseValues(object, keys(object));
      }

      /**
       * Creates an array of the own and inherited enumerable string keyed property
       * values of `object`.
       *
       * **Note:** Non-object values are coerced to objects.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Object
       * @param {Object} object The object to query.
       * @returns {Array} Returns the array of property values.
       * @example
       *
       * function Foo() {
       *   this.a = 1;
       *   this.b = 2;
       * }
       *
       * Foo.prototype.c = 3;
       *
       * _.valuesIn(new Foo);
       * // => [1, 2, 3] (iteration order is not guaranteed)
       */
      function valuesIn(object) {
        return object == null ? [] : baseValues(object, keysIn(object));
      }

      /*------------------------------------------------------------------------*/

      /**
       * Clamps `number` within the inclusive `lower` and `upper` bounds.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Number
       * @param {number} number The number to clamp.
       * @param {number} [lower] The lower bound.
       * @param {number} upper The upper bound.
       * @returns {number} Returns the clamped number.
       * @example
       *
       * _.clamp(-10, -5, 5);
       * // => -5
       *
       * _.clamp(10, -5, 5);
       * // => 5
       */
      function clamp(number, lower, upper) {
        if (upper === undefined$1) {
          upper = lower;
          lower = undefined$1;
        }
        if (upper !== undefined$1) {
          upper = toNumber(upper);
          upper = upper === upper ? upper : 0;
        }
        if (lower !== undefined$1) {
          lower = toNumber(lower);
          lower = lower === lower ? lower : 0;
        }
        return baseClamp(toNumber(number), lower, upper);
      }

      /**
       * Checks if `n` is between `start` and up to, but not including, `end`. If
       * `end` is not specified, it's set to `start` with `start` then set to `0`.
       * If `start` is greater than `end` the params are swapped to support
       * negative ranges.
       *
       * @static
       * @memberOf _
       * @since 3.3.0
       * @category Number
       * @param {number} number The number to check.
       * @param {number} [start=0] The start of the range.
       * @param {number} end The end of the range.
       * @returns {boolean} Returns `true` if `number` is in the range, else `false`.
       * @see _.range, _.rangeRight
       * @example
       *
       * _.inRange(3, 2, 4);
       * // => true
       *
       * _.inRange(4, 8);
       * // => true
       *
       * _.inRange(4, 2);
       * // => false
       *
       * _.inRange(2, 2);
       * // => false
       *
       * _.inRange(1.2, 2);
       * // => true
       *
       * _.inRange(5.2, 4);
       * // => false
       *
       * _.inRange(-3, -2, -6);
       * // => true
       */
      function inRange(number, start, end) {
        start = toFinite(start);
        if (end === undefined$1) {
          end = start;
          start = 0;
        } else {
          end = toFinite(end);
        }
        number = toNumber(number);
        return baseInRange(number, start, end);
      }

      /**
       * Produces a random number between the inclusive `lower` and `upper` bounds.
       * If only one argument is provided a number between `0` and the given number
       * is returned. If `floating` is `true`, or either `lower` or `upper` are
       * floats, a floating-point number is returned instead of an integer.
       *
       * **Note:** JavaScript follows the IEEE-754 standard for resolving
       * floating-point values which can produce unexpected results.
       *
       * @static
       * @memberOf _
       * @since 0.7.0
       * @category Number
       * @param {number} [lower=0] The lower bound.
       * @param {number} [upper=1] The upper bound.
       * @param {boolean} [floating] Specify returning a floating-point number.
       * @returns {number} Returns the random number.
       * @example
       *
       * _.random(0, 5);
       * // => an integer between 0 and 5
       *
       * _.random(5);
       * // => also an integer between 0 and 5
       *
       * _.random(5, true);
       * // => a floating-point number between 0 and 5
       *
       * _.random(1.2, 5.2);
       * // => a floating-point number between 1.2 and 5.2
       */
      function random(lower, upper, floating) {
        if (floating && typeof floating != 'boolean' && isIterateeCall(lower, upper, floating)) {
          upper = floating = undefined$1;
        }
        if (floating === undefined$1) {
          if (typeof upper == 'boolean') {
            floating = upper;
            upper = undefined$1;
          }
          else if (typeof lower == 'boolean') {
            floating = lower;
            lower = undefined$1;
          }
        }
        if (lower === undefined$1 && upper === undefined$1) {
          lower = 0;
          upper = 1;
        }
        else {
          lower = toFinite(lower);
          if (upper === undefined$1) {
            upper = lower;
            lower = 0;
          } else {
            upper = toFinite(upper);
          }
        }
        if (lower > upper) {
          var temp = lower;
          lower = upper;
          upper = temp;
        }
        if (floating || lower % 1 || upper % 1) {
          var rand = nativeRandom();
          return nativeMin(lower + (rand * (upper - lower + freeParseFloat('1e-' + ((rand + '').length - 1)))), upper);
        }
        return baseRandom(lower, upper);
      }

      /*------------------------------------------------------------------------*/

      /**
       * Converts `string` to [camel case](https://en.wikipedia.org/wiki/CamelCase).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the camel cased string.
       * @example
       *
       * _.camelCase('Foo Bar');
       * // => 'fooBar'
       *
       * _.camelCase('--foo-bar--');
       * // => 'fooBar'
       *
       * _.camelCase('__FOO_BAR__');
       * // => 'fooBar'
       */
      var camelCase = createCompounder(function(result, word, index) {
        word = word.toLowerCase();
        return result + (index ? capitalize(word) : word);
      });

      /**
       * Converts the first character of `string` to upper case and the remaining
       * to lower case.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to capitalize.
       * @returns {string} Returns the capitalized string.
       * @example
       *
       * _.capitalize('FRED');
       * // => 'Fred'
       */
      function capitalize(string) {
        return upperFirst(toString(string).toLowerCase());
      }

      /**
       * Deburrs `string` by converting
       * [Latin-1 Supplement](https://en.wikipedia.org/wiki/Latin-1_Supplement_(Unicode_block)#Character_table)
       * and [Latin Extended-A](https://en.wikipedia.org/wiki/Latin_Extended-A)
       * letters to basic Latin letters and removing
       * [combining diacritical marks](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to deburr.
       * @returns {string} Returns the deburred string.
       * @example
       *
       * _.deburr('déjà vu');
       * // => 'deja vu'
       */
      function deburr(string) {
        string = toString(string);
        return string && string.replace(reLatin, deburrLetter).replace(reComboMark, '');
      }

      /**
       * Checks if `string` ends with the given target string.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to inspect.
       * @param {string} [target] The string to search for.
       * @param {number} [position=string.length] The position to search up to.
       * @returns {boolean} Returns `true` if `string` ends with `target`,
       *  else `false`.
       * @example
       *
       * _.endsWith('abc', 'c');
       * // => true
       *
       * _.endsWith('abc', 'b');
       * // => false
       *
       * _.endsWith('abc', 'b', 2);
       * // => true
       */
      function endsWith(string, target, position) {
        string = toString(string);
        target = baseToString(target);

        var length = string.length;
        position = position === undefined$1
          ? length
          : baseClamp(toInteger(position), 0, length);

        var end = position;
        position -= target.length;
        return position >= 0 && string.slice(position, end) == target;
      }

      /**
       * Converts the characters "&", "<", ">", '"', and "'" in `string` to their
       * corresponding HTML entities.
       *
       * **Note:** No other characters are escaped. To escape additional
       * characters use a third-party library like [_he_](https://mths.be/he).
       *
       * Though the ">" character is escaped for symmetry, characters like
       * ">" and "/" don't need escaping in HTML and have no special meaning
       * unless they're part of a tag or unquoted attribute value. See
       * [Mathias Bynens's article](https://mathiasbynens.be/notes/ambiguous-ampersands)
       * (under "semi-related fun fact") for more details.
       *
       * When working with HTML you should always
       * [quote attribute values](http://wonko.com/post/html-escaping) to reduce
       * XSS vectors.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category String
       * @param {string} [string=''] The string to escape.
       * @returns {string} Returns the escaped string.
       * @example
       *
       * _.escape('fred, barney, & pebbles');
       * // => 'fred, barney, &amp; pebbles'
       */
      function escape(string) {
        string = toString(string);
        return (string && reHasUnescapedHtml.test(string))
          ? string.replace(reUnescapedHtml, escapeHtmlChar)
          : string;
      }

      /**
       * Escapes the `RegExp` special characters "^", "$", "\", ".", "*", "+",
       * "?", "(", ")", "[", "]", "{", "}", and "|" in `string`.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to escape.
       * @returns {string} Returns the escaped string.
       * @example
       *
       * _.escapeRegExp('[lodash](https://lodash.com/)');
       * // => '\[lodash\]\(https://lodash\.com/\)'
       */
      function escapeRegExp(string) {
        string = toString(string);
        return (string && reHasRegExpChar.test(string))
          ? string.replace(reRegExpChar, '\\$&')
          : string;
      }

      /**
       * Converts `string` to
       * [kebab case](https://en.wikipedia.org/wiki/Letter_case#Special_case_styles).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the kebab cased string.
       * @example
       *
       * _.kebabCase('Foo Bar');
       * // => 'foo-bar'
       *
       * _.kebabCase('fooBar');
       * // => 'foo-bar'
       *
       * _.kebabCase('__FOO_BAR__');
       * // => 'foo-bar'
       */
      var kebabCase = createCompounder(function(result, word, index) {
        return result + (index ? '-' : '') + word.toLowerCase();
      });

      /**
       * Converts `string`, as space separated words, to lower case.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the lower cased string.
       * @example
       *
       * _.lowerCase('--Foo-Bar--');
       * // => 'foo bar'
       *
       * _.lowerCase('fooBar');
       * // => 'foo bar'
       *
       * _.lowerCase('__FOO_BAR__');
       * // => 'foo bar'
       */
      var lowerCase = createCompounder(function(result, word, index) {
        return result + (index ? ' ' : '') + word.toLowerCase();
      });

      /**
       * Converts the first character of `string` to lower case.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the converted string.
       * @example
       *
       * _.lowerFirst('Fred');
       * // => 'fred'
       *
       * _.lowerFirst('FRED');
       * // => 'fRED'
       */
      var lowerFirst = createCaseFirst('toLowerCase');

      /**
       * Pads `string` on the left and right sides if it's shorter than `length`.
       * Padding characters are truncated if they can't be evenly divided by `length`.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to pad.
       * @param {number} [length=0] The padding length.
       * @param {string} [chars=' '] The string used as padding.
       * @returns {string} Returns the padded string.
       * @example
       *
       * _.pad('abc', 8);
       * // => '  abc   '
       *
       * _.pad('abc', 8, '_-');
       * // => '_-abc_-_'
       *
       * _.pad('abc', 3);
       * // => 'abc'
       */
      function pad(string, length, chars) {
        string = toString(string);
        length = toInteger(length);

        var strLength = length ? stringSize(string) : 0;
        if (!length || strLength >= length) {
          return string;
        }
        var mid = (length - strLength) / 2;
        return (
          createPadding(nativeFloor(mid), chars) +
          string +
          createPadding(nativeCeil(mid), chars)
        );
      }

      /**
       * Pads `string` on the right side if it's shorter than `length`. Padding
       * characters are truncated if they exceed `length`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to pad.
       * @param {number} [length=0] The padding length.
       * @param {string} [chars=' '] The string used as padding.
       * @returns {string} Returns the padded string.
       * @example
       *
       * _.padEnd('abc', 6);
       * // => 'abc   '
       *
       * _.padEnd('abc', 6, '_-');
       * // => 'abc_-_'
       *
       * _.padEnd('abc', 3);
       * // => 'abc'
       */
      function padEnd(string, length, chars) {
        string = toString(string);
        length = toInteger(length);

        var strLength = length ? stringSize(string) : 0;
        return (length && strLength < length)
          ? (string + createPadding(length - strLength, chars))
          : string;
      }

      /**
       * Pads `string` on the left side if it's shorter than `length`. Padding
       * characters are truncated if they exceed `length`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to pad.
       * @param {number} [length=0] The padding length.
       * @param {string} [chars=' '] The string used as padding.
       * @returns {string} Returns the padded string.
       * @example
       *
       * _.padStart('abc', 6);
       * // => '   abc'
       *
       * _.padStart('abc', 6, '_-');
       * // => '_-_abc'
       *
       * _.padStart('abc', 3);
       * // => 'abc'
       */
      function padStart(string, length, chars) {
        string = toString(string);
        length = toInteger(length);

        var strLength = length ? stringSize(string) : 0;
        return (length && strLength < length)
          ? (createPadding(length - strLength, chars) + string)
          : string;
      }

      /**
       * Converts `string` to an integer of the specified radix. If `radix` is
       * `undefined` or `0`, a `radix` of `10` is used unless `value` is a
       * hexadecimal, in which case a `radix` of `16` is used.
       *
       * **Note:** This method aligns with the
       * [ES5 implementation](https://es5.github.io/#x15.1.2.2) of `parseInt`.
       *
       * @static
       * @memberOf _
       * @since 1.1.0
       * @category String
       * @param {string} string The string to convert.
       * @param {number} [radix=10] The radix to interpret `value` by.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {number} Returns the converted integer.
       * @example
       *
       * _.parseInt('08');
       * // => 8
       *
       * _.map(['6', '08', '10'], _.parseInt);
       * // => [6, 8, 10]
       */
      function parseInt(string, radix, guard) {
        if (guard || radix == null) {
          radix = 0;
        } else if (radix) {
          radix = +radix;
        }
        return nativeParseInt(toString(string).replace(reTrimStart, ''), radix || 0);
      }

      /**
       * Repeats the given string `n` times.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to repeat.
       * @param {number} [n=1] The number of times to repeat the string.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {string} Returns the repeated string.
       * @example
       *
       * _.repeat('*', 3);
       * // => '***'
       *
       * _.repeat('abc', 2);
       * // => 'abcabc'
       *
       * _.repeat('abc', 0);
       * // => ''
       */
      function repeat(string, n, guard) {
        if ((guard ? isIterateeCall(string, n, guard) : n === undefined$1)) {
          n = 1;
        } else {
          n = toInteger(n);
        }
        return baseRepeat(toString(string), n);
      }

      /**
       * Replaces matches for `pattern` in `string` with `replacement`.
       *
       * **Note:** This method is based on
       * [`String#replace`](https://mdn.io/String/replace).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to modify.
       * @param {RegExp|string} pattern The pattern to replace.
       * @param {Function|string} replacement The match replacement.
       * @returns {string} Returns the modified string.
       * @example
       *
       * _.replace('Hi Fred', 'Fred', 'Barney');
       * // => 'Hi Barney'
       */
      function replace() {
        var args = arguments,
            string = toString(args[0]);

        return args.length < 3 ? string : string.replace(args[1], args[2]);
      }

      /**
       * Converts `string` to
       * [snake case](https://en.wikipedia.org/wiki/Snake_case).
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the snake cased string.
       * @example
       *
       * _.snakeCase('Foo Bar');
       * // => 'foo_bar'
       *
       * _.snakeCase('fooBar');
       * // => 'foo_bar'
       *
       * _.snakeCase('--FOO-BAR--');
       * // => 'foo_bar'
       */
      var snakeCase = createCompounder(function(result, word, index) {
        return result + (index ? '_' : '') + word.toLowerCase();
      });

      /**
       * Splits `string` by `separator`.
       *
       * **Note:** This method is based on
       * [`String#split`](https://mdn.io/String/split).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to split.
       * @param {RegExp|string} separator The separator pattern to split by.
       * @param {number} [limit] The length to truncate results to.
       * @returns {Array} Returns the string segments.
       * @example
       *
       * _.split('a-b-c', '-', 2);
       * // => ['a', 'b']
       */
      function split(string, separator, limit) {
        if (limit && typeof limit != 'number' && isIterateeCall(string, separator, limit)) {
          separator = limit = undefined$1;
        }
        limit = limit === undefined$1 ? MAX_ARRAY_LENGTH : limit >>> 0;
        if (!limit) {
          return [];
        }
        string = toString(string);
        if (string && (
              typeof separator == 'string' ||
              (separator != null && !isRegExp(separator))
            )) {
          separator = baseToString(separator);
          if (!separator && hasUnicode(string)) {
            return castSlice(stringToArray(string), 0, limit);
          }
        }
        return string.split(separator, limit);
      }

      /**
       * Converts `string` to
       * [start case](https://en.wikipedia.org/wiki/Letter_case#Stylistic_or_specialised_usage).
       *
       * @static
       * @memberOf _
       * @since 3.1.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the start cased string.
       * @example
       *
       * _.startCase('--foo-bar--');
       * // => 'Foo Bar'
       *
       * _.startCase('fooBar');
       * // => 'Foo Bar'
       *
       * _.startCase('__FOO_BAR__');
       * // => 'FOO BAR'
       */
      var startCase = createCompounder(function(result, word, index) {
        return result + (index ? ' ' : '') + upperFirst(word);
      });

      /**
       * Checks if `string` starts with the given target string.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to inspect.
       * @param {string} [target] The string to search for.
       * @param {number} [position=0] The position to search from.
       * @returns {boolean} Returns `true` if `string` starts with `target`,
       *  else `false`.
       * @example
       *
       * _.startsWith('abc', 'a');
       * // => true
       *
       * _.startsWith('abc', 'b');
       * // => false
       *
       * _.startsWith('abc', 'b', 1);
       * // => true
       */
      function startsWith(string, target, position) {
        string = toString(string);
        position = position == null
          ? 0
          : baseClamp(toInteger(position), 0, string.length);

        target = baseToString(target);
        return string.slice(position, position + target.length) == target;
      }

      /**
       * Creates a compiled template function that can interpolate data properties
       * in "interpolate" delimiters, HTML-escape interpolated data properties in
       * "escape" delimiters, and execute JavaScript in "evaluate" delimiters. Data
       * properties may be accessed as free variables in the template. If a setting
       * object is given, it takes precedence over `_.templateSettings` values.
       *
       * **Note:** In the development build `_.template` utilizes
       * [sourceURLs](http://www.html5rocks.com/en/tutorials/developertools/sourcemaps/#toc-sourceurl)
       * for easier debugging.
       *
       * For more information on precompiling templates see
       * [lodash's custom builds documentation](https://lodash.com/custom-builds).
       *
       * For more information on Chrome extension sandboxes see
       * [Chrome's extensions documentation](https://developer.chrome.com/extensions/sandboxingEval).
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category String
       * @param {string} [string=''] The template string.
       * @param {Object} [options={}] The options object.
       * @param {RegExp} [options.escape=_.templateSettings.escape]
       *  The HTML "escape" delimiter.
       * @param {RegExp} [options.evaluate=_.templateSettings.evaluate]
       *  The "evaluate" delimiter.
       * @param {Object} [options.imports=_.templateSettings.imports]
       *  An object to import into the template as free variables.
       * @param {RegExp} [options.interpolate=_.templateSettings.interpolate]
       *  The "interpolate" delimiter.
       * @param {string} [options.sourceURL='lodash.templateSources[n]']
       *  The sourceURL of the compiled template.
       * @param {string} [options.variable='obj']
       *  The data object variable name.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Function} Returns the compiled template function.
       * @example
       *
       * // Use the "interpolate" delimiter to create a compiled template.
       * var compiled = _.template('hello <%= user %>!');
       * compiled({ 'user': 'fred' });
       * // => 'hello fred!'
       *
       * // Use the HTML "escape" delimiter to escape data property values.
       * var compiled = _.template('<b><%- value %></b>');
       * compiled({ 'value': '<script>' });
       * // => '<b>&lt;script&gt;</b>'
       *
       * // Use the "evaluate" delimiter to execute JavaScript and generate HTML.
       * var compiled = _.template('<% _.forEach(users, function(user) { %><li><%- user %></li><% }); %>');
       * compiled({ 'users': ['fred', 'barney'] });
       * // => '<li>fred</li><li>barney</li>'
       *
       * // Use the internal `print` function in "evaluate" delimiters.
       * var compiled = _.template('<% print("hello " + user); %>!');
       * compiled({ 'user': 'barney' });
       * // => 'hello barney!'
       *
       * // Use the ES template literal delimiter as an "interpolate" delimiter.
       * // Disable support by replacing the "interpolate" delimiter.
       * var compiled = _.template('hello ${ user }!');
       * compiled({ 'user': 'pebbles' });
       * // => 'hello pebbles!'
       *
       * // Use backslashes to treat delimiters as plain text.
       * var compiled = _.template('<%= "\\<%- value %\\>" %>');
       * compiled({ 'value': 'ignored' });
       * // => '<%- value %>'
       *
       * // Use the `imports` option to import `jQuery` as `jq`.
       * var text = '<% jq.each(users, function(user) { %><li><%- user %></li><% }); %>';
       * var compiled = _.template(text, { 'imports': { 'jq': jQuery } });
       * compiled({ 'users': ['fred', 'barney'] });
       * // => '<li>fred</li><li>barney</li>'
       *
       * // Use the `sourceURL` option to specify a custom sourceURL for the template.
       * var compiled = _.template('hello <%= user %>!', { 'sourceURL': '/basic/greeting.jst' });
       * compiled(data);
       * // => Find the source of "greeting.jst" under the Sources tab or Resources panel of the web inspector.
       *
       * // Use the `variable` option to ensure a with-statement isn't used in the compiled template.
       * var compiled = _.template('hi <%= data.user %>!', { 'variable': 'data' });
       * compiled.source;
       * // => function(data) {
       * //   var __t, __p = '';
       * //   __p += 'hi ' + ((__t = ( data.user )) == null ? '' : __t) + '!';
       * //   return __p;
       * // }
       *
       * // Use custom template delimiters.
       * _.templateSettings.interpolate = /{{([\s\S]+?)}}/g;
       * var compiled = _.template('hello {{ user }}!');
       * compiled({ 'user': 'mustache' });
       * // => 'hello mustache!'
       *
       * // Use the `source` property to inline compiled templates for meaningful
       * // line numbers in error messages and stack traces.
       * fs.writeFileSync(path.join(process.cwd(), 'jst.js'), '\
       *   var JST = {\
       *     "main": ' + _.template(mainText).source + '\
       *   };\
       * ');
       */
      function template(string, options, guard) {
        // Based on John Resig's `tmpl` implementation
        // (http://ejohn.org/blog/javascript-micro-templating/)
        // and Laura Doktorova's doT.js (https://github.com/olado/doT).
        var settings = lodash.templateSettings;

        if (guard && isIterateeCall(string, options, guard)) {
          options = undefined$1;
        }
        string = toString(string);
        options = assignInWith({}, options, settings, customDefaultsAssignIn);

        var imports = assignInWith({}, options.imports, settings.imports, customDefaultsAssignIn),
            importsKeys = keys(imports),
            importsValues = baseValues(imports, importsKeys);

        var isEscaping,
            isEvaluating,
            index = 0,
            interpolate = options.interpolate || reNoMatch,
            source = "__p += '";

        // Compile the regexp to match each delimiter.
        var reDelimiters = RegExp(
          (options.escape || reNoMatch).source + '|' +
          interpolate.source + '|' +
          (interpolate === reInterpolate ? reEsTemplate : reNoMatch).source + '|' +
          (options.evaluate || reNoMatch).source + '|$'
        , 'g');

        // Use a sourceURL for easier debugging.
        // The sourceURL gets injected into the source that's eval-ed, so be careful
        // with lookup (in case of e.g. prototype pollution), and strip newlines if any.
        // A newline wouldn't be a valid sourceURL anyway, and it'd enable code injection.
        var sourceURL = '//# sourceURL=' +
          (hasOwnProperty.call(options, 'sourceURL')
            ? (options.sourceURL + '').replace(/[\r\n]/g, ' ')
            : ('lodash.templateSources[' + (++templateCounter) + ']')
          ) + '\n';

        string.replace(reDelimiters, function(match, escapeValue, interpolateValue, esTemplateValue, evaluateValue, offset) {
          interpolateValue || (interpolateValue = esTemplateValue);

          // Escape characters that can't be included in string literals.
          source += string.slice(index, offset).replace(reUnescapedString, escapeStringChar);

          // Replace delimiters with snippets.
          if (escapeValue) {
            isEscaping = true;
            source += "' +\n__e(" + escapeValue + ") +\n'";
          }
          if (evaluateValue) {
            isEvaluating = true;
            source += "';\n" + evaluateValue + ";\n__p += '";
          }
          if (interpolateValue) {
            source += "' +\n((__t = (" + interpolateValue + ")) == null ? '' : __t) +\n'";
          }
          index = offset + match.length;

          // The JS engine embedded in Adobe products needs `match` returned in
          // order to produce the correct `offset` value.
          return match;
        });

        source += "';\n";

        // If `variable` is not specified wrap a with-statement around the generated
        // code to add the data object to the top of the scope chain.
        // Like with sourceURL, we take care to not check the option's prototype,
        // as this configuration is a code injection vector.
        var variable = hasOwnProperty.call(options, 'variable') && options.variable;
        if (!variable) {
          source = 'with (obj) {\n' + source + '\n}\n';
        }
        // Cleanup code by stripping empty strings.
        source = (isEvaluating ? source.replace(reEmptyStringLeading, '') : source)
          .replace(reEmptyStringMiddle, '$1')
          .replace(reEmptyStringTrailing, '$1;');

        // Frame code as the function body.
        source = 'function(' + (variable || 'obj') + ') {\n' +
          (variable
            ? ''
            : 'obj || (obj = {});\n'
          ) +
          "var __t, __p = ''" +
          (isEscaping
             ? ', __e = _.escape'
             : ''
          ) +
          (isEvaluating
            ? ', __j = Array.prototype.join;\n' +
              "function print() { __p += __j.call(arguments, '') }\n"
            : ';\n'
          ) +
          source +
          'return __p\n}';

        var result = attempt(function() {
          return Function(importsKeys, sourceURL + 'return ' + source)
            .apply(undefined$1, importsValues);
        });

        // Provide the compiled function's source by its `toString` method or
        // the `source` property as a convenience for inlining compiled templates.
        result.source = source;
        if (isError(result)) {
          throw result;
        }
        return result;
      }

      /**
       * Converts `string`, as a whole, to lower case just like
       * [String#toLowerCase](https://mdn.io/toLowerCase).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the lower cased string.
       * @example
       *
       * _.toLower('--Foo-Bar--');
       * // => '--foo-bar--'
       *
       * _.toLower('fooBar');
       * // => 'foobar'
       *
       * _.toLower('__FOO_BAR__');
       * // => '__foo_bar__'
       */
      function toLower(value) {
        return toString(value).toLowerCase();
      }

      /**
       * Converts `string`, as a whole, to upper case just like
       * [String#toUpperCase](https://mdn.io/toUpperCase).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the upper cased string.
       * @example
       *
       * _.toUpper('--foo-bar--');
       * // => '--FOO-BAR--'
       *
       * _.toUpper('fooBar');
       * // => 'FOOBAR'
       *
       * _.toUpper('__foo_bar__');
       * // => '__FOO_BAR__'
       */
      function toUpper(value) {
        return toString(value).toUpperCase();
      }

      /**
       * Removes leading and trailing whitespace or specified characters from `string`.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to trim.
       * @param {string} [chars=whitespace] The characters to trim.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {string} Returns the trimmed string.
       * @example
       *
       * _.trim('  abc  ');
       * // => 'abc'
       *
       * _.trim('-_-abc-_-', '_-');
       * // => 'abc'
       *
       * _.map(['  foo  ', '  bar  '], _.trim);
       * // => ['foo', 'bar']
       */
      function trim(string, chars, guard) {
        string = toString(string);
        if (string && (guard || chars === undefined$1)) {
          return string.replace(reTrim, '');
        }
        if (!string || !(chars = baseToString(chars))) {
          return string;
        }
        var strSymbols = stringToArray(string),
            chrSymbols = stringToArray(chars),
            start = charsStartIndex(strSymbols, chrSymbols),
            end = charsEndIndex(strSymbols, chrSymbols) + 1;

        return castSlice(strSymbols, start, end).join('');
      }

      /**
       * Removes trailing whitespace or specified characters from `string`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to trim.
       * @param {string} [chars=whitespace] The characters to trim.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {string} Returns the trimmed string.
       * @example
       *
       * _.trimEnd('  abc  ');
       * // => '  abc'
       *
       * _.trimEnd('-_-abc-_-', '_-');
       * // => '-_-abc'
       */
      function trimEnd(string, chars, guard) {
        string = toString(string);
        if (string && (guard || chars === undefined$1)) {
          return string.replace(reTrimEnd, '');
        }
        if (!string || !(chars = baseToString(chars))) {
          return string;
        }
        var strSymbols = stringToArray(string),
            end = charsEndIndex(strSymbols, stringToArray(chars)) + 1;

        return castSlice(strSymbols, 0, end).join('');
      }

      /**
       * Removes leading whitespace or specified characters from `string`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to trim.
       * @param {string} [chars=whitespace] The characters to trim.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {string} Returns the trimmed string.
       * @example
       *
       * _.trimStart('  abc  ');
       * // => 'abc  '
       *
       * _.trimStart('-_-abc-_-', '_-');
       * // => 'abc-_-'
       */
      function trimStart(string, chars, guard) {
        string = toString(string);
        if (string && (guard || chars === undefined$1)) {
          return string.replace(reTrimStart, '');
        }
        if (!string || !(chars = baseToString(chars))) {
          return string;
        }
        var strSymbols = stringToArray(string),
            start = charsStartIndex(strSymbols, stringToArray(chars));

        return castSlice(strSymbols, start).join('');
      }

      /**
       * Truncates `string` if it's longer than the given maximum string length.
       * The last characters of the truncated string are replaced with the omission
       * string which defaults to "...".
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to truncate.
       * @param {Object} [options={}] The options object.
       * @param {number} [options.length=30] The maximum string length.
       * @param {string} [options.omission='...'] The string to indicate text is omitted.
       * @param {RegExp|string} [options.separator] The separator pattern to truncate to.
       * @returns {string} Returns the truncated string.
       * @example
       *
       * _.truncate('hi-diddly-ho there, neighborino');
       * // => 'hi-diddly-ho there, neighbo...'
       *
       * _.truncate('hi-diddly-ho there, neighborino', {
       *   'length': 24,
       *   'separator': ' '
       * });
       * // => 'hi-diddly-ho there,...'
       *
       * _.truncate('hi-diddly-ho there, neighborino', {
       *   'length': 24,
       *   'separator': /,? +/
       * });
       * // => 'hi-diddly-ho there...'
       *
       * _.truncate('hi-diddly-ho there, neighborino', {
       *   'omission': ' [...]'
       * });
       * // => 'hi-diddly-ho there, neig [...]'
       */
      function truncate(string, options) {
        var length = DEFAULT_TRUNC_LENGTH,
            omission = DEFAULT_TRUNC_OMISSION;

        if (isObject(options)) {
          var separator = 'separator' in options ? options.separator : separator;
          length = 'length' in options ? toInteger(options.length) : length;
          omission = 'omission' in options ? baseToString(options.omission) : omission;
        }
        string = toString(string);

        var strLength = string.length;
        if (hasUnicode(string)) {
          var strSymbols = stringToArray(string);
          strLength = strSymbols.length;
        }
        if (length >= strLength) {
          return string;
        }
        var end = length - stringSize(omission);
        if (end < 1) {
          return omission;
        }
        var result = strSymbols
          ? castSlice(strSymbols, 0, end).join('')
          : string.slice(0, end);

        if (separator === undefined$1) {
          return result + omission;
        }
        if (strSymbols) {
          end += (result.length - end);
        }
        if (isRegExp(separator)) {
          if (string.slice(end).search(separator)) {
            var match,
                substring = result;

            if (!separator.global) {
              separator = RegExp(separator.source, toString(reFlags.exec(separator)) + 'g');
            }
            separator.lastIndex = 0;
            while ((match = separator.exec(substring))) {
              var newEnd = match.index;
            }
            result = result.slice(0, newEnd === undefined$1 ? end : newEnd);
          }
        } else if (string.indexOf(baseToString(separator), end) != end) {
          var index = result.lastIndexOf(separator);
          if (index > -1) {
            result = result.slice(0, index);
          }
        }
        return result + omission;
      }

      /**
       * The inverse of `_.escape`; this method converts the HTML entities
       * `&amp;`, `&lt;`, `&gt;`, `&quot;`, and `&#39;` in `string` to
       * their corresponding characters.
       *
       * **Note:** No other HTML entities are unescaped. To unescape additional
       * HTML entities use a third-party library like [_he_](https://mths.be/he).
       *
       * @static
       * @memberOf _
       * @since 0.6.0
       * @category String
       * @param {string} [string=''] The string to unescape.
       * @returns {string} Returns the unescaped string.
       * @example
       *
       * _.unescape('fred, barney, &amp; pebbles');
       * // => 'fred, barney, & pebbles'
       */
      function unescape(string) {
        string = toString(string);
        return (string && reHasEscapedHtml.test(string))
          ? string.replace(reEscapedHtml, unescapeHtmlChar)
          : string;
      }

      /**
       * Converts `string`, as space separated words, to upper case.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the upper cased string.
       * @example
       *
       * _.upperCase('--foo-bar');
       * // => 'FOO BAR'
       *
       * _.upperCase('fooBar');
       * // => 'FOO BAR'
       *
       * _.upperCase('__foo_bar__');
       * // => 'FOO BAR'
       */
      var upperCase = createCompounder(function(result, word, index) {
        return result + (index ? ' ' : '') + word.toUpperCase();
      });

      /**
       * Converts the first character of `string` to upper case.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category String
       * @param {string} [string=''] The string to convert.
       * @returns {string} Returns the converted string.
       * @example
       *
       * _.upperFirst('fred');
       * // => 'Fred'
       *
       * _.upperFirst('FRED');
       * // => 'FRED'
       */
      var upperFirst = createCaseFirst('toUpperCase');

      /**
       * Splits `string` into an array of its words.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category String
       * @param {string} [string=''] The string to inspect.
       * @param {RegExp|string} [pattern] The pattern to match words.
       * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
       * @returns {Array} Returns the words of `string`.
       * @example
       *
       * _.words('fred, barney, & pebbles');
       * // => ['fred', 'barney', 'pebbles']
       *
       * _.words('fred, barney, & pebbles', /[^, ]+/g);
       * // => ['fred', 'barney', '&', 'pebbles']
       */
      function words(string, pattern, guard) {
        string = toString(string);
        pattern = guard ? undefined$1 : pattern;

        if (pattern === undefined$1) {
          return hasUnicodeWord(string) ? unicodeWords(string) : asciiWords(string);
        }
        return string.match(pattern) || [];
      }

      /*------------------------------------------------------------------------*/

      /**
       * Attempts to invoke `func`, returning either the result or the caught error
       * object. Any additional arguments are provided to `func` when it's invoked.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Util
       * @param {Function} func The function to attempt.
       * @param {...*} [args] The arguments to invoke `func` with.
       * @returns {*} Returns the `func` result or error object.
       * @example
       *
       * // Avoid throwing errors for invalid selectors.
       * var elements = _.attempt(function(selector) {
       *   return document.querySelectorAll(selector);
       * }, '>_>');
       *
       * if (_.isError(elements)) {
       *   elements = [];
       * }
       */
      var attempt = baseRest(function(func, args) {
        try {
          return apply(func, undefined$1, args);
        } catch (e) {
          return isError(e) ? e : new Error(e);
        }
      });

      /**
       * Binds methods of an object to the object itself, overwriting the existing
       * method.
       *
       * **Note:** This method doesn't set the "length" property of bound functions.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @param {Object} object The object to bind and assign the bound methods to.
       * @param {...(string|string[])} methodNames The object method names to bind.
       * @returns {Object} Returns `object`.
       * @example
       *
       * var view = {
       *   'label': 'docs',
       *   'click': function() {
       *     console.log('clicked ' + this.label);
       *   }
       * };
       *
       * _.bindAll(view, ['click']);
       * jQuery(element).on('click', view.click);
       * // => Logs 'clicked docs' when clicked.
       */
      var bindAll = flatRest(function(object, methodNames) {
        arrayEach(methodNames, function(key) {
          key = toKey(key);
          baseAssignValue(object, key, bind(object[key], object));
        });
        return object;
      });

      /**
       * Creates a function that iterates over `pairs` and invokes the corresponding
       * function of the first predicate to return truthy. The predicate-function
       * pairs are invoked with the `this` binding and arguments of the created
       * function.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {Array} pairs The predicate-function pairs.
       * @returns {Function} Returns the new composite function.
       * @example
       *
       * var func = _.cond([
       *   [_.matches({ 'a': 1 }),           _.constant('matches A')],
       *   [_.conforms({ 'b': _.isNumber }), _.constant('matches B')],
       *   [_.stubTrue,                      _.constant('no match')]
       * ]);
       *
       * func({ 'a': 1, 'b': 2 });
       * // => 'matches A'
       *
       * func({ 'a': 0, 'b': 1 });
       * // => 'matches B'
       *
       * func({ 'a': '1', 'b': '2' });
       * // => 'no match'
       */
      function cond(pairs) {
        var length = pairs == null ? 0 : pairs.length,
            toIteratee = getIteratee();

        pairs = !length ? [] : arrayMap(pairs, function(pair) {
          if (typeof pair[1] != 'function') {
            throw new TypeError(FUNC_ERROR_TEXT);
          }
          return [toIteratee(pair[0]), pair[1]];
        });

        return baseRest(function(args) {
          var index = -1;
          while (++index < length) {
            var pair = pairs[index];
            if (apply(pair[0], this, args)) {
              return apply(pair[1], this, args);
            }
          }
        });
      }

      /**
       * Creates a function that invokes the predicate properties of `source` with
       * the corresponding property values of a given object, returning `true` if
       * all predicates return truthy, else `false`.
       *
       * **Note:** The created function is equivalent to `_.conformsTo` with
       * `source` partially applied.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {Object} source The object of property predicates to conform to.
       * @returns {Function} Returns the new spec function.
       * @example
       *
       * var objects = [
       *   { 'a': 2, 'b': 1 },
       *   { 'a': 1, 'b': 2 }
       * ];
       *
       * _.filter(objects, _.conforms({ 'b': function(n) { return n > 1; } }));
       * // => [{ 'a': 1, 'b': 2 }]
       */
      function conforms(source) {
        return baseConforms(baseClone(source, CLONE_DEEP_FLAG));
      }

      /**
       * Creates a function that returns `value`.
       *
       * @static
       * @memberOf _
       * @since 2.4.0
       * @category Util
       * @param {*} value The value to return from the new function.
       * @returns {Function} Returns the new constant function.
       * @example
       *
       * var objects = _.times(2, _.constant({ 'a': 1 }));
       *
       * console.log(objects);
       * // => [{ 'a': 1 }, { 'a': 1 }]
       *
       * console.log(objects[0] === objects[1]);
       * // => true
       */
      function constant(value) {
        return function() {
          return value;
        };
      }

      /**
       * Checks `value` to determine whether a default value should be returned in
       * its place. The `defaultValue` is returned if `value` is `NaN`, `null`,
       * or `undefined`.
       *
       * @static
       * @memberOf _
       * @since 4.14.0
       * @category Util
       * @param {*} value The value to check.
       * @param {*} defaultValue The default value.
       * @returns {*} Returns the resolved value.
       * @example
       *
       * _.defaultTo(1, 10);
       * // => 1
       *
       * _.defaultTo(undefined, 10);
       * // => 10
       */
      function defaultTo(value, defaultValue) {
        return (value == null || value !== value) ? defaultValue : value;
      }

      /**
       * Creates a function that returns the result of invoking the given functions
       * with the `this` binding of the created function, where each successive
       * invocation is supplied the return value of the previous.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Util
       * @param {...(Function|Function[])} [funcs] The functions to invoke.
       * @returns {Function} Returns the new composite function.
       * @see _.flowRight
       * @example
       *
       * function square(n) {
       *   return n * n;
       * }
       *
       * var addSquare = _.flow([_.add, square]);
       * addSquare(1, 2);
       * // => 9
       */
      var flow = createFlow();

      /**
       * This method is like `_.flow` except that it creates a function that
       * invokes the given functions from right to left.
       *
       * @static
       * @since 3.0.0
       * @memberOf _
       * @category Util
       * @param {...(Function|Function[])} [funcs] The functions to invoke.
       * @returns {Function} Returns the new composite function.
       * @see _.flow
       * @example
       *
       * function square(n) {
       *   return n * n;
       * }
       *
       * var addSquare = _.flowRight([square, _.add]);
       * addSquare(1, 2);
       * // => 9
       */
      var flowRight = createFlow(true);

      /**
       * This method returns the first argument it receives.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @param {*} value Any value.
       * @returns {*} Returns `value`.
       * @example
       *
       * var object = { 'a': 1 };
       *
       * console.log(_.identity(object) === object);
       * // => true
       */
      function identity(value) {
        return value;
      }

      /**
       * Creates a function that invokes `func` with the arguments of the created
       * function. If `func` is a property name, the created function returns the
       * property value for a given element. If `func` is an array or object, the
       * created function returns `true` for elements that contain the equivalent
       * source properties, otherwise it returns `false`.
       *
       * @static
       * @since 4.0.0
       * @memberOf _
       * @category Util
       * @param {*} [func=_.identity] The value to convert to a callback.
       * @returns {Function} Returns the callback.
       * @example
       *
       * var users = [
       *   { 'user': 'barney', 'age': 36, 'active': true },
       *   { 'user': 'fred',   'age': 40, 'active': false }
       * ];
       *
       * // The `_.matches` iteratee shorthand.
       * _.filter(users, _.iteratee({ 'user': 'barney', 'active': true }));
       * // => [{ 'user': 'barney', 'age': 36, 'active': true }]
       *
       * // The `_.matchesProperty` iteratee shorthand.
       * _.filter(users, _.iteratee(['user', 'fred']));
       * // => [{ 'user': 'fred', 'age': 40 }]
       *
       * // The `_.property` iteratee shorthand.
       * _.map(users, _.iteratee('user'));
       * // => ['barney', 'fred']
       *
       * // Create custom iteratee shorthands.
       * _.iteratee = _.wrap(_.iteratee, function(iteratee, func) {
       *   return !_.isRegExp(func) ? iteratee(func) : function(string) {
       *     return func.test(string);
       *   };
       * });
       *
       * _.filter(['abc', 'def'], /ef/);
       * // => ['def']
       */
      function iteratee(func) {
        return baseIteratee(typeof func == 'function' ? func : baseClone(func, CLONE_DEEP_FLAG));
      }

      /**
       * Creates a function that performs a partial deep comparison between a given
       * object and `source`, returning `true` if the given object has equivalent
       * property values, else `false`.
       *
       * **Note:** The created function is equivalent to `_.isMatch` with `source`
       * partially applied.
       *
       * Partial comparisons will match empty array and empty object `source`
       * values against any array or object value, respectively. See `_.isEqual`
       * for a list of supported value comparisons.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Util
       * @param {Object} source The object of property values to match.
       * @returns {Function} Returns the new spec function.
       * @example
       *
       * var objects = [
       *   { 'a': 1, 'b': 2, 'c': 3 },
       *   { 'a': 4, 'b': 5, 'c': 6 }
       * ];
       *
       * _.filter(objects, _.matches({ 'a': 4, 'c': 6 }));
       * // => [{ 'a': 4, 'b': 5, 'c': 6 }]
       */
      function matches(source) {
        return baseMatches(baseClone(source, CLONE_DEEP_FLAG));
      }

      /**
       * Creates a function that performs a partial deep comparison between the
       * value at `path` of a given object to `srcValue`, returning `true` if the
       * object value is equivalent, else `false`.
       *
       * **Note:** Partial comparisons will match empty array and empty object
       * `srcValue` values against any array or object value, respectively. See
       * `_.isEqual` for a list of supported value comparisons.
       *
       * @static
       * @memberOf _
       * @since 3.2.0
       * @category Util
       * @param {Array|string} path The path of the property to get.
       * @param {*} srcValue The value to match.
       * @returns {Function} Returns the new spec function.
       * @example
       *
       * var objects = [
       *   { 'a': 1, 'b': 2, 'c': 3 },
       *   { 'a': 4, 'b': 5, 'c': 6 }
       * ];
       *
       * _.find(objects, _.matchesProperty('a', 4));
       * // => { 'a': 4, 'b': 5, 'c': 6 }
       */
      function matchesProperty(path, srcValue) {
        return baseMatchesProperty(path, baseClone(srcValue, CLONE_DEEP_FLAG));
      }

      /**
       * Creates a function that invokes the method at `path` of a given object.
       * Any additional arguments are provided to the invoked method.
       *
       * @static
       * @memberOf _
       * @since 3.7.0
       * @category Util
       * @param {Array|string} path The path of the method to invoke.
       * @param {...*} [args] The arguments to invoke the method with.
       * @returns {Function} Returns the new invoker function.
       * @example
       *
       * var objects = [
       *   { 'a': { 'b': _.constant(2) } },
       *   { 'a': { 'b': _.constant(1) } }
       * ];
       *
       * _.map(objects, _.method('a.b'));
       * // => [2, 1]
       *
       * _.map(objects, _.method(['a', 'b']));
       * // => [2, 1]
       */
      var method = baseRest(function(path, args) {
        return function(object) {
          return baseInvoke(object, path, args);
        };
      });

      /**
       * The opposite of `_.method`; this method creates a function that invokes
       * the method at a given path of `object`. Any additional arguments are
       * provided to the invoked method.
       *
       * @static
       * @memberOf _
       * @since 3.7.0
       * @category Util
       * @param {Object} object The object to query.
       * @param {...*} [args] The arguments to invoke the method with.
       * @returns {Function} Returns the new invoker function.
       * @example
       *
       * var array = _.times(3, _.constant),
       *     object = { 'a': array, 'b': array, 'c': array };
       *
       * _.map(['a[2]', 'c[0]'], _.methodOf(object));
       * // => [2, 0]
       *
       * _.map([['a', '2'], ['c', '0']], _.methodOf(object));
       * // => [2, 0]
       */
      var methodOf = baseRest(function(object, args) {
        return function(path) {
          return baseInvoke(object, path, args);
        };
      });

      /**
       * Adds all own enumerable string keyed function properties of a source
       * object to the destination object. If `object` is a function, then methods
       * are added to its prototype as well.
       *
       * **Note:** Use `_.runInContext` to create a pristine `lodash` function to
       * avoid conflicts caused by modifying the original.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @param {Function|Object} [object=lodash] The destination object.
       * @param {Object} source The object of functions to add.
       * @param {Object} [options={}] The options object.
       * @param {boolean} [options.chain=true] Specify whether mixins are chainable.
       * @returns {Function|Object} Returns `object`.
       * @example
       *
       * function vowels(string) {
       *   return _.filter(string, function(v) {
       *     return /[aeiou]/i.test(v);
       *   });
       * }
       *
       * _.mixin({ 'vowels': vowels });
       * _.vowels('fred');
       * // => ['e']
       *
       * _('fred').vowels().value();
       * // => ['e']
       *
       * _.mixin({ 'vowels': vowels }, { 'chain': false });
       * _('fred').vowels();
       * // => ['e']
       */
      function mixin(object, source, options) {
        var props = keys(source),
            methodNames = baseFunctions(source, props);

        if (options == null &&
            !(isObject(source) && (methodNames.length || !props.length))) {
          options = source;
          source = object;
          object = this;
          methodNames = baseFunctions(source, keys(source));
        }
        var chain = !(isObject(options) && 'chain' in options) || !!options.chain,
            isFunc = isFunction(object);

        arrayEach(methodNames, function(methodName) {
          var func = source[methodName];
          object[methodName] = func;
          if (isFunc) {
            object.prototype[methodName] = function() {
              var chainAll = this.__chain__;
              if (chain || chainAll) {
                var result = object(this.__wrapped__),
                    actions = result.__actions__ = copyArray(this.__actions__);

                actions.push({ 'func': func, 'args': arguments, 'thisArg': object });
                result.__chain__ = chainAll;
                return result;
              }
              return func.apply(object, arrayPush([this.value()], arguments));
            };
          }
        });

        return object;
      }

      /**
       * Reverts the `_` variable to its previous value and returns a reference to
       * the `lodash` function.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @returns {Function} Returns the `lodash` function.
       * @example
       *
       * var lodash = _.noConflict();
       */
      function noConflict() {
        if (root._ === this) {
          root._ = oldDash;
        }
        return this;
      }

      /**
       * This method returns `undefined`.
       *
       * @static
       * @memberOf _
       * @since 2.3.0
       * @category Util
       * @example
       *
       * _.times(2, _.noop);
       * // => [undefined, undefined]
       */
      function noop() {
        // No operation performed.
      }

      /**
       * Creates a function that gets the argument at index `n`. If `n` is negative,
       * the nth argument from the end is returned.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {number} [n=0] The index of the argument to return.
       * @returns {Function} Returns the new pass-thru function.
       * @example
       *
       * var func = _.nthArg(1);
       * func('a', 'b', 'c', 'd');
       * // => 'b'
       *
       * var func = _.nthArg(-2);
       * func('a', 'b', 'c', 'd');
       * // => 'c'
       */
      function nthArg(n) {
        n = toInteger(n);
        return baseRest(function(args) {
          return baseNth(args, n);
        });
      }

      /**
       * Creates a function that invokes `iteratees` with the arguments it receives
       * and returns their results.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {...(Function|Function[])} [iteratees=[_.identity]]
       *  The iteratees to invoke.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var func = _.over([Math.max, Math.min]);
       *
       * func(1, 2, 3, 4);
       * // => [4, 1]
       */
      var over = createOver(arrayMap);

      /**
       * Creates a function that checks if **all** of the `predicates` return
       * truthy when invoked with the arguments it receives.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {...(Function|Function[])} [predicates=[_.identity]]
       *  The predicates to check.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var func = _.overEvery([Boolean, isFinite]);
       *
       * func('1');
       * // => true
       *
       * func(null);
       * // => false
       *
       * func(NaN);
       * // => false
       */
      var overEvery = createOver(arrayEvery);

      /**
       * Creates a function that checks if **any** of the `predicates` return
       * truthy when invoked with the arguments it receives.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {...(Function|Function[])} [predicates=[_.identity]]
       *  The predicates to check.
       * @returns {Function} Returns the new function.
       * @example
       *
       * var func = _.overSome([Boolean, isFinite]);
       *
       * func('1');
       * // => true
       *
       * func(null);
       * // => true
       *
       * func(NaN);
       * // => false
       */
      var overSome = createOver(arraySome);

      /**
       * Creates a function that returns the value at `path` of a given object.
       *
       * @static
       * @memberOf _
       * @since 2.4.0
       * @category Util
       * @param {Array|string} path The path of the property to get.
       * @returns {Function} Returns the new accessor function.
       * @example
       *
       * var objects = [
       *   { 'a': { 'b': 2 } },
       *   { 'a': { 'b': 1 } }
       * ];
       *
       * _.map(objects, _.property('a.b'));
       * // => [2, 1]
       *
       * _.map(_.sortBy(objects, _.property(['a', 'b'])), 'a.b');
       * // => [1, 2]
       */
      function property(path) {
        return isKey(path) ? baseProperty(toKey(path)) : basePropertyDeep(path);
      }

      /**
       * The opposite of `_.property`; this method creates a function that returns
       * the value at a given path of `object`.
       *
       * @static
       * @memberOf _
       * @since 3.0.0
       * @category Util
       * @param {Object} object The object to query.
       * @returns {Function} Returns the new accessor function.
       * @example
       *
       * var array = [0, 1, 2],
       *     object = { 'a': array, 'b': array, 'c': array };
       *
       * _.map(['a[2]', 'c[0]'], _.propertyOf(object));
       * // => [2, 0]
       *
       * _.map([['a', '2'], ['c', '0']], _.propertyOf(object));
       * // => [2, 0]
       */
      function propertyOf(object) {
        return function(path) {
          return object == null ? undefined$1 : baseGet(object, path);
        };
      }

      /**
       * Creates an array of numbers (positive and/or negative) progressing from
       * `start` up to, but not including, `end`. A step of `-1` is used if a negative
       * `start` is specified without an `end` or `step`. If `end` is not specified,
       * it's set to `start` with `start` then set to `0`.
       *
       * **Note:** JavaScript follows the IEEE-754 standard for resolving
       * floating-point values which can produce unexpected results.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @param {number} [start=0] The start of the range.
       * @param {number} end The end of the range.
       * @param {number} [step=1] The value to increment or decrement by.
       * @returns {Array} Returns the range of numbers.
       * @see _.inRange, _.rangeRight
       * @example
       *
       * _.range(4);
       * // => [0, 1, 2, 3]
       *
       * _.range(-4);
       * // => [0, -1, -2, -3]
       *
       * _.range(1, 5);
       * // => [1, 2, 3, 4]
       *
       * _.range(0, 20, 5);
       * // => [0, 5, 10, 15]
       *
       * _.range(0, -4, -1);
       * // => [0, -1, -2, -3]
       *
       * _.range(1, 4, 0);
       * // => [1, 1, 1]
       *
       * _.range(0);
       * // => []
       */
      var range = createRange();

      /**
       * This method is like `_.range` except that it populates values in
       * descending order.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {number} [start=0] The start of the range.
       * @param {number} end The end of the range.
       * @param {number} [step=1] The value to increment or decrement by.
       * @returns {Array} Returns the range of numbers.
       * @see _.inRange, _.range
       * @example
       *
       * _.rangeRight(4);
       * // => [3, 2, 1, 0]
       *
       * _.rangeRight(-4);
       * // => [-3, -2, -1, 0]
       *
       * _.rangeRight(1, 5);
       * // => [4, 3, 2, 1]
       *
       * _.rangeRight(0, 20, 5);
       * // => [15, 10, 5, 0]
       *
       * _.rangeRight(0, -4, -1);
       * // => [-3, -2, -1, 0]
       *
       * _.rangeRight(1, 4, 0);
       * // => [1, 1, 1]
       *
       * _.rangeRight(0);
       * // => []
       */
      var rangeRight = createRange(true);

      /**
       * This method returns a new empty array.
       *
       * @static
       * @memberOf _
       * @since 4.13.0
       * @category Util
       * @returns {Array} Returns the new empty array.
       * @example
       *
       * var arrays = _.times(2, _.stubArray);
       *
       * console.log(arrays);
       * // => [[], []]
       *
       * console.log(arrays[0] === arrays[1]);
       * // => false
       */
      function stubArray() {
        return [];
      }

      /**
       * This method returns `false`.
       *
       * @static
       * @memberOf _
       * @since 4.13.0
       * @category Util
       * @returns {boolean} Returns `false`.
       * @example
       *
       * _.times(2, _.stubFalse);
       * // => [false, false]
       */
      function stubFalse() {
        return false;
      }

      /**
       * This method returns a new empty object.
       *
       * @static
       * @memberOf _
       * @since 4.13.0
       * @category Util
       * @returns {Object} Returns the new empty object.
       * @example
       *
       * var objects = _.times(2, _.stubObject);
       *
       * console.log(objects);
       * // => [{}, {}]
       *
       * console.log(objects[0] === objects[1]);
       * // => false
       */
      function stubObject() {
        return {};
      }

      /**
       * This method returns an empty string.
       *
       * @static
       * @memberOf _
       * @since 4.13.0
       * @category Util
       * @returns {string} Returns the empty string.
       * @example
       *
       * _.times(2, _.stubString);
       * // => ['', '']
       */
      function stubString() {
        return '';
      }

      /**
       * This method returns `true`.
       *
       * @static
       * @memberOf _
       * @since 4.13.0
       * @category Util
       * @returns {boolean} Returns `true`.
       * @example
       *
       * _.times(2, _.stubTrue);
       * // => [true, true]
       */
      function stubTrue() {
        return true;
      }

      /**
       * Invokes the iteratee `n` times, returning an array of the results of
       * each invocation. The iteratee is invoked with one argument; (index).
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @param {number} n The number of times to invoke `iteratee`.
       * @param {Function} [iteratee=_.identity] The function invoked per iteration.
       * @returns {Array} Returns the array of results.
       * @example
       *
       * _.times(3, String);
       * // => ['0', '1', '2']
       *
       *  _.times(4, _.constant(0));
       * // => [0, 0, 0, 0]
       */
      function times(n, iteratee) {
        n = toInteger(n);
        if (n < 1 || n > MAX_SAFE_INTEGER) {
          return [];
        }
        var index = MAX_ARRAY_LENGTH,
            length = nativeMin(n, MAX_ARRAY_LENGTH);

        iteratee = getIteratee(iteratee);
        n -= MAX_ARRAY_LENGTH;

        var result = baseTimes(length, iteratee);
        while (++index < n) {
          iteratee(index);
        }
        return result;
      }

      /**
       * Converts `value` to a property path array.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Util
       * @param {*} value The value to convert.
       * @returns {Array} Returns the new property path array.
       * @example
       *
       * _.toPath('a.b.c');
       * // => ['a', 'b', 'c']
       *
       * _.toPath('a[0].b.c');
       * // => ['a', '0', 'b', 'c']
       */
      function toPath(value) {
        if (isArray(value)) {
          return arrayMap(value, toKey);
        }
        return isSymbol(value) ? [value] : copyArray(stringToPath(toString(value)));
      }

      /**
       * Generates a unique ID. If `prefix` is given, the ID is appended to it.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Util
       * @param {string} [prefix=''] The value to prefix the ID with.
       * @returns {string} Returns the unique ID.
       * @example
       *
       * _.uniqueId('contact_');
       * // => 'contact_104'
       *
       * _.uniqueId();
       * // => '105'
       */
      function uniqueId(prefix) {
        var id = ++idCounter;
        return toString(prefix) + id;
      }

      /*------------------------------------------------------------------------*/

      /**
       * Adds two numbers.
       *
       * @static
       * @memberOf _
       * @since 3.4.0
       * @category Math
       * @param {number} augend The first number in an addition.
       * @param {number} addend The second number in an addition.
       * @returns {number} Returns the total.
       * @example
       *
       * _.add(6, 4);
       * // => 10
       */
      var add = createMathOperation(function(augend, addend) {
        return augend + addend;
      }, 0);

      /**
       * Computes `number` rounded up to `precision`.
       *
       * @static
       * @memberOf _
       * @since 3.10.0
       * @category Math
       * @param {number} number The number to round up.
       * @param {number} [precision=0] The precision to round up to.
       * @returns {number} Returns the rounded up number.
       * @example
       *
       * _.ceil(4.006);
       * // => 5
       *
       * _.ceil(6.004, 2);
       * // => 6.01
       *
       * _.ceil(6040, -2);
       * // => 6100
       */
      var ceil = createRound('ceil');

      /**
       * Divide two numbers.
       *
       * @static
       * @memberOf _
       * @since 4.7.0
       * @category Math
       * @param {number} dividend The first number in a division.
       * @param {number} divisor The second number in a division.
       * @returns {number} Returns the quotient.
       * @example
       *
       * _.divide(6, 4);
       * // => 1.5
       */
      var divide = createMathOperation(function(dividend, divisor) {
        return dividend / divisor;
      }, 1);

      /**
       * Computes `number` rounded down to `precision`.
       *
       * @static
       * @memberOf _
       * @since 3.10.0
       * @category Math
       * @param {number} number The number to round down.
       * @param {number} [precision=0] The precision to round down to.
       * @returns {number} Returns the rounded down number.
       * @example
       *
       * _.floor(4.006);
       * // => 4
       *
       * _.floor(0.046, 2);
       * // => 0.04
       *
       * _.floor(4060, -2);
       * // => 4000
       */
      var floor = createRound('floor');

      /**
       * Computes the maximum value of `array`. If `array` is empty or falsey,
       * `undefined` is returned.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Math
       * @param {Array} array The array to iterate over.
       * @returns {*} Returns the maximum value.
       * @example
       *
       * _.max([4, 2, 8, 6]);
       * // => 8
       *
       * _.max([]);
       * // => undefined
       */
      function max(array) {
        return (array && array.length)
          ? baseExtremum(array, identity, baseGt)
          : undefined$1;
      }

      /**
       * This method is like `_.max` except that it accepts `iteratee` which is
       * invoked for each element in `array` to generate the criterion by which
       * the value is ranked. The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Math
       * @param {Array} array The array to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {*} Returns the maximum value.
       * @example
       *
       * var objects = [{ 'n': 1 }, { 'n': 2 }];
       *
       * _.maxBy(objects, function(o) { return o.n; });
       * // => { 'n': 2 }
       *
       * // The `_.property` iteratee shorthand.
       * _.maxBy(objects, 'n');
       * // => { 'n': 2 }
       */
      function maxBy(array, iteratee) {
        return (array && array.length)
          ? baseExtremum(array, getIteratee(iteratee, 2), baseGt)
          : undefined$1;
      }

      /**
       * Computes the mean of the values in `array`.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Math
       * @param {Array} array The array to iterate over.
       * @returns {number} Returns the mean.
       * @example
       *
       * _.mean([4, 2, 8, 6]);
       * // => 5
       */
      function mean(array) {
        return baseMean(array, identity);
      }

      /**
       * This method is like `_.mean` except that it accepts `iteratee` which is
       * invoked for each element in `array` to generate the value to be averaged.
       * The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.7.0
       * @category Math
       * @param {Array} array The array to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {number} Returns the mean.
       * @example
       *
       * var objects = [{ 'n': 4 }, { 'n': 2 }, { 'n': 8 }, { 'n': 6 }];
       *
       * _.meanBy(objects, function(o) { return o.n; });
       * // => 5
       *
       * // The `_.property` iteratee shorthand.
       * _.meanBy(objects, 'n');
       * // => 5
       */
      function meanBy(array, iteratee) {
        return baseMean(array, getIteratee(iteratee, 2));
      }

      /**
       * Computes the minimum value of `array`. If `array` is empty or falsey,
       * `undefined` is returned.
       *
       * @static
       * @since 0.1.0
       * @memberOf _
       * @category Math
       * @param {Array} array The array to iterate over.
       * @returns {*} Returns the minimum value.
       * @example
       *
       * _.min([4, 2, 8, 6]);
       * // => 2
       *
       * _.min([]);
       * // => undefined
       */
      function min(array) {
        return (array && array.length)
          ? baseExtremum(array, identity, baseLt)
          : undefined$1;
      }

      /**
       * This method is like `_.min` except that it accepts `iteratee` which is
       * invoked for each element in `array` to generate the criterion by which
       * the value is ranked. The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Math
       * @param {Array} array The array to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {*} Returns the minimum value.
       * @example
       *
       * var objects = [{ 'n': 1 }, { 'n': 2 }];
       *
       * _.minBy(objects, function(o) { return o.n; });
       * // => { 'n': 1 }
       *
       * // The `_.property` iteratee shorthand.
       * _.minBy(objects, 'n');
       * // => { 'n': 1 }
       */
      function minBy(array, iteratee) {
        return (array && array.length)
          ? baseExtremum(array, getIteratee(iteratee, 2), baseLt)
          : undefined$1;
      }

      /**
       * Multiply two numbers.
       *
       * @static
       * @memberOf _
       * @since 4.7.0
       * @category Math
       * @param {number} multiplier The first number in a multiplication.
       * @param {number} multiplicand The second number in a multiplication.
       * @returns {number} Returns the product.
       * @example
       *
       * _.multiply(6, 4);
       * // => 24
       */
      var multiply = createMathOperation(function(multiplier, multiplicand) {
        return multiplier * multiplicand;
      }, 1);

      /**
       * Computes `number` rounded to `precision`.
       *
       * @static
       * @memberOf _
       * @since 3.10.0
       * @category Math
       * @param {number} number The number to round.
       * @param {number} [precision=0] The precision to round to.
       * @returns {number} Returns the rounded number.
       * @example
       *
       * _.round(4.006);
       * // => 4
       *
       * _.round(4.006, 2);
       * // => 4.01
       *
       * _.round(4060, -2);
       * // => 4100
       */
      var round = createRound('round');

      /**
       * Subtract two numbers.
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Math
       * @param {number} minuend The first number in a subtraction.
       * @param {number} subtrahend The second number in a subtraction.
       * @returns {number} Returns the difference.
       * @example
       *
       * _.subtract(6, 4);
       * // => 2
       */
      var subtract = createMathOperation(function(minuend, subtrahend) {
        return minuend - subtrahend;
      }, 0);

      /**
       * Computes the sum of the values in `array`.
       *
       * @static
       * @memberOf _
       * @since 3.4.0
       * @category Math
       * @param {Array} array The array to iterate over.
       * @returns {number} Returns the sum.
       * @example
       *
       * _.sum([4, 2, 8, 6]);
       * // => 20
       */
      function sum(array) {
        return (array && array.length)
          ? baseSum(array, identity)
          : 0;
      }

      /**
       * This method is like `_.sum` except that it accepts `iteratee` which is
       * invoked for each element in `array` to generate the value to be summed.
       * The iteratee is invoked with one argument: (value).
       *
       * @static
       * @memberOf _
       * @since 4.0.0
       * @category Math
       * @param {Array} array The array to iterate over.
       * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
       * @returns {number} Returns the sum.
       * @example
       *
       * var objects = [{ 'n': 4 }, { 'n': 2 }, { 'n': 8 }, { 'n': 6 }];
       *
       * _.sumBy(objects, function(o) { return o.n; });
       * // => 20
       *
       * // The `_.property` iteratee shorthand.
       * _.sumBy(objects, 'n');
       * // => 20
       */
      function sumBy(array, iteratee) {
        return (array && array.length)
          ? baseSum(array, getIteratee(iteratee, 2))
          : 0;
      }

      /*------------------------------------------------------------------------*/

      // Add methods that return wrapped values in chain sequences.
      lodash.after = after;
      lodash.ary = ary;
      lodash.assign = assign;
      lodash.assignIn = assignIn;
      lodash.assignInWith = assignInWith;
      lodash.assignWith = assignWith;
      lodash.at = at;
      lodash.before = before;
      lodash.bind = bind;
      lodash.bindAll = bindAll;
      lodash.bindKey = bindKey;
      lodash.castArray = castArray;
      lodash.chain = chain;
      lodash.chunk = chunk;
      lodash.compact = compact;
      lodash.concat = concat;
      lodash.cond = cond;
      lodash.conforms = conforms;
      lodash.constant = constant;
      lodash.countBy = countBy;
      lodash.create = create;
      lodash.curry = curry;
      lodash.curryRight = curryRight;
      lodash.debounce = debounce;
      lodash.defaults = defaults;
      lodash.defaultsDeep = defaultsDeep;
      lodash.defer = defer;
      lodash.delay = delay;
      lodash.difference = difference;
      lodash.differenceBy = differenceBy;
      lodash.differenceWith = differenceWith;
      lodash.drop = drop;
      lodash.dropRight = dropRight;
      lodash.dropRightWhile = dropRightWhile;
      lodash.dropWhile = dropWhile;
      lodash.fill = fill;
      lodash.filter = filter;
      lodash.flatMap = flatMap;
      lodash.flatMapDeep = flatMapDeep;
      lodash.flatMapDepth = flatMapDepth;
      lodash.flatten = flatten;
      lodash.flattenDeep = flattenDeep;
      lodash.flattenDepth = flattenDepth;
      lodash.flip = flip;
      lodash.flow = flow;
      lodash.flowRight = flowRight;
      lodash.fromPairs = fromPairs;
      lodash.functions = functions;
      lodash.functionsIn = functionsIn;
      lodash.groupBy = groupBy;
      lodash.initial = initial;
      lodash.intersection = intersection;
      lodash.intersectionBy = intersectionBy;
      lodash.intersectionWith = intersectionWith;
      lodash.invert = invert;
      lodash.invertBy = invertBy;
      lodash.invokeMap = invokeMap;
      lodash.iteratee = iteratee;
      lodash.keyBy = keyBy;
      lodash.keys = keys;
      lodash.keysIn = keysIn;
      lodash.map = map;
      lodash.mapKeys = mapKeys;
      lodash.mapValues = mapValues;
      lodash.matches = matches;
      lodash.matchesProperty = matchesProperty;
      lodash.memoize = memoize;
      lodash.merge = merge;
      lodash.mergeWith = mergeWith;
      lodash.method = method;
      lodash.methodOf = methodOf;
      lodash.mixin = mixin;
      lodash.negate = negate;
      lodash.nthArg = nthArg;
      lodash.omit = omit;
      lodash.omitBy = omitBy;
      lodash.once = once;
      lodash.orderBy = orderBy;
      lodash.over = over;
      lodash.overArgs = overArgs;
      lodash.overEvery = overEvery;
      lodash.overSome = overSome;
      lodash.partial = partial;
      lodash.partialRight = partialRight;
      lodash.partition = partition;
      lodash.pick = pick;
      lodash.pickBy = pickBy;
      lodash.property = property;
      lodash.propertyOf = propertyOf;
      lodash.pull = pull;
      lodash.pullAll = pullAll;
      lodash.pullAllBy = pullAllBy;
      lodash.pullAllWith = pullAllWith;
      lodash.pullAt = pullAt;
      lodash.range = range;
      lodash.rangeRight = rangeRight;
      lodash.rearg = rearg;
      lodash.reject = reject;
      lodash.remove = remove;
      lodash.rest = rest;
      lodash.reverse = reverse;
      lodash.sampleSize = sampleSize;
      lodash.set = set;
      lodash.setWith = setWith;
      lodash.shuffle = shuffle;
      lodash.slice = slice;
      lodash.sortBy = sortBy;
      lodash.sortedUniq = sortedUniq;
      lodash.sortedUniqBy = sortedUniqBy;
      lodash.split = split;
      lodash.spread = spread;
      lodash.tail = tail;
      lodash.take = take;
      lodash.takeRight = takeRight;
      lodash.takeRightWhile = takeRightWhile;
      lodash.takeWhile = takeWhile;
      lodash.tap = tap;
      lodash.throttle = throttle;
      lodash.thru = thru;
      lodash.toArray = toArray;
      lodash.toPairs = toPairs;
      lodash.toPairsIn = toPairsIn;
      lodash.toPath = toPath;
      lodash.toPlainObject = toPlainObject;
      lodash.transform = transform;
      lodash.unary = unary;
      lodash.union = union;
      lodash.unionBy = unionBy;
      lodash.unionWith = unionWith;
      lodash.uniq = uniq;
      lodash.uniqBy = uniqBy;
      lodash.uniqWith = uniqWith;
      lodash.unset = unset;
      lodash.unzip = unzip;
      lodash.unzipWith = unzipWith;
      lodash.update = update;
      lodash.updateWith = updateWith;
      lodash.values = values;
      lodash.valuesIn = valuesIn;
      lodash.without = without;
      lodash.words = words;
      lodash.wrap = wrap;
      lodash.xor = xor;
      lodash.xorBy = xorBy;
      lodash.xorWith = xorWith;
      lodash.zip = zip;
      lodash.zipObject = zipObject;
      lodash.zipObjectDeep = zipObjectDeep;
      lodash.zipWith = zipWith;

      // Add aliases.
      lodash.entries = toPairs;
      lodash.entriesIn = toPairsIn;
      lodash.extend = assignIn;
      lodash.extendWith = assignInWith;

      // Add methods to `lodash.prototype`.
      mixin(lodash, lodash);

      /*------------------------------------------------------------------------*/

      // Add methods that return unwrapped values in chain sequences.
      lodash.add = add;
      lodash.attempt = attempt;
      lodash.camelCase = camelCase;
      lodash.capitalize = capitalize;
      lodash.ceil = ceil;
      lodash.clamp = clamp;
      lodash.clone = clone;
      lodash.cloneDeep = cloneDeep;
      lodash.cloneDeepWith = cloneDeepWith;
      lodash.cloneWith = cloneWith;
      lodash.conformsTo = conformsTo;
      lodash.deburr = deburr;
      lodash.defaultTo = defaultTo;
      lodash.divide = divide;
      lodash.endsWith = endsWith;
      lodash.eq = eq;
      lodash.escape = escape;
      lodash.escapeRegExp = escapeRegExp;
      lodash.every = every;
      lodash.find = find;
      lodash.findIndex = findIndex;
      lodash.findKey = findKey;
      lodash.findLast = findLast;
      lodash.findLastIndex = findLastIndex;
      lodash.findLastKey = findLastKey;
      lodash.floor = floor;
      lodash.forEach = forEach;
      lodash.forEachRight = forEachRight;
      lodash.forIn = forIn;
      lodash.forInRight = forInRight;
      lodash.forOwn = forOwn;
      lodash.forOwnRight = forOwnRight;
      lodash.get = get;
      lodash.gt = gt;
      lodash.gte = gte;
      lodash.has = has;
      lodash.hasIn = hasIn;
      lodash.head = head;
      lodash.identity = identity;
      lodash.includes = includes;
      lodash.indexOf = indexOf;
      lodash.inRange = inRange;
      lodash.invoke = invoke;
      lodash.isArguments = isArguments;
      lodash.isArray = isArray;
      lodash.isArrayBuffer = isArrayBuffer;
      lodash.isArrayLike = isArrayLike;
      lodash.isArrayLikeObject = isArrayLikeObject;
      lodash.isBoolean = isBoolean;
      lodash.isBuffer = isBuffer;
      lodash.isDate = isDate;
      lodash.isElement = isElement;
      lodash.isEmpty = isEmpty;
      lodash.isEqual = isEqual;
      lodash.isEqualWith = isEqualWith;
      lodash.isError = isError;
      lodash.isFinite = isFinite;
      lodash.isFunction = isFunction;
      lodash.isInteger = isInteger;
      lodash.isLength = isLength;
      lodash.isMap = isMap;
      lodash.isMatch = isMatch;
      lodash.isMatchWith = isMatchWith;
      lodash.isNaN = isNaN;
      lodash.isNative = isNative;
      lodash.isNil = isNil;
      lodash.isNull = isNull;
      lodash.isNumber = isNumber;
      lodash.isObject = isObject;
      lodash.isObjectLike = isObjectLike;
      lodash.isPlainObject = isPlainObject;
      lodash.isRegExp = isRegExp;
      lodash.isSafeInteger = isSafeInteger;
      lodash.isSet = isSet;
      lodash.isString = isString;
      lodash.isSymbol = isSymbol;
      lodash.isTypedArray = isTypedArray;
      lodash.isUndefined = isUndefined;
      lodash.isWeakMap = isWeakMap;
      lodash.isWeakSet = isWeakSet;
      lodash.join = join;
      lodash.kebabCase = kebabCase;
      lodash.last = last;
      lodash.lastIndexOf = lastIndexOf;
      lodash.lowerCase = lowerCase;
      lodash.lowerFirst = lowerFirst;
      lodash.lt = lt;
      lodash.lte = lte;
      lodash.max = max;
      lodash.maxBy = maxBy;
      lodash.mean = mean;
      lodash.meanBy = meanBy;
      lodash.min = min;
      lodash.minBy = minBy;
      lodash.stubArray = stubArray;
      lodash.stubFalse = stubFalse;
      lodash.stubObject = stubObject;
      lodash.stubString = stubString;
      lodash.stubTrue = stubTrue;
      lodash.multiply = multiply;
      lodash.nth = nth;
      lodash.noConflict = noConflict;
      lodash.noop = noop;
      lodash.now = now;
      lodash.pad = pad;
      lodash.padEnd = padEnd;
      lodash.padStart = padStart;
      lodash.parseInt = parseInt;
      lodash.random = random;
      lodash.reduce = reduce;
      lodash.reduceRight = reduceRight;
      lodash.repeat = repeat;
      lodash.replace = replace;
      lodash.result = result;
      lodash.round = round;
      lodash.runInContext = runInContext;
      lodash.sample = sample;
      lodash.size = size;
      lodash.snakeCase = snakeCase;
      lodash.some = some;
      lodash.sortedIndex = sortedIndex;
      lodash.sortedIndexBy = sortedIndexBy;
      lodash.sortedIndexOf = sortedIndexOf;
      lodash.sortedLastIndex = sortedLastIndex;
      lodash.sortedLastIndexBy = sortedLastIndexBy;
      lodash.sortedLastIndexOf = sortedLastIndexOf;
      lodash.startCase = startCase;
      lodash.startsWith = startsWith;
      lodash.subtract = subtract;
      lodash.sum = sum;
      lodash.sumBy = sumBy;
      lodash.template = template;
      lodash.times = times;
      lodash.toFinite = toFinite;
      lodash.toInteger = toInteger;
      lodash.toLength = toLength;
      lodash.toLower = toLower;
      lodash.toNumber = toNumber;
      lodash.toSafeInteger = toSafeInteger;
      lodash.toString = toString;
      lodash.toUpper = toUpper;
      lodash.trim = trim;
      lodash.trimEnd = trimEnd;
      lodash.trimStart = trimStart;
      lodash.truncate = truncate;
      lodash.unescape = unescape;
      lodash.uniqueId = uniqueId;
      lodash.upperCase = upperCase;
      lodash.upperFirst = upperFirst;

      // Add aliases.
      lodash.each = forEach;
      lodash.eachRight = forEachRight;
      lodash.first = head;

      mixin(lodash, (function() {
        var source = {};
        baseForOwn(lodash, function(func, methodName) {
          if (!hasOwnProperty.call(lodash.prototype, methodName)) {
            source[methodName] = func;
          }
        });
        return source;
      }()), { 'chain': false });

      /*------------------------------------------------------------------------*/

      /**
       * The semantic version number.
       *
       * @static
       * @memberOf _
       * @type {string}
       */
      lodash.VERSION = VERSION;

      // Assign default placeholders.
      arrayEach(['bind', 'bindKey', 'curry', 'curryRight', 'partial', 'partialRight'], function(methodName) {
        lodash[methodName].placeholder = lodash;
      });

      // Add `LazyWrapper` methods for `_.drop` and `_.take` variants.
      arrayEach(['drop', 'take'], function(methodName, index) {
        LazyWrapper.prototype[methodName] = function(n) {
          n = n === undefined$1 ? 1 : nativeMax(toInteger(n), 0);

          var result = (this.__filtered__ && !index)
            ? new LazyWrapper(this)
            : this.clone();

          if (result.__filtered__) {
            result.__takeCount__ = nativeMin(n, result.__takeCount__);
          } else {
            result.__views__.push({
              'size': nativeMin(n, MAX_ARRAY_LENGTH),
              'type': methodName + (result.__dir__ < 0 ? 'Right' : '')
            });
          }
          return result;
        };

        LazyWrapper.prototype[methodName + 'Right'] = function(n) {
          return this.reverse()[methodName](n).reverse();
        };
      });

      // Add `LazyWrapper` methods that accept an `iteratee` value.
      arrayEach(['filter', 'map', 'takeWhile'], function(methodName, index) {
        var type = index + 1,
            isFilter = type == LAZY_FILTER_FLAG || type == LAZY_WHILE_FLAG;

        LazyWrapper.prototype[methodName] = function(iteratee) {
          var result = this.clone();
          result.__iteratees__.push({
            'iteratee': getIteratee(iteratee, 3),
            'type': type
          });
          result.__filtered__ = result.__filtered__ || isFilter;
          return result;
        };
      });

      // Add `LazyWrapper` methods for `_.head` and `_.last`.
      arrayEach(['head', 'last'], function(methodName, index) {
        var takeName = 'take' + (index ? 'Right' : '');

        LazyWrapper.prototype[methodName] = function() {
          return this[takeName](1).value()[0];
        };
      });

      // Add `LazyWrapper` methods for `_.initial` and `_.tail`.
      arrayEach(['initial', 'tail'], function(methodName, index) {
        var dropName = 'drop' + (index ? '' : 'Right');

        LazyWrapper.prototype[methodName] = function() {
          return this.__filtered__ ? new LazyWrapper(this) : this[dropName](1);
        };
      });

      LazyWrapper.prototype.compact = function() {
        return this.filter(identity);
      };

      LazyWrapper.prototype.find = function(predicate) {
        return this.filter(predicate).head();
      };

      LazyWrapper.prototype.findLast = function(predicate) {
        return this.reverse().find(predicate);
      };

      LazyWrapper.prototype.invokeMap = baseRest(function(path, args) {
        if (typeof path == 'function') {
          return new LazyWrapper(this);
        }
        return this.map(function(value) {
          return baseInvoke(value, path, args);
        });
      });

      LazyWrapper.prototype.reject = function(predicate) {
        return this.filter(negate(getIteratee(predicate)));
      };

      LazyWrapper.prototype.slice = function(start, end) {
        start = toInteger(start);

        var result = this;
        if (result.__filtered__ && (start > 0 || end < 0)) {
          return new LazyWrapper(result);
        }
        if (start < 0) {
          result = result.takeRight(-start);
        } else if (start) {
          result = result.drop(start);
        }
        if (end !== undefined$1) {
          end = toInteger(end);
          result = end < 0 ? result.dropRight(-end) : result.take(end - start);
        }
        return result;
      };

      LazyWrapper.prototype.takeRightWhile = function(predicate) {
        return this.reverse().takeWhile(predicate).reverse();
      };

      LazyWrapper.prototype.toArray = function() {
        return this.take(MAX_ARRAY_LENGTH);
      };

      // Add `LazyWrapper` methods to `lodash.prototype`.
      baseForOwn(LazyWrapper.prototype, function(func, methodName) {
        var checkIteratee = /^(?:filter|find|map|reject)|While$/.test(methodName),
            isTaker = /^(?:head|last)$/.test(methodName),
            lodashFunc = lodash[isTaker ? ('take' + (methodName == 'last' ? 'Right' : '')) : methodName],
            retUnwrapped = isTaker || /^find/.test(methodName);

        if (!lodashFunc) {
          return;
        }
        lodash.prototype[methodName] = function() {
          var value = this.__wrapped__,
              args = isTaker ? [1] : arguments,
              isLazy = value instanceof LazyWrapper,
              iteratee = args[0],
              useLazy = isLazy || isArray(value);

          var interceptor = function(value) {
            var result = lodashFunc.apply(lodash, arrayPush([value], args));
            return (isTaker && chainAll) ? result[0] : result;
          };

          if (useLazy && checkIteratee && typeof iteratee == 'function' && iteratee.length != 1) {
            // Avoid lazy use if the iteratee has a "length" value other than `1`.
            isLazy = useLazy = false;
          }
          var chainAll = this.__chain__,
              isHybrid = !!this.__actions__.length,
              isUnwrapped = retUnwrapped && !chainAll,
              onlyLazy = isLazy && !isHybrid;

          if (!retUnwrapped && useLazy) {
            value = onlyLazy ? value : new LazyWrapper(this);
            var result = func.apply(value, args);
            result.__actions__.push({ 'func': thru, 'args': [interceptor], 'thisArg': undefined$1 });
            return new LodashWrapper(result, chainAll);
          }
          if (isUnwrapped && onlyLazy) {
            return func.apply(this, args);
          }
          result = this.thru(interceptor);
          return isUnwrapped ? (isTaker ? result.value()[0] : result.value()) : result;
        };
      });

      // Add `Array` methods to `lodash.prototype`.
      arrayEach(['pop', 'push', 'shift', 'sort', 'splice', 'unshift'], function(methodName) {
        var func = arrayProto[methodName],
            chainName = /^(?:push|sort|unshift)$/.test(methodName) ? 'tap' : 'thru',
            retUnwrapped = /^(?:pop|shift)$/.test(methodName);

        lodash.prototype[methodName] = function() {
          var args = arguments;
          if (retUnwrapped && !this.__chain__) {
            var value = this.value();
            return func.apply(isArray(value) ? value : [], args);
          }
          return this[chainName](function(value) {
            return func.apply(isArray(value) ? value : [], args);
          });
        };
      });

      // Map minified method names to their real names.
      baseForOwn(LazyWrapper.prototype, function(func, methodName) {
        var lodashFunc = lodash[methodName];
        if (lodashFunc) {
          var key = lodashFunc.name + '';
          if (!hasOwnProperty.call(realNames, key)) {
            realNames[key] = [];
          }
          realNames[key].push({ 'name': methodName, 'func': lodashFunc });
        }
      });

      realNames[createHybrid(undefined$1, WRAP_BIND_KEY_FLAG).name] = [{
        'name': 'wrapper',
        'func': undefined$1
      }];

      // Add methods to `LazyWrapper`.
      LazyWrapper.prototype.clone = lazyClone;
      LazyWrapper.prototype.reverse = lazyReverse;
      LazyWrapper.prototype.value = lazyValue;

      // Add chain sequence methods to the `lodash` wrapper.
      lodash.prototype.at = wrapperAt;
      lodash.prototype.chain = wrapperChain;
      lodash.prototype.commit = wrapperCommit;
      lodash.prototype.next = wrapperNext;
      lodash.prototype.plant = wrapperPlant;
      lodash.prototype.reverse = wrapperReverse;
      lodash.prototype.toJSON = lodash.prototype.valueOf = lodash.prototype.value = wrapperValue;

      // Add lazy aliases.
      lodash.prototype.first = lodash.prototype.head;

      if (symIterator) {
        lodash.prototype[symIterator] = wrapperToIterator;
      }
      return lodash;
    });

    /*--------------------------------------------------------------------------*/

    // Export lodash.
    var _ = runInContext();

    // Some AMD build optimizers, like r.js, check for condition patterns like:
    if (freeModule) {
      // Export for Node.js.
      (freeModule.exports = _)._ = _;
      // Export for CommonJS support.
      freeExports._ = _;
    }
    else {
      // Export to the global object.
      root._ = _;
    }
  }.call(commonjsGlobal));
  });

  var author$2 = {
  	name: "The plotly.py team"
  };
  var bugs$2 = {
  	url: "https://github.com/plotly/plotly.py/issues"
  };
  var dependencies$2 = {
  	"@jupyter-widgets/base": "^2.0.0 || ^3.0.0",
  	lodash: "^4.17.4",
  	"plotly.js": "^1.54.1"
  };
  var description$2 = "The plotly JupyterLab extension";
  var devDependencies$2 = {
  	"ify-loader": "^1.1.0",
  	rimraf: "^2.6.1",
  	typescript: "~3.1.1",
  	webpack: "^3.10.0"
  };
  var files$2 = [
  	"src/**/*.js",
  	"dist/*.js",
  	"style/*.*"
  ];
  var homepage$2 = "https://github.com/plotly/plotly.py#readme";
  var jupyterlab = {
  	extension: "src/jupyterlab-plugin.js"
  };
  var keywords = [
  	"jupyter",
  	"widgets",
  	"ipython",
  	"ipywidgets",
  	"plotly"
  ];
  var license$2 = "MIT";
  var main$2 = "src/index.js";
  var name$2 = "plotlywidget";
  var repository$2 = {
  	type: "git",
  	url: "git+https://github.com/plotly/plotly.py.git"
  };
  var scripts$2 = {
  	build: "webpack",
  	clean: "rimraf dist/ && rimraf  ../../python/plotly/plotlywidget/static'",
  	test: "echo \"Error: no test specified\" && exit 1"
  };
  var version$3 = "4.8.1";
  var _package$4 = {
  	author: author$2,
  	bugs: bugs$2,
  	dependencies: dependencies$2,
  	description: description$2,
  	devDependencies: devDependencies$2,
  	files: files$2,
  	homepage: homepage$2,
  	jupyterlab: jupyterlab,
  	keywords: keywords,
  	license: license$2,
  	main: main$2,
  	name: name$2,
  	repository: repository$2,
  	scripts: scripts$2,
  	version: version$3
  };

  var _package$5 = /*#__PURE__*/Object.freeze({
    author: author$2,
    bugs: bugs$2,
    dependencies: dependencies$2,
    description: description$2,
    devDependencies: devDependencies$2,
    files: files$2,
    homepage: homepage$2,
    jupyterlab: jupyterlab,
    keywords: keywords,
    license: license$2,
    main: main$2,
    name: name$2,
    repository: repository$2,
    scripts: scripts$2,
    version: version$3,
    'default': _package$4
  });

  var require$$1 = getCjsExportFromNamespace(_package$5);

  window.PlotlyConfig = {MathJaxConfig: 'local'};

  var semver_range = "^" + require$$1.version;

  // Model
  // =====
  /**
   * A FigureModel holds a mirror copy of the state of a FigureWidget on
   * the Python side.  There is a one-to-one relationship between JavaScript
   * FigureModels and Python FigureWidgets. The JavaScript FigureModel is
   * initialized as soon as a Python FigureWidget initialized, this happens
   * even before the widget is first displayed in the Notebook
   * @type {widgets.DOMWidgetModel}
   */
  var FigureModel = lib$1.DOMWidgetModel.extend({

      defaults: lodash.extend(lib$1.DOMWidgetModel.prototype.defaults(), {
          // Model metadata
          // --------------
          _model_name: "FigureModel",
          _view_name: "FigureView",
          _model_module: "plotlywidget",
          _view_module: "plotlywidget",
          _view_module_version: semver_range,
          _model_module_version: semver_range,

          // Data and Layout
          // ---------------
          // The _data and _layout properties are synchronized with the
          // Python side on initialization only.  After initialization, these
          // properties are kept in sync through the use of the _py2js_*
          // messages
          _data: [],
          _layout: {},
          _config: {},

          // Python -> JS messages
          // ---------------------
          // Messages are implemented using trait properties. This is done so
          // that we can take advantage of ipywidget's binary serialization
          // protocol.
          //
          // Messages are sent by the Python side by assigning the message
          // contents to the appropriate _py2js_* property, and then immediately
          // setting it to None.  Messages are received by the JavaScript
          // side by registering property change callbacks in the initialize
          // methods for FigureModel and FigureView. e.g. (where this is a
          // FigureModel):
          //
          //      this.on('change:_py2js_addTraces', this.do_addTraces, this);
          //
          // Message handling methods, do_addTraces, are responsible for
          // performing the appropriate action if the message contents are
          // not null

          /**
           * @typedef {null|Object} Py2JsAddTracesMsg
           * @property {Array.<Object>} trace_data
           *  Array of traces to append to the end of the figure's current traces
           * @property {Number} trace_edit_id
           *  Edit ID to use when returning trace deltas using
           *  the _js2py_traceDeltas message.
           * @property {Number} layout_edit_id
           *  Edit ID to use when returning layout deltas using
           *  the _js2py_layoutDelta message.
           */
          _py2js_addTraces: null,

          /**
           * @typedef {null|Object} Py2JsDeleteTracesMsg
           * @property {Array.<Number>} delete_inds
           *  Array of indexes of traces to be deleted, in ascending order
           * @property {Number} trace_edit_id
           *  Edit ID to use when returning trace deltas using
           *  the _js2py_traceDeltas message.
           * @property {Number} layout_edit_id
           *  Edit ID to use when returning layout deltas using
           *  the _js2py_layoutDelta message.
           */
          _py2js_deleteTraces: null,

          /**
           * @typedef {null|Object} Py2JsMoveTracesMsg
           * @property {Array.<Number>} current_trace_inds
           *  Array of the current indexes of traces to be moved
           * @property {Array.<Number>} new_trace_inds
           *  Array of the new indexes that traces should be moved to.
           */
          _py2js_moveTraces: null,


          /**
           * @typedef {null|Object} Py2JsRestyleMsg
           * @property {Object} restyle_data
           *  Restyle data as accepted by Plotly.restyle
           * @property {null|Array.<Number>} restyle_traces
           *  Array of indexes of the traces that the resytle operation applies
           *  to, or null to apply the operation to all traces
           * @property {Number} trace_edit_id
           *  Edit ID to use when returning trace deltas using
           *  the _js2py_traceDeltas message
           * @property {Number} layout_edit_id
           *  Edit ID to use when returning layout deltas using
           *  the _js2py_layoutDelta message
           * @property {null|String} source_view_id
           *  view_id of the FigureView that triggered the original restyle
           *  event (e.g. by clicking the legend), or null if the restyle was
           *  triggered from Python
           */
          _py2js_restyle: null,

          /**
           * @typedef {null|Object} Py2JsRelayoutMsg
           * @property {Object} relayout_data
           *  Relayout data as accepted by Plotly.relayout
           * @property {Number} layout_edit_id
           *  Edit ID to use when returning layout deltas using
           *  the _js2py_layoutDelta message
           * @property {null|String} source_view_id
           *  view_id of the FigureView that triggered the original relayout
           *  event (e.g. by clicking the zoom button), or null if the
           *  relayout was triggered from Python
           */
          _py2js_relayout: null,

          /**
           * @typedef {null|Object} Py2JsUpdateMsg
           * @property {Object} style_data
           *  Style data as accepted by Plotly.update
           * @property {Object} layout_data
           *  Layout data as accepted by Plotly.update
           * @property {Array.<Number>} style_traces
           *  Array of indexes of the traces that the update operation applies
           *  to, or null to apply the operation to all traces
           * @property {Number} trace_edit_id
           *  Edit ID to use when returning trace deltas using
           *  the _js2py_traceDeltas message
           * @property {Number} layout_edit_id
           *  Edit ID to use when returning layout deltas using
           *  the _js2py_layoutDelta message
           * @property {null|String} source_view_id
           *  view_id of the FigureView that triggered the original update
           *  event (e.g. by clicking a button), or null if the update was
           *  triggered from Python
           */
          _py2js_update: null,

          /**
           * @typedef {null|Object} Py2JsAnimateMsg
           * @property {Object} style_data
           *  Style data as accepted by Plotly.animate
           * @property {Object} layout_data
           *  Layout data as accepted by Plotly.animate
           * @property {Array.<Number>} style_traces
           *  Array of indexes of the traces that the animate operation applies
           *  to, or null to apply the operation to all traces
           * @property {Object} animation_opts
           *  Animation options as accepted by Plotly.animate
           * @property {Number} trace_edit_id
           *  Edit ID to use when returning trace deltas using
           *  the _js2py_traceDeltas message
           * @property {Number} layout_edit_id
           *  Edit ID to use when returning layout deltas using
           *  the _js2py_layoutDelta message
           * @property {null|String} source_view_id
           *  view_id of the FigureView that triggered the original animate
           *  event (e.g. by clicking a button), or null if the update was
           *  triggered from Python
           */
          _py2js_animate: null,

          /**
           * @typedef {null|Object} Py2JsRemoveLayoutPropsMsg
           * @property {Array.<Array.<String|Number>>} remove_props
           *  Array of property paths to remove. Each propery path is an
           *  array of property names or array indexes that locate a property
           *  inside the _layout object
           */
          _py2js_removeLayoutProps: null,

          /**
           * @typedef {null|Object} Py2JsRemoveTracePropsMsg
           * @property {Number} remove_trace
           *  The index of the trace from which to remove properties
           * @property {Array.<Array.<String|Number>>} remove_props
           *  Array of property paths to remove. Each propery path is an
           *  array of property names or array indexes that locate a property
           *  inside the _data[remove_trace] object
           */
          _py2js_removeTraceProps: null,


          // JS -> Python messages
          // ---------------------
          // Messages are sent by the JavaScript side by assigning the
          // message contents to the appropriate _js2py_* property and then
          // calling the `touch` method on the view that triggered the
          // change. e.g. (where this is a FigureView):
          //
          //      this.model.set('_js2py_restyle', data);
          //      this.touch();
          //
          // The Python side is responsible for setting the property to None
          // after receiving the message.
          //
          // Message trigger logic is described in the corresponding
          // handle_plotly_* methods of FigureView

          /**
           * @typedef {null|Object} Js2PyRestyleMsg
           * @property {Object} style_data
           *  Style data that was passed to Plotly.restyle
           * @property {Array.<Number>} style_traces
           *  Array of indexes of the traces that the restyle operation
           *  was applied to, or null if applied to all traces
           * @property {String} source_view_id
           *  view_id of the FigureView that triggered the original restyle
           *  event (e.g. by clicking the legend)
           */
          _js2py_restyle: null,

          /**
           * @typedef {null|Object} Js2PyRelayoutMsg
           * @property {Object} relayout_data
           *  Relayout data that was passed to Plotly.relayout
           * @property {String} source_view_id
           *  view_id of the FigureView that triggered the original relayout
           *  event (e.g. by clicking the zoom button)
           */
          _js2py_relayout: null,

          /**
           * @typedef {null|Object} Js2PyUpdateMsg
           * @property {Object} style_data
           *  Style data that was passed to Plotly.update
           * @property {Object} layout_data
           *  Layout data that was passed to Plotly.update
           * @property {Array.<Number>} style_traces
           *  Array of indexes of the traces that the update operation applied
           *  to, or null if applied to all traces
           * @property {String} source_view_id
           *  view_id of the FigureView that triggered the original relayout
           *  event (e.g. by clicking the zoom button)
           */
          _js2py_update: null,

          /**
           * @typedef {null|Object} Js2PyLayoutDeltaMsg
           * @property {Object} layout_delta
           *  The layout delta object that contains all of the properties of
           *  _fullLayout that are not identical to those in the
           *  FigureModel's _layout property
           * @property {Number} layout_edit_id
           *  Edit ID of message that triggered the creation of layout delta
           */
          _js2py_layoutDelta: null,

          /**
           * @typedef {null|Object} Js2PyTraceDeltasMsg
           * @property {Array.<Object>} trace_deltas
           *  Array of trace delta objects. Each trace delta contains the
           *  trace's uid along with all of the properties of _fullData that
           *  are not identical to those in the FigureModel's _data property
           * @property {Number} trace_edit_id
           *  Edit ID of message that triggered the creation of trace deltas
           */
          _js2py_traceDeltas: null,


          /**
           * Object representing a collection of points for use in click, hover,
           * and selection events
           * @typedef {Object} Points
           * @property {Array.<Number>} trace_indexes
           *  Array of the trace index for each point
           * @property {Array.<Number>} point_indexes
           *  Array of the index of each point in its own trace
           * @property {null|Array.<Number>} xs
           *  Array of the x coordinate of each point (for cartesian trace types)
           *  or null (for non-cartesian trace types)
           * @property {null|Array.<Number>} ys
           *  Array of the y coordinate of each point (for cartesian trace types)
           *  or null (for non-cartesian trace types
           * @property {null|Array.<Number>} zs
           *  Array of the z coordinate of each point (for 3D cartesian
           *  trace types)
           *  or null (for non-3D-cartesian trace types)
           */

          /**
           * Object representing the state of the input devices during a
           * plotly event
           * @typedef {Object} InputDeviceState
           * @property {boolean} alt - true if alt key pressed,
           * false otherwise
           * @property {boolean} ctrl - true if ctrl key pressed,
           * false otherwise
           * @property {boolean} meta - true if meta key pressed,
           * false otherwise
           * @property {boolean} shift - true if shift key pressed,
           * false otherwise
           *
           * @property {boolean} button
           *  Indicates which button was pressed on the mouse to trigger the
           *  event.
           *    0: Main button pressed, usually the left button or the
           *       un-initialized state
           *    1: Auxiliary button pressed, usually the wheel button or
           *       the middle button (if present)
           *    2: Secondary button pressed, usually the right button
           *    3: Fourth button, typically the Browser Back button
           *    4: Fifth button, typically the Browser Forward button
           *
           * @property {boolean} buttons
           *  Indicates which buttons were pressed on the mouse when the event
           *  is triggered.
           *    0  : No button or un-initialized
           *    1  : Primary button (usually left)
           *    2  : Secondary button (usually right)
           *    4  : Auxilary button (usually middle or mouse wheel button)
           *    8  : 4th button (typically the "Browser Back" button)
           *    16 : 5th button (typically the "Browser Forward" button)
           *
           *  Combinations of buttons are represented by the sum of the codes
           *  above. e.g. a value of 7 indicates buttons 1 (primary),
           *  2 (secondary), and 4 (auxilary) were pressed during the event
           */

          /**
           * @typedef {Object} BoxSelectorState
           * @property {Array.<Number>} xrange
           *  Two element array containing the x-range of the box selection
           * @property {Array.<Number>} yrange
           *  Two element array containing the y-range of the box selection
           */

          /**
           * @typedef {Object} LassoSelectorState
           * @property {Array.<Number>} xs
           *  Array of the x-coordinates of the lasso selection region
           * @property {Array.<Number>} ys
           *  Array of the y-coordinates of the lasso selection region
           */

          /**
           * Object representing the state of the selection tool during a
           * plotly_select event
           * @typedef {Object} Selector
           * @property {String} type
           *  Selection type. One of: 'box', or 'lasso'
           * @property {BoxSelectorState|LassoSelectorState} selector_state
           */

          /**
           * @typedef {null|Object} Js2PyPointsCallbackMsg
           * @property {string} event_type
           *  Name of the triggering event. One of 'plotly_click',
           *  'plotly_hover', 'plotly_unhover', or 'plotly_selected'
           * @property {null|Points} points
           *  Points object for event
           * @property {null|InputDeviceState} device_state
           *  InputDeviceState object for event
           * @property {null|Selector} selector
           *  State of the selection tool for 'plotly_selected' events, null
           *  for other event types
           */
          _js2py_pointsCallback: null,

          // Message tracking
          // ----------------
          /**
           * @type {Number}
           * layout_edit_id of the last layout modification operation
           * requested by the Python side
           */
          _last_layout_edit_id: 0,

          /**
           * @type {Number}
           * trace_edit_id of the last trace modification operation
           * requested by the Python side
           */
          _last_trace_edit_id: 0
      }),

      /**
       * Initialize FigureModel. Called when the Python FigureWidget is first
       * constructed
       */
      initialize: function() {
          FigureModel.__super__.initialize.apply(this, arguments);

          this.on("change:_data", this.do_data, this);
          this.on("change:_layout", this.do_layout, this);
          this.on("change:_py2js_addTraces", this.do_addTraces, this);
          this.on("change:_py2js_deleteTraces", this.do_deleteTraces, this);
          this.on("change:_py2js_moveTraces", this.do_moveTraces, this);
          this.on("change:_py2js_restyle", this.do_restyle, this);
          this.on("change:_py2js_relayout", this.do_relayout, this);
          this.on("change:_py2js_update", this.do_update, this);
          this.on("change:_py2js_animate", this.do_animate, this);
          this.on("change:_py2js_removeLayoutProps",
              this.do_removeLayoutProps, this);
          this.on("change:_py2js_removeTraceProps",
              this.do_removeTraceProps, this);
      },

      /**
       * Input a trace index specification and return an Array of trace
       * indexes where:
       *
       *  - null|undefined -> Array of all traces
       *  - Trace index as Number -> Single element array of input index
       *  - Array of trace indexes -> Input array unchanged
       *
       * @param {undefined|null|Number|Array.<Number>} trace_indexes
       * @returns {Array.<Number>}
       *  Array of trace indexes
       * @private
       */
      _normalize_trace_indexes: function (trace_indexes) {
          if (trace_indexes === null || trace_indexes === undefined) {
              var numTraces = this.get("_data").length;
              trace_indexes = lodash.range(numTraces);
          }
          if (!Array.isArray(trace_indexes)) {
              // Make sure idx is an array
              trace_indexes = [trace_indexes];
          }
          return trace_indexes
      },

      /**
       * Log changes to the _data trait
       *
       * This should only happed on FigureModel initialization
       */
      do_data: function () {

      },

      /**
       * Log changes to the _layout trait
       *
       * This should only happed on FigureModel initialization
       */
      do_layout: function () {

      },

      /**
       * Handle addTraces message
       */
      do_addTraces: function () {
          // add trace to plot
          /** @type {Py2JsAddTracesMsg} */
          var msgData = this.get("_py2js_addTraces");

          if (msgData !== null) {
              var currentTraces = this.get("_data");
              var newTraces = msgData.trace_data;
              lodash.forEach(newTraces, function (newTrace) {
                  currentTraces.push(newTrace);
              });
          }
      },

      /**
       * Handle deleteTraces message
       */
      do_deleteTraces: function () {
          // remove traces from plot

          /** @type {Py2JsDeleteTracesMsg} */
          var msgData = this.get("_py2js_deleteTraces");

          if (msgData !== null) {
              var delete_inds = msgData.delete_inds;
              var tracesData = this.get("_data");

              // Remove del inds in reverse order so indexes remain valid
              // throughout loop
              delete_inds.slice().reverse().forEach(function (del_ind) {
                  tracesData.splice(del_ind, 1);
              });
          }
      },

      /**
       * Handle moveTraces message
       */
      do_moveTraces: function () {

          /** @type {Py2JsMoveTracesMsg} */
          var msgData = this.get("_py2js_moveTraces");

          if (msgData !== null) {
              var tracesData = this.get("_data");
              var currentInds = msgData.current_trace_inds;
              var newInds = msgData.new_trace_inds;

              performMoveTracesLike(tracesData, currentInds, newInds);
          }
      },

      /**
       * Handle restyle message
       */
      do_restyle: function () {

          /** @type {Py2JsRestyleMsg} */
          var msgData = this.get("_py2js_restyle");
          if (msgData !== null) {
              var restyleData = msgData.restyle_data;
              var restyleTraces = this._normalize_trace_indexes(
                  msgData.restyle_traces);
              performRestyleLike(this.get("_data"), restyleData, restyleTraces);
          }
      },

      /**
       * Handle relayout message
       */
      do_relayout: function () {

          /** @type {Py2JsRelayoutMsg} */
          var msgData = this.get("_py2js_relayout");

          if (msgData !== null) {
              performRelayoutLike(this.get("_layout"), msgData.relayout_data);
          }
      },

      /**
       * Handle update message
       */
      do_update: function() {

          /** @type {Py2JsUpdateMsg} */
          var msgData = this.get("_py2js_update");

          if (msgData !== null) {
              var style = msgData.style_data;
              var layout = msgData.layout_data;
              var styleTraces = this._normalize_trace_indexes(
                  msgData.style_traces);
              performRestyleLike(this.get("_data"), style, styleTraces);
              performRelayoutLike(this.get("_layout"), layout);
          }
      },

      /**
       * Handle animate message
       */
      do_animate: function () {

          /** @type {Py2JsAnimateMsg} */
          var msgData = this.get("_py2js_animate");
          if (msgData !== null) {

              var styles = msgData.style_data;
              var layout = msgData.layout_data;
              var trace_indexes = this._normalize_trace_indexes(
                  msgData.style_traces);

              for (var i = 0; i < styles.length; i++) {
                  var style = styles[i];
                  var trace_index = trace_indexes[i];
                  var trace = this.get("_data")[trace_index];
                  performRelayoutLike(trace, style);
              }

              performRelayoutLike(this.get("_layout"), layout);
          }
      },

      /**
       * Handle removeLayoutProps message
       */
      do_removeLayoutProps: function () {
          /** @type {Py2JsRemoveLayoutPropsMsg} */
          var msgData = this.get("_py2js_removeLayoutProps");

          if (msgData !== null) {
              var keyPaths = msgData.remove_props;
              var layout = this.get("_layout");
              performRemoveProps(layout, keyPaths);
          }
      },

      /**
       * Handle removeTraceProps message
       */
      do_removeTraceProps: function () {
          /** @type {Py2JsRemoveTracePropsMsg} */
          var msgData = this.get("_py2js_removeTraceProps");
          if (msgData !== null) {
              var keyPaths = msgData.remove_props;
              var traceIndex = msgData.remove_trace;
              var trace = this.get("_data")[traceIndex];

              performRemoveProps(trace, keyPaths);
          }
      }
  }, {
      serializers: lodash.extend({
          _data: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _layout: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_addTraces: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_deleteTraces: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_moveTraces: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_restyle: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_relayout: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_update: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_animate: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_removeLayoutProps: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _py2js_removeTraceProps: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _js2py_restyle: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _js2py_relayout: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _js2py_update: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _js2py_layoutDelta: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _js2py_traceDeltas: { deserialize: py2js_deserializer,
              serialize: js2py_serializer},
          _js2py_pointsCallback: { deserialize: py2js_deserializer,
              serialize: js2py_serializer}
      }, lib$1.DOMWidgetModel.serializers)
  });

  // View
  // ====
  /**
   * A FigureView manages the visual presentation of a single Plotly.js
   * figure for a single notebook output cell. Each FigureView has a
   * reference to FigureModel.  Multiple views may share a single model
   * instance, as is the case when a Python FigureWidget is displayed in
   * multiple notebook output cells.
   *
   * @type {widgets.DOMWidgetView}
   */
  var FigureView = lib$1.DOMWidgetView.extend({

      /**
       * The perform_render method is called by processPhosphorMessage
       * after the widget's DOM element has been attached to the notebook
       * output cell. This happens after the initialize of the
       * FigureModel, and it won't happen at all if the Python FigureWidget
       * is never displayed in a notebook output cell
       */
      perform_render: function() {

          var that = this;

          // Wire up message property callbacks
          // ----------------------------------
          // Python -> JS event properties
          this.model.on("change:_py2js_addTraces",
              this.do_addTraces, this);
          this.model.on("change:_py2js_deleteTraces",
              this.do_deleteTraces, this);
          this.model.on("change:_py2js_moveTraces",
              this.do_moveTraces, this);
          this.model.on("change:_py2js_restyle",
              this.do_restyle, this);
          this.model.on("change:_py2js_relayout",
              this.do_relayout, this);
          this.model.on("change:_py2js_update",
              this.do_update, this);
          this.model.on("change:_py2js_animate",
              this.do_animate, this);

          // MathJax configuration
          // ---------------------
          if (window.MathJax) {
              MathJax.Hub.Config({SVG: {font: "STIX-Web"}});
          }

          // Get message ids
          // ---------------------
          var layout_edit_id = this.model.get("_last_layout_edit_id");
          var trace_edit_id = this.model.get("_last_trace_edit_id");

          // Set view UID
          // ------------
          this.viewID = randstr();

          // Initialize Plotly.js figure
          // ---------------------------
          // We must clone the model's data and layout properties so that
          // the model is not directly mutated by the Plotly.js library.
          var initialTraces = lodash.cloneDeep(this.model.get("_data"));
          var initialLayout = lodash.cloneDeep(this.model.get("_layout"));
          var config = this.model.get("_config");

          Plotly.newPlot(that.el, initialTraces, initialLayout, config).then(
              function () {

                  // ### Send trace deltas ###
                  // We create an array of deltas corresponding to the new
                  // traces.
                  that._sendTraceDeltas(trace_edit_id);

                  // ### Send layout delta ###
                  that._sendLayoutDelta(layout_edit_id);

                  // Wire up plotly event callbacks
                  that.el.on("plotly_restyle",
                      function (update) {
                          that.handle_plotly_restyle(update);
                      });
                  that.el.on("plotly_relayout",
                      function (update) {
                          that.handle_plotly_relayout(update);
                      });
                  that.el.on("plotly_update",
                      function (update) {
                          that.handle_plotly_update(update);
                      });
                  that.el.on("plotly_click",
                      function (update) {
                          that.handle_plotly_click(update);
                      });
                  that.el.on("plotly_hover",
                      function (update) {
                          that.handle_plotly_hover(update);
                      });
                  that.el.on("plotly_unhover",
                      function (update) {
                          that.handle_plotly_unhover(update);
                      });
                  that.el.on("plotly_selected",
                      function (update) {
                          that.handle_plotly_selected(update);
                      });
                  that.el.on("plotly_deselect",
                      function (update) {
                          that.handle_plotly_deselect(update);
                      });
                  that.el.on("plotly_doubleclick",
                      function (update) {
                          that.handle_plotly_doubleclick(update);
                      });

                  // Emit event indicating that the widget has finished
                  // rendering
                  var event = new CustomEvent("plotlywidget-after-render",
                      { "detail": {"element": that.el, 'viewID': that.viewID}});

                  // Dispatch/Trigger/Fire the event
                  document.dispatchEvent(event);
              });
      },

      /**
       * Respond to phosphorjs events
       */
      processPhosphorMessage: function(msg) {
          FigureView.__super__.processPhosphorMessage.apply(this, arguments);
          var that = this;
          switch (msg.type) {
              case 'before-attach':
                  // Render an initial empty figure. This establishes with
                  // the page that the element will not be empty, avoiding
                  // some occasions where the dynamic sizing behavior leads
                  // to collapsed figure dimensions.
                  var axisHidden = {
                      showgrid: false, showline: false, tickvals: []};

                  Plotly.newPlot(that.el, [], {
                      xaxis: axisHidden, yaxis: axisHidden
                  });

                  window.addEventListener("resize", function(){
                      that.autosizeFigure();
                  });
                  break;
              case 'after-attach':
                  // Rendering actual figure in the after-attach event allows
                  // Plotly.js to size the figure to fill the available element
                  this.perform_render();
                  break;
              case 'resize':
                  this.autosizeFigure();
                  break
          }
      },

      autosizeFigure: function() {
          var that = this;
          var layout = that.model.get('_layout');
          if (lodash.isNil(layout) ||
              lodash.isNil(layout.width)) {
              Plotly.Plots.resize(that.el).then(function(){
                  var layout_edit_id = that.model.get(
                      "_last_layout_edit_id");
                  that._sendLayoutDelta(layout_edit_id);
              });
          }
      },

      /**
       * Purge Plotly.js data structures from the notebook output display
       * element when the view is destroyed
       */
      destroy: function() {
          Plotly.purge(this.el);
      },

      /**
       * Return the figure's _fullData array merged with its data array
       *
       * The merge ensures that for any properties that el._fullData and
       * el.data have in common, we return the version from el.data
       *
       * Named colorscales are one example of why this is needed. The el.data
       * array will hold named colorscale strings (e.g. 'Viridis'), while the
       * el._fullData array will hold the actual colorscale array. e.g.
       *
       *      el.data[0].marker.colorscale == 'Viridis' but
       *      el._fullData[0].marker.colorscale = [[..., ...], ...]
       *
       * Performing the merge allows our FigureModel to retain the 'Viridis'
       * string, rather than having it overridded by the colorscale array.
       *
       */
      getFullData: function () {
          return lodash.mergeWith({}, this.el._fullData, this.el.data,
              fullMergeCustomizer)
      },

      /**
       * Return the figure's _fullLayout object merged with its layout object
       *
       * See getFullData documentation for discussion of why the merge is
       * necessary
       */
      getFullLayout: function () {
          return lodash.mergeWith({}, this.el._fullLayout, this.el.layout,
              fullMergeCustomizer);
      },

      /**
       * Build Points data structure from data supplied by the plotly_click,
       * plotly_hover, or plotly_select events
       * @param {Object} data
       * @returns {null|Points}
       */
      buildPointsObject: function (data) {

          var pointsObject;
          if (data.hasOwnProperty("points")) {
              // Most cartesian plots
              var pointObjects = data["points"];
              var numPoints = pointObjects.length;
              pointsObject = {
                  "trace_indexes": new Array(numPoints),
                  "point_indexes": new Array(numPoints),
                  "xs": new Array(numPoints),
                  "ys": new Array(numPoints)};


                  for (var p = 0; p < numPoints; p++) {
                  pointsObject["trace_indexes"][p] =
                      pointObjects[p]["curveNumber"];
                  pointsObject["point_indexes"][p] =
                      pointObjects[p]["pointNumber"];
                  pointsObject["xs"][p] =
                      pointObjects[p]["x"];
                  pointsObject["ys"][p] =
                      pointObjects[p]["y"];
              }

              // Add z if present
              var hasZ = pointObjects[0] !==
                  undefined && pointObjects[0].hasOwnProperty("z");
              if (hasZ) {
                  pointsObject["zs"] = new Array(numPoints);
                  for (p = 0; p < numPoints; p++) {
                      pointsObject["zs"][p] = pointObjects[p]["z"];
                  }
              }

              return pointsObject
          } else {
              return null
          }
      },

      /**
       * Build InputDeviceState data structure from data supplied by the
       * plotly_click, plotly_hover, or plotly_select events
       * @param {Object} data
       * @returns {null|InputDeviceState}
       */
      buildInputDeviceStateObject: function (data) {
          var event = data["event"];
          if (event === undefined) {
              return null;
          } else {
              /** @type {InputDeviceState} */
              var inputDeviceState = {
                  // Keyboard modifiers
                  "alt": event["altKey"],
                  "ctrl": event["ctrlKey"],
                  "meta": event["metaKey"],
                  "shift": event["shiftKey"],

                  // Mouse buttons
                  "button": event["button"],
                  "buttons": event["buttons"]
              };
              return inputDeviceState
          }
      },

      /**
       * Build Selector data structure from data supplied by the
       * plotly_select event
       * @param data
       * @returns {null|Selector}
       */
      buildSelectorObject: function(data) {

          var selectorObject;

          if (data.hasOwnProperty("range")) {
              // Box selection
              selectorObject = {
                  type: "box",
                  selector_state: {
                      xrange: data["range"]["x"],
                      yrange: data["range"]["y"]
                  }
              };
          } else if (data.hasOwnProperty("lassoPoints")) {
              // Lasso selection
              selectorObject = {
                  type: "lasso",
                  selector_state: {
                      xs: data["lassoPoints"]["x"],
                      ys: data["lassoPoints"]["y"]
                  }
              };
          } else {
              selectorObject = null;
          }
          return selectorObject
      },

      /**
       * Handle ploty_restyle events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_restyle: function (data) {

          if (data === null || data === undefined) {
              // No data to report to the Python side
              return
          }

          if (data[0] && data[0].hasOwnProperty("_doNotReportToPy")) {
              // Restyle originated on the Python side
              return
          }

          // Unpack data
          var styleData = data[0];
          var styleTraces = data[1];

          // Construct restyle message to send to the Python side
          /** @type {Js2PyRestyleMsg} */
          var restyleMsg = {
              style_data: styleData,
              style_traces: styleTraces,
              source_view_id: this.viewID
          };

          this.model.set("_js2py_restyle", restyleMsg);
          this.touch();
      },

      /**
       * Handle plotly_relayout events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_relayout: function (data) {

          if (data === null || data === undefined) {
              // No data to report to the Python side
              return
          }

          if (data.hasOwnProperty("_doNotReportToPy")) {
              // Relayout originated on the Python side
              return
          }

          /** @type {Js2PyRelayoutMsg} */
          var relayoutMsg = {
              relayout_data: data,
              source_view_id: this.viewID
          };

          this.model.set("_js2py_relayout", relayoutMsg);
          this.touch();
      },

      /**
       * Handle plotly_update events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_update: function (data) {

          if (data === null || data === undefined) {
              // No data to report to the Python side
              return
          }

          if (data["data"] &&
              data["data"][0].hasOwnProperty("_doNotReportToPy")) {
              // Update originated on the Python side
              return
          }

          /** @type {Js2PyUpdateMsg} */
          var updateMsg = {
              style_data: data["data"][0],
              style_traces: data["data"][1],
              layout_data: data["layout"],
              source_view_id: this.viewID
          };

          // Log message
          this.model.set("_js2py_update", updateMsg);
          this.touch();
      },

      /**
       * Handle plotly_click events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_click: function (data) {
          this._send_points_callback_message(data, "plotly_click");
      },

      /**
       * Handle plotly_hover events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_hover: function (data) {
          this._send_points_callback_message(data, "plotly_hover");
      },

      /**
       * Handle plotly_unhover events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_unhover: function (data) {
          this._send_points_callback_message(data, "plotly_unhover");
      },

      /**
       * Handle plotly_selected events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_selected: function (data) {
          this._send_points_callback_message(data, "plotly_selected");
      },

      /**
       * Handle plotly_deselect events emitted by the Plotly.js library
       * @param data
       */
      handle_plotly_deselect: function (data) {
          data = {
              points : []
          };
          this._send_points_callback_message(data, "plotly_deselect");
      },

      /**
       * Build and send a points callback message to the Python side
       *
       * @param {Object} data
       *  data object as provided by the plotly_click, plotly_hover,
       *  plotly_unhover, or plotly_selected events
       * @param {String} event_type
       *  Name of the triggering event. One of 'plotly_click',
       *  'plotly_hover', 'plotly_unhover', or 'plotly_selected'
       * @private
       */
      _send_points_callback_message: function (data, event_type) {
          if (data === null || data === undefined) {
              // No data to report to the Python side
              return;
          }

          /** @type {Js2PyPointsCallbackMsg} */
          var pointsMsg = {
              event_type: event_type,
              points: this.buildPointsObject(data),
              device_state: this.buildInputDeviceStateObject(data),
              selector: this.buildSelectorObject(data)
          };

          if (pointsMsg["points"] !== null &&
              pointsMsg["points"] !== undefined) {

              this.model.set("_js2py_pointsCallback", pointsMsg);
              this.touch();
          }
      },

      /**
       * Stub for future handling of plotly_doubleclick
       * @param data
       */
      handle_plotly_doubleclick: function (data) {},


      /**
       * Handle Plotly.addTraces request
       */
      do_addTraces: function () {

          /** @type {Py2JsAddTracesMsg} */
          var msgData = this.model.get("_py2js_addTraces");

          if (msgData !== null) {

              // Save off original number of traces
              var prevNumTraces = this.el.data.length;

              var that = this;
              Plotly.addTraces(this.el, msgData.trace_data).then(function () {

                  // ### Send trace deltas ###
                  that._sendTraceDeltas(msgData.trace_edit_id);

                  // ### Send layout delta ###
                  var layout_edit_id = msgData.layout_edit_id;
                  that._sendLayoutDelta(layout_edit_id);
              });
          }
      },

      /**
       * Handle Plotly.deleteTraces request
       */
      do_deleteTraces: function () {

          /** @type {Py2JsDeleteTracesMsg} */
          var msgData = this.model.get("_py2js_deleteTraces");

          if (msgData  !== null){
              var delete_inds = msgData.delete_inds;
              var that = this;
              Plotly.deleteTraces(this.el, delete_inds).then(function () {

                  // ### Send trace deltas ###
                  var trace_edit_id = msgData.trace_edit_id;
                  that._sendTraceDeltas(trace_edit_id);

                  // ### Send layout delta ###
                  var layout_edit_id = msgData.layout_edit_id;
                  that._sendLayoutDelta(layout_edit_id);
              });
          }
      },

      /**
       * Handle Plotly.moveTraces request
       */
      do_moveTraces: function () {

          /** @type {Py2JsMoveTracesMsg} */
          var msgData = this.model.get("_py2js_moveTraces");

          if (msgData !== null){
              // Unpack message
              var currentInds = msgData.current_trace_inds;
              var newInds = msgData.new_trace_inds;

              // Check if the new trace indexes are actually different than
              // the current indexes
              var inds_equal = lodash.isEqual(currentInds, newInds);

              if (!inds_equal) {
                  Plotly.moveTraces(this.el, currentInds, newInds);
              }
          }
      },

      /**
       * Handle Plotly.restyle request
       */
      do_restyle: function () {

          /** @type {Py2JsRestyleMsg} */
          var msgData = this.model.get("_py2js_restyle");
          if (msgData !== null) {
              var restyleData = msgData.restyle_data;
              var traceIndexes = this.model._normalize_trace_indexes(
                  msgData.restyle_traces);

              restyleData["_doNotReportToPy"] = true;
              Plotly.restyle(this.el, restyleData, traceIndexes);

              // ### Send trace deltas ###
              // We create an array of deltas corresponding to the restyled
              // traces.
              this._sendTraceDeltas(msgData.trace_edit_id);

              // ### Send layout delta ###
              var layout_edit_id = msgData.layout_edit_id;
              this._sendLayoutDelta(layout_edit_id);
          }
      },

      /**
       * Handle Plotly.relayout request
       */
      do_relayout: function () {

          /** @type {Py2JsRelayoutMsg} */
          var msgData = this.model.get("_py2js_relayout");
          if (msgData !== null) {
              if (msgData.source_view_id !== this.viewID) {
                  var relayoutData = msgData.relayout_data;
                  relayoutData["_doNotReportToPy"] = true;
                  Plotly.relayout(this.el, msgData.relayout_data);
              }

              // ### Send layout delta ###
              var layout_edit_id = msgData.layout_edit_id;
              this._sendLayoutDelta(layout_edit_id);
          }
      },

      /**
       * Handle Plotly.update request
       */
      do_update: function () {

          /** @type {Py2JsUpdateMsg} */
          var msgData = this.model.get("_py2js_update");

          if (msgData !== null) {
              var style = msgData.style_data || {};
              var layout = msgData.layout_data || {};
              var traceIndexes = this.model._normalize_trace_indexes(
                  msgData.style_traces);

              style["_doNotReportToPy"] = true;
                  Plotly.update(this.el, style, layout, traceIndexes);

              // ### Send trace deltas ###
              // We create an array of deltas corresponding to the updated
              // traces.
              this._sendTraceDeltas(msgData.trace_edit_id);

              // ### Send layout delta ###
              var layout_edit_id = msgData.layout_edit_id;
              this._sendLayoutDelta(layout_edit_id);
          }
      },

      /**
       * Handle Plotly.animate request
       */
      do_animate: function() {

          /** @type {Py2JsAnimateMsg} */
          var msgData = this.model.get("_py2js_animate");

          if (msgData !== null) {

              // Unpack params
              // var animationData = msgData[0];
              var animationOpts = msgData.animation_opts;

              var styles = msgData.style_data;
              var layout = msgData.layout_data;
              var traceIndexes = this.model._normalize_trace_indexes(
                  msgData.style_traces);

              var animationData = {
                  data: styles,
                  layout: layout,
                  traces: traceIndexes
              };

              animationData["_doNotReportToPy"] = true;
              var that = this;

              Plotly.animate(this.el, animationData, animationOpts).then(
                  function () {

                      // ### Send trace deltas ###
                      // We create an array of deltas corresponding to the
                      // animated traces.
                      that._sendTraceDeltas(msgData.trace_edit_id);

                      // ### Send layout delta ###
                      var layout_edit_id = msgData.layout_edit_id;
                      that._sendLayoutDelta(layout_edit_id);
                  });

          }
      },

      /**
       * Construct layout delta object and send layoutDelta message to the
       * Python side
       *
       * @param layout_edit_id
       *  Edit ID of message that triggered the creation of the layout delta
       * @private
       */
      _sendLayoutDelta: function(layout_edit_id) {
          // ### Handle layout delta ###
          var layout_delta = createDeltaObject(
              this.getFullLayout(),
              this.model.get("_layout"));

          /** @type{Js2PyLayoutDeltaMsg} */
          var layoutDeltaMsg = {
              layout_delta: layout_delta,
              layout_edit_id: layout_edit_id};

          this.model.set("_js2py_layoutDelta", layoutDeltaMsg);
          this.touch();
      },

      /**
       * Construct trace deltas array for the requested trace indexes and
       * send traceDeltas message to the Python side
       *  Array of indexes of traces for which to compute deltas
       * @param trace_edit_id
       *  Edit ID of message that triggered the creation of trace deltas
       * @private
       */
      _sendTraceDeltas: function (trace_edit_id) {

          var trace_data = this.model.get("_data");
          var traceIndexes = lodash.range(trace_data.length);
          var trace_deltas = new Array(traceIndexes.length);

          var fullData = this.getFullData();
          for (var i = 0; i < traceIndexes.length; i++) {
              var traceInd = traceIndexes[i];
              trace_deltas[i] = createDeltaObject(
                  fullData[traceInd], trace_data[traceInd]);
          }

          /** @type{Js2PyTraceDeltasMsg} */
          var traceDeltasMsg = {
              trace_deltas: trace_deltas,
              trace_edit_id: trace_edit_id};

          this.model.set("_js2py_traceDeltas", traceDeltasMsg);
          this.touch();
      }
  });

  // Serialization
  /**
   * Create a mapping from numpy dtype strings to corresponding typed array
   * constructors
   */
  var numpy_dtype_to_typedarray_type = {
      int8: Int8Array,
      int16: Int16Array,
      int32: Int32Array,
      uint8: Uint8Array,
      uint16: Uint16Array,
      uint32: Uint32Array,
      float32: Float32Array,
      float64: Float64Array
  };

  function serializeTypedArray(v) {
      var numpyType;
      if (v instanceof Int8Array) {
          numpyType = 'int8';
      } else if (v instanceof Int16Array) {
          numpyType = 'int16';
      } else if (v instanceof Int32Array) {
          numpyType = 'int32';
      } else if (v instanceof Uint8Array) {
          numpyType = 'uint8';
      } else if (v instanceof Uint16Array) {
          numpyType = 'uint16';
      } else if (v instanceof Uint32Array) {
          numpyType = 'uint32';
      } else if (v instanceof Float32Array) {
          numpyType = 'float32';
      } else if (v instanceof Float64Array) {
          numpyType = 'float64';
      } else {
          // Don't understand it, return as is
          return v;
      }
      var res = {
          dtype: numpyType,
          shape: [v.length],
          value: v.buffer
      };
      return res
  }

  /**
   * ipywidget JavaScript -> Python serializer
   */
  function js2py_serializer(v, widgetManager) {
      var res;

      if (lodash.isTypedArray(v)) {
          res = serializeTypedArray(v);
      } else if (Array.isArray(v)) {
          // Serialize array elements recursively
          res = new Array(v.length);
          for (var i = 0; i < v.length; i++) {
              res[i] = js2py_serializer(v[i]);
          }
      } else if (lodash.isPlainObject(v)) {
          // Serialize object properties recursively
          res = {};
          for (var p in v) {
              if (v.hasOwnProperty(p)) {
                  res[p] = js2py_serializer(v[p]);
              }
          }
      } else if (v === undefined) {
          // Translate undefined into '_undefined_' sentinal string. The
          // Python _js_to_py deserializer will convert this into an
          // Undefined object
          res = "_undefined_";
      } else {
          // Primitive value to transfer directly
          res = v;
      }
      return res
  }

  /**
   * ipywidget Python -> Javascript deserializer
   */
  function py2js_deserializer(v, widgetManager) {
      var res;

      if (Array.isArray(v)) {
          // Deserialize array elements recursively
          res = new Array(v.length);
          for (var i = 0; i < v.length; i++) {
              res[i] = py2js_deserializer(v[i]);
          }
      } else if (lodash.isPlainObject(v)) {
          if ((lodash.has(v, 'value') || lodash.has(v, 'buffer')) &&
              lodash.has(v, 'dtype') &&
              lodash.has(v, 'shape')) {
              // Deserialize special buffer/dtype/shape objects into typed arrays
              // These objects correspond to numpy arrays on the Python side
              //
              // Note plotly.py<=3.1.1 called the buffer object `buffer`
              // This was renamed `value` in 3.2 to work around a naming conflict
              // when saving widget state to a notebook.
              var typedarray_type = numpy_dtype_to_typedarray_type[v.dtype];
              var buffer = lodash.has(v, 'value')? v.value.buffer: v.buffer.buffer;
              res = new typedarray_type(buffer);
          } else {
              // Deserialize object properties recursively
              res = {};
              for (var p in v) {
                  if (v.hasOwnProperty(p)) {
                      res[p] = py2js_deserializer(v[p]);
                  }
              }
          }
      } else if (v === "_undefined_") {
          // Convert the _undefined_ sentinal into undefined
          res = undefined;
      } else {
          // Accept primitive value directly
          res = v;
      }
      return res
  }

  /**
   * Return whether the input value is a typed array
   * @param potentialTypedArray
   *  Value to examine
   * @returns {boolean}
   */
  function isTypedArray$1(potentialTypedArray) {
      return ArrayBuffer.isView(potentialTypedArray) &&
          !(potentialTypedArray instanceof DataView);
  }

  /**
   * Customizer for use with lodash's mergeWith function
   *
   * The customizer ensures that typed arrays are not converted into standard
   * arrays during the recursive merge
   *
   * See: https://lodash.com/docs/latest#mergeWith
   */
  function fullMergeCustomizer(objValue, srcValue, key) {
      if (key[0] === '_') {
          // Don't recurse into private properties
          return null
      } else if (isTypedArray$1(srcValue)) {
          // Return typed arrays directly, don't recurse inside
          return srcValue
      }
  }

  /**
   * Reform a Plotly.relayout like operation on an input object
   *
   * @param {Object} parentObj
   *  The object that the relayout operation should be applied to
   * @param {Object} relayoutData
   *  An relayout object as accepted by Plotly.relayout
   *
   *  Examples:
   *      var d = {foo {bar [5, 10]}};
   *      performRelayoutLike(d, {'foo.bar': [0, 1]});
   *      d -> {foo: {bar: [0, 1]}}
   *
   *      var d = {foo {bar [5, 10]}};
   *      performRelayoutLike(d, {'baz': 34});
   *      d -> {foo: {bar: [5, 10]}, baz: 34}
   *
   *      var d = {foo: {bar: [5, 10]};
   *      performRelayoutLike(d, {'foo.baz[1]': 17});
   *      d -> {foo: {bar: [5, 17]}}
   *
   */
  function performRelayoutLike(parentObj, relayoutData) {
      // Perform a relayout style operation on a given parent object
      for (var rawKey in relayoutData) {
          if (!relayoutData.hasOwnProperty(rawKey)) {
              continue
          }

          // Extract value for this key
          var relayoutVal = relayoutData[rawKey];

          // Set property value
          if (relayoutVal === null) {
              lodash.unset(parentObj, rawKey);
          } else {
              lodash.set(parentObj, rawKey, relayoutVal);
          }
      }
  }

  /**
   * Perform a Plotly.restyle like operation on an input object array
   *
   * @param {Array.<Object>} parentArray
   *  The object that the restyle operation should be applied to
   * @param {Object} restyleData
   *  A restyle object as accepted by Plotly.restyle
   * @param {Array.<Number>} restyleTraces
   *  Array of indexes of the traces that the resytle operation applies to
   *
   *  Examples:
   *      var d = [{foo: {bar: 1}}, {}, {}]
   *      performRestyleLike(d, {'foo.bar': 2}, [0])
   *      d -> [{foo: {bar: 2}}, {}, {}]
   *
   *      var d = [{foo: {bar: 1}}, {}, {}]
   *      performRestyleLike(d, {'foo.bar': 2}, [0, 1, 2])
   *      d -> [{foo: {bar: 2}}, {foo: {bar: 2}}, {foo: {bar: 2}}]
   *
   *      var d = [{foo: {bar: 1}}, {}, {}]
   *      performRestyleLike(d, {'foo.bar': [2, 3, 4]}, [0, 1, 2])
   *      d -> [{foo: {bar: 2}}, {foo: {bar: 3}}, {foo: {bar: 4}}]
   *
   */
  function performRestyleLike(parentArray, restyleData, restyleTraces) {
      // Loop over the properties of restyleData
      for (var rawKey in restyleData) {
          if (!restyleData.hasOwnProperty(rawKey)) { continue }

          // Extract value for property and normalize into a value list
          var valArray = restyleData[rawKey];
          if (!Array.isArray(valArray)) {
              valArray = [valArray];
          }

          // Loop over the indexes of the traces being restyled
          for (var i = 0; i < restyleTraces.length; i++) {

              // Get trace object
              var traceInd = restyleTraces[i];
              var trace = parentArray[traceInd];

              // Extract value for this trace
              var singleVal = valArray[i % valArray.length];

              // Set property value
              if (singleVal === null) {
                  lodash.unset(trace, rawKey);
              } else if (singleVal !== undefined){
                  lodash.set(trace, rawKey, singleVal);
              }
          }
      }
  }

  /**
   * Perform a Plotly.moveTraces like operation on an input object array
   * @param parentArray
   *  The object that the moveTraces operation should be applied to
   * @param currentInds
   *  Array of the current indexes of traces to be moved
   * @param newInds
   *  Array of the new indexes that traces selected by currentInds should be
   *  moved to.
   *
   *  Examples:
   *      var d = [{foo: 0}, {foo: 1}, {foo: 2}]
   *      performMoveTracesLike(d, [0, 1], [2, 0])
   *      d -> [{foo: 1}, {foo: 2}, {foo: 0}]
   *
   *      var d = [{foo: 0}, {foo: 1}, {foo: 2}]
   *      performMoveTracesLike(d, [0, 2], [1, 2])
   *      d -> [{foo: 1}, {foo: 0}, {foo: 2}]
   */
  function performMoveTracesLike(parentArray, currentInds, newInds) {

      // ### Remove by currentInds in reverse order ###
      var movingTracesData = [];
      for (var ci = currentInds.length - 1; ci >= 0; ci--) {
          // Insert moving parentArray at beginning of the list
          movingTracesData.splice(0, 0, parentArray[currentInds[ci]]);
          parentArray.splice(currentInds[ci], 1);
      }

      // ### Sort newInds and movingTracesData by newInds ###
      var newIndexSortedArrays = lodash(newInds).zip(movingTracesData)
          .sortBy(0)
          .unzip()
          .value();

      newInds = newIndexSortedArrays[0];
      movingTracesData = newIndexSortedArrays[1];

      // ### Insert by newInds in forward order ###
      for (var ni = 0; ni < newInds.length; ni++) {
          parentArray.splice(newInds[ni], 0, movingTracesData[ni]);
      }
  }

  /**
   * Remove nested properties from a parent object
   * @param {Object} parentObj
   *  Parent object from which properties or nested properties should be removed
   * @param {Array.<Array.<Number|String>>} keyPaths
   *  Array of key paths for properties that should be removed. Each key path
   *  is an array of properties names or array indexes that reference a
   *  property to be removed
   *
   *  Examples:
   *      var d = {foo: [{bar: 0}, {bar: 1}], baz: 32}
   *      performRemoveProps(d, ['baz'])
   *      d -> {foo: [{bar: 0}, {bar: 1}]}
   *
   *      var d = {foo: [{bar: 0}, {bar: 1}], baz: 32}
   *      performRemoveProps(d, ['foo[1].bar', 'baz'])
   *      d -> {foo: [{bar: 0}, {}]}
   *
   */
  function performRemoveProps(parentObj, keyPaths) {

      for(var i=0; i < keyPaths.length; i++) {
          var keyPath = keyPaths[i];
          lodash.unset(parentObj, keyPath);
      }
  }


  /**
   * Return object that contains all properties in fullObj that are not
   * identical to the corresponding properties in removeObj
   *
   * Properties of fullObj and removeObj may be objects or arrays of objects
   *
   * Returned object is a deep clone of the properties of the input objects
   *
   * @param {Object} fullObj
   * @param {Object} removeObj
   *
   *  Examples:
   *      var fullD = {foo: [{bar: 0}, {bar: 1}], baz: 32}
   *      var removeD = {baz: 32}
   *      createDeltaObject(fullD, removeD)
   *          -> {foo: [{bar: 0}, {bar: 1}]}
   *
   *      var fullD = {foo: [{bar: 0}, {bar: 1}], baz: 32}
   *      var removeD = {baz: 45}
   *      createDeltaObject(fullD, removeD)
   *          -> {foo: [{bar: 0}, {bar: 1}], baz: 32}
   *
   *      var fullD = {foo: [{bar: 0}, {bar: 1}], baz: 32}
   *      var removeD = {foo: [{bar: 0}, {bar: 1}]}
   *      createDeltaObject(fullD, removeD)
   *          -> {baz: 32}
   *
   */
  function createDeltaObject(fullObj, removeObj) {

      // Initialize result as object or array
      var res;
      if(Array.isArray(fullObj)) {
          res = new Array(fullObj.length);
      } else {
          res = {};
      }

      // Initialize removeObj to empty object if not specified
      if (removeObj === null || removeObj === undefined) {
          removeObj = {};
      }

      // Iterate over object properties or array indices
      for (var p in fullObj) {
          if (p[0] !== "_" &&  // Don't consider private properties
              fullObj.hasOwnProperty(p) &&  // Exclude parent properties
              fullObj[p] !== null  // Exclude cases where fullObj doesn't
                                   // have the property
          ) {
              // Compute object equality
              var props_equal;
              props_equal = lodash.isEqual(fullObj[p], removeObj[p]);

              // Perform recursive comparison if props are not equal
              if (!props_equal || p === "uid") {  // Let uids through

                  // property has non-null value in fullObj that doesn't
                  // match the value in removeObj
                  var fullVal = fullObj[p];
                  if (removeObj.hasOwnProperty(p) &&
                      typeof fullVal === "object") {
                      // Recurse over object properties
                      if(Array.isArray(fullVal)) {

                          if (fullVal.length > 0 &&
                              typeof(fullVal[0]) === "object") {
                              // We have an object array
                              res[p] = new Array(fullVal.length);
                              for (var i = 0; i < fullVal.length; i++) {
                                  if (!Array.isArray(removeObj[p]) ||
                                      removeObj[p].length <= i) {

                                      res[p][i] = fullVal[i];
                                  } else {
                                      res[p][i] = createDeltaObject(fullVal[i],
                                          removeObj[p][i]);
                                  }
                              }
                          } else {
                              // We have a primitive array or typed array
                              res[p] = fullVal;
                          }
                      } else { // object
                          var full_obj = createDeltaObject(fullVal,
                              removeObj[p]);
                          if (Object.keys(full_obj).length > 0) {
                              // new object is not empty
                              res[p] = full_obj;
                          }
                      }
                  } else if (typeof fullVal === "object" &&
                      !Array.isArray(fullVal)) {
                      // Return 'clone' of fullVal
                      // We don't use a standard clone method so that we keep
                      // the special case handling of this method
                      res[p] = createDeltaObject(fullVal, {});

                  } else if (fullVal !== undefined &&
                      typeof fullVal !== 'function') {
                      // No recursion necessary, Just keep value from fullObj.
                      // But skip values with function type
                      res[p] = fullVal;
                  }
              }
          }
      }
      return res
  }

  function randstr(existing, bits, base, _recursion) {
      if(!base) base = 16;
      if(bits === undefined) bits = 24;
      if(bits <= 0) return '0';

      var digits = Math.log(Math.pow(2, bits)) / Math.log(base);
      var res = '';
      var i, b, x;

      for(i = 2; digits === Infinity; i *= 2) {
          digits = Math.log(Math.pow(2, bits / i)) / Math.log(base) * i;
      }

      var rem = digits - Math.floor(digits);

      for(i = 0; i < Math.floor(digits); i++) {
          x = Math.floor(Math.random() * base).toString(base);
          res = x + res;
      }

      if(rem) {
          b = Math.pow(base, rem);
          x = Math.floor(Math.random() * b).toString(base);
          res = x + res;
      }

      var parsed = parseInt(res, base);
      if((existing && existing[res]) ||
          (parsed !== Infinity && parsed >= Math.pow(2, bits))) {
          if(_recursion > 10) {
              lib.warn('randstr failed uniqueness');
              return res;
          }
          return randstr(existing, bits, base, (_recursion || 0) + 1);
      } else return res;
  }

  var Figure = {
      FigureView : FigureView,
      FigureModel: FigureModel
  };

  var src = createCommonjsModule(function (module) {
  // Export widget models and views, and the npm package version number.
  module.exports = Figure;
  module.exports['version'] = require$$1.version;
  });

  window.define("plotlywidget", function () {
    return src;
  });

  const load_ipython_extension$1 = function () {
    return Promise.resolve();
  };

  const id$1 = location.hash.replace("#", "");

  function handleDisplayData(message) {
    if (message) {
      const notebook = Jupyter.notebook;
      const output = notebook.get_cells()[0].output_area;

      output.append_output(message);

      // append_output should be executed after comm open message, as model created on this message
      // and get_model at append_output return null and display 'A Jupyter widget could not be displayed' this message
      if (output.element.innerText.startsWith("A Jupyter widget could not be displayed")) {
        output.element.innerText = "";
      }
    }
  }

  function handleStateMessage(message) {
    if (message) {
      const notebook = Jupyter.notebook;
      const widgetManager = notebook.kernel.widget_manager;
      widgetManager.set_state(message).then(_ => {
        notebook.get_cells().forEach(cell => {
          notebook.render_cell_output(cell);
        });
      });
    }
  }

  function handleSaveState() {
    const widgetManager = Jupyter.notebook.kernel.widget_manager;
    if (!widgetManager) {
      return;
    }
    try {
      widgetManager.get_state().then(state => {
        if (Object.keys(state.state).length != 0) {
          channel.send(SAVE_STATE_REQUEST, JSON.stringify(state));
        }
      });
    } catch (ex) {
      console.debug("Failed to serialize widgets state");
    }
  }

  function handleKernelState(message) {
    const state = message.state;
    Jupyter.notebook.kernel.set_kernel_state(state == "connected");
  }

  requirejs(["base/js/namespace", "base/js/utils"], function (namespace, utils) {
    window.Jupyter = namespace;

    const extensions = [
      extension_1(),
      load_ipython_extension$1()
    ];

    Promise.all(extensions).then(() => {

      const notebook = namespace.notebook;
      const kernel = notebook.kernel;
      const widgetManager = kernel.widget_manager;

      patch(kernel);
      patch$1(WidgetModel, widgetManager);

      channel.on(DISPLAY_DATA_RESPONSE, handleDisplayData);
      channel.on(SET_STATE_RESPONSE, handleStateMessage);
      channel.on(KERNEL_STATE_RESPONSE, handleKernelState);
      channel.on(COMM_MSG_RESPONSE, utils.debounse(handleSaveState, 100));
      channel.start();
    });
  });

}($));
